/* xg.c: Guile and Ruby bindings for gdk/gtk/pango, some of glib
 *   generated automatically from makexg.scm and xgdata.scm
 *   needs xen.h
 *
 *   HAVE_GTK_1 if 1.2.n
 *   GDK_DISABLE_DEPRECATED, GTK_DISABLE_DEPRECATED, and GDK_PIXBUF_DISABLE_DEPRECATED are handled together
 *   GDK_ENABLE_BROKEN and GTK_ENABLE_BROKEN are handled together
 *   PANGO_ENABLE_ENGINE and PANGO_ENABLE_BACKEND are handled together, and may be removed later
 *
 *   other flags:
 *     HAVE_GTK_FILE_SELECTION_GET_SELECTIONS
 *     HAVE_GTK_ICON_SET_GET_TYPE
 *
 * reference args are ignored if passed, resultant values are returned in a list.
 * null ptrs are passed and returned as #f, trailing "user_data" callback function arguments are optional (default: #f).
 * 'xg is added to *features*
 *
 * added funcs:
 *    (xm-version) -> date string.
 *    (c-array->list arr len) derefs each member of arr, returning lisp list
 *    (list->c-array lst ctype) packages each member of list as c-type "type" returning (wrapped) c array
 *    (|GdkCursor #:optional type ref_count) -> GdkCursor struct
 *    (|GdkPoint #:optional x y) -> GdkPoint struct
 *    (|GdkRectangle #:optional x y width height) -> GdkRectangle struct
 *    (|GdkSegment #:optional x1 y1 x2 y2) -> GdkSegment struct
 *    (|GdkSpan #:optional x y width) -> GdkSpan struct
 *    (|GtkTextIter) -> GtkTextIter struct
 *    (|GtkTextMark) -> GtkTextMark struct
 *    (|GtkTextChildAnchor) -> GtkTextChildAnchor struct
 *
 * omitted functions and macros:
 *     anything with a va_list or GtkArg* argument.  "..." args are ignored.
 *     most of the unusual keysym names
 *     all *_CLASS, *_IFACE macros
 *     deprecated macros that involve argument reordering
 *     deprecated struct field accessors, Pango struct field accessors
 *     win32-specific functions
 *
 * TODO: check out the g_signal handlers (gtk_signal_* is ok)
 * TODO: GdkEvent casts
 * TODO: struct print, more struct instance creators(?)
 * TODO: tie into libxm (configure.ac etc), Snd (snd-motif translation)
 * TODO: add unicode handlers from glib -- anything else?
 * TODO: unprotect *_remove, unprotect old upon reset callback
 * TODO: test suite (snd-test 24)
 *
 * HISTORY:
 *     13-Mar:    Gtk 2.0.0
 *     12-Mar:    support for GtkDestroyNotify callbacks
 *     27-Feb:    remove gtk_tree_view_column_cell_render, gtk_tree_view_column_cell_focus, 
 *                  gtk_tree_view_column_cell_draw_focus and gtk_tree_view_column_cell_set_dirty (privatized in 1.3.15)
 *                add (on HAVE-* switches) gtk_file_selection_get_selections, gtk_file_selection_set_select_multiple
 *                  and gtk_file_selection_get_select_multiple (new functions in 1.3.15)
 *                  also gtk_tree_path_new_first void and gtk_tree_model_get_iter_first (new names in 1.3.15)
 *     26-Feb:    Gtk 1.2.10 support, also 1.3.15
 *     25-Feb:    dialog example in libxm.html
 *                Ruby support via xg-ruby.c
 *     21-Feb:    #f=NULL throughout, gdk-pixbuf, GTypes.
 *     11-Feb-02: initial version.
 */

#if defined(HAVE_CONFIG_H)
  #include "config.h"
#endif

#include <glib.h>
#include <gdk/gdk.h>
#include <gdk/gdkkeysyms.h>
#include <gtk/gtk.h>
#if (!HAVE_GTK_1)
#include <glib-object.h>
#include <pango/pango.h>
#endif

#include <string.h>

#if USE_SND
  /* USE_SND causes xm to use Snd's error handlers which are much smarter than xen's fallback versions */
  #include "snd.h"
#else
  #include "xen.h"
#endif
#ifndef CALLOC
  #define CALLOC(a, b)  calloc((size_t)(a), (size_t)(b))
  #define FREE(a)       free(a)
#endif

/* -------------------------------- smob for GC -------------------------------- */
static XEN_OBJECT_TYPE xm_obj_tag;
#if HAVE_GUILE
static size_t xm_obj_free(XEN obj)
{
  FREE((void *)XEN_OBJECT_REF(obj));
  return(0);
}
#endif
#if HAVE_RUBY
static void *xm_obj_free(XEN obj)
{
  FREE((void *)obj);
  return(NULL);
}
#endif
static XEN make_xm_obj(void *ptr)
{
  XEN_MAKE_AND_RETURN_OBJECT(xm_obj_tag, ptr, 0, xm_obj_free);
}
static void define_xm_obj(void)
{
  xm_obj_tag = XEN_MAKE_OBJECT_TYPE("XmObj", sizeof(void *));
#if HAVE_GUILE
  scm_set_smob_free(xm_obj_tag, xm_obj_free);
#endif
}  

/* prefix for all names */
#if HAVE_GUILE
  #define XG_PRE "|"
  #define XG_POST ""
#else
/* for Ruby, XG PRE needs to be uppercase */
  #define XG_PRE "R"
  #define XG_POST ""
#endif

#define WRAP_FOR_XEN(Name, Value) XEN_LIST_2(C_STRING_TO_XEN_SYMBOL(Name), C_TO_XEN_ULONG((unsigned long)Value))
#define WRAP_P(Name, Value) (XEN_LIST_P(Value) && \
                            (XEN_LIST_LENGTH(Value) >= 2) && \
                            (XEN_SYMBOL_P(XEN_CAR(Value))) && \
                            (strcmp(Name, XEN_SYMBOL_TO_C_STRING(XEN_CAR(Value))) == 0))

#define XM_TYPE(Name, XType) \
  static XEN C_TO_XEN_ ## Name (XType val) {return(WRAP_FOR_XEN(#Name, val));} \
  static XType XEN_TO_C_ ## Name (XEN val) {return((XType)XEN_TO_C_ULONG(XEN_CADR(val)));} \
  static int XEN_ ## Name ## _P(XEN val) {return(WRAP_P(#Name, val));}

#define XM_TYPE_PTR(Name, XType) \
  static XEN C_TO_XEN_ ## Name (XType val) {if (val) return(WRAP_FOR_XEN(#Name, val)); return(XEN_FALSE);} \
  static XType XEN_TO_C_ ## Name (XEN val) {if (XEN_FALSE_P(val)) return(NULL); return((XType)XEN_TO_C_ULONG(XEN_CADR(val)));} \
  static int XEN_ ## Name ## _P(XEN val) {return(WRAP_P(#Name, val));} /* if NULL ok, should be explicit */

/* type checks for callback wrappers */
#define XEN_lambda2_P(Arg)  XEN_FALSE_P(Arg) || (XEN_PROCEDURE_P(Arg) && (XEN_REQUIRED_ARGS(Arg) == 2))
#define XEN_lambda3_P(Arg)  XEN_FALSE_P(Arg) || (XEN_PROCEDURE_P(Arg) && (XEN_REQUIRED_ARGS(Arg) == 3))
#define XEN_GdkInputFunction_P(Arg)  XEN_FALSE_P(Arg) || (XEN_PROCEDURE_P(Arg) && (XEN_REQUIRED_ARGS(Arg) == 3))
#define XEN_GtkCallback_P(Arg)  XEN_FALSE_P(Arg) || (XEN_PROCEDURE_P(Arg) && (XEN_REQUIRED_ARGS(Arg) == 2))
#define XEN_GtkTimeoutFunction_P(Arg)  XEN_FALSE_P(Arg) || (XEN_PROCEDURE_P(Arg) && (XEN_REQUIRED_ARGS(Arg) == 1))
#define XEN_GtkDestroyNotify_P(Arg)  XEN_FALSE_P(Arg) || (XEN_PROCEDURE_P(Arg) && (XEN_REQUIRED_ARGS(Arg) == 1))
#define XEN_GdkFilterFunc_P(Arg)  XEN_FALSE_P(Arg) || (XEN_PROCEDURE_P(Arg) && (XEN_REQUIRED_ARGS(Arg) == 3))
#define XEN_GdkEventFunc_P(Arg)  XEN_FALSE_P(Arg) || (XEN_PROCEDURE_P(Arg) && (XEN_REQUIRED_ARGS(Arg) == 2))
#define XEN_GdkSpanFunc_P(Arg)  XEN_FALSE_P(Arg) || (XEN_PROCEDURE_P(Arg) && (XEN_REQUIRED_ARGS(Arg) == 2))
#define XEN_GtkFunction_P(Arg)  XEN_FALSE_P(Arg) || (XEN_PROCEDURE_P(Arg) && (XEN_REQUIRED_ARGS(Arg) == 1))
#define XEN_GtkKeySnoopFunc_P(Arg)  XEN_FALSE_P(Arg) || (XEN_PROCEDURE_P(Arg) && (XEN_REQUIRED_ARGS(Arg) == 3))
#define XEN_GtkTranslateFunc_P(Arg)  XEN_FALSE_P(Arg) || (XEN_PROCEDURE_P(Arg) && (XEN_REQUIRED_ARGS(Arg) == 2))
#define XEN_GtkMenuPositionFunc_P(Arg)  XEN_FALSE_P(Arg) || (XEN_PROCEDURE_P(Arg) && (XEN_REQUIRED_ARGS(Arg) == 5))
#define XEN_GtkTreeModelForeachFunc_P(Arg)  XEN_FALSE_P(Arg) || (XEN_PROCEDURE_P(Arg) && (XEN_REQUIRED_ARGS(Arg) == 4))
#define XEN_GtkTreeSelectionForeachFunc_P(Arg)  XEN_FALSE_P(Arg) || (XEN_PROCEDURE_P(Arg) && (XEN_REQUIRED_ARGS(Arg) == 4))
#define XEN_GtkClipboardReceivedFunc_P(Arg)  XEN_FALSE_P(Arg) || (XEN_PROCEDURE_P(Arg) && (XEN_REQUIRED_ARGS(Arg) == 3))
#define XEN_GtkClipboardTextReceivedFunc_P(Arg)  XEN_FALSE_P(Arg) || (XEN_PROCEDURE_P(Arg) && (XEN_REQUIRED_ARGS(Arg) == 3))
#define XEN_GtkTreeViewColumnDropFunc_P(Arg)  XEN_FALSE_P(Arg) || (XEN_PROCEDURE_P(Arg) && (XEN_REQUIRED_ARGS(Arg) == 5))
#define XEN_GtkTreeViewMappingFunc_P(Arg)  XEN_FALSE_P(Arg) || (XEN_PROCEDURE_P(Arg) && (XEN_REQUIRED_ARGS(Arg) == 3))
#define XEN_GtkTreeViewSearchEqualFunc_P(Arg)  XEN_FALSE_P(Arg) || (XEN_PROCEDURE_P(Arg) && (XEN_REQUIRED_ARGS(Arg) == 5))
#define XEN_GtkTreeCellDataFunc_P(Arg)  XEN_FALSE_P(Arg) || (XEN_PROCEDURE_P(Arg) && (XEN_REQUIRED_ARGS(Arg) == 5))
#define XEN_GtkTreeIterCompareFunc_P(Arg)  XEN_FALSE_P(Arg) || (XEN_PROCEDURE_P(Arg) && (XEN_REQUIRED_ARGS(Arg) == 4))
#define XEN_GtkTreeSelectionFunc_P(Arg)  XEN_FALSE_P(Arg) || (XEN_PROCEDURE_P(Arg) && (XEN_REQUIRED_ARGS(Arg) == 5))
#define XEN_GtkClipboardGetFunc_P(Arg)  XEN_FALSE_P(Arg) || (XEN_PROCEDURE_P(Arg) && (XEN_REQUIRED_ARGS(Arg) == 4))
#define XEN_GtkClipboardClearFunc_P(Arg)  XEN_FALSE_P(Arg) || (XEN_PROCEDURE_P(Arg) && (XEN_REQUIRED_ARGS(Arg) == 2))
#define XEN_lambda_P(Arg) XEN_PROCEDURE_P(Arg)
#define XEN_GtkSignalFunc_P(Arg) XEN_PROCEDURE_P(Arg) && (XEN_REQUIRED_ARGS(Arg) == 2)
#define XEN_TO_C_lambda2(Arg) XEN_FALSE_P(Arg) ? NULL : gxg_child_func
#define XEN_TO_C_lambda3(Arg) XEN_FALSE_P(Arg) ? NULL : gxg_find_func
#define XEN_TO_C_GdkInputFunction(Arg) XEN_FALSE_P(Arg) ? NULL : gxg_input_func
#define XEN_TO_C_GtkCallback(Arg) XEN_FALSE_P(Arg) ? NULL : gxg_func2
#define XEN_TO_C_GtkTimeoutFunction(Arg) XEN_FALSE_P(Arg) ? NULL : gxg_timeout_func
#define XEN_TO_C_GtkDestroyNotify(Arg) XEN_FALSE_P(Arg) ? NULL : gxg_destroy_func
#define XEN_TO_C_GdkFilterFunc(Arg) XEN_FALSE_P(Arg) ? NULL : gxg_filter_func
#define XEN_TO_C_GdkEventFunc(Arg) XEN_FALSE_P(Arg) ? NULL : gxg_event_func
#define XEN_TO_C_GdkSpanFunc(Arg) XEN_FALSE_P(Arg) ? NULL : gxg_span_func
#define XEN_TO_C_GtkFunction(Arg) XEN_FALSE_P(Arg) ? NULL : gxg_func1
#define XEN_TO_C_GtkKeySnoopFunc(Arg) XEN_FALSE_P(Arg) ? NULL : gxg_snoop_func
#define XEN_TO_C_GtkTranslateFunc(Arg) XEN_FALSE_P(Arg) ? NULL : gxg_translate_func
#define XEN_TO_C_GtkMenuPositionFunc(Arg) XEN_FALSE_P(Arg) ? NULL : gxg_menu_position_func
#define XEN_TO_C_GtkTreeModelForeachFunc(Arg) XEN_FALSE_P(Arg) ? NULL : gxg_model_func
#define XEN_TO_C_GtkTreeSelectionForeachFunc(Arg) XEN_FALSE_P(Arg) ? NULL : gxg_tree_selection_func
#define XEN_TO_C_GtkClipboardReceivedFunc(Arg) XEN_FALSE_P(Arg) ? NULL : gxg_clip_received
#define XEN_TO_C_GtkClipboardTextReceivedFunc(Arg) XEN_FALSE_P(Arg) ? NULL : gxg_clip_text_received
#define XEN_TO_C_GtkTreeViewColumnDropFunc(Arg) XEN_FALSE_P(Arg) ? NULL : gxg_tree_column
#define XEN_TO_C_GtkTreeViewMappingFunc(Arg) XEN_FALSE_P(Arg) ? NULL : gxg_tree_mapping
#define XEN_TO_C_GtkTreeViewSearchEqualFunc(Arg) XEN_FALSE_P(Arg) ? NULL : gxg_tree_search
#define XEN_TO_C_GtkTreeCellDataFunc(Arg) XEN_FALSE_P(Arg) ? NULL : gxg_cell_data
#define XEN_TO_C_GtkTreeIterCompareFunc(Arg) XEN_FALSE_P(Arg) ? NULL : gxg_iter_compare
#define XEN_TO_C_GtkTreeSelectionFunc(Arg) XEN_FALSE_P(Arg) ? NULL : gxg_tree_selection
#define XEN_TO_C_GtkClipboardGetFunc(Arg) XEN_FALSE_P(Arg) ? NULL : gxg_clip_get
#define XEN_TO_C_GtkClipboardClearFunc(Arg) XEN_FALSE_P(Arg) ? NULL : gxg_clip_clear
#define XEN_TO_C_GtkSignalFunc(Arg) (GtkSignalFunc)gxg_func2
#define XEN_TO_C_lambda_data(Arg) (gpointer)gxg_ptr
#define XEN_lambda_data_P(Arg) 1
#define C_TO_XEN_GtkTreeViewSearchEqualFunc(Arg) WRAP_FOR_XEN("GtkTreeViewSearchEqualFunc", Arg)
#define C_TO_XEN_GtkTreeIterCompareFunc(Arg) WRAP_FOR_XEN("GtkTreeViewSearchEqualFunc", Arg)
#define C_TO_XEN_GtkTreeSelectionFunc(Arg) WRAP_FOR_XEN("GtkTreeSelectionFunc", Arg)
#define C_TO_XEN_GtkMenuPositionFunc(Arg) WRAP_FOR_XEN("GtkMenuPositionFunc", Arg)
#define C_TO_XEN_GtkDestroyNotify(Arg) WRAP_FOR_XEN("GtkDestroyNotify", Arg)
#define XEN_TO_C_GdkFilterReturn(Arg) XEN_TO_C_INT(Arg)


/* ---------------------------------------- types ---------------------------------------- */

XM_TYPE_PTR(GdkWindow_, GdkWindow*)
#if (!HAVE_GTK_1)
XM_TYPE_PTR(GtkAccelKey_, GtkAccelKey*)
XM_TYPE_PTR(GClosure_, GClosure*)
#endif
#define C_TO_XEN_gint(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_gint(Arg) XEN_TO_C_INT(Arg)
#define XEN_gint_P(Arg) XEN_INTEGER_P(Arg)
#define C_TO_XEN_GdkInputCondition(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_GdkInputCondition(Arg) XEN_TO_C_INT(Arg)
#define XEN_GdkInputCondition_P(Arg) XEN_INTEGER_P(Arg)
XM_TYPE_PTR(GtkWidget_, GtkWidget*)
XM_TYPE_PTR(GdkXEvent_, GdkXEvent*)
XM_TYPE_PTR(GdkEvent_, GdkEvent*)
#if (!HAVE_GTK_1)
XM_TYPE_PTR(GdkSpan_, GdkSpan*)
#endif
XM_TYPE_PTR(GdkEventKey_, GdkEventKey*)
#define C_TO_XEN_gchar_(Arg) C_TO_XEN_STRING(Arg)
#define XEN_TO_C_gchar_(Arg) XEN_TO_C_STRING(Arg)
#define XEN_gchar__P(Arg) XEN_STRING_P(Arg)
XM_TYPE_PTR(GtkMenu_, GtkMenu*)
XM_TYPE_PTR(gint_, gint*)
XM_TYPE_PTR(gboolean_, gboolean*)
#if (!HAVE_GTK_1)
XM_TYPE_PTR(GtkTreeModel_, GtkTreeModel*)
XM_TYPE_PTR(GtkTreePath_, GtkTreePath*)
XM_TYPE_PTR(GtkTreeIter_, GtkTreeIter*)
XM_TYPE_PTR(GtkClipboard_, GtkClipboard*)
#endif
XM_TYPE_PTR(GtkSelectionData_, GtkSelectionData*)
#if (!HAVE_GTK_1)
XM_TYPE_PTR(GtkTreeView_, GtkTreeView*)
XM_TYPE_PTR(GtkTreeViewColumn_, GtkTreeViewColumn*)
XM_TYPE_PTR(GtkCellRenderer_, GtkCellRenderer*)
XM_TYPE_PTR(GtkTreeSelection_, GtkTreeSelection*)
#endif
#define C_TO_XEN_gboolean(Arg) C_TO_XEN_BOOLEAN(Arg)
#define XEN_TO_C_gboolean(Arg) XEN_TO_C_BOOLEAN(Arg)
#define XEN_gboolean_P(Arg) XEN_BOOLEAN_P(Arg)
#define C_TO_XEN_guint(Arg) C_TO_XEN_ULONG(Arg)
#define XEN_TO_C_guint(Arg) XEN_TO_C_ULONG(Arg)
#define XEN_guint_P(Arg) XEN_ULONG_P(Arg)
#define C_TO_XEN_GType(Arg) C_TO_XEN_ULONG(Arg)
#define XEN_TO_C_GType(Arg) XEN_TO_C_ULONG(Arg)
#define XEN_GType_P(Arg) XEN_ULONG_P(Arg)
#define C_TO_XEN_GQuark(Arg) C_TO_XEN_ULONG(Arg)
#define XEN_TO_C_GQuark(Arg) XEN_TO_C_ULONG(Arg)
#define XEN_GQuark_P(Arg) XEN_ULONG_P(Arg)
#define C_TO_XEN_GSignalFlags(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_GSignalFlags(Arg) XEN_TO_C_INT(Arg)
#define XEN_GSignalFlags_P(Arg) XEN_INTEGER_P(Arg)
#if (!HAVE_GTK_1)
XM_TYPE(GSignalAccumulator, GSignalAccumulator)
#endif
XM_TYPE(gpointer, gpointer)
#if (!HAVE_GTK_1)
XM_TYPE(GSignalCMarshaller, GSignalCMarshaller)
XM_TYPE_PTR(GType_, GType*)
XM_TYPE_PTR(GValue_, GValue*)
XM_TYPE_PTR(GSignalQuery_, GSignalQuery*)
#endif
XM_TYPE_PTR(guint_, guint*)
XM_TYPE_PTR(GQuark_, GQuark*)
#if (!HAVE_GTK_1)
XM_TYPE_PTR(GSignalInvocationHint_, GSignalInvocationHint*)
#endif
#define C_TO_XEN_gulong(Arg) C_TO_XEN_ULONG(Arg)
#define XEN_TO_C_gulong(Arg) XEN_TO_C_ULONG(Arg)
#define XEN_gulong_P(Arg) XEN_ULONG_P(Arg)
#if (!HAVE_GTK_1)
XM_TYPE(GSignalEmissionHook, GSignalEmissionHook)
#endif
XM_TYPE(GDestroyNotify, GDestroyNotify)
#if (!HAVE_GTK_1)
XM_TYPE(GCallback, GCallback)
XM_TYPE(GClosureNotify, GClosureNotify)
#endif
#define C_TO_XEN_GConnectFlags(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_GConnectFlags(Arg) XEN_TO_C_INT(Arg)
#define XEN_GConnectFlags_P(Arg) XEN_INTEGER_P(Arg)
#define C_TO_XEN_GSignalMatchType(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_GSignalMatchType(Arg) XEN_TO_C_INT(Arg)
#define XEN_GSignalMatchType_P(Arg) XEN_INTEGER_P(Arg)
XM_TYPE_PTR(GdkColormap_, GdkColormap*)
XM_TYPE_PTR(GdkVisual_, GdkVisual*)
XM_TYPE_PTR(GdkColor_, GdkColor*)
XM_TYPE_PTR(GdkCursor_, GdkCursor*)
#define C_TO_XEN_GdkCursorType(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_GdkCursorType(Arg) XEN_TO_C_INT(Arg)
#define XEN_GdkCursorType_P(Arg) XEN_INTEGER_P(Arg)
XM_TYPE_PTR(GdkPixmap_, GdkPixmap*)
XM_TYPE_PTR(GdkDragContext_, GdkDragContext*)
#define C_TO_XEN_GdkDragAction(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_GdkDragAction(Arg) XEN_TO_C_INT(Arg)
#define XEN_GdkDragAction_P(Arg) XEN_INTEGER_P(Arg)
#define C_TO_XEN_guint32(Arg) C_TO_XEN_ULONG(Arg)
#define XEN_TO_C_guint32(Arg) XEN_TO_C_ULONG(Arg)
#define XEN_guint32_P(Arg) XEN_ULONG_P(Arg)
XM_TYPE(GdkAtom, GdkAtom)
XM_TYPE_PTR(GList_, GList*)
XM_TYPE_PTR(GdkDragProtocol_, GdkDragProtocol*)
#define C_TO_XEN_GdkDragProtocol(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_GdkDragProtocol(Arg) XEN_TO_C_INT(Arg)
#define XEN_GdkDragProtocol_P(Arg) XEN_INTEGER_P(Arg)
XM_TYPE_PTR(GdkDrawable_, GdkDrawable*)
XM_TYPE_PTR(GdkGC_, GdkGC*)
XM_TYPE_PTR(GdkPoint_, GdkPoint*)
XM_TYPE_PTR(GdkImage_, GdkImage*)
XM_TYPE_PTR(GdkSegment_, GdkSegment*)
#if (!HAVE_GTK_1)
XM_TYPE_PTR(PangoFont_, PangoFont*)
XM_TYPE_PTR(PangoGlyphString_, PangoGlyphString*)
XM_TYPE_PTR(PangoLayoutLine_, PangoLayoutLine*)
XM_TYPE_PTR(PangoLayout_, PangoLayout*)
#endif
XM_TYPE_PTR(GdkRegion_, GdkRegion*)
XM_TYPE_PTR(GdkModifierType_, GdkModifierType*)
#define C_TO_XEN_gdouble(Arg) C_TO_XEN_DOUBLE(Arg)
#define XEN_TO_C_gdouble(Arg) XEN_TO_C_DOUBLE(Arg)
#define XEN_gdouble_P(Arg) XEN_NUMBER_P(Arg)
XM_TYPE_PTR(gdouble_, gdouble*)
#define C_TO_XEN_GdkAxisUse(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_GdkAxisUse(Arg) XEN_TO_C_INT(Arg)
#define XEN_GdkAxisUse_P(Arg) XEN_INTEGER_P(Arg)
XM_TYPE_PTR(GdkGCValues_, GdkGCValues*)
#define C_TO_XEN_GdkGCValuesMask(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_GdkGCValuesMask(Arg) XEN_TO_C_INT(Arg)
#define XEN_GdkGCValuesMask_P(Arg) XEN_INTEGER_P(Arg)
#define C_TO_XEN_GdkFunction(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_GdkFunction(Arg) XEN_TO_C_INT(Arg)
#define XEN_GdkFunction_P(Arg) XEN_INTEGER_P(Arg)
#define C_TO_XEN_GdkFill(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_GdkFill(Arg) XEN_TO_C_INT(Arg)
#define XEN_GdkFill_P(Arg) XEN_INTEGER_P(Arg)
XM_TYPE_PTR(GdkBitmap_, GdkBitmap*)
XM_TYPE_PTR(GdkRectangle_, GdkRectangle*)
#define C_TO_XEN_GdkSubwindowMode(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_GdkSubwindowMode(Arg) XEN_TO_C_INT(Arg)
#define XEN_GdkSubwindowMode_P(Arg) XEN_INTEGER_P(Arg)
#define C_TO_XEN_GdkLineStyle(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_GdkLineStyle(Arg) XEN_TO_C_INT(Arg)
#define XEN_GdkLineStyle_P(Arg) XEN_INTEGER_P(Arg)
#define C_TO_XEN_GdkCapStyle(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_GdkCapStyle(Arg) XEN_TO_C_INT(Arg)
#define XEN_GdkCapStyle_P(Arg) XEN_INTEGER_P(Arg)
#define C_TO_XEN_GdkJoinStyle(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_GdkJoinStyle(Arg) XEN_TO_C_INT(Arg)
#define XEN_GdkJoinStyle_P(Arg) XEN_INTEGER_P(Arg)
XM_TYPE_PTR(gint8_, gint8*)
XM_TYPE_PTR(gchar___, gchar***)
#define C_TO_XEN_char_(Arg) C_TO_XEN_STRING(Arg)
#define XEN_TO_C_char_(Arg) XEN_TO_C_STRING(Arg)
#define XEN_char__P(Arg) XEN_STRING_P(Arg)
#define C_TO_XEN_GdkGrabStatus(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_GdkGrabStatus(Arg) XEN_TO_C_INT(Arg)
#define XEN_GdkGrabStatus_P(Arg) XEN_INTEGER_P(Arg)
#define C_TO_XEN_GdkEventMask(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_GdkEventMask(Arg) XEN_TO_C_INT(Arg)
#define XEN_GdkEventMask_P(Arg) XEN_INTEGER_P(Arg)
XM_TYPE_PTR(GdkWChar_, GdkWChar*)
#define C_TO_XEN_GdkImageType(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_GdkImageType(Arg) XEN_TO_C_INT(Arg)
#define XEN_GdkImageType_P(Arg) XEN_INTEGER_P(Arg)
#if (!HAVE_GTK_1)
XM_TYPE_PTR(GdkDevice_, GdkDevice*)
#endif
#define C_TO_XEN_GdkInputSource(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_GdkInputSource(Arg) XEN_TO_C_INT(Arg)
#define XEN_GdkInputSource_P(Arg) XEN_INTEGER_P(Arg)
#define C_TO_XEN_GdkInputMode(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_GdkInputMode(Arg) XEN_TO_C_INT(Arg)
#define XEN_GdkInputMode_P(Arg) XEN_INTEGER_P(Arg)
#define C_TO_XEN_GdkModifierType(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_GdkModifierType(Arg) XEN_TO_C_INT(Arg)
#define XEN_GdkModifierType_P(Arg) XEN_INTEGER_P(Arg)
XM_TYPE_PTR(GdkTimeCoord__, GdkTimeCoord**)
#define C_TO_XEN_GdkExtensionMode(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_GdkExtensionMode(Arg) XEN_TO_C_INT(Arg)
#define XEN_GdkExtensionMode_P(Arg) XEN_INTEGER_P(Arg)
#if (!HAVE_GTK_1)
XM_TYPE_PTR(GdkKeymap_, GdkKeymap*)
XM_TYPE_PTR(GdkKeymapKey_, GdkKeymapKey*)
#endif
#define C_TO_XEN_PangoDirection(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_PangoDirection(Arg) XEN_TO_C_INT(Arg)
#define XEN_PangoDirection_P(Arg) XEN_INTEGER_P(Arg)
#if (!HAVE_GTK_1)
XM_TYPE_PTR(PangoContext_, PangoContext*)
XM_TYPE_PTR(PangoAttribute_, PangoAttribute*)
XM_TYPE_PTR(GdkPixbuf_, GdkPixbuf*)
#endif
#define C_TO_XEN_int(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_int(Arg) XEN_TO_C_INT(Arg)
#define XEN_int_P(Arg) XEN_INTEGER_P(Arg)
#define C_TO_XEN_GdkRgbDither(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_GdkRgbDither(Arg) XEN_TO_C_INT(Arg)
#define XEN_GdkRgbDither_P(Arg) XEN_INTEGER_P(Arg)
#define C_TO_XEN_GdkPixbufAlphaMode(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_GdkPixbufAlphaMode(Arg) XEN_TO_C_INT(Arg)
#define XEN_GdkPixbufAlphaMode_P(Arg) XEN_INTEGER_P(Arg)
XM_TYPE_PTR(GdkBitmap__, GdkBitmap**)
XM_TYPE_PTR(gchar__, gchar**)
#define C_TO_XEN_GdkNativeWindow(Arg) C_TO_XEN_ULONG(Arg)
#define XEN_TO_C_GdkNativeWindow(Arg) XEN_TO_C_ULONG(Arg)
#define XEN_GdkNativeWindow_P(Arg) XEN_ULONG_P(Arg)
XM_TYPE_PTR(guchar_, guchar*)
#define C_TO_XEN_GdkPropMode(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_GdkPropMode(Arg) XEN_TO_C_INT(Arg)
#define XEN_GdkPropMode_P(Arg) XEN_INTEGER_P(Arg)
XM_TYPE_PTR(GdkAtom_, GdkAtom*)
#define C_TO_XEN_GdkFillRule(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_GdkFillRule(Arg) XEN_TO_C_INT(Arg)
#define XEN_GdkFillRule_P(Arg) XEN_INTEGER_P(Arg)
#define C_TO_XEN_GdkOverlapType(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_GdkOverlapType(Arg) XEN_TO_C_INT(Arg)
#define XEN_GdkOverlapType_P(Arg) XEN_INTEGER_P(Arg)
XM_TYPE_PTR(GdkRgbCmap_, GdkRgbCmap*)
XM_TYPE_PTR(guint32_, guint32*)
#define C_TO_XEN_GdkVisualType(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_GdkVisualType(Arg) XEN_TO_C_INT(Arg)
#define XEN_GdkVisualType_P(Arg) XEN_INTEGER_P(Arg)
XM_TYPE_PTR(GdkVisualType_, GdkVisualType*)
XM_TYPE_PTR(GdkWindowAttr_, GdkWindowAttr*)
#define C_TO_XEN_GdkWindowType(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_GdkWindowType(Arg) XEN_TO_C_INT(Arg)
#define XEN_GdkWindowType_P(Arg) XEN_INTEGER_P(Arg)
#define C_TO_XEN_GdkWindowState(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_GdkWindowState(Arg) XEN_TO_C_INT(Arg)
#define XEN_GdkWindowState_P(Arg) XEN_INTEGER_P(Arg)
#define C_TO_XEN_GdkWMDecoration(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_GdkWMDecoration(Arg) XEN_TO_C_INT(Arg)
#define XEN_GdkWMDecoration_P(Arg) XEN_INTEGER_P(Arg)
XM_TYPE_PTR(GdkWMDecoration_, GdkWMDecoration*)
#define C_TO_XEN_GdkWMFunction(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_GdkWMFunction(Arg) XEN_TO_C_INT(Arg)
#define XEN_GdkWMFunction_P(Arg) XEN_INTEGER_P(Arg)
#define C_TO_XEN_GdkWindowEdge(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_GdkWindowEdge(Arg) XEN_TO_C_INT(Arg)
#define XEN_GdkWindowEdge_P(Arg) XEN_INTEGER_P(Arg)
XM_TYPE_PTR(GdkGeometry_, GdkGeometry*)
#if (!HAVE_GTK_1)
XM_TYPE_PTR(GtkWindowGroup_, GtkWindowGroup*)
#endif
XM_TYPE_PTR(GtkWindow_, GtkWindow*)
#define C_TO_XEN_GdkWindowTypeHint(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_GdkWindowTypeHint(Arg) XEN_TO_C_INT(Arg)
#define XEN_GdkWindowTypeHint_P(Arg) XEN_INTEGER_P(Arg)
#define C_TO_XEN_GdkWindowHints(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_GdkWindowHints(Arg) XEN_TO_C_INT(Arg)
#define XEN_GdkWindowHints_P(Arg) XEN_INTEGER_P(Arg)
#if (!HAVE_GTK_1)
XM_TYPE_PTR(GdkPointerHooks_, GdkPointerHooks*)
XM_TYPE(GdkColorspace, GdkColorspace)
XM_TYPE_PTR(GError_, GError*)
XM_TYPE(GdkPixbufDestroyNotify, GdkPixbufDestroyNotify)
#endif
XM_TYPE_PTR(char__, char**)
XM_TYPE_PTR(guint8_, guint8*)
#define C_TO_XEN_guchar(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_guchar(Arg) XEN_TO_C_INT(Arg)
#define XEN_guchar_P(Arg) XEN_INTEGER_P(Arg)
#define C_TO_XEN_gfloat(Arg) C_TO_XEN_DOUBLE(Arg)
#define XEN_TO_C_gfloat(Arg) XEN_TO_C_DOUBLE(Arg)
#define XEN_gfloat_P(Arg) XEN_NUMBER_P(Arg)
#define C_TO_XEN_double(Arg) C_TO_XEN_DOUBLE(Arg)
#define XEN_TO_C_double(Arg) XEN_TO_C_DOUBLE(Arg)
#define XEN_double_P(Arg) XEN_NUMBER_P(Arg)
#if (!HAVE_GTK_1)
XM_TYPE(GdkInterpType, GdkInterpType)
XM_TYPE_PTR(GdkPixbufAnimation_, GdkPixbufAnimation*)
XM_TYPE_PTR(GdkPixbufAnimationIter_, GdkPixbufAnimationIter*)
XM_TYPE_PTR(GTimeVal_, GTimeVal*)
XM_TYPE_PTR(GdkPixbufLoader_, GdkPixbufLoader*)
XM_TYPE(gsize, gsize)
#endif
#define C_TO_XEN_GtkType(Arg) C_TO_XEN_ULONG(Arg)
#define XEN_TO_C_GtkType(Arg) XEN_TO_C_ULONG(Arg)
#define XEN_GtkType_P(Arg) XEN_ULONG_P(Arg)
XM_TYPE_PTR(GtkAccelGroup_, GtkAccelGroup*)
#define C_TO_XEN_GtkAccelFlags(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_GtkAccelFlags(Arg) XEN_TO_C_INT(Arg)
#define XEN_GtkAccelFlags_P(Arg) XEN_INTEGER_P(Arg)
#if (!HAVE_GTK_1)
XM_TYPE_PTR(GObject_, GObject*)
#endif
XM_TYPE_PTR(GSList_, GSList*)
#if (!HAVE_GTK_1)
XM_TYPE_PTR(GtkAccelGroupEntry_, GtkAccelGroupEntry*)
#endif
XM_TYPE_PTR(GtkAccelLabel_, GtkAccelLabel*)
#if (!HAVE_GTK_1)
XM_TYPE(GtkAccelMapForeach, GtkAccelMapForeach)
#endif
XM_TYPE_PTR(GScanner_, GScanner*)
#if (!HAVE_GTK_1)
XM_TYPE_PTR(GtkAccessible_, GtkAccessible*)
#endif
XM_TYPE_PTR(GtkObject_, GtkObject*)
XM_TYPE_PTR(GtkAdjustment_, GtkAdjustment*)
XM_TYPE_PTR(GtkAlignment_, GtkAlignment*)
#define C_TO_XEN_GtkArrowType(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_GtkArrowType(Arg) XEN_TO_C_INT(Arg)
#define XEN_GtkArrowType_P(Arg) XEN_INTEGER_P(Arg)
#define C_TO_XEN_GtkShadowType(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_GtkShadowType(Arg) XEN_TO_C_INT(Arg)
#define XEN_GtkShadowType_P(Arg) XEN_INTEGER_P(Arg)
XM_TYPE_PTR(GtkArrow_, GtkArrow*)
XM_TYPE_PTR(GtkAspectFrame_, GtkAspectFrame*)
#define C_TO_XEN_GtkButtonBoxStyle(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_GtkButtonBoxStyle(Arg) XEN_TO_C_INT(Arg)
#define XEN_GtkButtonBoxStyle_P(Arg) XEN_INTEGER_P(Arg)
XM_TYPE_PTR(GtkButtonBox_, GtkButtonBox*)
XM_TYPE_PTR(GtkBindingSet_, GtkBindingSet*)
#define C_TO_XEN_GtkPathType(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_GtkPathType(Arg) XEN_TO_C_INT(Arg)
#define XEN_GtkPathType_P(Arg) XEN_INTEGER_P(Arg)
#define C_TO_XEN_GtkPathPriorityType(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_GtkPathPriorityType(Arg) XEN_TO_C_INT(Arg)
#define XEN_GtkPathPriorityType_P(Arg) XEN_INTEGER_P(Arg)
XM_TYPE_PTR(GtkBin_, GtkBin*)
XM_TYPE_PTR(GtkBox_, GtkBox*)
XM_TYPE_PTR(GtkPackType_, GtkPackType*)
#define C_TO_XEN_GtkPackType(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_GtkPackType(Arg) XEN_TO_C_INT(Arg)
#define XEN_GtkPackType_P(Arg) XEN_INTEGER_P(Arg)
XM_TYPE_PTR(GtkButton_, GtkButton*)
#define C_TO_XEN_GtkReliefStyle(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_GtkReliefStyle(Arg) XEN_TO_C_INT(Arg)
#define XEN_GtkReliefStyle_P(Arg) XEN_INTEGER_P(Arg)
XM_TYPE_PTR(GtkCalendar_, GtkCalendar*)
#define C_TO_XEN_GtkCalendarDisplayOptions(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_GtkCalendarDisplayOptions(Arg) XEN_TO_C_INT(Arg)
#define XEN_GtkCalendarDisplayOptions_P(Arg) XEN_INTEGER_P(Arg)
#if (!HAVE_GTK_1)
XM_TYPE_PTR(GtkCellEditable_, GtkCellEditable*)
#endif
#define C_TO_XEN_GtkCellRendererState(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_GtkCellRendererState(Arg) XEN_TO_C_INT(Arg)
#define XEN_GtkCellRendererState_P(Arg) XEN_INTEGER_P(Arg)
#if (!HAVE_GTK_1)
XM_TYPE_PTR(GtkCellRendererText_, GtkCellRendererText*)
XM_TYPE_PTR(GtkCellRendererToggle_, GtkCellRendererToggle*)
#endif
XM_TYPE_PTR(GtkCheckMenuItem_, GtkCheckMenuItem*)
XM_TYPE_PTR(GtkTargetEntry_, GtkTargetEntry*)
XM_TYPE_PTR(GtkColorSelection_, GtkColorSelection*)
#define C_TO_XEN_guint16(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_guint16(Arg) XEN_TO_C_INT(Arg)
#define XEN_guint16_P(Arg) XEN_INTEGER_P(Arg)
#if (!HAVE_GTK_1)
XM_TYPE(GtkColorSelectionChangePaletteFunc, GtkColorSelectionChangePaletteFunc)
#endif
XM_TYPE_PTR(GtkCombo_, GtkCombo*)
XM_TYPE_PTR(GtkItem_, GtkItem*)
XM_TYPE_PTR(GtkContainer_, GtkContainer*)
#define C_TO_XEN_GtkResizeMode(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_GtkResizeMode(Arg) XEN_TO_C_INT(Arg)
#define XEN_GtkResizeMode_P(Arg) XEN_INTEGER_P(Arg)
XM_TYPE_PTR(GtkCurve_, GtkCurve*)
XM_TYPE_PTR(gfloat_, gfloat*)
#define C_TO_XEN_GtkCurveType(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_GtkCurveType(Arg) XEN_TO_C_INT(Arg)
#define XEN_GtkCurveType_P(Arg) XEN_INTEGER_P(Arg)
#define C_TO_XEN_GtkDialogFlags(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_GtkDialogFlags(Arg) XEN_TO_C_INT(Arg)
#define XEN_GtkDialogFlags_P(Arg) XEN_INTEGER_P(Arg)
XM_TYPE_PTR(GtkDialog_, GtkDialog*)
#define C_TO_XEN_GtkDestDefaults(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_GtkDestDefaults(Arg) XEN_TO_C_INT(Arg)
#define XEN_GtkDestDefaults_P(Arg) XEN_INTEGER_P(Arg)
XM_TYPE_PTR(GtkTargetList_, GtkTargetList*)
XM_TYPE_PTR(GtkEditable_, GtkEditable*)
XM_TYPE_PTR(GtkEntry_, GtkEntry*)
#define C_TO_XEN_gunichar(Arg) C_TO_XEN_ULONG(Arg)
#define XEN_TO_C_gunichar(Arg) XEN_TO_C_ULONG(Arg)
#define XEN_gunichar_P(Arg) XEN_ULONG_P(Arg)
XM_TYPE_PTR(GtkFileSelection_, GtkFileSelection*)
XM_TYPE_PTR(GtkFixed_, GtkFixed*)
XM_TYPE_PTR(GtkFontSelection_, GtkFontSelection*)
XM_TYPE_PTR(GtkFontSelectionDialog_, GtkFontSelectionDialog*)
XM_TYPE_PTR(GtkFrame_, GtkFrame*)
XM_TYPE_PTR(GtkHandleBox_, GtkHandleBox*)
#define C_TO_XEN_GtkPositionType(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_GtkPositionType(Arg) XEN_TO_C_INT(Arg)
#define XEN_GtkPositionType_P(Arg) XEN_INTEGER_P(Arg)
#if (!HAVE_GTK_1)
XM_TYPE_PTR(GtkIconFactory_, GtkIconFactory*)
XM_TYPE_PTR(GtkIconSet_, GtkIconSet*)
#endif
#define C_TO_XEN_GtkIconSize(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_GtkIconSize(Arg) XEN_TO_C_INT(Arg)
#define XEN_GtkIconSize_P(Arg) XEN_INTEGER_P(Arg)
XM_TYPE_PTR(GtkStyle_, GtkStyle*)
#define C_TO_XEN_GtkTextDirection(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_GtkTextDirection(Arg) XEN_TO_C_INT(Arg)
#define XEN_GtkTextDirection_P(Arg) XEN_INTEGER_P(Arg)
#define C_TO_XEN_GtkStateType(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_GtkStateType(Arg) XEN_TO_C_INT(Arg)
#define XEN_GtkStateType_P(Arg) XEN_INTEGER_P(Arg)
#if (!HAVE_GTK_1)
XM_TYPE_PTR(GtkIconSource_, GtkIconSource*)
XM_TYPE_PTR(GtkIconSize_, GtkIconSize*)
#endif
XM_TYPE_PTR(GtkImage_, GtkImage*)
#define C_TO_XEN_GtkImageType(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_GtkImageType(Arg) XEN_TO_C_INT(Arg)
#define XEN_GtkImageType_P(Arg) XEN_INTEGER_P(Arg)
#if (!HAVE_GTK_1)
XM_TYPE_PTR(GtkImageMenuItem_, GtkImageMenuItem*)
XM_TYPE_PTR(GtkIMContext_, GtkIMContext*)
XM_TYPE_PTR(PangoAttrList_, PangoAttrList*)
XM_TYPE_PTR(GtkIMContextSimple_, GtkIMContextSimple*)
#endif
XM_TYPE_PTR(guint16_, guint16*)
#if (!HAVE_GTK_1)
XM_TYPE_PTR(GtkIMMulticontext_, GtkIMMulticontext*)
#endif
XM_TYPE_PTR(GtkMenuShell_, GtkMenuShell*)
XM_TYPE_PTR(GtkItemFactory_, GtkItemFactory*)
XM_TYPE_PTR(GtkItemFactoryEntry_, GtkItemFactoryEntry*)
XM_TYPE_PTR(GtkLabel_, GtkLabel*)
#define C_TO_XEN_GtkJustification(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_GtkJustification(Arg) XEN_TO_C_INT(Arg)
#define XEN_GtkJustification_P(Arg) XEN_INTEGER_P(Arg)
XM_TYPE_PTR(GtkLayout_, GtkLayout*)
#if (!HAVE_GTK_1)
XM_TYPE_PTR(GtkListStore_, GtkListStore*)
#endif
XM_TYPE_PTR(int_, int*)
XM_TYPE_PTR(char___, char***)
#if (!HAVE_GTK_1)
XM_TYPE_PTR(PangoLanguage_, PangoLanguage*)
#endif
XM_TYPE(GtkCallbackMarshal, GtkCallbackMarshal)
#define XEN_xen_P(Arg) 1
#define XEN_TO_C_xen(Arg) ((gpointer)Arg)
XM_TYPE(GtkMenuDetachFunc, GtkMenuDetachFunc)
XM_TYPE_PTR(GtkMenuItem_, GtkMenuItem*)
#define C_TO_XEN_GtkMessageType(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_GtkMessageType(Arg) XEN_TO_C_INT(Arg)
#define XEN_GtkMessageType_P(Arg) XEN_INTEGER_P(Arg)
#define C_TO_XEN_GtkButtonsType(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_GtkButtonsType(Arg) XEN_TO_C_INT(Arg)
#define XEN_GtkButtonsType_P(Arg) XEN_INTEGER_P(Arg)
XM_TYPE_PTR(GtkMisc_, GtkMisc*)
XM_TYPE_PTR(GtkNotebook_, GtkNotebook*)
XM_TYPE_PTR(GtkOptionMenu_, GtkOptionMenu*)
XM_TYPE_PTR(GtkPaned_, GtkPaned*)
XM_TYPE_PTR(GtkPlug_, GtkPlug*)
XM_TYPE_PTR(GtkProgressBar_, GtkProgressBar*)
#define C_TO_XEN_GtkProgressBarOrientation(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_GtkProgressBarOrientation(Arg) XEN_TO_C_INT(Arg)
#define XEN_GtkProgressBarOrientation_P(Arg) XEN_INTEGER_P(Arg)
XM_TYPE_PTR(GtkRadioButton_, GtkRadioButton*)
XM_TYPE_PTR(GtkRadioMenuItem_, GtkRadioMenuItem*)
XM_TYPE_PTR(GtkRange_, GtkRange*)
#define C_TO_XEN_GtkUpdateType(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_GtkUpdateType(Arg) XEN_TO_C_INT(Arg)
#define XEN_GtkUpdateType_P(Arg) XEN_INTEGER_P(Arg)
#if (!HAVE_GTK_1)
XM_TYPE_PTR(GtkSettings_, GtkSettings*)
#endif
XM_TYPE_PTR(GtkRcStyle_, GtkRcStyle*)
XM_TYPE_PTR(GtkStateType_, GtkStateType*)
XM_TYPE_PTR(GtkPathPriorityType_, GtkPathPriorityType*)
XM_TYPE_PTR(GtkRuler_, GtkRuler*)
#define C_TO_XEN_GtkMetricType(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_GtkMetricType(Arg) XEN_TO_C_INT(Arg)
#define XEN_GtkMetricType_P(Arg) XEN_INTEGER_P(Arg)
XM_TYPE_PTR(GtkScale_, GtkScale*)
XM_TYPE_PTR(GtkScrolledWindow_, GtkScrolledWindow*)
#define C_TO_XEN_GtkPolicyType(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_GtkPolicyType(Arg) XEN_TO_C_INT(Arg)
#define XEN_GtkPolicyType_P(Arg) XEN_INTEGER_P(Arg)
#define C_TO_XEN_GtkCornerType(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_GtkCornerType(Arg) XEN_TO_C_INT(Arg)
#define XEN_GtkCornerType_P(Arg) XEN_INTEGER_P(Arg)
XM_TYPE_PTR(GdkEventSelection_, GdkEventSelection*)
XM_TYPE_PTR(GdkEventProperty_, GdkEventProperty*)
#if (!HAVE_GTK_1)
XM_TYPE_PTR(GParamSpec_, GParamSpec*)
XM_TYPE(GtkRcPropertyParser, GtkRcPropertyParser)
XM_TYPE_PTR(GString_, GString*)
XM_TYPE_PTR(GtkSettingsValue_, GtkSettingsValue*)
#endif
#define C_TO_XEN_glong(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_glong(Arg) XEN_TO_C_INT(Arg)
#define XEN_glong_P(Arg) XEN_INTEGER_P(Arg)
#if (!HAVE_GTK_1)
XM_TYPE_PTR(GtkSizeGroup_, GtkSizeGroup*)
#endif
#define C_TO_XEN_GtkSizeGroupMode(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_GtkSizeGroupMode(Arg) XEN_TO_C_INT(Arg)
#define XEN_GtkSizeGroupMode_P(Arg) XEN_INTEGER_P(Arg)
XM_TYPE_PTR(GtkSocket_, GtkSocket*)
XM_TYPE_PTR(GtkSpinButton_, GtkSpinButton*)
#define C_TO_XEN_GtkSpinButtonUpdatePolicy(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_GtkSpinButtonUpdatePolicy(Arg) XEN_TO_C_INT(Arg)
#define XEN_GtkSpinButtonUpdatePolicy_P(Arg) XEN_INTEGER_P(Arg)
#define C_TO_XEN_GtkSpinType(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_GtkSpinType(Arg) XEN_TO_C_INT(Arg)
#define XEN_GtkSpinType_P(Arg) XEN_INTEGER_P(Arg)
XM_TYPE_PTR(GtkStatusbar_, GtkStatusbar*)
#if (!HAVE_GTK_1)
XM_TYPE_PTR(GtkStockItem_, GtkStockItem*)
#endif
#define C_TO_XEN_GtkOrientation(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_GtkOrientation(Arg) XEN_TO_C_INT(Arg)
#define XEN_GtkOrientation_P(Arg) XEN_INTEGER_P(Arg)
#define C_TO_XEN_GtkExpanderStyle(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_GtkExpanderStyle(Arg) XEN_TO_C_INT(Arg)
#define XEN_GtkExpanderStyle_P(Arg) XEN_INTEGER_P(Arg)
#if (!HAVE_GTK_1)
XM_TYPE_PTR(GtkBorder_, GtkBorder*)
#endif
XM_TYPE_PTR(GtkTable_, GtkTable*)
#define C_TO_XEN_GtkAttachOptions(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_GtkAttachOptions(Arg) XEN_TO_C_INT(Arg)
#define XEN_GtkAttachOptions_P(Arg) XEN_INTEGER_P(Arg)
#if (!HAVE_GTK_1)
XM_TYPE_PTR(GtkTextBuffer_, GtkTextBuffer*)
XM_TYPE_PTR(GtkTextTagTable_, GtkTextTagTable*)
XM_TYPE_PTR(GtkTextIter_, GtkTextIter*)
XM_TYPE_PTR(GtkTextTag_, GtkTextTag*)
XM_TYPE_PTR(GtkTextChildAnchor_, GtkTextChildAnchor*)
XM_TYPE_PTR(GtkTextMark_, GtkTextMark*)
XM_TYPE_PTR(GtkTextAttributes_, GtkTextAttributes*)
XM_TYPE(GtkTextCharPredicate, GtkTextCharPredicate)
#endif
#define C_TO_XEN_GtkTextSearchFlags(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_GtkTextSearchFlags(Arg) XEN_TO_C_INT(Arg)
#define XEN_GtkTextSearchFlags_P(Arg) XEN_INTEGER_P(Arg)
#if (!HAVE_GTK_1)
XM_TYPE(GtkTextTagTableForeach, GtkTextTagTableForeach)
XM_TYPE_PTR(GtkTextView_, GtkTextView*)
#endif
#define C_TO_XEN_GtkTextWindowType(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_GtkTextWindowType(Arg) XEN_TO_C_INT(Arg)
#define XEN_GtkTextWindowType_P(Arg) XEN_INTEGER_P(Arg)
#define C_TO_XEN_GtkWrapMode(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_GtkWrapMode(Arg) XEN_TO_C_INT(Arg)
#define XEN_GtkWrapMode_P(Arg) XEN_INTEGER_P(Arg)
#if (!HAVE_GTK_1)
XM_TYPE_PTR(PangoTabArray_, PangoTabArray*)
#endif
XM_TYPE_PTR(GtkToggleButton_, GtkToggleButton*)
XM_TYPE_PTR(GtkToolbar_, GtkToolbar*)
#define C_TO_XEN_GtkToolbarChildType(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_GtkToolbarChildType(Arg) XEN_TO_C_INT(Arg)
#define XEN_GtkToolbarChildType_P(Arg) XEN_INTEGER_P(Arg)
#define C_TO_XEN_GtkToolbarStyle(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_GtkToolbarStyle(Arg) XEN_TO_C_INT(Arg)
#define XEN_GtkToolbarStyle_P(Arg) XEN_INTEGER_P(Arg)
XM_TYPE_PTR(GtkTooltips_, GtkTooltips*)
XM_TYPE_PTR(GtkTooltipsData_, GtkTooltipsData*)
#if (!HAVE_GTK_1)
XM_TYPE_PTR(GtkTreeDragSource_, GtkTreeDragSource*)
XM_TYPE_PTR(GtkTreeDragDest_, GtkTreeDragDest*)
XM_TYPE_PTR(GtkTreeRowReference_, GtkTreeRowReference*)
#endif
#define C_TO_XEN_GtkTreeModelFlags(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_GtkTreeModelFlags(Arg) XEN_TO_C_INT(Arg)
#define XEN_GtkTreeModelFlags_P(Arg) XEN_INTEGER_P(Arg)
#if (!HAVE_GTK_1)
XM_TYPE_PTR(GtkTreeModelSort_, GtkTreeModelSort*)
#endif
#define C_TO_XEN_GtkSelectionMode(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_GtkSelectionMode(Arg) XEN_TO_C_INT(Arg)
#define XEN_GtkSelectionMode_P(Arg) XEN_INTEGER_P(Arg)
#if (!HAVE_GTK_1)
XM_TYPE_PTR(GtkTreeModel__, GtkTreeModel**)
XM_TYPE_PTR(GtkTreeSortable_, GtkTreeSortable*)
#endif
XM_TYPE_PTR(GtkSortType_, GtkSortType*)
#define C_TO_XEN_GtkSortType(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_GtkSortType(Arg) XEN_TO_C_INT(Arg)
#define XEN_GtkSortType_P(Arg) XEN_INTEGER_P(Arg)
#if (!HAVE_GTK_1)
XM_TYPE_PTR(GtkTreeStore_, GtkTreeStore*)
#endif
#define C_TO_XEN_GtkTreeViewColumnSizing(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_GtkTreeViewColumnSizing(Arg) XEN_TO_C_INT(Arg)
#define XEN_GtkTreeViewColumnSizing_P(Arg) XEN_INTEGER_P(Arg)
#define C_TO_XEN_GtkTreeViewDropPosition(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_GtkTreeViewDropPosition(Arg) XEN_TO_C_INT(Arg)
#define XEN_GtkTreeViewDropPosition_P(Arg) XEN_INTEGER_P(Arg)
XM_TYPE_PTR(GtkViewport_, GtkViewport*)
XM_TYPE_PTR(GtkRequisition_, GtkRequisition*)
XM_TYPE_PTR(GtkAllocation_, GtkAllocation*)
#define C_TO_XEN_GtkDirectionType(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_GtkDirectionType(Arg) XEN_TO_C_INT(Arg)
#define XEN_GtkDirectionType_P(Arg) XEN_INTEGER_P(Arg)
#if (!HAVE_GTK_1)
XM_TYPE_PTR(AtkObject_, AtkObject*)
XM_TYPE_PTR(PangoFontDescription_, PangoFontDescription*)
#endif
XM_TYPE_PTR(GtkWidgetAuxInfo_, GtkWidgetAuxInfo*)
#define C_TO_XEN_GtkWindowType(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_GtkWindowType(Arg) XEN_TO_C_INT(Arg)
#define XEN_GtkWindowType_P(Arg) XEN_INTEGER_P(Arg)
#define C_TO_XEN_GtkWindowPosition(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_GtkWindowPosition(Arg) XEN_TO_C_INT(Arg)
#define XEN_GtkWindowPosition_P(Arg) XEN_INTEGER_P(Arg)
#define C_TO_XEN_GdkGravity(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_GdkGravity(Arg) XEN_TO_C_INT(Arg)
#define XEN_GdkGravity_P(Arg) XEN_INTEGER_P(Arg)
#if (!HAVE_GTK_1)
XM_TYPE_PTR(PangoColor_, PangoColor*)
#endif
#define C_TO_XEN_PangoAttrType(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_PangoAttrType(Arg) XEN_TO_C_INT(Arg)
#define XEN_PangoAttrType_P(Arg) XEN_INTEGER_P(Arg)
#define C_TO_XEN_PangoStyle(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_PangoStyle(Arg) XEN_TO_C_INT(Arg)
#define XEN_PangoStyle_P(Arg) XEN_INTEGER_P(Arg)
#define C_TO_XEN_PangoWeight(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_PangoWeight(Arg) XEN_TO_C_INT(Arg)
#define XEN_PangoWeight_P(Arg) XEN_INTEGER_P(Arg)
#define C_TO_XEN_PangoVariant(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_PangoVariant(Arg) XEN_TO_C_INT(Arg)
#define XEN_PangoVariant_P(Arg) XEN_INTEGER_P(Arg)
#define C_TO_XEN_PangoStretch(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_PangoStretch(Arg) XEN_TO_C_INT(Arg)
#define XEN_PangoStretch_P(Arg) XEN_INTEGER_P(Arg)
#define C_TO_XEN_PangoUnderline(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_PangoUnderline(Arg) XEN_TO_C_INT(Arg)
#define XEN_PangoUnderline_P(Arg) XEN_INTEGER_P(Arg)
#if (!HAVE_GTK_1)
XM_TYPE_PTR(PangoRectangle_, PangoRectangle*)
XM_TYPE_PTR(PangoAttrIterator_, PangoAttrIterator*)
XM_TYPE_PTR(PangoAttrList__, PangoAttrList**)
XM_TYPE_PTR(gunichar_, gunichar*)
XM_TYPE_PTR(PangoAnalysis_, PangoAnalysis*)
XM_TYPE_PTR(PangoLogAttr_, PangoLogAttr*)
XM_TYPE_PTR(PangoFontFamily__, PangoFontFamily**)
XM_TYPE_PTR(PangoFontset_, PangoFontset*)
XM_TYPE_PTR(PangoFontMetrics_, PangoFontMetrics*)
XM_TYPE_PTR(PangoCoverage_, PangoCoverage*)
#endif
#define C_TO_XEN_PangoCoverageLevel(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_PangoCoverageLevel(Arg) XEN_TO_C_INT(Arg)
#define XEN_PangoCoverageLevel_P(Arg) XEN_INTEGER_P(Arg)
#if (!HAVE_GTK_1)
XM_TYPE_PTR(PangoFontDescription__, PangoFontDescription**)
#endif
#define C_TO_XEN_PangoFontMask(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_PangoFontMask(Arg) XEN_TO_C_INT(Arg)
#define XEN_PangoFontMask_P(Arg) XEN_INTEGER_P(Arg)
#if (!HAVE_GTK_1)
XM_TYPE_PTR(PangoFontFace_, PangoFontFace*)
XM_TYPE_PTR(PangoEngineShape_, PangoEngineShape*)
#endif
#define C_TO_XEN_PangoGlyph(Arg) C_TO_XEN_ULONG(Arg)
#define XEN_TO_C_PangoGlyph(Arg) XEN_TO_C_ULONG(Arg)
#define XEN_PangoGlyph_P(Arg) XEN_ULONG_P(Arg)
#if (!HAVE_GTK_1)
XM_TYPE_PTR(PangoFontMap_, PangoFontMap*)
XM_TYPE_PTR(PangoItem_, PangoItem*)
#endif
#define C_TO_XEN_PangoWrapMode(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_PangoWrapMode(Arg) XEN_TO_C_INT(Arg)
#define XEN_PangoWrapMode_P(Arg) XEN_INTEGER_P(Arg)
#define C_TO_XEN_PangoAlignment(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_PangoAlignment(Arg) XEN_TO_C_INT(Arg)
#define XEN_PangoAlignment_P(Arg) XEN_INTEGER_P(Arg)
#if (!HAVE_GTK_1)
XM_TYPE_PTR(PangoLayoutIter_, PangoLayoutIter*)
XM_TYPE_PTR(PangoLayoutRun_, PangoLayoutRun*)
#endif
#define C_TO_XEN_GdkEventType(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_GdkEventType(Arg) XEN_TO_C_INT(Arg)
#define XEN_GdkEventType_P(Arg) XEN_INTEGER_P(Arg)
#define C_TO_XEN_gint8(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_gint8(Arg) XEN_TO_C_INT(Arg)
#define XEN_gint8_P(Arg) XEN_INTEGER_P(Arg)
XM_TYPE_PTR(GdkEventAny_, GdkEventAny*)
XM_TYPE_PTR(GdkEventExpose_, GdkEventExpose*)
XM_TYPE_PTR(GdkEventNoExpose_, GdkEventNoExpose*)
#define C_TO_XEN_GdkVisibilityState(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_GdkVisibilityState(Arg) XEN_TO_C_INT(Arg)
#define XEN_GdkVisibilityState_P(Arg) XEN_INTEGER_P(Arg)
XM_TYPE_PTR(GdkEventVisibility_, GdkEventVisibility*)
#define C_TO_XEN_gint16(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_gint16(Arg) XEN_TO_C_INT(Arg)
#define XEN_gint16_P(Arg) XEN_INTEGER_P(Arg)
XM_TYPE_PTR(GdkEventMotion_, GdkEventMotion*)
XM_TYPE_PTR(GdkEventButton_, GdkEventButton*)
#define C_TO_XEN_GdkScrollDirection(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_GdkScrollDirection(Arg) XEN_TO_C_INT(Arg)
#define XEN_GdkScrollDirection_P(Arg) XEN_INTEGER_P(Arg)
#if (!HAVE_GTK_1)
XM_TYPE_PTR(GdkEventScroll_, GdkEventScroll*)
#endif
#define C_TO_XEN_guint8(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_guint8(Arg) XEN_TO_C_INT(Arg)
#define XEN_guint8_P(Arg) XEN_INTEGER_P(Arg)
#define C_TO_XEN_GdkCrossingMode(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_GdkCrossingMode(Arg) XEN_TO_C_INT(Arg)
#define XEN_GdkCrossingMode_P(Arg) XEN_INTEGER_P(Arg)
#define C_TO_XEN_GdkNotifyType(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_GdkNotifyType(Arg) XEN_TO_C_INT(Arg)
#define XEN_GdkNotifyType_P(Arg) XEN_INTEGER_P(Arg)
XM_TYPE_PTR(GdkEventCrossing_, GdkEventCrossing*)
XM_TYPE_PTR(GdkEventFocus_, GdkEventFocus*)
XM_TYPE_PTR(GdkEventConfigure_, GdkEventConfigure*)
XM_TYPE_PTR(GdkEventProximity_, GdkEventProximity*)
#define C_TO_XEN_GdkSettingAction(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_GdkSettingAction(Arg) XEN_TO_C_INT(Arg)
#define XEN_GdkSettingAction_P(Arg) XEN_INTEGER_P(Arg)
#if (!HAVE_GTK_1)
XM_TYPE_PTR(GdkEventSetting_, GdkEventSetting*)
XM_TYPE_PTR(GdkEventWindowState_, GdkEventWindowState*)
#endif
#define C_TO_XEN_gshort(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_gshort(Arg) XEN_TO_C_INT(Arg)
#define XEN_gshort_P(Arg) XEN_INTEGER_P(Arg)
XM_TYPE_PTR(GdkEventDND_, GdkEventDND*)
#define C_TO_XEN_GdkByteOrder(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_GdkByteOrder(Arg) XEN_TO_C_INT(Arg)
#define XEN_GdkByteOrder_P(Arg) XEN_INTEGER_P(Arg)
#if (!HAVE_GTK_1)
XM_TYPE_PTR(GdkDeviceAxis_, GdkDeviceAxis*)
#endif
XM_TYPE_PTR(GdkDeviceKey_, GdkDeviceKey*)
#if (!HAVE_GTK_1)
XM_TYPE_PTR(GdkWindowObject_, GdkWindowObject*)
XM_TYPE_PTR(GtkCellRendererPixbuf_, GtkCellRendererPixbuf*)
#endif
XM_TYPE_PTR(GtkColorSelectionDialog_, GtkColorSelectionDialog*)
XM_TYPE_PTR(GtkDrawingArea_, GtkDrawingArea*)
XM_TYPE_PTR(GtkWidget__, GtkWidget**)
XM_TYPE_PTR(GtkGammaCurve_, GtkGammaCurve*)
XM_TYPE_PTR(GtkInputDialog_, GtkInputDialog*)
#if (!HAVE_GTK_1)
XM_TYPE_PTR(GtkInvisible_, GtkInvisible*)
#endif
XM_TYPE(GtkItemFactoryCallback, GtkItemFactoryCallback)
#if (!HAVE_GTK_1)
XM_TYPE_PTR(GtkLabelSelectionInfo_, GtkLabelSelectionInfo*)
XM_TYPE_PTR(GtkMessageDialog_, GtkMessageDialog*)
#endif
XM_TYPE_PTR(GtkNotebookPage_, GtkNotebookPage*)
#if (!HAVE_GTK_1)
XM_TYPE_PTR(GtkRangeLayout_, GtkRangeLayout*)
XM_TYPE_PTR(GtkRangeStepTimer_, GtkRangeStepTimer*)
#endif
XM_TYPE_PTR(GtkRulerMetric_, GtkRulerMetric*)
XM_TYPE_PTR(GData_, GData*)
#if (!HAVE_GTK_1)
XM_TYPE_PTR(GtkRcContext_, GtkRcContext*)
#endif
XM_TYPE_PTR(GdkGC__, GdkGC**)
XM_TYPE_PTR(GdkPixmap__, GdkPixmap**)
#if (!HAVE_GTK_1)
XM_TYPE_PTR(GArray_, GArray*)
#endif
XM_TYPE_PTR(GtkTableRowCol_, GtkTableRowCol*)
#if (!HAVE_GTK_1)
XM_TYPE_PTR(GtkTextBTree_, GtkTextBTree*)
XM_TYPE_PTR(GtkTextLogAttrCache_, GtkTextLogAttrCache*)
#endif
static XEN XEN_GdkColormap__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GdkColormap_", val)));}
static XEN XEN_GdkDragContext__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GdkDragContext_", val)));}
static XEN XEN_GdkDrawable__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GdkDrawable_", val)));}
static XEN XEN_GdkGC__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GdkGC_", val)));}
static XEN XEN_GdkImage__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GdkImage_", val)));}
#if (!HAVE_GTK_1)
static XEN XEN_GdkDevice__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GdkDevice_", val)));}
static XEN XEN_GdkKeymap__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GdkKeymap_", val)));}
#endif
static XEN XEN_GdkPixmap__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GdkPixmap_", val)));}
static XEN XEN_GdkVisual__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GdkVisual_", val)));}
static XEN XEN_GdkWindow__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GdkWindow_", val)));}
#if (!HAVE_GTK_1)
static XEN XEN_GdkPixbuf__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GdkPixbuf_", val)));}
static XEN XEN_GdkPixbufAnimation__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GdkPixbufAnimation_", val)));}
static XEN XEN_GdkPixbufAnimationIter__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GdkPixbufAnimationIter_", val)));}
static XEN XEN_GdkPixbufLoader__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GdkPixbufLoader_", val)));}
static XEN XEN_GtkVBox__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkVBox_", val)));}
#endif
static XEN XEN_GtkAccelGroup__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkAccelGroup_", val)));}
static XEN XEN_GtkAccelLabel__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkAccelLabel_", val)));}
#if (!HAVE_GTK_1)
static XEN XEN_GtkAccessible__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkAccessible_", val)));}
#endif
static XEN XEN_GtkAdjustment__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkAdjustment_", val)));}
static XEN XEN_GtkAlignment__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkAlignment_", val)));}
static XEN XEN_GtkArrow__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkArrow_", val)));}
static XEN XEN_GtkAspectFrame__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkAspectFrame_", val)));}
static XEN XEN_GtkButtonBox__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkButtonBox_", val)));}
static XEN XEN_GtkBin__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkBin_", val)));}
static XEN XEN_GtkBox__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkBox_", val)));}
static XEN XEN_GtkButton__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkButton_", val)));}
static XEN XEN_GtkCalendar__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkCalendar_", val)));}
#if (!HAVE_GTK_1)
static XEN XEN_GtkCellEditable__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkCellEditable_", val)));}
static XEN XEN_GtkCellRenderer__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkCellRenderer_", val)));}
static XEN XEN_GtkCellRendererPixbuf__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkCellRendererPixbuf_", val)));}
static XEN XEN_GtkCellRendererText__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkCellRendererText_", val)));}
static XEN XEN_GtkCellRendererToggle__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkCellRendererToggle_", val)));}
static XEN XEN_GtkCheckButton__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkCheckButton_", val)));}
#endif
static XEN XEN_GtkCheckMenuItem__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkCheckMenuItem_", val)));}
static XEN XEN_GtkColorSelectionDialog__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkColorSelectionDialog_", val)));}
static XEN XEN_GtkColorSelection__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkColorSelection_", val)));}
static XEN XEN_GtkCombo__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkCombo_", val)));}
static XEN XEN_GtkContainer__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkContainer_", val)));}
static XEN XEN_GtkCurve__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkCurve_", val)));}
static XEN XEN_GtkDialog__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkDialog_", val)));}
static XEN XEN_GtkDrawingArea__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkDrawingArea_", val)));}
static XEN XEN_GtkEditable__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkEditable_", val)));}
static XEN XEN_GtkEntry__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkEntry_", val)));}
#if (!HAVE_GTK_1)
static XEN XEN_GtkEventBox__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkEventBox_", val)));}
#endif
static XEN XEN_GtkFileSelection__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkFileSelection_", val)));}
static XEN XEN_GtkFixed__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkFixed_", val)));}
static XEN XEN_GtkFontSelection__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkFontSelection_", val)));}
static XEN XEN_GtkFontSelectionDialog__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkFontSelectionDialog_", val)));}
static XEN XEN_GtkFrame__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkFrame_", val)));}
static XEN XEN_GtkGammaCurve__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkGammaCurve_", val)));}
static XEN XEN_GtkHandleBox__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkHandleBox_", val)));}
#if (!HAVE_GTK_1)
static XEN XEN_GtkHButtonBox__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkHButtonBox_", val)));}
static XEN XEN_GtkHBox__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkHBox_", val)));}
static XEN XEN_GtkHPaned__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkHPaned_", val)));}
static XEN XEN_GtkHRuler__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkHRuler_", val)));}
static XEN XEN_GtkHScale__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkHScale_", val)));}
static XEN XEN_GtkHScrollbar__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkHScrollbar_", val)));}
static XEN XEN_GtkHSeparator__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkHSeparator_", val)));}
static XEN XEN_GtkIconFactory__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkIconFactory_", val)));}
#endif
static XEN XEN_GtkImage__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkImage_", val)));}
#if (!HAVE_GTK_1)
static XEN XEN_GtkImageMenuItem__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkImageMenuItem_", val)));}
static XEN XEN_GtkIMContext__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkIMContext_", val)));}
static XEN XEN_GtkIMContextSimple__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkIMContextSimple_", val)));}
static XEN XEN_GtkIMMulticontext__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkIMMulticontext_", val)));}
#endif
static XEN XEN_GtkInputDialog__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkInputDialog_", val)));}
#if (!HAVE_GTK_1)
static XEN XEN_GtkInvisible__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkInvisible_", val)));}
#endif
static XEN XEN_GtkItemFactory__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkItemFactory_", val)));}
static XEN XEN_GtkItem__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkItem_", val)));}
static XEN XEN_GtkLabel__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkLabel_", val)));}
static XEN XEN_GtkLayout__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkLayout_", val)));}
#if (!HAVE_GTK_1)
static XEN XEN_GtkListStore__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkListStore_", val)));}
static XEN XEN_GtkMenuBar__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkMenuBar_", val)));}
#endif
static XEN XEN_GtkMenu__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkMenu_", val)));}
static XEN XEN_GtkMenuItem__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkMenuItem_", val)));}
static XEN XEN_GtkMenuShell__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkMenuShell_", val)));}
#if (!HAVE_GTK_1)
static XEN XEN_GtkMessageDialog__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkMessageDialog_", val)));}
#endif
static XEN XEN_GtkMisc__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkMisc_", val)));}
static XEN XEN_GtkNotebook__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkNotebook_", val)));}
static XEN XEN_GtkObject__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkObject_", val)));}
static XEN XEN_GtkOptionMenu__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkOptionMenu_", val)));}
static XEN XEN_GtkPaned__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkPaned_", val)));}
static XEN XEN_GtkPlug__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkPlug_", val)));}
static XEN XEN_GtkProgressBar__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkProgressBar_", val)));}
static XEN XEN_GtkRadioButton__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkRadioButton_", val)));}
static XEN XEN_GtkRadioMenuItem__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkRadioMenuItem_", val)));}
static XEN XEN_GtkRange__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkRange_", val)));}
static XEN XEN_GtkRcStyle__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkRcStyle_", val)));}
static XEN XEN_GtkRuler__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkRuler_", val)));}
static XEN XEN_GtkScale__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkScale_", val)));}
#if (!HAVE_GTK_1)
static XEN XEN_GtkScrollbar__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkScrollbar_", val)));}
#endif
static XEN XEN_GtkScrolledWindow__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkScrolledWindow_", val)));}
#if (!HAVE_GTK_1)
static XEN XEN_GtkSeparator__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkSeparator_", val)));}
static XEN XEN_GtkSeparatorMenuItem__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkSeparatorMenuItem_", val)));}
static XEN XEN_GtkSettings__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkSettings_", val)));}
static XEN XEN_GtkSizeGroup__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkSizeGroup_", val)));}
#endif
static XEN XEN_GtkSocket__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkSocket_", val)));}
static XEN XEN_GtkSpinButton__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkSpinButton_", val)));}
static XEN XEN_GtkStatusbar__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkStatusbar_", val)));}
static XEN XEN_GtkStyle__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkStyle_", val)));}
static XEN XEN_GtkTable__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkTable_", val)));}
#if (!HAVE_GTK_1)
static XEN XEN_GtkTearoffMenuItem__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkTearoffMenuItem_", val)));}
static XEN XEN_GtkTextBuffer__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkTextBuffer_", val)));}
static XEN XEN_GtkTextChildAnchor__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkTextChildAnchor_", val)));}
static XEN XEN_GtkTextMark__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkTextMark_", val)));}
static XEN XEN_GtkTextTag__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkTextTag_", val)));}
static XEN XEN_GtkTextTagTable__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkTextTagTable_", val)));}
static XEN XEN_GtkTextView__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkTextView_", val)));}
static XEN XEN_GtkThemeEngine__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkThemeEngine_", val)));}
#endif
static XEN XEN_GtkToggleButton__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkToggleButton_", val)));}
static XEN XEN_GtkToolbar__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkToolbar_", val)));}
static XEN XEN_GtkTooltips__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkTooltips_", val)));}
#if (!HAVE_GTK_1)
static XEN XEN_GtkTreeDragSource__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkTreeDragSource_", val)));}
static XEN XEN_GtkTreeDragDest__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkTreeDragDest_", val)));}
static XEN XEN_GtkTreeModel__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkTreeModel_", val)));}
static XEN XEN_GtkTreeModelSort__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkTreeModelSort_", val)));}
static XEN XEN_GtkTreeSelection__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkTreeSelection_", val)));}
static XEN XEN_GtkTreeSortable__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkTreeSortable_", val)));}
static XEN XEN_GtkTreeStore__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkTreeStore_", val)));}
static XEN XEN_GtkTreeViewColumn__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkTreeViewColumn_", val)));}
static XEN XEN_GtkTreeView__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkTreeView_", val)));}
static XEN XEN_GtkVButtonBox__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkVButtonBox_", val)));}
#endif
static XEN XEN_GtkViewport__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkViewport_", val)));}
#if (!HAVE_GTK_1)
static XEN XEN_GtkVPaned__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkVPaned_", val)));}
static XEN XEN_GtkVRuler__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkVRuler_", val)));}
static XEN XEN_GtkVScale__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkVScale_", val)));}
static XEN XEN_GtkVScrollbar__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkVScrollbar_", val)));}
static XEN XEN_GtkVSeparator__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkVSeparator_", val)));}
#endif
static XEN XEN_GtkWidget__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkWidget_", val)));}
static XEN XEN_GtkWindow__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkWindow_", val)));}
#if (!HAVE_GTK_1)
static XEN XEN_GtkWindowGroup__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkWindowGroup_", val)));}
static XEN XEN_PangoContext__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("PangoContext_", val)));}
static XEN XEN_PangoFontFace__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("PangoFontFace_", val)));}
static XEN XEN_PangoFont__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("PangoFont_", val)));}
static XEN XEN_PangoFontMap__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("PangoFontMap_", val)));}
static XEN XEN_PangoLayout__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("PangoLayout_", val)));}
#endif
#if PANGO_ENABLE_ENGINE && PANGO_ENABLE_BACKEND
#if (!HAVE_GTK_1)
XM_TYPE_PTR(PangoEngineInfo_, PangoEngineInfo*)
XM_TYPE_PTR(PangoEngine_, PangoEngine*)
XM_TYPE_PTR(PangoFontFamily_, PangoFontFamily*)
XM_TYPE_PTR(PangoFontFace__, PangoFontFace**)
static XEN XEN_PangoEngineInfo__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("PangoEngineInfo_", val)));}
static XEN XEN_PangoEngine__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("PangoEngine_", val)));}
static XEN XEN_PangoFontFamily__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("PangoFontFamily_", val)));}
static XEN XEN_PangoFontFace___p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("PangoFontFace__", val)));}
#endif
#endif

#if (!(defined(GDK_DISABLE_DEPRECATED))) && (!(defined(GTK_DISABLE_DEPRECATED))) && (!(defined(GDK_PIXBUF_DISABLE_DEPRECATED)))
XM_TYPE_PTR(gulong_, gulong*)
XM_TYPE_PTR(GdkFont_, GdkFont*)
#define C_TO_XEN_gchar(Arg) C_TO_XEN_CHAR(Arg)
#define XEN_TO_C_gchar(Arg) XEN_TO_C_CHAR(Arg)
#define XEN_gchar_P(Arg) XEN_CHAR_P(Arg)
XM_TYPE(GdkWChar, GdkWChar)
XM_TYPE(GdkDestroyNotify, GdkDestroyNotify)
XM_TYPE_PTR(GtkCList_, GtkCList*)
#define C_TO_XEN_GtkVisibility(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_GtkVisibility(Arg) XEN_TO_C_INT(Arg)
#define XEN_GtkVisibility_P(Arg) XEN_INTEGER_P(Arg)
#define C_TO_XEN_GtkCellType(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_GtkCellType(Arg) XEN_TO_C_INT(Arg)
#define XEN_GtkCellType_P(Arg) XEN_INTEGER_P(Arg)
XM_TYPE(GtkCListCompareFunc, GtkCListCompareFunc)
XM_TYPE_PTR(GList__, GList**)
XM_TYPE_PTR(GtkCTree_, GtkCTree*)
XM_TYPE_PTR(GtkCTreeNode_, GtkCTreeNode*)
XM_TYPE_PTR(GNode_, GNode*)
XM_TYPE(GtkCTreeGNodeFunc, GtkCTreeGNodeFunc)
XM_TYPE(GtkCTreeFunc, GtkCTreeFunc)
XM_TYPE_PTR(GtkCTreeRow_, GtkCTreeRow*)
XM_TYPE(GCompareFunc, GCompareFunc)
#define C_TO_XEN_GtkCTreeLineStyle(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_GtkCTreeLineStyle(Arg) XEN_TO_C_INT(Arg)
#define XEN_GtkCTreeLineStyle_P(Arg) XEN_INTEGER_P(Arg)
#define C_TO_XEN_GtkCTreeExpanderStyle(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_GtkCTreeExpanderStyle(Arg) XEN_TO_C_INT(Arg)
#define XEN_GtkCTreeExpanderStyle_P(Arg) XEN_INTEGER_P(Arg)
XM_TYPE(GtkCTreeCompareDragFunc, GtkCTreeCompareDragFunc)
XM_TYPE_PTR(GtkMenuEntry_, GtkMenuEntry*)
XM_TYPE_PTR(GtkList_, GtkList*)
#define C_TO_XEN_GtkScrollType(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_GtkScrollType(Arg) XEN_TO_C_INT(Arg)
#define XEN_GtkScrollType_P(Arg) XEN_INTEGER_P(Arg)
XM_TYPE_PTR(GtkListItem_, GtkListItem*)
#if (!HAVE_GTK_1)
XM_TYPE_PTR(GtkOldEditable_, GtkOldEditable*)
#endif
XM_TYPE_PTR(GtkPixmap_, GtkPixmap*)
XM_TYPE_PTR(GtkPreview_, GtkPreview*)
#define C_TO_XEN_GtkPreviewType(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_GtkPreviewType(Arg) XEN_TO_C_INT(Arg)
#define XEN_GtkPreviewType_P(Arg) XEN_INTEGER_P(Arg)
XM_TYPE_PTR(GtkPreviewInfo_, GtkPreviewInfo*)
#define C_TO_XEN_GtkProgressBarStyle(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_GtkProgressBarStyle(Arg) XEN_TO_C_INT(Arg)
#define XEN_GtkProgressBarStyle_P(Arg) XEN_INTEGER_P(Arg)
XM_TYPE_PTR(GtkProgress_, GtkProgress*)
#define C_TO_XEN_GtkSignalRunType(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_GtkSignalRunType(Arg) XEN_TO_C_INT(Arg)
#define XEN_GtkSignalRunType_P(Arg) XEN_INTEGER_P(Arg)
XM_TYPE(GtkSignalMarshaller, GtkSignalMarshaller)
XM_TYPE_PTR(GtkType_, GtkType*)
XM_TYPE_PTR(GtkTipsQuery_, GtkTipsQuery*)
static XEN XEN_gulong__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("gulong_", val)));}
static XEN XEN_GdkFont__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GdkFont_", val)));}
static XEN XEN_gchar_p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("gchar", val)));}
static XEN XEN_GdkWChar_p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GdkWChar", val)));}
static XEN XEN_GdkDestroyNotify_p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GdkDestroyNotify", val)));}
static XEN XEN_GtkCList__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkCList_", val)));}
static XEN XEN_GtkVisibility_p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkVisibility", val)));}
static XEN XEN_GtkCellType_p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkCellType", val)));}
static XEN XEN_GtkCListCompareFunc_p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkCListCompareFunc", val)));}
static XEN XEN_GList___p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GList__", val)));}
static XEN XEN_GtkCTree__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkCTree_", val)));}
static XEN XEN_GtkCTreeNode__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkCTreeNode_", val)));}
static XEN XEN_GNode__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GNode_", val)));}
static XEN XEN_GtkCTreeGNodeFunc_p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkCTreeGNodeFunc", val)));}
static XEN XEN_GtkCTreeFunc_p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkCTreeFunc", val)));}
static XEN XEN_GtkCTreeRow__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkCTreeRow_", val)));}
static XEN XEN_GCompareFunc_p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GCompareFunc", val)));}
static XEN XEN_GtkCTreeLineStyle_p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkCTreeLineStyle", val)));}
static XEN XEN_GtkCTreeExpanderStyle_p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkCTreeExpanderStyle", val)));}
static XEN XEN_GtkCTreeCompareDragFunc_p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkCTreeCompareDragFunc", val)));}
static XEN XEN_GtkMenuEntry__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkMenuEntry_", val)));}
static XEN XEN_GtkList__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkList_", val)));}
static XEN XEN_GtkScrollType_p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkScrollType", val)));}
static XEN XEN_GtkListItem__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkListItem_", val)));}
#if (!HAVE_GTK_1)
static XEN XEN_GtkOldEditable__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkOldEditable_", val)));}
#endif
static XEN XEN_GtkPixmap__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkPixmap_", val)));}
static XEN XEN_GtkPreview__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkPreview_", val)));}
static XEN XEN_GtkPreviewType_p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkPreviewType", val)));}
static XEN XEN_GtkPreviewInfo__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkPreviewInfo_", val)));}
static XEN XEN_GtkProgressBarStyle_p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkProgressBarStyle", val)));}
static XEN XEN_GtkProgress__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkProgress_", val)));}
static XEN XEN_lambda_p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("lambda", val)));}
static XEN XEN_GtkSignalRunType_p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkSignalRunType", val)));}
static XEN XEN_GtkSignalMarshaller_p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkSignalMarshaller", val)));}
static XEN XEN_GtkType__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkType_", val)));}
static XEN XEN_GtkTipsQuery__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkTipsQuery_", val)));}
#endif

#if GDK_ENABLE_BROKEN && GTK_ENABLE_BROKEN
XM_TYPE_PTR(GtkText_, GtkText*)
XM_TYPE_PTR(GtkTree_, GtkTree*)
#define C_TO_XEN_GtkTreeViewMode(Arg) C_TO_XEN_INT(Arg)
#define XEN_TO_C_GtkTreeViewMode(Arg) XEN_TO_C_INT(Arg)
#define XEN_GtkTreeViewMode_P(Arg) XEN_INTEGER_P(Arg)
XM_TYPE_PTR(GtkTreeItem_, GtkTreeItem*)
static XEN XEN_GtkText__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkText_", val)));}
static XEN XEN_GtkTree__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkTree_", val)));}
static XEN XEN_GtkTreeViewMode_p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkTreeViewMode", val)));}
static XEN XEN_GtkTreeItem__p(XEN val) {return(C_TO_XEN_BOOLEAN(WRAP_P("GtkTreeItem_", val)));}
#endif

/* -------------------------------- gc protection -------------------------------- */

static XEN xm_protected = XEN_FALSE;
static int xm_protected_size = 0;
static XEN xm_gc_table = XEN_FALSE;
static int last_xm_unprotect = -1;

static int xm_protect(XEN obj)
{
  int i, new_size;
  XEN new_table;
  XEN *older, *newer;
  older = XEN_VECTOR_ELEMENTS(xm_protected);
  if (last_xm_unprotect >= 0)
    {
      i = last_xm_unprotect;
      if (XEN_FALSE_P(older[i]))
	{
	  older[i] = obj;
	  last_xm_unprotect = -1;
	  return(i);
	}
      last_xm_unprotect = -1;
    }
  for (i = 0; i < xm_protected_size; i++)
    if (XEN_FALSE_P(older[i]))
      {
	older[i] = obj;
	return(i);
      }
  new_size = xm_protected_size * 2;
  new_table = XEN_MAKE_VECTOR(new_size, XEN_FALSE);
  newer = XEN_VECTOR_ELEMENTS(new_table);
  for (i = 0; i < xm_protected_size; i++)
    {
      newer[i] = older[i];
      older[i] = XEN_FALSE;
    }
  newer[xm_protected_size] = obj;
  XEN_VECTOR_SET(xm_gc_table, 0, new_table);
  i = xm_protected_size;
  xm_protected_size = new_size;
  xm_protected = new_table;
  return(i);
}

static void xm_unprotect_idler(guint id)
{
  int i;
  XEN *velts;
  XEN cur, idler;
  velts = XEN_VECTOR_ELEMENTS(xm_protected);
  for (i = 0; i < xm_protected_size; i++)
    {
      cur = velts[i];
      if ((XEN_LIST_P(cur)) && (XEN_LIST_LENGTH(cur) == 3) && (XEN_LIST_P(XEN_CADDR(cur))))
        {
          idler = XEN_CADDR(cur);
          if ((XEN_SYMBOL_P(XEN_CAR(idler))) &&
              (strcmp("idler", XEN_SYMBOL_TO_C_STRING(XEN_CAR(idler))) == 0) &&
              (id == XEN_TO_C_INT(XEN_CADR(idler))))
            {
              velts[i] = XEN_FALSE;
              last_xm_unprotect = i;
              return;
            }}}
}
static void xm_unprotect_at(int ind)
{
  XEN *velts;
  velts = XEN_VECTOR_ELEMENTS(xm_protected);
  velts[ind] = XEN_FALSE;
  last_xm_unprotect = ind;
}



/* ---------------------------------------- callback handlers ---------------------------------------- */


#if (!HAVE_GTK_1)
static gboolean gxg_child_func(GdkWindow* window, gpointer func_data)
{
  return(XEN_TO_C_gboolean(XEN_CALL_2(XEN_CAR((XEN)func_data),
                                      C_TO_XEN_GdkWindow_(window),
                                      XEN_CADR((XEN)func_data),
                                      __FUNCTION__)));
}

static gboolean gxg_find_func(GtkAccelKey* key, GClosure* closure, gpointer func_data)
{
  return(XEN_TO_C_gboolean(XEN_CALL_3(XEN_CAR((XEN)func_data),
                                      C_TO_XEN_GtkAccelKey_(key),
                                      C_TO_XEN_GClosure_(closure),
                                      XEN_CADR((XEN)func_data),
                                      __FUNCTION__)));
}
#endif

static void gxg_input_func(gpointer func_data, gint fd, GdkInputCondition condition)
{
  XEN_CALL_3(XEN_CAR((XEN)func_data),
             XEN_CADR((XEN)func_data),
             C_TO_XEN_gint(fd),
             C_TO_XEN_GdkInputCondition(condition),
             __FUNCTION__);
}

static void gxg_func2(GtkWidget* w, gpointer func_data)
{
  XEN_CALL_2(XEN_CAR((XEN)func_data),
             C_TO_XEN_GtkWidget_(w),
             XEN_CADR((XEN)func_data),
             __FUNCTION__);
}

#if (!HAVE_GTK_1)
static gint gxg_timeout_func(gpointer func_data)
{
  int call_again = 0;
  XEN result = XEN_FALSE;
  result = XEN_CALL_1(XEN_CAR((XEN)func_data),
                      XEN_CADR((XEN)func_data),
                      __FUNCTION__);
  if (XEN_BOOLEAN_P(result)) call_again = XEN_TO_C_BOOLEAN(result); else call_again = XEN_TO_C_INT_OR_ELSE(result, 0);
  if (!call_again) xm_unprotect_at(XEN_TO_C_INT(XEN_CADDR((XEN)func_data)));
  return(call_again);
}
#endif

static void gxg_destroy_func(gpointer func_data)
{
  XEN_CALL_1(XEN_CADDDR((XEN)func_data),
             XEN_CADR((XEN)func_data),
             __FUNCTION__);
}

static GdkFilterReturn gxg_filter_func(GdkXEvent* xevent, GdkEvent* event, gpointer func_data)
{
  return(XEN_TO_C_GdkFilterReturn(XEN_CALL_3(XEN_CAR((XEN)func_data),
                                             C_TO_XEN_GdkXEvent_(xevent),
                                             C_TO_XEN_GdkEvent_(event),
                                             XEN_CADR((XEN)func_data),
                                             __FUNCTION__)));
}

static void gxg_event_func(GdkEvent* event, gpointer func_data)
{
  XEN_CALL_2(XEN_CAR((XEN)func_data),
             C_TO_XEN_GdkEvent_(event),
             XEN_CADR((XEN)func_data),
             __FUNCTION__);
}

#if (!HAVE_GTK_1)
static void gxg_span_func(GdkSpan* span, gpointer func_data)
{
  XEN_CALL_2(XEN_CAR((XEN)func_data),
             C_TO_XEN_GdkSpan_(span),
             XEN_CADR((XEN)func_data),
             __FUNCTION__);
}
#endif

static gboolean gxg_func1(gpointer func_data)
{
  return(XEN_TO_C_gboolean(XEN_CALL_1(XEN_CAR((XEN)func_data),
                                      XEN_CADR((XEN)func_data),
                                      __FUNCTION__)));
}

static gint gxg_snoop_func(GtkWidget* widget, GdkEventKey* event, gpointer func_data)
{
  return(XEN_TO_C_gint(XEN_CALL_3(XEN_CAR((XEN)func_data),
                                  C_TO_XEN_GtkWidget_(widget),
                                  C_TO_XEN_GdkEventKey_(event),
                                  XEN_CADR((XEN)func_data),
                                  __FUNCTION__)));
}

static gchar* gxg_translate_func(const gchar* path, gpointer func_data)
{
  return(XEN_TO_C_gchar_(XEN_CALL_2(XEN_CAR((XEN)func_data),
                                    C_TO_XEN_gchar_(path),
                                    XEN_CADR((XEN)func_data),
                                    __FUNCTION__)));
}

static void gxg_menu_position_func(GtkMenu* menu, gint* x, gint* y, gboolean* push, gpointer func_data)
{
  XEN_CALL_5(XEN_CAR((XEN)func_data),
             C_TO_XEN_GtkMenu_(menu),
             C_TO_XEN_gint_(x),
             C_TO_XEN_gint_(y),
             C_TO_XEN_gboolean_(push),
             XEN_CADR((XEN)func_data),
             __FUNCTION__);
}

#if (!HAVE_GTK_1)
static gboolean gxg_model_func(GtkTreeModel* model, GtkTreePath* path, GtkTreeIter* iter, gpointer func_data)
{
  return(XEN_TO_C_gboolean(XEN_CALL_4(XEN_CAR((XEN)func_data),
                                      C_TO_XEN_GtkTreeModel_(model),
                                      C_TO_XEN_GtkTreePath_(path),
                                      C_TO_XEN_GtkTreeIter_(iter),
                                      XEN_CADR((XEN)func_data),
                                      __FUNCTION__)));
}

static void gxg_tree_selection_func(GtkTreeModel* model, GtkTreePath* path, GtkTreeIter* iter, gpointer func_data)
{
  XEN_CALL_4(XEN_CAR((XEN)func_data),
             C_TO_XEN_GtkTreeModel_(model),
             C_TO_XEN_GtkTreePath_(path),
             C_TO_XEN_GtkTreeIter_(iter),
             XEN_CADR((XEN)func_data),
             __FUNCTION__);
}

static void gxg_clip_received(GtkClipboard* clipboard, GtkSelectionData* selection_data, gpointer func_data)
{
  XEN_CALL_3(XEN_CAR((XEN)func_data),
             C_TO_XEN_GtkClipboard_(clipboard),
             C_TO_XEN_GtkSelectionData_(selection_data),
             XEN_CADR((XEN)func_data),
             __FUNCTION__);
}

static void gxg_clip_text_received(GtkClipboard* clipboard, const gchar* text, gpointer func_data)
{
  XEN_CALL_3(XEN_CAR((XEN)func_data),
             C_TO_XEN_GtkClipboard_(clipboard),
             C_TO_XEN_gchar_(text),
             XEN_CADR((XEN)func_data),
             __FUNCTION__);
}

static gboolean gxg_tree_column(GtkTreeView* tree_view, GtkTreeViewColumn* column, GtkTreeViewColumn* prev_column, GtkTreeViewColumn* next_column, gpointer func_data)
{
  return(XEN_TO_C_gboolean(XEN_CALL_5(XEN_CAR((XEN)func_data),
                                      C_TO_XEN_GtkTreeView_(tree_view),
                                      C_TO_XEN_GtkTreeViewColumn_(column),
                                      C_TO_XEN_GtkTreeViewColumn_(prev_column),
                                      C_TO_XEN_GtkTreeViewColumn_(next_column),
                                      XEN_CADR((XEN)func_data),
                                      __FUNCTION__)));
}

static void gxg_tree_mapping(GtkTreeView* tree_view, GtkTreePath* path, gpointer func_data)
{
  XEN_CALL_3(XEN_CAR((XEN)func_data),
             C_TO_XEN_GtkTreeView_(tree_view),
             C_TO_XEN_GtkTreePath_(path),
             XEN_CADR((XEN)func_data),
             __FUNCTION__);
}

static gboolean gxg_tree_search(GtkTreeModel* model, gint column, const gchar* key, GtkTreeIter* iter, gpointer func_data)
{
  return(XEN_TO_C_gboolean(XEN_CALL_5(XEN_CAR((XEN)func_data),
                                      C_TO_XEN_GtkTreeModel_(model),
                                      C_TO_XEN_gint(column),
                                      C_TO_XEN_gchar_(key),
                                      C_TO_XEN_GtkTreeIter_(iter),
                                      XEN_CADR((XEN)func_data),
                                      __FUNCTION__)));
}

static void gxg_cell_data(GtkTreeViewColumn* tree_column, GtkCellRenderer* cell, GtkTreeModel* tree_model, GtkTreeIter* iter, gpointer func_data)
{
  XEN_CALL_5(XEN_CAR((XEN)func_data),
             C_TO_XEN_GtkTreeViewColumn_(tree_column),
             C_TO_XEN_GtkCellRenderer_(cell),
             C_TO_XEN_GtkTreeModel_(tree_model),
             C_TO_XEN_GtkTreeIter_(iter),
             XEN_CADR((XEN)func_data),
             __FUNCTION__);
}

static gint gxg_iter_compare(GtkTreeModel* model, GtkTreeIter* a, GtkTreeIter* b, gpointer func_data)
{
  return(XEN_TO_C_gint(XEN_CALL_4(XEN_CAR((XEN)func_data),
                                  C_TO_XEN_GtkTreeModel_(model),
                                  C_TO_XEN_GtkTreeIter_(a),
                                  C_TO_XEN_GtkTreeIter_(b),
                                  XEN_CADR((XEN)func_data),
                                  __FUNCTION__)));
}

static gboolean gxg_tree_selection(GtkTreeSelection* selection, GtkTreeModel* model, GtkTreePath* path, gboolean path_currently_selected, gpointer func_data)
{
  return(XEN_TO_C_gboolean(XEN_CALL_5(XEN_CAR((XEN)func_data),
                                      C_TO_XEN_GtkTreeSelection_(selection),
                                      C_TO_XEN_GtkTreeModel_(model),
                                      C_TO_XEN_GtkTreePath_(path),
                                      C_TO_XEN_gboolean(path_currently_selected),
                                      XEN_CADR((XEN)func_data),
                                      __FUNCTION__)));
}

static void gxg_clip_get(GtkClipboard* clipboard, GtkSelectionData* selection_data, guint info, gpointer func_data)
{
  XEN_CALL_4(XEN_CAR((XEN)func_data),
             C_TO_XEN_GtkClipboard_(clipboard),
             C_TO_XEN_GtkSelectionData_(selection_data),
             C_TO_XEN_guint(info),
             XEN_CADR((XEN)func_data),
             __FUNCTION__);
}

static void gxg_clip_clear(GtkClipboard* clipboard, gpointer func_data)
{
  XEN_CALL_2(XEN_CADDR((XEN)func_data),
             C_TO_XEN_GtkClipboard_(clipboard),
             XEN_CADR((XEN)func_data),
             __FUNCTION__);
}
#endif

static void gxg_func3(GtkWidget *w, GdkEventAny *ev, gpointer data)
{
  XEN_CALL_3(XEN_CAR((XEN)data),
             C_TO_XEN_GtkWidget_(w),
             C_TO_XEN_GdkEventAny_(ev),
             XEN_CADR((XEN)data),
             __FUNCTION__);
}



/* ---------------------------------------- functions ---------------------------------------- */


#if (!HAVE_GTK_1)
static XEN gxg_g_type_name(XEN type)
{
  #define H_g_type_name "gchar* g_type_name(GType type)"
  XEN_ASSERT_TYPE(XEN_GType_P(type), type, 1, "g_type_name", "GType");
  return(C_TO_XEN_gchar_(g_type_name(XEN_TO_C_GType(type))));
}

static XEN gxg_g_type_qname(XEN type)
{
  #define H_g_type_qname "GQuark g_type_qname(GType type)"
  XEN_ASSERT_TYPE(XEN_GType_P(type), type, 1, "g_type_qname", "GType");
  return(C_TO_XEN_GQuark(g_type_qname(XEN_TO_C_GType(type))));
}

static XEN gxg_g_type_from_name(XEN name)
{
  #define H_g_type_from_name "GType g_type_from_name(gchar* name)"
  XEN_ASSERT_TYPE(XEN_gchar__P(name), name, 1, "g_type_from_name", "gchar*");
  return(C_TO_XEN_GType(g_type_from_name(XEN_TO_C_gchar_(name))));
}

static XEN gxg_g_type_parent(XEN type)
{
  #define H_g_type_parent "GType g_type_parent(GType type)"
  XEN_ASSERT_TYPE(XEN_GType_P(type), type, 1, "g_type_parent", "GType");
  return(C_TO_XEN_GType(g_type_parent(XEN_TO_C_GType(type))));
}

static XEN gxg_g_type_is_a(XEN type, XEN is_a_type)
{
  #define H_g_type_is_a "gboolean g_type_is_a(GType type, GType is_a_type)"
  XEN_ASSERT_TYPE(XEN_GType_P(type), type, 1, "g_type_is_a", "GType");
  XEN_ASSERT_TYPE(XEN_GType_P(is_a_type), is_a_type, 2, "g_type_is_a", "GType");
  return(C_TO_XEN_gboolean(g_type_is_a(XEN_TO_C_GType(type), XEN_TO_C_GType(is_a_type))));
}

static XEN gxg_g_signal_newv(XEN arglist)
{
  #define H_g_signal_newv "guint g_signal_newv(gchar* signal_name, GType itype, GSignalFlags signal_flags, \
GClosure* class_closure, GSignalAccumulator accumulator, gpointer accu_data, GSignalCMarshaller c_marshaller, \
GType return_type, guint n_params, GType* param_types)"
  XEN signal_name, itype, signal_flags, class_closure, accumulator, accu_data, c_marshaller, return_type, n_params, param_types;
  signal_name = XEN_LIST_REF(arglist, 0);
  itype = XEN_LIST_REF(arglist, 1);
  signal_flags = XEN_LIST_REF(arglist, 2);
  class_closure = XEN_LIST_REF(arglist, 3);
  accumulator = XEN_LIST_REF(arglist, 4);
  accu_data = XEN_LIST_REF(arglist, 5);
  c_marshaller = XEN_LIST_REF(arglist, 6);
  return_type = XEN_LIST_REF(arglist, 7);
  n_params = XEN_LIST_REF(arglist, 8);
  param_types = XEN_LIST_REF(arglist, 9);
  XEN_ASSERT_TYPE(XEN_gchar__P(signal_name), signal_name, 1, "g_signal_newv", "gchar*");
  XEN_ASSERT_TYPE(XEN_GType_P(itype), itype, 2, "g_signal_newv", "GType");
  XEN_ASSERT_TYPE(XEN_GSignalFlags_P(signal_flags), signal_flags, 3, "g_signal_newv", "GSignalFlags");
  XEN_ASSERT_TYPE(XEN_GClosure__P(class_closure), class_closure, 4, "g_signal_newv", "GClosure*");
  XEN_ASSERT_TYPE(XEN_GSignalAccumulator_P(accumulator), accumulator, 5, "g_signal_newv", "GSignalAccumulator");
  XEN_ASSERT_TYPE(XEN_gpointer_P(accu_data), accu_data, 6, "g_signal_newv", "gpointer");
  XEN_ASSERT_TYPE(XEN_GSignalCMarshaller_P(c_marshaller), c_marshaller, 7, "g_signal_newv", "GSignalCMarshaller");
  XEN_ASSERT_TYPE(XEN_GType_P(return_type), return_type, 8, "g_signal_newv", "GType");
  XEN_ASSERT_TYPE(XEN_guint_P(n_params), n_params, 9, "g_signal_newv", "guint");
  XEN_ASSERT_TYPE(XEN_GType__P(param_types), param_types, 10, "g_signal_newv", "GType*");
  return(C_TO_XEN_guint(g_signal_newv(XEN_TO_C_gchar_(signal_name), XEN_TO_C_GType(itype), XEN_TO_C_GSignalFlags(signal_flags), 
                                      XEN_TO_C_GClosure_(class_closure), XEN_TO_C_GSignalAccumulator(accumulator), XEN_TO_C_gpointer(accu_data), 
                                      XEN_TO_C_GSignalCMarshaller(c_marshaller), XEN_TO_C_GType(return_type), XEN_TO_C_guint(n_params), 
                                      XEN_TO_C_GType_(param_types))));
}

static XEN gxg_g_signal_new(XEN signal_name, XEN itype, XEN signal_flags, XEN class_offset, XEN accumulator, XEN accu_data, XEN c_marshaller, XEN return_type, XEN n_params)
{
  #define H_g_signal_new "guint g_signal_new(gchar* signal_name, GType itype, GSignalFlags signal_flags, \
guint class_offset, GSignalAccumulator accumulator, gpointer accu_data, GSignalCMarshaller c_marshaller, \
GType return_type, guint n_params, ...)"
  XEN_ASSERT_TYPE(XEN_gchar__P(signal_name), signal_name, 1, "g_signal_new", "gchar*");
  XEN_ASSERT_TYPE(XEN_GType_P(itype), itype, 2, "g_signal_new", "GType");
  XEN_ASSERT_TYPE(XEN_GSignalFlags_P(signal_flags), signal_flags, 3, "g_signal_new", "GSignalFlags");
  XEN_ASSERT_TYPE(XEN_guint_P(class_offset), class_offset, 4, "g_signal_new", "guint");
  XEN_ASSERT_TYPE(XEN_GSignalAccumulator_P(accumulator), accumulator, 5, "g_signal_new", "GSignalAccumulator");
  XEN_ASSERT_TYPE(XEN_gpointer_P(accu_data), accu_data, 6, "g_signal_new", "gpointer");
  XEN_ASSERT_TYPE(XEN_GSignalCMarshaller_P(c_marshaller), c_marshaller, 7, "g_signal_new", "GSignalCMarshaller");
  XEN_ASSERT_TYPE(XEN_GType_P(return_type), return_type, 8, "g_signal_new", "GType");
  XEN_ASSERT_TYPE(XEN_guint_P(n_params), n_params, 9, "g_signal_new", "guint");
  return(C_TO_XEN_guint(g_signal_new(XEN_TO_C_gchar_(signal_name), XEN_TO_C_GType(itype), XEN_TO_C_GSignalFlags(signal_flags), 
                                     XEN_TO_C_guint(class_offset), XEN_TO_C_GSignalAccumulator(accumulator), XEN_TO_C_gpointer(accu_data), 
                                     XEN_TO_C_GSignalCMarshaller(c_marshaller), XEN_TO_C_GType(return_type), XEN_TO_C_guint(n_params))));
}

static XEN gxg_g_signal_emitv(XEN instance_and_params, XEN signal_id, XEN detail, XEN return_value)
{
  #define H_g_signal_emitv "void g_signal_emitv(GValue* instance_and_params, guint signal_id, GQuark detail, \
GValue* return_value)"
  XEN_ASSERT_TYPE(XEN_GValue__P(instance_and_params), instance_and_params, 1, "g_signal_emitv", "GValue*");
  XEN_ASSERT_TYPE(XEN_guint_P(signal_id), signal_id, 2, "g_signal_emitv", "guint");
  XEN_ASSERT_TYPE(XEN_GQuark_P(detail), detail, 3, "g_signal_emitv", "GQuark");
  XEN_ASSERT_TYPE(XEN_GValue__P(return_value), return_value, 4, "g_signal_emitv", "GValue*");
  g_signal_emitv(XEN_TO_C_GValue_(instance_and_params), XEN_TO_C_guint(signal_id), XEN_TO_C_GQuark(detail), XEN_TO_C_GValue_(return_value));
  return(XEN_FALSE);
}

static XEN gxg_g_signal_emit(XEN instance, XEN signal_id, XEN detail)
{
  #define H_g_signal_emit "void g_signal_emit(gpointer instance, guint signal_id, GQuark detail, ...)"
  XEN_ASSERT_TYPE(XEN_gpointer_P(instance), instance, 1, "g_signal_emit", "gpointer");
  XEN_ASSERT_TYPE(XEN_guint_P(signal_id), signal_id, 2, "g_signal_emit", "guint");
  XEN_ASSERT_TYPE(XEN_GQuark_P(detail), detail, 3, "g_signal_emit", "GQuark");
  g_signal_emit(XEN_TO_C_gpointer(instance), XEN_TO_C_guint(signal_id), XEN_TO_C_GQuark(detail));
  return(XEN_FALSE);
}

static XEN gxg_g_signal_emit_by_name(XEN instance, XEN detailed_signal)
{
  #define H_g_signal_emit_by_name "void g_signal_emit_by_name(gpointer instance, gchar* detailed_signal, \
...)"
  XEN_ASSERT_TYPE(XEN_gpointer_P(instance), instance, 1, "g_signal_emit_by_name", "gpointer");
  XEN_ASSERT_TYPE(XEN_gchar__P(detailed_signal), detailed_signal, 2, "g_signal_emit_by_name", "gchar*");
  g_signal_emit_by_name(XEN_TO_C_gpointer(instance), XEN_TO_C_gchar_(detailed_signal));
  return(XEN_FALSE);
}

static XEN gxg_g_signal_lookup(XEN name, XEN itype)
{
  #define H_g_signal_lookup "guint g_signal_lookup(gchar* name, GType itype)"
  XEN_ASSERT_TYPE(XEN_gchar__P(name), name, 1, "g_signal_lookup", "gchar*");
  XEN_ASSERT_TYPE(XEN_GType_P(itype), itype, 2, "g_signal_lookup", "GType");
  return(C_TO_XEN_guint(g_signal_lookup(XEN_TO_C_gchar_(name), XEN_TO_C_GType(itype))));
}

static XEN gxg_g_signal_name(XEN signal_id)
{
  #define H_g_signal_name "gchar* g_signal_name(guint signal_id)"
  XEN_ASSERT_TYPE(XEN_guint_P(signal_id), signal_id, 1, "g_signal_name", "guint");
  return(C_TO_XEN_gchar_(g_signal_name(XEN_TO_C_guint(signal_id))));
}

static XEN gxg_g_signal_query(XEN signal_id, XEN query)
{
  #define H_g_signal_query "void g_signal_query(guint signal_id, GSignalQuery* query)"
  XEN_ASSERT_TYPE(XEN_guint_P(signal_id), signal_id, 1, "g_signal_query", "guint");
  XEN_ASSERT_TYPE(XEN_GSignalQuery__P(query), query, 2, "g_signal_query", "GSignalQuery*");
  g_signal_query(XEN_TO_C_guint(signal_id), XEN_TO_C_GSignalQuery_(query));
  return(XEN_FALSE);
}

static XEN gxg_g_signal_list_ids(XEN itype, XEN n_ids)
{
  #define H_g_signal_list_ids "guint* g_signal_list_ids(GType itype, guint* n_ids)"
  XEN_ASSERT_TYPE(XEN_GType_P(itype), itype, 1, "g_signal_list_ids", "GType");
  XEN_ASSERT_TYPE(XEN_guint__P(n_ids), n_ids, 2, "g_signal_list_ids", "guint*");
  return(C_TO_XEN_guint_(g_signal_list_ids(XEN_TO_C_GType(itype), XEN_TO_C_guint_(n_ids))));
}

static XEN gxg_g_signal_parse_name(XEN detailed_signal, XEN itype, XEN signal_id_p, XEN detail_p, XEN force_detail_quark)
{
  #define H_g_signal_parse_name "gboolean g_signal_parse_name(gchar* detailed_signal, GType itype, guint* signal_id_p, \
GQuark* detail_p, gboolean force_detail_quark)"
  XEN_ASSERT_TYPE(XEN_gchar__P(detailed_signal), detailed_signal, 1, "g_signal_parse_name", "gchar*");
  XEN_ASSERT_TYPE(XEN_GType_P(itype), itype, 2, "g_signal_parse_name", "GType");
  XEN_ASSERT_TYPE(XEN_guint__P(signal_id_p), signal_id_p, 3, "g_signal_parse_name", "guint*");
  XEN_ASSERT_TYPE(XEN_GQuark__P(detail_p), detail_p, 4, "g_signal_parse_name", "GQuark*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(force_detail_quark), force_detail_quark, 5, "g_signal_parse_name", "gboolean");
  return(C_TO_XEN_gboolean(g_signal_parse_name(XEN_TO_C_gchar_(detailed_signal), XEN_TO_C_GType(itype), XEN_TO_C_guint_(signal_id_p), 
                                               XEN_TO_C_GQuark_(detail_p), XEN_TO_C_gboolean(force_detail_quark))));
}

static XEN gxg_g_signal_get_invocation_hint(XEN instance)
{
  #define H_g_signal_get_invocation_hint "GSignalInvocationHint* g_signal_get_invocation_hint(gpointer instance)"
  XEN_ASSERT_TYPE(XEN_gpointer_P(instance), instance, 1, "g_signal_get_invocation_hint", "gpointer");
  return(C_TO_XEN_GSignalInvocationHint_(g_signal_get_invocation_hint(XEN_TO_C_gpointer(instance))));
}

static XEN gxg_g_signal_stop_emission(XEN instance, XEN signal_id, XEN detail)
{
  #define H_g_signal_stop_emission "void g_signal_stop_emission(gpointer instance, guint signal_id, GQuark detail)"
  XEN_ASSERT_TYPE(XEN_gpointer_P(instance), instance, 1, "g_signal_stop_emission", "gpointer");
  XEN_ASSERT_TYPE(XEN_guint_P(signal_id), signal_id, 2, "g_signal_stop_emission", "guint");
  XEN_ASSERT_TYPE(XEN_GQuark_P(detail), detail, 3, "g_signal_stop_emission", "GQuark");
  g_signal_stop_emission(XEN_TO_C_gpointer(instance), XEN_TO_C_guint(signal_id), XEN_TO_C_GQuark(detail));
  return(XEN_FALSE);
}

static XEN gxg_g_signal_stop_emission_by_name(XEN instance, XEN detailed_signal)
{
  #define H_g_signal_stop_emission_by_name "void g_signal_stop_emission_by_name(gpointer instance, gchar* detailed_signal)"
  XEN_ASSERT_TYPE(XEN_gpointer_P(instance), instance, 1, "g_signal_stop_emission_by_name", "gpointer");
  XEN_ASSERT_TYPE(XEN_gchar__P(detailed_signal), detailed_signal, 2, "g_signal_stop_emission_by_name", "gchar*");
  g_signal_stop_emission_by_name(XEN_TO_C_gpointer(instance), XEN_TO_C_gchar_(detailed_signal));
  return(XEN_FALSE);
}

static XEN gxg_g_signal_add_emission_hook(XEN signal_id, XEN quark, XEN hook_func, XEN hook_data, XEN data_destroy)
{
  #define H_g_signal_add_emission_hook "gulong g_signal_add_emission_hook(guint signal_id, GQuark quark, \
GSignalEmissionHook hook_func, gpointer hook_data, GDestroyNotify data_destroy)"
  XEN_ASSERT_TYPE(XEN_guint_P(signal_id), signal_id, 1, "g_signal_add_emission_hook", "guint");
  XEN_ASSERT_TYPE(XEN_GQuark_P(quark), quark, 2, "g_signal_add_emission_hook", "GQuark");
  XEN_ASSERT_TYPE(XEN_GSignalEmissionHook_P(hook_func), hook_func, 3, "g_signal_add_emission_hook", "GSignalEmissionHook");
  XEN_ASSERT_TYPE(XEN_gpointer_P(hook_data), hook_data, 4, "g_signal_add_emission_hook", "gpointer");
  XEN_ASSERT_TYPE(XEN_GDestroyNotify_P(data_destroy), data_destroy, 5, "g_signal_add_emission_hook", "GDestroyNotify");
  return(C_TO_XEN_gulong(g_signal_add_emission_hook(XEN_TO_C_guint(signal_id), XEN_TO_C_GQuark(quark), XEN_TO_C_GSignalEmissionHook(hook_func), 
                                                    XEN_TO_C_gpointer(hook_data), XEN_TO_C_GDestroyNotify(data_destroy))));
}

static XEN gxg_g_signal_remove_emission_hook(XEN signal_id, XEN hook_id)
{
  #define H_g_signal_remove_emission_hook "void g_signal_remove_emission_hook(guint signal_id, gulong hook_id)"
  XEN_ASSERT_TYPE(XEN_guint_P(signal_id), signal_id, 1, "g_signal_remove_emission_hook", "guint");
  XEN_ASSERT_TYPE(XEN_gulong_P(hook_id), hook_id, 2, "g_signal_remove_emission_hook", "gulong");
  g_signal_remove_emission_hook(XEN_TO_C_guint(signal_id), XEN_TO_C_gulong(hook_id));
  return(XEN_FALSE);
}

static XEN gxg_g_signal_has_handler_pending(XEN instance, XEN signal_id, XEN detail, XEN may_be_blocked)
{
  #define H_g_signal_has_handler_pending "gboolean g_signal_has_handler_pending(gpointer instance, guint signal_id, \
GQuark detail, gboolean may_be_blocked)"
  XEN_ASSERT_TYPE(XEN_gpointer_P(instance), instance, 1, "g_signal_has_handler_pending", "gpointer");
  XEN_ASSERT_TYPE(XEN_guint_P(signal_id), signal_id, 2, "g_signal_has_handler_pending", "guint");
  XEN_ASSERT_TYPE(XEN_GQuark_P(detail), detail, 3, "g_signal_has_handler_pending", "GQuark");
  XEN_ASSERT_TYPE(XEN_gboolean_P(may_be_blocked), may_be_blocked, 4, "g_signal_has_handler_pending", "gboolean");
  return(C_TO_XEN_gboolean(g_signal_has_handler_pending(XEN_TO_C_gpointer(instance), XEN_TO_C_guint(signal_id), XEN_TO_C_GQuark(detail), 
                                                        XEN_TO_C_gboolean(may_be_blocked))));
}

static XEN gxg_g_signal_connect_closure_by_id(XEN instance, XEN signal_id, XEN detail, XEN closure, XEN after)
{
  #define H_g_signal_connect_closure_by_id "gulong g_signal_connect_closure_by_id(gpointer instance, \
guint signal_id, GQuark detail, GClosure* closure, gboolean after)"
  XEN_ASSERT_TYPE(XEN_gpointer_P(instance), instance, 1, "g_signal_connect_closure_by_id", "gpointer");
  XEN_ASSERT_TYPE(XEN_guint_P(signal_id), signal_id, 2, "g_signal_connect_closure_by_id", "guint");
  XEN_ASSERT_TYPE(XEN_GQuark_P(detail), detail, 3, "g_signal_connect_closure_by_id", "GQuark");
  XEN_ASSERT_TYPE(XEN_GClosure__P(closure), closure, 4, "g_signal_connect_closure_by_id", "GClosure*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(after), after, 5, "g_signal_connect_closure_by_id", "gboolean");
  return(C_TO_XEN_gulong(g_signal_connect_closure_by_id(XEN_TO_C_gpointer(instance), XEN_TO_C_guint(signal_id), XEN_TO_C_GQuark(detail), 
                                                        XEN_TO_C_GClosure_(closure), XEN_TO_C_gboolean(after))));
}

static XEN gxg_g_signal_connect_closure(XEN instance, XEN detailed_signal, XEN closure, XEN after)
{
  #define H_g_signal_connect_closure "gulong g_signal_connect_closure(gpointer instance, gchar* detailed_signal, \
GClosure* closure, gboolean after)"
  XEN_ASSERT_TYPE(XEN_gpointer_P(instance), instance, 1, "g_signal_connect_closure", "gpointer");
  XEN_ASSERT_TYPE(XEN_gchar__P(detailed_signal), detailed_signal, 2, "g_signal_connect_closure", "gchar*");
  XEN_ASSERT_TYPE(XEN_GClosure__P(closure), closure, 3, "g_signal_connect_closure", "GClosure*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(after), after, 4, "g_signal_connect_closure", "gboolean");
  return(C_TO_XEN_gulong(g_signal_connect_closure(XEN_TO_C_gpointer(instance), XEN_TO_C_gchar_(detailed_signal), XEN_TO_C_GClosure_(closure), 
                                                  XEN_TO_C_gboolean(after))));
}

static XEN gxg_g_signal_connect_data(XEN instance, XEN detailed_signal, XEN c_handler, XEN data, XEN destroy_data, XEN connect_flags)
{
  #define H_g_signal_connect_data "gulong g_signal_connect_data(gpointer instance, gchar* detailed_signal, \
GCallback c_handler, gpointer data, GClosureNotify destroy_data, GConnectFlags connect_flags)"
  XEN_ASSERT_TYPE(XEN_gpointer_P(instance), instance, 1, "g_signal_connect_data", "gpointer");
  XEN_ASSERT_TYPE(XEN_gchar__P(detailed_signal), detailed_signal, 2, "g_signal_connect_data", "gchar*");
  XEN_ASSERT_TYPE(XEN_GCallback_P(c_handler), c_handler, 3, "g_signal_connect_data", "GCallback");
  XEN_ASSERT_TYPE(XEN_gpointer_P(data), data, 4, "g_signal_connect_data", "gpointer");
  XEN_ASSERT_TYPE(XEN_GClosureNotify_P(destroy_data), destroy_data, 5, "g_signal_connect_data", "GClosureNotify");
  XEN_ASSERT_TYPE(XEN_GConnectFlags_P(connect_flags), connect_flags, 6, "g_signal_connect_data", "GConnectFlags");
  return(C_TO_XEN_gulong(g_signal_connect_data(XEN_TO_C_gpointer(instance), XEN_TO_C_gchar_(detailed_signal), XEN_TO_C_GCallback(c_handler), 
                                               XEN_TO_C_gpointer(data), XEN_TO_C_GClosureNotify(destroy_data), XEN_TO_C_GConnectFlags(connect_flags))));
}

static XEN gxg_g_signal_handler_block(XEN instance, XEN handler_id)
{
  #define H_g_signal_handler_block "void g_signal_handler_block(gpointer instance, gulong handler_id)"
  XEN_ASSERT_TYPE(XEN_gpointer_P(instance), instance, 1, "g_signal_handler_block", "gpointer");
  XEN_ASSERT_TYPE(XEN_gulong_P(handler_id), handler_id, 2, "g_signal_handler_block", "gulong");
  g_signal_handler_block(XEN_TO_C_gpointer(instance), XEN_TO_C_gulong(handler_id));
  return(XEN_FALSE);
}

static XEN gxg_g_signal_handler_unblock(XEN instance, XEN handler_id)
{
  #define H_g_signal_handler_unblock "void g_signal_handler_unblock(gpointer instance, gulong handler_id)"
  XEN_ASSERT_TYPE(XEN_gpointer_P(instance), instance, 1, "g_signal_handler_unblock", "gpointer");
  XEN_ASSERT_TYPE(XEN_gulong_P(handler_id), handler_id, 2, "g_signal_handler_unblock", "gulong");
  g_signal_handler_unblock(XEN_TO_C_gpointer(instance), XEN_TO_C_gulong(handler_id));
  return(XEN_FALSE);
}

static XEN gxg_g_signal_handler_disconnect(XEN instance, XEN handler_id)
{
  #define H_g_signal_handler_disconnect "void g_signal_handler_disconnect(gpointer instance, gulong handler_id)"
  XEN_ASSERT_TYPE(XEN_gpointer_P(instance), instance, 1, "g_signal_handler_disconnect", "gpointer");
  XEN_ASSERT_TYPE(XEN_gulong_P(handler_id), handler_id, 2, "g_signal_handler_disconnect", "gulong");
  g_signal_handler_disconnect(XEN_TO_C_gpointer(instance), XEN_TO_C_gulong(handler_id));
  return(XEN_FALSE);
}

static XEN gxg_g_signal_handler_is_connected(XEN instance, XEN handler_id)
{
  #define H_g_signal_handler_is_connected "gboolean g_signal_handler_is_connected(gpointer instance, \
gulong handler_id)"
  XEN_ASSERT_TYPE(XEN_gpointer_P(instance), instance, 1, "g_signal_handler_is_connected", "gpointer");
  XEN_ASSERT_TYPE(XEN_gulong_P(handler_id), handler_id, 2, "g_signal_handler_is_connected", "gulong");
  return(C_TO_XEN_gboolean(g_signal_handler_is_connected(XEN_TO_C_gpointer(instance), XEN_TO_C_gulong(handler_id))));
}

static XEN gxg_g_signal_handler_find(XEN instance, XEN mask, XEN signal_id, XEN detail, XEN closure, XEN func, XEN data)
{
  #define H_g_signal_handler_find "gulong g_signal_handler_find(gpointer instance, GSignalMatchType mask, \
guint signal_id, GQuark detail, GClosure* closure, gpointer func, gpointer data)"
  XEN_ASSERT_TYPE(XEN_gpointer_P(instance), instance, 1, "g_signal_handler_find", "gpointer");
  XEN_ASSERT_TYPE(XEN_GSignalMatchType_P(mask), mask, 2, "g_signal_handler_find", "GSignalMatchType");
  XEN_ASSERT_TYPE(XEN_guint_P(signal_id), signal_id, 3, "g_signal_handler_find", "guint");
  XEN_ASSERT_TYPE(XEN_GQuark_P(detail), detail, 4, "g_signal_handler_find", "GQuark");
  XEN_ASSERT_TYPE(XEN_GClosure__P(closure), closure, 5, "g_signal_handler_find", "GClosure*");
  XEN_ASSERT_TYPE(XEN_gpointer_P(func), func, 6, "g_signal_handler_find", "gpointer");
  XEN_ASSERT_TYPE(XEN_gpointer_P(data), data, 7, "g_signal_handler_find", "gpointer");
  return(C_TO_XEN_gulong(g_signal_handler_find(XEN_TO_C_gpointer(instance), XEN_TO_C_GSignalMatchType(mask), XEN_TO_C_guint(signal_id), 
                                               XEN_TO_C_GQuark(detail), XEN_TO_C_GClosure_(closure), XEN_TO_C_gpointer(func), 
                                               XEN_TO_C_gpointer(data))));
}

static XEN gxg_g_signal_handlers_block_matched(XEN instance, XEN mask, XEN signal_id, XEN detail, XEN closure, XEN func, XEN data)
{
  #define H_g_signal_handlers_block_matched "guint g_signal_handlers_block_matched(gpointer instance, \
GSignalMatchType mask, guint signal_id, GQuark detail, GClosure* closure, gpointer func, gpointer data)"
  XEN_ASSERT_TYPE(XEN_gpointer_P(instance), instance, 1, "g_signal_handlers_block_matched", "gpointer");
  XEN_ASSERT_TYPE(XEN_GSignalMatchType_P(mask), mask, 2, "g_signal_handlers_block_matched", "GSignalMatchType");
  XEN_ASSERT_TYPE(XEN_guint_P(signal_id), signal_id, 3, "g_signal_handlers_block_matched", "guint");
  XEN_ASSERT_TYPE(XEN_GQuark_P(detail), detail, 4, "g_signal_handlers_block_matched", "GQuark");
  XEN_ASSERT_TYPE(XEN_GClosure__P(closure), closure, 5, "g_signal_handlers_block_matched", "GClosure*");
  XEN_ASSERT_TYPE(XEN_gpointer_P(func), func, 6, "g_signal_handlers_block_matched", "gpointer");
  XEN_ASSERT_TYPE(XEN_gpointer_P(data), data, 7, "g_signal_handlers_block_matched", "gpointer");
  return(C_TO_XEN_guint(g_signal_handlers_block_matched(XEN_TO_C_gpointer(instance), XEN_TO_C_GSignalMatchType(mask), XEN_TO_C_guint(signal_id), 
                                                        XEN_TO_C_GQuark(detail), XEN_TO_C_GClosure_(closure), XEN_TO_C_gpointer(func), 
                                                        XEN_TO_C_gpointer(data))));
}

static XEN gxg_g_signal_handlers_unblock_matched(XEN instance, XEN mask, XEN signal_id, XEN detail, XEN closure, XEN func, XEN data)
{
  #define H_g_signal_handlers_unblock_matched "guint g_signal_handlers_unblock_matched(gpointer instance, \
GSignalMatchType mask, guint signal_id, GQuark detail, GClosure* closure, gpointer func, gpointer data)"
  XEN_ASSERT_TYPE(XEN_gpointer_P(instance), instance, 1, "g_signal_handlers_unblock_matched", "gpointer");
  XEN_ASSERT_TYPE(XEN_GSignalMatchType_P(mask), mask, 2, "g_signal_handlers_unblock_matched", "GSignalMatchType");
  XEN_ASSERT_TYPE(XEN_guint_P(signal_id), signal_id, 3, "g_signal_handlers_unblock_matched", "guint");
  XEN_ASSERT_TYPE(XEN_GQuark_P(detail), detail, 4, "g_signal_handlers_unblock_matched", "GQuark");
  XEN_ASSERT_TYPE(XEN_GClosure__P(closure), closure, 5, "g_signal_handlers_unblock_matched", "GClosure*");
  XEN_ASSERT_TYPE(XEN_gpointer_P(func), func, 6, "g_signal_handlers_unblock_matched", "gpointer");
  XEN_ASSERT_TYPE(XEN_gpointer_P(data), data, 7, "g_signal_handlers_unblock_matched", "gpointer");
  return(C_TO_XEN_guint(g_signal_handlers_unblock_matched(XEN_TO_C_gpointer(instance), XEN_TO_C_GSignalMatchType(mask), XEN_TO_C_guint(signal_id), 
                                                          XEN_TO_C_GQuark(detail), XEN_TO_C_GClosure_(closure), XEN_TO_C_gpointer(func), 
                                                          XEN_TO_C_gpointer(data))));
}

static XEN gxg_g_signal_handlers_disconnect_matched(XEN instance, XEN mask, XEN signal_id, XEN detail, XEN closure, XEN func, XEN data)
{
  #define H_g_signal_handlers_disconnect_matched "guint g_signal_handlers_disconnect_matched(gpointer instance, \
GSignalMatchType mask, guint signal_id, GQuark detail, GClosure* closure, gpointer func, gpointer data)"
  XEN_ASSERT_TYPE(XEN_gpointer_P(instance), instance, 1, "g_signal_handlers_disconnect_matched", "gpointer");
  XEN_ASSERT_TYPE(XEN_GSignalMatchType_P(mask), mask, 2, "g_signal_handlers_disconnect_matched", "GSignalMatchType");
  XEN_ASSERT_TYPE(XEN_guint_P(signal_id), signal_id, 3, "g_signal_handlers_disconnect_matched", "guint");
  XEN_ASSERT_TYPE(XEN_GQuark_P(detail), detail, 4, "g_signal_handlers_disconnect_matched", "GQuark");
  XEN_ASSERT_TYPE(XEN_GClosure__P(closure), closure, 5, "g_signal_handlers_disconnect_matched", "GClosure*");
  XEN_ASSERT_TYPE(XEN_gpointer_P(func), func, 6, "g_signal_handlers_disconnect_matched", "gpointer");
  XEN_ASSERT_TYPE(XEN_gpointer_P(data), data, 7, "g_signal_handlers_disconnect_matched", "gpointer");
  return(C_TO_XEN_guint(g_signal_handlers_disconnect_matched(XEN_TO_C_gpointer(instance), XEN_TO_C_GSignalMatchType(mask), 
                                                             XEN_TO_C_guint(signal_id), XEN_TO_C_GQuark(detail), XEN_TO_C_GClosure_(closure), 
                                                             XEN_TO_C_gpointer(func), XEN_TO_C_gpointer(data))));
}

static XEN gxg_g_signal_handlers_destroy(XEN instance)
{
  #define H_g_signal_handlers_destroy "void g_signal_handlers_destroy(gpointer instance)"
  XEN_ASSERT_TYPE(XEN_gpointer_P(instance), instance, 1, "g_signal_handlers_destroy", "gpointer");
  g_signal_handlers_destroy(XEN_TO_C_gpointer(instance));
  return(XEN_FALSE);
}

static XEN gxg__g_signals_destroy(XEN itype)
{
  #define H__g_signals_destroy "void _g_signals_destroy(GType itype)"
  XEN_ASSERT_TYPE(XEN_GType_P(itype), itype, 1, "_g_signals_destroy", "GType");
  _g_signals_destroy(XEN_TO_C_GType(itype));
  return(XEN_FALSE);
}

static XEN gxg_g_object_ref(XEN object)
{
  #define H_g_object_ref "gpointer g_object_ref(gpointer object)"
  XEN_ASSERT_TYPE(XEN_gpointer_P(object), object, 1, "g_object_ref", "gpointer");
  return(C_TO_XEN_gpointer(g_object_ref(XEN_TO_C_gpointer(object))));
}

static XEN gxg_g_object_unref(XEN object)
{
  #define H_g_object_unref "void g_object_unref(gpointer object)"
  XEN_ASSERT_TYPE(XEN_gpointer_P(object), object, 1, "g_object_unref", "gpointer");
  g_object_unref(XEN_TO_C_gpointer(object));
  return(XEN_FALSE);
}

static XEN gxg_gdk_colormap_get_type(void)
{
  #define H_gdk_colormap_get_type "GType gdk_colormap_get_type( void)"
  return(C_TO_XEN_GType(gdk_colormap_get_type()));
}
#endif

static XEN gxg_gdk_colormap_new(XEN visual, XEN allocate)
{
  #define H_gdk_colormap_new "GdkColormap* gdk_colormap_new(GdkVisual* visual, gboolean allocate)"
  XEN_ASSERT_TYPE(XEN_GdkVisual__P(visual), visual, 1, "gdk_colormap_new", "GdkVisual*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(allocate), allocate, 2, "gdk_colormap_new", "gboolean");
  return(C_TO_XEN_GdkColormap_(gdk_colormap_new(XEN_TO_C_GdkVisual_(visual), XEN_TO_C_gboolean(allocate))));
}

static XEN gxg_gdk_colormap_get_system(void)
{
  #define H_gdk_colormap_get_system "GdkColormap* gdk_colormap_get_system( void)"
  return(C_TO_XEN_GdkColormap_(gdk_colormap_get_system()));
}

static XEN gxg_gdk_colormap_alloc_colors(XEN colormap, XEN colors, XEN ncolors, XEN writeable, XEN best_match, XEN success)
{
  #define H_gdk_colormap_alloc_colors "gint gdk_colormap_alloc_colors(GdkColormap* colormap, GdkColor* colors, \
gint ncolors, gboolean writeable, gboolean best_match, gboolean* success)"
  XEN_ASSERT_TYPE(XEN_GdkColormap__P(colormap), colormap, 1, "gdk_colormap_alloc_colors", "GdkColormap*");
  XEN_ASSERT_TYPE(XEN_GdkColor__P(colors), colors, 2, "gdk_colormap_alloc_colors", "GdkColor*");
  XEN_ASSERT_TYPE(XEN_gint_P(ncolors), ncolors, 3, "gdk_colormap_alloc_colors", "gint");
  XEN_ASSERT_TYPE(XEN_gboolean_P(writeable), writeable, 4, "gdk_colormap_alloc_colors", "gboolean");
  XEN_ASSERT_TYPE(XEN_gboolean_P(best_match), best_match, 5, "gdk_colormap_alloc_colors", "gboolean");
  XEN_ASSERT_TYPE(XEN_gboolean__P(success), success, 6, "gdk_colormap_alloc_colors", "gboolean*");
  return(C_TO_XEN_gint(gdk_colormap_alloc_colors(XEN_TO_C_GdkColormap_(colormap), XEN_TO_C_GdkColor_(colors), XEN_TO_C_gint(ncolors), 
                                                 XEN_TO_C_gboolean(writeable), XEN_TO_C_gboolean(best_match), XEN_TO_C_gboolean_(success))));
}

static XEN gxg_gdk_colormap_alloc_color(XEN colormap, XEN color, XEN writeable, XEN best_match)
{
  #define H_gdk_colormap_alloc_color "gboolean gdk_colormap_alloc_color(GdkColormap* colormap, GdkColor* color, \
gboolean writeable, gboolean best_match)"
  XEN_ASSERT_TYPE(XEN_GdkColormap__P(colormap), colormap, 1, "gdk_colormap_alloc_color", "GdkColormap*");
  XEN_ASSERT_TYPE(XEN_GdkColor__P(color), color, 2, "gdk_colormap_alloc_color", "GdkColor*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(writeable), writeable, 3, "gdk_colormap_alloc_color", "gboolean");
  XEN_ASSERT_TYPE(XEN_gboolean_P(best_match), best_match, 4, "gdk_colormap_alloc_color", "gboolean");
  return(C_TO_XEN_gboolean(gdk_colormap_alloc_color(XEN_TO_C_GdkColormap_(colormap), XEN_TO_C_GdkColor_(color), XEN_TO_C_gboolean(writeable), 
                                                    XEN_TO_C_gboolean(best_match))));
}

static XEN gxg_gdk_colormap_free_colors(XEN colormap, XEN colors, XEN ncolors)
{
  #define H_gdk_colormap_free_colors "void gdk_colormap_free_colors(GdkColormap* colormap, GdkColor* colors, \
gint ncolors)"
  XEN_ASSERT_TYPE(XEN_GdkColormap__P(colormap), colormap, 1, "gdk_colormap_free_colors", "GdkColormap*");
  XEN_ASSERT_TYPE(XEN_GdkColor__P(colors), colors, 2, "gdk_colormap_free_colors", "GdkColor*");
  XEN_ASSERT_TYPE(XEN_gint_P(ncolors), ncolors, 3, "gdk_colormap_free_colors", "gint");
  gdk_colormap_free_colors(XEN_TO_C_GdkColormap_(colormap), XEN_TO_C_GdkColor_(colors), XEN_TO_C_gint(ncolors));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gdk_colormap_query_color(XEN colormap, XEN pixel, XEN result)
{
  #define H_gdk_colormap_query_color "void gdk_colormap_query_color(GdkColormap* colormap, gulong pixel, \
GdkColor* result)"
  XEN_ASSERT_TYPE(XEN_GdkColormap__P(colormap), colormap, 1, "gdk_colormap_query_color", "GdkColormap*");
  XEN_ASSERT_TYPE(XEN_gulong_P(pixel), pixel, 2, "gdk_colormap_query_color", "gulong");
  XEN_ASSERT_TYPE(XEN_GdkColor__P(result), result, 3, "gdk_colormap_query_color", "GdkColor*");
  gdk_colormap_query_color(XEN_TO_C_GdkColormap_(colormap), XEN_TO_C_gulong(pixel), XEN_TO_C_GdkColor_(result));
  return(XEN_FALSE);
}
#endif

static XEN gxg_gdk_colormap_get_visual(XEN colormap)
{
  #define H_gdk_colormap_get_visual "GdkVisual* gdk_colormap_get_visual(GdkColormap* colormap)"
  XEN_ASSERT_TYPE(XEN_GdkColormap__P(colormap), colormap, 1, "gdk_colormap_get_visual", "GdkColormap*");
  return(C_TO_XEN_GdkVisual_(gdk_colormap_get_visual(XEN_TO_C_GdkColormap_(colormap))));
}

static XEN gxg_gdk_color_copy(XEN color)
{
  #define H_gdk_color_copy "GdkColor* gdk_color_copy(GdkColor* color)"
  XEN_ASSERT_TYPE(XEN_GdkColor__P(color), color, 1, "gdk_color_copy", "GdkColor*");
  return(C_TO_XEN_GdkColor_(gdk_color_copy(XEN_TO_C_GdkColor_(color))));
}

static XEN gxg_gdk_color_free(XEN color)
{
  #define H_gdk_color_free "void gdk_color_free(GdkColor* color)"
  XEN_ASSERT_TYPE(XEN_GdkColor__P(color), color, 1, "gdk_color_free", "GdkColor*");
  gdk_color_free(XEN_TO_C_GdkColor_(color));
  return(XEN_FALSE);
}

static XEN gxg_gdk_color_parse(XEN spec, XEN color)
{
  #define H_gdk_color_parse "gint gdk_color_parse(gchar* spec, GdkColor* color)"
  XEN_ASSERT_TYPE(XEN_gchar__P(spec), spec, 1, "gdk_color_parse", "gchar*");
  XEN_ASSERT_TYPE(XEN_GdkColor__P(color), color, 2, "gdk_color_parse", "GdkColor*");
  return(C_TO_XEN_gint(gdk_color_parse(XEN_TO_C_gchar_(spec), XEN_TO_C_GdkColor_(color))));
}

#if (!HAVE_GTK_1)
static XEN gxg_gdk_color_hash(XEN colora)
{
  #define H_gdk_color_hash "guint gdk_color_hash(GdkColor* colora)"
  XEN_ASSERT_TYPE(XEN_GdkColor__P(colora), colora, 1, "gdk_color_hash", "GdkColor*");
  return(C_TO_XEN_guint(gdk_color_hash(XEN_TO_C_GdkColor_(colora))));
}
#endif

static XEN gxg_gdk_color_equal(XEN colora, XEN colorb)
{
  #define H_gdk_color_equal "gboolean gdk_color_equal(GdkColor* colora, GdkColor* colorb)"
  XEN_ASSERT_TYPE(XEN_GdkColor__P(colora), colora, 1, "gdk_color_equal", "GdkColor*");
  XEN_ASSERT_TYPE(XEN_GdkColor__P(colorb), colorb, 2, "gdk_color_equal", "GdkColor*");
  return(C_TO_XEN_gboolean(gdk_color_equal(XEN_TO_C_GdkColor_(colora), XEN_TO_C_GdkColor_(colorb))));
}

#if (!HAVE_GTK_1)
static XEN gxg_gdk_color_get_type(void)
{
  #define H_gdk_color_get_type "GType gdk_color_get_type( void)"
  return(C_TO_XEN_GType(gdk_color_get_type()));
}

static XEN gxg_gdk_cursor_get_type(void)
{
  #define H_gdk_cursor_get_type "GType gdk_cursor_get_type( void)"
  return(C_TO_XEN_GType(gdk_cursor_get_type()));
}
#endif

static XEN gxg_gdk_cursor_new(XEN cursor_type)
{
  #define H_gdk_cursor_new "GdkCursor* gdk_cursor_new(GdkCursorType cursor_type)"
  XEN_ASSERT_TYPE(XEN_GdkCursorType_P(cursor_type), cursor_type, 1, "gdk_cursor_new", "GdkCursorType");
  return(C_TO_XEN_GdkCursor_(gdk_cursor_new(XEN_TO_C_GdkCursorType(cursor_type))));
}

static XEN gxg_gdk_cursor_new_from_pixmap(XEN source, XEN mask, XEN fg, XEN bg, XEN x, XEN y)
{
  #define H_gdk_cursor_new_from_pixmap "GdkCursor* gdk_cursor_new_from_pixmap(GdkPixmap* source, GdkPixmap* mask, \
GdkColor* fg, GdkColor* bg, gint x, gint y)"
  XEN_ASSERT_TYPE(XEN_GdkPixmap__P(source), source, 1, "gdk_cursor_new_from_pixmap", "GdkPixmap*");
  XEN_ASSERT_TYPE(XEN_GdkPixmap__P(mask), mask, 2, "gdk_cursor_new_from_pixmap", "GdkPixmap*");
  XEN_ASSERT_TYPE(XEN_GdkColor__P(fg), fg, 3, "gdk_cursor_new_from_pixmap", "GdkColor*");
  XEN_ASSERT_TYPE(XEN_GdkColor__P(bg), bg, 4, "gdk_cursor_new_from_pixmap", "GdkColor*");
  XEN_ASSERT_TYPE(XEN_gint_P(x), x, 5, "gdk_cursor_new_from_pixmap", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(y), y, 6, "gdk_cursor_new_from_pixmap", "gint");
  return(C_TO_XEN_GdkCursor_(gdk_cursor_new_from_pixmap(XEN_TO_C_GdkPixmap_(source), XEN_TO_C_GdkPixmap_(mask), XEN_TO_C_GdkColor_(fg), 
                                                        XEN_TO_C_GdkColor_(bg), XEN_TO_C_gint(x), XEN_TO_C_gint(y))));
}

#if (!HAVE_GTK_1)
static XEN gxg_gdk_cursor_ref(XEN cursor)
{
  #define H_gdk_cursor_ref "GdkCursor* gdk_cursor_ref(GdkCursor* cursor)"
  XEN_ASSERT_TYPE(XEN_GdkCursor__P(cursor), cursor, 1, "gdk_cursor_ref", "GdkCursor*");
  return(C_TO_XEN_GdkCursor_(gdk_cursor_ref(XEN_TO_C_GdkCursor_(cursor))));
}

static XEN gxg_gdk_cursor_unref(XEN cursor)
{
  #define H_gdk_cursor_unref "void gdk_cursor_unref(GdkCursor* cursor)"
  XEN_ASSERT_TYPE(XEN_GdkCursor__P(cursor), cursor, 1, "gdk_cursor_unref", "GdkCursor*");
  gdk_cursor_unref(XEN_TO_C_GdkCursor_(cursor));
  return(XEN_FALSE);
}

static XEN gxg_gdk_drag_context_get_type(void)
{
  #define H_gdk_drag_context_get_type "GType gdk_drag_context_get_type( void)"
  return(C_TO_XEN_GType(gdk_drag_context_get_type()));
}
#endif

static XEN gxg_gdk_drag_context_new(void)
{
  #define H_gdk_drag_context_new "GdkDragContext* gdk_drag_context_new( void)"
  return(C_TO_XEN_GdkDragContext_(gdk_drag_context_new()));
}

static XEN gxg_gdk_drag_status(XEN context, XEN action, XEN time)
{
  #define H_gdk_drag_status "void gdk_drag_status(GdkDragContext* context, GdkDragAction action, guint32 time)"
  XEN_ASSERT_TYPE(XEN_GdkDragContext__P(context), context, 1, "gdk_drag_status", "GdkDragContext*");
  XEN_ASSERT_TYPE(XEN_GdkDragAction_P(action), action, 2, "gdk_drag_status", "GdkDragAction");
  XEN_ASSERT_TYPE(XEN_guint32_P(time), time, 3, "gdk_drag_status", "guint32");
  gdk_drag_status(XEN_TO_C_GdkDragContext_(context), XEN_TO_C_GdkDragAction(action), XEN_TO_C_guint32(time));
  return(XEN_FALSE);
}

static XEN gxg_gdk_drop_reply(XEN context, XEN ok, XEN time)
{
  #define H_gdk_drop_reply "void gdk_drop_reply(GdkDragContext* context, gboolean ok, guint32 time)"
  XEN_ASSERT_TYPE(XEN_GdkDragContext__P(context), context, 1, "gdk_drop_reply", "GdkDragContext*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(ok), ok, 2, "gdk_drop_reply", "gboolean");
  XEN_ASSERT_TYPE(XEN_guint32_P(time), time, 3, "gdk_drop_reply", "guint32");
  gdk_drop_reply(XEN_TO_C_GdkDragContext_(context), XEN_TO_C_gboolean(ok), XEN_TO_C_guint32(time));
  return(XEN_FALSE);
}

static XEN gxg_gdk_drop_finish(XEN context, XEN success, XEN time)
{
  #define H_gdk_drop_finish "void gdk_drop_finish(GdkDragContext* context, gboolean success, guint32 time)"
  XEN_ASSERT_TYPE(XEN_GdkDragContext__P(context), context, 1, "gdk_drop_finish", "GdkDragContext*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(success), success, 2, "gdk_drop_finish", "gboolean");
  XEN_ASSERT_TYPE(XEN_guint32_P(time), time, 3, "gdk_drop_finish", "guint32");
  gdk_drop_finish(XEN_TO_C_GdkDragContext_(context), XEN_TO_C_gboolean(success), XEN_TO_C_guint32(time));
  return(XEN_FALSE);
}

static XEN gxg_gdk_drag_get_selection(XEN context)
{
  #define H_gdk_drag_get_selection "GdkAtom gdk_drag_get_selection(GdkDragContext* context)"
  XEN_ASSERT_TYPE(XEN_GdkDragContext__P(context), context, 1, "gdk_drag_get_selection", "GdkDragContext*");
  return(C_TO_XEN_GdkAtom(gdk_drag_get_selection(XEN_TO_C_GdkDragContext_(context))));
}

static XEN gxg_gdk_drag_begin(XEN window, XEN targets)
{
  #define H_gdk_drag_begin "GdkDragContext* gdk_drag_begin(GdkWindow* window, GList* targets)"
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 1, "gdk_drag_begin", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_GList__P(targets), targets, 2, "gdk_drag_begin", "GList*");
  return(C_TO_XEN_GdkDragContext_(gdk_drag_begin(XEN_TO_C_GdkWindow_(window), XEN_TO_C_GList_(targets))));
}

static XEN gxg_gdk_drag_get_protocol(XEN xid, XEN protocol)
{
  #define H_gdk_drag_get_protocol "guint32 gdk_drag_get_protocol(guint32 xid, GdkDragProtocol* protocol)"
  XEN_ASSERT_TYPE(XEN_guint32_P(xid), xid, 1, "gdk_drag_get_protocol", "guint32");
  XEN_ASSERT_TYPE(XEN_GdkDragProtocol__P(protocol), protocol, 2, "gdk_drag_get_protocol", "GdkDragProtocol*");
  return(C_TO_XEN_guint32(gdk_drag_get_protocol(XEN_TO_C_guint32(xid), XEN_TO_C_GdkDragProtocol_(protocol))));
}

static XEN gxg_gdk_drag_find_window(XEN context, XEN drag_window, XEN x_root, XEN y_root, XEN dest_window, XEN protocol)
{
  #define H_gdk_drag_find_window "void gdk_drag_find_window(GdkDragContext* context, GdkWindow* drag_window, \
gint x_root, gint y_root, GdkWindow** [dest_window], GdkDragProtocol* [protocol])"
  GdkWindow* ref_dest_window;
  GdkDragProtocol ref_protocol;
  XEN_ASSERT_TYPE(XEN_GdkDragContext__P(context), context, 1, "gdk_drag_find_window", "GdkDragContext*");
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(drag_window), drag_window, 2, "gdk_drag_find_window", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_gint_P(x_root), x_root, 3, "gdk_drag_find_window", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(y_root), y_root, 4, "gdk_drag_find_window", "gint");
  gdk_drag_find_window(XEN_TO_C_GdkDragContext_(context), XEN_TO_C_GdkWindow_(drag_window), XEN_TO_C_gint(x_root), XEN_TO_C_gint(y_root), 
                       &ref_dest_window, &ref_protocol);
  return(XEN_LIST_2(C_TO_XEN_GdkWindow_(ref_dest_window), C_TO_XEN_GdkDragProtocol(ref_protocol)));
}

static XEN gxg_gdk_drag_motion(XEN context, XEN dest_window, XEN protocol, XEN x_root, XEN y_root, XEN suggested_action, XEN possible_actions, XEN time)
{
  #define H_gdk_drag_motion "gboolean gdk_drag_motion(GdkDragContext* context, GdkWindow* dest_window, \
GdkDragProtocol protocol, gint x_root, gint y_root, GdkDragAction suggested_action, GdkDragAction possible_actions, \
guint32 time)"
  XEN_ASSERT_TYPE(XEN_GdkDragContext__P(context), context, 1, "gdk_drag_motion", "GdkDragContext*");
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(dest_window), dest_window, 2, "gdk_drag_motion", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_GdkDragProtocol_P(protocol), protocol, 3, "gdk_drag_motion", "GdkDragProtocol");
  XEN_ASSERT_TYPE(XEN_gint_P(x_root), x_root, 4, "gdk_drag_motion", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(y_root), y_root, 5, "gdk_drag_motion", "gint");
  XEN_ASSERT_TYPE(XEN_GdkDragAction_P(suggested_action), suggested_action, 6, "gdk_drag_motion", "GdkDragAction");
  XEN_ASSERT_TYPE(XEN_GdkDragAction_P(possible_actions), possible_actions, 7, "gdk_drag_motion", "GdkDragAction");
  XEN_ASSERT_TYPE(XEN_guint32_P(time), time, 8, "gdk_drag_motion", "guint32");
  return(C_TO_XEN_gboolean(gdk_drag_motion(XEN_TO_C_GdkDragContext_(context), XEN_TO_C_GdkWindow_(dest_window), XEN_TO_C_GdkDragProtocol(protocol), 
                                           XEN_TO_C_gint(x_root), XEN_TO_C_gint(y_root), XEN_TO_C_GdkDragAction(suggested_action), 
                                           XEN_TO_C_GdkDragAction(possible_actions), XEN_TO_C_guint32(time))));
}

static XEN gxg_gdk_drag_drop(XEN context, XEN time)
{
  #define H_gdk_drag_drop "void gdk_drag_drop(GdkDragContext* context, guint32 time)"
  XEN_ASSERT_TYPE(XEN_GdkDragContext__P(context), context, 1, "gdk_drag_drop", "GdkDragContext*");
  XEN_ASSERT_TYPE(XEN_guint32_P(time), time, 2, "gdk_drag_drop", "guint32");
  gdk_drag_drop(XEN_TO_C_GdkDragContext_(context), XEN_TO_C_guint32(time));
  return(XEN_FALSE);
}

static XEN gxg_gdk_drag_abort(XEN context, XEN time)
{
  #define H_gdk_drag_abort "void gdk_drag_abort(GdkDragContext* context, guint32 time)"
  XEN_ASSERT_TYPE(XEN_GdkDragContext__P(context), context, 1, "gdk_drag_abort", "GdkDragContext*");
  XEN_ASSERT_TYPE(XEN_guint32_P(time), time, 2, "gdk_drag_abort", "guint32");
  gdk_drag_abort(XEN_TO_C_GdkDragContext_(context), XEN_TO_C_guint32(time));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gdk_drawable_get_type(void)
{
  #define H_gdk_drawable_get_type "GType gdk_drawable_get_type( void)"
  return(C_TO_XEN_GType(gdk_drawable_get_type()));
}

static XEN gxg_gdk_drawable_get_size(XEN drawable, XEN width, XEN height)
{
  #define H_gdk_drawable_get_size "void gdk_drawable_get_size(GdkDrawable* drawable, gint* [width], gint* [height])"
  gint ref_width;
  gint ref_height;
  XEN_ASSERT_TYPE(XEN_GdkDrawable__P(drawable), drawable, 1, "gdk_drawable_get_size", "GdkDrawable*");
  gdk_drawable_get_size(XEN_TO_C_GdkDrawable_(drawable), &ref_width, &ref_height);
  return(XEN_LIST_2(C_TO_XEN_gint(ref_width), C_TO_XEN_gint(ref_height)));
}

static XEN gxg_gdk_drawable_set_colormap(XEN drawable, XEN colormap)
{
  #define H_gdk_drawable_set_colormap "void gdk_drawable_set_colormap(GdkDrawable* drawable, GdkColormap* colormap)"
  XEN_ASSERT_TYPE(XEN_GdkDrawable__P(drawable), drawable, 1, "gdk_drawable_set_colormap", "GdkDrawable*");
  XEN_ASSERT_TYPE(XEN_GdkColormap__P(colormap), colormap, 2, "gdk_drawable_set_colormap", "GdkColormap*");
  gdk_drawable_set_colormap(XEN_TO_C_GdkDrawable_(drawable), XEN_TO_C_GdkColormap_(colormap));
  return(XEN_FALSE);
}

static XEN gxg_gdk_drawable_get_colormap(XEN drawable)
{
  #define H_gdk_drawable_get_colormap "GdkColormap* gdk_drawable_get_colormap(GdkDrawable* drawable)"
  XEN_ASSERT_TYPE(XEN_GdkDrawable__P(drawable), drawable, 1, "gdk_drawable_get_colormap", "GdkDrawable*");
  return(C_TO_XEN_GdkColormap_(gdk_drawable_get_colormap(XEN_TO_C_GdkDrawable_(drawable))));
}

static XEN gxg_gdk_drawable_get_visual(XEN drawable)
{
  #define H_gdk_drawable_get_visual "GdkVisual* gdk_drawable_get_visual(GdkDrawable* drawable)"
  XEN_ASSERT_TYPE(XEN_GdkDrawable__P(drawable), drawable, 1, "gdk_drawable_get_visual", "GdkDrawable*");
  return(C_TO_XEN_GdkVisual_(gdk_drawable_get_visual(XEN_TO_C_GdkDrawable_(drawable))));
}

static XEN gxg_gdk_drawable_get_depth(XEN drawable)
{
  #define H_gdk_drawable_get_depth "gint gdk_drawable_get_depth(GdkDrawable* drawable)"
  XEN_ASSERT_TYPE(XEN_GdkDrawable__P(drawable), drawable, 1, "gdk_drawable_get_depth", "GdkDrawable*");
  return(C_TO_XEN_gint(gdk_drawable_get_depth(XEN_TO_C_GdkDrawable_(drawable))));
}
#endif

static XEN gxg_gdk_draw_point(XEN drawable, XEN gc, XEN x, XEN y)
{
  #define H_gdk_draw_point "void gdk_draw_point(GdkDrawable* drawable, GdkGC* gc, gint x, gint y)"
  XEN_ASSERT_TYPE(XEN_GdkDrawable__P(drawable), drawable, 1, "gdk_draw_point", "GdkDrawable*");
  XEN_ASSERT_TYPE(XEN_GdkGC__P(gc), gc, 2, "gdk_draw_point", "GdkGC*");
  XEN_ASSERT_TYPE(XEN_gint_P(x), x, 3, "gdk_draw_point", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(y), y, 4, "gdk_draw_point", "gint");
  gdk_draw_point(XEN_TO_C_GdkDrawable_(drawable), XEN_TO_C_GdkGC_(gc), XEN_TO_C_gint(x), XEN_TO_C_gint(y));
  return(XEN_FALSE);
}

static XEN gxg_gdk_draw_line(XEN drawable, XEN gc, XEN x1, XEN y1, XEN x2, XEN y2)
{
  #define H_gdk_draw_line "void gdk_draw_line(GdkDrawable* drawable, GdkGC* gc, gint x1, gint y1, gint x2, \
gint y2)"
  XEN_ASSERT_TYPE(XEN_GdkDrawable__P(drawable), drawable, 1, "gdk_draw_line", "GdkDrawable*");
  XEN_ASSERT_TYPE(XEN_GdkGC__P(gc), gc, 2, "gdk_draw_line", "GdkGC*");
  XEN_ASSERT_TYPE(XEN_gint_P(x1), x1, 3, "gdk_draw_line", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(y1), y1, 4, "gdk_draw_line", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(x2), x2, 5, "gdk_draw_line", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(y2), y2, 6, "gdk_draw_line", "gint");
  gdk_draw_line(XEN_TO_C_GdkDrawable_(drawable), XEN_TO_C_GdkGC_(gc), XEN_TO_C_gint(x1), XEN_TO_C_gint(y1), XEN_TO_C_gint(x2), 
                XEN_TO_C_gint(y2));
  return(XEN_FALSE);
}

static XEN gxg_gdk_draw_rectangle(XEN drawable, XEN gc, XEN filled, XEN x, XEN y, XEN width, XEN height)
{
  #define H_gdk_draw_rectangle "void gdk_draw_rectangle(GdkDrawable* drawable, GdkGC* gc, gint filled, \
gint x, gint y, gint width, gint height)"
  XEN_ASSERT_TYPE(XEN_GdkDrawable__P(drawable), drawable, 1, "gdk_draw_rectangle", "GdkDrawable*");
  XEN_ASSERT_TYPE(XEN_GdkGC__P(gc), gc, 2, "gdk_draw_rectangle", "GdkGC*");
  XEN_ASSERT_TYPE(XEN_gint_P(filled), filled, 3, "gdk_draw_rectangle", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(x), x, 4, "gdk_draw_rectangle", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(y), y, 5, "gdk_draw_rectangle", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(width), width, 6, "gdk_draw_rectangle", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(height), height, 7, "gdk_draw_rectangle", "gint");
  gdk_draw_rectangle(XEN_TO_C_GdkDrawable_(drawable), XEN_TO_C_GdkGC_(gc), XEN_TO_C_gint(filled), XEN_TO_C_gint(x), XEN_TO_C_gint(y), 
                     XEN_TO_C_gint(width), XEN_TO_C_gint(height));
  return(XEN_FALSE);
}

static XEN gxg_gdk_draw_arc(XEN drawable, XEN gc, XEN filled, XEN x, XEN y, XEN width, XEN height, XEN angle1, XEN angle2)
{
  #define H_gdk_draw_arc "void gdk_draw_arc(GdkDrawable* drawable, GdkGC* gc, gint filled, gint x, gint y, \
gint width, gint height, gint angle1, gint angle2)"
  XEN_ASSERT_TYPE(XEN_GdkDrawable__P(drawable), drawable, 1, "gdk_draw_arc", "GdkDrawable*");
  XEN_ASSERT_TYPE(XEN_GdkGC__P(gc), gc, 2, "gdk_draw_arc", "GdkGC*");
  XEN_ASSERT_TYPE(XEN_gint_P(filled), filled, 3, "gdk_draw_arc", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(x), x, 4, "gdk_draw_arc", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(y), y, 5, "gdk_draw_arc", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(width), width, 6, "gdk_draw_arc", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(height), height, 7, "gdk_draw_arc", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(angle1), angle1, 8, "gdk_draw_arc", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(angle2), angle2, 9, "gdk_draw_arc", "gint");
  gdk_draw_arc(XEN_TO_C_GdkDrawable_(drawable), XEN_TO_C_GdkGC_(gc), XEN_TO_C_gint(filled), XEN_TO_C_gint(x), XEN_TO_C_gint(y), 
               XEN_TO_C_gint(width), XEN_TO_C_gint(height), XEN_TO_C_gint(angle1), XEN_TO_C_gint(angle2));
  return(XEN_FALSE);
}

static XEN gxg_gdk_draw_polygon(XEN drawable, XEN gc, XEN filled, XEN points, XEN npoints)
{
  #define H_gdk_draw_polygon "void gdk_draw_polygon(GdkDrawable* drawable, GdkGC* gc, gint filled, GdkPoint* points, \
gint npoints)"
  XEN_ASSERT_TYPE(XEN_GdkDrawable__P(drawable), drawable, 1, "gdk_draw_polygon", "GdkDrawable*");
  XEN_ASSERT_TYPE(XEN_GdkGC__P(gc), gc, 2, "gdk_draw_polygon", "GdkGC*");
  XEN_ASSERT_TYPE(XEN_gint_P(filled), filled, 3, "gdk_draw_polygon", "gint");
  XEN_ASSERT_TYPE(XEN_GdkPoint__P(points), points, 4, "gdk_draw_polygon", "GdkPoint*");
  XEN_ASSERT_TYPE(XEN_gint_P(npoints), npoints, 5, "gdk_draw_polygon", "gint");
  gdk_draw_polygon(XEN_TO_C_GdkDrawable_(drawable), XEN_TO_C_GdkGC_(gc), XEN_TO_C_gint(filled), XEN_TO_C_GdkPoint_(points), 
                   XEN_TO_C_gint(npoints));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gdk_draw_drawable(XEN drawable, XEN gc, XEN src, XEN xsrc, XEN ysrc, XEN xdest, XEN ydest, XEN width, XEN height)
{
  #define H_gdk_draw_drawable "void gdk_draw_drawable(GdkDrawable* drawable, GdkGC* gc, GdkDrawable* src, \
gint xsrc, gint ysrc, gint xdest, gint ydest, gint width, gint height)"
  XEN_ASSERT_TYPE(XEN_GdkDrawable__P(drawable), drawable, 1, "gdk_draw_drawable", "GdkDrawable*");
  XEN_ASSERT_TYPE(XEN_GdkGC__P(gc), gc, 2, "gdk_draw_drawable", "GdkGC*");
  XEN_ASSERT_TYPE(XEN_GdkDrawable__P(src), src, 3, "gdk_draw_drawable", "GdkDrawable*");
  XEN_ASSERT_TYPE(XEN_gint_P(xsrc), xsrc, 4, "gdk_draw_drawable", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(ysrc), ysrc, 5, "gdk_draw_drawable", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(xdest), xdest, 6, "gdk_draw_drawable", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(ydest), ydest, 7, "gdk_draw_drawable", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(width), width, 8, "gdk_draw_drawable", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(height), height, 9, "gdk_draw_drawable", "gint");
  gdk_draw_drawable(XEN_TO_C_GdkDrawable_(drawable), XEN_TO_C_GdkGC_(gc), XEN_TO_C_GdkDrawable_(src), XEN_TO_C_gint(xsrc), 
                    XEN_TO_C_gint(ysrc), XEN_TO_C_gint(xdest), XEN_TO_C_gint(ydest), XEN_TO_C_gint(width), XEN_TO_C_gint(height));
  return(XEN_FALSE);
}
#endif

static XEN gxg_gdk_draw_image(XEN drawable, XEN gc, XEN image, XEN xsrc, XEN ysrc, XEN xdest, XEN ydest, XEN width, XEN height)
{
  #define H_gdk_draw_image "void gdk_draw_image(GdkDrawable* drawable, GdkGC* gc, GdkImage* image, gint xsrc, \
gint ysrc, gint xdest, gint ydest, gint width, gint height)"
  XEN_ASSERT_TYPE(XEN_GdkDrawable__P(drawable), drawable, 1, "gdk_draw_image", "GdkDrawable*");
  XEN_ASSERT_TYPE(XEN_GdkGC__P(gc), gc, 2, "gdk_draw_image", "GdkGC*");
  XEN_ASSERT_TYPE(XEN_GdkImage__P(image), image, 3, "gdk_draw_image", "GdkImage*");
  XEN_ASSERT_TYPE(XEN_gint_P(xsrc), xsrc, 4, "gdk_draw_image", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(ysrc), ysrc, 5, "gdk_draw_image", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(xdest), xdest, 6, "gdk_draw_image", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(ydest), ydest, 7, "gdk_draw_image", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(width), width, 8, "gdk_draw_image", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(height), height, 9, "gdk_draw_image", "gint");
  gdk_draw_image(XEN_TO_C_GdkDrawable_(drawable), XEN_TO_C_GdkGC_(gc), XEN_TO_C_GdkImage_(image), XEN_TO_C_gint(xsrc), XEN_TO_C_gint(ysrc), 
                 XEN_TO_C_gint(xdest), XEN_TO_C_gint(ydest), XEN_TO_C_gint(width), XEN_TO_C_gint(height));
  return(XEN_FALSE);
}

static XEN gxg_gdk_draw_points(XEN drawable, XEN gc, XEN points, XEN npoints)
{
  #define H_gdk_draw_points "void gdk_draw_points(GdkDrawable* drawable, GdkGC* gc, GdkPoint* points, \
gint npoints)"
  XEN_ASSERT_TYPE(XEN_GdkDrawable__P(drawable), drawable, 1, "gdk_draw_points", "GdkDrawable*");
  XEN_ASSERT_TYPE(XEN_GdkGC__P(gc), gc, 2, "gdk_draw_points", "GdkGC*");
  XEN_ASSERT_TYPE(XEN_GdkPoint__P(points), points, 3, "gdk_draw_points", "GdkPoint*");
  XEN_ASSERT_TYPE(XEN_gint_P(npoints), npoints, 4, "gdk_draw_points", "gint");
  gdk_draw_points(XEN_TO_C_GdkDrawable_(drawable), XEN_TO_C_GdkGC_(gc), XEN_TO_C_GdkPoint_(points), XEN_TO_C_gint(npoints));
  return(XEN_FALSE);
}

static XEN gxg_gdk_draw_segments(XEN drawable, XEN gc, XEN segs, XEN nsegs)
{
  #define H_gdk_draw_segments "void gdk_draw_segments(GdkDrawable* drawable, GdkGC* gc, GdkSegment* segs, \
gint nsegs)"
  XEN_ASSERT_TYPE(XEN_GdkDrawable__P(drawable), drawable, 1, "gdk_draw_segments", "GdkDrawable*");
  XEN_ASSERT_TYPE(XEN_GdkGC__P(gc), gc, 2, "gdk_draw_segments", "GdkGC*");
  XEN_ASSERT_TYPE(XEN_GdkSegment__P(segs), segs, 3, "gdk_draw_segments", "GdkSegment*");
  XEN_ASSERT_TYPE(XEN_gint_P(nsegs), nsegs, 4, "gdk_draw_segments", "gint");
  gdk_draw_segments(XEN_TO_C_GdkDrawable_(drawable), XEN_TO_C_GdkGC_(gc), XEN_TO_C_GdkSegment_(segs), XEN_TO_C_gint(nsegs));
  return(XEN_FALSE);
}

static XEN gxg_gdk_draw_lines(XEN drawable, XEN gc, XEN points, XEN npoints)
{
  #define H_gdk_draw_lines "void gdk_draw_lines(GdkDrawable* drawable, GdkGC* gc, GdkPoint* points, gint npoints)"
  XEN_ASSERT_TYPE(XEN_GdkDrawable__P(drawable), drawable, 1, "gdk_draw_lines", "GdkDrawable*");
  XEN_ASSERT_TYPE(XEN_GdkGC__P(gc), gc, 2, "gdk_draw_lines", "GdkGC*");
  XEN_ASSERT_TYPE(XEN_GdkPoint__P(points), points, 3, "gdk_draw_lines", "GdkPoint*");
  XEN_ASSERT_TYPE(XEN_gint_P(npoints), npoints, 4, "gdk_draw_lines", "gint");
  gdk_draw_lines(XEN_TO_C_GdkDrawable_(drawable), XEN_TO_C_GdkGC_(gc), XEN_TO_C_GdkPoint_(points), XEN_TO_C_gint(npoints));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gdk_draw_glyphs(XEN drawable, XEN gc, XEN font, XEN x, XEN y, XEN glyphs)
{
  #define H_gdk_draw_glyphs "void gdk_draw_glyphs(GdkDrawable* drawable, GdkGC* gc, PangoFont* font, \
gint x, gint y, PangoGlyphString* glyphs)"
  XEN_ASSERT_TYPE(XEN_GdkDrawable__P(drawable), drawable, 1, "gdk_draw_glyphs", "GdkDrawable*");
  XEN_ASSERT_TYPE(XEN_GdkGC__P(gc), gc, 2, "gdk_draw_glyphs", "GdkGC*");
  XEN_ASSERT_TYPE(XEN_PangoFont__P(font), font, 3, "gdk_draw_glyphs", "PangoFont*");
  XEN_ASSERT_TYPE(XEN_gint_P(x), x, 4, "gdk_draw_glyphs", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(y), y, 5, "gdk_draw_glyphs", "gint");
  XEN_ASSERT_TYPE(XEN_PangoGlyphString__P(glyphs), glyphs, 6, "gdk_draw_glyphs", "PangoGlyphString*");
  gdk_draw_glyphs(XEN_TO_C_GdkDrawable_(drawable), XEN_TO_C_GdkGC_(gc), XEN_TO_C_PangoFont_(font), XEN_TO_C_gint(x), XEN_TO_C_gint(y), 
                  XEN_TO_C_PangoGlyphString_(glyphs));
  return(XEN_FALSE);
}

static XEN gxg_gdk_draw_layout_line(XEN drawable, XEN gc, XEN x, XEN y, XEN line)
{
  #define H_gdk_draw_layout_line "void gdk_draw_layout_line(GdkDrawable* drawable, GdkGC* gc, gint x, \
gint y, PangoLayoutLine* line)"
  XEN_ASSERT_TYPE(XEN_GdkDrawable__P(drawable), drawable, 1, "gdk_draw_layout_line", "GdkDrawable*");
  XEN_ASSERT_TYPE(XEN_GdkGC__P(gc), gc, 2, "gdk_draw_layout_line", "GdkGC*");
  XEN_ASSERT_TYPE(XEN_gint_P(x), x, 3, "gdk_draw_layout_line", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(y), y, 4, "gdk_draw_layout_line", "gint");
  XEN_ASSERT_TYPE(XEN_PangoLayoutLine__P(line), line, 5, "gdk_draw_layout_line", "PangoLayoutLine*");
  gdk_draw_layout_line(XEN_TO_C_GdkDrawable_(drawable), XEN_TO_C_GdkGC_(gc), XEN_TO_C_gint(x), XEN_TO_C_gint(y), XEN_TO_C_PangoLayoutLine_(line));
  return(XEN_FALSE);
}

static XEN gxg_gdk_draw_layout(XEN drawable, XEN gc, XEN x, XEN y, XEN layout)
{
  #define H_gdk_draw_layout "void gdk_draw_layout(GdkDrawable* drawable, GdkGC* gc, gint x, gint y, PangoLayout* layout)"
  XEN_ASSERT_TYPE(XEN_GdkDrawable__P(drawable), drawable, 1, "gdk_draw_layout", "GdkDrawable*");
  XEN_ASSERT_TYPE(XEN_GdkGC__P(gc), gc, 2, "gdk_draw_layout", "GdkGC*");
  XEN_ASSERT_TYPE(XEN_gint_P(x), x, 3, "gdk_draw_layout", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(y), y, 4, "gdk_draw_layout", "gint");
  XEN_ASSERT_TYPE(XEN_PangoLayout__P(layout), layout, 5, "gdk_draw_layout", "PangoLayout*");
  gdk_draw_layout(XEN_TO_C_GdkDrawable_(drawable), XEN_TO_C_GdkGC_(gc), XEN_TO_C_gint(x), XEN_TO_C_gint(y), XEN_TO_C_PangoLayout_(layout));
  return(XEN_FALSE);
}

static XEN gxg_gdk_draw_layout_line_with_colors(XEN drawable, XEN gc, XEN x, XEN y, XEN line, XEN foreground, XEN background)
{
  #define H_gdk_draw_layout_line_with_colors "void gdk_draw_layout_line_with_colors(GdkDrawable* drawable, \
GdkGC* gc, gint x, gint y, PangoLayoutLine* line, GdkColor* foreground, GdkColor* background)"
  XEN_ASSERT_TYPE(XEN_GdkDrawable__P(drawable), drawable, 1, "gdk_draw_layout_line_with_colors", "GdkDrawable*");
  XEN_ASSERT_TYPE(XEN_GdkGC__P(gc), gc, 2, "gdk_draw_layout_line_with_colors", "GdkGC*");
  XEN_ASSERT_TYPE(XEN_gint_P(x), x, 3, "gdk_draw_layout_line_with_colors", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(y), y, 4, "gdk_draw_layout_line_with_colors", "gint");
  XEN_ASSERT_TYPE(XEN_PangoLayoutLine__P(line), line, 5, "gdk_draw_layout_line_with_colors", "PangoLayoutLine*");
  XEN_ASSERT_TYPE(XEN_GdkColor__P(foreground) || XEN_FALSE_P(foreground), foreground, 6, "gdk_draw_layout_line_with_colors", "GdkColor*");
  XEN_ASSERT_TYPE(XEN_GdkColor__P(background) || XEN_FALSE_P(background), background, 7, "gdk_draw_layout_line_with_colors", "GdkColor*");
  gdk_draw_layout_line_with_colors(XEN_TO_C_GdkDrawable_(drawable), XEN_TO_C_GdkGC_(gc), XEN_TO_C_gint(x), XEN_TO_C_gint(y), 
                                   XEN_TO_C_PangoLayoutLine_(line), XEN_TO_C_GdkColor_(foreground), XEN_TO_C_GdkColor_(background));
  return(XEN_FALSE);
}

static XEN gxg_gdk_draw_layout_with_colors(XEN drawable, XEN gc, XEN x, XEN y, XEN layout, XEN foreground, XEN background)
{
  #define H_gdk_draw_layout_with_colors "void gdk_draw_layout_with_colors(GdkDrawable* drawable, GdkGC* gc, \
gint x, gint y, PangoLayout* layout, GdkColor* foreground, GdkColor* background)"
  XEN_ASSERT_TYPE(XEN_GdkDrawable__P(drawable), drawable, 1, "gdk_draw_layout_with_colors", "GdkDrawable*");
  XEN_ASSERT_TYPE(XEN_GdkGC__P(gc), gc, 2, "gdk_draw_layout_with_colors", "GdkGC*");
  XEN_ASSERT_TYPE(XEN_gint_P(x), x, 3, "gdk_draw_layout_with_colors", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(y), y, 4, "gdk_draw_layout_with_colors", "gint");
  XEN_ASSERT_TYPE(XEN_PangoLayout__P(layout), layout, 5, "gdk_draw_layout_with_colors", "PangoLayout*");
  XEN_ASSERT_TYPE(XEN_GdkColor__P(foreground) || XEN_FALSE_P(foreground), foreground, 6, "gdk_draw_layout_with_colors", "GdkColor*");
  XEN_ASSERT_TYPE(XEN_GdkColor__P(background) || XEN_FALSE_P(background), background, 7, "gdk_draw_layout_with_colors", "GdkColor*");
  gdk_draw_layout_with_colors(XEN_TO_C_GdkDrawable_(drawable), XEN_TO_C_GdkGC_(gc), XEN_TO_C_gint(x), XEN_TO_C_gint(y), XEN_TO_C_PangoLayout_(layout), 
                              XEN_TO_C_GdkColor_(foreground), XEN_TO_C_GdkColor_(background));
  return(XEN_FALSE);
}

static XEN gxg_gdk_drawable_get_image(XEN drawable, XEN x, XEN y, XEN width, XEN height)
{
  #define H_gdk_drawable_get_image "GdkImage* gdk_drawable_get_image(GdkDrawable* drawable, gint x, gint y, \
gint width, gint height)"
  XEN_ASSERT_TYPE(XEN_GdkDrawable__P(drawable), drawable, 1, "gdk_drawable_get_image", "GdkDrawable*");
  XEN_ASSERT_TYPE(XEN_gint_P(x), x, 2, "gdk_drawable_get_image", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(y), y, 3, "gdk_drawable_get_image", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(width), width, 4, "gdk_drawable_get_image", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(height), height, 5, "gdk_drawable_get_image", "gint");
  return(C_TO_XEN_GdkImage_(gdk_drawable_get_image(XEN_TO_C_GdkDrawable_(drawable), XEN_TO_C_gint(x), XEN_TO_C_gint(y), XEN_TO_C_gint(width), 
                                                   XEN_TO_C_gint(height))));
}

static XEN gxg_gdk_drawable_get_clip_region(XEN drawable)
{
  #define H_gdk_drawable_get_clip_region "GdkRegion* gdk_drawable_get_clip_region(GdkDrawable* drawable)"
  XEN_ASSERT_TYPE(XEN_GdkDrawable__P(drawable), drawable, 1, "gdk_drawable_get_clip_region", "GdkDrawable*");
  return(C_TO_XEN_GdkRegion_(gdk_drawable_get_clip_region(XEN_TO_C_GdkDrawable_(drawable))));
}

static XEN gxg_gdk_drawable_get_visible_region(XEN drawable)
{
  #define H_gdk_drawable_get_visible_region "GdkRegion* gdk_drawable_get_visible_region(GdkDrawable* drawable)"
  XEN_ASSERT_TYPE(XEN_GdkDrawable__P(drawable), drawable, 1, "gdk_drawable_get_visible_region", "GdkDrawable*");
  return(C_TO_XEN_GdkRegion_(gdk_drawable_get_visible_region(XEN_TO_C_GdkDrawable_(drawable))));
}

static XEN gxg_gdk_cursor_type_get_type(void)
{
  #define H_gdk_cursor_type_get_type "GType gdk_cursor_type_get_type( void)"
  return(C_TO_XEN_GType(gdk_cursor_type_get_type()));
}

static XEN gxg_gdk_drag_action_get_type(void)
{
  #define H_gdk_drag_action_get_type "GType gdk_drag_action_get_type( void)"
  return(C_TO_XEN_GType(gdk_drag_action_get_type()));
}

static XEN gxg_gdk_drag_protocol_get_type(void)
{
  #define H_gdk_drag_protocol_get_type "GType gdk_drag_protocol_get_type( void)"
  return(C_TO_XEN_GType(gdk_drag_protocol_get_type()));
}

static XEN gxg_gdk_filter_return_get_type(void)
{
  #define H_gdk_filter_return_get_type "GType gdk_filter_return_get_type( void)"
  return(C_TO_XEN_GType(gdk_filter_return_get_type()));
}

static XEN gxg_gdk_event_type_get_type(void)
{
  #define H_gdk_event_type_get_type "GType gdk_event_type_get_type( void)"
  return(C_TO_XEN_GType(gdk_event_type_get_type()));
}

static XEN gxg_gdk_event_mask_get_type(void)
{
  #define H_gdk_event_mask_get_type "GType gdk_event_mask_get_type( void)"
  return(C_TO_XEN_GType(gdk_event_mask_get_type()));
}

static XEN gxg_gdk_visibility_state_get_type(void)
{
  #define H_gdk_visibility_state_get_type "GType gdk_visibility_state_get_type( void)"
  return(C_TO_XEN_GType(gdk_visibility_state_get_type()));
}

static XEN gxg_gdk_scroll_direction_get_type(void)
{
  #define H_gdk_scroll_direction_get_type "GType gdk_scroll_direction_get_type( void)"
  return(C_TO_XEN_GType(gdk_scroll_direction_get_type()));
}

static XEN gxg_gdk_notify_type_get_type(void)
{
  #define H_gdk_notify_type_get_type "GType gdk_notify_type_get_type( void)"
  return(C_TO_XEN_GType(gdk_notify_type_get_type()));
}

static XEN gxg_gdk_crossing_mode_get_type(void)
{
  #define H_gdk_crossing_mode_get_type "GType gdk_crossing_mode_get_type( void)"
  return(C_TO_XEN_GType(gdk_crossing_mode_get_type()));
}

static XEN gxg_gdk_property_state_get_type(void)
{
  #define H_gdk_property_state_get_type "GType gdk_property_state_get_type( void)"
  return(C_TO_XEN_GType(gdk_property_state_get_type()));
}

static XEN gxg_gdk_window_state_get_type(void)
{
  #define H_gdk_window_state_get_type "GType gdk_window_state_get_type( void)"
  return(C_TO_XEN_GType(gdk_window_state_get_type()));
}

static XEN gxg_gdk_setting_action_get_type(void)
{
  #define H_gdk_setting_action_get_type "GType gdk_setting_action_get_type( void)"
  return(C_TO_XEN_GType(gdk_setting_action_get_type()));
}

static XEN gxg_gdk_font_type_get_type(void)
{
  #define H_gdk_font_type_get_type "GType gdk_font_type_get_type( void)"
  return(C_TO_XEN_GType(gdk_font_type_get_type()));
}

static XEN gxg_gdk_cap_style_get_type(void)
{
  #define H_gdk_cap_style_get_type "GType gdk_cap_style_get_type( void)"
  return(C_TO_XEN_GType(gdk_cap_style_get_type()));
}

static XEN gxg_gdk_fill_get_type(void)
{
  #define H_gdk_fill_get_type "GType gdk_fill_get_type( void)"
  return(C_TO_XEN_GType(gdk_fill_get_type()));
}

static XEN gxg_gdk_function_get_type(void)
{
  #define H_gdk_function_get_type "GType gdk_function_get_type( void)"
  return(C_TO_XEN_GType(gdk_function_get_type()));
}

static XEN gxg_gdk_join_style_get_type(void)
{
  #define H_gdk_join_style_get_type "GType gdk_join_style_get_type( void)"
  return(C_TO_XEN_GType(gdk_join_style_get_type()));
}

static XEN gxg_gdk_line_style_get_type(void)
{
  #define H_gdk_line_style_get_type "GType gdk_line_style_get_type( void)"
  return(C_TO_XEN_GType(gdk_line_style_get_type()));
}

static XEN gxg_gdk_subwindow_mode_get_type(void)
{
  #define H_gdk_subwindow_mode_get_type "GType gdk_subwindow_mode_get_type( void)"
  return(C_TO_XEN_GType(gdk_subwindow_mode_get_type()));
}

static XEN gxg_gdk_gc_values_mask_get_type(void)
{
  #define H_gdk_gc_values_mask_get_type "GType gdk_gc_values_mask_get_type( void)"
  return(C_TO_XEN_GType(gdk_gc_values_mask_get_type()));
}

static XEN gxg_gdk_image_type_get_type(void)
{
  #define H_gdk_image_type_get_type "GType gdk_image_type_get_type( void)"
  return(C_TO_XEN_GType(gdk_image_type_get_type()));
}

static XEN gxg_gdk_extension_mode_get_type(void)
{
  #define H_gdk_extension_mode_get_type "GType gdk_extension_mode_get_type( void)"
  return(C_TO_XEN_GType(gdk_extension_mode_get_type()));
}

static XEN gxg_gdk_input_source_get_type(void)
{
  #define H_gdk_input_source_get_type "GType gdk_input_source_get_type( void)"
  return(C_TO_XEN_GType(gdk_input_source_get_type()));
}

static XEN gxg_gdk_input_mode_get_type(void)
{
  #define H_gdk_input_mode_get_type "GType gdk_input_mode_get_type( void)"
  return(C_TO_XEN_GType(gdk_input_mode_get_type()));
}

static XEN gxg_gdk_axis_use_get_type(void)
{
  #define H_gdk_axis_use_get_type "GType gdk_axis_use_get_type( void)"
  return(C_TO_XEN_GType(gdk_axis_use_get_type()));
}

static XEN gxg_gdk_prop_mode_get_type(void)
{
  #define H_gdk_prop_mode_get_type "GType gdk_prop_mode_get_type( void)"
  return(C_TO_XEN_GType(gdk_prop_mode_get_type()));
}

static XEN gxg_gdk_fill_rule_get_type(void)
{
  #define H_gdk_fill_rule_get_type "GType gdk_fill_rule_get_type( void)"
  return(C_TO_XEN_GType(gdk_fill_rule_get_type()));
}

static XEN gxg_gdk_overlap_type_get_type(void)
{
  #define H_gdk_overlap_type_get_type "GType gdk_overlap_type_get_type( void)"
  return(C_TO_XEN_GType(gdk_overlap_type_get_type()));
}

static XEN gxg_gdk_rgb_dither_get_type(void)
{
  #define H_gdk_rgb_dither_get_type "GType gdk_rgb_dither_get_type( void)"
  return(C_TO_XEN_GType(gdk_rgb_dither_get_type()));
}

static XEN gxg_gdk_byte_order_get_type(void)
{
  #define H_gdk_byte_order_get_type "GType gdk_byte_order_get_type( void)"
  return(C_TO_XEN_GType(gdk_byte_order_get_type()));
}

static XEN gxg_gdk_modifier_type_get_type(void)
{
  #define H_gdk_modifier_type_get_type "GType gdk_modifier_type_get_type( void)"
  return(C_TO_XEN_GType(gdk_modifier_type_get_type()));
}

static XEN gxg_gdk_input_condition_get_type(void)
{
  #define H_gdk_input_condition_get_type "GType gdk_input_condition_get_type( void)"
  return(C_TO_XEN_GType(gdk_input_condition_get_type()));
}

static XEN gxg_gdk_status_get_type(void)
{
  #define H_gdk_status_get_type "GType gdk_status_get_type( void)"
  return(C_TO_XEN_GType(gdk_status_get_type()));
}

static XEN gxg_gdk_grab_status_get_type(void)
{
  #define H_gdk_grab_status_get_type "GType gdk_grab_status_get_type( void)"
  return(C_TO_XEN_GType(gdk_grab_status_get_type()));
}

static XEN gxg_gdk_visual_type_get_type(void)
{
  #define H_gdk_visual_type_get_type "GType gdk_visual_type_get_type( void)"
  return(C_TO_XEN_GType(gdk_visual_type_get_type()));
}

static XEN gxg_gdk_window_class_get_type(void)
{
  #define H_gdk_window_class_get_type "GType gdk_window_class_get_type( void)"
  return(C_TO_XEN_GType(gdk_window_class_get_type()));
}

static XEN gxg_gdk_window_type_get_type(void)
{
  #define H_gdk_window_type_get_type "GType gdk_window_type_get_type( void)"
  return(C_TO_XEN_GType(gdk_window_type_get_type()));
}

static XEN gxg_gdk_window_attributes_type_get_type(void)
{
  #define H_gdk_window_attributes_type_get_type "GType gdk_window_attributes_type_get_type( void)"
  return(C_TO_XEN_GType(gdk_window_attributes_type_get_type()));
}

static XEN gxg_gdk_window_hints_get_type(void)
{
  #define H_gdk_window_hints_get_type "GType gdk_window_hints_get_type( void)"
  return(C_TO_XEN_GType(gdk_window_hints_get_type()));
}

static XEN gxg_gdk_window_type_hint_get_type(void)
{
  #define H_gdk_window_type_hint_get_type "GType gdk_window_type_hint_get_type( void)"
  return(C_TO_XEN_GType(gdk_window_type_hint_get_type()));
}

static XEN gxg_gdk_wm_decoration_get_type(void)
{
  #define H_gdk_wm_decoration_get_type "GType gdk_wm_decoration_get_type( void)"
  return(C_TO_XEN_GType(gdk_wm_decoration_get_type()));
}

static XEN gxg_gdk_wm_function_get_type(void)
{
  #define H_gdk_wm_function_get_type "GType gdk_wm_function_get_type( void)"
  return(C_TO_XEN_GType(gdk_wm_function_get_type()));
}

static XEN gxg_gdk_gravity_get_type(void)
{
  #define H_gdk_gravity_get_type "GType gdk_gravity_get_type( void)"
  return(C_TO_XEN_GType(gdk_gravity_get_type()));
}

static XEN gxg_gdk_window_edge_get_type(void)
{
  #define H_gdk_window_edge_get_type "GType gdk_window_edge_get_type( void)"
  return(C_TO_XEN_GType(gdk_window_edge_get_type()));
}

static XEN gxg_gdk_event_get_type(void)
{
  #define H_gdk_event_get_type "GType gdk_event_get_type( void)"
  return(C_TO_XEN_GType(gdk_event_get_type()));
}
#endif

static XEN gxg_gdk_events_pending(void)
{
  #define H_gdk_events_pending "gboolean gdk_events_pending( void)"
  return(C_TO_XEN_gboolean(gdk_events_pending()));
}

static XEN gxg_gdk_event_get(void)
{
  #define H_gdk_event_get "GdkEvent* gdk_event_get( void)"
  return(C_TO_XEN_GdkEvent_(gdk_event_get()));
}

static XEN gxg_gdk_event_peek(void)
{
  #define H_gdk_event_peek "GdkEvent* gdk_event_peek( void)"
  return(C_TO_XEN_GdkEvent_(gdk_event_peek()));
}

static XEN gxg_gdk_event_get_graphics_expose(XEN window)
{
  #define H_gdk_event_get_graphics_expose "GdkEvent* gdk_event_get_graphics_expose(GdkWindow* window)"
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 1, "gdk_event_get_graphics_expose", "GdkWindow*");
  return(C_TO_XEN_GdkEvent_(gdk_event_get_graphics_expose(XEN_TO_C_GdkWindow_(window))));
}

static XEN gxg_gdk_event_put(XEN event)
{
  #define H_gdk_event_put "void gdk_event_put(GdkEvent* event)"
  XEN_ASSERT_TYPE(XEN_GdkEvent__P(event), event, 1, "gdk_event_put", "GdkEvent*");
  gdk_event_put(XEN_TO_C_GdkEvent_(event));
  return(XEN_FALSE);
}

static XEN gxg_gdk_event_copy(XEN event)
{
  #define H_gdk_event_copy "GdkEvent* gdk_event_copy(GdkEvent* event)"
  XEN_ASSERT_TYPE(XEN_GdkEvent__P(event), event, 1, "gdk_event_copy", "GdkEvent*");
  return(C_TO_XEN_GdkEvent_(gdk_event_copy(XEN_TO_C_GdkEvent_(event))));
}

static XEN gxg_gdk_event_free(XEN event)
{
  #define H_gdk_event_free "void gdk_event_free(GdkEvent* event)"
  XEN_ASSERT_TYPE(XEN_GdkEvent__P(event), event, 1, "gdk_event_free", "GdkEvent*");
  gdk_event_free(XEN_TO_C_GdkEvent_(event));
  return(XEN_FALSE);
}

static XEN gxg_gdk_event_get_time(XEN event)
{
  #define H_gdk_event_get_time "guint32 gdk_event_get_time(GdkEvent* event)"
  XEN_ASSERT_TYPE(XEN_GdkEvent__P(event) || XEN_FALSE_P(event), event, 1, "gdk_event_get_time", "GdkEvent*");
  return(C_TO_XEN_guint32(gdk_event_get_time(XEN_TO_C_GdkEvent_(event))));
}

#if (!HAVE_GTK_1)
static XEN gxg_gdk_event_get_state(XEN event, XEN state)
{
  #define H_gdk_event_get_state "gboolean gdk_event_get_state(GdkEvent* event, GdkModifierType* state)"
  XEN_ASSERT_TYPE(XEN_GdkEvent__P(event), event, 1, "gdk_event_get_state", "GdkEvent*");
  XEN_ASSERT_TYPE(XEN_GdkModifierType__P(state), state, 2, "gdk_event_get_state", "GdkModifierType*");
  return(C_TO_XEN_gboolean(gdk_event_get_state(XEN_TO_C_GdkEvent_(event), XEN_TO_C_GdkModifierType_(state))));
}

static XEN gxg_gdk_event_get_coords(XEN event, XEN x_win, XEN y_win)
{
  #define H_gdk_event_get_coords "gboolean gdk_event_get_coords(GdkEvent* event, gdouble* [x_win], gdouble* [y_win])"
  gdouble ref_x_win;
  gdouble ref_y_win;
  XEN_ASSERT_TYPE(XEN_GdkEvent__P(event), event, 1, "gdk_event_get_coords", "GdkEvent*");
  {
    XEN result = XEN_FALSE;
    result = C_TO_XEN_gboolean(gdk_event_get_coords(XEN_TO_C_GdkEvent_(event), &ref_x_win, &ref_y_win));
    return(XEN_LIST_3(result, C_TO_XEN_gdouble(ref_x_win), C_TO_XEN_gdouble(ref_y_win)));
   }
}

static XEN gxg_gdk_event_get_root_coords(XEN event, XEN x_root, XEN y_root)
{
  #define H_gdk_event_get_root_coords "gboolean gdk_event_get_root_coords(GdkEvent* event, gdouble* x_root, \
gdouble* y_root)"
  XEN_ASSERT_TYPE(XEN_GdkEvent__P(event), event, 1, "gdk_event_get_root_coords", "GdkEvent*");
  XEN_ASSERT_TYPE(XEN_gdouble__P(x_root), x_root, 2, "gdk_event_get_root_coords", "gdouble*");
  XEN_ASSERT_TYPE(XEN_gdouble__P(y_root), y_root, 3, "gdk_event_get_root_coords", "gdouble*");
  return(C_TO_XEN_gboolean(gdk_event_get_root_coords(XEN_TO_C_GdkEvent_(event), XEN_TO_C_gdouble_(x_root), XEN_TO_C_gdouble_(y_root))));
}

static XEN gxg_gdk_event_get_axis(XEN event, XEN axis_use, XEN value)
{
  #define H_gdk_event_get_axis "gboolean gdk_event_get_axis(GdkEvent* event, GdkAxisUse axis_use, gdouble* [value])"
  gdouble ref_value;
  XEN_ASSERT_TYPE(XEN_GdkEvent__P(event), event, 1, "gdk_event_get_axis", "GdkEvent*");
  XEN_ASSERT_TYPE(XEN_GdkAxisUse_P(axis_use), axis_use, 2, "gdk_event_get_axis", "GdkAxisUse");
  {
    XEN result = XEN_FALSE;
    result = C_TO_XEN_gboolean(gdk_event_get_axis(XEN_TO_C_GdkEvent_(event), XEN_TO_C_GdkAxisUse(axis_use), &ref_value));
    return(XEN_LIST_2(result, C_TO_XEN_gdouble(ref_value)));
   }
}
#endif

static XEN gxg_gdk_event_handler_set(XEN func, XEN func_data, XEN notify)
{
  #define H_gdk_event_handler_set "void gdk_event_handler_set(GdkEventFunc func, lambda_data func_data, \
GDestroyNotify notify)"
  XEN_ASSERT_TYPE(XEN_GdkEventFunc_P(func), func, 1, "gdk_event_handler_set", "GdkEventFunc");
  XEN_ASSERT_TYPE(XEN_lambda_data_P(func_data), func_data, 2, "gdk_event_handler_set", "lambda_data");
  XEN_ASSERT_TYPE(XEN_GDestroyNotify_P(notify), notify, 3, "gdk_event_handler_set", "GDestroyNotify");
  {
    XEN gxg_ptr = XEN_LIST_5(func, func_data, XEN_FALSE, XEN_FALSE, XEN_FALSE);
    xm_protect(gxg_ptr);
    gdk_event_handler_set(XEN_TO_C_GdkEventFunc(func), XEN_TO_C_lambda_data(func_data), XEN_TO_C_GDestroyNotify(notify));
    return(XEN_FALSE);
   }
}

static XEN gxg_gdk_set_show_events(XEN show_events)
{
  #define H_gdk_set_show_events "void gdk_set_show_events(gboolean show_events)"
  XEN_ASSERT_TYPE(XEN_gboolean_P(show_events), show_events, 1, "gdk_set_show_events", "gboolean");
  gdk_set_show_events(XEN_TO_C_gboolean(show_events));
  return(XEN_FALSE);
}

static XEN gxg_gdk_get_show_events(void)
{
  #define H_gdk_get_show_events "gboolean gdk_get_show_events( void)"
  return(C_TO_XEN_gboolean(gdk_get_show_events()));
}

static XEN gxg_gdk_add_client_message_filter(XEN message_type, XEN func, XEN func_data)
{
  #define H_gdk_add_client_message_filter "void gdk_add_client_message_filter(GdkAtom message_type, GdkFilterFunc func, \
lambda_data func_data)"
  XEN_ASSERT_TYPE(XEN_GdkAtom_P(message_type), message_type, 1, "gdk_add_client_message_filter", "GdkAtom");
  XEN_ASSERT_TYPE(XEN_GdkFilterFunc_P(func), func, 2, "gdk_add_client_message_filter", "GdkFilterFunc");
  if (XEN_NOT_BOUND_P(func_data)) func_data = XEN_FALSE; 
  else XEN_ASSERT_TYPE(XEN_lambda_data_P(func_data), func_data, 3, "gdk_add_client_message_filter", "lambda_data");
  {
    XEN gxg_ptr = XEN_LIST_5(func, func_data, XEN_FALSE, XEN_FALSE, XEN_FALSE);
    xm_protect(gxg_ptr);
    gdk_add_client_message_filter(XEN_TO_C_GdkAtom(message_type), XEN_TO_C_GdkFilterFunc(func), XEN_TO_C_lambda_data(func_data));
    return(XEN_FALSE);
   }
}

#if (!HAVE_GTK_1)
static XEN gxg_gdk_setting_get(XEN name, XEN value)
{
  #define H_gdk_setting_get "gboolean gdk_setting_get(gchar* name, GValue* value)"
  XEN_ASSERT_TYPE(XEN_gchar__P(name), name, 1, "gdk_setting_get", "gchar*");
  XEN_ASSERT_TYPE(XEN_GValue__P(value), value, 2, "gdk_setting_get", "GValue*");
  return(C_TO_XEN_gboolean(gdk_setting_get(XEN_TO_C_gchar_(name), XEN_TO_C_GValue_(value))));
}

static XEN gxg_gdk_gc_get_type(void)
{
  #define H_gdk_gc_get_type "GType gdk_gc_get_type( void)"
  return(C_TO_XEN_GType(gdk_gc_get_type()));
}
#endif

static XEN gxg_gdk_gc_new(XEN drawable)
{
  #define H_gdk_gc_new "GdkGC* gdk_gc_new(GdkDrawable* drawable)"
  XEN_ASSERT_TYPE(XEN_GdkDrawable__P(drawable), drawable, 1, "gdk_gc_new", "GdkDrawable*");
  return(C_TO_XEN_GdkGC_(gdk_gc_new(XEN_TO_C_GdkDrawable_(drawable))));
}

static XEN gxg_gdk_gc_new_with_values(XEN drawable, XEN values, XEN values_mask)
{
  #define H_gdk_gc_new_with_values "GdkGC* gdk_gc_new_with_values(GdkDrawable* drawable, GdkGCValues* values, \
GdkGCValuesMask values_mask)"
  XEN_ASSERT_TYPE(XEN_GdkDrawable__P(drawable), drawable, 1, "gdk_gc_new_with_values", "GdkDrawable*");
  XEN_ASSERT_TYPE(XEN_GdkGCValues__P(values), values, 2, "gdk_gc_new_with_values", "GdkGCValues*");
  XEN_ASSERT_TYPE(XEN_GdkGCValuesMask_P(values_mask), values_mask, 3, "gdk_gc_new_with_values", "GdkGCValuesMask");
  return(C_TO_XEN_GdkGC_(gdk_gc_new_with_values(XEN_TO_C_GdkDrawable_(drawable), XEN_TO_C_GdkGCValues_(values), XEN_TO_C_GdkGCValuesMask(values_mask))));
}

static XEN gxg_gdk_gc_get_values(XEN gc, XEN values)
{
  #define H_gdk_gc_get_values "void gdk_gc_get_values(GdkGC* gc, GdkGCValues* values)"
  XEN_ASSERT_TYPE(XEN_GdkGC__P(gc), gc, 1, "gdk_gc_get_values", "GdkGC*");
  XEN_ASSERT_TYPE(XEN_GdkGCValues__P(values), values, 2, "gdk_gc_get_values", "GdkGCValues*");
  gdk_gc_get_values(XEN_TO_C_GdkGC_(gc), XEN_TO_C_GdkGCValues_(values));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gdk_gc_set_values(XEN gc, XEN values, XEN values_mask)
{
  #define H_gdk_gc_set_values "void gdk_gc_set_values(GdkGC* gc, GdkGCValues* values, GdkGCValuesMask values_mask)"
  XEN_ASSERT_TYPE(XEN_GdkGC__P(gc), gc, 1, "gdk_gc_set_values", "GdkGC*");
  XEN_ASSERT_TYPE(XEN_GdkGCValues__P(values), values, 2, "gdk_gc_set_values", "GdkGCValues*");
  XEN_ASSERT_TYPE(XEN_GdkGCValuesMask_P(values_mask), values_mask, 3, "gdk_gc_set_values", "GdkGCValuesMask");
  gdk_gc_set_values(XEN_TO_C_GdkGC_(gc), XEN_TO_C_GdkGCValues_(values), XEN_TO_C_GdkGCValuesMask(values_mask));
  return(XEN_FALSE);
}
#endif

static XEN gxg_gdk_gc_set_foreground(XEN gc, XEN color)
{
  #define H_gdk_gc_set_foreground "void gdk_gc_set_foreground(GdkGC* gc, GdkColor* color)"
  XEN_ASSERT_TYPE(XEN_GdkGC__P(gc), gc, 1, "gdk_gc_set_foreground", "GdkGC*");
  XEN_ASSERT_TYPE(XEN_GdkColor__P(color), color, 2, "gdk_gc_set_foreground", "GdkColor*");
  gdk_gc_set_foreground(XEN_TO_C_GdkGC_(gc), XEN_TO_C_GdkColor_(color));
  return(XEN_FALSE);
}

static XEN gxg_gdk_gc_set_background(XEN gc, XEN color)
{
  #define H_gdk_gc_set_background "void gdk_gc_set_background(GdkGC* gc, GdkColor* color)"
  XEN_ASSERT_TYPE(XEN_GdkGC__P(gc), gc, 1, "gdk_gc_set_background", "GdkGC*");
  XEN_ASSERT_TYPE(XEN_GdkColor__P(color), color, 2, "gdk_gc_set_background", "GdkColor*");
  gdk_gc_set_background(XEN_TO_C_GdkGC_(gc), XEN_TO_C_GdkColor_(color));
  return(XEN_FALSE);
}

static XEN gxg_gdk_gc_set_function(XEN gc, XEN function)
{
  #define H_gdk_gc_set_function "void gdk_gc_set_function(GdkGC* gc, GdkFunction function)"
  XEN_ASSERT_TYPE(XEN_GdkGC__P(gc), gc, 1, "gdk_gc_set_function", "GdkGC*");
  XEN_ASSERT_TYPE(XEN_GdkFunction_P(function), function, 2, "gdk_gc_set_function", "GdkFunction");
  gdk_gc_set_function(XEN_TO_C_GdkGC_(gc), XEN_TO_C_GdkFunction(function));
  return(XEN_FALSE);
}

static XEN gxg_gdk_gc_set_fill(XEN gc, XEN fill)
{
  #define H_gdk_gc_set_fill "void gdk_gc_set_fill(GdkGC* gc, GdkFill fill)"
  XEN_ASSERT_TYPE(XEN_GdkGC__P(gc), gc, 1, "gdk_gc_set_fill", "GdkGC*");
  XEN_ASSERT_TYPE(XEN_GdkFill_P(fill), fill, 2, "gdk_gc_set_fill", "GdkFill");
  gdk_gc_set_fill(XEN_TO_C_GdkGC_(gc), XEN_TO_C_GdkFill(fill));
  return(XEN_FALSE);
}

static XEN gxg_gdk_gc_set_tile(XEN gc, XEN tile)
{
  #define H_gdk_gc_set_tile "void gdk_gc_set_tile(GdkGC* gc, GdkPixmap* tile)"
  XEN_ASSERT_TYPE(XEN_GdkGC__P(gc), gc, 1, "gdk_gc_set_tile", "GdkGC*");
  XEN_ASSERT_TYPE(XEN_GdkPixmap__P(tile), tile, 2, "gdk_gc_set_tile", "GdkPixmap*");
  gdk_gc_set_tile(XEN_TO_C_GdkGC_(gc), XEN_TO_C_GdkPixmap_(tile));
  return(XEN_FALSE);
}

static XEN gxg_gdk_gc_set_stipple(XEN gc, XEN stipple)
{
  #define H_gdk_gc_set_stipple "void gdk_gc_set_stipple(GdkGC* gc, GdkPixmap* stipple)"
  XEN_ASSERT_TYPE(XEN_GdkGC__P(gc), gc, 1, "gdk_gc_set_stipple", "GdkGC*");
  XEN_ASSERT_TYPE(XEN_GdkPixmap__P(stipple), stipple, 2, "gdk_gc_set_stipple", "GdkPixmap*");
  gdk_gc_set_stipple(XEN_TO_C_GdkGC_(gc), XEN_TO_C_GdkPixmap_(stipple));
  return(XEN_FALSE);
}

static XEN gxg_gdk_gc_set_ts_origin(XEN gc, XEN x, XEN y)
{
  #define H_gdk_gc_set_ts_origin "void gdk_gc_set_ts_origin(GdkGC* gc, gint x, gint y)"
  XEN_ASSERT_TYPE(XEN_GdkGC__P(gc), gc, 1, "gdk_gc_set_ts_origin", "GdkGC*");
  XEN_ASSERT_TYPE(XEN_gint_P(x), x, 2, "gdk_gc_set_ts_origin", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(y), y, 3, "gdk_gc_set_ts_origin", "gint");
  gdk_gc_set_ts_origin(XEN_TO_C_GdkGC_(gc), XEN_TO_C_gint(x), XEN_TO_C_gint(y));
  return(XEN_FALSE);
}

static XEN gxg_gdk_gc_set_clip_origin(XEN gc, XEN x, XEN y)
{
  #define H_gdk_gc_set_clip_origin "void gdk_gc_set_clip_origin(GdkGC* gc, gint x, gint y)"
  XEN_ASSERT_TYPE(XEN_GdkGC__P(gc), gc, 1, "gdk_gc_set_clip_origin", "GdkGC*");
  XEN_ASSERT_TYPE(XEN_gint_P(x), x, 2, "gdk_gc_set_clip_origin", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(y), y, 3, "gdk_gc_set_clip_origin", "gint");
  gdk_gc_set_clip_origin(XEN_TO_C_GdkGC_(gc), XEN_TO_C_gint(x), XEN_TO_C_gint(y));
  return(XEN_FALSE);
}

static XEN gxg_gdk_gc_set_clip_mask(XEN gc, XEN mask)
{
  #define H_gdk_gc_set_clip_mask "void gdk_gc_set_clip_mask(GdkGC* gc, GdkBitmap* mask)"
  XEN_ASSERT_TYPE(XEN_GdkGC__P(gc), gc, 1, "gdk_gc_set_clip_mask", "GdkGC*");
  XEN_ASSERT_TYPE(XEN_GdkBitmap__P(mask), mask, 2, "gdk_gc_set_clip_mask", "GdkBitmap*");
  gdk_gc_set_clip_mask(XEN_TO_C_GdkGC_(gc), XEN_TO_C_GdkBitmap_(mask));
  return(XEN_FALSE);
}

static XEN gxg_gdk_gc_set_clip_rectangle(XEN gc, XEN rectangle)
{
  #define H_gdk_gc_set_clip_rectangle "void gdk_gc_set_clip_rectangle(GdkGC* gc, GdkRectangle* rectangle)"
  XEN_ASSERT_TYPE(XEN_GdkGC__P(gc), gc, 1, "gdk_gc_set_clip_rectangle", "GdkGC*");
  XEN_ASSERT_TYPE(XEN_GdkRectangle__P(rectangle), rectangle, 2, "gdk_gc_set_clip_rectangle", "GdkRectangle*");
  gdk_gc_set_clip_rectangle(XEN_TO_C_GdkGC_(gc), XEN_TO_C_GdkRectangle_(rectangle));
  return(XEN_FALSE);
}

static XEN gxg_gdk_gc_set_clip_region(XEN gc, XEN region)
{
  #define H_gdk_gc_set_clip_region "void gdk_gc_set_clip_region(GdkGC* gc, GdkRegion* region)"
  XEN_ASSERT_TYPE(XEN_GdkGC__P(gc), gc, 1, "gdk_gc_set_clip_region", "GdkGC*");
  XEN_ASSERT_TYPE(XEN_GdkRegion__P(region), region, 2, "gdk_gc_set_clip_region", "GdkRegion*");
  gdk_gc_set_clip_region(XEN_TO_C_GdkGC_(gc), XEN_TO_C_GdkRegion_(region));
  return(XEN_FALSE);
}

static XEN gxg_gdk_gc_set_subwindow(XEN gc, XEN mode)
{
  #define H_gdk_gc_set_subwindow "void gdk_gc_set_subwindow(GdkGC* gc, GdkSubwindowMode mode)"
  XEN_ASSERT_TYPE(XEN_GdkGC__P(gc), gc, 1, "gdk_gc_set_subwindow", "GdkGC*");
  XEN_ASSERT_TYPE(XEN_GdkSubwindowMode_P(mode), mode, 2, "gdk_gc_set_subwindow", "GdkSubwindowMode");
  gdk_gc_set_subwindow(XEN_TO_C_GdkGC_(gc), XEN_TO_C_GdkSubwindowMode(mode));
  return(XEN_FALSE);
}

static XEN gxg_gdk_gc_set_exposures(XEN gc, XEN exposures)
{
  #define H_gdk_gc_set_exposures "void gdk_gc_set_exposures(GdkGC* gc, gboolean exposures)"
  XEN_ASSERT_TYPE(XEN_GdkGC__P(gc), gc, 1, "gdk_gc_set_exposures", "GdkGC*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(exposures), exposures, 2, "gdk_gc_set_exposures", "gboolean");
  gdk_gc_set_exposures(XEN_TO_C_GdkGC_(gc), XEN_TO_C_gboolean(exposures));
  return(XEN_FALSE);
}

static XEN gxg_gdk_gc_set_line_attributes(XEN gc, XEN line_width, XEN line_style, XEN cap_style, XEN join_style)
{
  #define H_gdk_gc_set_line_attributes "void gdk_gc_set_line_attributes(GdkGC* gc, gint line_width, GdkLineStyle line_style, \
GdkCapStyle cap_style, GdkJoinStyle join_style)"
  XEN_ASSERT_TYPE(XEN_GdkGC__P(gc), gc, 1, "gdk_gc_set_line_attributes", "GdkGC*");
  XEN_ASSERT_TYPE(XEN_gint_P(line_width), line_width, 2, "gdk_gc_set_line_attributes", "gint");
  XEN_ASSERT_TYPE(XEN_GdkLineStyle_P(line_style), line_style, 3, "gdk_gc_set_line_attributes", "GdkLineStyle");
  XEN_ASSERT_TYPE(XEN_GdkCapStyle_P(cap_style), cap_style, 4, "gdk_gc_set_line_attributes", "GdkCapStyle");
  XEN_ASSERT_TYPE(XEN_GdkJoinStyle_P(join_style), join_style, 5, "gdk_gc_set_line_attributes", "GdkJoinStyle");
  gdk_gc_set_line_attributes(XEN_TO_C_GdkGC_(gc), XEN_TO_C_gint(line_width), XEN_TO_C_GdkLineStyle(line_style), XEN_TO_C_GdkCapStyle(cap_style), 
                             XEN_TO_C_GdkJoinStyle(join_style));
  return(XEN_FALSE);
}

static XEN gxg_gdk_gc_set_dashes(XEN gc, XEN dash_offset, XEN dash_list, XEN n)
{
  #define H_gdk_gc_set_dashes "void gdk_gc_set_dashes(GdkGC* gc, gint dash_offset, gint8* dash_list, \
gint n)"
  XEN_ASSERT_TYPE(XEN_GdkGC__P(gc), gc, 1, "gdk_gc_set_dashes", "GdkGC*");
  XEN_ASSERT_TYPE(XEN_gint_P(dash_offset), dash_offset, 2, "gdk_gc_set_dashes", "gint");
  XEN_ASSERT_TYPE(XEN_gint8__P(dash_list), dash_list, 3, "gdk_gc_set_dashes", "gint8*");
  XEN_ASSERT_TYPE(XEN_gint_P(n), n, 4, "gdk_gc_set_dashes", "gint");
  gdk_gc_set_dashes(XEN_TO_C_GdkGC_(gc), XEN_TO_C_gint(dash_offset), XEN_TO_C_gint8_(dash_list), XEN_TO_C_gint(n));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gdk_gc_offset(XEN gc, XEN x_offset, XEN y_offset)
{
  #define H_gdk_gc_offset "void gdk_gc_offset(GdkGC* gc, gint x_offset, gint y_offset)"
  XEN_ASSERT_TYPE(XEN_GdkGC__P(gc), gc, 1, "gdk_gc_offset", "GdkGC*");
  XEN_ASSERT_TYPE(XEN_gint_P(x_offset), x_offset, 2, "gdk_gc_offset", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(y_offset), y_offset, 3, "gdk_gc_offset", "gint");
  gdk_gc_offset(XEN_TO_C_GdkGC_(gc), XEN_TO_C_gint(x_offset), XEN_TO_C_gint(y_offset));
  return(XEN_FALSE);
}
#endif

static XEN gxg_gdk_gc_copy(XEN dst_gc, XEN src_gc)
{
  #define H_gdk_gc_copy "void gdk_gc_copy(GdkGC* dst_gc, GdkGC* src_gc)"
  XEN_ASSERT_TYPE(XEN_GdkGC__P(dst_gc), dst_gc, 1, "gdk_gc_copy", "GdkGC*");
  XEN_ASSERT_TYPE(XEN_GdkGC__P(src_gc), src_gc, 2, "gdk_gc_copy", "GdkGC*");
  gdk_gc_copy(XEN_TO_C_GdkGC_(dst_gc), XEN_TO_C_GdkGC_(src_gc));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gdk_gc_set_colormap(XEN gc, XEN colormap)
{
  #define H_gdk_gc_set_colormap "void gdk_gc_set_colormap(GdkGC* gc, GdkColormap* colormap)"
  XEN_ASSERT_TYPE(XEN_GdkGC__P(gc), gc, 1, "gdk_gc_set_colormap", "GdkGC*");
  XEN_ASSERT_TYPE(XEN_GdkColormap__P(colormap), colormap, 2, "gdk_gc_set_colormap", "GdkColormap*");
  gdk_gc_set_colormap(XEN_TO_C_GdkGC_(gc), XEN_TO_C_GdkColormap_(colormap));
  return(XEN_FALSE);
}

static XEN gxg_gdk_gc_get_colormap(XEN gc)
{
  #define H_gdk_gc_get_colormap "GdkColormap* gdk_gc_get_colormap(GdkGC* gc)"
  XEN_ASSERT_TYPE(XEN_GdkGC__P(gc), gc, 1, "gdk_gc_get_colormap", "GdkGC*");
  return(C_TO_XEN_GdkColormap_(gdk_gc_get_colormap(XEN_TO_C_GdkGC_(gc))));
}

static XEN gxg_gdk_gc_set_rgb_fg_color(XEN gc, XEN color)
{
  #define H_gdk_gc_set_rgb_fg_color "void gdk_gc_set_rgb_fg_color(GdkGC* gc, GdkColor* color)"
  XEN_ASSERT_TYPE(XEN_GdkGC__P(gc), gc, 1, "gdk_gc_set_rgb_fg_color", "GdkGC*");
  XEN_ASSERT_TYPE(XEN_GdkColor__P(color), color, 2, "gdk_gc_set_rgb_fg_color", "GdkColor*");
  gdk_gc_set_rgb_fg_color(XEN_TO_C_GdkGC_(gc), XEN_TO_C_GdkColor_(color));
  return(XEN_FALSE);
}

static XEN gxg_gdk_gc_set_rgb_bg_color(XEN gc, XEN color)
{
  #define H_gdk_gc_set_rgb_bg_color "void gdk_gc_set_rgb_bg_color(GdkGC* gc, GdkColor* color)"
  XEN_ASSERT_TYPE(XEN_GdkGC__P(gc), gc, 1, "gdk_gc_set_rgb_bg_color", "GdkGC*");
  XEN_ASSERT_TYPE(XEN_GdkColor__P(color), color, 2, "gdk_gc_set_rgb_bg_color", "GdkColor*");
  gdk_gc_set_rgb_bg_color(XEN_TO_C_GdkGC_(gc), XEN_TO_C_GdkColor_(color));
  return(XEN_FALSE);
}
#endif

static XEN gxg_gdk_init(XEN argc, XEN argv)
{
  #define H_gdk_init "void gdk_init(gint* argc, gchar*** argv)"
  XEN_ASSERT_TYPE(XEN_gint__P(argc), argc, 1, "gdk_init", "gint*");
  XEN_ASSERT_TYPE(XEN_gchar____P(argv), argv, 2, "gdk_init", "gchar***");
  gdk_init(XEN_TO_C_gint_(argc), XEN_TO_C_gchar___(argv));
  return(XEN_FALSE);
}

static XEN gxg_gdk_init_check(XEN argc, XEN argv)
{
  #define H_gdk_init_check "gboolean gdk_init_check(gint* argc, gchar*** argv)"
  XEN_ASSERT_TYPE(XEN_gint__P(argc), argc, 1, "gdk_init_check", "gint*");
  XEN_ASSERT_TYPE(XEN_gchar____P(argv), argv, 2, "gdk_init_check", "gchar***");
  return(C_TO_XEN_gboolean(gdk_init_check(XEN_TO_C_gint_(argc), XEN_TO_C_gchar___(argv))));
}

static XEN gxg_gdk_set_locale(void)
{
  #define H_gdk_set_locale "gchar* gdk_set_locale( void)"
  return(C_TO_XEN_gchar_(gdk_set_locale()));
}

#if (!HAVE_GTK_1)
static XEN gxg_gdk_get_program_class(void)
{
  #define H_gdk_get_program_class "char* gdk_get_program_class( void)"
  return(C_TO_XEN_char_(gdk_get_program_class()));
}

static XEN gxg_gdk_set_program_class(XEN program_class)
{
  #define H_gdk_set_program_class "void gdk_set_program_class(char* program_class)"
  XEN_ASSERT_TYPE(XEN_char__P(program_class), program_class, 1, "gdk_set_program_class", "char*");
  gdk_set_program_class(XEN_TO_C_char_(program_class));
  return(XEN_FALSE);
}
#endif

static XEN gxg_gdk_error_trap_push(void)
{
  #define H_gdk_error_trap_push "void gdk_error_trap_push( void)"
  gdk_error_trap_push();
  return(XEN_FALSE);
}

static XEN gxg_gdk_error_trap_pop(void)
{
  #define H_gdk_error_trap_pop "gint gdk_error_trap_pop( void)"
  return(C_TO_XEN_gint(gdk_error_trap_pop()));
}

static XEN gxg_gdk_get_display(void)
{
  #define H_gdk_get_display "gchar* gdk_get_display( void)"
  return(C_TO_XEN_gchar_(gdk_get_display()));
}

static XEN gxg_gdk_pointer_grab(XEN window, XEN owner_events, XEN event_mask, XEN confine_to, XEN cursor, XEN time)
{
  #define H_gdk_pointer_grab "GdkGrabStatus gdk_pointer_grab(GdkWindow* window, gboolean owner_events, \
GdkEventMask event_mask, GdkWindow* confine_to, GdkCursor* cursor, guint32 time)"
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 1, "gdk_pointer_grab", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(owner_events), owner_events, 2, "gdk_pointer_grab", "gboolean");
  XEN_ASSERT_TYPE(XEN_GdkEventMask_P(event_mask), event_mask, 3, "gdk_pointer_grab", "GdkEventMask");
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(confine_to), confine_to, 4, "gdk_pointer_grab", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_GdkCursor__P(cursor), cursor, 5, "gdk_pointer_grab", "GdkCursor*");
  XEN_ASSERT_TYPE(XEN_guint32_P(time), time, 6, "gdk_pointer_grab", "guint32");
  return(C_TO_XEN_GdkGrabStatus(gdk_pointer_grab(XEN_TO_C_GdkWindow_(window), XEN_TO_C_gboolean(owner_events), XEN_TO_C_GdkEventMask(event_mask), 
                                                 XEN_TO_C_GdkWindow_(confine_to), XEN_TO_C_GdkCursor_(cursor), XEN_TO_C_guint32(time))));
}

static XEN gxg_gdk_pointer_ungrab(XEN time)
{
  #define H_gdk_pointer_ungrab "void gdk_pointer_ungrab(guint32 time)"
  XEN_ASSERT_TYPE(XEN_guint32_P(time), time, 1, "gdk_pointer_ungrab", "guint32");
  gdk_pointer_ungrab(XEN_TO_C_guint32(time));
  return(XEN_FALSE);
}

static XEN gxg_gdk_keyboard_grab(XEN window, XEN owner_events, XEN time)
{
  #define H_gdk_keyboard_grab "GdkGrabStatus gdk_keyboard_grab(GdkWindow* window, gboolean owner_events, \
guint32 time)"
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 1, "gdk_keyboard_grab", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(owner_events), owner_events, 2, "gdk_keyboard_grab", "gboolean");
  XEN_ASSERT_TYPE(XEN_guint32_P(time), time, 3, "gdk_keyboard_grab", "guint32");
  return(C_TO_XEN_GdkGrabStatus(gdk_keyboard_grab(XEN_TO_C_GdkWindow_(window), XEN_TO_C_gboolean(owner_events), XEN_TO_C_guint32(time))));
}

static XEN gxg_gdk_keyboard_ungrab(XEN time)
{
  #define H_gdk_keyboard_ungrab "void gdk_keyboard_ungrab(guint32 time)"
  XEN_ASSERT_TYPE(XEN_guint32_P(time), time, 1, "gdk_keyboard_ungrab", "guint32");
  gdk_keyboard_ungrab(XEN_TO_C_guint32(time));
  return(XEN_FALSE);
}

static XEN gxg_gdk_pointer_is_grabbed(void)
{
  #define H_gdk_pointer_is_grabbed "gboolean gdk_pointer_is_grabbed( void)"
  return(C_TO_XEN_gboolean(gdk_pointer_is_grabbed()));
}

static XEN gxg_gdk_screen_width(void)
{
  #define H_gdk_screen_width "gint gdk_screen_width( void)"
  return(C_TO_XEN_gint(gdk_screen_width()));
}

static XEN gxg_gdk_screen_height(void)
{
  #define H_gdk_screen_height "gint gdk_screen_height( void)"
  return(C_TO_XEN_gint(gdk_screen_height()));
}

static XEN gxg_gdk_screen_width_mm(void)
{
  #define H_gdk_screen_width_mm "gint gdk_screen_width_mm( void)"
  return(C_TO_XEN_gint(gdk_screen_width_mm()));
}

static XEN gxg_gdk_screen_height_mm(void)
{
  #define H_gdk_screen_height_mm "gint gdk_screen_height_mm( void)"
  return(C_TO_XEN_gint(gdk_screen_height_mm()));
}

static XEN gxg_gdk_flush(void)
{
  #define H_gdk_flush "void gdk_flush( void)"
  gdk_flush();
  return(XEN_FALSE);
}

static XEN gxg_gdk_beep(void)
{
  #define H_gdk_beep "void gdk_beep( void)"
  gdk_beep();
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gdk_set_double_click_time(XEN msec)
{
  #define H_gdk_set_double_click_time "void gdk_set_double_click_time(guint msec)"
  XEN_ASSERT_TYPE(XEN_guint_P(msec), msec, 1, "gdk_set_double_click_time", "guint");
  gdk_set_double_click_time(XEN_TO_C_guint(msec));
  return(XEN_FALSE);
}
#endif

static XEN gxg_gdk_rectangle_intersect(XEN src1, XEN src2, XEN dest)
{
  #define H_gdk_rectangle_intersect "gboolean gdk_rectangle_intersect(GdkRectangle* src1, GdkRectangle* src2, \
GdkRectangle* dest)"
  XEN_ASSERT_TYPE(XEN_GdkRectangle__P(src1), src1, 1, "gdk_rectangle_intersect", "GdkRectangle*");
  XEN_ASSERT_TYPE(XEN_GdkRectangle__P(src2), src2, 2, "gdk_rectangle_intersect", "GdkRectangle*");
  XEN_ASSERT_TYPE(XEN_GdkRectangle__P(dest), dest, 3, "gdk_rectangle_intersect", "GdkRectangle*");
  return(C_TO_XEN_gboolean(gdk_rectangle_intersect(XEN_TO_C_GdkRectangle_(src1), XEN_TO_C_GdkRectangle_(src2), XEN_TO_C_GdkRectangle_(dest))));
}

static XEN gxg_gdk_rectangle_union(XEN src1, XEN src2, XEN dest)
{
  #define H_gdk_rectangle_union "void gdk_rectangle_union(GdkRectangle* src1, GdkRectangle* src2, GdkRectangle* dest)"
  XEN_ASSERT_TYPE(XEN_GdkRectangle__P(src1), src1, 1, "gdk_rectangle_union", "GdkRectangle*");
  XEN_ASSERT_TYPE(XEN_GdkRectangle__P(src2), src2, 2, "gdk_rectangle_union", "GdkRectangle*");
  XEN_ASSERT_TYPE(XEN_GdkRectangle__P(dest), dest, 3, "gdk_rectangle_union", "GdkRectangle*");
  gdk_rectangle_union(XEN_TO_C_GdkRectangle_(src1), XEN_TO_C_GdkRectangle_(src2), XEN_TO_C_GdkRectangle_(dest));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gdk_rectangle_get_type(void)
{
  #define H_gdk_rectangle_get_type "GType gdk_rectangle_get_type( void)"
  return(C_TO_XEN_GType(gdk_rectangle_get_type()));
}
#endif

static XEN gxg_gdk_wcstombs(XEN src)
{
  #define H_gdk_wcstombs "gchar* gdk_wcstombs(GdkWChar* src)"
  XEN_ASSERT_TYPE(XEN_GdkWChar__P(src), src, 1, "gdk_wcstombs", "GdkWChar*");
  return(C_TO_XEN_gchar_(gdk_wcstombs(XEN_TO_C_GdkWChar_(src))));
}

static XEN gxg_gdk_mbstowcs(XEN dest, XEN src, XEN dest_max)
{
  #define H_gdk_mbstowcs "gint gdk_mbstowcs(GdkWChar* dest, gchar* src, gint dest_max)"
  XEN_ASSERT_TYPE(XEN_GdkWChar__P(dest), dest, 1, "gdk_mbstowcs", "GdkWChar*");
  XEN_ASSERT_TYPE(XEN_gchar__P(src), src, 2, "gdk_mbstowcs", "gchar*");
  XEN_ASSERT_TYPE(XEN_gint_P(dest_max), dest_max, 3, "gdk_mbstowcs", "gint");
  return(C_TO_XEN_gint(gdk_mbstowcs(XEN_TO_C_GdkWChar_(dest), XEN_TO_C_gchar_(src), XEN_TO_C_gint(dest_max))));
}

static XEN gxg_gdk_event_send_clientmessage_toall(XEN event)
{
  #define H_gdk_event_send_clientmessage_toall "void gdk_event_send_clientmessage_toall(GdkEvent* event)"
  XEN_ASSERT_TYPE(XEN_GdkEvent__P(event), event, 1, "gdk_event_send_clientmessage_toall", "GdkEvent*");
  gdk_event_send_clientmessage_toall(XEN_TO_C_GdkEvent_(event));
  return(XEN_FALSE);
}

static XEN gxg_gdk_event_send_client_message(XEN event, XEN xid)
{
  #define H_gdk_event_send_client_message "gboolean gdk_event_send_client_message(GdkEvent* event, guint32 xid)"
  XEN_ASSERT_TYPE(XEN_GdkEvent__P(event), event, 1, "gdk_event_send_client_message", "GdkEvent*");
  XEN_ASSERT_TYPE(XEN_guint32_P(xid), xid, 2, "gdk_event_send_client_message", "guint32");
  return(C_TO_XEN_gboolean(gdk_event_send_client_message(XEN_TO_C_GdkEvent_(event), XEN_TO_C_guint32(xid))));
}

static XEN gxg_gdk_threads_enter(void)
{
  #define H_gdk_threads_enter "void gdk_threads_enter( void)"
  gdk_threads_enter();
  return(XEN_FALSE);
}

static XEN gxg_gdk_threads_leave(void)
{
  #define H_gdk_threads_leave "void gdk_threads_leave( void)"
  gdk_threads_leave();
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gdk_threads_init(void)
{
  #define H_gdk_threads_init "void gdk_threads_init( void)"
  gdk_threads_init();
  return(XEN_FALSE);
}

static XEN gxg_gdk_image_get_type(void)
{
  #define H_gdk_image_get_type "GType gdk_image_get_type( void)"
  return(C_TO_XEN_GType(gdk_image_get_type()));
}
#endif

static XEN gxg_gdk_image_new(XEN type, XEN visual, XEN width, XEN height)
{
  #define H_gdk_image_new "GdkImage* gdk_image_new(GdkImageType type, GdkVisual* visual, gint width, \
gint height)"
  XEN_ASSERT_TYPE(XEN_GdkImageType_P(type), type, 1, "gdk_image_new", "GdkImageType");
  XEN_ASSERT_TYPE(XEN_GdkVisual__P(visual), visual, 2, "gdk_image_new", "GdkVisual*");
  XEN_ASSERT_TYPE(XEN_gint_P(width), width, 3, "gdk_image_new", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(height), height, 4, "gdk_image_new", "gint");
  return(C_TO_XEN_GdkImage_(gdk_image_new(XEN_TO_C_GdkImageType(type), XEN_TO_C_GdkVisual_(visual), XEN_TO_C_gint(width), 
                                          XEN_TO_C_gint(height))));
}

#if (!HAVE_GTK_1)
static XEN gxg_gdk_device_get_type(void)
{
  #define H_gdk_device_get_type "GType gdk_device_get_type( void)"
  return(C_TO_XEN_GType(gdk_device_get_type()));
}

static XEN gxg_gdk_devices_list(void)
{
  #define H_gdk_devices_list "GList* gdk_devices_list( void)"
  return(C_TO_XEN_GList_(gdk_devices_list()));
}

static XEN gxg_gdk_device_set_source(XEN device, XEN source)
{
  #define H_gdk_device_set_source "void gdk_device_set_source(GdkDevice* device, GdkInputSource source)"
  XEN_ASSERT_TYPE(XEN_GdkDevice__P(device), device, 1, "gdk_device_set_source", "GdkDevice*");
  XEN_ASSERT_TYPE(XEN_GdkInputSource_P(source), source, 2, "gdk_device_set_source", "GdkInputSource");
  gdk_device_set_source(XEN_TO_C_GdkDevice_(device), XEN_TO_C_GdkInputSource(source));
  return(XEN_FALSE);
}

static XEN gxg_gdk_device_set_mode(XEN device, XEN mode)
{
  #define H_gdk_device_set_mode "gboolean gdk_device_set_mode(GdkDevice* device, GdkInputMode mode)"
  XEN_ASSERT_TYPE(XEN_GdkDevice__P(device), device, 1, "gdk_device_set_mode", "GdkDevice*");
  XEN_ASSERT_TYPE(XEN_GdkInputMode_P(mode), mode, 2, "gdk_device_set_mode", "GdkInputMode");
  return(C_TO_XEN_gboolean(gdk_device_set_mode(XEN_TO_C_GdkDevice_(device), XEN_TO_C_GdkInputMode(mode))));
}

static XEN gxg_gdk_device_set_key(XEN device, XEN index, XEN keyval, XEN modifiers)
{
  #define H_gdk_device_set_key "void gdk_device_set_key(GdkDevice* device, guint index, guint keyval, \
GdkModifierType modifiers)"
  XEN_ASSERT_TYPE(XEN_GdkDevice__P(device), device, 1, "gdk_device_set_key", "GdkDevice*");
  XEN_ASSERT_TYPE(XEN_guint_P(index), index, 2, "gdk_device_set_key", "guint");
  XEN_ASSERT_TYPE(XEN_guint_P(keyval), keyval, 3, "gdk_device_set_key", "guint");
  XEN_ASSERT_TYPE(XEN_GdkModifierType_P(modifiers), modifiers, 4, "gdk_device_set_key", "GdkModifierType");
  gdk_device_set_key(XEN_TO_C_GdkDevice_(device), XEN_TO_C_guint(index), XEN_TO_C_guint(keyval), XEN_TO_C_GdkModifierType(modifiers));
  return(XEN_FALSE);
}

static XEN gxg_gdk_device_set_axis_use(XEN device, XEN index, XEN use)
{
  #define H_gdk_device_set_axis_use "void gdk_device_set_axis_use(GdkDevice* device, guint index, GdkAxisUse use)"
  XEN_ASSERT_TYPE(XEN_GdkDevice__P(device), device, 1, "gdk_device_set_axis_use", "GdkDevice*");
  XEN_ASSERT_TYPE(XEN_guint_P(index), index, 2, "gdk_device_set_axis_use", "guint");
  XEN_ASSERT_TYPE(XEN_GdkAxisUse_P(use), use, 3, "gdk_device_set_axis_use", "GdkAxisUse");
  gdk_device_set_axis_use(XEN_TO_C_GdkDevice_(device), XEN_TO_C_guint(index), XEN_TO_C_GdkAxisUse(use));
  return(XEN_FALSE);
}

static XEN gxg_gdk_device_get_state(XEN device, XEN window, XEN axes, XEN mask)
{
  #define H_gdk_device_get_state "void gdk_device_get_state(GdkDevice* device, GdkWindow* window, gdouble* axes, \
GdkModifierType* mask)"
  XEN_ASSERT_TYPE(XEN_GdkDevice__P(device), device, 1, "gdk_device_get_state", "GdkDevice*");
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 2, "gdk_device_get_state", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_gdouble__P(axes), axes, 3, "gdk_device_get_state", "gdouble*");
  XEN_ASSERT_TYPE(XEN_GdkModifierType__P(mask), mask, 4, "gdk_device_get_state", "GdkModifierType*");
  gdk_device_get_state(XEN_TO_C_GdkDevice_(device), XEN_TO_C_GdkWindow_(window), XEN_TO_C_gdouble_(axes), XEN_TO_C_GdkModifierType_(mask));
  return(XEN_FALSE);
}

static XEN gxg_gdk_device_get_history(XEN device, XEN window, XEN start, XEN stop, XEN events, XEN n_events)
{
  #define H_gdk_device_get_history "gboolean gdk_device_get_history(GdkDevice* device, GdkWindow* window, \
guint32 start, guint32 stop, GdkTimeCoord*** [events], gint* [n_events])"
  GdkTimeCoord** ref_events;
  gint ref_n_events;
  XEN_ASSERT_TYPE(XEN_GdkDevice__P(device), device, 1, "gdk_device_get_history", "GdkDevice*");
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 2, "gdk_device_get_history", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_guint32_P(start), start, 3, "gdk_device_get_history", "guint32");
  XEN_ASSERT_TYPE(XEN_guint32_P(stop), stop, 4, "gdk_device_get_history", "guint32");
  {
    XEN result = XEN_FALSE;
    result = C_TO_XEN_gboolean(gdk_device_get_history(XEN_TO_C_GdkDevice_(device), XEN_TO_C_GdkWindow_(window), XEN_TO_C_guint32(start), 
                                                      XEN_TO_C_guint32(stop), &ref_events, &ref_n_events));
    return(XEN_LIST_3(result, C_TO_XEN_GdkTimeCoord__(ref_events), C_TO_XEN_gint(ref_n_events)));
   }
}

static XEN gxg_gdk_device_free_history(XEN events, XEN n_events)
{
  #define H_gdk_device_free_history "void gdk_device_free_history(GdkTimeCoord** events, gint n_events)"
  XEN_ASSERT_TYPE(XEN_GdkTimeCoord___P(events), events, 1, "gdk_device_free_history", "GdkTimeCoord**");
  XEN_ASSERT_TYPE(XEN_gint_P(n_events), n_events, 2, "gdk_device_free_history", "gint");
  gdk_device_free_history(XEN_TO_C_GdkTimeCoord__(events), XEN_TO_C_gint(n_events));
  return(XEN_FALSE);
}

static XEN gxg_gdk_device_get_axis(XEN device, XEN axes, XEN use, XEN value)
{
  #define H_gdk_device_get_axis "gboolean gdk_device_get_axis(GdkDevice* device, gdouble* axes, GdkAxisUse use, \
gdouble* value)"
  XEN_ASSERT_TYPE(XEN_GdkDevice__P(device), device, 1, "gdk_device_get_axis", "GdkDevice*");
  XEN_ASSERT_TYPE(XEN_gdouble__P(axes), axes, 2, "gdk_device_get_axis", "gdouble*");
  XEN_ASSERT_TYPE(XEN_GdkAxisUse_P(use), use, 3, "gdk_device_get_axis", "GdkAxisUse");
  XEN_ASSERT_TYPE(XEN_gdouble__P(value), value, 4, "gdk_device_get_axis", "gdouble*");
  return(C_TO_XEN_gboolean(gdk_device_get_axis(XEN_TO_C_GdkDevice_(device), XEN_TO_C_gdouble_(axes), XEN_TO_C_GdkAxisUse(use), 
                                               XEN_TO_C_gdouble_(value))));
}
#endif

static XEN gxg_gdk_input_set_extension_events(XEN window, XEN mask, XEN mode)
{
  #define H_gdk_input_set_extension_events "void gdk_input_set_extension_events(GdkWindow* window, gint mask, \
GdkExtensionMode mode)"
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 1, "gdk_input_set_extension_events", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_gint_P(mask), mask, 2, "gdk_input_set_extension_events", "gint");
  XEN_ASSERT_TYPE(XEN_GdkExtensionMode_P(mode), mode, 3, "gdk_input_set_extension_events", "GdkExtensionMode");
  gdk_input_set_extension_events(XEN_TO_C_GdkWindow_(window), XEN_TO_C_gint(mask), XEN_TO_C_GdkExtensionMode(mode));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gdk_device_get_core_pointer(void)
{
  #define H_gdk_device_get_core_pointer "GdkDevice* gdk_device_get_core_pointer( void)"
  return(C_TO_XEN_GdkDevice_(gdk_device_get_core_pointer()));
}

static XEN gxg_gdk_keymap_get_type(void)
{
  #define H_gdk_keymap_get_type "GType gdk_keymap_get_type( void)"
  return(C_TO_XEN_GType(gdk_keymap_get_type()));
}

static XEN gxg_gdk_keymap_get_default(void)
{
  #define H_gdk_keymap_get_default "GdkKeymap* gdk_keymap_get_default( void)"
  return(C_TO_XEN_GdkKeymap_(gdk_keymap_get_default()));
}

static XEN gxg_gdk_keymap_lookup_key(XEN keymap, XEN key)
{
  #define H_gdk_keymap_lookup_key "guint gdk_keymap_lookup_key(GdkKeymap* keymap, GdkKeymapKey* key)"
  XEN_ASSERT_TYPE(XEN_GdkKeymap__P(keymap), keymap, 1, "gdk_keymap_lookup_key", "GdkKeymap*");
  XEN_ASSERT_TYPE(XEN_GdkKeymapKey__P(key), key, 2, "gdk_keymap_lookup_key", "GdkKeymapKey*");
  return(C_TO_XEN_guint(gdk_keymap_lookup_key(XEN_TO_C_GdkKeymap_(keymap), XEN_TO_C_GdkKeymapKey_(key))));
}

static XEN gxg_gdk_keymap_get_entries_for_keyval(XEN keymap, XEN keyval, XEN keys, XEN n_keys)
{
  #define H_gdk_keymap_get_entries_for_keyval "gboolean gdk_keymap_get_entries_for_keyval(GdkKeymap* keymap, \
guint keyval, GdkKeymapKey** [keys], gint* [n_keys])"
  GdkKeymapKey* ref_keys;
  gint ref_n_keys;
  XEN_ASSERT_TYPE(XEN_GdkKeymap__P(keymap), keymap, 1, "gdk_keymap_get_entries_for_keyval", "GdkKeymap*");
  XEN_ASSERT_TYPE(XEN_guint_P(keyval), keyval, 2, "gdk_keymap_get_entries_for_keyval", "guint");
  {
    XEN result = XEN_FALSE;
    result = C_TO_XEN_gboolean(gdk_keymap_get_entries_for_keyval(XEN_TO_C_GdkKeymap_(keymap), XEN_TO_C_guint(keyval), &ref_keys, 
                                                                 &ref_n_keys));
    return(XEN_LIST_3(result, C_TO_XEN_GdkKeymapKey_(ref_keys), C_TO_XEN_gint(ref_n_keys)));
   }
}

static XEN gxg_gdk_keymap_get_entries_for_keycode(XEN keymap, XEN hardware_keycode, XEN keys, XEN keyvals, XEN n_entries)
{
  #define H_gdk_keymap_get_entries_for_keycode "gboolean gdk_keymap_get_entries_for_keycode(GdkKeymap* keymap, \
guint hardware_keycode, GdkKeymapKey** [keys], guint** [keyvals], gint* [n_entries])"
  GdkKeymapKey* ref_keys;
  guint* ref_keyvals;
  gint ref_n_entries;
  XEN_ASSERT_TYPE(XEN_GdkKeymap__P(keymap), keymap, 1, "gdk_keymap_get_entries_for_keycode", "GdkKeymap*");
  XEN_ASSERT_TYPE(XEN_guint_P(hardware_keycode), hardware_keycode, 2, "gdk_keymap_get_entries_for_keycode", "guint");
  {
    XEN result = XEN_FALSE;
    result = C_TO_XEN_gboolean(gdk_keymap_get_entries_for_keycode(XEN_TO_C_GdkKeymap_(keymap), XEN_TO_C_guint(hardware_keycode), 
                                                                  &ref_keys, &ref_keyvals, &ref_n_entries));
    return(XEN_LIST_4(result, C_TO_XEN_GdkKeymapKey_(ref_keys), C_TO_XEN_guint_(ref_keyvals), C_TO_XEN_gint(ref_n_entries)));
   }
}

static XEN gxg_gdk_keymap_get_direction(XEN keymap)
{
  #define H_gdk_keymap_get_direction "PangoDirection gdk_keymap_get_direction(GdkKeymap* keymap)"
  XEN_ASSERT_TYPE(XEN_GdkKeymap__P(keymap), keymap, 1, "gdk_keymap_get_direction", "GdkKeymap*");
  return(C_TO_XEN_PangoDirection(gdk_keymap_get_direction(XEN_TO_C_GdkKeymap_(keymap))));
}
#endif

static XEN gxg_gdk_keyval_name(XEN keyval)
{
  #define H_gdk_keyval_name "gchar* gdk_keyval_name(guint keyval)"
  XEN_ASSERT_TYPE(XEN_guint_P(keyval), keyval, 1, "gdk_keyval_name", "guint");
  return(C_TO_XEN_gchar_(gdk_keyval_name(XEN_TO_C_guint(keyval))));
}

static XEN gxg_gdk_keyval_from_name(XEN keyval_name)
{
  #define H_gdk_keyval_from_name "guint gdk_keyval_from_name(gchar* keyval_name)"
  XEN_ASSERT_TYPE(XEN_gchar__P(keyval_name), keyval_name, 1, "gdk_keyval_from_name", "gchar*");
  return(C_TO_XEN_guint(gdk_keyval_from_name(XEN_TO_C_gchar_(keyval_name))));
}

#if (!HAVE_GTK_1)
static XEN gxg_gdk_keyval_convert_case(XEN symbol, XEN lower, XEN upper)
{
  #define H_gdk_keyval_convert_case "void gdk_keyval_convert_case(guint symbol, guint* [lower], guint* [upper])"
  guint ref_lower;
  guint ref_upper;
  XEN_ASSERT_TYPE(XEN_guint_P(symbol), symbol, 1, "gdk_keyval_convert_case", "guint");
  gdk_keyval_convert_case(XEN_TO_C_guint(symbol), &ref_lower, &ref_upper);
  return(XEN_LIST_2(C_TO_XEN_guint(ref_lower), C_TO_XEN_guint(ref_upper)));
}
#endif

static XEN gxg_gdk_keyval_to_upper(XEN keyval)
{
  #define H_gdk_keyval_to_upper "guint gdk_keyval_to_upper(guint keyval)"
  XEN_ASSERT_TYPE(XEN_guint_P(keyval), keyval, 1, "gdk_keyval_to_upper", "guint");
  return(C_TO_XEN_guint(gdk_keyval_to_upper(XEN_TO_C_guint(keyval))));
}

static XEN gxg_gdk_keyval_to_lower(XEN keyval)
{
  #define H_gdk_keyval_to_lower "guint gdk_keyval_to_lower(guint keyval)"
  XEN_ASSERT_TYPE(XEN_guint_P(keyval), keyval, 1, "gdk_keyval_to_lower", "guint");
  return(C_TO_XEN_guint(gdk_keyval_to_lower(XEN_TO_C_guint(keyval))));
}

static XEN gxg_gdk_keyval_is_upper(XEN keyval)
{
  #define H_gdk_keyval_is_upper "gboolean gdk_keyval_is_upper(guint keyval)"
  XEN_ASSERT_TYPE(XEN_guint_P(keyval), keyval, 1, "gdk_keyval_is_upper", "guint");
  return(C_TO_XEN_gboolean(gdk_keyval_is_upper(XEN_TO_C_guint(keyval))));
}

static XEN gxg_gdk_keyval_is_lower(XEN keyval)
{
  #define H_gdk_keyval_is_lower "gboolean gdk_keyval_is_lower(guint keyval)"
  XEN_ASSERT_TYPE(XEN_guint_P(keyval), keyval, 1, "gdk_keyval_is_lower", "guint");
  return(C_TO_XEN_gboolean(gdk_keyval_is_lower(XEN_TO_C_guint(keyval))));
}

#if (!HAVE_GTK_1)
static XEN gxg_gdk_keyval_to_unicode(XEN keyval)
{
  #define H_gdk_keyval_to_unicode "guint32 gdk_keyval_to_unicode(guint keyval)"
  XEN_ASSERT_TYPE(XEN_guint_P(keyval), keyval, 1, "gdk_keyval_to_unicode", "guint");
  return(C_TO_XEN_guint32(gdk_keyval_to_unicode(XEN_TO_C_guint(keyval))));
}

static XEN gxg_gdk_unicode_to_keyval(XEN wc)
{
  #define H_gdk_unicode_to_keyval "guint gdk_unicode_to_keyval(guint32 wc)"
  XEN_ASSERT_TYPE(XEN_guint32_P(wc), wc, 1, "gdk_unicode_to_keyval", "guint32");
  return(C_TO_XEN_guint(gdk_unicode_to_keyval(XEN_TO_C_guint32(wc))));
}

static XEN gxg_gdk_pango_context_get(void)
{
  #define H_gdk_pango_context_get "PangoContext* gdk_pango_context_get( void)"
  return(C_TO_XEN_PangoContext_(gdk_pango_context_get()));
}

static XEN gxg_gdk_pango_context_set_colormap(XEN context, XEN colormap)
{
  #define H_gdk_pango_context_set_colormap "void gdk_pango_context_set_colormap(PangoContext* context, \
GdkColormap* colormap)"
  XEN_ASSERT_TYPE(XEN_PangoContext__P(context), context, 1, "gdk_pango_context_set_colormap", "PangoContext*");
  XEN_ASSERT_TYPE(XEN_GdkColormap__P(colormap), colormap, 2, "gdk_pango_context_set_colormap", "GdkColormap*");
  gdk_pango_context_set_colormap(XEN_TO_C_PangoContext_(context), XEN_TO_C_GdkColormap_(colormap));
  return(XEN_FALSE);
}

static XEN gxg_gdk_pango_layout_line_get_clip_region(XEN line, XEN x_origin, XEN y_origin, XEN index_ranges, XEN n_ranges)
{
  #define H_gdk_pango_layout_line_get_clip_region "GdkRegion* gdk_pango_layout_line_get_clip_region(PangoLayoutLine* line, \
gint x_origin, gint y_origin, gint* index_ranges, gint n_ranges)"
  XEN_ASSERT_TYPE(XEN_PangoLayoutLine__P(line), line, 1, "gdk_pango_layout_line_get_clip_region", "PangoLayoutLine*");
  XEN_ASSERT_TYPE(XEN_gint_P(x_origin), x_origin, 2, "gdk_pango_layout_line_get_clip_region", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(y_origin), y_origin, 3, "gdk_pango_layout_line_get_clip_region", "gint");
  XEN_ASSERT_TYPE(XEN_gint__P(index_ranges), index_ranges, 4, "gdk_pango_layout_line_get_clip_region", "gint*");
  XEN_ASSERT_TYPE(XEN_gint_P(n_ranges), n_ranges, 5, "gdk_pango_layout_line_get_clip_region", "gint");
  return(C_TO_XEN_GdkRegion_(gdk_pango_layout_line_get_clip_region(XEN_TO_C_PangoLayoutLine_(line), XEN_TO_C_gint(x_origin), 
                                                                   XEN_TO_C_gint(y_origin), XEN_TO_C_gint_(index_ranges), 
                                                                   XEN_TO_C_gint(n_ranges))));
}

static XEN gxg_gdk_pango_layout_get_clip_region(XEN layout, XEN x_origin, XEN y_origin, XEN index_ranges, XEN n_ranges)
{
  #define H_gdk_pango_layout_get_clip_region "GdkRegion* gdk_pango_layout_get_clip_region(PangoLayout* layout, \
gint x_origin, gint y_origin, gint* index_ranges, gint n_ranges)"
  XEN_ASSERT_TYPE(XEN_PangoLayout__P(layout), layout, 1, "gdk_pango_layout_get_clip_region", "PangoLayout*");
  XEN_ASSERT_TYPE(XEN_gint_P(x_origin), x_origin, 2, "gdk_pango_layout_get_clip_region", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(y_origin), y_origin, 3, "gdk_pango_layout_get_clip_region", "gint");
  XEN_ASSERT_TYPE(XEN_gint__P(index_ranges), index_ranges, 4, "gdk_pango_layout_get_clip_region", "gint*");
  XEN_ASSERT_TYPE(XEN_gint_P(n_ranges), n_ranges, 5, "gdk_pango_layout_get_clip_region", "gint");
  return(C_TO_XEN_GdkRegion_(gdk_pango_layout_get_clip_region(XEN_TO_C_PangoLayout_(layout), XEN_TO_C_gint(x_origin), XEN_TO_C_gint(y_origin), 
                                                              XEN_TO_C_gint_(index_ranges), XEN_TO_C_gint(n_ranges))));
}

static XEN gxg_gdk_pango_attr_stipple_new(XEN stipple)
{
  #define H_gdk_pango_attr_stipple_new "PangoAttribute* gdk_pango_attr_stipple_new(GdkBitmap* stipple)"
  XEN_ASSERT_TYPE(XEN_GdkBitmap__P(stipple), stipple, 1, "gdk_pango_attr_stipple_new", "GdkBitmap*");
  return(C_TO_XEN_PangoAttribute_(gdk_pango_attr_stipple_new(XEN_TO_C_GdkBitmap_(stipple))));
}

static XEN gxg_gdk_pango_attr_embossed_new(XEN embossed)
{
  #define H_gdk_pango_attr_embossed_new "PangoAttribute* gdk_pango_attr_embossed_new(gboolean embossed)"
  XEN_ASSERT_TYPE(XEN_gboolean_P(embossed), embossed, 1, "gdk_pango_attr_embossed_new", "gboolean");
  return(C_TO_XEN_PangoAttribute_(gdk_pango_attr_embossed_new(XEN_TO_C_gboolean(embossed))));
}

static XEN gxg_gdk_pixbuf_render_threshold_alpha(XEN pixbuf, XEN bitmap, XEN src_x, XEN src_y, XEN dest_x, XEN dest_y, XEN width, XEN height, XEN alpha_threshold)
{
  #define H_gdk_pixbuf_render_threshold_alpha "void gdk_pixbuf_render_threshold_alpha(GdkPixbuf* pixbuf, \
GdkBitmap* bitmap, int src_x, int src_y, int dest_x, int dest_y, int width, int height, int alpha_threshold)"
  XEN_ASSERT_TYPE(XEN_GdkPixbuf__P(pixbuf), pixbuf, 1, "gdk_pixbuf_render_threshold_alpha", "GdkPixbuf*");
  XEN_ASSERT_TYPE(XEN_GdkBitmap__P(bitmap), bitmap, 2, "gdk_pixbuf_render_threshold_alpha", "GdkBitmap*");
  XEN_ASSERT_TYPE(XEN_int_P(src_x), src_x, 3, "gdk_pixbuf_render_threshold_alpha", "int");
  XEN_ASSERT_TYPE(XEN_int_P(src_y), src_y, 4, "gdk_pixbuf_render_threshold_alpha", "int");
  XEN_ASSERT_TYPE(XEN_int_P(dest_x), dest_x, 5, "gdk_pixbuf_render_threshold_alpha", "int");
  XEN_ASSERT_TYPE(XEN_int_P(dest_y), dest_y, 6, "gdk_pixbuf_render_threshold_alpha", "int");
  XEN_ASSERT_TYPE(XEN_int_P(width), width, 7, "gdk_pixbuf_render_threshold_alpha", "int");
  XEN_ASSERT_TYPE(XEN_int_P(height), height, 8, "gdk_pixbuf_render_threshold_alpha", "int");
  XEN_ASSERT_TYPE(XEN_int_P(alpha_threshold), alpha_threshold, 9, "gdk_pixbuf_render_threshold_alpha", "int");
  gdk_pixbuf_render_threshold_alpha(XEN_TO_C_GdkPixbuf_(pixbuf), XEN_TO_C_GdkBitmap_(bitmap), XEN_TO_C_int(src_x), XEN_TO_C_int(src_y), 
                                    XEN_TO_C_int(dest_x), XEN_TO_C_int(dest_y), XEN_TO_C_int(width), XEN_TO_C_int(height), 
                                    XEN_TO_C_int(alpha_threshold));
  return(XEN_FALSE);
}

static XEN gxg_gdk_pixbuf_render_to_drawable(XEN arglist)
{
  #define H_gdk_pixbuf_render_to_drawable "void gdk_pixbuf_render_to_drawable(GdkPixbuf* pixbuf, GdkDrawable* drawable, \
GdkGC* gc, int src_x, int src_y, int dest_x, int dest_y, int width, int height, GdkRgbDither dither, \
int x_dither, int y_dither)"
  XEN pixbuf, drawable, gc, src_x, src_y, dest_x, dest_y, width, height, dither, x_dither, y_dither;
  pixbuf = XEN_LIST_REF(arglist, 0);
  drawable = XEN_LIST_REF(arglist, 1);
  gc = XEN_LIST_REF(arglist, 2);
  src_x = XEN_LIST_REF(arglist, 3);
  src_y = XEN_LIST_REF(arglist, 4);
  dest_x = XEN_LIST_REF(arglist, 5);
  dest_y = XEN_LIST_REF(arglist, 6);
  width = XEN_LIST_REF(arglist, 7);
  height = XEN_LIST_REF(arglist, 8);
  dither = XEN_LIST_REF(arglist, 9);
  x_dither = XEN_LIST_REF(arglist, 10);
  y_dither = XEN_LIST_REF(arglist, 11);
  XEN_ASSERT_TYPE(XEN_GdkPixbuf__P(pixbuf), pixbuf, 1, "gdk_pixbuf_render_to_drawable", "GdkPixbuf*");
  XEN_ASSERT_TYPE(XEN_GdkDrawable__P(drawable), drawable, 2, "gdk_pixbuf_render_to_drawable", "GdkDrawable*");
  XEN_ASSERT_TYPE(XEN_GdkGC__P(gc), gc, 3, "gdk_pixbuf_render_to_drawable", "GdkGC*");
  XEN_ASSERT_TYPE(XEN_int_P(src_x), src_x, 4, "gdk_pixbuf_render_to_drawable", "int");
  XEN_ASSERT_TYPE(XEN_int_P(src_y), src_y, 5, "gdk_pixbuf_render_to_drawable", "int");
  XEN_ASSERT_TYPE(XEN_int_P(dest_x), dest_x, 6, "gdk_pixbuf_render_to_drawable", "int");
  XEN_ASSERT_TYPE(XEN_int_P(dest_y), dest_y, 7, "gdk_pixbuf_render_to_drawable", "int");
  XEN_ASSERT_TYPE(XEN_int_P(width), width, 8, "gdk_pixbuf_render_to_drawable", "int");
  XEN_ASSERT_TYPE(XEN_int_P(height), height, 9, "gdk_pixbuf_render_to_drawable", "int");
  XEN_ASSERT_TYPE(XEN_GdkRgbDither_P(dither), dither, 10, "gdk_pixbuf_render_to_drawable", "GdkRgbDither");
  XEN_ASSERT_TYPE(XEN_int_P(x_dither), x_dither, 11, "gdk_pixbuf_render_to_drawable", "int");
  XEN_ASSERT_TYPE(XEN_int_P(y_dither), y_dither, 12, "gdk_pixbuf_render_to_drawable", "int");
  gdk_pixbuf_render_to_drawable(XEN_TO_C_GdkPixbuf_(pixbuf), XEN_TO_C_GdkDrawable_(drawable), XEN_TO_C_GdkGC_(gc), XEN_TO_C_int(src_x), 
                                XEN_TO_C_int(src_y), XEN_TO_C_int(dest_x), XEN_TO_C_int(dest_y), XEN_TO_C_int(width), XEN_TO_C_int(height), 
                                XEN_TO_C_GdkRgbDither(dither), XEN_TO_C_int(x_dither), XEN_TO_C_int(y_dither));
  return(XEN_FALSE);
}

static XEN gxg_gdk_pixbuf_render_to_drawable_alpha(XEN arglist)
{
  #define H_gdk_pixbuf_render_to_drawable_alpha "void gdk_pixbuf_render_to_drawable_alpha(GdkPixbuf* pixbuf, \
GdkDrawable* drawable, int src_x, int src_y, int dest_x, int dest_y, int width, int height, GdkPixbufAlphaMode alpha_mode, \
int alpha_threshold, GdkRgbDither dither, int x_dither, int y_dither)"
  XEN pixbuf, drawable, src_x, src_y, dest_x, dest_y, width, height, alpha_mode, alpha_threshold, dither, x_dither, y_dither;
  pixbuf = XEN_LIST_REF(arglist, 0);
  drawable = XEN_LIST_REF(arglist, 1);
  src_x = XEN_LIST_REF(arglist, 2);
  src_y = XEN_LIST_REF(arglist, 3);
  dest_x = XEN_LIST_REF(arglist, 4);
  dest_y = XEN_LIST_REF(arglist, 5);
  width = XEN_LIST_REF(arglist, 6);
  height = XEN_LIST_REF(arglist, 7);
  alpha_mode = XEN_LIST_REF(arglist, 8);
  alpha_threshold = XEN_LIST_REF(arglist, 9);
  dither = XEN_LIST_REF(arglist, 10);
  x_dither = XEN_LIST_REF(arglist, 11);
  y_dither = XEN_LIST_REF(arglist, 12);
  XEN_ASSERT_TYPE(XEN_GdkPixbuf__P(pixbuf), pixbuf, 1, "gdk_pixbuf_render_to_drawable_alpha", "GdkPixbuf*");
  XEN_ASSERT_TYPE(XEN_GdkDrawable__P(drawable), drawable, 2, "gdk_pixbuf_render_to_drawable_alpha", "GdkDrawable*");
  XEN_ASSERT_TYPE(XEN_int_P(src_x), src_x, 3, "gdk_pixbuf_render_to_drawable_alpha", "int");
  XEN_ASSERT_TYPE(XEN_int_P(src_y), src_y, 4, "gdk_pixbuf_render_to_drawable_alpha", "int");
  XEN_ASSERT_TYPE(XEN_int_P(dest_x), dest_x, 5, "gdk_pixbuf_render_to_drawable_alpha", "int");
  XEN_ASSERT_TYPE(XEN_int_P(dest_y), dest_y, 6, "gdk_pixbuf_render_to_drawable_alpha", "int");
  XEN_ASSERT_TYPE(XEN_int_P(width), width, 7, "gdk_pixbuf_render_to_drawable_alpha", "int");
  XEN_ASSERT_TYPE(XEN_int_P(height), height, 8, "gdk_pixbuf_render_to_drawable_alpha", "int");
  XEN_ASSERT_TYPE(XEN_GdkPixbufAlphaMode_P(alpha_mode), alpha_mode, 9, "gdk_pixbuf_render_to_drawable_alpha", "GdkPixbufAlphaMode");
  XEN_ASSERT_TYPE(XEN_int_P(alpha_threshold), alpha_threshold, 10, "gdk_pixbuf_render_to_drawable_alpha", "int");
  XEN_ASSERT_TYPE(XEN_GdkRgbDither_P(dither), dither, 11, "gdk_pixbuf_render_to_drawable_alpha", "GdkRgbDither");
  XEN_ASSERT_TYPE(XEN_int_P(x_dither), x_dither, 12, "gdk_pixbuf_render_to_drawable_alpha", "int");
  XEN_ASSERT_TYPE(XEN_int_P(y_dither), y_dither, 13, "gdk_pixbuf_render_to_drawable_alpha", "int");
  gdk_pixbuf_render_to_drawable_alpha(XEN_TO_C_GdkPixbuf_(pixbuf), XEN_TO_C_GdkDrawable_(drawable), XEN_TO_C_int(src_x), 
                                      XEN_TO_C_int(src_y), XEN_TO_C_int(dest_x), XEN_TO_C_int(dest_y), XEN_TO_C_int(width), 
                                      XEN_TO_C_int(height), XEN_TO_C_GdkPixbufAlphaMode(alpha_mode), XEN_TO_C_int(alpha_threshold), 
                                      XEN_TO_C_GdkRgbDither(dither), XEN_TO_C_int(x_dither), XEN_TO_C_int(y_dither));
  return(XEN_FALSE);
}

static XEN gxg_gdk_pixbuf_render_pixmap_and_mask_for_colormap(XEN pixbuf, XEN colormap, XEN pixmap_return, XEN mask_return, XEN alpha_threshold)
{
  #define H_gdk_pixbuf_render_pixmap_and_mask_for_colormap "void gdk_pixbuf_render_pixmap_and_mask_for_colormap(GdkPixbuf* pixbuf, \
GdkColormap* colormap, GdkPixmap** [pixmap_return], GdkBitmap** [mask_return], int alpha_threshold)"
  GdkPixmap* ref_pixmap_return;
  GdkBitmap* ref_mask_return;
  XEN_ASSERT_TYPE(XEN_GdkPixbuf__P(pixbuf), pixbuf, 1, "gdk_pixbuf_render_pixmap_and_mask_for_colormap", "GdkPixbuf*");
  XEN_ASSERT_TYPE(XEN_GdkColormap__P(colormap), colormap, 2, "gdk_pixbuf_render_pixmap_and_mask_for_colormap", "GdkColormap*");
  XEN_ASSERT_TYPE(XEN_int_P(alpha_threshold), alpha_threshold, 5, "gdk_pixbuf_render_pixmap_and_mask_for_colormap", "int");
  gdk_pixbuf_render_pixmap_and_mask_for_colormap(XEN_TO_C_GdkPixbuf_(pixbuf), XEN_TO_C_GdkColormap_(colormap), &ref_pixmap_return, 
                                                 &ref_mask_return, XEN_TO_C_int(alpha_threshold));
  return(XEN_LIST_2(C_TO_XEN_GdkPixmap_(ref_pixmap_return), C_TO_XEN_GdkBitmap_(ref_mask_return)));
}

static XEN gxg_gdk_pixbuf_render_pixmap_and_mask(XEN pixbuf, XEN pixmap_return, XEN mask_return, XEN alpha_threshold)
{
  #define H_gdk_pixbuf_render_pixmap_and_mask "void gdk_pixbuf_render_pixmap_and_mask(GdkPixbuf* pixbuf, \
GdkPixmap** [pixmap_return], GdkBitmap** [mask_return], int alpha_threshold)"
  GdkPixmap* ref_pixmap_return;
  GdkBitmap* ref_mask_return;
  XEN_ASSERT_TYPE(XEN_GdkPixbuf__P(pixbuf), pixbuf, 1, "gdk_pixbuf_render_pixmap_and_mask", "GdkPixbuf*");
  XEN_ASSERT_TYPE(XEN_int_P(alpha_threshold), alpha_threshold, 4, "gdk_pixbuf_render_pixmap_and_mask", "int");
  gdk_pixbuf_render_pixmap_and_mask(XEN_TO_C_GdkPixbuf_(pixbuf), &ref_pixmap_return, &ref_mask_return, XEN_TO_C_int(alpha_threshold));
  return(XEN_LIST_2(C_TO_XEN_GdkPixmap_(ref_pixmap_return), C_TO_XEN_GdkBitmap_(ref_mask_return)));
}

static XEN gxg_gdk_pixbuf_get_from_drawable(XEN dest, XEN src, XEN cmap, XEN src_x, XEN src_y, XEN dest_x, XEN dest_y, XEN width, XEN height)
{
  #define H_gdk_pixbuf_get_from_drawable "GdkPixbuf* gdk_pixbuf_get_from_drawable(GdkPixbuf* dest, GdkDrawable* src, \
GdkColormap* cmap, int src_x, int src_y, int dest_x, int dest_y, int width, int height)"
  XEN_ASSERT_TYPE(XEN_GdkPixbuf__P(dest), dest, 1, "gdk_pixbuf_get_from_drawable", "GdkPixbuf*");
  XEN_ASSERT_TYPE(XEN_GdkDrawable__P(src), src, 2, "gdk_pixbuf_get_from_drawable", "GdkDrawable*");
  XEN_ASSERT_TYPE(XEN_GdkColormap__P(cmap), cmap, 3, "gdk_pixbuf_get_from_drawable", "GdkColormap*");
  XEN_ASSERT_TYPE(XEN_int_P(src_x), src_x, 4, "gdk_pixbuf_get_from_drawable", "int");
  XEN_ASSERT_TYPE(XEN_int_P(src_y), src_y, 5, "gdk_pixbuf_get_from_drawable", "int");
  XEN_ASSERT_TYPE(XEN_int_P(dest_x), dest_x, 6, "gdk_pixbuf_get_from_drawable", "int");
  XEN_ASSERT_TYPE(XEN_int_P(dest_y), dest_y, 7, "gdk_pixbuf_get_from_drawable", "int");
  XEN_ASSERT_TYPE(XEN_int_P(width), width, 8, "gdk_pixbuf_get_from_drawable", "int");
  XEN_ASSERT_TYPE(XEN_int_P(height), height, 9, "gdk_pixbuf_get_from_drawable", "int");
  return(C_TO_XEN_GdkPixbuf_(gdk_pixbuf_get_from_drawable(XEN_TO_C_GdkPixbuf_(dest), XEN_TO_C_GdkDrawable_(src), XEN_TO_C_GdkColormap_(cmap), 
                                                          XEN_TO_C_int(src_x), XEN_TO_C_int(src_y), XEN_TO_C_int(dest_x), 
                                                          XEN_TO_C_int(dest_y), XEN_TO_C_int(width), XEN_TO_C_int(height))));
}

static XEN gxg_gdk_pixbuf_get_from_image(XEN dest, XEN src, XEN cmap, XEN src_x, XEN src_y, XEN dest_x, XEN dest_y, XEN width, XEN height)
{
  #define H_gdk_pixbuf_get_from_image "GdkPixbuf* gdk_pixbuf_get_from_image(GdkPixbuf* dest, GdkImage* src, \
GdkColormap* cmap, int src_x, int src_y, int dest_x, int dest_y, int width, int height)"
  XEN_ASSERT_TYPE(XEN_GdkPixbuf__P(dest), dest, 1, "gdk_pixbuf_get_from_image", "GdkPixbuf*");
  XEN_ASSERT_TYPE(XEN_GdkImage__P(src), src, 2, "gdk_pixbuf_get_from_image", "GdkImage*");
  XEN_ASSERT_TYPE(XEN_GdkColormap__P(cmap), cmap, 3, "gdk_pixbuf_get_from_image", "GdkColormap*");
  XEN_ASSERT_TYPE(XEN_int_P(src_x), src_x, 4, "gdk_pixbuf_get_from_image", "int");
  XEN_ASSERT_TYPE(XEN_int_P(src_y), src_y, 5, "gdk_pixbuf_get_from_image", "int");
  XEN_ASSERT_TYPE(XEN_int_P(dest_x), dest_x, 6, "gdk_pixbuf_get_from_image", "int");
  XEN_ASSERT_TYPE(XEN_int_P(dest_y), dest_y, 7, "gdk_pixbuf_get_from_image", "int");
  XEN_ASSERT_TYPE(XEN_int_P(width), width, 8, "gdk_pixbuf_get_from_image", "int");
  XEN_ASSERT_TYPE(XEN_int_P(height), height, 9, "gdk_pixbuf_get_from_image", "int");
  return(C_TO_XEN_GdkPixbuf_(gdk_pixbuf_get_from_image(XEN_TO_C_GdkPixbuf_(dest), XEN_TO_C_GdkImage_(src), XEN_TO_C_GdkColormap_(cmap), 
                                                       XEN_TO_C_int(src_x), XEN_TO_C_int(src_y), XEN_TO_C_int(dest_x), XEN_TO_C_int(dest_y), 
                                                       XEN_TO_C_int(width), XEN_TO_C_int(height))));
}

static XEN gxg_gdk_pixmap_get_type(void)
{
  #define H_gdk_pixmap_get_type "GType gdk_pixmap_get_type( void)"
  return(C_TO_XEN_GType(gdk_pixmap_get_type()));
}
#endif

static XEN gxg_gdk_pixmap_new(XEN window, XEN width, XEN height, XEN depth)
{
  #define H_gdk_pixmap_new "GdkPixmap* gdk_pixmap_new(GdkWindow* window, gint width, gint height, gint depth)"
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 1, "gdk_pixmap_new", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_gint_P(width), width, 2, "gdk_pixmap_new", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(height), height, 3, "gdk_pixmap_new", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(depth), depth, 4, "gdk_pixmap_new", "gint");
  return(C_TO_XEN_GdkPixmap_(gdk_pixmap_new(XEN_TO_C_GdkWindow_(window), XEN_TO_C_gint(width), XEN_TO_C_gint(height), XEN_TO_C_gint(depth))));
}

static XEN gxg_gdk_bitmap_create_from_data(XEN window, XEN data, XEN width, XEN height)
{
  #define H_gdk_bitmap_create_from_data "GdkBitmap* gdk_bitmap_create_from_data(GdkWindow* window, gchar* data, \
gint width, gint height)"
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 1, "gdk_bitmap_create_from_data", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_gchar__P(data), data, 2, "gdk_bitmap_create_from_data", "gchar*");
  XEN_ASSERT_TYPE(XEN_gint_P(width), width, 3, "gdk_bitmap_create_from_data", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(height), height, 4, "gdk_bitmap_create_from_data", "gint");
  return(C_TO_XEN_GdkBitmap_(gdk_bitmap_create_from_data(XEN_TO_C_GdkWindow_(window), XEN_TO_C_gchar_(data), XEN_TO_C_gint(width), 
                                                         XEN_TO_C_gint(height))));
}

static XEN gxg_gdk_pixmap_create_from_data(XEN window, XEN data, XEN width, XEN height, XEN depth, XEN fg, XEN bg)
{
  #define H_gdk_pixmap_create_from_data "GdkPixmap* gdk_pixmap_create_from_data(GdkWindow* window, gchar* data, \
gint width, gint height, gint depth, GdkColor* fg, GdkColor* bg)"
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 1, "gdk_pixmap_create_from_data", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_gchar__P(data), data, 2, "gdk_pixmap_create_from_data", "gchar*");
  XEN_ASSERT_TYPE(XEN_gint_P(width), width, 3, "gdk_pixmap_create_from_data", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(height), height, 4, "gdk_pixmap_create_from_data", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(depth), depth, 5, "gdk_pixmap_create_from_data", "gint");
  XEN_ASSERT_TYPE(XEN_GdkColor__P(fg), fg, 6, "gdk_pixmap_create_from_data", "GdkColor*");
  XEN_ASSERT_TYPE(XEN_GdkColor__P(bg), bg, 7, "gdk_pixmap_create_from_data", "GdkColor*");
  return(C_TO_XEN_GdkPixmap_(gdk_pixmap_create_from_data(XEN_TO_C_GdkWindow_(window), XEN_TO_C_gchar_(data), XEN_TO_C_gint(width), 
                                                         XEN_TO_C_gint(height), XEN_TO_C_gint(depth), XEN_TO_C_GdkColor_(fg), 
                                                         XEN_TO_C_GdkColor_(bg))));
}

static XEN gxg_gdk_pixmap_create_from_xpm(XEN window, XEN mask, XEN transparent_color, XEN filename)
{
  #define H_gdk_pixmap_create_from_xpm "GdkPixmap* gdk_pixmap_create_from_xpm(GdkWindow* window, GdkBitmap** mask, \
GdkColor* transparent_color, gchar* filename)"
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 1, "gdk_pixmap_create_from_xpm", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_GdkBitmap___P(mask), mask, 2, "gdk_pixmap_create_from_xpm", "GdkBitmap**");
  XEN_ASSERT_TYPE(XEN_GdkColor__P(transparent_color), transparent_color, 3, "gdk_pixmap_create_from_xpm", "GdkColor*");
  XEN_ASSERT_TYPE(XEN_gchar__P(filename), filename, 4, "gdk_pixmap_create_from_xpm", "gchar*");
  return(C_TO_XEN_GdkPixmap_(gdk_pixmap_create_from_xpm(XEN_TO_C_GdkWindow_(window), XEN_TO_C_GdkBitmap__(mask), XEN_TO_C_GdkColor_(transparent_color), 
                                                        XEN_TO_C_gchar_(filename))));
}

static XEN gxg_gdk_pixmap_colormap_create_from_xpm(XEN window, XEN colormap, XEN mask, XEN transparent_color, XEN filename)
{
  #define H_gdk_pixmap_colormap_create_from_xpm "GdkPixmap* gdk_pixmap_colormap_create_from_xpm(GdkWindow* window, \
GdkColormap* colormap, GdkBitmap** mask, GdkColor* transparent_color, gchar* filename)"
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 1, "gdk_pixmap_colormap_create_from_xpm", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_GdkColormap__P(colormap), colormap, 2, "gdk_pixmap_colormap_create_from_xpm", "GdkColormap*");
  XEN_ASSERT_TYPE(XEN_GdkBitmap___P(mask), mask, 3, "gdk_pixmap_colormap_create_from_xpm", "GdkBitmap**");
  XEN_ASSERT_TYPE(XEN_GdkColor__P(transparent_color), transparent_color, 4, "gdk_pixmap_colormap_create_from_xpm", "GdkColor*");
  XEN_ASSERT_TYPE(XEN_gchar__P(filename), filename, 5, "gdk_pixmap_colormap_create_from_xpm", "gchar*");
  return(C_TO_XEN_GdkPixmap_(gdk_pixmap_colormap_create_from_xpm(XEN_TO_C_GdkWindow_(window), XEN_TO_C_GdkColormap_(colormap), 
                                                                 XEN_TO_C_GdkBitmap__(mask), XEN_TO_C_GdkColor_(transparent_color), 
                                                                 XEN_TO_C_gchar_(filename))));
}

static XEN gxg_gdk_pixmap_create_from_xpm_d(XEN window, XEN mask, XEN transparent_color, XEN data)
{
  #define H_gdk_pixmap_create_from_xpm_d "GdkPixmap* gdk_pixmap_create_from_xpm_d(GdkWindow* window, \
GdkBitmap** mask, GdkColor* transparent_color, gchar** data)"
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 1, "gdk_pixmap_create_from_xpm_d", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_GdkBitmap___P(mask), mask, 2, "gdk_pixmap_create_from_xpm_d", "GdkBitmap**");
  XEN_ASSERT_TYPE(XEN_GdkColor__P(transparent_color), transparent_color, 3, "gdk_pixmap_create_from_xpm_d", "GdkColor*");
  XEN_ASSERT_TYPE(XEN_gchar___P(data), data, 4, "gdk_pixmap_create_from_xpm_d", "gchar**");
  return(C_TO_XEN_GdkPixmap_(gdk_pixmap_create_from_xpm_d(XEN_TO_C_GdkWindow_(window), XEN_TO_C_GdkBitmap__(mask), XEN_TO_C_GdkColor_(transparent_color), 
                                                          XEN_TO_C_gchar__(data))));
}

static XEN gxg_gdk_pixmap_colormap_create_from_xpm_d(XEN window, XEN colormap, XEN mask, XEN transparent_color, XEN data)
{
  #define H_gdk_pixmap_colormap_create_from_xpm_d "GdkPixmap* gdk_pixmap_colormap_create_from_xpm_d(GdkWindow* window, \
GdkColormap* colormap, GdkBitmap** mask, GdkColor* transparent_color, gchar** data)"
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 1, "gdk_pixmap_colormap_create_from_xpm_d", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_GdkColormap__P(colormap), colormap, 2, "gdk_pixmap_colormap_create_from_xpm_d", "GdkColormap*");
  XEN_ASSERT_TYPE(XEN_GdkBitmap___P(mask), mask, 3, "gdk_pixmap_colormap_create_from_xpm_d", "GdkBitmap**");
  XEN_ASSERT_TYPE(XEN_GdkColor__P(transparent_color), transparent_color, 4, "gdk_pixmap_colormap_create_from_xpm_d", "GdkColor*");
  XEN_ASSERT_TYPE(XEN_gchar___P(data), data, 5, "gdk_pixmap_colormap_create_from_xpm_d", "gchar**");
  return(C_TO_XEN_GdkPixmap_(gdk_pixmap_colormap_create_from_xpm_d(XEN_TO_C_GdkWindow_(window), XEN_TO_C_GdkColormap_(colormap), 
                                                                   XEN_TO_C_GdkBitmap__(mask), XEN_TO_C_GdkColor_(transparent_color), 
                                                                   XEN_TO_C_gchar__(data))));
}

#if (!HAVE_GTK_1)
static XEN gxg_gdk_pixmap_foreign_new(XEN anid)
{
  #define H_gdk_pixmap_foreign_new "GdkPixmap* gdk_pixmap_foreign_new(GdkNativeWindow anid)"
  XEN_ASSERT_TYPE(XEN_GdkNativeWindow_P(anid), anid, 1, "gdk_pixmap_foreign_new", "GdkNativeWindow");
  return(C_TO_XEN_GdkPixmap_(gdk_pixmap_foreign_new(XEN_TO_C_GdkNativeWindow(anid))));
}

static XEN gxg_gdk_pixmap_lookup(XEN anid)
{
  #define H_gdk_pixmap_lookup "GdkPixmap* gdk_pixmap_lookup(GdkNativeWindow anid)"
  XEN_ASSERT_TYPE(XEN_GdkNativeWindow_P(anid), anid, 1, "gdk_pixmap_lookup", "GdkNativeWindow");
  return(C_TO_XEN_GdkPixmap_(gdk_pixmap_lookup(XEN_TO_C_GdkNativeWindow(anid))));
}
#endif

static XEN gxg_gdk_atom_intern(XEN atom_name, XEN only_if_exists)
{
  #define H_gdk_atom_intern "GdkAtom gdk_atom_intern(gchar* atom_name, gboolean only_if_exists)"
  XEN_ASSERT_TYPE(XEN_gchar__P(atom_name), atom_name, 1, "gdk_atom_intern", "gchar*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(only_if_exists), only_if_exists, 2, "gdk_atom_intern", "gboolean");
  return(C_TO_XEN_GdkAtom(gdk_atom_intern(XEN_TO_C_gchar_(atom_name), XEN_TO_C_gboolean(only_if_exists))));
}

static XEN gxg_gdk_atom_name(XEN atom)
{
  #define H_gdk_atom_name "gchar* gdk_atom_name(GdkAtom atom)"
  XEN_ASSERT_TYPE(XEN_GdkAtom_P(atom), atom, 1, "gdk_atom_name", "GdkAtom");
  return(C_TO_XEN_gchar_(gdk_atom_name(XEN_TO_C_GdkAtom(atom))));
}

static XEN gxg_gdk_property_get(XEN arglist)
{
  #define H_gdk_property_get "gboolean gdk_property_get(GdkWindow* window, GdkAtom property, GdkAtom type, \
gulong offset, gulong length, gint pdelete, GdkAtom* [actual_property_type], gint* [actual_format], gint* [actual_length], \
guchar** [data])"
  GdkAtom ref_actual_property_type;
  gint ref_actual_format;
  gint ref_actual_length;
  guchar* ref_data;
  XEN window, property, type, offset, length, pdelete;
  window = XEN_LIST_REF(arglist, 0);
  property = XEN_LIST_REF(arglist, 1);
  type = XEN_LIST_REF(arglist, 2);
  offset = XEN_LIST_REF(arglist, 3);
  length = XEN_LIST_REF(arglist, 4);
  pdelete = XEN_LIST_REF(arglist, 5);
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 1, "gdk_property_get", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_GdkAtom_P(property), property, 2, "gdk_property_get", "GdkAtom");
  XEN_ASSERT_TYPE(XEN_GdkAtom_P(type), type, 3, "gdk_property_get", "GdkAtom");
  XEN_ASSERT_TYPE(XEN_gulong_P(offset), offset, 4, "gdk_property_get", "gulong");
  XEN_ASSERT_TYPE(XEN_gulong_P(length), length, 5, "gdk_property_get", "gulong");
  XEN_ASSERT_TYPE(XEN_gint_P(pdelete), pdelete, 6, "gdk_property_get", "gint");
  {
    XEN result = XEN_FALSE;
    result = C_TO_XEN_gboolean(gdk_property_get(XEN_TO_C_GdkWindow_(window), XEN_TO_C_GdkAtom(property), XEN_TO_C_GdkAtom(type), 
                                                XEN_TO_C_gulong(offset), XEN_TO_C_gulong(length), XEN_TO_C_gint(pdelete), 
                                                &ref_actual_property_type, &ref_actual_format, &ref_actual_length, &ref_data));
    return(XEN_LIST_5(result, C_TO_XEN_GdkAtom(ref_actual_property_type), C_TO_XEN_gint(ref_actual_format), C_TO_XEN_gint(ref_actual_length), C_TO_XEN_guchar_(ref_data)));
   }
}

static XEN gxg_gdk_property_change(XEN window, XEN property, XEN type, XEN format, XEN mode, XEN data, XEN nelements)
{
  #define H_gdk_property_change "void gdk_property_change(GdkWindow* window, GdkAtom property, GdkAtom type, \
gint format, GdkPropMode mode, guchar* data, gint nelements)"
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 1, "gdk_property_change", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_GdkAtom_P(property), property, 2, "gdk_property_change", "GdkAtom");
  XEN_ASSERT_TYPE(XEN_GdkAtom_P(type), type, 3, "gdk_property_change", "GdkAtom");
  XEN_ASSERT_TYPE(XEN_gint_P(format), format, 4, "gdk_property_change", "gint");
  XEN_ASSERT_TYPE(XEN_GdkPropMode_P(mode), mode, 5, "gdk_property_change", "GdkPropMode");
  XEN_ASSERT_TYPE(XEN_guchar__P(data), data, 6, "gdk_property_change", "guchar*");
  XEN_ASSERT_TYPE(XEN_gint_P(nelements), nelements, 7, "gdk_property_change", "gint");
  gdk_property_change(XEN_TO_C_GdkWindow_(window), XEN_TO_C_GdkAtom(property), XEN_TO_C_GdkAtom(type), XEN_TO_C_gint(format), 
                      XEN_TO_C_GdkPropMode(mode), XEN_TO_C_guchar_(data), XEN_TO_C_gint(nelements));
  return(XEN_FALSE);
}

static XEN gxg_gdk_property_delete(XEN window, XEN property)
{
  #define H_gdk_property_delete "void gdk_property_delete(GdkWindow* window, GdkAtom property)"
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 1, "gdk_property_delete", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_GdkAtom_P(property), property, 2, "gdk_property_delete", "GdkAtom");
  gdk_property_delete(XEN_TO_C_GdkWindow_(window), XEN_TO_C_GdkAtom(property));
  return(XEN_FALSE);
}

static XEN gxg_gdk_text_property_to_text_list(XEN encoding, XEN format, XEN text, XEN length, XEN list)
{
  #define H_gdk_text_property_to_text_list "gint gdk_text_property_to_text_list(GdkAtom encoding, gint format, \
guchar* text, gint length, gchar*** list)"
  XEN_ASSERT_TYPE(XEN_GdkAtom_P(encoding), encoding, 1, "gdk_text_property_to_text_list", "GdkAtom");
  XEN_ASSERT_TYPE(XEN_gint_P(format), format, 2, "gdk_text_property_to_text_list", "gint");
  XEN_ASSERT_TYPE(XEN_guchar__P(text), text, 3, "gdk_text_property_to_text_list", "guchar*");
  XEN_ASSERT_TYPE(XEN_gint_P(length), length, 4, "gdk_text_property_to_text_list", "gint");
  XEN_ASSERT_TYPE(XEN_gchar____P(list), list, 5, "gdk_text_property_to_text_list", "gchar***");
  return(C_TO_XEN_gint(gdk_text_property_to_text_list(XEN_TO_C_GdkAtom(encoding), XEN_TO_C_gint(format), XEN_TO_C_guchar_(text), 
                                                      XEN_TO_C_gint(length), XEN_TO_C_gchar___(list))));
}

#if (!HAVE_GTK_1)
static XEN gxg_gdk_text_property_to_utf8_list(XEN encoding, XEN format, XEN text, XEN length, XEN list)
{
  #define H_gdk_text_property_to_utf8_list "gint gdk_text_property_to_utf8_list(GdkAtom encoding, gint format, \
guchar* text, gint length, gchar*** list)"
  XEN_ASSERT_TYPE(XEN_GdkAtom_P(encoding), encoding, 1, "gdk_text_property_to_utf8_list", "GdkAtom");
  XEN_ASSERT_TYPE(XEN_gint_P(format), format, 2, "gdk_text_property_to_utf8_list", "gint");
  XEN_ASSERT_TYPE(XEN_guchar__P(text), text, 3, "gdk_text_property_to_utf8_list", "guchar*");
  XEN_ASSERT_TYPE(XEN_gint_P(length), length, 4, "gdk_text_property_to_utf8_list", "gint");
  XEN_ASSERT_TYPE(XEN_gchar____P(list), list, 5, "gdk_text_property_to_utf8_list", "gchar***");
  return(C_TO_XEN_gint(gdk_text_property_to_utf8_list(XEN_TO_C_GdkAtom(encoding), XEN_TO_C_gint(format), XEN_TO_C_guchar_(text), 
                                                      XEN_TO_C_gint(length), XEN_TO_C_gchar___(list))));
}

static XEN gxg_gdk_utf8_to_string_target(XEN str)
{
  #define H_gdk_utf8_to_string_target "gchar* gdk_utf8_to_string_target(gchar* str)"
  XEN_ASSERT_TYPE(XEN_gchar__P(str), str, 1, "gdk_utf8_to_string_target", "gchar*");
  return(C_TO_XEN_gchar_(gdk_utf8_to_string_target(XEN_TO_C_gchar_(str))));
}

static XEN gxg_gdk_utf8_to_compound_text(XEN str, XEN encoding, XEN format, XEN ctext, XEN length)
{
  #define H_gdk_utf8_to_compound_text "gboolean gdk_utf8_to_compound_text(gchar* str, GdkAtom* encoding, \
gint* format, guchar** [ctext], gint* [length])"
  guchar* ref_ctext;
  gint ref_length;
  XEN_ASSERT_TYPE(XEN_gchar__P(str), str, 1, "gdk_utf8_to_compound_text", "gchar*");
  XEN_ASSERT_TYPE(XEN_GdkAtom__P(encoding), encoding, 2, "gdk_utf8_to_compound_text", "GdkAtom*");
  XEN_ASSERT_TYPE(XEN_gint__P(format), format, 3, "gdk_utf8_to_compound_text", "gint*");
  {
    XEN result = XEN_FALSE;
    result = C_TO_XEN_gboolean(gdk_utf8_to_compound_text(XEN_TO_C_gchar_(str), XEN_TO_C_GdkAtom_(encoding), XEN_TO_C_gint_(format), 
                                                         &ref_ctext, &ref_length));
    return(XEN_LIST_3(result, C_TO_XEN_guchar_(ref_ctext), C_TO_XEN_gint(ref_length)));
   }
}
#endif

static XEN gxg_gdk_free_text_list(XEN list)
{
  #define H_gdk_free_text_list "void gdk_free_text_list(gchar** list)"
  XEN_ASSERT_TYPE(XEN_gchar___P(list), list, 1, "gdk_free_text_list", "gchar**");
  gdk_free_text_list(XEN_TO_C_gchar__(list));
  return(XEN_FALSE);
}

static XEN gxg_gdk_string_to_compound_text(XEN str, XEN encoding, XEN format, XEN ctext, XEN length)
{
  #define H_gdk_string_to_compound_text "gint gdk_string_to_compound_text(gchar* str, GdkAtom* encoding, \
gint* format, guchar** [ctext], gint* [length])"
  guchar* ref_ctext;
  gint ref_length;
  XEN_ASSERT_TYPE(XEN_gchar__P(str), str, 1, "gdk_string_to_compound_text", "gchar*");
  XEN_ASSERT_TYPE(XEN_GdkAtom__P(encoding), encoding, 2, "gdk_string_to_compound_text", "GdkAtom*");
  XEN_ASSERT_TYPE(XEN_gint__P(format), format, 3, "gdk_string_to_compound_text", "gint*");
  {
    XEN result = XEN_FALSE;
    result = C_TO_XEN_gint(gdk_string_to_compound_text(XEN_TO_C_gchar_(str), XEN_TO_C_GdkAtom_(encoding), XEN_TO_C_gint_(format), 
                                                       &ref_ctext, &ref_length));
    return(XEN_LIST_3(result, C_TO_XEN_guchar_(ref_ctext), C_TO_XEN_gint(ref_length)));
   }
}

static XEN gxg_gdk_free_compound_text(XEN ctext)
{
  #define H_gdk_free_compound_text "void gdk_free_compound_text(guchar* ctext)"
  XEN_ASSERT_TYPE(XEN_guchar__P(ctext), ctext, 1, "gdk_free_compound_text", "guchar*");
  gdk_free_compound_text(XEN_TO_C_guchar_(ctext));
  return(XEN_FALSE);
}

static XEN gxg_gdk_region_new(void)
{
  #define H_gdk_region_new "GdkRegion* gdk_region_new( void)"
  return(C_TO_XEN_GdkRegion_(gdk_region_new()));
}

static XEN gxg_gdk_region_polygon(XEN points, XEN npoints, XEN fill_rule)
{
  #define H_gdk_region_polygon "GdkRegion* gdk_region_polygon(GdkPoint* points, gint npoints, GdkFillRule fill_rule)"
  XEN_ASSERT_TYPE(XEN_GdkPoint__P(points), points, 1, "gdk_region_polygon", "GdkPoint*");
  XEN_ASSERT_TYPE(XEN_gint_P(npoints), npoints, 2, "gdk_region_polygon", "gint");
  XEN_ASSERT_TYPE(XEN_GdkFillRule_P(fill_rule), fill_rule, 3, "gdk_region_polygon", "GdkFillRule");
  return(C_TO_XEN_GdkRegion_(gdk_region_polygon(XEN_TO_C_GdkPoint_(points), XEN_TO_C_gint(npoints), XEN_TO_C_GdkFillRule(fill_rule))));
}

#if (!HAVE_GTK_1)
static XEN gxg_gdk_region_copy(XEN region)
{
  #define H_gdk_region_copy "GdkRegion* gdk_region_copy(GdkRegion* region)"
  XEN_ASSERT_TYPE(XEN_GdkRegion__P(region), region, 1, "gdk_region_copy", "GdkRegion*");
  return(C_TO_XEN_GdkRegion_(gdk_region_copy(XEN_TO_C_GdkRegion_(region))));
}

static XEN gxg_gdk_region_rectangle(XEN rectangle)
{
  #define H_gdk_region_rectangle "GdkRegion* gdk_region_rectangle(GdkRectangle* rectangle)"
  XEN_ASSERT_TYPE(XEN_GdkRectangle__P(rectangle), rectangle, 1, "gdk_region_rectangle", "GdkRectangle*");
  return(C_TO_XEN_GdkRegion_(gdk_region_rectangle(XEN_TO_C_GdkRectangle_(rectangle))));
}
#endif

static XEN gxg_gdk_region_destroy(XEN region)
{
  #define H_gdk_region_destroy "void gdk_region_destroy(GdkRegion* region)"
  XEN_ASSERT_TYPE(XEN_GdkRegion__P(region), region, 1, "gdk_region_destroy", "GdkRegion*");
  gdk_region_destroy(XEN_TO_C_GdkRegion_(region));
  return(XEN_FALSE);
}

static XEN gxg_gdk_region_get_clipbox(XEN region, XEN rectangle)
{
  #define H_gdk_region_get_clipbox "void gdk_region_get_clipbox(GdkRegion* region, GdkRectangle* rectangle)"
  XEN_ASSERT_TYPE(XEN_GdkRegion__P(region), region, 1, "gdk_region_get_clipbox", "GdkRegion*");
  XEN_ASSERT_TYPE(XEN_GdkRectangle__P(rectangle), rectangle, 2, "gdk_region_get_clipbox", "GdkRectangle*");
  gdk_region_get_clipbox(XEN_TO_C_GdkRegion_(region), XEN_TO_C_GdkRectangle_(rectangle));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gdk_region_get_rectangles(XEN region, XEN rectangles, XEN n_rectangles)
{
  #define H_gdk_region_get_rectangles "void gdk_region_get_rectangles(GdkRegion* region, GdkRectangle** [rectangles], \
gint* [n_rectangles])"
  GdkRectangle* ref_rectangles;
  gint ref_n_rectangles;
  XEN_ASSERT_TYPE(XEN_GdkRegion__P(region), region, 1, "gdk_region_get_rectangles", "GdkRegion*");
  gdk_region_get_rectangles(XEN_TO_C_GdkRegion_(region), &ref_rectangles, &ref_n_rectangles);
  return(XEN_LIST_2(C_TO_XEN_GdkRectangle_(ref_rectangles), C_TO_XEN_gint(ref_n_rectangles)));
}
#endif

static XEN gxg_gdk_region_empty(XEN region)
{
  #define H_gdk_region_empty "gboolean gdk_region_empty(GdkRegion* region)"
  XEN_ASSERT_TYPE(XEN_GdkRegion__P(region), region, 1, "gdk_region_empty", "GdkRegion*");
  return(C_TO_XEN_gboolean(gdk_region_empty(XEN_TO_C_GdkRegion_(region))));
}

static XEN gxg_gdk_region_equal(XEN region1, XEN region2)
{
  #define H_gdk_region_equal "gboolean gdk_region_equal(GdkRegion* region1, GdkRegion* region2)"
  XEN_ASSERT_TYPE(XEN_GdkRegion__P(region1), region1, 1, "gdk_region_equal", "GdkRegion*");
  XEN_ASSERT_TYPE(XEN_GdkRegion__P(region2), region2, 2, "gdk_region_equal", "GdkRegion*");
  return(C_TO_XEN_gboolean(gdk_region_equal(XEN_TO_C_GdkRegion_(region1), XEN_TO_C_GdkRegion_(region2))));
}

static XEN gxg_gdk_region_point_in(XEN region, XEN x, XEN y)
{
  #define H_gdk_region_point_in "gboolean gdk_region_point_in(GdkRegion* region, int x, int y)"
  XEN_ASSERT_TYPE(XEN_GdkRegion__P(region), region, 1, "gdk_region_point_in", "GdkRegion*");
  XEN_ASSERT_TYPE(XEN_int_P(x), x, 2, "gdk_region_point_in", "int");
  XEN_ASSERT_TYPE(XEN_int_P(y), y, 3, "gdk_region_point_in", "int");
  return(C_TO_XEN_gboolean(gdk_region_point_in(XEN_TO_C_GdkRegion_(region), XEN_TO_C_int(x), XEN_TO_C_int(y))));
}

static XEN gxg_gdk_region_rect_in(XEN region, XEN rect)
{
  #define H_gdk_region_rect_in "GdkOverlapType gdk_region_rect_in(GdkRegion* region, GdkRectangle* rect)"
  XEN_ASSERT_TYPE(XEN_GdkRegion__P(region), region, 1, "gdk_region_rect_in", "GdkRegion*");
  XEN_ASSERT_TYPE(XEN_GdkRectangle__P(rect), rect, 2, "gdk_region_rect_in", "GdkRectangle*");
  return(C_TO_XEN_GdkOverlapType(gdk_region_rect_in(XEN_TO_C_GdkRegion_(region), XEN_TO_C_GdkRectangle_(rect))));
}

static XEN gxg_gdk_region_offset(XEN region, XEN dx, XEN dy)
{
  #define H_gdk_region_offset "void gdk_region_offset(GdkRegion* region, gint dx, gint dy)"
  XEN_ASSERT_TYPE(XEN_GdkRegion__P(region), region, 1, "gdk_region_offset", "GdkRegion*");
  XEN_ASSERT_TYPE(XEN_gint_P(dx), dx, 2, "gdk_region_offset", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(dy), dy, 3, "gdk_region_offset", "gint");
  gdk_region_offset(XEN_TO_C_GdkRegion_(region), XEN_TO_C_gint(dx), XEN_TO_C_gint(dy));
  return(XEN_FALSE);
}

static XEN gxg_gdk_region_shrink(XEN region, XEN dx, XEN dy)
{
  #define H_gdk_region_shrink "void gdk_region_shrink(GdkRegion* region, gint dx, gint dy)"
  XEN_ASSERT_TYPE(XEN_GdkRegion__P(region), region, 1, "gdk_region_shrink", "GdkRegion*");
  XEN_ASSERT_TYPE(XEN_gint_P(dx), dx, 2, "gdk_region_shrink", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(dy), dy, 3, "gdk_region_shrink", "gint");
  gdk_region_shrink(XEN_TO_C_GdkRegion_(region), XEN_TO_C_gint(dx), XEN_TO_C_gint(dy));
  return(XEN_FALSE);
}

static XEN gxg_gdk_region_union_with_rect(XEN region, XEN rect)
{
  #define H_gdk_region_union_with_rect "void gdk_region_union_with_rect(GdkRegion* region, GdkRectangle* rect)"
  XEN_ASSERT_TYPE(XEN_GdkRegion__P(region), region, 1, "gdk_region_union_with_rect", "GdkRegion*");
  XEN_ASSERT_TYPE(XEN_GdkRectangle__P(rect), rect, 2, "gdk_region_union_with_rect", "GdkRectangle*");
  gdk_region_union_with_rect(XEN_TO_C_GdkRegion_(region), XEN_TO_C_GdkRectangle_(rect));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gdk_region_intersect(XEN source1, XEN source2)
{
  #define H_gdk_region_intersect "void gdk_region_intersect(GdkRegion* source1, GdkRegion* source2)"
  XEN_ASSERT_TYPE(XEN_GdkRegion__P(source1), source1, 1, "gdk_region_intersect", "GdkRegion*");
  XEN_ASSERT_TYPE(XEN_GdkRegion__P(source2), source2, 2, "gdk_region_intersect", "GdkRegion*");
  gdk_region_intersect(XEN_TO_C_GdkRegion_(source1), XEN_TO_C_GdkRegion_(source2));
  return(XEN_FALSE);
}

static XEN gxg_gdk_region_union(XEN source1, XEN source2)
{
  #define H_gdk_region_union "void gdk_region_union(GdkRegion* source1, GdkRegion* source2)"
  XEN_ASSERT_TYPE(XEN_GdkRegion__P(source1), source1, 1, "gdk_region_union", "GdkRegion*");
  XEN_ASSERT_TYPE(XEN_GdkRegion__P(source2), source2, 2, "gdk_region_union", "GdkRegion*");
  gdk_region_union(XEN_TO_C_GdkRegion_(source1), XEN_TO_C_GdkRegion_(source2));
  return(XEN_FALSE);
}

static XEN gxg_gdk_region_subtract(XEN source1, XEN source2)
{
  #define H_gdk_region_subtract "void gdk_region_subtract(GdkRegion* source1, GdkRegion* source2)"
  XEN_ASSERT_TYPE(XEN_GdkRegion__P(source1), source1, 1, "gdk_region_subtract", "GdkRegion*");
  XEN_ASSERT_TYPE(XEN_GdkRegion__P(source2), source2, 2, "gdk_region_subtract", "GdkRegion*");
  gdk_region_subtract(XEN_TO_C_GdkRegion_(source1), XEN_TO_C_GdkRegion_(source2));
  return(XEN_FALSE);
}

static XEN gxg_gdk_region_xor(XEN source1, XEN source2)
{
  #define H_gdk_region_xor "void gdk_region_xor(GdkRegion* source1, GdkRegion* source2)"
  XEN_ASSERT_TYPE(XEN_GdkRegion__P(source1), source1, 1, "gdk_region_xor", "GdkRegion*");
  XEN_ASSERT_TYPE(XEN_GdkRegion__P(source2), source2, 2, "gdk_region_xor", "GdkRegion*");
  gdk_region_xor(XEN_TO_C_GdkRegion_(source1), XEN_TO_C_GdkRegion_(source2));
  return(XEN_FALSE);
}

static XEN gxg_gdk_region_spans_intersect_foreach(XEN region, XEN spans, XEN n_spans, XEN sorted, XEN func, XEN func_data)
{
  #define H_gdk_region_spans_intersect_foreach "void gdk_region_spans_intersect_foreach(GdkRegion* region, \
GdkSpan* spans, int n_spans, gboolean sorted, GdkSpanFunc func, lambda_data func_data)"
  XEN_ASSERT_TYPE(XEN_GdkRegion__P(region), region, 1, "gdk_region_spans_intersect_foreach", "GdkRegion*");
  XEN_ASSERT_TYPE(XEN_GdkSpan__P(spans), spans, 2, "gdk_region_spans_intersect_foreach", "GdkSpan*");
  XEN_ASSERT_TYPE(XEN_int_P(n_spans), n_spans, 3, "gdk_region_spans_intersect_foreach", "int");
  XEN_ASSERT_TYPE(XEN_gboolean_P(sorted), sorted, 4, "gdk_region_spans_intersect_foreach", "gboolean");
  XEN_ASSERT_TYPE(XEN_GdkSpanFunc_P(func), func, 5, "gdk_region_spans_intersect_foreach", "GdkSpanFunc");
  if (XEN_NOT_BOUND_P(func_data)) func_data = XEN_FALSE; 
  else XEN_ASSERT_TYPE(XEN_lambda_data_P(func_data), func_data, 6, "gdk_region_spans_intersect_foreach", "lambda_data");
  {
    int loc;
    XEN gxg_ptr = XEN_LIST_5(func, func_data, XEN_FALSE, XEN_FALSE, XEN_FALSE);
    loc = xm_protect(gxg_ptr);
    XEN_LIST_SET(gxg_ptr, 2, C_TO_XEN_INT(loc));
    gdk_region_spans_intersect_foreach(XEN_TO_C_GdkRegion_(region), XEN_TO_C_GdkSpan_(spans), XEN_TO_C_int(n_spans), XEN_TO_C_gboolean(sorted), 
                                   XEN_TO_C_GdkSpanFunc(func), XEN_TO_C_lambda_data(func_data));
    xm_unprotect_at(loc);
    return(XEN_FALSE);
   }
}

static XEN gxg_gdk_rgb_find_color(XEN colormap, XEN color)
{
  #define H_gdk_rgb_find_color "void gdk_rgb_find_color(GdkColormap* colormap, GdkColor* color)"
  XEN_ASSERT_TYPE(XEN_GdkColormap__P(colormap), colormap, 1, "gdk_rgb_find_color", "GdkColormap*");
  XEN_ASSERT_TYPE(XEN_GdkColor__P(color), color, 2, "gdk_rgb_find_color", "GdkColor*");
  gdk_rgb_find_color(XEN_TO_C_GdkColormap_(colormap), XEN_TO_C_GdkColor_(color));
  return(XEN_FALSE);
}
#endif

static XEN gxg_gdk_draw_rgb_image(XEN drawable, XEN gc, XEN x, XEN y, XEN width, XEN height, XEN dith, XEN rgb_buf, XEN rowstride)
{
  #define H_gdk_draw_rgb_image "void gdk_draw_rgb_image(GdkDrawable* drawable, GdkGC* gc, gint x, gint y, \
gint width, gint height, GdkRgbDither dith, guchar* rgb_buf, gint rowstride)"
  XEN_ASSERT_TYPE(XEN_GdkDrawable__P(drawable), drawable, 1, "gdk_draw_rgb_image", "GdkDrawable*");
  XEN_ASSERT_TYPE(XEN_GdkGC__P(gc), gc, 2, "gdk_draw_rgb_image", "GdkGC*");
  XEN_ASSERT_TYPE(XEN_gint_P(x), x, 3, "gdk_draw_rgb_image", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(y), y, 4, "gdk_draw_rgb_image", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(width), width, 5, "gdk_draw_rgb_image", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(height), height, 6, "gdk_draw_rgb_image", "gint");
  XEN_ASSERT_TYPE(XEN_GdkRgbDither_P(dith), dith, 7, "gdk_draw_rgb_image", "GdkRgbDither");
  XEN_ASSERT_TYPE(XEN_guchar__P(rgb_buf), rgb_buf, 8, "gdk_draw_rgb_image", "guchar*");
  XEN_ASSERT_TYPE(XEN_gint_P(rowstride), rowstride, 9, "gdk_draw_rgb_image", "gint");
  gdk_draw_rgb_image(XEN_TO_C_GdkDrawable_(drawable), XEN_TO_C_GdkGC_(gc), XEN_TO_C_gint(x), XEN_TO_C_gint(y), XEN_TO_C_gint(width), 
                     XEN_TO_C_gint(height), XEN_TO_C_GdkRgbDither(dith), XEN_TO_C_guchar_(rgb_buf), XEN_TO_C_gint(rowstride));
  return(XEN_FALSE);
}

static XEN gxg_gdk_draw_rgb_image_dithalign(XEN arglist)
{
  #define H_gdk_draw_rgb_image_dithalign "void gdk_draw_rgb_image_dithalign(GdkDrawable* drawable, GdkGC* gc, \
gint x, gint y, gint width, gint height, GdkRgbDither dith, guchar* rgb_buf, gint rowstride, gint xdith, \
gint ydith)"
  XEN drawable, gc, x, y, width, height, dith, rgb_buf, rowstride, xdith, ydith;
  drawable = XEN_LIST_REF(arglist, 0);
  gc = XEN_LIST_REF(arglist, 1);
  x = XEN_LIST_REF(arglist, 2);
  y = XEN_LIST_REF(arglist, 3);
  width = XEN_LIST_REF(arglist, 4);
  height = XEN_LIST_REF(arglist, 5);
  dith = XEN_LIST_REF(arglist, 6);
  rgb_buf = XEN_LIST_REF(arglist, 7);
  rowstride = XEN_LIST_REF(arglist, 8);
  xdith = XEN_LIST_REF(arglist, 9);
  ydith = XEN_LIST_REF(arglist, 10);
  XEN_ASSERT_TYPE(XEN_GdkDrawable__P(drawable), drawable, 1, "gdk_draw_rgb_image_dithalign", "GdkDrawable*");
  XEN_ASSERT_TYPE(XEN_GdkGC__P(gc), gc, 2, "gdk_draw_rgb_image_dithalign", "GdkGC*");
  XEN_ASSERT_TYPE(XEN_gint_P(x), x, 3, "gdk_draw_rgb_image_dithalign", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(y), y, 4, "gdk_draw_rgb_image_dithalign", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(width), width, 5, "gdk_draw_rgb_image_dithalign", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(height), height, 6, "gdk_draw_rgb_image_dithalign", "gint");
  XEN_ASSERT_TYPE(XEN_GdkRgbDither_P(dith), dith, 7, "gdk_draw_rgb_image_dithalign", "GdkRgbDither");
  XEN_ASSERT_TYPE(XEN_guchar__P(rgb_buf), rgb_buf, 8, "gdk_draw_rgb_image_dithalign", "guchar*");
  XEN_ASSERT_TYPE(XEN_gint_P(rowstride), rowstride, 9, "gdk_draw_rgb_image_dithalign", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(xdith), xdith, 10, "gdk_draw_rgb_image_dithalign", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(ydith), ydith, 11, "gdk_draw_rgb_image_dithalign", "gint");
  gdk_draw_rgb_image_dithalign(XEN_TO_C_GdkDrawable_(drawable), XEN_TO_C_GdkGC_(gc), XEN_TO_C_gint(x), XEN_TO_C_gint(y), 
                               XEN_TO_C_gint(width), XEN_TO_C_gint(height), XEN_TO_C_GdkRgbDither(dith), XEN_TO_C_guchar_(rgb_buf), 
                               XEN_TO_C_gint(rowstride), XEN_TO_C_gint(xdith), XEN_TO_C_gint(ydith));
  return(XEN_FALSE);
}

static XEN gxg_gdk_draw_rgb_32_image(XEN drawable, XEN gc, XEN x, XEN y, XEN width, XEN height, XEN dith, XEN buf, XEN rowstride)
{
  #define H_gdk_draw_rgb_32_image "void gdk_draw_rgb_32_image(GdkDrawable* drawable, GdkGC* gc, gint x, \
gint y, gint width, gint height, GdkRgbDither dith, guchar* buf, gint rowstride)"
  XEN_ASSERT_TYPE(XEN_GdkDrawable__P(drawable), drawable, 1, "gdk_draw_rgb_32_image", "GdkDrawable*");
  XEN_ASSERT_TYPE(XEN_GdkGC__P(gc), gc, 2, "gdk_draw_rgb_32_image", "GdkGC*");
  XEN_ASSERT_TYPE(XEN_gint_P(x), x, 3, "gdk_draw_rgb_32_image", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(y), y, 4, "gdk_draw_rgb_32_image", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(width), width, 5, "gdk_draw_rgb_32_image", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(height), height, 6, "gdk_draw_rgb_32_image", "gint");
  XEN_ASSERT_TYPE(XEN_GdkRgbDither_P(dith), dith, 7, "gdk_draw_rgb_32_image", "GdkRgbDither");
  XEN_ASSERT_TYPE(XEN_guchar__P(buf), buf, 8, "gdk_draw_rgb_32_image", "guchar*");
  XEN_ASSERT_TYPE(XEN_gint_P(rowstride), rowstride, 9, "gdk_draw_rgb_32_image", "gint");
  gdk_draw_rgb_32_image(XEN_TO_C_GdkDrawable_(drawable), XEN_TO_C_GdkGC_(gc), XEN_TO_C_gint(x), XEN_TO_C_gint(y), XEN_TO_C_gint(width), 
                        XEN_TO_C_gint(height), XEN_TO_C_GdkRgbDither(dith), XEN_TO_C_guchar_(buf), XEN_TO_C_gint(rowstride));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gdk_draw_rgb_32_image_dithalign(XEN arglist)
{
  #define H_gdk_draw_rgb_32_image_dithalign "void gdk_draw_rgb_32_image_dithalign(GdkDrawable* drawable, \
GdkGC* gc, gint x, gint y, gint width, gint height, GdkRgbDither dith, guchar* buf, gint rowstride, gint xdith, \
gint ydith)"
  XEN drawable, gc, x, y, width, height, dith, buf, rowstride, xdith, ydith;
  drawable = XEN_LIST_REF(arglist, 0);
  gc = XEN_LIST_REF(arglist, 1);
  x = XEN_LIST_REF(arglist, 2);
  y = XEN_LIST_REF(arglist, 3);
  width = XEN_LIST_REF(arglist, 4);
  height = XEN_LIST_REF(arglist, 5);
  dith = XEN_LIST_REF(arglist, 6);
  buf = XEN_LIST_REF(arglist, 7);
  rowstride = XEN_LIST_REF(arglist, 8);
  xdith = XEN_LIST_REF(arglist, 9);
  ydith = XEN_LIST_REF(arglist, 10);
  XEN_ASSERT_TYPE(XEN_GdkDrawable__P(drawable), drawable, 1, "gdk_draw_rgb_32_image_dithalign", "GdkDrawable*");
  XEN_ASSERT_TYPE(XEN_GdkGC__P(gc), gc, 2, "gdk_draw_rgb_32_image_dithalign", "GdkGC*");
  XEN_ASSERT_TYPE(XEN_gint_P(x), x, 3, "gdk_draw_rgb_32_image_dithalign", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(y), y, 4, "gdk_draw_rgb_32_image_dithalign", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(width), width, 5, "gdk_draw_rgb_32_image_dithalign", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(height), height, 6, "gdk_draw_rgb_32_image_dithalign", "gint");
  XEN_ASSERT_TYPE(XEN_GdkRgbDither_P(dith), dith, 7, "gdk_draw_rgb_32_image_dithalign", "GdkRgbDither");
  XEN_ASSERT_TYPE(XEN_guchar__P(buf), buf, 8, "gdk_draw_rgb_32_image_dithalign", "guchar*");
  XEN_ASSERT_TYPE(XEN_gint_P(rowstride), rowstride, 9, "gdk_draw_rgb_32_image_dithalign", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(xdith), xdith, 10, "gdk_draw_rgb_32_image_dithalign", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(ydith), ydith, 11, "gdk_draw_rgb_32_image_dithalign", "gint");
  gdk_draw_rgb_32_image_dithalign(XEN_TO_C_GdkDrawable_(drawable), XEN_TO_C_GdkGC_(gc), XEN_TO_C_gint(x), XEN_TO_C_gint(y), 
                                  XEN_TO_C_gint(width), XEN_TO_C_gint(height), XEN_TO_C_GdkRgbDither(dith), XEN_TO_C_guchar_(buf), 
                                  XEN_TO_C_gint(rowstride), XEN_TO_C_gint(xdith), XEN_TO_C_gint(ydith));
  return(XEN_FALSE);
}
#endif

static XEN gxg_gdk_draw_gray_image(XEN drawable, XEN gc, XEN x, XEN y, XEN width, XEN height, XEN dith, XEN buf, XEN rowstride)
{
  #define H_gdk_draw_gray_image "void gdk_draw_gray_image(GdkDrawable* drawable, GdkGC* gc, gint x, gint y, \
gint width, gint height, GdkRgbDither dith, guchar* buf, gint rowstride)"
  XEN_ASSERT_TYPE(XEN_GdkDrawable__P(drawable), drawable, 1, "gdk_draw_gray_image", "GdkDrawable*");
  XEN_ASSERT_TYPE(XEN_GdkGC__P(gc), gc, 2, "gdk_draw_gray_image", "GdkGC*");
  XEN_ASSERT_TYPE(XEN_gint_P(x), x, 3, "gdk_draw_gray_image", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(y), y, 4, "gdk_draw_gray_image", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(width), width, 5, "gdk_draw_gray_image", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(height), height, 6, "gdk_draw_gray_image", "gint");
  XEN_ASSERT_TYPE(XEN_GdkRgbDither_P(dith), dith, 7, "gdk_draw_gray_image", "GdkRgbDither");
  XEN_ASSERT_TYPE(XEN_guchar__P(buf), buf, 8, "gdk_draw_gray_image", "guchar*");
  XEN_ASSERT_TYPE(XEN_gint_P(rowstride), rowstride, 9, "gdk_draw_gray_image", "gint");
  gdk_draw_gray_image(XEN_TO_C_GdkDrawable_(drawable), XEN_TO_C_GdkGC_(gc), XEN_TO_C_gint(x), XEN_TO_C_gint(y), XEN_TO_C_gint(width), 
                      XEN_TO_C_gint(height), XEN_TO_C_GdkRgbDither(dith), XEN_TO_C_guchar_(buf), XEN_TO_C_gint(rowstride));
  return(XEN_FALSE);
}

static XEN gxg_gdk_draw_indexed_image(XEN arglist)
{
  #define H_gdk_draw_indexed_image "void gdk_draw_indexed_image(GdkDrawable* drawable, GdkGC* gc, gint x, \
gint y, gint width, gint height, GdkRgbDither dith, guchar* buf, gint rowstride, GdkRgbCmap* cmap)"
  XEN drawable, gc, x, y, width, height, dith, buf, rowstride, cmap;
  drawable = XEN_LIST_REF(arglist, 0);
  gc = XEN_LIST_REF(arglist, 1);
  x = XEN_LIST_REF(arglist, 2);
  y = XEN_LIST_REF(arglist, 3);
  width = XEN_LIST_REF(arglist, 4);
  height = XEN_LIST_REF(arglist, 5);
  dith = XEN_LIST_REF(arglist, 6);
  buf = XEN_LIST_REF(arglist, 7);
  rowstride = XEN_LIST_REF(arglist, 8);
  cmap = XEN_LIST_REF(arglist, 9);
  XEN_ASSERT_TYPE(XEN_GdkDrawable__P(drawable), drawable, 1, "gdk_draw_indexed_image", "GdkDrawable*");
  XEN_ASSERT_TYPE(XEN_GdkGC__P(gc), gc, 2, "gdk_draw_indexed_image", "GdkGC*");
  XEN_ASSERT_TYPE(XEN_gint_P(x), x, 3, "gdk_draw_indexed_image", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(y), y, 4, "gdk_draw_indexed_image", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(width), width, 5, "gdk_draw_indexed_image", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(height), height, 6, "gdk_draw_indexed_image", "gint");
  XEN_ASSERT_TYPE(XEN_GdkRgbDither_P(dith), dith, 7, "gdk_draw_indexed_image", "GdkRgbDither");
  XEN_ASSERT_TYPE(XEN_guchar__P(buf), buf, 8, "gdk_draw_indexed_image", "guchar*");
  XEN_ASSERT_TYPE(XEN_gint_P(rowstride), rowstride, 9, "gdk_draw_indexed_image", "gint");
  XEN_ASSERT_TYPE(XEN_GdkRgbCmap__P(cmap), cmap, 10, "gdk_draw_indexed_image", "GdkRgbCmap*");
  gdk_draw_indexed_image(XEN_TO_C_GdkDrawable_(drawable), XEN_TO_C_GdkGC_(gc), XEN_TO_C_gint(x), XEN_TO_C_gint(y), XEN_TO_C_gint(width), 
                         XEN_TO_C_gint(height), XEN_TO_C_GdkRgbDither(dith), XEN_TO_C_guchar_(buf), XEN_TO_C_gint(rowstride), 
                         XEN_TO_C_GdkRgbCmap_(cmap));
  return(XEN_FALSE);
}

static XEN gxg_gdk_rgb_cmap_new(XEN colors, XEN n_colors)
{
  #define H_gdk_rgb_cmap_new "GdkRgbCmap* gdk_rgb_cmap_new(guint32* colors, gint n_colors)"
  XEN_ASSERT_TYPE(XEN_guint32__P(colors), colors, 1, "gdk_rgb_cmap_new", "guint32*");
  XEN_ASSERT_TYPE(XEN_gint_P(n_colors), n_colors, 2, "gdk_rgb_cmap_new", "gint");
  return(C_TO_XEN_GdkRgbCmap_(gdk_rgb_cmap_new(XEN_TO_C_guint32_(colors), XEN_TO_C_gint(n_colors))));
}

static XEN gxg_gdk_rgb_cmap_free(XEN cmap)
{
  #define H_gdk_rgb_cmap_free "void gdk_rgb_cmap_free(GdkRgbCmap* cmap)"
  XEN_ASSERT_TYPE(XEN_GdkRgbCmap__P(cmap), cmap, 1, "gdk_rgb_cmap_free", "GdkRgbCmap*");
  gdk_rgb_cmap_free(XEN_TO_C_GdkRgbCmap_(cmap));
  return(XEN_FALSE);
}

static XEN gxg_gdk_rgb_ditherable(void)
{
  #define H_gdk_rgb_ditherable "gboolean gdk_rgb_ditherable( void)"
  return(C_TO_XEN_gboolean(gdk_rgb_ditherable()));
}

static XEN gxg_gdk_rgb_set_verbose(XEN verbose)
{
  #define H_gdk_rgb_set_verbose "void gdk_rgb_set_verbose(gboolean verbose)"
  XEN_ASSERT_TYPE(XEN_gboolean_P(verbose), verbose, 1, "gdk_rgb_set_verbose", "gboolean");
  gdk_rgb_set_verbose(XEN_TO_C_gboolean(verbose));
  return(XEN_FALSE);
}

static XEN gxg_gdk_rgb_set_install(XEN install)
{
  #define H_gdk_rgb_set_install "void gdk_rgb_set_install(gboolean install)"
  XEN_ASSERT_TYPE(XEN_gboolean_P(install), install, 1, "gdk_rgb_set_install", "gboolean");
  gdk_rgb_set_install(XEN_TO_C_gboolean(install));
  return(XEN_FALSE);
}

static XEN gxg_gdk_rgb_set_min_colors(XEN min_colors)
{
  #define H_gdk_rgb_set_min_colors "void gdk_rgb_set_min_colors(gint min_colors)"
  XEN_ASSERT_TYPE(XEN_gint_P(min_colors), min_colors, 1, "gdk_rgb_set_min_colors", "gint");
  gdk_rgb_set_min_colors(XEN_TO_C_gint(min_colors));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gdk_rgb_get_colormap(void)
{
  #define H_gdk_rgb_get_colormap "GdkColormap* gdk_rgb_get_colormap( void)"
  return(C_TO_XEN_GdkColormap_(gdk_rgb_get_colormap()));
}
#endif

static XEN gxg_gdk_rgb_get_visual(void)
{
  #define H_gdk_rgb_get_visual "GdkVisual* gdk_rgb_get_visual( void)"
  return(C_TO_XEN_GdkVisual_(gdk_rgb_get_visual()));
}

static XEN gxg_gdk_selection_owner_set(XEN owner, XEN selection, XEN time, XEN send_event)
{
  #define H_gdk_selection_owner_set "gboolean gdk_selection_owner_set(GdkWindow* owner, GdkAtom selection, \
guint32 time, gboolean send_event)"
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(owner), owner, 1, "gdk_selection_owner_set", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_GdkAtom_P(selection), selection, 2, "gdk_selection_owner_set", "GdkAtom");
  XEN_ASSERT_TYPE(XEN_guint32_P(time), time, 3, "gdk_selection_owner_set", "guint32");
  XEN_ASSERT_TYPE(XEN_gboolean_P(send_event), send_event, 4, "gdk_selection_owner_set", "gboolean");
  return(C_TO_XEN_gboolean(gdk_selection_owner_set(XEN_TO_C_GdkWindow_(owner), XEN_TO_C_GdkAtom(selection), XEN_TO_C_guint32(time), 
                                                   XEN_TO_C_gboolean(send_event))));
}

static XEN gxg_gdk_selection_owner_get(XEN selection)
{
  #define H_gdk_selection_owner_get "GdkWindow* gdk_selection_owner_get(GdkAtom selection)"
  XEN_ASSERT_TYPE(XEN_GdkAtom_P(selection), selection, 1, "gdk_selection_owner_get", "GdkAtom");
  return(C_TO_XEN_GdkWindow_(gdk_selection_owner_get(XEN_TO_C_GdkAtom(selection))));
}

static XEN gxg_gdk_selection_convert(XEN requestor, XEN selection, XEN target, XEN time)
{
  #define H_gdk_selection_convert "void gdk_selection_convert(GdkWindow* requestor, GdkAtom selection, \
GdkAtom target, guint32 time)"
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(requestor), requestor, 1, "gdk_selection_convert", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_GdkAtom_P(selection), selection, 2, "gdk_selection_convert", "GdkAtom");
  XEN_ASSERT_TYPE(XEN_GdkAtom_P(target), target, 3, "gdk_selection_convert", "GdkAtom");
  XEN_ASSERT_TYPE(XEN_guint32_P(time), time, 4, "gdk_selection_convert", "guint32");
  gdk_selection_convert(XEN_TO_C_GdkWindow_(requestor), XEN_TO_C_GdkAtom(selection), XEN_TO_C_GdkAtom(target), XEN_TO_C_guint32(time));
  return(XEN_FALSE);
}

static XEN gxg_gdk_selection_property_get(XEN requestor, XEN data, XEN prop_type, XEN prop_format)
{
  #define H_gdk_selection_property_get "gboolean gdk_selection_property_get(GdkWindow* requestor, guchar** [data], \
GdkAtom* [prop_type], gint* [prop_format])"
  guchar* ref_data;
  GdkAtom ref_prop_type;
  gint ref_prop_format;
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(requestor), requestor, 1, "gdk_selection_property_get", "GdkWindow*");
  {
    XEN result = XEN_FALSE;
    result = C_TO_XEN_gboolean(gdk_selection_property_get(XEN_TO_C_GdkWindow_(requestor), &ref_data, &ref_prop_type, &ref_prop_format));
    return(XEN_LIST_4(result, C_TO_XEN_guchar_(ref_data), C_TO_XEN_GdkAtom(ref_prop_type), C_TO_XEN_gint(ref_prop_format)));
   }
}

static XEN gxg_gdk_selection_send_notify(XEN requestor, XEN selection, XEN target, XEN property, XEN time)
{
  #define H_gdk_selection_send_notify "void gdk_selection_send_notify(guint32 requestor, GdkAtom selection, \
GdkAtom target, GdkAtom property, guint32 time)"
  XEN_ASSERT_TYPE(XEN_guint32_P(requestor), requestor, 1, "gdk_selection_send_notify", "guint32");
  XEN_ASSERT_TYPE(XEN_GdkAtom_P(selection), selection, 2, "gdk_selection_send_notify", "GdkAtom");
  XEN_ASSERT_TYPE(XEN_GdkAtom_P(target), target, 3, "gdk_selection_send_notify", "GdkAtom");
  XEN_ASSERT_TYPE(XEN_GdkAtom_P(property), property, 4, "gdk_selection_send_notify", "GdkAtom");
  XEN_ASSERT_TYPE(XEN_guint32_P(time), time, 5, "gdk_selection_send_notify", "guint32");
  gdk_selection_send_notify(XEN_TO_C_guint32(requestor), XEN_TO_C_GdkAtom(selection), XEN_TO_C_GdkAtom(target), XEN_TO_C_GdkAtom(property), 
                            XEN_TO_C_guint32(time));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gdk_visual_get_type(void)
{
  #define H_gdk_visual_get_type "GType gdk_visual_get_type( void)"
  return(C_TO_XEN_GType(gdk_visual_get_type()));
}
#endif

static XEN gxg_gdk_visual_get_best_depth(void)
{
  #define H_gdk_visual_get_best_depth "gint gdk_visual_get_best_depth( void)"
  return(C_TO_XEN_gint(gdk_visual_get_best_depth()));
}

static XEN gxg_gdk_visual_get_best_type(void)
{
  #define H_gdk_visual_get_best_type "GdkVisualType gdk_visual_get_best_type( void)"
  return(C_TO_XEN_GdkVisualType(gdk_visual_get_best_type()));
}

static XEN gxg_gdk_visual_get_system(void)
{
  #define H_gdk_visual_get_system "GdkVisual* gdk_visual_get_system( void)"
  return(C_TO_XEN_GdkVisual_(gdk_visual_get_system()));
}

static XEN gxg_gdk_visual_get_best(void)
{
  #define H_gdk_visual_get_best "GdkVisual* gdk_visual_get_best( void)"
  return(C_TO_XEN_GdkVisual_(gdk_visual_get_best()));
}

static XEN gxg_gdk_visual_get_best_with_depth(XEN depth)
{
  #define H_gdk_visual_get_best_with_depth "GdkVisual* gdk_visual_get_best_with_depth(gint depth)"
  XEN_ASSERT_TYPE(XEN_gint_P(depth), depth, 1, "gdk_visual_get_best_with_depth", "gint");
  return(C_TO_XEN_GdkVisual_(gdk_visual_get_best_with_depth(XEN_TO_C_gint(depth))));
}

static XEN gxg_gdk_visual_get_best_with_type(XEN visual_type)
{
  #define H_gdk_visual_get_best_with_type "GdkVisual* gdk_visual_get_best_with_type(GdkVisualType visual_type)"
  XEN_ASSERT_TYPE(XEN_GdkVisualType_P(visual_type), visual_type, 1, "gdk_visual_get_best_with_type", "GdkVisualType");
  return(C_TO_XEN_GdkVisual_(gdk_visual_get_best_with_type(XEN_TO_C_GdkVisualType(visual_type))));
}

static XEN gxg_gdk_visual_get_best_with_both(XEN depth, XEN visual_type)
{
  #define H_gdk_visual_get_best_with_both "GdkVisual* gdk_visual_get_best_with_both(gint depth, GdkVisualType visual_type)"
  XEN_ASSERT_TYPE(XEN_gint_P(depth), depth, 1, "gdk_visual_get_best_with_both", "gint");
  XEN_ASSERT_TYPE(XEN_GdkVisualType_P(visual_type), visual_type, 2, "gdk_visual_get_best_with_both", "GdkVisualType");
  return(C_TO_XEN_GdkVisual_(gdk_visual_get_best_with_both(XEN_TO_C_gint(depth), XEN_TO_C_GdkVisualType(visual_type))));
}

static XEN gxg_gdk_query_depths(XEN depths, XEN count)
{
  #define H_gdk_query_depths "void gdk_query_depths(gint** [depths], gint* [count])"
  gint* ref_depths;
  gint ref_count;
  gdk_query_depths(&ref_depths, &ref_count);
  return(XEN_LIST_2(C_TO_XEN_gint_(ref_depths), C_TO_XEN_gint(ref_count)));
}

static XEN gxg_gdk_query_visual_types(XEN visual_types, XEN count)
{
  #define H_gdk_query_visual_types "void gdk_query_visual_types(GdkVisualType** [visual_types], gint* [count])"
  GdkVisualType* ref_visual_types;
  gint ref_count;
  gdk_query_visual_types(&ref_visual_types, &ref_count);
  return(XEN_LIST_2(C_TO_XEN_GdkVisualType_(ref_visual_types), C_TO_XEN_gint(ref_count)));
}

static XEN gxg_gdk_list_visuals(void)
{
  #define H_gdk_list_visuals "GList* gdk_list_visuals( void)"
  return(C_TO_XEN_GList_(gdk_list_visuals()));
}

#if (!HAVE_GTK_1)
static XEN gxg_gdk_window_object_get_type(void)
{
  #define H_gdk_window_object_get_type "GType gdk_window_object_get_type( void)"
  return(C_TO_XEN_GType(gdk_window_object_get_type()));
}
#endif

static XEN gxg_gdk_window_new(XEN parent, XEN attributes, XEN attributes_mask)
{
  #define H_gdk_window_new "GdkWindow* gdk_window_new(GdkWindow* parent, GdkWindowAttr* attributes, gint attributes_mask)"
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(parent), parent, 1, "gdk_window_new", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_GdkWindowAttr__P(attributes), attributes, 2, "gdk_window_new", "GdkWindowAttr*");
  XEN_ASSERT_TYPE(XEN_gint_P(attributes_mask), attributes_mask, 3, "gdk_window_new", "gint");
  return(C_TO_XEN_GdkWindow_(gdk_window_new(XEN_TO_C_GdkWindow_(parent), XEN_TO_C_GdkWindowAttr_(attributes), XEN_TO_C_gint(attributes_mask))));
}

static XEN gxg_gdk_window_destroy(XEN window)
{
  #define H_gdk_window_destroy "void gdk_window_destroy(GdkWindow* window)"
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 1, "gdk_window_destroy", "GdkWindow*");
  gdk_window_destroy(XEN_TO_C_GdkWindow_(window));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gdk_window_get_window_type(XEN window)
{
  #define H_gdk_window_get_window_type "GdkWindowType gdk_window_get_window_type(GdkWindow* window)"
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 1, "gdk_window_get_window_type", "GdkWindow*");
  return(C_TO_XEN_GdkWindowType(gdk_window_get_window_type(XEN_TO_C_GdkWindow_(window))));
}
#endif

static XEN gxg_gdk_window_at_pointer(XEN win_x, XEN win_y)
{
  #define H_gdk_window_at_pointer "GdkWindow* gdk_window_at_pointer(gint* [win_x], gint* [win_y])"
  gint ref_win_x;
  gint ref_win_y;
  {
    XEN result = XEN_FALSE;
    result = C_TO_XEN_GdkWindow_(gdk_window_at_pointer(&ref_win_x, &ref_win_y));
    return(XEN_LIST_3(result, C_TO_XEN_gint(ref_win_x), C_TO_XEN_gint(ref_win_y)));
   }
}

static XEN gxg_gdk_window_show(XEN window)
{
  #define H_gdk_window_show "void gdk_window_show(GdkWindow* window)"
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 1, "gdk_window_show", "GdkWindow*");
  gdk_window_show(XEN_TO_C_GdkWindow_(window));
  return(XEN_FALSE);
}

static XEN gxg_gdk_window_hide(XEN window)
{
  #define H_gdk_window_hide "void gdk_window_hide(GdkWindow* window)"
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 1, "gdk_window_hide", "GdkWindow*");
  gdk_window_hide(XEN_TO_C_GdkWindow_(window));
  return(XEN_FALSE);
}

static XEN gxg_gdk_window_withdraw(XEN window)
{
  #define H_gdk_window_withdraw "void gdk_window_withdraw(GdkWindow* window)"
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 1, "gdk_window_withdraw", "GdkWindow*");
  gdk_window_withdraw(XEN_TO_C_GdkWindow_(window));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gdk_window_show_unraised(XEN window)
{
  #define H_gdk_window_show_unraised "void gdk_window_show_unraised(GdkWindow* window)"
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 1, "gdk_window_show_unraised", "GdkWindow*");
  gdk_window_show_unraised(XEN_TO_C_GdkWindow_(window));
  return(XEN_FALSE);
}
#endif

static XEN gxg_gdk_window_move(XEN window, XEN x, XEN y)
{
  #define H_gdk_window_move "void gdk_window_move(GdkWindow* window, gint x, gint y)"
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 1, "gdk_window_move", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_gint_P(x), x, 2, "gdk_window_move", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(y), y, 3, "gdk_window_move", "gint");
  gdk_window_move(XEN_TO_C_GdkWindow_(window), XEN_TO_C_gint(x), XEN_TO_C_gint(y));
  return(XEN_FALSE);
}

static XEN gxg_gdk_window_resize(XEN window, XEN width, XEN height)
{
  #define H_gdk_window_resize "void gdk_window_resize(GdkWindow* window, gint width, gint height)"
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 1, "gdk_window_resize", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_gint_P(width), width, 2, "gdk_window_resize", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(height), height, 3, "gdk_window_resize", "gint");
  gdk_window_resize(XEN_TO_C_GdkWindow_(window), XEN_TO_C_gint(width), XEN_TO_C_gint(height));
  return(XEN_FALSE);
}

static XEN gxg_gdk_window_move_resize(XEN window, XEN x, XEN y, XEN width, XEN height)
{
  #define H_gdk_window_move_resize "void gdk_window_move_resize(GdkWindow* window, gint x, gint y, gint width, \
gint height)"
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 1, "gdk_window_move_resize", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_gint_P(x), x, 2, "gdk_window_move_resize", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(y), y, 3, "gdk_window_move_resize", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(width), width, 4, "gdk_window_move_resize", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(height), height, 5, "gdk_window_move_resize", "gint");
  gdk_window_move_resize(XEN_TO_C_GdkWindow_(window), XEN_TO_C_gint(x), XEN_TO_C_gint(y), XEN_TO_C_gint(width), XEN_TO_C_gint(height));
  return(XEN_FALSE);
}

static XEN gxg_gdk_window_reparent(XEN window, XEN new_parent, XEN x, XEN y)
{
  #define H_gdk_window_reparent "void gdk_window_reparent(GdkWindow* window, GdkWindow* new_parent, gint x, \
gint y)"
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 1, "gdk_window_reparent", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(new_parent), new_parent, 2, "gdk_window_reparent", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_gint_P(x), x, 3, "gdk_window_reparent", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(y), y, 4, "gdk_window_reparent", "gint");
  gdk_window_reparent(XEN_TO_C_GdkWindow_(window), XEN_TO_C_GdkWindow_(new_parent), XEN_TO_C_gint(x), XEN_TO_C_gint(y));
  return(XEN_FALSE);
}

static XEN gxg_gdk_window_clear(XEN window)
{
  #define H_gdk_window_clear "void gdk_window_clear(GdkWindow* window)"
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 1, "gdk_window_clear", "GdkWindow*");
  gdk_window_clear(XEN_TO_C_GdkWindow_(window));
  return(XEN_FALSE);
}

static XEN gxg_gdk_window_clear_area(XEN window, XEN x, XEN y, XEN width, XEN height)
{
  #define H_gdk_window_clear_area "void gdk_window_clear_area(GdkWindow* window, gint x, gint y, gint width, \
gint height)"
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 1, "gdk_window_clear_area", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_gint_P(x), x, 2, "gdk_window_clear_area", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(y), y, 3, "gdk_window_clear_area", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(width), width, 4, "gdk_window_clear_area", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(height), height, 5, "gdk_window_clear_area", "gint");
  gdk_window_clear_area(XEN_TO_C_GdkWindow_(window), XEN_TO_C_gint(x), XEN_TO_C_gint(y), XEN_TO_C_gint(width), XEN_TO_C_gint(height));
  return(XEN_FALSE);
}

static XEN gxg_gdk_window_clear_area_e(XEN window, XEN x, XEN y, XEN width, XEN height)
{
  #define H_gdk_window_clear_area_e "void gdk_window_clear_area_e(GdkWindow* window, gint x, gint y, \
gint width, gint height)"
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 1, "gdk_window_clear_area_e", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_gint_P(x), x, 2, "gdk_window_clear_area_e", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(y), y, 3, "gdk_window_clear_area_e", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(width), width, 4, "gdk_window_clear_area_e", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(height), height, 5, "gdk_window_clear_area_e", "gint");
  gdk_window_clear_area_e(XEN_TO_C_GdkWindow_(window), XEN_TO_C_gint(x), XEN_TO_C_gint(y), XEN_TO_C_gint(width), XEN_TO_C_gint(height));
  return(XEN_FALSE);
}

static XEN gxg_gdk_window_raise(XEN window)
{
  #define H_gdk_window_raise "void gdk_window_raise(GdkWindow* window)"
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 1, "gdk_window_raise", "GdkWindow*");
  gdk_window_raise(XEN_TO_C_GdkWindow_(window));
  return(XEN_FALSE);
}

static XEN gxg_gdk_window_lower(XEN window)
{
  #define H_gdk_window_lower "void gdk_window_lower(GdkWindow* window)"
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 1, "gdk_window_lower", "GdkWindow*");
  gdk_window_lower(XEN_TO_C_GdkWindow_(window));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gdk_window_focus(XEN window, XEN timestamp)
{
  #define H_gdk_window_focus "void gdk_window_focus(GdkWindow* window, guint32 timestamp)"
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 1, "gdk_window_focus", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_guint32_P(timestamp), timestamp, 2, "gdk_window_focus", "guint32");
  gdk_window_focus(XEN_TO_C_GdkWindow_(window), XEN_TO_C_guint32(timestamp));
  return(XEN_FALSE);
}
#endif

static XEN gxg_gdk_window_set_user_data(XEN window, XEN user_data)
{
  #define H_gdk_window_set_user_data "void gdk_window_set_user_data(GdkWindow* window, gpointer user_data)"
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window) || XEN_FALSE_P(window), window, 1, "gdk_window_set_user_data", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_gpointer_P(user_data), user_data, 2, "gdk_window_set_user_data", "gpointer");
  gdk_window_set_user_data(XEN_TO_C_GdkWindow_(window), XEN_TO_C_gpointer(user_data));
  return(XEN_FALSE);
}

static XEN gxg_gdk_window_set_override_redirect(XEN window, XEN override_redirect)
{
  #define H_gdk_window_set_override_redirect "void gdk_window_set_override_redirect(GdkWindow* window, \
gboolean override_redirect)"
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 1, "gdk_window_set_override_redirect", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(override_redirect), override_redirect, 2, "gdk_window_set_override_redirect", "gboolean");
  gdk_window_set_override_redirect(XEN_TO_C_GdkWindow_(window), XEN_TO_C_gboolean(override_redirect));
  return(XEN_FALSE);
}

static XEN gxg_gdk_window_add_filter(XEN window, XEN func, XEN func_data)
{
  #define H_gdk_window_add_filter "void gdk_window_add_filter(GdkWindow* window, GdkFilterFunc func, \
lambda_data func_data)"
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 1, "gdk_window_add_filter", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_GdkFilterFunc_P(func), func, 2, "gdk_window_add_filter", "GdkFilterFunc");
  if (XEN_NOT_BOUND_P(func_data)) func_data = XEN_FALSE; 
  else XEN_ASSERT_TYPE(XEN_lambda_data_P(func_data), func_data, 3, "gdk_window_add_filter", "lambda_data");
  {
    XEN gxg_ptr = XEN_LIST_5(func, func_data, XEN_FALSE, XEN_FALSE, XEN_FALSE);
    xm_protect(gxg_ptr);
    gdk_window_add_filter(XEN_TO_C_GdkWindow_(window), XEN_TO_C_GdkFilterFunc(func), XEN_TO_C_lambda_data(func_data));
    return(XEN_FALSE);
   }
}

static XEN gxg_gdk_window_remove_filter(XEN window, XEN func, XEN func_data)
{
  #define H_gdk_window_remove_filter "void gdk_window_remove_filter(GdkWindow* window, GdkFilterFunc func, \
lambda_data func_data)"
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 1, "gdk_window_remove_filter", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_GdkFilterFunc_P(func), func, 2, "gdk_window_remove_filter", "GdkFilterFunc");
  if (XEN_NOT_BOUND_P(func_data)) func_data = XEN_FALSE; 
  else XEN_ASSERT_TYPE(XEN_lambda_data_P(func_data), func_data, 3, "gdk_window_remove_filter", "lambda_data");
  {
    XEN gxg_ptr = XEN_LIST_5(func, func_data, XEN_FALSE, XEN_FALSE, XEN_FALSE);
    xm_protect(gxg_ptr);
    gdk_window_remove_filter(XEN_TO_C_GdkWindow_(window), XEN_TO_C_GdkFilterFunc(func), XEN_TO_C_lambda_data(func_data));
    return(XEN_FALSE);
   }
}

#if (!HAVE_GTK_1)
static XEN gxg_gdk_window_scroll(XEN window, XEN dx, XEN dy)
{
  #define H_gdk_window_scroll "void gdk_window_scroll(GdkWindow* window, gint dx, gint dy)"
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 1, "gdk_window_scroll", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_gint_P(dx), dx, 2, "gdk_window_scroll", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(dy), dy, 3, "gdk_window_scroll", "gint");
  gdk_window_scroll(XEN_TO_C_GdkWindow_(window), XEN_TO_C_gint(dx), XEN_TO_C_gint(dy));
  return(XEN_FALSE);
}
#endif

static XEN gxg_gdk_window_shape_combine_mask(XEN window, XEN mask, XEN x, XEN y)
{
  #define H_gdk_window_shape_combine_mask "void gdk_window_shape_combine_mask(GdkWindow* window, GdkBitmap* mask, \
gint x, gint y)"
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 1, "gdk_window_shape_combine_mask", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_GdkBitmap__P(mask), mask, 2, "gdk_window_shape_combine_mask", "GdkBitmap*");
  XEN_ASSERT_TYPE(XEN_gint_P(x), x, 3, "gdk_window_shape_combine_mask", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(y), y, 4, "gdk_window_shape_combine_mask", "gint");
  gdk_window_shape_combine_mask(XEN_TO_C_GdkWindow_(window), XEN_TO_C_GdkBitmap_(mask), XEN_TO_C_gint(x), XEN_TO_C_gint(y));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gdk_window_shape_combine_region(XEN window, XEN shape_region, XEN offset_x, XEN offset_y)
{
  #define H_gdk_window_shape_combine_region "void gdk_window_shape_combine_region(GdkWindow* window, \
GdkRegion* shape_region, gint offset_x, gint offset_y)"
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 1, "gdk_window_shape_combine_region", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_GdkRegion__P(shape_region), shape_region, 2, "gdk_window_shape_combine_region", "GdkRegion*");
  XEN_ASSERT_TYPE(XEN_gint_P(offset_x), offset_x, 3, "gdk_window_shape_combine_region", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(offset_y), offset_y, 4, "gdk_window_shape_combine_region", "gint");
  gdk_window_shape_combine_region(XEN_TO_C_GdkWindow_(window), XEN_TO_C_GdkRegion_(shape_region), XEN_TO_C_gint(offset_x), 
                                  XEN_TO_C_gint(offset_y));
  return(XEN_FALSE);
}
#endif

static XEN gxg_gdk_window_set_child_shapes(XEN window)
{
  #define H_gdk_window_set_child_shapes "void gdk_window_set_child_shapes(GdkWindow* window)"
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 1, "gdk_window_set_child_shapes", "GdkWindow*");
  gdk_window_set_child_shapes(XEN_TO_C_GdkWindow_(window));
  return(XEN_FALSE);
}

static XEN gxg_gdk_window_merge_child_shapes(XEN window)
{
  #define H_gdk_window_merge_child_shapes "void gdk_window_merge_child_shapes(GdkWindow* window)"
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 1, "gdk_window_merge_child_shapes", "GdkWindow*");
  gdk_window_merge_child_shapes(XEN_TO_C_GdkWindow_(window));
  return(XEN_FALSE);
}

static XEN gxg_gdk_window_is_visible(XEN window)
{
  #define H_gdk_window_is_visible "gboolean gdk_window_is_visible(GdkWindow* window)"
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 1, "gdk_window_is_visible", "GdkWindow*");
  return(C_TO_XEN_gboolean(gdk_window_is_visible(XEN_TO_C_GdkWindow_(window))));
}

static XEN gxg_gdk_window_is_viewable(XEN window)
{
  #define H_gdk_window_is_viewable "gboolean gdk_window_is_viewable(GdkWindow* window)"
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 1, "gdk_window_is_viewable", "GdkWindow*");
  return(C_TO_XEN_gboolean(gdk_window_is_viewable(XEN_TO_C_GdkWindow_(window))));
}

#if (!HAVE_GTK_1)
static XEN gxg_gdk_window_get_state(XEN window)
{
  #define H_gdk_window_get_state "GdkWindowState gdk_window_get_state(GdkWindow* window)"
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 1, "gdk_window_get_state", "GdkWindow*");
  return(C_TO_XEN_GdkWindowState(gdk_window_get_state(XEN_TO_C_GdkWindow_(window))));
}
#endif

static XEN gxg_gdk_window_set_static_gravities(XEN window, XEN use_static)
{
  #define H_gdk_window_set_static_gravities "gboolean gdk_window_set_static_gravities(GdkWindow* window, \
gboolean use_static)"
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 1, "gdk_window_set_static_gravities", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(use_static), use_static, 2, "gdk_window_set_static_gravities", "gboolean");
  return(C_TO_XEN_gboolean(gdk_window_set_static_gravities(XEN_TO_C_GdkWindow_(window), XEN_TO_C_gboolean(use_static))));
}

#if (!HAVE_GTK_1)
static XEN gxg_gdk_window_foreign_new(XEN anid)
{
  #define H_gdk_window_foreign_new "GdkWindow* gdk_window_foreign_new(GdkNativeWindow anid)"
  XEN_ASSERT_TYPE(XEN_GdkNativeWindow_P(anid), anid, 1, "gdk_window_foreign_new", "GdkNativeWindow");
  return(C_TO_XEN_GdkWindow_(gdk_window_foreign_new(XEN_TO_C_GdkNativeWindow(anid))));
}

static XEN gxg_gdk_window_lookup(XEN anid)
{
  #define H_gdk_window_lookup "GdkWindow* gdk_window_lookup(GdkNativeWindow anid)"
  XEN_ASSERT_TYPE(XEN_GdkNativeWindow_P(anid), anid, 1, "gdk_window_lookup", "GdkNativeWindow");
  return(C_TO_XEN_GdkWindow_(gdk_window_lookup(XEN_TO_C_GdkNativeWindow(anid))));
}
#endif

static XEN gxg_gdk_window_get_root_origin(XEN window, XEN x, XEN y)
{
  #define H_gdk_window_get_root_origin "void gdk_window_get_root_origin(GdkWindow* window, gint* [x], \
gint* [y])"
  gint ref_x;
  gint ref_y;
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 1, "gdk_window_get_root_origin", "GdkWindow*");
  gdk_window_get_root_origin(XEN_TO_C_GdkWindow_(window), &ref_x, &ref_y);
  return(XEN_LIST_2(C_TO_XEN_gint(ref_x), C_TO_XEN_gint(ref_y)));
}

#if (!HAVE_GTK_1)
static XEN gxg_gdk_window_get_frame_extents(XEN window, XEN rect)
{
  #define H_gdk_window_get_frame_extents "void gdk_window_get_frame_extents(GdkWindow* window, GdkRectangle* rect)"
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 1, "gdk_window_get_frame_extents", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_GdkRectangle__P(rect), rect, 2, "gdk_window_get_frame_extents", "GdkRectangle*");
  gdk_window_get_frame_extents(XEN_TO_C_GdkWindow_(window), XEN_TO_C_GdkRectangle_(rect));
  return(XEN_FALSE);
}
#endif

static XEN gxg_gdk_window_get_pointer(XEN window, XEN x, XEN y, XEN mask)
{
  #define H_gdk_window_get_pointer "GdkWindow* gdk_window_get_pointer(GdkWindow* window, gint* [x], gint* [y], \
GdkModifierType* [mask])"
  gint ref_x;
  gint ref_y;
  GdkModifierType ref_mask;
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 1, "gdk_window_get_pointer", "GdkWindow*");
  {
    XEN result = XEN_FALSE;
    result = C_TO_XEN_GdkWindow_(gdk_window_get_pointer(XEN_TO_C_GdkWindow_(window), &ref_x, &ref_y, &ref_mask));
    return(XEN_LIST_4(result, C_TO_XEN_gint(ref_x), C_TO_XEN_gint(ref_y), C_TO_XEN_GdkModifierType(ref_mask)));
   }
}

static XEN gxg_gdk_window_get_parent(XEN window)
{
  #define H_gdk_window_get_parent "GdkWindow* gdk_window_get_parent(GdkWindow* window)"
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 1, "gdk_window_get_parent", "GdkWindow*");
  return(C_TO_XEN_GdkWindow_(gdk_window_get_parent(XEN_TO_C_GdkWindow_(window))));
}

static XEN gxg_gdk_window_get_toplevel(XEN window)
{
  #define H_gdk_window_get_toplevel "GdkWindow* gdk_window_get_toplevel(GdkWindow* window)"
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 1, "gdk_window_get_toplevel", "GdkWindow*");
  return(C_TO_XEN_GdkWindow_(gdk_window_get_toplevel(XEN_TO_C_GdkWindow_(window))));
}

static XEN gxg_gdk_window_get_children(XEN window)
{
  #define H_gdk_window_get_children "GList* gdk_window_get_children(GdkWindow* window)"
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 1, "gdk_window_get_children", "GdkWindow*");
  return(C_TO_XEN_GList_(gdk_window_get_children(XEN_TO_C_GdkWindow_(window))));
}

#if (!HAVE_GTK_1)
static XEN gxg_gdk_window_peek_children(XEN window)
{
  #define H_gdk_window_peek_children "GList* gdk_window_peek_children(GdkWindow* window)"
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 1, "gdk_window_peek_children", "GdkWindow*");
  return(C_TO_XEN_GList_(gdk_window_peek_children(XEN_TO_C_GdkWindow_(window))));
}
#endif

static XEN gxg_gdk_window_get_events(XEN window)
{
  #define H_gdk_window_get_events "GdkEventMask gdk_window_get_events(GdkWindow* window)"
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 1, "gdk_window_get_events", "GdkWindow*");
  return(C_TO_XEN_GdkEventMask(gdk_window_get_events(XEN_TO_C_GdkWindow_(window))));
}

static XEN gxg_gdk_window_set_events(XEN window, XEN event_mask)
{
  #define H_gdk_window_set_events "void gdk_window_set_events(GdkWindow* window, GdkEventMask event_mask)"
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 1, "gdk_window_set_events", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_GdkEventMask_P(event_mask), event_mask, 2, "gdk_window_set_events", "GdkEventMask");
  gdk_window_set_events(XEN_TO_C_GdkWindow_(window), XEN_TO_C_GdkEventMask(event_mask));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gdk_window_set_icon_list(XEN window, XEN pixbufs)
{
  #define H_gdk_window_set_icon_list "void gdk_window_set_icon_list(GdkWindow* window, GList* pixbufs)"
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 1, "gdk_window_set_icon_list", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_GList__P(pixbufs), pixbufs, 2, "gdk_window_set_icon_list", "GList*");
  gdk_window_set_icon_list(XEN_TO_C_GdkWindow_(window), XEN_TO_C_GList_(pixbufs));
  return(XEN_FALSE);
}
#endif

static XEN gxg_gdk_window_set_icon(XEN window, XEN icon_window, XEN pixmap, XEN mask)
{
  #define H_gdk_window_set_icon "void gdk_window_set_icon(GdkWindow* window, GdkWindow* icon_window, \
GdkPixmap* pixmap, GdkBitmap* mask)"
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 1, "gdk_window_set_icon", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(icon_window), icon_window, 2, "gdk_window_set_icon", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_GdkPixmap__P(pixmap), pixmap, 3, "gdk_window_set_icon", "GdkPixmap*");
  XEN_ASSERT_TYPE(XEN_GdkBitmap__P(mask), mask, 4, "gdk_window_set_icon", "GdkBitmap*");
  gdk_window_set_icon(XEN_TO_C_GdkWindow_(window), XEN_TO_C_GdkWindow_(icon_window), XEN_TO_C_GdkPixmap_(pixmap), XEN_TO_C_GdkBitmap_(mask));
  return(XEN_FALSE);
}

static XEN gxg_gdk_window_set_icon_name(XEN window, XEN name)
{
  #define H_gdk_window_set_icon_name "void gdk_window_set_icon_name(GdkWindow* window, gchar* name)"
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 1, "gdk_window_set_icon_name", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_gchar__P(name), name, 2, "gdk_window_set_icon_name", "gchar*");
  gdk_window_set_icon_name(XEN_TO_C_GdkWindow_(window), XEN_TO_C_gchar_(name));
  return(XEN_FALSE);
}

static XEN gxg_gdk_window_set_group(XEN window, XEN leader)
{
  #define H_gdk_window_set_group "void gdk_window_set_group(GdkWindow* window, GdkWindow* leader)"
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 1, "gdk_window_set_group", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(leader), leader, 2, "gdk_window_set_group", "GdkWindow*");
  gdk_window_set_group(XEN_TO_C_GdkWindow_(window), XEN_TO_C_GdkWindow_(leader));
  return(XEN_FALSE);
}

static XEN gxg_gdk_window_set_decorations(XEN window, XEN decorations)
{
  #define H_gdk_window_set_decorations "void gdk_window_set_decorations(GdkWindow* window, GdkWMDecoration decorations)"
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 1, "gdk_window_set_decorations", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_GdkWMDecoration_P(decorations), decorations, 2, "gdk_window_set_decorations", "GdkWMDecoration");
  gdk_window_set_decorations(XEN_TO_C_GdkWindow_(window), XEN_TO_C_GdkWMDecoration(decorations));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gdk_window_get_decorations(XEN window, XEN decorations)
{
  #define H_gdk_window_get_decorations "gboolean gdk_window_get_decorations(GdkWindow* window, GdkWMDecoration* decorations)"
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 1, "gdk_window_get_decorations", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_GdkWMDecoration__P(decorations), decorations, 2, "gdk_window_get_decorations", "GdkWMDecoration*");
  return(C_TO_XEN_gboolean(gdk_window_get_decorations(XEN_TO_C_GdkWindow_(window), XEN_TO_C_GdkWMDecoration_(decorations))));
}
#endif

static XEN gxg_gdk_window_set_functions(XEN window, XEN functions)
{
  #define H_gdk_window_set_functions "void gdk_window_set_functions(GdkWindow* window, GdkWMFunction functions)"
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 1, "gdk_window_set_functions", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_GdkWMFunction_P(functions), functions, 2, "gdk_window_set_functions", "GdkWMFunction");
  gdk_window_set_functions(XEN_TO_C_GdkWindow_(window), XEN_TO_C_GdkWMFunction(functions));
  return(XEN_FALSE);
}

static XEN gxg_gdk_window_get_toplevels(void)
{
  #define H_gdk_window_get_toplevels "GList* gdk_window_get_toplevels( void)"
  return(C_TO_XEN_GList_(gdk_window_get_toplevels()));
}

#if (!HAVE_GTK_1)
static XEN gxg_gdk_window_iconify(XEN window)
{
  #define H_gdk_window_iconify "void gdk_window_iconify(GdkWindow* window)"
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 1, "gdk_window_iconify", "GdkWindow*");
  gdk_window_iconify(XEN_TO_C_GdkWindow_(window));
  return(XEN_FALSE);
}

static XEN gxg_gdk_window_deiconify(XEN window)
{
  #define H_gdk_window_deiconify "void gdk_window_deiconify(GdkWindow* window)"
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 1, "gdk_window_deiconify", "GdkWindow*");
  gdk_window_deiconify(XEN_TO_C_GdkWindow_(window));
  return(XEN_FALSE);
}

static XEN gxg_gdk_window_stick(XEN window)
{
  #define H_gdk_window_stick "void gdk_window_stick(GdkWindow* window)"
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 1, "gdk_window_stick", "GdkWindow*");
  gdk_window_stick(XEN_TO_C_GdkWindow_(window));
  return(XEN_FALSE);
}

static XEN gxg_gdk_window_unstick(XEN window)
{
  #define H_gdk_window_unstick "void gdk_window_unstick(GdkWindow* window)"
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 1, "gdk_window_unstick", "GdkWindow*");
  gdk_window_unstick(XEN_TO_C_GdkWindow_(window));
  return(XEN_FALSE);
}

static XEN gxg_gdk_window_maximize(XEN window)
{
  #define H_gdk_window_maximize "void gdk_window_maximize(GdkWindow* window)"
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 1, "gdk_window_maximize", "GdkWindow*");
  gdk_window_maximize(XEN_TO_C_GdkWindow_(window));
  return(XEN_FALSE);
}

static XEN gxg_gdk_window_unmaximize(XEN window)
{
  #define H_gdk_window_unmaximize "void gdk_window_unmaximize(GdkWindow* window)"
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 1, "gdk_window_unmaximize", "GdkWindow*");
  gdk_window_unmaximize(XEN_TO_C_GdkWindow_(window));
  return(XEN_FALSE);
}
#endif

static XEN gxg_gdk_window_register_dnd(XEN window)
{
  #define H_gdk_window_register_dnd "void gdk_window_register_dnd(GdkWindow* window)"
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 1, "gdk_window_register_dnd", "GdkWindow*");
  gdk_window_register_dnd(XEN_TO_C_GdkWindow_(window));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gdk_window_begin_resize_drag(XEN window, XEN edge, XEN button, XEN root_x, XEN root_y, XEN timestamp)
{
  #define H_gdk_window_begin_resize_drag "void gdk_window_begin_resize_drag(GdkWindow* window, GdkWindowEdge edge, \
gint button, gint root_x, gint root_y, guint32 timestamp)"
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 1, "gdk_window_begin_resize_drag", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_GdkWindowEdge_P(edge), edge, 2, "gdk_window_begin_resize_drag", "GdkWindowEdge");
  XEN_ASSERT_TYPE(XEN_gint_P(button), button, 3, "gdk_window_begin_resize_drag", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(root_x), root_x, 4, "gdk_window_begin_resize_drag", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(root_y), root_y, 5, "gdk_window_begin_resize_drag", "gint");
  XEN_ASSERT_TYPE(XEN_guint32_P(timestamp), timestamp, 6, "gdk_window_begin_resize_drag", "guint32");
  gdk_window_begin_resize_drag(XEN_TO_C_GdkWindow_(window), XEN_TO_C_GdkWindowEdge(edge), XEN_TO_C_gint(button), XEN_TO_C_gint(root_x), 
                               XEN_TO_C_gint(root_y), XEN_TO_C_guint32(timestamp));
  return(XEN_FALSE);
}

static XEN gxg_gdk_window_begin_move_drag(XEN window, XEN button, XEN root_x, XEN root_y, XEN timestamp)
{
  #define H_gdk_window_begin_move_drag "void gdk_window_begin_move_drag(GdkWindow* window, gint button, \
gint root_x, gint root_y, guint32 timestamp)"
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 1, "gdk_window_begin_move_drag", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_gint_P(button), button, 2, "gdk_window_begin_move_drag", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(root_x), root_x, 3, "gdk_window_begin_move_drag", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(root_y), root_y, 4, "gdk_window_begin_move_drag", "gint");
  XEN_ASSERT_TYPE(XEN_guint32_P(timestamp), timestamp, 5, "gdk_window_begin_move_drag", "guint32");
  gdk_window_begin_move_drag(XEN_TO_C_GdkWindow_(window), XEN_TO_C_gint(button), XEN_TO_C_gint(root_x), XEN_TO_C_gint(root_y), 
                             XEN_TO_C_guint32(timestamp));
  return(XEN_FALSE);
}

static XEN gxg_gdk_window_invalidate_rect(XEN window, XEN rect, XEN invalidate_children)
{
  #define H_gdk_window_invalidate_rect "void gdk_window_invalidate_rect(GdkWindow* window, GdkRectangle* rect, \
gboolean invalidate_children)"
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 1, "gdk_window_invalidate_rect", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_GdkRectangle__P(rect), rect, 2, "gdk_window_invalidate_rect", "GdkRectangle*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(invalidate_children), invalidate_children, 3, "gdk_window_invalidate_rect", "gboolean");
  gdk_window_invalidate_rect(XEN_TO_C_GdkWindow_(window), XEN_TO_C_GdkRectangle_(rect), XEN_TO_C_gboolean(invalidate_children));
  return(XEN_FALSE);
}

static XEN gxg_gdk_window_invalidate_region(XEN window, XEN region, XEN invalidate_children)
{
  #define H_gdk_window_invalidate_region "void gdk_window_invalidate_region(GdkWindow* window, GdkRegion* region, \
gboolean invalidate_children)"
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 1, "gdk_window_invalidate_region", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_GdkRegion__P(region), region, 2, "gdk_window_invalidate_region", "GdkRegion*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(invalidate_children), invalidate_children, 3, "gdk_window_invalidate_region", "gboolean");
  gdk_window_invalidate_region(XEN_TO_C_GdkWindow_(window), XEN_TO_C_GdkRegion_(region), XEN_TO_C_gboolean(invalidate_children));
  return(XEN_FALSE);
}

static XEN gxg_gdk_window_invalidate_maybe_recurse(XEN window, XEN region, XEN func, XEN func_data)
{
  #define H_gdk_window_invalidate_maybe_recurse "void gdk_window_invalidate_maybe_recurse(GdkWindow* window, \
GdkRegion* region, lambda2 func, lambda_data func_data)"
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 1, "gdk_window_invalidate_maybe_recurse", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_GdkRegion__P(region), region, 2, "gdk_window_invalidate_maybe_recurse", "GdkRegion*");
  XEN_ASSERT_TYPE(XEN_lambda2_P(func), func, 3, "gdk_window_invalidate_maybe_recurse", "lambda2");
  if (XEN_NOT_BOUND_P(func_data)) func_data = XEN_FALSE; 
  else XEN_ASSERT_TYPE(XEN_lambda_data_P(func_data), func_data, 4, "gdk_window_invalidate_maybe_recurse", "lambda_data");
  {
    int loc;
    XEN gxg_ptr = XEN_LIST_5(func, func_data, XEN_FALSE, XEN_FALSE, XEN_FALSE);
    loc = xm_protect(gxg_ptr);
    XEN_LIST_SET(gxg_ptr, 2, C_TO_XEN_INT(loc));
    gdk_window_invalidate_maybe_recurse(XEN_TO_C_GdkWindow_(window), XEN_TO_C_GdkRegion_(region), XEN_TO_C_lambda2(func), XEN_TO_C_lambda_data(func_data));
    xm_unprotect_at(loc);
    return(XEN_FALSE);
   }
}

static XEN gxg_gdk_window_get_update_area(XEN window)
{
  #define H_gdk_window_get_update_area "GdkRegion* gdk_window_get_update_area(GdkWindow* window)"
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 1, "gdk_window_get_update_area", "GdkWindow*");
  return(C_TO_XEN_GdkRegion_(gdk_window_get_update_area(XEN_TO_C_GdkWindow_(window))));
}

static XEN gxg_gdk_window_freeze_updates(XEN window)
{
  #define H_gdk_window_freeze_updates "void gdk_window_freeze_updates(GdkWindow* window)"
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 1, "gdk_window_freeze_updates", "GdkWindow*");
  gdk_window_freeze_updates(XEN_TO_C_GdkWindow_(window));
  return(XEN_FALSE);
}

static XEN gxg_gdk_window_thaw_updates(XEN window)
{
  #define H_gdk_window_thaw_updates "void gdk_window_thaw_updates(GdkWindow* window)"
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 1, "gdk_window_thaw_updates", "GdkWindow*");
  gdk_window_thaw_updates(XEN_TO_C_GdkWindow_(window));
  return(XEN_FALSE);
}

static XEN gxg_gdk_window_process_all_updates(void)
{
  #define H_gdk_window_process_all_updates "void gdk_window_process_all_updates( void)"
  gdk_window_process_all_updates();
  return(XEN_FALSE);
}

static XEN gxg_gdk_window_process_updates(XEN window, XEN update_children)
{
  #define H_gdk_window_process_updates "void gdk_window_process_updates(GdkWindow* window, gboolean update_children)"
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 1, "gdk_window_process_updates", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(update_children), update_children, 2, "gdk_window_process_updates", "gboolean");
  gdk_window_process_updates(XEN_TO_C_GdkWindow_(window), XEN_TO_C_gboolean(update_children));
  return(XEN_FALSE);
}

static XEN gxg_gdk_window_set_debug_updates(XEN setting)
{
  #define H_gdk_window_set_debug_updates "void gdk_window_set_debug_updates(gboolean setting)"
  XEN_ASSERT_TYPE(XEN_gboolean_P(setting), setting, 1, "gdk_window_set_debug_updates", "gboolean");
  gdk_window_set_debug_updates(XEN_TO_C_gboolean(setting));
  return(XEN_FALSE);
}

static XEN gxg_gdk_window_constrain_size(XEN geometry, XEN flags, XEN width, XEN height, XEN new_width, XEN new_height)
{
  #define H_gdk_window_constrain_size "void gdk_window_constrain_size(GdkGeometry* geometry, guint flags, \
gint width, gint height, gint* [new_width], gint* [new_height])"
  gint ref_new_width;
  gint ref_new_height;
  XEN_ASSERT_TYPE(XEN_GdkGeometry__P(geometry), geometry, 1, "gdk_window_constrain_size", "GdkGeometry*");
  XEN_ASSERT_TYPE(XEN_guint_P(flags), flags, 2, "gdk_window_constrain_size", "guint");
  XEN_ASSERT_TYPE(XEN_gint_P(width), width, 3, "gdk_window_constrain_size", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(height), height, 4, "gdk_window_constrain_size", "gint");
  gdk_window_constrain_size(XEN_TO_C_GdkGeometry_(geometry), XEN_TO_C_guint(flags), XEN_TO_C_gint(width), XEN_TO_C_gint(height), 
                            &ref_new_width, &ref_new_height);
  return(XEN_LIST_2(C_TO_XEN_gint(ref_new_width), C_TO_XEN_gint(ref_new_height)));
}

static XEN gxg_gdk_window_get_internal_paint_info(XEN window, XEN real_drawable, XEN x_offset, XEN y_offset)
{
  #define H_gdk_window_get_internal_paint_info "void gdk_window_get_internal_paint_info(GdkWindow* window, \
GdkDrawable** [real_drawable], gint* [x_offset], gint* [y_offset])"
  GdkDrawable* ref_real_drawable;
  gint ref_x_offset;
  gint ref_y_offset;
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 1, "gdk_window_get_internal_paint_info", "GdkWindow*");
  gdk_window_get_internal_paint_info(XEN_TO_C_GdkWindow_(window), &ref_real_drawable, &ref_x_offset, &ref_y_offset);
  return(XEN_LIST_3(C_TO_XEN_GdkDrawable_(ref_real_drawable), C_TO_XEN_gint(ref_x_offset), C_TO_XEN_gint(ref_y_offset)));
}

static XEN gxg_gtk_window_group_add_window(XEN window_group, XEN window)
{
  #define H_gtk_window_group_add_window "void gtk_window_group_add_window(GtkWindowGroup* window_group, \
GtkWindow* window)"
  XEN_ASSERT_TYPE(XEN_GtkWindowGroup__P(window_group), window_group, 1, "gtk_window_group_add_window", "GtkWindowGroup*");
  XEN_ASSERT_TYPE(XEN_GtkWindow__P(window), window, 2, "gtk_window_group_add_window", "GtkWindow*");
  gtk_window_group_add_window(XEN_TO_C_GtkWindowGroup_(window_group), XEN_TO_C_GtkWindow_(window));
  return(XEN_FALSE);
}

static XEN gxg_gtk_window_group_remove_window(XEN window_group, XEN window)
{
  #define H_gtk_window_group_remove_window "void gtk_window_group_remove_window(GtkWindowGroup* window_group, \
GtkWindow* window)"
  XEN_ASSERT_TYPE(XEN_GtkWindowGroup__P(window_group), window_group, 1, "gtk_window_group_remove_window", "GtkWindowGroup*");
  XEN_ASSERT_TYPE(XEN_GtkWindow__P(window), window, 2, "gtk_window_group_remove_window", "GtkWindow*");
  gtk_window_group_remove_window(XEN_TO_C_GtkWindowGroup_(window_group), XEN_TO_C_GtkWindow_(window));
  return(XEN_FALSE);
}
#endif

static XEN gxg_gtk_window_remove_embedded_xid(XEN window, XEN xid)
{
  #define H_gtk_window_remove_embedded_xid "void gtk_window_remove_embedded_xid(GtkWindow* window, guint xid)"
  XEN_ASSERT_TYPE(XEN_GtkWindow__P(window), window, 1, "gtk_window_remove_embedded_xid", "GtkWindow*");
  XEN_ASSERT_TYPE(XEN_guint_P(xid), xid, 2, "gtk_window_remove_embedded_xid", "guint");
  gtk_window_remove_embedded_xid(XEN_TO_C_GtkWindow_(window), XEN_TO_C_guint(xid));
  return(XEN_FALSE);
}

static XEN gxg_gtk_window_add_embedded_xid(XEN window, XEN xid)
{
  #define H_gtk_window_add_embedded_xid "void gtk_window_add_embedded_xid(GtkWindow* window, guint xid)"
  XEN_ASSERT_TYPE(XEN_GtkWindow__P(window), window, 1, "gtk_window_add_embedded_xid", "GtkWindow*");
  XEN_ASSERT_TYPE(XEN_guint_P(xid), xid, 2, "gtk_window_add_embedded_xid", "guint");
  gtk_window_add_embedded_xid(XEN_TO_C_GtkWindow_(window), XEN_TO_C_guint(xid));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gdk_window_set_type_hint(XEN window, XEN hint)
{
  #define H_gdk_window_set_type_hint "void gdk_window_set_type_hint(GdkWindow* window, GdkWindowTypeHint hint)"
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 1, "gdk_window_set_type_hint", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_GdkWindowTypeHint_P(hint), hint, 2, "gdk_window_set_type_hint", "GdkWindowTypeHint");
  gdk_window_set_type_hint(XEN_TO_C_GdkWindow_(window), XEN_TO_C_GdkWindowTypeHint(hint));
  return(XEN_FALSE);
}

static XEN gxg_gdk_window_set_modal_hint(XEN window, XEN modal)
{
  #define H_gdk_window_set_modal_hint "void gdk_window_set_modal_hint(GdkWindow* window, gboolean modal)"
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 1, "gdk_window_set_modal_hint", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(modal), modal, 2, "gdk_window_set_modal_hint", "gboolean");
  gdk_window_set_modal_hint(XEN_TO_C_GdkWindow_(window), XEN_TO_C_gboolean(modal));
  return(XEN_FALSE);
}
#endif

static XEN gxg_gdk_window_set_geometry_hints(XEN window, XEN geometry, XEN geom_mask)
{
  #define H_gdk_window_set_geometry_hints "void gdk_window_set_geometry_hints(GdkWindow* window, GdkGeometry* geometry, \
GdkWindowHints geom_mask)"
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 1, "gdk_window_set_geometry_hints", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_GdkGeometry__P(geometry), geometry, 2, "gdk_window_set_geometry_hints", "GdkGeometry*");
  XEN_ASSERT_TYPE(XEN_GdkWindowHints_P(geom_mask), geom_mask, 3, "gdk_window_set_geometry_hints", "GdkWindowHints");
  gdk_window_set_geometry_hints(XEN_TO_C_GdkWindow_(window), XEN_TO_C_GdkGeometry_(geometry), XEN_TO_C_GdkWindowHints(geom_mask));
  return(XEN_FALSE);
}

static XEN gxg_gdk_set_sm_client_id(XEN sm_client_id)
{
  #define H_gdk_set_sm_client_id "void gdk_set_sm_client_id(gchar* sm_client_id)"
  XEN_ASSERT_TYPE(XEN_gchar__P(sm_client_id), sm_client_id, 1, "gdk_set_sm_client_id", "gchar*");
  gdk_set_sm_client_id(XEN_TO_C_gchar_(sm_client_id));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gdk_window_begin_paint_rect(XEN window, XEN rectangle)
{
  #define H_gdk_window_begin_paint_rect "void gdk_window_begin_paint_rect(GdkWindow* window, GdkRectangle* rectangle)"
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 1, "gdk_window_begin_paint_rect", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_GdkRectangle__P(rectangle), rectangle, 2, "gdk_window_begin_paint_rect", "GdkRectangle*");
  gdk_window_begin_paint_rect(XEN_TO_C_GdkWindow_(window), XEN_TO_C_GdkRectangle_(rectangle));
  return(XEN_FALSE);
}

static XEN gxg_gdk_window_begin_paint_region(XEN window, XEN region)
{
  #define H_gdk_window_begin_paint_region "void gdk_window_begin_paint_region(GdkWindow* window, GdkRegion* region)"
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 1, "gdk_window_begin_paint_region", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_GdkRegion__P(region), region, 2, "gdk_window_begin_paint_region", "GdkRegion*");
  gdk_window_begin_paint_region(XEN_TO_C_GdkWindow_(window), XEN_TO_C_GdkRegion_(region));
  return(XEN_FALSE);
}

static XEN gxg_gdk_window_end_paint(XEN window)
{
  #define H_gdk_window_end_paint "void gdk_window_end_paint(GdkWindow* window)"
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 1, "gdk_window_end_paint", "GdkWindow*");
  gdk_window_end_paint(XEN_TO_C_GdkWindow_(window));
  return(XEN_FALSE);
}
#endif

static XEN gxg_gdk_window_set_title(XEN window, XEN title)
{
  #define H_gdk_window_set_title "void gdk_window_set_title(GdkWindow* window, gchar* title)"
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 1, "gdk_window_set_title", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_gchar__P(title), title, 2, "gdk_window_set_title", "gchar*");
  gdk_window_set_title(XEN_TO_C_GdkWindow_(window), XEN_TO_C_gchar_(title));
  return(XEN_FALSE);
}

static XEN gxg_gdk_window_set_role(XEN window, XEN role)
{
  #define H_gdk_window_set_role "void gdk_window_set_role(GdkWindow* window, gchar* role)"
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 1, "gdk_window_set_role", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_gchar__P(role), role, 2, "gdk_window_set_role", "gchar*");
  gdk_window_set_role(XEN_TO_C_GdkWindow_(window), XEN_TO_C_gchar_(role));
  return(XEN_FALSE);
}

static XEN gxg_gdk_window_set_transient_for(XEN window, XEN parent)
{
  #define H_gdk_window_set_transient_for "void gdk_window_set_transient_for(GdkWindow* window, GdkWindow* parent)"
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 1, "gdk_window_set_transient_for", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(parent), parent, 2, "gdk_window_set_transient_for", "GdkWindow*");
  gdk_window_set_transient_for(XEN_TO_C_GdkWindow_(window), XEN_TO_C_GdkWindow_(parent));
  return(XEN_FALSE);
}

static XEN gxg_gdk_window_set_background(XEN window, XEN color)
{
  #define H_gdk_window_set_background "void gdk_window_set_background(GdkWindow* window, GdkColor* color)"
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 1, "gdk_window_set_background", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_GdkColor__P(color), color, 2, "gdk_window_set_background", "GdkColor*");
  gdk_window_set_background(XEN_TO_C_GdkWindow_(window), XEN_TO_C_GdkColor_(color));
  return(XEN_FALSE);
}

static XEN gxg_gdk_window_set_back_pixmap(XEN window, XEN pixmap, XEN parent_relative)
{
  #define H_gdk_window_set_back_pixmap "void gdk_window_set_back_pixmap(GdkWindow* window, GdkPixmap* pixmap, \
gboolean parent_relative)"
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 1, "gdk_window_set_back_pixmap", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_GdkPixmap__P(pixmap), pixmap, 2, "gdk_window_set_back_pixmap", "GdkPixmap*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(parent_relative), parent_relative, 3, "gdk_window_set_back_pixmap", "gboolean");
  gdk_window_set_back_pixmap(XEN_TO_C_GdkWindow_(window), XEN_TO_C_GdkPixmap_(pixmap), XEN_TO_C_gboolean(parent_relative));
  return(XEN_FALSE);
}

static XEN gxg_gdk_window_set_cursor(XEN window, XEN cursor)
{
  #define H_gdk_window_set_cursor "void gdk_window_set_cursor(GdkWindow* window, GdkCursor* cursor)"
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 1, "gdk_window_set_cursor", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_GdkCursor__P(cursor), cursor, 2, "gdk_window_set_cursor", "GdkCursor*");
  gdk_window_set_cursor(XEN_TO_C_GdkWindow_(window), XEN_TO_C_GdkCursor_(cursor));
  return(XEN_FALSE);
}

static XEN gxg_gdk_window_get_user_data(XEN window, XEN data)
{
  #define H_gdk_window_get_user_data "void gdk_window_get_user_data(GdkWindow* window, gpointer* [data])"
  gpointer ref_data;
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 1, "gdk_window_get_user_data", "GdkWindow*");
  gdk_window_get_user_data(XEN_TO_C_GdkWindow_(window), &ref_data);
  return(XEN_LIST_1(C_TO_XEN_gpointer(ref_data)));
}

static XEN gxg_gdk_window_get_geometry(XEN window, XEN x, XEN y, XEN width, XEN height, XEN depth)
{
  #define H_gdk_window_get_geometry "void gdk_window_get_geometry(GdkWindow* window, gint* x, gint* y, \
gint* width, gint* height, gint* depth)"
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 1, "gdk_window_get_geometry", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_gint__P(x), x, 2, "gdk_window_get_geometry", "gint*");
  XEN_ASSERT_TYPE(XEN_gint__P(y), y, 3, "gdk_window_get_geometry", "gint*");
  XEN_ASSERT_TYPE(XEN_gint__P(width), width, 4, "gdk_window_get_geometry", "gint*");
  XEN_ASSERT_TYPE(XEN_gint__P(height), height, 5, "gdk_window_get_geometry", "gint*");
  XEN_ASSERT_TYPE(XEN_gint__P(depth), depth, 6, "gdk_window_get_geometry", "gint*");
  gdk_window_get_geometry(XEN_TO_C_GdkWindow_(window), XEN_TO_C_gint_(x), XEN_TO_C_gint_(y), XEN_TO_C_gint_(width), XEN_TO_C_gint_(height), 
                          XEN_TO_C_gint_(depth));
  return(XEN_FALSE);
}

static XEN gxg_gdk_window_get_position(XEN window, XEN x, XEN y)
{
  #define H_gdk_window_get_position "void gdk_window_get_position(GdkWindow* window, gint* [x], gint* [y])"
  gint ref_x;
  gint ref_y;
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 1, "gdk_window_get_position", "GdkWindow*");
  gdk_window_get_position(XEN_TO_C_GdkWindow_(window), &ref_x, &ref_y);
  return(XEN_LIST_2(C_TO_XEN_gint(ref_x), C_TO_XEN_gint(ref_y)));
}

static XEN gxg_gdk_window_get_origin(XEN window, XEN x, XEN y)
{
  #define H_gdk_window_get_origin "gint gdk_window_get_origin(GdkWindow* window, gint* x, gint* y)"
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 1, "gdk_window_get_origin", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_gint__P(x), x, 2, "gdk_window_get_origin", "gint*");
  XEN_ASSERT_TYPE(XEN_gint__P(y), y, 3, "gdk_window_get_origin", "gint*");
  return(C_TO_XEN_gint(gdk_window_get_origin(XEN_TO_C_GdkWindow_(window), XEN_TO_C_gint_(x), XEN_TO_C_gint_(y))));
}

#if (!HAVE_GTK_1)
static XEN gxg_gdk_set_pointer_hooks(XEN new_hooks)
{
  #define H_gdk_set_pointer_hooks "GdkPointerHooks* gdk_set_pointer_hooks(GdkPointerHooks* new_hooks)"
  XEN_ASSERT_TYPE(XEN_GdkPointerHooks__P(new_hooks) || XEN_FALSE_P(new_hooks), new_hooks, 1, "gdk_set_pointer_hooks", "GdkPointerHooks*");
  return(C_TO_XEN_GdkPointerHooks_(gdk_set_pointer_hooks(XEN_TO_C_GdkPointerHooks_(new_hooks))));
}

static XEN gxg_gdk_get_default_root_window(void)
{
  #define H_gdk_get_default_root_window "GdkWindow* gdk_get_default_root_window( void)"
  return(C_TO_XEN_GdkWindow_(gdk_get_default_root_window()));
}

static XEN gxg_gdk_pixbuf_error_quark(void)
{
  #define H_gdk_pixbuf_error_quark "GQuark gdk_pixbuf_error_quark( void)"
  return(C_TO_XEN_GQuark(gdk_pixbuf_error_quark()));
}

static XEN gxg_gdk_pixbuf_get_type(void)
{
  #define H_gdk_pixbuf_get_type "GType gdk_pixbuf_get_type( void)"
  return(C_TO_XEN_GType(gdk_pixbuf_get_type()));
}

static XEN gxg_gdk_pixbuf_get_colorspace(XEN pixbuf)
{
  #define H_gdk_pixbuf_get_colorspace "GdkColorspace gdk_pixbuf_get_colorspace(GdkPixbuf* pixbuf)"
  XEN_ASSERT_TYPE(XEN_GdkPixbuf__P(pixbuf), pixbuf, 1, "gdk_pixbuf_get_colorspace", "GdkPixbuf*");
  return(C_TO_XEN_GdkColorspace(gdk_pixbuf_get_colorspace(XEN_TO_C_GdkPixbuf_(pixbuf))));
}

static XEN gxg_gdk_pixbuf_get_n_channels(XEN pixbuf)
{
  #define H_gdk_pixbuf_get_n_channels "int gdk_pixbuf_get_n_channels(GdkPixbuf* pixbuf)"
  XEN_ASSERT_TYPE(XEN_GdkPixbuf__P(pixbuf), pixbuf, 1, "gdk_pixbuf_get_n_channels", "GdkPixbuf*");
  return(C_TO_XEN_int(gdk_pixbuf_get_n_channels(XEN_TO_C_GdkPixbuf_(pixbuf))));
}

static XEN gxg_gdk_pixbuf_get_has_alpha(XEN pixbuf)
{
  #define H_gdk_pixbuf_get_has_alpha "gboolean gdk_pixbuf_get_has_alpha(GdkPixbuf* pixbuf)"
  XEN_ASSERT_TYPE(XEN_GdkPixbuf__P(pixbuf), pixbuf, 1, "gdk_pixbuf_get_has_alpha", "GdkPixbuf*");
  return(C_TO_XEN_gboolean(gdk_pixbuf_get_has_alpha(XEN_TO_C_GdkPixbuf_(pixbuf))));
}

static XEN gxg_gdk_pixbuf_get_bits_per_sample(XEN pixbuf)
{
  #define H_gdk_pixbuf_get_bits_per_sample "int gdk_pixbuf_get_bits_per_sample(GdkPixbuf* pixbuf)"
  XEN_ASSERT_TYPE(XEN_GdkPixbuf__P(pixbuf), pixbuf, 1, "gdk_pixbuf_get_bits_per_sample", "GdkPixbuf*");
  return(C_TO_XEN_int(gdk_pixbuf_get_bits_per_sample(XEN_TO_C_GdkPixbuf_(pixbuf))));
}

static XEN gxg_gdk_pixbuf_get_pixels(XEN pixbuf)
{
  #define H_gdk_pixbuf_get_pixels "guchar* gdk_pixbuf_get_pixels(GdkPixbuf* pixbuf)"
  XEN_ASSERT_TYPE(XEN_GdkPixbuf__P(pixbuf), pixbuf, 1, "gdk_pixbuf_get_pixels", "GdkPixbuf*");
  return(C_TO_XEN_guchar_(gdk_pixbuf_get_pixels(XEN_TO_C_GdkPixbuf_(pixbuf))));
}

static XEN gxg_gdk_pixbuf_get_width(XEN pixbuf)
{
  #define H_gdk_pixbuf_get_width "int gdk_pixbuf_get_width(GdkPixbuf* pixbuf)"
  XEN_ASSERT_TYPE(XEN_GdkPixbuf__P(pixbuf), pixbuf, 1, "gdk_pixbuf_get_width", "GdkPixbuf*");
  return(C_TO_XEN_int(gdk_pixbuf_get_width(XEN_TO_C_GdkPixbuf_(pixbuf))));
}

static XEN gxg_gdk_pixbuf_get_height(XEN pixbuf)
{
  #define H_gdk_pixbuf_get_height "int gdk_pixbuf_get_height(GdkPixbuf* pixbuf)"
  XEN_ASSERT_TYPE(XEN_GdkPixbuf__P(pixbuf), pixbuf, 1, "gdk_pixbuf_get_height", "GdkPixbuf*");
  return(C_TO_XEN_int(gdk_pixbuf_get_height(XEN_TO_C_GdkPixbuf_(pixbuf))));
}

static XEN gxg_gdk_pixbuf_get_rowstride(XEN pixbuf)
{
  #define H_gdk_pixbuf_get_rowstride "int gdk_pixbuf_get_rowstride(GdkPixbuf* pixbuf)"
  XEN_ASSERT_TYPE(XEN_GdkPixbuf__P(pixbuf), pixbuf, 1, "gdk_pixbuf_get_rowstride", "GdkPixbuf*");
  return(C_TO_XEN_int(gdk_pixbuf_get_rowstride(XEN_TO_C_GdkPixbuf_(pixbuf))));
}

static XEN gxg_gdk_pixbuf_new(XEN colorspace, XEN has_alpha, XEN bits_per_sample, XEN width, XEN height)
{
  #define H_gdk_pixbuf_new "GdkPixbuf* gdk_pixbuf_new(GdkColorspace colorspace, gboolean has_alpha, int bits_per_sample, \
int width, int height)"
  XEN_ASSERT_TYPE(XEN_GdkColorspace_P(colorspace), colorspace, 1, "gdk_pixbuf_new", "GdkColorspace");
  XEN_ASSERT_TYPE(XEN_gboolean_P(has_alpha), has_alpha, 2, "gdk_pixbuf_new", "gboolean");
  XEN_ASSERT_TYPE(XEN_int_P(bits_per_sample), bits_per_sample, 3, "gdk_pixbuf_new", "int");
  XEN_ASSERT_TYPE(XEN_int_P(width), width, 4, "gdk_pixbuf_new", "int");
  XEN_ASSERT_TYPE(XEN_int_P(height), height, 5, "gdk_pixbuf_new", "int");
  return(C_TO_XEN_GdkPixbuf_(gdk_pixbuf_new(XEN_TO_C_GdkColorspace(colorspace), XEN_TO_C_gboolean(has_alpha), XEN_TO_C_int(bits_per_sample), 
                                            XEN_TO_C_int(width), XEN_TO_C_int(height))));
}

static XEN gxg_gdk_pixbuf_copy(XEN pixbuf)
{
  #define H_gdk_pixbuf_copy "GdkPixbuf* gdk_pixbuf_copy(GdkPixbuf* pixbuf)"
  XEN_ASSERT_TYPE(XEN_GdkPixbuf__P(pixbuf), pixbuf, 1, "gdk_pixbuf_copy", "GdkPixbuf*");
  return(C_TO_XEN_GdkPixbuf_(gdk_pixbuf_copy(XEN_TO_C_GdkPixbuf_(pixbuf))));
}

static XEN gxg_gdk_pixbuf_new_subpixbuf(XEN src_pixbuf, XEN src_x, XEN src_y, XEN width, XEN height)
{
  #define H_gdk_pixbuf_new_subpixbuf "GdkPixbuf* gdk_pixbuf_new_subpixbuf(GdkPixbuf* src_pixbuf, int src_x, \
int src_y, int width, int height)"
  XEN_ASSERT_TYPE(XEN_GdkPixbuf__P(src_pixbuf), src_pixbuf, 1, "gdk_pixbuf_new_subpixbuf", "GdkPixbuf*");
  XEN_ASSERT_TYPE(XEN_int_P(src_x), src_x, 2, "gdk_pixbuf_new_subpixbuf", "int");
  XEN_ASSERT_TYPE(XEN_int_P(src_y), src_y, 3, "gdk_pixbuf_new_subpixbuf", "int");
  XEN_ASSERT_TYPE(XEN_int_P(width), width, 4, "gdk_pixbuf_new_subpixbuf", "int");
  XEN_ASSERT_TYPE(XEN_int_P(height), height, 5, "gdk_pixbuf_new_subpixbuf", "int");
  return(C_TO_XEN_GdkPixbuf_(gdk_pixbuf_new_subpixbuf(XEN_TO_C_GdkPixbuf_(src_pixbuf), XEN_TO_C_int(src_x), XEN_TO_C_int(src_y), 
                                                      XEN_TO_C_int(width), XEN_TO_C_int(height))));
}

static XEN gxg_gdk_pixbuf_new_from_file(XEN filename, XEN error)
{
  #define H_gdk_pixbuf_new_from_file "GdkPixbuf* gdk_pixbuf_new_from_file(char* filename, GError** [error])"
  GError* ref_error;
  XEN_ASSERT_TYPE(XEN_char__P(filename), filename, 1, "gdk_pixbuf_new_from_file", "char*");
  {
    XEN result = XEN_FALSE;
    result = C_TO_XEN_GdkPixbuf_(gdk_pixbuf_new_from_file(XEN_TO_C_char_(filename), &ref_error));
    return(XEN_LIST_2(result, C_TO_XEN_GError_(ref_error)));
   }
}

static XEN gxg_gdk_pixbuf_new_from_data(XEN data, XEN colorspace, XEN has_alpha, XEN bits_per_sample, XEN width, XEN height, XEN rowstride, XEN destroy_fn, XEN destroy_fn_data)
{
  #define H_gdk_pixbuf_new_from_data "GdkPixbuf* gdk_pixbuf_new_from_data(guchar* data, GdkColorspace colorspace, \
gboolean has_alpha, int bits_per_sample, int width, int height, int rowstride, GdkPixbufDestroyNotify destroy_fn, \
gpointer destroy_fn_data)"
  XEN_ASSERT_TYPE(XEN_guchar__P(data), data, 1, "gdk_pixbuf_new_from_data", "guchar*");
  XEN_ASSERT_TYPE(XEN_GdkColorspace_P(colorspace), colorspace, 2, "gdk_pixbuf_new_from_data", "GdkColorspace");
  XEN_ASSERT_TYPE(XEN_gboolean_P(has_alpha), has_alpha, 3, "gdk_pixbuf_new_from_data", "gboolean");
  XEN_ASSERT_TYPE(XEN_int_P(bits_per_sample), bits_per_sample, 4, "gdk_pixbuf_new_from_data", "int");
  XEN_ASSERT_TYPE(XEN_int_P(width), width, 5, "gdk_pixbuf_new_from_data", "int");
  XEN_ASSERT_TYPE(XEN_int_P(height), height, 6, "gdk_pixbuf_new_from_data", "int");
  XEN_ASSERT_TYPE(XEN_int_P(rowstride), rowstride, 7, "gdk_pixbuf_new_from_data", "int");
  XEN_ASSERT_TYPE(XEN_GdkPixbufDestroyNotify_P(destroy_fn), destroy_fn, 8, "gdk_pixbuf_new_from_data", "GdkPixbufDestroyNotify");
  XEN_ASSERT_TYPE(XEN_gpointer_P(destroy_fn_data), destroy_fn_data, 9, "gdk_pixbuf_new_from_data", "gpointer");
  return(C_TO_XEN_GdkPixbuf_(gdk_pixbuf_new_from_data(XEN_TO_C_guchar_(data), XEN_TO_C_GdkColorspace(colorspace), XEN_TO_C_gboolean(has_alpha), 
                                                      XEN_TO_C_int(bits_per_sample), XEN_TO_C_int(width), XEN_TO_C_int(height), 
                                                      XEN_TO_C_int(rowstride), XEN_TO_C_GdkPixbufDestroyNotify(destroy_fn), 
                                                      XEN_TO_C_gpointer(destroy_fn_data))));
}

static XEN gxg_gdk_pixbuf_new_from_xpm_data(XEN data)
{
  #define H_gdk_pixbuf_new_from_xpm_data "GdkPixbuf* gdk_pixbuf_new_from_xpm_data(char** data)"
  XEN_ASSERT_TYPE(XEN_char___P(data), data, 1, "gdk_pixbuf_new_from_xpm_data", "char**");
  return(C_TO_XEN_GdkPixbuf_(gdk_pixbuf_new_from_xpm_data((const char**)XEN_TO_C_char__(data))));
}

static XEN gxg_gdk_pixbuf_new_from_inline(XEN data_length, XEN data, XEN copy_pixels, XEN error)
{
  #define H_gdk_pixbuf_new_from_inline "GdkPixbuf* gdk_pixbuf_new_from_inline(gint data_length, guint8* data, \
gboolean copy_pixels, GError** [error])"
  GError* ref_error;
  XEN_ASSERT_TYPE(XEN_gint_P(data_length), data_length, 1, "gdk_pixbuf_new_from_inline", "gint");
  XEN_ASSERT_TYPE(XEN_guint8__P(data), data, 2, "gdk_pixbuf_new_from_inline", "guint8*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(copy_pixels), copy_pixels, 3, "gdk_pixbuf_new_from_inline", "gboolean");
  {
    XEN result = XEN_FALSE;
    result = C_TO_XEN_GdkPixbuf_(gdk_pixbuf_new_from_inline(XEN_TO_C_gint(data_length), XEN_TO_C_guint8_(data), XEN_TO_C_gboolean(copy_pixels), 
                                                            &ref_error));
    return(XEN_LIST_2(result, C_TO_XEN_GError_(ref_error)));
   }
}

static XEN gxg_gdk_pixbuf_fill(XEN pixbuf, XEN pixel)
{
  #define H_gdk_pixbuf_fill "void gdk_pixbuf_fill(GdkPixbuf* pixbuf, guint32 pixel)"
  XEN_ASSERT_TYPE(XEN_GdkPixbuf__P(pixbuf), pixbuf, 1, "gdk_pixbuf_fill", "GdkPixbuf*");
  XEN_ASSERT_TYPE(XEN_guint32_P(pixel), pixel, 2, "gdk_pixbuf_fill", "guint32");
  gdk_pixbuf_fill(XEN_TO_C_GdkPixbuf_(pixbuf), XEN_TO_C_guint32(pixel));
  return(XEN_FALSE);
}

static XEN gxg_gdk_pixbuf_save(XEN pixbuf, XEN filename, XEN type, XEN error)
{
  #define H_gdk_pixbuf_save "gboolean gdk_pixbuf_save(GdkPixbuf* pixbuf, char* filename, char* type, \
GError** [error], ...)"
  GError* ref_error;
  XEN_ASSERT_TYPE(XEN_GdkPixbuf__P(pixbuf), pixbuf, 1, "gdk_pixbuf_save", "GdkPixbuf*");
  XEN_ASSERT_TYPE(XEN_char__P(filename), filename, 2, "gdk_pixbuf_save", "char*");
  XEN_ASSERT_TYPE(XEN_char__P(type), type, 3, "gdk_pixbuf_save", "char*");
  {
    XEN result = XEN_FALSE;
    result = C_TO_XEN_gboolean(gdk_pixbuf_save(XEN_TO_C_GdkPixbuf_(pixbuf), XEN_TO_C_char_(filename), XEN_TO_C_char_(type), 
                                               &ref_error));
    return(XEN_LIST_2(result, C_TO_XEN_GError_(ref_error)));
   }
}

static XEN gxg_gdk_pixbuf_savev(XEN pixbuf, XEN filename, XEN type, XEN option_keys, XEN option_values, XEN error)
{
  #define H_gdk_pixbuf_savev "gboolean gdk_pixbuf_savev(GdkPixbuf* pixbuf, char* filename, char* type, \
char** option_keys, char** option_values, GError** [error])"
  GError* ref_error;
  XEN_ASSERT_TYPE(XEN_GdkPixbuf__P(pixbuf), pixbuf, 1, "gdk_pixbuf_savev", "GdkPixbuf*");
  XEN_ASSERT_TYPE(XEN_char__P(filename), filename, 2, "gdk_pixbuf_savev", "char*");
  XEN_ASSERT_TYPE(XEN_char__P(type), type, 3, "gdk_pixbuf_savev", "char*");
  XEN_ASSERT_TYPE(XEN_char___P(option_keys), option_keys, 4, "gdk_pixbuf_savev", "char**");
  XEN_ASSERT_TYPE(XEN_char___P(option_values), option_values, 5, "gdk_pixbuf_savev", "char**");
  {
    XEN result = XEN_FALSE;
    result = C_TO_XEN_gboolean(gdk_pixbuf_savev(XEN_TO_C_GdkPixbuf_(pixbuf), XEN_TO_C_char_(filename), XEN_TO_C_char_(type), 
                                                XEN_TO_C_char__(option_keys), XEN_TO_C_char__(option_values), &ref_error));
    return(XEN_LIST_2(result, C_TO_XEN_GError_(ref_error)));
   }
}

static XEN gxg_gdk_pixbuf_add_alpha(XEN pixbuf, XEN substitute_color, XEN r, XEN g, XEN b)
{
  #define H_gdk_pixbuf_add_alpha "GdkPixbuf* gdk_pixbuf_add_alpha(GdkPixbuf* pixbuf, gboolean substitute_color, \
guchar r, guchar g, guchar b)"
  XEN_ASSERT_TYPE(XEN_GdkPixbuf__P(pixbuf), pixbuf, 1, "gdk_pixbuf_add_alpha", "GdkPixbuf*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(substitute_color), substitute_color, 2, "gdk_pixbuf_add_alpha", "gboolean");
  XEN_ASSERT_TYPE(XEN_guchar_P(r), r, 3, "gdk_pixbuf_add_alpha", "guchar");
  XEN_ASSERT_TYPE(XEN_guchar_P(g), g, 4, "gdk_pixbuf_add_alpha", "guchar");
  XEN_ASSERT_TYPE(XEN_guchar_P(b), b, 5, "gdk_pixbuf_add_alpha", "guchar");
  return(C_TO_XEN_GdkPixbuf_(gdk_pixbuf_add_alpha(XEN_TO_C_GdkPixbuf_(pixbuf), XEN_TO_C_gboolean(substitute_color), XEN_TO_C_guchar(r), 
                                                  XEN_TO_C_guchar(g), XEN_TO_C_guchar(b))));
}

static XEN gxg_gdk_pixbuf_copy_area(XEN src_pixbuf, XEN src_x, XEN src_y, XEN width, XEN height, XEN dest_pixbuf, XEN dest_x, XEN dest_y)
{
  #define H_gdk_pixbuf_copy_area "void gdk_pixbuf_copy_area(GdkPixbuf* src_pixbuf, int src_x, int src_y, \
int width, int height, GdkPixbuf* dest_pixbuf, int dest_x, int dest_y)"
  XEN_ASSERT_TYPE(XEN_GdkPixbuf__P(src_pixbuf), src_pixbuf, 1, "gdk_pixbuf_copy_area", "GdkPixbuf*");
  XEN_ASSERT_TYPE(XEN_int_P(src_x), src_x, 2, "gdk_pixbuf_copy_area", "int");
  XEN_ASSERT_TYPE(XEN_int_P(src_y), src_y, 3, "gdk_pixbuf_copy_area", "int");
  XEN_ASSERT_TYPE(XEN_int_P(width), width, 4, "gdk_pixbuf_copy_area", "int");
  XEN_ASSERT_TYPE(XEN_int_P(height), height, 5, "gdk_pixbuf_copy_area", "int");
  XEN_ASSERT_TYPE(XEN_GdkPixbuf__P(dest_pixbuf), dest_pixbuf, 6, "gdk_pixbuf_copy_area", "GdkPixbuf*");
  XEN_ASSERT_TYPE(XEN_int_P(dest_x), dest_x, 7, "gdk_pixbuf_copy_area", "int");
  XEN_ASSERT_TYPE(XEN_int_P(dest_y), dest_y, 8, "gdk_pixbuf_copy_area", "int");
  gdk_pixbuf_copy_area(XEN_TO_C_GdkPixbuf_(src_pixbuf), XEN_TO_C_int(src_x), XEN_TO_C_int(src_y), XEN_TO_C_int(width), XEN_TO_C_int(height), 
                       XEN_TO_C_GdkPixbuf_(dest_pixbuf), XEN_TO_C_int(dest_x), XEN_TO_C_int(dest_y));
  return(XEN_FALSE);
}

static XEN gxg_gdk_pixbuf_saturate_and_pixelate(XEN src, XEN dest, XEN saturation, XEN pixelate)
{
  #define H_gdk_pixbuf_saturate_and_pixelate "void gdk_pixbuf_saturate_and_pixelate(GdkPixbuf* src, GdkPixbuf* dest, \
gfloat saturation, gboolean pixelate)"
  XEN_ASSERT_TYPE(XEN_GdkPixbuf__P(src), src, 1, "gdk_pixbuf_saturate_and_pixelate", "GdkPixbuf*");
  XEN_ASSERT_TYPE(XEN_GdkPixbuf__P(dest), dest, 2, "gdk_pixbuf_saturate_and_pixelate", "GdkPixbuf*");
  XEN_ASSERT_TYPE(XEN_gfloat_P(saturation), saturation, 3, "gdk_pixbuf_saturate_and_pixelate", "gfloat");
  XEN_ASSERT_TYPE(XEN_gboolean_P(pixelate), pixelate, 4, "gdk_pixbuf_saturate_and_pixelate", "gboolean");
  gdk_pixbuf_saturate_and_pixelate(XEN_TO_C_GdkPixbuf_(src), XEN_TO_C_GdkPixbuf_(dest), XEN_TO_C_gfloat(saturation), XEN_TO_C_gboolean(pixelate));
  return(XEN_FALSE);
}

static XEN gxg_gdk_pixbuf_scale(XEN arglist)
{
  #define H_gdk_pixbuf_scale "void gdk_pixbuf_scale(GdkPixbuf* src, GdkPixbuf* dest, int dest_x, int dest_y, \
int dest_width, int dest_height, double offset_x, double offset_y, double scale_x, double scale_y, GdkInterpType interp_type)"
  XEN src, dest, dest_x, dest_y, dest_width, dest_height, offset_x, offset_y, scale_x, scale_y, interp_type;
  src = XEN_LIST_REF(arglist, 0);
  dest = XEN_LIST_REF(arglist, 1);
  dest_x = XEN_LIST_REF(arglist, 2);
  dest_y = XEN_LIST_REF(arglist, 3);
  dest_width = XEN_LIST_REF(arglist, 4);
  dest_height = XEN_LIST_REF(arglist, 5);
  offset_x = XEN_LIST_REF(arglist, 6);
  offset_y = XEN_LIST_REF(arglist, 7);
  scale_x = XEN_LIST_REF(arglist, 8);
  scale_y = XEN_LIST_REF(arglist, 9);
  interp_type = XEN_LIST_REF(arglist, 10);
  XEN_ASSERT_TYPE(XEN_GdkPixbuf__P(src), src, 1, "gdk_pixbuf_scale", "GdkPixbuf*");
  XEN_ASSERT_TYPE(XEN_GdkPixbuf__P(dest), dest, 2, "gdk_pixbuf_scale", "GdkPixbuf*");
  XEN_ASSERT_TYPE(XEN_int_P(dest_x), dest_x, 3, "gdk_pixbuf_scale", "int");
  XEN_ASSERT_TYPE(XEN_int_P(dest_y), dest_y, 4, "gdk_pixbuf_scale", "int");
  XEN_ASSERT_TYPE(XEN_int_P(dest_width), dest_width, 5, "gdk_pixbuf_scale", "int");
  XEN_ASSERT_TYPE(XEN_int_P(dest_height), dest_height, 6, "gdk_pixbuf_scale", "int");
  XEN_ASSERT_TYPE(XEN_double_P(offset_x), offset_x, 7, "gdk_pixbuf_scale", "double");
  XEN_ASSERT_TYPE(XEN_double_P(offset_y), offset_y, 8, "gdk_pixbuf_scale", "double");
  XEN_ASSERT_TYPE(XEN_double_P(scale_x), scale_x, 9, "gdk_pixbuf_scale", "double");
  XEN_ASSERT_TYPE(XEN_double_P(scale_y), scale_y, 10, "gdk_pixbuf_scale", "double");
  XEN_ASSERT_TYPE(XEN_GdkInterpType_P(interp_type), interp_type, 11, "gdk_pixbuf_scale", "GdkInterpType");
  gdk_pixbuf_scale(XEN_TO_C_GdkPixbuf_(src), XEN_TO_C_GdkPixbuf_(dest), XEN_TO_C_int(dest_x), XEN_TO_C_int(dest_y), XEN_TO_C_int(dest_width), 
                   XEN_TO_C_int(dest_height), XEN_TO_C_double(offset_x), XEN_TO_C_double(offset_y), XEN_TO_C_double(scale_x), 
                   XEN_TO_C_double(scale_y), XEN_TO_C_GdkInterpType(interp_type));
  return(XEN_FALSE);
}

static XEN gxg_gdk_pixbuf_composite(XEN arglist)
{
  #define H_gdk_pixbuf_composite "void gdk_pixbuf_composite(GdkPixbuf* src, GdkPixbuf* dest, int dest_x, \
int dest_y, int dest_width, int dest_height, double offset_x, double offset_y, double scale_x, double scale_y, \
GdkInterpType interp_type, int overall_alpha)"
  XEN src, dest, dest_x, dest_y, dest_width, dest_height, offset_x, offset_y, scale_x, scale_y, interp_type, overall_alpha;
  src = XEN_LIST_REF(arglist, 0);
  dest = XEN_LIST_REF(arglist, 1);
  dest_x = XEN_LIST_REF(arglist, 2);
  dest_y = XEN_LIST_REF(arglist, 3);
  dest_width = XEN_LIST_REF(arglist, 4);
  dest_height = XEN_LIST_REF(arglist, 5);
  offset_x = XEN_LIST_REF(arglist, 6);
  offset_y = XEN_LIST_REF(arglist, 7);
  scale_x = XEN_LIST_REF(arglist, 8);
  scale_y = XEN_LIST_REF(arglist, 9);
  interp_type = XEN_LIST_REF(arglist, 10);
  overall_alpha = XEN_LIST_REF(arglist, 11);
  XEN_ASSERT_TYPE(XEN_GdkPixbuf__P(src), src, 1, "gdk_pixbuf_composite", "GdkPixbuf*");
  XEN_ASSERT_TYPE(XEN_GdkPixbuf__P(dest), dest, 2, "gdk_pixbuf_composite", "GdkPixbuf*");
  XEN_ASSERT_TYPE(XEN_int_P(dest_x), dest_x, 3, "gdk_pixbuf_composite", "int");
  XEN_ASSERT_TYPE(XEN_int_P(dest_y), dest_y, 4, "gdk_pixbuf_composite", "int");
  XEN_ASSERT_TYPE(XEN_int_P(dest_width), dest_width, 5, "gdk_pixbuf_composite", "int");
  XEN_ASSERT_TYPE(XEN_int_P(dest_height), dest_height, 6, "gdk_pixbuf_composite", "int");
  XEN_ASSERT_TYPE(XEN_double_P(offset_x), offset_x, 7, "gdk_pixbuf_composite", "double");
  XEN_ASSERT_TYPE(XEN_double_P(offset_y), offset_y, 8, "gdk_pixbuf_composite", "double");
  XEN_ASSERT_TYPE(XEN_double_P(scale_x), scale_x, 9, "gdk_pixbuf_composite", "double");
  XEN_ASSERT_TYPE(XEN_double_P(scale_y), scale_y, 10, "gdk_pixbuf_composite", "double");
  XEN_ASSERT_TYPE(XEN_GdkInterpType_P(interp_type), interp_type, 11, "gdk_pixbuf_composite", "GdkInterpType");
  XEN_ASSERT_TYPE(XEN_int_P(overall_alpha), overall_alpha, 12, "gdk_pixbuf_composite", "int");
  gdk_pixbuf_composite(XEN_TO_C_GdkPixbuf_(src), XEN_TO_C_GdkPixbuf_(dest), XEN_TO_C_int(dest_x), XEN_TO_C_int(dest_y), XEN_TO_C_int(dest_width), 
                       XEN_TO_C_int(dest_height), XEN_TO_C_double(offset_x), XEN_TO_C_double(offset_y), XEN_TO_C_double(scale_x), 
                       XEN_TO_C_double(scale_y), XEN_TO_C_GdkInterpType(interp_type), XEN_TO_C_int(overall_alpha));
  return(XEN_FALSE);
}

static XEN gxg_gdk_pixbuf_composite_color(XEN arglist)
{
  #define H_gdk_pixbuf_composite_color "void gdk_pixbuf_composite_color(GdkPixbuf* src, GdkPixbuf* dest, \
int dest_x, int dest_y, int dest_width, int dest_height, double offset_x, double offset_y, double scale_x, \
double scale_y, GdkInterpType interp_type, int overall_alpha, int check_x, int check_y, int check_size, \
guint32 color1, guint32 color2)"
  XEN src, dest, dest_x, dest_y, dest_width, dest_height, offset_x, offset_y, scale_x, scale_y, interp_type, overall_alpha, check_x, check_y, check_size, color1, color2;
  src = XEN_LIST_REF(arglist, 0);
  dest = XEN_LIST_REF(arglist, 1);
  dest_x = XEN_LIST_REF(arglist, 2);
  dest_y = XEN_LIST_REF(arglist, 3);
  dest_width = XEN_LIST_REF(arglist, 4);
  dest_height = XEN_LIST_REF(arglist, 5);
  offset_x = XEN_LIST_REF(arglist, 6);
  offset_y = XEN_LIST_REF(arglist, 7);
  scale_x = XEN_LIST_REF(arglist, 8);
  scale_y = XEN_LIST_REF(arglist, 9);
  interp_type = XEN_LIST_REF(arglist, 10);
  overall_alpha = XEN_LIST_REF(arglist, 11);
  check_x = XEN_LIST_REF(arglist, 12);
  check_y = XEN_LIST_REF(arglist, 13);
  check_size = XEN_LIST_REF(arglist, 14);
  color1 = XEN_LIST_REF(arglist, 15);
  color2 = XEN_LIST_REF(arglist, 16);
  XEN_ASSERT_TYPE(XEN_GdkPixbuf__P(src), src, 1, "gdk_pixbuf_composite_color", "GdkPixbuf*");
  XEN_ASSERT_TYPE(XEN_GdkPixbuf__P(dest), dest, 2, "gdk_pixbuf_composite_color", "GdkPixbuf*");
  XEN_ASSERT_TYPE(XEN_int_P(dest_x), dest_x, 3, "gdk_pixbuf_composite_color", "int");
  XEN_ASSERT_TYPE(XEN_int_P(dest_y), dest_y, 4, "gdk_pixbuf_composite_color", "int");
  XEN_ASSERT_TYPE(XEN_int_P(dest_width), dest_width, 5, "gdk_pixbuf_composite_color", "int");
  XEN_ASSERT_TYPE(XEN_int_P(dest_height), dest_height, 6, "gdk_pixbuf_composite_color", "int");
  XEN_ASSERT_TYPE(XEN_double_P(offset_x), offset_x, 7, "gdk_pixbuf_composite_color", "double");
  XEN_ASSERT_TYPE(XEN_double_P(offset_y), offset_y, 8, "gdk_pixbuf_composite_color", "double");
  XEN_ASSERT_TYPE(XEN_double_P(scale_x), scale_x, 9, "gdk_pixbuf_composite_color", "double");
  XEN_ASSERT_TYPE(XEN_double_P(scale_y), scale_y, 10, "gdk_pixbuf_composite_color", "double");
  XEN_ASSERT_TYPE(XEN_GdkInterpType_P(interp_type), interp_type, 11, "gdk_pixbuf_composite_color", "GdkInterpType");
  XEN_ASSERT_TYPE(XEN_int_P(overall_alpha), overall_alpha, 12, "gdk_pixbuf_composite_color", "int");
  XEN_ASSERT_TYPE(XEN_int_P(check_x), check_x, 13, "gdk_pixbuf_composite_color", "int");
  XEN_ASSERT_TYPE(XEN_int_P(check_y), check_y, 14, "gdk_pixbuf_composite_color", "int");
  XEN_ASSERT_TYPE(XEN_int_P(check_size), check_size, 15, "gdk_pixbuf_composite_color", "int");
  XEN_ASSERT_TYPE(XEN_guint32_P(color1), color1, 16, "gdk_pixbuf_composite_color", "guint32");
  XEN_ASSERT_TYPE(XEN_guint32_P(color2), color2, 17, "gdk_pixbuf_composite_color", "guint32");
  gdk_pixbuf_composite_color(XEN_TO_C_GdkPixbuf_(src), XEN_TO_C_GdkPixbuf_(dest), XEN_TO_C_int(dest_x), XEN_TO_C_int(dest_y), 
                             XEN_TO_C_int(dest_width), XEN_TO_C_int(dest_height), XEN_TO_C_double(offset_x), XEN_TO_C_double(offset_y), 
                             XEN_TO_C_double(scale_x), XEN_TO_C_double(scale_y), XEN_TO_C_GdkInterpType(interp_type), XEN_TO_C_int(overall_alpha), 
                             XEN_TO_C_int(check_x), XEN_TO_C_int(check_y), XEN_TO_C_int(check_size), XEN_TO_C_guint32(color1), 
                             XEN_TO_C_guint32(color2));
  return(XEN_FALSE);
}

static XEN gxg_gdk_pixbuf_scale_simple(XEN src, XEN dest_width, XEN dest_height, XEN interp_type)
{
  #define H_gdk_pixbuf_scale_simple "GdkPixbuf* gdk_pixbuf_scale_simple(GdkPixbuf* src, int dest_width, \
int dest_height, GdkInterpType interp_type)"
  XEN_ASSERT_TYPE(XEN_GdkPixbuf__P(src), src, 1, "gdk_pixbuf_scale_simple", "GdkPixbuf*");
  XEN_ASSERT_TYPE(XEN_int_P(dest_width), dest_width, 2, "gdk_pixbuf_scale_simple", "int");
  XEN_ASSERT_TYPE(XEN_int_P(dest_height), dest_height, 3, "gdk_pixbuf_scale_simple", "int");
  XEN_ASSERT_TYPE(XEN_GdkInterpType_P(interp_type), interp_type, 4, "gdk_pixbuf_scale_simple", "GdkInterpType");
  return(C_TO_XEN_GdkPixbuf_(gdk_pixbuf_scale_simple(XEN_TO_C_GdkPixbuf_(src), XEN_TO_C_int(dest_width), XEN_TO_C_int(dest_height), 
                                                     XEN_TO_C_GdkInterpType(interp_type))));
}

static XEN gxg_gdk_pixbuf_composite_color_simple(XEN src, XEN dest_width, XEN dest_height, XEN interp_type, XEN overall_alpha, XEN check_size, XEN color1, XEN color2)
{
  #define H_gdk_pixbuf_composite_color_simple "GdkPixbuf* gdk_pixbuf_composite_color_simple(GdkPixbuf* src, \
int dest_width, int dest_height, GdkInterpType interp_type, int overall_alpha, int check_size, guint32 color1, \
guint32 color2)"
  XEN_ASSERT_TYPE(XEN_GdkPixbuf__P(src), src, 1, "gdk_pixbuf_composite_color_simple", "GdkPixbuf*");
  XEN_ASSERT_TYPE(XEN_int_P(dest_width), dest_width, 2, "gdk_pixbuf_composite_color_simple", "int");
  XEN_ASSERT_TYPE(XEN_int_P(dest_height), dest_height, 3, "gdk_pixbuf_composite_color_simple", "int");
  XEN_ASSERT_TYPE(XEN_GdkInterpType_P(interp_type), interp_type, 4, "gdk_pixbuf_composite_color_simple", "GdkInterpType");
  XEN_ASSERT_TYPE(XEN_int_P(overall_alpha), overall_alpha, 5, "gdk_pixbuf_composite_color_simple", "int");
  XEN_ASSERT_TYPE(XEN_int_P(check_size), check_size, 6, "gdk_pixbuf_composite_color_simple", "int");
  XEN_ASSERT_TYPE(XEN_guint32_P(color1), color1, 7, "gdk_pixbuf_composite_color_simple", "guint32");
  XEN_ASSERT_TYPE(XEN_guint32_P(color2), color2, 8, "gdk_pixbuf_composite_color_simple", "guint32");
  return(C_TO_XEN_GdkPixbuf_(gdk_pixbuf_composite_color_simple(XEN_TO_C_GdkPixbuf_(src), XEN_TO_C_int(dest_width), XEN_TO_C_int(dest_height), 
                                                               XEN_TO_C_GdkInterpType(interp_type), XEN_TO_C_int(overall_alpha), 
                                                               XEN_TO_C_int(check_size), XEN_TO_C_guint32(color1), XEN_TO_C_guint32(color2))));
}

static XEN gxg_gdk_pixbuf_animation_get_type(void)
{
  #define H_gdk_pixbuf_animation_get_type "GType gdk_pixbuf_animation_get_type( void)"
  return(C_TO_XEN_GType(gdk_pixbuf_animation_get_type()));
}

static XEN gxg_gdk_pixbuf_animation_new_from_file(XEN filename, XEN error)
{
  #define H_gdk_pixbuf_animation_new_from_file "GdkPixbufAnimation* gdk_pixbuf_animation_new_from_file(char* filename, \
GError** [error])"
  GError* ref_error;
  XEN_ASSERT_TYPE(XEN_char__P(filename), filename, 1, "gdk_pixbuf_animation_new_from_file", "char*");
  {
    XEN result = XEN_FALSE;
    result = C_TO_XEN_GdkPixbufAnimation_(gdk_pixbuf_animation_new_from_file(XEN_TO_C_char_(filename), &ref_error));
    return(XEN_LIST_2(result, C_TO_XEN_GError_(ref_error)));
   }
}

static XEN gxg_gdk_pixbuf_animation_get_width(XEN animation)
{
  #define H_gdk_pixbuf_animation_get_width "int gdk_pixbuf_animation_get_width(GdkPixbufAnimation* animation)"
  XEN_ASSERT_TYPE(XEN_GdkPixbufAnimation__P(animation), animation, 1, "gdk_pixbuf_animation_get_width", "GdkPixbufAnimation*");
  return(C_TO_XEN_int(gdk_pixbuf_animation_get_width(XEN_TO_C_GdkPixbufAnimation_(animation))));
}

static XEN gxg_gdk_pixbuf_animation_get_height(XEN animation)
{
  #define H_gdk_pixbuf_animation_get_height "int gdk_pixbuf_animation_get_height(GdkPixbufAnimation* animation)"
  XEN_ASSERT_TYPE(XEN_GdkPixbufAnimation__P(animation), animation, 1, "gdk_pixbuf_animation_get_height", "GdkPixbufAnimation*");
  return(C_TO_XEN_int(gdk_pixbuf_animation_get_height(XEN_TO_C_GdkPixbufAnimation_(animation))));
}

static XEN gxg_gdk_pixbuf_animation_is_static_image(XEN animation)
{
  #define H_gdk_pixbuf_animation_is_static_image "gboolean gdk_pixbuf_animation_is_static_image(GdkPixbufAnimation* animation)"
  XEN_ASSERT_TYPE(XEN_GdkPixbufAnimation__P(animation), animation, 1, "gdk_pixbuf_animation_is_static_image", "GdkPixbufAnimation*");
  return(C_TO_XEN_gboolean(gdk_pixbuf_animation_is_static_image(XEN_TO_C_GdkPixbufAnimation_(animation))));
}

static XEN gxg_gdk_pixbuf_animation_get_static_image(XEN animation)
{
  #define H_gdk_pixbuf_animation_get_static_image "GdkPixbuf* gdk_pixbuf_animation_get_static_image(GdkPixbufAnimation* animation)"
  XEN_ASSERT_TYPE(XEN_GdkPixbufAnimation__P(animation), animation, 1, "gdk_pixbuf_animation_get_static_image", "GdkPixbufAnimation*");
  return(C_TO_XEN_GdkPixbuf_(gdk_pixbuf_animation_get_static_image(XEN_TO_C_GdkPixbufAnimation_(animation))));
}

static XEN gxg_gdk_pixbuf_animation_get_iter(XEN animation, XEN start_time)
{
  #define H_gdk_pixbuf_animation_get_iter "GdkPixbufAnimationIter* gdk_pixbuf_animation_get_iter(GdkPixbufAnimation* animation, \
GTimeVal* start_time)"
  XEN_ASSERT_TYPE(XEN_GdkPixbufAnimation__P(animation), animation, 1, "gdk_pixbuf_animation_get_iter", "GdkPixbufAnimation*");
  XEN_ASSERT_TYPE(XEN_GTimeVal__P(start_time), start_time, 2, "gdk_pixbuf_animation_get_iter", "GTimeVal*");
  return(C_TO_XEN_GdkPixbufAnimationIter_(gdk_pixbuf_animation_get_iter(XEN_TO_C_GdkPixbufAnimation_(animation), XEN_TO_C_GTimeVal_(start_time))));
}

static XEN gxg_gdk_pixbuf_animation_iter_get_type(void)
{
  #define H_gdk_pixbuf_animation_iter_get_type "GType gdk_pixbuf_animation_iter_get_type( void)"
  return(C_TO_XEN_GType(gdk_pixbuf_animation_iter_get_type()));
}

static XEN gxg_gdk_pixbuf_animation_iter_get_delay_time(XEN iter)
{
  #define H_gdk_pixbuf_animation_iter_get_delay_time "int gdk_pixbuf_animation_iter_get_delay_time(GdkPixbufAnimationIter* iter)"
  XEN_ASSERT_TYPE(XEN_GdkPixbufAnimationIter__P(iter), iter, 1, "gdk_pixbuf_animation_iter_get_delay_time", "GdkPixbufAnimationIter*");
  return(C_TO_XEN_int(gdk_pixbuf_animation_iter_get_delay_time(XEN_TO_C_GdkPixbufAnimationIter_(iter))));
}

static XEN gxg_gdk_pixbuf_animation_iter_get_pixbuf(XEN iter)
{
  #define H_gdk_pixbuf_animation_iter_get_pixbuf "GdkPixbuf* gdk_pixbuf_animation_iter_get_pixbuf(GdkPixbufAnimationIter* iter)"
  XEN_ASSERT_TYPE(XEN_GdkPixbufAnimationIter__P(iter), iter, 1, "gdk_pixbuf_animation_iter_get_pixbuf", "GdkPixbufAnimationIter*");
  return(C_TO_XEN_GdkPixbuf_(gdk_pixbuf_animation_iter_get_pixbuf(XEN_TO_C_GdkPixbufAnimationIter_(iter))));
}

static XEN gxg_gdk_pixbuf_animation_iter_on_currently_loading_frame(XEN iter)
{
  #define H_gdk_pixbuf_animation_iter_on_currently_loading_frame "gboolean gdk_pixbuf_animation_iter_on_currently_loading_frame(GdkPixbufAnimationIter* iter)"
  XEN_ASSERT_TYPE(XEN_GdkPixbufAnimationIter__P(iter), iter, 1, "gdk_pixbuf_animation_iter_on_currently_loading_frame", "GdkPixbufAnimationIter*");
  return(C_TO_XEN_gboolean(gdk_pixbuf_animation_iter_on_currently_loading_frame(XEN_TO_C_GdkPixbufAnimationIter_(iter))));
}

static XEN gxg_gdk_pixbuf_animation_iter_advance(XEN iter, XEN current_time)
{
  #define H_gdk_pixbuf_animation_iter_advance "gboolean gdk_pixbuf_animation_iter_advance(GdkPixbufAnimationIter* iter, \
GTimeVal* current_time)"
  XEN_ASSERT_TYPE(XEN_GdkPixbufAnimationIter__P(iter), iter, 1, "gdk_pixbuf_animation_iter_advance", "GdkPixbufAnimationIter*");
  XEN_ASSERT_TYPE(XEN_GTimeVal__P(current_time), current_time, 2, "gdk_pixbuf_animation_iter_advance", "GTimeVal*");
  return(C_TO_XEN_gboolean(gdk_pixbuf_animation_iter_advance(XEN_TO_C_GdkPixbufAnimationIter_(iter), XEN_TO_C_GTimeVal_(current_time))));
}

static XEN gxg_gdk_pixbuf_get_option(XEN pixbuf, XEN key)
{
  #define H_gdk_pixbuf_get_option "gchar* gdk_pixbuf_get_option(GdkPixbuf* pixbuf, gchar* key)"
  XEN_ASSERT_TYPE(XEN_GdkPixbuf__P(pixbuf), pixbuf, 1, "gdk_pixbuf_get_option", "GdkPixbuf*");
  XEN_ASSERT_TYPE(XEN_gchar__P(key), key, 2, "gdk_pixbuf_get_option", "gchar*");
  return(C_TO_XEN_gchar_(gdk_pixbuf_get_option(XEN_TO_C_GdkPixbuf_(pixbuf), XEN_TO_C_gchar_(key))));
}

static XEN gxg_gdk_pixbuf_loader_get_type(void)
{
  #define H_gdk_pixbuf_loader_get_type "GType gdk_pixbuf_loader_get_type( void)"
  return(C_TO_XEN_GType(gdk_pixbuf_loader_get_type()));
}

static XEN gxg_gdk_pixbuf_loader_new(void)
{
  #define H_gdk_pixbuf_loader_new "GdkPixbufLoader* gdk_pixbuf_loader_new( void)"
  return(C_TO_XEN_GdkPixbufLoader_(gdk_pixbuf_loader_new()));
}

static XEN gxg_gdk_pixbuf_loader_new_with_type(XEN image_type, XEN error)
{
  #define H_gdk_pixbuf_loader_new_with_type "GdkPixbufLoader* gdk_pixbuf_loader_new_with_type(char* image_type, \
GError** [error])"
  GError* ref_error;
  XEN_ASSERT_TYPE(XEN_char__P(image_type), image_type, 1, "gdk_pixbuf_loader_new_with_type", "char*");
  {
    XEN result = XEN_FALSE;
    result = C_TO_XEN_GdkPixbufLoader_(gdk_pixbuf_loader_new_with_type(XEN_TO_C_char_(image_type), &ref_error));
    return(XEN_LIST_2(result, C_TO_XEN_GError_(ref_error)));
   }
}

static XEN gxg_gdk_pixbuf_loader_write(XEN loader, XEN buf, XEN count, XEN error)
{
  #define H_gdk_pixbuf_loader_write "gboolean gdk_pixbuf_loader_write(GdkPixbufLoader* loader, guchar* buf, \
gsize count, GError** [error])"
  GError* ref_error;
  XEN_ASSERT_TYPE(XEN_GdkPixbufLoader__P(loader), loader, 1, "gdk_pixbuf_loader_write", "GdkPixbufLoader*");
  XEN_ASSERT_TYPE(XEN_guchar__P(buf), buf, 2, "gdk_pixbuf_loader_write", "guchar*");
  XEN_ASSERT_TYPE(XEN_gsize_P(count), count, 3, "gdk_pixbuf_loader_write", "gsize");
  {
    XEN result = XEN_FALSE;
    result = C_TO_XEN_gboolean(gdk_pixbuf_loader_write(XEN_TO_C_GdkPixbufLoader_(loader), XEN_TO_C_guchar_(buf), XEN_TO_C_gsize(count), 
                                                       &ref_error));
    return(XEN_LIST_2(result, C_TO_XEN_GError_(ref_error)));
   }
}

static XEN gxg_gdk_pixbuf_loader_get_pixbuf(XEN loader)
{
  #define H_gdk_pixbuf_loader_get_pixbuf "GdkPixbuf* gdk_pixbuf_loader_get_pixbuf(GdkPixbufLoader* loader)"
  XEN_ASSERT_TYPE(XEN_GdkPixbufLoader__P(loader), loader, 1, "gdk_pixbuf_loader_get_pixbuf", "GdkPixbufLoader*");
  return(C_TO_XEN_GdkPixbuf_(gdk_pixbuf_loader_get_pixbuf(XEN_TO_C_GdkPixbufLoader_(loader))));
}

static XEN gxg_gdk_pixbuf_loader_get_animation(XEN loader)
{
  #define H_gdk_pixbuf_loader_get_animation "GdkPixbufAnimation* gdk_pixbuf_loader_get_animation(GdkPixbufLoader* loader)"
  XEN_ASSERT_TYPE(XEN_GdkPixbufLoader__P(loader), loader, 1, "gdk_pixbuf_loader_get_animation", "GdkPixbufLoader*");
  return(C_TO_XEN_GdkPixbufAnimation_(gdk_pixbuf_loader_get_animation(XEN_TO_C_GdkPixbufLoader_(loader))));
}

static XEN gxg_gdk_pixbuf_loader_close(XEN loader, XEN error)
{
  #define H_gdk_pixbuf_loader_close "gboolean gdk_pixbuf_loader_close(GdkPixbufLoader* loader, GError** [error])"
  GError* ref_error;
  XEN_ASSERT_TYPE(XEN_GdkPixbufLoader__P(loader), loader, 1, "gdk_pixbuf_loader_close", "GdkPixbufLoader*");
  {
    XEN result = XEN_FALSE;
    result = C_TO_XEN_gboolean(gdk_pixbuf_loader_close(XEN_TO_C_GdkPixbufLoader_(loader), &ref_error));
    return(XEN_LIST_2(result, C_TO_XEN_GError_(ref_error)));
   }
}

static XEN gxg_gdk_pixbuf_alpha_mode_get_type(void)
{
  #define H_gdk_pixbuf_alpha_mode_get_type "GType gdk_pixbuf_alpha_mode_get_type( void)"
  return(C_TO_XEN_GType(gdk_pixbuf_alpha_mode_get_type()));
}

static XEN gxg_gdk_colorspace_get_type(void)
{
  #define H_gdk_colorspace_get_type "GType gdk_colorspace_get_type( void)"
  return(C_TO_XEN_GType(gdk_colorspace_get_type()));
}

static XEN gxg_gdk_pixbuf_error_get_type(void)
{
  #define H_gdk_pixbuf_error_get_type "GType gdk_pixbuf_error_get_type( void)"
  return(C_TO_XEN_GType(gdk_pixbuf_error_get_type()));
}

static XEN gxg_gdk_interp_type_get_type(void)
{
  #define H_gdk_interp_type_get_type "GType gdk_interp_type_get_type( void)"
  return(C_TO_XEN_GType(gdk_interp_type_get_type()));
}
#endif

static XEN gxg_gtk_vbox_get_type(void)
{
  #define H_gtk_vbox_get_type "GtkType gtk_vbox_get_type( void)"
  return(C_TO_XEN_GtkType(gtk_vbox_get_type()));
}

static XEN gxg_gtk_vbox_new(XEN homogeneous, XEN spacing)
{
  #define H_gtk_vbox_new "GtkWidget* gtk_vbox_new(gboolean homogeneous, gint spacing)"
  XEN_ASSERT_TYPE(XEN_gboolean_P(homogeneous), homogeneous, 1, "gtk_vbox_new", "gboolean");
  XEN_ASSERT_TYPE(XEN_gint_P(spacing), spacing, 2, "gtk_vbox_new", "gint");
  return(C_TO_XEN_GtkWidget_(gtk_vbox_new(XEN_TO_C_gboolean(homogeneous), XEN_TO_C_gint(spacing))));
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_accel_group_get_type(void)
{
  #define H_gtk_accel_group_get_type "GType gtk_accel_group_get_type( void)"
  return(C_TO_XEN_GType(gtk_accel_group_get_type()));
}
#endif

static XEN gxg_gtk_accel_group_new(void)
{
  #define H_gtk_accel_group_new "GtkAccelGroup* gtk_accel_group_new( void)"
  return(C_TO_XEN_GtkAccelGroup_(gtk_accel_group_new()));
}

static XEN gxg_gtk_accel_group_lock(XEN accel_group)
{
  #define H_gtk_accel_group_lock "void gtk_accel_group_lock(GtkAccelGroup* accel_group)"
  XEN_ASSERT_TYPE(XEN_GtkAccelGroup__P(accel_group), accel_group, 1, "gtk_accel_group_lock", "GtkAccelGroup*");
  gtk_accel_group_lock(XEN_TO_C_GtkAccelGroup_(accel_group));
  return(XEN_FALSE);
}

static XEN gxg_gtk_accel_group_unlock(XEN accel_group)
{
  #define H_gtk_accel_group_unlock "void gtk_accel_group_unlock(GtkAccelGroup* accel_group)"
  XEN_ASSERT_TYPE(XEN_GtkAccelGroup__P(accel_group), accel_group, 1, "gtk_accel_group_unlock", "GtkAccelGroup*");
  gtk_accel_group_unlock(XEN_TO_C_GtkAccelGroup_(accel_group));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_accel_group_connect(XEN accel_group, XEN accel_key, XEN accel_mods, XEN accel_flags, XEN closure)
{
  #define H_gtk_accel_group_connect "void gtk_accel_group_connect(GtkAccelGroup* accel_group, guint accel_key, \
GdkModifierType accel_mods, GtkAccelFlags accel_flags, GClosure* closure)"
  XEN_ASSERT_TYPE(XEN_GtkAccelGroup__P(accel_group), accel_group, 1, "gtk_accel_group_connect", "GtkAccelGroup*");
  XEN_ASSERT_TYPE(XEN_guint_P(accel_key), accel_key, 2, "gtk_accel_group_connect", "guint");
  XEN_ASSERT_TYPE(XEN_GdkModifierType_P(accel_mods), accel_mods, 3, "gtk_accel_group_connect", "GdkModifierType");
  XEN_ASSERT_TYPE(XEN_GtkAccelFlags_P(accel_flags), accel_flags, 4, "gtk_accel_group_connect", "GtkAccelFlags");
  XEN_ASSERT_TYPE(XEN_GClosure__P(closure), closure, 5, "gtk_accel_group_connect", "GClosure*");
  gtk_accel_group_connect(XEN_TO_C_GtkAccelGroup_(accel_group), XEN_TO_C_guint(accel_key), XEN_TO_C_GdkModifierType(accel_mods), 
                          XEN_TO_C_GtkAccelFlags(accel_flags), XEN_TO_C_GClosure_(closure));
  return(XEN_FALSE);
}

static XEN gxg_gtk_accel_group_connect_by_path(XEN accel_group, XEN accel_path, XEN closure)
{
  #define H_gtk_accel_group_connect_by_path "void gtk_accel_group_connect_by_path(GtkAccelGroup* accel_group, \
gchar* accel_path, GClosure* closure)"
  XEN_ASSERT_TYPE(XEN_GtkAccelGroup__P(accel_group), accel_group, 1, "gtk_accel_group_connect_by_path", "GtkAccelGroup*");
  XEN_ASSERT_TYPE(XEN_gchar__P(accel_path), accel_path, 2, "gtk_accel_group_connect_by_path", "gchar*");
  XEN_ASSERT_TYPE(XEN_GClosure__P(closure), closure, 3, "gtk_accel_group_connect_by_path", "GClosure*");
  gtk_accel_group_connect_by_path(XEN_TO_C_GtkAccelGroup_(accel_group), XEN_TO_C_gchar_(accel_path), XEN_TO_C_GClosure_(closure));
  return(XEN_FALSE);
}

static XEN gxg_gtk_accel_group_disconnect(XEN accel_group, XEN closure)
{
  #define H_gtk_accel_group_disconnect "gboolean gtk_accel_group_disconnect(GtkAccelGroup* accel_group, \
GClosure* closure)"
  XEN_ASSERT_TYPE(XEN_GtkAccelGroup__P(accel_group), accel_group, 1, "gtk_accel_group_disconnect", "GtkAccelGroup*");
  XEN_ASSERT_TYPE(XEN_GClosure__P(closure), closure, 2, "gtk_accel_group_disconnect", "GClosure*");
  return(C_TO_XEN_gboolean(gtk_accel_group_disconnect(XEN_TO_C_GtkAccelGroup_(accel_group), XEN_TO_C_GClosure_(closure))));
}

static XEN gxg_gtk_accel_group_disconnect_key(XEN accel_group, XEN accel_key, XEN accel_mods)
{
  #define H_gtk_accel_group_disconnect_key "gboolean gtk_accel_group_disconnect_key(GtkAccelGroup* accel_group, \
guint accel_key, GdkModifierType accel_mods)"
  XEN_ASSERT_TYPE(XEN_GtkAccelGroup__P(accel_group), accel_group, 1, "gtk_accel_group_disconnect_key", "GtkAccelGroup*");
  XEN_ASSERT_TYPE(XEN_guint_P(accel_key), accel_key, 2, "gtk_accel_group_disconnect_key", "guint");
  XEN_ASSERT_TYPE(XEN_GdkModifierType_P(accel_mods), accel_mods, 3, "gtk_accel_group_disconnect_key", "GdkModifierType");
  return(C_TO_XEN_gboolean(gtk_accel_group_disconnect_key(XEN_TO_C_GtkAccelGroup_(accel_group), XEN_TO_C_guint(accel_key), 
                                                          XEN_TO_C_GdkModifierType(accel_mods))));
}

static XEN gxg_gtk_accel_groups_activate(XEN object, XEN accel_key, XEN accel_mods)
{
  #define H_gtk_accel_groups_activate "gboolean gtk_accel_groups_activate(GObject* object, guint accel_key, \
GdkModifierType accel_mods)"
  XEN_ASSERT_TYPE(XEN_GObject__P(object), object, 1, "gtk_accel_groups_activate", "GObject*");
  XEN_ASSERT_TYPE(XEN_guint_P(accel_key), accel_key, 2, "gtk_accel_groups_activate", "guint");
  XEN_ASSERT_TYPE(XEN_GdkModifierType_P(accel_mods), accel_mods, 3, "gtk_accel_groups_activate", "GdkModifierType");
  return(C_TO_XEN_gboolean(gtk_accel_groups_activate(XEN_TO_C_GObject_(object), XEN_TO_C_guint(accel_key), XEN_TO_C_GdkModifierType(accel_mods))));
}

static XEN gxg_gtk_accel_groups_from_object(XEN object)
{
  #define H_gtk_accel_groups_from_object "GSList* gtk_accel_groups_from_object(GObject* object)"
  XEN_ASSERT_TYPE(XEN_GObject__P(object), object, 1, "gtk_accel_groups_from_object", "GObject*");
  return(C_TO_XEN_GSList_(gtk_accel_groups_from_object(XEN_TO_C_GObject_(object))));
}

static XEN gxg_gtk_accel_group_find(XEN accel_group, XEN func, XEN func_data)
{
  #define H_gtk_accel_group_find "GtkAccelKey* gtk_accel_group_find(GtkAccelGroup* accel_group, lambda3 func, \
lambda_data func_data)"
  XEN_ASSERT_TYPE(XEN_GtkAccelGroup__P(accel_group), accel_group, 1, "gtk_accel_group_find", "GtkAccelGroup*");
  XEN_ASSERT_TYPE(XEN_lambda3_P(func), func, 2, "gtk_accel_group_find", "lambda3");
  if (XEN_NOT_BOUND_P(func_data)) func_data = XEN_FALSE; 
  else XEN_ASSERT_TYPE(XEN_lambda_data_P(func_data), func_data, 3, "gtk_accel_group_find", "lambda_data");
  {
    XEN result = XEN_FALSE;
    int loc;
    XEN gxg_ptr = XEN_LIST_5(func, func_data, XEN_FALSE, XEN_FALSE, XEN_FALSE);
    loc = xm_protect(gxg_ptr);
    XEN_LIST_SET(gxg_ptr, 2, C_TO_XEN_INT(loc));
    result = C_TO_XEN_GtkAccelKey_(gtk_accel_group_find(XEN_TO_C_GtkAccelGroup_(accel_group), XEN_TO_C_lambda3(func), XEN_TO_C_lambda_data(func_data)));
    xm_unprotect_at(loc);
    return(result);
   }
}

static XEN gxg_gtk_accel_group_from_accel_closure(XEN closure)
{
  #define H_gtk_accel_group_from_accel_closure "GtkAccelGroup* gtk_accel_group_from_accel_closure(GClosure* closure)"
  XEN_ASSERT_TYPE(XEN_GClosure__P(closure), closure, 1, "gtk_accel_group_from_accel_closure", "GClosure*");
  return(C_TO_XEN_GtkAccelGroup_(gtk_accel_group_from_accel_closure(XEN_TO_C_GClosure_(closure))));
}
#endif

static XEN gxg_gtk_accelerator_valid(XEN keyval, XEN modifiers)
{
  #define H_gtk_accelerator_valid "gboolean gtk_accelerator_valid(guint keyval, GdkModifierType modifiers)"
  XEN_ASSERT_TYPE(XEN_guint_P(keyval), keyval, 1, "gtk_accelerator_valid", "guint");
  XEN_ASSERT_TYPE(XEN_GdkModifierType_P(modifiers), modifiers, 2, "gtk_accelerator_valid", "GdkModifierType");
  return(C_TO_XEN_gboolean(gtk_accelerator_valid(XEN_TO_C_guint(keyval), XEN_TO_C_GdkModifierType(modifiers))));
}

static XEN gxg_gtk_accelerator_parse(XEN accelerator, XEN accelerator_key, XEN accelerator_mods)
{
  #define H_gtk_accelerator_parse "void gtk_accelerator_parse(gchar* accelerator, guint* [accelerator_key], \
GdkModifierType* [accelerator_mods])"
  guint ref_accelerator_key;
  GdkModifierType ref_accelerator_mods;
  XEN_ASSERT_TYPE(XEN_gchar__P(accelerator), accelerator, 1, "gtk_accelerator_parse", "gchar*");
  gtk_accelerator_parse(XEN_TO_C_gchar_(accelerator), &ref_accelerator_key, &ref_accelerator_mods);
  return(XEN_LIST_2(C_TO_XEN_guint(ref_accelerator_key), C_TO_XEN_GdkModifierType(ref_accelerator_mods)));
}

static XEN gxg_gtk_accelerator_name(XEN accelerator_key, XEN accelerator_mods)
{
  #define H_gtk_accelerator_name "gchar* gtk_accelerator_name(guint accelerator_key, GdkModifierType accelerator_mods)"
  XEN_ASSERT_TYPE(XEN_guint_P(accelerator_key), accelerator_key, 1, "gtk_accelerator_name", "guint");
  XEN_ASSERT_TYPE(XEN_GdkModifierType_P(accelerator_mods), accelerator_mods, 2, "gtk_accelerator_name", "GdkModifierType");
  return(C_TO_XEN_gchar_(gtk_accelerator_name(XEN_TO_C_guint(accelerator_key), XEN_TO_C_GdkModifierType(accelerator_mods))));
}

static XEN gxg_gtk_accelerator_set_default_mod_mask(XEN default_mod_mask)
{
  #define H_gtk_accelerator_set_default_mod_mask "void gtk_accelerator_set_default_mod_mask(GdkModifierType default_mod_mask)"
  XEN_ASSERT_TYPE(XEN_GdkModifierType_P(default_mod_mask), default_mod_mask, 1, "gtk_accelerator_set_default_mod_mask", "GdkModifierType");
  gtk_accelerator_set_default_mod_mask(XEN_TO_C_GdkModifierType(default_mod_mask));
  return(XEN_FALSE);
}

static XEN gxg_gtk_accelerator_get_default_mod_mask(void)
{
  #define H_gtk_accelerator_get_default_mod_mask "guint gtk_accelerator_get_default_mod_mask( void)"
  return(C_TO_XEN_guint(gtk_accelerator_get_default_mod_mask()));
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_accel_group_query(XEN accel_group, XEN accel_key, XEN accel_mods, XEN n_entries)
{
  #define H_gtk_accel_group_query "GtkAccelGroupEntry* gtk_accel_group_query(GtkAccelGroup* accel_group, \
guint accel_key, GdkModifierType accel_mods, guint* [n_entries])"
  guint ref_n_entries;
  XEN_ASSERT_TYPE(XEN_GtkAccelGroup__P(accel_group), accel_group, 1, "gtk_accel_group_query", "GtkAccelGroup*");
  XEN_ASSERT_TYPE(XEN_guint_P(accel_key), accel_key, 2, "gtk_accel_group_query", "guint");
  XEN_ASSERT_TYPE(XEN_GdkModifierType_P(accel_mods), accel_mods, 3, "gtk_accel_group_query", "GdkModifierType");
  {
    XEN result = XEN_FALSE;
    result = C_TO_XEN_GtkAccelGroupEntry_(gtk_accel_group_query(XEN_TO_C_GtkAccelGroup_(accel_group), XEN_TO_C_guint(accel_key), 
                                                                XEN_TO_C_GdkModifierType(accel_mods), &ref_n_entries));
    return(XEN_LIST_2(result, C_TO_XEN_guint(ref_n_entries)));
   }
}
#endif

static XEN gxg_gtk_accel_label_get_type(void)
{
  #define H_gtk_accel_label_get_type "GtkType gtk_accel_label_get_type( void)"
  return(C_TO_XEN_GtkType(gtk_accel_label_get_type()));
}

static XEN gxg_gtk_accel_label_new(XEN string)
{
  #define H_gtk_accel_label_new "GtkWidget* gtk_accel_label_new(gchar* string)"
  XEN_ASSERT_TYPE(XEN_gchar__P(string), string, 1, "gtk_accel_label_new", "gchar*");
  return(C_TO_XEN_GtkWidget_(gtk_accel_label_new(XEN_TO_C_gchar_(string))));
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_accel_label_get_accel_widget(XEN accel_label)
{
  #define H_gtk_accel_label_get_accel_widget "GtkWidget* gtk_accel_label_get_accel_widget(GtkAccelLabel* accel_label)"
  XEN_ASSERT_TYPE(XEN_GtkAccelLabel__P(accel_label), accel_label, 1, "gtk_accel_label_get_accel_widget", "GtkAccelLabel*");
  return(C_TO_XEN_GtkWidget_(gtk_accel_label_get_accel_widget(XEN_TO_C_GtkAccelLabel_(accel_label))));
}
#endif

static XEN gxg_gtk_accel_label_get_accel_width(XEN accel_label)
{
  #define H_gtk_accel_label_get_accel_width "guint gtk_accel_label_get_accel_width(GtkAccelLabel* accel_label)"
  XEN_ASSERT_TYPE(XEN_GtkAccelLabel__P(accel_label), accel_label, 1, "gtk_accel_label_get_accel_width", "GtkAccelLabel*");
  return(C_TO_XEN_guint(gtk_accel_label_get_accel_width(XEN_TO_C_GtkAccelLabel_(accel_label))));
}

static XEN gxg_gtk_accel_label_set_accel_widget(XEN accel_label, XEN accel_widget)
{
  #define H_gtk_accel_label_set_accel_widget "void gtk_accel_label_set_accel_widget(GtkAccelLabel* accel_label, \
GtkWidget* accel_widget)"
  XEN_ASSERT_TYPE(XEN_GtkAccelLabel__P(accel_label), accel_label, 1, "gtk_accel_label_set_accel_widget", "GtkAccelLabel*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(accel_widget), accel_widget, 2, "gtk_accel_label_set_accel_widget", "GtkWidget*");
  gtk_accel_label_set_accel_widget(XEN_TO_C_GtkAccelLabel_(accel_label), XEN_TO_C_GtkWidget_(accel_widget));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_accel_label_set_accel_closure(XEN accel_label, XEN closure)
{
  #define H_gtk_accel_label_set_accel_closure "void gtk_accel_label_set_accel_closure(GtkAccelLabel* accel_label, \
GClosure* closure)"
  XEN_ASSERT_TYPE(XEN_GtkAccelLabel__P(accel_label), accel_label, 1, "gtk_accel_label_set_accel_closure", "GtkAccelLabel*");
  XEN_ASSERT_TYPE(XEN_GClosure__P(closure), closure, 2, "gtk_accel_label_set_accel_closure", "GClosure*");
  gtk_accel_label_set_accel_closure(XEN_TO_C_GtkAccelLabel_(accel_label), XEN_TO_C_GClosure_(closure));
  return(XEN_FALSE);
}
#endif

static XEN gxg_gtk_accel_label_refetch(XEN accel_label)
{
  #define H_gtk_accel_label_refetch "gboolean gtk_accel_label_refetch(GtkAccelLabel* accel_label)"
  XEN_ASSERT_TYPE(XEN_GtkAccelLabel__P(accel_label), accel_label, 1, "gtk_accel_label_refetch", "GtkAccelLabel*");
  return(C_TO_XEN_gboolean(gtk_accel_label_refetch(XEN_TO_C_GtkAccelLabel_(accel_label))));
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_accel_map_add_entry(XEN accel_path, XEN accel_key, XEN accel_mods)
{
  #define H_gtk_accel_map_add_entry "void gtk_accel_map_add_entry(gchar* accel_path, guint accel_key, \
GdkModifierType accel_mods)"
  XEN_ASSERT_TYPE(XEN_gchar__P(accel_path), accel_path, 1, "gtk_accel_map_add_entry", "gchar*");
  XEN_ASSERT_TYPE(XEN_guint_P(accel_key), accel_key, 2, "gtk_accel_map_add_entry", "guint");
  XEN_ASSERT_TYPE(XEN_GdkModifierType_P(accel_mods), accel_mods, 3, "gtk_accel_map_add_entry", "GdkModifierType");
  gtk_accel_map_add_entry(XEN_TO_C_gchar_(accel_path), XEN_TO_C_guint(accel_key), XEN_TO_C_GdkModifierType(accel_mods));
  return(XEN_FALSE);
}

static XEN gxg_gtk_accel_map_lookup_entry(XEN accel_path, XEN key)
{
  #define H_gtk_accel_map_lookup_entry "gboolean gtk_accel_map_lookup_entry(gchar* accel_path, GtkAccelKey* key)"
  XEN_ASSERT_TYPE(XEN_gchar__P(accel_path), accel_path, 1, "gtk_accel_map_lookup_entry", "gchar*");
  XEN_ASSERT_TYPE(XEN_GtkAccelKey__P(key), key, 2, "gtk_accel_map_lookup_entry", "GtkAccelKey*");
  return(C_TO_XEN_gboolean(gtk_accel_map_lookup_entry(XEN_TO_C_gchar_(accel_path), XEN_TO_C_GtkAccelKey_(key))));
}

static XEN gxg_gtk_accel_map_change_entry(XEN accel_path, XEN accel_key, XEN accel_mods, XEN replace)
{
  #define H_gtk_accel_map_change_entry "gboolean gtk_accel_map_change_entry(gchar* accel_path, guint accel_key, \
GdkModifierType accel_mods, gboolean replace)"
  XEN_ASSERT_TYPE(XEN_gchar__P(accel_path), accel_path, 1, "gtk_accel_map_change_entry", "gchar*");
  XEN_ASSERT_TYPE(XEN_guint_P(accel_key), accel_key, 2, "gtk_accel_map_change_entry", "guint");
  XEN_ASSERT_TYPE(XEN_GdkModifierType_P(accel_mods), accel_mods, 3, "gtk_accel_map_change_entry", "GdkModifierType");
  XEN_ASSERT_TYPE(XEN_gboolean_P(replace), replace, 4, "gtk_accel_map_change_entry", "gboolean");
  return(C_TO_XEN_gboolean(gtk_accel_map_change_entry(XEN_TO_C_gchar_(accel_path), XEN_TO_C_guint(accel_key), XEN_TO_C_GdkModifierType(accel_mods), 
                                                      XEN_TO_C_gboolean(replace))));
}

static XEN gxg_gtk_accel_map_load(XEN file_name)
{
  #define H_gtk_accel_map_load "void gtk_accel_map_load(gchar* file_name)"
  XEN_ASSERT_TYPE(XEN_gchar__P(file_name), file_name, 1, "gtk_accel_map_load", "gchar*");
  gtk_accel_map_load(XEN_TO_C_gchar_(file_name));
  return(XEN_FALSE);
}

static XEN gxg_gtk_accel_map_save(XEN file_name)
{
  #define H_gtk_accel_map_save "void gtk_accel_map_save(gchar* file_name)"
  XEN_ASSERT_TYPE(XEN_gchar__P(file_name), file_name, 1, "gtk_accel_map_save", "gchar*");
  gtk_accel_map_save(XEN_TO_C_gchar_(file_name));
  return(XEN_FALSE);
}

static XEN gxg_gtk_accel_map_foreach(XEN data, XEN foreach_func)
{
  #define H_gtk_accel_map_foreach "void gtk_accel_map_foreach(gpointer data, GtkAccelMapForeach foreach_func)"
  XEN_ASSERT_TYPE(XEN_gpointer_P(data), data, 1, "gtk_accel_map_foreach", "gpointer");
  XEN_ASSERT_TYPE(XEN_GtkAccelMapForeach_P(foreach_func), foreach_func, 2, "gtk_accel_map_foreach", "GtkAccelMapForeach");
  gtk_accel_map_foreach(XEN_TO_C_gpointer(data), XEN_TO_C_GtkAccelMapForeach(foreach_func));
  return(XEN_FALSE);
}

static XEN gxg_gtk_accel_map_load_fd(XEN fd)
{
  #define H_gtk_accel_map_load_fd "void gtk_accel_map_load_fd(gint fd)"
  XEN_ASSERT_TYPE(XEN_gint_P(fd), fd, 1, "gtk_accel_map_load_fd", "gint");
  gtk_accel_map_load_fd(XEN_TO_C_gint(fd));
  return(XEN_FALSE);
}

static XEN gxg_gtk_accel_map_load_scanner(XEN scanner)
{
  #define H_gtk_accel_map_load_scanner "void gtk_accel_map_load_scanner(GScanner* scanner)"
  XEN_ASSERT_TYPE(XEN_GScanner__P(scanner), scanner, 1, "gtk_accel_map_load_scanner", "GScanner*");
  gtk_accel_map_load_scanner(XEN_TO_C_GScanner_(scanner));
  return(XEN_FALSE);
}

static XEN gxg_gtk_accel_map_save_fd(XEN fd)
{
  #define H_gtk_accel_map_save_fd "void gtk_accel_map_save_fd(gint fd)"
  XEN_ASSERT_TYPE(XEN_gint_P(fd), fd, 1, "gtk_accel_map_save_fd", "gint");
  gtk_accel_map_save_fd(XEN_TO_C_gint(fd));
  return(XEN_FALSE);
}

static XEN gxg_gtk_accel_map_add_filter(XEN filter_pattern)
{
  #define H_gtk_accel_map_add_filter "void gtk_accel_map_add_filter(gchar* filter_pattern)"
  XEN_ASSERT_TYPE(XEN_gchar__P(filter_pattern), filter_pattern, 1, "gtk_accel_map_add_filter", "gchar*");
  gtk_accel_map_add_filter(XEN_TO_C_gchar_(filter_pattern));
  return(XEN_FALSE);
}

static XEN gxg_gtk_accel_map_foreach_unfiltered(XEN data, XEN foreach_func)
{
  #define H_gtk_accel_map_foreach_unfiltered "void gtk_accel_map_foreach_unfiltered(gpointer data, GtkAccelMapForeach foreach_func)"
  XEN_ASSERT_TYPE(XEN_gpointer_P(data), data, 1, "gtk_accel_map_foreach_unfiltered", "gpointer");
  XEN_ASSERT_TYPE(XEN_GtkAccelMapForeach_P(foreach_func), foreach_func, 2, "gtk_accel_map_foreach_unfiltered", "GtkAccelMapForeach");
  gtk_accel_map_foreach_unfiltered(XEN_TO_C_gpointer(data), XEN_TO_C_GtkAccelMapForeach(foreach_func));
  return(XEN_FALSE);
}

static XEN gxg_gtk_accessible_get_type(void)
{
  #define H_gtk_accessible_get_type "GtkType gtk_accessible_get_type( void)"
  return(C_TO_XEN_GtkType(gtk_accessible_get_type()));
}

static XEN gxg_gtk_accessible_connect_widget_destroyed(XEN accessible)
{
  #define H_gtk_accessible_connect_widget_destroyed "void gtk_accessible_connect_widget_destroyed(GtkAccessible* accessible)"
  XEN_ASSERT_TYPE(XEN_GtkAccessible__P(accessible), accessible, 1, "gtk_accessible_connect_widget_destroyed", "GtkAccessible*");
  gtk_accessible_connect_widget_destroyed(XEN_TO_C_GtkAccessible_(accessible));
  return(XEN_FALSE);
}
#endif

static XEN gxg_gtk_adjustment_get_type(void)
{
  #define H_gtk_adjustment_get_type "GtkType gtk_adjustment_get_type( void)"
  return(C_TO_XEN_GtkType(gtk_adjustment_get_type()));
}

static XEN gxg_gtk_adjustment_new(XEN value, XEN lower, XEN upper, XEN step_increment, XEN page_increment, XEN page_size)
{
  #define H_gtk_adjustment_new "GtkObject* gtk_adjustment_new(gdouble value, gdouble lower, gdouble upper, \
gdouble step_increment, gdouble page_increment, gdouble page_size)"
  XEN_ASSERT_TYPE(XEN_gdouble_P(value), value, 1, "gtk_adjustment_new", "gdouble");
  XEN_ASSERT_TYPE(XEN_gdouble_P(lower), lower, 2, "gtk_adjustment_new", "gdouble");
  XEN_ASSERT_TYPE(XEN_gdouble_P(upper), upper, 3, "gtk_adjustment_new", "gdouble");
  XEN_ASSERT_TYPE(XEN_gdouble_P(step_increment), step_increment, 4, "gtk_adjustment_new", "gdouble");
  XEN_ASSERT_TYPE(XEN_gdouble_P(page_increment), page_increment, 5, "gtk_adjustment_new", "gdouble");
  XEN_ASSERT_TYPE(XEN_gdouble_P(page_size), page_size, 6, "gtk_adjustment_new", "gdouble");
  return(C_TO_XEN_GtkObject_(gtk_adjustment_new(XEN_TO_C_gdouble(value), XEN_TO_C_gdouble(lower), XEN_TO_C_gdouble(upper), 
                                                XEN_TO_C_gdouble(step_increment), XEN_TO_C_gdouble(page_increment), XEN_TO_C_gdouble(page_size))));
}

static XEN gxg_gtk_adjustment_changed(XEN adjustment)
{
  #define H_gtk_adjustment_changed "void gtk_adjustment_changed(GtkAdjustment* adjustment)"
  XEN_ASSERT_TYPE(XEN_GtkAdjustment__P(adjustment), adjustment, 1, "gtk_adjustment_changed", "GtkAdjustment*");
  gtk_adjustment_changed(XEN_TO_C_GtkAdjustment_(adjustment));
  return(XEN_FALSE);
}

static XEN gxg_gtk_adjustment_value_changed(XEN adjustment)
{
  #define H_gtk_adjustment_value_changed "void gtk_adjustment_value_changed(GtkAdjustment* adjustment)"
  XEN_ASSERT_TYPE(XEN_GtkAdjustment__P(adjustment), adjustment, 1, "gtk_adjustment_value_changed", "GtkAdjustment*");
  gtk_adjustment_value_changed(XEN_TO_C_GtkAdjustment_(adjustment));
  return(XEN_FALSE);
}

static XEN gxg_gtk_adjustment_clamp_page(XEN adjustment, XEN lower, XEN upper)
{
  #define H_gtk_adjustment_clamp_page "void gtk_adjustment_clamp_page(GtkAdjustment* adjustment, gdouble lower, \
gdouble upper)"
  XEN_ASSERT_TYPE(XEN_GtkAdjustment__P(adjustment), adjustment, 1, "gtk_adjustment_clamp_page", "GtkAdjustment*");
  XEN_ASSERT_TYPE(XEN_gdouble_P(lower), lower, 2, "gtk_adjustment_clamp_page", "gdouble");
  XEN_ASSERT_TYPE(XEN_gdouble_P(upper), upper, 3, "gtk_adjustment_clamp_page", "gdouble");
  gtk_adjustment_clamp_page(XEN_TO_C_GtkAdjustment_(adjustment), XEN_TO_C_gdouble(lower), XEN_TO_C_gdouble(upper));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_adjustment_get_value(XEN adjustment)
{
  #define H_gtk_adjustment_get_value "gdouble gtk_adjustment_get_value(GtkAdjustment* adjustment)"
  XEN_ASSERT_TYPE(XEN_GtkAdjustment__P(adjustment), adjustment, 1, "gtk_adjustment_get_value", "GtkAdjustment*");
  return(C_TO_XEN_gdouble(gtk_adjustment_get_value(XEN_TO_C_GtkAdjustment_(adjustment))));
}
#endif

static XEN gxg_gtk_adjustment_set_value(XEN adjustment, XEN value)
{
  #define H_gtk_adjustment_set_value "void gtk_adjustment_set_value(GtkAdjustment* adjustment, gdouble value)"
  XEN_ASSERT_TYPE(XEN_GtkAdjustment__P(adjustment), adjustment, 1, "gtk_adjustment_set_value", "GtkAdjustment*");
  XEN_ASSERT_TYPE(XEN_gdouble_P(value), value, 2, "gtk_adjustment_set_value", "gdouble");
  gtk_adjustment_set_value(XEN_TO_C_GtkAdjustment_(adjustment), XEN_TO_C_gdouble(value));
  return(XEN_FALSE);
}

static XEN gxg_gtk_alignment_get_type(void)
{
  #define H_gtk_alignment_get_type "GtkType gtk_alignment_get_type( void)"
  return(C_TO_XEN_GtkType(gtk_alignment_get_type()));
}

static XEN gxg_gtk_alignment_new(XEN xalign, XEN yalign, XEN xscale, XEN yscale)
{
  #define H_gtk_alignment_new "GtkWidget* gtk_alignment_new(gfloat xalign, gfloat yalign, gfloat xscale, \
gfloat yscale)"
  XEN_ASSERT_TYPE(XEN_gfloat_P(xalign), xalign, 1, "gtk_alignment_new", "gfloat");
  XEN_ASSERT_TYPE(XEN_gfloat_P(yalign), yalign, 2, "gtk_alignment_new", "gfloat");
  XEN_ASSERT_TYPE(XEN_gfloat_P(xscale), xscale, 3, "gtk_alignment_new", "gfloat");
  XEN_ASSERT_TYPE(XEN_gfloat_P(yscale), yscale, 4, "gtk_alignment_new", "gfloat");
  return(C_TO_XEN_GtkWidget_(gtk_alignment_new(XEN_TO_C_gfloat(xalign), XEN_TO_C_gfloat(yalign), XEN_TO_C_gfloat(xscale), 
                                               XEN_TO_C_gfloat(yscale))));
}

static XEN gxg_gtk_alignment_set(XEN alignment, XEN xalign, XEN yalign, XEN xscale, XEN yscale)
{
  #define H_gtk_alignment_set "void gtk_alignment_set(GtkAlignment* alignment, gfloat xalign, gfloat yalign, \
gfloat xscale, gfloat yscale)"
  XEN_ASSERT_TYPE(XEN_GtkAlignment__P(alignment), alignment, 1, "gtk_alignment_set", "GtkAlignment*");
  XEN_ASSERT_TYPE(XEN_gfloat_P(xalign), xalign, 2, "gtk_alignment_set", "gfloat");
  XEN_ASSERT_TYPE(XEN_gfloat_P(yalign), yalign, 3, "gtk_alignment_set", "gfloat");
  XEN_ASSERT_TYPE(XEN_gfloat_P(xscale), xscale, 4, "gtk_alignment_set", "gfloat");
  XEN_ASSERT_TYPE(XEN_gfloat_P(yscale), yscale, 5, "gtk_alignment_set", "gfloat");
  gtk_alignment_set(XEN_TO_C_GtkAlignment_(alignment), XEN_TO_C_gfloat(xalign), XEN_TO_C_gfloat(yalign), XEN_TO_C_gfloat(xscale), 
                    XEN_TO_C_gfloat(yscale));
  return(XEN_FALSE);
}

static XEN gxg_gtk_arrow_get_type(void)
{
  #define H_gtk_arrow_get_type "GtkType gtk_arrow_get_type( void)"
  return(C_TO_XEN_GtkType(gtk_arrow_get_type()));
}

static XEN gxg_gtk_arrow_new(XEN arrow_type, XEN shadow_type)
{
  #define H_gtk_arrow_new "GtkWidget* gtk_arrow_new(GtkArrowType arrow_type, GtkShadowType shadow_type)"
  XEN_ASSERT_TYPE(XEN_GtkArrowType_P(arrow_type), arrow_type, 1, "gtk_arrow_new", "GtkArrowType");
  XEN_ASSERT_TYPE(XEN_GtkShadowType_P(shadow_type), shadow_type, 2, "gtk_arrow_new", "GtkShadowType");
  return(C_TO_XEN_GtkWidget_(gtk_arrow_new(XEN_TO_C_GtkArrowType(arrow_type), XEN_TO_C_GtkShadowType(shadow_type))));
}

static XEN gxg_gtk_arrow_set(XEN arrow, XEN arrow_type, XEN shadow_type)
{
  #define H_gtk_arrow_set "void gtk_arrow_set(GtkArrow* arrow, GtkArrowType arrow_type, GtkShadowType shadow_type)"
  XEN_ASSERT_TYPE(XEN_GtkArrow__P(arrow), arrow, 1, "gtk_arrow_set", "GtkArrow*");
  XEN_ASSERT_TYPE(XEN_GtkArrowType_P(arrow_type), arrow_type, 2, "gtk_arrow_set", "GtkArrowType");
  XEN_ASSERT_TYPE(XEN_GtkShadowType_P(shadow_type), shadow_type, 3, "gtk_arrow_set", "GtkShadowType");
  gtk_arrow_set(XEN_TO_C_GtkArrow_(arrow), XEN_TO_C_GtkArrowType(arrow_type), XEN_TO_C_GtkShadowType(shadow_type));
  return(XEN_FALSE);
}

static XEN gxg_gtk_aspect_frame_get_type(void)
{
  #define H_gtk_aspect_frame_get_type "GtkType gtk_aspect_frame_get_type( void)"
  return(C_TO_XEN_GtkType(gtk_aspect_frame_get_type()));
}

static XEN gxg_gtk_aspect_frame_new(XEN label, XEN xalign, XEN yalign, XEN ratio, XEN obey_child)
{
  #define H_gtk_aspect_frame_new "GtkWidget* gtk_aspect_frame_new(gchar* label, gfloat xalign, gfloat yalign, \
gfloat ratio, gboolean obey_child)"
  XEN_ASSERT_TYPE(XEN_gchar__P(label), label, 1, "gtk_aspect_frame_new", "gchar*");
  XEN_ASSERT_TYPE(XEN_gfloat_P(xalign), xalign, 2, "gtk_aspect_frame_new", "gfloat");
  XEN_ASSERT_TYPE(XEN_gfloat_P(yalign), yalign, 3, "gtk_aspect_frame_new", "gfloat");
  XEN_ASSERT_TYPE(XEN_gfloat_P(ratio), ratio, 4, "gtk_aspect_frame_new", "gfloat");
  XEN_ASSERT_TYPE(XEN_gboolean_P(obey_child), obey_child, 5, "gtk_aspect_frame_new", "gboolean");
  return(C_TO_XEN_GtkWidget_(gtk_aspect_frame_new(XEN_TO_C_gchar_(label), XEN_TO_C_gfloat(xalign), XEN_TO_C_gfloat(yalign), 
                                                  XEN_TO_C_gfloat(ratio), XEN_TO_C_gboolean(obey_child))));
}

static XEN gxg_gtk_aspect_frame_set(XEN aspect_frame, XEN xalign, XEN yalign, XEN ratio, XEN obey_child)
{
  #define H_gtk_aspect_frame_set "void gtk_aspect_frame_set(GtkAspectFrame* aspect_frame, gfloat xalign, \
gfloat yalign, gfloat ratio, gboolean obey_child)"
  XEN_ASSERT_TYPE(XEN_GtkAspectFrame__P(aspect_frame), aspect_frame, 1, "gtk_aspect_frame_set", "GtkAspectFrame*");
  XEN_ASSERT_TYPE(XEN_gfloat_P(xalign), xalign, 2, "gtk_aspect_frame_set", "gfloat");
  XEN_ASSERT_TYPE(XEN_gfloat_P(yalign), yalign, 3, "gtk_aspect_frame_set", "gfloat");
  XEN_ASSERT_TYPE(XEN_gfloat_P(ratio), ratio, 4, "gtk_aspect_frame_set", "gfloat");
  XEN_ASSERT_TYPE(XEN_gboolean_P(obey_child), obey_child, 5, "gtk_aspect_frame_set", "gboolean");
  gtk_aspect_frame_set(XEN_TO_C_GtkAspectFrame_(aspect_frame), XEN_TO_C_gfloat(xalign), XEN_TO_C_gfloat(yalign), XEN_TO_C_gfloat(ratio), 
                       XEN_TO_C_gboolean(obey_child));
  return(XEN_FALSE);
}

static XEN gxg_gtk_button_box_get_type(void)
{
  #define H_gtk_button_box_get_type "GtkType gtk_button_box_get_type( void)"
  return(C_TO_XEN_GtkType(gtk_button_box_get_type()));
}

static XEN gxg_gtk_button_box_get_layout(XEN widget)
{
  #define H_gtk_button_box_get_layout "GtkButtonBoxStyle gtk_button_box_get_layout(GtkButtonBox* widget)"
  XEN_ASSERT_TYPE(XEN_GtkButtonBox__P(widget), widget, 1, "gtk_button_box_get_layout", "GtkButtonBox*");
  return(C_TO_XEN_GtkButtonBoxStyle(gtk_button_box_get_layout(XEN_TO_C_GtkButtonBox_(widget))));
}

static XEN gxg_gtk_button_box_set_layout(XEN widget, XEN layout_style)
{
  #define H_gtk_button_box_set_layout "void gtk_button_box_set_layout(GtkButtonBox* widget, GtkButtonBoxStyle layout_style)"
  XEN_ASSERT_TYPE(XEN_GtkButtonBox__P(widget), widget, 1, "gtk_button_box_set_layout", "GtkButtonBox*");
  XEN_ASSERT_TYPE(XEN_GtkButtonBoxStyle_P(layout_style), layout_style, 2, "gtk_button_box_set_layout", "GtkButtonBoxStyle");
  gtk_button_box_set_layout(XEN_TO_C_GtkButtonBox_(widget), XEN_TO_C_GtkButtonBoxStyle(layout_style));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_button_box_set_child_secondary(XEN widget, XEN child, XEN is_secondary)
{
  #define H_gtk_button_box_set_child_secondary "void gtk_button_box_set_child_secondary(GtkButtonBox* widget, \
GtkWidget* child, gboolean is_secondary)"
  XEN_ASSERT_TYPE(XEN_GtkButtonBox__P(widget), widget, 1, "gtk_button_box_set_child_secondary", "GtkButtonBox*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(child), child, 2, "gtk_button_box_set_child_secondary", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(is_secondary), is_secondary, 3, "gtk_button_box_set_child_secondary", "gboolean");
  gtk_button_box_set_child_secondary(XEN_TO_C_GtkButtonBox_(widget), XEN_TO_C_GtkWidget_(child), XEN_TO_C_gboolean(is_secondary));
  return(XEN_FALSE);
}
#endif

static XEN gxg_gtk_binding_set_new(XEN set_name)
{
  #define H_gtk_binding_set_new "GtkBindingSet* gtk_binding_set_new(gchar* set_name)"
  XEN_ASSERT_TYPE(XEN_gchar__P(set_name), set_name, 1, "gtk_binding_set_new", "gchar*");
  return(C_TO_XEN_GtkBindingSet_(gtk_binding_set_new(XEN_TO_C_gchar_(set_name))));
}

static XEN gxg_gtk_binding_set_by_class(XEN object_class)
{
  #define H_gtk_binding_set_by_class "GtkBindingSet* gtk_binding_set_by_class(gpointer object_class)"
  XEN_ASSERT_TYPE(XEN_gpointer_P(object_class), object_class, 1, "gtk_binding_set_by_class", "gpointer");
  return(C_TO_XEN_GtkBindingSet_(gtk_binding_set_by_class(XEN_TO_C_gpointer(object_class))));
}

static XEN gxg_gtk_binding_set_find(XEN set_name)
{
  #define H_gtk_binding_set_find "GtkBindingSet* gtk_binding_set_find(gchar* set_name)"
  XEN_ASSERT_TYPE(XEN_gchar__P(set_name), set_name, 1, "gtk_binding_set_find", "gchar*");
  return(C_TO_XEN_GtkBindingSet_(gtk_binding_set_find(XEN_TO_C_gchar_(set_name))));
}

static XEN gxg_gtk_bindings_activate(XEN object, XEN keyval, XEN modifiers)
{
  #define H_gtk_bindings_activate "gboolean gtk_bindings_activate(GtkObject* object, guint keyval, GdkModifierType modifiers)"
  XEN_ASSERT_TYPE(XEN_GtkObject__P(object), object, 1, "gtk_bindings_activate", "GtkObject*");
  XEN_ASSERT_TYPE(XEN_guint_P(keyval), keyval, 2, "gtk_bindings_activate", "guint");
  XEN_ASSERT_TYPE(XEN_GdkModifierType_P(modifiers), modifiers, 3, "gtk_bindings_activate", "GdkModifierType");
  return(C_TO_XEN_gboolean(gtk_bindings_activate(XEN_TO_C_GtkObject_(object), XEN_TO_C_guint(keyval), XEN_TO_C_GdkModifierType(modifiers))));
}

static XEN gxg_gtk_binding_set_activate(XEN binding_set, XEN keyval, XEN modifiers, XEN object)
{
  #define H_gtk_binding_set_activate "gboolean gtk_binding_set_activate(GtkBindingSet* binding_set, guint keyval, \
GdkModifierType modifiers, GtkObject* object)"
  XEN_ASSERT_TYPE(XEN_GtkBindingSet__P(binding_set), binding_set, 1, "gtk_binding_set_activate", "GtkBindingSet*");
  XEN_ASSERT_TYPE(XEN_guint_P(keyval), keyval, 2, "gtk_binding_set_activate", "guint");
  XEN_ASSERT_TYPE(XEN_GdkModifierType_P(modifiers), modifiers, 3, "gtk_binding_set_activate", "GdkModifierType");
  XEN_ASSERT_TYPE(XEN_GtkObject__P(object), object, 4, "gtk_binding_set_activate", "GtkObject*");
  return(C_TO_XEN_gboolean(gtk_binding_set_activate(XEN_TO_C_GtkBindingSet_(binding_set), XEN_TO_C_guint(keyval), XEN_TO_C_GdkModifierType(modifiers), 
                                                    XEN_TO_C_GtkObject_(object))));
}

static XEN gxg_gtk_binding_entry_clear(XEN binding_set, XEN keyval, XEN modifiers)
{
  #define H_gtk_binding_entry_clear "void gtk_binding_entry_clear(GtkBindingSet* binding_set, guint keyval, \
GdkModifierType modifiers)"
  XEN_ASSERT_TYPE(XEN_GtkBindingSet__P(binding_set), binding_set, 1, "gtk_binding_entry_clear", "GtkBindingSet*");
  XEN_ASSERT_TYPE(XEN_guint_P(keyval), keyval, 2, "gtk_binding_entry_clear", "guint");
  XEN_ASSERT_TYPE(XEN_GdkModifierType_P(modifiers), modifiers, 3, "gtk_binding_entry_clear", "GdkModifierType");
  gtk_binding_entry_clear(XEN_TO_C_GtkBindingSet_(binding_set), XEN_TO_C_guint(keyval), XEN_TO_C_GdkModifierType(modifiers));
  return(XEN_FALSE);
}

static XEN gxg_gtk_binding_entry_add_signal(XEN binding_set, XEN keyval, XEN modifiers, XEN signal_name, XEN n_args)
{
  #define H_gtk_binding_entry_add_signal "void gtk_binding_entry_add_signal(GtkBindingSet* binding_set, \
guint keyval, GdkModifierType modifiers, gchar* signal_name, guint n_args, ...)"
  XEN_ASSERT_TYPE(XEN_GtkBindingSet__P(binding_set), binding_set, 1, "gtk_binding_entry_add_signal", "GtkBindingSet*");
  XEN_ASSERT_TYPE(XEN_guint_P(keyval), keyval, 2, "gtk_binding_entry_add_signal", "guint");
  XEN_ASSERT_TYPE(XEN_GdkModifierType_P(modifiers), modifiers, 3, "gtk_binding_entry_add_signal", "GdkModifierType");
  XEN_ASSERT_TYPE(XEN_gchar__P(signal_name), signal_name, 4, "gtk_binding_entry_add_signal", "gchar*");
  XEN_ASSERT_TYPE(XEN_guint_P(n_args), n_args, 5, "gtk_binding_entry_add_signal", "guint");
  gtk_binding_entry_add_signal(XEN_TO_C_GtkBindingSet_(binding_set), XEN_TO_C_guint(keyval), XEN_TO_C_GdkModifierType(modifiers), 
                               XEN_TO_C_gchar_(signal_name), XEN_TO_C_guint(n_args));
  return(XEN_FALSE);
}

static XEN gxg_gtk_binding_set_add_path(XEN binding_set, XEN path_type, XEN path_pattern, XEN priority)
{
  #define H_gtk_binding_set_add_path "void gtk_binding_set_add_path(GtkBindingSet* binding_set, GtkPathType path_type, \
gchar* path_pattern, GtkPathPriorityType priority)"
  XEN_ASSERT_TYPE(XEN_GtkBindingSet__P(binding_set), binding_set, 1, "gtk_binding_set_add_path", "GtkBindingSet*");
  XEN_ASSERT_TYPE(XEN_GtkPathType_P(path_type), path_type, 2, "gtk_binding_set_add_path", "GtkPathType");
  XEN_ASSERT_TYPE(XEN_gchar__P(path_pattern), path_pattern, 3, "gtk_binding_set_add_path", "gchar*");
  XEN_ASSERT_TYPE(XEN_GtkPathPriorityType_P(priority), priority, 4, "gtk_binding_set_add_path", "GtkPathPriorityType");
  gtk_binding_set_add_path(XEN_TO_C_GtkBindingSet_(binding_set), XEN_TO_C_GtkPathType(path_type), XEN_TO_C_gchar_(path_pattern), 
                           XEN_TO_C_GtkPathPriorityType(priority));
  return(XEN_FALSE);
}

static XEN gxg_gtk_binding_entry_remove(XEN binding_set, XEN keyval, XEN modifiers)
{
  #define H_gtk_binding_entry_remove "void gtk_binding_entry_remove(GtkBindingSet* binding_set, guint keyval, \
GdkModifierType modifiers)"
  XEN_ASSERT_TYPE(XEN_GtkBindingSet__P(binding_set), binding_set, 1, "gtk_binding_entry_remove", "GtkBindingSet*");
  XEN_ASSERT_TYPE(XEN_guint_P(keyval), keyval, 2, "gtk_binding_entry_remove", "guint");
  XEN_ASSERT_TYPE(XEN_GdkModifierType_P(modifiers), modifiers, 3, "gtk_binding_entry_remove", "GdkModifierType");
  gtk_binding_entry_remove(XEN_TO_C_GtkBindingSet_(binding_set), XEN_TO_C_guint(keyval), XEN_TO_C_GdkModifierType(modifiers));
  return(XEN_FALSE);
}

static XEN gxg_gtk_binding_entry_add_signall(XEN binding_set, XEN keyval, XEN modifiers, XEN signal_name, XEN binding_args)
{
  #define H_gtk_binding_entry_add_signall "void gtk_binding_entry_add_signall(GtkBindingSet* binding_set, \
guint keyval, GdkModifierType modifiers, gchar* signal_name, GSList* binding_args)"
  XEN_ASSERT_TYPE(XEN_GtkBindingSet__P(binding_set), binding_set, 1, "gtk_binding_entry_add_signall", "GtkBindingSet*");
  XEN_ASSERT_TYPE(XEN_guint_P(keyval), keyval, 2, "gtk_binding_entry_add_signall", "guint");
  XEN_ASSERT_TYPE(XEN_GdkModifierType_P(modifiers), modifiers, 3, "gtk_binding_entry_add_signall", "GdkModifierType");
  XEN_ASSERT_TYPE(XEN_gchar__P(signal_name), signal_name, 4, "gtk_binding_entry_add_signall", "gchar*");
  XEN_ASSERT_TYPE(XEN_GSList__P(binding_args), binding_args, 5, "gtk_binding_entry_add_signall", "GSList*");
  gtk_binding_entry_add_signall(XEN_TO_C_GtkBindingSet_(binding_set), XEN_TO_C_guint(keyval), XEN_TO_C_GdkModifierType(modifiers), 
                                XEN_TO_C_gchar_(signal_name), XEN_TO_C_GSList_(binding_args));
  return(XEN_FALSE);
}

static XEN gxg_gtk_binding_parse_binding(XEN scanner)
{
  #define H_gtk_binding_parse_binding "guint gtk_binding_parse_binding(GScanner* scanner)"
  XEN_ASSERT_TYPE(XEN_GScanner__P(scanner), scanner, 1, "gtk_binding_parse_binding", "GScanner*");
  return(C_TO_XEN_guint(gtk_binding_parse_binding(XEN_TO_C_GScanner_(scanner))));
}

static XEN gxg_gtk_bin_get_type(void)
{
  #define H_gtk_bin_get_type "GtkType gtk_bin_get_type( void)"
  return(C_TO_XEN_GtkType(gtk_bin_get_type()));
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_bin_get_child(XEN bin)
{
  #define H_gtk_bin_get_child "GtkWidget* gtk_bin_get_child(GtkBin* bin)"
  XEN_ASSERT_TYPE(XEN_GtkBin__P(bin), bin, 1, "gtk_bin_get_child", "GtkBin*");
  return(C_TO_XEN_GtkWidget_(gtk_bin_get_child(XEN_TO_C_GtkBin_(bin))));
}
#endif

static XEN gxg_gtk_box_get_type(void)
{
  #define H_gtk_box_get_type "GtkType gtk_box_get_type( void)"
  return(C_TO_XEN_GtkType(gtk_box_get_type()));
}

static XEN gxg_gtk_box_pack_start(XEN box, XEN child, XEN expand, XEN fill, XEN padding)
{
  #define H_gtk_box_pack_start "void gtk_box_pack_start(GtkBox* box, GtkWidget* child, gboolean expand, \
gboolean fill, guint padding)"
  XEN_ASSERT_TYPE(XEN_GtkBox__P(box), box, 1, "gtk_box_pack_start", "GtkBox*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(child), child, 2, "gtk_box_pack_start", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(expand), expand, 3, "gtk_box_pack_start", "gboolean");
  XEN_ASSERT_TYPE(XEN_gboolean_P(fill), fill, 4, "gtk_box_pack_start", "gboolean");
  XEN_ASSERT_TYPE(XEN_guint_P(padding), padding, 5, "gtk_box_pack_start", "guint");
  gtk_box_pack_start(XEN_TO_C_GtkBox_(box), XEN_TO_C_GtkWidget_(child), XEN_TO_C_gboolean(expand), XEN_TO_C_gboolean(fill), 
                     XEN_TO_C_guint(padding));
  return(XEN_FALSE);
}

static XEN gxg_gtk_box_pack_end(XEN box, XEN child, XEN expand, XEN fill, XEN padding)
{
  #define H_gtk_box_pack_end "void gtk_box_pack_end(GtkBox* box, GtkWidget* child, gboolean expand, gboolean fill, \
guint padding)"
  XEN_ASSERT_TYPE(XEN_GtkBox__P(box), box, 1, "gtk_box_pack_end", "GtkBox*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(child), child, 2, "gtk_box_pack_end", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(expand), expand, 3, "gtk_box_pack_end", "gboolean");
  XEN_ASSERT_TYPE(XEN_gboolean_P(fill), fill, 4, "gtk_box_pack_end", "gboolean");
  XEN_ASSERT_TYPE(XEN_guint_P(padding), padding, 5, "gtk_box_pack_end", "guint");
  gtk_box_pack_end(XEN_TO_C_GtkBox_(box), XEN_TO_C_GtkWidget_(child), XEN_TO_C_gboolean(expand), XEN_TO_C_gboolean(fill), 
                   XEN_TO_C_guint(padding));
  return(XEN_FALSE);
}

static XEN gxg_gtk_box_pack_start_defaults(XEN box, XEN widget)
{
  #define H_gtk_box_pack_start_defaults "void gtk_box_pack_start_defaults(GtkBox* box, GtkWidget* widget)"
  XEN_ASSERT_TYPE(XEN_GtkBox__P(box), box, 1, "gtk_box_pack_start_defaults", "GtkBox*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 2, "gtk_box_pack_start_defaults", "GtkWidget*");
  gtk_box_pack_start_defaults(XEN_TO_C_GtkBox_(box), XEN_TO_C_GtkWidget_(widget));
  return(XEN_FALSE);
}

static XEN gxg_gtk_box_pack_end_defaults(XEN box, XEN widget)
{
  #define H_gtk_box_pack_end_defaults "void gtk_box_pack_end_defaults(GtkBox* box, GtkWidget* widget)"
  XEN_ASSERT_TYPE(XEN_GtkBox__P(box), box, 1, "gtk_box_pack_end_defaults", "GtkBox*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 2, "gtk_box_pack_end_defaults", "GtkWidget*");
  gtk_box_pack_end_defaults(XEN_TO_C_GtkBox_(box), XEN_TO_C_GtkWidget_(widget));
  return(XEN_FALSE);
}

static XEN gxg_gtk_box_set_homogeneous(XEN box, XEN homogeneous)
{
  #define H_gtk_box_set_homogeneous "void gtk_box_set_homogeneous(GtkBox* box, gboolean homogeneous)"
  XEN_ASSERT_TYPE(XEN_GtkBox__P(box), box, 1, "gtk_box_set_homogeneous", "GtkBox*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(homogeneous), homogeneous, 2, "gtk_box_set_homogeneous", "gboolean");
  gtk_box_set_homogeneous(XEN_TO_C_GtkBox_(box), XEN_TO_C_gboolean(homogeneous));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_box_get_homogeneous(XEN box)
{
  #define H_gtk_box_get_homogeneous "gboolean gtk_box_get_homogeneous(GtkBox* box)"
  XEN_ASSERT_TYPE(XEN_GtkBox__P(box), box, 1, "gtk_box_get_homogeneous", "GtkBox*");
  return(C_TO_XEN_gboolean(gtk_box_get_homogeneous(XEN_TO_C_GtkBox_(box))));
}
#endif

static XEN gxg_gtk_box_set_spacing(XEN box, XEN spacing)
{
  #define H_gtk_box_set_spacing "void gtk_box_set_spacing(GtkBox* box, gint spacing)"
  XEN_ASSERT_TYPE(XEN_GtkBox__P(box), box, 1, "gtk_box_set_spacing", "GtkBox*");
  XEN_ASSERT_TYPE(XEN_gint_P(spacing), spacing, 2, "gtk_box_set_spacing", "gint");
  gtk_box_set_spacing(XEN_TO_C_GtkBox_(box), XEN_TO_C_gint(spacing));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_box_get_spacing(XEN box)
{
  #define H_gtk_box_get_spacing "gint gtk_box_get_spacing(GtkBox* box)"
  XEN_ASSERT_TYPE(XEN_GtkBox__P(box), box, 1, "gtk_box_get_spacing", "GtkBox*");
  return(C_TO_XEN_gint(gtk_box_get_spacing(XEN_TO_C_GtkBox_(box))));
}
#endif

static XEN gxg_gtk_box_reorder_child(XEN box, XEN child, XEN position)
{
  #define H_gtk_box_reorder_child "void gtk_box_reorder_child(GtkBox* box, GtkWidget* child, gint position)"
  XEN_ASSERT_TYPE(XEN_GtkBox__P(box), box, 1, "gtk_box_reorder_child", "GtkBox*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(child), child, 2, "gtk_box_reorder_child", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_gint_P(position), position, 3, "gtk_box_reorder_child", "gint");
  gtk_box_reorder_child(XEN_TO_C_GtkBox_(box), XEN_TO_C_GtkWidget_(child), XEN_TO_C_gint(position));
  return(XEN_FALSE);
}

static XEN gxg_gtk_box_query_child_packing(XEN box, XEN child, XEN expand, XEN fill, XEN padding, XEN pack_type)
{
  #define H_gtk_box_query_child_packing "void gtk_box_query_child_packing(GtkBox* box, GtkWidget* child, \
gboolean* expand, gboolean* fill, guint* padding, GtkPackType* pack_type)"
  XEN_ASSERT_TYPE(XEN_GtkBox__P(box), box, 1, "gtk_box_query_child_packing", "GtkBox*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(child), child, 2, "gtk_box_query_child_packing", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_gboolean__P(expand), expand, 3, "gtk_box_query_child_packing", "gboolean*");
  XEN_ASSERT_TYPE(XEN_gboolean__P(fill), fill, 4, "gtk_box_query_child_packing", "gboolean*");
  XEN_ASSERT_TYPE(XEN_guint__P(padding), padding, 5, "gtk_box_query_child_packing", "guint*");
  XEN_ASSERT_TYPE(XEN_GtkPackType__P(pack_type), pack_type, 6, "gtk_box_query_child_packing", "GtkPackType*");
  gtk_box_query_child_packing(XEN_TO_C_GtkBox_(box), XEN_TO_C_GtkWidget_(child), XEN_TO_C_gboolean_(expand), XEN_TO_C_gboolean_(fill), 
                              XEN_TO_C_guint_(padding), XEN_TO_C_GtkPackType_(pack_type));
  return(XEN_FALSE);
}

static XEN gxg_gtk_box_set_child_packing(XEN box, XEN child, XEN expand, XEN fill, XEN padding, XEN pack_type)
{
  #define H_gtk_box_set_child_packing "void gtk_box_set_child_packing(GtkBox* box, GtkWidget* child, \
gboolean expand, gboolean fill, guint padding, GtkPackType pack_type)"
  XEN_ASSERT_TYPE(XEN_GtkBox__P(box), box, 1, "gtk_box_set_child_packing", "GtkBox*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(child), child, 2, "gtk_box_set_child_packing", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(expand), expand, 3, "gtk_box_set_child_packing", "gboolean");
  XEN_ASSERT_TYPE(XEN_gboolean_P(fill), fill, 4, "gtk_box_set_child_packing", "gboolean");
  XEN_ASSERT_TYPE(XEN_guint_P(padding), padding, 5, "gtk_box_set_child_packing", "guint");
  XEN_ASSERT_TYPE(XEN_GtkPackType_P(pack_type), pack_type, 6, "gtk_box_set_child_packing", "GtkPackType");
  gtk_box_set_child_packing(XEN_TO_C_GtkBox_(box), XEN_TO_C_GtkWidget_(child), XEN_TO_C_gboolean(expand), XEN_TO_C_gboolean(fill), 
                            XEN_TO_C_guint(padding), XEN_TO_C_GtkPackType(pack_type));
  return(XEN_FALSE);
}

static XEN gxg_gtk_button_get_type(void)
{
  #define H_gtk_button_get_type "GtkType gtk_button_get_type( void)"
  return(C_TO_XEN_GtkType(gtk_button_get_type()));
}

static XEN gxg_gtk_button_new(void)
{
  #define H_gtk_button_new "GtkWidget* gtk_button_new( void)"
  return(C_TO_XEN_GtkWidget_(gtk_button_new()));
}

static XEN gxg_gtk_button_new_with_label(XEN label)
{
  #define H_gtk_button_new_with_label "GtkWidget* gtk_button_new_with_label(gchar* label)"
  XEN_ASSERT_TYPE(XEN_gchar__P(label), label, 1, "gtk_button_new_with_label", "gchar*");
  return(C_TO_XEN_GtkWidget_(gtk_button_new_with_label(XEN_TO_C_gchar_(label))));
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_button_new_from_stock(XEN stock_id)
{
  #define H_gtk_button_new_from_stock "GtkWidget* gtk_button_new_from_stock(gchar* stock_id)"
  XEN_ASSERT_TYPE(XEN_gchar__P(stock_id), stock_id, 1, "gtk_button_new_from_stock", "gchar*");
  return(C_TO_XEN_GtkWidget_(gtk_button_new_from_stock(XEN_TO_C_gchar_(stock_id))));
}

static XEN gxg_gtk_button_new_with_mnemonic(XEN label)
{
  #define H_gtk_button_new_with_mnemonic "GtkWidget* gtk_button_new_with_mnemonic(gchar* label)"
  XEN_ASSERT_TYPE(XEN_gchar__P(label), label, 1, "gtk_button_new_with_mnemonic", "gchar*");
  return(C_TO_XEN_GtkWidget_(gtk_button_new_with_mnemonic(XEN_TO_C_gchar_(label))));
}
#endif

static XEN gxg_gtk_button_pressed(XEN button)
{
  #define H_gtk_button_pressed "void gtk_button_pressed(GtkButton* button)"
  XEN_ASSERT_TYPE(XEN_GtkButton__P(button), button, 1, "gtk_button_pressed", "GtkButton*");
  gtk_button_pressed(XEN_TO_C_GtkButton_(button));
  return(XEN_FALSE);
}

static XEN gxg_gtk_button_released(XEN button)
{
  #define H_gtk_button_released "void gtk_button_released(GtkButton* button)"
  XEN_ASSERT_TYPE(XEN_GtkButton__P(button), button, 1, "gtk_button_released", "GtkButton*");
  gtk_button_released(XEN_TO_C_GtkButton_(button));
  return(XEN_FALSE);
}

static XEN gxg_gtk_button_clicked(XEN button)
{
  #define H_gtk_button_clicked "void gtk_button_clicked(GtkButton* button)"
  XEN_ASSERT_TYPE(XEN_GtkButton__P(button), button, 1, "gtk_button_clicked", "GtkButton*");
  gtk_button_clicked(XEN_TO_C_GtkButton_(button));
  return(XEN_FALSE);
}

static XEN gxg_gtk_button_enter(XEN button)
{
  #define H_gtk_button_enter "void gtk_button_enter(GtkButton* button)"
  XEN_ASSERT_TYPE(XEN_GtkButton__P(button), button, 1, "gtk_button_enter", "GtkButton*");
  gtk_button_enter(XEN_TO_C_GtkButton_(button));
  return(XEN_FALSE);
}

static XEN gxg_gtk_button_leave(XEN button)
{
  #define H_gtk_button_leave "void gtk_button_leave(GtkButton* button)"
  XEN_ASSERT_TYPE(XEN_GtkButton__P(button), button, 1, "gtk_button_leave", "GtkButton*");
  gtk_button_leave(XEN_TO_C_GtkButton_(button));
  return(XEN_FALSE);
}

static XEN gxg_gtk_button_set_relief(XEN button, XEN newstyle)
{
  #define H_gtk_button_set_relief "void gtk_button_set_relief(GtkButton* button, GtkReliefStyle newstyle)"
  XEN_ASSERT_TYPE(XEN_GtkButton__P(button), button, 1, "gtk_button_set_relief", "GtkButton*");
  XEN_ASSERT_TYPE(XEN_GtkReliefStyle_P(newstyle), newstyle, 2, "gtk_button_set_relief", "GtkReliefStyle");
  gtk_button_set_relief(XEN_TO_C_GtkButton_(button), XEN_TO_C_GtkReliefStyle(newstyle));
  return(XEN_FALSE);
}

static XEN gxg_gtk_button_get_relief(XEN button)
{
  #define H_gtk_button_get_relief "GtkReliefStyle gtk_button_get_relief(GtkButton* button)"
  XEN_ASSERT_TYPE(XEN_GtkButton__P(button), button, 1, "gtk_button_get_relief", "GtkButton*");
  return(C_TO_XEN_GtkReliefStyle(gtk_button_get_relief(XEN_TO_C_GtkButton_(button))));
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_button_set_label(XEN button, XEN label)
{
  #define H_gtk_button_set_label "void gtk_button_set_label(GtkButton* button, gchar* label)"
  XEN_ASSERT_TYPE(XEN_GtkButton__P(button), button, 1, "gtk_button_set_label", "GtkButton*");
  XEN_ASSERT_TYPE(XEN_gchar__P(label), label, 2, "gtk_button_set_label", "gchar*");
  gtk_button_set_label(XEN_TO_C_GtkButton_(button), XEN_TO_C_gchar_(label));
  return(XEN_FALSE);
}

static XEN gxg_gtk_button_get_label(XEN button)
{
  #define H_gtk_button_get_label "gchar* gtk_button_get_label(GtkButton* button)"
  XEN_ASSERT_TYPE(XEN_GtkButton__P(button), button, 1, "gtk_button_get_label", "GtkButton*");
  return(C_TO_XEN_gchar_(gtk_button_get_label(XEN_TO_C_GtkButton_(button))));
}

static XEN gxg_gtk_button_set_use_underline(XEN button, XEN use_underline)
{
  #define H_gtk_button_set_use_underline "void gtk_button_set_use_underline(GtkButton* button, gboolean use_underline)"
  XEN_ASSERT_TYPE(XEN_GtkButton__P(button), button, 1, "gtk_button_set_use_underline", "GtkButton*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(use_underline), use_underline, 2, "gtk_button_set_use_underline", "gboolean");
  gtk_button_set_use_underline(XEN_TO_C_GtkButton_(button), XEN_TO_C_gboolean(use_underline));
  return(XEN_FALSE);
}

static XEN gxg_gtk_button_get_use_underline(XEN button)
{
  #define H_gtk_button_get_use_underline "gboolean gtk_button_get_use_underline(GtkButton* button)"
  XEN_ASSERT_TYPE(XEN_GtkButton__P(button), button, 1, "gtk_button_get_use_underline", "GtkButton*");
  return(C_TO_XEN_gboolean(gtk_button_get_use_underline(XEN_TO_C_GtkButton_(button))));
}

static XEN gxg_gtk_button_set_use_stock(XEN button, XEN use_stock)
{
  #define H_gtk_button_set_use_stock "void gtk_button_set_use_stock(GtkButton* button, gboolean use_stock)"
  XEN_ASSERT_TYPE(XEN_GtkButton__P(button), button, 1, "gtk_button_set_use_stock", "GtkButton*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(use_stock), use_stock, 2, "gtk_button_set_use_stock", "gboolean");
  gtk_button_set_use_stock(XEN_TO_C_GtkButton_(button), XEN_TO_C_gboolean(use_stock));
  return(XEN_FALSE);
}

static XEN gxg_gtk_button_get_use_stock(XEN button)
{
  #define H_gtk_button_get_use_stock "gboolean gtk_button_get_use_stock(GtkButton* button)"
  XEN_ASSERT_TYPE(XEN_GtkButton__P(button), button, 1, "gtk_button_get_use_stock", "GtkButton*");
  return(C_TO_XEN_gboolean(gtk_button_get_use_stock(XEN_TO_C_GtkButton_(button))));
}
#endif

static XEN gxg_gtk_calendar_get_type(void)
{
  #define H_gtk_calendar_get_type "GtkType gtk_calendar_get_type( void)"
  return(C_TO_XEN_GtkType(gtk_calendar_get_type()));
}

static XEN gxg_gtk_calendar_new(void)
{
  #define H_gtk_calendar_new "GtkWidget* gtk_calendar_new( void)"
  return(C_TO_XEN_GtkWidget_(gtk_calendar_new()));
}

static XEN gxg_gtk_calendar_select_month(XEN calendar, XEN month, XEN year)
{
  #define H_gtk_calendar_select_month "gboolean gtk_calendar_select_month(GtkCalendar* calendar, guint month, \
guint year)"
  XEN_ASSERT_TYPE(XEN_GtkCalendar__P(calendar), calendar, 1, "gtk_calendar_select_month", "GtkCalendar*");
  XEN_ASSERT_TYPE(XEN_guint_P(month), month, 2, "gtk_calendar_select_month", "guint");
  XEN_ASSERT_TYPE(XEN_guint_P(year), year, 3, "gtk_calendar_select_month", "guint");
  return(C_TO_XEN_gboolean(gtk_calendar_select_month(XEN_TO_C_GtkCalendar_(calendar), XEN_TO_C_guint(month), XEN_TO_C_guint(year))));
}

static XEN gxg_gtk_calendar_select_day(XEN calendar, XEN day)
{
  #define H_gtk_calendar_select_day "void gtk_calendar_select_day(GtkCalendar* calendar, guint day)"
  XEN_ASSERT_TYPE(XEN_GtkCalendar__P(calendar), calendar, 1, "gtk_calendar_select_day", "GtkCalendar*");
  XEN_ASSERT_TYPE(XEN_guint_P(day), day, 2, "gtk_calendar_select_day", "guint");
  gtk_calendar_select_day(XEN_TO_C_GtkCalendar_(calendar), XEN_TO_C_guint(day));
  return(XEN_FALSE);
}

static XEN gxg_gtk_calendar_mark_day(XEN calendar, XEN day)
{
  #define H_gtk_calendar_mark_day "gboolean gtk_calendar_mark_day(GtkCalendar* calendar, guint day)"
  XEN_ASSERT_TYPE(XEN_GtkCalendar__P(calendar), calendar, 1, "gtk_calendar_mark_day", "GtkCalendar*");
  XEN_ASSERT_TYPE(XEN_guint_P(day), day, 2, "gtk_calendar_mark_day", "guint");
  return(C_TO_XEN_gboolean(gtk_calendar_mark_day(XEN_TO_C_GtkCalendar_(calendar), XEN_TO_C_guint(day))));
}

static XEN gxg_gtk_calendar_unmark_day(XEN calendar, XEN day)
{
  #define H_gtk_calendar_unmark_day "gboolean gtk_calendar_unmark_day(GtkCalendar* calendar, guint day)"
  XEN_ASSERT_TYPE(XEN_GtkCalendar__P(calendar), calendar, 1, "gtk_calendar_unmark_day", "GtkCalendar*");
  XEN_ASSERT_TYPE(XEN_guint_P(day), day, 2, "gtk_calendar_unmark_day", "guint");
  return(C_TO_XEN_gboolean(gtk_calendar_unmark_day(XEN_TO_C_GtkCalendar_(calendar), XEN_TO_C_guint(day))));
}

static XEN gxg_gtk_calendar_clear_marks(XEN calendar)
{
  #define H_gtk_calendar_clear_marks "void gtk_calendar_clear_marks(GtkCalendar* calendar)"
  XEN_ASSERT_TYPE(XEN_GtkCalendar__P(calendar), calendar, 1, "gtk_calendar_clear_marks", "GtkCalendar*");
  gtk_calendar_clear_marks(XEN_TO_C_GtkCalendar_(calendar));
  return(XEN_FALSE);
}

static XEN gxg_gtk_calendar_display_options(XEN calendar, XEN flags)
{
  #define H_gtk_calendar_display_options "void gtk_calendar_display_options(GtkCalendar* calendar, GtkCalendarDisplayOptions flags)"
  XEN_ASSERT_TYPE(XEN_GtkCalendar__P(calendar), calendar, 1, "gtk_calendar_display_options", "GtkCalendar*");
  XEN_ASSERT_TYPE(XEN_GtkCalendarDisplayOptions_P(flags), flags, 2, "gtk_calendar_display_options", "GtkCalendarDisplayOptions");
  gtk_calendar_display_options(XEN_TO_C_GtkCalendar_(calendar), XEN_TO_C_GtkCalendarDisplayOptions(flags));
  return(XEN_FALSE);
}

static XEN gxg_gtk_calendar_get_date(XEN calendar, XEN year, XEN month, XEN day)
{
  #define H_gtk_calendar_get_date "void gtk_calendar_get_date(GtkCalendar* calendar, guint* year, guint* month, \
guint* day)"
  XEN_ASSERT_TYPE(XEN_GtkCalendar__P(calendar), calendar, 1, "gtk_calendar_get_date", "GtkCalendar*");
  XEN_ASSERT_TYPE(XEN_guint__P(year), year, 2, "gtk_calendar_get_date", "guint*");
  XEN_ASSERT_TYPE(XEN_guint__P(month), month, 3, "gtk_calendar_get_date", "guint*");
  XEN_ASSERT_TYPE(XEN_guint__P(day), day, 4, "gtk_calendar_get_date", "guint*");
  gtk_calendar_get_date(XEN_TO_C_GtkCalendar_(calendar), XEN_TO_C_guint_(year), XEN_TO_C_guint_(month), XEN_TO_C_guint_(day));
  return(XEN_FALSE);
}

static XEN gxg_gtk_calendar_freeze(XEN calendar)
{
  #define H_gtk_calendar_freeze "void gtk_calendar_freeze(GtkCalendar* calendar)"
  XEN_ASSERT_TYPE(XEN_GtkCalendar__P(calendar), calendar, 1, "gtk_calendar_freeze", "GtkCalendar*");
  gtk_calendar_freeze(XEN_TO_C_GtkCalendar_(calendar));
  return(XEN_FALSE);
}

static XEN gxg_gtk_calendar_thaw(XEN calendar)
{
  #define H_gtk_calendar_thaw "void gtk_calendar_thaw(GtkCalendar* calendar)"
  XEN_ASSERT_TYPE(XEN_GtkCalendar__P(calendar), calendar, 1, "gtk_calendar_thaw", "GtkCalendar*");
  gtk_calendar_thaw(XEN_TO_C_GtkCalendar_(calendar));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_cell_editable_get_type(void)
{
  #define H_gtk_cell_editable_get_type "GType gtk_cell_editable_get_type( void)"
  return(C_TO_XEN_GType(gtk_cell_editable_get_type()));
}

static XEN gxg_gtk_cell_editable_start_editing(XEN cell_editable, XEN event)
{
  #define H_gtk_cell_editable_start_editing "void gtk_cell_editable_start_editing(GtkCellEditable* cell_editable, \
GdkEvent* event)"
  XEN_ASSERT_TYPE(XEN_GtkCellEditable__P(cell_editable), cell_editable, 1, "gtk_cell_editable_start_editing", "GtkCellEditable*");
  XEN_ASSERT_TYPE(XEN_GdkEvent__P(event) || XEN_FALSE_P(event), event, 2, "gtk_cell_editable_start_editing", "GdkEvent*");
  gtk_cell_editable_start_editing(XEN_TO_C_GtkCellEditable_(cell_editable), XEN_TO_C_GdkEvent_(event));
  return(XEN_FALSE);
}

static XEN gxg_gtk_cell_editable_editing_done(XEN cell_editable)
{
  #define H_gtk_cell_editable_editing_done "void gtk_cell_editable_editing_done(GtkCellEditable* cell_editable)"
  XEN_ASSERT_TYPE(XEN_GtkCellEditable__P(cell_editable), cell_editable, 1, "gtk_cell_editable_editing_done", "GtkCellEditable*");
  gtk_cell_editable_editing_done(XEN_TO_C_GtkCellEditable_(cell_editable));
  return(XEN_FALSE);
}

static XEN gxg_gtk_cell_editable_remove_widget(XEN cell_editable)
{
  #define H_gtk_cell_editable_remove_widget "void gtk_cell_editable_remove_widget(GtkCellEditable* cell_editable)"
  XEN_ASSERT_TYPE(XEN_GtkCellEditable__P(cell_editable), cell_editable, 1, "gtk_cell_editable_remove_widget", "GtkCellEditable*");
  gtk_cell_editable_remove_widget(XEN_TO_C_GtkCellEditable_(cell_editable));
  return(XEN_FALSE);
}

static XEN gxg_gtk_cell_renderer_get_type(void)
{
  #define H_gtk_cell_renderer_get_type "GType gtk_cell_renderer_get_type( void)"
  return(C_TO_XEN_GType(gtk_cell_renderer_get_type()));
}

static XEN gxg_gtk_cell_renderer_get_size(XEN cell, XEN widget, XEN cell_area, XEN x_offset, XEN y_offset, XEN width, XEN height)
{
  #define H_gtk_cell_renderer_get_size "void gtk_cell_renderer_get_size(GtkCellRenderer* cell, GtkWidget* widget, \
GdkRectangle* cell_area, gint* [x_offset], gint* [y_offset], gint* [width], gint* [height])"
  gint ref_x_offset;
  gint ref_y_offset;
  gint ref_width;
  gint ref_height;
  XEN_ASSERT_TYPE(XEN_GtkCellRenderer__P(cell), cell, 1, "gtk_cell_renderer_get_size", "GtkCellRenderer*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 2, "gtk_cell_renderer_get_size", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_GdkRectangle__P(cell_area) || XEN_FALSE_P(cell_area), cell_area, 3, "gtk_cell_renderer_get_size", "GdkRectangle*");
  gtk_cell_renderer_get_size(XEN_TO_C_GtkCellRenderer_(cell), XEN_TO_C_GtkWidget_(widget), XEN_TO_C_GdkRectangle_(cell_area), 
                             &ref_x_offset, &ref_y_offset, &ref_width, &ref_height);
  return(XEN_LIST_4(C_TO_XEN_gint(ref_x_offset), C_TO_XEN_gint(ref_y_offset), C_TO_XEN_gint(ref_width), C_TO_XEN_gint(ref_height)));
}

static XEN gxg_gtk_cell_renderer_render(XEN cell, XEN window, XEN widget, XEN background_area, XEN cell_area, XEN expose_area, XEN flags)
{
  #define H_gtk_cell_renderer_render "void gtk_cell_renderer_render(GtkCellRenderer* cell, GdkWindow* window, \
GtkWidget* widget, GdkRectangle* background_area, GdkRectangle* cell_area, GdkRectangle* expose_area, \
GtkCellRendererState flags)"
  XEN_ASSERT_TYPE(XEN_GtkCellRenderer__P(cell), cell, 1, "gtk_cell_renderer_render", "GtkCellRenderer*");
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 2, "gtk_cell_renderer_render", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 3, "gtk_cell_renderer_render", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_GdkRectangle__P(background_area), background_area, 4, "gtk_cell_renderer_render", "GdkRectangle*");
  XEN_ASSERT_TYPE(XEN_GdkRectangle__P(cell_area), cell_area, 5, "gtk_cell_renderer_render", "GdkRectangle*");
  XEN_ASSERT_TYPE(XEN_GdkRectangle__P(expose_area), expose_area, 6, "gtk_cell_renderer_render", "GdkRectangle*");
  XEN_ASSERT_TYPE(XEN_GtkCellRendererState_P(flags), flags, 7, "gtk_cell_renderer_render", "GtkCellRendererState");
  gtk_cell_renderer_render(XEN_TO_C_GtkCellRenderer_(cell), XEN_TO_C_GdkWindow_(window), XEN_TO_C_GtkWidget_(widget), XEN_TO_C_GdkRectangle_(background_area), 
                           XEN_TO_C_GdkRectangle_(cell_area), XEN_TO_C_GdkRectangle_(expose_area), XEN_TO_C_GtkCellRendererState(flags));
  return(XEN_FALSE);
}

static XEN gxg_gtk_cell_renderer_activate(XEN cell, XEN event, XEN widget, XEN path, XEN background_area, XEN cell_area, XEN flags)
{
  #define H_gtk_cell_renderer_activate "gboolean gtk_cell_renderer_activate(GtkCellRenderer* cell, GdkEvent* event, \
GtkWidget* widget, gchar* path, GdkRectangle* background_area, GdkRectangle* cell_area, GtkCellRendererState flags)"
  XEN_ASSERT_TYPE(XEN_GtkCellRenderer__P(cell), cell, 1, "gtk_cell_renderer_activate", "GtkCellRenderer*");
  XEN_ASSERT_TYPE(XEN_GdkEvent__P(event), event, 2, "gtk_cell_renderer_activate", "GdkEvent*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 3, "gtk_cell_renderer_activate", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_gchar__P(path), path, 4, "gtk_cell_renderer_activate", "gchar*");
  XEN_ASSERT_TYPE(XEN_GdkRectangle__P(background_area), background_area, 5, "gtk_cell_renderer_activate", "GdkRectangle*");
  XEN_ASSERT_TYPE(XEN_GdkRectangle__P(cell_area), cell_area, 6, "gtk_cell_renderer_activate", "GdkRectangle*");
  XEN_ASSERT_TYPE(XEN_GtkCellRendererState_P(flags), flags, 7, "gtk_cell_renderer_activate", "GtkCellRendererState");
  return(C_TO_XEN_gboolean(gtk_cell_renderer_activate(XEN_TO_C_GtkCellRenderer_(cell), XEN_TO_C_GdkEvent_(event), XEN_TO_C_GtkWidget_(widget), 
                                                      XEN_TO_C_gchar_(path), XEN_TO_C_GdkRectangle_(background_area), XEN_TO_C_GdkRectangle_(cell_area), 
                                                      XEN_TO_C_GtkCellRendererState(flags))));
}

static XEN gxg_gtk_cell_renderer_start_editing(XEN cell, XEN event, XEN widget, XEN path, XEN background_area, XEN cell_area, XEN flags)
{
  #define H_gtk_cell_renderer_start_editing "GtkCellEditable* gtk_cell_renderer_start_editing(GtkCellRenderer* cell, \
GdkEvent* event, GtkWidget* widget, gchar* path, GdkRectangle* background_area, GdkRectangle* cell_area, \
GtkCellRendererState flags)"
  XEN_ASSERT_TYPE(XEN_GtkCellRenderer__P(cell), cell, 1, "gtk_cell_renderer_start_editing", "GtkCellRenderer*");
  XEN_ASSERT_TYPE(XEN_GdkEvent__P(event), event, 2, "gtk_cell_renderer_start_editing", "GdkEvent*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 3, "gtk_cell_renderer_start_editing", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_gchar__P(path), path, 4, "gtk_cell_renderer_start_editing", "gchar*");
  XEN_ASSERT_TYPE(XEN_GdkRectangle__P(background_area), background_area, 5, "gtk_cell_renderer_start_editing", "GdkRectangle*");
  XEN_ASSERT_TYPE(XEN_GdkRectangle__P(cell_area), cell_area, 6, "gtk_cell_renderer_start_editing", "GdkRectangle*");
  XEN_ASSERT_TYPE(XEN_GtkCellRendererState_P(flags), flags, 7, "gtk_cell_renderer_start_editing", "GtkCellRendererState");
  return(C_TO_XEN_GtkCellEditable_(gtk_cell_renderer_start_editing(XEN_TO_C_GtkCellRenderer_(cell), XEN_TO_C_GdkEvent_(event), 
                                                                   XEN_TO_C_GtkWidget_(widget), XEN_TO_C_gchar_(path), XEN_TO_C_GdkRectangle_(background_area), 
                                                                   XEN_TO_C_GdkRectangle_(cell_area), XEN_TO_C_GtkCellRendererState(flags))));
}

static XEN gxg_gtk_cell_renderer_set_fixed_size(XEN cell, XEN width, XEN height)
{
  #define H_gtk_cell_renderer_set_fixed_size "void gtk_cell_renderer_set_fixed_size(GtkCellRenderer* cell, \
gint width, gint height)"
  XEN_ASSERT_TYPE(XEN_GtkCellRenderer__P(cell), cell, 1, "gtk_cell_renderer_set_fixed_size", "GtkCellRenderer*");
  XEN_ASSERT_TYPE(XEN_gint_P(width), width, 2, "gtk_cell_renderer_set_fixed_size", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(height), height, 3, "gtk_cell_renderer_set_fixed_size", "gint");
  gtk_cell_renderer_set_fixed_size(XEN_TO_C_GtkCellRenderer_(cell), XEN_TO_C_gint(width), XEN_TO_C_gint(height));
  return(XEN_FALSE);
}

static XEN gxg_gtk_cell_renderer_get_fixed_size(XEN cell, XEN width, XEN height)
{
  #define H_gtk_cell_renderer_get_fixed_size "void gtk_cell_renderer_get_fixed_size(GtkCellRenderer* cell, \
gint* [width], gint* [height])"
  gint ref_width;
  gint ref_height;
  XEN_ASSERT_TYPE(XEN_GtkCellRenderer__P(cell), cell, 1, "gtk_cell_renderer_get_fixed_size", "GtkCellRenderer*");
  gtk_cell_renderer_get_fixed_size(XEN_TO_C_GtkCellRenderer_(cell), &ref_width, &ref_height);
  return(XEN_LIST_2(C_TO_XEN_gint(ref_width), C_TO_XEN_gint(ref_height)));
}

static XEN gxg_gtk_cell_renderer_pixbuf_get_type(void)
{
  #define H_gtk_cell_renderer_pixbuf_get_type "GtkType gtk_cell_renderer_pixbuf_get_type( void)"
  return(C_TO_XEN_GtkType(gtk_cell_renderer_pixbuf_get_type()));
}

static XEN gxg_gtk_cell_renderer_pixbuf_new(void)
{
  #define H_gtk_cell_renderer_pixbuf_new "GtkCellRenderer* gtk_cell_renderer_pixbuf_new( void)"
  return(C_TO_XEN_GtkCellRenderer_(gtk_cell_renderer_pixbuf_new()));
}

static XEN gxg_gtk_cell_renderer_text_get_type(void)
{
  #define H_gtk_cell_renderer_text_get_type "GtkType gtk_cell_renderer_text_get_type( void)"
  return(C_TO_XEN_GtkType(gtk_cell_renderer_text_get_type()));
}

static XEN gxg_gtk_cell_renderer_text_new(void)
{
  #define H_gtk_cell_renderer_text_new "GtkCellRenderer* gtk_cell_renderer_text_new( void)"
  return(C_TO_XEN_GtkCellRenderer_(gtk_cell_renderer_text_new()));
}

static XEN gxg_gtk_cell_renderer_text_set_fixed_height_from_font(XEN renderer, XEN number_of_rows)
{
  #define H_gtk_cell_renderer_text_set_fixed_height_from_font "void gtk_cell_renderer_text_set_fixed_height_from_font(GtkCellRendererText* renderer, \
gint number_of_rows)"
  XEN_ASSERT_TYPE(XEN_GtkCellRendererText__P(renderer), renderer, 1, "gtk_cell_renderer_text_set_fixed_height_from_font", "GtkCellRendererText*");
  XEN_ASSERT_TYPE(XEN_gint_P(number_of_rows), number_of_rows, 2, "gtk_cell_renderer_text_set_fixed_height_from_font", "gint");
  gtk_cell_renderer_text_set_fixed_height_from_font(XEN_TO_C_GtkCellRendererText_(renderer), XEN_TO_C_gint(number_of_rows));
  return(XEN_FALSE);
}

static XEN gxg_gtk_cell_renderer_toggle_get_type(void)
{
  #define H_gtk_cell_renderer_toggle_get_type "GtkType gtk_cell_renderer_toggle_get_type( void)"
  return(C_TO_XEN_GtkType(gtk_cell_renderer_toggle_get_type()));
}

static XEN gxg_gtk_cell_renderer_toggle_new(void)
{
  #define H_gtk_cell_renderer_toggle_new "GtkCellRenderer* gtk_cell_renderer_toggle_new( void)"
  return(C_TO_XEN_GtkCellRenderer_(gtk_cell_renderer_toggle_new()));
}

static XEN gxg_gtk_cell_renderer_toggle_get_radio(XEN toggle)
{
  #define H_gtk_cell_renderer_toggle_get_radio "gboolean gtk_cell_renderer_toggle_get_radio(GtkCellRendererToggle* toggle)"
  XEN_ASSERT_TYPE(XEN_GtkCellRendererToggle__P(toggle), toggle, 1, "gtk_cell_renderer_toggle_get_radio", "GtkCellRendererToggle*");
  return(C_TO_XEN_gboolean(gtk_cell_renderer_toggle_get_radio(XEN_TO_C_GtkCellRendererToggle_(toggle))));
}

static XEN gxg_gtk_cell_renderer_toggle_set_radio(XEN toggle, XEN radio)
{
  #define H_gtk_cell_renderer_toggle_set_radio "void gtk_cell_renderer_toggle_set_radio(GtkCellRendererToggle* toggle, \
gboolean radio)"
  XEN_ASSERT_TYPE(XEN_GtkCellRendererToggle__P(toggle), toggle, 1, "gtk_cell_renderer_toggle_set_radio", "GtkCellRendererToggle*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(radio), radio, 2, "gtk_cell_renderer_toggle_set_radio", "gboolean");
  gtk_cell_renderer_toggle_set_radio(XEN_TO_C_GtkCellRendererToggle_(toggle), XEN_TO_C_gboolean(radio));
  return(XEN_FALSE);
}

static XEN gxg_gtk_cell_renderer_toggle_get_active(XEN toggle)
{
  #define H_gtk_cell_renderer_toggle_get_active "gboolean gtk_cell_renderer_toggle_get_active(GtkCellRendererToggle* toggle)"
  XEN_ASSERT_TYPE(XEN_GtkCellRendererToggle__P(toggle), toggle, 1, "gtk_cell_renderer_toggle_get_active", "GtkCellRendererToggle*");
  return(C_TO_XEN_gboolean(gtk_cell_renderer_toggle_get_active(XEN_TO_C_GtkCellRendererToggle_(toggle))));
}

static XEN gxg_gtk_cell_renderer_toggle_set_active(XEN toggle, XEN setting)
{
  #define H_gtk_cell_renderer_toggle_set_active "void gtk_cell_renderer_toggle_set_active(GtkCellRendererToggle* toggle, \
gboolean setting)"
  XEN_ASSERT_TYPE(XEN_GtkCellRendererToggle__P(toggle), toggle, 1, "gtk_cell_renderer_toggle_set_active", "GtkCellRendererToggle*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(setting), setting, 2, "gtk_cell_renderer_toggle_set_active", "gboolean");
  gtk_cell_renderer_toggle_set_active(XEN_TO_C_GtkCellRendererToggle_(toggle), XEN_TO_C_gboolean(setting));
  return(XEN_FALSE);
}
#endif

static XEN gxg_gtk_check_button_get_type(void)
{
  #define H_gtk_check_button_get_type "GtkType gtk_check_button_get_type( void)"
  return(C_TO_XEN_GtkType(gtk_check_button_get_type()));
}

static XEN gxg_gtk_check_button_new(void)
{
  #define H_gtk_check_button_new "GtkWidget* gtk_check_button_new( void)"
  return(C_TO_XEN_GtkWidget_(gtk_check_button_new()));
}

static XEN gxg_gtk_check_button_new_with_label(XEN label)
{
  #define H_gtk_check_button_new_with_label "GtkWidget* gtk_check_button_new_with_label(gchar* label)"
  XEN_ASSERT_TYPE(XEN_gchar__P(label), label, 1, "gtk_check_button_new_with_label", "gchar*");
  return(C_TO_XEN_GtkWidget_(gtk_check_button_new_with_label(XEN_TO_C_gchar_(label))));
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_check_button_new_with_mnemonic(XEN label)
{
  #define H_gtk_check_button_new_with_mnemonic "GtkWidget* gtk_check_button_new_with_mnemonic(gchar* label)"
  XEN_ASSERT_TYPE(XEN_gchar__P(label), label, 1, "gtk_check_button_new_with_mnemonic", "gchar*");
  return(C_TO_XEN_GtkWidget_(gtk_check_button_new_with_mnemonic(XEN_TO_C_gchar_(label))));
}
#endif

static XEN gxg_gtk_check_menu_item_get_type(void)
{
  #define H_gtk_check_menu_item_get_type "GtkType gtk_check_menu_item_get_type( void)"
  return(C_TO_XEN_GtkType(gtk_check_menu_item_get_type()));
}

static XEN gxg_gtk_check_menu_item_new(void)
{
  #define H_gtk_check_menu_item_new "GtkWidget* gtk_check_menu_item_new( void)"
  return(C_TO_XEN_GtkWidget_(gtk_check_menu_item_new()));
}

static XEN gxg_gtk_check_menu_item_new_with_label(XEN label)
{
  #define H_gtk_check_menu_item_new_with_label "GtkWidget* gtk_check_menu_item_new_with_label(gchar* label)"
  XEN_ASSERT_TYPE(XEN_gchar__P(label), label, 1, "gtk_check_menu_item_new_with_label", "gchar*");
  return(C_TO_XEN_GtkWidget_(gtk_check_menu_item_new_with_label(XEN_TO_C_gchar_(label))));
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_check_menu_item_new_with_mnemonic(XEN label)
{
  #define H_gtk_check_menu_item_new_with_mnemonic "GtkWidget* gtk_check_menu_item_new_with_mnemonic(gchar* label)"
  XEN_ASSERT_TYPE(XEN_gchar__P(label), label, 1, "gtk_check_menu_item_new_with_mnemonic", "gchar*");
  return(C_TO_XEN_GtkWidget_(gtk_check_menu_item_new_with_mnemonic(XEN_TO_C_gchar_(label))));
}
#endif

static XEN gxg_gtk_check_menu_item_set_active(XEN check_menu_item, XEN is_active)
{
  #define H_gtk_check_menu_item_set_active "void gtk_check_menu_item_set_active(GtkCheckMenuItem* check_menu_item, \
gboolean is_active)"
  XEN_ASSERT_TYPE(XEN_GtkCheckMenuItem__P(check_menu_item), check_menu_item, 1, "gtk_check_menu_item_set_active", "GtkCheckMenuItem*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(is_active), is_active, 2, "gtk_check_menu_item_set_active", "gboolean");
  gtk_check_menu_item_set_active(XEN_TO_C_GtkCheckMenuItem_(check_menu_item), XEN_TO_C_gboolean(is_active));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_check_menu_item_get_active(XEN check_menu_item)
{
  #define H_gtk_check_menu_item_get_active "gboolean gtk_check_menu_item_get_active(GtkCheckMenuItem* check_menu_item)"
  XEN_ASSERT_TYPE(XEN_GtkCheckMenuItem__P(check_menu_item), check_menu_item, 1, "gtk_check_menu_item_get_active", "GtkCheckMenuItem*");
  return(C_TO_XEN_gboolean(gtk_check_menu_item_get_active(XEN_TO_C_GtkCheckMenuItem_(check_menu_item))));
}
#endif

static XEN gxg_gtk_check_menu_item_toggled(XEN check_menu_item)
{
  #define H_gtk_check_menu_item_toggled "void gtk_check_menu_item_toggled(GtkCheckMenuItem* check_menu_item)"
  XEN_ASSERT_TYPE(XEN_GtkCheckMenuItem__P(check_menu_item), check_menu_item, 1, "gtk_check_menu_item_toggled", "GtkCheckMenuItem*");
  gtk_check_menu_item_toggled(XEN_TO_C_GtkCheckMenuItem_(check_menu_item));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_check_menu_item_set_inconsistent(XEN check_menu_item, XEN setting)
{
  #define H_gtk_check_menu_item_set_inconsistent "void gtk_check_menu_item_set_inconsistent(GtkCheckMenuItem* check_menu_item, \
gboolean setting)"
  XEN_ASSERT_TYPE(XEN_GtkCheckMenuItem__P(check_menu_item), check_menu_item, 1, "gtk_check_menu_item_set_inconsistent", "GtkCheckMenuItem*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(setting), setting, 2, "gtk_check_menu_item_set_inconsistent", "gboolean");
  gtk_check_menu_item_set_inconsistent(XEN_TO_C_GtkCheckMenuItem_(check_menu_item), XEN_TO_C_gboolean(setting));
  return(XEN_FALSE);
}

static XEN gxg_gtk_check_menu_item_get_inconsistent(XEN check_menu_item)
{
  #define H_gtk_check_menu_item_get_inconsistent "gboolean gtk_check_menu_item_get_inconsistent(GtkCheckMenuItem* check_menu_item)"
  XEN_ASSERT_TYPE(XEN_GtkCheckMenuItem__P(check_menu_item), check_menu_item, 1, "gtk_check_menu_item_get_inconsistent", "GtkCheckMenuItem*");
  return(C_TO_XEN_gboolean(gtk_check_menu_item_get_inconsistent(XEN_TO_C_GtkCheckMenuItem_(check_menu_item))));
}

static XEN gxg_gtk_clipboard_get(XEN selection)
{
  #define H_gtk_clipboard_get "GtkClipboard* gtk_clipboard_get(GdkAtom selection)"
  XEN_ASSERT_TYPE(XEN_GdkAtom_P(selection), selection, 1, "gtk_clipboard_get", "GdkAtom");
  return(C_TO_XEN_GtkClipboard_(gtk_clipboard_get(XEN_TO_C_GdkAtom(selection))));
}

static XEN gxg_gtk_clipboard_set_with_data(XEN clipboard, XEN targets, XEN n_targets, XEN func, XEN clear_func, XEN func_data)
{
  #define H_gtk_clipboard_set_with_data "gboolean gtk_clipboard_set_with_data(GtkClipboard* clipboard, \
GtkTargetEntry* targets, guint n_targets, GtkClipboardGetFunc func, GtkClipboardClearFunc clear_func, \
lambda_data func_data)"
  XEN_ASSERT_TYPE(XEN_GtkClipboard__P(clipboard), clipboard, 1, "gtk_clipboard_set_with_data", "GtkClipboard*");
  XEN_ASSERT_TYPE(XEN_GtkTargetEntry__P(targets), targets, 2, "gtk_clipboard_set_with_data", "GtkTargetEntry*");
  XEN_ASSERT_TYPE(XEN_guint_P(n_targets), n_targets, 3, "gtk_clipboard_set_with_data", "guint");
  XEN_ASSERT_TYPE(XEN_GtkClipboardGetFunc_P(func), func, 4, "gtk_clipboard_set_with_data", "GtkClipboardGetFunc");
  XEN_ASSERT_TYPE(XEN_GtkClipboardClearFunc_P(clear_func), clear_func, 5, "gtk_clipboard_set_with_data", "GtkClipboardClearFunc");
  if (XEN_NOT_BOUND_P(func_data)) func_data = XEN_FALSE; 
  else XEN_ASSERT_TYPE(XEN_lambda_data_P(func_data), func_data, 6, "gtk_clipboard_set_with_data", "lambda_data");
  {
    XEN result = XEN_FALSE;
    XEN gxg_ptr = XEN_LIST_5(func, func_data, XEN_FALSE, XEN_FALSE, XEN_FALSE);
    xm_protect(gxg_ptr);
    XEN_LIST_SET(gxg_ptr, 2, clear_func);
    result = C_TO_XEN_gboolean(gtk_clipboard_set_with_data(XEN_TO_C_GtkClipboard_(clipboard), XEN_TO_C_GtkTargetEntry_(targets), 
                                                           XEN_TO_C_guint(n_targets), XEN_TO_C_GtkClipboardGetFunc(func), 
                                                           XEN_TO_C_GtkClipboardClearFunc(clear_func), XEN_TO_C_lambda_data(func_data)));
    return(result);
   }
}

static XEN gxg_gtk_clipboard_get_owner(XEN clipboard)
{
  #define H_gtk_clipboard_get_owner "GObject* gtk_clipboard_get_owner(GtkClipboard* clipboard)"
  XEN_ASSERT_TYPE(XEN_GtkClipboard__P(clipboard), clipboard, 1, "gtk_clipboard_get_owner", "GtkClipboard*");
  return(C_TO_XEN_GObject_(gtk_clipboard_get_owner(XEN_TO_C_GtkClipboard_(clipboard))));
}

static XEN gxg_gtk_clipboard_clear(XEN clipboard)
{
  #define H_gtk_clipboard_clear "void gtk_clipboard_clear(GtkClipboard* clipboard)"
  XEN_ASSERT_TYPE(XEN_GtkClipboard__P(clipboard), clipboard, 1, "gtk_clipboard_clear", "GtkClipboard*");
  gtk_clipboard_clear(XEN_TO_C_GtkClipboard_(clipboard));
  return(XEN_FALSE);
}

static XEN gxg_gtk_clipboard_set_text(XEN clipboard, XEN text, XEN len)
{
  #define H_gtk_clipboard_set_text "void gtk_clipboard_set_text(GtkClipboard* clipboard, gchar* text, \
gint len)"
  XEN_ASSERT_TYPE(XEN_GtkClipboard__P(clipboard), clipboard, 1, "gtk_clipboard_set_text", "GtkClipboard*");
  XEN_ASSERT_TYPE(XEN_gchar__P(text), text, 2, "gtk_clipboard_set_text", "gchar*");
  XEN_ASSERT_TYPE(XEN_gint_P(len), len, 3, "gtk_clipboard_set_text", "gint");
  gtk_clipboard_set_text(XEN_TO_C_GtkClipboard_(clipboard), XEN_TO_C_gchar_(text), XEN_TO_C_gint(len));
  return(XEN_FALSE);
}

static XEN gxg_gtk_clipboard_request_contents(XEN clipboard, XEN target, XEN func, XEN func_data)
{
  #define H_gtk_clipboard_request_contents "void gtk_clipboard_request_contents(GtkClipboard* clipboard, \
GdkAtom target, GtkClipboardReceivedFunc func, lambda_data func_data)"
  XEN_ASSERT_TYPE(XEN_GtkClipboard__P(clipboard), clipboard, 1, "gtk_clipboard_request_contents", "GtkClipboard*");
  XEN_ASSERT_TYPE(XEN_GdkAtom_P(target), target, 2, "gtk_clipboard_request_contents", "GdkAtom");
  XEN_ASSERT_TYPE(XEN_GtkClipboardReceivedFunc_P(func), func, 3, "gtk_clipboard_request_contents", "GtkClipboardReceivedFunc");
  if (XEN_NOT_BOUND_P(func_data)) func_data = XEN_FALSE; 
  else XEN_ASSERT_TYPE(XEN_lambda_data_P(func_data), func_data, 4, "gtk_clipboard_request_contents", "lambda_data");
  {
    int loc;
    XEN gxg_ptr = XEN_LIST_5(func, func_data, XEN_FALSE, XEN_FALSE, XEN_FALSE);
    loc = xm_protect(gxg_ptr);
    XEN_LIST_SET(gxg_ptr, 2, C_TO_XEN_INT(loc));
    gtk_clipboard_request_contents(XEN_TO_C_GtkClipboard_(clipboard), XEN_TO_C_GdkAtom(target), XEN_TO_C_GtkClipboardReceivedFunc(func), 
                               XEN_TO_C_lambda_data(func_data));
    xm_unprotect_at(loc);
    return(XEN_FALSE);
   }
}

static XEN gxg_gtk_clipboard_request_text(XEN clipboard, XEN func, XEN func_data)
{
  #define H_gtk_clipboard_request_text "void gtk_clipboard_request_text(GtkClipboard* clipboard, GtkClipboardTextReceivedFunc func, \
lambda_data func_data)"
  XEN_ASSERT_TYPE(XEN_GtkClipboard__P(clipboard), clipboard, 1, "gtk_clipboard_request_text", "GtkClipboard*");
  XEN_ASSERT_TYPE(XEN_GtkClipboardTextReceivedFunc_P(func), func, 2, "gtk_clipboard_request_text", "GtkClipboardTextReceivedFunc");
  if (XEN_NOT_BOUND_P(func_data)) func_data = XEN_FALSE; 
  else XEN_ASSERT_TYPE(XEN_lambda_data_P(func_data), func_data, 3, "gtk_clipboard_request_text", "lambda_data");
  {
    int loc;
    XEN gxg_ptr = XEN_LIST_5(func, func_data, XEN_FALSE, XEN_FALSE, XEN_FALSE);
    loc = xm_protect(gxg_ptr);
    XEN_LIST_SET(gxg_ptr, 2, C_TO_XEN_INT(loc));
    gtk_clipboard_request_text(XEN_TO_C_GtkClipboard_(clipboard), XEN_TO_C_GtkClipboardTextReceivedFunc(func), XEN_TO_C_lambda_data(func_data));
    xm_unprotect_at(loc);
    return(XEN_FALSE);
   }
}

static XEN gxg_gtk_clipboard_wait_for_contents(XEN clipboard, XEN target)
{
  #define H_gtk_clipboard_wait_for_contents "GtkSelectionData* gtk_clipboard_wait_for_contents(GtkClipboard* clipboard, \
GdkAtom target)"
  XEN_ASSERT_TYPE(XEN_GtkClipboard__P(clipboard), clipboard, 1, "gtk_clipboard_wait_for_contents", "GtkClipboard*");
  XEN_ASSERT_TYPE(XEN_GdkAtom_P(target), target, 2, "gtk_clipboard_wait_for_contents", "GdkAtom");
  return(C_TO_XEN_GtkSelectionData_(gtk_clipboard_wait_for_contents(XEN_TO_C_GtkClipboard_(clipboard), XEN_TO_C_GdkAtom(target))));
}

static XEN gxg_gtk_clipboard_wait_for_text(XEN clipboard)
{
  #define H_gtk_clipboard_wait_for_text "gchar* gtk_clipboard_wait_for_text(GtkClipboard* clipboard)"
  XEN_ASSERT_TYPE(XEN_GtkClipboard__P(clipboard), clipboard, 1, "gtk_clipboard_wait_for_text", "GtkClipboard*");
  return(C_TO_XEN_gchar_(gtk_clipboard_wait_for_text(XEN_TO_C_GtkClipboard_(clipboard))));
}

static XEN gxg_gtk_clipboard_wait_is_text_available(XEN clipboard)
{
  #define H_gtk_clipboard_wait_is_text_available "gboolean gtk_clipboard_wait_is_text_available(GtkClipboard* clipboard)"
  XEN_ASSERT_TYPE(XEN_GtkClipboard__P(clipboard), clipboard, 1, "gtk_clipboard_wait_is_text_available", "GtkClipboard*");
  return(C_TO_XEN_gboolean(gtk_clipboard_wait_is_text_available(XEN_TO_C_GtkClipboard_(clipboard))));
}
#endif

static XEN gxg_gtk_color_selection_dialog_get_type(void)
{
  #define H_gtk_color_selection_dialog_get_type "GtkType gtk_color_selection_dialog_get_type( void)"
  return(C_TO_XEN_GtkType(gtk_color_selection_dialog_get_type()));
}

static XEN gxg_gtk_color_selection_dialog_new(XEN title)
{
  #define H_gtk_color_selection_dialog_new "GtkWidget* gtk_color_selection_dialog_new(gchar* title)"
  XEN_ASSERT_TYPE(XEN_gchar__P(title), title, 1, "gtk_color_selection_dialog_new", "gchar*");
  return(C_TO_XEN_GtkWidget_(gtk_color_selection_dialog_new(XEN_TO_C_gchar_(title))));
}

static XEN gxg_gtk_color_selection_get_type(void)
{
  #define H_gtk_color_selection_get_type "GtkType gtk_color_selection_get_type( void)"
  return(C_TO_XEN_GtkType(gtk_color_selection_get_type()));
}

static XEN gxg_gtk_color_selection_new(void)
{
  #define H_gtk_color_selection_new "GtkWidget* gtk_color_selection_new( void)"
  return(C_TO_XEN_GtkWidget_(gtk_color_selection_new()));
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_color_selection_get_has_opacity_control(XEN colorsel)
{
  #define H_gtk_color_selection_get_has_opacity_control "gboolean gtk_color_selection_get_has_opacity_control(GtkColorSelection* colorsel)"
  XEN_ASSERT_TYPE(XEN_GtkColorSelection__P(colorsel), colorsel, 1, "gtk_color_selection_get_has_opacity_control", "GtkColorSelection*");
  return(C_TO_XEN_gboolean(gtk_color_selection_get_has_opacity_control(XEN_TO_C_GtkColorSelection_(colorsel))));
}

static XEN gxg_gtk_color_selection_set_has_opacity_control(XEN colorsel, XEN has_opacity)
{
  #define H_gtk_color_selection_set_has_opacity_control "void gtk_color_selection_set_has_opacity_control(GtkColorSelection* colorsel, \
gboolean has_opacity)"
  XEN_ASSERT_TYPE(XEN_GtkColorSelection__P(colorsel), colorsel, 1, "gtk_color_selection_set_has_opacity_control", "GtkColorSelection*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(has_opacity), has_opacity, 2, "gtk_color_selection_set_has_opacity_control", "gboolean");
  gtk_color_selection_set_has_opacity_control(XEN_TO_C_GtkColorSelection_(colorsel), XEN_TO_C_gboolean(has_opacity));
  return(XEN_FALSE);
}

static XEN gxg_gtk_color_selection_get_has_palette(XEN colorsel)
{
  #define H_gtk_color_selection_get_has_palette "gboolean gtk_color_selection_get_has_palette(GtkColorSelection* colorsel)"
  XEN_ASSERT_TYPE(XEN_GtkColorSelection__P(colorsel), colorsel, 1, "gtk_color_selection_get_has_palette", "GtkColorSelection*");
  return(C_TO_XEN_gboolean(gtk_color_selection_get_has_palette(XEN_TO_C_GtkColorSelection_(colorsel))));
}

static XEN gxg_gtk_color_selection_set_has_palette(XEN colorsel, XEN has_palette)
{
  #define H_gtk_color_selection_set_has_palette "void gtk_color_selection_set_has_palette(GtkColorSelection* colorsel, \
gboolean has_palette)"
  XEN_ASSERT_TYPE(XEN_GtkColorSelection__P(colorsel), colorsel, 1, "gtk_color_selection_set_has_palette", "GtkColorSelection*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(has_palette), has_palette, 2, "gtk_color_selection_set_has_palette", "gboolean");
  gtk_color_selection_set_has_palette(XEN_TO_C_GtkColorSelection_(colorsel), XEN_TO_C_gboolean(has_palette));
  return(XEN_FALSE);
}

static XEN gxg_gtk_color_selection_set_current_color(XEN colorsel, XEN color)
{
  #define H_gtk_color_selection_set_current_color "void gtk_color_selection_set_current_color(GtkColorSelection* colorsel, \
GdkColor* color)"
  XEN_ASSERT_TYPE(XEN_GtkColorSelection__P(colorsel), colorsel, 1, "gtk_color_selection_set_current_color", "GtkColorSelection*");
  XEN_ASSERT_TYPE(XEN_GdkColor__P(color), color, 2, "gtk_color_selection_set_current_color", "GdkColor*");
  gtk_color_selection_set_current_color(XEN_TO_C_GtkColorSelection_(colorsel), XEN_TO_C_GdkColor_(color));
  return(XEN_FALSE);
}

static XEN gxg_gtk_color_selection_set_current_alpha(XEN colorsel, XEN alpha)
{
  #define H_gtk_color_selection_set_current_alpha "void gtk_color_selection_set_current_alpha(GtkColorSelection* colorsel, \
guint16 alpha)"
  XEN_ASSERT_TYPE(XEN_GtkColorSelection__P(colorsel), colorsel, 1, "gtk_color_selection_set_current_alpha", "GtkColorSelection*");
  XEN_ASSERT_TYPE(XEN_guint16_P(alpha), alpha, 2, "gtk_color_selection_set_current_alpha", "guint16");
  gtk_color_selection_set_current_alpha(XEN_TO_C_GtkColorSelection_(colorsel), XEN_TO_C_guint16(alpha));
  return(XEN_FALSE);
}

static XEN gxg_gtk_color_selection_get_current_color(XEN colorsel, XEN color)
{
  #define H_gtk_color_selection_get_current_color "void gtk_color_selection_get_current_color(GtkColorSelection* colorsel, \
GdkColor* color)"
  XEN_ASSERT_TYPE(XEN_GtkColorSelection__P(colorsel), colorsel, 1, "gtk_color_selection_get_current_color", "GtkColorSelection*");
  XEN_ASSERT_TYPE(XEN_GdkColor__P(color), color, 2, "gtk_color_selection_get_current_color", "GdkColor*");
  gtk_color_selection_get_current_color(XEN_TO_C_GtkColorSelection_(colorsel), XEN_TO_C_GdkColor_(color));
  return(XEN_FALSE);
}

static XEN gxg_gtk_color_selection_get_current_alpha(XEN colorsel)
{
  #define H_gtk_color_selection_get_current_alpha "guint16 gtk_color_selection_get_current_alpha(GtkColorSelection* colorsel)"
  XEN_ASSERT_TYPE(XEN_GtkColorSelection__P(colorsel), colorsel, 1, "gtk_color_selection_get_current_alpha", "GtkColorSelection*");
  return(C_TO_XEN_guint16(gtk_color_selection_get_current_alpha(XEN_TO_C_GtkColorSelection_(colorsel))));
}

static XEN gxg_gtk_color_selection_set_previous_color(XEN colorsel, XEN color)
{
  #define H_gtk_color_selection_set_previous_color "void gtk_color_selection_set_previous_color(GtkColorSelection* colorsel, \
GdkColor* color)"
  XEN_ASSERT_TYPE(XEN_GtkColorSelection__P(colorsel), colorsel, 1, "gtk_color_selection_set_previous_color", "GtkColorSelection*");
  XEN_ASSERT_TYPE(XEN_GdkColor__P(color), color, 2, "gtk_color_selection_set_previous_color", "GdkColor*");
  gtk_color_selection_set_previous_color(XEN_TO_C_GtkColorSelection_(colorsel), XEN_TO_C_GdkColor_(color));
  return(XEN_FALSE);
}

static XEN gxg_gtk_color_selection_set_previous_alpha(XEN colorsel, XEN alpha)
{
  #define H_gtk_color_selection_set_previous_alpha "void gtk_color_selection_set_previous_alpha(GtkColorSelection* colorsel, \
guint16 alpha)"
  XEN_ASSERT_TYPE(XEN_GtkColorSelection__P(colorsel), colorsel, 1, "gtk_color_selection_set_previous_alpha", "GtkColorSelection*");
  XEN_ASSERT_TYPE(XEN_guint16_P(alpha), alpha, 2, "gtk_color_selection_set_previous_alpha", "guint16");
  gtk_color_selection_set_previous_alpha(XEN_TO_C_GtkColorSelection_(colorsel), XEN_TO_C_guint16(alpha));
  return(XEN_FALSE);
}

static XEN gxg_gtk_color_selection_get_previous_color(XEN colorsel, XEN color)
{
  #define H_gtk_color_selection_get_previous_color "void gtk_color_selection_get_previous_color(GtkColorSelection* colorsel, \
GdkColor* color)"
  XEN_ASSERT_TYPE(XEN_GtkColorSelection__P(colorsel), colorsel, 1, "gtk_color_selection_get_previous_color", "GtkColorSelection*");
  XEN_ASSERT_TYPE(XEN_GdkColor__P(color), color, 2, "gtk_color_selection_get_previous_color", "GdkColor*");
  gtk_color_selection_get_previous_color(XEN_TO_C_GtkColorSelection_(colorsel), XEN_TO_C_GdkColor_(color));
  return(XEN_FALSE);
}

static XEN gxg_gtk_color_selection_get_previous_alpha(XEN colorsel)
{
  #define H_gtk_color_selection_get_previous_alpha "guint16 gtk_color_selection_get_previous_alpha(GtkColorSelection* colorsel)"
  XEN_ASSERT_TYPE(XEN_GtkColorSelection__P(colorsel), colorsel, 1, "gtk_color_selection_get_previous_alpha", "GtkColorSelection*");
  return(C_TO_XEN_guint16(gtk_color_selection_get_previous_alpha(XEN_TO_C_GtkColorSelection_(colorsel))));
}

static XEN gxg_gtk_color_selection_is_adjusting(XEN colorsel)
{
  #define H_gtk_color_selection_is_adjusting "gboolean gtk_color_selection_is_adjusting(GtkColorSelection* colorsel)"
  XEN_ASSERT_TYPE(XEN_GtkColorSelection__P(colorsel), colorsel, 1, "gtk_color_selection_is_adjusting", "GtkColorSelection*");
  return(C_TO_XEN_gboolean(gtk_color_selection_is_adjusting(XEN_TO_C_GtkColorSelection_(colorsel))));
}

static XEN gxg_gtk_color_selection_palette_from_string(XEN str, XEN colors, XEN n_colors)
{
  #define H_gtk_color_selection_palette_from_string "gboolean gtk_color_selection_palette_from_string(gchar* str, \
GdkColor** [colors], gint* [n_colors])"
  GdkColor* ref_colors;
  gint ref_n_colors;
  XEN_ASSERT_TYPE(XEN_gchar__P(str), str, 1, "gtk_color_selection_palette_from_string", "gchar*");
  {
    XEN result = XEN_FALSE;
    result = C_TO_XEN_gboolean(gtk_color_selection_palette_from_string(XEN_TO_C_gchar_(str), &ref_colors, &ref_n_colors));
    return(XEN_LIST_3(result, C_TO_XEN_GdkColor_(ref_colors), C_TO_XEN_gint(ref_n_colors)));
   }
}

static XEN gxg_gtk_color_selection_palette_to_string(XEN colors, XEN n_colors)
{
  #define H_gtk_color_selection_palette_to_string "gchar* gtk_color_selection_palette_to_string(GdkColor* colors, \
gint n_colors)"
  XEN_ASSERT_TYPE(XEN_GdkColor__P(colors), colors, 1, "gtk_color_selection_palette_to_string", "GdkColor*");
  XEN_ASSERT_TYPE(XEN_gint_P(n_colors), n_colors, 2, "gtk_color_selection_palette_to_string", "gint");
  return(C_TO_XEN_gchar_(gtk_color_selection_palette_to_string(XEN_TO_C_GdkColor_(colors), XEN_TO_C_gint(n_colors))));
}

static XEN gxg_gtk_color_selection_set_change_palette_hook(XEN func)
{
  #define H_gtk_color_selection_set_change_palette_hook "GtkColorSelectionChangePaletteFunc gtk_color_selection_set_change_palette_hook(GtkColorSelectionChangePaletteFunc func)"
  XEN_ASSERT_TYPE(XEN_GtkColorSelectionChangePaletteFunc_P(func), func, 1, "gtk_color_selection_set_change_palette_hook", "GtkColorSelectionChangePaletteFunc");
  return(C_TO_XEN_GtkColorSelectionChangePaletteFunc(gtk_color_selection_set_change_palette_hook(XEN_TO_C_GtkColorSelectionChangePaletteFunc(func))));
}
#endif

static XEN gxg_gtk_combo_get_type(void)
{
  #define H_gtk_combo_get_type "GtkType gtk_combo_get_type( void)"
  return(C_TO_XEN_GtkType(gtk_combo_get_type()));
}

static XEN gxg_gtk_combo_new(void)
{
  #define H_gtk_combo_new "GtkWidget* gtk_combo_new( void)"
  return(C_TO_XEN_GtkWidget_(gtk_combo_new()));
}

static XEN gxg_gtk_combo_set_value_in_list(XEN combo, XEN val, XEN ok_if_empty)
{
  #define H_gtk_combo_set_value_in_list "void gtk_combo_set_value_in_list(GtkCombo* combo, gboolean val, \
gboolean ok_if_empty)"
  XEN_ASSERT_TYPE(XEN_GtkCombo__P(combo), combo, 1, "gtk_combo_set_value_in_list", "GtkCombo*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(val), val, 2, "gtk_combo_set_value_in_list", "gboolean");
  XEN_ASSERT_TYPE(XEN_gboolean_P(ok_if_empty), ok_if_empty, 3, "gtk_combo_set_value_in_list", "gboolean");
  gtk_combo_set_value_in_list(XEN_TO_C_GtkCombo_(combo), XEN_TO_C_gboolean(val), XEN_TO_C_gboolean(ok_if_empty));
  return(XEN_FALSE);
}

static XEN gxg_gtk_combo_set_use_arrows(XEN combo, XEN val)
{
  #define H_gtk_combo_set_use_arrows "void gtk_combo_set_use_arrows(GtkCombo* combo, gboolean val)"
  XEN_ASSERT_TYPE(XEN_GtkCombo__P(combo), combo, 1, "gtk_combo_set_use_arrows", "GtkCombo*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(val), val, 2, "gtk_combo_set_use_arrows", "gboolean");
  gtk_combo_set_use_arrows(XEN_TO_C_GtkCombo_(combo), XEN_TO_C_gboolean(val));
  return(XEN_FALSE);
}

static XEN gxg_gtk_combo_set_use_arrows_always(XEN combo, XEN val)
{
  #define H_gtk_combo_set_use_arrows_always "void gtk_combo_set_use_arrows_always(GtkCombo* combo, gboolean val)"
  XEN_ASSERT_TYPE(XEN_GtkCombo__P(combo), combo, 1, "gtk_combo_set_use_arrows_always", "GtkCombo*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(val), val, 2, "gtk_combo_set_use_arrows_always", "gboolean");
  gtk_combo_set_use_arrows_always(XEN_TO_C_GtkCombo_(combo), XEN_TO_C_gboolean(val));
  return(XEN_FALSE);
}

static XEN gxg_gtk_combo_set_case_sensitive(XEN combo, XEN val)
{
  #define H_gtk_combo_set_case_sensitive "void gtk_combo_set_case_sensitive(GtkCombo* combo, gboolean val)"
  XEN_ASSERT_TYPE(XEN_GtkCombo__P(combo), combo, 1, "gtk_combo_set_case_sensitive", "GtkCombo*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(val), val, 2, "gtk_combo_set_case_sensitive", "gboolean");
  gtk_combo_set_case_sensitive(XEN_TO_C_GtkCombo_(combo), XEN_TO_C_gboolean(val));
  return(XEN_FALSE);
}

static XEN gxg_gtk_combo_set_item_string(XEN combo, XEN item, XEN item_value)
{
  #define H_gtk_combo_set_item_string "void gtk_combo_set_item_string(GtkCombo* combo, GtkItem* item, \
gchar* item_value)"
  XEN_ASSERT_TYPE(XEN_GtkCombo__P(combo), combo, 1, "gtk_combo_set_item_string", "GtkCombo*");
  XEN_ASSERT_TYPE(XEN_GtkItem__P(item), item, 2, "gtk_combo_set_item_string", "GtkItem*");
  XEN_ASSERT_TYPE(XEN_gchar__P(item_value), item_value, 3, "gtk_combo_set_item_string", "gchar*");
  gtk_combo_set_item_string(XEN_TO_C_GtkCombo_(combo), XEN_TO_C_GtkItem_(item), XEN_TO_C_gchar_(item_value));
  return(XEN_FALSE);
}

static XEN gxg_gtk_combo_set_popdown_strings(XEN combo, XEN strings)
{
  #define H_gtk_combo_set_popdown_strings "void gtk_combo_set_popdown_strings(GtkCombo* combo, GList* strings)"
  XEN_ASSERT_TYPE(XEN_GtkCombo__P(combo), combo, 1, "gtk_combo_set_popdown_strings", "GtkCombo*");
  XEN_ASSERT_TYPE(XEN_GList__P(strings), strings, 2, "gtk_combo_set_popdown_strings", "GList*");
  gtk_combo_set_popdown_strings(XEN_TO_C_GtkCombo_(combo), XEN_TO_C_GList_(strings));
  return(XEN_FALSE);
}

static XEN gxg_gtk_combo_disable_activate(XEN combo)
{
  #define H_gtk_combo_disable_activate "void gtk_combo_disable_activate(GtkCombo* combo)"
  XEN_ASSERT_TYPE(XEN_GtkCombo__P(combo), combo, 1, "gtk_combo_disable_activate", "GtkCombo*");
  gtk_combo_disable_activate(XEN_TO_C_GtkCombo_(combo));
  return(XEN_FALSE);
}

static XEN gxg_GTK_IS_RESIZE_CONTAINER(XEN widget)
{
  #define H_GTK_IS_RESIZE_CONTAINER "gboolean GTK_IS_RESIZE_CONTAINER(GtkWidget* widget)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "GTK_IS_RESIZE_CONTAINER", "GtkWidget*");
  return(C_TO_XEN_gboolean(GTK_IS_RESIZE_CONTAINER(XEN_TO_C_GtkWidget_(widget))));
}

static XEN gxg_gtk_container_get_type(void)
{
  #define H_gtk_container_get_type "GtkType gtk_container_get_type( void)"
  return(C_TO_XEN_GtkType(gtk_container_get_type()));
}

static XEN gxg_gtk_container_set_border_width(XEN container, XEN border_width)
{
  #define H_gtk_container_set_border_width "void gtk_container_set_border_width(GtkContainer* container, \
guint border_width)"
  XEN_ASSERT_TYPE(XEN_GtkContainer__P(container), container, 1, "gtk_container_set_border_width", "GtkContainer*");
  XEN_ASSERT_TYPE(XEN_guint_P(border_width), border_width, 2, "gtk_container_set_border_width", "guint");
  gtk_container_set_border_width(XEN_TO_C_GtkContainer_(container), XEN_TO_C_guint(border_width));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_container_get_border_width(XEN container)
{
  #define H_gtk_container_get_border_width "guint gtk_container_get_border_width(GtkContainer* container)"
  XEN_ASSERT_TYPE(XEN_GtkContainer__P(container), container, 1, "gtk_container_get_border_width", "GtkContainer*");
  return(C_TO_XEN_guint(gtk_container_get_border_width(XEN_TO_C_GtkContainer_(container))));
}
#endif

static XEN gxg_gtk_container_add(XEN container, XEN widget)
{
  #define H_gtk_container_add "void gtk_container_add(GtkContainer* container, GtkWidget* widget)"
  XEN_ASSERT_TYPE(XEN_GtkContainer__P(container), container, 1, "gtk_container_add", "GtkContainer*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 2, "gtk_container_add", "GtkWidget*");
  gtk_container_add(XEN_TO_C_GtkContainer_(container), XEN_TO_C_GtkWidget_(widget));
  return(XEN_FALSE);
}

static XEN gxg_gtk_container_remove(XEN container, XEN widget)
{
  #define H_gtk_container_remove "void gtk_container_remove(GtkContainer* container, GtkWidget* widget)"
  XEN_ASSERT_TYPE(XEN_GtkContainer__P(container), container, 1, "gtk_container_remove", "GtkContainer*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 2, "gtk_container_remove", "GtkWidget*");
  gtk_container_remove(XEN_TO_C_GtkContainer_(container), XEN_TO_C_GtkWidget_(widget));
  return(XEN_FALSE);
}

static XEN gxg_gtk_container_set_resize_mode(XEN container, XEN resize_mode)
{
  #define H_gtk_container_set_resize_mode "void gtk_container_set_resize_mode(GtkContainer* container, \
GtkResizeMode resize_mode)"
  XEN_ASSERT_TYPE(XEN_GtkContainer__P(container), container, 1, "gtk_container_set_resize_mode", "GtkContainer*");
  XEN_ASSERT_TYPE(XEN_GtkResizeMode_P(resize_mode), resize_mode, 2, "gtk_container_set_resize_mode", "GtkResizeMode");
  gtk_container_set_resize_mode(XEN_TO_C_GtkContainer_(container), XEN_TO_C_GtkResizeMode(resize_mode));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_container_get_resize_mode(XEN container)
{
  #define H_gtk_container_get_resize_mode "GtkResizeMode gtk_container_get_resize_mode(GtkContainer* container)"
  XEN_ASSERT_TYPE(XEN_GtkContainer__P(container), container, 1, "gtk_container_get_resize_mode", "GtkContainer*");
  return(C_TO_XEN_GtkResizeMode(gtk_container_get_resize_mode(XEN_TO_C_GtkContainer_(container))));
}
#endif

static XEN gxg_gtk_container_check_resize(XEN container)
{
  #define H_gtk_container_check_resize "void gtk_container_check_resize(GtkContainer* container)"
  XEN_ASSERT_TYPE(XEN_GtkContainer__P(container), container, 1, "gtk_container_check_resize", "GtkContainer*");
  gtk_container_check_resize(XEN_TO_C_GtkContainer_(container));
  return(XEN_FALSE);
}

static XEN gxg_gtk_container_foreach(XEN container, XEN func, XEN func_data)
{
  #define H_gtk_container_foreach "void gtk_container_foreach(GtkContainer* container, GtkCallback func, \
lambda_data func_data)"
  XEN_ASSERT_TYPE(XEN_GtkContainer__P(container), container, 1, "gtk_container_foreach", "GtkContainer*");
  XEN_ASSERT_TYPE(XEN_GtkCallback_P(func), func, 2, "gtk_container_foreach", "GtkCallback");
  if (XEN_NOT_BOUND_P(func_data)) func_data = XEN_FALSE; 
  else XEN_ASSERT_TYPE(XEN_lambda_data_P(func_data), func_data, 3, "gtk_container_foreach", "lambda_data");
  {
    int loc;
    XEN gxg_ptr = XEN_LIST_5(func, func_data, XEN_FALSE, XEN_FALSE, XEN_FALSE);
    loc = xm_protect(gxg_ptr);
    XEN_LIST_SET(gxg_ptr, 2, C_TO_XEN_INT(loc));
    gtk_container_foreach(XEN_TO_C_GtkContainer_(container), XEN_TO_C_GtkCallback(func), XEN_TO_C_lambda_data(func_data));
    xm_unprotect_at(loc);
    return(XEN_FALSE);
   }
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_container_get_children(XEN container)
{
  #define H_gtk_container_get_children "GList* gtk_container_get_children(GtkContainer* container)"
  XEN_ASSERT_TYPE(XEN_GtkContainer__P(container), container, 1, "gtk_container_get_children", "GtkContainer*");
  return(C_TO_XEN_GList_(gtk_container_get_children(XEN_TO_C_GtkContainer_(container))));
}
#endif

static XEN gxg_gtk_curve_get_type(void)
{
  #define H_gtk_curve_get_type "GtkType gtk_curve_get_type( void)"
  return(C_TO_XEN_GtkType(gtk_curve_get_type()));
}

static XEN gxg_gtk_curve_new(void)
{
  #define H_gtk_curve_new "GtkWidget* gtk_curve_new( void)"
  return(C_TO_XEN_GtkWidget_(gtk_curve_new()));
}

static XEN gxg_gtk_curve_reset(XEN curve)
{
  #define H_gtk_curve_reset "void gtk_curve_reset(GtkCurve* curve)"
  XEN_ASSERT_TYPE(XEN_GtkCurve__P(curve), curve, 1, "gtk_curve_reset", "GtkCurve*");
  gtk_curve_reset(XEN_TO_C_GtkCurve_(curve));
  return(XEN_FALSE);
}

static XEN gxg_gtk_curve_set_gamma(XEN curve, XEN gamma)
{
  #define H_gtk_curve_set_gamma "void gtk_curve_set_gamma(GtkCurve* curve, gfloat gamma)"
  XEN_ASSERT_TYPE(XEN_GtkCurve__P(curve), curve, 1, "gtk_curve_set_gamma", "GtkCurve*");
  XEN_ASSERT_TYPE(XEN_gfloat_P(gamma), gamma, 2, "gtk_curve_set_gamma", "gfloat");
  gtk_curve_set_gamma(XEN_TO_C_GtkCurve_(curve), XEN_TO_C_gfloat(gamma));
  return(XEN_FALSE);
}

static XEN gxg_gtk_curve_set_range(XEN curve, XEN min_x, XEN max_x, XEN min_y, XEN max_y)
{
  #define H_gtk_curve_set_range "void gtk_curve_set_range(GtkCurve* curve, gfloat min_x, gfloat max_x, \
gfloat min_y, gfloat max_y)"
  XEN_ASSERT_TYPE(XEN_GtkCurve__P(curve), curve, 1, "gtk_curve_set_range", "GtkCurve*");
  XEN_ASSERT_TYPE(XEN_gfloat_P(min_x), min_x, 2, "gtk_curve_set_range", "gfloat");
  XEN_ASSERT_TYPE(XEN_gfloat_P(max_x), max_x, 3, "gtk_curve_set_range", "gfloat");
  XEN_ASSERT_TYPE(XEN_gfloat_P(min_y), min_y, 4, "gtk_curve_set_range", "gfloat");
  XEN_ASSERT_TYPE(XEN_gfloat_P(max_y), max_y, 5, "gtk_curve_set_range", "gfloat");
  gtk_curve_set_range(XEN_TO_C_GtkCurve_(curve), XEN_TO_C_gfloat(min_x), XEN_TO_C_gfloat(max_x), XEN_TO_C_gfloat(min_y), 
                      XEN_TO_C_gfloat(max_y));
  return(XEN_FALSE);
}

static XEN gxg_gtk_curve_get_vector(XEN curve, XEN veclen, XEN vector)
{
  #define H_gtk_curve_get_vector "void gtk_curve_get_vector(GtkCurve* curve, int veclen, gfloat* vector)"
  XEN_ASSERT_TYPE(XEN_GtkCurve__P(curve), curve, 1, "gtk_curve_get_vector", "GtkCurve*");
  XEN_ASSERT_TYPE(XEN_int_P(veclen), veclen, 2, "gtk_curve_get_vector", "int");
  XEN_ASSERT_TYPE(XEN_gfloat__P(vector), vector, 3, "gtk_curve_get_vector", "gfloat*");
  gtk_curve_get_vector(XEN_TO_C_GtkCurve_(curve), XEN_TO_C_int(veclen), XEN_TO_C_gfloat_(vector));
  return(XEN_FALSE);
}

static XEN gxg_gtk_curve_set_vector(XEN curve, XEN veclen, XEN vector)
{
  #define H_gtk_curve_set_vector "void gtk_curve_set_vector(GtkCurve* curve, int veclen, gfloat* vector)"
  XEN_ASSERT_TYPE(XEN_GtkCurve__P(curve), curve, 1, "gtk_curve_set_vector", "GtkCurve*");
  XEN_ASSERT_TYPE(XEN_int_P(veclen), veclen, 2, "gtk_curve_set_vector", "int");
  XEN_ASSERT_TYPE(XEN_gfloat__P(vector), vector, 3, "gtk_curve_set_vector", "gfloat*");
  gtk_curve_set_vector(XEN_TO_C_GtkCurve_(curve), XEN_TO_C_int(veclen), XEN_TO_C_gfloat_(vector));
  return(XEN_FALSE);
}

static XEN gxg_gtk_curve_set_curve_type(XEN curve, XEN type)
{
  #define H_gtk_curve_set_curve_type "void gtk_curve_set_curve_type(GtkCurve* curve, GtkCurveType type)"
  XEN_ASSERT_TYPE(XEN_GtkCurve__P(curve), curve, 1, "gtk_curve_set_curve_type", "GtkCurve*");
  XEN_ASSERT_TYPE(XEN_GtkCurveType_P(type), type, 2, "gtk_curve_set_curve_type", "GtkCurveType");
  gtk_curve_set_curve_type(XEN_TO_C_GtkCurve_(curve), XEN_TO_C_GtkCurveType(type));
  return(XEN_FALSE);
}

static XEN gxg_gtk_dialog_get_type(void)
{
  #define H_gtk_dialog_get_type "GtkType gtk_dialog_get_type( void)"
  return(C_TO_XEN_GtkType(gtk_dialog_get_type()));
}

static XEN gxg_gtk_dialog_new(void)
{
  #define H_gtk_dialog_new "GtkWidget* gtk_dialog_new( void)"
  return(C_TO_XEN_GtkWidget_(gtk_dialog_new()));
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_dialog_new_with_buttons(XEN title, XEN parent, XEN flags, XEN first_button_text)
{
  #define H_gtk_dialog_new_with_buttons "GtkWidget* gtk_dialog_new_with_buttons(gchar* title, GtkWindow* parent, \
GtkDialogFlags flags, gchar* first_button_text, ...)"
  XEN_ASSERT_TYPE(XEN_gchar__P(title), title, 1, "gtk_dialog_new_with_buttons", "gchar*");
  XEN_ASSERT_TYPE(XEN_GtkWindow__P(parent) || XEN_FALSE_P(parent), parent, 2, "gtk_dialog_new_with_buttons", "GtkWindow*");
  XEN_ASSERT_TYPE(XEN_GtkDialogFlags_P(flags), flags, 3, "gtk_dialog_new_with_buttons", "GtkDialogFlags");
  XEN_ASSERT_TYPE(XEN_gchar__P(first_button_text), first_button_text, 4, "gtk_dialog_new_with_buttons", "gchar*");
  return(C_TO_XEN_GtkWidget_(gtk_dialog_new_with_buttons(XEN_TO_C_gchar_(title), XEN_TO_C_GtkWindow_(parent), XEN_TO_C_GtkDialogFlags(flags), 
                                                         XEN_TO_C_gchar_(first_button_text))));
}

static XEN gxg_gtk_dialog_add_action_widget(XEN dialog, XEN child, XEN response_id)
{
  #define H_gtk_dialog_add_action_widget "void gtk_dialog_add_action_widget(GtkDialog* dialog, GtkWidget* child, \
gint response_id)"
  XEN_ASSERT_TYPE(XEN_GtkDialog__P(dialog), dialog, 1, "gtk_dialog_add_action_widget", "GtkDialog*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(child), child, 2, "gtk_dialog_add_action_widget", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_gint_P(response_id), response_id, 3, "gtk_dialog_add_action_widget", "gint");
  gtk_dialog_add_action_widget(XEN_TO_C_GtkDialog_(dialog), XEN_TO_C_GtkWidget_(child), XEN_TO_C_gint(response_id));
  return(XEN_FALSE);
}

static XEN gxg_gtk_dialog_add_button(XEN dialog, XEN button_text, XEN response_id)
{
  #define H_gtk_dialog_add_button "GtkWidget* gtk_dialog_add_button(GtkDialog* dialog, gchar* button_text, \
gint response_id)"
  XEN_ASSERT_TYPE(XEN_GtkDialog__P(dialog), dialog, 1, "gtk_dialog_add_button", "GtkDialog*");
  XEN_ASSERT_TYPE(XEN_gchar__P(button_text), button_text, 2, "gtk_dialog_add_button", "gchar*");
  XEN_ASSERT_TYPE(XEN_gint_P(response_id), response_id, 3, "gtk_dialog_add_button", "gint");
  return(C_TO_XEN_GtkWidget_(gtk_dialog_add_button(XEN_TO_C_GtkDialog_(dialog), XEN_TO_C_gchar_(button_text), XEN_TO_C_gint(response_id))));
}

static XEN gxg_gtk_dialog_add_buttons(XEN dialog, XEN first_button_text)
{
  #define H_gtk_dialog_add_buttons "void gtk_dialog_add_buttons(GtkDialog* dialog, gchar* first_button_text, \
...)"
  XEN_ASSERT_TYPE(XEN_GtkDialog__P(dialog), dialog, 1, "gtk_dialog_add_buttons", "GtkDialog*");
  XEN_ASSERT_TYPE(XEN_gchar__P(first_button_text), first_button_text, 2, "gtk_dialog_add_buttons", "gchar*");
  gtk_dialog_add_buttons(XEN_TO_C_GtkDialog_(dialog), XEN_TO_C_gchar_(first_button_text));
  return(XEN_FALSE);
}

static XEN gxg_gtk_dialog_set_response_sensitive(XEN dialog, XEN response_id, XEN setting)
{
  #define H_gtk_dialog_set_response_sensitive "void gtk_dialog_set_response_sensitive(GtkDialog* dialog, \
gint response_id, gboolean setting)"
  XEN_ASSERT_TYPE(XEN_GtkDialog__P(dialog), dialog, 1, "gtk_dialog_set_response_sensitive", "GtkDialog*");
  XEN_ASSERT_TYPE(XEN_gint_P(response_id), response_id, 2, "gtk_dialog_set_response_sensitive", "gint");
  XEN_ASSERT_TYPE(XEN_gboolean_P(setting), setting, 3, "gtk_dialog_set_response_sensitive", "gboolean");
  gtk_dialog_set_response_sensitive(XEN_TO_C_GtkDialog_(dialog), XEN_TO_C_gint(response_id), XEN_TO_C_gboolean(setting));
  return(XEN_FALSE);
}

static XEN gxg_gtk_dialog_set_default_response(XEN dialog, XEN response_id)
{
  #define H_gtk_dialog_set_default_response "void gtk_dialog_set_default_response(GtkDialog* dialog, \
gint response_id)"
  XEN_ASSERT_TYPE(XEN_GtkDialog__P(dialog), dialog, 1, "gtk_dialog_set_default_response", "GtkDialog*");
  XEN_ASSERT_TYPE(XEN_gint_P(response_id), response_id, 2, "gtk_dialog_set_default_response", "gint");
  gtk_dialog_set_default_response(XEN_TO_C_GtkDialog_(dialog), XEN_TO_C_gint(response_id));
  return(XEN_FALSE);
}

static XEN gxg_gtk_dialog_set_has_separator(XEN dialog, XEN setting)
{
  #define H_gtk_dialog_set_has_separator "void gtk_dialog_set_has_separator(GtkDialog* dialog, gboolean setting)"
  XEN_ASSERT_TYPE(XEN_GtkDialog__P(dialog), dialog, 1, "gtk_dialog_set_has_separator", "GtkDialog*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(setting), setting, 2, "gtk_dialog_set_has_separator", "gboolean");
  gtk_dialog_set_has_separator(XEN_TO_C_GtkDialog_(dialog), XEN_TO_C_gboolean(setting));
  return(XEN_FALSE);
}

static XEN gxg_gtk_dialog_get_has_separator(XEN dialog)
{
  #define H_gtk_dialog_get_has_separator "gboolean gtk_dialog_get_has_separator(GtkDialog* dialog)"
  XEN_ASSERT_TYPE(XEN_GtkDialog__P(dialog), dialog, 1, "gtk_dialog_get_has_separator", "GtkDialog*");
  return(C_TO_XEN_gboolean(gtk_dialog_get_has_separator(XEN_TO_C_GtkDialog_(dialog))));
}

static XEN gxg_gtk_dialog_response(XEN dialog, XEN response_id)
{
  #define H_gtk_dialog_response "void gtk_dialog_response(GtkDialog* dialog, gint response_id)"
  XEN_ASSERT_TYPE(XEN_GtkDialog__P(dialog), dialog, 1, "gtk_dialog_response", "GtkDialog*");
  XEN_ASSERT_TYPE(XEN_gint_P(response_id), response_id, 2, "gtk_dialog_response", "gint");
  gtk_dialog_response(XEN_TO_C_GtkDialog_(dialog), XEN_TO_C_gint(response_id));
  return(XEN_FALSE);
}

static XEN gxg_gtk_dialog_run(XEN dialog)
{
  #define H_gtk_dialog_run "gint gtk_dialog_run(GtkDialog* dialog)"
  XEN_ASSERT_TYPE(XEN_GtkDialog__P(dialog), dialog, 1, "gtk_dialog_run", "GtkDialog*");
  return(C_TO_XEN_gint(gtk_dialog_run(XEN_TO_C_GtkDialog_(dialog))));
}
#endif

static XEN gxg_gtk_drag_get_data(XEN widget, XEN context, XEN target, XEN time)
{
  #define H_gtk_drag_get_data "void gtk_drag_get_data(GtkWidget* widget, GdkDragContext* context, GdkAtom target, \
guint32 time)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_drag_get_data", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_GdkDragContext__P(context), context, 2, "gtk_drag_get_data", "GdkDragContext*");
  XEN_ASSERT_TYPE(XEN_GdkAtom_P(target), target, 3, "gtk_drag_get_data", "GdkAtom");
  XEN_ASSERT_TYPE(XEN_guint32_P(time), time, 4, "gtk_drag_get_data", "guint32");
  gtk_drag_get_data(XEN_TO_C_GtkWidget_(widget), XEN_TO_C_GdkDragContext_(context), XEN_TO_C_GdkAtom(target), XEN_TO_C_guint32(time));
  return(XEN_FALSE);
}

static XEN gxg_gtk_drag_finish(XEN context, XEN success, XEN del, XEN time)
{
  #define H_gtk_drag_finish "void gtk_drag_finish(GdkDragContext* context, gboolean success, gboolean del, \
guint32 time)"
  XEN_ASSERT_TYPE(XEN_GdkDragContext__P(context), context, 1, "gtk_drag_finish", "GdkDragContext*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(success), success, 2, "gtk_drag_finish", "gboolean");
  XEN_ASSERT_TYPE(XEN_gboolean_P(del), del, 3, "gtk_drag_finish", "gboolean");
  XEN_ASSERT_TYPE(XEN_guint32_P(time), time, 4, "gtk_drag_finish", "guint32");
  gtk_drag_finish(XEN_TO_C_GdkDragContext_(context), XEN_TO_C_gboolean(success), XEN_TO_C_gboolean(del), XEN_TO_C_guint32(time));
  return(XEN_FALSE);
}

static XEN gxg_gtk_drag_get_source_widget(XEN context)
{
  #define H_gtk_drag_get_source_widget "GtkWidget* gtk_drag_get_source_widget(GdkDragContext* context)"
  XEN_ASSERT_TYPE(XEN_GdkDragContext__P(context), context, 1, "gtk_drag_get_source_widget", "GdkDragContext*");
  return(C_TO_XEN_GtkWidget_(gtk_drag_get_source_widget(XEN_TO_C_GdkDragContext_(context))));
}

static XEN gxg_gtk_drag_highlight(XEN widget)
{
  #define H_gtk_drag_highlight "void gtk_drag_highlight(GtkWidget* widget)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_drag_highlight", "GtkWidget*");
  gtk_drag_highlight(XEN_TO_C_GtkWidget_(widget));
  return(XEN_FALSE);
}

static XEN gxg_gtk_drag_unhighlight(XEN widget)
{
  #define H_gtk_drag_unhighlight "void gtk_drag_unhighlight(GtkWidget* widget)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_drag_unhighlight", "GtkWidget*");
  gtk_drag_unhighlight(XEN_TO_C_GtkWidget_(widget));
  return(XEN_FALSE);
}

static XEN gxg_gtk_drag_dest_set(XEN widget, XEN flags, XEN targets, XEN n_targets, XEN actions)
{
  #define H_gtk_drag_dest_set "void gtk_drag_dest_set(GtkWidget* widget, GtkDestDefaults flags, GtkTargetEntry* targets, \
gint n_targets, GdkDragAction actions)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_drag_dest_set", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_GtkDestDefaults_P(flags), flags, 2, "gtk_drag_dest_set", "GtkDestDefaults");
  XEN_ASSERT_TYPE(XEN_GtkTargetEntry__P(targets), targets, 3, "gtk_drag_dest_set", "GtkTargetEntry*");
  XEN_ASSERT_TYPE(XEN_gint_P(n_targets), n_targets, 4, "gtk_drag_dest_set", "gint");
  XEN_ASSERT_TYPE(XEN_GdkDragAction_P(actions), actions, 5, "gtk_drag_dest_set", "GdkDragAction");
  gtk_drag_dest_set(XEN_TO_C_GtkWidget_(widget), XEN_TO_C_GtkDestDefaults(flags), XEN_TO_C_GtkTargetEntry_(targets), XEN_TO_C_gint(n_targets), 
                    XEN_TO_C_GdkDragAction(actions));
  return(XEN_FALSE);
}

static XEN gxg_gtk_drag_dest_set_proxy(XEN widget, XEN proxy_window, XEN protocol, XEN use_coordinates)
{
  #define H_gtk_drag_dest_set_proxy "void gtk_drag_dest_set_proxy(GtkWidget* widget, GdkWindow* proxy_window, \
GdkDragProtocol protocol, gboolean use_coordinates)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_drag_dest_set_proxy", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(proxy_window), proxy_window, 2, "gtk_drag_dest_set_proxy", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_GdkDragProtocol_P(protocol), protocol, 3, "gtk_drag_dest_set_proxy", "GdkDragProtocol");
  XEN_ASSERT_TYPE(XEN_gboolean_P(use_coordinates), use_coordinates, 4, "gtk_drag_dest_set_proxy", "gboolean");
  gtk_drag_dest_set_proxy(XEN_TO_C_GtkWidget_(widget), XEN_TO_C_GdkWindow_(proxy_window), XEN_TO_C_GdkDragProtocol(protocol), 
                          XEN_TO_C_gboolean(use_coordinates));
  return(XEN_FALSE);
}

static XEN gxg_gtk_drag_dest_unset(XEN widget)
{
  #define H_gtk_drag_dest_unset "void gtk_drag_dest_unset(GtkWidget* widget)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_drag_dest_unset", "GtkWidget*");
  gtk_drag_dest_unset(XEN_TO_C_GtkWidget_(widget));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_drag_dest_find_target(XEN widget, XEN context, XEN target_list)
{
  #define H_gtk_drag_dest_find_target "GdkAtom gtk_drag_dest_find_target(GtkWidget* widget, GdkDragContext* context, \
GtkTargetList* target_list)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_drag_dest_find_target", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_GdkDragContext__P(context), context, 2, "gtk_drag_dest_find_target", "GdkDragContext*");
  XEN_ASSERT_TYPE(XEN_GtkTargetList__P(target_list) || XEN_FALSE_P(target_list), target_list, 3, "gtk_drag_dest_find_target", "GtkTargetList*");
  return(C_TO_XEN_GdkAtom(gtk_drag_dest_find_target(XEN_TO_C_GtkWidget_(widget), XEN_TO_C_GdkDragContext_(context), XEN_TO_C_GtkTargetList_(target_list))));
}

static XEN gxg_gtk_drag_dest_get_target_list(XEN widget)
{
  #define H_gtk_drag_dest_get_target_list "GtkTargetList* gtk_drag_dest_get_target_list(GtkWidget* widget)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_drag_dest_get_target_list", "GtkWidget*");
  return(C_TO_XEN_GtkTargetList_(gtk_drag_dest_get_target_list(XEN_TO_C_GtkWidget_(widget))));
}

static XEN gxg_gtk_drag_dest_set_target_list(XEN widget, XEN target_list)
{
  #define H_gtk_drag_dest_set_target_list "void gtk_drag_dest_set_target_list(GtkWidget* widget, GtkTargetList* target_list)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_drag_dest_set_target_list", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_GtkTargetList__P(target_list) || XEN_FALSE_P(target_list), target_list, 2, "gtk_drag_dest_set_target_list", "GtkTargetList*");
  gtk_drag_dest_set_target_list(XEN_TO_C_GtkWidget_(widget), XEN_TO_C_GtkTargetList_(target_list));
  return(XEN_FALSE);
}
#endif

static XEN gxg_gtk_drag_source_set(XEN widget, XEN start_button_mask, XEN targets, XEN n_targets, XEN actions)
{
  #define H_gtk_drag_source_set "void gtk_drag_source_set(GtkWidget* widget, GdkModifierType start_button_mask, \
GtkTargetEntry* targets, gint n_targets, GdkDragAction actions)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_drag_source_set", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_GdkModifierType_P(start_button_mask), start_button_mask, 2, "gtk_drag_source_set", "GdkModifierType");
  XEN_ASSERT_TYPE(XEN_GtkTargetEntry__P(targets), targets, 3, "gtk_drag_source_set", "GtkTargetEntry*");
  XEN_ASSERT_TYPE(XEN_gint_P(n_targets), n_targets, 4, "gtk_drag_source_set", "gint");
  XEN_ASSERT_TYPE(XEN_GdkDragAction_P(actions), actions, 5, "gtk_drag_source_set", "GdkDragAction");
  gtk_drag_source_set(XEN_TO_C_GtkWidget_(widget), XEN_TO_C_GdkModifierType(start_button_mask), XEN_TO_C_GtkTargetEntry_(targets), 
                      XEN_TO_C_gint(n_targets), XEN_TO_C_GdkDragAction(actions));
  return(XEN_FALSE);
}

static XEN gxg_gtk_drag_source_unset(XEN widget)
{
  #define H_gtk_drag_source_unset "void gtk_drag_source_unset(GtkWidget* widget)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_drag_source_unset", "GtkWidget*");
  gtk_drag_source_unset(XEN_TO_C_GtkWidget_(widget));
  return(XEN_FALSE);
}

static XEN gxg_gtk_drag_source_set_icon(XEN widget, XEN colormap, XEN pixmap, XEN mask)
{
  #define H_gtk_drag_source_set_icon "void gtk_drag_source_set_icon(GtkWidget* widget, GdkColormap* colormap, \
GdkPixmap* pixmap, GdkBitmap* mask)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_drag_source_set_icon", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_GdkColormap__P(colormap), colormap, 2, "gtk_drag_source_set_icon", "GdkColormap*");
  XEN_ASSERT_TYPE(XEN_GdkPixmap__P(pixmap), pixmap, 3, "gtk_drag_source_set_icon", "GdkPixmap*");
  XEN_ASSERT_TYPE(XEN_GdkBitmap__P(mask), mask, 4, "gtk_drag_source_set_icon", "GdkBitmap*");
  gtk_drag_source_set_icon(XEN_TO_C_GtkWidget_(widget), XEN_TO_C_GdkColormap_(colormap), XEN_TO_C_GdkPixmap_(pixmap), XEN_TO_C_GdkBitmap_(mask));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_drag_source_set_icon_pixbuf(XEN widget, XEN pixbuf)
{
  #define H_gtk_drag_source_set_icon_pixbuf "void gtk_drag_source_set_icon_pixbuf(GtkWidget* widget, \
GdkPixbuf* pixbuf)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_drag_source_set_icon_pixbuf", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_GdkPixbuf__P(pixbuf), pixbuf, 2, "gtk_drag_source_set_icon_pixbuf", "GdkPixbuf*");
  gtk_drag_source_set_icon_pixbuf(XEN_TO_C_GtkWidget_(widget), XEN_TO_C_GdkPixbuf_(pixbuf));
  return(XEN_FALSE);
}

static XEN gxg_gtk_drag_source_set_icon_stock(XEN widget, XEN stock_id)
{
  #define H_gtk_drag_source_set_icon_stock "void gtk_drag_source_set_icon_stock(GtkWidget* widget, gchar* stock_id)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_drag_source_set_icon_stock", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_gchar__P(stock_id), stock_id, 2, "gtk_drag_source_set_icon_stock", "gchar*");
  gtk_drag_source_set_icon_stock(XEN_TO_C_GtkWidget_(widget), XEN_TO_C_gchar_(stock_id));
  return(XEN_FALSE);
}
#endif

static XEN gxg_gtk_drag_begin(XEN widget, XEN targets, XEN actions, XEN button, XEN event)
{
  #define H_gtk_drag_begin "GdkDragContext* gtk_drag_begin(GtkWidget* widget, GtkTargetList* targets, \
GdkDragAction actions, gint button, GdkEvent* event)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_drag_begin", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_GtkTargetList__P(targets), targets, 2, "gtk_drag_begin", "GtkTargetList*");
  XEN_ASSERT_TYPE(XEN_GdkDragAction_P(actions), actions, 3, "gtk_drag_begin", "GdkDragAction");
  XEN_ASSERT_TYPE(XEN_gint_P(button), button, 4, "gtk_drag_begin", "gint");
  XEN_ASSERT_TYPE(XEN_GdkEvent__P(event), event, 5, "gtk_drag_begin", "GdkEvent*");
  return(C_TO_XEN_GdkDragContext_(gtk_drag_begin(XEN_TO_C_GtkWidget_(widget), XEN_TO_C_GtkTargetList_(targets), XEN_TO_C_GdkDragAction(actions), 
                                                 XEN_TO_C_gint(button), XEN_TO_C_GdkEvent_(event))));
}

static XEN gxg_gtk_drag_set_icon_widget(XEN context, XEN widget, XEN hot_x, XEN hot_y)
{
  #define H_gtk_drag_set_icon_widget "void gtk_drag_set_icon_widget(GdkDragContext* context, GtkWidget* widget, \
gint hot_x, gint hot_y)"
  XEN_ASSERT_TYPE(XEN_GdkDragContext__P(context), context, 1, "gtk_drag_set_icon_widget", "GdkDragContext*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 2, "gtk_drag_set_icon_widget", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_gint_P(hot_x), hot_x, 3, "gtk_drag_set_icon_widget", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(hot_y), hot_y, 4, "gtk_drag_set_icon_widget", "gint");
  gtk_drag_set_icon_widget(XEN_TO_C_GdkDragContext_(context), XEN_TO_C_GtkWidget_(widget), XEN_TO_C_gint(hot_x), XEN_TO_C_gint(hot_y));
  return(XEN_FALSE);
}

static XEN gxg_gtk_drag_set_icon_pixmap(XEN context, XEN colormap, XEN pixmap, XEN mask, XEN hot_x, XEN hot_y)
{
  #define H_gtk_drag_set_icon_pixmap "void gtk_drag_set_icon_pixmap(GdkDragContext* context, GdkColormap* colormap, \
GdkPixmap* pixmap, GdkBitmap* mask, gint hot_x, gint hot_y)"
  XEN_ASSERT_TYPE(XEN_GdkDragContext__P(context), context, 1, "gtk_drag_set_icon_pixmap", "GdkDragContext*");
  XEN_ASSERT_TYPE(XEN_GdkColormap__P(colormap), colormap, 2, "gtk_drag_set_icon_pixmap", "GdkColormap*");
  XEN_ASSERT_TYPE(XEN_GdkPixmap__P(pixmap), pixmap, 3, "gtk_drag_set_icon_pixmap", "GdkPixmap*");
  XEN_ASSERT_TYPE(XEN_GdkBitmap__P(mask), mask, 4, "gtk_drag_set_icon_pixmap", "GdkBitmap*");
  XEN_ASSERT_TYPE(XEN_gint_P(hot_x), hot_x, 5, "gtk_drag_set_icon_pixmap", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(hot_y), hot_y, 6, "gtk_drag_set_icon_pixmap", "gint");
  gtk_drag_set_icon_pixmap(XEN_TO_C_GdkDragContext_(context), XEN_TO_C_GdkColormap_(colormap), XEN_TO_C_GdkPixmap_(pixmap), 
                           XEN_TO_C_GdkBitmap_(mask), XEN_TO_C_gint(hot_x), XEN_TO_C_gint(hot_y));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_drag_set_icon_pixbuf(XEN context, XEN pixbuf, XEN hot_x, XEN hot_y)
{
  #define H_gtk_drag_set_icon_pixbuf "void gtk_drag_set_icon_pixbuf(GdkDragContext* context, GdkPixbuf* pixbuf, \
gint hot_x, gint hot_y)"
  XEN_ASSERT_TYPE(XEN_GdkDragContext__P(context), context, 1, "gtk_drag_set_icon_pixbuf", "GdkDragContext*");
  XEN_ASSERT_TYPE(XEN_GdkPixbuf__P(pixbuf), pixbuf, 2, "gtk_drag_set_icon_pixbuf", "GdkPixbuf*");
  XEN_ASSERT_TYPE(XEN_gint_P(hot_x), hot_x, 3, "gtk_drag_set_icon_pixbuf", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(hot_y), hot_y, 4, "gtk_drag_set_icon_pixbuf", "gint");
  gtk_drag_set_icon_pixbuf(XEN_TO_C_GdkDragContext_(context), XEN_TO_C_GdkPixbuf_(pixbuf), XEN_TO_C_gint(hot_x), XEN_TO_C_gint(hot_y));
  return(XEN_FALSE);
}

static XEN gxg_gtk_drag_set_icon_stock(XEN context, XEN stock_id, XEN hot_x, XEN hot_y)
{
  #define H_gtk_drag_set_icon_stock "void gtk_drag_set_icon_stock(GdkDragContext* context, gchar* stock_id, \
gint hot_x, gint hot_y)"
  XEN_ASSERT_TYPE(XEN_GdkDragContext__P(context), context, 1, "gtk_drag_set_icon_stock", "GdkDragContext*");
  XEN_ASSERT_TYPE(XEN_gchar__P(stock_id), stock_id, 2, "gtk_drag_set_icon_stock", "gchar*");
  XEN_ASSERT_TYPE(XEN_gint_P(hot_x), hot_x, 3, "gtk_drag_set_icon_stock", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(hot_y), hot_y, 4, "gtk_drag_set_icon_stock", "gint");
  gtk_drag_set_icon_stock(XEN_TO_C_GdkDragContext_(context), XEN_TO_C_gchar_(stock_id), XEN_TO_C_gint(hot_x), XEN_TO_C_gint(hot_y));
  return(XEN_FALSE);
}
#endif

static XEN gxg_gtk_drag_set_icon_default(XEN context)
{
  #define H_gtk_drag_set_icon_default "void gtk_drag_set_icon_default(GdkDragContext* context)"
  XEN_ASSERT_TYPE(XEN_GdkDragContext__P(context), context, 1, "gtk_drag_set_icon_default", "GdkDragContext*");
  gtk_drag_set_icon_default(XEN_TO_C_GdkDragContext_(context));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_drag_check_threshold(XEN widget, XEN start_x, XEN start_y, XEN current_x, XEN current_y)
{
  #define H_gtk_drag_check_threshold "gboolean gtk_drag_check_threshold(GtkWidget* widget, gint start_x, \
gint start_y, gint current_x, gint current_y)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_drag_check_threshold", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_gint_P(start_x), start_x, 2, "gtk_drag_check_threshold", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(start_y), start_y, 3, "gtk_drag_check_threshold", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(current_x), current_x, 4, "gtk_drag_check_threshold", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(current_y), current_y, 5, "gtk_drag_check_threshold", "gint");
  return(C_TO_XEN_gboolean(gtk_drag_check_threshold(XEN_TO_C_GtkWidget_(widget), XEN_TO_C_gint(start_x), XEN_TO_C_gint(start_y), 
                                                    XEN_TO_C_gint(current_x), XEN_TO_C_gint(current_y))));
}
#endif

static XEN gxg_gtk_drawing_area_get_type(void)
{
  #define H_gtk_drawing_area_get_type "GtkType gtk_drawing_area_get_type( void)"
  return(C_TO_XEN_GtkType(gtk_drawing_area_get_type()));
}

static XEN gxg_gtk_drawing_area_new(void)
{
  #define H_gtk_drawing_area_new "GtkWidget* gtk_drawing_area_new( void)"
  return(C_TO_XEN_GtkWidget_(gtk_drawing_area_new()));
}

static XEN gxg_gtk_editable_get_type(void)
{
  #define H_gtk_editable_get_type "GtkType gtk_editable_get_type( void)"
  return(C_TO_XEN_GtkType(gtk_editable_get_type()));
}

static XEN gxg_gtk_editable_select_region(XEN editable, XEN start, XEN end)
{
  #define H_gtk_editable_select_region "void gtk_editable_select_region(GtkEditable* editable, gint start, \
gint end)"
  XEN_ASSERT_TYPE(XEN_GtkEditable__P(editable), editable, 1, "gtk_editable_select_region", "GtkEditable*");
  XEN_ASSERT_TYPE(XEN_gint_P(start), start, 2, "gtk_editable_select_region", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(end), end, 3, "gtk_editable_select_region", "gint");
  gtk_editable_select_region(XEN_TO_C_GtkEditable_(editable), XEN_TO_C_gint(start), XEN_TO_C_gint(end));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_editable_get_selection_bounds(XEN editable, XEN start, XEN end)
{
  #define H_gtk_editable_get_selection_bounds "gboolean gtk_editable_get_selection_bounds(GtkEditable* editable, \
gint* start, gint* end)"
  XEN_ASSERT_TYPE(XEN_GtkEditable__P(editable), editable, 1, "gtk_editable_get_selection_bounds", "GtkEditable*");
  XEN_ASSERT_TYPE(XEN_gint__P(start), start, 2, "gtk_editable_get_selection_bounds", "gint*");
  XEN_ASSERT_TYPE(XEN_gint__P(end), end, 3, "gtk_editable_get_selection_bounds", "gint*");
  return(C_TO_XEN_gboolean(gtk_editable_get_selection_bounds(XEN_TO_C_GtkEditable_(editable), XEN_TO_C_gint_(start), XEN_TO_C_gint_(end))));
}
#endif

static XEN gxg_gtk_editable_insert_text(XEN editable, XEN new_text, XEN new_text_length, XEN position)
{
  #define H_gtk_editable_insert_text "void gtk_editable_insert_text(GtkEditable* editable, gchar* new_text, \
gint new_text_length, gint* position)"
  XEN_ASSERT_TYPE(XEN_GtkEditable__P(editable), editable, 1, "gtk_editable_insert_text", "GtkEditable*");
  XEN_ASSERT_TYPE(XEN_gchar__P(new_text), new_text, 2, "gtk_editable_insert_text", "gchar*");
  XEN_ASSERT_TYPE(XEN_gint_P(new_text_length), new_text_length, 3, "gtk_editable_insert_text", "gint");
  XEN_ASSERT_TYPE(XEN_gint__P(position), position, 4, "gtk_editable_insert_text", "gint*");
  gtk_editable_insert_text(XEN_TO_C_GtkEditable_(editable), XEN_TO_C_gchar_(new_text), XEN_TO_C_gint(new_text_length), XEN_TO_C_gint_(position));
  return(XEN_FALSE);
}

static XEN gxg_gtk_editable_delete_text(XEN editable, XEN start_pos, XEN end_pos)
{
  #define H_gtk_editable_delete_text "void gtk_editable_delete_text(GtkEditable* editable, gint start_pos, \
gint end_pos)"
  XEN_ASSERT_TYPE(XEN_GtkEditable__P(editable), editable, 1, "gtk_editable_delete_text", "GtkEditable*");
  XEN_ASSERT_TYPE(XEN_gint_P(start_pos), start_pos, 2, "gtk_editable_delete_text", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(end_pos), end_pos, 3, "gtk_editable_delete_text", "gint");
  gtk_editable_delete_text(XEN_TO_C_GtkEditable_(editable), XEN_TO_C_gint(start_pos), XEN_TO_C_gint(end_pos));
  return(XEN_FALSE);
}

static XEN gxg_gtk_editable_get_chars(XEN editable, XEN start_pos, XEN end_pos)
{
  #define H_gtk_editable_get_chars "gchar* gtk_editable_get_chars(GtkEditable* editable, gint start_pos, \
gint end_pos)"
  XEN_ASSERT_TYPE(XEN_GtkEditable__P(editable), editable, 1, "gtk_editable_get_chars", "GtkEditable*");
  XEN_ASSERT_TYPE(XEN_gint_P(start_pos), start_pos, 2, "gtk_editable_get_chars", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(end_pos), end_pos, 3, "gtk_editable_get_chars", "gint");
  return(C_TO_XEN_gchar_(gtk_editable_get_chars(XEN_TO_C_GtkEditable_(editable), XEN_TO_C_gint(start_pos), XEN_TO_C_gint(end_pos))));
}

static XEN gxg_gtk_editable_cut_clipboard(XEN editable)
{
  #define H_gtk_editable_cut_clipboard "void gtk_editable_cut_clipboard(GtkEditable* editable)"
  XEN_ASSERT_TYPE(XEN_GtkEditable__P(editable), editable, 1, "gtk_editable_cut_clipboard", "GtkEditable*");
  gtk_editable_cut_clipboard(XEN_TO_C_GtkEditable_(editable));
  return(XEN_FALSE);
}

static XEN gxg_gtk_editable_copy_clipboard(XEN editable)
{
  #define H_gtk_editable_copy_clipboard "void gtk_editable_copy_clipboard(GtkEditable* editable)"
  XEN_ASSERT_TYPE(XEN_GtkEditable__P(editable), editable, 1, "gtk_editable_copy_clipboard", "GtkEditable*");
  gtk_editable_copy_clipboard(XEN_TO_C_GtkEditable_(editable));
  return(XEN_FALSE);
}

static XEN gxg_gtk_editable_paste_clipboard(XEN editable)
{
  #define H_gtk_editable_paste_clipboard "void gtk_editable_paste_clipboard(GtkEditable* editable)"
  XEN_ASSERT_TYPE(XEN_GtkEditable__P(editable), editable, 1, "gtk_editable_paste_clipboard", "GtkEditable*");
  gtk_editable_paste_clipboard(XEN_TO_C_GtkEditable_(editable));
  return(XEN_FALSE);
}

static XEN gxg_gtk_editable_delete_selection(XEN editable)
{
  #define H_gtk_editable_delete_selection "void gtk_editable_delete_selection(GtkEditable* editable)"
  XEN_ASSERT_TYPE(XEN_GtkEditable__P(editable), editable, 1, "gtk_editable_delete_selection", "GtkEditable*");
  gtk_editable_delete_selection(XEN_TO_C_GtkEditable_(editable));
  return(XEN_FALSE);
}

static XEN gxg_gtk_editable_set_position(XEN editable, XEN position)
{
  #define H_gtk_editable_set_position "void gtk_editable_set_position(GtkEditable* editable, gint position)"
  XEN_ASSERT_TYPE(XEN_GtkEditable__P(editable), editable, 1, "gtk_editable_set_position", "GtkEditable*");
  XEN_ASSERT_TYPE(XEN_gint_P(position), position, 2, "gtk_editable_set_position", "gint");
  gtk_editable_set_position(XEN_TO_C_GtkEditable_(editable), XEN_TO_C_gint(position));
  return(XEN_FALSE);
}

static XEN gxg_gtk_editable_get_position(XEN editable)
{
  #define H_gtk_editable_get_position "gint gtk_editable_get_position(GtkEditable* editable)"
  XEN_ASSERT_TYPE(XEN_GtkEditable__P(editable), editable, 1, "gtk_editable_get_position", "GtkEditable*");
  return(C_TO_XEN_gint(gtk_editable_get_position(XEN_TO_C_GtkEditable_(editable))));
}

static XEN gxg_gtk_editable_set_editable(XEN editable, XEN is_editable)
{
  #define H_gtk_editable_set_editable "void gtk_editable_set_editable(GtkEditable* editable, gboolean is_editable)"
  XEN_ASSERT_TYPE(XEN_GtkEditable__P(editable), editable, 1, "gtk_editable_set_editable", "GtkEditable*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(is_editable), is_editable, 2, "gtk_editable_set_editable", "gboolean");
  gtk_editable_set_editable(XEN_TO_C_GtkEditable_(editable), XEN_TO_C_gboolean(is_editable));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_editable_get_editable(XEN editable)
{
  #define H_gtk_editable_get_editable "gboolean gtk_editable_get_editable(GtkEditable* editable)"
  XEN_ASSERT_TYPE(XEN_GtkEditable__P(editable), editable, 1, "gtk_editable_get_editable", "GtkEditable*");
  return(C_TO_XEN_gboolean(gtk_editable_get_editable(XEN_TO_C_GtkEditable_(editable))));
}
#endif

static XEN gxg_gtk_entry_get_type(void)
{
  #define H_gtk_entry_get_type "GtkType gtk_entry_get_type( void)"
  return(C_TO_XEN_GtkType(gtk_entry_get_type()));
}

static XEN gxg_gtk_entry_new(void)
{
  #define H_gtk_entry_new "GtkWidget* gtk_entry_new( void)"
  return(C_TO_XEN_GtkWidget_(gtk_entry_new()));
}

static XEN gxg_gtk_entry_set_visibility(XEN entry, XEN visible)
{
  #define H_gtk_entry_set_visibility "void gtk_entry_set_visibility(GtkEntry* entry, gboolean visible)"
  XEN_ASSERT_TYPE(XEN_GtkEntry__P(entry), entry, 1, "gtk_entry_set_visibility", "GtkEntry*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(visible), visible, 2, "gtk_entry_set_visibility", "gboolean");
  gtk_entry_set_visibility(XEN_TO_C_GtkEntry_(entry), XEN_TO_C_gboolean(visible));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_entry_get_visibility(XEN entry)
{
  #define H_gtk_entry_get_visibility "gboolean gtk_entry_get_visibility(GtkEntry* entry)"
  XEN_ASSERT_TYPE(XEN_GtkEntry__P(entry), entry, 1, "gtk_entry_get_visibility", "GtkEntry*");
  return(C_TO_XEN_gboolean(gtk_entry_get_visibility(XEN_TO_C_GtkEntry_(entry))));
}

static XEN gxg_gtk_entry_set_invisible_char(XEN entry, XEN ch)
{
  #define H_gtk_entry_set_invisible_char "void gtk_entry_set_invisible_char(GtkEntry* entry, gunichar ch)"
  XEN_ASSERT_TYPE(XEN_GtkEntry__P(entry), entry, 1, "gtk_entry_set_invisible_char", "GtkEntry*");
  XEN_ASSERT_TYPE(XEN_gunichar_P(ch), ch, 2, "gtk_entry_set_invisible_char", "gunichar");
  gtk_entry_set_invisible_char(XEN_TO_C_GtkEntry_(entry), XEN_TO_C_gunichar(ch));
  return(XEN_FALSE);
}

static XEN gxg_gtk_entry_get_invisible_char(XEN entry)
{
  #define H_gtk_entry_get_invisible_char "gunichar gtk_entry_get_invisible_char(GtkEntry* entry)"
  XEN_ASSERT_TYPE(XEN_GtkEntry__P(entry), entry, 1, "gtk_entry_get_invisible_char", "GtkEntry*");
  return(C_TO_XEN_gunichar(gtk_entry_get_invisible_char(XEN_TO_C_GtkEntry_(entry))));
}

static XEN gxg_gtk_entry_set_has_frame(XEN entry, XEN setting)
{
  #define H_gtk_entry_set_has_frame "void gtk_entry_set_has_frame(GtkEntry* entry, gboolean setting)"
  XEN_ASSERT_TYPE(XEN_GtkEntry__P(entry), entry, 1, "gtk_entry_set_has_frame", "GtkEntry*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(setting), setting, 2, "gtk_entry_set_has_frame", "gboolean");
  gtk_entry_set_has_frame(XEN_TO_C_GtkEntry_(entry), XEN_TO_C_gboolean(setting));
  return(XEN_FALSE);
}

static XEN gxg_gtk_entry_get_has_frame(XEN entry)
{
  #define H_gtk_entry_get_has_frame "gboolean gtk_entry_get_has_frame(GtkEntry* entry)"
  XEN_ASSERT_TYPE(XEN_GtkEntry__P(entry), entry, 1, "gtk_entry_get_has_frame", "GtkEntry*");
  return(C_TO_XEN_gboolean(gtk_entry_get_has_frame(XEN_TO_C_GtkEntry_(entry))));
}
#endif

static XEN gxg_gtk_entry_set_max_length(XEN entry, XEN max)
{
  #define H_gtk_entry_set_max_length "void gtk_entry_set_max_length(GtkEntry* entry, gint max)"
  XEN_ASSERT_TYPE(XEN_GtkEntry__P(entry), entry, 1, "gtk_entry_set_max_length", "GtkEntry*");
  XEN_ASSERT_TYPE(XEN_gint_P(max), max, 2, "gtk_entry_set_max_length", "gint");
  gtk_entry_set_max_length(XEN_TO_C_GtkEntry_(entry), XEN_TO_C_gint(max));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_entry_get_max_length(XEN entry)
{
  #define H_gtk_entry_get_max_length "gint gtk_entry_get_max_length(GtkEntry* entry)"
  XEN_ASSERT_TYPE(XEN_GtkEntry__P(entry), entry, 1, "gtk_entry_get_max_length", "GtkEntry*");
  return(C_TO_XEN_gint(gtk_entry_get_max_length(XEN_TO_C_GtkEntry_(entry))));
}

static XEN gxg_gtk_entry_set_activates_default(XEN entry, XEN setting)
{
  #define H_gtk_entry_set_activates_default "void gtk_entry_set_activates_default(GtkEntry* entry, gboolean setting)"
  XEN_ASSERT_TYPE(XEN_GtkEntry__P(entry), entry, 1, "gtk_entry_set_activates_default", "GtkEntry*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(setting), setting, 2, "gtk_entry_set_activates_default", "gboolean");
  gtk_entry_set_activates_default(XEN_TO_C_GtkEntry_(entry), XEN_TO_C_gboolean(setting));
  return(XEN_FALSE);
}

static XEN gxg_gtk_entry_get_activates_default(XEN entry)
{
  #define H_gtk_entry_get_activates_default "gboolean gtk_entry_get_activates_default(GtkEntry* entry)"
  XEN_ASSERT_TYPE(XEN_GtkEntry__P(entry), entry, 1, "gtk_entry_get_activates_default", "GtkEntry*");
  return(C_TO_XEN_gboolean(gtk_entry_get_activates_default(XEN_TO_C_GtkEntry_(entry))));
}

static XEN gxg_gtk_entry_set_width_chars(XEN entry, XEN n_chars)
{
  #define H_gtk_entry_set_width_chars "void gtk_entry_set_width_chars(GtkEntry* entry, gint n_chars)"
  XEN_ASSERT_TYPE(XEN_GtkEntry__P(entry), entry, 1, "gtk_entry_set_width_chars", "GtkEntry*");
  XEN_ASSERT_TYPE(XEN_gint_P(n_chars), n_chars, 2, "gtk_entry_set_width_chars", "gint");
  gtk_entry_set_width_chars(XEN_TO_C_GtkEntry_(entry), XEN_TO_C_gint(n_chars));
  return(XEN_FALSE);
}

static XEN gxg_gtk_entry_get_width_chars(XEN entry)
{
  #define H_gtk_entry_get_width_chars "gint gtk_entry_get_width_chars(GtkEntry* entry)"
  XEN_ASSERT_TYPE(XEN_GtkEntry__P(entry), entry, 1, "gtk_entry_get_width_chars", "GtkEntry*");
  return(C_TO_XEN_gint(gtk_entry_get_width_chars(XEN_TO_C_GtkEntry_(entry))));
}
#endif

static XEN gxg_gtk_entry_set_text(XEN entry, XEN text)
{
  #define H_gtk_entry_set_text "void gtk_entry_set_text(GtkEntry* entry, gchar* text)"
  XEN_ASSERT_TYPE(XEN_GtkEntry__P(entry), entry, 1, "gtk_entry_set_text", "GtkEntry*");
  XEN_ASSERT_TYPE(XEN_gchar__P(text), text, 2, "gtk_entry_set_text", "gchar*");
  gtk_entry_set_text(XEN_TO_C_GtkEntry_(entry), XEN_TO_C_gchar_(text));
  return(XEN_FALSE);
}

static XEN gxg_gtk_entry_get_text(XEN entry)
{
  #define H_gtk_entry_get_text "gchar* gtk_entry_get_text(GtkEntry* entry)"
  XEN_ASSERT_TYPE(XEN_GtkEntry__P(entry), entry, 1, "gtk_entry_get_text", "GtkEntry*");
  return(C_TO_XEN_gchar_(gtk_entry_get_text(XEN_TO_C_GtkEntry_(entry))));
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_entry_get_layout(XEN entry)
{
  #define H_gtk_entry_get_layout "PangoLayout* gtk_entry_get_layout(GtkEntry* entry)"
  XEN_ASSERT_TYPE(XEN_GtkEntry__P(entry), entry, 1, "gtk_entry_get_layout", "GtkEntry*");
  return(C_TO_XEN_PangoLayout_(gtk_entry_get_layout(XEN_TO_C_GtkEntry_(entry))));
}

static XEN gxg_gtk_entry_get_layout_offsets(XEN entry, XEN x, XEN y)
{
  #define H_gtk_entry_get_layout_offsets "void gtk_entry_get_layout_offsets(GtkEntry* entry, gint* [x], \
gint* [y])"
  gint ref_x;
  gint ref_y;
  XEN_ASSERT_TYPE(XEN_GtkEntry__P(entry), entry, 1, "gtk_entry_get_layout_offsets", "GtkEntry*");
  gtk_entry_get_layout_offsets(XEN_TO_C_GtkEntry_(entry), &ref_x, &ref_y);
  return(XEN_LIST_2(C_TO_XEN_gint(ref_x), C_TO_XEN_gint(ref_y)));
}
#endif

static XEN gxg_gtk_event_box_get_type(void)
{
  #define H_gtk_event_box_get_type "GtkType gtk_event_box_get_type( void)"
  return(C_TO_XEN_GtkType(gtk_event_box_get_type()));
}

static XEN gxg_gtk_event_box_new(void)
{
  #define H_gtk_event_box_new "GtkWidget* gtk_event_box_new( void)"
  return(C_TO_XEN_GtkWidget_(gtk_event_box_new()));
}

static XEN gxg_gtk_file_selection_get_type(void)
{
  #define H_gtk_file_selection_get_type "GtkType gtk_file_selection_get_type( void)"
  return(C_TO_XEN_GtkType(gtk_file_selection_get_type()));
}

static XEN gxg_gtk_file_selection_new(XEN title)
{
  #define H_gtk_file_selection_new "GtkWidget* gtk_file_selection_new(gchar* title)"
  XEN_ASSERT_TYPE(XEN_gchar__P(title), title, 1, "gtk_file_selection_new", "gchar*");
  return(C_TO_XEN_GtkWidget_(gtk_file_selection_new(XEN_TO_C_gchar_(title))));
}

static XEN gxg_gtk_file_selection_set_filename(XEN filesel, XEN filename)
{
  #define H_gtk_file_selection_set_filename "void gtk_file_selection_set_filename(GtkFileSelection* filesel, \
gchar* filename)"
  XEN_ASSERT_TYPE(XEN_GtkFileSelection__P(filesel), filesel, 1, "gtk_file_selection_set_filename", "GtkFileSelection*");
  XEN_ASSERT_TYPE(XEN_gchar__P(filename), filename, 2, "gtk_file_selection_set_filename", "gchar*");
  gtk_file_selection_set_filename(XEN_TO_C_GtkFileSelection_(filesel), XEN_TO_C_gchar_(filename));
  return(XEN_FALSE);
}

static XEN gxg_gtk_file_selection_get_filename(XEN filesel)
{
  #define H_gtk_file_selection_get_filename "gchar* gtk_file_selection_get_filename(GtkFileSelection* filesel)"
  XEN_ASSERT_TYPE(XEN_GtkFileSelection__P(filesel), filesel, 1, "gtk_file_selection_get_filename", "GtkFileSelection*");
  return(C_TO_XEN_gchar_(gtk_file_selection_get_filename(XEN_TO_C_GtkFileSelection_(filesel))));
}

static XEN gxg_gtk_file_selection_complete(XEN filesel, XEN pattern)
{
  #define H_gtk_file_selection_complete "void gtk_file_selection_complete(GtkFileSelection* filesel, \
gchar* pattern)"
  XEN_ASSERT_TYPE(XEN_GtkFileSelection__P(filesel), filesel, 1, "gtk_file_selection_complete", "GtkFileSelection*");
  XEN_ASSERT_TYPE(XEN_gchar__P(pattern), pattern, 2, "gtk_file_selection_complete", "gchar*");
  gtk_file_selection_complete(XEN_TO_C_GtkFileSelection_(filesel), XEN_TO_C_gchar_(pattern));
  return(XEN_FALSE);
}

static XEN gxg_gtk_file_selection_show_fileop_buttons(XEN filesel)
{
  #define H_gtk_file_selection_show_fileop_buttons "void gtk_file_selection_show_fileop_buttons(GtkFileSelection* filesel)"
  XEN_ASSERT_TYPE(XEN_GtkFileSelection__P(filesel), filesel, 1, "gtk_file_selection_show_fileop_buttons", "GtkFileSelection*");
  gtk_file_selection_show_fileop_buttons(XEN_TO_C_GtkFileSelection_(filesel));
  return(XEN_FALSE);
}

static XEN gxg_gtk_file_selection_hide_fileop_buttons(XEN filesel)
{
  #define H_gtk_file_selection_hide_fileop_buttons "void gtk_file_selection_hide_fileop_buttons(GtkFileSelection* filesel)"
  XEN_ASSERT_TYPE(XEN_GtkFileSelection__P(filesel), filesel, 1, "gtk_file_selection_hide_fileop_buttons", "GtkFileSelection*");
  gtk_file_selection_hide_fileop_buttons(XEN_TO_C_GtkFileSelection_(filesel));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
#if HAVE_GTK_FILE_SELECTION_GET_SELECTIONS
static XEN gxg_gtk_file_selection_get_selections(XEN filesel)
{
  #define H_gtk_file_selection_get_selections "gchar** gtk_file_selection_get_selections(GtkFileSelection* filesel)"
  XEN_ASSERT_TYPE(XEN_GtkFileSelection__P(filesel), filesel, 1, "gtk_file_selection_get_selections", "GtkFileSelection*");
  return(C_TO_XEN_gchar__(gtk_file_selection_get_selections(XEN_TO_C_GtkFileSelection_(filesel))));
}
#endif

#if HAVE_GTK_FILE_SELECTION_GET_SELECTIONS
static XEN gxg_gtk_file_selection_set_select_multiple(XEN filesel, XEN select_multiple)
{
  #define H_gtk_file_selection_set_select_multiple "void gtk_file_selection_set_select_multiple(GtkFileSelection* filesel, \
gboolean select_multiple)"
  XEN_ASSERT_TYPE(XEN_GtkFileSelection__P(filesel), filesel, 1, "gtk_file_selection_set_select_multiple", "GtkFileSelection*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(select_multiple), select_multiple, 2, "gtk_file_selection_set_select_multiple", "gboolean");
  gtk_file_selection_set_select_multiple(XEN_TO_C_GtkFileSelection_(filesel), XEN_TO_C_gboolean(select_multiple));
  return(XEN_FALSE);
}
#endif

#if HAVE_GTK_FILE_SELECTION_GET_SELECTIONS
static XEN gxg_gtk_file_selection_get_select_multiple(XEN filesel)
{
  #define H_gtk_file_selection_get_select_multiple "gboolean gtk_file_selection_get_select_multiple(GtkFileSelection* filesel)"
  XEN_ASSERT_TYPE(XEN_GtkFileSelection__P(filesel), filesel, 1, "gtk_file_selection_get_select_multiple", "GtkFileSelection*");
  return(C_TO_XEN_gboolean(gtk_file_selection_get_select_multiple(XEN_TO_C_GtkFileSelection_(filesel))));
}
#endif

#if HAVE_GTK_ICON_SET_GET_TYPE
static XEN gxg_gtk_icon_set_get_type(void)
{
  #define H_gtk_icon_set_get_type "GType gtk_icon_set_get_type( void)"
  return(C_TO_XEN_GType(gtk_icon_set_get_type()));
}
#endif

#if HAVE_GTK_ICON_SET_GET_TYPE
static XEN gxg_gtk_icon_source_get_type(void)
{
  #define H_gtk_icon_source_get_type "GType gtk_icon_source_get_type( void)"
  return(C_TO_XEN_GType(gtk_icon_source_get_type()));
}
#endif

#if HAVE_GTK_ICON_SET_GET_TYPE
static XEN gxg_gtk_selection_data_get_type(void)
{
  #define H_gtk_selection_data_get_type "GType gtk_selection_data_get_type( void)"
  return(C_TO_XEN_GType(gtk_selection_data_get_type()));
}
#endif

#if HAVE_GTK_ICON_SET_GET_TYPE
static XEN gxg_gtk_border_get_type(void)
{
  #define H_gtk_border_get_type "GType gtk_border_get_type( void)"
  return(C_TO_XEN_GType(gtk_border_get_type()));
}
#endif

#if HAVE_GTK_ICON_SET_GET_TYPE
static XEN gxg_gtk_tree_path_get_type(void)
{
  #define H_gtk_tree_path_get_type "GType gtk_tree_path_get_type( void)"
  return(C_TO_XEN_GType(gtk_tree_path_get_type()));
}
#endif

#if HAVE_GTK_ICON_SET_GET_TYPE
static XEN gxg_gtk_tree_iter_get_type(void)
{
  #define H_gtk_tree_iter_get_type "GType gtk_tree_iter_get_type( void)"
  return(C_TO_XEN_GType(gtk_tree_iter_get_type()));
}
#endif

#if HAVE_GTK_ICON_SET_GET_TYPE
static XEN gxg_gtk_identifier_get_type(void)
{
  #define H_gtk_identifier_get_type "GType gtk_identifier_get_type( void)"
  return(C_TO_XEN_GType(gtk_identifier_get_type()));
}
#endif

#if HAVE_GTK_ICON_SET_GET_TYPE
static XEN gxg_gtk_requisition_get_type(void)
{
  #define H_gtk_requisition_get_type "GType gtk_requisition_get_type( void)"
  return(C_TO_XEN_GType(gtk_requisition_get_type()));
}
#endif
#endif

static XEN gxg_gtk_fixed_get_type(void)
{
  #define H_gtk_fixed_get_type "GtkType gtk_fixed_get_type( void)"
  return(C_TO_XEN_GtkType(gtk_fixed_get_type()));
}

static XEN gxg_gtk_fixed_new(void)
{
  #define H_gtk_fixed_new "GtkWidget* gtk_fixed_new( void)"
  return(C_TO_XEN_GtkWidget_(gtk_fixed_new()));
}

static XEN gxg_gtk_fixed_put(XEN fixed, XEN widget, XEN x, XEN y)
{
  #define H_gtk_fixed_put "void gtk_fixed_put(GtkFixed* fixed, GtkWidget* widget, gint x, gint y)"
  XEN_ASSERT_TYPE(XEN_GtkFixed__P(fixed), fixed, 1, "gtk_fixed_put", "GtkFixed*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 2, "gtk_fixed_put", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_gint_P(x), x, 3, "gtk_fixed_put", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(y), y, 4, "gtk_fixed_put", "gint");
  gtk_fixed_put(XEN_TO_C_GtkFixed_(fixed), XEN_TO_C_GtkWidget_(widget), XEN_TO_C_gint(x), XEN_TO_C_gint(y));
  return(XEN_FALSE);
}

static XEN gxg_gtk_fixed_move(XEN fixed, XEN widget, XEN x, XEN y)
{
  #define H_gtk_fixed_move "void gtk_fixed_move(GtkFixed* fixed, GtkWidget* widget, gint x, gint y)"
  XEN_ASSERT_TYPE(XEN_GtkFixed__P(fixed), fixed, 1, "gtk_fixed_move", "GtkFixed*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 2, "gtk_fixed_move", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_gint_P(x), x, 3, "gtk_fixed_move", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(y), y, 4, "gtk_fixed_move", "gint");
  gtk_fixed_move(XEN_TO_C_GtkFixed_(fixed), XEN_TO_C_GtkWidget_(widget), XEN_TO_C_gint(x), XEN_TO_C_gint(y));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_fixed_set_has_window(XEN fixed, XEN has_window)
{
  #define H_gtk_fixed_set_has_window "void gtk_fixed_set_has_window(GtkFixed* fixed, gboolean has_window)"
  XEN_ASSERT_TYPE(XEN_GtkFixed__P(fixed), fixed, 1, "gtk_fixed_set_has_window", "GtkFixed*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(has_window), has_window, 2, "gtk_fixed_set_has_window", "gboolean");
  gtk_fixed_set_has_window(XEN_TO_C_GtkFixed_(fixed), XEN_TO_C_gboolean(has_window));
  return(XEN_FALSE);
}

static XEN gxg_gtk_fixed_get_has_window(XEN fixed)
{
  #define H_gtk_fixed_get_has_window "gboolean gtk_fixed_get_has_window(GtkFixed* fixed)"
  XEN_ASSERT_TYPE(XEN_GtkFixed__P(fixed), fixed, 1, "gtk_fixed_get_has_window", "GtkFixed*");
  return(C_TO_XEN_gboolean(gtk_fixed_get_has_window(XEN_TO_C_GtkFixed_(fixed))));
}
#endif

static XEN gxg_gtk_font_selection_get_type(void)
{
  #define H_gtk_font_selection_get_type "GtkType gtk_font_selection_get_type( void)"
  return(C_TO_XEN_GtkType(gtk_font_selection_get_type()));
}

static XEN gxg_gtk_font_selection_new(void)
{
  #define H_gtk_font_selection_new "GtkWidget* gtk_font_selection_new( void)"
  return(C_TO_XEN_GtkWidget_(gtk_font_selection_new()));
}

static XEN gxg_gtk_font_selection_get_font_name(XEN fontsel)
{
  #define H_gtk_font_selection_get_font_name "gchar* gtk_font_selection_get_font_name(GtkFontSelection* fontsel)"
  XEN_ASSERT_TYPE(XEN_GtkFontSelection__P(fontsel), fontsel, 1, "gtk_font_selection_get_font_name", "GtkFontSelection*");
  return(C_TO_XEN_gchar_(gtk_font_selection_get_font_name(XEN_TO_C_GtkFontSelection_(fontsel))));
}

static XEN gxg_gtk_font_selection_set_font_name(XEN fontsel, XEN fontname)
{
  #define H_gtk_font_selection_set_font_name "gboolean gtk_font_selection_set_font_name(GtkFontSelection* fontsel, \
gchar* fontname)"
  XEN_ASSERT_TYPE(XEN_GtkFontSelection__P(fontsel), fontsel, 1, "gtk_font_selection_set_font_name", "GtkFontSelection*");
  XEN_ASSERT_TYPE(XEN_gchar__P(fontname), fontname, 2, "gtk_font_selection_set_font_name", "gchar*");
  return(C_TO_XEN_gboolean(gtk_font_selection_set_font_name(XEN_TO_C_GtkFontSelection_(fontsel), XEN_TO_C_gchar_(fontname))));
}

static XEN gxg_gtk_font_selection_get_preview_text(XEN fontsel)
{
  #define H_gtk_font_selection_get_preview_text "gchar* gtk_font_selection_get_preview_text(GtkFontSelection* fontsel)"
  XEN_ASSERT_TYPE(XEN_GtkFontSelection__P(fontsel), fontsel, 1, "gtk_font_selection_get_preview_text", "GtkFontSelection*");
  return(C_TO_XEN_gchar_(gtk_font_selection_get_preview_text(XEN_TO_C_GtkFontSelection_(fontsel))));
}

static XEN gxg_gtk_font_selection_set_preview_text(XEN fontsel, XEN text)
{
  #define H_gtk_font_selection_set_preview_text "void gtk_font_selection_set_preview_text(GtkFontSelection* fontsel, \
gchar* text)"
  XEN_ASSERT_TYPE(XEN_GtkFontSelection__P(fontsel), fontsel, 1, "gtk_font_selection_set_preview_text", "GtkFontSelection*");
  XEN_ASSERT_TYPE(XEN_gchar__P(text), text, 2, "gtk_font_selection_set_preview_text", "gchar*");
  gtk_font_selection_set_preview_text(XEN_TO_C_GtkFontSelection_(fontsel), XEN_TO_C_gchar_(text));
  return(XEN_FALSE);
}

static XEN gxg_gtk_font_selection_dialog_get_type(void)
{
  #define H_gtk_font_selection_dialog_get_type "GtkType gtk_font_selection_dialog_get_type( void)"
  return(C_TO_XEN_GtkType(gtk_font_selection_dialog_get_type()));
}

static XEN gxg_gtk_font_selection_dialog_new(XEN title)
{
  #define H_gtk_font_selection_dialog_new "GtkWidget* gtk_font_selection_dialog_new(gchar* title)"
  XEN_ASSERT_TYPE(XEN_gchar__P(title), title, 1, "gtk_font_selection_dialog_new", "gchar*");
  return(C_TO_XEN_GtkWidget_(gtk_font_selection_dialog_new(XEN_TO_C_gchar_(title))));
}

static XEN gxg_gtk_font_selection_dialog_get_font_name(XEN fsd)
{
  #define H_gtk_font_selection_dialog_get_font_name "gchar* gtk_font_selection_dialog_get_font_name(GtkFontSelectionDialog* fsd)"
  XEN_ASSERT_TYPE(XEN_GtkFontSelectionDialog__P(fsd), fsd, 1, "gtk_font_selection_dialog_get_font_name", "GtkFontSelectionDialog*");
  return(C_TO_XEN_gchar_(gtk_font_selection_dialog_get_font_name(XEN_TO_C_GtkFontSelectionDialog_(fsd))));
}

static XEN gxg_gtk_font_selection_dialog_set_font_name(XEN fsd, XEN fontname)
{
  #define H_gtk_font_selection_dialog_set_font_name "gboolean gtk_font_selection_dialog_set_font_name(GtkFontSelectionDialog* fsd, \
gchar* fontname)"
  XEN_ASSERT_TYPE(XEN_GtkFontSelectionDialog__P(fsd), fsd, 1, "gtk_font_selection_dialog_set_font_name", "GtkFontSelectionDialog*");
  XEN_ASSERT_TYPE(XEN_gchar__P(fontname), fontname, 2, "gtk_font_selection_dialog_set_font_name", "gchar*");
  return(C_TO_XEN_gboolean(gtk_font_selection_dialog_set_font_name(XEN_TO_C_GtkFontSelectionDialog_(fsd), XEN_TO_C_gchar_(fontname))));
}

static XEN gxg_gtk_font_selection_dialog_get_preview_text(XEN fsd)
{
  #define H_gtk_font_selection_dialog_get_preview_text "gchar* gtk_font_selection_dialog_get_preview_text(GtkFontSelectionDialog* fsd)"
  XEN_ASSERT_TYPE(XEN_GtkFontSelectionDialog__P(fsd), fsd, 1, "gtk_font_selection_dialog_get_preview_text", "GtkFontSelectionDialog*");
  return(C_TO_XEN_gchar_(gtk_font_selection_dialog_get_preview_text(XEN_TO_C_GtkFontSelectionDialog_(fsd))));
}

static XEN gxg_gtk_font_selection_dialog_set_preview_text(XEN fsd, XEN text)
{
  #define H_gtk_font_selection_dialog_set_preview_text "void gtk_font_selection_dialog_set_preview_text(GtkFontSelectionDialog* fsd, \
gchar* text)"
  XEN_ASSERT_TYPE(XEN_GtkFontSelectionDialog__P(fsd), fsd, 1, "gtk_font_selection_dialog_set_preview_text", "GtkFontSelectionDialog*");
  XEN_ASSERT_TYPE(XEN_gchar__P(text), text, 2, "gtk_font_selection_dialog_set_preview_text", "gchar*");
  gtk_font_selection_dialog_set_preview_text(XEN_TO_C_GtkFontSelectionDialog_(fsd), XEN_TO_C_gchar_(text));
  return(XEN_FALSE);
}

static XEN gxg_gtk_frame_get_type(void)
{
  #define H_gtk_frame_get_type "GtkType gtk_frame_get_type( void)"
  return(C_TO_XEN_GtkType(gtk_frame_get_type()));
}

static XEN gxg_gtk_frame_new(XEN label)
{
  #define H_gtk_frame_new "GtkWidget* gtk_frame_new(gchar* label)"
  XEN_ASSERT_TYPE(XEN_gchar__P(label), label, 1, "gtk_frame_new", "gchar*");
  return(C_TO_XEN_GtkWidget_(gtk_frame_new(XEN_TO_C_gchar_(label))));
}

static XEN gxg_gtk_frame_set_label(XEN frame, XEN label)
{
  #define H_gtk_frame_set_label "void gtk_frame_set_label(GtkFrame* frame, gchar* label)"
  XEN_ASSERT_TYPE(XEN_GtkFrame__P(frame), frame, 1, "gtk_frame_set_label", "GtkFrame*");
  XEN_ASSERT_TYPE(XEN_gchar__P(label), label, 2, "gtk_frame_set_label", "gchar*");
  gtk_frame_set_label(XEN_TO_C_GtkFrame_(frame), XEN_TO_C_gchar_(label));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_frame_get_label(XEN frame)
{
  #define H_gtk_frame_get_label "gchar* gtk_frame_get_label(GtkFrame* frame)"
  XEN_ASSERT_TYPE(XEN_GtkFrame__P(frame), frame, 1, "gtk_frame_get_label", "GtkFrame*");
  return(C_TO_XEN_gchar_(gtk_frame_get_label(XEN_TO_C_GtkFrame_(frame))));
}

static XEN gxg_gtk_frame_set_label_widget(XEN frame, XEN label_widget)
{
  #define H_gtk_frame_set_label_widget "void gtk_frame_set_label_widget(GtkFrame* frame, GtkWidget* label_widget)"
  XEN_ASSERT_TYPE(XEN_GtkFrame__P(frame), frame, 1, "gtk_frame_set_label_widget", "GtkFrame*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(label_widget), label_widget, 2, "gtk_frame_set_label_widget", "GtkWidget*");
  gtk_frame_set_label_widget(XEN_TO_C_GtkFrame_(frame), XEN_TO_C_GtkWidget_(label_widget));
  return(XEN_FALSE);
}

static XEN gxg_gtk_frame_get_label_widget(XEN frame)
{
  #define H_gtk_frame_get_label_widget "GtkWidget* gtk_frame_get_label_widget(GtkFrame* frame)"
  XEN_ASSERT_TYPE(XEN_GtkFrame__P(frame), frame, 1, "gtk_frame_get_label_widget", "GtkFrame*");
  return(C_TO_XEN_GtkWidget_(gtk_frame_get_label_widget(XEN_TO_C_GtkFrame_(frame))));
}
#endif

static XEN gxg_gtk_frame_set_label_align(XEN frame, XEN xalign, XEN yalign)
{
  #define H_gtk_frame_set_label_align "void gtk_frame_set_label_align(GtkFrame* frame, gfloat xalign, \
gfloat yalign)"
  XEN_ASSERT_TYPE(XEN_GtkFrame__P(frame), frame, 1, "gtk_frame_set_label_align", "GtkFrame*");
  XEN_ASSERT_TYPE(XEN_gfloat_P(xalign), xalign, 2, "gtk_frame_set_label_align", "gfloat");
  XEN_ASSERT_TYPE(XEN_gfloat_P(yalign), yalign, 3, "gtk_frame_set_label_align", "gfloat");
  gtk_frame_set_label_align(XEN_TO_C_GtkFrame_(frame), XEN_TO_C_gfloat(xalign), XEN_TO_C_gfloat(yalign));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_frame_get_label_align(XEN frame, XEN xalign, XEN yalign)
{
  #define H_gtk_frame_get_label_align "void gtk_frame_get_label_align(GtkFrame* frame, gfloat* [xalign], \
gfloat* [yalign])"
  gfloat ref_xalign;
  gfloat ref_yalign;
  XEN_ASSERT_TYPE(XEN_GtkFrame__P(frame), frame, 1, "gtk_frame_get_label_align", "GtkFrame*");
  gtk_frame_get_label_align(XEN_TO_C_GtkFrame_(frame), &ref_xalign, &ref_yalign);
  return(XEN_LIST_2(C_TO_XEN_gfloat(ref_xalign), C_TO_XEN_gfloat(ref_yalign)));
}
#endif

static XEN gxg_gtk_frame_set_shadow_type(XEN frame, XEN type)
{
  #define H_gtk_frame_set_shadow_type "void gtk_frame_set_shadow_type(GtkFrame* frame, GtkShadowType type)"
  XEN_ASSERT_TYPE(XEN_GtkFrame__P(frame), frame, 1, "gtk_frame_set_shadow_type", "GtkFrame*");
  XEN_ASSERT_TYPE(XEN_GtkShadowType_P(type), type, 2, "gtk_frame_set_shadow_type", "GtkShadowType");
  gtk_frame_set_shadow_type(XEN_TO_C_GtkFrame_(frame), XEN_TO_C_GtkShadowType(type));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_frame_get_shadow_type(XEN frame)
{
  #define H_gtk_frame_get_shadow_type "GtkShadowType gtk_frame_get_shadow_type(GtkFrame* frame)"
  XEN_ASSERT_TYPE(XEN_GtkFrame__P(frame), frame, 1, "gtk_frame_get_shadow_type", "GtkFrame*");
  return(C_TO_XEN_GtkShadowType(gtk_frame_get_shadow_type(XEN_TO_C_GtkFrame_(frame))));
}
#endif

static XEN gxg_gtk_gamma_curve_get_type(void)
{
  #define H_gtk_gamma_curve_get_type "GtkType gtk_gamma_curve_get_type( void)"
  return(C_TO_XEN_GtkType(gtk_gamma_curve_get_type()));
}

static XEN gxg_gtk_gamma_curve_new(void)
{
  #define H_gtk_gamma_curve_new "GtkWidget* gtk_gamma_curve_new( void)"
  return(C_TO_XEN_GtkWidget_(gtk_gamma_curve_new()));
}

static XEN gxg_gtk_gc_get(XEN depth, XEN colormap, XEN values, XEN values_mask)
{
  #define H_gtk_gc_get "GdkGC* gtk_gc_get(gint depth, GdkColormap* colormap, GdkGCValues* values, GdkGCValuesMask values_mask)"
  XEN_ASSERT_TYPE(XEN_gint_P(depth), depth, 1, "gtk_gc_get", "gint");
  XEN_ASSERT_TYPE(XEN_GdkColormap__P(colormap), colormap, 2, "gtk_gc_get", "GdkColormap*");
  XEN_ASSERT_TYPE(XEN_GdkGCValues__P(values), values, 3, "gtk_gc_get", "GdkGCValues*");
  XEN_ASSERT_TYPE(XEN_GdkGCValuesMask_P(values_mask), values_mask, 4, "gtk_gc_get", "GdkGCValuesMask");
  return(C_TO_XEN_GdkGC_(gtk_gc_get(XEN_TO_C_gint(depth), XEN_TO_C_GdkColormap_(colormap), XEN_TO_C_GdkGCValues_(values), 
                                    XEN_TO_C_GdkGCValuesMask(values_mask))));
}

static XEN gxg_gtk_gc_release(XEN gc)
{
  #define H_gtk_gc_release "void gtk_gc_release(GdkGC* gc)"
  XEN_ASSERT_TYPE(XEN_GdkGC__P(gc), gc, 1, "gtk_gc_release", "GdkGC*");
  gtk_gc_release(XEN_TO_C_GdkGC_(gc));
  return(XEN_FALSE);
}

static XEN gxg_gtk_handle_box_get_type(void)
{
  #define H_gtk_handle_box_get_type "GtkType gtk_handle_box_get_type( void)"
  return(C_TO_XEN_GtkType(gtk_handle_box_get_type()));
}

static XEN gxg_gtk_handle_box_new(void)
{
  #define H_gtk_handle_box_new "GtkWidget* gtk_handle_box_new( void)"
  return(C_TO_XEN_GtkWidget_(gtk_handle_box_new()));
}

static XEN gxg_gtk_handle_box_set_shadow_type(XEN handle_box, XEN type)
{
  #define H_gtk_handle_box_set_shadow_type "void gtk_handle_box_set_shadow_type(GtkHandleBox* handle_box, \
GtkShadowType type)"
  XEN_ASSERT_TYPE(XEN_GtkHandleBox__P(handle_box), handle_box, 1, "gtk_handle_box_set_shadow_type", "GtkHandleBox*");
  XEN_ASSERT_TYPE(XEN_GtkShadowType_P(type), type, 2, "gtk_handle_box_set_shadow_type", "GtkShadowType");
  gtk_handle_box_set_shadow_type(XEN_TO_C_GtkHandleBox_(handle_box), XEN_TO_C_GtkShadowType(type));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_handle_box_get_shadow_type(XEN handle_box)
{
  #define H_gtk_handle_box_get_shadow_type "GtkShadowType gtk_handle_box_get_shadow_type(GtkHandleBox* handle_box)"
  XEN_ASSERT_TYPE(XEN_GtkHandleBox__P(handle_box), handle_box, 1, "gtk_handle_box_get_shadow_type", "GtkHandleBox*");
  return(C_TO_XEN_GtkShadowType(gtk_handle_box_get_shadow_type(XEN_TO_C_GtkHandleBox_(handle_box))));
}
#endif

static XEN gxg_gtk_handle_box_set_handle_position(XEN handle_box, XEN position)
{
  #define H_gtk_handle_box_set_handle_position "void gtk_handle_box_set_handle_position(GtkHandleBox* handle_box, \
GtkPositionType position)"
  XEN_ASSERT_TYPE(XEN_GtkHandleBox__P(handle_box), handle_box, 1, "gtk_handle_box_set_handle_position", "GtkHandleBox*");
  XEN_ASSERT_TYPE(XEN_GtkPositionType_P(position), position, 2, "gtk_handle_box_set_handle_position", "GtkPositionType");
  gtk_handle_box_set_handle_position(XEN_TO_C_GtkHandleBox_(handle_box), XEN_TO_C_GtkPositionType(position));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_handle_box_get_handle_position(XEN handle_box)
{
  #define H_gtk_handle_box_get_handle_position "GtkPositionType gtk_handle_box_get_handle_position(GtkHandleBox* handle_box)"
  XEN_ASSERT_TYPE(XEN_GtkHandleBox__P(handle_box), handle_box, 1, "gtk_handle_box_get_handle_position", "GtkHandleBox*");
  return(C_TO_XEN_GtkPositionType(gtk_handle_box_get_handle_position(XEN_TO_C_GtkHandleBox_(handle_box))));
}
#endif

static XEN gxg_gtk_handle_box_set_snap_edge(XEN handle_box, XEN edge)
{
  #define H_gtk_handle_box_set_snap_edge "void gtk_handle_box_set_snap_edge(GtkHandleBox* handle_box, \
GtkPositionType edge)"
  XEN_ASSERT_TYPE(XEN_GtkHandleBox__P(handle_box), handle_box, 1, "gtk_handle_box_set_snap_edge", "GtkHandleBox*");
  XEN_ASSERT_TYPE(XEN_GtkPositionType_P(edge), edge, 2, "gtk_handle_box_set_snap_edge", "GtkPositionType");
  gtk_handle_box_set_snap_edge(XEN_TO_C_GtkHandleBox_(handle_box), XEN_TO_C_GtkPositionType(edge));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_handle_box_get_snap_edge(XEN handle_box)
{
  #define H_gtk_handle_box_get_snap_edge "GtkPositionType gtk_handle_box_get_snap_edge(GtkHandleBox* handle_box)"
  XEN_ASSERT_TYPE(XEN_GtkHandleBox__P(handle_box), handle_box, 1, "gtk_handle_box_get_snap_edge", "GtkHandleBox*");
  return(C_TO_XEN_GtkPositionType(gtk_handle_box_get_snap_edge(XEN_TO_C_GtkHandleBox_(handle_box))));
}
#endif

static XEN gxg_gtk_hbutton_box_get_type(void)
{
  #define H_gtk_hbutton_box_get_type "GtkType gtk_hbutton_box_get_type( void)"
  return(C_TO_XEN_GtkType(gtk_hbutton_box_get_type()));
}

static XEN gxg_gtk_hbutton_box_new(void)
{
  #define H_gtk_hbutton_box_new "GtkWidget* gtk_hbutton_box_new( void)"
  return(C_TO_XEN_GtkWidget_(gtk_hbutton_box_new()));
}

static XEN gxg_gtk_hbox_get_type(void)
{
  #define H_gtk_hbox_get_type "GtkType gtk_hbox_get_type( void)"
  return(C_TO_XEN_GtkType(gtk_hbox_get_type()));
}

static XEN gxg_gtk_hbox_new(XEN homogeneous, XEN spacing)
{
  #define H_gtk_hbox_new "GtkWidget* gtk_hbox_new(gboolean homogeneous, gint spacing)"
  XEN_ASSERT_TYPE(XEN_gboolean_P(homogeneous), homogeneous, 1, "gtk_hbox_new", "gboolean");
  XEN_ASSERT_TYPE(XEN_gint_P(spacing), spacing, 2, "gtk_hbox_new", "gint");
  return(C_TO_XEN_GtkWidget_(gtk_hbox_new(XEN_TO_C_gboolean(homogeneous), XEN_TO_C_gint(spacing))));
}

static XEN gxg_gtk_hpaned_get_type(void)
{
  #define H_gtk_hpaned_get_type "GtkType gtk_hpaned_get_type( void)"
  return(C_TO_XEN_GtkType(gtk_hpaned_get_type()));
}

static XEN gxg_gtk_hpaned_new(void)
{
  #define H_gtk_hpaned_new "GtkWidget* gtk_hpaned_new( void)"
  return(C_TO_XEN_GtkWidget_(gtk_hpaned_new()));
}

static XEN gxg_gtk_hruler_get_type(void)
{
  #define H_gtk_hruler_get_type "GtkType gtk_hruler_get_type( void)"
  return(C_TO_XEN_GtkType(gtk_hruler_get_type()));
}

static XEN gxg_gtk_hruler_new(void)
{
  #define H_gtk_hruler_new "GtkWidget* gtk_hruler_new( void)"
  return(C_TO_XEN_GtkWidget_(gtk_hruler_new()));
}

static XEN gxg_gtk_hscale_get_type(void)
{
  #define H_gtk_hscale_get_type "GtkType gtk_hscale_get_type( void)"
  return(C_TO_XEN_GtkType(gtk_hscale_get_type()));
}

static XEN gxg_gtk_hscale_new(XEN adjustment)
{
  #define H_gtk_hscale_new "GtkWidget* gtk_hscale_new(GtkAdjustment* adjustment)"
  XEN_ASSERT_TYPE(XEN_GtkAdjustment__P(adjustment), adjustment, 1, "gtk_hscale_new", "GtkAdjustment*");
  return(C_TO_XEN_GtkWidget_(gtk_hscale_new(XEN_TO_C_GtkAdjustment_(adjustment))));
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_hscale_new_with_range(XEN min, XEN max, XEN step)
{
  #define H_gtk_hscale_new_with_range "GtkWidget* gtk_hscale_new_with_range(gdouble min, gdouble max, \
gdouble step)"
  XEN_ASSERT_TYPE(XEN_gdouble_P(min), min, 1, "gtk_hscale_new_with_range", "gdouble");
  XEN_ASSERT_TYPE(XEN_gdouble_P(max), max, 2, "gtk_hscale_new_with_range", "gdouble");
  XEN_ASSERT_TYPE(XEN_gdouble_P(step), step, 3, "gtk_hscale_new_with_range", "gdouble");
  return(C_TO_XEN_GtkWidget_(gtk_hscale_new_with_range(XEN_TO_C_gdouble(min), XEN_TO_C_gdouble(max), XEN_TO_C_gdouble(step))));
}
#endif

static XEN gxg_gtk_hscrollbar_get_type(void)
{
  #define H_gtk_hscrollbar_get_type "GtkType gtk_hscrollbar_get_type( void)"
  return(C_TO_XEN_GtkType(gtk_hscrollbar_get_type()));
}

static XEN gxg_gtk_hscrollbar_new(XEN adjustment)
{
  #define H_gtk_hscrollbar_new "GtkWidget* gtk_hscrollbar_new(GtkAdjustment* adjustment)"
  XEN_ASSERT_TYPE(XEN_GtkAdjustment__P(adjustment), adjustment, 1, "gtk_hscrollbar_new", "GtkAdjustment*");
  return(C_TO_XEN_GtkWidget_(gtk_hscrollbar_new(XEN_TO_C_GtkAdjustment_(adjustment))));
}

static XEN gxg_gtk_hseparator_get_type(void)
{
  #define H_gtk_hseparator_get_type "GtkType gtk_hseparator_get_type( void)"
  return(C_TO_XEN_GtkType(gtk_hseparator_get_type()));
}

static XEN gxg_gtk_hseparator_new(void)
{
  #define H_gtk_hseparator_new "GtkWidget* gtk_hseparator_new( void)"
  return(C_TO_XEN_GtkWidget_(gtk_hseparator_new()));
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_icon_factory_get_type(void)
{
  #define H_gtk_icon_factory_get_type "GType gtk_icon_factory_get_type( void)"
  return(C_TO_XEN_GType(gtk_icon_factory_get_type()));
}

static XEN gxg_gtk_icon_factory_new(void)
{
  #define H_gtk_icon_factory_new "GtkIconFactory* gtk_icon_factory_new( void)"
  return(C_TO_XEN_GtkIconFactory_(gtk_icon_factory_new()));
}

static XEN gxg_gtk_icon_factory_add(XEN factory, XEN stock_id, XEN icon_set)
{
  #define H_gtk_icon_factory_add "void gtk_icon_factory_add(GtkIconFactory* factory, gchar* stock_id, \
GtkIconSet* icon_set)"
  XEN_ASSERT_TYPE(XEN_GtkIconFactory__P(factory), factory, 1, "gtk_icon_factory_add", "GtkIconFactory*");
  XEN_ASSERT_TYPE(XEN_gchar__P(stock_id), stock_id, 2, "gtk_icon_factory_add", "gchar*");
  XEN_ASSERT_TYPE(XEN_GtkIconSet__P(icon_set), icon_set, 3, "gtk_icon_factory_add", "GtkIconSet*");
  gtk_icon_factory_add(XEN_TO_C_GtkIconFactory_(factory), XEN_TO_C_gchar_(stock_id), XEN_TO_C_GtkIconSet_(icon_set));
  return(XEN_FALSE);
}

static XEN gxg_gtk_icon_factory_lookup(XEN factory, XEN stock_id)
{
  #define H_gtk_icon_factory_lookup "GtkIconSet* gtk_icon_factory_lookup(GtkIconFactory* factory, gchar* stock_id)"
  XEN_ASSERT_TYPE(XEN_GtkIconFactory__P(factory), factory, 1, "gtk_icon_factory_lookup", "GtkIconFactory*");
  XEN_ASSERT_TYPE(XEN_gchar__P(stock_id), stock_id, 2, "gtk_icon_factory_lookup", "gchar*");
  return(C_TO_XEN_GtkIconSet_(gtk_icon_factory_lookup(XEN_TO_C_GtkIconFactory_(factory), XEN_TO_C_gchar_(stock_id))));
}

static XEN gxg_gtk_icon_factory_add_default(XEN factory)
{
  #define H_gtk_icon_factory_add_default "void gtk_icon_factory_add_default(GtkIconFactory* factory)"
  XEN_ASSERT_TYPE(XEN_GtkIconFactory__P(factory), factory, 1, "gtk_icon_factory_add_default", "GtkIconFactory*");
  gtk_icon_factory_add_default(XEN_TO_C_GtkIconFactory_(factory));
  return(XEN_FALSE);
}

static XEN gxg_gtk_icon_factory_remove_default(XEN factory)
{
  #define H_gtk_icon_factory_remove_default "void gtk_icon_factory_remove_default(GtkIconFactory* factory)"
  XEN_ASSERT_TYPE(XEN_GtkIconFactory__P(factory), factory, 1, "gtk_icon_factory_remove_default", "GtkIconFactory*");
  gtk_icon_factory_remove_default(XEN_TO_C_GtkIconFactory_(factory));
  return(XEN_FALSE);
}

static XEN gxg_gtk_icon_factory_lookup_default(XEN stock_id)
{
  #define H_gtk_icon_factory_lookup_default "GtkIconSet* gtk_icon_factory_lookup_default(gchar* stock_id)"
  XEN_ASSERT_TYPE(XEN_gchar__P(stock_id), stock_id, 1, "gtk_icon_factory_lookup_default", "gchar*");
  return(C_TO_XEN_GtkIconSet_(gtk_icon_factory_lookup_default(XEN_TO_C_gchar_(stock_id))));
}

static XEN gxg_gtk_icon_size_lookup(XEN size, XEN width, XEN height)
{
  #define H_gtk_icon_size_lookup "gboolean gtk_icon_size_lookup(GtkIconSize size, gint* [width], gint* [height])"
  gint ref_width;
  gint ref_height;
  XEN_ASSERT_TYPE(XEN_GtkIconSize_P(size), size, 1, "gtk_icon_size_lookup", "GtkIconSize");
  {
    XEN result = XEN_FALSE;
    result = C_TO_XEN_gboolean(gtk_icon_size_lookup(XEN_TO_C_GtkIconSize(size), &ref_width, &ref_height));
    return(XEN_LIST_3(result, C_TO_XEN_gint(ref_width), C_TO_XEN_gint(ref_height)));
   }
}

static XEN gxg_gtk_icon_size_register(XEN name, XEN width, XEN height)
{
  #define H_gtk_icon_size_register "GtkIconSize gtk_icon_size_register(gchar* name, gint width, gint height)"
  XEN_ASSERT_TYPE(XEN_gchar__P(name), name, 1, "gtk_icon_size_register", "gchar*");
  XEN_ASSERT_TYPE(XEN_gint_P(width), width, 2, "gtk_icon_size_register", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(height), height, 3, "gtk_icon_size_register", "gint");
  return(C_TO_XEN_GtkIconSize(gtk_icon_size_register(XEN_TO_C_gchar_(name), XEN_TO_C_gint(width), XEN_TO_C_gint(height))));
}

static XEN gxg_gtk_icon_size_register_alias(XEN alias, XEN target)
{
  #define H_gtk_icon_size_register_alias "void gtk_icon_size_register_alias(gchar* alias, GtkIconSize target)"
  XEN_ASSERT_TYPE(XEN_gchar__P(alias), alias, 1, "gtk_icon_size_register_alias", "gchar*");
  XEN_ASSERT_TYPE(XEN_GtkIconSize_P(target), target, 2, "gtk_icon_size_register_alias", "GtkIconSize");
  gtk_icon_size_register_alias(XEN_TO_C_gchar_(alias), XEN_TO_C_GtkIconSize(target));
  return(XEN_FALSE);
}

static XEN gxg_gtk_icon_size_from_name(XEN name)
{
  #define H_gtk_icon_size_from_name "GtkIconSize gtk_icon_size_from_name(gchar* name)"
  XEN_ASSERT_TYPE(XEN_gchar__P(name), name, 1, "gtk_icon_size_from_name", "gchar*");
  return(C_TO_XEN_GtkIconSize(gtk_icon_size_from_name(XEN_TO_C_gchar_(name))));
}

static XEN gxg_gtk_icon_size_get_name(XEN size)
{
  #define H_gtk_icon_size_get_name "gchar* gtk_icon_size_get_name(GtkIconSize size)"
  XEN_ASSERT_TYPE(XEN_GtkIconSize_P(size), size, 1, "gtk_icon_size_get_name", "GtkIconSize");
  return(C_TO_XEN_gchar_(gtk_icon_size_get_name(XEN_TO_C_GtkIconSize(size))));
}

static XEN gxg_gtk_icon_set_new(void)
{
  #define H_gtk_icon_set_new "GtkIconSet* gtk_icon_set_new( void)"
  return(C_TO_XEN_GtkIconSet_(gtk_icon_set_new()));
}

static XEN gxg_gtk_icon_set_new_from_pixbuf(XEN pixbuf)
{
  #define H_gtk_icon_set_new_from_pixbuf "GtkIconSet* gtk_icon_set_new_from_pixbuf(GdkPixbuf* pixbuf)"
  XEN_ASSERT_TYPE(XEN_GdkPixbuf__P(pixbuf), pixbuf, 1, "gtk_icon_set_new_from_pixbuf", "GdkPixbuf*");
  return(C_TO_XEN_GtkIconSet_(gtk_icon_set_new_from_pixbuf(XEN_TO_C_GdkPixbuf_(pixbuf))));
}

static XEN gxg_gtk_icon_set_ref(XEN icon_set)
{
  #define H_gtk_icon_set_ref "GtkIconSet* gtk_icon_set_ref(GtkIconSet* icon_set)"
  XEN_ASSERT_TYPE(XEN_GtkIconSet__P(icon_set), icon_set, 1, "gtk_icon_set_ref", "GtkIconSet*");
  return(C_TO_XEN_GtkIconSet_(gtk_icon_set_ref(XEN_TO_C_GtkIconSet_(icon_set))));
}

static XEN gxg_gtk_icon_set_unref(XEN icon_set)
{
  #define H_gtk_icon_set_unref "void gtk_icon_set_unref(GtkIconSet* icon_set)"
  XEN_ASSERT_TYPE(XEN_GtkIconSet__P(icon_set), icon_set, 1, "gtk_icon_set_unref", "GtkIconSet*");
  gtk_icon_set_unref(XEN_TO_C_GtkIconSet_(icon_set));
  return(XEN_FALSE);
}

static XEN gxg_gtk_icon_set_copy(XEN icon_set)
{
  #define H_gtk_icon_set_copy "GtkIconSet* gtk_icon_set_copy(GtkIconSet* icon_set)"
  XEN_ASSERT_TYPE(XEN_GtkIconSet__P(icon_set), icon_set, 1, "gtk_icon_set_copy", "GtkIconSet*");
  return(C_TO_XEN_GtkIconSet_(gtk_icon_set_copy(XEN_TO_C_GtkIconSet_(icon_set))));
}

static XEN gxg_gtk_icon_set_render_icon(XEN icon_set, XEN style, XEN direction, XEN state, XEN size, XEN widget, XEN detail)
{
  #define H_gtk_icon_set_render_icon "GdkPixbuf* gtk_icon_set_render_icon(GtkIconSet* icon_set, GtkStyle* style, \
GtkTextDirection direction, GtkStateType state, GtkIconSize size, GtkWidget* widget, char* detail)"
  XEN_ASSERT_TYPE(XEN_GtkIconSet__P(icon_set), icon_set, 1, "gtk_icon_set_render_icon", "GtkIconSet*");
  XEN_ASSERT_TYPE(XEN_GtkStyle__P(style) || XEN_FALSE_P(style), style, 2, "gtk_icon_set_render_icon", "GtkStyle*");
  XEN_ASSERT_TYPE(XEN_GtkTextDirection_P(direction), direction, 3, "gtk_icon_set_render_icon", "GtkTextDirection");
  XEN_ASSERT_TYPE(XEN_GtkStateType_P(state), state, 4, "gtk_icon_set_render_icon", "GtkStateType");
  XEN_ASSERT_TYPE(XEN_GtkIconSize_P(size), size, 5, "gtk_icon_set_render_icon", "GtkIconSize");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget) || XEN_FALSE_P(widget), widget, 6, "gtk_icon_set_render_icon", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_char__P(detail), detail, 7, "gtk_icon_set_render_icon", "char*");
  return(C_TO_XEN_GdkPixbuf_(gtk_icon_set_render_icon(XEN_TO_C_GtkIconSet_(icon_set), XEN_TO_C_GtkStyle_(style), XEN_TO_C_GtkTextDirection(direction), 
                                                      XEN_TO_C_GtkStateType(state), XEN_TO_C_GtkIconSize(size), XEN_TO_C_GtkWidget_(widget), 
                                                      XEN_TO_C_char_(detail))));
}

static XEN gxg_gtk_icon_set_add_source(XEN icon_set, XEN source)
{
  #define H_gtk_icon_set_add_source "void gtk_icon_set_add_source(GtkIconSet* icon_set, GtkIconSource* source)"
  XEN_ASSERT_TYPE(XEN_GtkIconSet__P(icon_set), icon_set, 1, "gtk_icon_set_add_source", "GtkIconSet*");
  XEN_ASSERT_TYPE(XEN_GtkIconSource__P(source), source, 2, "gtk_icon_set_add_source", "GtkIconSource*");
  gtk_icon_set_add_source(XEN_TO_C_GtkIconSet_(icon_set), XEN_TO_C_GtkIconSource_(source));
  return(XEN_FALSE);
}

static XEN gxg_gtk_icon_set_get_sizes(XEN icon_set, XEN sizes, XEN n_sizes)
{
  #define H_gtk_icon_set_get_sizes "void gtk_icon_set_get_sizes(GtkIconSet* icon_set, GtkIconSize** [sizes], \
gint* [n_sizes])"
  GtkIconSize* ref_sizes;
  gint ref_n_sizes;
  XEN_ASSERT_TYPE(XEN_GtkIconSet__P(icon_set), icon_set, 1, "gtk_icon_set_get_sizes", "GtkIconSet*");
  gtk_icon_set_get_sizes(XEN_TO_C_GtkIconSet_(icon_set), &ref_sizes, &ref_n_sizes);
  return(XEN_LIST_2(C_TO_XEN_GtkIconSize_(ref_sizes), C_TO_XEN_gint(ref_n_sizes)));
}

static XEN gxg_gtk_icon_source_new(void)
{
  #define H_gtk_icon_source_new "GtkIconSource* gtk_icon_source_new( void)"
  return(C_TO_XEN_GtkIconSource_(gtk_icon_source_new()));
}

static XEN gxg_gtk_icon_source_copy(XEN source)
{
  #define H_gtk_icon_source_copy "GtkIconSource* gtk_icon_source_copy(GtkIconSource* source)"
  XEN_ASSERT_TYPE(XEN_GtkIconSource__P(source), source, 1, "gtk_icon_source_copy", "GtkIconSource*");
  return(C_TO_XEN_GtkIconSource_(gtk_icon_source_copy(XEN_TO_C_GtkIconSource_(source))));
}

static XEN gxg_gtk_icon_source_free(XEN source)
{
  #define H_gtk_icon_source_free "void gtk_icon_source_free(GtkIconSource* source)"
  XEN_ASSERT_TYPE(XEN_GtkIconSource__P(source), source, 1, "gtk_icon_source_free", "GtkIconSource*");
  gtk_icon_source_free(XEN_TO_C_GtkIconSource_(source));
  return(XEN_FALSE);
}

static XEN gxg_gtk_icon_source_set_filename(XEN source, XEN filename)
{
  #define H_gtk_icon_source_set_filename "void gtk_icon_source_set_filename(GtkIconSource* source, gchar* filename)"
  XEN_ASSERT_TYPE(XEN_GtkIconSource__P(source), source, 1, "gtk_icon_source_set_filename", "GtkIconSource*");
  XEN_ASSERT_TYPE(XEN_gchar__P(filename), filename, 2, "gtk_icon_source_set_filename", "gchar*");
  gtk_icon_source_set_filename(XEN_TO_C_GtkIconSource_(source), XEN_TO_C_gchar_(filename));
  return(XEN_FALSE);
}

static XEN gxg_gtk_icon_source_set_pixbuf(XEN source, XEN pixbuf)
{
  #define H_gtk_icon_source_set_pixbuf "void gtk_icon_source_set_pixbuf(GtkIconSource* source, GdkPixbuf* pixbuf)"
  XEN_ASSERT_TYPE(XEN_GtkIconSource__P(source), source, 1, "gtk_icon_source_set_pixbuf", "GtkIconSource*");
  XEN_ASSERT_TYPE(XEN_GdkPixbuf__P(pixbuf), pixbuf, 2, "gtk_icon_source_set_pixbuf", "GdkPixbuf*");
  gtk_icon_source_set_pixbuf(XEN_TO_C_GtkIconSource_(source), XEN_TO_C_GdkPixbuf_(pixbuf));
  return(XEN_FALSE);
}

static XEN gxg_gtk_icon_source_get_filename(XEN source)
{
  #define H_gtk_icon_source_get_filename "gchar* gtk_icon_source_get_filename(GtkIconSource* source)"
  XEN_ASSERT_TYPE(XEN_GtkIconSource__P(source), source, 1, "gtk_icon_source_get_filename", "GtkIconSource*");
  return(C_TO_XEN_gchar_(gtk_icon_source_get_filename(XEN_TO_C_GtkIconSource_(source))));
}

static XEN gxg_gtk_icon_source_get_pixbuf(XEN source)
{
  #define H_gtk_icon_source_get_pixbuf "GdkPixbuf* gtk_icon_source_get_pixbuf(GtkIconSource* source)"
  XEN_ASSERT_TYPE(XEN_GtkIconSource__P(source), source, 1, "gtk_icon_source_get_pixbuf", "GtkIconSource*");
  return(C_TO_XEN_GdkPixbuf_(gtk_icon_source_get_pixbuf(XEN_TO_C_GtkIconSource_(source))));
}

static XEN gxg_gtk_icon_source_set_direction_wildcarded(XEN source, XEN setting)
{
  #define H_gtk_icon_source_set_direction_wildcarded "void gtk_icon_source_set_direction_wildcarded(GtkIconSource* source, \
gboolean setting)"
  XEN_ASSERT_TYPE(XEN_GtkIconSource__P(source), source, 1, "gtk_icon_source_set_direction_wildcarded", "GtkIconSource*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(setting), setting, 2, "gtk_icon_source_set_direction_wildcarded", "gboolean");
  gtk_icon_source_set_direction_wildcarded(XEN_TO_C_GtkIconSource_(source), XEN_TO_C_gboolean(setting));
  return(XEN_FALSE);
}

static XEN gxg_gtk_icon_source_set_state_wildcarded(XEN source, XEN setting)
{
  #define H_gtk_icon_source_set_state_wildcarded "void gtk_icon_source_set_state_wildcarded(GtkIconSource* source, \
gboolean setting)"
  XEN_ASSERT_TYPE(XEN_GtkIconSource__P(source), source, 1, "gtk_icon_source_set_state_wildcarded", "GtkIconSource*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(setting), setting, 2, "gtk_icon_source_set_state_wildcarded", "gboolean");
  gtk_icon_source_set_state_wildcarded(XEN_TO_C_GtkIconSource_(source), XEN_TO_C_gboolean(setting));
  return(XEN_FALSE);
}

static XEN gxg_gtk_icon_source_set_size_wildcarded(XEN source, XEN setting)
{
  #define H_gtk_icon_source_set_size_wildcarded "void gtk_icon_source_set_size_wildcarded(GtkIconSource* source, \
gboolean setting)"
  XEN_ASSERT_TYPE(XEN_GtkIconSource__P(source), source, 1, "gtk_icon_source_set_size_wildcarded", "GtkIconSource*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(setting), setting, 2, "gtk_icon_source_set_size_wildcarded", "gboolean");
  gtk_icon_source_set_size_wildcarded(XEN_TO_C_GtkIconSource_(source), XEN_TO_C_gboolean(setting));
  return(XEN_FALSE);
}

static XEN gxg_gtk_icon_source_get_size_wildcarded(XEN source)
{
  #define H_gtk_icon_source_get_size_wildcarded "gboolean gtk_icon_source_get_size_wildcarded(GtkIconSource* source)"
  XEN_ASSERT_TYPE(XEN_GtkIconSource__P(source), source, 1, "gtk_icon_source_get_size_wildcarded", "GtkIconSource*");
  return(C_TO_XEN_gboolean(gtk_icon_source_get_size_wildcarded(XEN_TO_C_GtkIconSource_(source))));
}

static XEN gxg_gtk_icon_source_get_state_wildcarded(XEN source)
{
  #define H_gtk_icon_source_get_state_wildcarded "gboolean gtk_icon_source_get_state_wildcarded(GtkIconSource* source)"
  XEN_ASSERT_TYPE(XEN_GtkIconSource__P(source), source, 1, "gtk_icon_source_get_state_wildcarded", "GtkIconSource*");
  return(C_TO_XEN_gboolean(gtk_icon_source_get_state_wildcarded(XEN_TO_C_GtkIconSource_(source))));
}

static XEN gxg_gtk_icon_source_get_direction_wildcarded(XEN source)
{
  #define H_gtk_icon_source_get_direction_wildcarded "gboolean gtk_icon_source_get_direction_wildcarded(GtkIconSource* source)"
  XEN_ASSERT_TYPE(XEN_GtkIconSource__P(source), source, 1, "gtk_icon_source_get_direction_wildcarded", "GtkIconSource*");
  return(C_TO_XEN_gboolean(gtk_icon_source_get_direction_wildcarded(XEN_TO_C_GtkIconSource_(source))));
}

static XEN gxg_gtk_icon_source_set_direction(XEN source, XEN direction)
{
  #define H_gtk_icon_source_set_direction "void gtk_icon_source_set_direction(GtkIconSource* source, \
GtkTextDirection direction)"
  XEN_ASSERT_TYPE(XEN_GtkIconSource__P(source), source, 1, "gtk_icon_source_set_direction", "GtkIconSource*");
  XEN_ASSERT_TYPE(XEN_GtkTextDirection_P(direction), direction, 2, "gtk_icon_source_set_direction", "GtkTextDirection");
  gtk_icon_source_set_direction(XEN_TO_C_GtkIconSource_(source), XEN_TO_C_GtkTextDirection(direction));
  return(XEN_FALSE);
}

static XEN gxg_gtk_icon_source_set_state(XEN source, XEN state)
{
  #define H_gtk_icon_source_set_state "void gtk_icon_source_set_state(GtkIconSource* source, GtkStateType state)"
  XEN_ASSERT_TYPE(XEN_GtkIconSource__P(source), source, 1, "gtk_icon_source_set_state", "GtkIconSource*");
  XEN_ASSERT_TYPE(XEN_GtkStateType_P(state), state, 2, "gtk_icon_source_set_state", "GtkStateType");
  gtk_icon_source_set_state(XEN_TO_C_GtkIconSource_(source), XEN_TO_C_GtkStateType(state));
  return(XEN_FALSE);
}

static XEN gxg_gtk_icon_source_set_size(XEN source, XEN size)
{
  #define H_gtk_icon_source_set_size "void gtk_icon_source_set_size(GtkIconSource* source, GtkIconSize size)"
  XEN_ASSERT_TYPE(XEN_GtkIconSource__P(source), source, 1, "gtk_icon_source_set_size", "GtkIconSource*");
  XEN_ASSERT_TYPE(XEN_GtkIconSize_P(size), size, 2, "gtk_icon_source_set_size", "GtkIconSize");
  gtk_icon_source_set_size(XEN_TO_C_GtkIconSource_(source), XEN_TO_C_GtkIconSize(size));
  return(XEN_FALSE);
}

static XEN gxg_gtk_icon_source_get_direction(XEN source)
{
  #define H_gtk_icon_source_get_direction "GtkTextDirection gtk_icon_source_get_direction(GtkIconSource* source)"
  XEN_ASSERT_TYPE(XEN_GtkIconSource__P(source), source, 1, "gtk_icon_source_get_direction", "GtkIconSource*");
  return(C_TO_XEN_GtkTextDirection(gtk_icon_source_get_direction(XEN_TO_C_GtkIconSource_(source))));
}

static XEN gxg_gtk_icon_source_get_state(XEN source)
{
  #define H_gtk_icon_source_get_state "GtkStateType gtk_icon_source_get_state(GtkIconSource* source)"
  XEN_ASSERT_TYPE(XEN_GtkIconSource__P(source), source, 1, "gtk_icon_source_get_state", "GtkIconSource*");
  return(C_TO_XEN_GtkStateType(gtk_icon_source_get_state(XEN_TO_C_GtkIconSource_(source))));
}

static XEN gxg_gtk_icon_source_get_size(XEN source)
{
  #define H_gtk_icon_source_get_size "GtkIconSize gtk_icon_source_get_size(GtkIconSource* source)"
  XEN_ASSERT_TYPE(XEN_GtkIconSource__P(source), source, 1, "gtk_icon_source_get_size", "GtkIconSource*");
  return(C_TO_XEN_GtkIconSize(gtk_icon_source_get_size(XEN_TO_C_GtkIconSource_(source))));
}
#endif

static XEN gxg_gtk_image_get_type(void)
{
  #define H_gtk_image_get_type "GtkType gtk_image_get_type( void)"
  return(C_TO_XEN_GtkType(gtk_image_get_type()));
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_image_new(void)
{
  #define H_gtk_image_new "GtkWidget* gtk_image_new( void)"
  return(C_TO_XEN_GtkWidget_(gtk_image_new()));
}

static XEN gxg_gtk_image_new_from_pixmap(XEN pixmap, XEN mask)
{
  #define H_gtk_image_new_from_pixmap "GtkWidget* gtk_image_new_from_pixmap(GdkPixmap* pixmap, GdkBitmap* mask)"
  XEN_ASSERT_TYPE(XEN_GdkPixmap__P(pixmap) || XEN_FALSE_P(pixmap), pixmap, 1, "gtk_image_new_from_pixmap", "GdkPixmap*");
  XEN_ASSERT_TYPE(XEN_GdkBitmap__P(mask) || XEN_FALSE_P(mask), mask, 2, "gtk_image_new_from_pixmap", "GdkBitmap*");
  return(C_TO_XEN_GtkWidget_(gtk_image_new_from_pixmap(XEN_TO_C_GdkPixmap_(pixmap), XEN_TO_C_GdkBitmap_(mask))));
}

static XEN gxg_gtk_image_new_from_image(XEN image, XEN mask)
{
  #define H_gtk_image_new_from_image "GtkWidget* gtk_image_new_from_image(GdkImage* image, GdkBitmap* mask)"
  XEN_ASSERT_TYPE(XEN_GdkImage__P(image) || XEN_FALSE_P(image), image, 1, "gtk_image_new_from_image", "GdkImage*");
  XEN_ASSERT_TYPE(XEN_GdkBitmap__P(mask) || XEN_FALSE_P(mask), mask, 2, "gtk_image_new_from_image", "GdkBitmap*");
  return(C_TO_XEN_GtkWidget_(gtk_image_new_from_image(XEN_TO_C_GdkImage_(image), XEN_TO_C_GdkBitmap_(mask))));
}

static XEN gxg_gtk_image_new_from_file(XEN filename)
{
  #define H_gtk_image_new_from_file "GtkWidget* gtk_image_new_from_file(gchar* filename)"
  XEN_ASSERT_TYPE(XEN_gchar__P(filename), filename, 1, "gtk_image_new_from_file", "gchar*");
  return(C_TO_XEN_GtkWidget_(gtk_image_new_from_file(XEN_TO_C_gchar_(filename))));
}

static XEN gxg_gtk_image_new_from_pixbuf(XEN pixbuf)
{
  #define H_gtk_image_new_from_pixbuf "GtkWidget* gtk_image_new_from_pixbuf(GdkPixbuf* pixbuf)"
  XEN_ASSERT_TYPE(XEN_GdkPixbuf__P(pixbuf) || XEN_FALSE_P(pixbuf), pixbuf, 1, "gtk_image_new_from_pixbuf", "GdkPixbuf*");
  return(C_TO_XEN_GtkWidget_(gtk_image_new_from_pixbuf(XEN_TO_C_GdkPixbuf_(pixbuf))));
}

static XEN gxg_gtk_image_new_from_stock(XEN stock_id, XEN size)
{
  #define H_gtk_image_new_from_stock "GtkWidget* gtk_image_new_from_stock(gchar* stock_id, GtkIconSize size)"
  XEN_ASSERT_TYPE(XEN_gchar__P(stock_id), stock_id, 1, "gtk_image_new_from_stock", "gchar*");
  XEN_ASSERT_TYPE(XEN_GtkIconSize_P(size), size, 2, "gtk_image_new_from_stock", "GtkIconSize");
  return(C_TO_XEN_GtkWidget_(gtk_image_new_from_stock(XEN_TO_C_gchar_(stock_id), XEN_TO_C_GtkIconSize(size))));
}

static XEN gxg_gtk_image_new_from_icon_set(XEN icon_set, XEN size)
{
  #define H_gtk_image_new_from_icon_set "GtkWidget* gtk_image_new_from_icon_set(GtkIconSet* icon_set, \
GtkIconSize size)"
  XEN_ASSERT_TYPE(XEN_GtkIconSet__P(icon_set), icon_set, 1, "gtk_image_new_from_icon_set", "GtkIconSet*");
  XEN_ASSERT_TYPE(XEN_GtkIconSize_P(size), size, 2, "gtk_image_new_from_icon_set", "GtkIconSize");
  return(C_TO_XEN_GtkWidget_(gtk_image_new_from_icon_set(XEN_TO_C_GtkIconSet_(icon_set), XEN_TO_C_GtkIconSize(size))));
}

static XEN gxg_gtk_image_new_from_animation(XEN animation)
{
  #define H_gtk_image_new_from_animation "GtkWidget* gtk_image_new_from_animation(GdkPixbufAnimation* animation)"
  XEN_ASSERT_TYPE(XEN_GdkPixbufAnimation__P(animation), animation, 1, "gtk_image_new_from_animation", "GdkPixbufAnimation*");
  return(C_TO_XEN_GtkWidget_(gtk_image_new_from_animation(XEN_TO_C_GdkPixbufAnimation_(animation))));
}

static XEN gxg_gtk_image_set_from_pixmap(XEN image, XEN pixmap, XEN mask)
{
  #define H_gtk_image_set_from_pixmap "void gtk_image_set_from_pixmap(GtkImage* image, GdkPixmap* pixmap, \
GdkBitmap* mask)"
  XEN_ASSERT_TYPE(XEN_GtkImage__P(image), image, 1, "gtk_image_set_from_pixmap", "GtkImage*");
  XEN_ASSERT_TYPE(XEN_GdkPixmap__P(pixmap) || XEN_FALSE_P(pixmap), pixmap, 2, "gtk_image_set_from_pixmap", "GdkPixmap*");
  XEN_ASSERT_TYPE(XEN_GdkBitmap__P(mask) || XEN_FALSE_P(mask), mask, 3, "gtk_image_set_from_pixmap", "GdkBitmap*");
  gtk_image_set_from_pixmap(XEN_TO_C_GtkImage_(image), XEN_TO_C_GdkPixmap_(pixmap), XEN_TO_C_GdkBitmap_(mask));
  return(XEN_FALSE);
}

static XEN gxg_gtk_image_set_from_image(XEN image, XEN gdk_image, XEN mask)
{
  #define H_gtk_image_set_from_image "void gtk_image_set_from_image(GtkImage* image, GdkImage* gdk_image, \
GdkBitmap* mask)"
  XEN_ASSERT_TYPE(XEN_GtkImage__P(image), image, 1, "gtk_image_set_from_image", "GtkImage*");
  XEN_ASSERT_TYPE(XEN_GdkImage__P(gdk_image) || XEN_FALSE_P(gdk_image), gdk_image, 2, "gtk_image_set_from_image", "GdkImage*");
  XEN_ASSERT_TYPE(XEN_GdkBitmap__P(mask) || XEN_FALSE_P(mask), mask, 3, "gtk_image_set_from_image", "GdkBitmap*");
  gtk_image_set_from_image(XEN_TO_C_GtkImage_(image), XEN_TO_C_GdkImage_(gdk_image), XEN_TO_C_GdkBitmap_(mask));
  return(XEN_FALSE);
}

static XEN gxg_gtk_image_set_from_file(XEN image, XEN filename)
{
  #define H_gtk_image_set_from_file "void gtk_image_set_from_file(GtkImage* image, gchar* filename)"
  XEN_ASSERT_TYPE(XEN_GtkImage__P(image), image, 1, "gtk_image_set_from_file", "GtkImage*");
  XEN_ASSERT_TYPE(XEN_gchar__P(filename), filename, 2, "gtk_image_set_from_file", "gchar*");
  gtk_image_set_from_file(XEN_TO_C_GtkImage_(image), XEN_TO_C_gchar_(filename));
  return(XEN_FALSE);
}

static XEN gxg_gtk_image_set_from_pixbuf(XEN image, XEN pixbuf)
{
  #define H_gtk_image_set_from_pixbuf "void gtk_image_set_from_pixbuf(GtkImage* image, GdkPixbuf* pixbuf)"
  XEN_ASSERT_TYPE(XEN_GtkImage__P(image), image, 1, "gtk_image_set_from_pixbuf", "GtkImage*");
  XEN_ASSERT_TYPE(XEN_GdkPixbuf__P(pixbuf) || XEN_FALSE_P(pixbuf), pixbuf, 2, "gtk_image_set_from_pixbuf", "GdkPixbuf*");
  gtk_image_set_from_pixbuf(XEN_TO_C_GtkImage_(image), XEN_TO_C_GdkPixbuf_(pixbuf));
  return(XEN_FALSE);
}

static XEN gxg_gtk_image_set_from_stock(XEN image, XEN stock_id, XEN size)
{
  #define H_gtk_image_set_from_stock "void gtk_image_set_from_stock(GtkImage* image, gchar* stock_id, \
GtkIconSize size)"
  XEN_ASSERT_TYPE(XEN_GtkImage__P(image), image, 1, "gtk_image_set_from_stock", "GtkImage*");
  XEN_ASSERT_TYPE(XEN_gchar__P(stock_id), stock_id, 2, "gtk_image_set_from_stock", "gchar*");
  XEN_ASSERT_TYPE(XEN_GtkIconSize_P(size), size, 3, "gtk_image_set_from_stock", "GtkIconSize");
  gtk_image_set_from_stock(XEN_TO_C_GtkImage_(image), XEN_TO_C_gchar_(stock_id), XEN_TO_C_GtkIconSize(size));
  return(XEN_FALSE);
}

static XEN gxg_gtk_image_set_from_icon_set(XEN image, XEN icon_set, XEN size)
{
  #define H_gtk_image_set_from_icon_set "void gtk_image_set_from_icon_set(GtkImage* image, GtkIconSet* icon_set, \
GtkIconSize size)"
  XEN_ASSERT_TYPE(XEN_GtkImage__P(image), image, 1, "gtk_image_set_from_icon_set", "GtkImage*");
  XEN_ASSERT_TYPE(XEN_GtkIconSet__P(icon_set), icon_set, 2, "gtk_image_set_from_icon_set", "GtkIconSet*");
  XEN_ASSERT_TYPE(XEN_GtkIconSize_P(size), size, 3, "gtk_image_set_from_icon_set", "GtkIconSize");
  gtk_image_set_from_icon_set(XEN_TO_C_GtkImage_(image), XEN_TO_C_GtkIconSet_(icon_set), XEN_TO_C_GtkIconSize(size));
  return(XEN_FALSE);
}

static XEN gxg_gtk_image_set_from_animation(XEN image, XEN animation)
{
  #define H_gtk_image_set_from_animation "void gtk_image_set_from_animation(GtkImage* image, GdkPixbufAnimation* animation)"
  XEN_ASSERT_TYPE(XEN_GtkImage__P(image), image, 1, "gtk_image_set_from_animation", "GtkImage*");
  XEN_ASSERT_TYPE(XEN_GdkPixbufAnimation__P(animation) || XEN_FALSE_P(animation), animation, 2, "gtk_image_set_from_animation", "GdkPixbufAnimation*");
  gtk_image_set_from_animation(XEN_TO_C_GtkImage_(image), XEN_TO_C_GdkPixbufAnimation_(animation));
  return(XEN_FALSE);
}

static XEN gxg_gtk_image_get_storage_type(XEN image)
{
  #define H_gtk_image_get_storage_type "GtkImageType gtk_image_get_storage_type(GtkImage* image)"
  XEN_ASSERT_TYPE(XEN_GtkImage__P(image), image, 1, "gtk_image_get_storage_type", "GtkImage*");
  return(C_TO_XEN_GtkImageType(gtk_image_get_storage_type(XEN_TO_C_GtkImage_(image))));
}

static XEN gxg_gtk_image_get_pixmap(XEN image, XEN pixmap, XEN mask)
{
  #define H_gtk_image_get_pixmap "void gtk_image_get_pixmap(GtkImage* image, GdkPixmap** [pixmap], GdkBitmap** [mask])"
  GdkPixmap* ref_pixmap;
  GdkBitmap* ref_mask;
  XEN_ASSERT_TYPE(XEN_GtkImage__P(image), image, 1, "gtk_image_get_pixmap", "GtkImage*");
  gtk_image_get_pixmap(XEN_TO_C_GtkImage_(image), &ref_pixmap, &ref_mask);
  return(XEN_LIST_2(C_TO_XEN_GdkPixmap_(ref_pixmap), C_TO_XEN_GdkBitmap_(ref_mask)));
}

static XEN gxg_gtk_image_get_image(XEN image, XEN gdk_image, XEN mask)
{
  #define H_gtk_image_get_image "void gtk_image_get_image(GtkImage* image, GdkImage** [gdk_image], GdkBitmap** [mask])"
  GdkImage* ref_gdk_image;
  GdkBitmap* ref_mask;
  XEN_ASSERT_TYPE(XEN_GtkImage__P(image), image, 1, "gtk_image_get_image", "GtkImage*");
  gtk_image_get_image(XEN_TO_C_GtkImage_(image), &ref_gdk_image, &ref_mask);
  return(XEN_LIST_2(C_TO_XEN_GdkImage_(ref_gdk_image), C_TO_XEN_GdkBitmap_(ref_mask)));
}

static XEN gxg_gtk_image_get_pixbuf(XEN image)
{
  #define H_gtk_image_get_pixbuf "GdkPixbuf* gtk_image_get_pixbuf(GtkImage* image)"
  XEN_ASSERT_TYPE(XEN_GtkImage__P(image), image, 1, "gtk_image_get_pixbuf", "GtkImage*");
  return(C_TO_XEN_GdkPixbuf_(gtk_image_get_pixbuf(XEN_TO_C_GtkImage_(image))));
}

static XEN gxg_gtk_image_get_stock(XEN image, XEN stock_id, XEN size)
{
  #define H_gtk_image_get_stock "void gtk_image_get_stock(GtkImage* image, gchar** [stock_id], GtkIconSize* [size])"
  gchar* ref_stock_id;
  GtkIconSize ref_size;
  XEN_ASSERT_TYPE(XEN_GtkImage__P(image), image, 1, "gtk_image_get_stock", "GtkImage*");
  gtk_image_get_stock(XEN_TO_C_GtkImage_(image), &ref_stock_id, &ref_size);
  return(XEN_LIST_2(C_TO_XEN_gchar_(ref_stock_id), C_TO_XEN_GtkIconSize(ref_size)));
}

static XEN gxg_gtk_image_get_icon_set(XEN image, XEN icon_set, XEN size)
{
  #define H_gtk_image_get_icon_set "void gtk_image_get_icon_set(GtkImage* image, GtkIconSet** [icon_set], \
GtkIconSize* [size])"
  GtkIconSet* ref_icon_set;
  GtkIconSize ref_size;
  XEN_ASSERT_TYPE(XEN_GtkImage__P(image), image, 1, "gtk_image_get_icon_set", "GtkImage*");
  gtk_image_get_icon_set(XEN_TO_C_GtkImage_(image), &ref_icon_set, &ref_size);
  return(XEN_LIST_2(C_TO_XEN_GtkIconSet_(ref_icon_set), C_TO_XEN_GtkIconSize(ref_size)));
}

static XEN gxg_gtk_image_get_animation(XEN image)
{
  #define H_gtk_image_get_animation "GdkPixbufAnimation* gtk_image_get_animation(GtkImage* image)"
  XEN_ASSERT_TYPE(XEN_GtkImage__P(image), image, 1, "gtk_image_get_animation", "GtkImage*");
  return(C_TO_XEN_GdkPixbufAnimation_(gtk_image_get_animation(XEN_TO_C_GtkImage_(image))));
}

static XEN gxg_gtk_image_menu_item_get_type(void)
{
  #define H_gtk_image_menu_item_get_type "GtkType gtk_image_menu_item_get_type( void)"
  return(C_TO_XEN_GtkType(gtk_image_menu_item_get_type()));
}

static XEN gxg_gtk_image_menu_item_new(void)
{
  #define H_gtk_image_menu_item_new "GtkWidget* gtk_image_menu_item_new( void)"
  return(C_TO_XEN_GtkWidget_(gtk_image_menu_item_new()));
}

static XEN gxg_gtk_image_menu_item_new_with_label(XEN label)
{
  #define H_gtk_image_menu_item_new_with_label "GtkWidget* gtk_image_menu_item_new_with_label(gchar* label)"
  XEN_ASSERT_TYPE(XEN_gchar__P(label), label, 1, "gtk_image_menu_item_new_with_label", "gchar*");
  return(C_TO_XEN_GtkWidget_(gtk_image_menu_item_new_with_label(XEN_TO_C_gchar_(label))));
}

static XEN gxg_gtk_image_menu_item_new_with_mnemonic(XEN label)
{
  #define H_gtk_image_menu_item_new_with_mnemonic "GtkWidget* gtk_image_menu_item_new_with_mnemonic(gchar* label)"
  XEN_ASSERT_TYPE(XEN_gchar__P(label), label, 1, "gtk_image_menu_item_new_with_mnemonic", "gchar*");
  return(C_TO_XEN_GtkWidget_(gtk_image_menu_item_new_with_mnemonic(XEN_TO_C_gchar_(label))));
}

static XEN gxg_gtk_image_menu_item_new_from_stock(XEN stock_id, XEN accel_group)
{
  #define H_gtk_image_menu_item_new_from_stock "GtkWidget* gtk_image_menu_item_new_from_stock(gchar* stock_id, \
GtkAccelGroup* accel_group)"
  XEN_ASSERT_TYPE(XEN_gchar__P(stock_id), stock_id, 1, "gtk_image_menu_item_new_from_stock", "gchar*");
  XEN_ASSERT_TYPE(XEN_GtkAccelGroup__P(accel_group), accel_group, 2, "gtk_image_menu_item_new_from_stock", "GtkAccelGroup*");
  return(C_TO_XEN_GtkWidget_(gtk_image_menu_item_new_from_stock(XEN_TO_C_gchar_(stock_id), XEN_TO_C_GtkAccelGroup_(accel_group))));
}

static XEN gxg_gtk_image_menu_item_set_image(XEN image_menu_item, XEN image)
{
  #define H_gtk_image_menu_item_set_image "void gtk_image_menu_item_set_image(GtkImageMenuItem* image_menu_item, \
GtkWidget* image)"
  XEN_ASSERT_TYPE(XEN_GtkImageMenuItem__P(image_menu_item), image_menu_item, 1, "gtk_image_menu_item_set_image", "GtkImageMenuItem*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(image), image, 2, "gtk_image_menu_item_set_image", "GtkWidget*");
  gtk_image_menu_item_set_image(XEN_TO_C_GtkImageMenuItem_(image_menu_item), XEN_TO_C_GtkWidget_(image));
  return(XEN_FALSE);
}

static XEN gxg_gtk_image_menu_item_get_image(XEN image_menu_item)
{
  #define H_gtk_image_menu_item_get_image "GtkWidget* gtk_image_menu_item_get_image(GtkImageMenuItem* image_menu_item)"
  XEN_ASSERT_TYPE(XEN_GtkImageMenuItem__P(image_menu_item), image_menu_item, 1, "gtk_image_menu_item_get_image", "GtkImageMenuItem*");
  return(C_TO_XEN_GtkWidget_(gtk_image_menu_item_get_image(XEN_TO_C_GtkImageMenuItem_(image_menu_item))));
}

static XEN gxg_gtk_im_context_get_type(void)
{
  #define H_gtk_im_context_get_type "GtkType gtk_im_context_get_type( void)"
  return(C_TO_XEN_GtkType(gtk_im_context_get_type()));
}

static XEN gxg_gtk_im_context_set_client_window(XEN context, XEN window)
{
  #define H_gtk_im_context_set_client_window "void gtk_im_context_set_client_window(GtkIMContext* context, \
GdkWindow* window)"
  XEN_ASSERT_TYPE(XEN_GtkIMContext__P(context), context, 1, "gtk_im_context_set_client_window", "GtkIMContext*");
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window) || XEN_FALSE_P(window), window, 2, "gtk_im_context_set_client_window", "GdkWindow*");
  gtk_im_context_set_client_window(XEN_TO_C_GtkIMContext_(context), XEN_TO_C_GdkWindow_(window));
  return(XEN_FALSE);
}

static XEN gxg_gtk_im_context_get_preedit_string(XEN context, XEN str, XEN attrs, XEN cursor_pos)
{
  #define H_gtk_im_context_get_preedit_string "void gtk_im_context_get_preedit_string(GtkIMContext* context, \
gchar** [str], PangoAttrList** [attrs], gint* [cursor_pos])"
  gchar* ref_str;
  PangoAttrList* ref_attrs;
  gint ref_cursor_pos;
  XEN_ASSERT_TYPE(XEN_GtkIMContext__P(context), context, 1, "gtk_im_context_get_preedit_string", "GtkIMContext*");
  gtk_im_context_get_preedit_string(XEN_TO_C_GtkIMContext_(context), &ref_str, &ref_attrs, &ref_cursor_pos);
  return(XEN_LIST_3(C_TO_XEN_gchar_(ref_str), C_TO_XEN_PangoAttrList_(ref_attrs), C_TO_XEN_gint(ref_cursor_pos)));
}

static XEN gxg_gtk_im_context_filter_keypress(XEN context, XEN event)
{
  #define H_gtk_im_context_filter_keypress "gboolean gtk_im_context_filter_keypress(GtkIMContext* context, \
GdkEventKey* event)"
  XEN_ASSERT_TYPE(XEN_GtkIMContext__P(context), context, 1, "gtk_im_context_filter_keypress", "GtkIMContext*");
  XEN_ASSERT_TYPE(XEN_GdkEventKey__P(event), event, 2, "gtk_im_context_filter_keypress", "GdkEventKey*");
  return(C_TO_XEN_gboolean(gtk_im_context_filter_keypress(XEN_TO_C_GtkIMContext_(context), XEN_TO_C_GdkEventKey_(event))));
}

static XEN gxg_gtk_im_context_focus_in(XEN context)
{
  #define H_gtk_im_context_focus_in "void gtk_im_context_focus_in(GtkIMContext* context)"
  XEN_ASSERT_TYPE(XEN_GtkIMContext__P(context), context, 1, "gtk_im_context_focus_in", "GtkIMContext*");
  gtk_im_context_focus_in(XEN_TO_C_GtkIMContext_(context));
  return(XEN_FALSE);
}

static XEN gxg_gtk_im_context_focus_out(XEN context)
{
  #define H_gtk_im_context_focus_out "void gtk_im_context_focus_out(GtkIMContext* context)"
  XEN_ASSERT_TYPE(XEN_GtkIMContext__P(context), context, 1, "gtk_im_context_focus_out", "GtkIMContext*");
  gtk_im_context_focus_out(XEN_TO_C_GtkIMContext_(context));
  return(XEN_FALSE);
}

static XEN gxg_gtk_im_context_reset(XEN context)
{
  #define H_gtk_im_context_reset "void gtk_im_context_reset(GtkIMContext* context)"
  XEN_ASSERT_TYPE(XEN_GtkIMContext__P(context), context, 1, "gtk_im_context_reset", "GtkIMContext*");
  gtk_im_context_reset(XEN_TO_C_GtkIMContext_(context));
  return(XEN_FALSE);
}

static XEN gxg_gtk_im_context_set_cursor_location(XEN context, XEN area)
{
  #define H_gtk_im_context_set_cursor_location "void gtk_im_context_set_cursor_location(GtkIMContext* context, \
GdkRectangle* area)"
  XEN_ASSERT_TYPE(XEN_GtkIMContext__P(context), context, 1, "gtk_im_context_set_cursor_location", "GtkIMContext*");
  XEN_ASSERT_TYPE(XEN_GdkRectangle__P(area), area, 2, "gtk_im_context_set_cursor_location", "GdkRectangle*");
  gtk_im_context_set_cursor_location(XEN_TO_C_GtkIMContext_(context), XEN_TO_C_GdkRectangle_(area));
  return(XEN_FALSE);
}

static XEN gxg_gtk_im_context_set_use_preedit(XEN context, XEN use_preedit)
{
  #define H_gtk_im_context_set_use_preedit "void gtk_im_context_set_use_preedit(GtkIMContext* context, \
gboolean use_preedit)"
  XEN_ASSERT_TYPE(XEN_GtkIMContext__P(context), context, 1, "gtk_im_context_set_use_preedit", "GtkIMContext*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(use_preedit), use_preedit, 2, "gtk_im_context_set_use_preedit", "gboolean");
  gtk_im_context_set_use_preedit(XEN_TO_C_GtkIMContext_(context), XEN_TO_C_gboolean(use_preedit));
  return(XEN_FALSE);
}

static XEN gxg_gtk_im_context_set_surrounding(XEN context, XEN text, XEN len, XEN cursor_index)
{
  #define H_gtk_im_context_set_surrounding "void gtk_im_context_set_surrounding(GtkIMContext* context, \
gchar* text, gint len, gint cursor_index)"
  XEN_ASSERT_TYPE(XEN_GtkIMContext__P(context), context, 1, "gtk_im_context_set_surrounding", "GtkIMContext*");
  XEN_ASSERT_TYPE(XEN_gchar__P(text), text, 2, "gtk_im_context_set_surrounding", "gchar*");
  XEN_ASSERT_TYPE(XEN_gint_P(len), len, 3, "gtk_im_context_set_surrounding", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(cursor_index), cursor_index, 4, "gtk_im_context_set_surrounding", "gint");
  gtk_im_context_set_surrounding(XEN_TO_C_GtkIMContext_(context), XEN_TO_C_gchar_(text), XEN_TO_C_gint(len), XEN_TO_C_gint(cursor_index));
  return(XEN_FALSE);
}

static XEN gxg_gtk_im_context_get_surrounding(XEN context, XEN text, XEN cursor_index)
{
  #define H_gtk_im_context_get_surrounding "gboolean gtk_im_context_get_surrounding(GtkIMContext* context, \
gchar** [text], gint* [cursor_index])"
  gchar* ref_text;
  gint ref_cursor_index;
  XEN_ASSERT_TYPE(XEN_GtkIMContext__P(context), context, 1, "gtk_im_context_get_surrounding", "GtkIMContext*");
  {
    XEN result = XEN_FALSE;
    result = C_TO_XEN_gboolean(gtk_im_context_get_surrounding(XEN_TO_C_GtkIMContext_(context), &ref_text, &ref_cursor_index));
    return(XEN_LIST_3(result, C_TO_XEN_gchar_(ref_text), C_TO_XEN_gint(ref_cursor_index)));
   }
}

static XEN gxg_gtk_im_context_delete_surrounding(XEN context, XEN offset, XEN n_chars)
{
  #define H_gtk_im_context_delete_surrounding "gboolean gtk_im_context_delete_surrounding(GtkIMContext* context, \
gint offset, gint n_chars)"
  XEN_ASSERT_TYPE(XEN_GtkIMContext__P(context), context, 1, "gtk_im_context_delete_surrounding", "GtkIMContext*");
  XEN_ASSERT_TYPE(XEN_gint_P(offset), offset, 2, "gtk_im_context_delete_surrounding", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(n_chars), n_chars, 3, "gtk_im_context_delete_surrounding", "gint");
  return(C_TO_XEN_gboolean(gtk_im_context_delete_surrounding(XEN_TO_C_GtkIMContext_(context), XEN_TO_C_gint(offset), XEN_TO_C_gint(n_chars))));
}

static XEN gxg_gtk_im_context_simple_get_type(void)
{
  #define H_gtk_im_context_simple_get_type "GtkType gtk_im_context_simple_get_type( void)"
  return(C_TO_XEN_GtkType(gtk_im_context_simple_get_type()));
}

static XEN gxg_gtk_im_context_simple_new(void)
{
  #define H_gtk_im_context_simple_new "GtkIMContext* gtk_im_context_simple_new( void)"
  return(C_TO_XEN_GtkIMContext_(gtk_im_context_simple_new()));
}

static XEN gxg_gtk_im_context_simple_add_table(XEN context_simple, XEN data, XEN max_seq_len, XEN n_seqs)
{
  #define H_gtk_im_context_simple_add_table "void gtk_im_context_simple_add_table(GtkIMContextSimple* context_simple, \
guint16* data, gint max_seq_len, gint n_seqs)"
  XEN_ASSERT_TYPE(XEN_GtkIMContextSimple__P(context_simple), context_simple, 1, "gtk_im_context_simple_add_table", "GtkIMContextSimple*");
  XEN_ASSERT_TYPE(XEN_guint16__P(data), data, 2, "gtk_im_context_simple_add_table", "guint16*");
  XEN_ASSERT_TYPE(XEN_gint_P(max_seq_len), max_seq_len, 3, "gtk_im_context_simple_add_table", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(n_seqs), n_seqs, 4, "gtk_im_context_simple_add_table", "gint");
  gtk_im_context_simple_add_table(XEN_TO_C_GtkIMContextSimple_(context_simple), XEN_TO_C_guint16_(data), XEN_TO_C_gint(max_seq_len), 
                                  XEN_TO_C_gint(n_seqs));
  return(XEN_FALSE);
}

static XEN gxg_gtk_im_multicontext_get_type(void)
{
  #define H_gtk_im_multicontext_get_type "GtkType gtk_im_multicontext_get_type( void)"
  return(C_TO_XEN_GtkType(gtk_im_multicontext_get_type()));
}

static XEN gxg_gtk_im_multicontext_new(void)
{
  #define H_gtk_im_multicontext_new "GtkIMContext* gtk_im_multicontext_new( void)"
  return(C_TO_XEN_GtkIMContext_(gtk_im_multicontext_new()));
}

static XEN gxg_gtk_im_multicontext_append_menuitems(XEN context, XEN menushell)
{
  #define H_gtk_im_multicontext_append_menuitems "void gtk_im_multicontext_append_menuitems(GtkIMMulticontext* context, \
GtkMenuShell* menushell)"
  XEN_ASSERT_TYPE(XEN_GtkIMMulticontext__P(context), context, 1, "gtk_im_multicontext_append_menuitems", "GtkIMMulticontext*");
  XEN_ASSERT_TYPE(XEN_GtkMenuShell__P(menushell), menushell, 2, "gtk_im_multicontext_append_menuitems", "GtkMenuShell*");
  gtk_im_multicontext_append_menuitems(XEN_TO_C_GtkIMMulticontext_(context), XEN_TO_C_GtkMenuShell_(menushell));
  return(XEN_FALSE);
}
#endif

static XEN gxg_gtk_input_dialog_get_type(void)
{
  #define H_gtk_input_dialog_get_type "GtkType gtk_input_dialog_get_type( void)"
  return(C_TO_XEN_GtkType(gtk_input_dialog_get_type()));
}

static XEN gxg_gtk_input_dialog_new(void)
{
  #define H_gtk_input_dialog_new "GtkWidget* gtk_input_dialog_new( void)"
  return(C_TO_XEN_GtkWidget_(gtk_input_dialog_new()));
}

static XEN gxg_gtk_invisible_get_type(void)
{
  #define H_gtk_invisible_get_type "GtkType gtk_invisible_get_type( void)"
  return(C_TO_XEN_GtkType(gtk_invisible_get_type()));
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_invisible_new(void)
{
  #define H_gtk_invisible_new "GtkWidget* gtk_invisible_new( void)"
  return(C_TO_XEN_GtkWidget_(gtk_invisible_new()));
}
#endif

static XEN gxg_gtk_item_factory_get_type(void)
{
  #define H_gtk_item_factory_get_type "GtkType gtk_item_factory_get_type( void)"
  return(C_TO_XEN_GtkType(gtk_item_factory_get_type()));
}

static XEN gxg_gtk_item_factory_new(XEN container_type, XEN path, XEN accel_group)
{
  #define H_gtk_item_factory_new "GtkItemFactory* gtk_item_factory_new(GtkType container_type, gchar* path, \
GtkAccelGroup* accel_group)"
  XEN_ASSERT_TYPE(XEN_GtkType_P(container_type), container_type, 1, "gtk_item_factory_new", "GtkType");
  XEN_ASSERT_TYPE(XEN_gchar__P(path), path, 2, "gtk_item_factory_new", "gchar*");
  XEN_ASSERT_TYPE(XEN_GtkAccelGroup__P(accel_group) || XEN_FALSE_P(accel_group), accel_group, 3, "gtk_item_factory_new", "GtkAccelGroup*");
  return(C_TO_XEN_GtkItemFactory_(gtk_item_factory_new(XEN_TO_C_GtkType(container_type), XEN_TO_C_gchar_(path), XEN_TO_C_GtkAccelGroup_(accel_group))));
}

static XEN gxg_gtk_item_factory_construct(XEN ifactory, XEN container_type, XEN path, XEN accel_group)
{
  #define H_gtk_item_factory_construct "void gtk_item_factory_construct(GtkItemFactory* ifactory, GtkType container_type, \
gchar* path, GtkAccelGroup* accel_group)"
  XEN_ASSERT_TYPE(XEN_GtkItemFactory__P(ifactory), ifactory, 1, "gtk_item_factory_construct", "GtkItemFactory*");
  XEN_ASSERT_TYPE(XEN_GtkType_P(container_type), container_type, 2, "gtk_item_factory_construct", "GtkType");
  XEN_ASSERT_TYPE(XEN_gchar__P(path), path, 3, "gtk_item_factory_construct", "gchar*");
  XEN_ASSERT_TYPE(XEN_GtkAccelGroup__P(accel_group) || XEN_FALSE_P(accel_group), accel_group, 4, "gtk_item_factory_construct", "GtkAccelGroup*");
  gtk_item_factory_construct(XEN_TO_C_GtkItemFactory_(ifactory), XEN_TO_C_GtkType(container_type), XEN_TO_C_gchar_(path), 
                             XEN_TO_C_GtkAccelGroup_(accel_group));
  return(XEN_FALSE);
}

static XEN gxg_gtk_item_factory_add_foreign(XEN accel_widget, XEN full_path, XEN accel_group, XEN keyval, XEN modifiers)
{
  #define H_gtk_item_factory_add_foreign "void gtk_item_factory_add_foreign(GtkWidget* accel_widget, \
gchar* full_path, GtkAccelGroup* accel_group, guint keyval, GdkModifierType modifiers)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(accel_widget), accel_widget, 1, "gtk_item_factory_add_foreign", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_gchar__P(full_path), full_path, 2, "gtk_item_factory_add_foreign", "gchar*");
  XEN_ASSERT_TYPE(XEN_GtkAccelGroup__P(accel_group), accel_group, 3, "gtk_item_factory_add_foreign", "GtkAccelGroup*");
  XEN_ASSERT_TYPE(XEN_guint_P(keyval), keyval, 4, "gtk_item_factory_add_foreign", "guint");
  XEN_ASSERT_TYPE(XEN_GdkModifierType_P(modifiers), modifiers, 5, "gtk_item_factory_add_foreign", "GdkModifierType");
  gtk_item_factory_add_foreign(XEN_TO_C_GtkWidget_(accel_widget), XEN_TO_C_gchar_(full_path), XEN_TO_C_GtkAccelGroup_(accel_group), 
                               XEN_TO_C_guint(keyval), XEN_TO_C_GdkModifierType(modifiers));
  return(XEN_FALSE);
}

static XEN gxg_gtk_item_factory_from_widget(XEN widget)
{
  #define H_gtk_item_factory_from_widget "GtkItemFactory* gtk_item_factory_from_widget(GtkWidget* widget)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_item_factory_from_widget", "GtkWidget*");
  return(C_TO_XEN_GtkItemFactory_(gtk_item_factory_from_widget(XEN_TO_C_GtkWidget_(widget))));
}

static XEN gxg_gtk_item_factory_path_from_widget(XEN widget)
{
  #define H_gtk_item_factory_path_from_widget "gchar* gtk_item_factory_path_from_widget(GtkWidget* widget)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_item_factory_path_from_widget", "GtkWidget*");
  return(C_TO_XEN_gchar_(gtk_item_factory_path_from_widget(XEN_TO_C_GtkWidget_(widget))));
}

static XEN gxg_gtk_item_factory_get_item(XEN ifactory, XEN path)
{
  #define H_gtk_item_factory_get_item "GtkWidget* gtk_item_factory_get_item(GtkItemFactory* ifactory, \
gchar* path)"
  XEN_ASSERT_TYPE(XEN_GtkItemFactory__P(ifactory), ifactory, 1, "gtk_item_factory_get_item", "GtkItemFactory*");
  XEN_ASSERT_TYPE(XEN_gchar__P(path), path, 2, "gtk_item_factory_get_item", "gchar*");
  return(C_TO_XEN_GtkWidget_(gtk_item_factory_get_item(XEN_TO_C_GtkItemFactory_(ifactory), XEN_TO_C_gchar_(path))));
}

static XEN gxg_gtk_item_factory_get_widget(XEN ifactory, XEN path)
{
  #define H_gtk_item_factory_get_widget "GtkWidget* gtk_item_factory_get_widget(GtkItemFactory* ifactory, \
gchar* path)"
  XEN_ASSERT_TYPE(XEN_GtkItemFactory__P(ifactory), ifactory, 1, "gtk_item_factory_get_widget", "GtkItemFactory*");
  XEN_ASSERT_TYPE(XEN_gchar__P(path), path, 2, "gtk_item_factory_get_widget", "gchar*");
  return(C_TO_XEN_GtkWidget_(gtk_item_factory_get_widget(XEN_TO_C_GtkItemFactory_(ifactory), XEN_TO_C_gchar_(path))));
}

static XEN gxg_gtk_item_factory_get_widget_by_action(XEN ifactory, XEN action)
{
  #define H_gtk_item_factory_get_widget_by_action "GtkWidget* gtk_item_factory_get_widget_by_action(GtkItemFactory* ifactory, \
guint action)"
  XEN_ASSERT_TYPE(XEN_GtkItemFactory__P(ifactory), ifactory, 1, "gtk_item_factory_get_widget_by_action", "GtkItemFactory*");
  XEN_ASSERT_TYPE(XEN_guint_P(action), action, 2, "gtk_item_factory_get_widget_by_action", "guint");
  return(C_TO_XEN_GtkWidget_(gtk_item_factory_get_widget_by_action(XEN_TO_C_GtkItemFactory_(ifactory), XEN_TO_C_guint(action))));
}

static XEN gxg_gtk_item_factory_get_item_by_action(XEN ifactory, XEN action)
{
  #define H_gtk_item_factory_get_item_by_action "GtkWidget* gtk_item_factory_get_item_by_action(GtkItemFactory* ifactory, \
guint action)"
  XEN_ASSERT_TYPE(XEN_GtkItemFactory__P(ifactory), ifactory, 1, "gtk_item_factory_get_item_by_action", "GtkItemFactory*");
  XEN_ASSERT_TYPE(XEN_guint_P(action), action, 2, "gtk_item_factory_get_item_by_action", "guint");
  return(C_TO_XEN_GtkWidget_(gtk_item_factory_get_item_by_action(XEN_TO_C_GtkItemFactory_(ifactory), XEN_TO_C_guint(action))));
}

static XEN gxg_gtk_item_factory_create_item(XEN ifactory, XEN entry, XEN callback_data, XEN callback_type)
{
  #define H_gtk_item_factory_create_item "void gtk_item_factory_create_item(GtkItemFactory* ifactory, \
GtkItemFactoryEntry* entry, gpointer callback_data, guint callback_type)"
  XEN_ASSERT_TYPE(XEN_GtkItemFactory__P(ifactory), ifactory, 1, "gtk_item_factory_create_item", "GtkItemFactory*");
  XEN_ASSERT_TYPE(XEN_GtkItemFactoryEntry__P(entry), entry, 2, "gtk_item_factory_create_item", "GtkItemFactoryEntry*");
  XEN_ASSERT_TYPE(XEN_gpointer_P(callback_data), callback_data, 3, "gtk_item_factory_create_item", "gpointer");
  XEN_ASSERT_TYPE(XEN_guint_P(callback_type), callback_type, 4, "gtk_item_factory_create_item", "guint");
  gtk_item_factory_create_item(XEN_TO_C_GtkItemFactory_(ifactory), XEN_TO_C_GtkItemFactoryEntry_(entry), XEN_TO_C_gpointer(callback_data), 
                               XEN_TO_C_guint(callback_type));
  return(XEN_FALSE);
}

static XEN gxg_gtk_item_factory_create_items(XEN ifactory, XEN n_entries, XEN entries, XEN callback_data)
{
  #define H_gtk_item_factory_create_items "void gtk_item_factory_create_items(GtkItemFactory* ifactory, \
guint n_entries, GtkItemFactoryEntry* entries, gpointer callback_data)"
  XEN_ASSERT_TYPE(XEN_GtkItemFactory__P(ifactory), ifactory, 1, "gtk_item_factory_create_items", "GtkItemFactory*");
  XEN_ASSERT_TYPE(XEN_guint_P(n_entries), n_entries, 2, "gtk_item_factory_create_items", "guint");
  XEN_ASSERT_TYPE(XEN_GtkItemFactoryEntry__P(entries), entries, 3, "gtk_item_factory_create_items", "GtkItemFactoryEntry*");
  XEN_ASSERT_TYPE(XEN_gpointer_P(callback_data), callback_data, 4, "gtk_item_factory_create_items", "gpointer");
  gtk_item_factory_create_items(XEN_TO_C_GtkItemFactory_(ifactory), XEN_TO_C_guint(n_entries), XEN_TO_C_GtkItemFactoryEntry_(entries), 
                                XEN_TO_C_gpointer(callback_data));
  return(XEN_FALSE);
}

static XEN gxg_gtk_item_factory_delete_item(XEN ifactory, XEN path)
{
  #define H_gtk_item_factory_delete_item "void gtk_item_factory_delete_item(GtkItemFactory* ifactory, \
gchar* path)"
  XEN_ASSERT_TYPE(XEN_GtkItemFactory__P(ifactory), ifactory, 1, "gtk_item_factory_delete_item", "GtkItemFactory*");
  XEN_ASSERT_TYPE(XEN_gchar__P(path), path, 2, "gtk_item_factory_delete_item", "gchar*");
  gtk_item_factory_delete_item(XEN_TO_C_GtkItemFactory_(ifactory), XEN_TO_C_gchar_(path));
  return(XEN_FALSE);
}

static XEN gxg_gtk_item_factory_delete_entry(XEN ifactory, XEN entry)
{
  #define H_gtk_item_factory_delete_entry "void gtk_item_factory_delete_entry(GtkItemFactory* ifactory, \
GtkItemFactoryEntry* entry)"
  XEN_ASSERT_TYPE(XEN_GtkItemFactory__P(ifactory), ifactory, 1, "gtk_item_factory_delete_entry", "GtkItemFactory*");
  XEN_ASSERT_TYPE(XEN_GtkItemFactoryEntry__P(entry), entry, 2, "gtk_item_factory_delete_entry", "GtkItemFactoryEntry*");
  gtk_item_factory_delete_entry(XEN_TO_C_GtkItemFactory_(ifactory), XEN_TO_C_GtkItemFactoryEntry_(entry));
  return(XEN_FALSE);
}

static XEN gxg_gtk_item_factory_delete_entries(XEN ifactory, XEN n_entries, XEN entries)
{
  #define H_gtk_item_factory_delete_entries "void gtk_item_factory_delete_entries(GtkItemFactory* ifactory, \
guint n_entries, GtkItemFactoryEntry* entries)"
  XEN_ASSERT_TYPE(XEN_GtkItemFactory__P(ifactory), ifactory, 1, "gtk_item_factory_delete_entries", "GtkItemFactory*");
  XEN_ASSERT_TYPE(XEN_guint_P(n_entries), n_entries, 2, "gtk_item_factory_delete_entries", "guint");
  XEN_ASSERT_TYPE(XEN_GtkItemFactoryEntry__P(entries), entries, 3, "gtk_item_factory_delete_entries", "GtkItemFactoryEntry*");
  gtk_item_factory_delete_entries(XEN_TO_C_GtkItemFactory_(ifactory), XEN_TO_C_guint(n_entries), XEN_TO_C_GtkItemFactoryEntry_(entries));
  return(XEN_FALSE);
}

static XEN gxg_gtk_item_factory_popup(XEN ifactory, XEN x, XEN y, XEN mouse_button, XEN time)
{
  #define H_gtk_item_factory_popup "void gtk_item_factory_popup(GtkItemFactory* ifactory, guint x, guint y, \
guint mouse_button, guint32 time)"
  XEN_ASSERT_TYPE(XEN_GtkItemFactory__P(ifactory), ifactory, 1, "gtk_item_factory_popup", "GtkItemFactory*");
  XEN_ASSERT_TYPE(XEN_guint_P(x), x, 2, "gtk_item_factory_popup", "guint");
  XEN_ASSERT_TYPE(XEN_guint_P(y), y, 3, "gtk_item_factory_popup", "guint");
  XEN_ASSERT_TYPE(XEN_guint_P(mouse_button), mouse_button, 4, "gtk_item_factory_popup", "guint");
  XEN_ASSERT_TYPE(XEN_guint32_P(time), time, 5, "gtk_item_factory_popup", "guint32");
  gtk_item_factory_popup(XEN_TO_C_GtkItemFactory_(ifactory), XEN_TO_C_guint(x), XEN_TO_C_guint(y), XEN_TO_C_guint(mouse_button), 
                         XEN_TO_C_guint32(time));
  return(XEN_FALSE);
}

static XEN gxg_gtk_item_factory_popup_with_data(XEN ifactory, XEN func_data, XEN destroy, XEN x, XEN y, XEN mouse_button, XEN time)
{
  #define H_gtk_item_factory_popup_with_data "void gtk_item_factory_popup_with_data(GtkItemFactory* ifactory, \
gpointer func_data, GtkDestroyNotify destroy, guint x, guint y, guint mouse_button, guint32 time)"
  XEN_ASSERT_TYPE(XEN_GtkItemFactory__P(ifactory), ifactory, 1, "gtk_item_factory_popup_with_data", "GtkItemFactory*");
  XEN_ASSERT_TYPE(XEN_gpointer_P(func_data), func_data, 2, "gtk_item_factory_popup_with_data", "gpointer");
  XEN_ASSERT_TYPE(XEN_GtkDestroyNotify_P(destroy), destroy, 3, "gtk_item_factory_popup_with_data", "GtkDestroyNotify");
  XEN_ASSERT_TYPE(XEN_guint_P(x), x, 4, "gtk_item_factory_popup_with_data", "guint");
  XEN_ASSERT_TYPE(XEN_guint_P(y), y, 5, "gtk_item_factory_popup_with_data", "guint");
  XEN_ASSERT_TYPE(XEN_guint_P(mouse_button), mouse_button, 6, "gtk_item_factory_popup_with_data", "guint");
  XEN_ASSERT_TYPE(XEN_guint32_P(time), time, 7, "gtk_item_factory_popup_with_data", "guint32");
  {
    XEN gxg_ptr = XEN_LIST_5(XEN_FALSE, func_data, XEN_FALSE, XEN_FALSE, XEN_FALSE);
    xm_protect(gxg_ptr);
    XEN_LIST_SET(gxg_ptr, 3, destroy);
    gtk_item_factory_popup_with_data(XEN_TO_C_GtkItemFactory_(ifactory), XEN_TO_C_gpointer(func_data), XEN_TO_C_GtkDestroyNotify(destroy), 
                                 XEN_TO_C_guint(x), XEN_TO_C_guint(y), XEN_TO_C_guint(mouse_button), XEN_TO_C_guint32(time));
    return(XEN_FALSE);
   }
}

static XEN gxg_gtk_item_factory_popup_data(XEN ifactory)
{
  #define H_gtk_item_factory_popup_data "gpointer gtk_item_factory_popup_data(GtkItemFactory* ifactory)"
  XEN_ASSERT_TYPE(XEN_GtkItemFactory__P(ifactory), ifactory, 1, "gtk_item_factory_popup_data", "GtkItemFactory*");
  return(C_TO_XEN_gpointer(gtk_item_factory_popup_data(XEN_TO_C_GtkItemFactory_(ifactory))));
}

static XEN gxg_gtk_item_factory_popup_data_from_widget(XEN widget)
{
  #define H_gtk_item_factory_popup_data_from_widget "gpointer gtk_item_factory_popup_data_from_widget(GtkWidget* widget)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_item_factory_popup_data_from_widget", "GtkWidget*");
  return(C_TO_XEN_gpointer(gtk_item_factory_popup_data_from_widget(XEN_TO_C_GtkWidget_(widget))));
}

static XEN gxg_gtk_item_factory_set_translate_func(XEN ifactory, XEN func, XEN func_data, XEN notify)
{
  #define H_gtk_item_factory_set_translate_func "void gtk_item_factory_set_translate_func(GtkItemFactory* ifactory, \
GtkTranslateFunc func, lambda_data func_data, GtkDestroyNotify notify)"
  XEN_ASSERT_TYPE(XEN_GtkItemFactory__P(ifactory), ifactory, 1, "gtk_item_factory_set_translate_func", "GtkItemFactory*");
  XEN_ASSERT_TYPE(XEN_GtkTranslateFunc_P(func), func, 2, "gtk_item_factory_set_translate_func", "GtkTranslateFunc");
  XEN_ASSERT_TYPE(XEN_lambda_data_P(func_data), func_data, 3, "gtk_item_factory_set_translate_func", "lambda_data");
  XEN_ASSERT_TYPE(XEN_GtkDestroyNotify_P(notify), notify, 4, "gtk_item_factory_set_translate_func", "GtkDestroyNotify");
  {
    XEN gxg_ptr = XEN_LIST_5(func, func_data, XEN_FALSE, XEN_FALSE, XEN_FALSE);
    xm_protect(gxg_ptr);
    XEN_LIST_SET(gxg_ptr, 3, notify);
    gtk_item_factory_set_translate_func(XEN_TO_C_GtkItemFactory_(ifactory), XEN_TO_C_GtkTranslateFunc(func), XEN_TO_C_lambda_data(func_data), 
                                    XEN_TO_C_GtkDestroyNotify(notify));
    return(XEN_FALSE);
   }
}

static XEN gxg_gtk_item_get_type(void)
{
  #define H_gtk_item_get_type "GtkType gtk_item_get_type( void)"
  return(C_TO_XEN_GtkType(gtk_item_get_type()));
}

static XEN gxg_gtk_item_select(XEN item)
{
  #define H_gtk_item_select "void gtk_item_select(GtkItem* item)"
  XEN_ASSERT_TYPE(XEN_GtkItem__P(item), item, 1, "gtk_item_select", "GtkItem*");
  gtk_item_select(XEN_TO_C_GtkItem_(item));
  return(XEN_FALSE);
}

static XEN gxg_gtk_item_deselect(XEN item)
{
  #define H_gtk_item_deselect "void gtk_item_deselect(GtkItem* item)"
  XEN_ASSERT_TYPE(XEN_GtkItem__P(item), item, 1, "gtk_item_deselect", "GtkItem*");
  gtk_item_deselect(XEN_TO_C_GtkItem_(item));
  return(XEN_FALSE);
}

static XEN gxg_gtk_item_toggle(XEN item)
{
  #define H_gtk_item_toggle "void gtk_item_toggle(GtkItem* item)"
  XEN_ASSERT_TYPE(XEN_GtkItem__P(item), item, 1, "gtk_item_toggle", "GtkItem*");
  gtk_item_toggle(XEN_TO_C_GtkItem_(item));
  return(XEN_FALSE);
}

static XEN gxg_gtk_label_get_type(void)
{
  #define H_gtk_label_get_type "GtkType gtk_label_get_type( void)"
  return(C_TO_XEN_GtkType(gtk_label_get_type()));
}

static XEN gxg_gtk_label_new(XEN str)
{
  #define H_gtk_label_new "GtkWidget* gtk_label_new(char* str)"
  XEN_ASSERT_TYPE(XEN_char__P(str), str, 1, "gtk_label_new", "char*");
  return(C_TO_XEN_GtkWidget_(gtk_label_new(XEN_TO_C_char_(str))));
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_label_new_with_mnemonic(XEN str)
{
  #define H_gtk_label_new_with_mnemonic "GtkWidget* gtk_label_new_with_mnemonic(char* str)"
  XEN_ASSERT_TYPE(XEN_char__P(str), str, 1, "gtk_label_new_with_mnemonic", "char*");
  return(C_TO_XEN_GtkWidget_(gtk_label_new_with_mnemonic(XEN_TO_C_char_(str))));
}
#endif

static XEN gxg_gtk_label_set_text(XEN label, XEN str)
{
  #define H_gtk_label_set_text "void gtk_label_set_text(GtkLabel* label, char* str)"
  XEN_ASSERT_TYPE(XEN_GtkLabel__P(label), label, 1, "gtk_label_set_text", "GtkLabel*");
  XEN_ASSERT_TYPE(XEN_char__P(str), str, 2, "gtk_label_set_text", "char*");
  gtk_label_set_text(XEN_TO_C_GtkLabel_(label), XEN_TO_C_char_(str));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_label_get_text(XEN label)
{
  #define H_gtk_label_get_text "gchar* gtk_label_get_text(GtkLabel* label)"
  XEN_ASSERT_TYPE(XEN_GtkLabel__P(label), label, 1, "gtk_label_get_text", "GtkLabel*");
  return(C_TO_XEN_gchar_(gtk_label_get_text(XEN_TO_C_GtkLabel_(label))));
}

static XEN gxg_gtk_label_set_attributes(XEN label, XEN attrs)
{
  #define H_gtk_label_set_attributes "void gtk_label_set_attributes(GtkLabel* label, PangoAttrList* attrs)"
  XEN_ASSERT_TYPE(XEN_GtkLabel__P(label), label, 1, "gtk_label_set_attributes", "GtkLabel*");
  XEN_ASSERT_TYPE(XEN_PangoAttrList__P(attrs), attrs, 2, "gtk_label_set_attributes", "PangoAttrList*");
  gtk_label_set_attributes(XEN_TO_C_GtkLabel_(label), XEN_TO_C_PangoAttrList_(attrs));
  return(XEN_FALSE);
}

static XEN gxg_gtk_label_get_attributes(XEN label)
{
  #define H_gtk_label_get_attributes "PangoAttrList* gtk_label_get_attributes(GtkLabel* label)"
  XEN_ASSERT_TYPE(XEN_GtkLabel__P(label), label, 1, "gtk_label_get_attributes", "GtkLabel*");
  return(C_TO_XEN_PangoAttrList_(gtk_label_get_attributes(XEN_TO_C_GtkLabel_(label))));
}

static XEN gxg_gtk_label_set_label(XEN label, XEN str)
{
  #define H_gtk_label_set_label "void gtk_label_set_label(GtkLabel* label, gchar* str)"
  XEN_ASSERT_TYPE(XEN_GtkLabel__P(label), label, 1, "gtk_label_set_label", "GtkLabel*");
  XEN_ASSERT_TYPE(XEN_gchar__P(str), str, 2, "gtk_label_set_label", "gchar*");
  gtk_label_set_label(XEN_TO_C_GtkLabel_(label), XEN_TO_C_gchar_(str));
  return(XEN_FALSE);
}

static XEN gxg_gtk_label_get_label(XEN label)
{
  #define H_gtk_label_get_label "gchar* gtk_label_get_label(GtkLabel* label)"
  XEN_ASSERT_TYPE(XEN_GtkLabel__P(label), label, 1, "gtk_label_get_label", "GtkLabel*");
  return(C_TO_XEN_gchar_(gtk_label_get_label(XEN_TO_C_GtkLabel_(label))));
}

static XEN gxg_gtk_label_set_markup(XEN label, XEN str)
{
  #define H_gtk_label_set_markup "void gtk_label_set_markup(GtkLabel* label, gchar* str)"
  XEN_ASSERT_TYPE(XEN_GtkLabel__P(label), label, 1, "gtk_label_set_markup", "GtkLabel*");
  XEN_ASSERT_TYPE(XEN_gchar__P(str), str, 2, "gtk_label_set_markup", "gchar*");
  gtk_label_set_markup(XEN_TO_C_GtkLabel_(label), XEN_TO_C_gchar_(str));
  return(XEN_FALSE);
}

static XEN gxg_gtk_label_set_use_markup(XEN label, XEN setting)
{
  #define H_gtk_label_set_use_markup "void gtk_label_set_use_markup(GtkLabel* label, gboolean setting)"
  XEN_ASSERT_TYPE(XEN_GtkLabel__P(label), label, 1, "gtk_label_set_use_markup", "GtkLabel*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(setting), setting, 2, "gtk_label_set_use_markup", "gboolean");
  gtk_label_set_use_markup(XEN_TO_C_GtkLabel_(label), XEN_TO_C_gboolean(setting));
  return(XEN_FALSE);
}

static XEN gxg_gtk_label_get_use_markup(XEN label)
{
  #define H_gtk_label_get_use_markup "gboolean gtk_label_get_use_markup(GtkLabel* label)"
  XEN_ASSERT_TYPE(XEN_GtkLabel__P(label), label, 1, "gtk_label_get_use_markup", "GtkLabel*");
  return(C_TO_XEN_gboolean(gtk_label_get_use_markup(XEN_TO_C_GtkLabel_(label))));
}

static XEN gxg_gtk_label_set_use_underline(XEN label, XEN setting)
{
  #define H_gtk_label_set_use_underline "void gtk_label_set_use_underline(GtkLabel* label, gboolean setting)"
  XEN_ASSERT_TYPE(XEN_GtkLabel__P(label), label, 1, "gtk_label_set_use_underline", "GtkLabel*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(setting), setting, 2, "gtk_label_set_use_underline", "gboolean");
  gtk_label_set_use_underline(XEN_TO_C_GtkLabel_(label), XEN_TO_C_gboolean(setting));
  return(XEN_FALSE);
}

static XEN gxg_gtk_label_get_use_underline(XEN label)
{
  #define H_gtk_label_get_use_underline "gboolean gtk_label_get_use_underline(GtkLabel* label)"
  XEN_ASSERT_TYPE(XEN_GtkLabel__P(label), label, 1, "gtk_label_get_use_underline", "GtkLabel*");
  return(C_TO_XEN_gboolean(gtk_label_get_use_underline(XEN_TO_C_GtkLabel_(label))));
}

static XEN gxg_gtk_label_set_markup_with_mnemonic(XEN label, XEN str)
{
  #define H_gtk_label_set_markup_with_mnemonic "void gtk_label_set_markup_with_mnemonic(GtkLabel* label, \
gchar* str)"
  XEN_ASSERT_TYPE(XEN_GtkLabel__P(label), label, 1, "gtk_label_set_markup_with_mnemonic", "GtkLabel*");
  XEN_ASSERT_TYPE(XEN_gchar__P(str), str, 2, "gtk_label_set_markup_with_mnemonic", "gchar*");
  gtk_label_set_markup_with_mnemonic(XEN_TO_C_GtkLabel_(label), XEN_TO_C_gchar_(str));
  return(XEN_FALSE);
}

static XEN gxg_gtk_label_get_mnemonic_keyval(XEN label)
{
  #define H_gtk_label_get_mnemonic_keyval "guint gtk_label_get_mnemonic_keyval(GtkLabel* label)"
  XEN_ASSERT_TYPE(XEN_GtkLabel__P(label), label, 1, "gtk_label_get_mnemonic_keyval", "GtkLabel*");
  return(C_TO_XEN_guint(gtk_label_get_mnemonic_keyval(XEN_TO_C_GtkLabel_(label))));
}

static XEN gxg_gtk_label_set_mnemonic_widget(XEN label, XEN widget)
{
  #define H_gtk_label_set_mnemonic_widget "void gtk_label_set_mnemonic_widget(GtkLabel* label, GtkWidget* widget)"
  XEN_ASSERT_TYPE(XEN_GtkLabel__P(label), label, 1, "gtk_label_set_mnemonic_widget", "GtkLabel*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 2, "gtk_label_set_mnemonic_widget", "GtkWidget*");
  gtk_label_set_mnemonic_widget(XEN_TO_C_GtkLabel_(label), XEN_TO_C_GtkWidget_(widget));
  return(XEN_FALSE);
}

static XEN gxg_gtk_label_get_mnemonic_widget(XEN label)
{
  #define H_gtk_label_get_mnemonic_widget "GtkWidget* gtk_label_get_mnemonic_widget(GtkLabel* label)"
  XEN_ASSERT_TYPE(XEN_GtkLabel__P(label), label, 1, "gtk_label_get_mnemonic_widget", "GtkLabel*");
  return(C_TO_XEN_GtkWidget_(gtk_label_get_mnemonic_widget(XEN_TO_C_GtkLabel_(label))));
}

static XEN gxg_gtk_label_set_text_with_mnemonic(XEN label, XEN str)
{
  #define H_gtk_label_set_text_with_mnemonic "void gtk_label_set_text_with_mnemonic(GtkLabel* label, \
gchar* str)"
  XEN_ASSERT_TYPE(XEN_GtkLabel__P(label), label, 1, "gtk_label_set_text_with_mnemonic", "GtkLabel*");
  XEN_ASSERT_TYPE(XEN_gchar__P(str), str, 2, "gtk_label_set_text_with_mnemonic", "gchar*");
  gtk_label_set_text_with_mnemonic(XEN_TO_C_GtkLabel_(label), XEN_TO_C_gchar_(str));
  return(XEN_FALSE);
}
#endif

static XEN gxg_gtk_label_set_justify(XEN label, XEN jtype)
{
  #define H_gtk_label_set_justify "void gtk_label_set_justify(GtkLabel* label, GtkJustification jtype)"
  XEN_ASSERT_TYPE(XEN_GtkLabel__P(label), label, 1, "gtk_label_set_justify", "GtkLabel*");
  XEN_ASSERT_TYPE(XEN_GtkJustification_P(jtype), jtype, 2, "gtk_label_set_justify", "GtkJustification");
  gtk_label_set_justify(XEN_TO_C_GtkLabel_(label), XEN_TO_C_GtkJustification(jtype));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_label_get_justify(XEN label)
{
  #define H_gtk_label_get_justify "GtkJustification gtk_label_get_justify(GtkLabel* label)"
  XEN_ASSERT_TYPE(XEN_GtkLabel__P(label), label, 1, "gtk_label_get_justify", "GtkLabel*");
  return(C_TO_XEN_GtkJustification(gtk_label_get_justify(XEN_TO_C_GtkLabel_(label))));
}
#endif

static XEN gxg_gtk_label_set_pattern(XEN label, XEN pattern)
{
  #define H_gtk_label_set_pattern "void gtk_label_set_pattern(GtkLabel* label, gchar* pattern)"
  XEN_ASSERT_TYPE(XEN_GtkLabel__P(label), label, 1, "gtk_label_set_pattern", "GtkLabel*");
  XEN_ASSERT_TYPE(XEN_gchar__P(pattern), pattern, 2, "gtk_label_set_pattern", "gchar*");
  gtk_label_set_pattern(XEN_TO_C_GtkLabel_(label), XEN_TO_C_gchar_(pattern));
  return(XEN_FALSE);
}

static XEN gxg_gtk_label_set_line_wrap(XEN label, XEN wrap)
{
  #define H_gtk_label_set_line_wrap "void gtk_label_set_line_wrap(GtkLabel* label, gboolean wrap)"
  XEN_ASSERT_TYPE(XEN_GtkLabel__P(label), label, 1, "gtk_label_set_line_wrap", "GtkLabel*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(wrap), wrap, 2, "gtk_label_set_line_wrap", "gboolean");
  gtk_label_set_line_wrap(XEN_TO_C_GtkLabel_(label), XEN_TO_C_gboolean(wrap));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_label_get_line_wrap(XEN label)
{
  #define H_gtk_label_get_line_wrap "gboolean gtk_label_get_line_wrap(GtkLabel* label)"
  XEN_ASSERT_TYPE(XEN_GtkLabel__P(label), label, 1, "gtk_label_get_line_wrap", "GtkLabel*");
  return(C_TO_XEN_gboolean(gtk_label_get_line_wrap(XEN_TO_C_GtkLabel_(label))));
}

static XEN gxg_gtk_label_set_selectable(XEN label, XEN setting)
{
  #define H_gtk_label_set_selectable "void gtk_label_set_selectable(GtkLabel* label, gboolean setting)"
  XEN_ASSERT_TYPE(XEN_GtkLabel__P(label), label, 1, "gtk_label_set_selectable", "GtkLabel*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(setting), setting, 2, "gtk_label_set_selectable", "gboolean");
  gtk_label_set_selectable(XEN_TO_C_GtkLabel_(label), XEN_TO_C_gboolean(setting));
  return(XEN_FALSE);
}

static XEN gxg_gtk_label_get_selectable(XEN label)
{
  #define H_gtk_label_get_selectable "gboolean gtk_label_get_selectable(GtkLabel* label)"
  XEN_ASSERT_TYPE(XEN_GtkLabel__P(label), label, 1, "gtk_label_get_selectable", "GtkLabel*");
  return(C_TO_XEN_gboolean(gtk_label_get_selectable(XEN_TO_C_GtkLabel_(label))));
}

static XEN gxg_gtk_label_select_region(XEN label, XEN start_offset, XEN end_offset)
{
  #define H_gtk_label_select_region "void gtk_label_select_region(GtkLabel* label, gint start_offset, \
gint end_offset)"
  XEN_ASSERT_TYPE(XEN_GtkLabel__P(label), label, 1, "gtk_label_select_region", "GtkLabel*");
  XEN_ASSERT_TYPE(XEN_gint_P(start_offset), start_offset, 2, "gtk_label_select_region", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(end_offset), end_offset, 3, "gtk_label_select_region", "gint");
  gtk_label_select_region(XEN_TO_C_GtkLabel_(label), XEN_TO_C_gint(start_offset), XEN_TO_C_gint(end_offset));
  return(XEN_FALSE);
}

static XEN gxg_gtk_label_get_selection_bounds(XEN label, XEN start, XEN end)
{
  #define H_gtk_label_get_selection_bounds "gboolean gtk_label_get_selection_bounds(GtkLabel* label, \
gint* [start], gint* [end])"
  gint ref_start;
  gint ref_end;
  XEN_ASSERT_TYPE(XEN_GtkLabel__P(label), label, 1, "gtk_label_get_selection_bounds", "GtkLabel*");
  {
    XEN result = XEN_FALSE;
    result = C_TO_XEN_gboolean(gtk_label_get_selection_bounds(XEN_TO_C_GtkLabel_(label), &ref_start, &ref_end));
    return(XEN_LIST_3(result, C_TO_XEN_gint(ref_start), C_TO_XEN_gint(ref_end)));
   }
}

static XEN gxg_gtk_label_get_layout(XEN label)
{
  #define H_gtk_label_get_layout "PangoLayout* gtk_label_get_layout(GtkLabel* label)"
  XEN_ASSERT_TYPE(XEN_GtkLabel__P(label), label, 1, "gtk_label_get_layout", "GtkLabel*");
  return(C_TO_XEN_PangoLayout_(gtk_label_get_layout(XEN_TO_C_GtkLabel_(label))));
}

static XEN gxg_gtk_label_get_layout_offsets(XEN label, XEN x, XEN y)
{
  #define H_gtk_label_get_layout_offsets "void gtk_label_get_layout_offsets(GtkLabel* label, gint* [x], \
gint* [y])"
  gint ref_x;
  gint ref_y;
  XEN_ASSERT_TYPE(XEN_GtkLabel__P(label), label, 1, "gtk_label_get_layout_offsets", "GtkLabel*");
  gtk_label_get_layout_offsets(XEN_TO_C_GtkLabel_(label), &ref_x, &ref_y);
  return(XEN_LIST_2(C_TO_XEN_gint(ref_x), C_TO_XEN_gint(ref_y)));
}
#endif

static XEN gxg_gtk_layout_get_type(void)
{
  #define H_gtk_layout_get_type "GtkType gtk_layout_get_type( void)"
  return(C_TO_XEN_GtkType(gtk_layout_get_type()));
}

static XEN gxg_gtk_layout_new(XEN hadjustment, XEN vadjustment)
{
  #define H_gtk_layout_new "GtkWidget* gtk_layout_new(GtkAdjustment* hadjustment, GtkAdjustment* vadjustment)"
  XEN_ASSERT_TYPE(XEN_GtkAdjustment__P(hadjustment), hadjustment, 1, "gtk_layout_new", "GtkAdjustment*");
  XEN_ASSERT_TYPE(XEN_GtkAdjustment__P(vadjustment), vadjustment, 2, "gtk_layout_new", "GtkAdjustment*");
  return(C_TO_XEN_GtkWidget_(gtk_layout_new(XEN_TO_C_GtkAdjustment_(hadjustment), XEN_TO_C_GtkAdjustment_(vadjustment))));
}

static XEN gxg_gtk_layout_put(XEN layout, XEN child_widget, XEN x, XEN y)
{
  #define H_gtk_layout_put "void gtk_layout_put(GtkLayout* layout, GtkWidget* child_widget, gint x, gint y)"
  XEN_ASSERT_TYPE(XEN_GtkLayout__P(layout), layout, 1, "gtk_layout_put", "GtkLayout*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(child_widget), child_widget, 2, "gtk_layout_put", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_gint_P(x), x, 3, "gtk_layout_put", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(y), y, 4, "gtk_layout_put", "gint");
  gtk_layout_put(XEN_TO_C_GtkLayout_(layout), XEN_TO_C_GtkWidget_(child_widget), XEN_TO_C_gint(x), XEN_TO_C_gint(y));
  return(XEN_FALSE);
}

static XEN gxg_gtk_layout_move(XEN layout, XEN child_widget, XEN x, XEN y)
{
  #define H_gtk_layout_move "void gtk_layout_move(GtkLayout* layout, GtkWidget* child_widget, gint x, \
gint y)"
  XEN_ASSERT_TYPE(XEN_GtkLayout__P(layout), layout, 1, "gtk_layout_move", "GtkLayout*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(child_widget), child_widget, 2, "gtk_layout_move", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_gint_P(x), x, 3, "gtk_layout_move", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(y), y, 4, "gtk_layout_move", "gint");
  gtk_layout_move(XEN_TO_C_GtkLayout_(layout), XEN_TO_C_GtkWidget_(child_widget), XEN_TO_C_gint(x), XEN_TO_C_gint(y));
  return(XEN_FALSE);
}

static XEN gxg_gtk_layout_set_size(XEN layout, XEN width, XEN height)
{
  #define H_gtk_layout_set_size "void gtk_layout_set_size(GtkLayout* layout, guint width, guint height)"
  XEN_ASSERT_TYPE(XEN_GtkLayout__P(layout), layout, 1, "gtk_layout_set_size", "GtkLayout*");
  XEN_ASSERT_TYPE(XEN_guint_P(width), width, 2, "gtk_layout_set_size", "guint");
  XEN_ASSERT_TYPE(XEN_guint_P(height), height, 3, "gtk_layout_set_size", "guint");
  gtk_layout_set_size(XEN_TO_C_GtkLayout_(layout), XEN_TO_C_guint(width), XEN_TO_C_guint(height));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_layout_get_size(XEN layout, XEN width, XEN height)
{
  #define H_gtk_layout_get_size "void gtk_layout_get_size(GtkLayout* layout, guint* [width], guint* [height])"
  guint ref_width;
  guint ref_height;
  XEN_ASSERT_TYPE(XEN_GtkLayout__P(layout), layout, 1, "gtk_layout_get_size", "GtkLayout*");
  gtk_layout_get_size(XEN_TO_C_GtkLayout_(layout), &ref_width, &ref_height);
  return(XEN_LIST_2(C_TO_XEN_guint(ref_width), C_TO_XEN_guint(ref_height)));
}
#endif

static XEN gxg_gtk_layout_get_hadjustment(XEN layout)
{
  #define H_gtk_layout_get_hadjustment "GtkAdjustment* gtk_layout_get_hadjustment(GtkLayout* layout)"
  XEN_ASSERT_TYPE(XEN_GtkLayout__P(layout), layout, 1, "gtk_layout_get_hadjustment", "GtkLayout*");
  return(C_TO_XEN_GtkAdjustment_(gtk_layout_get_hadjustment(XEN_TO_C_GtkLayout_(layout))));
}

static XEN gxg_gtk_layout_get_vadjustment(XEN layout)
{
  #define H_gtk_layout_get_vadjustment "GtkAdjustment* gtk_layout_get_vadjustment(GtkLayout* layout)"
  XEN_ASSERT_TYPE(XEN_GtkLayout__P(layout), layout, 1, "gtk_layout_get_vadjustment", "GtkLayout*");
  return(C_TO_XEN_GtkAdjustment_(gtk_layout_get_vadjustment(XEN_TO_C_GtkLayout_(layout))));
}

static XEN gxg_gtk_layout_set_hadjustment(XEN layout, XEN adjustment)
{
  #define H_gtk_layout_set_hadjustment "void gtk_layout_set_hadjustment(GtkLayout* layout, GtkAdjustment* adjustment)"
  XEN_ASSERT_TYPE(XEN_GtkLayout__P(layout), layout, 1, "gtk_layout_set_hadjustment", "GtkLayout*");
  XEN_ASSERT_TYPE(XEN_GtkAdjustment__P(adjustment), adjustment, 2, "gtk_layout_set_hadjustment", "GtkAdjustment*");
  gtk_layout_set_hadjustment(XEN_TO_C_GtkLayout_(layout), XEN_TO_C_GtkAdjustment_(adjustment));
  return(XEN_FALSE);
}

static XEN gxg_gtk_layout_set_vadjustment(XEN layout, XEN adjustment)
{
  #define H_gtk_layout_set_vadjustment "void gtk_layout_set_vadjustment(GtkLayout* layout, GtkAdjustment* adjustment)"
  XEN_ASSERT_TYPE(XEN_GtkLayout__P(layout), layout, 1, "gtk_layout_set_vadjustment", "GtkLayout*");
  XEN_ASSERT_TYPE(XEN_GtkAdjustment__P(adjustment), adjustment, 2, "gtk_layout_set_vadjustment", "GtkAdjustment*");
  gtk_layout_set_vadjustment(XEN_TO_C_GtkLayout_(layout), XEN_TO_C_GtkAdjustment_(adjustment));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_list_store_get_type(void)
{
  #define H_gtk_list_store_get_type "GtkType gtk_list_store_get_type( void)"
  return(C_TO_XEN_GtkType(gtk_list_store_get_type()));
}

static XEN gxg_gtk_list_store_new(XEN n_columns)
{
  #define H_gtk_list_store_new "GtkListStore* gtk_list_store_new(gint n_columns, ...)"
  XEN_ASSERT_TYPE(XEN_gint_P(n_columns), n_columns, 1, "gtk_list_store_new", "gint");
  return(C_TO_XEN_GtkListStore_(gtk_list_store_new(XEN_TO_C_gint(n_columns))));
}

static XEN gxg_gtk_list_store_newv(XEN n_columns, XEN types)
{
  #define H_gtk_list_store_newv "GtkListStore* gtk_list_store_newv(gint n_columns, GType* types)"
  XEN_ASSERT_TYPE(XEN_gint_P(n_columns), n_columns, 1, "gtk_list_store_newv", "gint");
  XEN_ASSERT_TYPE(XEN_GType__P(types), types, 2, "gtk_list_store_newv", "GType*");
  return(C_TO_XEN_GtkListStore_(gtk_list_store_newv(XEN_TO_C_gint(n_columns), XEN_TO_C_GType_(types))));
}

static XEN gxg_gtk_list_store_set_column_types(XEN list_store, XEN n_columns, XEN types)
{
  #define H_gtk_list_store_set_column_types "void gtk_list_store_set_column_types(GtkListStore* list_store, \
gint n_columns, GType* types)"
  XEN_ASSERT_TYPE(XEN_GtkListStore__P(list_store), list_store, 1, "gtk_list_store_set_column_types", "GtkListStore*");
  XEN_ASSERT_TYPE(XEN_gint_P(n_columns), n_columns, 2, "gtk_list_store_set_column_types", "gint");
  XEN_ASSERT_TYPE(XEN_GType__P(types), types, 3, "gtk_list_store_set_column_types", "GType*");
  gtk_list_store_set_column_types(XEN_TO_C_GtkListStore_(list_store), XEN_TO_C_gint(n_columns), XEN_TO_C_GType_(types));
  return(XEN_FALSE);
}

static XEN gxg_gtk_list_store_set_value(XEN list_store, XEN iter, XEN column, XEN value)
{
  #define H_gtk_list_store_set_value "void gtk_list_store_set_value(GtkListStore* list_store, GtkTreeIter* iter, \
gint column, GValue* value)"
  XEN_ASSERT_TYPE(XEN_GtkListStore__P(list_store), list_store, 1, "gtk_list_store_set_value", "GtkListStore*");
  XEN_ASSERT_TYPE(XEN_GtkTreeIter__P(iter), iter, 2, "gtk_list_store_set_value", "GtkTreeIter*");
  XEN_ASSERT_TYPE(XEN_gint_P(column), column, 3, "gtk_list_store_set_value", "gint");
  XEN_ASSERT_TYPE(XEN_GValue__P(value), value, 4, "gtk_list_store_set_value", "GValue*");
  gtk_list_store_set_value(XEN_TO_C_GtkListStore_(list_store), XEN_TO_C_GtkTreeIter_(iter), XEN_TO_C_gint(column), XEN_TO_C_GValue_(value));
  return(XEN_FALSE);
}

static XEN gxg_gtk_list_store_set(XEN list_store, XEN iter)
{
  #define H_gtk_list_store_set "void gtk_list_store_set(GtkListStore* list_store, GtkTreeIter* iter, \
...)"
  XEN_ASSERT_TYPE(XEN_GtkListStore__P(list_store), list_store, 1, "gtk_list_store_set", "GtkListStore*");
  XEN_ASSERT_TYPE(XEN_GtkTreeIter__P(iter), iter, 2, "gtk_list_store_set", "GtkTreeIter*");
  gtk_list_store_set(XEN_TO_C_GtkListStore_(list_store), XEN_TO_C_GtkTreeIter_(iter));
  return(XEN_FALSE);
}

static XEN gxg_gtk_list_store_remove(XEN list_store, XEN iter)
{
  #define H_gtk_list_store_remove "void gtk_list_store_remove(GtkListStore* list_store, GtkTreeIter* iter)"
  XEN_ASSERT_TYPE(XEN_GtkListStore__P(list_store), list_store, 1, "gtk_list_store_remove", "GtkListStore*");
  XEN_ASSERT_TYPE(XEN_GtkTreeIter__P(iter), iter, 2, "gtk_list_store_remove", "GtkTreeIter*");
  gtk_list_store_remove(XEN_TO_C_GtkListStore_(list_store), XEN_TO_C_GtkTreeIter_(iter));
  return(XEN_FALSE);
}

static XEN gxg_gtk_list_store_insert(XEN list_store, XEN iter, XEN position)
{
  #define H_gtk_list_store_insert "void gtk_list_store_insert(GtkListStore* list_store, GtkTreeIter* iter, \
gint position)"
  XEN_ASSERT_TYPE(XEN_GtkListStore__P(list_store), list_store, 1, "gtk_list_store_insert", "GtkListStore*");
  XEN_ASSERT_TYPE(XEN_GtkTreeIter__P(iter), iter, 2, "gtk_list_store_insert", "GtkTreeIter*");
  XEN_ASSERT_TYPE(XEN_gint_P(position), position, 3, "gtk_list_store_insert", "gint");
  gtk_list_store_insert(XEN_TO_C_GtkListStore_(list_store), XEN_TO_C_GtkTreeIter_(iter), XEN_TO_C_gint(position));
  return(XEN_FALSE);
}

static XEN gxg_gtk_list_store_insert_before(XEN list_store, XEN iter, XEN sibling)
{
  #define H_gtk_list_store_insert_before "void gtk_list_store_insert_before(GtkListStore* list_store, \
GtkTreeIter* iter, GtkTreeIter* sibling)"
  XEN_ASSERT_TYPE(XEN_GtkListStore__P(list_store), list_store, 1, "gtk_list_store_insert_before", "GtkListStore*");
  XEN_ASSERT_TYPE(XEN_GtkTreeIter__P(iter), iter, 2, "gtk_list_store_insert_before", "GtkTreeIter*");
  XEN_ASSERT_TYPE(XEN_GtkTreeIter__P(sibling) || XEN_FALSE_P(sibling), sibling, 3, "gtk_list_store_insert_before", "GtkTreeIter*");
  gtk_list_store_insert_before(XEN_TO_C_GtkListStore_(list_store), XEN_TO_C_GtkTreeIter_(iter), XEN_TO_C_GtkTreeIter_(sibling));
  return(XEN_FALSE);
}

static XEN gxg_gtk_list_store_insert_after(XEN list_store, XEN iter, XEN sibling)
{
  #define H_gtk_list_store_insert_after "void gtk_list_store_insert_after(GtkListStore* list_store, GtkTreeIter* iter, \
GtkTreeIter* sibling)"
  XEN_ASSERT_TYPE(XEN_GtkListStore__P(list_store), list_store, 1, "gtk_list_store_insert_after", "GtkListStore*");
  XEN_ASSERT_TYPE(XEN_GtkTreeIter__P(iter), iter, 2, "gtk_list_store_insert_after", "GtkTreeIter*");
  XEN_ASSERT_TYPE(XEN_GtkTreeIter__P(sibling) || XEN_FALSE_P(sibling), sibling, 3, "gtk_list_store_insert_after", "GtkTreeIter*");
  gtk_list_store_insert_after(XEN_TO_C_GtkListStore_(list_store), XEN_TO_C_GtkTreeIter_(iter), XEN_TO_C_GtkTreeIter_(sibling));
  return(XEN_FALSE);
}

static XEN gxg_gtk_list_store_prepend(XEN list_store, XEN iter)
{
  #define H_gtk_list_store_prepend "void gtk_list_store_prepend(GtkListStore* list_store, GtkTreeIter* iter)"
  XEN_ASSERT_TYPE(XEN_GtkListStore__P(list_store), list_store, 1, "gtk_list_store_prepend", "GtkListStore*");
  XEN_ASSERT_TYPE(XEN_GtkTreeIter__P(iter), iter, 2, "gtk_list_store_prepend", "GtkTreeIter*");
  gtk_list_store_prepend(XEN_TO_C_GtkListStore_(list_store), XEN_TO_C_GtkTreeIter_(iter));
  return(XEN_FALSE);
}

static XEN gxg_gtk_list_store_append(XEN list_store, XEN iter)
{
  #define H_gtk_list_store_append "void gtk_list_store_append(GtkListStore* list_store, GtkTreeIter* iter)"
  XEN_ASSERT_TYPE(XEN_GtkListStore__P(list_store), list_store, 1, "gtk_list_store_append", "GtkListStore*");
  XEN_ASSERT_TYPE(XEN_GtkTreeIter__P(iter), iter, 2, "gtk_list_store_append", "GtkTreeIter*");
  gtk_list_store_append(XEN_TO_C_GtkListStore_(list_store), XEN_TO_C_GtkTreeIter_(iter));
  return(XEN_FALSE);
}

static XEN gxg_gtk_list_store_clear(XEN list_store)
{
  #define H_gtk_list_store_clear "void gtk_list_store_clear(GtkListStore* list_store)"
  XEN_ASSERT_TYPE(XEN_GtkListStore__P(list_store), list_store, 1, "gtk_list_store_clear", "GtkListStore*");
  gtk_list_store_clear(XEN_TO_C_GtkListStore_(list_store));
  return(XEN_FALSE);
}
#endif

static XEN gxg_gtk_check_version(XEN required_major, XEN required_minor, XEN required_micro)
{
  #define H_gtk_check_version "gchar* gtk_check_version(guint required_major, guint required_minor, guint required_micro)"
  XEN_ASSERT_TYPE(XEN_guint_P(required_major), required_major, 1, "gtk_check_version", "guint");
  XEN_ASSERT_TYPE(XEN_guint_P(required_minor), required_minor, 2, "gtk_check_version", "guint");
  XEN_ASSERT_TYPE(XEN_guint_P(required_micro), required_micro, 3, "gtk_check_version", "guint");
  return(C_TO_XEN_gchar_(gtk_check_version(XEN_TO_C_guint(required_major), XEN_TO_C_guint(required_minor), XEN_TO_C_guint(required_micro))));
}

static XEN gxg_gtk_init(XEN argc, XEN argv)
{
  #define H_gtk_init "void gtk_init(int* argc, char*** argv)"
  XEN_ASSERT_TYPE(XEN_int__P(argc), argc, 1, "gtk_init", "int*");
  XEN_ASSERT_TYPE(XEN_char____P(argv), argv, 2, "gtk_init", "char***");
  gtk_init(XEN_TO_C_int_(argc), XEN_TO_C_char___(argv));
  return(XEN_FALSE);
}

static XEN gxg_gtk_init_check(XEN argc, XEN argv)
{
  #define H_gtk_init_check "gboolean gtk_init_check(int* argc, char*** argv)"
  XEN_ASSERT_TYPE(XEN_int__P(argc), argc, 1, "gtk_init_check", "int*");
  XEN_ASSERT_TYPE(XEN_char____P(argv), argv, 2, "gtk_init_check", "char***");
  return(C_TO_XEN_gboolean(gtk_init_check(XEN_TO_C_int_(argc), XEN_TO_C_char___(argv))));
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_disable_setlocale(void)
{
  #define H_gtk_disable_setlocale "void gtk_disable_setlocale( void)"
  gtk_disable_setlocale();
  return(XEN_FALSE);
}
#endif

static XEN gxg_gtk_set_locale(void)
{
  #define H_gtk_set_locale "gchar* gtk_set_locale( void)"
  return(C_TO_XEN_gchar_(gtk_set_locale()));
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_get_default_language(void)
{
  #define H_gtk_get_default_language "PangoLanguage* gtk_get_default_language( void)"
  return(C_TO_XEN_PangoLanguage_(gtk_get_default_language()));
}
#endif

static XEN gxg_gtk_events_pending(void)
{
  #define H_gtk_events_pending "gint gtk_events_pending( void)"
  return(C_TO_XEN_gint(gtk_events_pending()));
}

static XEN gxg_gtk_main_do_event(XEN event)
{
  #define H_gtk_main_do_event "void gtk_main_do_event(GdkEvent* event)"
  XEN_ASSERT_TYPE(XEN_GdkEvent__P(event), event, 1, "gtk_main_do_event", "GdkEvent*");
  gtk_main_do_event(XEN_TO_C_GdkEvent_(event));
  return(XEN_FALSE);
}

static XEN gxg_gtk_main(void)
{
  #define H_gtk_main "void gtk_main( void)"
  gtk_main();
  return(XEN_FALSE);
}

static XEN gxg_gtk_main_level(void)
{
  #define H_gtk_main_level "guint gtk_main_level( void)"
  return(C_TO_XEN_guint(gtk_main_level()));
}

static XEN gxg_gtk_main_quit(void)
{
  #define H_gtk_main_quit "void gtk_main_quit( void)"
  gtk_main_quit();
  return(XEN_FALSE);
}

static XEN gxg_gtk_main_iteration(void)
{
  #define H_gtk_main_iteration "gboolean gtk_main_iteration( void)"
  return(C_TO_XEN_gboolean(gtk_main_iteration()));
}

static XEN gxg_gtk_main_iteration_do(XEN blocking)
{
  #define H_gtk_main_iteration_do "gboolean gtk_main_iteration_do(gboolean blocking)"
  XEN_ASSERT_TYPE(XEN_gboolean_P(blocking), blocking, 1, "gtk_main_iteration_do", "gboolean");
  return(C_TO_XEN_gboolean(gtk_main_iteration_do(XEN_TO_C_gboolean(blocking))));
}

static XEN gxg_gtk_true(void)
{
  #define H_gtk_true "gboolean gtk_true( void)"
  return(C_TO_XEN_gboolean(gtk_true()));
}

static XEN gxg_gtk_false(void)
{
  #define H_gtk_false "gboolean gtk_false( void)"
  return(C_TO_XEN_gboolean(gtk_false()));
}

static XEN gxg_gtk_grab_add(XEN widget)
{
  #define H_gtk_grab_add "void gtk_grab_add(GtkWidget* widget)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_grab_add", "GtkWidget*");
  gtk_grab_add(XEN_TO_C_GtkWidget_(widget));
  return(XEN_FALSE);
}

static XEN gxg_gtk_grab_get_current(void)
{
  #define H_gtk_grab_get_current "GtkWidget* gtk_grab_get_current( void)"
  return(C_TO_XEN_GtkWidget_(gtk_grab_get_current()));
}

static XEN gxg_gtk_grab_remove(XEN widget)
{
  #define H_gtk_grab_remove "void gtk_grab_remove(GtkWidget* widget)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_grab_remove", "GtkWidget*");
  gtk_grab_remove(XEN_TO_C_GtkWidget_(widget));
  return(XEN_FALSE);
}

static XEN gxg_gtk_init_add(XEN func, XEN func_data)
{
  #define H_gtk_init_add "void gtk_init_add(GtkFunction func, lambda_data func_data)"
  XEN_ASSERT_TYPE(XEN_GtkFunction_P(func), func, 1, "gtk_init_add", "GtkFunction");
  if (XEN_NOT_BOUND_P(func_data)) func_data = XEN_FALSE; 
  else XEN_ASSERT_TYPE(XEN_lambda_data_P(func_data), func_data, 2, "gtk_init_add", "lambda_data");
  {
    int loc;
    XEN gxg_ptr = XEN_LIST_5(func, func_data, XEN_FALSE, XEN_FALSE, XEN_FALSE);
    loc = xm_protect(gxg_ptr);
    XEN_LIST_SET(gxg_ptr, 2, C_TO_XEN_INT(loc));
    gtk_init_add(XEN_TO_C_GtkFunction(func), XEN_TO_C_lambda_data(func_data));
    XEN_LIST_SET(gxg_ptr, 2, XEN_LIST_3(C_STRING_TO_XEN_SYMBOL("idler"), XEN_FALSE, C_TO_XEN_INT(loc)));
    return(XEN_FALSE);
   }
}

static XEN gxg_gtk_quit_add_destroy(XEN main_level, XEN object)
{
  #define H_gtk_quit_add_destroy "void gtk_quit_add_destroy(guint main_level, GtkObject* object)"
  XEN_ASSERT_TYPE(XEN_guint_P(main_level), main_level, 1, "gtk_quit_add_destroy", "guint");
  XEN_ASSERT_TYPE(XEN_GtkObject__P(object), object, 2, "gtk_quit_add_destroy", "GtkObject*");
  gtk_quit_add_destroy(XEN_TO_C_guint(main_level), XEN_TO_C_GtkObject_(object));
  return(XEN_FALSE);
}

static XEN gxg_gtk_quit_add(XEN main_level, XEN func, XEN func_data)
{
  #define H_gtk_quit_add "guint gtk_quit_add(guint main_level, GtkFunction func, lambda_data func_data)"
  XEN_ASSERT_TYPE(XEN_guint_P(main_level), main_level, 1, "gtk_quit_add", "guint");
  XEN_ASSERT_TYPE(XEN_GtkFunction_P(func), func, 2, "gtk_quit_add", "GtkFunction");
  if (XEN_NOT_BOUND_P(func_data)) func_data = XEN_FALSE; 
  else XEN_ASSERT_TYPE(XEN_lambda_data_P(func_data), func_data, 3, "gtk_quit_add", "lambda_data");
  {
    XEN result = XEN_FALSE;
    int loc;
    XEN gxg_ptr = XEN_LIST_5(func, func_data, XEN_FALSE, XEN_FALSE, XEN_FALSE);
    loc = xm_protect(gxg_ptr);
    XEN_LIST_SET(gxg_ptr, 2, C_TO_XEN_INT(loc));
    result = C_TO_XEN_guint(gtk_quit_add(XEN_TO_C_guint(main_level), XEN_TO_C_GtkFunction(func), XEN_TO_C_lambda_data(func_data)));
    XEN_LIST_SET(gxg_ptr, 2, XEN_LIST_3(C_STRING_TO_XEN_SYMBOL("idler"), result, C_TO_XEN_INT(loc)));
    return(result);
   }
}

static XEN gxg_gtk_quit_add_full(XEN main_level, XEN func, XEN marshal, XEN func_data, XEN destroy)
{
  #define H_gtk_quit_add_full "guint gtk_quit_add_full(guint main_level, GtkFunction func, GtkCallbackMarshal marshal, \
lambda_data func_data, GtkDestroyNotify destroy)"
  XEN_ASSERT_TYPE(XEN_guint_P(main_level), main_level, 1, "gtk_quit_add_full", "guint");
  XEN_ASSERT_TYPE(XEN_GtkFunction_P(func), func, 2, "gtk_quit_add_full", "GtkFunction");
  XEN_ASSERT_TYPE(XEN_GtkCallbackMarshal_P(marshal), marshal, 3, "gtk_quit_add_full", "GtkCallbackMarshal");
  XEN_ASSERT_TYPE(XEN_lambda_data_P(func_data), func_data, 4, "gtk_quit_add_full", "lambda_data");
  XEN_ASSERT_TYPE(XEN_GtkDestroyNotify_P(destroy), destroy, 5, "gtk_quit_add_full", "GtkDestroyNotify");
  {
    XEN result = XEN_FALSE;
    int loc;
    XEN gxg_ptr = XEN_LIST_5(func, func_data, XEN_FALSE, XEN_FALSE, XEN_FALSE);
    loc = xm_protect(gxg_ptr);
    XEN_LIST_SET(gxg_ptr, 2, C_TO_XEN_INT(loc));
    XEN_LIST_SET(gxg_ptr, 3, destroy);
    result = C_TO_XEN_guint(gtk_quit_add_full(XEN_TO_C_guint(main_level), XEN_TO_C_GtkFunction(func), XEN_TO_C_GtkCallbackMarshal(marshal), 
                                              XEN_TO_C_lambda_data(func_data), XEN_TO_C_GtkDestroyNotify(destroy)));
    XEN_LIST_SET(gxg_ptr, 2, XEN_LIST_3(C_STRING_TO_XEN_SYMBOL("idler"), result, C_TO_XEN_INT(loc)));
    return(result);
   }
}

static XEN gxg_gtk_quit_remove(XEN quit_handler_id)
{
  #define H_gtk_quit_remove "void gtk_quit_remove(guint quit_handler_id)"
  XEN_ASSERT_TYPE(XEN_guint_P(quit_handler_id), quit_handler_id, 1, "gtk_quit_remove", "guint");
  gtk_quit_remove(XEN_TO_C_guint(quit_handler_id));
  xm_unprotect_at(XEN_TO_C_INT(XEN_CADDR(quit_handler_id)));
  return(XEN_FALSE);
}

static XEN gxg_gtk_quit_remove_by_data(XEN data)
{
  #define H_gtk_quit_remove_by_data "void gtk_quit_remove_by_data(xen data)"
  XEN_ASSERT_TYPE(XEN_xen_P(data), data, 1, "gtk_quit_remove_by_data", "xen");
  gtk_quit_remove_by_data(XEN_TO_C_xen(data));
  xm_unprotect_at(XEN_TO_C_INT(XEN_CADDR(data)));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_timeout_add(XEN interval, XEN func, XEN func_data)
{
  #define H_gtk_timeout_add "guint gtk_timeout_add(guint32 interval, GtkTimeoutFunction func, lambda_data func_data)"
  XEN_ASSERT_TYPE(XEN_guint32_P(interval), interval, 1, "gtk_timeout_add", "guint32");
  XEN_ASSERT_TYPE(XEN_GtkTimeoutFunction_P(func), func, 2, "gtk_timeout_add", "GtkTimeoutFunction");
  if (XEN_NOT_BOUND_P(func_data)) func_data = XEN_FALSE; 
  else XEN_ASSERT_TYPE(XEN_lambda_data_P(func_data), func_data, 3, "gtk_timeout_add", "lambda_data");
  {
    XEN result = XEN_FALSE;
    int loc;
    XEN gxg_ptr = XEN_LIST_5(func, func_data, XEN_FALSE, XEN_FALSE, XEN_FALSE);
    loc = xm_protect(gxg_ptr);
    XEN_LIST_SET(gxg_ptr, 2, C_TO_XEN_INT(loc));
    result = C_TO_XEN_guint(gtk_timeout_add(XEN_TO_C_guint32(interval), XEN_TO_C_GtkTimeoutFunction(func), XEN_TO_C_lambda_data(func_data)));
    return(result);
   }
}

static XEN gxg_gtk_timeout_add_full(XEN interval, XEN func, XEN marshal, XEN func_data, XEN destroy)
{
  #define H_gtk_timeout_add_full "guint gtk_timeout_add_full(guint32 interval, GtkTimeoutFunction func, \
GtkCallbackMarshal marshal, lambda_data func_data, GtkDestroyNotify destroy)"
  XEN_ASSERT_TYPE(XEN_guint32_P(interval), interval, 1, "gtk_timeout_add_full", "guint32");
  XEN_ASSERT_TYPE(XEN_GtkTimeoutFunction_P(func), func, 2, "gtk_timeout_add_full", "GtkTimeoutFunction");
  XEN_ASSERT_TYPE(XEN_GtkCallbackMarshal_P(marshal), marshal, 3, "gtk_timeout_add_full", "GtkCallbackMarshal");
  XEN_ASSERT_TYPE(XEN_lambda_data_P(func_data), func_data, 4, "gtk_timeout_add_full", "lambda_data");
  XEN_ASSERT_TYPE(XEN_GtkDestroyNotify_P(destroy), destroy, 5, "gtk_timeout_add_full", "GtkDestroyNotify");
  {
    XEN result = XEN_FALSE;
    int loc;
    XEN gxg_ptr = XEN_LIST_5(func, func_data, XEN_FALSE, XEN_FALSE, XEN_FALSE);
    loc = xm_protect(gxg_ptr);
    XEN_LIST_SET(gxg_ptr, 2, C_TO_XEN_INT(loc));
    XEN_LIST_SET(gxg_ptr, 3, destroy);
    result = C_TO_XEN_guint(gtk_timeout_add_full(XEN_TO_C_guint32(interval), XEN_TO_C_GtkTimeoutFunction(func), XEN_TO_C_GtkCallbackMarshal(marshal), 
                                                 XEN_TO_C_lambda_data(func_data), XEN_TO_C_GtkDestroyNotify(destroy)));
    return(result);
   }
}
#endif

static XEN gxg_gtk_timeout_remove(XEN timeout_handler_id)
{
  #define H_gtk_timeout_remove "void gtk_timeout_remove(guint timeout_handler_id)"
  XEN_ASSERT_TYPE(XEN_guint_P(timeout_handler_id), timeout_handler_id, 1, "gtk_timeout_remove", "guint");
  gtk_timeout_remove(XEN_TO_C_guint(timeout_handler_id));
  return(XEN_FALSE);
}

static XEN gxg_gtk_idle_add(XEN func, XEN func_data)
{
  #define H_gtk_idle_add "guint gtk_idle_add(GtkFunction func, lambda_data func_data)"
  XEN_ASSERT_TYPE(XEN_GtkFunction_P(func), func, 1, "gtk_idle_add", "GtkFunction");
  if (XEN_NOT_BOUND_P(func_data)) func_data = XEN_FALSE; 
  else XEN_ASSERT_TYPE(XEN_lambda_data_P(func_data), func_data, 2, "gtk_idle_add", "lambda_data");
  {
    XEN result = XEN_FALSE;
    int loc;
    XEN gxg_ptr = XEN_LIST_5(func, func_data, XEN_FALSE, XEN_FALSE, XEN_FALSE);
    loc = xm_protect(gxg_ptr);
    XEN_LIST_SET(gxg_ptr, 2, C_TO_XEN_INT(loc));
    result = C_TO_XEN_guint(gtk_idle_add(XEN_TO_C_GtkFunction(func), XEN_TO_C_lambda_data(func_data)));
    XEN_LIST_SET(gxg_ptr, 2, XEN_LIST_3(C_STRING_TO_XEN_SYMBOL("idler"), result, C_TO_XEN_INT(loc)));
    return(result);
   }
}

static XEN gxg_gtk_idle_add_priority(XEN priority, XEN func, XEN func_data)
{
  #define H_gtk_idle_add_priority "guint gtk_idle_add_priority(gint priority, GtkFunction func, lambda_data func_data)"
  XEN_ASSERT_TYPE(XEN_gint_P(priority), priority, 1, "gtk_idle_add_priority", "gint");
  XEN_ASSERT_TYPE(XEN_GtkFunction_P(func), func, 2, "gtk_idle_add_priority", "GtkFunction");
  if (XEN_NOT_BOUND_P(func_data)) func_data = XEN_FALSE; 
  else XEN_ASSERT_TYPE(XEN_lambda_data_P(func_data), func_data, 3, "gtk_idle_add_priority", "lambda_data");
  {
    XEN result = XEN_FALSE;
    int loc;
    XEN gxg_ptr = XEN_LIST_5(func, func_data, XEN_FALSE, XEN_FALSE, XEN_FALSE);
    loc = xm_protect(gxg_ptr);
    XEN_LIST_SET(gxg_ptr, 2, C_TO_XEN_INT(loc));
    result = C_TO_XEN_guint(gtk_idle_add_priority(XEN_TO_C_gint(priority), XEN_TO_C_GtkFunction(func), XEN_TO_C_lambda_data(func_data)));
    XEN_LIST_SET(gxg_ptr, 2, XEN_LIST_3(C_STRING_TO_XEN_SYMBOL("idler"), result, C_TO_XEN_INT(loc)));
    return(result);
   }
}

static XEN gxg_gtk_idle_add_full(XEN priority, XEN func, XEN marshal, XEN func_data, XEN destroy)
{
  #define H_gtk_idle_add_full "guint gtk_idle_add_full(gint priority, GtkFunction func, GtkCallbackMarshal marshal, \
lambda_data func_data, GtkDestroyNotify destroy)"
  XEN_ASSERT_TYPE(XEN_gint_P(priority), priority, 1, "gtk_idle_add_full", "gint");
  XEN_ASSERT_TYPE(XEN_GtkFunction_P(func), func, 2, "gtk_idle_add_full", "GtkFunction");
  XEN_ASSERT_TYPE(XEN_GtkCallbackMarshal_P(marshal), marshal, 3, "gtk_idle_add_full", "GtkCallbackMarshal");
  XEN_ASSERT_TYPE(XEN_lambda_data_P(func_data), func_data, 4, "gtk_idle_add_full", "lambda_data");
  XEN_ASSERT_TYPE(XEN_GtkDestroyNotify_P(destroy), destroy, 5, "gtk_idle_add_full", "GtkDestroyNotify");
  {
    XEN result = XEN_FALSE;
    int loc;
    XEN gxg_ptr = XEN_LIST_5(func, func_data, XEN_FALSE, XEN_FALSE, XEN_FALSE);
    loc = xm_protect(gxg_ptr);
    XEN_LIST_SET(gxg_ptr, 2, C_TO_XEN_INT(loc));
    XEN_LIST_SET(gxg_ptr, 3, destroy);
    result = C_TO_XEN_guint(gtk_idle_add_full(XEN_TO_C_gint(priority), XEN_TO_C_GtkFunction(func), XEN_TO_C_GtkCallbackMarshal(marshal), 
                                              XEN_TO_C_lambda_data(func_data), XEN_TO_C_GtkDestroyNotify(destroy)));
    XEN_LIST_SET(gxg_ptr, 2, XEN_LIST_3(C_STRING_TO_XEN_SYMBOL("idler"), result, C_TO_XEN_INT(loc)));
    return(result);
   }
}

static XEN gxg_gtk_idle_remove(XEN idle_handler_id)
{
  #define H_gtk_idle_remove "void gtk_idle_remove(guint idle_handler_id)"
  XEN_ASSERT_TYPE(XEN_guint_P(idle_handler_id), idle_handler_id, 1, "gtk_idle_remove", "guint");
  gtk_idle_remove(XEN_TO_C_guint(idle_handler_id));
  xm_unprotect_idler(XEN_TO_C_guint(idle_handler_id));
  return(XEN_FALSE);
}

static XEN gxg_gtk_idle_remove_by_data(XEN data)
{
  #define H_gtk_idle_remove_by_data "void gtk_idle_remove_by_data(xen data)"
  XEN_ASSERT_TYPE(XEN_xen_P(data), data, 1, "gtk_idle_remove_by_data", "xen");
  gtk_idle_remove_by_data(XEN_TO_C_xen(data));
  xm_unprotect_at(XEN_TO_C_INT(XEN_CADDR(data)));
  return(XEN_FALSE);
}

static XEN gxg_gtk_input_add_full(XEN source, XEN condition, XEN func, XEN marshal, XEN func_data, XEN destroy)
{
  #define H_gtk_input_add_full "guint gtk_input_add_full(gint source, GdkInputCondition condition, GdkInputFunction func, \
GtkCallbackMarshal marshal, lambda_data func_data, GtkDestroyNotify destroy)"
  XEN_ASSERT_TYPE(XEN_gint_P(source), source, 1, "gtk_input_add_full", "gint");
  XEN_ASSERT_TYPE(XEN_GdkInputCondition_P(condition), condition, 2, "gtk_input_add_full", "GdkInputCondition");
  XEN_ASSERT_TYPE(XEN_GdkInputFunction_P(func), func, 3, "gtk_input_add_full", "GdkInputFunction");
  XEN_ASSERT_TYPE(XEN_GtkCallbackMarshal_P(marshal), marshal, 4, "gtk_input_add_full", "GtkCallbackMarshal");
  XEN_ASSERT_TYPE(XEN_lambda_data_P(func_data), func_data, 5, "gtk_input_add_full", "lambda_data");
  XEN_ASSERT_TYPE(XEN_GtkDestroyNotify_P(destroy), destroy, 6, "gtk_input_add_full", "GtkDestroyNotify");
  {
    XEN result = XEN_FALSE;
    int loc;
    XEN gxg_ptr = XEN_LIST_5(func, func_data, XEN_FALSE, XEN_FALSE, XEN_FALSE);
    loc = xm_protect(gxg_ptr);
    XEN_LIST_SET(gxg_ptr, 2, C_TO_XEN_INT(loc));
    XEN_LIST_SET(gxg_ptr, 3, destroy);
    result = C_TO_XEN_guint(gtk_input_add_full(XEN_TO_C_gint(source), XEN_TO_C_GdkInputCondition(condition), XEN_TO_C_GdkInputFunction(func), 
                                               XEN_TO_C_GtkCallbackMarshal(marshal), XEN_TO_C_lambda_data(func_data), XEN_TO_C_GtkDestroyNotify(destroy)));
    XEN_LIST_SET(gxg_ptr, 2, XEN_LIST_3(C_STRING_TO_XEN_SYMBOL("idler"), result, C_TO_XEN_INT(loc)));
    return(result);
   }
}

static XEN gxg_gtk_input_remove(XEN input_handler_id)
{
  #define H_gtk_input_remove "void gtk_input_remove(guint input_handler_id)"
  XEN_ASSERT_TYPE(XEN_guint_P(input_handler_id), input_handler_id, 1, "gtk_input_remove", "guint");
  gtk_input_remove(XEN_TO_C_guint(input_handler_id));
  xm_unprotect_at(XEN_TO_C_INT(XEN_CADDR(input_handler_id)));
  return(XEN_FALSE);
}

static XEN gxg_gtk_key_snooper_install(XEN func, XEN func_data)
{
  #define H_gtk_key_snooper_install "guint gtk_key_snooper_install(GtkKeySnoopFunc func, lambda_data func_data)"
  XEN_ASSERT_TYPE(XEN_GtkKeySnoopFunc_P(func), func, 1, "gtk_key_snooper_install", "GtkKeySnoopFunc");
  if (XEN_NOT_BOUND_P(func_data)) func_data = XEN_FALSE; 
  else XEN_ASSERT_TYPE(XEN_lambda_data_P(func_data), func_data, 2, "gtk_key_snooper_install", "lambda_data");
  {
    XEN result = XEN_FALSE;
    int loc;
    XEN gxg_ptr = XEN_LIST_5(func, func_data, XEN_FALSE, XEN_FALSE, XEN_FALSE);
    loc = xm_protect(gxg_ptr);
    XEN_LIST_SET(gxg_ptr, 2, C_TO_XEN_INT(loc));
    result = C_TO_XEN_guint(gtk_key_snooper_install(XEN_TO_C_GtkKeySnoopFunc(func), XEN_TO_C_lambda_data(func_data)));
    XEN_LIST_SET(gxg_ptr, 2, XEN_LIST_3(C_STRING_TO_XEN_SYMBOL("idler"), result, C_TO_XEN_INT(loc)));
    return(result);
   }
}

static XEN gxg_gtk_key_snooper_remove(XEN snooper_handler_id)
{
  #define H_gtk_key_snooper_remove "void gtk_key_snooper_remove(guint snooper_handler_id)"
  XEN_ASSERT_TYPE(XEN_guint_P(snooper_handler_id), snooper_handler_id, 1, "gtk_key_snooper_remove", "guint");
  gtk_key_snooper_remove(XEN_TO_C_guint(snooper_handler_id));
  xm_unprotect_at(XEN_TO_C_INT(XEN_CADDR(snooper_handler_id)));
  return(XEN_FALSE);
}

static XEN gxg_gtk_get_current_event(void)
{
  #define H_gtk_get_current_event "GdkEvent* gtk_get_current_event( void)"
  return(C_TO_XEN_GdkEvent_(gtk_get_current_event()));
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_get_current_event_time(void)
{
  #define H_gtk_get_current_event_time "guint32 gtk_get_current_event_time( void)"
  return(C_TO_XEN_guint32(gtk_get_current_event_time()));
}

static XEN gxg_gtk_get_current_event_state(XEN state)
{
  #define H_gtk_get_current_event_state "gboolean gtk_get_current_event_state(GdkModifierType* [state])"
  GdkModifierType ref_state;
  {
    XEN result = XEN_FALSE;
    result = C_TO_XEN_gboolean(gtk_get_current_event_state(&ref_state));
    return(XEN_LIST_2(result, C_TO_XEN_GdkModifierType(ref_state)));
   }
}
#endif

static XEN gxg_gtk_get_event_widget(XEN event)
{
  #define H_gtk_get_event_widget "GtkWidget* gtk_get_event_widget(GdkEvent* event)"
  XEN_ASSERT_TYPE(XEN_GdkEvent__P(event) || XEN_FALSE_P(event), event, 1, "gtk_get_event_widget", "GdkEvent*");
  return(C_TO_XEN_GtkWidget_(gtk_get_event_widget(XEN_TO_C_GdkEvent_(event))));
}

static XEN gxg_gtk_propagate_event(XEN widget, XEN event)
{
  #define H_gtk_propagate_event "void gtk_propagate_event(GtkWidget* widget, GdkEvent* event)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_propagate_event", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_GdkEvent__P(event), event, 2, "gtk_propagate_event", "GdkEvent*");
  gtk_propagate_event(XEN_TO_C_GtkWidget_(widget), XEN_TO_C_GdkEvent_(event));
  return(XEN_FALSE);
}

static XEN gxg_gtk_menu_bar_get_type(void)
{
  #define H_gtk_menu_bar_get_type "GtkType gtk_menu_bar_get_type( void)"
  return(C_TO_XEN_GtkType(gtk_menu_bar_get_type()));
}

static XEN gxg_gtk_menu_bar_new(void)
{
  #define H_gtk_menu_bar_new "GtkWidget* gtk_menu_bar_new( void)"
  return(C_TO_XEN_GtkWidget_(gtk_menu_bar_new()));
}

static XEN gxg_gtk_menu_get_type(void)
{
  #define H_gtk_menu_get_type "GtkType gtk_menu_get_type( void)"
  return(C_TO_XEN_GtkType(gtk_menu_get_type()));
}

static XEN gxg_gtk_menu_new(void)
{
  #define H_gtk_menu_new "GtkWidget* gtk_menu_new( void)"
  return(C_TO_XEN_GtkWidget_(gtk_menu_new()));
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_menu_popup(XEN menu, XEN parent_menu_shell, XEN parent_menu_item, XEN func, XEN func_data, XEN button, XEN activate_time)
{
  #define H_gtk_menu_popup "void gtk_menu_popup(GtkMenu* menu, GtkWidget* parent_menu_shell, GtkWidget* parent_menu_item, \
GtkMenuPositionFunc func, lambda_data func_data, guint button, guint32 activate_time)"
  XEN_ASSERT_TYPE(XEN_GtkMenu__P(menu), menu, 1, "gtk_menu_popup", "GtkMenu*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(parent_menu_shell), parent_menu_shell, 2, "gtk_menu_popup", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(parent_menu_item), parent_menu_item, 3, "gtk_menu_popup", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_GtkMenuPositionFunc_P(func), func, 4, "gtk_menu_popup", "GtkMenuPositionFunc");
  XEN_ASSERT_TYPE(XEN_lambda_data_P(func_data), func_data, 5, "gtk_menu_popup", "lambda_data");
  XEN_ASSERT_TYPE(XEN_guint_P(button), button, 6, "gtk_menu_popup", "guint");
  XEN_ASSERT_TYPE(XEN_guint32_P(activate_time), activate_time, 7, "gtk_menu_popup", "guint32");
  {
    XEN gxg_ptr = XEN_LIST_5(func, func_data, XEN_FALSE, XEN_FALSE, XEN_FALSE);
    xm_protect(gxg_ptr);
    gtk_menu_popup(XEN_TO_C_GtkMenu_(menu), XEN_TO_C_GtkWidget_(parent_menu_shell), XEN_TO_C_GtkWidget_(parent_menu_item), XEN_TO_C_GtkMenuPositionFunc(func), 
               XEN_TO_C_lambda_data(func_data), XEN_TO_C_guint(button), XEN_TO_C_guint32(activate_time));
    return(XEN_FALSE);
   }
}
#endif

static XEN gxg_gtk_menu_reposition(XEN menu)
{
  #define H_gtk_menu_reposition "void gtk_menu_reposition(GtkMenu* menu)"
  XEN_ASSERT_TYPE(XEN_GtkMenu__P(menu), menu, 1, "gtk_menu_reposition", "GtkMenu*");
  gtk_menu_reposition(XEN_TO_C_GtkMenu_(menu));
  return(XEN_FALSE);
}

static XEN gxg_gtk_menu_popdown(XEN menu)
{
  #define H_gtk_menu_popdown "void gtk_menu_popdown(GtkMenu* menu)"
  XEN_ASSERT_TYPE(XEN_GtkMenu__P(menu), menu, 1, "gtk_menu_popdown", "GtkMenu*");
  gtk_menu_popdown(XEN_TO_C_GtkMenu_(menu));
  return(XEN_FALSE);
}

static XEN gxg_gtk_menu_get_active(XEN menu)
{
  #define H_gtk_menu_get_active "GtkWidget* gtk_menu_get_active(GtkMenu* menu)"
  XEN_ASSERT_TYPE(XEN_GtkMenu__P(menu), menu, 1, "gtk_menu_get_active", "GtkMenu*");
  return(C_TO_XEN_GtkWidget_(gtk_menu_get_active(XEN_TO_C_GtkMenu_(menu))));
}

static XEN gxg_gtk_menu_set_active(XEN menu, XEN index)
{
  #define H_gtk_menu_set_active "void gtk_menu_set_active(GtkMenu* menu, guint index)"
  XEN_ASSERT_TYPE(XEN_GtkMenu__P(menu), menu, 1, "gtk_menu_set_active", "GtkMenu*");
  XEN_ASSERT_TYPE(XEN_guint_P(index), index, 2, "gtk_menu_set_active", "guint");
  gtk_menu_set_active(XEN_TO_C_GtkMenu_(menu), XEN_TO_C_guint(index));
  return(XEN_FALSE);
}

static XEN gxg_gtk_menu_set_accel_group(XEN menu, XEN accel_group)
{
  #define H_gtk_menu_set_accel_group "void gtk_menu_set_accel_group(GtkMenu* menu, GtkAccelGroup* accel_group)"
  XEN_ASSERT_TYPE(XEN_GtkMenu__P(menu), menu, 1, "gtk_menu_set_accel_group", "GtkMenu*");
  XEN_ASSERT_TYPE(XEN_GtkAccelGroup__P(accel_group), accel_group, 2, "gtk_menu_set_accel_group", "GtkAccelGroup*");
  gtk_menu_set_accel_group(XEN_TO_C_GtkMenu_(menu), XEN_TO_C_GtkAccelGroup_(accel_group));
  return(XEN_FALSE);
}

static XEN gxg_gtk_menu_get_accel_group(XEN menu)
{
  #define H_gtk_menu_get_accel_group "GtkAccelGroup* gtk_menu_get_accel_group(GtkMenu* menu)"
  XEN_ASSERT_TYPE(XEN_GtkMenu__P(menu), menu, 1, "gtk_menu_get_accel_group", "GtkMenu*");
  return(C_TO_XEN_GtkAccelGroup_(gtk_menu_get_accel_group(XEN_TO_C_GtkMenu_(menu))));
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_menu_set_accel_path(XEN menu, XEN accel_path)
{
  #define H_gtk_menu_set_accel_path "void gtk_menu_set_accel_path(GtkMenu* menu, gchar* accel_path)"
  XEN_ASSERT_TYPE(XEN_GtkMenu__P(menu), menu, 1, "gtk_menu_set_accel_path", "GtkMenu*");
  XEN_ASSERT_TYPE(XEN_gchar__P(accel_path), accel_path, 2, "gtk_menu_set_accel_path", "gchar*");
  gtk_menu_set_accel_path(XEN_TO_C_GtkMenu_(menu), XEN_TO_C_gchar_(accel_path));
  return(XEN_FALSE);
}
#endif

static XEN gxg_gtk_menu_attach_to_widget(XEN menu, XEN attach_widget, XEN detacher)
{
  #define H_gtk_menu_attach_to_widget "void gtk_menu_attach_to_widget(GtkMenu* menu, GtkWidget* attach_widget, \
GtkMenuDetachFunc detacher)"
  XEN_ASSERT_TYPE(XEN_GtkMenu__P(menu), menu, 1, "gtk_menu_attach_to_widget", "GtkMenu*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(attach_widget), attach_widget, 2, "gtk_menu_attach_to_widget", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_GtkMenuDetachFunc_P(detacher), detacher, 3, "gtk_menu_attach_to_widget", "GtkMenuDetachFunc");
  gtk_menu_attach_to_widget(XEN_TO_C_GtkMenu_(menu), XEN_TO_C_GtkWidget_(attach_widget), XEN_TO_C_GtkMenuDetachFunc(detacher));
  return(XEN_FALSE);
}

static XEN gxg_gtk_menu_detach(XEN menu)
{
  #define H_gtk_menu_detach "void gtk_menu_detach(GtkMenu* menu)"
  XEN_ASSERT_TYPE(XEN_GtkMenu__P(menu), menu, 1, "gtk_menu_detach", "GtkMenu*");
  gtk_menu_detach(XEN_TO_C_GtkMenu_(menu));
  return(XEN_FALSE);
}

static XEN gxg_gtk_menu_get_attach_widget(XEN menu)
{
  #define H_gtk_menu_get_attach_widget "GtkWidget* gtk_menu_get_attach_widget(GtkMenu* menu)"
  XEN_ASSERT_TYPE(XEN_GtkMenu__P(menu), menu, 1, "gtk_menu_get_attach_widget", "GtkMenu*");
  return(C_TO_XEN_GtkWidget_(gtk_menu_get_attach_widget(XEN_TO_C_GtkMenu_(menu))));
}

static XEN gxg_gtk_menu_set_tearoff_state(XEN menu, XEN torn_off)
{
  #define H_gtk_menu_set_tearoff_state "void gtk_menu_set_tearoff_state(GtkMenu* menu, gboolean torn_off)"
  XEN_ASSERT_TYPE(XEN_GtkMenu__P(menu), menu, 1, "gtk_menu_set_tearoff_state", "GtkMenu*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(torn_off), torn_off, 2, "gtk_menu_set_tearoff_state", "gboolean");
  gtk_menu_set_tearoff_state(XEN_TO_C_GtkMenu_(menu), XEN_TO_C_gboolean(torn_off));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_menu_get_tearoff_state(XEN menu)
{
  #define H_gtk_menu_get_tearoff_state "gboolean gtk_menu_get_tearoff_state(GtkMenu* menu)"
  XEN_ASSERT_TYPE(XEN_GtkMenu__P(menu), menu, 1, "gtk_menu_get_tearoff_state", "GtkMenu*");
  return(C_TO_XEN_gboolean(gtk_menu_get_tearoff_state(XEN_TO_C_GtkMenu_(menu))));
}
#endif

static XEN gxg_gtk_menu_set_title(XEN menu, XEN title)
{
  #define H_gtk_menu_set_title "void gtk_menu_set_title(GtkMenu* menu, gchar* title)"
  XEN_ASSERT_TYPE(XEN_GtkMenu__P(menu), menu, 1, "gtk_menu_set_title", "GtkMenu*");
  XEN_ASSERT_TYPE(XEN_gchar__P(title), title, 2, "gtk_menu_set_title", "gchar*");
  gtk_menu_set_title(XEN_TO_C_GtkMenu_(menu), XEN_TO_C_gchar_(title));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_menu_get_title(XEN menu)
{
  #define H_gtk_menu_get_title "gchar* gtk_menu_get_title(GtkMenu* menu)"
  XEN_ASSERT_TYPE(XEN_GtkMenu__P(menu), menu, 1, "gtk_menu_get_title", "GtkMenu*");
  return(C_TO_XEN_gchar_(gtk_menu_get_title(XEN_TO_C_GtkMenu_(menu))));
}
#endif

static XEN gxg_gtk_menu_reorder_child(XEN menu, XEN child, XEN position)
{
  #define H_gtk_menu_reorder_child "void gtk_menu_reorder_child(GtkMenu* menu, GtkWidget* child, gint position)"
  XEN_ASSERT_TYPE(XEN_GtkMenu__P(menu), menu, 1, "gtk_menu_reorder_child", "GtkMenu*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(child), child, 2, "gtk_menu_reorder_child", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_gint_P(position), position, 3, "gtk_menu_reorder_child", "gint");
  gtk_menu_reorder_child(XEN_TO_C_GtkMenu_(menu), XEN_TO_C_GtkWidget_(child), XEN_TO_C_gint(position));
  return(XEN_FALSE);
}

static XEN gxg_gtk_menu_item_get_type(void)
{
  #define H_gtk_menu_item_get_type "GtkType gtk_menu_item_get_type( void)"
  return(C_TO_XEN_GtkType(gtk_menu_item_get_type()));
}

static XEN gxg_gtk_menu_item_new(void)
{
  #define H_gtk_menu_item_new "GtkWidget* gtk_menu_item_new( void)"
  return(C_TO_XEN_GtkWidget_(gtk_menu_item_new()));
}

static XEN gxg_gtk_menu_item_new_with_label(XEN label)
{
  #define H_gtk_menu_item_new_with_label "GtkWidget* gtk_menu_item_new_with_label(gchar* label)"
  XEN_ASSERT_TYPE(XEN_gchar__P(label), label, 1, "gtk_menu_item_new_with_label", "gchar*");
  return(C_TO_XEN_GtkWidget_(gtk_menu_item_new_with_label(XEN_TO_C_gchar_(label))));
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_menu_item_new_with_mnemonic(XEN label)
{
  #define H_gtk_menu_item_new_with_mnemonic "GtkWidget* gtk_menu_item_new_with_mnemonic(gchar* label)"
  XEN_ASSERT_TYPE(XEN_gchar__P(label), label, 1, "gtk_menu_item_new_with_mnemonic", "gchar*");
  return(C_TO_XEN_GtkWidget_(gtk_menu_item_new_with_mnemonic(XEN_TO_C_gchar_(label))));
}
#endif

static XEN gxg_gtk_menu_item_set_submenu(XEN menu_item, XEN submenu)
{
  #define H_gtk_menu_item_set_submenu "void gtk_menu_item_set_submenu(GtkMenuItem* menu_item, GtkWidget* submenu)"
  XEN_ASSERT_TYPE(XEN_GtkMenuItem__P(menu_item), menu_item, 1, "gtk_menu_item_set_submenu", "GtkMenuItem*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(submenu), submenu, 2, "gtk_menu_item_set_submenu", "GtkWidget*");
  gtk_menu_item_set_submenu(XEN_TO_C_GtkMenuItem_(menu_item), XEN_TO_C_GtkWidget_(submenu));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_menu_item_get_submenu(XEN menu_item)
{
  #define H_gtk_menu_item_get_submenu "GtkWidget* gtk_menu_item_get_submenu(GtkMenuItem* menu_item)"
  XEN_ASSERT_TYPE(XEN_GtkMenuItem__P(menu_item), menu_item, 1, "gtk_menu_item_get_submenu", "GtkMenuItem*");
  return(C_TO_XEN_GtkWidget_(gtk_menu_item_get_submenu(XEN_TO_C_GtkMenuItem_(menu_item))));
}
#endif

static XEN gxg_gtk_menu_item_remove_submenu(XEN menu_item)
{
  #define H_gtk_menu_item_remove_submenu "void gtk_menu_item_remove_submenu(GtkMenuItem* menu_item)"
  XEN_ASSERT_TYPE(XEN_GtkMenuItem__P(menu_item), menu_item, 1, "gtk_menu_item_remove_submenu", "GtkMenuItem*");
  gtk_menu_item_remove_submenu(XEN_TO_C_GtkMenuItem_(menu_item));
  return(XEN_FALSE);
}

static XEN gxg_gtk_menu_item_select(XEN menu_item)
{
  #define H_gtk_menu_item_select "void gtk_menu_item_select(GtkMenuItem* menu_item)"
  XEN_ASSERT_TYPE(XEN_GtkMenuItem__P(menu_item), menu_item, 1, "gtk_menu_item_select", "GtkMenuItem*");
  gtk_menu_item_select(XEN_TO_C_GtkMenuItem_(menu_item));
  return(XEN_FALSE);
}

static XEN gxg_gtk_menu_item_deselect(XEN menu_item)
{
  #define H_gtk_menu_item_deselect "void gtk_menu_item_deselect(GtkMenuItem* menu_item)"
  XEN_ASSERT_TYPE(XEN_GtkMenuItem__P(menu_item), menu_item, 1, "gtk_menu_item_deselect", "GtkMenuItem*");
  gtk_menu_item_deselect(XEN_TO_C_GtkMenuItem_(menu_item));
  return(XEN_FALSE);
}

static XEN gxg_gtk_menu_item_activate(XEN menu_item)
{
  #define H_gtk_menu_item_activate "void gtk_menu_item_activate(GtkMenuItem* menu_item)"
  XEN_ASSERT_TYPE(XEN_GtkMenuItem__P(menu_item), menu_item, 1, "gtk_menu_item_activate", "GtkMenuItem*");
  gtk_menu_item_activate(XEN_TO_C_GtkMenuItem_(menu_item));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_menu_item_toggle_size_request(XEN menu_item, XEN requisition)
{
  #define H_gtk_menu_item_toggle_size_request "void gtk_menu_item_toggle_size_request(GtkMenuItem* menu_item, \
gint* requisition)"
  XEN_ASSERT_TYPE(XEN_GtkMenuItem__P(menu_item), menu_item, 1, "gtk_menu_item_toggle_size_request", "GtkMenuItem*");
  XEN_ASSERT_TYPE(XEN_gint__P(requisition), requisition, 2, "gtk_menu_item_toggle_size_request", "gint*");
  gtk_menu_item_toggle_size_request(XEN_TO_C_GtkMenuItem_(menu_item), XEN_TO_C_gint_(requisition));
  return(XEN_FALSE);
}

static XEN gxg_gtk_menu_item_toggle_size_allocate(XEN menu_item, XEN allocation)
{
  #define H_gtk_menu_item_toggle_size_allocate "void gtk_menu_item_toggle_size_allocate(GtkMenuItem* menu_item, \
gint allocation)"
  XEN_ASSERT_TYPE(XEN_GtkMenuItem__P(menu_item), menu_item, 1, "gtk_menu_item_toggle_size_allocate", "GtkMenuItem*");
  XEN_ASSERT_TYPE(XEN_gint_P(allocation), allocation, 2, "gtk_menu_item_toggle_size_allocate", "gint");
  gtk_menu_item_toggle_size_allocate(XEN_TO_C_GtkMenuItem_(menu_item), XEN_TO_C_gint(allocation));
  return(XEN_FALSE);
}

static XEN gxg_gtk_menu_item_set_right_justified(XEN menu_item, XEN right_justified)
{
  #define H_gtk_menu_item_set_right_justified "void gtk_menu_item_set_right_justified(GtkMenuItem* menu_item, \
gboolean right_justified)"
  XEN_ASSERT_TYPE(XEN_GtkMenuItem__P(menu_item), menu_item, 1, "gtk_menu_item_set_right_justified", "GtkMenuItem*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(right_justified), right_justified, 2, "gtk_menu_item_set_right_justified", "gboolean");
  gtk_menu_item_set_right_justified(XEN_TO_C_GtkMenuItem_(menu_item), XEN_TO_C_gboolean(right_justified));
  return(XEN_FALSE);
}

static XEN gxg_gtk_menu_item_get_right_justified(XEN menu_item)
{
  #define H_gtk_menu_item_get_right_justified "gboolean gtk_menu_item_get_right_justified(GtkMenuItem* menu_item)"
  XEN_ASSERT_TYPE(XEN_GtkMenuItem__P(menu_item), menu_item, 1, "gtk_menu_item_get_right_justified", "GtkMenuItem*");
  return(C_TO_XEN_gboolean(gtk_menu_item_get_right_justified(XEN_TO_C_GtkMenuItem_(menu_item))));
}

static XEN gxg_gtk_menu_item_set_accel_path(XEN menu_item, XEN accel_path)
{
  #define H_gtk_menu_item_set_accel_path "void gtk_menu_item_set_accel_path(GtkMenuItem* menu_item, gchar* accel_path)"
  XEN_ASSERT_TYPE(XEN_GtkMenuItem__P(menu_item), menu_item, 1, "gtk_menu_item_set_accel_path", "GtkMenuItem*");
  XEN_ASSERT_TYPE(XEN_gchar__P(accel_path), accel_path, 2, "gtk_menu_item_set_accel_path", "gchar*");
  gtk_menu_item_set_accel_path(XEN_TO_C_GtkMenuItem_(menu_item), XEN_TO_C_gchar_(accel_path));
  return(XEN_FALSE);
}
#endif

static XEN gxg_gtk_menu_shell_get_type(void)
{
  #define H_gtk_menu_shell_get_type "GtkType gtk_menu_shell_get_type( void)"
  return(C_TO_XEN_GtkType(gtk_menu_shell_get_type()));
}

static XEN gxg_gtk_menu_shell_append(XEN menu_shell, XEN child)
{
  #define H_gtk_menu_shell_append "void gtk_menu_shell_append(GtkMenuShell* menu_shell, GtkWidget* child)"
  XEN_ASSERT_TYPE(XEN_GtkMenuShell__P(menu_shell), menu_shell, 1, "gtk_menu_shell_append", "GtkMenuShell*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(child), child, 2, "gtk_menu_shell_append", "GtkWidget*");
  gtk_menu_shell_append(XEN_TO_C_GtkMenuShell_(menu_shell), XEN_TO_C_GtkWidget_(child));
  return(XEN_FALSE);
}

static XEN gxg_gtk_menu_shell_prepend(XEN menu_shell, XEN child)
{
  #define H_gtk_menu_shell_prepend "void gtk_menu_shell_prepend(GtkMenuShell* menu_shell, GtkWidget* child)"
  XEN_ASSERT_TYPE(XEN_GtkMenuShell__P(menu_shell), menu_shell, 1, "gtk_menu_shell_prepend", "GtkMenuShell*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(child), child, 2, "gtk_menu_shell_prepend", "GtkWidget*");
  gtk_menu_shell_prepend(XEN_TO_C_GtkMenuShell_(menu_shell), XEN_TO_C_GtkWidget_(child));
  return(XEN_FALSE);
}

static XEN gxg_gtk_menu_shell_insert(XEN menu_shell, XEN child, XEN position)
{
  #define H_gtk_menu_shell_insert "void gtk_menu_shell_insert(GtkMenuShell* menu_shell, GtkWidget* child, \
gint position)"
  XEN_ASSERT_TYPE(XEN_GtkMenuShell__P(menu_shell), menu_shell, 1, "gtk_menu_shell_insert", "GtkMenuShell*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(child), child, 2, "gtk_menu_shell_insert", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_gint_P(position), position, 3, "gtk_menu_shell_insert", "gint");
  gtk_menu_shell_insert(XEN_TO_C_GtkMenuShell_(menu_shell), XEN_TO_C_GtkWidget_(child), XEN_TO_C_gint(position));
  return(XEN_FALSE);
}

static XEN gxg_gtk_menu_shell_deactivate(XEN menu_shell)
{
  #define H_gtk_menu_shell_deactivate "void gtk_menu_shell_deactivate(GtkMenuShell* menu_shell)"
  XEN_ASSERT_TYPE(XEN_GtkMenuShell__P(menu_shell), menu_shell, 1, "gtk_menu_shell_deactivate", "GtkMenuShell*");
  gtk_menu_shell_deactivate(XEN_TO_C_GtkMenuShell_(menu_shell));
  return(XEN_FALSE);
}

static XEN gxg_gtk_menu_shell_select_item(XEN menu_shell, XEN menu_item)
{
  #define H_gtk_menu_shell_select_item "void gtk_menu_shell_select_item(GtkMenuShell* menu_shell, GtkWidget* menu_item)"
  XEN_ASSERT_TYPE(XEN_GtkMenuShell__P(menu_shell), menu_shell, 1, "gtk_menu_shell_select_item", "GtkMenuShell*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(menu_item), menu_item, 2, "gtk_menu_shell_select_item", "GtkWidget*");
  gtk_menu_shell_select_item(XEN_TO_C_GtkMenuShell_(menu_shell), XEN_TO_C_GtkWidget_(menu_item));
  return(XEN_FALSE);
}

static XEN gxg_gtk_menu_shell_deselect(XEN menu_shell)
{
  #define H_gtk_menu_shell_deselect "void gtk_menu_shell_deselect(GtkMenuShell* menu_shell)"
  XEN_ASSERT_TYPE(XEN_GtkMenuShell__P(menu_shell), menu_shell, 1, "gtk_menu_shell_deselect", "GtkMenuShell*");
  gtk_menu_shell_deselect(XEN_TO_C_GtkMenuShell_(menu_shell));
  return(XEN_FALSE);
}

static XEN gxg_gtk_menu_shell_activate_item(XEN menu_shell, XEN menu_item, XEN force_deactivate)
{
  #define H_gtk_menu_shell_activate_item "void gtk_menu_shell_activate_item(GtkMenuShell* menu_shell, \
GtkWidget* menu_item, gboolean force_deactivate)"
  XEN_ASSERT_TYPE(XEN_GtkMenuShell__P(menu_shell), menu_shell, 1, "gtk_menu_shell_activate_item", "GtkMenuShell*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(menu_item), menu_item, 2, "gtk_menu_shell_activate_item", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(force_deactivate), force_deactivate, 3, "gtk_menu_shell_activate_item", "gboolean");
  gtk_menu_shell_activate_item(XEN_TO_C_GtkMenuShell_(menu_shell), XEN_TO_C_GtkWidget_(menu_item), XEN_TO_C_gboolean(force_deactivate));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_message_dialog_get_type(void)
{
  #define H_gtk_message_dialog_get_type "GtkType gtk_message_dialog_get_type( void)"
  return(C_TO_XEN_GtkType(gtk_message_dialog_get_type()));
}

static XEN gxg_gtk_message_dialog_new(XEN parent, XEN flags, XEN type, XEN buttons, XEN message_format)
{
  #define H_gtk_message_dialog_new "GtkWidget* gtk_message_dialog_new(GtkWindow* parent, GtkDialogFlags flags, \
GtkMessageType type, GtkButtonsType buttons, gchar* message_format, ...)"
  XEN_ASSERT_TYPE(XEN_GtkWindow__P(parent), parent, 1, "gtk_message_dialog_new", "GtkWindow*");
  XEN_ASSERT_TYPE(XEN_GtkDialogFlags_P(flags), flags, 2, "gtk_message_dialog_new", "GtkDialogFlags");
  XEN_ASSERT_TYPE(XEN_GtkMessageType_P(type), type, 3, "gtk_message_dialog_new", "GtkMessageType");
  XEN_ASSERT_TYPE(XEN_GtkButtonsType_P(buttons), buttons, 4, "gtk_message_dialog_new", "GtkButtonsType");
  XEN_ASSERT_TYPE(XEN_gchar__P(message_format), message_format, 5, "gtk_message_dialog_new", "gchar*");
  return(C_TO_XEN_GtkWidget_(gtk_message_dialog_new(XEN_TO_C_GtkWindow_(parent), XEN_TO_C_GtkDialogFlags(flags), XEN_TO_C_GtkMessageType(type), 
                                                    XEN_TO_C_GtkButtonsType(buttons), XEN_TO_C_gchar_(message_format))));
}
#endif

static XEN gxg_gtk_misc_get_type(void)
{
  #define H_gtk_misc_get_type "GtkType gtk_misc_get_type( void)"
  return(C_TO_XEN_GtkType(gtk_misc_get_type()));
}

static XEN gxg_gtk_misc_set_alignment(XEN misc, XEN xalign, XEN yalign)
{
  #define H_gtk_misc_set_alignment "void gtk_misc_set_alignment(GtkMisc* misc, gfloat xalign, gfloat yalign)"
  XEN_ASSERT_TYPE(XEN_GtkMisc__P(misc), misc, 1, "gtk_misc_set_alignment", "GtkMisc*");
  XEN_ASSERT_TYPE(XEN_gfloat_P(xalign), xalign, 2, "gtk_misc_set_alignment", "gfloat");
  XEN_ASSERT_TYPE(XEN_gfloat_P(yalign), yalign, 3, "gtk_misc_set_alignment", "gfloat");
  gtk_misc_set_alignment(XEN_TO_C_GtkMisc_(misc), XEN_TO_C_gfloat(xalign), XEN_TO_C_gfloat(yalign));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_misc_get_alignment(XEN misc, XEN xalign, XEN yalign)
{
  #define H_gtk_misc_get_alignment "void gtk_misc_get_alignment(GtkMisc* misc, gfloat* [xalign], gfloat* [yalign])"
  gfloat ref_xalign;
  gfloat ref_yalign;
  XEN_ASSERT_TYPE(XEN_GtkMisc__P(misc), misc, 1, "gtk_misc_get_alignment", "GtkMisc*");
  gtk_misc_get_alignment(XEN_TO_C_GtkMisc_(misc), &ref_xalign, &ref_yalign);
  return(XEN_LIST_2(C_TO_XEN_gfloat(ref_xalign), C_TO_XEN_gfloat(ref_yalign)));
}
#endif

static XEN gxg_gtk_misc_set_padding(XEN misc, XEN xpad, XEN ypad)
{
  #define H_gtk_misc_set_padding "void gtk_misc_set_padding(GtkMisc* misc, gint xpad, gint ypad)"
  XEN_ASSERT_TYPE(XEN_GtkMisc__P(misc), misc, 1, "gtk_misc_set_padding", "GtkMisc*");
  XEN_ASSERT_TYPE(XEN_gint_P(xpad), xpad, 2, "gtk_misc_set_padding", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(ypad), ypad, 3, "gtk_misc_set_padding", "gint");
  gtk_misc_set_padding(XEN_TO_C_GtkMisc_(misc), XEN_TO_C_gint(xpad), XEN_TO_C_gint(ypad));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_misc_get_padding(XEN misc, XEN xpad, XEN ypad)
{
  #define H_gtk_misc_get_padding "void gtk_misc_get_padding(GtkMisc* misc, gint* xpad, gint* ypad)"
  XEN_ASSERT_TYPE(XEN_GtkMisc__P(misc), misc, 1, "gtk_misc_get_padding", "GtkMisc*");
  XEN_ASSERT_TYPE(XEN_gint__P(xpad), xpad, 2, "gtk_misc_get_padding", "gint*");
  XEN_ASSERT_TYPE(XEN_gint__P(ypad), ypad, 3, "gtk_misc_get_padding", "gint*");
  gtk_misc_get_padding(XEN_TO_C_GtkMisc_(misc), XEN_TO_C_gint_(xpad), XEN_TO_C_gint_(ypad));
  return(XEN_FALSE);
}
#endif

static XEN gxg_gtk_notebook_get_type(void)
{
  #define H_gtk_notebook_get_type "GtkType gtk_notebook_get_type( void)"
  return(C_TO_XEN_GtkType(gtk_notebook_get_type()));
}

static XEN gxg_gtk_notebook_new(void)
{
  #define H_gtk_notebook_new "GtkWidget* gtk_notebook_new( void)"
  return(C_TO_XEN_GtkWidget_(gtk_notebook_new()));
}

static XEN gxg_gtk_notebook_append_page(XEN notebook, XEN child, XEN tab_label)
{
  #define H_gtk_notebook_append_page "void gtk_notebook_append_page(GtkNotebook* notebook, GtkWidget* child, \
GtkWidget* tab_label)"
  XEN_ASSERT_TYPE(XEN_GtkNotebook__P(notebook), notebook, 1, "gtk_notebook_append_page", "GtkNotebook*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(child), child, 2, "gtk_notebook_append_page", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(tab_label) || XEN_FALSE_P(tab_label), tab_label, 3, "gtk_notebook_append_page", "GtkWidget*");
  gtk_notebook_append_page(XEN_TO_C_GtkNotebook_(notebook), XEN_TO_C_GtkWidget_(child), XEN_TO_C_GtkWidget_(tab_label));
  return(XEN_FALSE);
}

static XEN gxg_gtk_notebook_append_page_menu(XEN notebook, XEN child, XEN tab_label, XEN menu_label)
{
  #define H_gtk_notebook_append_page_menu "void gtk_notebook_append_page_menu(GtkNotebook* notebook, \
GtkWidget* child, GtkWidget* tab_label, GtkWidget* menu_label)"
  XEN_ASSERT_TYPE(XEN_GtkNotebook__P(notebook), notebook, 1, "gtk_notebook_append_page_menu", "GtkNotebook*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(child), child, 2, "gtk_notebook_append_page_menu", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(tab_label) || XEN_FALSE_P(tab_label), tab_label, 3, "gtk_notebook_append_page_menu", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(menu_label) || XEN_FALSE_P(menu_label), menu_label, 4, "gtk_notebook_append_page_menu", "GtkWidget*");
  gtk_notebook_append_page_menu(XEN_TO_C_GtkNotebook_(notebook), XEN_TO_C_GtkWidget_(child), XEN_TO_C_GtkWidget_(tab_label), 
                                XEN_TO_C_GtkWidget_(menu_label));
  return(XEN_FALSE);
}

static XEN gxg_gtk_notebook_prepend_page(XEN notebook, XEN child, XEN tab_label)
{
  #define H_gtk_notebook_prepend_page "void gtk_notebook_prepend_page(GtkNotebook* notebook, GtkWidget* child, \
GtkWidget* tab_label)"
  XEN_ASSERT_TYPE(XEN_GtkNotebook__P(notebook), notebook, 1, "gtk_notebook_prepend_page", "GtkNotebook*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(child), child, 2, "gtk_notebook_prepend_page", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(tab_label) || XEN_FALSE_P(tab_label), tab_label, 3, "gtk_notebook_prepend_page", "GtkWidget*");
  gtk_notebook_prepend_page(XEN_TO_C_GtkNotebook_(notebook), XEN_TO_C_GtkWidget_(child), XEN_TO_C_GtkWidget_(tab_label));
  return(XEN_FALSE);
}

static XEN gxg_gtk_notebook_prepend_page_menu(XEN notebook, XEN child, XEN tab_label, XEN menu_label)
{
  #define H_gtk_notebook_prepend_page_menu "void gtk_notebook_prepend_page_menu(GtkNotebook* notebook, \
GtkWidget* child, GtkWidget* tab_label, GtkWidget* menu_label)"
  XEN_ASSERT_TYPE(XEN_GtkNotebook__P(notebook), notebook, 1, "gtk_notebook_prepend_page_menu", "GtkNotebook*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(child), child, 2, "gtk_notebook_prepend_page_menu", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(tab_label) || XEN_FALSE_P(tab_label), tab_label, 3, "gtk_notebook_prepend_page_menu", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(menu_label) || XEN_FALSE_P(menu_label), menu_label, 4, "gtk_notebook_prepend_page_menu", "GtkWidget*");
  gtk_notebook_prepend_page_menu(XEN_TO_C_GtkNotebook_(notebook), XEN_TO_C_GtkWidget_(child), XEN_TO_C_GtkWidget_(tab_label), 
                                 XEN_TO_C_GtkWidget_(menu_label));
  return(XEN_FALSE);
}

static XEN gxg_gtk_notebook_insert_page(XEN notebook, XEN child, XEN tab_label, XEN position)
{
  #define H_gtk_notebook_insert_page "void gtk_notebook_insert_page(GtkNotebook* notebook, GtkWidget* child, \
GtkWidget* tab_label, gint position)"
  XEN_ASSERT_TYPE(XEN_GtkNotebook__P(notebook), notebook, 1, "gtk_notebook_insert_page", "GtkNotebook*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(child), child, 2, "gtk_notebook_insert_page", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(tab_label) || XEN_FALSE_P(tab_label), tab_label, 3, "gtk_notebook_insert_page", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_gint_P(position), position, 4, "gtk_notebook_insert_page", "gint");
  gtk_notebook_insert_page(XEN_TO_C_GtkNotebook_(notebook), XEN_TO_C_GtkWidget_(child), XEN_TO_C_GtkWidget_(tab_label), XEN_TO_C_gint(position));
  return(XEN_FALSE);
}

static XEN gxg_gtk_notebook_insert_page_menu(XEN notebook, XEN child, XEN tab_label, XEN menu_label, XEN position)
{
  #define H_gtk_notebook_insert_page_menu "void gtk_notebook_insert_page_menu(GtkNotebook* notebook, \
GtkWidget* child, GtkWidget* tab_label, GtkWidget* menu_label, gint position)"
  XEN_ASSERT_TYPE(XEN_GtkNotebook__P(notebook), notebook, 1, "gtk_notebook_insert_page_menu", "GtkNotebook*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(child), child, 2, "gtk_notebook_insert_page_menu", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(tab_label) || XEN_FALSE_P(tab_label), tab_label, 3, "gtk_notebook_insert_page_menu", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(menu_label) || XEN_FALSE_P(menu_label), menu_label, 4, "gtk_notebook_insert_page_menu", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_gint_P(position), position, 5, "gtk_notebook_insert_page_menu", "gint");
  gtk_notebook_insert_page_menu(XEN_TO_C_GtkNotebook_(notebook), XEN_TO_C_GtkWidget_(child), XEN_TO_C_GtkWidget_(tab_label), 
                                XEN_TO_C_GtkWidget_(menu_label), XEN_TO_C_gint(position));
  return(XEN_FALSE);
}

static XEN gxg_gtk_notebook_remove_page(XEN notebook, XEN page_num)
{
  #define H_gtk_notebook_remove_page "void gtk_notebook_remove_page(GtkNotebook* notebook, gint page_num)"
  XEN_ASSERT_TYPE(XEN_GtkNotebook__P(notebook), notebook, 1, "gtk_notebook_remove_page", "GtkNotebook*");
  XEN_ASSERT_TYPE(XEN_gint_P(page_num), page_num, 2, "gtk_notebook_remove_page", "gint");
  gtk_notebook_remove_page(XEN_TO_C_GtkNotebook_(notebook), XEN_TO_C_gint(page_num));
  return(XEN_FALSE);
}

static XEN gxg_gtk_notebook_get_current_page(XEN notebook)
{
  #define H_gtk_notebook_get_current_page "gint gtk_notebook_get_current_page(GtkNotebook* notebook)"
  XEN_ASSERT_TYPE(XEN_GtkNotebook__P(notebook), notebook, 1, "gtk_notebook_get_current_page", "GtkNotebook*");
  return(C_TO_XEN_gint(gtk_notebook_get_current_page(XEN_TO_C_GtkNotebook_(notebook))));
}

static XEN gxg_gtk_notebook_get_nth_page(XEN notebook, XEN page_num)
{
  #define H_gtk_notebook_get_nth_page "GtkWidget* gtk_notebook_get_nth_page(GtkNotebook* notebook, gint page_num)"
  XEN_ASSERT_TYPE(XEN_GtkNotebook__P(notebook), notebook, 1, "gtk_notebook_get_nth_page", "GtkNotebook*");
  XEN_ASSERT_TYPE(XEN_gint_P(page_num), page_num, 2, "gtk_notebook_get_nth_page", "gint");
  return(C_TO_XEN_GtkWidget_(gtk_notebook_get_nth_page(XEN_TO_C_GtkNotebook_(notebook), XEN_TO_C_gint(page_num))));
}

static XEN gxg_gtk_notebook_page_num(XEN notebook, XEN child)
{
  #define H_gtk_notebook_page_num "gint gtk_notebook_page_num(GtkNotebook* notebook, GtkWidget* child)"
  XEN_ASSERT_TYPE(XEN_GtkNotebook__P(notebook), notebook, 1, "gtk_notebook_page_num", "GtkNotebook*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(child), child, 2, "gtk_notebook_page_num", "GtkWidget*");
  return(C_TO_XEN_gint(gtk_notebook_page_num(XEN_TO_C_GtkNotebook_(notebook), XEN_TO_C_GtkWidget_(child))));
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_notebook_set_current_page(XEN notebook, XEN page_num)
{
  #define H_gtk_notebook_set_current_page "void gtk_notebook_set_current_page(GtkNotebook* notebook, \
gint page_num)"
  XEN_ASSERT_TYPE(XEN_GtkNotebook__P(notebook), notebook, 1, "gtk_notebook_set_current_page", "GtkNotebook*");
  XEN_ASSERT_TYPE(XEN_gint_P(page_num), page_num, 2, "gtk_notebook_set_current_page", "gint");
  gtk_notebook_set_current_page(XEN_TO_C_GtkNotebook_(notebook), XEN_TO_C_gint(page_num));
  return(XEN_FALSE);
}
#endif

static XEN gxg_gtk_notebook_next_page(XEN notebook)
{
  #define H_gtk_notebook_next_page "void gtk_notebook_next_page(GtkNotebook* notebook)"
  XEN_ASSERT_TYPE(XEN_GtkNotebook__P(notebook), notebook, 1, "gtk_notebook_next_page", "GtkNotebook*");
  gtk_notebook_next_page(XEN_TO_C_GtkNotebook_(notebook));
  return(XEN_FALSE);
}

static XEN gxg_gtk_notebook_prev_page(XEN notebook)
{
  #define H_gtk_notebook_prev_page "void gtk_notebook_prev_page(GtkNotebook* notebook)"
  XEN_ASSERT_TYPE(XEN_GtkNotebook__P(notebook), notebook, 1, "gtk_notebook_prev_page", "GtkNotebook*");
  gtk_notebook_prev_page(XEN_TO_C_GtkNotebook_(notebook));
  return(XEN_FALSE);
}

static XEN gxg_gtk_notebook_set_show_border(XEN notebook, XEN show_border)
{
  #define H_gtk_notebook_set_show_border "void gtk_notebook_set_show_border(GtkNotebook* notebook, gboolean show_border)"
  XEN_ASSERT_TYPE(XEN_GtkNotebook__P(notebook), notebook, 1, "gtk_notebook_set_show_border", "GtkNotebook*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(show_border), show_border, 2, "gtk_notebook_set_show_border", "gboolean");
  gtk_notebook_set_show_border(XEN_TO_C_GtkNotebook_(notebook), XEN_TO_C_gboolean(show_border));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_notebook_get_show_border(XEN notebook)
{
  #define H_gtk_notebook_get_show_border "gboolean gtk_notebook_get_show_border(GtkNotebook* notebook)"
  XEN_ASSERT_TYPE(XEN_GtkNotebook__P(notebook), notebook, 1, "gtk_notebook_get_show_border", "GtkNotebook*");
  return(C_TO_XEN_gboolean(gtk_notebook_get_show_border(XEN_TO_C_GtkNotebook_(notebook))));
}
#endif

static XEN gxg_gtk_notebook_set_show_tabs(XEN notebook, XEN show_tabs)
{
  #define H_gtk_notebook_set_show_tabs "void gtk_notebook_set_show_tabs(GtkNotebook* notebook, gboolean show_tabs)"
  XEN_ASSERT_TYPE(XEN_GtkNotebook__P(notebook), notebook, 1, "gtk_notebook_set_show_tabs", "GtkNotebook*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(show_tabs), show_tabs, 2, "gtk_notebook_set_show_tabs", "gboolean");
  gtk_notebook_set_show_tabs(XEN_TO_C_GtkNotebook_(notebook), XEN_TO_C_gboolean(show_tabs));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_notebook_get_show_tabs(XEN notebook)
{
  #define H_gtk_notebook_get_show_tabs "gboolean gtk_notebook_get_show_tabs(GtkNotebook* notebook)"
  XEN_ASSERT_TYPE(XEN_GtkNotebook__P(notebook), notebook, 1, "gtk_notebook_get_show_tabs", "GtkNotebook*");
  return(C_TO_XEN_gboolean(gtk_notebook_get_show_tabs(XEN_TO_C_GtkNotebook_(notebook))));
}
#endif

static XEN gxg_gtk_notebook_set_tab_pos(XEN notebook, XEN pos)
{
  #define H_gtk_notebook_set_tab_pos "void gtk_notebook_set_tab_pos(GtkNotebook* notebook, GtkPositionType pos)"
  XEN_ASSERT_TYPE(XEN_GtkNotebook__P(notebook), notebook, 1, "gtk_notebook_set_tab_pos", "GtkNotebook*");
  XEN_ASSERT_TYPE(XEN_GtkPositionType_P(pos), pos, 2, "gtk_notebook_set_tab_pos", "GtkPositionType");
  gtk_notebook_set_tab_pos(XEN_TO_C_GtkNotebook_(notebook), XEN_TO_C_GtkPositionType(pos));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_notebook_get_tab_pos(XEN notebook)
{
  #define H_gtk_notebook_get_tab_pos "GtkPositionType gtk_notebook_get_tab_pos(GtkNotebook* notebook)"
  XEN_ASSERT_TYPE(XEN_GtkNotebook__P(notebook), notebook, 1, "gtk_notebook_get_tab_pos", "GtkNotebook*");
  return(C_TO_XEN_GtkPositionType(gtk_notebook_get_tab_pos(XEN_TO_C_GtkNotebook_(notebook))));
}
#endif

static XEN gxg_gtk_notebook_set_scrollable(XEN notebook, XEN scrollable)
{
  #define H_gtk_notebook_set_scrollable "void gtk_notebook_set_scrollable(GtkNotebook* notebook, gboolean scrollable)"
  XEN_ASSERT_TYPE(XEN_GtkNotebook__P(notebook), notebook, 1, "gtk_notebook_set_scrollable", "GtkNotebook*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(scrollable), scrollable, 2, "gtk_notebook_set_scrollable", "gboolean");
  gtk_notebook_set_scrollable(XEN_TO_C_GtkNotebook_(notebook), XEN_TO_C_gboolean(scrollable));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_notebook_get_scrollable(XEN notebook)
{
  #define H_gtk_notebook_get_scrollable "gboolean gtk_notebook_get_scrollable(GtkNotebook* notebook)"
  XEN_ASSERT_TYPE(XEN_GtkNotebook__P(notebook), notebook, 1, "gtk_notebook_get_scrollable", "GtkNotebook*");
  return(C_TO_XEN_gboolean(gtk_notebook_get_scrollable(XEN_TO_C_GtkNotebook_(notebook))));
}
#endif

static XEN gxg_gtk_notebook_popup_enable(XEN notebook)
{
  #define H_gtk_notebook_popup_enable "void gtk_notebook_popup_enable(GtkNotebook* notebook)"
  XEN_ASSERT_TYPE(XEN_GtkNotebook__P(notebook), notebook, 1, "gtk_notebook_popup_enable", "GtkNotebook*");
  gtk_notebook_popup_enable(XEN_TO_C_GtkNotebook_(notebook));
  return(XEN_FALSE);
}

static XEN gxg_gtk_notebook_popup_disable(XEN notebook)
{
  #define H_gtk_notebook_popup_disable "void gtk_notebook_popup_disable(GtkNotebook* notebook)"
  XEN_ASSERT_TYPE(XEN_GtkNotebook__P(notebook), notebook, 1, "gtk_notebook_popup_disable", "GtkNotebook*");
  gtk_notebook_popup_disable(XEN_TO_C_GtkNotebook_(notebook));
  return(XEN_FALSE);
}

static XEN gxg_gtk_notebook_get_tab_label(XEN notebook, XEN child)
{
  #define H_gtk_notebook_get_tab_label "GtkWidget* gtk_notebook_get_tab_label(GtkNotebook* notebook, \
GtkWidget* child)"
  XEN_ASSERT_TYPE(XEN_GtkNotebook__P(notebook), notebook, 1, "gtk_notebook_get_tab_label", "GtkNotebook*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(child), child, 2, "gtk_notebook_get_tab_label", "GtkWidget*");
  return(C_TO_XEN_GtkWidget_(gtk_notebook_get_tab_label(XEN_TO_C_GtkNotebook_(notebook), XEN_TO_C_GtkWidget_(child))));
}

static XEN gxg_gtk_notebook_set_tab_label(XEN notebook, XEN child, XEN tab_label)
{
  #define H_gtk_notebook_set_tab_label "void gtk_notebook_set_tab_label(GtkNotebook* notebook, GtkWidget* child, \
GtkWidget* tab_label)"
  XEN_ASSERT_TYPE(XEN_GtkNotebook__P(notebook), notebook, 1, "gtk_notebook_set_tab_label", "GtkNotebook*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(child), child, 2, "gtk_notebook_set_tab_label", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(tab_label) || XEN_FALSE_P(tab_label), tab_label, 3, "gtk_notebook_set_tab_label", "GtkWidget*");
  gtk_notebook_set_tab_label(XEN_TO_C_GtkNotebook_(notebook), XEN_TO_C_GtkWidget_(child), XEN_TO_C_GtkWidget_(tab_label));
  return(XEN_FALSE);
}

static XEN gxg_gtk_notebook_set_tab_label_text(XEN notebook, XEN child, XEN tab_text)
{
  #define H_gtk_notebook_set_tab_label_text "void gtk_notebook_set_tab_label_text(GtkNotebook* notebook, \
GtkWidget* child, gchar* tab_text)"
  XEN_ASSERT_TYPE(XEN_GtkNotebook__P(notebook), notebook, 1, "gtk_notebook_set_tab_label_text", "GtkNotebook*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(child), child, 2, "gtk_notebook_set_tab_label_text", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_gchar__P(tab_text), tab_text, 3, "gtk_notebook_set_tab_label_text", "gchar*");
  gtk_notebook_set_tab_label_text(XEN_TO_C_GtkNotebook_(notebook), XEN_TO_C_GtkWidget_(child), XEN_TO_C_gchar_(tab_text));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_notebook_get_tab_label_text(XEN notebook, XEN child)
{
  #define H_gtk_notebook_get_tab_label_text "gchar* gtk_notebook_get_tab_label_text(GtkNotebook* notebook, \
GtkWidget* child)"
  XEN_ASSERT_TYPE(XEN_GtkNotebook__P(notebook), notebook, 1, "gtk_notebook_get_tab_label_text", "GtkNotebook*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(child), child, 2, "gtk_notebook_get_tab_label_text", "GtkWidget*");
  return(C_TO_XEN_gchar_(gtk_notebook_get_tab_label_text(XEN_TO_C_GtkNotebook_(notebook), XEN_TO_C_GtkWidget_(child))));
}
#endif

static XEN gxg_gtk_notebook_get_menu_label(XEN notebook, XEN child)
{
  #define H_gtk_notebook_get_menu_label "GtkWidget* gtk_notebook_get_menu_label(GtkNotebook* notebook, \
GtkWidget* child)"
  XEN_ASSERT_TYPE(XEN_GtkNotebook__P(notebook), notebook, 1, "gtk_notebook_get_menu_label", "GtkNotebook*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(child), child, 2, "gtk_notebook_get_menu_label", "GtkWidget*");
  return(C_TO_XEN_GtkWidget_(gtk_notebook_get_menu_label(XEN_TO_C_GtkNotebook_(notebook), XEN_TO_C_GtkWidget_(child))));
}

static XEN gxg_gtk_notebook_set_menu_label(XEN notebook, XEN child, XEN menu_label)
{
  #define H_gtk_notebook_set_menu_label "void gtk_notebook_set_menu_label(GtkNotebook* notebook, GtkWidget* child, \
GtkWidget* menu_label)"
  XEN_ASSERT_TYPE(XEN_GtkNotebook__P(notebook), notebook, 1, "gtk_notebook_set_menu_label", "GtkNotebook*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(child), child, 2, "gtk_notebook_set_menu_label", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(menu_label), menu_label, 3, "gtk_notebook_set_menu_label", "GtkWidget*");
  gtk_notebook_set_menu_label(XEN_TO_C_GtkNotebook_(notebook), XEN_TO_C_GtkWidget_(child), XEN_TO_C_GtkWidget_(menu_label));
  return(XEN_FALSE);
}

static XEN gxg_gtk_notebook_set_menu_label_text(XEN notebook, XEN child, XEN menu_text)
{
  #define H_gtk_notebook_set_menu_label_text "void gtk_notebook_set_menu_label_text(GtkNotebook* notebook, \
GtkWidget* child, gchar* menu_text)"
  XEN_ASSERT_TYPE(XEN_GtkNotebook__P(notebook), notebook, 1, "gtk_notebook_set_menu_label_text", "GtkNotebook*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(child), child, 2, "gtk_notebook_set_menu_label_text", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_gchar__P(menu_text), menu_text, 3, "gtk_notebook_set_menu_label_text", "gchar*");
  gtk_notebook_set_menu_label_text(XEN_TO_C_GtkNotebook_(notebook), XEN_TO_C_GtkWidget_(child), XEN_TO_C_gchar_(menu_text));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_notebook_get_menu_label_text(XEN notebook, XEN child)
{
  #define H_gtk_notebook_get_menu_label_text "gchar* gtk_notebook_get_menu_label_text(GtkNotebook* notebook, \
GtkWidget* child)"
  XEN_ASSERT_TYPE(XEN_GtkNotebook__P(notebook), notebook, 1, "gtk_notebook_get_menu_label_text", "GtkNotebook*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(child), child, 2, "gtk_notebook_get_menu_label_text", "GtkWidget*");
  return(C_TO_XEN_gchar_(gtk_notebook_get_menu_label_text(XEN_TO_C_GtkNotebook_(notebook), XEN_TO_C_GtkWidget_(child))));
}
#endif

static XEN gxg_gtk_notebook_query_tab_label_packing(XEN notebook, XEN child, XEN expand, XEN fill, XEN pack_type)
{
  #define H_gtk_notebook_query_tab_label_packing "void gtk_notebook_query_tab_label_packing(GtkNotebook* notebook, \
GtkWidget* child, gboolean* expand, gboolean* fill, GtkPackType* pack_type)"
  XEN_ASSERT_TYPE(XEN_GtkNotebook__P(notebook), notebook, 1, "gtk_notebook_query_tab_label_packing", "GtkNotebook*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(child), child, 2, "gtk_notebook_query_tab_label_packing", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_gboolean__P(expand), expand, 3, "gtk_notebook_query_tab_label_packing", "gboolean*");
  XEN_ASSERT_TYPE(XEN_gboolean__P(fill), fill, 4, "gtk_notebook_query_tab_label_packing", "gboolean*");
  XEN_ASSERT_TYPE(XEN_GtkPackType__P(pack_type), pack_type, 5, "gtk_notebook_query_tab_label_packing", "GtkPackType*");
  gtk_notebook_query_tab_label_packing(XEN_TO_C_GtkNotebook_(notebook), XEN_TO_C_GtkWidget_(child), XEN_TO_C_gboolean_(expand), 
                                       XEN_TO_C_gboolean_(fill), XEN_TO_C_GtkPackType_(pack_type));
  return(XEN_FALSE);
}

static XEN gxg_gtk_notebook_set_tab_label_packing(XEN notebook, XEN child, XEN expand, XEN fill, XEN pack_type)
{
  #define H_gtk_notebook_set_tab_label_packing "void gtk_notebook_set_tab_label_packing(GtkNotebook* notebook, \
GtkWidget* child, gboolean expand, gboolean fill, GtkPackType pack_type)"
  XEN_ASSERT_TYPE(XEN_GtkNotebook__P(notebook), notebook, 1, "gtk_notebook_set_tab_label_packing", "GtkNotebook*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(child), child, 2, "gtk_notebook_set_tab_label_packing", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(expand), expand, 3, "gtk_notebook_set_tab_label_packing", "gboolean");
  XEN_ASSERT_TYPE(XEN_gboolean_P(fill), fill, 4, "gtk_notebook_set_tab_label_packing", "gboolean");
  XEN_ASSERT_TYPE(XEN_GtkPackType_P(pack_type), pack_type, 5, "gtk_notebook_set_tab_label_packing", "GtkPackType");
  gtk_notebook_set_tab_label_packing(XEN_TO_C_GtkNotebook_(notebook), XEN_TO_C_GtkWidget_(child), XEN_TO_C_gboolean(expand), 
                                     XEN_TO_C_gboolean(fill), XEN_TO_C_GtkPackType(pack_type));
  return(XEN_FALSE);
}

static XEN gxg_gtk_notebook_reorder_child(XEN notebook, XEN child, XEN position)
{
  #define H_gtk_notebook_reorder_child "void gtk_notebook_reorder_child(GtkNotebook* notebook, GtkWidget* child, \
gint position)"
  XEN_ASSERT_TYPE(XEN_GtkNotebook__P(notebook), notebook, 1, "gtk_notebook_reorder_child", "GtkNotebook*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(child), child, 2, "gtk_notebook_reorder_child", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_gint_P(position), position, 3, "gtk_notebook_reorder_child", "gint");
  gtk_notebook_reorder_child(XEN_TO_C_GtkNotebook_(notebook), XEN_TO_C_GtkWidget_(child), XEN_TO_C_gint(position));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_GTK_OBJECT_TYPE_NAME(XEN object)
{
  #define H_GTK_OBJECT_TYPE_NAME "char* GTK_OBJECT_TYPE_NAME(GtkObject* object)"
  XEN_ASSERT_TYPE(XEN_GtkObject__P(object), object, 1, "GTK_OBJECT_TYPE_NAME", "GtkObject*");
  return(C_TO_XEN_char_(GTK_OBJECT_TYPE_NAME(XEN_TO_C_GtkObject_(object))));
}
#endif

static XEN gxg_GTK_OBJECT_FLAGS(XEN obj)
{
  #define H_GTK_OBJECT_FLAGS "int GTK_OBJECT_FLAGS(GtkObject* obj)"
  XEN_ASSERT_TYPE(XEN_GtkObject__P(obj), obj, 1, "GTK_OBJECT_FLAGS", "GtkObject*");
  return(C_TO_XEN_int(GTK_OBJECT_FLAGS(XEN_TO_C_GtkObject_(obj))));
}

static XEN gxg_GTK_OBJECT_FLOATING(XEN obj)
{
  #define H_GTK_OBJECT_FLOATING "gboolean GTK_OBJECT_FLOATING(GtkObject* obj)"
  XEN_ASSERT_TYPE(XEN_GtkObject__P(obj), obj, 1, "GTK_OBJECT_FLOATING", "GtkObject*");
  return(C_TO_XEN_gboolean(GTK_OBJECT_FLOATING(XEN_TO_C_GtkObject_(obj))));
}

static XEN gxg_GTK_OBJECT_SET_FLAGS(XEN obj, XEN flag)
{
  #define H_GTK_OBJECT_SET_FLAGS "void GTK_OBJECT_SET_FLAGS(GtkObject* obj, int flag)"
  XEN_ASSERT_TYPE(XEN_GtkObject__P(obj), obj, 1, "GTK_OBJECT_SET_FLAGS", "GtkObject*");
  XEN_ASSERT_TYPE(XEN_int_P(flag), flag, 2, "GTK_OBJECT_SET_FLAGS", "int");
  GTK_OBJECT_SET_FLAGS(XEN_TO_C_GtkObject_(obj), XEN_TO_C_int(flag));
  return(XEN_FALSE);
}

static XEN gxg_GTK_OBJECT_UNSET_FLAGS(XEN obj, XEN flag)
{
  #define H_GTK_OBJECT_UNSET_FLAGS "void GTK_OBJECT_UNSET_FLAGS(GtkObject* obj, int flag)"
  XEN_ASSERT_TYPE(XEN_GtkObject__P(obj), obj, 1, "GTK_OBJECT_UNSET_FLAGS", "GtkObject*");
  XEN_ASSERT_TYPE(XEN_int_P(flag), flag, 2, "GTK_OBJECT_UNSET_FLAGS", "int");
  GTK_OBJECT_UNSET_FLAGS(XEN_TO_C_GtkObject_(obj), XEN_TO_C_int(flag));
  return(XEN_FALSE);
}

static XEN gxg_gtk_object_get_type(void)
{
  #define H_gtk_object_get_type "GtkType gtk_object_get_type( void)"
  return(C_TO_XEN_GtkType(gtk_object_get_type()));
}

static XEN gxg_gtk_object_new(XEN type, XEN first_property_name)
{
  #define H_gtk_object_new "GtkObject* gtk_object_new(GtkType type, gchar* first_property_name, ...)"
  XEN_ASSERT_TYPE(XEN_GtkType_P(type), type, 1, "gtk_object_new", "GtkType");
  XEN_ASSERT_TYPE(XEN_gchar__P(first_property_name), first_property_name, 2, "gtk_object_new", "gchar*");
  return(C_TO_XEN_GtkObject_(gtk_object_new(XEN_TO_C_GtkType(type), XEN_TO_C_gchar_(first_property_name))));
}

static XEN gxg_gtk_object_sink(XEN object)
{
  #define H_gtk_object_sink "void gtk_object_sink(GtkObject* object)"
  XEN_ASSERT_TYPE(XEN_GtkObject__P(object), object, 1, "gtk_object_sink", "GtkObject*");
  gtk_object_sink(XEN_TO_C_GtkObject_(object));
  return(XEN_FALSE);
}

static XEN gxg_gtk_object_destroy(XEN object)
{
  #define H_gtk_object_destroy "void gtk_object_destroy(GtkObject* object)"
  XEN_ASSERT_TYPE(XEN_GtkObject__P(object), object, 1, "gtk_object_destroy", "GtkObject*");
  gtk_object_destroy(XEN_TO_C_GtkObject_(object));
  return(XEN_FALSE);
}

static XEN gxg_gtk_option_menu_get_type(void)
{
  #define H_gtk_option_menu_get_type "GtkType gtk_option_menu_get_type( void)"
  return(C_TO_XEN_GtkType(gtk_option_menu_get_type()));
}

static XEN gxg_gtk_option_menu_new(void)
{
  #define H_gtk_option_menu_new "GtkWidget* gtk_option_menu_new( void)"
  return(C_TO_XEN_GtkWidget_(gtk_option_menu_new()));
}

static XEN gxg_gtk_option_menu_get_menu(XEN option_menu)
{
  #define H_gtk_option_menu_get_menu "GtkWidget* gtk_option_menu_get_menu(GtkOptionMenu* option_menu)"
  XEN_ASSERT_TYPE(XEN_GtkOptionMenu__P(option_menu), option_menu, 1, "gtk_option_menu_get_menu", "GtkOptionMenu*");
  return(C_TO_XEN_GtkWidget_(gtk_option_menu_get_menu(XEN_TO_C_GtkOptionMenu_(option_menu))));
}

static XEN gxg_gtk_option_menu_set_menu(XEN option_menu, XEN menu)
{
  #define H_gtk_option_menu_set_menu "void gtk_option_menu_set_menu(GtkOptionMenu* option_menu, GtkWidget* menu)"
  XEN_ASSERT_TYPE(XEN_GtkOptionMenu__P(option_menu), option_menu, 1, "gtk_option_menu_set_menu", "GtkOptionMenu*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(menu), menu, 2, "gtk_option_menu_set_menu", "GtkWidget*");
  gtk_option_menu_set_menu(XEN_TO_C_GtkOptionMenu_(option_menu), XEN_TO_C_GtkWidget_(menu));
  return(XEN_FALSE);
}

static XEN gxg_gtk_option_menu_remove_menu(XEN option_menu)
{
  #define H_gtk_option_menu_remove_menu "void gtk_option_menu_remove_menu(GtkOptionMenu* option_menu)"
  XEN_ASSERT_TYPE(XEN_GtkOptionMenu__P(option_menu), option_menu, 1, "gtk_option_menu_remove_menu", "GtkOptionMenu*");
  gtk_option_menu_remove_menu(XEN_TO_C_GtkOptionMenu_(option_menu));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_option_menu_get_history(XEN option_menu)
{
  #define H_gtk_option_menu_get_history "gint gtk_option_menu_get_history(GtkOptionMenu* option_menu)"
  XEN_ASSERT_TYPE(XEN_GtkOptionMenu__P(option_menu), option_menu, 1, "gtk_option_menu_get_history", "GtkOptionMenu*");
  return(C_TO_XEN_gint(gtk_option_menu_get_history(XEN_TO_C_GtkOptionMenu_(option_menu))));
}
#endif

static XEN gxg_gtk_option_menu_set_history(XEN option_menu, XEN index)
{
  #define H_gtk_option_menu_set_history "void gtk_option_menu_set_history(GtkOptionMenu* option_menu, \
guint index)"
  XEN_ASSERT_TYPE(XEN_GtkOptionMenu__P(option_menu), option_menu, 1, "gtk_option_menu_set_history", "GtkOptionMenu*");
  XEN_ASSERT_TYPE(XEN_guint_P(index), index, 2, "gtk_option_menu_set_history", "guint");
  gtk_option_menu_set_history(XEN_TO_C_GtkOptionMenu_(option_menu), XEN_TO_C_guint(index));
  return(XEN_FALSE);
}

static XEN gxg_gtk_paned_get_type(void)
{
  #define H_gtk_paned_get_type "GtkType gtk_paned_get_type( void)"
  return(C_TO_XEN_GtkType(gtk_paned_get_type()));
}

static XEN gxg_gtk_paned_add1(XEN paned, XEN child)
{
  #define H_gtk_paned_add1 "void gtk_paned_add1(GtkPaned* paned, GtkWidget* child)"
  XEN_ASSERT_TYPE(XEN_GtkPaned__P(paned), paned, 1, "gtk_paned_add1", "GtkPaned*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(child), child, 2, "gtk_paned_add1", "GtkWidget*");
  gtk_paned_add1(XEN_TO_C_GtkPaned_(paned), XEN_TO_C_GtkWidget_(child));
  return(XEN_FALSE);
}

static XEN gxg_gtk_paned_add2(XEN paned, XEN child)
{
  #define H_gtk_paned_add2 "void gtk_paned_add2(GtkPaned* paned, GtkWidget* child)"
  XEN_ASSERT_TYPE(XEN_GtkPaned__P(paned), paned, 1, "gtk_paned_add2", "GtkPaned*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(child), child, 2, "gtk_paned_add2", "GtkWidget*");
  gtk_paned_add2(XEN_TO_C_GtkPaned_(paned), XEN_TO_C_GtkWidget_(child));
  return(XEN_FALSE);
}

static XEN gxg_gtk_paned_pack1(XEN paned, XEN child, XEN resize, XEN shrink)
{
  #define H_gtk_paned_pack1 "void gtk_paned_pack1(GtkPaned* paned, GtkWidget* child, gboolean resize, \
gboolean shrink)"
  XEN_ASSERT_TYPE(XEN_GtkPaned__P(paned), paned, 1, "gtk_paned_pack1", "GtkPaned*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(child), child, 2, "gtk_paned_pack1", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(resize), resize, 3, "gtk_paned_pack1", "gboolean");
  XEN_ASSERT_TYPE(XEN_gboolean_P(shrink), shrink, 4, "gtk_paned_pack1", "gboolean");
  gtk_paned_pack1(XEN_TO_C_GtkPaned_(paned), XEN_TO_C_GtkWidget_(child), XEN_TO_C_gboolean(resize), XEN_TO_C_gboolean(shrink));
  return(XEN_FALSE);
}

static XEN gxg_gtk_paned_pack2(XEN paned, XEN child, XEN resize, XEN shrink)
{
  #define H_gtk_paned_pack2 "void gtk_paned_pack2(GtkPaned* paned, GtkWidget* child, gboolean resize, \
gboolean shrink)"
  XEN_ASSERT_TYPE(XEN_GtkPaned__P(paned), paned, 1, "gtk_paned_pack2", "GtkPaned*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(child), child, 2, "gtk_paned_pack2", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(resize), resize, 3, "gtk_paned_pack2", "gboolean");
  XEN_ASSERT_TYPE(XEN_gboolean_P(shrink), shrink, 4, "gtk_paned_pack2", "gboolean");
  gtk_paned_pack2(XEN_TO_C_GtkPaned_(paned), XEN_TO_C_GtkWidget_(child), XEN_TO_C_gboolean(resize), XEN_TO_C_gboolean(shrink));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_paned_get_position(XEN paned)
{
  #define H_gtk_paned_get_position "gint gtk_paned_get_position(GtkPaned* paned)"
  XEN_ASSERT_TYPE(XEN_GtkPaned__P(paned), paned, 1, "gtk_paned_get_position", "GtkPaned*");
  return(C_TO_XEN_gint(gtk_paned_get_position(XEN_TO_C_GtkPaned_(paned))));
}
#endif

static XEN gxg_gtk_paned_set_position(XEN paned, XEN position)
{
  #define H_gtk_paned_set_position "void gtk_paned_set_position(GtkPaned* paned, gint position)"
  XEN_ASSERT_TYPE(XEN_GtkPaned__P(paned), paned, 1, "gtk_paned_set_position", "GtkPaned*");
  XEN_ASSERT_TYPE(XEN_gint_P(position), position, 2, "gtk_paned_set_position", "gint");
  gtk_paned_set_position(XEN_TO_C_GtkPaned_(paned), XEN_TO_C_gint(position));
  return(XEN_FALSE);
}

static XEN gxg_gtk_paned_compute_position(XEN paned, XEN allocation, XEN child1_req, XEN child2_req)
{
  #define H_gtk_paned_compute_position "void gtk_paned_compute_position(GtkPaned* paned, gint allocation, \
gint child1_req, gint child2_req)"
  XEN_ASSERT_TYPE(XEN_GtkPaned__P(paned), paned, 1, "gtk_paned_compute_position", "GtkPaned*");
  XEN_ASSERT_TYPE(XEN_gint_P(allocation), allocation, 2, "gtk_paned_compute_position", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(child1_req), child1_req, 3, "gtk_paned_compute_position", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(child2_req), child2_req, 4, "gtk_paned_compute_position", "gint");
  gtk_paned_compute_position(XEN_TO_C_GtkPaned_(paned), XEN_TO_C_gint(allocation), XEN_TO_C_gint(child1_req), XEN_TO_C_gint(child2_req));
  return(XEN_FALSE);
}

static XEN gxg_gtk_plug_get_type(void)
{
  #define H_gtk_plug_get_type "GtkType gtk_plug_get_type( void)"
  return(C_TO_XEN_GtkType(gtk_plug_get_type()));
}

static XEN gxg_gtk_plug_construct(XEN plug, XEN socket_id)
{
  #define H_gtk_plug_construct "void gtk_plug_construct(GtkPlug* plug, GdkNativeWindow socket_id)"
  XEN_ASSERT_TYPE(XEN_GtkPlug__P(plug), plug, 1, "gtk_plug_construct", "GtkPlug*");
  XEN_ASSERT_TYPE(XEN_GdkNativeWindow_P(socket_id), socket_id, 2, "gtk_plug_construct", "GdkNativeWindow");
  gtk_plug_construct(XEN_TO_C_GtkPlug_(plug), XEN_TO_C_GdkNativeWindow(socket_id));
  return(XEN_FALSE);
}

static XEN gxg_gtk_plug_new(XEN socket_id)
{
  #define H_gtk_plug_new "GtkWidget* gtk_plug_new(GdkNativeWindow socket_id)"
  XEN_ASSERT_TYPE(XEN_GdkNativeWindow_P(socket_id), socket_id, 1, "gtk_plug_new", "GdkNativeWindow");
  return(C_TO_XEN_GtkWidget_(gtk_plug_new(XEN_TO_C_GdkNativeWindow(socket_id))));
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_plug_get_id(XEN plug)
{
  #define H_gtk_plug_get_id "GdkNativeWindow gtk_plug_get_id(GtkPlug* plug)"
  XEN_ASSERT_TYPE(XEN_GtkPlug__P(plug), plug, 1, "gtk_plug_get_id", "GtkPlug*");
  return(C_TO_XEN_GdkNativeWindow(gtk_plug_get_id(XEN_TO_C_GtkPlug_(plug))));
}
#endif

static XEN gxg_gtk_progress_bar_get_type(void)
{
  #define H_gtk_progress_bar_get_type "GtkType gtk_progress_bar_get_type( void)"
  return(C_TO_XEN_GtkType(gtk_progress_bar_get_type()));
}

static XEN gxg_gtk_progress_bar_new(void)
{
  #define H_gtk_progress_bar_new "GtkWidget* gtk_progress_bar_new( void)"
  return(C_TO_XEN_GtkWidget_(gtk_progress_bar_new()));
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_progress_bar_pulse(XEN pbar)
{
  #define H_gtk_progress_bar_pulse "void gtk_progress_bar_pulse(GtkProgressBar* pbar)"
  XEN_ASSERT_TYPE(XEN_GtkProgressBar__P(pbar), pbar, 1, "gtk_progress_bar_pulse", "GtkProgressBar*");
  gtk_progress_bar_pulse(XEN_TO_C_GtkProgressBar_(pbar));
  return(XEN_FALSE);
}

static XEN gxg_gtk_progress_bar_set_text(XEN pbar, XEN text)
{
  #define H_gtk_progress_bar_set_text "void gtk_progress_bar_set_text(GtkProgressBar* pbar, gchar* text)"
  XEN_ASSERT_TYPE(XEN_GtkProgressBar__P(pbar), pbar, 1, "gtk_progress_bar_set_text", "GtkProgressBar*");
  XEN_ASSERT_TYPE(XEN_gchar__P(text), text, 2, "gtk_progress_bar_set_text", "gchar*");
  gtk_progress_bar_set_text(XEN_TO_C_GtkProgressBar_(pbar), XEN_TO_C_gchar_(text));
  return(XEN_FALSE);
}

static XEN gxg_gtk_progress_bar_set_fraction(XEN pbar, XEN fraction)
{
  #define H_gtk_progress_bar_set_fraction "void gtk_progress_bar_set_fraction(GtkProgressBar* pbar, gdouble fraction)"
  XEN_ASSERT_TYPE(XEN_GtkProgressBar__P(pbar), pbar, 1, "gtk_progress_bar_set_fraction", "GtkProgressBar*");
  XEN_ASSERT_TYPE(XEN_gdouble_P(fraction), fraction, 2, "gtk_progress_bar_set_fraction", "gdouble");
  gtk_progress_bar_set_fraction(XEN_TO_C_GtkProgressBar_(pbar), XEN_TO_C_gdouble(fraction));
  return(XEN_FALSE);
}

static XEN gxg_gtk_progress_bar_set_pulse_step(XEN pbar, XEN fraction)
{
  #define H_gtk_progress_bar_set_pulse_step "void gtk_progress_bar_set_pulse_step(GtkProgressBar* pbar, \
gdouble fraction)"
  XEN_ASSERT_TYPE(XEN_GtkProgressBar__P(pbar), pbar, 1, "gtk_progress_bar_set_pulse_step", "GtkProgressBar*");
  XEN_ASSERT_TYPE(XEN_gdouble_P(fraction), fraction, 2, "gtk_progress_bar_set_pulse_step", "gdouble");
  gtk_progress_bar_set_pulse_step(XEN_TO_C_GtkProgressBar_(pbar), XEN_TO_C_gdouble(fraction));
  return(XEN_FALSE);
}
#endif

static XEN gxg_gtk_progress_bar_set_orientation(XEN pbar, XEN orientation)
{
  #define H_gtk_progress_bar_set_orientation "void gtk_progress_bar_set_orientation(GtkProgressBar* pbar, \
GtkProgressBarOrientation orientation)"
  XEN_ASSERT_TYPE(XEN_GtkProgressBar__P(pbar), pbar, 1, "gtk_progress_bar_set_orientation", "GtkProgressBar*");
  XEN_ASSERT_TYPE(XEN_GtkProgressBarOrientation_P(orientation), orientation, 2, "gtk_progress_bar_set_orientation", "GtkProgressBarOrientation");
  gtk_progress_bar_set_orientation(XEN_TO_C_GtkProgressBar_(pbar), XEN_TO_C_GtkProgressBarOrientation(orientation));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_progress_bar_get_text(XEN pbar)
{
  #define H_gtk_progress_bar_get_text "gchar* gtk_progress_bar_get_text(GtkProgressBar* pbar)"
  XEN_ASSERT_TYPE(XEN_GtkProgressBar__P(pbar), pbar, 1, "gtk_progress_bar_get_text", "GtkProgressBar*");
  return(C_TO_XEN_gchar_(gtk_progress_bar_get_text(XEN_TO_C_GtkProgressBar_(pbar))));
}

static XEN gxg_gtk_progress_bar_get_fraction(XEN pbar)
{
  #define H_gtk_progress_bar_get_fraction "gdouble gtk_progress_bar_get_fraction(GtkProgressBar* pbar)"
  XEN_ASSERT_TYPE(XEN_GtkProgressBar__P(pbar), pbar, 1, "gtk_progress_bar_get_fraction", "GtkProgressBar*");
  return(C_TO_XEN_gdouble(gtk_progress_bar_get_fraction(XEN_TO_C_GtkProgressBar_(pbar))));
}

static XEN gxg_gtk_progress_bar_get_pulse_step(XEN pbar)
{
  #define H_gtk_progress_bar_get_pulse_step "gdouble gtk_progress_bar_get_pulse_step(GtkProgressBar* pbar)"
  XEN_ASSERT_TYPE(XEN_GtkProgressBar__P(pbar), pbar, 1, "gtk_progress_bar_get_pulse_step", "GtkProgressBar*");
  return(C_TO_XEN_gdouble(gtk_progress_bar_get_pulse_step(XEN_TO_C_GtkProgressBar_(pbar))));
}

static XEN gxg_gtk_progress_bar_get_orientation(XEN pbar)
{
  #define H_gtk_progress_bar_get_orientation "GtkProgressBarOrientation gtk_progress_bar_get_orientation(GtkProgressBar* pbar)"
  XEN_ASSERT_TYPE(XEN_GtkProgressBar__P(pbar), pbar, 1, "gtk_progress_bar_get_orientation", "GtkProgressBar*");
  return(C_TO_XEN_GtkProgressBarOrientation(gtk_progress_bar_get_orientation(XEN_TO_C_GtkProgressBar_(pbar))));
}
#endif

static XEN gxg_gtk_radio_button_get_type(void)
{
  #define H_gtk_radio_button_get_type "GtkType gtk_radio_button_get_type( void)"
  return(C_TO_XEN_GtkType(gtk_radio_button_get_type()));
}

static XEN gxg_gtk_radio_button_new(XEN group)
{
  #define H_gtk_radio_button_new "GtkWidget* gtk_radio_button_new(GSList* group)"
  XEN_ASSERT_TYPE(XEN_GSList__P(group), group, 1, "gtk_radio_button_new", "GSList*");
  return(C_TO_XEN_GtkWidget_(gtk_radio_button_new(XEN_TO_C_GSList_(group))));
}

static XEN gxg_gtk_radio_button_new_from_widget(XEN group)
{
  #define H_gtk_radio_button_new_from_widget "GtkWidget* gtk_radio_button_new_from_widget(GtkRadioButton* group)"
  XEN_ASSERT_TYPE(XEN_GtkRadioButton__P(group), group, 1, "gtk_radio_button_new_from_widget", "GtkRadioButton*");
  return(C_TO_XEN_GtkWidget_(gtk_radio_button_new_from_widget(XEN_TO_C_GtkRadioButton_(group))));
}

static XEN gxg_gtk_radio_button_new_with_label(XEN group, XEN label)
{
  #define H_gtk_radio_button_new_with_label "GtkWidget* gtk_radio_button_new_with_label(GSList* group, \
gchar* label)"
  XEN_ASSERT_TYPE(XEN_GSList__P(group), group, 1, "gtk_radio_button_new_with_label", "GSList*");
  XEN_ASSERT_TYPE(XEN_gchar__P(label), label, 2, "gtk_radio_button_new_with_label", "gchar*");
  return(C_TO_XEN_GtkWidget_(gtk_radio_button_new_with_label(XEN_TO_C_GSList_(group), XEN_TO_C_gchar_(label))));
}

static XEN gxg_gtk_radio_button_new_with_label_from_widget(XEN group, XEN label)
{
  #define H_gtk_radio_button_new_with_label_from_widget "GtkWidget* gtk_radio_button_new_with_label_from_widget(GtkRadioButton* group, \
gchar* label)"
  XEN_ASSERT_TYPE(XEN_GtkRadioButton__P(group), group, 1, "gtk_radio_button_new_with_label_from_widget", "GtkRadioButton*");
  XEN_ASSERT_TYPE(XEN_gchar__P(label), label, 2, "gtk_radio_button_new_with_label_from_widget", "gchar*");
  return(C_TO_XEN_GtkWidget_(gtk_radio_button_new_with_label_from_widget(XEN_TO_C_GtkRadioButton_(group), XEN_TO_C_gchar_(label))));
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_radio_button_new_with_mnemonic(XEN group, XEN label)
{
  #define H_gtk_radio_button_new_with_mnemonic "GtkWidget* gtk_radio_button_new_with_mnemonic(GSList* group, \
gchar* label)"
  XEN_ASSERT_TYPE(XEN_GSList__P(group), group, 1, "gtk_radio_button_new_with_mnemonic", "GSList*");
  XEN_ASSERT_TYPE(XEN_gchar__P(label), label, 2, "gtk_radio_button_new_with_mnemonic", "gchar*");
  return(C_TO_XEN_GtkWidget_(gtk_radio_button_new_with_mnemonic(XEN_TO_C_GSList_(group), XEN_TO_C_gchar_(label))));
}

static XEN gxg_gtk_radio_button_new_with_mnemonic_from_widget(XEN group, XEN label)
{
  #define H_gtk_radio_button_new_with_mnemonic_from_widget "GtkWidget* gtk_radio_button_new_with_mnemonic_from_widget(GtkRadioButton* group, \
gchar* label)"
  XEN_ASSERT_TYPE(XEN_GtkRadioButton__P(group), group, 1, "gtk_radio_button_new_with_mnemonic_from_widget", "GtkRadioButton*");
  XEN_ASSERT_TYPE(XEN_gchar__P(label), label, 2, "gtk_radio_button_new_with_mnemonic_from_widget", "gchar*");
  return(C_TO_XEN_GtkWidget_(gtk_radio_button_new_with_mnemonic_from_widget(XEN_TO_C_GtkRadioButton_(group), XEN_TO_C_gchar_(label))));
}

static XEN gxg_gtk_radio_button_get_group(XEN radio_button)
{
  #define H_gtk_radio_button_get_group "GSList* gtk_radio_button_get_group(GtkRadioButton* radio_button)"
  XEN_ASSERT_TYPE(XEN_GtkRadioButton__P(radio_button), radio_button, 1, "gtk_radio_button_get_group", "GtkRadioButton*");
  return(C_TO_XEN_GSList_(gtk_radio_button_get_group(XEN_TO_C_GtkRadioButton_(radio_button))));
}
#endif

static XEN gxg_gtk_radio_button_set_group(XEN radio_button, XEN group)
{
  #define H_gtk_radio_button_set_group "void gtk_radio_button_set_group(GtkRadioButton* radio_button, \
GSList* group)"
  XEN_ASSERT_TYPE(XEN_GtkRadioButton__P(radio_button), radio_button, 1, "gtk_radio_button_set_group", "GtkRadioButton*");
  XEN_ASSERT_TYPE(XEN_GSList__P(group), group, 2, "gtk_radio_button_set_group", "GSList*");
  gtk_radio_button_set_group(XEN_TO_C_GtkRadioButton_(radio_button), XEN_TO_C_GSList_(group));
  return(XEN_FALSE);
}

static XEN gxg_gtk_radio_menu_item_get_type(void)
{
  #define H_gtk_radio_menu_item_get_type "GtkType gtk_radio_menu_item_get_type( void)"
  return(C_TO_XEN_GtkType(gtk_radio_menu_item_get_type()));
}

static XEN gxg_gtk_radio_menu_item_new(XEN group)
{
  #define H_gtk_radio_menu_item_new "GtkWidget* gtk_radio_menu_item_new(GSList* group)"
  XEN_ASSERT_TYPE(XEN_GSList__P(group), group, 1, "gtk_radio_menu_item_new", "GSList*");
  return(C_TO_XEN_GtkWidget_(gtk_radio_menu_item_new(XEN_TO_C_GSList_(group))));
}

static XEN gxg_gtk_radio_menu_item_new_with_label(XEN group, XEN label)
{
  #define H_gtk_radio_menu_item_new_with_label "GtkWidget* gtk_radio_menu_item_new_with_label(GSList* group, \
gchar* label)"
  XEN_ASSERT_TYPE(XEN_GSList__P(group), group, 1, "gtk_radio_menu_item_new_with_label", "GSList*");
  XEN_ASSERT_TYPE(XEN_gchar__P(label), label, 2, "gtk_radio_menu_item_new_with_label", "gchar*");
  return(C_TO_XEN_GtkWidget_(gtk_radio_menu_item_new_with_label(XEN_TO_C_GSList_(group), XEN_TO_C_gchar_(label))));
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_radio_menu_item_new_with_mnemonic(XEN group, XEN label)
{
  #define H_gtk_radio_menu_item_new_with_mnemonic "GtkWidget* gtk_radio_menu_item_new_with_mnemonic(GSList* group, \
gchar* label)"
  XEN_ASSERT_TYPE(XEN_GSList__P(group), group, 1, "gtk_radio_menu_item_new_with_mnemonic", "GSList*");
  XEN_ASSERT_TYPE(XEN_gchar__P(label), label, 2, "gtk_radio_menu_item_new_with_mnemonic", "gchar*");
  return(C_TO_XEN_GtkWidget_(gtk_radio_menu_item_new_with_mnemonic(XEN_TO_C_GSList_(group), XEN_TO_C_gchar_(label))));
}

static XEN gxg_gtk_radio_menu_item_get_group(XEN radio_menu_item)
{
  #define H_gtk_radio_menu_item_get_group "GSList* gtk_radio_menu_item_get_group(GtkRadioMenuItem* radio_menu_item)"
  XEN_ASSERT_TYPE(XEN_GtkRadioMenuItem__P(radio_menu_item), radio_menu_item, 1, "gtk_radio_menu_item_get_group", "GtkRadioMenuItem*");
  return(C_TO_XEN_GSList_(gtk_radio_menu_item_get_group(XEN_TO_C_GtkRadioMenuItem_(radio_menu_item))));
}
#endif

static XEN gxg_gtk_radio_menu_item_set_group(XEN radio_menu_item, XEN group)
{
  #define H_gtk_radio_menu_item_set_group "void gtk_radio_menu_item_set_group(GtkRadioMenuItem* radio_menu_item, \
GSList* group)"
  XEN_ASSERT_TYPE(XEN_GtkRadioMenuItem__P(radio_menu_item), radio_menu_item, 1, "gtk_radio_menu_item_set_group", "GtkRadioMenuItem*");
  XEN_ASSERT_TYPE(XEN_GSList__P(group), group, 2, "gtk_radio_menu_item_set_group", "GSList*");
  gtk_radio_menu_item_set_group(XEN_TO_C_GtkRadioMenuItem_(radio_menu_item), XEN_TO_C_GSList_(group));
  return(XEN_FALSE);
}

static XEN gxg_gtk_range_get_type(void)
{
  #define H_gtk_range_get_type "GtkType gtk_range_get_type( void)"
  return(C_TO_XEN_GtkType(gtk_range_get_type()));
}

static XEN gxg_gtk_range_set_update_policy(XEN range, XEN policy)
{
  #define H_gtk_range_set_update_policy "void gtk_range_set_update_policy(GtkRange* range, GtkUpdateType policy)"
  XEN_ASSERT_TYPE(XEN_GtkRange__P(range), range, 1, "gtk_range_set_update_policy", "GtkRange*");
  XEN_ASSERT_TYPE(XEN_GtkUpdateType_P(policy), policy, 2, "gtk_range_set_update_policy", "GtkUpdateType");
  gtk_range_set_update_policy(XEN_TO_C_GtkRange_(range), XEN_TO_C_GtkUpdateType(policy));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_range_get_update_policy(XEN range)
{
  #define H_gtk_range_get_update_policy "GtkUpdateType gtk_range_get_update_policy(GtkRange* range)"
  XEN_ASSERT_TYPE(XEN_GtkRange__P(range), range, 1, "gtk_range_get_update_policy", "GtkRange*");
  return(C_TO_XEN_GtkUpdateType(gtk_range_get_update_policy(XEN_TO_C_GtkRange_(range))));
}
#endif

static XEN gxg_gtk_range_set_adjustment(XEN range, XEN adjustment)
{
  #define H_gtk_range_set_adjustment "void gtk_range_set_adjustment(GtkRange* range, GtkAdjustment* adjustment)"
  XEN_ASSERT_TYPE(XEN_GtkRange__P(range), range, 1, "gtk_range_set_adjustment", "GtkRange*");
  XEN_ASSERT_TYPE(XEN_GtkAdjustment__P(adjustment), adjustment, 2, "gtk_range_set_adjustment", "GtkAdjustment*");
  gtk_range_set_adjustment(XEN_TO_C_GtkRange_(range), XEN_TO_C_GtkAdjustment_(adjustment));
  return(XEN_FALSE);
}

static XEN gxg_gtk_range_get_adjustment(XEN range)
{
  #define H_gtk_range_get_adjustment "GtkAdjustment* gtk_range_get_adjustment(GtkRange* range)"
  XEN_ASSERT_TYPE(XEN_GtkRange__P(range), range, 1, "gtk_range_get_adjustment", "GtkRange*");
  return(C_TO_XEN_GtkAdjustment_(gtk_range_get_adjustment(XEN_TO_C_GtkRange_(range))));
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_range_set_inverted(XEN range, XEN setting)
{
  #define H_gtk_range_set_inverted "void gtk_range_set_inverted(GtkRange* range, gboolean setting)"
  XEN_ASSERT_TYPE(XEN_GtkRange__P(range), range, 1, "gtk_range_set_inverted", "GtkRange*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(setting), setting, 2, "gtk_range_set_inverted", "gboolean");
  gtk_range_set_inverted(XEN_TO_C_GtkRange_(range), XEN_TO_C_gboolean(setting));
  return(XEN_FALSE);
}

static XEN gxg_gtk_range_get_inverted(XEN range)
{
  #define H_gtk_range_get_inverted "gboolean gtk_range_get_inverted(GtkRange* range)"
  XEN_ASSERT_TYPE(XEN_GtkRange__P(range), range, 1, "gtk_range_get_inverted", "GtkRange*");
  return(C_TO_XEN_gboolean(gtk_range_get_inverted(XEN_TO_C_GtkRange_(range))));
}

static XEN gxg_gtk_range_set_increments(XEN range, XEN step, XEN page)
{
  #define H_gtk_range_set_increments "void gtk_range_set_increments(GtkRange* range, gdouble step, gdouble page)"
  XEN_ASSERT_TYPE(XEN_GtkRange__P(range), range, 1, "gtk_range_set_increments", "GtkRange*");
  XEN_ASSERT_TYPE(XEN_gdouble_P(step), step, 2, "gtk_range_set_increments", "gdouble");
  XEN_ASSERT_TYPE(XEN_gdouble_P(page), page, 3, "gtk_range_set_increments", "gdouble");
  gtk_range_set_increments(XEN_TO_C_GtkRange_(range), XEN_TO_C_gdouble(step), XEN_TO_C_gdouble(page));
  return(XEN_FALSE);
}

static XEN gxg_gtk_range_set_range(XEN range, XEN min, XEN max)
{
  #define H_gtk_range_set_range "void gtk_range_set_range(GtkRange* range, gdouble min, gdouble max)"
  XEN_ASSERT_TYPE(XEN_GtkRange__P(range), range, 1, "gtk_range_set_range", "GtkRange*");
  XEN_ASSERT_TYPE(XEN_gdouble_P(min), min, 2, "gtk_range_set_range", "gdouble");
  XEN_ASSERT_TYPE(XEN_gdouble_P(max), max, 3, "gtk_range_set_range", "gdouble");
  gtk_range_set_range(XEN_TO_C_GtkRange_(range), XEN_TO_C_gdouble(min), XEN_TO_C_gdouble(max));
  return(XEN_FALSE);
}

static XEN gxg_gtk_range_set_value(XEN range, XEN value)
{
  #define H_gtk_range_set_value "void gtk_range_set_value(GtkRange* range, gdouble value)"
  XEN_ASSERT_TYPE(XEN_GtkRange__P(range), range, 1, "gtk_range_set_value", "GtkRange*");
  XEN_ASSERT_TYPE(XEN_gdouble_P(value), value, 2, "gtk_range_set_value", "gdouble");
  gtk_range_set_value(XEN_TO_C_GtkRange_(range), XEN_TO_C_gdouble(value));
  return(XEN_FALSE);
}

static XEN gxg_gtk_range_get_value(XEN range)
{
  #define H_gtk_range_get_value "gdouble gtk_range_get_value(GtkRange* range)"
  XEN_ASSERT_TYPE(XEN_GtkRange__P(range), range, 1, "gtk_range_get_value", "GtkRange*");
  return(C_TO_XEN_gdouble(gtk_range_get_value(XEN_TO_C_GtkRange_(range))));
}
#endif

static XEN gxg_gtk_rc_add_default_file(XEN filename)
{
  #define H_gtk_rc_add_default_file "void gtk_rc_add_default_file(gchar* filename)"
  XEN_ASSERT_TYPE(XEN_gchar__P(filename), filename, 1, "gtk_rc_add_default_file", "gchar*");
  gtk_rc_add_default_file(XEN_TO_C_gchar_(filename));
  return(XEN_FALSE);
}

static XEN gxg_gtk_rc_set_default_files(XEN filenames)
{
  #define H_gtk_rc_set_default_files "void gtk_rc_set_default_files(gchar** filenames)"
  XEN_ASSERT_TYPE(XEN_gchar___P(filenames), filenames, 1, "gtk_rc_set_default_files", "gchar**");
  gtk_rc_set_default_files(XEN_TO_C_gchar__(filenames));
  return(XEN_FALSE);
}

static XEN gxg_gtk_rc_get_default_files(void)
{
  #define H_gtk_rc_get_default_files "gchar** gtk_rc_get_default_files( void)"
  return(C_TO_XEN_gchar__(gtk_rc_get_default_files()));
}

static XEN gxg_gtk_rc_get_style(XEN widget)
{
  #define H_gtk_rc_get_style "GtkStyle* gtk_rc_get_style(GtkWidget* widget)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_rc_get_style", "GtkWidget*");
  return(C_TO_XEN_GtkStyle_(gtk_rc_get_style(XEN_TO_C_GtkWidget_(widget))));
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_rc_get_style_by_paths(XEN settings, XEN widget_path, XEN class_path, XEN type)
{
  #define H_gtk_rc_get_style_by_paths "GtkStyle* gtk_rc_get_style_by_paths(GtkSettings* settings, char* widget_path, \
char* class_path, GType type)"
  XEN_ASSERT_TYPE(XEN_GtkSettings__P(settings), settings, 1, "gtk_rc_get_style_by_paths", "GtkSettings*");
  XEN_ASSERT_TYPE(XEN_char__P(widget_path), widget_path, 2, "gtk_rc_get_style_by_paths", "char*");
  XEN_ASSERT_TYPE(XEN_char__P(class_path), class_path, 3, "gtk_rc_get_style_by_paths", "char*");
  XEN_ASSERT_TYPE(XEN_GType_P(type), type, 4, "gtk_rc_get_style_by_paths", "GType");
  return(C_TO_XEN_GtkStyle_(gtk_rc_get_style_by_paths(XEN_TO_C_GtkSettings_(settings), XEN_TO_C_char_(widget_path), XEN_TO_C_char_(class_path), 
                                                      XEN_TO_C_GType(type))));
}

static XEN gxg_gtk_rc_reparse_all_for_settings(XEN settings, XEN force_load)
{
  #define H_gtk_rc_reparse_all_for_settings "gboolean gtk_rc_reparse_all_for_settings(GtkSettings* settings, \
gboolean force_load)"
  XEN_ASSERT_TYPE(XEN_GtkSettings__P(settings), settings, 1, "gtk_rc_reparse_all_for_settings", "GtkSettings*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(force_load), force_load, 2, "gtk_rc_reparse_all_for_settings", "gboolean");
  return(C_TO_XEN_gboolean(gtk_rc_reparse_all_for_settings(XEN_TO_C_GtkSettings_(settings), XEN_TO_C_gboolean(force_load))));
}

static XEN gxg_gtk_rc_find_pixmap_in_path(XEN settings, XEN scanner, XEN pixmap_file)
{
  #define H_gtk_rc_find_pixmap_in_path "gchar* gtk_rc_find_pixmap_in_path(GtkSettings* settings, GScanner* scanner, \
gchar* pixmap_file)"
  XEN_ASSERT_TYPE(XEN_GtkSettings__P(settings), settings, 1, "gtk_rc_find_pixmap_in_path", "GtkSettings*");
  XEN_ASSERT_TYPE(XEN_GScanner__P(scanner) || XEN_FALSE_P(scanner), scanner, 2, "gtk_rc_find_pixmap_in_path", "GScanner*");
  XEN_ASSERT_TYPE(XEN_gchar__P(pixmap_file), pixmap_file, 3, "gtk_rc_find_pixmap_in_path", "gchar*");
  return(C_TO_XEN_gchar_(gtk_rc_find_pixmap_in_path(XEN_TO_C_GtkSettings_(settings), XEN_TO_C_GScanner_(scanner), XEN_TO_C_gchar_(pixmap_file))));
}
#endif

static XEN gxg_gtk_rc_parse(XEN filename)
{
  #define H_gtk_rc_parse "void gtk_rc_parse(gchar* filename)"
  XEN_ASSERT_TYPE(XEN_gchar__P(filename), filename, 1, "gtk_rc_parse", "gchar*");
  gtk_rc_parse(XEN_TO_C_gchar_(filename));
  return(XEN_FALSE);
}

static XEN gxg_gtk_rc_parse_string(XEN rc_string)
{
  #define H_gtk_rc_parse_string "void gtk_rc_parse_string(gchar* rc_string)"
  XEN_ASSERT_TYPE(XEN_gchar__P(rc_string), rc_string, 1, "gtk_rc_parse_string", "gchar*");
  gtk_rc_parse_string(XEN_TO_C_gchar_(rc_string));
  return(XEN_FALSE);
}

static XEN gxg_gtk_rc_reparse_all(void)
{
  #define H_gtk_rc_reparse_all "gboolean gtk_rc_reparse_all( void)"
  return(C_TO_XEN_gboolean(gtk_rc_reparse_all()));
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_rc_style_get_type(void)
{
  #define H_gtk_rc_style_get_type "GType gtk_rc_style_get_type( void)"
  return(C_TO_XEN_GType(gtk_rc_style_get_type()));
}
#endif

static XEN gxg_gtk_rc_style_new(void)
{
  #define H_gtk_rc_style_new "GtkRcStyle* gtk_rc_style_new( void)"
  return(C_TO_XEN_GtkRcStyle_(gtk_rc_style_new()));
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_rc_style_copy(XEN orig)
{
  #define H_gtk_rc_style_copy "GtkRcStyle* gtk_rc_style_copy(GtkRcStyle* orig)"
  XEN_ASSERT_TYPE(XEN_GtkRcStyle__P(orig), orig, 1, "gtk_rc_style_copy", "GtkRcStyle*");
  return(C_TO_XEN_GtkRcStyle_(gtk_rc_style_copy(XEN_TO_C_GtkRcStyle_(orig))));
}
#endif

static XEN gxg_gtk_rc_style_ref(XEN rc_style)
{
  #define H_gtk_rc_style_ref "void gtk_rc_style_ref(GtkRcStyle* rc_style)"
  XEN_ASSERT_TYPE(XEN_GtkRcStyle__P(rc_style), rc_style, 1, "gtk_rc_style_ref", "GtkRcStyle*");
  gtk_rc_style_ref(XEN_TO_C_GtkRcStyle_(rc_style));
  return(XEN_FALSE);
}

static XEN gxg_gtk_rc_style_unref(XEN rc_style)
{
  #define H_gtk_rc_style_unref "void gtk_rc_style_unref(GtkRcStyle* rc_style)"
  XEN_ASSERT_TYPE(XEN_GtkRcStyle__P(rc_style), rc_style, 1, "gtk_rc_style_unref", "GtkRcStyle*");
  gtk_rc_style_unref(XEN_TO_C_GtkRcStyle_(rc_style));
  return(XEN_FALSE);
}

static XEN gxg_gtk_rc_find_module_in_path(XEN module_file)
{
  #define H_gtk_rc_find_module_in_path "gchar* gtk_rc_find_module_in_path(gchar* module_file)"
  XEN_ASSERT_TYPE(XEN_gchar__P(module_file), module_file, 1, "gtk_rc_find_module_in_path", "gchar*");
  return(C_TO_XEN_gchar_(gtk_rc_find_module_in_path(XEN_TO_C_gchar_(module_file))));
}

static XEN gxg_gtk_rc_get_theme_dir(void)
{
  #define H_gtk_rc_get_theme_dir "gchar* gtk_rc_get_theme_dir( void)"
  return(C_TO_XEN_gchar_(gtk_rc_get_theme_dir()));
}

static XEN gxg_gtk_rc_get_module_dir(void)
{
  #define H_gtk_rc_get_module_dir "gchar* gtk_rc_get_module_dir( void)"
  return(C_TO_XEN_gchar_(gtk_rc_get_module_dir()));
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_rc_get_im_module_path(void)
{
  #define H_gtk_rc_get_im_module_path "gchar* gtk_rc_get_im_module_path( void)"
  return(C_TO_XEN_gchar_(gtk_rc_get_im_module_path()));
}

static XEN gxg_gtk_rc_get_im_module_file(void)
{
  #define H_gtk_rc_get_im_module_file "gchar* gtk_rc_get_im_module_file( void)"
  return(C_TO_XEN_gchar_(gtk_rc_get_im_module_file()));
}

static XEN gxg_gtk_rc_scanner_new(void)
{
  #define H_gtk_rc_scanner_new "GScanner* gtk_rc_scanner_new( void)"
  return(C_TO_XEN_GScanner_(gtk_rc_scanner_new()));
}
#endif

static XEN gxg_gtk_rc_parse_color(XEN scanner, XEN color)
{
  #define H_gtk_rc_parse_color "guint gtk_rc_parse_color(GScanner* scanner, GdkColor* color)"
  XEN_ASSERT_TYPE(XEN_GScanner__P(scanner), scanner, 1, "gtk_rc_parse_color", "GScanner*");
  XEN_ASSERT_TYPE(XEN_GdkColor__P(color), color, 2, "gtk_rc_parse_color", "GdkColor*");
  return(C_TO_XEN_guint(gtk_rc_parse_color(XEN_TO_C_GScanner_(scanner), XEN_TO_C_GdkColor_(color))));
}

static XEN gxg_gtk_rc_parse_state(XEN scanner, XEN state)
{
  #define H_gtk_rc_parse_state "guint gtk_rc_parse_state(GScanner* scanner, GtkStateType* state)"
  XEN_ASSERT_TYPE(XEN_GScanner__P(scanner), scanner, 1, "gtk_rc_parse_state", "GScanner*");
  XEN_ASSERT_TYPE(XEN_GtkStateType__P(state), state, 2, "gtk_rc_parse_state", "GtkStateType*");
  return(C_TO_XEN_guint(gtk_rc_parse_state(XEN_TO_C_GScanner_(scanner), XEN_TO_C_GtkStateType_(state))));
}

static XEN gxg_gtk_rc_parse_priority(XEN scanner, XEN priority)
{
  #define H_gtk_rc_parse_priority "guint gtk_rc_parse_priority(GScanner* scanner, GtkPathPriorityType* priority)"
  XEN_ASSERT_TYPE(XEN_GScanner__P(scanner), scanner, 1, "gtk_rc_parse_priority", "GScanner*");
  XEN_ASSERT_TYPE(XEN_GtkPathPriorityType__P(priority), priority, 2, "gtk_rc_parse_priority", "GtkPathPriorityType*");
  return(C_TO_XEN_guint(gtk_rc_parse_priority(XEN_TO_C_GScanner_(scanner), XEN_TO_C_GtkPathPriorityType_(priority))));
}

static XEN gxg_gtk_ruler_get_type(void)
{
  #define H_gtk_ruler_get_type "GtkType gtk_ruler_get_type( void)"
  return(C_TO_XEN_GtkType(gtk_ruler_get_type()));
}

static XEN gxg_gtk_ruler_set_metric(XEN ruler, XEN metric)
{
  #define H_gtk_ruler_set_metric "void gtk_ruler_set_metric(GtkRuler* ruler, GtkMetricType metric)"
  XEN_ASSERT_TYPE(XEN_GtkRuler__P(ruler), ruler, 1, "gtk_ruler_set_metric", "GtkRuler*");
  XEN_ASSERT_TYPE(XEN_GtkMetricType_P(metric), metric, 2, "gtk_ruler_set_metric", "GtkMetricType");
  gtk_ruler_set_metric(XEN_TO_C_GtkRuler_(ruler), XEN_TO_C_GtkMetricType(metric));
  return(XEN_FALSE);
}

static XEN gxg_gtk_ruler_set_range(XEN ruler, XEN lower, XEN upper, XEN position, XEN max_size)
{
  #define H_gtk_ruler_set_range "void gtk_ruler_set_range(GtkRuler* ruler, gdouble lower, gdouble upper, \
gdouble position, gdouble max_size)"
  XEN_ASSERT_TYPE(XEN_GtkRuler__P(ruler), ruler, 1, "gtk_ruler_set_range", "GtkRuler*");
  XEN_ASSERT_TYPE(XEN_gdouble_P(lower), lower, 2, "gtk_ruler_set_range", "gdouble");
  XEN_ASSERT_TYPE(XEN_gdouble_P(upper), upper, 3, "gtk_ruler_set_range", "gdouble");
  XEN_ASSERT_TYPE(XEN_gdouble_P(position), position, 4, "gtk_ruler_set_range", "gdouble");
  XEN_ASSERT_TYPE(XEN_gdouble_P(max_size), max_size, 5, "gtk_ruler_set_range", "gdouble");
  gtk_ruler_set_range(XEN_TO_C_GtkRuler_(ruler), XEN_TO_C_gdouble(lower), XEN_TO_C_gdouble(upper), XEN_TO_C_gdouble(position), 
                      XEN_TO_C_gdouble(max_size));
  return(XEN_FALSE);
}

static XEN gxg_gtk_ruler_draw_ticks(XEN ruler)
{
  #define H_gtk_ruler_draw_ticks "void gtk_ruler_draw_ticks(GtkRuler* ruler)"
  XEN_ASSERT_TYPE(XEN_GtkRuler__P(ruler), ruler, 1, "gtk_ruler_draw_ticks", "GtkRuler*");
  gtk_ruler_draw_ticks(XEN_TO_C_GtkRuler_(ruler));
  return(XEN_FALSE);
}

static XEN gxg_gtk_ruler_draw_pos(XEN ruler)
{
  #define H_gtk_ruler_draw_pos "void gtk_ruler_draw_pos(GtkRuler* ruler)"
  XEN_ASSERT_TYPE(XEN_GtkRuler__P(ruler), ruler, 1, "gtk_ruler_draw_pos", "GtkRuler*");
  gtk_ruler_draw_pos(XEN_TO_C_GtkRuler_(ruler));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_ruler_get_metric(XEN ruler)
{
  #define H_gtk_ruler_get_metric "GtkMetricType gtk_ruler_get_metric(GtkRuler* ruler)"
  XEN_ASSERT_TYPE(XEN_GtkRuler__P(ruler), ruler, 1, "gtk_ruler_get_metric", "GtkRuler*");
  return(C_TO_XEN_GtkMetricType(gtk_ruler_get_metric(XEN_TO_C_GtkRuler_(ruler))));
}

static XEN gxg_gtk_ruler_get_range(XEN ruler, XEN lower, XEN upper, XEN position, XEN max_size)
{
  #define H_gtk_ruler_get_range "void gtk_ruler_get_range(GtkRuler* ruler, gdouble* [lower], gdouble* [upper], \
gdouble* [position], gdouble* [max_size])"
  gdouble ref_lower;
  gdouble ref_upper;
  gdouble ref_position;
  gdouble ref_max_size;
  XEN_ASSERT_TYPE(XEN_GtkRuler__P(ruler), ruler, 1, "gtk_ruler_get_range", "GtkRuler*");
  gtk_ruler_get_range(XEN_TO_C_GtkRuler_(ruler), &ref_lower, &ref_upper, &ref_position, &ref_max_size);
  return(XEN_LIST_4(C_TO_XEN_gdouble(ref_lower), C_TO_XEN_gdouble(ref_upper), C_TO_XEN_gdouble(ref_position), C_TO_XEN_gdouble(ref_max_size)));
}
#endif

static XEN gxg_gtk_scale_get_type(void)
{
  #define H_gtk_scale_get_type "GtkType gtk_scale_get_type( void)"
  return(C_TO_XEN_GtkType(gtk_scale_get_type()));
}

static XEN gxg_gtk_scale_set_digits(XEN scale, XEN digits)
{
  #define H_gtk_scale_set_digits "void gtk_scale_set_digits(GtkScale* scale, gint digits)"
  XEN_ASSERT_TYPE(XEN_GtkScale__P(scale), scale, 1, "gtk_scale_set_digits", "GtkScale*");
  XEN_ASSERT_TYPE(XEN_gint_P(digits), digits, 2, "gtk_scale_set_digits", "gint");
  gtk_scale_set_digits(XEN_TO_C_GtkScale_(scale), XEN_TO_C_gint(digits));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_scale_get_digits(XEN scale)
{
  #define H_gtk_scale_get_digits "gint gtk_scale_get_digits(GtkScale* scale)"
  XEN_ASSERT_TYPE(XEN_GtkScale__P(scale), scale, 1, "gtk_scale_get_digits", "GtkScale*");
  return(C_TO_XEN_gint(gtk_scale_get_digits(XEN_TO_C_GtkScale_(scale))));
}
#endif

static XEN gxg_gtk_scale_set_draw_value(XEN scale, XEN draw_value)
{
  #define H_gtk_scale_set_draw_value "void gtk_scale_set_draw_value(GtkScale* scale, gboolean draw_value)"
  XEN_ASSERT_TYPE(XEN_GtkScale__P(scale), scale, 1, "gtk_scale_set_draw_value", "GtkScale*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(draw_value), draw_value, 2, "gtk_scale_set_draw_value", "gboolean");
  gtk_scale_set_draw_value(XEN_TO_C_GtkScale_(scale), XEN_TO_C_gboolean(draw_value));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_scale_get_draw_value(XEN scale)
{
  #define H_gtk_scale_get_draw_value "gboolean gtk_scale_get_draw_value(GtkScale* scale)"
  XEN_ASSERT_TYPE(XEN_GtkScale__P(scale), scale, 1, "gtk_scale_get_draw_value", "GtkScale*");
  return(C_TO_XEN_gboolean(gtk_scale_get_draw_value(XEN_TO_C_GtkScale_(scale))));
}
#endif

static XEN gxg_gtk_scale_set_value_pos(XEN scale, XEN pos)
{
  #define H_gtk_scale_set_value_pos "void gtk_scale_set_value_pos(GtkScale* scale, GtkPositionType pos)"
  XEN_ASSERT_TYPE(XEN_GtkScale__P(scale), scale, 1, "gtk_scale_set_value_pos", "GtkScale*");
  XEN_ASSERT_TYPE(XEN_GtkPositionType_P(pos), pos, 2, "gtk_scale_set_value_pos", "GtkPositionType");
  gtk_scale_set_value_pos(XEN_TO_C_GtkScale_(scale), XEN_TO_C_GtkPositionType(pos));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_scale_get_value_pos(XEN scale)
{
  #define H_gtk_scale_get_value_pos "GtkPositionType gtk_scale_get_value_pos(GtkScale* scale)"
  XEN_ASSERT_TYPE(XEN_GtkScale__P(scale), scale, 1, "gtk_scale_get_value_pos", "GtkScale*");
  return(C_TO_XEN_GtkPositionType(gtk_scale_get_value_pos(XEN_TO_C_GtkScale_(scale))));
}
#endif

static XEN gxg_gtk_scrollbar_get_type(void)
{
  #define H_gtk_scrollbar_get_type "GtkType gtk_scrollbar_get_type( void)"
  return(C_TO_XEN_GtkType(gtk_scrollbar_get_type()));
}

static XEN gxg_gtk_scrolled_window_get_type(void)
{
  #define H_gtk_scrolled_window_get_type "GtkType gtk_scrolled_window_get_type( void)"
  return(C_TO_XEN_GtkType(gtk_scrolled_window_get_type()));
}

static XEN gxg_gtk_scrolled_window_new(XEN hadjustment, XEN vadjustment)
{
  #define H_gtk_scrolled_window_new "GtkWidget* gtk_scrolled_window_new(GtkAdjustment* hadjustment, GtkAdjustment* vadjustment)"
  XEN_ASSERT_TYPE(XEN_GtkAdjustment__P(hadjustment), hadjustment, 1, "gtk_scrolled_window_new", "GtkAdjustment*");
  XEN_ASSERT_TYPE(XEN_GtkAdjustment__P(vadjustment), vadjustment, 2, "gtk_scrolled_window_new", "GtkAdjustment*");
  return(C_TO_XEN_GtkWidget_(gtk_scrolled_window_new(XEN_TO_C_GtkAdjustment_(hadjustment), XEN_TO_C_GtkAdjustment_(vadjustment))));
}

static XEN gxg_gtk_scrolled_window_set_hadjustment(XEN scrolled_window, XEN hadjustment)
{
  #define H_gtk_scrolled_window_set_hadjustment "void gtk_scrolled_window_set_hadjustment(GtkScrolledWindow* scrolled_window, \
GtkAdjustment* hadjustment)"
  XEN_ASSERT_TYPE(XEN_GtkScrolledWindow__P(scrolled_window), scrolled_window, 1, "gtk_scrolled_window_set_hadjustment", "GtkScrolledWindow*");
  XEN_ASSERT_TYPE(XEN_GtkAdjustment__P(hadjustment), hadjustment, 2, "gtk_scrolled_window_set_hadjustment", "GtkAdjustment*");
  gtk_scrolled_window_set_hadjustment(XEN_TO_C_GtkScrolledWindow_(scrolled_window), XEN_TO_C_GtkAdjustment_(hadjustment));
  return(XEN_FALSE);
}

static XEN gxg_gtk_scrolled_window_set_vadjustment(XEN scrolled_window, XEN hadjustment)
{
  #define H_gtk_scrolled_window_set_vadjustment "void gtk_scrolled_window_set_vadjustment(GtkScrolledWindow* scrolled_window, \
GtkAdjustment* hadjustment)"
  XEN_ASSERT_TYPE(XEN_GtkScrolledWindow__P(scrolled_window), scrolled_window, 1, "gtk_scrolled_window_set_vadjustment", "GtkScrolledWindow*");
  XEN_ASSERT_TYPE(XEN_GtkAdjustment__P(hadjustment), hadjustment, 2, "gtk_scrolled_window_set_vadjustment", "GtkAdjustment*");
  gtk_scrolled_window_set_vadjustment(XEN_TO_C_GtkScrolledWindow_(scrolled_window), XEN_TO_C_GtkAdjustment_(hadjustment));
  return(XEN_FALSE);
}

static XEN gxg_gtk_scrolled_window_get_hadjustment(XEN scrolled_window)
{
  #define H_gtk_scrolled_window_get_hadjustment "GtkAdjustment* gtk_scrolled_window_get_hadjustment(GtkScrolledWindow* scrolled_window)"
  XEN_ASSERT_TYPE(XEN_GtkScrolledWindow__P(scrolled_window), scrolled_window, 1, "gtk_scrolled_window_get_hadjustment", "GtkScrolledWindow*");
  return(C_TO_XEN_GtkAdjustment_(gtk_scrolled_window_get_hadjustment(XEN_TO_C_GtkScrolledWindow_(scrolled_window))));
}

static XEN gxg_gtk_scrolled_window_get_vadjustment(XEN scrolled_window)
{
  #define H_gtk_scrolled_window_get_vadjustment "GtkAdjustment* gtk_scrolled_window_get_vadjustment(GtkScrolledWindow* scrolled_window)"
  XEN_ASSERT_TYPE(XEN_GtkScrolledWindow__P(scrolled_window), scrolled_window, 1, "gtk_scrolled_window_get_vadjustment", "GtkScrolledWindow*");
  return(C_TO_XEN_GtkAdjustment_(gtk_scrolled_window_get_vadjustment(XEN_TO_C_GtkScrolledWindow_(scrolled_window))));
}

static XEN gxg_gtk_scrolled_window_set_policy(XEN scrolled_window, XEN hscrollbar_policy, XEN vscrollbar_policy)
{
  #define H_gtk_scrolled_window_set_policy "void gtk_scrolled_window_set_policy(GtkScrolledWindow* scrolled_window, \
GtkPolicyType hscrollbar_policy, GtkPolicyType vscrollbar_policy)"
  XEN_ASSERT_TYPE(XEN_GtkScrolledWindow__P(scrolled_window), scrolled_window, 1, "gtk_scrolled_window_set_policy", "GtkScrolledWindow*");
  XEN_ASSERT_TYPE(XEN_GtkPolicyType_P(hscrollbar_policy), hscrollbar_policy, 2, "gtk_scrolled_window_set_policy", "GtkPolicyType");
  XEN_ASSERT_TYPE(XEN_GtkPolicyType_P(vscrollbar_policy), vscrollbar_policy, 3, "gtk_scrolled_window_set_policy", "GtkPolicyType");
  gtk_scrolled_window_set_policy(XEN_TO_C_GtkScrolledWindow_(scrolled_window), XEN_TO_C_GtkPolicyType(hscrollbar_policy), 
                                 XEN_TO_C_GtkPolicyType(vscrollbar_policy));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_scrolled_window_get_policy(XEN scrolled_window, XEN hscrollbar_policy, XEN vscrollbar_policy)
{
  #define H_gtk_scrolled_window_get_policy "void gtk_scrolled_window_get_policy(GtkScrolledWindow* scrolled_window, \
GtkPolicyType* [hscrollbar_policy], GtkPolicyType* [vscrollbar_policy])"
  GtkPolicyType ref_hscrollbar_policy;
  GtkPolicyType ref_vscrollbar_policy;
  XEN_ASSERT_TYPE(XEN_GtkScrolledWindow__P(scrolled_window), scrolled_window, 1, "gtk_scrolled_window_get_policy", "GtkScrolledWindow*");
  gtk_scrolled_window_get_policy(XEN_TO_C_GtkScrolledWindow_(scrolled_window), &ref_hscrollbar_policy, &ref_vscrollbar_policy);
  return(XEN_LIST_2(C_TO_XEN_GtkPolicyType(ref_hscrollbar_policy), C_TO_XEN_GtkPolicyType(ref_vscrollbar_policy)));
}
#endif

static XEN gxg_gtk_scrolled_window_set_placement(XEN scrolled_window, XEN window_placement)
{
  #define H_gtk_scrolled_window_set_placement "void gtk_scrolled_window_set_placement(GtkScrolledWindow* scrolled_window, \
GtkCornerType window_placement)"
  XEN_ASSERT_TYPE(XEN_GtkScrolledWindow__P(scrolled_window), scrolled_window, 1, "gtk_scrolled_window_set_placement", "GtkScrolledWindow*");
  XEN_ASSERT_TYPE(XEN_GtkCornerType_P(window_placement), window_placement, 2, "gtk_scrolled_window_set_placement", "GtkCornerType");
  gtk_scrolled_window_set_placement(XEN_TO_C_GtkScrolledWindow_(scrolled_window), XEN_TO_C_GtkCornerType(window_placement));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_scrolled_window_get_placement(XEN scrolled_window)
{
  #define H_gtk_scrolled_window_get_placement "GtkCornerType gtk_scrolled_window_get_placement(GtkScrolledWindow* scrolled_window)"
  XEN_ASSERT_TYPE(XEN_GtkScrolledWindow__P(scrolled_window), scrolled_window, 1, "gtk_scrolled_window_get_placement", "GtkScrolledWindow*");
  return(C_TO_XEN_GtkCornerType(gtk_scrolled_window_get_placement(XEN_TO_C_GtkScrolledWindow_(scrolled_window))));
}

static XEN gxg_gtk_scrolled_window_set_shadow_type(XEN scrolled_window, XEN type)
{
  #define H_gtk_scrolled_window_set_shadow_type "void gtk_scrolled_window_set_shadow_type(GtkScrolledWindow* scrolled_window, \
GtkShadowType type)"
  XEN_ASSERT_TYPE(XEN_GtkScrolledWindow__P(scrolled_window), scrolled_window, 1, "gtk_scrolled_window_set_shadow_type", "GtkScrolledWindow*");
  XEN_ASSERT_TYPE(XEN_GtkShadowType_P(type), type, 2, "gtk_scrolled_window_set_shadow_type", "GtkShadowType");
  gtk_scrolled_window_set_shadow_type(XEN_TO_C_GtkScrolledWindow_(scrolled_window), XEN_TO_C_GtkShadowType(type));
  return(XEN_FALSE);
}

static XEN gxg_gtk_scrolled_window_get_shadow_type(XEN scrolled_window)
{
  #define H_gtk_scrolled_window_get_shadow_type "GtkShadowType gtk_scrolled_window_get_shadow_type(GtkScrolledWindow* scrolled_window)"
  XEN_ASSERT_TYPE(XEN_GtkScrolledWindow__P(scrolled_window), scrolled_window, 1, "gtk_scrolled_window_get_shadow_type", "GtkScrolledWindow*");
  return(C_TO_XEN_GtkShadowType(gtk_scrolled_window_get_shadow_type(XEN_TO_C_GtkScrolledWindow_(scrolled_window))));
}
#endif

static XEN gxg_gtk_scrolled_window_add_with_viewport(XEN scrolled_window, XEN child)
{
  #define H_gtk_scrolled_window_add_with_viewport "void gtk_scrolled_window_add_with_viewport(GtkScrolledWindow* scrolled_window, \
GtkWidget* child)"
  XEN_ASSERT_TYPE(XEN_GtkScrolledWindow__P(scrolled_window), scrolled_window, 1, "gtk_scrolled_window_add_with_viewport", "GtkScrolledWindow*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(child), child, 2, "gtk_scrolled_window_add_with_viewport", "GtkWidget*");
  gtk_scrolled_window_add_with_viewport(XEN_TO_C_GtkScrolledWindow_(scrolled_window), XEN_TO_C_GtkWidget_(child));
  return(XEN_FALSE);
}

static XEN gxg_gtk_target_list_new(XEN targets, XEN ntargets)
{
  #define H_gtk_target_list_new "GtkTargetList* gtk_target_list_new(GtkTargetEntry* targets, guint ntargets)"
  XEN_ASSERT_TYPE(XEN_GtkTargetEntry__P(targets), targets, 1, "gtk_target_list_new", "GtkTargetEntry*");
  XEN_ASSERT_TYPE(XEN_guint_P(ntargets), ntargets, 2, "gtk_target_list_new", "guint");
  return(C_TO_XEN_GtkTargetList_(gtk_target_list_new(XEN_TO_C_GtkTargetEntry_(targets), XEN_TO_C_guint(ntargets))));
}

static XEN gxg_gtk_target_list_ref(XEN list)
{
  #define H_gtk_target_list_ref "void gtk_target_list_ref(GtkTargetList* list)"
  XEN_ASSERT_TYPE(XEN_GtkTargetList__P(list), list, 1, "gtk_target_list_ref", "GtkTargetList*");
  gtk_target_list_ref(XEN_TO_C_GtkTargetList_(list));
  return(XEN_FALSE);
}

static XEN gxg_gtk_target_list_unref(XEN list)
{
  #define H_gtk_target_list_unref "void gtk_target_list_unref(GtkTargetList* list)"
  XEN_ASSERT_TYPE(XEN_GtkTargetList__P(list), list, 1, "gtk_target_list_unref", "GtkTargetList*");
  gtk_target_list_unref(XEN_TO_C_GtkTargetList_(list));
  return(XEN_FALSE);
}

static XEN gxg_gtk_target_list_add(XEN list, XEN target, XEN flags, XEN info)
{
  #define H_gtk_target_list_add "void gtk_target_list_add(GtkTargetList* list, GdkAtom target, guint flags, \
guint info)"
  XEN_ASSERT_TYPE(XEN_GtkTargetList__P(list), list, 1, "gtk_target_list_add", "GtkTargetList*");
  XEN_ASSERT_TYPE(XEN_GdkAtom_P(target), target, 2, "gtk_target_list_add", "GdkAtom");
  XEN_ASSERT_TYPE(XEN_guint_P(flags), flags, 3, "gtk_target_list_add", "guint");
  XEN_ASSERT_TYPE(XEN_guint_P(info), info, 4, "gtk_target_list_add", "guint");
  gtk_target_list_add(XEN_TO_C_GtkTargetList_(list), XEN_TO_C_GdkAtom(target), XEN_TO_C_guint(flags), XEN_TO_C_guint(info));
  return(XEN_FALSE);
}

static XEN gxg_gtk_target_list_add_table(XEN list, XEN targets, XEN ntargets)
{
  #define H_gtk_target_list_add_table "void gtk_target_list_add_table(GtkTargetList* list, GtkTargetEntry* targets, \
guint ntargets)"
  XEN_ASSERT_TYPE(XEN_GtkTargetList__P(list), list, 1, "gtk_target_list_add_table", "GtkTargetList*");
  XEN_ASSERT_TYPE(XEN_GtkTargetEntry__P(targets), targets, 2, "gtk_target_list_add_table", "GtkTargetEntry*");
  XEN_ASSERT_TYPE(XEN_guint_P(ntargets), ntargets, 3, "gtk_target_list_add_table", "guint");
  gtk_target_list_add_table(XEN_TO_C_GtkTargetList_(list), XEN_TO_C_GtkTargetEntry_(targets), XEN_TO_C_guint(ntargets));
  return(XEN_FALSE);
}

static XEN gxg_gtk_target_list_remove(XEN list, XEN target)
{
  #define H_gtk_target_list_remove "void gtk_target_list_remove(GtkTargetList* list, GdkAtom target)"
  XEN_ASSERT_TYPE(XEN_GtkTargetList__P(list), list, 1, "gtk_target_list_remove", "GtkTargetList*");
  XEN_ASSERT_TYPE(XEN_GdkAtom_P(target), target, 2, "gtk_target_list_remove", "GdkAtom");
  gtk_target_list_remove(XEN_TO_C_GtkTargetList_(list), XEN_TO_C_GdkAtom(target));
  return(XEN_FALSE);
}

static XEN gxg_gtk_target_list_find(XEN list, XEN target, XEN info)
{
  #define H_gtk_target_list_find "gboolean gtk_target_list_find(GtkTargetList* list, GdkAtom target, \
guint* info)"
  XEN_ASSERT_TYPE(XEN_GtkTargetList__P(list), list, 1, "gtk_target_list_find", "GtkTargetList*");
  XEN_ASSERT_TYPE(XEN_GdkAtom_P(target), target, 2, "gtk_target_list_find", "GdkAtom");
  XEN_ASSERT_TYPE(XEN_guint__P(info), info, 3, "gtk_target_list_find", "guint*");
  return(C_TO_XEN_gboolean(gtk_target_list_find(XEN_TO_C_GtkTargetList_(list), XEN_TO_C_GdkAtom(target), XEN_TO_C_guint_(info))));
}

static XEN gxg_gtk_selection_owner_set(XEN widget, XEN selection, XEN time)
{
  #define H_gtk_selection_owner_set "gboolean gtk_selection_owner_set(GtkWidget* widget, GdkAtom selection, \
guint32 time)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_selection_owner_set", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_GdkAtom_P(selection), selection, 2, "gtk_selection_owner_set", "GdkAtom");
  XEN_ASSERT_TYPE(XEN_guint32_P(time), time, 3, "gtk_selection_owner_set", "guint32");
  return(C_TO_XEN_gboolean(gtk_selection_owner_set(XEN_TO_C_GtkWidget_(widget), XEN_TO_C_GdkAtom(selection), XEN_TO_C_guint32(time))));
}

static XEN gxg_gtk_selection_add_target(XEN widget, XEN selection, XEN target, XEN info)
{
  #define H_gtk_selection_add_target "void gtk_selection_add_target(GtkWidget* widget, GdkAtom selection, \
GdkAtom target, guint info)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_selection_add_target", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_GdkAtom_P(selection), selection, 2, "gtk_selection_add_target", "GdkAtom");
  XEN_ASSERT_TYPE(XEN_GdkAtom_P(target), target, 3, "gtk_selection_add_target", "GdkAtom");
  XEN_ASSERT_TYPE(XEN_guint_P(info), info, 4, "gtk_selection_add_target", "guint");
  gtk_selection_add_target(XEN_TO_C_GtkWidget_(widget), XEN_TO_C_GdkAtom(selection), XEN_TO_C_GdkAtom(target), XEN_TO_C_guint(info));
  return(XEN_FALSE);
}

static XEN gxg_gtk_selection_add_targets(XEN widget, XEN selection, XEN targets, XEN ntargets)
{
  #define H_gtk_selection_add_targets "void gtk_selection_add_targets(GtkWidget* widget, GdkAtom selection, \
GtkTargetEntry* targets, guint ntargets)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_selection_add_targets", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_GdkAtom_P(selection), selection, 2, "gtk_selection_add_targets", "GdkAtom");
  XEN_ASSERT_TYPE(XEN_GtkTargetEntry__P(targets), targets, 3, "gtk_selection_add_targets", "GtkTargetEntry*");
  XEN_ASSERT_TYPE(XEN_guint_P(ntargets), ntargets, 4, "gtk_selection_add_targets", "guint");
  gtk_selection_add_targets(XEN_TO_C_GtkWidget_(widget), XEN_TO_C_GdkAtom(selection), XEN_TO_C_GtkTargetEntry_(targets), 
                            XEN_TO_C_guint(ntargets));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_selection_clear_targets(XEN widget, XEN selection)
{
  #define H_gtk_selection_clear_targets "void gtk_selection_clear_targets(GtkWidget* widget, GdkAtom selection)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_selection_clear_targets", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_GdkAtom_P(selection), selection, 2, "gtk_selection_clear_targets", "GdkAtom");
  gtk_selection_clear_targets(XEN_TO_C_GtkWidget_(widget), XEN_TO_C_GdkAtom(selection));
  return(XEN_FALSE);
}
#endif

static XEN gxg_gtk_selection_convert(XEN widget, XEN selection, XEN target, XEN time)
{
  #define H_gtk_selection_convert "gboolean gtk_selection_convert(GtkWidget* widget, GdkAtom selection, \
GdkAtom target, guint32 time)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_selection_convert", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_GdkAtom_P(selection), selection, 2, "gtk_selection_convert", "GdkAtom");
  XEN_ASSERT_TYPE(XEN_GdkAtom_P(target), target, 3, "gtk_selection_convert", "GdkAtom");
  XEN_ASSERT_TYPE(XEN_guint32_P(time), time, 4, "gtk_selection_convert", "guint32");
  return(C_TO_XEN_gboolean(gtk_selection_convert(XEN_TO_C_GtkWidget_(widget), XEN_TO_C_GdkAtom(selection), XEN_TO_C_GdkAtom(target), 
                                                 XEN_TO_C_guint32(time))));
}

static XEN gxg_gtk_selection_data_set(XEN selection_data, XEN type, XEN format, XEN data, XEN length)
{
  #define H_gtk_selection_data_set "void gtk_selection_data_set(GtkSelectionData* selection_data, GdkAtom type, \
gint format, guchar* data, gint length)"
  XEN_ASSERT_TYPE(XEN_GtkSelectionData__P(selection_data), selection_data, 1, "gtk_selection_data_set", "GtkSelectionData*");
  XEN_ASSERT_TYPE(XEN_GdkAtom_P(type), type, 2, "gtk_selection_data_set", "GdkAtom");
  XEN_ASSERT_TYPE(XEN_gint_P(format), format, 3, "gtk_selection_data_set", "gint");
  XEN_ASSERT_TYPE(XEN_guchar__P(data), data, 4, "gtk_selection_data_set", "guchar*");
  XEN_ASSERT_TYPE(XEN_gint_P(length), length, 5, "gtk_selection_data_set", "gint");
  gtk_selection_data_set(XEN_TO_C_GtkSelectionData_(selection_data), XEN_TO_C_GdkAtom(type), XEN_TO_C_gint(format), XEN_TO_C_guchar_(data), 
                         XEN_TO_C_gint(length));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_selection_data_set_text(XEN selection_data, XEN str, XEN len)
{
  #define H_gtk_selection_data_set_text "gboolean gtk_selection_data_set_text(GtkSelectionData* selection_data, \
gchar* str, gint len)"
  XEN_ASSERT_TYPE(XEN_GtkSelectionData__P(selection_data), selection_data, 1, "gtk_selection_data_set_text", "GtkSelectionData*");
  XEN_ASSERT_TYPE(XEN_gchar__P(str), str, 2, "gtk_selection_data_set_text", "gchar*");
  XEN_ASSERT_TYPE(XEN_gint_P(len), len, 3, "gtk_selection_data_set_text", "gint");
  return(C_TO_XEN_gboolean(gtk_selection_data_set_text(XEN_TO_C_GtkSelectionData_(selection_data), XEN_TO_C_gchar_(str), 
                                                       XEN_TO_C_gint(len))));
}

static XEN gxg_gtk_selection_data_get_text(XEN selection_data)
{
  #define H_gtk_selection_data_get_text "guchar* gtk_selection_data_get_text(GtkSelectionData* selection_data)"
  XEN_ASSERT_TYPE(XEN_GtkSelectionData__P(selection_data), selection_data, 1, "gtk_selection_data_get_text", "GtkSelectionData*");
  return(C_TO_XEN_guchar_(gtk_selection_data_get_text(XEN_TO_C_GtkSelectionData_(selection_data))));
}

static XEN gxg_gtk_selection_data_get_targets(XEN selection_data, XEN targets, XEN n_atoms)
{
  #define H_gtk_selection_data_get_targets "gboolean gtk_selection_data_get_targets(GtkSelectionData* selection_data, \
GdkAtom** [targets], gint* [n_atoms])"
  GdkAtom* ref_targets;
  gint ref_n_atoms;
  XEN_ASSERT_TYPE(XEN_GtkSelectionData__P(selection_data), selection_data, 1, "gtk_selection_data_get_targets", "GtkSelectionData*");
  {
    XEN result = XEN_FALSE;
    result = C_TO_XEN_gboolean(gtk_selection_data_get_targets(XEN_TO_C_GtkSelectionData_(selection_data), &ref_targets, &ref_n_atoms));
    return(XEN_LIST_3(result, C_TO_XEN_GdkAtom_(ref_targets), C_TO_XEN_gint(ref_n_atoms)));
   }
}

static XEN gxg_gtk_selection_data_targets_include_text(XEN selection_data)
{
  #define H_gtk_selection_data_targets_include_text "gboolean gtk_selection_data_targets_include_text(GtkSelectionData* selection_data)"
  XEN_ASSERT_TYPE(XEN_GtkSelectionData__P(selection_data), selection_data, 1, "gtk_selection_data_targets_include_text", "GtkSelectionData*");
  return(C_TO_XEN_gboolean(gtk_selection_data_targets_include_text(XEN_TO_C_GtkSelectionData_(selection_data))));
}
#endif

static XEN gxg_gtk_selection_remove_all(XEN widget)
{
  #define H_gtk_selection_remove_all "void gtk_selection_remove_all(GtkWidget* widget)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_selection_remove_all", "GtkWidget*");
  gtk_selection_remove_all(XEN_TO_C_GtkWidget_(widget));
  return(XEN_FALSE);
}

static XEN gxg_gtk_selection_clear(XEN widget, XEN event)
{
  #define H_gtk_selection_clear "gboolean gtk_selection_clear(GtkWidget* widget, GdkEventSelection* event)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_selection_clear", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_GdkEventSelection__P(event), event, 2, "gtk_selection_clear", "GdkEventSelection*");
  return(C_TO_XEN_gboolean(gtk_selection_clear(XEN_TO_C_GtkWidget_(widget), XEN_TO_C_GdkEventSelection_(event))));
}

static XEN gxg_gtk_selection_request(XEN widget, XEN event)
{
  #define H_gtk_selection_request "gboolean gtk_selection_request(GtkWidget* widget, GdkEventSelection* event)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_selection_request", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_GdkEventSelection__P(event), event, 2, "gtk_selection_request", "GdkEventSelection*");
  return(C_TO_XEN_gboolean(gtk_selection_request(XEN_TO_C_GtkWidget_(widget), XEN_TO_C_GdkEventSelection_(event))));
}

static XEN gxg_gtk_selection_incr_event(XEN window, XEN event)
{
  #define H_gtk_selection_incr_event "gboolean gtk_selection_incr_event(GdkWindow* window, GdkEventProperty* event)"
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 1, "gtk_selection_incr_event", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_GdkEventProperty__P(event), event, 2, "gtk_selection_incr_event", "GdkEventProperty*");
  return(C_TO_XEN_gboolean(gtk_selection_incr_event(XEN_TO_C_GdkWindow_(window), XEN_TO_C_GdkEventProperty_(event))));
}

static XEN gxg_gtk_selection_notify(XEN widget, XEN event)
{
  #define H_gtk_selection_notify "gboolean gtk_selection_notify(GtkWidget* widget, GdkEventSelection* event)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_selection_notify", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_GdkEventSelection__P(event), event, 2, "gtk_selection_notify", "GdkEventSelection*");
  return(C_TO_XEN_gboolean(gtk_selection_notify(XEN_TO_C_GtkWidget_(widget), XEN_TO_C_GdkEventSelection_(event))));
}

static XEN gxg_gtk_selection_property_notify(XEN widget, XEN event)
{
  #define H_gtk_selection_property_notify "gboolean gtk_selection_property_notify(GtkWidget* widget, \
GdkEventProperty* event)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_selection_property_notify", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_GdkEventProperty__P(event), event, 2, "gtk_selection_property_notify", "GdkEventProperty*");
  return(C_TO_XEN_gboolean(gtk_selection_property_notify(XEN_TO_C_GtkWidget_(widget), XEN_TO_C_GdkEventProperty_(event))));
}

static XEN gxg_gtk_selection_data_copy(XEN data)
{
  #define H_gtk_selection_data_copy "GtkSelectionData* gtk_selection_data_copy(GtkSelectionData* data)"
  XEN_ASSERT_TYPE(XEN_GtkSelectionData__P(data), data, 1, "gtk_selection_data_copy", "GtkSelectionData*");
  return(C_TO_XEN_GtkSelectionData_(gtk_selection_data_copy(XEN_TO_C_GtkSelectionData_(data))));
}

static XEN gxg_gtk_selection_data_free(XEN data)
{
  #define H_gtk_selection_data_free "void gtk_selection_data_free(GtkSelectionData* data)"
  XEN_ASSERT_TYPE(XEN_GtkSelectionData__P(data), data, 1, "gtk_selection_data_free", "GtkSelectionData*");
  gtk_selection_data_free(XEN_TO_C_GtkSelectionData_(data));
  return(XEN_FALSE);
}

static XEN gxg_gtk_separator_get_type(void)
{
  #define H_gtk_separator_get_type "GtkType gtk_separator_get_type( void)"
  return(C_TO_XEN_GtkType(gtk_separator_get_type()));
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_separator_menu_item_get_type(void)
{
  #define H_gtk_separator_menu_item_get_type "GType gtk_separator_menu_item_get_type( void)"
  return(C_TO_XEN_GType(gtk_separator_menu_item_get_type()));
}

static XEN gxg_gtk_separator_menu_item_new(void)
{
  #define H_gtk_separator_menu_item_new "GtkWidget* gtk_separator_menu_item_new( void)"
  return(C_TO_XEN_GtkWidget_(gtk_separator_menu_item_new()));
}

static XEN gxg_gtk_settings_get_type(void)
{
  #define H_gtk_settings_get_type "GType gtk_settings_get_type( void)"
  return(C_TO_XEN_GType(gtk_settings_get_type()));
}

static XEN gxg_gtk_settings_get_default(void)
{
  #define H_gtk_settings_get_default "GtkSettings* gtk_settings_get_default( void)"
  return(C_TO_XEN_GtkSettings_(gtk_settings_get_default()));
}

static XEN gxg_gtk_settings_install_property(XEN pspec)
{
  #define H_gtk_settings_install_property "void gtk_settings_install_property(GParamSpec* pspec)"
  XEN_ASSERT_TYPE(XEN_GParamSpec__P(pspec), pspec, 1, "gtk_settings_install_property", "GParamSpec*");
  gtk_settings_install_property(XEN_TO_C_GParamSpec_(pspec));
  return(XEN_FALSE);
}

static XEN gxg_gtk_settings_install_property_parser(XEN pspec, XEN parser)
{
  #define H_gtk_settings_install_property_parser "void gtk_settings_install_property_parser(GParamSpec* pspec, \
GtkRcPropertyParser parser)"
  XEN_ASSERT_TYPE(XEN_GParamSpec__P(pspec), pspec, 1, "gtk_settings_install_property_parser", "GParamSpec*");
  XEN_ASSERT_TYPE(XEN_GtkRcPropertyParser_P(parser), parser, 2, "gtk_settings_install_property_parser", "GtkRcPropertyParser");
  gtk_settings_install_property_parser(XEN_TO_C_GParamSpec_(pspec), XEN_TO_C_GtkRcPropertyParser(parser));
  return(XEN_FALSE);
}

static XEN gxg_gtk_rc_property_parse_color(XEN pspec, XEN gstring, XEN property_value)
{
  #define H_gtk_rc_property_parse_color "gboolean gtk_rc_property_parse_color(GParamSpec* pspec, GString* gstring, \
GValue* property_value)"
  XEN_ASSERT_TYPE(XEN_GParamSpec__P(pspec), pspec, 1, "gtk_rc_property_parse_color", "GParamSpec*");
  XEN_ASSERT_TYPE(XEN_GString__P(gstring), gstring, 2, "gtk_rc_property_parse_color", "GString*");
  XEN_ASSERT_TYPE(XEN_GValue__P(property_value), property_value, 3, "gtk_rc_property_parse_color", "GValue*");
  return(C_TO_XEN_gboolean(gtk_rc_property_parse_color(XEN_TO_C_GParamSpec_(pspec), XEN_TO_C_GString_(gstring), XEN_TO_C_GValue_(property_value))));
}

static XEN gxg_gtk_rc_property_parse_enum(XEN pspec, XEN gstring, XEN property_value)
{
  #define H_gtk_rc_property_parse_enum "gboolean gtk_rc_property_parse_enum(GParamSpec* pspec, GString* gstring, \
GValue* property_value)"
  XEN_ASSERT_TYPE(XEN_GParamSpec__P(pspec), pspec, 1, "gtk_rc_property_parse_enum", "GParamSpec*");
  XEN_ASSERT_TYPE(XEN_GString__P(gstring), gstring, 2, "gtk_rc_property_parse_enum", "GString*");
  XEN_ASSERT_TYPE(XEN_GValue__P(property_value), property_value, 3, "gtk_rc_property_parse_enum", "GValue*");
  return(C_TO_XEN_gboolean(gtk_rc_property_parse_enum(XEN_TO_C_GParamSpec_(pspec), XEN_TO_C_GString_(gstring), XEN_TO_C_GValue_(property_value))));
}

static XEN gxg_gtk_rc_property_parse_flags(XEN pspec, XEN gstring, XEN property_value)
{
  #define H_gtk_rc_property_parse_flags "gboolean gtk_rc_property_parse_flags(GParamSpec* pspec, GString* gstring, \
GValue* property_value)"
  XEN_ASSERT_TYPE(XEN_GParamSpec__P(pspec), pspec, 1, "gtk_rc_property_parse_flags", "GParamSpec*");
  XEN_ASSERT_TYPE(XEN_GString__P(gstring), gstring, 2, "gtk_rc_property_parse_flags", "GString*");
  XEN_ASSERT_TYPE(XEN_GValue__P(property_value), property_value, 3, "gtk_rc_property_parse_flags", "GValue*");
  return(C_TO_XEN_gboolean(gtk_rc_property_parse_flags(XEN_TO_C_GParamSpec_(pspec), XEN_TO_C_GString_(gstring), XEN_TO_C_GValue_(property_value))));
}

static XEN gxg_gtk_rc_property_parse_requisition(XEN pspec, XEN gstring, XEN property_value)
{
  #define H_gtk_rc_property_parse_requisition "gboolean gtk_rc_property_parse_requisition(GParamSpec* pspec, \
GString* gstring, GValue* property_value)"
  XEN_ASSERT_TYPE(XEN_GParamSpec__P(pspec), pspec, 1, "gtk_rc_property_parse_requisition", "GParamSpec*");
  XEN_ASSERT_TYPE(XEN_GString__P(gstring), gstring, 2, "gtk_rc_property_parse_requisition", "GString*");
  XEN_ASSERT_TYPE(XEN_GValue__P(property_value), property_value, 3, "gtk_rc_property_parse_requisition", "GValue*");
  return(C_TO_XEN_gboolean(gtk_rc_property_parse_requisition(XEN_TO_C_GParamSpec_(pspec), XEN_TO_C_GString_(gstring), XEN_TO_C_GValue_(property_value))));
}

static XEN gxg_gtk_rc_property_parse_border(XEN pspec, XEN gstring, XEN property_value)
{
  #define H_gtk_rc_property_parse_border "gboolean gtk_rc_property_parse_border(GParamSpec* pspec, GString* gstring, \
GValue* property_value)"
  XEN_ASSERT_TYPE(XEN_GParamSpec__P(pspec), pspec, 1, "gtk_rc_property_parse_border", "GParamSpec*");
  XEN_ASSERT_TYPE(XEN_GString__P(gstring), gstring, 2, "gtk_rc_property_parse_border", "GString*");
  XEN_ASSERT_TYPE(XEN_GValue__P(property_value), property_value, 3, "gtk_rc_property_parse_border", "GValue*");
  return(C_TO_XEN_gboolean(gtk_rc_property_parse_border(XEN_TO_C_GParamSpec_(pspec), XEN_TO_C_GString_(gstring), XEN_TO_C_GValue_(property_value))));
}

static XEN gxg_gtk_settings_set_property_value(XEN settings, XEN name, XEN svalue)
{
  #define H_gtk_settings_set_property_value "void gtk_settings_set_property_value(GtkSettings* settings, \
gchar* name, GtkSettingsValue* svalue)"
  XEN_ASSERT_TYPE(XEN_GtkSettings__P(settings), settings, 1, "gtk_settings_set_property_value", "GtkSettings*");
  XEN_ASSERT_TYPE(XEN_gchar__P(name), name, 2, "gtk_settings_set_property_value", "gchar*");
  XEN_ASSERT_TYPE(XEN_GtkSettingsValue__P(svalue), svalue, 3, "gtk_settings_set_property_value", "GtkSettingsValue*");
  gtk_settings_set_property_value(XEN_TO_C_GtkSettings_(settings), XEN_TO_C_gchar_(name), XEN_TO_C_GtkSettingsValue_(svalue));
  return(XEN_FALSE);
}

static XEN gxg_gtk_settings_set_string_property(XEN settings, XEN name, XEN v_string, XEN origin)
{
  #define H_gtk_settings_set_string_property "void gtk_settings_set_string_property(GtkSettings* settings, \
gchar* name, gchar* v_string, gchar* origin)"
  XEN_ASSERT_TYPE(XEN_GtkSettings__P(settings), settings, 1, "gtk_settings_set_string_property", "GtkSettings*");
  XEN_ASSERT_TYPE(XEN_gchar__P(name), name, 2, "gtk_settings_set_string_property", "gchar*");
  XEN_ASSERT_TYPE(XEN_gchar__P(v_string), v_string, 3, "gtk_settings_set_string_property", "gchar*");
  XEN_ASSERT_TYPE(XEN_gchar__P(origin), origin, 4, "gtk_settings_set_string_property", "gchar*");
  gtk_settings_set_string_property(XEN_TO_C_GtkSettings_(settings), XEN_TO_C_gchar_(name), XEN_TO_C_gchar_(v_string), XEN_TO_C_gchar_(origin));
  return(XEN_FALSE);
}

static XEN gxg_gtk_settings_set_long_property(XEN settings, XEN name, XEN v_long, XEN origin)
{
  #define H_gtk_settings_set_long_property "void gtk_settings_set_long_property(GtkSettings* settings, \
gchar* name, glong v_long, gchar* origin)"
  XEN_ASSERT_TYPE(XEN_GtkSettings__P(settings), settings, 1, "gtk_settings_set_long_property", "GtkSettings*");
  XEN_ASSERT_TYPE(XEN_gchar__P(name), name, 2, "gtk_settings_set_long_property", "gchar*");
  XEN_ASSERT_TYPE(XEN_glong_P(v_long), v_long, 3, "gtk_settings_set_long_property", "glong");
  XEN_ASSERT_TYPE(XEN_gchar__P(origin), origin, 4, "gtk_settings_set_long_property", "gchar*");
  gtk_settings_set_long_property(XEN_TO_C_GtkSettings_(settings), XEN_TO_C_gchar_(name), XEN_TO_C_glong(v_long), XEN_TO_C_gchar_(origin));
  return(XEN_FALSE);
}

static XEN gxg_gtk_settings_set_double_property(XEN settings, XEN name, XEN v_double, XEN origin)
{
  #define H_gtk_settings_set_double_property "void gtk_settings_set_double_property(GtkSettings* settings, \
gchar* name, gdouble v_double, gchar* origin)"
  XEN_ASSERT_TYPE(XEN_GtkSettings__P(settings), settings, 1, "gtk_settings_set_double_property", "GtkSettings*");
  XEN_ASSERT_TYPE(XEN_gchar__P(name), name, 2, "gtk_settings_set_double_property", "gchar*");
  XEN_ASSERT_TYPE(XEN_gdouble_P(v_double), v_double, 3, "gtk_settings_set_double_property", "gdouble");
  XEN_ASSERT_TYPE(XEN_gchar__P(origin), origin, 4, "gtk_settings_set_double_property", "gchar*");
  gtk_settings_set_double_property(XEN_TO_C_GtkSettings_(settings), XEN_TO_C_gchar_(name), XEN_TO_C_gdouble(v_double), XEN_TO_C_gchar_(origin));
  return(XEN_FALSE);
}

static XEN gxg_gtk_size_group_get_type(void)
{
  #define H_gtk_size_group_get_type "GType gtk_size_group_get_type( void)"
  return(C_TO_XEN_GType(gtk_size_group_get_type()));
}

static XEN gxg_gtk_size_group_new(XEN mode)
{
  #define H_gtk_size_group_new "GtkSizeGroup* gtk_size_group_new(GtkSizeGroupMode mode)"
  XEN_ASSERT_TYPE(XEN_GtkSizeGroupMode_P(mode), mode, 1, "gtk_size_group_new", "GtkSizeGroupMode");
  return(C_TO_XEN_GtkSizeGroup_(gtk_size_group_new(XEN_TO_C_GtkSizeGroupMode(mode))));
}

static XEN gxg_gtk_size_group_set_mode(XEN size_group, XEN mode)
{
  #define H_gtk_size_group_set_mode "void gtk_size_group_set_mode(GtkSizeGroup* size_group, GtkSizeGroupMode mode)"
  XEN_ASSERT_TYPE(XEN_GtkSizeGroup__P(size_group), size_group, 1, "gtk_size_group_set_mode", "GtkSizeGroup*");
  XEN_ASSERT_TYPE(XEN_GtkSizeGroupMode_P(mode), mode, 2, "gtk_size_group_set_mode", "GtkSizeGroupMode");
  gtk_size_group_set_mode(XEN_TO_C_GtkSizeGroup_(size_group), XEN_TO_C_GtkSizeGroupMode(mode));
  return(XEN_FALSE);
}

static XEN gxg_gtk_size_group_get_mode(XEN size_group)
{
  #define H_gtk_size_group_get_mode "GtkSizeGroupMode gtk_size_group_get_mode(GtkSizeGroup* size_group)"
  XEN_ASSERT_TYPE(XEN_GtkSizeGroup__P(size_group), size_group, 1, "gtk_size_group_get_mode", "GtkSizeGroup*");
  return(C_TO_XEN_GtkSizeGroupMode(gtk_size_group_get_mode(XEN_TO_C_GtkSizeGroup_(size_group))));
}

static XEN gxg_gtk_size_group_add_widget(XEN size_group, XEN widget)
{
  #define H_gtk_size_group_add_widget "void gtk_size_group_add_widget(GtkSizeGroup* size_group, GtkWidget* widget)"
  XEN_ASSERT_TYPE(XEN_GtkSizeGroup__P(size_group), size_group, 1, "gtk_size_group_add_widget", "GtkSizeGroup*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 2, "gtk_size_group_add_widget", "GtkWidget*");
  gtk_size_group_add_widget(XEN_TO_C_GtkSizeGroup_(size_group), XEN_TO_C_GtkWidget_(widget));
  return(XEN_FALSE);
}

static XEN gxg_gtk_size_group_remove_widget(XEN size_group, XEN widget)
{
  #define H_gtk_size_group_remove_widget "void gtk_size_group_remove_widget(GtkSizeGroup* size_group, \
GtkWidget* widget)"
  XEN_ASSERT_TYPE(XEN_GtkSizeGroup__P(size_group), size_group, 1, "gtk_size_group_remove_widget", "GtkSizeGroup*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 2, "gtk_size_group_remove_widget", "GtkWidget*");
  gtk_size_group_remove_widget(XEN_TO_C_GtkSizeGroup_(size_group), XEN_TO_C_GtkWidget_(widget));
  return(XEN_FALSE);
}
#endif

static XEN gxg_gtk_socket_new(void)
{
  #define H_gtk_socket_new "GtkWidget* gtk_socket_new( void)"
  return(C_TO_XEN_GtkWidget_(gtk_socket_new()));
}

static XEN gxg_gtk_socket_get_type(void)
{
  #define H_gtk_socket_get_type "GtkType gtk_socket_get_type( void)"
  return(C_TO_XEN_GtkType(gtk_socket_get_type()));
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_socket_add_id(XEN socket, XEN window_id)
{
  #define H_gtk_socket_add_id "void gtk_socket_add_id(GtkSocket* socket, GdkNativeWindow window_id)"
  XEN_ASSERT_TYPE(XEN_GtkSocket__P(socket), socket, 1, "gtk_socket_add_id", "GtkSocket*");
  XEN_ASSERT_TYPE(XEN_GdkNativeWindow_P(window_id), window_id, 2, "gtk_socket_add_id", "GdkNativeWindow");
  gtk_socket_add_id(XEN_TO_C_GtkSocket_(socket), XEN_TO_C_GdkNativeWindow(window_id));
  return(XEN_FALSE);
}

static XEN gxg_gtk_socket_get_id(XEN socket)
{
  #define H_gtk_socket_get_id "GdkNativeWindow gtk_socket_get_id(GtkSocket* socket)"
  XEN_ASSERT_TYPE(XEN_GtkSocket__P(socket), socket, 1, "gtk_socket_get_id", "GtkSocket*");
  return(C_TO_XEN_GdkNativeWindow(gtk_socket_get_id(XEN_TO_C_GtkSocket_(socket))));
}
#endif

static XEN gxg_gtk_spin_button_get_type(void)
{
  #define H_gtk_spin_button_get_type "GtkType gtk_spin_button_get_type( void)"
  return(C_TO_XEN_GtkType(gtk_spin_button_get_type()));
}

static XEN gxg_gtk_spin_button_configure(XEN spin_button, XEN adjustment, XEN climb_rate, XEN digits)
{
  #define H_gtk_spin_button_configure "void gtk_spin_button_configure(GtkSpinButton* spin_button, GtkAdjustment* adjustment, \
gdouble climb_rate, guint digits)"
  XEN_ASSERT_TYPE(XEN_GtkSpinButton__P(spin_button), spin_button, 1, "gtk_spin_button_configure", "GtkSpinButton*");
  XEN_ASSERT_TYPE(XEN_GtkAdjustment__P(adjustment), adjustment, 2, "gtk_spin_button_configure", "GtkAdjustment*");
  XEN_ASSERT_TYPE(XEN_gdouble_P(climb_rate), climb_rate, 3, "gtk_spin_button_configure", "gdouble");
  XEN_ASSERT_TYPE(XEN_guint_P(digits), digits, 4, "gtk_spin_button_configure", "guint");
  gtk_spin_button_configure(XEN_TO_C_GtkSpinButton_(spin_button), XEN_TO_C_GtkAdjustment_(adjustment), XEN_TO_C_gdouble(climb_rate), 
                            XEN_TO_C_guint(digits));
  return(XEN_FALSE);
}

static XEN gxg_gtk_spin_button_new(XEN adjustment, XEN climb_rate, XEN digits)
{
  #define H_gtk_spin_button_new "GtkWidget* gtk_spin_button_new(GtkAdjustment* adjustment, gdouble climb_rate, \
guint digits)"
  XEN_ASSERT_TYPE(XEN_GtkAdjustment__P(adjustment), adjustment, 1, "gtk_spin_button_new", "GtkAdjustment*");
  XEN_ASSERT_TYPE(XEN_gdouble_P(climb_rate), climb_rate, 2, "gtk_spin_button_new", "gdouble");
  XEN_ASSERT_TYPE(XEN_guint_P(digits), digits, 3, "gtk_spin_button_new", "guint");
  return(C_TO_XEN_GtkWidget_(gtk_spin_button_new(XEN_TO_C_GtkAdjustment_(adjustment), XEN_TO_C_gdouble(climb_rate), XEN_TO_C_guint(digits))));
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_spin_button_new_with_range(XEN min, XEN max, XEN step)
{
  #define H_gtk_spin_button_new_with_range "GtkWidget* gtk_spin_button_new_with_range(gdouble min, gdouble max, \
gdouble step)"
  XEN_ASSERT_TYPE(XEN_gdouble_P(min), min, 1, "gtk_spin_button_new_with_range", "gdouble");
  XEN_ASSERT_TYPE(XEN_gdouble_P(max), max, 2, "gtk_spin_button_new_with_range", "gdouble");
  XEN_ASSERT_TYPE(XEN_gdouble_P(step), step, 3, "gtk_spin_button_new_with_range", "gdouble");
  return(C_TO_XEN_GtkWidget_(gtk_spin_button_new_with_range(XEN_TO_C_gdouble(min), XEN_TO_C_gdouble(max), XEN_TO_C_gdouble(step))));
}
#endif

static XEN gxg_gtk_spin_button_set_adjustment(XEN spin_button, XEN adjustment)
{
  #define H_gtk_spin_button_set_adjustment "void gtk_spin_button_set_adjustment(GtkSpinButton* spin_button, \
GtkAdjustment* adjustment)"
  XEN_ASSERT_TYPE(XEN_GtkSpinButton__P(spin_button), spin_button, 1, "gtk_spin_button_set_adjustment", "GtkSpinButton*");
  XEN_ASSERT_TYPE(XEN_GtkAdjustment__P(adjustment), adjustment, 2, "gtk_spin_button_set_adjustment", "GtkAdjustment*");
  gtk_spin_button_set_adjustment(XEN_TO_C_GtkSpinButton_(spin_button), XEN_TO_C_GtkAdjustment_(adjustment));
  return(XEN_FALSE);
}

static XEN gxg_gtk_spin_button_get_adjustment(XEN spin_button)
{
  #define H_gtk_spin_button_get_adjustment "GtkAdjustment* gtk_spin_button_get_adjustment(GtkSpinButton* spin_button)"
  XEN_ASSERT_TYPE(XEN_GtkSpinButton__P(spin_button), spin_button, 1, "gtk_spin_button_get_adjustment", "GtkSpinButton*");
  return(C_TO_XEN_GtkAdjustment_(gtk_spin_button_get_adjustment(XEN_TO_C_GtkSpinButton_(spin_button))));
}

static XEN gxg_gtk_spin_button_set_digits(XEN spin_button, XEN digits)
{
  #define H_gtk_spin_button_set_digits "void gtk_spin_button_set_digits(GtkSpinButton* spin_button, guint digits)"
  XEN_ASSERT_TYPE(XEN_GtkSpinButton__P(spin_button), spin_button, 1, "gtk_spin_button_set_digits", "GtkSpinButton*");
  XEN_ASSERT_TYPE(XEN_guint_P(digits), digits, 2, "gtk_spin_button_set_digits", "guint");
  gtk_spin_button_set_digits(XEN_TO_C_GtkSpinButton_(spin_button), XEN_TO_C_guint(digits));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_spin_button_get_digits(XEN spin_button)
{
  #define H_gtk_spin_button_get_digits "guint gtk_spin_button_get_digits(GtkSpinButton* spin_button)"
  XEN_ASSERT_TYPE(XEN_GtkSpinButton__P(spin_button), spin_button, 1, "gtk_spin_button_get_digits", "GtkSpinButton*");
  return(C_TO_XEN_guint(gtk_spin_button_get_digits(XEN_TO_C_GtkSpinButton_(spin_button))));
}

static XEN gxg_gtk_spin_button_set_increments(XEN spin_button, XEN step, XEN page)
{
  #define H_gtk_spin_button_set_increments "void gtk_spin_button_set_increments(GtkSpinButton* spin_button, \
gdouble step, gdouble page)"
  XEN_ASSERT_TYPE(XEN_GtkSpinButton__P(spin_button), spin_button, 1, "gtk_spin_button_set_increments", "GtkSpinButton*");
  XEN_ASSERT_TYPE(XEN_gdouble_P(step), step, 2, "gtk_spin_button_set_increments", "gdouble");
  XEN_ASSERT_TYPE(XEN_gdouble_P(page), page, 3, "gtk_spin_button_set_increments", "gdouble");
  gtk_spin_button_set_increments(XEN_TO_C_GtkSpinButton_(spin_button), XEN_TO_C_gdouble(step), XEN_TO_C_gdouble(page));
  return(XEN_FALSE);
}

static XEN gxg_gtk_spin_button_get_increments(XEN spin_button, XEN step, XEN page)
{
  #define H_gtk_spin_button_get_increments "void gtk_spin_button_get_increments(GtkSpinButton* spin_button, \
gdouble* [step], gdouble* [page])"
  gdouble ref_step;
  gdouble ref_page;
  XEN_ASSERT_TYPE(XEN_GtkSpinButton__P(spin_button), spin_button, 1, "gtk_spin_button_get_increments", "GtkSpinButton*");
  gtk_spin_button_get_increments(XEN_TO_C_GtkSpinButton_(spin_button), &ref_step, &ref_page);
  return(XEN_LIST_2(C_TO_XEN_gdouble(ref_step), C_TO_XEN_gdouble(ref_page)));
}

static XEN gxg_gtk_spin_button_set_range(XEN spin_button, XEN min, XEN max)
{
  #define H_gtk_spin_button_set_range "void gtk_spin_button_set_range(GtkSpinButton* spin_button, gdouble min, \
gdouble max)"
  XEN_ASSERT_TYPE(XEN_GtkSpinButton__P(spin_button), spin_button, 1, "gtk_spin_button_set_range", "GtkSpinButton*");
  XEN_ASSERT_TYPE(XEN_gdouble_P(min), min, 2, "gtk_spin_button_set_range", "gdouble");
  XEN_ASSERT_TYPE(XEN_gdouble_P(max), max, 3, "gtk_spin_button_set_range", "gdouble");
  gtk_spin_button_set_range(XEN_TO_C_GtkSpinButton_(spin_button), XEN_TO_C_gdouble(min), XEN_TO_C_gdouble(max));
  return(XEN_FALSE);
}

static XEN gxg_gtk_spin_button_get_range(XEN spin_button, XEN min, XEN max)
{
  #define H_gtk_spin_button_get_range "void gtk_spin_button_get_range(GtkSpinButton* spin_button, gdouble* [min], \
gdouble* [max])"
  gdouble ref_min;
  gdouble ref_max;
  XEN_ASSERT_TYPE(XEN_GtkSpinButton__P(spin_button), spin_button, 1, "gtk_spin_button_get_range", "GtkSpinButton*");
  gtk_spin_button_get_range(XEN_TO_C_GtkSpinButton_(spin_button), &ref_min, &ref_max);
  return(XEN_LIST_2(C_TO_XEN_gdouble(ref_min), C_TO_XEN_gdouble(ref_max)));
}

static XEN gxg_gtk_spin_button_get_value(XEN spin_button)
{
  #define H_gtk_spin_button_get_value "gdouble gtk_spin_button_get_value(GtkSpinButton* spin_button)"
  XEN_ASSERT_TYPE(XEN_GtkSpinButton__P(spin_button), spin_button, 1, "gtk_spin_button_get_value", "GtkSpinButton*");
  return(C_TO_XEN_gdouble(gtk_spin_button_get_value(XEN_TO_C_GtkSpinButton_(spin_button))));
}
#endif

static XEN gxg_gtk_spin_button_get_value_as_int(XEN spin_button)
{
  #define H_gtk_spin_button_get_value_as_int "gint gtk_spin_button_get_value_as_int(GtkSpinButton* spin_button)"
  XEN_ASSERT_TYPE(XEN_GtkSpinButton__P(spin_button), spin_button, 1, "gtk_spin_button_get_value_as_int", "GtkSpinButton*");
  return(C_TO_XEN_gint(gtk_spin_button_get_value_as_int(XEN_TO_C_GtkSpinButton_(spin_button))));
}

static XEN gxg_gtk_spin_button_set_value(XEN spin_button, XEN value)
{
  #define H_gtk_spin_button_set_value "void gtk_spin_button_set_value(GtkSpinButton* spin_button, gdouble value)"
  XEN_ASSERT_TYPE(XEN_GtkSpinButton__P(spin_button), spin_button, 1, "gtk_spin_button_set_value", "GtkSpinButton*");
  XEN_ASSERT_TYPE(XEN_gdouble_P(value), value, 2, "gtk_spin_button_set_value", "gdouble");
  gtk_spin_button_set_value(XEN_TO_C_GtkSpinButton_(spin_button), XEN_TO_C_gdouble(value));
  return(XEN_FALSE);
}

static XEN gxg_gtk_spin_button_set_update_policy(XEN spin_button, XEN policy)
{
  #define H_gtk_spin_button_set_update_policy "void gtk_spin_button_set_update_policy(GtkSpinButton* spin_button, \
GtkSpinButtonUpdatePolicy policy)"
  XEN_ASSERT_TYPE(XEN_GtkSpinButton__P(spin_button), spin_button, 1, "gtk_spin_button_set_update_policy", "GtkSpinButton*");
  XEN_ASSERT_TYPE(XEN_GtkSpinButtonUpdatePolicy_P(policy), policy, 2, "gtk_spin_button_set_update_policy", "GtkSpinButtonUpdatePolicy");
  gtk_spin_button_set_update_policy(XEN_TO_C_GtkSpinButton_(spin_button), XEN_TO_C_GtkSpinButtonUpdatePolicy(policy));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_spin_button_get_update_policy(XEN spin_button)
{
  #define H_gtk_spin_button_get_update_policy "GtkSpinButtonUpdatePolicy gtk_spin_button_get_update_policy(GtkSpinButton* spin_button)"
  XEN_ASSERT_TYPE(XEN_GtkSpinButton__P(spin_button), spin_button, 1, "gtk_spin_button_get_update_policy", "GtkSpinButton*");
  return(C_TO_XEN_GtkSpinButtonUpdatePolicy(gtk_spin_button_get_update_policy(XEN_TO_C_GtkSpinButton_(spin_button))));
}
#endif

static XEN gxg_gtk_spin_button_set_numeric(XEN spin_button, XEN numeric)
{
  #define H_gtk_spin_button_set_numeric "void gtk_spin_button_set_numeric(GtkSpinButton* spin_button, \
gboolean numeric)"
  XEN_ASSERT_TYPE(XEN_GtkSpinButton__P(spin_button), spin_button, 1, "gtk_spin_button_set_numeric", "GtkSpinButton*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(numeric), numeric, 2, "gtk_spin_button_set_numeric", "gboolean");
  gtk_spin_button_set_numeric(XEN_TO_C_GtkSpinButton_(spin_button), XEN_TO_C_gboolean(numeric));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_spin_button_get_numeric(XEN spin_button)
{
  #define H_gtk_spin_button_get_numeric "gboolean gtk_spin_button_get_numeric(GtkSpinButton* spin_button)"
  XEN_ASSERT_TYPE(XEN_GtkSpinButton__P(spin_button), spin_button, 1, "gtk_spin_button_get_numeric", "GtkSpinButton*");
  return(C_TO_XEN_gboolean(gtk_spin_button_get_numeric(XEN_TO_C_GtkSpinButton_(spin_button))));
}
#endif

static XEN gxg_gtk_spin_button_spin(XEN spin_button, XEN direction, XEN increment)
{
  #define H_gtk_spin_button_spin "void gtk_spin_button_spin(GtkSpinButton* spin_button, GtkSpinType direction, \
gdouble increment)"
  XEN_ASSERT_TYPE(XEN_GtkSpinButton__P(spin_button), spin_button, 1, "gtk_spin_button_spin", "GtkSpinButton*");
  XEN_ASSERT_TYPE(XEN_GtkSpinType_P(direction), direction, 2, "gtk_spin_button_spin", "GtkSpinType");
  XEN_ASSERT_TYPE(XEN_gdouble_P(increment), increment, 3, "gtk_spin_button_spin", "gdouble");
  gtk_spin_button_spin(XEN_TO_C_GtkSpinButton_(spin_button), XEN_TO_C_GtkSpinType(direction), XEN_TO_C_gdouble(increment));
  return(XEN_FALSE);
}

static XEN gxg_gtk_spin_button_set_wrap(XEN spin_button, XEN wrap)
{
  #define H_gtk_spin_button_set_wrap "void gtk_spin_button_set_wrap(GtkSpinButton* spin_button, gboolean wrap)"
  XEN_ASSERT_TYPE(XEN_GtkSpinButton__P(spin_button), spin_button, 1, "gtk_spin_button_set_wrap", "GtkSpinButton*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(wrap), wrap, 2, "gtk_spin_button_set_wrap", "gboolean");
  gtk_spin_button_set_wrap(XEN_TO_C_GtkSpinButton_(spin_button), XEN_TO_C_gboolean(wrap));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_spin_button_get_wrap(XEN spin_button)
{
  #define H_gtk_spin_button_get_wrap "gboolean gtk_spin_button_get_wrap(GtkSpinButton* spin_button)"
  XEN_ASSERT_TYPE(XEN_GtkSpinButton__P(spin_button), spin_button, 1, "gtk_spin_button_get_wrap", "GtkSpinButton*");
  return(C_TO_XEN_gboolean(gtk_spin_button_get_wrap(XEN_TO_C_GtkSpinButton_(spin_button))));
}
#endif

static XEN gxg_gtk_spin_button_set_snap_to_ticks(XEN spin_button, XEN snap_to_ticks)
{
  #define H_gtk_spin_button_set_snap_to_ticks "void gtk_spin_button_set_snap_to_ticks(GtkSpinButton* spin_button, \
gboolean snap_to_ticks)"
  XEN_ASSERT_TYPE(XEN_GtkSpinButton__P(spin_button), spin_button, 1, "gtk_spin_button_set_snap_to_ticks", "GtkSpinButton*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(snap_to_ticks), snap_to_ticks, 2, "gtk_spin_button_set_snap_to_ticks", "gboolean");
  gtk_spin_button_set_snap_to_ticks(XEN_TO_C_GtkSpinButton_(spin_button), XEN_TO_C_gboolean(snap_to_ticks));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_spin_button_get_snap_to_ticks(XEN spin_button)
{
  #define H_gtk_spin_button_get_snap_to_ticks "gboolean gtk_spin_button_get_snap_to_ticks(GtkSpinButton* spin_button)"
  XEN_ASSERT_TYPE(XEN_GtkSpinButton__P(spin_button), spin_button, 1, "gtk_spin_button_get_snap_to_ticks", "GtkSpinButton*");
  return(C_TO_XEN_gboolean(gtk_spin_button_get_snap_to_ticks(XEN_TO_C_GtkSpinButton_(spin_button))));
}
#endif

static XEN gxg_gtk_spin_button_update(XEN spin_button)
{
  #define H_gtk_spin_button_update "void gtk_spin_button_update(GtkSpinButton* spin_button)"
  XEN_ASSERT_TYPE(XEN_GtkSpinButton__P(spin_button), spin_button, 1, "gtk_spin_button_update", "GtkSpinButton*");
  gtk_spin_button_update(XEN_TO_C_GtkSpinButton_(spin_button));
  return(XEN_FALSE);
}

static XEN gxg_gtk_statusbar_get_type(void)
{
  #define H_gtk_statusbar_get_type "GtkType gtk_statusbar_get_type( void)"
  return(C_TO_XEN_GtkType(gtk_statusbar_get_type()));
}

static XEN gxg_gtk_statusbar_new(void)
{
  #define H_gtk_statusbar_new "GtkWidget* gtk_statusbar_new( void)"
  return(C_TO_XEN_GtkWidget_(gtk_statusbar_new()));
}

static XEN gxg_gtk_statusbar_get_context_id(XEN statusbar, XEN context_description)
{
  #define H_gtk_statusbar_get_context_id "guint gtk_statusbar_get_context_id(GtkStatusbar* statusbar, \
gchar* context_description)"
  XEN_ASSERT_TYPE(XEN_GtkStatusbar__P(statusbar), statusbar, 1, "gtk_statusbar_get_context_id", "GtkStatusbar*");
  XEN_ASSERT_TYPE(XEN_gchar__P(context_description), context_description, 2, "gtk_statusbar_get_context_id", "gchar*");
  return(C_TO_XEN_guint(gtk_statusbar_get_context_id(XEN_TO_C_GtkStatusbar_(statusbar), XEN_TO_C_gchar_(context_description))));
}

static XEN gxg_gtk_statusbar_push(XEN statusbar, XEN context_id, XEN text)
{
  #define H_gtk_statusbar_push "guint gtk_statusbar_push(GtkStatusbar* statusbar, guint context_id, gchar* text)"
  XEN_ASSERT_TYPE(XEN_GtkStatusbar__P(statusbar), statusbar, 1, "gtk_statusbar_push", "GtkStatusbar*");
  XEN_ASSERT_TYPE(XEN_guint_P(context_id), context_id, 2, "gtk_statusbar_push", "guint");
  XEN_ASSERT_TYPE(XEN_gchar__P(text), text, 3, "gtk_statusbar_push", "gchar*");
  return(C_TO_XEN_guint(gtk_statusbar_push(XEN_TO_C_GtkStatusbar_(statusbar), XEN_TO_C_guint(context_id), XEN_TO_C_gchar_(text))));
}

static XEN gxg_gtk_statusbar_pop(XEN statusbar, XEN context_id)
{
  #define H_gtk_statusbar_pop "void gtk_statusbar_pop(GtkStatusbar* statusbar, guint context_id)"
  XEN_ASSERT_TYPE(XEN_GtkStatusbar__P(statusbar), statusbar, 1, "gtk_statusbar_pop", "GtkStatusbar*");
  XEN_ASSERT_TYPE(XEN_guint_P(context_id), context_id, 2, "gtk_statusbar_pop", "guint");
  gtk_statusbar_pop(XEN_TO_C_GtkStatusbar_(statusbar), XEN_TO_C_guint(context_id));
  return(XEN_FALSE);
}

static XEN gxg_gtk_statusbar_remove(XEN statusbar, XEN context_id, XEN message_id)
{
  #define H_gtk_statusbar_remove "void gtk_statusbar_remove(GtkStatusbar* statusbar, guint context_id, \
guint message_id)"
  XEN_ASSERT_TYPE(XEN_GtkStatusbar__P(statusbar), statusbar, 1, "gtk_statusbar_remove", "GtkStatusbar*");
  XEN_ASSERT_TYPE(XEN_guint_P(context_id), context_id, 2, "gtk_statusbar_remove", "guint");
  XEN_ASSERT_TYPE(XEN_guint_P(message_id), message_id, 3, "gtk_statusbar_remove", "guint");
  gtk_statusbar_remove(XEN_TO_C_GtkStatusbar_(statusbar), XEN_TO_C_guint(context_id), XEN_TO_C_guint(message_id));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_statusbar_set_has_resize_grip(XEN statusbar, XEN setting)
{
  #define H_gtk_statusbar_set_has_resize_grip "void gtk_statusbar_set_has_resize_grip(GtkStatusbar* statusbar, \
gboolean setting)"
  XEN_ASSERT_TYPE(XEN_GtkStatusbar__P(statusbar), statusbar, 1, "gtk_statusbar_set_has_resize_grip", "GtkStatusbar*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(setting), setting, 2, "gtk_statusbar_set_has_resize_grip", "gboolean");
  gtk_statusbar_set_has_resize_grip(XEN_TO_C_GtkStatusbar_(statusbar), XEN_TO_C_gboolean(setting));
  return(XEN_FALSE);
}

static XEN gxg_gtk_statusbar_get_has_resize_grip(XEN statusbar)
{
  #define H_gtk_statusbar_get_has_resize_grip "gboolean gtk_statusbar_get_has_resize_grip(GtkStatusbar* statusbar)"
  XEN_ASSERT_TYPE(XEN_GtkStatusbar__P(statusbar), statusbar, 1, "gtk_statusbar_get_has_resize_grip", "GtkStatusbar*");
  return(C_TO_XEN_gboolean(gtk_statusbar_get_has_resize_grip(XEN_TO_C_GtkStatusbar_(statusbar))));
}

static XEN gxg_gtk_stock_add(XEN items, XEN n_items)
{
  #define H_gtk_stock_add "void gtk_stock_add(GtkStockItem* items, guint n_items)"
  XEN_ASSERT_TYPE(XEN_GtkStockItem__P(items), items, 1, "gtk_stock_add", "GtkStockItem*");
  XEN_ASSERT_TYPE(XEN_guint_P(n_items), n_items, 2, "gtk_stock_add", "guint");
  gtk_stock_add(XEN_TO_C_GtkStockItem_(items), XEN_TO_C_guint(n_items));
  return(XEN_FALSE);
}

static XEN gxg_gtk_stock_add_static(XEN items, XEN n_items)
{
  #define H_gtk_stock_add_static "void gtk_stock_add_static(GtkStockItem* items, guint n_items)"
  XEN_ASSERT_TYPE(XEN_GtkStockItem__P(items), items, 1, "gtk_stock_add_static", "GtkStockItem*");
  XEN_ASSERT_TYPE(XEN_guint_P(n_items), n_items, 2, "gtk_stock_add_static", "guint");
  gtk_stock_add_static(XEN_TO_C_GtkStockItem_(items), XEN_TO_C_guint(n_items));
  return(XEN_FALSE);
}

static XEN gxg_gtk_stock_lookup(XEN stock_id, XEN item)
{
  #define H_gtk_stock_lookup "gboolean gtk_stock_lookup(gchar* stock_id, GtkStockItem* item)"
  XEN_ASSERT_TYPE(XEN_gchar__P(stock_id), stock_id, 1, "gtk_stock_lookup", "gchar*");
  XEN_ASSERT_TYPE(XEN_GtkStockItem__P(item), item, 2, "gtk_stock_lookup", "GtkStockItem*");
  return(C_TO_XEN_gboolean(gtk_stock_lookup(XEN_TO_C_gchar_(stock_id), XEN_TO_C_GtkStockItem_(item))));
}

static XEN gxg_gtk_stock_list_ids(void)
{
  #define H_gtk_stock_list_ids "GSList* gtk_stock_list_ids( void)"
  return(C_TO_XEN_GSList_(gtk_stock_list_ids()));
}

static XEN gxg_gtk_stock_item_copy(XEN item)
{
  #define H_gtk_stock_item_copy "GtkStockItem* gtk_stock_item_copy(GtkStockItem* item)"
  XEN_ASSERT_TYPE(XEN_GtkStockItem__P(item), item, 1, "gtk_stock_item_copy", "GtkStockItem*");
  return(C_TO_XEN_GtkStockItem_(gtk_stock_item_copy(XEN_TO_C_GtkStockItem_(item))));
}

static XEN gxg_gtk_stock_item_free(XEN item)
{
  #define H_gtk_stock_item_free "void gtk_stock_item_free(GtkStockItem* item)"
  XEN_ASSERT_TYPE(XEN_GtkStockItem__P(item), item, 1, "gtk_stock_item_free", "GtkStockItem*");
  gtk_stock_item_free(XEN_TO_C_GtkStockItem_(item));
  return(XEN_FALSE);
}

static XEN gxg_gtk_style_get_type(void)
{
  #define H_gtk_style_get_type "GType gtk_style_get_type( void)"
  return(C_TO_XEN_GType(gtk_style_get_type()));
}
#endif

static XEN gxg_gtk_style_new(void)
{
  #define H_gtk_style_new "GtkStyle* gtk_style_new( void)"
  return(C_TO_XEN_GtkStyle_(gtk_style_new()));
}

static XEN gxg_gtk_style_copy(XEN style)
{
  #define H_gtk_style_copy "GtkStyle* gtk_style_copy(GtkStyle* style)"
  XEN_ASSERT_TYPE(XEN_GtkStyle__P(style), style, 1, "gtk_style_copy", "GtkStyle*");
  return(C_TO_XEN_GtkStyle_(gtk_style_copy(XEN_TO_C_GtkStyle_(style))));
}

static XEN gxg_gtk_style_attach(XEN style, XEN window)
{
  #define H_gtk_style_attach "GtkStyle* gtk_style_attach(GtkStyle* style, GdkWindow* window)"
  XEN_ASSERT_TYPE(XEN_GtkStyle__P(style), style, 1, "gtk_style_attach", "GtkStyle*");
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 2, "gtk_style_attach", "GdkWindow*");
  return(C_TO_XEN_GtkStyle_(gtk_style_attach(XEN_TO_C_GtkStyle_(style), XEN_TO_C_GdkWindow_(window))));
}

static XEN gxg_gtk_style_detach(XEN style)
{
  #define H_gtk_style_detach "void gtk_style_detach(GtkStyle* style)"
  XEN_ASSERT_TYPE(XEN_GtkStyle__P(style), style, 1, "gtk_style_detach", "GtkStyle*");
  gtk_style_detach(XEN_TO_C_GtkStyle_(style));
  return(XEN_FALSE);
}

static XEN gxg_gtk_style_set_background(XEN style, XEN window, XEN state_type)
{
  #define H_gtk_style_set_background "void gtk_style_set_background(GtkStyle* style, GdkWindow* window, \
GtkStateType state_type)"
  XEN_ASSERT_TYPE(XEN_GtkStyle__P(style), style, 1, "gtk_style_set_background", "GtkStyle*");
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 2, "gtk_style_set_background", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_GtkStateType_P(state_type), state_type, 3, "gtk_style_set_background", "GtkStateType");
  gtk_style_set_background(XEN_TO_C_GtkStyle_(style), XEN_TO_C_GdkWindow_(window), XEN_TO_C_GtkStateType(state_type));
  return(XEN_FALSE);
}

static XEN gxg_gtk_style_apply_default_background(XEN style, XEN window, XEN set_bg, XEN state_type, XEN area, XEN x, XEN y, XEN width, XEN height)
{
  #define H_gtk_style_apply_default_background "void gtk_style_apply_default_background(GtkStyle* style, \
GdkWindow* window, gboolean set_bg, GtkStateType state_type, GdkRectangle* area, gint x, gint y, gint width, \
gint height)"
  XEN_ASSERT_TYPE(XEN_GtkStyle__P(style), style, 1, "gtk_style_apply_default_background", "GtkStyle*");
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 2, "gtk_style_apply_default_background", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(set_bg), set_bg, 3, "gtk_style_apply_default_background", "gboolean");
  XEN_ASSERT_TYPE(XEN_GtkStateType_P(state_type), state_type, 4, "gtk_style_apply_default_background", "GtkStateType");
  XEN_ASSERT_TYPE(XEN_GdkRectangle__P(area), area, 5, "gtk_style_apply_default_background", "GdkRectangle*");
  XEN_ASSERT_TYPE(XEN_gint_P(x), x, 6, "gtk_style_apply_default_background", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(y), y, 7, "gtk_style_apply_default_background", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(width), width, 8, "gtk_style_apply_default_background", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(height), height, 9, "gtk_style_apply_default_background", "gint");
  gtk_style_apply_default_background(XEN_TO_C_GtkStyle_(style), XEN_TO_C_GdkWindow_(window), XEN_TO_C_gboolean(set_bg), XEN_TO_C_GtkStateType(state_type), 
                                     XEN_TO_C_GdkRectangle_(area), XEN_TO_C_gint(x), XEN_TO_C_gint(y), XEN_TO_C_gint(width), 
                                     XEN_TO_C_gint(height));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_style_lookup_icon_set(XEN style, XEN stock_id)
{
  #define H_gtk_style_lookup_icon_set "GtkIconSet* gtk_style_lookup_icon_set(GtkStyle* style, gchar* stock_id)"
  XEN_ASSERT_TYPE(XEN_GtkStyle__P(style), style, 1, "gtk_style_lookup_icon_set", "GtkStyle*");
  XEN_ASSERT_TYPE(XEN_gchar__P(stock_id), stock_id, 2, "gtk_style_lookup_icon_set", "gchar*");
  return(C_TO_XEN_GtkIconSet_(gtk_style_lookup_icon_set(XEN_TO_C_GtkStyle_(style), XEN_TO_C_gchar_(stock_id))));
}

static XEN gxg_gtk_style_render_icon(XEN style, XEN source, XEN direction, XEN state, XEN size, XEN widget, XEN detail)
{
  #define H_gtk_style_render_icon "GdkPixbuf* gtk_style_render_icon(GtkStyle* style, GtkIconSource* source, \
GtkTextDirection direction, GtkStateType state, GtkIconSize size, GtkWidget* widget, gchar* detail)"
  XEN_ASSERT_TYPE(XEN_GtkStyle__P(style), style, 1, "gtk_style_render_icon", "GtkStyle*");
  XEN_ASSERT_TYPE(XEN_GtkIconSource__P(source), source, 2, "gtk_style_render_icon", "GtkIconSource*");
  XEN_ASSERT_TYPE(XEN_GtkTextDirection_P(direction), direction, 3, "gtk_style_render_icon", "GtkTextDirection");
  XEN_ASSERT_TYPE(XEN_GtkStateType_P(state), state, 4, "gtk_style_render_icon", "GtkStateType");
  XEN_ASSERT_TYPE(XEN_GtkIconSize_P(size), size, 5, "gtk_style_render_icon", "GtkIconSize");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 6, "gtk_style_render_icon", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_gchar__P(detail), detail, 7, "gtk_style_render_icon", "gchar*");
  return(C_TO_XEN_GdkPixbuf_(gtk_style_render_icon(XEN_TO_C_GtkStyle_(style), XEN_TO_C_GtkIconSource_(source), XEN_TO_C_GtkTextDirection(direction), 
                                                   XEN_TO_C_GtkStateType(state), XEN_TO_C_GtkIconSize(size), XEN_TO_C_GtkWidget_(widget), 
                                                   XEN_TO_C_gchar_(detail))));
}
#endif

static XEN gxg_gtk_paint_hline(XEN style, XEN window, XEN state_type, XEN area, XEN widget, XEN detail, XEN x1, XEN x2, XEN y)
{
  #define H_gtk_paint_hline "void gtk_paint_hline(GtkStyle* style, GdkWindow* window, GtkStateType state_type, \
GdkRectangle* area, GtkWidget* widget, gchar* detail, gint x1, gint x2, gint y)"
  XEN_ASSERT_TYPE(XEN_GtkStyle__P(style), style, 1, "gtk_paint_hline", "GtkStyle*");
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 2, "gtk_paint_hline", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_GtkStateType_P(state_type), state_type, 3, "gtk_paint_hline", "GtkStateType");
  XEN_ASSERT_TYPE(XEN_GdkRectangle__P(area), area, 4, "gtk_paint_hline", "GdkRectangle*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 5, "gtk_paint_hline", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_gchar__P(detail), detail, 6, "gtk_paint_hline", "gchar*");
  XEN_ASSERT_TYPE(XEN_gint_P(x1), x1, 7, "gtk_paint_hline", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(x2), x2, 8, "gtk_paint_hline", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(y), y, 9, "gtk_paint_hline", "gint");
  gtk_paint_hline(XEN_TO_C_GtkStyle_(style), XEN_TO_C_GdkWindow_(window), XEN_TO_C_GtkStateType(state_type), XEN_TO_C_GdkRectangle_(area), 
                  XEN_TO_C_GtkWidget_(widget), XEN_TO_C_gchar_(detail), XEN_TO_C_gint(x1), XEN_TO_C_gint(x2), XEN_TO_C_gint(y));
  return(XEN_FALSE);
}

static XEN gxg_gtk_paint_vline(XEN style, XEN window, XEN state_type, XEN area, XEN widget, XEN detail, XEN y1, XEN y2, XEN x)
{
  #define H_gtk_paint_vline "void gtk_paint_vline(GtkStyle* style, GdkWindow* window, GtkStateType state_type, \
GdkRectangle* area, GtkWidget* widget, gchar* detail, gint y1, gint y2, gint x)"
  XEN_ASSERT_TYPE(XEN_GtkStyle__P(style), style, 1, "gtk_paint_vline", "GtkStyle*");
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 2, "gtk_paint_vline", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_GtkStateType_P(state_type), state_type, 3, "gtk_paint_vline", "GtkStateType");
  XEN_ASSERT_TYPE(XEN_GdkRectangle__P(area), area, 4, "gtk_paint_vline", "GdkRectangle*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 5, "gtk_paint_vline", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_gchar__P(detail), detail, 6, "gtk_paint_vline", "gchar*");
  XEN_ASSERT_TYPE(XEN_gint_P(y1), y1, 7, "gtk_paint_vline", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(y2), y2, 8, "gtk_paint_vline", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(x), x, 9, "gtk_paint_vline", "gint");
  gtk_paint_vline(XEN_TO_C_GtkStyle_(style), XEN_TO_C_GdkWindow_(window), XEN_TO_C_GtkStateType(state_type), XEN_TO_C_GdkRectangle_(area), 
                  XEN_TO_C_GtkWidget_(widget), XEN_TO_C_gchar_(detail), XEN_TO_C_gint(y1), XEN_TO_C_gint(y2), XEN_TO_C_gint(x));
  return(XEN_FALSE);
}

static XEN gxg_gtk_paint_shadow(XEN arglist)
{
  #define H_gtk_paint_shadow "void gtk_paint_shadow(GtkStyle* style, GdkWindow* window, GtkStateType state_type, \
GtkShadowType shadow_type, GdkRectangle* area, GtkWidget* widget, gchar* detail, gint x, gint y, gint width, \
gint height)"
  XEN style, window, state_type, shadow_type, area, widget, detail, x, y, width, height;
  style = XEN_LIST_REF(arglist, 0);
  window = XEN_LIST_REF(arglist, 1);
  state_type = XEN_LIST_REF(arglist, 2);
  shadow_type = XEN_LIST_REF(arglist, 3);
  area = XEN_LIST_REF(arglist, 4);
  widget = XEN_LIST_REF(arglist, 5);
  detail = XEN_LIST_REF(arglist, 6);
  x = XEN_LIST_REF(arglist, 7);
  y = XEN_LIST_REF(arglist, 8);
  width = XEN_LIST_REF(arglist, 9);
  height = XEN_LIST_REF(arglist, 10);
  XEN_ASSERT_TYPE(XEN_GtkStyle__P(style), style, 1, "gtk_paint_shadow", "GtkStyle*");
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 2, "gtk_paint_shadow", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_GtkStateType_P(state_type), state_type, 3, "gtk_paint_shadow", "GtkStateType");
  XEN_ASSERT_TYPE(XEN_GtkShadowType_P(shadow_type), shadow_type, 4, "gtk_paint_shadow", "GtkShadowType");
  XEN_ASSERT_TYPE(XEN_GdkRectangle__P(area), area, 5, "gtk_paint_shadow", "GdkRectangle*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 6, "gtk_paint_shadow", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_gchar__P(detail), detail, 7, "gtk_paint_shadow", "gchar*");
  XEN_ASSERT_TYPE(XEN_gint_P(x), x, 8, "gtk_paint_shadow", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(y), y, 9, "gtk_paint_shadow", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(width), width, 10, "gtk_paint_shadow", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(height), height, 11, "gtk_paint_shadow", "gint");
  gtk_paint_shadow(XEN_TO_C_GtkStyle_(style), XEN_TO_C_GdkWindow_(window), XEN_TO_C_GtkStateType(state_type), XEN_TO_C_GtkShadowType(shadow_type), 
                   XEN_TO_C_GdkRectangle_(area), XEN_TO_C_GtkWidget_(widget), XEN_TO_C_gchar_(detail), XEN_TO_C_gint(x), 
                   XEN_TO_C_gint(y), XEN_TO_C_gint(width), XEN_TO_C_gint(height));
  return(XEN_FALSE);
}

static XEN gxg_gtk_paint_polygon(XEN arglist)
{
  #define H_gtk_paint_polygon "void gtk_paint_polygon(GtkStyle* style, GdkWindow* window, GtkStateType state_type, \
GtkShadowType shadow_type, GdkRectangle* area, GtkWidget* widget, gchar* detail, GdkPoint* points, gint npoints, \
gboolean fill)"
  XEN style, window, state_type, shadow_type, area, widget, detail, points, npoints, fill;
  style = XEN_LIST_REF(arglist, 0);
  window = XEN_LIST_REF(arglist, 1);
  state_type = XEN_LIST_REF(arglist, 2);
  shadow_type = XEN_LIST_REF(arglist, 3);
  area = XEN_LIST_REF(arglist, 4);
  widget = XEN_LIST_REF(arglist, 5);
  detail = XEN_LIST_REF(arglist, 6);
  points = XEN_LIST_REF(arglist, 7);
  npoints = XEN_LIST_REF(arglist, 8);
  fill = XEN_LIST_REF(arglist, 9);
  XEN_ASSERT_TYPE(XEN_GtkStyle__P(style), style, 1, "gtk_paint_polygon", "GtkStyle*");
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 2, "gtk_paint_polygon", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_GtkStateType_P(state_type), state_type, 3, "gtk_paint_polygon", "GtkStateType");
  XEN_ASSERT_TYPE(XEN_GtkShadowType_P(shadow_type), shadow_type, 4, "gtk_paint_polygon", "GtkShadowType");
  XEN_ASSERT_TYPE(XEN_GdkRectangle__P(area), area, 5, "gtk_paint_polygon", "GdkRectangle*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 6, "gtk_paint_polygon", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_gchar__P(detail), detail, 7, "gtk_paint_polygon", "gchar*");
  XEN_ASSERT_TYPE(XEN_GdkPoint__P(points), points, 8, "gtk_paint_polygon", "GdkPoint*");
  XEN_ASSERT_TYPE(XEN_gint_P(npoints), npoints, 9, "gtk_paint_polygon", "gint");
  XEN_ASSERT_TYPE(XEN_gboolean_P(fill), fill, 10, "gtk_paint_polygon", "gboolean");
  gtk_paint_polygon(XEN_TO_C_GtkStyle_(style), XEN_TO_C_GdkWindow_(window), XEN_TO_C_GtkStateType(state_type), XEN_TO_C_GtkShadowType(shadow_type), 
                    XEN_TO_C_GdkRectangle_(area), XEN_TO_C_GtkWidget_(widget), XEN_TO_C_gchar_(detail), XEN_TO_C_GdkPoint_(points), 
                    XEN_TO_C_gint(npoints), XEN_TO_C_gboolean(fill));
  return(XEN_FALSE);
}

static XEN gxg_gtk_paint_arrow(XEN arglist)
{
  #define H_gtk_paint_arrow "void gtk_paint_arrow(GtkStyle* style, GdkWindow* window, GtkStateType state_type, \
GtkShadowType shadow_type, GdkRectangle* area, GtkWidget* widget, gchar* detail, GtkArrowType arrow_type, \
gboolean fill, gint x, gint y, gint width, gint height)"
  XEN style, window, state_type, shadow_type, area, widget, detail, arrow_type, fill, x, y, width, height;
  style = XEN_LIST_REF(arglist, 0);
  window = XEN_LIST_REF(arglist, 1);
  state_type = XEN_LIST_REF(arglist, 2);
  shadow_type = XEN_LIST_REF(arglist, 3);
  area = XEN_LIST_REF(arglist, 4);
  widget = XEN_LIST_REF(arglist, 5);
  detail = XEN_LIST_REF(arglist, 6);
  arrow_type = XEN_LIST_REF(arglist, 7);
  fill = XEN_LIST_REF(arglist, 8);
  x = XEN_LIST_REF(arglist, 9);
  y = XEN_LIST_REF(arglist, 10);
  width = XEN_LIST_REF(arglist, 11);
  height = XEN_LIST_REF(arglist, 12);
  XEN_ASSERT_TYPE(XEN_GtkStyle__P(style), style, 1, "gtk_paint_arrow", "GtkStyle*");
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 2, "gtk_paint_arrow", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_GtkStateType_P(state_type), state_type, 3, "gtk_paint_arrow", "GtkStateType");
  XEN_ASSERT_TYPE(XEN_GtkShadowType_P(shadow_type), shadow_type, 4, "gtk_paint_arrow", "GtkShadowType");
  XEN_ASSERT_TYPE(XEN_GdkRectangle__P(area), area, 5, "gtk_paint_arrow", "GdkRectangle*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 6, "gtk_paint_arrow", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_gchar__P(detail), detail, 7, "gtk_paint_arrow", "gchar*");
  XEN_ASSERT_TYPE(XEN_GtkArrowType_P(arrow_type), arrow_type, 8, "gtk_paint_arrow", "GtkArrowType");
  XEN_ASSERT_TYPE(XEN_gboolean_P(fill), fill, 9, "gtk_paint_arrow", "gboolean");
  XEN_ASSERT_TYPE(XEN_gint_P(x), x, 10, "gtk_paint_arrow", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(y), y, 11, "gtk_paint_arrow", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(width), width, 12, "gtk_paint_arrow", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(height), height, 13, "gtk_paint_arrow", "gint");
  gtk_paint_arrow(XEN_TO_C_GtkStyle_(style), XEN_TO_C_GdkWindow_(window), XEN_TO_C_GtkStateType(state_type), XEN_TO_C_GtkShadowType(shadow_type), 
                  XEN_TO_C_GdkRectangle_(area), XEN_TO_C_GtkWidget_(widget), XEN_TO_C_gchar_(detail), XEN_TO_C_GtkArrowType(arrow_type), 
                  XEN_TO_C_gboolean(fill), XEN_TO_C_gint(x), XEN_TO_C_gint(y), XEN_TO_C_gint(width), XEN_TO_C_gint(height));
  return(XEN_FALSE);
}

static XEN gxg_gtk_paint_diamond(XEN arglist)
{
  #define H_gtk_paint_diamond "void gtk_paint_diamond(GtkStyle* style, GdkWindow* window, GtkStateType state_type, \
GtkShadowType shadow_type, GdkRectangle* area, GtkWidget* widget, gchar* detail, gint x, gint y, gint width, \
gint height)"
  XEN style, window, state_type, shadow_type, area, widget, detail, x, y, width, height;
  style = XEN_LIST_REF(arglist, 0);
  window = XEN_LIST_REF(arglist, 1);
  state_type = XEN_LIST_REF(arglist, 2);
  shadow_type = XEN_LIST_REF(arglist, 3);
  area = XEN_LIST_REF(arglist, 4);
  widget = XEN_LIST_REF(arglist, 5);
  detail = XEN_LIST_REF(arglist, 6);
  x = XEN_LIST_REF(arglist, 7);
  y = XEN_LIST_REF(arglist, 8);
  width = XEN_LIST_REF(arglist, 9);
  height = XEN_LIST_REF(arglist, 10);
  XEN_ASSERT_TYPE(XEN_GtkStyle__P(style), style, 1, "gtk_paint_diamond", "GtkStyle*");
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 2, "gtk_paint_diamond", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_GtkStateType_P(state_type), state_type, 3, "gtk_paint_diamond", "GtkStateType");
  XEN_ASSERT_TYPE(XEN_GtkShadowType_P(shadow_type), shadow_type, 4, "gtk_paint_diamond", "GtkShadowType");
  XEN_ASSERT_TYPE(XEN_GdkRectangle__P(area), area, 5, "gtk_paint_diamond", "GdkRectangle*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 6, "gtk_paint_diamond", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_gchar__P(detail), detail, 7, "gtk_paint_diamond", "gchar*");
  XEN_ASSERT_TYPE(XEN_gint_P(x), x, 8, "gtk_paint_diamond", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(y), y, 9, "gtk_paint_diamond", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(width), width, 10, "gtk_paint_diamond", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(height), height, 11, "gtk_paint_diamond", "gint");
  gtk_paint_diamond(XEN_TO_C_GtkStyle_(style), XEN_TO_C_GdkWindow_(window), XEN_TO_C_GtkStateType(state_type), XEN_TO_C_GtkShadowType(shadow_type), 
                    XEN_TO_C_GdkRectangle_(area), XEN_TO_C_GtkWidget_(widget), XEN_TO_C_gchar_(detail), XEN_TO_C_gint(x), 
                    XEN_TO_C_gint(y), XEN_TO_C_gint(width), XEN_TO_C_gint(height));
  return(XEN_FALSE);
}

static XEN gxg_gtk_paint_box(XEN arglist)
{
  #define H_gtk_paint_box "void gtk_paint_box(GtkStyle* style, GdkWindow* window, GtkStateType state_type, \
GtkShadowType shadow_type, GdkRectangle* area, GtkWidget* widget, gchar* detail, gint x, gint y, gint width, \
gint height)"
  XEN style, window, state_type, shadow_type, area, widget, detail, x, y, width, height;
  style = XEN_LIST_REF(arglist, 0);
  window = XEN_LIST_REF(arglist, 1);
  state_type = XEN_LIST_REF(arglist, 2);
  shadow_type = XEN_LIST_REF(arglist, 3);
  area = XEN_LIST_REF(arglist, 4);
  widget = XEN_LIST_REF(arglist, 5);
  detail = XEN_LIST_REF(arglist, 6);
  x = XEN_LIST_REF(arglist, 7);
  y = XEN_LIST_REF(arglist, 8);
  width = XEN_LIST_REF(arglist, 9);
  height = XEN_LIST_REF(arglist, 10);
  XEN_ASSERT_TYPE(XEN_GtkStyle__P(style), style, 1, "gtk_paint_box", "GtkStyle*");
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 2, "gtk_paint_box", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_GtkStateType_P(state_type), state_type, 3, "gtk_paint_box", "GtkStateType");
  XEN_ASSERT_TYPE(XEN_GtkShadowType_P(shadow_type), shadow_type, 4, "gtk_paint_box", "GtkShadowType");
  XEN_ASSERT_TYPE(XEN_GdkRectangle__P(area), area, 5, "gtk_paint_box", "GdkRectangle*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 6, "gtk_paint_box", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_gchar__P(detail), detail, 7, "gtk_paint_box", "gchar*");
  XEN_ASSERT_TYPE(XEN_gint_P(x), x, 8, "gtk_paint_box", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(y), y, 9, "gtk_paint_box", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(width), width, 10, "gtk_paint_box", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(height), height, 11, "gtk_paint_box", "gint");
  gtk_paint_box(XEN_TO_C_GtkStyle_(style), XEN_TO_C_GdkWindow_(window), XEN_TO_C_GtkStateType(state_type), XEN_TO_C_GtkShadowType(shadow_type), 
                XEN_TO_C_GdkRectangle_(area), XEN_TO_C_GtkWidget_(widget), XEN_TO_C_gchar_(detail), XEN_TO_C_gint(x), XEN_TO_C_gint(y), 
                XEN_TO_C_gint(width), XEN_TO_C_gint(height));
  return(XEN_FALSE);
}

static XEN gxg_gtk_paint_flat_box(XEN arglist)
{
  #define H_gtk_paint_flat_box "void gtk_paint_flat_box(GtkStyle* style, GdkWindow* window, GtkStateType state_type, \
GtkShadowType shadow_type, GdkRectangle* area, GtkWidget* widget, gchar* detail, gint x, gint y, gint width, \
gint height)"
  XEN style, window, state_type, shadow_type, area, widget, detail, x, y, width, height;
  style = XEN_LIST_REF(arglist, 0);
  window = XEN_LIST_REF(arglist, 1);
  state_type = XEN_LIST_REF(arglist, 2);
  shadow_type = XEN_LIST_REF(arglist, 3);
  area = XEN_LIST_REF(arglist, 4);
  widget = XEN_LIST_REF(arglist, 5);
  detail = XEN_LIST_REF(arglist, 6);
  x = XEN_LIST_REF(arglist, 7);
  y = XEN_LIST_REF(arglist, 8);
  width = XEN_LIST_REF(arglist, 9);
  height = XEN_LIST_REF(arglist, 10);
  XEN_ASSERT_TYPE(XEN_GtkStyle__P(style), style, 1, "gtk_paint_flat_box", "GtkStyle*");
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 2, "gtk_paint_flat_box", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_GtkStateType_P(state_type), state_type, 3, "gtk_paint_flat_box", "GtkStateType");
  XEN_ASSERT_TYPE(XEN_GtkShadowType_P(shadow_type), shadow_type, 4, "gtk_paint_flat_box", "GtkShadowType");
  XEN_ASSERT_TYPE(XEN_GdkRectangle__P(area), area, 5, "gtk_paint_flat_box", "GdkRectangle*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 6, "gtk_paint_flat_box", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_gchar__P(detail), detail, 7, "gtk_paint_flat_box", "gchar*");
  XEN_ASSERT_TYPE(XEN_gint_P(x), x, 8, "gtk_paint_flat_box", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(y), y, 9, "gtk_paint_flat_box", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(width), width, 10, "gtk_paint_flat_box", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(height), height, 11, "gtk_paint_flat_box", "gint");
  gtk_paint_flat_box(XEN_TO_C_GtkStyle_(style), XEN_TO_C_GdkWindow_(window), XEN_TO_C_GtkStateType(state_type), XEN_TO_C_GtkShadowType(shadow_type), 
                     XEN_TO_C_GdkRectangle_(area), XEN_TO_C_GtkWidget_(widget), XEN_TO_C_gchar_(detail), XEN_TO_C_gint(x), 
                     XEN_TO_C_gint(y), XEN_TO_C_gint(width), XEN_TO_C_gint(height));
  return(XEN_FALSE);
}

static XEN gxg_gtk_paint_check(XEN arglist)
{
  #define H_gtk_paint_check "void gtk_paint_check(GtkStyle* style, GdkWindow* window, GtkStateType state_type, \
GtkShadowType shadow_type, GdkRectangle* area, GtkWidget* widget, gchar* detail, gint x, gint y, gint width, \
gint height)"
  XEN style, window, state_type, shadow_type, area, widget, detail, x, y, width, height;
  style = XEN_LIST_REF(arglist, 0);
  window = XEN_LIST_REF(arglist, 1);
  state_type = XEN_LIST_REF(arglist, 2);
  shadow_type = XEN_LIST_REF(arglist, 3);
  area = XEN_LIST_REF(arglist, 4);
  widget = XEN_LIST_REF(arglist, 5);
  detail = XEN_LIST_REF(arglist, 6);
  x = XEN_LIST_REF(arglist, 7);
  y = XEN_LIST_REF(arglist, 8);
  width = XEN_LIST_REF(arglist, 9);
  height = XEN_LIST_REF(arglist, 10);
  XEN_ASSERT_TYPE(XEN_GtkStyle__P(style), style, 1, "gtk_paint_check", "GtkStyle*");
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 2, "gtk_paint_check", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_GtkStateType_P(state_type), state_type, 3, "gtk_paint_check", "GtkStateType");
  XEN_ASSERT_TYPE(XEN_GtkShadowType_P(shadow_type), shadow_type, 4, "gtk_paint_check", "GtkShadowType");
  XEN_ASSERT_TYPE(XEN_GdkRectangle__P(area), area, 5, "gtk_paint_check", "GdkRectangle*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 6, "gtk_paint_check", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_gchar__P(detail), detail, 7, "gtk_paint_check", "gchar*");
  XEN_ASSERT_TYPE(XEN_gint_P(x), x, 8, "gtk_paint_check", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(y), y, 9, "gtk_paint_check", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(width), width, 10, "gtk_paint_check", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(height), height, 11, "gtk_paint_check", "gint");
  gtk_paint_check(XEN_TO_C_GtkStyle_(style), XEN_TO_C_GdkWindow_(window), XEN_TO_C_GtkStateType(state_type), XEN_TO_C_GtkShadowType(shadow_type), 
                  XEN_TO_C_GdkRectangle_(area), XEN_TO_C_GtkWidget_(widget), XEN_TO_C_gchar_(detail), XEN_TO_C_gint(x), XEN_TO_C_gint(y), 
                  XEN_TO_C_gint(width), XEN_TO_C_gint(height));
  return(XEN_FALSE);
}

static XEN gxg_gtk_paint_option(XEN arglist)
{
  #define H_gtk_paint_option "void gtk_paint_option(GtkStyle* style, GdkWindow* window, GtkStateType state_type, \
GtkShadowType shadow_type, GdkRectangle* area, GtkWidget* widget, gchar* detail, gint x, gint y, gint width, \
gint height)"
  XEN style, window, state_type, shadow_type, area, widget, detail, x, y, width, height;
  style = XEN_LIST_REF(arglist, 0);
  window = XEN_LIST_REF(arglist, 1);
  state_type = XEN_LIST_REF(arglist, 2);
  shadow_type = XEN_LIST_REF(arglist, 3);
  area = XEN_LIST_REF(arglist, 4);
  widget = XEN_LIST_REF(arglist, 5);
  detail = XEN_LIST_REF(arglist, 6);
  x = XEN_LIST_REF(arglist, 7);
  y = XEN_LIST_REF(arglist, 8);
  width = XEN_LIST_REF(arglist, 9);
  height = XEN_LIST_REF(arglist, 10);
  XEN_ASSERT_TYPE(XEN_GtkStyle__P(style), style, 1, "gtk_paint_option", "GtkStyle*");
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 2, "gtk_paint_option", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_GtkStateType_P(state_type), state_type, 3, "gtk_paint_option", "GtkStateType");
  XEN_ASSERT_TYPE(XEN_GtkShadowType_P(shadow_type), shadow_type, 4, "gtk_paint_option", "GtkShadowType");
  XEN_ASSERT_TYPE(XEN_GdkRectangle__P(area), area, 5, "gtk_paint_option", "GdkRectangle*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 6, "gtk_paint_option", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_gchar__P(detail), detail, 7, "gtk_paint_option", "gchar*");
  XEN_ASSERT_TYPE(XEN_gint_P(x), x, 8, "gtk_paint_option", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(y), y, 9, "gtk_paint_option", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(width), width, 10, "gtk_paint_option", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(height), height, 11, "gtk_paint_option", "gint");
  gtk_paint_option(XEN_TO_C_GtkStyle_(style), XEN_TO_C_GdkWindow_(window), XEN_TO_C_GtkStateType(state_type), XEN_TO_C_GtkShadowType(shadow_type), 
                   XEN_TO_C_GdkRectangle_(area), XEN_TO_C_GtkWidget_(widget), XEN_TO_C_gchar_(detail), XEN_TO_C_gint(x), 
                   XEN_TO_C_gint(y), XEN_TO_C_gint(width), XEN_TO_C_gint(height));
  return(XEN_FALSE);
}

static XEN gxg_gtk_paint_tab(XEN arglist)
{
  #define H_gtk_paint_tab "void gtk_paint_tab(GtkStyle* style, GdkWindow* window, GtkStateType state_type, \
GtkShadowType shadow_type, GdkRectangle* area, GtkWidget* widget, gchar* detail, gint x, gint y, gint width, \
gint height)"
  XEN style, window, state_type, shadow_type, area, widget, detail, x, y, width, height;
  style = XEN_LIST_REF(arglist, 0);
  window = XEN_LIST_REF(arglist, 1);
  state_type = XEN_LIST_REF(arglist, 2);
  shadow_type = XEN_LIST_REF(arglist, 3);
  area = XEN_LIST_REF(arglist, 4);
  widget = XEN_LIST_REF(arglist, 5);
  detail = XEN_LIST_REF(arglist, 6);
  x = XEN_LIST_REF(arglist, 7);
  y = XEN_LIST_REF(arglist, 8);
  width = XEN_LIST_REF(arglist, 9);
  height = XEN_LIST_REF(arglist, 10);
  XEN_ASSERT_TYPE(XEN_GtkStyle__P(style), style, 1, "gtk_paint_tab", "GtkStyle*");
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 2, "gtk_paint_tab", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_GtkStateType_P(state_type), state_type, 3, "gtk_paint_tab", "GtkStateType");
  XEN_ASSERT_TYPE(XEN_GtkShadowType_P(shadow_type), shadow_type, 4, "gtk_paint_tab", "GtkShadowType");
  XEN_ASSERT_TYPE(XEN_GdkRectangle__P(area), area, 5, "gtk_paint_tab", "GdkRectangle*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 6, "gtk_paint_tab", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_gchar__P(detail), detail, 7, "gtk_paint_tab", "gchar*");
  XEN_ASSERT_TYPE(XEN_gint_P(x), x, 8, "gtk_paint_tab", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(y), y, 9, "gtk_paint_tab", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(width), width, 10, "gtk_paint_tab", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(height), height, 11, "gtk_paint_tab", "gint");
  gtk_paint_tab(XEN_TO_C_GtkStyle_(style), XEN_TO_C_GdkWindow_(window), XEN_TO_C_GtkStateType(state_type), XEN_TO_C_GtkShadowType(shadow_type), 
                XEN_TO_C_GdkRectangle_(area), XEN_TO_C_GtkWidget_(widget), XEN_TO_C_gchar_(detail), XEN_TO_C_gint(x), XEN_TO_C_gint(y), 
                XEN_TO_C_gint(width), XEN_TO_C_gint(height));
  return(XEN_FALSE);
}

static XEN gxg_gtk_paint_shadow_gap(XEN arglist)
{
  #define H_gtk_paint_shadow_gap "void gtk_paint_shadow_gap(GtkStyle* style, GdkWindow* window, GtkStateType state_type, \
GtkShadowType shadow_type, GdkRectangle* area, GtkWidget* widget, gchar* detail, gint x, gint y, gint width, \
gint height, GtkPositionType gap_side, gint gap_x, gint gap_width)"
  XEN style, window, state_type, shadow_type, area, widget, detail, x, y, width, height, gap_side, gap_x, gap_width;
  style = XEN_LIST_REF(arglist, 0);
  window = XEN_LIST_REF(arglist, 1);
  state_type = XEN_LIST_REF(arglist, 2);
  shadow_type = XEN_LIST_REF(arglist, 3);
  area = XEN_LIST_REF(arglist, 4);
  widget = XEN_LIST_REF(arglist, 5);
  detail = XEN_LIST_REF(arglist, 6);
  x = XEN_LIST_REF(arglist, 7);
  y = XEN_LIST_REF(arglist, 8);
  width = XEN_LIST_REF(arglist, 9);
  height = XEN_LIST_REF(arglist, 10);
  gap_side = XEN_LIST_REF(arglist, 11);
  gap_x = XEN_LIST_REF(arglist, 12);
  gap_width = XEN_LIST_REF(arglist, 13);
  XEN_ASSERT_TYPE(XEN_GtkStyle__P(style), style, 1, "gtk_paint_shadow_gap", "GtkStyle*");
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 2, "gtk_paint_shadow_gap", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_GtkStateType_P(state_type), state_type, 3, "gtk_paint_shadow_gap", "GtkStateType");
  XEN_ASSERT_TYPE(XEN_GtkShadowType_P(shadow_type), shadow_type, 4, "gtk_paint_shadow_gap", "GtkShadowType");
  XEN_ASSERT_TYPE(XEN_GdkRectangle__P(area), area, 5, "gtk_paint_shadow_gap", "GdkRectangle*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 6, "gtk_paint_shadow_gap", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_gchar__P(detail), detail, 7, "gtk_paint_shadow_gap", "gchar*");
  XEN_ASSERT_TYPE(XEN_gint_P(x), x, 8, "gtk_paint_shadow_gap", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(y), y, 9, "gtk_paint_shadow_gap", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(width), width, 10, "gtk_paint_shadow_gap", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(height), height, 11, "gtk_paint_shadow_gap", "gint");
  XEN_ASSERT_TYPE(XEN_GtkPositionType_P(gap_side), gap_side, 12, "gtk_paint_shadow_gap", "GtkPositionType");
  XEN_ASSERT_TYPE(XEN_gint_P(gap_x), gap_x, 13, "gtk_paint_shadow_gap", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(gap_width), gap_width, 14, "gtk_paint_shadow_gap", "gint");
  gtk_paint_shadow_gap(XEN_TO_C_GtkStyle_(style), XEN_TO_C_GdkWindow_(window), XEN_TO_C_GtkStateType(state_type), XEN_TO_C_GtkShadowType(shadow_type), 
                       XEN_TO_C_GdkRectangle_(area), XEN_TO_C_GtkWidget_(widget), XEN_TO_C_gchar_(detail), XEN_TO_C_gint(x), 
                       XEN_TO_C_gint(y), XEN_TO_C_gint(width), XEN_TO_C_gint(height), XEN_TO_C_GtkPositionType(gap_side), 
                       XEN_TO_C_gint(gap_x), XEN_TO_C_gint(gap_width));
  return(XEN_FALSE);
}

static XEN gxg_gtk_paint_box_gap(XEN arglist)
{
  #define H_gtk_paint_box_gap "void gtk_paint_box_gap(GtkStyle* style, GdkWindow* window, GtkStateType state_type, \
GtkShadowType shadow_type, GdkRectangle* area, GtkWidget* widget, gchar* detail, gint x, gint y, gint width, \
gint height, GtkPositionType gap_side, gint gap_x, gint gap_width)"
  XEN style, window, state_type, shadow_type, area, widget, detail, x, y, width, height, gap_side, gap_x, gap_width;
  style = XEN_LIST_REF(arglist, 0);
  window = XEN_LIST_REF(arglist, 1);
  state_type = XEN_LIST_REF(arglist, 2);
  shadow_type = XEN_LIST_REF(arglist, 3);
  area = XEN_LIST_REF(arglist, 4);
  widget = XEN_LIST_REF(arglist, 5);
  detail = XEN_LIST_REF(arglist, 6);
  x = XEN_LIST_REF(arglist, 7);
  y = XEN_LIST_REF(arglist, 8);
  width = XEN_LIST_REF(arglist, 9);
  height = XEN_LIST_REF(arglist, 10);
  gap_side = XEN_LIST_REF(arglist, 11);
  gap_x = XEN_LIST_REF(arglist, 12);
  gap_width = XEN_LIST_REF(arglist, 13);
  XEN_ASSERT_TYPE(XEN_GtkStyle__P(style), style, 1, "gtk_paint_box_gap", "GtkStyle*");
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 2, "gtk_paint_box_gap", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_GtkStateType_P(state_type), state_type, 3, "gtk_paint_box_gap", "GtkStateType");
  XEN_ASSERT_TYPE(XEN_GtkShadowType_P(shadow_type), shadow_type, 4, "gtk_paint_box_gap", "GtkShadowType");
  XEN_ASSERT_TYPE(XEN_GdkRectangle__P(area), area, 5, "gtk_paint_box_gap", "GdkRectangle*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 6, "gtk_paint_box_gap", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_gchar__P(detail), detail, 7, "gtk_paint_box_gap", "gchar*");
  XEN_ASSERT_TYPE(XEN_gint_P(x), x, 8, "gtk_paint_box_gap", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(y), y, 9, "gtk_paint_box_gap", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(width), width, 10, "gtk_paint_box_gap", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(height), height, 11, "gtk_paint_box_gap", "gint");
  XEN_ASSERT_TYPE(XEN_GtkPositionType_P(gap_side), gap_side, 12, "gtk_paint_box_gap", "GtkPositionType");
  XEN_ASSERT_TYPE(XEN_gint_P(gap_x), gap_x, 13, "gtk_paint_box_gap", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(gap_width), gap_width, 14, "gtk_paint_box_gap", "gint");
  gtk_paint_box_gap(XEN_TO_C_GtkStyle_(style), XEN_TO_C_GdkWindow_(window), XEN_TO_C_GtkStateType(state_type), XEN_TO_C_GtkShadowType(shadow_type), 
                    XEN_TO_C_GdkRectangle_(area), XEN_TO_C_GtkWidget_(widget), XEN_TO_C_gchar_(detail), XEN_TO_C_gint(x), 
                    XEN_TO_C_gint(y), XEN_TO_C_gint(width), XEN_TO_C_gint(height), XEN_TO_C_GtkPositionType(gap_side), XEN_TO_C_gint(gap_x), 
                    XEN_TO_C_gint(gap_width));
  return(XEN_FALSE);
}

static XEN gxg_gtk_paint_extension(XEN arglist)
{
  #define H_gtk_paint_extension "void gtk_paint_extension(GtkStyle* style, GdkWindow* window, GtkStateType state_type, \
GtkShadowType shadow_type, GdkRectangle* area, GtkWidget* widget, gchar* detail, gint x, gint y, gint width, \
gint height, GtkPositionType gap_side)"
  XEN style, window, state_type, shadow_type, area, widget, detail, x, y, width, height, gap_side;
  style = XEN_LIST_REF(arglist, 0);
  window = XEN_LIST_REF(arglist, 1);
  state_type = XEN_LIST_REF(arglist, 2);
  shadow_type = XEN_LIST_REF(arglist, 3);
  area = XEN_LIST_REF(arglist, 4);
  widget = XEN_LIST_REF(arglist, 5);
  detail = XEN_LIST_REF(arglist, 6);
  x = XEN_LIST_REF(arglist, 7);
  y = XEN_LIST_REF(arglist, 8);
  width = XEN_LIST_REF(arglist, 9);
  height = XEN_LIST_REF(arglist, 10);
  gap_side = XEN_LIST_REF(arglist, 11);
  XEN_ASSERT_TYPE(XEN_GtkStyle__P(style), style, 1, "gtk_paint_extension", "GtkStyle*");
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 2, "gtk_paint_extension", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_GtkStateType_P(state_type), state_type, 3, "gtk_paint_extension", "GtkStateType");
  XEN_ASSERT_TYPE(XEN_GtkShadowType_P(shadow_type), shadow_type, 4, "gtk_paint_extension", "GtkShadowType");
  XEN_ASSERT_TYPE(XEN_GdkRectangle__P(area), area, 5, "gtk_paint_extension", "GdkRectangle*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 6, "gtk_paint_extension", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_gchar__P(detail), detail, 7, "gtk_paint_extension", "gchar*");
  XEN_ASSERT_TYPE(XEN_gint_P(x), x, 8, "gtk_paint_extension", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(y), y, 9, "gtk_paint_extension", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(width), width, 10, "gtk_paint_extension", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(height), height, 11, "gtk_paint_extension", "gint");
  XEN_ASSERT_TYPE(XEN_GtkPositionType_P(gap_side), gap_side, 12, "gtk_paint_extension", "GtkPositionType");
  gtk_paint_extension(XEN_TO_C_GtkStyle_(style), XEN_TO_C_GdkWindow_(window), XEN_TO_C_GtkStateType(state_type), XEN_TO_C_GtkShadowType(shadow_type), 
                      XEN_TO_C_GdkRectangle_(area), XEN_TO_C_GtkWidget_(widget), XEN_TO_C_gchar_(detail), XEN_TO_C_gint(x), 
                      XEN_TO_C_gint(y), XEN_TO_C_gint(width), XEN_TO_C_gint(height), XEN_TO_C_GtkPositionType(gap_side));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_paint_focus(XEN arglist)
{
  #define H_gtk_paint_focus "void gtk_paint_focus(GtkStyle* style, GdkWindow* window, GtkStateType state_type, \
GdkRectangle* area, GtkWidget* widget, gchar* detail, gint x, gint y, gint width, gint height)"
  XEN style, window, state_type, area, widget, detail, x, y, width, height;
  style = XEN_LIST_REF(arglist, 0);
  window = XEN_LIST_REF(arglist, 1);
  state_type = XEN_LIST_REF(arglist, 2);
  area = XEN_LIST_REF(arglist, 3);
  widget = XEN_LIST_REF(arglist, 4);
  detail = XEN_LIST_REF(arglist, 5);
  x = XEN_LIST_REF(arglist, 6);
  y = XEN_LIST_REF(arglist, 7);
  width = XEN_LIST_REF(arglist, 8);
  height = XEN_LIST_REF(arglist, 9);
  XEN_ASSERT_TYPE(XEN_GtkStyle__P(style), style, 1, "gtk_paint_focus", "GtkStyle*");
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 2, "gtk_paint_focus", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_GtkStateType_P(state_type), state_type, 3, "gtk_paint_focus", "GtkStateType");
  XEN_ASSERT_TYPE(XEN_GdkRectangle__P(area), area, 4, "gtk_paint_focus", "GdkRectangle*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 5, "gtk_paint_focus", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_gchar__P(detail), detail, 6, "gtk_paint_focus", "gchar*");
  XEN_ASSERT_TYPE(XEN_gint_P(x), x, 7, "gtk_paint_focus", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(y), y, 8, "gtk_paint_focus", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(width), width, 9, "gtk_paint_focus", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(height), height, 10, "gtk_paint_focus", "gint");
  gtk_paint_focus(XEN_TO_C_GtkStyle_(style), XEN_TO_C_GdkWindow_(window), XEN_TO_C_GtkStateType(state_type), XEN_TO_C_GdkRectangle_(area), 
                  XEN_TO_C_GtkWidget_(widget), XEN_TO_C_gchar_(detail), XEN_TO_C_gint(x), XEN_TO_C_gint(y), XEN_TO_C_gint(width), 
                  XEN_TO_C_gint(height));
  return(XEN_FALSE);
}
#endif

static XEN gxg_gtk_paint_slider(XEN arglist)
{
  #define H_gtk_paint_slider "void gtk_paint_slider(GtkStyle* style, GdkWindow* window, GtkStateType state_type, \
GtkShadowType shadow_type, GdkRectangle* area, GtkWidget* widget, gchar* detail, gint x, gint y, gint width, \
gint height, GtkOrientation orientation)"
  XEN style, window, state_type, shadow_type, area, widget, detail, x, y, width, height, orientation;
  style = XEN_LIST_REF(arglist, 0);
  window = XEN_LIST_REF(arglist, 1);
  state_type = XEN_LIST_REF(arglist, 2);
  shadow_type = XEN_LIST_REF(arglist, 3);
  area = XEN_LIST_REF(arglist, 4);
  widget = XEN_LIST_REF(arglist, 5);
  detail = XEN_LIST_REF(arglist, 6);
  x = XEN_LIST_REF(arglist, 7);
  y = XEN_LIST_REF(arglist, 8);
  width = XEN_LIST_REF(arglist, 9);
  height = XEN_LIST_REF(arglist, 10);
  orientation = XEN_LIST_REF(arglist, 11);
  XEN_ASSERT_TYPE(XEN_GtkStyle__P(style), style, 1, "gtk_paint_slider", "GtkStyle*");
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 2, "gtk_paint_slider", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_GtkStateType_P(state_type), state_type, 3, "gtk_paint_slider", "GtkStateType");
  XEN_ASSERT_TYPE(XEN_GtkShadowType_P(shadow_type), shadow_type, 4, "gtk_paint_slider", "GtkShadowType");
  XEN_ASSERT_TYPE(XEN_GdkRectangle__P(area), area, 5, "gtk_paint_slider", "GdkRectangle*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 6, "gtk_paint_slider", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_gchar__P(detail), detail, 7, "gtk_paint_slider", "gchar*");
  XEN_ASSERT_TYPE(XEN_gint_P(x), x, 8, "gtk_paint_slider", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(y), y, 9, "gtk_paint_slider", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(width), width, 10, "gtk_paint_slider", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(height), height, 11, "gtk_paint_slider", "gint");
  XEN_ASSERT_TYPE(XEN_GtkOrientation_P(orientation), orientation, 12, "gtk_paint_slider", "GtkOrientation");
  gtk_paint_slider(XEN_TO_C_GtkStyle_(style), XEN_TO_C_GdkWindow_(window), XEN_TO_C_GtkStateType(state_type), XEN_TO_C_GtkShadowType(shadow_type), 
                   XEN_TO_C_GdkRectangle_(area), XEN_TO_C_GtkWidget_(widget), XEN_TO_C_gchar_(detail), XEN_TO_C_gint(x), 
                   XEN_TO_C_gint(y), XEN_TO_C_gint(width), XEN_TO_C_gint(height), XEN_TO_C_GtkOrientation(orientation));
  return(XEN_FALSE);
}

static XEN gxg_gtk_paint_handle(XEN arglist)
{
  #define H_gtk_paint_handle "void gtk_paint_handle(GtkStyle* style, GdkWindow* window, GtkStateType state_type, \
GtkShadowType shadow_type, GdkRectangle* area, GtkWidget* widget, gchar* detail, gint x, gint y, gint width, \
gint height, GtkOrientation orientation)"
  XEN style, window, state_type, shadow_type, area, widget, detail, x, y, width, height, orientation;
  style = XEN_LIST_REF(arglist, 0);
  window = XEN_LIST_REF(arglist, 1);
  state_type = XEN_LIST_REF(arglist, 2);
  shadow_type = XEN_LIST_REF(arglist, 3);
  area = XEN_LIST_REF(arglist, 4);
  widget = XEN_LIST_REF(arglist, 5);
  detail = XEN_LIST_REF(arglist, 6);
  x = XEN_LIST_REF(arglist, 7);
  y = XEN_LIST_REF(arglist, 8);
  width = XEN_LIST_REF(arglist, 9);
  height = XEN_LIST_REF(arglist, 10);
  orientation = XEN_LIST_REF(arglist, 11);
  XEN_ASSERT_TYPE(XEN_GtkStyle__P(style), style, 1, "gtk_paint_handle", "GtkStyle*");
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 2, "gtk_paint_handle", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_GtkStateType_P(state_type), state_type, 3, "gtk_paint_handle", "GtkStateType");
  XEN_ASSERT_TYPE(XEN_GtkShadowType_P(shadow_type), shadow_type, 4, "gtk_paint_handle", "GtkShadowType");
  XEN_ASSERT_TYPE(XEN_GdkRectangle__P(area), area, 5, "gtk_paint_handle", "GdkRectangle*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 6, "gtk_paint_handle", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_gchar__P(detail), detail, 7, "gtk_paint_handle", "gchar*");
  XEN_ASSERT_TYPE(XEN_gint_P(x), x, 8, "gtk_paint_handle", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(y), y, 9, "gtk_paint_handle", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(width), width, 10, "gtk_paint_handle", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(height), height, 11, "gtk_paint_handle", "gint");
  XEN_ASSERT_TYPE(XEN_GtkOrientation_P(orientation), orientation, 12, "gtk_paint_handle", "GtkOrientation");
  gtk_paint_handle(XEN_TO_C_GtkStyle_(style), XEN_TO_C_GdkWindow_(window), XEN_TO_C_GtkStateType(state_type), XEN_TO_C_GtkShadowType(shadow_type), 
                   XEN_TO_C_GdkRectangle_(area), XEN_TO_C_GtkWidget_(widget), XEN_TO_C_gchar_(detail), XEN_TO_C_gint(x), 
                   XEN_TO_C_gint(y), XEN_TO_C_gint(width), XEN_TO_C_gint(height), XEN_TO_C_GtkOrientation(orientation));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_paint_expander(XEN style, XEN window, XEN state_type, XEN area, XEN widget, XEN detail, XEN x, XEN y, XEN expander_style)
{
  #define H_gtk_paint_expander "void gtk_paint_expander(GtkStyle* style, GdkWindow* window, GtkStateType state_type, \
GdkRectangle* area, GtkWidget* widget, gchar* detail, gint x, gint y, GtkExpanderStyle expander_style)"
  XEN_ASSERT_TYPE(XEN_GtkStyle__P(style), style, 1, "gtk_paint_expander", "GtkStyle*");
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 2, "gtk_paint_expander", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_GtkStateType_P(state_type), state_type, 3, "gtk_paint_expander", "GtkStateType");
  XEN_ASSERT_TYPE(XEN_GdkRectangle__P(area), area, 4, "gtk_paint_expander", "GdkRectangle*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 5, "gtk_paint_expander", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_gchar__P(detail), detail, 6, "gtk_paint_expander", "gchar*");
  XEN_ASSERT_TYPE(XEN_gint_P(x), x, 7, "gtk_paint_expander", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(y), y, 8, "gtk_paint_expander", "gint");
  XEN_ASSERT_TYPE(XEN_GtkExpanderStyle_P(expander_style), expander_style, 9, "gtk_paint_expander", "GtkExpanderStyle");
  gtk_paint_expander(XEN_TO_C_GtkStyle_(style), XEN_TO_C_GdkWindow_(window), XEN_TO_C_GtkStateType(state_type), XEN_TO_C_GdkRectangle_(area), 
                     XEN_TO_C_GtkWidget_(widget), XEN_TO_C_gchar_(detail), XEN_TO_C_gint(x), XEN_TO_C_gint(y), XEN_TO_C_GtkExpanderStyle(expander_style));
  return(XEN_FALSE);
}

static XEN gxg_gtk_paint_layout(XEN arglist)
{
  #define H_gtk_paint_layout "void gtk_paint_layout(GtkStyle* style, GdkWindow* window, GtkStateType state_type, \
gboolean use_text, GdkRectangle* area, GtkWidget* widget, gchar* detail, gint x, gint y, PangoLayout* layout)"
  XEN style, window, state_type, use_text, area, widget, detail, x, y, layout;
  style = XEN_LIST_REF(arglist, 0);
  window = XEN_LIST_REF(arglist, 1);
  state_type = XEN_LIST_REF(arglist, 2);
  use_text = XEN_LIST_REF(arglist, 3);
  area = XEN_LIST_REF(arglist, 4);
  widget = XEN_LIST_REF(arglist, 5);
  detail = XEN_LIST_REF(arglist, 6);
  x = XEN_LIST_REF(arglist, 7);
  y = XEN_LIST_REF(arglist, 8);
  layout = XEN_LIST_REF(arglist, 9);
  XEN_ASSERT_TYPE(XEN_GtkStyle__P(style), style, 1, "gtk_paint_layout", "GtkStyle*");
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 2, "gtk_paint_layout", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_GtkStateType_P(state_type), state_type, 3, "gtk_paint_layout", "GtkStateType");
  XEN_ASSERT_TYPE(XEN_gboolean_P(use_text), use_text, 4, "gtk_paint_layout", "gboolean");
  XEN_ASSERT_TYPE(XEN_GdkRectangle__P(area), area, 5, "gtk_paint_layout", "GdkRectangle*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 6, "gtk_paint_layout", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_gchar__P(detail), detail, 7, "gtk_paint_layout", "gchar*");
  XEN_ASSERT_TYPE(XEN_gint_P(x), x, 8, "gtk_paint_layout", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(y), y, 9, "gtk_paint_layout", "gint");
  XEN_ASSERT_TYPE(XEN_PangoLayout__P(layout), layout, 10, "gtk_paint_layout", "PangoLayout*");
  gtk_paint_layout(XEN_TO_C_GtkStyle_(style), XEN_TO_C_GdkWindow_(window), XEN_TO_C_GtkStateType(state_type), XEN_TO_C_gboolean(use_text), 
                   XEN_TO_C_GdkRectangle_(area), XEN_TO_C_GtkWidget_(widget), XEN_TO_C_gchar_(detail), XEN_TO_C_gint(x), 
                   XEN_TO_C_gint(y), XEN_TO_C_PangoLayout_(layout));
  return(XEN_FALSE);
}

static XEN gxg_gtk_paint_resize_grip(XEN arglist)
{
  #define H_gtk_paint_resize_grip "void gtk_paint_resize_grip(GtkStyle* style, GdkWindow* window, GtkStateType state_type, \
GdkRectangle* area, GtkWidget* widget, gchar* detail, GdkWindowEdge edge, gint x, gint y, gint width, \
gint height)"
  XEN style, window, state_type, area, widget, detail, edge, x, y, width, height;
  style = XEN_LIST_REF(arglist, 0);
  window = XEN_LIST_REF(arglist, 1);
  state_type = XEN_LIST_REF(arglist, 2);
  area = XEN_LIST_REF(arglist, 3);
  widget = XEN_LIST_REF(arglist, 4);
  detail = XEN_LIST_REF(arglist, 5);
  edge = XEN_LIST_REF(arglist, 6);
  x = XEN_LIST_REF(arglist, 7);
  y = XEN_LIST_REF(arglist, 8);
  width = XEN_LIST_REF(arglist, 9);
  height = XEN_LIST_REF(arglist, 10);
  XEN_ASSERT_TYPE(XEN_GtkStyle__P(style), style, 1, "gtk_paint_resize_grip", "GtkStyle*");
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 2, "gtk_paint_resize_grip", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_GtkStateType_P(state_type), state_type, 3, "gtk_paint_resize_grip", "GtkStateType");
  XEN_ASSERT_TYPE(XEN_GdkRectangle__P(area), area, 4, "gtk_paint_resize_grip", "GdkRectangle*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 5, "gtk_paint_resize_grip", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_gchar__P(detail), detail, 6, "gtk_paint_resize_grip", "gchar*");
  XEN_ASSERT_TYPE(XEN_GdkWindowEdge_P(edge), edge, 7, "gtk_paint_resize_grip", "GdkWindowEdge");
  XEN_ASSERT_TYPE(XEN_gint_P(x), x, 8, "gtk_paint_resize_grip", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(y), y, 9, "gtk_paint_resize_grip", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(width), width, 10, "gtk_paint_resize_grip", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(height), height, 11, "gtk_paint_resize_grip", "gint");
  gtk_paint_resize_grip(XEN_TO_C_GtkStyle_(style), XEN_TO_C_GdkWindow_(window), XEN_TO_C_GtkStateType(state_type), XEN_TO_C_GdkRectangle_(area), 
                        XEN_TO_C_GtkWidget_(widget), XEN_TO_C_gchar_(detail), XEN_TO_C_GdkWindowEdge(edge), XEN_TO_C_gint(x), 
                        XEN_TO_C_gint(y), XEN_TO_C_gint(width), XEN_TO_C_gint(height));
  return(XEN_FALSE);
}

static XEN gxg_gtk_border_copy(XEN border)
{
  #define H_gtk_border_copy "GtkBorder* gtk_border_copy(GtkBorder* border)"
  XEN_ASSERT_TYPE(XEN_GtkBorder__P(border), border, 1, "gtk_border_copy", "GtkBorder*");
  return(C_TO_XEN_GtkBorder_(gtk_border_copy(XEN_TO_C_GtkBorder_(border))));
}

static XEN gxg_gtk_border_free(XEN border)
{
  #define H_gtk_border_free "void gtk_border_free(GtkBorder* border)"
  XEN_ASSERT_TYPE(XEN_GtkBorder__P(border), border, 1, "gtk_border_free", "GtkBorder*");
  gtk_border_free(XEN_TO_C_GtkBorder_(border));
  return(XEN_FALSE);
}
#endif

static XEN gxg_gtk_table_get_type(void)
{
  #define H_gtk_table_get_type "GtkType gtk_table_get_type( void)"
  return(C_TO_XEN_GtkType(gtk_table_get_type()));
}

static XEN gxg_gtk_table_new(XEN rows, XEN columns, XEN homogeneous)
{
  #define H_gtk_table_new "GtkWidget* gtk_table_new(guint rows, guint columns, gboolean homogeneous)"
  XEN_ASSERT_TYPE(XEN_guint_P(rows), rows, 1, "gtk_table_new", "guint");
  XEN_ASSERT_TYPE(XEN_guint_P(columns), columns, 2, "gtk_table_new", "guint");
  XEN_ASSERT_TYPE(XEN_gboolean_P(homogeneous), homogeneous, 3, "gtk_table_new", "gboolean");
  return(C_TO_XEN_GtkWidget_(gtk_table_new(XEN_TO_C_guint(rows), XEN_TO_C_guint(columns), XEN_TO_C_gboolean(homogeneous))));
}

static XEN gxg_gtk_table_resize(XEN table, XEN rows, XEN columns)
{
  #define H_gtk_table_resize "void gtk_table_resize(GtkTable* table, guint rows, guint columns)"
  XEN_ASSERT_TYPE(XEN_GtkTable__P(table), table, 1, "gtk_table_resize", "GtkTable*");
  XEN_ASSERT_TYPE(XEN_guint_P(rows), rows, 2, "gtk_table_resize", "guint");
  XEN_ASSERT_TYPE(XEN_guint_P(columns), columns, 3, "gtk_table_resize", "guint");
  gtk_table_resize(XEN_TO_C_GtkTable_(table), XEN_TO_C_guint(rows), XEN_TO_C_guint(columns));
  return(XEN_FALSE);
}

static XEN gxg_gtk_table_attach(XEN arglist)
{
  #define H_gtk_table_attach "void gtk_table_attach(GtkTable* table, GtkWidget* child, guint left_attach, \
guint right_attach, guint top_attach, guint bottom_attach, GtkAttachOptions xoptions, GtkAttachOptions yoptions, \
guint xpadding, guint ypadding)"
  XEN table, child, left_attach, right_attach, top_attach, bottom_attach, xoptions, yoptions, xpadding, ypadding;
  table = XEN_LIST_REF(arglist, 0);
  child = XEN_LIST_REF(arglist, 1);
  left_attach = XEN_LIST_REF(arglist, 2);
  right_attach = XEN_LIST_REF(arglist, 3);
  top_attach = XEN_LIST_REF(arglist, 4);
  bottom_attach = XEN_LIST_REF(arglist, 5);
  xoptions = XEN_LIST_REF(arglist, 6);
  yoptions = XEN_LIST_REF(arglist, 7);
  xpadding = XEN_LIST_REF(arglist, 8);
  ypadding = XEN_LIST_REF(arglist, 9);
  XEN_ASSERT_TYPE(XEN_GtkTable__P(table), table, 1, "gtk_table_attach", "GtkTable*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(child), child, 2, "gtk_table_attach", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_guint_P(left_attach), left_attach, 3, "gtk_table_attach", "guint");
  XEN_ASSERT_TYPE(XEN_guint_P(right_attach), right_attach, 4, "gtk_table_attach", "guint");
  XEN_ASSERT_TYPE(XEN_guint_P(top_attach), top_attach, 5, "gtk_table_attach", "guint");
  XEN_ASSERT_TYPE(XEN_guint_P(bottom_attach), bottom_attach, 6, "gtk_table_attach", "guint");
  XEN_ASSERT_TYPE(XEN_GtkAttachOptions_P(xoptions), xoptions, 7, "gtk_table_attach", "GtkAttachOptions");
  XEN_ASSERT_TYPE(XEN_GtkAttachOptions_P(yoptions), yoptions, 8, "gtk_table_attach", "GtkAttachOptions");
  XEN_ASSERT_TYPE(XEN_guint_P(xpadding), xpadding, 9, "gtk_table_attach", "guint");
  XEN_ASSERT_TYPE(XEN_guint_P(ypadding), ypadding, 10, "gtk_table_attach", "guint");
  gtk_table_attach(XEN_TO_C_GtkTable_(table), XEN_TO_C_GtkWidget_(child), XEN_TO_C_guint(left_attach), XEN_TO_C_guint(right_attach), 
                   XEN_TO_C_guint(top_attach), XEN_TO_C_guint(bottom_attach), XEN_TO_C_GtkAttachOptions(xoptions), XEN_TO_C_GtkAttachOptions(yoptions), 
                   XEN_TO_C_guint(xpadding), XEN_TO_C_guint(ypadding));
  return(XEN_FALSE);
}

static XEN gxg_gtk_table_attach_defaults(XEN table, XEN widget, XEN left_attach, XEN right_attach, XEN top_attach, XEN bottom_attach)
{
  #define H_gtk_table_attach_defaults "void gtk_table_attach_defaults(GtkTable* table, GtkWidget* widget, \
guint left_attach, guint right_attach, guint top_attach, guint bottom_attach)"
  XEN_ASSERT_TYPE(XEN_GtkTable__P(table), table, 1, "gtk_table_attach_defaults", "GtkTable*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 2, "gtk_table_attach_defaults", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_guint_P(left_attach), left_attach, 3, "gtk_table_attach_defaults", "guint");
  XEN_ASSERT_TYPE(XEN_guint_P(right_attach), right_attach, 4, "gtk_table_attach_defaults", "guint");
  XEN_ASSERT_TYPE(XEN_guint_P(top_attach), top_attach, 5, "gtk_table_attach_defaults", "guint");
  XEN_ASSERT_TYPE(XEN_guint_P(bottom_attach), bottom_attach, 6, "gtk_table_attach_defaults", "guint");
  gtk_table_attach_defaults(XEN_TO_C_GtkTable_(table), XEN_TO_C_GtkWidget_(widget), XEN_TO_C_guint(left_attach), XEN_TO_C_guint(right_attach), 
                            XEN_TO_C_guint(top_attach), XEN_TO_C_guint(bottom_attach));
  return(XEN_FALSE);
}

static XEN gxg_gtk_table_set_row_spacing(XEN table, XEN row, XEN spacing)
{
  #define H_gtk_table_set_row_spacing "void gtk_table_set_row_spacing(GtkTable* table, guint row, guint spacing)"
  XEN_ASSERT_TYPE(XEN_GtkTable__P(table), table, 1, "gtk_table_set_row_spacing", "GtkTable*");
  XEN_ASSERT_TYPE(XEN_guint_P(row), row, 2, "gtk_table_set_row_spacing", "guint");
  XEN_ASSERT_TYPE(XEN_guint_P(spacing), spacing, 3, "gtk_table_set_row_spacing", "guint");
  gtk_table_set_row_spacing(XEN_TO_C_GtkTable_(table), XEN_TO_C_guint(row), XEN_TO_C_guint(spacing));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_table_get_row_spacing(XEN table, XEN row)
{
  #define H_gtk_table_get_row_spacing "guint gtk_table_get_row_spacing(GtkTable* table, guint row)"
  XEN_ASSERT_TYPE(XEN_GtkTable__P(table), table, 1, "gtk_table_get_row_spacing", "GtkTable*");
  XEN_ASSERT_TYPE(XEN_guint_P(row), row, 2, "gtk_table_get_row_spacing", "guint");
  return(C_TO_XEN_guint(gtk_table_get_row_spacing(XEN_TO_C_GtkTable_(table), XEN_TO_C_guint(row))));
}
#endif

static XEN gxg_gtk_table_set_col_spacing(XEN table, XEN column, XEN spacing)
{
  #define H_gtk_table_set_col_spacing "void gtk_table_set_col_spacing(GtkTable* table, guint column, \
guint spacing)"
  XEN_ASSERT_TYPE(XEN_GtkTable__P(table), table, 1, "gtk_table_set_col_spacing", "GtkTable*");
  XEN_ASSERT_TYPE(XEN_guint_P(column), column, 2, "gtk_table_set_col_spacing", "guint");
  XEN_ASSERT_TYPE(XEN_guint_P(spacing), spacing, 3, "gtk_table_set_col_spacing", "guint");
  gtk_table_set_col_spacing(XEN_TO_C_GtkTable_(table), XEN_TO_C_guint(column), XEN_TO_C_guint(spacing));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_table_get_col_spacing(XEN table, XEN column)
{
  #define H_gtk_table_get_col_spacing "guint gtk_table_get_col_spacing(GtkTable* table, guint column)"
  XEN_ASSERT_TYPE(XEN_GtkTable__P(table), table, 1, "gtk_table_get_col_spacing", "GtkTable*");
  XEN_ASSERT_TYPE(XEN_guint_P(column), column, 2, "gtk_table_get_col_spacing", "guint");
  return(C_TO_XEN_guint(gtk_table_get_col_spacing(XEN_TO_C_GtkTable_(table), XEN_TO_C_guint(column))));
}
#endif

static XEN gxg_gtk_table_set_row_spacings(XEN table, XEN spacing)
{
  #define H_gtk_table_set_row_spacings "void gtk_table_set_row_spacings(GtkTable* table, guint spacing)"
  XEN_ASSERT_TYPE(XEN_GtkTable__P(table), table, 1, "gtk_table_set_row_spacings", "GtkTable*");
  XEN_ASSERT_TYPE(XEN_guint_P(spacing), spacing, 2, "gtk_table_set_row_spacings", "guint");
  gtk_table_set_row_spacings(XEN_TO_C_GtkTable_(table), XEN_TO_C_guint(spacing));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_table_get_default_row_spacing(XEN table)
{
  #define H_gtk_table_get_default_row_spacing "guint gtk_table_get_default_row_spacing(GtkTable* table)"
  XEN_ASSERT_TYPE(XEN_GtkTable__P(table), table, 1, "gtk_table_get_default_row_spacing", "GtkTable*");
  return(C_TO_XEN_guint(gtk_table_get_default_row_spacing(XEN_TO_C_GtkTable_(table))));
}
#endif

static XEN gxg_gtk_table_set_col_spacings(XEN table, XEN spacing)
{
  #define H_gtk_table_set_col_spacings "void gtk_table_set_col_spacings(GtkTable* table, guint spacing)"
  XEN_ASSERT_TYPE(XEN_GtkTable__P(table), table, 1, "gtk_table_set_col_spacings", "GtkTable*");
  XEN_ASSERT_TYPE(XEN_guint_P(spacing), spacing, 2, "gtk_table_set_col_spacings", "guint");
  gtk_table_set_col_spacings(XEN_TO_C_GtkTable_(table), XEN_TO_C_guint(spacing));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_table_get_default_col_spacing(XEN table)
{
  #define H_gtk_table_get_default_col_spacing "guint gtk_table_get_default_col_spacing(GtkTable* table)"
  XEN_ASSERT_TYPE(XEN_GtkTable__P(table), table, 1, "gtk_table_get_default_col_spacing", "GtkTable*");
  return(C_TO_XEN_guint(gtk_table_get_default_col_spacing(XEN_TO_C_GtkTable_(table))));
}
#endif

static XEN gxg_gtk_table_set_homogeneous(XEN table, XEN homogeneous)
{
  #define H_gtk_table_set_homogeneous "void gtk_table_set_homogeneous(GtkTable* table, gboolean homogeneous)"
  XEN_ASSERT_TYPE(XEN_GtkTable__P(table), table, 1, "gtk_table_set_homogeneous", "GtkTable*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(homogeneous), homogeneous, 2, "gtk_table_set_homogeneous", "gboolean");
  gtk_table_set_homogeneous(XEN_TO_C_GtkTable_(table), XEN_TO_C_gboolean(homogeneous));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_table_get_homogeneous(XEN table)
{
  #define H_gtk_table_get_homogeneous "gboolean gtk_table_get_homogeneous(GtkTable* table)"
  XEN_ASSERT_TYPE(XEN_GtkTable__P(table), table, 1, "gtk_table_get_homogeneous", "GtkTable*");
  return(C_TO_XEN_gboolean(gtk_table_get_homogeneous(XEN_TO_C_GtkTable_(table))));
}
#endif

static XEN gxg_gtk_tearoff_menu_item_get_type(void)
{
  #define H_gtk_tearoff_menu_item_get_type "GtkType gtk_tearoff_menu_item_get_type( void)"
  return(C_TO_XEN_GtkType(gtk_tearoff_menu_item_get_type()));
}

static XEN gxg_gtk_tearoff_menu_item_new(void)
{
  #define H_gtk_tearoff_menu_item_new "GtkWidget* gtk_tearoff_menu_item_new( void)"
  return(C_TO_XEN_GtkWidget_(gtk_tearoff_menu_item_new()));
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_text_buffer_get_type(void)
{
  #define H_gtk_text_buffer_get_type "GType gtk_text_buffer_get_type( void)"
  return(C_TO_XEN_GType(gtk_text_buffer_get_type()));
}

static XEN gxg_gtk_text_buffer_new(XEN table)
{
  #define H_gtk_text_buffer_new "GtkTextBuffer* gtk_text_buffer_new(GtkTextTagTable* table)"
  XEN_ASSERT_TYPE(XEN_GtkTextTagTable__P(table), table, 1, "gtk_text_buffer_new", "GtkTextTagTable*");
  return(C_TO_XEN_GtkTextBuffer_(gtk_text_buffer_new(XEN_TO_C_GtkTextTagTable_(table))));
}

static XEN gxg_gtk_text_buffer_get_line_count(XEN buffer)
{
  #define H_gtk_text_buffer_get_line_count "gint gtk_text_buffer_get_line_count(GtkTextBuffer* buffer)"
  XEN_ASSERT_TYPE(XEN_GtkTextBuffer__P(buffer), buffer, 1, "gtk_text_buffer_get_line_count", "GtkTextBuffer*");
  return(C_TO_XEN_gint(gtk_text_buffer_get_line_count(XEN_TO_C_GtkTextBuffer_(buffer))));
}

static XEN gxg_gtk_text_buffer_get_char_count(XEN buffer)
{
  #define H_gtk_text_buffer_get_char_count "gint gtk_text_buffer_get_char_count(GtkTextBuffer* buffer)"
  XEN_ASSERT_TYPE(XEN_GtkTextBuffer__P(buffer), buffer, 1, "gtk_text_buffer_get_char_count", "GtkTextBuffer*");
  return(C_TO_XEN_gint(gtk_text_buffer_get_char_count(XEN_TO_C_GtkTextBuffer_(buffer))));
}

static XEN gxg_gtk_text_buffer_get_tag_table(XEN buffer)
{
  #define H_gtk_text_buffer_get_tag_table "GtkTextTagTable* gtk_text_buffer_get_tag_table(GtkTextBuffer* buffer)"
  XEN_ASSERT_TYPE(XEN_GtkTextBuffer__P(buffer), buffer, 1, "gtk_text_buffer_get_tag_table", "GtkTextBuffer*");
  return(C_TO_XEN_GtkTextTagTable_(gtk_text_buffer_get_tag_table(XEN_TO_C_GtkTextBuffer_(buffer))));
}

static XEN gxg_gtk_text_buffer_set_text(XEN buffer, XEN text, XEN len)
{
  #define H_gtk_text_buffer_set_text "void gtk_text_buffer_set_text(GtkTextBuffer* buffer, gchar* text, \
gint len)"
  XEN_ASSERT_TYPE(XEN_GtkTextBuffer__P(buffer), buffer, 1, "gtk_text_buffer_set_text", "GtkTextBuffer*");
  XEN_ASSERT_TYPE(XEN_gchar__P(text), text, 2, "gtk_text_buffer_set_text", "gchar*");
  XEN_ASSERT_TYPE(XEN_gint_P(len), len, 3, "gtk_text_buffer_set_text", "gint");
  gtk_text_buffer_set_text(XEN_TO_C_GtkTextBuffer_(buffer), XEN_TO_C_gchar_(text), XEN_TO_C_gint(len));
  return(XEN_FALSE);
}

static XEN gxg_gtk_text_buffer_insert(XEN buffer, XEN iter, XEN text, XEN len)
{
  #define H_gtk_text_buffer_insert "void gtk_text_buffer_insert(GtkTextBuffer* buffer, GtkTextIter* iter, \
gchar* text, gint len)"
  XEN_ASSERT_TYPE(XEN_GtkTextBuffer__P(buffer), buffer, 1, "gtk_text_buffer_insert", "GtkTextBuffer*");
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(iter), iter, 2, "gtk_text_buffer_insert", "GtkTextIter*");
  XEN_ASSERT_TYPE(XEN_gchar__P(text), text, 3, "gtk_text_buffer_insert", "gchar*");
  XEN_ASSERT_TYPE(XEN_gint_P(len), len, 4, "gtk_text_buffer_insert", "gint");
  gtk_text_buffer_insert(XEN_TO_C_GtkTextBuffer_(buffer), XEN_TO_C_GtkTextIter_(iter), XEN_TO_C_gchar_(text), XEN_TO_C_gint(len));
  return(XEN_FALSE);
}

static XEN gxg_gtk_text_buffer_insert_at_cursor(XEN buffer, XEN text, XEN len)
{
  #define H_gtk_text_buffer_insert_at_cursor "void gtk_text_buffer_insert_at_cursor(GtkTextBuffer* buffer, \
gchar* text, gint len)"
  XEN_ASSERT_TYPE(XEN_GtkTextBuffer__P(buffer), buffer, 1, "gtk_text_buffer_insert_at_cursor", "GtkTextBuffer*");
  XEN_ASSERT_TYPE(XEN_gchar__P(text), text, 2, "gtk_text_buffer_insert_at_cursor", "gchar*");
  XEN_ASSERT_TYPE(XEN_gint_P(len), len, 3, "gtk_text_buffer_insert_at_cursor", "gint");
  gtk_text_buffer_insert_at_cursor(XEN_TO_C_GtkTextBuffer_(buffer), XEN_TO_C_gchar_(text), XEN_TO_C_gint(len));
  return(XEN_FALSE);
}

static XEN gxg_gtk_text_buffer_insert_interactive(XEN buffer, XEN iter, XEN text, XEN len, XEN default_editable)
{
  #define H_gtk_text_buffer_insert_interactive "gboolean gtk_text_buffer_insert_interactive(GtkTextBuffer* buffer, \
GtkTextIter* iter, gchar* text, gint len, gboolean default_editable)"
  XEN_ASSERT_TYPE(XEN_GtkTextBuffer__P(buffer), buffer, 1, "gtk_text_buffer_insert_interactive", "GtkTextBuffer*");
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(iter), iter, 2, "gtk_text_buffer_insert_interactive", "GtkTextIter*");
  XEN_ASSERT_TYPE(XEN_gchar__P(text), text, 3, "gtk_text_buffer_insert_interactive", "gchar*");
  XEN_ASSERT_TYPE(XEN_gint_P(len), len, 4, "gtk_text_buffer_insert_interactive", "gint");
  XEN_ASSERT_TYPE(XEN_gboolean_P(default_editable), default_editable, 5, "gtk_text_buffer_insert_interactive", "gboolean");
  return(C_TO_XEN_gboolean(gtk_text_buffer_insert_interactive(XEN_TO_C_GtkTextBuffer_(buffer), XEN_TO_C_GtkTextIter_(iter), 
                                                              XEN_TO_C_gchar_(text), XEN_TO_C_gint(len), XEN_TO_C_gboolean(default_editable))));
}

static XEN gxg_gtk_text_buffer_insert_interactive_at_cursor(XEN buffer, XEN text, XEN len, XEN default_editable)
{
  #define H_gtk_text_buffer_insert_interactive_at_cursor "gboolean gtk_text_buffer_insert_interactive_at_cursor(GtkTextBuffer* buffer, \
gchar* text, gint len, gboolean default_editable)"
  XEN_ASSERT_TYPE(XEN_GtkTextBuffer__P(buffer), buffer, 1, "gtk_text_buffer_insert_interactive_at_cursor", "GtkTextBuffer*");
  XEN_ASSERT_TYPE(XEN_gchar__P(text), text, 2, "gtk_text_buffer_insert_interactive_at_cursor", "gchar*");
  XEN_ASSERT_TYPE(XEN_gint_P(len), len, 3, "gtk_text_buffer_insert_interactive_at_cursor", "gint");
  XEN_ASSERT_TYPE(XEN_gboolean_P(default_editable), default_editable, 4, "gtk_text_buffer_insert_interactive_at_cursor", "gboolean");
  return(C_TO_XEN_gboolean(gtk_text_buffer_insert_interactive_at_cursor(XEN_TO_C_GtkTextBuffer_(buffer), XEN_TO_C_gchar_(text), 
                                                                        XEN_TO_C_gint(len), XEN_TO_C_gboolean(default_editable))));
}

static XEN gxg_gtk_text_buffer_insert_range(XEN buffer, XEN iter, XEN start, XEN end)
{
  #define H_gtk_text_buffer_insert_range "void gtk_text_buffer_insert_range(GtkTextBuffer* buffer, GtkTextIter* iter, \
GtkTextIter* start, GtkTextIter* end)"
  XEN_ASSERT_TYPE(XEN_GtkTextBuffer__P(buffer), buffer, 1, "gtk_text_buffer_insert_range", "GtkTextBuffer*");
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(iter), iter, 2, "gtk_text_buffer_insert_range", "GtkTextIter*");
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(start), start, 3, "gtk_text_buffer_insert_range", "GtkTextIter*");
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(end), end, 4, "gtk_text_buffer_insert_range", "GtkTextIter*");
  gtk_text_buffer_insert_range(XEN_TO_C_GtkTextBuffer_(buffer), XEN_TO_C_GtkTextIter_(iter), XEN_TO_C_GtkTextIter_(start), 
                               XEN_TO_C_GtkTextIter_(end));
  return(XEN_FALSE);
}

static XEN gxg_gtk_text_buffer_insert_range_interactive(XEN buffer, XEN iter, XEN start, XEN end, XEN default_editable)
{
  #define H_gtk_text_buffer_insert_range_interactive "gboolean gtk_text_buffer_insert_range_interactive(GtkTextBuffer* buffer, \
GtkTextIter* iter, GtkTextIter* start, GtkTextIter* end, gboolean default_editable)"
  XEN_ASSERT_TYPE(XEN_GtkTextBuffer__P(buffer), buffer, 1, "gtk_text_buffer_insert_range_interactive", "GtkTextBuffer*");
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(iter), iter, 2, "gtk_text_buffer_insert_range_interactive", "GtkTextIter*");
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(start), start, 3, "gtk_text_buffer_insert_range_interactive", "GtkTextIter*");
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(end), end, 4, "gtk_text_buffer_insert_range_interactive", "GtkTextIter*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(default_editable), default_editable, 5, "gtk_text_buffer_insert_range_interactive", "gboolean");
  return(C_TO_XEN_gboolean(gtk_text_buffer_insert_range_interactive(XEN_TO_C_GtkTextBuffer_(buffer), XEN_TO_C_GtkTextIter_(iter), 
                                                                    XEN_TO_C_GtkTextIter_(start), XEN_TO_C_GtkTextIter_(end), 
                                                                    XEN_TO_C_gboolean(default_editable))));
}

static XEN gxg_gtk_text_buffer_insert_with_tags(XEN buffer, XEN iter, XEN text, XEN len, XEN first_tag)
{
  #define H_gtk_text_buffer_insert_with_tags "void gtk_text_buffer_insert_with_tags(GtkTextBuffer* buffer, \
GtkTextIter* iter, gchar* text, gint len, GtkTextTag* first_tag, ...)"
  XEN_ASSERT_TYPE(XEN_GtkTextBuffer__P(buffer), buffer, 1, "gtk_text_buffer_insert_with_tags", "GtkTextBuffer*");
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(iter), iter, 2, "gtk_text_buffer_insert_with_tags", "GtkTextIter*");
  XEN_ASSERT_TYPE(XEN_gchar__P(text), text, 3, "gtk_text_buffer_insert_with_tags", "gchar*");
  XEN_ASSERT_TYPE(XEN_gint_P(len), len, 4, "gtk_text_buffer_insert_with_tags", "gint");
  XEN_ASSERT_TYPE(XEN_GtkTextTag__P(first_tag), first_tag, 5, "gtk_text_buffer_insert_with_tags", "GtkTextTag*");
  gtk_text_buffer_insert_with_tags(XEN_TO_C_GtkTextBuffer_(buffer), XEN_TO_C_GtkTextIter_(iter), XEN_TO_C_gchar_(text), XEN_TO_C_gint(len), 
                                   XEN_TO_C_GtkTextTag_(first_tag));
  return(XEN_FALSE);
}

static XEN gxg_gtk_text_buffer_insert_with_tags_by_name(XEN buffer, XEN iter, XEN text, XEN len, XEN first_tag_name)
{
  #define H_gtk_text_buffer_insert_with_tags_by_name "void gtk_text_buffer_insert_with_tags_by_name(GtkTextBuffer* buffer, \
GtkTextIter* iter, gchar* text, gint len, gchar* first_tag_name, ...)"
  XEN_ASSERT_TYPE(XEN_GtkTextBuffer__P(buffer), buffer, 1, "gtk_text_buffer_insert_with_tags_by_name", "GtkTextBuffer*");
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(iter), iter, 2, "gtk_text_buffer_insert_with_tags_by_name", "GtkTextIter*");
  XEN_ASSERT_TYPE(XEN_gchar__P(text), text, 3, "gtk_text_buffer_insert_with_tags_by_name", "gchar*");
  XEN_ASSERT_TYPE(XEN_gint_P(len), len, 4, "gtk_text_buffer_insert_with_tags_by_name", "gint");
  XEN_ASSERT_TYPE(XEN_gchar__P(first_tag_name), first_tag_name, 5, "gtk_text_buffer_insert_with_tags_by_name", "gchar*");
  gtk_text_buffer_insert_with_tags_by_name(XEN_TO_C_GtkTextBuffer_(buffer), XEN_TO_C_GtkTextIter_(iter), XEN_TO_C_gchar_(text), 
                                           XEN_TO_C_gint(len), XEN_TO_C_gchar_(first_tag_name));
  return(XEN_FALSE);
}

static XEN gxg_gtk_text_buffer_delete(XEN buffer, XEN start, XEN end)
{
  #define H_gtk_text_buffer_delete "void gtk_text_buffer_delete(GtkTextBuffer* buffer, GtkTextIter* start, \
GtkTextIter* end)"
  XEN_ASSERT_TYPE(XEN_GtkTextBuffer__P(buffer), buffer, 1, "gtk_text_buffer_delete", "GtkTextBuffer*");
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(start), start, 2, "gtk_text_buffer_delete", "GtkTextIter*");
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(end), end, 3, "gtk_text_buffer_delete", "GtkTextIter*");
  gtk_text_buffer_delete(XEN_TO_C_GtkTextBuffer_(buffer), XEN_TO_C_GtkTextIter_(start), XEN_TO_C_GtkTextIter_(end));
  return(XEN_FALSE);
}

static XEN gxg_gtk_text_buffer_delete_interactive(XEN buffer, XEN start_iter, XEN end_iter, XEN default_editable)
{
  #define H_gtk_text_buffer_delete_interactive "gboolean gtk_text_buffer_delete_interactive(GtkTextBuffer* buffer, \
GtkTextIter* start_iter, GtkTextIter* end_iter, gboolean default_editable)"
  XEN_ASSERT_TYPE(XEN_GtkTextBuffer__P(buffer), buffer, 1, "gtk_text_buffer_delete_interactive", "GtkTextBuffer*");
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(start_iter), start_iter, 2, "gtk_text_buffer_delete_interactive", "GtkTextIter*");
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(end_iter), end_iter, 3, "gtk_text_buffer_delete_interactive", "GtkTextIter*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(default_editable), default_editable, 4, "gtk_text_buffer_delete_interactive", "gboolean");
  return(C_TO_XEN_gboolean(gtk_text_buffer_delete_interactive(XEN_TO_C_GtkTextBuffer_(buffer), XEN_TO_C_GtkTextIter_(start_iter), 
                                                              XEN_TO_C_GtkTextIter_(end_iter), XEN_TO_C_gboolean(default_editable))));
}

static XEN gxg_gtk_text_buffer_get_text(XEN buffer, XEN start, XEN end, XEN include_hidden_chars)
{
  #define H_gtk_text_buffer_get_text "gchar* gtk_text_buffer_get_text(GtkTextBuffer* buffer, GtkTextIter* start, \
GtkTextIter* end, gboolean include_hidden_chars)"
  XEN_ASSERT_TYPE(XEN_GtkTextBuffer__P(buffer), buffer, 1, "gtk_text_buffer_get_text", "GtkTextBuffer*");
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(start), start, 2, "gtk_text_buffer_get_text", "GtkTextIter*");
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(end), end, 3, "gtk_text_buffer_get_text", "GtkTextIter*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(include_hidden_chars), include_hidden_chars, 4, "gtk_text_buffer_get_text", "gboolean");
  return(C_TO_XEN_gchar_(gtk_text_buffer_get_text(XEN_TO_C_GtkTextBuffer_(buffer), XEN_TO_C_GtkTextIter_(start), XEN_TO_C_GtkTextIter_(end), 
                                                  XEN_TO_C_gboolean(include_hidden_chars))));
}

static XEN gxg_gtk_text_buffer_get_slice(XEN buffer, XEN start, XEN end, XEN include_hidden_chars)
{
  #define H_gtk_text_buffer_get_slice "gchar* gtk_text_buffer_get_slice(GtkTextBuffer* buffer, GtkTextIter* start, \
GtkTextIter* end, gboolean include_hidden_chars)"
  XEN_ASSERT_TYPE(XEN_GtkTextBuffer__P(buffer), buffer, 1, "gtk_text_buffer_get_slice", "GtkTextBuffer*");
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(start), start, 2, "gtk_text_buffer_get_slice", "GtkTextIter*");
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(end), end, 3, "gtk_text_buffer_get_slice", "GtkTextIter*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(include_hidden_chars), include_hidden_chars, 4, "gtk_text_buffer_get_slice", "gboolean");
  return(C_TO_XEN_gchar_(gtk_text_buffer_get_slice(XEN_TO_C_GtkTextBuffer_(buffer), XEN_TO_C_GtkTextIter_(start), XEN_TO_C_GtkTextIter_(end), 
                                                   XEN_TO_C_gboolean(include_hidden_chars))));
}

static XEN gxg_gtk_text_buffer_insert_pixbuf(XEN buffer, XEN iter, XEN pixbuf)
{
  #define H_gtk_text_buffer_insert_pixbuf "void gtk_text_buffer_insert_pixbuf(GtkTextBuffer* buffer, \
GtkTextIter* iter, GdkPixbuf* pixbuf)"
  XEN_ASSERT_TYPE(XEN_GtkTextBuffer__P(buffer), buffer, 1, "gtk_text_buffer_insert_pixbuf", "GtkTextBuffer*");
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(iter), iter, 2, "gtk_text_buffer_insert_pixbuf", "GtkTextIter*");
  XEN_ASSERT_TYPE(XEN_GdkPixbuf__P(pixbuf), pixbuf, 3, "gtk_text_buffer_insert_pixbuf", "GdkPixbuf*");
  gtk_text_buffer_insert_pixbuf(XEN_TO_C_GtkTextBuffer_(buffer), XEN_TO_C_GtkTextIter_(iter), XEN_TO_C_GdkPixbuf_(pixbuf));
  return(XEN_FALSE);
}

static XEN gxg_gtk_text_buffer_insert_child_anchor(XEN buffer, XEN iter, XEN anchor)
{
  #define H_gtk_text_buffer_insert_child_anchor "void gtk_text_buffer_insert_child_anchor(GtkTextBuffer* buffer, \
GtkTextIter* iter, GtkTextChildAnchor* anchor)"
  XEN_ASSERT_TYPE(XEN_GtkTextBuffer__P(buffer), buffer, 1, "gtk_text_buffer_insert_child_anchor", "GtkTextBuffer*");
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(iter), iter, 2, "gtk_text_buffer_insert_child_anchor", "GtkTextIter*");
  XEN_ASSERT_TYPE(XEN_GtkTextChildAnchor__P(anchor), anchor, 3, "gtk_text_buffer_insert_child_anchor", "GtkTextChildAnchor*");
  gtk_text_buffer_insert_child_anchor(XEN_TO_C_GtkTextBuffer_(buffer), XEN_TO_C_GtkTextIter_(iter), XEN_TO_C_GtkTextChildAnchor_(anchor));
  return(XEN_FALSE);
}

static XEN gxg_gtk_text_buffer_create_child_anchor(XEN buffer, XEN iter)
{
  #define H_gtk_text_buffer_create_child_anchor "GtkTextChildAnchor* gtk_text_buffer_create_child_anchor(GtkTextBuffer* buffer, \
GtkTextIter* iter)"
  XEN_ASSERT_TYPE(XEN_GtkTextBuffer__P(buffer), buffer, 1, "gtk_text_buffer_create_child_anchor", "GtkTextBuffer*");
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(iter), iter, 2, "gtk_text_buffer_create_child_anchor", "GtkTextIter*");
  return(C_TO_XEN_GtkTextChildAnchor_(gtk_text_buffer_create_child_anchor(XEN_TO_C_GtkTextBuffer_(buffer), XEN_TO_C_GtkTextIter_(iter))));
}

static XEN gxg_gtk_text_buffer_create_mark(XEN buffer, XEN mark_name, XEN where, XEN left_gravity)
{
  #define H_gtk_text_buffer_create_mark "GtkTextMark* gtk_text_buffer_create_mark(GtkTextBuffer* buffer, \
gchar* mark_name, GtkTextIter* where, gboolean left_gravity)"
  XEN_ASSERT_TYPE(XEN_GtkTextBuffer__P(buffer), buffer, 1, "gtk_text_buffer_create_mark", "GtkTextBuffer*");
  XEN_ASSERT_TYPE(XEN_gchar__P(mark_name), mark_name, 2, "gtk_text_buffer_create_mark", "gchar*");
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(where), where, 3, "gtk_text_buffer_create_mark", "GtkTextIter*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(left_gravity), left_gravity, 4, "gtk_text_buffer_create_mark", "gboolean");
  return(C_TO_XEN_GtkTextMark_(gtk_text_buffer_create_mark(XEN_TO_C_GtkTextBuffer_(buffer), XEN_TO_C_gchar_(mark_name), XEN_TO_C_GtkTextIter_(where), 
                                                           XEN_TO_C_gboolean(left_gravity))));
}

static XEN gxg_gtk_text_buffer_move_mark(XEN buffer, XEN mark, XEN where)
{
  #define H_gtk_text_buffer_move_mark "void gtk_text_buffer_move_mark(GtkTextBuffer* buffer, GtkTextMark* mark, \
GtkTextIter* where)"
  XEN_ASSERT_TYPE(XEN_GtkTextBuffer__P(buffer), buffer, 1, "gtk_text_buffer_move_mark", "GtkTextBuffer*");
  XEN_ASSERT_TYPE(XEN_GtkTextMark__P(mark), mark, 2, "gtk_text_buffer_move_mark", "GtkTextMark*");
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(where), where, 3, "gtk_text_buffer_move_mark", "GtkTextIter*");
  gtk_text_buffer_move_mark(XEN_TO_C_GtkTextBuffer_(buffer), XEN_TO_C_GtkTextMark_(mark), XEN_TO_C_GtkTextIter_(where));
  return(XEN_FALSE);
}

static XEN gxg_gtk_text_buffer_delete_mark(XEN buffer, XEN mark)
{
  #define H_gtk_text_buffer_delete_mark "void gtk_text_buffer_delete_mark(GtkTextBuffer* buffer, GtkTextMark* mark)"
  XEN_ASSERT_TYPE(XEN_GtkTextBuffer__P(buffer), buffer, 1, "gtk_text_buffer_delete_mark", "GtkTextBuffer*");
  XEN_ASSERT_TYPE(XEN_GtkTextMark__P(mark), mark, 2, "gtk_text_buffer_delete_mark", "GtkTextMark*");
  gtk_text_buffer_delete_mark(XEN_TO_C_GtkTextBuffer_(buffer), XEN_TO_C_GtkTextMark_(mark));
  return(XEN_FALSE);
}

static XEN gxg_gtk_text_buffer_get_mark(XEN buffer, XEN name)
{
  #define H_gtk_text_buffer_get_mark "GtkTextMark* gtk_text_buffer_get_mark(GtkTextBuffer* buffer, gchar* name)"
  XEN_ASSERT_TYPE(XEN_GtkTextBuffer__P(buffer), buffer, 1, "gtk_text_buffer_get_mark", "GtkTextBuffer*");
  XEN_ASSERT_TYPE(XEN_gchar__P(name), name, 2, "gtk_text_buffer_get_mark", "gchar*");
  return(C_TO_XEN_GtkTextMark_(gtk_text_buffer_get_mark(XEN_TO_C_GtkTextBuffer_(buffer), XEN_TO_C_gchar_(name))));
}

static XEN gxg_gtk_text_buffer_move_mark_by_name(XEN buffer, XEN name, XEN where)
{
  #define H_gtk_text_buffer_move_mark_by_name "void gtk_text_buffer_move_mark_by_name(GtkTextBuffer* buffer, \
gchar* name, GtkTextIter* where)"
  XEN_ASSERT_TYPE(XEN_GtkTextBuffer__P(buffer), buffer, 1, "gtk_text_buffer_move_mark_by_name", "GtkTextBuffer*");
  XEN_ASSERT_TYPE(XEN_gchar__P(name), name, 2, "gtk_text_buffer_move_mark_by_name", "gchar*");
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(where), where, 3, "gtk_text_buffer_move_mark_by_name", "GtkTextIter*");
  gtk_text_buffer_move_mark_by_name(XEN_TO_C_GtkTextBuffer_(buffer), XEN_TO_C_gchar_(name), XEN_TO_C_GtkTextIter_(where));
  return(XEN_FALSE);
}

static XEN gxg_gtk_text_buffer_delete_mark_by_name(XEN buffer, XEN name)
{
  #define H_gtk_text_buffer_delete_mark_by_name "void gtk_text_buffer_delete_mark_by_name(GtkTextBuffer* buffer, \
gchar* name)"
  XEN_ASSERT_TYPE(XEN_GtkTextBuffer__P(buffer), buffer, 1, "gtk_text_buffer_delete_mark_by_name", "GtkTextBuffer*");
  XEN_ASSERT_TYPE(XEN_gchar__P(name), name, 2, "gtk_text_buffer_delete_mark_by_name", "gchar*");
  gtk_text_buffer_delete_mark_by_name(XEN_TO_C_GtkTextBuffer_(buffer), XEN_TO_C_gchar_(name));
  return(XEN_FALSE);
}

static XEN gxg_gtk_text_buffer_get_insert(XEN buffer)
{
  #define H_gtk_text_buffer_get_insert "GtkTextMark* gtk_text_buffer_get_insert(GtkTextBuffer* buffer)"
  XEN_ASSERT_TYPE(XEN_GtkTextBuffer__P(buffer), buffer, 1, "gtk_text_buffer_get_insert", "GtkTextBuffer*");
  return(C_TO_XEN_GtkTextMark_(gtk_text_buffer_get_insert(XEN_TO_C_GtkTextBuffer_(buffer))));
}

static XEN gxg_gtk_text_buffer_get_selection_bound(XEN buffer)
{
  #define H_gtk_text_buffer_get_selection_bound "GtkTextMark* gtk_text_buffer_get_selection_bound(GtkTextBuffer* buffer)"
  XEN_ASSERT_TYPE(XEN_GtkTextBuffer__P(buffer), buffer, 1, "gtk_text_buffer_get_selection_bound", "GtkTextBuffer*");
  return(C_TO_XEN_GtkTextMark_(gtk_text_buffer_get_selection_bound(XEN_TO_C_GtkTextBuffer_(buffer))));
}

static XEN gxg_gtk_text_buffer_place_cursor(XEN buffer, XEN where)
{
  #define H_gtk_text_buffer_place_cursor "void gtk_text_buffer_place_cursor(GtkTextBuffer* buffer, GtkTextIter* where)"
  XEN_ASSERT_TYPE(XEN_GtkTextBuffer__P(buffer), buffer, 1, "gtk_text_buffer_place_cursor", "GtkTextBuffer*");
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(where), where, 2, "gtk_text_buffer_place_cursor", "GtkTextIter*");
  gtk_text_buffer_place_cursor(XEN_TO_C_GtkTextBuffer_(buffer), XEN_TO_C_GtkTextIter_(where));
  return(XEN_FALSE);
}

static XEN gxg_gtk_text_buffer_apply_tag(XEN buffer, XEN tag, XEN start, XEN end)
{
  #define H_gtk_text_buffer_apply_tag "void gtk_text_buffer_apply_tag(GtkTextBuffer* buffer, GtkTextTag* tag, \
GtkTextIter* start, GtkTextIter* end)"
  XEN_ASSERT_TYPE(XEN_GtkTextBuffer__P(buffer), buffer, 1, "gtk_text_buffer_apply_tag", "GtkTextBuffer*");
  XEN_ASSERT_TYPE(XEN_GtkTextTag__P(tag), tag, 2, "gtk_text_buffer_apply_tag", "GtkTextTag*");
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(start), start, 3, "gtk_text_buffer_apply_tag", "GtkTextIter*");
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(end), end, 4, "gtk_text_buffer_apply_tag", "GtkTextIter*");
  gtk_text_buffer_apply_tag(XEN_TO_C_GtkTextBuffer_(buffer), XEN_TO_C_GtkTextTag_(tag), XEN_TO_C_GtkTextIter_(start), XEN_TO_C_GtkTextIter_(end));
  return(XEN_FALSE);
}

static XEN gxg_gtk_text_buffer_remove_tag(XEN buffer, XEN tag, XEN start, XEN end)
{
  #define H_gtk_text_buffer_remove_tag "void gtk_text_buffer_remove_tag(GtkTextBuffer* buffer, GtkTextTag* tag, \
GtkTextIter* start, GtkTextIter* end)"
  XEN_ASSERT_TYPE(XEN_GtkTextBuffer__P(buffer), buffer, 1, "gtk_text_buffer_remove_tag", "GtkTextBuffer*");
  XEN_ASSERT_TYPE(XEN_GtkTextTag__P(tag), tag, 2, "gtk_text_buffer_remove_tag", "GtkTextTag*");
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(start), start, 3, "gtk_text_buffer_remove_tag", "GtkTextIter*");
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(end), end, 4, "gtk_text_buffer_remove_tag", "GtkTextIter*");
  gtk_text_buffer_remove_tag(XEN_TO_C_GtkTextBuffer_(buffer), XEN_TO_C_GtkTextTag_(tag), XEN_TO_C_GtkTextIter_(start), XEN_TO_C_GtkTextIter_(end));
  return(XEN_FALSE);
}

static XEN gxg_gtk_text_buffer_apply_tag_by_name(XEN buffer, XEN name, XEN start, XEN end)
{
  #define H_gtk_text_buffer_apply_tag_by_name "void gtk_text_buffer_apply_tag_by_name(GtkTextBuffer* buffer, \
gchar* name, GtkTextIter* start, GtkTextIter* end)"
  XEN_ASSERT_TYPE(XEN_GtkTextBuffer__P(buffer), buffer, 1, "gtk_text_buffer_apply_tag_by_name", "GtkTextBuffer*");
  XEN_ASSERT_TYPE(XEN_gchar__P(name), name, 2, "gtk_text_buffer_apply_tag_by_name", "gchar*");
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(start), start, 3, "gtk_text_buffer_apply_tag_by_name", "GtkTextIter*");
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(end), end, 4, "gtk_text_buffer_apply_tag_by_name", "GtkTextIter*");
  gtk_text_buffer_apply_tag_by_name(XEN_TO_C_GtkTextBuffer_(buffer), XEN_TO_C_gchar_(name), XEN_TO_C_GtkTextIter_(start), 
                                    XEN_TO_C_GtkTextIter_(end));
  return(XEN_FALSE);
}

static XEN gxg_gtk_text_buffer_remove_tag_by_name(XEN buffer, XEN name, XEN start, XEN end)
{
  #define H_gtk_text_buffer_remove_tag_by_name "void gtk_text_buffer_remove_tag_by_name(GtkTextBuffer* buffer, \
gchar* name, GtkTextIter* start, GtkTextIter* end)"
  XEN_ASSERT_TYPE(XEN_GtkTextBuffer__P(buffer), buffer, 1, "gtk_text_buffer_remove_tag_by_name", "GtkTextBuffer*");
  XEN_ASSERT_TYPE(XEN_gchar__P(name), name, 2, "gtk_text_buffer_remove_tag_by_name", "gchar*");
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(start), start, 3, "gtk_text_buffer_remove_tag_by_name", "GtkTextIter*");
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(end), end, 4, "gtk_text_buffer_remove_tag_by_name", "GtkTextIter*");
  gtk_text_buffer_remove_tag_by_name(XEN_TO_C_GtkTextBuffer_(buffer), XEN_TO_C_gchar_(name), XEN_TO_C_GtkTextIter_(start), 
                                     XEN_TO_C_GtkTextIter_(end));
  return(XEN_FALSE);
}

static XEN gxg_gtk_text_buffer_remove_all_tags(XEN buffer, XEN start, XEN end)
{
  #define H_gtk_text_buffer_remove_all_tags "void gtk_text_buffer_remove_all_tags(GtkTextBuffer* buffer, \
GtkTextIter* start, GtkTextIter* end)"
  XEN_ASSERT_TYPE(XEN_GtkTextBuffer__P(buffer), buffer, 1, "gtk_text_buffer_remove_all_tags", "GtkTextBuffer*");
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(start), start, 2, "gtk_text_buffer_remove_all_tags", "GtkTextIter*");
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(end), end, 3, "gtk_text_buffer_remove_all_tags", "GtkTextIter*");
  gtk_text_buffer_remove_all_tags(XEN_TO_C_GtkTextBuffer_(buffer), XEN_TO_C_GtkTextIter_(start), XEN_TO_C_GtkTextIter_(end));
  return(XEN_FALSE);
}

static XEN gxg_gtk_text_buffer_create_tag(XEN buffer, XEN tag_name, XEN first_property_name)
{
  #define H_gtk_text_buffer_create_tag "GtkTextTag* gtk_text_buffer_create_tag(GtkTextBuffer* buffer, \
gchar* tag_name, gchar* first_property_name, ...)"
  XEN_ASSERT_TYPE(XEN_GtkTextBuffer__P(buffer), buffer, 1, "gtk_text_buffer_create_tag", "GtkTextBuffer*");
  XEN_ASSERT_TYPE(XEN_gchar__P(tag_name), tag_name, 2, "gtk_text_buffer_create_tag", "gchar*");
  XEN_ASSERT_TYPE(XEN_gchar__P(first_property_name), first_property_name, 3, "gtk_text_buffer_create_tag", "gchar*");
  return(C_TO_XEN_GtkTextTag_(gtk_text_buffer_create_tag(XEN_TO_C_GtkTextBuffer_(buffer), XEN_TO_C_gchar_(tag_name), XEN_TO_C_gchar_(first_property_name))));
}

static XEN gxg_gtk_text_buffer_get_iter_at_line_offset(XEN buffer, XEN iter, XEN line_number, XEN char_offset)
{
  #define H_gtk_text_buffer_get_iter_at_line_offset "void gtk_text_buffer_get_iter_at_line_offset(GtkTextBuffer* buffer, \
GtkTextIter* iter, gint line_number, gint char_offset)"
  XEN_ASSERT_TYPE(XEN_GtkTextBuffer__P(buffer), buffer, 1, "gtk_text_buffer_get_iter_at_line_offset", "GtkTextBuffer*");
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(iter), iter, 2, "gtk_text_buffer_get_iter_at_line_offset", "GtkTextIter*");
  XEN_ASSERT_TYPE(XEN_gint_P(line_number), line_number, 3, "gtk_text_buffer_get_iter_at_line_offset", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(char_offset), char_offset, 4, "gtk_text_buffer_get_iter_at_line_offset", "gint");
  gtk_text_buffer_get_iter_at_line_offset(XEN_TO_C_GtkTextBuffer_(buffer), XEN_TO_C_GtkTextIter_(iter), XEN_TO_C_gint(line_number), 
                                          XEN_TO_C_gint(char_offset));
  return(XEN_FALSE);
}

static XEN gxg_gtk_text_buffer_get_iter_at_line_index(XEN buffer, XEN iter, XEN line_number, XEN byte_index)
{
  #define H_gtk_text_buffer_get_iter_at_line_index "void gtk_text_buffer_get_iter_at_line_index(GtkTextBuffer* buffer, \
GtkTextIter* iter, gint line_number, gint byte_index)"
  XEN_ASSERT_TYPE(XEN_GtkTextBuffer__P(buffer), buffer, 1, "gtk_text_buffer_get_iter_at_line_index", "GtkTextBuffer*");
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(iter), iter, 2, "gtk_text_buffer_get_iter_at_line_index", "GtkTextIter*");
  XEN_ASSERT_TYPE(XEN_gint_P(line_number), line_number, 3, "gtk_text_buffer_get_iter_at_line_index", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(byte_index), byte_index, 4, "gtk_text_buffer_get_iter_at_line_index", "gint");
  gtk_text_buffer_get_iter_at_line_index(XEN_TO_C_GtkTextBuffer_(buffer), XEN_TO_C_GtkTextIter_(iter), XEN_TO_C_gint(line_number), 
                                         XEN_TO_C_gint(byte_index));
  return(XEN_FALSE);
}

static XEN gxg_gtk_text_buffer_get_iter_at_offset(XEN buffer, XEN iter, XEN char_offset)
{
  #define H_gtk_text_buffer_get_iter_at_offset "void gtk_text_buffer_get_iter_at_offset(GtkTextBuffer* buffer, \
GtkTextIter* iter, gint char_offset)"
  XEN_ASSERT_TYPE(XEN_GtkTextBuffer__P(buffer), buffer, 1, "gtk_text_buffer_get_iter_at_offset", "GtkTextBuffer*");
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(iter), iter, 2, "gtk_text_buffer_get_iter_at_offset", "GtkTextIter*");
  XEN_ASSERT_TYPE(XEN_gint_P(char_offset), char_offset, 3, "gtk_text_buffer_get_iter_at_offset", "gint");
  gtk_text_buffer_get_iter_at_offset(XEN_TO_C_GtkTextBuffer_(buffer), XEN_TO_C_GtkTextIter_(iter), XEN_TO_C_gint(char_offset));
  return(XEN_FALSE);
}

static XEN gxg_gtk_text_buffer_get_iter_at_line(XEN buffer, XEN iter, XEN line_number)
{
  #define H_gtk_text_buffer_get_iter_at_line "void gtk_text_buffer_get_iter_at_line(GtkTextBuffer* buffer, \
GtkTextIter* iter, gint line_number)"
  XEN_ASSERT_TYPE(XEN_GtkTextBuffer__P(buffer), buffer, 1, "gtk_text_buffer_get_iter_at_line", "GtkTextBuffer*");
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(iter), iter, 2, "gtk_text_buffer_get_iter_at_line", "GtkTextIter*");
  XEN_ASSERT_TYPE(XEN_gint_P(line_number), line_number, 3, "gtk_text_buffer_get_iter_at_line", "gint");
  gtk_text_buffer_get_iter_at_line(XEN_TO_C_GtkTextBuffer_(buffer), XEN_TO_C_GtkTextIter_(iter), XEN_TO_C_gint(line_number));
  return(XEN_FALSE);
}

static XEN gxg_gtk_text_buffer_get_start_iter(XEN buffer, XEN iter)
{
  #define H_gtk_text_buffer_get_start_iter "void gtk_text_buffer_get_start_iter(GtkTextBuffer* buffer, \
GtkTextIter* iter)"
  XEN_ASSERT_TYPE(XEN_GtkTextBuffer__P(buffer), buffer, 1, "gtk_text_buffer_get_start_iter", "GtkTextBuffer*");
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(iter), iter, 2, "gtk_text_buffer_get_start_iter", "GtkTextIter*");
  gtk_text_buffer_get_start_iter(XEN_TO_C_GtkTextBuffer_(buffer), XEN_TO_C_GtkTextIter_(iter));
  return(XEN_FALSE);
}

static XEN gxg_gtk_text_buffer_get_end_iter(XEN buffer, XEN iter)
{
  #define H_gtk_text_buffer_get_end_iter "void gtk_text_buffer_get_end_iter(GtkTextBuffer* buffer, GtkTextIter* iter)"
  XEN_ASSERT_TYPE(XEN_GtkTextBuffer__P(buffer), buffer, 1, "gtk_text_buffer_get_end_iter", "GtkTextBuffer*");
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(iter), iter, 2, "gtk_text_buffer_get_end_iter", "GtkTextIter*");
  gtk_text_buffer_get_end_iter(XEN_TO_C_GtkTextBuffer_(buffer), XEN_TO_C_GtkTextIter_(iter));
  return(XEN_FALSE);
}

static XEN gxg_gtk_text_buffer_get_bounds(XEN buffer, XEN start, XEN end)
{
  #define H_gtk_text_buffer_get_bounds "void gtk_text_buffer_get_bounds(GtkTextBuffer* buffer, GtkTextIter* start, \
GtkTextIter* end)"
  XEN_ASSERT_TYPE(XEN_GtkTextBuffer__P(buffer), buffer, 1, "gtk_text_buffer_get_bounds", "GtkTextBuffer*");
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(start), start, 2, "gtk_text_buffer_get_bounds", "GtkTextIter*");
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(end), end, 3, "gtk_text_buffer_get_bounds", "GtkTextIter*");
  gtk_text_buffer_get_bounds(XEN_TO_C_GtkTextBuffer_(buffer), XEN_TO_C_GtkTextIter_(start), XEN_TO_C_GtkTextIter_(end));
  return(XEN_FALSE);
}

static XEN gxg_gtk_text_buffer_get_iter_at_mark(XEN buffer, XEN iter, XEN mark)
{
  #define H_gtk_text_buffer_get_iter_at_mark "void gtk_text_buffer_get_iter_at_mark(GtkTextBuffer* buffer, \
GtkTextIter* iter, GtkTextMark* mark)"
  XEN_ASSERT_TYPE(XEN_GtkTextBuffer__P(buffer), buffer, 1, "gtk_text_buffer_get_iter_at_mark", "GtkTextBuffer*");
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(iter), iter, 2, "gtk_text_buffer_get_iter_at_mark", "GtkTextIter*");
  XEN_ASSERT_TYPE(XEN_GtkTextMark__P(mark), mark, 3, "gtk_text_buffer_get_iter_at_mark", "GtkTextMark*");
  gtk_text_buffer_get_iter_at_mark(XEN_TO_C_GtkTextBuffer_(buffer), XEN_TO_C_GtkTextIter_(iter), XEN_TO_C_GtkTextMark_(mark));
  return(XEN_FALSE);
}

static XEN gxg_gtk_text_buffer_get_iter_at_child_anchor(XEN buffer, XEN iter, XEN anchor)
{
  #define H_gtk_text_buffer_get_iter_at_child_anchor "void gtk_text_buffer_get_iter_at_child_anchor(GtkTextBuffer* buffer, \
GtkTextIter* iter, GtkTextChildAnchor* anchor)"
  XEN_ASSERT_TYPE(XEN_GtkTextBuffer__P(buffer), buffer, 1, "gtk_text_buffer_get_iter_at_child_anchor", "GtkTextBuffer*");
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(iter), iter, 2, "gtk_text_buffer_get_iter_at_child_anchor", "GtkTextIter*");
  XEN_ASSERT_TYPE(XEN_GtkTextChildAnchor__P(anchor), anchor, 3, "gtk_text_buffer_get_iter_at_child_anchor", "GtkTextChildAnchor*");
  gtk_text_buffer_get_iter_at_child_anchor(XEN_TO_C_GtkTextBuffer_(buffer), XEN_TO_C_GtkTextIter_(iter), XEN_TO_C_GtkTextChildAnchor_(anchor));
  return(XEN_FALSE);
}

static XEN gxg_gtk_text_buffer_get_modified(XEN buffer)
{
  #define H_gtk_text_buffer_get_modified "gboolean gtk_text_buffer_get_modified(GtkTextBuffer* buffer)"
  XEN_ASSERT_TYPE(XEN_GtkTextBuffer__P(buffer), buffer, 1, "gtk_text_buffer_get_modified", "GtkTextBuffer*");
  return(C_TO_XEN_gboolean(gtk_text_buffer_get_modified(XEN_TO_C_GtkTextBuffer_(buffer))));
}

static XEN gxg_gtk_text_buffer_set_modified(XEN buffer, XEN setting)
{
  #define H_gtk_text_buffer_set_modified "void gtk_text_buffer_set_modified(GtkTextBuffer* buffer, gboolean setting)"
  XEN_ASSERT_TYPE(XEN_GtkTextBuffer__P(buffer), buffer, 1, "gtk_text_buffer_set_modified", "GtkTextBuffer*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(setting), setting, 2, "gtk_text_buffer_set_modified", "gboolean");
  gtk_text_buffer_set_modified(XEN_TO_C_GtkTextBuffer_(buffer), XEN_TO_C_gboolean(setting));
  return(XEN_FALSE);
}

static XEN gxg_gtk_text_buffer_add_selection_clipboard(XEN buffer, XEN clipboard)
{
  #define H_gtk_text_buffer_add_selection_clipboard "void gtk_text_buffer_add_selection_clipboard(GtkTextBuffer* buffer, \
GtkClipboard* clipboard)"
  XEN_ASSERT_TYPE(XEN_GtkTextBuffer__P(buffer), buffer, 1, "gtk_text_buffer_add_selection_clipboard", "GtkTextBuffer*");
  XEN_ASSERT_TYPE(XEN_GtkClipboard__P(clipboard), clipboard, 2, "gtk_text_buffer_add_selection_clipboard", "GtkClipboard*");
  gtk_text_buffer_add_selection_clipboard(XEN_TO_C_GtkTextBuffer_(buffer), XEN_TO_C_GtkClipboard_(clipboard));
  return(XEN_FALSE);
}

static XEN gxg_gtk_text_buffer_remove_selection_clipboard(XEN buffer, XEN clipboard)
{
  #define H_gtk_text_buffer_remove_selection_clipboard "void gtk_text_buffer_remove_selection_clipboard(GtkTextBuffer* buffer, \
GtkClipboard* clipboard)"
  XEN_ASSERT_TYPE(XEN_GtkTextBuffer__P(buffer), buffer, 1, "gtk_text_buffer_remove_selection_clipboard", "GtkTextBuffer*");
  XEN_ASSERT_TYPE(XEN_GtkClipboard__P(clipboard), clipboard, 2, "gtk_text_buffer_remove_selection_clipboard", "GtkClipboard*");
  gtk_text_buffer_remove_selection_clipboard(XEN_TO_C_GtkTextBuffer_(buffer), XEN_TO_C_GtkClipboard_(clipboard));
  return(XEN_FALSE);
}

static XEN gxg_gtk_text_buffer_cut_clipboard(XEN buffer, XEN clipboard, XEN default_editable)
{
  #define H_gtk_text_buffer_cut_clipboard "void gtk_text_buffer_cut_clipboard(GtkTextBuffer* buffer, \
GtkClipboard* clipboard, gboolean default_editable)"
  XEN_ASSERT_TYPE(XEN_GtkTextBuffer__P(buffer), buffer, 1, "gtk_text_buffer_cut_clipboard", "GtkTextBuffer*");
  XEN_ASSERT_TYPE(XEN_GtkClipboard__P(clipboard), clipboard, 2, "gtk_text_buffer_cut_clipboard", "GtkClipboard*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(default_editable), default_editable, 3, "gtk_text_buffer_cut_clipboard", "gboolean");
  gtk_text_buffer_cut_clipboard(XEN_TO_C_GtkTextBuffer_(buffer), XEN_TO_C_GtkClipboard_(clipboard), XEN_TO_C_gboolean(default_editable));
  return(XEN_FALSE);
}

static XEN gxg_gtk_text_buffer_copy_clipboard(XEN buffer, XEN clipboard)
{
  #define H_gtk_text_buffer_copy_clipboard "void gtk_text_buffer_copy_clipboard(GtkTextBuffer* buffer, \
GtkClipboard* clipboard)"
  XEN_ASSERT_TYPE(XEN_GtkTextBuffer__P(buffer), buffer, 1, "gtk_text_buffer_copy_clipboard", "GtkTextBuffer*");
  XEN_ASSERT_TYPE(XEN_GtkClipboard__P(clipboard), clipboard, 2, "gtk_text_buffer_copy_clipboard", "GtkClipboard*");
  gtk_text_buffer_copy_clipboard(XEN_TO_C_GtkTextBuffer_(buffer), XEN_TO_C_GtkClipboard_(clipboard));
  return(XEN_FALSE);
}

static XEN gxg_gtk_text_buffer_paste_clipboard(XEN buffer, XEN clipboard, XEN override_location, XEN default_editable)
{
  #define H_gtk_text_buffer_paste_clipboard "void gtk_text_buffer_paste_clipboard(GtkTextBuffer* buffer, \
GtkClipboard* clipboard, GtkTextIter* override_location, gboolean default_editable)"
  XEN_ASSERT_TYPE(XEN_GtkTextBuffer__P(buffer), buffer, 1, "gtk_text_buffer_paste_clipboard", "GtkTextBuffer*");
  XEN_ASSERT_TYPE(XEN_GtkClipboard__P(clipboard), clipboard, 2, "gtk_text_buffer_paste_clipboard", "GtkClipboard*");
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(override_location) || XEN_FALSE_P(override_location), override_location, 3, "gtk_text_buffer_paste_clipboard", "GtkTextIter*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(default_editable), default_editable, 4, "gtk_text_buffer_paste_clipboard", "gboolean");
  gtk_text_buffer_paste_clipboard(XEN_TO_C_GtkTextBuffer_(buffer), XEN_TO_C_GtkClipboard_(clipboard), XEN_TO_C_GtkTextIter_(override_location), 
                                  XEN_TO_C_gboolean(default_editable));
  return(XEN_FALSE);
}

static XEN gxg_gtk_text_buffer_get_selection_bounds(XEN buffer, XEN start, XEN end)
{
  #define H_gtk_text_buffer_get_selection_bounds "gboolean gtk_text_buffer_get_selection_bounds(GtkTextBuffer* buffer, \
GtkTextIter* start, GtkTextIter* end)"
  XEN_ASSERT_TYPE(XEN_GtkTextBuffer__P(buffer), buffer, 1, "gtk_text_buffer_get_selection_bounds", "GtkTextBuffer*");
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(start), start, 2, "gtk_text_buffer_get_selection_bounds", "GtkTextIter*");
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(end), end, 3, "gtk_text_buffer_get_selection_bounds", "GtkTextIter*");
  return(C_TO_XEN_gboolean(gtk_text_buffer_get_selection_bounds(XEN_TO_C_GtkTextBuffer_(buffer), XEN_TO_C_GtkTextIter_(start), 
                                                                XEN_TO_C_GtkTextIter_(end))));
}

static XEN gxg_gtk_text_buffer_delete_selection(XEN buffer, XEN interactive, XEN default_editable)
{
  #define H_gtk_text_buffer_delete_selection "gboolean gtk_text_buffer_delete_selection(GtkTextBuffer* buffer, \
gboolean interactive, gboolean default_editable)"
  XEN_ASSERT_TYPE(XEN_GtkTextBuffer__P(buffer), buffer, 1, "gtk_text_buffer_delete_selection", "GtkTextBuffer*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(interactive), interactive, 2, "gtk_text_buffer_delete_selection", "gboolean");
  XEN_ASSERT_TYPE(XEN_gboolean_P(default_editable), default_editable, 3, "gtk_text_buffer_delete_selection", "gboolean");
  return(C_TO_XEN_gboolean(gtk_text_buffer_delete_selection(XEN_TO_C_GtkTextBuffer_(buffer), XEN_TO_C_gboolean(interactive), 
                                                            XEN_TO_C_gboolean(default_editable))));
}

static XEN gxg_gtk_text_buffer_begin_user_action(XEN buffer)
{
  #define H_gtk_text_buffer_begin_user_action "void gtk_text_buffer_begin_user_action(GtkTextBuffer* buffer)"
  XEN_ASSERT_TYPE(XEN_GtkTextBuffer__P(buffer), buffer, 1, "gtk_text_buffer_begin_user_action", "GtkTextBuffer*");
  gtk_text_buffer_begin_user_action(XEN_TO_C_GtkTextBuffer_(buffer));
  return(XEN_FALSE);
}

static XEN gxg_gtk_text_buffer_end_user_action(XEN buffer)
{
  #define H_gtk_text_buffer_end_user_action "void gtk_text_buffer_end_user_action(GtkTextBuffer* buffer)"
  XEN_ASSERT_TYPE(XEN_GtkTextBuffer__P(buffer), buffer, 1, "gtk_text_buffer_end_user_action", "GtkTextBuffer*");
  gtk_text_buffer_end_user_action(XEN_TO_C_GtkTextBuffer_(buffer));
  return(XEN_FALSE);
}

static XEN gxg_gtk_text_child_anchor_get_type(void)
{
  #define H_gtk_text_child_anchor_get_type "GType gtk_text_child_anchor_get_type( void)"
  return(C_TO_XEN_GType(gtk_text_child_anchor_get_type()));
}

static XEN gxg_gtk_text_child_anchor_new(void)
{
  #define H_gtk_text_child_anchor_new "GtkTextChildAnchor* gtk_text_child_anchor_new( void)"
  return(C_TO_XEN_GtkTextChildAnchor_(gtk_text_child_anchor_new()));
}

static XEN gxg_gtk_text_child_anchor_get_widgets(XEN anchor)
{
  #define H_gtk_text_child_anchor_get_widgets "GList* gtk_text_child_anchor_get_widgets(GtkTextChildAnchor* anchor)"
  XEN_ASSERT_TYPE(XEN_GtkTextChildAnchor__P(anchor), anchor, 1, "gtk_text_child_anchor_get_widgets", "GtkTextChildAnchor*");
  return(C_TO_XEN_GList_(gtk_text_child_anchor_get_widgets(XEN_TO_C_GtkTextChildAnchor_(anchor))));
}

static XEN gxg_gtk_text_child_anchor_get_deleted(XEN anchor)
{
  #define H_gtk_text_child_anchor_get_deleted "gboolean gtk_text_child_anchor_get_deleted(GtkTextChildAnchor* anchor)"
  XEN_ASSERT_TYPE(XEN_GtkTextChildAnchor__P(anchor), anchor, 1, "gtk_text_child_anchor_get_deleted", "GtkTextChildAnchor*");
  return(C_TO_XEN_gboolean(gtk_text_child_anchor_get_deleted(XEN_TO_C_GtkTextChildAnchor_(anchor))));
}

static XEN gxg_gtk_text_iter_get_buffer(XEN iter)
{
  #define H_gtk_text_iter_get_buffer "GtkTextBuffer* gtk_text_iter_get_buffer(GtkTextIter* iter)"
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(iter), iter, 1, "gtk_text_iter_get_buffer", "GtkTextIter*");
  return(C_TO_XEN_GtkTextBuffer_(gtk_text_iter_get_buffer(XEN_TO_C_GtkTextIter_(iter))));
}

static XEN gxg_gtk_text_iter_copy(XEN iter)
{
  #define H_gtk_text_iter_copy "GtkTextIter* gtk_text_iter_copy(GtkTextIter* iter)"
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(iter), iter, 1, "gtk_text_iter_copy", "GtkTextIter*");
  return(C_TO_XEN_GtkTextIter_(gtk_text_iter_copy(XEN_TO_C_GtkTextIter_(iter))));
}

static XEN gxg_gtk_text_iter_free(XEN iter)
{
  #define H_gtk_text_iter_free "void gtk_text_iter_free(GtkTextIter* iter)"
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(iter), iter, 1, "gtk_text_iter_free", "GtkTextIter*");
  gtk_text_iter_free(XEN_TO_C_GtkTextIter_(iter));
  return(XEN_FALSE);
}

static XEN gxg_gtk_text_iter_get_offset(XEN iter)
{
  #define H_gtk_text_iter_get_offset "gint gtk_text_iter_get_offset(GtkTextIter* iter)"
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(iter), iter, 1, "gtk_text_iter_get_offset", "GtkTextIter*");
  return(C_TO_XEN_gint(gtk_text_iter_get_offset(XEN_TO_C_GtkTextIter_(iter))));
}

static XEN gxg_gtk_text_iter_get_line(XEN iter)
{
  #define H_gtk_text_iter_get_line "gint gtk_text_iter_get_line(GtkTextIter* iter)"
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(iter), iter, 1, "gtk_text_iter_get_line", "GtkTextIter*");
  return(C_TO_XEN_gint(gtk_text_iter_get_line(XEN_TO_C_GtkTextIter_(iter))));
}

static XEN gxg_gtk_text_iter_get_line_offset(XEN iter)
{
  #define H_gtk_text_iter_get_line_offset "gint gtk_text_iter_get_line_offset(GtkTextIter* iter)"
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(iter), iter, 1, "gtk_text_iter_get_line_offset", "GtkTextIter*");
  return(C_TO_XEN_gint(gtk_text_iter_get_line_offset(XEN_TO_C_GtkTextIter_(iter))));
}

static XEN gxg_gtk_text_iter_get_line_index(XEN iter)
{
  #define H_gtk_text_iter_get_line_index "gint gtk_text_iter_get_line_index(GtkTextIter* iter)"
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(iter), iter, 1, "gtk_text_iter_get_line_index", "GtkTextIter*");
  return(C_TO_XEN_gint(gtk_text_iter_get_line_index(XEN_TO_C_GtkTextIter_(iter))));
}

static XEN gxg_gtk_text_iter_get_visible_line_offset(XEN iter)
{
  #define H_gtk_text_iter_get_visible_line_offset "gint gtk_text_iter_get_visible_line_offset(GtkTextIter* iter)"
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(iter), iter, 1, "gtk_text_iter_get_visible_line_offset", "GtkTextIter*");
  return(C_TO_XEN_gint(gtk_text_iter_get_visible_line_offset(XEN_TO_C_GtkTextIter_(iter))));
}

static XEN gxg_gtk_text_iter_get_visible_line_index(XEN iter)
{
  #define H_gtk_text_iter_get_visible_line_index "gint gtk_text_iter_get_visible_line_index(GtkTextIter* iter)"
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(iter), iter, 1, "gtk_text_iter_get_visible_line_index", "GtkTextIter*");
  return(C_TO_XEN_gint(gtk_text_iter_get_visible_line_index(XEN_TO_C_GtkTextIter_(iter))));
}

static XEN gxg_gtk_text_iter_get_char(XEN iter)
{
  #define H_gtk_text_iter_get_char "gunichar gtk_text_iter_get_char(GtkTextIter* iter)"
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(iter), iter, 1, "gtk_text_iter_get_char", "GtkTextIter*");
  return(C_TO_XEN_gunichar(gtk_text_iter_get_char(XEN_TO_C_GtkTextIter_(iter))));
}

static XEN gxg_gtk_text_iter_get_slice(XEN start, XEN end)
{
  #define H_gtk_text_iter_get_slice "gchar* gtk_text_iter_get_slice(GtkTextIter* start, GtkTextIter* end)"
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(start), start, 1, "gtk_text_iter_get_slice", "GtkTextIter*");
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(end), end, 2, "gtk_text_iter_get_slice", "GtkTextIter*");
  return(C_TO_XEN_gchar_(gtk_text_iter_get_slice(XEN_TO_C_GtkTextIter_(start), XEN_TO_C_GtkTextIter_(end))));
}

static XEN gxg_gtk_text_iter_get_text(XEN start, XEN end)
{
  #define H_gtk_text_iter_get_text "gchar* gtk_text_iter_get_text(GtkTextIter* start, GtkTextIter* end)"
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(start), start, 1, "gtk_text_iter_get_text", "GtkTextIter*");
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(end), end, 2, "gtk_text_iter_get_text", "GtkTextIter*");
  return(C_TO_XEN_gchar_(gtk_text_iter_get_text(XEN_TO_C_GtkTextIter_(start), XEN_TO_C_GtkTextIter_(end))));
}

static XEN gxg_gtk_text_iter_get_visible_slice(XEN start, XEN end)
{
  #define H_gtk_text_iter_get_visible_slice "gchar* gtk_text_iter_get_visible_slice(GtkTextIter* start, \
GtkTextIter* end)"
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(start), start, 1, "gtk_text_iter_get_visible_slice", "GtkTextIter*");
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(end), end, 2, "gtk_text_iter_get_visible_slice", "GtkTextIter*");
  return(C_TO_XEN_gchar_(gtk_text_iter_get_visible_slice(XEN_TO_C_GtkTextIter_(start), XEN_TO_C_GtkTextIter_(end))));
}

static XEN gxg_gtk_text_iter_get_visible_text(XEN start, XEN end)
{
  #define H_gtk_text_iter_get_visible_text "gchar* gtk_text_iter_get_visible_text(GtkTextIter* start, \
GtkTextIter* end)"
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(start), start, 1, "gtk_text_iter_get_visible_text", "GtkTextIter*");
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(end), end, 2, "gtk_text_iter_get_visible_text", "GtkTextIter*");
  return(C_TO_XEN_gchar_(gtk_text_iter_get_visible_text(XEN_TO_C_GtkTextIter_(start), XEN_TO_C_GtkTextIter_(end))));
}

static XEN gxg_gtk_text_iter_get_pixbuf(XEN iter)
{
  #define H_gtk_text_iter_get_pixbuf "GdkPixbuf* gtk_text_iter_get_pixbuf(GtkTextIter* iter)"
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(iter), iter, 1, "gtk_text_iter_get_pixbuf", "GtkTextIter*");
  return(C_TO_XEN_GdkPixbuf_(gtk_text_iter_get_pixbuf(XEN_TO_C_GtkTextIter_(iter))));
}

static XEN gxg_gtk_text_iter_get_marks(XEN iter)
{
  #define H_gtk_text_iter_get_marks "GSList* gtk_text_iter_get_marks(GtkTextIter* iter)"
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(iter), iter, 1, "gtk_text_iter_get_marks", "GtkTextIter*");
  return(C_TO_XEN_GSList_(gtk_text_iter_get_marks(XEN_TO_C_GtkTextIter_(iter))));
}

static XEN gxg_gtk_text_iter_get_child_anchor(XEN iter)
{
  #define H_gtk_text_iter_get_child_anchor "GtkTextChildAnchor* gtk_text_iter_get_child_anchor(GtkTextIter* iter)"
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(iter), iter, 1, "gtk_text_iter_get_child_anchor", "GtkTextIter*");
  return(C_TO_XEN_GtkTextChildAnchor_(gtk_text_iter_get_child_anchor(XEN_TO_C_GtkTextIter_(iter))));
}

static XEN gxg_gtk_text_iter_get_toggled_tags(XEN iter, XEN toggled_on)
{
  #define H_gtk_text_iter_get_toggled_tags "GSList* gtk_text_iter_get_toggled_tags(GtkTextIter* iter, \
gboolean toggled_on)"
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(iter), iter, 1, "gtk_text_iter_get_toggled_tags", "GtkTextIter*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(toggled_on), toggled_on, 2, "gtk_text_iter_get_toggled_tags", "gboolean");
  return(C_TO_XEN_GSList_(gtk_text_iter_get_toggled_tags(XEN_TO_C_GtkTextIter_(iter), XEN_TO_C_gboolean(toggled_on))));
}

static XEN gxg_gtk_text_iter_begins_tag(XEN iter, XEN tag)
{
  #define H_gtk_text_iter_begins_tag "gboolean gtk_text_iter_begins_tag(GtkTextIter* iter, GtkTextTag* tag)"
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(iter), iter, 1, "gtk_text_iter_begins_tag", "GtkTextIter*");
  XEN_ASSERT_TYPE(XEN_GtkTextTag__P(tag) || XEN_FALSE_P(tag), tag, 2, "gtk_text_iter_begins_tag", "GtkTextTag*");
  return(C_TO_XEN_gboolean(gtk_text_iter_begins_tag(XEN_TO_C_GtkTextIter_(iter), XEN_TO_C_GtkTextTag_(tag))));
}

static XEN gxg_gtk_text_iter_ends_tag(XEN iter, XEN tag)
{
  #define H_gtk_text_iter_ends_tag "gboolean gtk_text_iter_ends_tag(GtkTextIter* iter, GtkTextTag* tag)"
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(iter), iter, 1, "gtk_text_iter_ends_tag", "GtkTextIter*");
  XEN_ASSERT_TYPE(XEN_GtkTextTag__P(tag) || XEN_FALSE_P(tag), tag, 2, "gtk_text_iter_ends_tag", "GtkTextTag*");
  return(C_TO_XEN_gboolean(gtk_text_iter_ends_tag(XEN_TO_C_GtkTextIter_(iter), XEN_TO_C_GtkTextTag_(tag))));
}

static XEN gxg_gtk_text_iter_toggles_tag(XEN iter, XEN tag)
{
  #define H_gtk_text_iter_toggles_tag "gboolean gtk_text_iter_toggles_tag(GtkTextIter* iter, GtkTextTag* tag)"
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(iter), iter, 1, "gtk_text_iter_toggles_tag", "GtkTextIter*");
  XEN_ASSERT_TYPE(XEN_GtkTextTag__P(tag), tag, 2, "gtk_text_iter_toggles_tag", "GtkTextTag*");
  return(C_TO_XEN_gboolean(gtk_text_iter_toggles_tag(XEN_TO_C_GtkTextIter_(iter), XEN_TO_C_GtkTextTag_(tag))));
}

static XEN gxg_gtk_text_iter_has_tag(XEN iter, XEN tag)
{
  #define H_gtk_text_iter_has_tag "gboolean gtk_text_iter_has_tag(GtkTextIter* iter, GtkTextTag* tag)"
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(iter), iter, 1, "gtk_text_iter_has_tag", "GtkTextIter*");
  XEN_ASSERT_TYPE(XEN_GtkTextTag__P(tag), tag, 2, "gtk_text_iter_has_tag", "GtkTextTag*");
  return(C_TO_XEN_gboolean(gtk_text_iter_has_tag(XEN_TO_C_GtkTextIter_(iter), XEN_TO_C_GtkTextTag_(tag))));
}

static XEN gxg_gtk_text_iter_get_tags(XEN iter)
{
  #define H_gtk_text_iter_get_tags "GSList* gtk_text_iter_get_tags(GtkTextIter* iter)"
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(iter), iter, 1, "gtk_text_iter_get_tags", "GtkTextIter*");
  return(C_TO_XEN_GSList_(gtk_text_iter_get_tags(XEN_TO_C_GtkTextIter_(iter))));
}

static XEN gxg_gtk_text_iter_editable(XEN iter, XEN default_setting)
{
  #define H_gtk_text_iter_editable "gboolean gtk_text_iter_editable(GtkTextIter* iter, gboolean default_setting)"
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(iter), iter, 1, "gtk_text_iter_editable", "GtkTextIter*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(default_setting), default_setting, 2, "gtk_text_iter_editable", "gboolean");
  return(C_TO_XEN_gboolean(gtk_text_iter_editable(XEN_TO_C_GtkTextIter_(iter), XEN_TO_C_gboolean(default_setting))));
}

static XEN gxg_gtk_text_iter_can_insert(XEN iter, XEN default_editability)
{
  #define H_gtk_text_iter_can_insert "gboolean gtk_text_iter_can_insert(GtkTextIter* iter, gboolean default_editability)"
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(iter), iter, 1, "gtk_text_iter_can_insert", "GtkTextIter*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(default_editability), default_editability, 2, "gtk_text_iter_can_insert", "gboolean");
  return(C_TO_XEN_gboolean(gtk_text_iter_can_insert(XEN_TO_C_GtkTextIter_(iter), XEN_TO_C_gboolean(default_editability))));
}

static XEN gxg_gtk_text_iter_starts_word(XEN iter)
{
  #define H_gtk_text_iter_starts_word "gboolean gtk_text_iter_starts_word(GtkTextIter* iter)"
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(iter), iter, 1, "gtk_text_iter_starts_word", "GtkTextIter*");
  return(C_TO_XEN_gboolean(gtk_text_iter_starts_word(XEN_TO_C_GtkTextIter_(iter))));
}

static XEN gxg_gtk_text_iter_ends_word(XEN iter)
{
  #define H_gtk_text_iter_ends_word "gboolean gtk_text_iter_ends_word(GtkTextIter* iter)"
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(iter), iter, 1, "gtk_text_iter_ends_word", "GtkTextIter*");
  return(C_TO_XEN_gboolean(gtk_text_iter_ends_word(XEN_TO_C_GtkTextIter_(iter))));
}

static XEN gxg_gtk_text_iter_inside_word(XEN iter)
{
  #define H_gtk_text_iter_inside_word "gboolean gtk_text_iter_inside_word(GtkTextIter* iter)"
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(iter), iter, 1, "gtk_text_iter_inside_word", "GtkTextIter*");
  return(C_TO_XEN_gboolean(gtk_text_iter_inside_word(XEN_TO_C_GtkTextIter_(iter))));
}

static XEN gxg_gtk_text_iter_starts_sentence(XEN iter)
{
  #define H_gtk_text_iter_starts_sentence "gboolean gtk_text_iter_starts_sentence(GtkTextIter* iter)"
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(iter), iter, 1, "gtk_text_iter_starts_sentence", "GtkTextIter*");
  return(C_TO_XEN_gboolean(gtk_text_iter_starts_sentence(XEN_TO_C_GtkTextIter_(iter))));
}

static XEN gxg_gtk_text_iter_ends_sentence(XEN iter)
{
  #define H_gtk_text_iter_ends_sentence "gboolean gtk_text_iter_ends_sentence(GtkTextIter* iter)"
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(iter), iter, 1, "gtk_text_iter_ends_sentence", "GtkTextIter*");
  return(C_TO_XEN_gboolean(gtk_text_iter_ends_sentence(XEN_TO_C_GtkTextIter_(iter))));
}

static XEN gxg_gtk_text_iter_inside_sentence(XEN iter)
{
  #define H_gtk_text_iter_inside_sentence "gboolean gtk_text_iter_inside_sentence(GtkTextIter* iter)"
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(iter), iter, 1, "gtk_text_iter_inside_sentence", "GtkTextIter*");
  return(C_TO_XEN_gboolean(gtk_text_iter_inside_sentence(XEN_TO_C_GtkTextIter_(iter))));
}

static XEN gxg_gtk_text_iter_starts_line(XEN iter)
{
  #define H_gtk_text_iter_starts_line "gboolean gtk_text_iter_starts_line(GtkTextIter* iter)"
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(iter), iter, 1, "gtk_text_iter_starts_line", "GtkTextIter*");
  return(C_TO_XEN_gboolean(gtk_text_iter_starts_line(XEN_TO_C_GtkTextIter_(iter))));
}

static XEN gxg_gtk_text_iter_ends_line(XEN iter)
{
  #define H_gtk_text_iter_ends_line "gboolean gtk_text_iter_ends_line(GtkTextIter* iter)"
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(iter), iter, 1, "gtk_text_iter_ends_line", "GtkTextIter*");
  return(C_TO_XEN_gboolean(gtk_text_iter_ends_line(XEN_TO_C_GtkTextIter_(iter))));
}

static XEN gxg_gtk_text_iter_is_cursor_position(XEN iter)
{
  #define H_gtk_text_iter_is_cursor_position "gboolean gtk_text_iter_is_cursor_position(GtkTextIter* iter)"
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(iter), iter, 1, "gtk_text_iter_is_cursor_position", "GtkTextIter*");
  return(C_TO_XEN_gboolean(gtk_text_iter_is_cursor_position(XEN_TO_C_GtkTextIter_(iter))));
}

static XEN gxg_gtk_text_iter_get_chars_in_line(XEN iter)
{
  #define H_gtk_text_iter_get_chars_in_line "gint gtk_text_iter_get_chars_in_line(GtkTextIter* iter)"
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(iter), iter, 1, "gtk_text_iter_get_chars_in_line", "GtkTextIter*");
  return(C_TO_XEN_gint(gtk_text_iter_get_chars_in_line(XEN_TO_C_GtkTextIter_(iter))));
}

static XEN gxg_gtk_text_iter_get_bytes_in_line(XEN iter)
{
  #define H_gtk_text_iter_get_bytes_in_line "gint gtk_text_iter_get_bytes_in_line(GtkTextIter* iter)"
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(iter), iter, 1, "gtk_text_iter_get_bytes_in_line", "GtkTextIter*");
  return(C_TO_XEN_gint(gtk_text_iter_get_bytes_in_line(XEN_TO_C_GtkTextIter_(iter))));
}

static XEN gxg_gtk_text_iter_get_attributes(XEN iter, XEN values)
{
  #define H_gtk_text_iter_get_attributes "gboolean gtk_text_iter_get_attributes(GtkTextIter* iter, GtkTextAttributes* values)"
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(iter), iter, 1, "gtk_text_iter_get_attributes", "GtkTextIter*");
  XEN_ASSERT_TYPE(XEN_GtkTextAttributes__P(values), values, 2, "gtk_text_iter_get_attributes", "GtkTextAttributes*");
  return(C_TO_XEN_gboolean(gtk_text_iter_get_attributes(XEN_TO_C_GtkTextIter_(iter), XEN_TO_C_GtkTextAttributes_(values))));
}

static XEN gxg_gtk_text_iter_get_language(XEN iter)
{
  #define H_gtk_text_iter_get_language "PangoLanguage* gtk_text_iter_get_language(GtkTextIter* iter)"
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(iter), iter, 1, "gtk_text_iter_get_language", "GtkTextIter*");
  return(C_TO_XEN_PangoLanguage_(gtk_text_iter_get_language(XEN_TO_C_GtkTextIter_(iter))));
}

static XEN gxg_gtk_text_iter_is_end(XEN iter)
{
  #define H_gtk_text_iter_is_end "gboolean gtk_text_iter_is_end(GtkTextIter* iter)"
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(iter), iter, 1, "gtk_text_iter_is_end", "GtkTextIter*");
  return(C_TO_XEN_gboolean(gtk_text_iter_is_end(XEN_TO_C_GtkTextIter_(iter))));
}

static XEN gxg_gtk_text_iter_is_start(XEN iter)
{
  #define H_gtk_text_iter_is_start "gboolean gtk_text_iter_is_start(GtkTextIter* iter)"
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(iter), iter, 1, "gtk_text_iter_is_start", "GtkTextIter*");
  return(C_TO_XEN_gboolean(gtk_text_iter_is_start(XEN_TO_C_GtkTextIter_(iter))));
}

static XEN gxg_gtk_text_iter_forward_char(XEN iter)
{
  #define H_gtk_text_iter_forward_char "gboolean gtk_text_iter_forward_char(GtkTextIter* iter)"
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(iter), iter, 1, "gtk_text_iter_forward_char", "GtkTextIter*");
  return(C_TO_XEN_gboolean(gtk_text_iter_forward_char(XEN_TO_C_GtkTextIter_(iter))));
}

static XEN gxg_gtk_text_iter_backward_char(XEN iter)
{
  #define H_gtk_text_iter_backward_char "gboolean gtk_text_iter_backward_char(GtkTextIter* iter)"
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(iter), iter, 1, "gtk_text_iter_backward_char", "GtkTextIter*");
  return(C_TO_XEN_gboolean(gtk_text_iter_backward_char(XEN_TO_C_GtkTextIter_(iter))));
}

static XEN gxg_gtk_text_iter_forward_chars(XEN iter, XEN count)
{
  #define H_gtk_text_iter_forward_chars "gboolean gtk_text_iter_forward_chars(GtkTextIter* iter, gint count)"
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(iter), iter, 1, "gtk_text_iter_forward_chars", "GtkTextIter*");
  XEN_ASSERT_TYPE(XEN_gint_P(count), count, 2, "gtk_text_iter_forward_chars", "gint");
  return(C_TO_XEN_gboolean(gtk_text_iter_forward_chars(XEN_TO_C_GtkTextIter_(iter), XEN_TO_C_gint(count))));
}

static XEN gxg_gtk_text_iter_backward_chars(XEN iter, XEN count)
{
  #define H_gtk_text_iter_backward_chars "gboolean gtk_text_iter_backward_chars(GtkTextIter* iter, gint count)"
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(iter), iter, 1, "gtk_text_iter_backward_chars", "GtkTextIter*");
  XEN_ASSERT_TYPE(XEN_gint_P(count), count, 2, "gtk_text_iter_backward_chars", "gint");
  return(C_TO_XEN_gboolean(gtk_text_iter_backward_chars(XEN_TO_C_GtkTextIter_(iter), XEN_TO_C_gint(count))));
}

static XEN gxg_gtk_text_iter_forward_line(XEN iter)
{
  #define H_gtk_text_iter_forward_line "gboolean gtk_text_iter_forward_line(GtkTextIter* iter)"
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(iter), iter, 1, "gtk_text_iter_forward_line", "GtkTextIter*");
  return(C_TO_XEN_gboolean(gtk_text_iter_forward_line(XEN_TO_C_GtkTextIter_(iter))));
}

static XEN gxg_gtk_text_iter_backward_line(XEN iter)
{
  #define H_gtk_text_iter_backward_line "gboolean gtk_text_iter_backward_line(GtkTextIter* iter)"
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(iter), iter, 1, "gtk_text_iter_backward_line", "GtkTextIter*");
  return(C_TO_XEN_gboolean(gtk_text_iter_backward_line(XEN_TO_C_GtkTextIter_(iter))));
}

static XEN gxg_gtk_text_iter_forward_lines(XEN iter, XEN count)
{
  #define H_gtk_text_iter_forward_lines "gboolean gtk_text_iter_forward_lines(GtkTextIter* iter, gint count)"
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(iter), iter, 1, "gtk_text_iter_forward_lines", "GtkTextIter*");
  XEN_ASSERT_TYPE(XEN_gint_P(count), count, 2, "gtk_text_iter_forward_lines", "gint");
  return(C_TO_XEN_gboolean(gtk_text_iter_forward_lines(XEN_TO_C_GtkTextIter_(iter), XEN_TO_C_gint(count))));
}

static XEN gxg_gtk_text_iter_backward_lines(XEN iter, XEN count)
{
  #define H_gtk_text_iter_backward_lines "gboolean gtk_text_iter_backward_lines(GtkTextIter* iter, gint count)"
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(iter), iter, 1, "gtk_text_iter_backward_lines", "GtkTextIter*");
  XEN_ASSERT_TYPE(XEN_gint_P(count), count, 2, "gtk_text_iter_backward_lines", "gint");
  return(C_TO_XEN_gboolean(gtk_text_iter_backward_lines(XEN_TO_C_GtkTextIter_(iter), XEN_TO_C_gint(count))));
}

static XEN gxg_gtk_text_iter_forward_word_end(XEN iter)
{
  #define H_gtk_text_iter_forward_word_end "gboolean gtk_text_iter_forward_word_end(GtkTextIter* iter)"
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(iter), iter, 1, "gtk_text_iter_forward_word_end", "GtkTextIter*");
  return(C_TO_XEN_gboolean(gtk_text_iter_forward_word_end(XEN_TO_C_GtkTextIter_(iter))));
}

static XEN gxg_gtk_text_iter_backward_word_start(XEN iter)
{
  #define H_gtk_text_iter_backward_word_start "gboolean gtk_text_iter_backward_word_start(GtkTextIter* iter)"
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(iter), iter, 1, "gtk_text_iter_backward_word_start", "GtkTextIter*");
  return(C_TO_XEN_gboolean(gtk_text_iter_backward_word_start(XEN_TO_C_GtkTextIter_(iter))));
}

static XEN gxg_gtk_text_iter_forward_word_ends(XEN iter, XEN count)
{
  #define H_gtk_text_iter_forward_word_ends "gboolean gtk_text_iter_forward_word_ends(GtkTextIter* iter, \
gint count)"
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(iter), iter, 1, "gtk_text_iter_forward_word_ends", "GtkTextIter*");
  XEN_ASSERT_TYPE(XEN_gint_P(count), count, 2, "gtk_text_iter_forward_word_ends", "gint");
  return(C_TO_XEN_gboolean(gtk_text_iter_forward_word_ends(XEN_TO_C_GtkTextIter_(iter), XEN_TO_C_gint(count))));
}

static XEN gxg_gtk_text_iter_backward_word_starts(XEN iter, XEN count)
{
  #define H_gtk_text_iter_backward_word_starts "gboolean gtk_text_iter_backward_word_starts(GtkTextIter* iter, \
gint count)"
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(iter), iter, 1, "gtk_text_iter_backward_word_starts", "GtkTextIter*");
  XEN_ASSERT_TYPE(XEN_gint_P(count), count, 2, "gtk_text_iter_backward_word_starts", "gint");
  return(C_TO_XEN_gboolean(gtk_text_iter_backward_word_starts(XEN_TO_C_GtkTextIter_(iter), XEN_TO_C_gint(count))));
}

static XEN gxg_gtk_text_iter_forward_sentence_end(XEN iter)
{
  #define H_gtk_text_iter_forward_sentence_end "gboolean gtk_text_iter_forward_sentence_end(GtkTextIter* iter)"
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(iter), iter, 1, "gtk_text_iter_forward_sentence_end", "GtkTextIter*");
  return(C_TO_XEN_gboolean(gtk_text_iter_forward_sentence_end(XEN_TO_C_GtkTextIter_(iter))));
}

static XEN gxg_gtk_text_iter_backward_sentence_start(XEN iter)
{
  #define H_gtk_text_iter_backward_sentence_start "gboolean gtk_text_iter_backward_sentence_start(GtkTextIter* iter)"
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(iter), iter, 1, "gtk_text_iter_backward_sentence_start", "GtkTextIter*");
  return(C_TO_XEN_gboolean(gtk_text_iter_backward_sentence_start(XEN_TO_C_GtkTextIter_(iter))));
}

static XEN gxg_gtk_text_iter_forward_sentence_ends(XEN iter, XEN count)
{
  #define H_gtk_text_iter_forward_sentence_ends "gboolean gtk_text_iter_forward_sentence_ends(GtkTextIter* iter, \
gint count)"
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(iter), iter, 1, "gtk_text_iter_forward_sentence_ends", "GtkTextIter*");
  XEN_ASSERT_TYPE(XEN_gint_P(count), count, 2, "gtk_text_iter_forward_sentence_ends", "gint");
  return(C_TO_XEN_gboolean(gtk_text_iter_forward_sentence_ends(XEN_TO_C_GtkTextIter_(iter), XEN_TO_C_gint(count))));
}

static XEN gxg_gtk_text_iter_backward_sentence_starts(XEN iter, XEN count)
{
  #define H_gtk_text_iter_backward_sentence_starts "gboolean gtk_text_iter_backward_sentence_starts(GtkTextIter* iter, \
gint count)"
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(iter), iter, 1, "gtk_text_iter_backward_sentence_starts", "GtkTextIter*");
  XEN_ASSERT_TYPE(XEN_gint_P(count), count, 2, "gtk_text_iter_backward_sentence_starts", "gint");
  return(C_TO_XEN_gboolean(gtk_text_iter_backward_sentence_starts(XEN_TO_C_GtkTextIter_(iter), XEN_TO_C_gint(count))));
}

static XEN gxg_gtk_text_iter_forward_cursor_position(XEN iter)
{
  #define H_gtk_text_iter_forward_cursor_position "gboolean gtk_text_iter_forward_cursor_position(GtkTextIter* iter)"
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(iter), iter, 1, "gtk_text_iter_forward_cursor_position", "GtkTextIter*");
  return(C_TO_XEN_gboolean(gtk_text_iter_forward_cursor_position(XEN_TO_C_GtkTextIter_(iter))));
}

static XEN gxg_gtk_text_iter_backward_cursor_position(XEN iter)
{
  #define H_gtk_text_iter_backward_cursor_position "gboolean gtk_text_iter_backward_cursor_position(GtkTextIter* iter)"
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(iter), iter, 1, "gtk_text_iter_backward_cursor_position", "GtkTextIter*");
  return(C_TO_XEN_gboolean(gtk_text_iter_backward_cursor_position(XEN_TO_C_GtkTextIter_(iter))));
}

static XEN gxg_gtk_text_iter_forward_cursor_positions(XEN iter, XEN count)
{
  #define H_gtk_text_iter_forward_cursor_positions "gboolean gtk_text_iter_forward_cursor_positions(GtkTextIter* iter, \
gint count)"
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(iter), iter, 1, "gtk_text_iter_forward_cursor_positions", "GtkTextIter*");
  XEN_ASSERT_TYPE(XEN_gint_P(count), count, 2, "gtk_text_iter_forward_cursor_positions", "gint");
  return(C_TO_XEN_gboolean(gtk_text_iter_forward_cursor_positions(XEN_TO_C_GtkTextIter_(iter), XEN_TO_C_gint(count))));
}

static XEN gxg_gtk_text_iter_backward_cursor_positions(XEN iter, XEN count)
{
  #define H_gtk_text_iter_backward_cursor_positions "gboolean gtk_text_iter_backward_cursor_positions(GtkTextIter* iter, \
gint count)"
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(iter), iter, 1, "gtk_text_iter_backward_cursor_positions", "GtkTextIter*");
  XEN_ASSERT_TYPE(XEN_gint_P(count), count, 2, "gtk_text_iter_backward_cursor_positions", "gint");
  return(C_TO_XEN_gboolean(gtk_text_iter_backward_cursor_positions(XEN_TO_C_GtkTextIter_(iter), XEN_TO_C_gint(count))));
}

static XEN gxg_gtk_text_iter_set_offset(XEN iter, XEN char_offset)
{
  #define H_gtk_text_iter_set_offset "void gtk_text_iter_set_offset(GtkTextIter* iter, gint char_offset)"
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(iter), iter, 1, "gtk_text_iter_set_offset", "GtkTextIter*");
  XEN_ASSERT_TYPE(XEN_gint_P(char_offset), char_offset, 2, "gtk_text_iter_set_offset", "gint");
  gtk_text_iter_set_offset(XEN_TO_C_GtkTextIter_(iter), XEN_TO_C_gint(char_offset));
  return(XEN_FALSE);
}

static XEN gxg_gtk_text_iter_set_line(XEN iter, XEN line_number)
{
  #define H_gtk_text_iter_set_line "void gtk_text_iter_set_line(GtkTextIter* iter, gint line_number)"
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(iter), iter, 1, "gtk_text_iter_set_line", "GtkTextIter*");
  XEN_ASSERT_TYPE(XEN_gint_P(line_number), line_number, 2, "gtk_text_iter_set_line", "gint");
  gtk_text_iter_set_line(XEN_TO_C_GtkTextIter_(iter), XEN_TO_C_gint(line_number));
  return(XEN_FALSE);
}

static XEN gxg_gtk_text_iter_set_line_offset(XEN iter, XEN char_on_line)
{
  #define H_gtk_text_iter_set_line_offset "void gtk_text_iter_set_line_offset(GtkTextIter* iter, gint char_on_line)"
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(iter), iter, 1, "gtk_text_iter_set_line_offset", "GtkTextIter*");
  XEN_ASSERT_TYPE(XEN_gint_P(char_on_line), char_on_line, 2, "gtk_text_iter_set_line_offset", "gint");
  gtk_text_iter_set_line_offset(XEN_TO_C_GtkTextIter_(iter), XEN_TO_C_gint(char_on_line));
  return(XEN_FALSE);
}

static XEN gxg_gtk_text_iter_set_line_index(XEN iter, XEN byte_on_line)
{
  #define H_gtk_text_iter_set_line_index "void gtk_text_iter_set_line_index(GtkTextIter* iter, gint byte_on_line)"
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(iter), iter, 1, "gtk_text_iter_set_line_index", "GtkTextIter*");
  XEN_ASSERT_TYPE(XEN_gint_P(byte_on_line), byte_on_line, 2, "gtk_text_iter_set_line_index", "gint");
  gtk_text_iter_set_line_index(XEN_TO_C_GtkTextIter_(iter), XEN_TO_C_gint(byte_on_line));
  return(XEN_FALSE);
}

static XEN gxg_gtk_text_iter_forward_to_end(XEN iter)
{
  #define H_gtk_text_iter_forward_to_end "void gtk_text_iter_forward_to_end(GtkTextIter* iter)"
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(iter), iter, 1, "gtk_text_iter_forward_to_end", "GtkTextIter*");
  gtk_text_iter_forward_to_end(XEN_TO_C_GtkTextIter_(iter));
  return(XEN_FALSE);
}

static XEN gxg_gtk_text_iter_forward_to_line_end(XEN iter)
{
  #define H_gtk_text_iter_forward_to_line_end "gboolean gtk_text_iter_forward_to_line_end(GtkTextIter* iter)"
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(iter), iter, 1, "gtk_text_iter_forward_to_line_end", "GtkTextIter*");
  return(C_TO_XEN_gboolean(gtk_text_iter_forward_to_line_end(XEN_TO_C_GtkTextIter_(iter))));
}

static XEN gxg_gtk_text_iter_set_visible_line_offset(XEN iter, XEN char_on_line)
{
  #define H_gtk_text_iter_set_visible_line_offset "void gtk_text_iter_set_visible_line_offset(GtkTextIter* iter, \
gint char_on_line)"
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(iter), iter, 1, "gtk_text_iter_set_visible_line_offset", "GtkTextIter*");
  XEN_ASSERT_TYPE(XEN_gint_P(char_on_line), char_on_line, 2, "gtk_text_iter_set_visible_line_offset", "gint");
  gtk_text_iter_set_visible_line_offset(XEN_TO_C_GtkTextIter_(iter), XEN_TO_C_gint(char_on_line));
  return(XEN_FALSE);
}

static XEN gxg_gtk_text_iter_set_visible_line_index(XEN iter, XEN byte_on_line)
{
  #define H_gtk_text_iter_set_visible_line_index "void gtk_text_iter_set_visible_line_index(GtkTextIter* iter, \
gint byte_on_line)"
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(iter), iter, 1, "gtk_text_iter_set_visible_line_index", "GtkTextIter*");
  XEN_ASSERT_TYPE(XEN_gint_P(byte_on_line), byte_on_line, 2, "gtk_text_iter_set_visible_line_index", "gint");
  gtk_text_iter_set_visible_line_index(XEN_TO_C_GtkTextIter_(iter), XEN_TO_C_gint(byte_on_line));
  return(XEN_FALSE);
}

static XEN gxg_gtk_text_iter_forward_to_tag_toggle(XEN iter, XEN tag)
{
  #define H_gtk_text_iter_forward_to_tag_toggle "gboolean gtk_text_iter_forward_to_tag_toggle(GtkTextIter* iter, \
GtkTextTag* tag)"
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(iter), iter, 1, "gtk_text_iter_forward_to_tag_toggle", "GtkTextIter*");
  XEN_ASSERT_TYPE(XEN_GtkTextTag__P(tag) || XEN_FALSE_P(tag), tag, 2, "gtk_text_iter_forward_to_tag_toggle", "GtkTextTag*");
  return(C_TO_XEN_gboolean(gtk_text_iter_forward_to_tag_toggle(XEN_TO_C_GtkTextIter_(iter), XEN_TO_C_GtkTextTag_(tag))));
}

static XEN gxg_gtk_text_iter_backward_to_tag_toggle(XEN iter, XEN tag)
{
  #define H_gtk_text_iter_backward_to_tag_toggle "gboolean gtk_text_iter_backward_to_tag_toggle(GtkTextIter* iter, \
GtkTextTag* tag)"
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(iter), iter, 1, "gtk_text_iter_backward_to_tag_toggle", "GtkTextIter*");
  XEN_ASSERT_TYPE(XEN_GtkTextTag__P(tag) || XEN_FALSE_P(tag), tag, 2, "gtk_text_iter_backward_to_tag_toggle", "GtkTextTag*");
  return(C_TO_XEN_gboolean(gtk_text_iter_backward_to_tag_toggle(XEN_TO_C_GtkTextIter_(iter), XEN_TO_C_GtkTextTag_(tag))));
}

static XEN gxg_gtk_text_iter_forward_find_char(XEN iter, XEN pred, XEN user_data, XEN limit)
{
  #define H_gtk_text_iter_forward_find_char "gboolean gtk_text_iter_forward_find_char(GtkTextIter* iter, \
GtkTextCharPredicate pred, gpointer user_data, GtkTextIter* limit)"
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(iter), iter, 1, "gtk_text_iter_forward_find_char", "GtkTextIter*");
  XEN_ASSERT_TYPE(XEN_GtkTextCharPredicate_P(pred), pred, 2, "gtk_text_iter_forward_find_char", "GtkTextCharPredicate");
  XEN_ASSERT_TYPE(XEN_gpointer_P(user_data), user_data, 3, "gtk_text_iter_forward_find_char", "gpointer");
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(limit) || XEN_FALSE_P(limit), limit, 4, "gtk_text_iter_forward_find_char", "GtkTextIter*");
  return(C_TO_XEN_gboolean(gtk_text_iter_forward_find_char(XEN_TO_C_GtkTextIter_(iter), XEN_TO_C_GtkTextCharPredicate(pred), 
                                                           XEN_TO_C_gpointer(user_data), XEN_TO_C_GtkTextIter_(limit))));
}

static XEN gxg_gtk_text_iter_backward_find_char(XEN iter, XEN pred, XEN user_data, XEN limit)
{
  #define H_gtk_text_iter_backward_find_char "gboolean gtk_text_iter_backward_find_char(GtkTextIter* iter, \
GtkTextCharPredicate pred, gpointer user_data, GtkTextIter* limit)"
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(iter), iter, 1, "gtk_text_iter_backward_find_char", "GtkTextIter*");
  XEN_ASSERT_TYPE(XEN_GtkTextCharPredicate_P(pred), pred, 2, "gtk_text_iter_backward_find_char", "GtkTextCharPredicate");
  XEN_ASSERT_TYPE(XEN_gpointer_P(user_data), user_data, 3, "gtk_text_iter_backward_find_char", "gpointer");
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(limit) || XEN_FALSE_P(limit), limit, 4, "gtk_text_iter_backward_find_char", "GtkTextIter*");
  return(C_TO_XEN_gboolean(gtk_text_iter_backward_find_char(XEN_TO_C_GtkTextIter_(iter), XEN_TO_C_GtkTextCharPredicate(pred), 
                                                            XEN_TO_C_gpointer(user_data), XEN_TO_C_GtkTextIter_(limit))));
}

static XEN gxg_gtk_text_iter_forward_search(XEN iter, XEN str, XEN flags, XEN match_start, XEN match_end, XEN limit)
{
  #define H_gtk_text_iter_forward_search "gboolean gtk_text_iter_forward_search(GtkTextIter* iter, gchar* str, \
GtkTextSearchFlags flags, GtkTextIter* match_start, GtkTextIter* match_end, GtkTextIter* limit)"
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(iter), iter, 1, "gtk_text_iter_forward_search", "GtkTextIter*");
  XEN_ASSERT_TYPE(XEN_gchar__P(str), str, 2, "gtk_text_iter_forward_search", "gchar*");
  XEN_ASSERT_TYPE(XEN_GtkTextSearchFlags_P(flags), flags, 3, "gtk_text_iter_forward_search", "GtkTextSearchFlags");
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(match_start) || XEN_FALSE_P(match_start), match_start, 4, "gtk_text_iter_forward_search", "GtkTextIter*");
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(match_end) || XEN_FALSE_P(match_end), match_end, 5, "gtk_text_iter_forward_search", "GtkTextIter*");
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(limit) || XEN_FALSE_P(limit), limit, 6, "gtk_text_iter_forward_search", "GtkTextIter*");
  return(C_TO_XEN_gboolean(gtk_text_iter_forward_search(XEN_TO_C_GtkTextIter_(iter), XEN_TO_C_gchar_(str), XEN_TO_C_GtkTextSearchFlags(flags), 
                                                        XEN_TO_C_GtkTextIter_(match_start), XEN_TO_C_GtkTextIter_(match_end), 
                                                        XEN_TO_C_GtkTextIter_(limit))));
}

static XEN gxg_gtk_text_iter_backward_search(XEN iter, XEN str, XEN flags, XEN match_start, XEN match_end, XEN limit)
{
  #define H_gtk_text_iter_backward_search "gboolean gtk_text_iter_backward_search(GtkTextIter* iter, \
gchar* str, GtkTextSearchFlags flags, GtkTextIter* match_start, GtkTextIter* match_end, GtkTextIter* limit)"
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(iter), iter, 1, "gtk_text_iter_backward_search", "GtkTextIter*");
  XEN_ASSERT_TYPE(XEN_gchar__P(str), str, 2, "gtk_text_iter_backward_search", "gchar*");
  XEN_ASSERT_TYPE(XEN_GtkTextSearchFlags_P(flags), flags, 3, "gtk_text_iter_backward_search", "GtkTextSearchFlags");
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(match_start) || XEN_FALSE_P(match_start), match_start, 4, "gtk_text_iter_backward_search", "GtkTextIter*");
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(match_end) || XEN_FALSE_P(match_end), match_end, 5, "gtk_text_iter_backward_search", "GtkTextIter*");
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(limit) || XEN_FALSE_P(limit), limit, 6, "gtk_text_iter_backward_search", "GtkTextIter*");
  return(C_TO_XEN_gboolean(gtk_text_iter_backward_search(XEN_TO_C_GtkTextIter_(iter), XEN_TO_C_gchar_(str), XEN_TO_C_GtkTextSearchFlags(flags), 
                                                         XEN_TO_C_GtkTextIter_(match_start), XEN_TO_C_GtkTextIter_(match_end), 
                                                         XEN_TO_C_GtkTextIter_(limit))));
}

static XEN gxg_gtk_text_iter_equal(XEN lhs, XEN rhs)
{
  #define H_gtk_text_iter_equal "gboolean gtk_text_iter_equal(GtkTextIter* lhs, GtkTextIter* rhs)"
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(lhs), lhs, 1, "gtk_text_iter_equal", "GtkTextIter*");
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(rhs), rhs, 2, "gtk_text_iter_equal", "GtkTextIter*");
  return(C_TO_XEN_gboolean(gtk_text_iter_equal(XEN_TO_C_GtkTextIter_(lhs), XEN_TO_C_GtkTextIter_(rhs))));
}

static XEN gxg_gtk_text_iter_compare(XEN lhs, XEN rhs)
{
  #define H_gtk_text_iter_compare "gint gtk_text_iter_compare(GtkTextIter* lhs, GtkTextIter* rhs)"
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(lhs), lhs, 1, "gtk_text_iter_compare", "GtkTextIter*");
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(rhs), rhs, 2, "gtk_text_iter_compare", "GtkTextIter*");
  return(C_TO_XEN_gint(gtk_text_iter_compare(XEN_TO_C_GtkTextIter_(lhs), XEN_TO_C_GtkTextIter_(rhs))));
}

static XEN gxg_gtk_text_iter_in_range(XEN iter, XEN start, XEN end)
{
  #define H_gtk_text_iter_in_range "gboolean gtk_text_iter_in_range(GtkTextIter* iter, GtkTextIter* start, \
GtkTextIter* end)"
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(iter), iter, 1, "gtk_text_iter_in_range", "GtkTextIter*");
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(start), start, 2, "gtk_text_iter_in_range", "GtkTextIter*");
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(end), end, 3, "gtk_text_iter_in_range", "GtkTextIter*");
  return(C_TO_XEN_gboolean(gtk_text_iter_in_range(XEN_TO_C_GtkTextIter_(iter), XEN_TO_C_GtkTextIter_(start), XEN_TO_C_GtkTextIter_(end))));
}

static XEN gxg_gtk_text_iter_order(XEN first, XEN second)
{
  #define H_gtk_text_iter_order "void gtk_text_iter_order(GtkTextIter* first, GtkTextIter* second)"
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(first), first, 1, "gtk_text_iter_order", "GtkTextIter*");
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(second), second, 2, "gtk_text_iter_order", "GtkTextIter*");
  gtk_text_iter_order(XEN_TO_C_GtkTextIter_(first), XEN_TO_C_GtkTextIter_(second));
  return(XEN_FALSE);
}

static XEN gxg_gtk_text_mark_get_type(void)
{
  #define H_gtk_text_mark_get_type "GType gtk_text_mark_get_type( void)"
  return(C_TO_XEN_GType(gtk_text_mark_get_type()));
}

static XEN gxg_gtk_text_mark_set_visible(XEN mark, XEN setting)
{
  #define H_gtk_text_mark_set_visible "void gtk_text_mark_set_visible(GtkTextMark* mark, gboolean setting)"
  XEN_ASSERT_TYPE(XEN_GtkTextMark__P(mark), mark, 1, "gtk_text_mark_set_visible", "GtkTextMark*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(setting), setting, 2, "gtk_text_mark_set_visible", "gboolean");
  gtk_text_mark_set_visible(XEN_TO_C_GtkTextMark_(mark), XEN_TO_C_gboolean(setting));
  return(XEN_FALSE);
}

static XEN gxg_gtk_text_mark_get_visible(XEN mark)
{
  #define H_gtk_text_mark_get_visible "gboolean gtk_text_mark_get_visible(GtkTextMark* mark)"
  XEN_ASSERT_TYPE(XEN_GtkTextMark__P(mark), mark, 1, "gtk_text_mark_get_visible", "GtkTextMark*");
  return(C_TO_XEN_gboolean(gtk_text_mark_get_visible(XEN_TO_C_GtkTextMark_(mark))));
}

static XEN gxg_gtk_text_mark_get_name(XEN mark)
{
  #define H_gtk_text_mark_get_name "char* gtk_text_mark_get_name(GtkTextMark* mark)"
  XEN_ASSERT_TYPE(XEN_GtkTextMark__P(mark), mark, 1, "gtk_text_mark_get_name", "GtkTextMark*");
  return(C_TO_XEN_char_(gtk_text_mark_get_name(XEN_TO_C_GtkTextMark_(mark))));
}

static XEN gxg_gtk_text_mark_get_deleted(XEN mark)
{
  #define H_gtk_text_mark_get_deleted "gboolean gtk_text_mark_get_deleted(GtkTextMark* mark)"
  XEN_ASSERT_TYPE(XEN_GtkTextMark__P(mark), mark, 1, "gtk_text_mark_get_deleted", "GtkTextMark*");
  return(C_TO_XEN_gboolean(gtk_text_mark_get_deleted(XEN_TO_C_GtkTextMark_(mark))));
}

static XEN gxg_gtk_text_mark_get_buffer(XEN mark)
{
  #define H_gtk_text_mark_get_buffer "GtkTextBuffer* gtk_text_mark_get_buffer(GtkTextMark* mark)"
  XEN_ASSERT_TYPE(XEN_GtkTextMark__P(mark), mark, 1, "gtk_text_mark_get_buffer", "GtkTextMark*");
  return(C_TO_XEN_GtkTextBuffer_(gtk_text_mark_get_buffer(XEN_TO_C_GtkTextMark_(mark))));
}

static XEN gxg_gtk_text_mark_get_left_gravity(XEN mark)
{
  #define H_gtk_text_mark_get_left_gravity "gboolean gtk_text_mark_get_left_gravity(GtkTextMark* mark)"
  XEN_ASSERT_TYPE(XEN_GtkTextMark__P(mark), mark, 1, "gtk_text_mark_get_left_gravity", "GtkTextMark*");
  return(C_TO_XEN_gboolean(gtk_text_mark_get_left_gravity(XEN_TO_C_GtkTextMark_(mark))));
}

static XEN gxg_gtk_text_tag_get_type(void)
{
  #define H_gtk_text_tag_get_type "GType gtk_text_tag_get_type( void)"
  return(C_TO_XEN_GType(gtk_text_tag_get_type()));
}

static XEN gxg_gtk_text_tag_new(XEN name)
{
  #define H_gtk_text_tag_new "GtkTextTag* gtk_text_tag_new(gchar* name)"
  XEN_ASSERT_TYPE(XEN_gchar__P(name), name, 1, "gtk_text_tag_new", "gchar*");
  return(C_TO_XEN_GtkTextTag_(gtk_text_tag_new(XEN_TO_C_gchar_(name))));
}

static XEN gxg_gtk_text_tag_get_priority(XEN tag)
{
  #define H_gtk_text_tag_get_priority "gint gtk_text_tag_get_priority(GtkTextTag* tag)"
  XEN_ASSERT_TYPE(XEN_GtkTextTag__P(tag), tag, 1, "gtk_text_tag_get_priority", "GtkTextTag*");
  return(C_TO_XEN_gint(gtk_text_tag_get_priority(XEN_TO_C_GtkTextTag_(tag))));
}

static XEN gxg_gtk_text_tag_set_priority(XEN tag, XEN priority)
{
  #define H_gtk_text_tag_set_priority "void gtk_text_tag_set_priority(GtkTextTag* tag, gint priority)"
  XEN_ASSERT_TYPE(XEN_GtkTextTag__P(tag), tag, 1, "gtk_text_tag_set_priority", "GtkTextTag*");
  XEN_ASSERT_TYPE(XEN_gint_P(priority), priority, 2, "gtk_text_tag_set_priority", "gint");
  gtk_text_tag_set_priority(XEN_TO_C_GtkTextTag_(tag), XEN_TO_C_gint(priority));
  return(XEN_FALSE);
}

static XEN gxg_gtk_text_tag_event(XEN tag, XEN event_object, XEN event, XEN iter)
{
  #define H_gtk_text_tag_event "gboolean gtk_text_tag_event(GtkTextTag* tag, GObject* event_object, GdkEvent* event, \
GtkTextIter* iter)"
  XEN_ASSERT_TYPE(XEN_GtkTextTag__P(tag), tag, 1, "gtk_text_tag_event", "GtkTextTag*");
  XEN_ASSERT_TYPE(XEN_GObject__P(event_object), event_object, 2, "gtk_text_tag_event", "GObject*");
  XEN_ASSERT_TYPE(XEN_GdkEvent__P(event), event, 3, "gtk_text_tag_event", "GdkEvent*");
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(iter), iter, 4, "gtk_text_tag_event", "GtkTextIter*");
  return(C_TO_XEN_gboolean(gtk_text_tag_event(XEN_TO_C_GtkTextTag_(tag), XEN_TO_C_GObject_(event_object), XEN_TO_C_GdkEvent_(event), 
                                              XEN_TO_C_GtkTextIter_(iter))));
}

static XEN gxg_gtk_text_attributes_new(void)
{
  #define H_gtk_text_attributes_new "GtkTextAttributes* gtk_text_attributes_new( void)"
  return(C_TO_XEN_GtkTextAttributes_(gtk_text_attributes_new()));
}

static XEN gxg_gtk_text_attributes_copy(XEN src)
{
  #define H_gtk_text_attributes_copy "GtkTextAttributes* gtk_text_attributes_copy(GtkTextAttributes* src)"
  XEN_ASSERT_TYPE(XEN_GtkTextAttributes__P(src), src, 1, "gtk_text_attributes_copy", "GtkTextAttributes*");
  return(C_TO_XEN_GtkTextAttributes_(gtk_text_attributes_copy(XEN_TO_C_GtkTextAttributes_(src))));
}

static XEN gxg_gtk_text_attributes_copy_values(XEN src, XEN dest)
{
  #define H_gtk_text_attributes_copy_values "void gtk_text_attributes_copy_values(GtkTextAttributes* src, \
GtkTextAttributes* dest)"
  XEN_ASSERT_TYPE(XEN_GtkTextAttributes__P(src), src, 1, "gtk_text_attributes_copy_values", "GtkTextAttributes*");
  XEN_ASSERT_TYPE(XEN_GtkTextAttributes__P(dest), dest, 2, "gtk_text_attributes_copy_values", "GtkTextAttributes*");
  gtk_text_attributes_copy_values(XEN_TO_C_GtkTextAttributes_(src), XEN_TO_C_GtkTextAttributes_(dest));
  return(XEN_FALSE);
}

static XEN gxg_gtk_text_attributes_unref(XEN values)
{
  #define H_gtk_text_attributes_unref "void gtk_text_attributes_unref(GtkTextAttributes* values)"
  XEN_ASSERT_TYPE(XEN_GtkTextAttributes__P(values), values, 1, "gtk_text_attributes_unref", "GtkTextAttributes*");
  gtk_text_attributes_unref(XEN_TO_C_GtkTextAttributes_(values));
  return(XEN_FALSE);
}

static XEN gxg_gtk_text_attributes_ref(XEN values)
{
  #define H_gtk_text_attributes_ref "void gtk_text_attributes_ref(GtkTextAttributes* values)"
  XEN_ASSERT_TYPE(XEN_GtkTextAttributes__P(values), values, 1, "gtk_text_attributes_ref", "GtkTextAttributes*");
  gtk_text_attributes_ref(XEN_TO_C_GtkTextAttributes_(values));
  return(XEN_FALSE);
}

static XEN gxg_gtk_text_tag_table_get_type(void)
{
  #define H_gtk_text_tag_table_get_type "GType gtk_text_tag_table_get_type( void)"
  return(C_TO_XEN_GType(gtk_text_tag_table_get_type()));
}

static XEN gxg_gtk_text_tag_table_new(void)
{
  #define H_gtk_text_tag_table_new "GtkTextTagTable* gtk_text_tag_table_new( void)"
  return(C_TO_XEN_GtkTextTagTable_(gtk_text_tag_table_new()));
}

static XEN gxg_gtk_text_tag_table_add(XEN table, XEN tag)
{
  #define H_gtk_text_tag_table_add "void gtk_text_tag_table_add(GtkTextTagTable* table, GtkTextTag* tag)"
  XEN_ASSERT_TYPE(XEN_GtkTextTagTable__P(table), table, 1, "gtk_text_tag_table_add", "GtkTextTagTable*");
  XEN_ASSERT_TYPE(XEN_GtkTextTag__P(tag), tag, 2, "gtk_text_tag_table_add", "GtkTextTag*");
  gtk_text_tag_table_add(XEN_TO_C_GtkTextTagTable_(table), XEN_TO_C_GtkTextTag_(tag));
  return(XEN_FALSE);
}

static XEN gxg_gtk_text_tag_table_remove(XEN table, XEN tag)
{
  #define H_gtk_text_tag_table_remove "void gtk_text_tag_table_remove(GtkTextTagTable* table, GtkTextTag* tag)"
  XEN_ASSERT_TYPE(XEN_GtkTextTagTable__P(table), table, 1, "gtk_text_tag_table_remove", "GtkTextTagTable*");
  XEN_ASSERT_TYPE(XEN_GtkTextTag__P(tag), tag, 2, "gtk_text_tag_table_remove", "GtkTextTag*");
  gtk_text_tag_table_remove(XEN_TO_C_GtkTextTagTable_(table), XEN_TO_C_GtkTextTag_(tag));
  return(XEN_FALSE);
}

static XEN gxg_gtk_text_tag_table_lookup(XEN table, XEN name)
{
  #define H_gtk_text_tag_table_lookup "GtkTextTag* gtk_text_tag_table_lookup(GtkTextTagTable* table, \
gchar* name)"
  XEN_ASSERT_TYPE(XEN_GtkTextTagTable__P(table), table, 1, "gtk_text_tag_table_lookup", "GtkTextTagTable*");
  XEN_ASSERT_TYPE(XEN_gchar__P(name), name, 2, "gtk_text_tag_table_lookup", "gchar*");
  return(C_TO_XEN_GtkTextTag_(gtk_text_tag_table_lookup(XEN_TO_C_GtkTextTagTable_(table), XEN_TO_C_gchar_(name))));
}

static XEN gxg_gtk_text_tag_table_foreach(XEN table, XEN func, XEN data)
{
  #define H_gtk_text_tag_table_foreach "void gtk_text_tag_table_foreach(GtkTextTagTable* table, GtkTextTagTableForeach func, \
gpointer data)"
  XEN_ASSERT_TYPE(XEN_GtkTextTagTable__P(table), table, 1, "gtk_text_tag_table_foreach", "GtkTextTagTable*");
  XEN_ASSERT_TYPE(XEN_GtkTextTagTableForeach_P(func), func, 2, "gtk_text_tag_table_foreach", "GtkTextTagTableForeach");
  XEN_ASSERT_TYPE(XEN_gpointer_P(data), data, 3, "gtk_text_tag_table_foreach", "gpointer");
  gtk_text_tag_table_foreach(XEN_TO_C_GtkTextTagTable_(table), XEN_TO_C_GtkTextTagTableForeach(func), XEN_TO_C_gpointer(data));
  return(XEN_FALSE);
}

static XEN gxg_gtk_text_tag_table_get_size(XEN table)
{
  #define H_gtk_text_tag_table_get_size "gint gtk_text_tag_table_get_size(GtkTextTagTable* table)"
  XEN_ASSERT_TYPE(XEN_GtkTextTagTable__P(table), table, 1, "gtk_text_tag_table_get_size", "GtkTextTagTable*");
  return(C_TO_XEN_gint(gtk_text_tag_table_get_size(XEN_TO_C_GtkTextTagTable_(table))));
}

static XEN gxg_gtk_text_view_get_type(void)
{
  #define H_gtk_text_view_get_type "GtkType gtk_text_view_get_type( void)"
  return(C_TO_XEN_GtkType(gtk_text_view_get_type()));
}

static XEN gxg_gtk_text_view_new(void)
{
  #define H_gtk_text_view_new "GtkWidget* gtk_text_view_new( void)"
  return(C_TO_XEN_GtkWidget_(gtk_text_view_new()));
}

static XEN gxg_gtk_text_view_new_with_buffer(XEN buffer)
{
  #define H_gtk_text_view_new_with_buffer "GtkWidget* gtk_text_view_new_with_buffer(GtkTextBuffer* buffer)"
  XEN_ASSERT_TYPE(XEN_GtkTextBuffer__P(buffer), buffer, 1, "gtk_text_view_new_with_buffer", "GtkTextBuffer*");
  return(C_TO_XEN_GtkWidget_(gtk_text_view_new_with_buffer(XEN_TO_C_GtkTextBuffer_(buffer))));
}

static XEN gxg_gtk_text_view_set_buffer(XEN text_view, XEN buffer)
{
  #define H_gtk_text_view_set_buffer "void gtk_text_view_set_buffer(GtkTextView* text_view, GtkTextBuffer* buffer)"
  XEN_ASSERT_TYPE(XEN_GtkTextView__P(text_view), text_view, 1, "gtk_text_view_set_buffer", "GtkTextView*");
  XEN_ASSERT_TYPE(XEN_GtkTextBuffer__P(buffer), buffer, 2, "gtk_text_view_set_buffer", "GtkTextBuffer*");
  gtk_text_view_set_buffer(XEN_TO_C_GtkTextView_(text_view), XEN_TO_C_GtkTextBuffer_(buffer));
  return(XEN_FALSE);
}

static XEN gxg_gtk_text_view_get_buffer(XEN text_view)
{
  #define H_gtk_text_view_get_buffer "GtkTextBuffer* gtk_text_view_get_buffer(GtkTextView* text_view)"
  XEN_ASSERT_TYPE(XEN_GtkTextView__P(text_view), text_view, 1, "gtk_text_view_get_buffer", "GtkTextView*");
  return(C_TO_XEN_GtkTextBuffer_(gtk_text_view_get_buffer(XEN_TO_C_GtkTextView_(text_view))));
}

static XEN gxg_gtk_text_view_scroll_to_iter(XEN text_view, XEN iter, XEN within_margin, XEN use_align, XEN xalign, XEN yalign)
{
  #define H_gtk_text_view_scroll_to_iter "gboolean gtk_text_view_scroll_to_iter(GtkTextView* text_view, \
GtkTextIter* iter, gdouble within_margin, gboolean use_align, gdouble xalign, gdouble yalign)"
  XEN_ASSERT_TYPE(XEN_GtkTextView__P(text_view), text_view, 1, "gtk_text_view_scroll_to_iter", "GtkTextView*");
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(iter), iter, 2, "gtk_text_view_scroll_to_iter", "GtkTextIter*");
  XEN_ASSERT_TYPE(XEN_gdouble_P(within_margin), within_margin, 3, "gtk_text_view_scroll_to_iter", "gdouble");
  XEN_ASSERT_TYPE(XEN_gboolean_P(use_align), use_align, 4, "gtk_text_view_scroll_to_iter", "gboolean");
  XEN_ASSERT_TYPE(XEN_gdouble_P(xalign), xalign, 5, "gtk_text_view_scroll_to_iter", "gdouble");
  XEN_ASSERT_TYPE(XEN_gdouble_P(yalign), yalign, 6, "gtk_text_view_scroll_to_iter", "gdouble");
  return(C_TO_XEN_gboolean(gtk_text_view_scroll_to_iter(XEN_TO_C_GtkTextView_(text_view), XEN_TO_C_GtkTextIter_(iter), XEN_TO_C_gdouble(within_margin), 
                                                        XEN_TO_C_gboolean(use_align), XEN_TO_C_gdouble(xalign), XEN_TO_C_gdouble(yalign))));
}

static XEN gxg_gtk_text_view_scroll_to_mark(XEN text_view, XEN mark, XEN within_margin, XEN use_align, XEN xalign, XEN yalign)
{
  #define H_gtk_text_view_scroll_to_mark "void gtk_text_view_scroll_to_mark(GtkTextView* text_view, GtkTextMark* mark, \
gdouble within_margin, gboolean use_align, gdouble xalign, gdouble yalign)"
  XEN_ASSERT_TYPE(XEN_GtkTextView__P(text_view), text_view, 1, "gtk_text_view_scroll_to_mark", "GtkTextView*");
  XEN_ASSERT_TYPE(XEN_GtkTextMark__P(mark), mark, 2, "gtk_text_view_scroll_to_mark", "GtkTextMark*");
  XEN_ASSERT_TYPE(XEN_gdouble_P(within_margin), within_margin, 3, "gtk_text_view_scroll_to_mark", "gdouble");
  XEN_ASSERT_TYPE(XEN_gboolean_P(use_align), use_align, 4, "gtk_text_view_scroll_to_mark", "gboolean");
  XEN_ASSERT_TYPE(XEN_gdouble_P(xalign), xalign, 5, "gtk_text_view_scroll_to_mark", "gdouble");
  XEN_ASSERT_TYPE(XEN_gdouble_P(yalign), yalign, 6, "gtk_text_view_scroll_to_mark", "gdouble");
  gtk_text_view_scroll_to_mark(XEN_TO_C_GtkTextView_(text_view), XEN_TO_C_GtkTextMark_(mark), XEN_TO_C_gdouble(within_margin), 
                               XEN_TO_C_gboolean(use_align), XEN_TO_C_gdouble(xalign), XEN_TO_C_gdouble(yalign));
  return(XEN_FALSE);
}

static XEN gxg_gtk_text_view_scroll_mark_onscreen(XEN text_view, XEN mark)
{
  #define H_gtk_text_view_scroll_mark_onscreen "void gtk_text_view_scroll_mark_onscreen(GtkTextView* text_view, \
GtkTextMark* mark)"
  XEN_ASSERT_TYPE(XEN_GtkTextView__P(text_view), text_view, 1, "gtk_text_view_scroll_mark_onscreen", "GtkTextView*");
  XEN_ASSERT_TYPE(XEN_GtkTextMark__P(mark), mark, 2, "gtk_text_view_scroll_mark_onscreen", "GtkTextMark*");
  gtk_text_view_scroll_mark_onscreen(XEN_TO_C_GtkTextView_(text_view), XEN_TO_C_GtkTextMark_(mark));
  return(XEN_FALSE);
}

static XEN gxg_gtk_text_view_move_mark_onscreen(XEN text_view, XEN mark)
{
  #define H_gtk_text_view_move_mark_onscreen "gboolean gtk_text_view_move_mark_onscreen(GtkTextView* text_view, \
GtkTextMark* mark)"
  XEN_ASSERT_TYPE(XEN_GtkTextView__P(text_view), text_view, 1, "gtk_text_view_move_mark_onscreen", "GtkTextView*");
  XEN_ASSERT_TYPE(XEN_GtkTextMark__P(mark), mark, 2, "gtk_text_view_move_mark_onscreen", "GtkTextMark*");
  return(C_TO_XEN_gboolean(gtk_text_view_move_mark_onscreen(XEN_TO_C_GtkTextView_(text_view), XEN_TO_C_GtkTextMark_(mark))));
}

static XEN gxg_gtk_text_view_place_cursor_onscreen(XEN text_view)
{
  #define H_gtk_text_view_place_cursor_onscreen "gboolean gtk_text_view_place_cursor_onscreen(GtkTextView* text_view)"
  XEN_ASSERT_TYPE(XEN_GtkTextView__P(text_view), text_view, 1, "gtk_text_view_place_cursor_onscreen", "GtkTextView*");
  return(C_TO_XEN_gboolean(gtk_text_view_place_cursor_onscreen(XEN_TO_C_GtkTextView_(text_view))));
}

static XEN gxg_gtk_text_view_get_visible_rect(XEN text_view, XEN visible_rect)
{
  #define H_gtk_text_view_get_visible_rect "void gtk_text_view_get_visible_rect(GtkTextView* text_view, \
GdkRectangle* visible_rect)"
  XEN_ASSERT_TYPE(XEN_GtkTextView__P(text_view), text_view, 1, "gtk_text_view_get_visible_rect", "GtkTextView*");
  XEN_ASSERT_TYPE(XEN_GdkRectangle__P(visible_rect), visible_rect, 2, "gtk_text_view_get_visible_rect", "GdkRectangle*");
  gtk_text_view_get_visible_rect(XEN_TO_C_GtkTextView_(text_view), XEN_TO_C_GdkRectangle_(visible_rect));
  return(XEN_FALSE);
}

static XEN gxg_gtk_text_view_set_cursor_visible(XEN text_view, XEN setting)
{
  #define H_gtk_text_view_set_cursor_visible "void gtk_text_view_set_cursor_visible(GtkTextView* text_view, \
gboolean setting)"
  XEN_ASSERT_TYPE(XEN_GtkTextView__P(text_view), text_view, 1, "gtk_text_view_set_cursor_visible", "GtkTextView*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(setting), setting, 2, "gtk_text_view_set_cursor_visible", "gboolean");
  gtk_text_view_set_cursor_visible(XEN_TO_C_GtkTextView_(text_view), XEN_TO_C_gboolean(setting));
  return(XEN_FALSE);
}

static XEN gxg_gtk_text_view_get_cursor_visible(XEN text_view)
{
  #define H_gtk_text_view_get_cursor_visible "gboolean gtk_text_view_get_cursor_visible(GtkTextView* text_view)"
  XEN_ASSERT_TYPE(XEN_GtkTextView__P(text_view), text_view, 1, "gtk_text_view_get_cursor_visible", "GtkTextView*");
  return(C_TO_XEN_gboolean(gtk_text_view_get_cursor_visible(XEN_TO_C_GtkTextView_(text_view))));
}

static XEN gxg_gtk_text_view_get_iter_location(XEN text_view, XEN iter, XEN location)
{
  #define H_gtk_text_view_get_iter_location "void gtk_text_view_get_iter_location(GtkTextView* text_view, \
GtkTextIter* iter, GdkRectangle* location)"
  XEN_ASSERT_TYPE(XEN_GtkTextView__P(text_view), text_view, 1, "gtk_text_view_get_iter_location", "GtkTextView*");
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(iter), iter, 2, "gtk_text_view_get_iter_location", "GtkTextIter*");
  XEN_ASSERT_TYPE(XEN_GdkRectangle__P(location), location, 3, "gtk_text_view_get_iter_location", "GdkRectangle*");
  gtk_text_view_get_iter_location(XEN_TO_C_GtkTextView_(text_view), XEN_TO_C_GtkTextIter_(iter), XEN_TO_C_GdkRectangle_(location));
  return(XEN_FALSE);
}

static XEN gxg_gtk_text_view_get_iter_at_location(XEN text_view, XEN iter, XEN x, XEN y)
{
  #define H_gtk_text_view_get_iter_at_location "void gtk_text_view_get_iter_at_location(GtkTextView* text_view, \
GtkTextIter* iter, gint x, gint y)"
  XEN_ASSERT_TYPE(XEN_GtkTextView__P(text_view), text_view, 1, "gtk_text_view_get_iter_at_location", "GtkTextView*");
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(iter), iter, 2, "gtk_text_view_get_iter_at_location", "GtkTextIter*");
  XEN_ASSERT_TYPE(XEN_gint_P(x), x, 3, "gtk_text_view_get_iter_at_location", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(y), y, 4, "gtk_text_view_get_iter_at_location", "gint");
  gtk_text_view_get_iter_at_location(XEN_TO_C_GtkTextView_(text_view), XEN_TO_C_GtkTextIter_(iter), XEN_TO_C_gint(x), XEN_TO_C_gint(y));
  return(XEN_FALSE);
}

static XEN gxg_gtk_text_view_get_line_yrange(XEN text_view, XEN iter, XEN y, XEN height)
{
  #define H_gtk_text_view_get_line_yrange "void gtk_text_view_get_line_yrange(GtkTextView* text_view, \
GtkTextIter* iter, gint* y, gint* height)"
  XEN_ASSERT_TYPE(XEN_GtkTextView__P(text_view), text_view, 1, "gtk_text_view_get_line_yrange", "GtkTextView*");
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(iter), iter, 2, "gtk_text_view_get_line_yrange", "GtkTextIter*");
  XEN_ASSERT_TYPE(XEN_gint__P(y), y, 3, "gtk_text_view_get_line_yrange", "gint*");
  XEN_ASSERT_TYPE(XEN_gint__P(height), height, 4, "gtk_text_view_get_line_yrange", "gint*");
  gtk_text_view_get_line_yrange(XEN_TO_C_GtkTextView_(text_view), XEN_TO_C_GtkTextIter_(iter), XEN_TO_C_gint_(y), XEN_TO_C_gint_(height));
  return(XEN_FALSE);
}

static XEN gxg_gtk_text_view_get_line_at_y(XEN text_view, XEN target_iter, XEN y, XEN line_top)
{
  #define H_gtk_text_view_get_line_at_y "void gtk_text_view_get_line_at_y(GtkTextView* text_view, GtkTextIter* target_iter, \
gint y, gint* [line_top])"
  gint ref_line_top;
  XEN_ASSERT_TYPE(XEN_GtkTextView__P(text_view), text_view, 1, "gtk_text_view_get_line_at_y", "GtkTextView*");
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(target_iter), target_iter, 2, "gtk_text_view_get_line_at_y", "GtkTextIter*");
  XEN_ASSERT_TYPE(XEN_gint_P(y), y, 3, "gtk_text_view_get_line_at_y", "gint");
  gtk_text_view_get_line_at_y(XEN_TO_C_GtkTextView_(text_view), XEN_TO_C_GtkTextIter_(target_iter), XEN_TO_C_gint(y), &ref_line_top);
  return(XEN_LIST_1(C_TO_XEN_gint(ref_line_top)));
}

static XEN gxg_gtk_text_view_buffer_to_window_coords(XEN text_view, XEN win, XEN buffer_x, XEN buffer_y, XEN window_x, XEN window_y)
{
  #define H_gtk_text_view_buffer_to_window_coords "void gtk_text_view_buffer_to_window_coords(GtkTextView* text_view, \
GtkTextWindowType win, gint buffer_x, gint buffer_y, gint* [window_x], gint* [window_y])"
  gint ref_window_x;
  gint ref_window_y;
  XEN_ASSERT_TYPE(XEN_GtkTextView__P(text_view), text_view, 1, "gtk_text_view_buffer_to_window_coords", "GtkTextView*");
  XEN_ASSERT_TYPE(XEN_GtkTextWindowType_P(win), win, 2, "gtk_text_view_buffer_to_window_coords", "GtkTextWindowType");
  XEN_ASSERT_TYPE(XEN_gint_P(buffer_x), buffer_x, 3, "gtk_text_view_buffer_to_window_coords", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(buffer_y), buffer_y, 4, "gtk_text_view_buffer_to_window_coords", "gint");
  gtk_text_view_buffer_to_window_coords(XEN_TO_C_GtkTextView_(text_view), XEN_TO_C_GtkTextWindowType(win), XEN_TO_C_gint(buffer_x), 
                                        XEN_TO_C_gint(buffer_y), &ref_window_x, &ref_window_y);
  return(XEN_LIST_2(C_TO_XEN_gint(ref_window_x), C_TO_XEN_gint(ref_window_y)));
}

static XEN gxg_gtk_text_view_window_to_buffer_coords(XEN text_view, XEN win, XEN window_x, XEN window_y, XEN buffer_x, XEN buffer_y)
{
  #define H_gtk_text_view_window_to_buffer_coords "void gtk_text_view_window_to_buffer_coords(GtkTextView* text_view, \
GtkTextWindowType win, gint window_x, gint window_y, gint* [buffer_x], gint* [buffer_y])"
  gint ref_buffer_x;
  gint ref_buffer_y;
  XEN_ASSERT_TYPE(XEN_GtkTextView__P(text_view), text_view, 1, "gtk_text_view_window_to_buffer_coords", "GtkTextView*");
  XEN_ASSERT_TYPE(XEN_GtkTextWindowType_P(win), win, 2, "gtk_text_view_window_to_buffer_coords", "GtkTextWindowType");
  XEN_ASSERT_TYPE(XEN_gint_P(window_x), window_x, 3, "gtk_text_view_window_to_buffer_coords", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(window_y), window_y, 4, "gtk_text_view_window_to_buffer_coords", "gint");
  gtk_text_view_window_to_buffer_coords(XEN_TO_C_GtkTextView_(text_view), XEN_TO_C_GtkTextWindowType(win), XEN_TO_C_gint(window_x), 
                                        XEN_TO_C_gint(window_y), &ref_buffer_x, &ref_buffer_y);
  return(XEN_LIST_2(C_TO_XEN_gint(ref_buffer_x), C_TO_XEN_gint(ref_buffer_y)));
}

static XEN gxg_gtk_text_view_get_window(XEN text_view, XEN win)
{
  #define H_gtk_text_view_get_window "GdkWindow* gtk_text_view_get_window(GtkTextView* text_view, GtkTextWindowType win)"
  XEN_ASSERT_TYPE(XEN_GtkTextView__P(text_view), text_view, 1, "gtk_text_view_get_window", "GtkTextView*");
  XEN_ASSERT_TYPE(XEN_GtkTextWindowType_P(win), win, 2, "gtk_text_view_get_window", "GtkTextWindowType");
  return(C_TO_XEN_GdkWindow_(gtk_text_view_get_window(XEN_TO_C_GtkTextView_(text_view), XEN_TO_C_GtkTextWindowType(win))));
}

static XEN gxg_gtk_text_view_get_window_type(XEN text_view, XEN window)
{
  #define H_gtk_text_view_get_window_type "GtkTextWindowType gtk_text_view_get_window_type(GtkTextView* text_view, \
GdkWindow* window)"
  XEN_ASSERT_TYPE(XEN_GtkTextView__P(text_view), text_view, 1, "gtk_text_view_get_window_type", "GtkTextView*");
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 2, "gtk_text_view_get_window_type", "GdkWindow*");
  return(C_TO_XEN_GtkTextWindowType(gtk_text_view_get_window_type(XEN_TO_C_GtkTextView_(text_view), XEN_TO_C_GdkWindow_(window))));
}

static XEN gxg_gtk_text_view_set_border_window_size(XEN text_view, XEN type, XEN size)
{
  #define H_gtk_text_view_set_border_window_size "void gtk_text_view_set_border_window_size(GtkTextView* text_view, \
GtkTextWindowType type, gint size)"
  XEN_ASSERT_TYPE(XEN_GtkTextView__P(text_view), text_view, 1, "gtk_text_view_set_border_window_size", "GtkTextView*");
  XEN_ASSERT_TYPE(XEN_GtkTextWindowType_P(type), type, 2, "gtk_text_view_set_border_window_size", "GtkTextWindowType");
  XEN_ASSERT_TYPE(XEN_gint_P(size), size, 3, "gtk_text_view_set_border_window_size", "gint");
  gtk_text_view_set_border_window_size(XEN_TO_C_GtkTextView_(text_view), XEN_TO_C_GtkTextWindowType(type), XEN_TO_C_gint(size));
  return(XEN_FALSE);
}

static XEN gxg_gtk_text_view_get_border_window_size(XEN text_view, XEN type)
{
  #define H_gtk_text_view_get_border_window_size "gint gtk_text_view_get_border_window_size(GtkTextView* text_view, \
GtkTextWindowType type)"
  XEN_ASSERT_TYPE(XEN_GtkTextView__P(text_view), text_view, 1, "gtk_text_view_get_border_window_size", "GtkTextView*");
  XEN_ASSERT_TYPE(XEN_GtkTextWindowType_P(type), type, 2, "gtk_text_view_get_border_window_size", "GtkTextWindowType");
  return(C_TO_XEN_gint(gtk_text_view_get_border_window_size(XEN_TO_C_GtkTextView_(text_view), XEN_TO_C_GtkTextWindowType(type))));
}

static XEN gxg_gtk_text_view_forward_display_line(XEN text_view, XEN iter)
{
  #define H_gtk_text_view_forward_display_line "gboolean gtk_text_view_forward_display_line(GtkTextView* text_view, \
GtkTextIter* iter)"
  XEN_ASSERT_TYPE(XEN_GtkTextView__P(text_view), text_view, 1, "gtk_text_view_forward_display_line", "GtkTextView*");
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(iter), iter, 2, "gtk_text_view_forward_display_line", "GtkTextIter*");
  return(C_TO_XEN_gboolean(gtk_text_view_forward_display_line(XEN_TO_C_GtkTextView_(text_view), XEN_TO_C_GtkTextIter_(iter))));
}

static XEN gxg_gtk_text_view_backward_display_line(XEN text_view, XEN iter)
{
  #define H_gtk_text_view_backward_display_line "gboolean gtk_text_view_backward_display_line(GtkTextView* text_view, \
GtkTextIter* iter)"
  XEN_ASSERT_TYPE(XEN_GtkTextView__P(text_view), text_view, 1, "gtk_text_view_backward_display_line", "GtkTextView*");
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(iter), iter, 2, "gtk_text_view_backward_display_line", "GtkTextIter*");
  return(C_TO_XEN_gboolean(gtk_text_view_backward_display_line(XEN_TO_C_GtkTextView_(text_view), XEN_TO_C_GtkTextIter_(iter))));
}

static XEN gxg_gtk_text_view_forward_display_line_end(XEN text_view, XEN iter)
{
  #define H_gtk_text_view_forward_display_line_end "gboolean gtk_text_view_forward_display_line_end(GtkTextView* text_view, \
GtkTextIter* iter)"
  XEN_ASSERT_TYPE(XEN_GtkTextView__P(text_view), text_view, 1, "gtk_text_view_forward_display_line_end", "GtkTextView*");
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(iter), iter, 2, "gtk_text_view_forward_display_line_end", "GtkTextIter*");
  return(C_TO_XEN_gboolean(gtk_text_view_forward_display_line_end(XEN_TO_C_GtkTextView_(text_view), XEN_TO_C_GtkTextIter_(iter))));
}

static XEN gxg_gtk_text_view_backward_display_line_start(XEN text_view, XEN iter)
{
  #define H_gtk_text_view_backward_display_line_start "gboolean gtk_text_view_backward_display_line_start(GtkTextView* text_view, \
GtkTextIter* iter)"
  XEN_ASSERT_TYPE(XEN_GtkTextView__P(text_view), text_view, 1, "gtk_text_view_backward_display_line_start", "GtkTextView*");
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(iter), iter, 2, "gtk_text_view_backward_display_line_start", "GtkTextIter*");
  return(C_TO_XEN_gboolean(gtk_text_view_backward_display_line_start(XEN_TO_C_GtkTextView_(text_view), XEN_TO_C_GtkTextIter_(iter))));
}

static XEN gxg_gtk_text_view_starts_display_line(XEN text_view, XEN iter)
{
  #define H_gtk_text_view_starts_display_line "gboolean gtk_text_view_starts_display_line(GtkTextView* text_view, \
GtkTextIter* iter)"
  XEN_ASSERT_TYPE(XEN_GtkTextView__P(text_view), text_view, 1, "gtk_text_view_starts_display_line", "GtkTextView*");
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(iter), iter, 2, "gtk_text_view_starts_display_line", "GtkTextIter*");
  return(C_TO_XEN_gboolean(gtk_text_view_starts_display_line(XEN_TO_C_GtkTextView_(text_view), XEN_TO_C_GtkTextIter_(iter))));
}

static XEN gxg_gtk_text_view_move_visually(XEN text_view, XEN iter, XEN count)
{
  #define H_gtk_text_view_move_visually "gboolean gtk_text_view_move_visually(GtkTextView* text_view, \
GtkTextIter* iter, gint count)"
  XEN_ASSERT_TYPE(XEN_GtkTextView__P(text_view), text_view, 1, "gtk_text_view_move_visually", "GtkTextView*");
  XEN_ASSERT_TYPE(XEN_GtkTextIter__P(iter), iter, 2, "gtk_text_view_move_visually", "GtkTextIter*");
  XEN_ASSERT_TYPE(XEN_gint_P(count), count, 3, "gtk_text_view_move_visually", "gint");
  return(C_TO_XEN_gboolean(gtk_text_view_move_visually(XEN_TO_C_GtkTextView_(text_view), XEN_TO_C_GtkTextIter_(iter), XEN_TO_C_gint(count))));
}

static XEN gxg_gtk_text_view_add_child_at_anchor(XEN text_view, XEN child, XEN anchor)
{
  #define H_gtk_text_view_add_child_at_anchor "void gtk_text_view_add_child_at_anchor(GtkTextView* text_view, \
GtkWidget* child, GtkTextChildAnchor* anchor)"
  XEN_ASSERT_TYPE(XEN_GtkTextView__P(text_view), text_view, 1, "gtk_text_view_add_child_at_anchor", "GtkTextView*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(child), child, 2, "gtk_text_view_add_child_at_anchor", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_GtkTextChildAnchor__P(anchor), anchor, 3, "gtk_text_view_add_child_at_anchor", "GtkTextChildAnchor*");
  gtk_text_view_add_child_at_anchor(XEN_TO_C_GtkTextView_(text_view), XEN_TO_C_GtkWidget_(child), XEN_TO_C_GtkTextChildAnchor_(anchor));
  return(XEN_FALSE);
}

static XEN gxg_gtk_text_view_add_child_in_window(XEN text_view, XEN child, XEN which_window, XEN xpos, XEN ypos)
{
  #define H_gtk_text_view_add_child_in_window "void gtk_text_view_add_child_in_window(GtkTextView* text_view, \
GtkWidget* child, GtkTextWindowType which_window, gint xpos, gint ypos)"
  XEN_ASSERT_TYPE(XEN_GtkTextView__P(text_view), text_view, 1, "gtk_text_view_add_child_in_window", "GtkTextView*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(child), child, 2, "gtk_text_view_add_child_in_window", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_GtkTextWindowType_P(which_window), which_window, 3, "gtk_text_view_add_child_in_window", "GtkTextWindowType");
  XEN_ASSERT_TYPE(XEN_gint_P(xpos), xpos, 4, "gtk_text_view_add_child_in_window", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(ypos), ypos, 5, "gtk_text_view_add_child_in_window", "gint");
  gtk_text_view_add_child_in_window(XEN_TO_C_GtkTextView_(text_view), XEN_TO_C_GtkWidget_(child), XEN_TO_C_GtkTextWindowType(which_window), 
                                    XEN_TO_C_gint(xpos), XEN_TO_C_gint(ypos));
  return(XEN_FALSE);
}

static XEN gxg_gtk_text_view_move_child(XEN text_view, XEN child, XEN xpos, XEN ypos)
{
  #define H_gtk_text_view_move_child "void gtk_text_view_move_child(GtkTextView* text_view, GtkWidget* child, \
gint xpos, gint ypos)"
  XEN_ASSERT_TYPE(XEN_GtkTextView__P(text_view), text_view, 1, "gtk_text_view_move_child", "GtkTextView*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(child), child, 2, "gtk_text_view_move_child", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_gint_P(xpos), xpos, 3, "gtk_text_view_move_child", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(ypos), ypos, 4, "gtk_text_view_move_child", "gint");
  gtk_text_view_move_child(XEN_TO_C_GtkTextView_(text_view), XEN_TO_C_GtkWidget_(child), XEN_TO_C_gint(xpos), XEN_TO_C_gint(ypos));
  return(XEN_FALSE);
}

static XEN gxg_gtk_text_view_set_wrap_mode(XEN text_view, XEN wrap_mode)
{
  #define H_gtk_text_view_set_wrap_mode "void gtk_text_view_set_wrap_mode(GtkTextView* text_view, GtkWrapMode wrap_mode)"
  XEN_ASSERT_TYPE(XEN_GtkTextView__P(text_view), text_view, 1, "gtk_text_view_set_wrap_mode", "GtkTextView*");
  XEN_ASSERT_TYPE(XEN_GtkWrapMode_P(wrap_mode), wrap_mode, 2, "gtk_text_view_set_wrap_mode", "GtkWrapMode");
  gtk_text_view_set_wrap_mode(XEN_TO_C_GtkTextView_(text_view), XEN_TO_C_GtkWrapMode(wrap_mode));
  return(XEN_FALSE);
}

static XEN gxg_gtk_text_view_get_wrap_mode(XEN text_view)
{
  #define H_gtk_text_view_get_wrap_mode "GtkWrapMode gtk_text_view_get_wrap_mode(GtkTextView* text_view)"
  XEN_ASSERT_TYPE(XEN_GtkTextView__P(text_view), text_view, 1, "gtk_text_view_get_wrap_mode", "GtkTextView*");
  return(C_TO_XEN_GtkWrapMode(gtk_text_view_get_wrap_mode(XEN_TO_C_GtkTextView_(text_view))));
}

static XEN gxg_gtk_text_view_set_editable(XEN text_view, XEN setting)
{
  #define H_gtk_text_view_set_editable "void gtk_text_view_set_editable(GtkTextView* text_view, gboolean setting)"
  XEN_ASSERT_TYPE(XEN_GtkTextView__P(text_view), text_view, 1, "gtk_text_view_set_editable", "GtkTextView*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(setting), setting, 2, "gtk_text_view_set_editable", "gboolean");
  gtk_text_view_set_editable(XEN_TO_C_GtkTextView_(text_view), XEN_TO_C_gboolean(setting));
  return(XEN_FALSE);
}

static XEN gxg_gtk_text_view_get_editable(XEN text_view)
{
  #define H_gtk_text_view_get_editable "gboolean gtk_text_view_get_editable(GtkTextView* text_view)"
  XEN_ASSERT_TYPE(XEN_GtkTextView__P(text_view), text_view, 1, "gtk_text_view_get_editable", "GtkTextView*");
  return(C_TO_XEN_gboolean(gtk_text_view_get_editable(XEN_TO_C_GtkTextView_(text_view))));
}

static XEN gxg_gtk_text_view_set_pixels_above_lines(XEN text_view, XEN pixels_above_lines)
{
  #define H_gtk_text_view_set_pixels_above_lines "void gtk_text_view_set_pixels_above_lines(GtkTextView* text_view, \
gint pixels_above_lines)"
  XEN_ASSERT_TYPE(XEN_GtkTextView__P(text_view), text_view, 1, "gtk_text_view_set_pixels_above_lines", "GtkTextView*");
  XEN_ASSERT_TYPE(XEN_gint_P(pixels_above_lines), pixels_above_lines, 2, "gtk_text_view_set_pixels_above_lines", "gint");
  gtk_text_view_set_pixels_above_lines(XEN_TO_C_GtkTextView_(text_view), XEN_TO_C_gint(pixels_above_lines));
  return(XEN_FALSE);
}

static XEN gxg_gtk_text_view_get_pixels_above_lines(XEN text_view)
{
  #define H_gtk_text_view_get_pixels_above_lines "gint gtk_text_view_get_pixels_above_lines(GtkTextView* text_view)"
  XEN_ASSERT_TYPE(XEN_GtkTextView__P(text_view), text_view, 1, "gtk_text_view_get_pixels_above_lines", "GtkTextView*");
  return(C_TO_XEN_gint(gtk_text_view_get_pixels_above_lines(XEN_TO_C_GtkTextView_(text_view))));
}

static XEN gxg_gtk_text_view_set_pixels_below_lines(XEN text_view, XEN pixels_below_lines)
{
  #define H_gtk_text_view_set_pixels_below_lines "void gtk_text_view_set_pixels_below_lines(GtkTextView* text_view, \
gint pixels_below_lines)"
  XEN_ASSERT_TYPE(XEN_GtkTextView__P(text_view), text_view, 1, "gtk_text_view_set_pixels_below_lines", "GtkTextView*");
  XEN_ASSERT_TYPE(XEN_gint_P(pixels_below_lines), pixels_below_lines, 2, "gtk_text_view_set_pixels_below_lines", "gint");
  gtk_text_view_set_pixels_below_lines(XEN_TO_C_GtkTextView_(text_view), XEN_TO_C_gint(pixels_below_lines));
  return(XEN_FALSE);
}

static XEN gxg_gtk_text_view_get_pixels_below_lines(XEN text_view)
{
  #define H_gtk_text_view_get_pixels_below_lines "gint gtk_text_view_get_pixels_below_lines(GtkTextView* text_view)"
  XEN_ASSERT_TYPE(XEN_GtkTextView__P(text_view), text_view, 1, "gtk_text_view_get_pixels_below_lines", "GtkTextView*");
  return(C_TO_XEN_gint(gtk_text_view_get_pixels_below_lines(XEN_TO_C_GtkTextView_(text_view))));
}

static XEN gxg_gtk_text_view_set_pixels_inside_wrap(XEN text_view, XEN pixels_inside_wrap)
{
  #define H_gtk_text_view_set_pixels_inside_wrap "void gtk_text_view_set_pixels_inside_wrap(GtkTextView* text_view, \
gint pixels_inside_wrap)"
  XEN_ASSERT_TYPE(XEN_GtkTextView__P(text_view), text_view, 1, "gtk_text_view_set_pixels_inside_wrap", "GtkTextView*");
  XEN_ASSERT_TYPE(XEN_gint_P(pixels_inside_wrap), pixels_inside_wrap, 2, "gtk_text_view_set_pixels_inside_wrap", "gint");
  gtk_text_view_set_pixels_inside_wrap(XEN_TO_C_GtkTextView_(text_view), XEN_TO_C_gint(pixels_inside_wrap));
  return(XEN_FALSE);
}

static XEN gxg_gtk_text_view_get_pixels_inside_wrap(XEN text_view)
{
  #define H_gtk_text_view_get_pixels_inside_wrap "gint gtk_text_view_get_pixels_inside_wrap(GtkTextView* text_view)"
  XEN_ASSERT_TYPE(XEN_GtkTextView__P(text_view), text_view, 1, "gtk_text_view_get_pixels_inside_wrap", "GtkTextView*");
  return(C_TO_XEN_gint(gtk_text_view_get_pixels_inside_wrap(XEN_TO_C_GtkTextView_(text_view))));
}

static XEN gxg_gtk_text_view_set_justification(XEN text_view, XEN justification)
{
  #define H_gtk_text_view_set_justification "void gtk_text_view_set_justification(GtkTextView* text_view, \
GtkJustification justification)"
  XEN_ASSERT_TYPE(XEN_GtkTextView__P(text_view), text_view, 1, "gtk_text_view_set_justification", "GtkTextView*");
  XEN_ASSERT_TYPE(XEN_GtkJustification_P(justification), justification, 2, "gtk_text_view_set_justification", "GtkJustification");
  gtk_text_view_set_justification(XEN_TO_C_GtkTextView_(text_view), XEN_TO_C_GtkJustification(justification));
  return(XEN_FALSE);
}

static XEN gxg_gtk_text_view_get_justification(XEN text_view)
{
  #define H_gtk_text_view_get_justification "GtkJustification gtk_text_view_get_justification(GtkTextView* text_view)"
  XEN_ASSERT_TYPE(XEN_GtkTextView__P(text_view), text_view, 1, "gtk_text_view_get_justification", "GtkTextView*");
  return(C_TO_XEN_GtkJustification(gtk_text_view_get_justification(XEN_TO_C_GtkTextView_(text_view))));
}

static XEN gxg_gtk_text_view_set_left_margin(XEN text_view, XEN left_margin)
{
  #define H_gtk_text_view_set_left_margin "void gtk_text_view_set_left_margin(GtkTextView* text_view, \
gint left_margin)"
  XEN_ASSERT_TYPE(XEN_GtkTextView__P(text_view), text_view, 1, "gtk_text_view_set_left_margin", "GtkTextView*");
  XEN_ASSERT_TYPE(XEN_gint_P(left_margin), left_margin, 2, "gtk_text_view_set_left_margin", "gint");
  gtk_text_view_set_left_margin(XEN_TO_C_GtkTextView_(text_view), XEN_TO_C_gint(left_margin));
  return(XEN_FALSE);
}

static XEN gxg_gtk_text_view_get_left_margin(XEN text_view)
{
  #define H_gtk_text_view_get_left_margin "gint gtk_text_view_get_left_margin(GtkTextView* text_view)"
  XEN_ASSERT_TYPE(XEN_GtkTextView__P(text_view), text_view, 1, "gtk_text_view_get_left_margin", "GtkTextView*");
  return(C_TO_XEN_gint(gtk_text_view_get_left_margin(XEN_TO_C_GtkTextView_(text_view))));
}

static XEN gxg_gtk_text_view_set_right_margin(XEN text_view, XEN right_margin)
{
  #define H_gtk_text_view_set_right_margin "void gtk_text_view_set_right_margin(GtkTextView* text_view, \
gint right_margin)"
  XEN_ASSERT_TYPE(XEN_GtkTextView__P(text_view), text_view, 1, "gtk_text_view_set_right_margin", "GtkTextView*");
  XEN_ASSERT_TYPE(XEN_gint_P(right_margin), right_margin, 2, "gtk_text_view_set_right_margin", "gint");
  gtk_text_view_set_right_margin(XEN_TO_C_GtkTextView_(text_view), XEN_TO_C_gint(right_margin));
  return(XEN_FALSE);
}

static XEN gxg_gtk_text_view_get_right_margin(XEN text_view)
{
  #define H_gtk_text_view_get_right_margin "gint gtk_text_view_get_right_margin(GtkTextView* text_view)"
  XEN_ASSERT_TYPE(XEN_GtkTextView__P(text_view), text_view, 1, "gtk_text_view_get_right_margin", "GtkTextView*");
  return(C_TO_XEN_gint(gtk_text_view_get_right_margin(XEN_TO_C_GtkTextView_(text_view))));
}

static XEN gxg_gtk_text_view_set_indent(XEN text_view, XEN indent)
{
  #define H_gtk_text_view_set_indent "void gtk_text_view_set_indent(GtkTextView* text_view, gint indent)"
  XEN_ASSERT_TYPE(XEN_GtkTextView__P(text_view), text_view, 1, "gtk_text_view_set_indent", "GtkTextView*");
  XEN_ASSERT_TYPE(XEN_gint_P(indent), indent, 2, "gtk_text_view_set_indent", "gint");
  gtk_text_view_set_indent(XEN_TO_C_GtkTextView_(text_view), XEN_TO_C_gint(indent));
  return(XEN_FALSE);
}

static XEN gxg_gtk_text_view_get_indent(XEN text_view)
{
  #define H_gtk_text_view_get_indent "gint gtk_text_view_get_indent(GtkTextView* text_view)"
  XEN_ASSERT_TYPE(XEN_GtkTextView__P(text_view), text_view, 1, "gtk_text_view_get_indent", "GtkTextView*");
  return(C_TO_XEN_gint(gtk_text_view_get_indent(XEN_TO_C_GtkTextView_(text_view))));
}

static XEN gxg_gtk_text_view_set_tabs(XEN text_view, XEN tabs)
{
  #define H_gtk_text_view_set_tabs "void gtk_text_view_set_tabs(GtkTextView* text_view, PangoTabArray* tabs)"
  XEN_ASSERT_TYPE(XEN_GtkTextView__P(text_view), text_view, 1, "gtk_text_view_set_tabs", "GtkTextView*");
  XEN_ASSERT_TYPE(XEN_PangoTabArray__P(tabs), tabs, 2, "gtk_text_view_set_tabs", "PangoTabArray*");
  gtk_text_view_set_tabs(XEN_TO_C_GtkTextView_(text_view), XEN_TO_C_PangoTabArray_(tabs));
  return(XEN_FALSE);
}

static XEN gxg_gtk_text_view_get_tabs(XEN text_view)
{
  #define H_gtk_text_view_get_tabs "PangoTabArray* gtk_text_view_get_tabs(GtkTextView* text_view)"
  XEN_ASSERT_TYPE(XEN_GtkTextView__P(text_view), text_view, 1, "gtk_text_view_get_tabs", "GtkTextView*");
  return(C_TO_XEN_PangoTabArray_(gtk_text_view_get_tabs(XEN_TO_C_GtkTextView_(text_view))));
}

static XEN gxg_gtk_text_view_get_default_attributes(XEN text_view)
{
  #define H_gtk_text_view_get_default_attributes "GtkTextAttributes* gtk_text_view_get_default_attributes(GtkTextView* text_view)"
  XEN_ASSERT_TYPE(XEN_GtkTextView__P(text_view), text_view, 1, "gtk_text_view_get_default_attributes", "GtkTextView*");
  return(C_TO_XEN_GtkTextAttributes_(gtk_text_view_get_default_attributes(XEN_TO_C_GtkTextView_(text_view))));
}
#endif

static XEN gxg_gtk_toggle_button_get_type(void)
{
  #define H_gtk_toggle_button_get_type "GtkType gtk_toggle_button_get_type( void)"
  return(C_TO_XEN_GtkType(gtk_toggle_button_get_type()));
}

static XEN gxg_gtk_toggle_button_new(void)
{
  #define H_gtk_toggle_button_new "GtkWidget* gtk_toggle_button_new( void)"
  return(C_TO_XEN_GtkWidget_(gtk_toggle_button_new()));
}

static XEN gxg_gtk_toggle_button_new_with_label(XEN label)
{
  #define H_gtk_toggle_button_new_with_label "GtkWidget* gtk_toggle_button_new_with_label(gchar* label)"
  XEN_ASSERT_TYPE(XEN_gchar__P(label), label, 1, "gtk_toggle_button_new_with_label", "gchar*");
  return(C_TO_XEN_GtkWidget_(gtk_toggle_button_new_with_label(XEN_TO_C_gchar_(label))));
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_toggle_button_new_with_mnemonic(XEN label)
{
  #define H_gtk_toggle_button_new_with_mnemonic "GtkWidget* gtk_toggle_button_new_with_mnemonic(gchar* label)"
  XEN_ASSERT_TYPE(XEN_gchar__P(label), label, 1, "gtk_toggle_button_new_with_mnemonic", "gchar*");
  return(C_TO_XEN_GtkWidget_(gtk_toggle_button_new_with_mnemonic(XEN_TO_C_gchar_(label))));
}
#endif

static XEN gxg_gtk_toggle_button_set_mode(XEN toggle_button, XEN draw_indicator)
{
  #define H_gtk_toggle_button_set_mode "void gtk_toggle_button_set_mode(GtkToggleButton* toggle_button, \
gboolean draw_indicator)"
  XEN_ASSERT_TYPE(XEN_GtkToggleButton__P(toggle_button), toggle_button, 1, "gtk_toggle_button_set_mode", "GtkToggleButton*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(draw_indicator), draw_indicator, 2, "gtk_toggle_button_set_mode", "gboolean");
  gtk_toggle_button_set_mode(XEN_TO_C_GtkToggleButton_(toggle_button), XEN_TO_C_gboolean(draw_indicator));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_toggle_button_get_mode(XEN toggle_button)
{
  #define H_gtk_toggle_button_get_mode "gboolean gtk_toggle_button_get_mode(GtkToggleButton* toggle_button)"
  XEN_ASSERT_TYPE(XEN_GtkToggleButton__P(toggle_button), toggle_button, 1, "gtk_toggle_button_get_mode", "GtkToggleButton*");
  return(C_TO_XEN_gboolean(gtk_toggle_button_get_mode(XEN_TO_C_GtkToggleButton_(toggle_button))));
}
#endif

static XEN gxg_gtk_toggle_button_set_active(XEN toggle_button, XEN is_active)
{
  #define H_gtk_toggle_button_set_active "void gtk_toggle_button_set_active(GtkToggleButton* toggle_button, \
gboolean is_active)"
  XEN_ASSERT_TYPE(XEN_GtkToggleButton__P(toggle_button), toggle_button, 1, "gtk_toggle_button_set_active", "GtkToggleButton*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(is_active), is_active, 2, "gtk_toggle_button_set_active", "gboolean");
  gtk_toggle_button_set_active(XEN_TO_C_GtkToggleButton_(toggle_button), XEN_TO_C_gboolean(is_active));
  return(XEN_FALSE);
}

static XEN gxg_gtk_toggle_button_get_active(XEN toggle_button)
{
  #define H_gtk_toggle_button_get_active "gboolean gtk_toggle_button_get_active(GtkToggleButton* toggle_button)"
  XEN_ASSERT_TYPE(XEN_GtkToggleButton__P(toggle_button), toggle_button, 1, "gtk_toggle_button_get_active", "GtkToggleButton*");
  return(C_TO_XEN_gboolean(gtk_toggle_button_get_active(XEN_TO_C_GtkToggleButton_(toggle_button))));
}

static XEN gxg_gtk_toggle_button_toggled(XEN toggle_button)
{
  #define H_gtk_toggle_button_toggled "void gtk_toggle_button_toggled(GtkToggleButton* toggle_button)"
  XEN_ASSERT_TYPE(XEN_GtkToggleButton__P(toggle_button), toggle_button, 1, "gtk_toggle_button_toggled", "GtkToggleButton*");
  gtk_toggle_button_toggled(XEN_TO_C_GtkToggleButton_(toggle_button));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_toggle_button_set_inconsistent(XEN toggle_button, XEN setting)
{
  #define H_gtk_toggle_button_set_inconsistent "void gtk_toggle_button_set_inconsistent(GtkToggleButton* toggle_button, \
gboolean setting)"
  XEN_ASSERT_TYPE(XEN_GtkToggleButton__P(toggle_button), toggle_button, 1, "gtk_toggle_button_set_inconsistent", "GtkToggleButton*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(setting), setting, 2, "gtk_toggle_button_set_inconsistent", "gboolean");
  gtk_toggle_button_set_inconsistent(XEN_TO_C_GtkToggleButton_(toggle_button), XEN_TO_C_gboolean(setting));
  return(XEN_FALSE);
}

static XEN gxg_gtk_toggle_button_get_inconsistent(XEN toggle_button)
{
  #define H_gtk_toggle_button_get_inconsistent "gboolean gtk_toggle_button_get_inconsistent(GtkToggleButton* toggle_button)"
  XEN_ASSERT_TYPE(XEN_GtkToggleButton__P(toggle_button), toggle_button, 1, "gtk_toggle_button_get_inconsistent", "GtkToggleButton*");
  return(C_TO_XEN_gboolean(gtk_toggle_button_get_inconsistent(XEN_TO_C_GtkToggleButton_(toggle_button))));
}
#endif

static XEN gxg_gtk_toolbar_get_type(void)
{
  #define H_gtk_toolbar_get_type "GtkType gtk_toolbar_get_type( void)"
  return(C_TO_XEN_GtkType(gtk_toolbar_get_type()));
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_toolbar_new(void)
{
  #define H_gtk_toolbar_new "GtkWidget* gtk_toolbar_new( void)"
  return(C_TO_XEN_GtkWidget_(gtk_toolbar_new()));
}
#endif

static XEN gxg_gtk_toolbar_append_item(XEN toolbar, XEN text, XEN tooltip_text, XEN tooltip_private_text, XEN icon, XEN func, XEN func_data)
{
  #define H_gtk_toolbar_append_item "GtkWidget* gtk_toolbar_append_item(GtkToolbar* toolbar, char* text, \
char* tooltip_text, char* tooltip_private_text, GtkWidget* icon, GtkSignalFunc func, lambda_data func_data)"
  XEN_ASSERT_TYPE(XEN_GtkToolbar__P(toolbar), toolbar, 1, "gtk_toolbar_append_item", "GtkToolbar*");
  XEN_ASSERT_TYPE(XEN_char__P(text), text, 2, "gtk_toolbar_append_item", "char*");
  XEN_ASSERT_TYPE(XEN_char__P(tooltip_text), tooltip_text, 3, "gtk_toolbar_append_item", "char*");
  XEN_ASSERT_TYPE(XEN_char__P(tooltip_private_text), tooltip_private_text, 4, "gtk_toolbar_append_item", "char*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(icon), icon, 5, "gtk_toolbar_append_item", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_GtkSignalFunc_P(func), func, 6, "gtk_toolbar_append_item", "GtkSignalFunc");
  if (XEN_NOT_BOUND_P(func_data)) func_data = XEN_FALSE; 
  else XEN_ASSERT_TYPE(XEN_lambda_data_P(func_data), func_data, 7, "gtk_toolbar_append_item", "lambda_data");
  {
    XEN result = XEN_FALSE;
    int loc;
    XEN gxg_ptr = XEN_LIST_5(func, func_data, XEN_FALSE, XEN_FALSE, XEN_FALSE);
    loc = xm_protect(gxg_ptr);
    XEN_LIST_SET(gxg_ptr, 2, C_TO_XEN_INT(loc));
    result = C_TO_XEN_GtkWidget_(gtk_toolbar_append_item(XEN_TO_C_GtkToolbar_(toolbar), XEN_TO_C_char_(text), XEN_TO_C_char_(tooltip_text), 
                                                         XEN_TO_C_char_(tooltip_private_text), XEN_TO_C_GtkWidget_(icon), 
                                                         XEN_TO_C_GtkSignalFunc(func), XEN_TO_C_lambda_data(func_data)));
    return(result);
   }
}

static XEN gxg_gtk_toolbar_prepend_item(XEN toolbar, XEN text, XEN tooltip_text, XEN tooltip_private_text, XEN icon, XEN func, XEN func_data)
{
  #define H_gtk_toolbar_prepend_item "GtkWidget* gtk_toolbar_prepend_item(GtkToolbar* toolbar, char* text, \
char* tooltip_text, char* tooltip_private_text, GtkWidget* icon, GtkSignalFunc func, lambda_data func_data)"
  XEN_ASSERT_TYPE(XEN_GtkToolbar__P(toolbar), toolbar, 1, "gtk_toolbar_prepend_item", "GtkToolbar*");
  XEN_ASSERT_TYPE(XEN_char__P(text), text, 2, "gtk_toolbar_prepend_item", "char*");
  XEN_ASSERT_TYPE(XEN_char__P(tooltip_text), tooltip_text, 3, "gtk_toolbar_prepend_item", "char*");
  XEN_ASSERT_TYPE(XEN_char__P(tooltip_private_text), tooltip_private_text, 4, "gtk_toolbar_prepend_item", "char*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(icon), icon, 5, "gtk_toolbar_prepend_item", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_GtkSignalFunc_P(func), func, 6, "gtk_toolbar_prepend_item", "GtkSignalFunc");
  if (XEN_NOT_BOUND_P(func_data)) func_data = XEN_FALSE; 
  else XEN_ASSERT_TYPE(XEN_lambda_data_P(func_data), func_data, 7, "gtk_toolbar_prepend_item", "lambda_data");
  {
    XEN result = XEN_FALSE;
    int loc;
    XEN gxg_ptr = XEN_LIST_5(func, func_data, XEN_FALSE, XEN_FALSE, XEN_FALSE);
    loc = xm_protect(gxg_ptr);
    XEN_LIST_SET(gxg_ptr, 2, C_TO_XEN_INT(loc));
    result = C_TO_XEN_GtkWidget_(gtk_toolbar_prepend_item(XEN_TO_C_GtkToolbar_(toolbar), XEN_TO_C_char_(text), XEN_TO_C_char_(tooltip_text), 
                                                          XEN_TO_C_char_(tooltip_private_text), XEN_TO_C_GtkWidget_(icon), 
                                                          XEN_TO_C_GtkSignalFunc(func), XEN_TO_C_lambda_data(func_data)));
    return(result);
   }
}

static XEN gxg_gtk_toolbar_insert_item(XEN toolbar, XEN text, XEN tooltip_text, XEN tooltip_private_text, XEN icon, XEN func, XEN func_data, XEN position)
{
  #define H_gtk_toolbar_insert_item "GtkWidget* gtk_toolbar_insert_item(GtkToolbar* toolbar, char* text, \
char* tooltip_text, char* tooltip_private_text, GtkWidget* icon, GtkSignalFunc func, lambda_data func_data, \
gint position)"
  XEN_ASSERT_TYPE(XEN_GtkToolbar__P(toolbar), toolbar, 1, "gtk_toolbar_insert_item", "GtkToolbar*");
  XEN_ASSERT_TYPE(XEN_char__P(text), text, 2, "gtk_toolbar_insert_item", "char*");
  XEN_ASSERT_TYPE(XEN_char__P(tooltip_text), tooltip_text, 3, "gtk_toolbar_insert_item", "char*");
  XEN_ASSERT_TYPE(XEN_char__P(tooltip_private_text), tooltip_private_text, 4, "gtk_toolbar_insert_item", "char*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(icon), icon, 5, "gtk_toolbar_insert_item", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_GtkSignalFunc_P(func), func, 6, "gtk_toolbar_insert_item", "GtkSignalFunc");
  XEN_ASSERT_TYPE(XEN_lambda_data_P(func_data), func_data, 7, "gtk_toolbar_insert_item", "lambda_data");
  XEN_ASSERT_TYPE(XEN_gint_P(position), position, 8, "gtk_toolbar_insert_item", "gint");
  {
    XEN result = XEN_FALSE;
    int loc;
    XEN gxg_ptr = XEN_LIST_5(func, func_data, XEN_FALSE, XEN_FALSE, XEN_FALSE);
    loc = xm_protect(gxg_ptr);
    XEN_LIST_SET(gxg_ptr, 2, C_TO_XEN_INT(loc));
    result = C_TO_XEN_GtkWidget_(gtk_toolbar_insert_item(XEN_TO_C_GtkToolbar_(toolbar), XEN_TO_C_char_(text), XEN_TO_C_char_(tooltip_text), 
                                                         XEN_TO_C_char_(tooltip_private_text), XEN_TO_C_GtkWidget_(icon), 
                                                         XEN_TO_C_GtkSignalFunc(func), XEN_TO_C_lambda_data(func_data), XEN_TO_C_gint(position)));
    return(result);
   }
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_toolbar_insert_stock(XEN toolbar, XEN stock_id, XEN tooltip_text, XEN tooltip_private_text, XEN func, XEN func_data, XEN position)
{
  #define H_gtk_toolbar_insert_stock "GtkWidget* gtk_toolbar_insert_stock(GtkToolbar* toolbar, gchar* stock_id, \
char* tooltip_text, char* tooltip_private_text, GtkSignalFunc func, lambda_data func_data, gint position)"
  XEN_ASSERT_TYPE(XEN_GtkToolbar__P(toolbar), toolbar, 1, "gtk_toolbar_insert_stock", "GtkToolbar*");
  XEN_ASSERT_TYPE(XEN_gchar__P(stock_id), stock_id, 2, "gtk_toolbar_insert_stock", "gchar*");
  XEN_ASSERT_TYPE(XEN_char__P(tooltip_text), tooltip_text, 3, "gtk_toolbar_insert_stock", "char*");
  XEN_ASSERT_TYPE(XEN_char__P(tooltip_private_text), tooltip_private_text, 4, "gtk_toolbar_insert_stock", "char*");
  XEN_ASSERT_TYPE(XEN_GtkSignalFunc_P(func), func, 5, "gtk_toolbar_insert_stock", "GtkSignalFunc");
  XEN_ASSERT_TYPE(XEN_lambda_data_P(func_data), func_data, 6, "gtk_toolbar_insert_stock", "lambda_data");
  XEN_ASSERT_TYPE(XEN_gint_P(position), position, 7, "gtk_toolbar_insert_stock", "gint");
  {
    XEN result = XEN_FALSE;
    int loc;
    XEN gxg_ptr = XEN_LIST_5(func, func_data, XEN_FALSE, XEN_FALSE, XEN_FALSE);
    loc = xm_protect(gxg_ptr);
    XEN_LIST_SET(gxg_ptr, 2, C_TO_XEN_INT(loc));
    result = C_TO_XEN_GtkWidget_(gtk_toolbar_insert_stock(XEN_TO_C_GtkToolbar_(toolbar), XEN_TO_C_gchar_(stock_id), XEN_TO_C_char_(tooltip_text), 
                                                          XEN_TO_C_char_(tooltip_private_text), XEN_TO_C_GtkSignalFunc(func), 
                                                          XEN_TO_C_lambda_data(func_data), XEN_TO_C_gint(position)));
    return(result);
   }
}
#endif

static XEN gxg_gtk_toolbar_append_space(XEN toolbar)
{
  #define H_gtk_toolbar_append_space "void gtk_toolbar_append_space(GtkToolbar* toolbar)"
  XEN_ASSERT_TYPE(XEN_GtkToolbar__P(toolbar), toolbar, 1, "gtk_toolbar_append_space", "GtkToolbar*");
  gtk_toolbar_append_space(XEN_TO_C_GtkToolbar_(toolbar));
  return(XEN_FALSE);
}

static XEN gxg_gtk_toolbar_prepend_space(XEN toolbar)
{
  #define H_gtk_toolbar_prepend_space "void gtk_toolbar_prepend_space(GtkToolbar* toolbar)"
  XEN_ASSERT_TYPE(XEN_GtkToolbar__P(toolbar), toolbar, 1, "gtk_toolbar_prepend_space", "GtkToolbar*");
  gtk_toolbar_prepend_space(XEN_TO_C_GtkToolbar_(toolbar));
  return(XEN_FALSE);
}

static XEN gxg_gtk_toolbar_insert_space(XEN toolbar, XEN position)
{
  #define H_gtk_toolbar_insert_space "void gtk_toolbar_insert_space(GtkToolbar* toolbar, gint position)"
  XEN_ASSERT_TYPE(XEN_GtkToolbar__P(toolbar), toolbar, 1, "gtk_toolbar_insert_space", "GtkToolbar*");
  XEN_ASSERT_TYPE(XEN_gint_P(position), position, 2, "gtk_toolbar_insert_space", "gint");
  gtk_toolbar_insert_space(XEN_TO_C_GtkToolbar_(toolbar), XEN_TO_C_gint(position));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_toolbar_remove_space(XEN toolbar, XEN position)
{
  #define H_gtk_toolbar_remove_space "void gtk_toolbar_remove_space(GtkToolbar* toolbar, gint position)"
  XEN_ASSERT_TYPE(XEN_GtkToolbar__P(toolbar), toolbar, 1, "gtk_toolbar_remove_space", "GtkToolbar*");
  XEN_ASSERT_TYPE(XEN_gint_P(position), position, 2, "gtk_toolbar_remove_space", "gint");
  gtk_toolbar_remove_space(XEN_TO_C_GtkToolbar_(toolbar), XEN_TO_C_gint(position));
  return(XEN_FALSE);
}
#endif

static XEN gxg_gtk_toolbar_append_element(XEN toolbar, XEN type, XEN widget, XEN text, XEN tooltip_text, XEN tooltip_private_text, XEN icon, XEN func, XEN func_data)
{
  #define H_gtk_toolbar_append_element "GtkWidget* gtk_toolbar_append_element(GtkToolbar* toolbar, GtkToolbarChildType type, \
GtkWidget* widget, char* text, char* tooltip_text, char* tooltip_private_text, GtkWidget* icon, GtkSignalFunc func, \
lambda_data func_data)"
  XEN_ASSERT_TYPE(XEN_GtkToolbar__P(toolbar), toolbar, 1, "gtk_toolbar_append_element", "GtkToolbar*");
  XEN_ASSERT_TYPE(XEN_GtkToolbarChildType_P(type), type, 2, "gtk_toolbar_append_element", "GtkToolbarChildType");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 3, "gtk_toolbar_append_element", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_char__P(text), text, 4, "gtk_toolbar_append_element", "char*");
  XEN_ASSERT_TYPE(XEN_char__P(tooltip_text), tooltip_text, 5, "gtk_toolbar_append_element", "char*");
  XEN_ASSERT_TYPE(XEN_char__P(tooltip_private_text), tooltip_private_text, 6, "gtk_toolbar_append_element", "char*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(icon), icon, 7, "gtk_toolbar_append_element", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_GtkSignalFunc_P(func), func, 8, "gtk_toolbar_append_element", "GtkSignalFunc");
  if (XEN_NOT_BOUND_P(func_data)) func_data = XEN_FALSE; 
  else XEN_ASSERT_TYPE(XEN_lambda_data_P(func_data), func_data, 9, "gtk_toolbar_append_element", "lambda_data");
  {
    XEN result = XEN_FALSE;
    int loc;
    XEN gxg_ptr = XEN_LIST_5(func, func_data, XEN_FALSE, XEN_FALSE, XEN_FALSE);
    loc = xm_protect(gxg_ptr);
    XEN_LIST_SET(gxg_ptr, 2, C_TO_XEN_INT(loc));
    result = C_TO_XEN_GtkWidget_(gtk_toolbar_append_element(XEN_TO_C_GtkToolbar_(toolbar), XEN_TO_C_GtkToolbarChildType(type), 
                                                            XEN_TO_C_GtkWidget_(widget), XEN_TO_C_char_(text), XEN_TO_C_char_(tooltip_text), 
                                                            XEN_TO_C_char_(tooltip_private_text), XEN_TO_C_GtkWidget_(icon), 
                                                            XEN_TO_C_GtkSignalFunc(func), XEN_TO_C_lambda_data(func_data)));
    return(result);
   }
}

static XEN gxg_gtk_toolbar_prepend_element(XEN toolbar, XEN type, XEN widget, XEN text, XEN tooltip_text, XEN tooltip_private_text, XEN icon, XEN func, XEN func_data)
{
  #define H_gtk_toolbar_prepend_element "GtkWidget* gtk_toolbar_prepend_element(GtkToolbar* toolbar, \
GtkToolbarChildType type, GtkWidget* widget, char* text, char* tooltip_text, char* tooltip_private_text, \
GtkWidget* icon, GtkSignalFunc func, lambda_data func_data)"
  XEN_ASSERT_TYPE(XEN_GtkToolbar__P(toolbar), toolbar, 1, "gtk_toolbar_prepend_element", "GtkToolbar*");
  XEN_ASSERT_TYPE(XEN_GtkToolbarChildType_P(type), type, 2, "gtk_toolbar_prepend_element", "GtkToolbarChildType");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 3, "gtk_toolbar_prepend_element", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_char__P(text), text, 4, "gtk_toolbar_prepend_element", "char*");
  XEN_ASSERT_TYPE(XEN_char__P(tooltip_text), tooltip_text, 5, "gtk_toolbar_prepend_element", "char*");
  XEN_ASSERT_TYPE(XEN_char__P(tooltip_private_text), tooltip_private_text, 6, "gtk_toolbar_prepend_element", "char*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(icon), icon, 7, "gtk_toolbar_prepend_element", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_GtkSignalFunc_P(func), func, 8, "gtk_toolbar_prepend_element", "GtkSignalFunc");
  if (XEN_NOT_BOUND_P(func_data)) func_data = XEN_FALSE; 
  else XEN_ASSERT_TYPE(XEN_lambda_data_P(func_data), func_data, 9, "gtk_toolbar_prepend_element", "lambda_data");
  {
    XEN result = XEN_FALSE;
    int loc;
    XEN gxg_ptr = XEN_LIST_5(func, func_data, XEN_FALSE, XEN_FALSE, XEN_FALSE);
    loc = xm_protect(gxg_ptr);
    XEN_LIST_SET(gxg_ptr, 2, C_TO_XEN_INT(loc));
    result = C_TO_XEN_GtkWidget_(gtk_toolbar_prepend_element(XEN_TO_C_GtkToolbar_(toolbar), XEN_TO_C_GtkToolbarChildType(type), 
                                                             XEN_TO_C_GtkWidget_(widget), XEN_TO_C_char_(text), XEN_TO_C_char_(tooltip_text), 
                                                             XEN_TO_C_char_(tooltip_private_text), XEN_TO_C_GtkWidget_(icon), 
                                                             XEN_TO_C_GtkSignalFunc(func), XEN_TO_C_lambda_data(func_data)));
    return(result);
   }
}

static XEN gxg_gtk_toolbar_insert_element(XEN arglist)
{
  #define H_gtk_toolbar_insert_element "GtkWidget* gtk_toolbar_insert_element(GtkToolbar* toolbar, GtkToolbarChildType type, \
GtkWidget* widget, char* text, char* tooltip_text, char* tooltip_private_text, GtkWidget* icon, GtkSignalFunc func, \
lambda_data func_data, gint position)"
  XEN toolbar, type, widget, text, tooltip_text, tooltip_private_text, icon, func, func_data, position;
  toolbar = XEN_LIST_REF(arglist, 0);
  type = XEN_LIST_REF(arglist, 1);
  widget = XEN_LIST_REF(arglist, 2);
  text = XEN_LIST_REF(arglist, 3);
  tooltip_text = XEN_LIST_REF(arglist, 4);
  tooltip_private_text = XEN_LIST_REF(arglist, 5);
  icon = XEN_LIST_REF(arglist, 6);
  func = XEN_LIST_REF(arglist, 7);
  func_data = XEN_LIST_REF(arglist, 8);
  position = XEN_LIST_REF(arglist, 9);
  XEN_ASSERT_TYPE(XEN_GtkToolbar__P(toolbar), toolbar, 1, "gtk_toolbar_insert_element", "GtkToolbar*");
  XEN_ASSERT_TYPE(XEN_GtkToolbarChildType_P(type), type, 2, "gtk_toolbar_insert_element", "GtkToolbarChildType");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget) || XEN_FALSE_P(widget), widget, 3, "gtk_toolbar_insert_element", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_char__P(text), text, 4, "gtk_toolbar_insert_element", "char*");
  XEN_ASSERT_TYPE(XEN_char__P(tooltip_text), tooltip_text, 5, "gtk_toolbar_insert_element", "char*");
  XEN_ASSERT_TYPE(XEN_char__P(tooltip_private_text), tooltip_private_text, 6, "gtk_toolbar_insert_element", "char*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(icon), icon, 7, "gtk_toolbar_insert_element", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_GtkSignalFunc_P(func), func, 8, "gtk_toolbar_insert_element", "GtkSignalFunc");
  XEN_ASSERT_TYPE(XEN_lambda_data_P(func_data), func_data, 9, "gtk_toolbar_insert_element", "lambda_data");
  XEN_ASSERT_TYPE(XEN_gint_P(position), position, 10, "gtk_toolbar_insert_element", "gint");
  {
    XEN result = XEN_FALSE;
    int loc;
    XEN gxg_ptr = XEN_LIST_5(func, func_data, XEN_FALSE, XEN_FALSE, XEN_FALSE);
    loc = xm_protect(gxg_ptr);
    XEN_LIST_SET(gxg_ptr, 2, C_TO_XEN_INT(loc));
    result = C_TO_XEN_GtkWidget_(gtk_toolbar_insert_element(XEN_TO_C_GtkToolbar_(toolbar), XEN_TO_C_GtkToolbarChildType(type), 
                                                            XEN_TO_C_GtkWidget_(widget), XEN_TO_C_char_(text), XEN_TO_C_char_(tooltip_text), 
                                                            XEN_TO_C_char_(tooltip_private_text), XEN_TO_C_GtkWidget_(icon), 
                                                            XEN_TO_C_GtkSignalFunc(func), XEN_TO_C_lambda_data(func_data), 
                                                            XEN_TO_C_gint(position)));
    return(result);
   }
}

static XEN gxg_gtk_toolbar_append_widget(XEN toolbar, XEN widget, XEN tooltip_text, XEN tooltip_private_text)
{
  #define H_gtk_toolbar_append_widget "void gtk_toolbar_append_widget(GtkToolbar* toolbar, GtkWidget* widget, \
char* tooltip_text, char* tooltip_private_text)"
  XEN_ASSERT_TYPE(XEN_GtkToolbar__P(toolbar), toolbar, 1, "gtk_toolbar_append_widget", "GtkToolbar*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 2, "gtk_toolbar_append_widget", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_char__P(tooltip_text), tooltip_text, 3, "gtk_toolbar_append_widget", "char*");
  XEN_ASSERT_TYPE(XEN_char__P(tooltip_private_text), tooltip_private_text, 4, "gtk_toolbar_append_widget", "char*");
  gtk_toolbar_append_widget(XEN_TO_C_GtkToolbar_(toolbar), XEN_TO_C_GtkWidget_(widget), XEN_TO_C_char_(tooltip_text), XEN_TO_C_char_(tooltip_private_text));
  return(XEN_FALSE);
}

static XEN gxg_gtk_toolbar_prepend_widget(XEN toolbar, XEN widget, XEN tooltip_text, XEN tooltip_private_text)
{
  #define H_gtk_toolbar_prepend_widget "void gtk_toolbar_prepend_widget(GtkToolbar* toolbar, GtkWidget* widget, \
char* tooltip_text, char* tooltip_private_text)"
  XEN_ASSERT_TYPE(XEN_GtkToolbar__P(toolbar), toolbar, 1, "gtk_toolbar_prepend_widget", "GtkToolbar*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 2, "gtk_toolbar_prepend_widget", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_char__P(tooltip_text), tooltip_text, 3, "gtk_toolbar_prepend_widget", "char*");
  XEN_ASSERT_TYPE(XEN_char__P(tooltip_private_text), tooltip_private_text, 4, "gtk_toolbar_prepend_widget", "char*");
  gtk_toolbar_prepend_widget(XEN_TO_C_GtkToolbar_(toolbar), XEN_TO_C_GtkWidget_(widget), XEN_TO_C_char_(tooltip_text), XEN_TO_C_char_(tooltip_private_text));
  return(XEN_FALSE);
}

static XEN gxg_gtk_toolbar_insert_widget(XEN toolbar, XEN widget, XEN tooltip_text, XEN tooltip_private_text, XEN position)
{
  #define H_gtk_toolbar_insert_widget "void gtk_toolbar_insert_widget(GtkToolbar* toolbar, GtkWidget* widget, \
char* tooltip_text, char* tooltip_private_text, gint position)"
  XEN_ASSERT_TYPE(XEN_GtkToolbar__P(toolbar), toolbar, 1, "gtk_toolbar_insert_widget", "GtkToolbar*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 2, "gtk_toolbar_insert_widget", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_char__P(tooltip_text), tooltip_text, 3, "gtk_toolbar_insert_widget", "char*");
  XEN_ASSERT_TYPE(XEN_char__P(tooltip_private_text), tooltip_private_text, 4, "gtk_toolbar_insert_widget", "char*");
  XEN_ASSERT_TYPE(XEN_gint_P(position), position, 5, "gtk_toolbar_insert_widget", "gint");
  gtk_toolbar_insert_widget(XEN_TO_C_GtkToolbar_(toolbar), XEN_TO_C_GtkWidget_(widget), XEN_TO_C_char_(tooltip_text), XEN_TO_C_char_(tooltip_private_text), 
                            XEN_TO_C_gint(position));
  return(XEN_FALSE);
}

static XEN gxg_gtk_toolbar_set_orientation(XEN toolbar, XEN orientation)
{
  #define H_gtk_toolbar_set_orientation "void gtk_toolbar_set_orientation(GtkToolbar* toolbar, GtkOrientation orientation)"
  XEN_ASSERT_TYPE(XEN_GtkToolbar__P(toolbar), toolbar, 1, "gtk_toolbar_set_orientation", "GtkToolbar*");
  XEN_ASSERT_TYPE(XEN_GtkOrientation_P(orientation), orientation, 2, "gtk_toolbar_set_orientation", "GtkOrientation");
  gtk_toolbar_set_orientation(XEN_TO_C_GtkToolbar_(toolbar), XEN_TO_C_GtkOrientation(orientation));
  return(XEN_FALSE);
}

static XEN gxg_gtk_toolbar_set_style(XEN toolbar, XEN style)
{
  #define H_gtk_toolbar_set_style "void gtk_toolbar_set_style(GtkToolbar* toolbar, GtkToolbarStyle style)"
  XEN_ASSERT_TYPE(XEN_GtkToolbar__P(toolbar), toolbar, 1, "gtk_toolbar_set_style", "GtkToolbar*");
  XEN_ASSERT_TYPE(XEN_GtkToolbarStyle_P(style), style, 2, "gtk_toolbar_set_style", "GtkToolbarStyle");
  gtk_toolbar_set_style(XEN_TO_C_GtkToolbar_(toolbar), XEN_TO_C_GtkToolbarStyle(style));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_toolbar_set_icon_size(XEN toolbar, XEN icon_size)
{
  #define H_gtk_toolbar_set_icon_size "void gtk_toolbar_set_icon_size(GtkToolbar* toolbar, GtkIconSize icon_size)"
  XEN_ASSERT_TYPE(XEN_GtkToolbar__P(toolbar), toolbar, 1, "gtk_toolbar_set_icon_size", "GtkToolbar*");
  XEN_ASSERT_TYPE(XEN_GtkIconSize_P(icon_size), icon_size, 2, "gtk_toolbar_set_icon_size", "GtkIconSize");
  gtk_toolbar_set_icon_size(XEN_TO_C_GtkToolbar_(toolbar), XEN_TO_C_GtkIconSize(icon_size));
  return(XEN_FALSE);
}
#endif

static XEN gxg_gtk_toolbar_set_tooltips(XEN toolbar, XEN enable)
{
  #define H_gtk_toolbar_set_tooltips "void gtk_toolbar_set_tooltips(GtkToolbar* toolbar, gboolean enable)"
  XEN_ASSERT_TYPE(XEN_GtkToolbar__P(toolbar), toolbar, 1, "gtk_toolbar_set_tooltips", "GtkToolbar*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(enable), enable, 2, "gtk_toolbar_set_tooltips", "gboolean");
  gtk_toolbar_set_tooltips(XEN_TO_C_GtkToolbar_(toolbar), XEN_TO_C_gboolean(enable));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_toolbar_unset_style(XEN toolbar)
{
  #define H_gtk_toolbar_unset_style "void gtk_toolbar_unset_style(GtkToolbar* toolbar)"
  XEN_ASSERT_TYPE(XEN_GtkToolbar__P(toolbar), toolbar, 1, "gtk_toolbar_unset_style", "GtkToolbar*");
  gtk_toolbar_unset_style(XEN_TO_C_GtkToolbar_(toolbar));
  return(XEN_FALSE);
}

static XEN gxg_gtk_toolbar_unset_icon_size(XEN toolbar)
{
  #define H_gtk_toolbar_unset_icon_size "void gtk_toolbar_unset_icon_size(GtkToolbar* toolbar)"
  XEN_ASSERT_TYPE(XEN_GtkToolbar__P(toolbar), toolbar, 1, "gtk_toolbar_unset_icon_size", "GtkToolbar*");
  gtk_toolbar_unset_icon_size(XEN_TO_C_GtkToolbar_(toolbar));
  return(XEN_FALSE);
}

static XEN gxg_gtk_toolbar_get_orientation(XEN toolbar)
{
  #define H_gtk_toolbar_get_orientation "GtkOrientation gtk_toolbar_get_orientation(GtkToolbar* toolbar)"
  XEN_ASSERT_TYPE(XEN_GtkToolbar__P(toolbar), toolbar, 1, "gtk_toolbar_get_orientation", "GtkToolbar*");
  return(C_TO_XEN_GtkOrientation(gtk_toolbar_get_orientation(XEN_TO_C_GtkToolbar_(toolbar))));
}

static XEN gxg_gtk_toolbar_get_style(XEN toolbar)
{
  #define H_gtk_toolbar_get_style "GtkToolbarStyle gtk_toolbar_get_style(GtkToolbar* toolbar)"
  XEN_ASSERT_TYPE(XEN_GtkToolbar__P(toolbar), toolbar, 1, "gtk_toolbar_get_style", "GtkToolbar*");
  return(C_TO_XEN_GtkToolbarStyle(gtk_toolbar_get_style(XEN_TO_C_GtkToolbar_(toolbar))));
}

static XEN gxg_gtk_toolbar_get_icon_size(XEN toolbar)
{
  #define H_gtk_toolbar_get_icon_size "GtkIconSize gtk_toolbar_get_icon_size(GtkToolbar* toolbar)"
  XEN_ASSERT_TYPE(XEN_GtkToolbar__P(toolbar), toolbar, 1, "gtk_toolbar_get_icon_size", "GtkToolbar*");
  return(C_TO_XEN_GtkIconSize(gtk_toolbar_get_icon_size(XEN_TO_C_GtkToolbar_(toolbar))));
}

static XEN gxg_gtk_toolbar_get_tooltips(XEN toolbar)
{
  #define H_gtk_toolbar_get_tooltips "gboolean gtk_toolbar_get_tooltips(GtkToolbar* toolbar)"
  XEN_ASSERT_TYPE(XEN_GtkToolbar__P(toolbar), toolbar, 1, "gtk_toolbar_get_tooltips", "GtkToolbar*");
  return(C_TO_XEN_gboolean(gtk_toolbar_get_tooltips(XEN_TO_C_GtkToolbar_(toolbar))));
}
#endif

static XEN gxg_gtk_tooltips_get_type(void)
{
  #define H_gtk_tooltips_get_type "GtkType gtk_tooltips_get_type( void)"
  return(C_TO_XEN_GtkType(gtk_tooltips_get_type()));
}

static XEN gxg_gtk_tooltips_new(void)
{
  #define H_gtk_tooltips_new "GtkTooltips* gtk_tooltips_new( void)"
  return(C_TO_XEN_GtkTooltips_(gtk_tooltips_new()));
}

static XEN gxg_gtk_tooltips_enable(XEN tooltips)
{
  #define H_gtk_tooltips_enable "void gtk_tooltips_enable(GtkTooltips* tooltips)"
  XEN_ASSERT_TYPE(XEN_GtkTooltips__P(tooltips), tooltips, 1, "gtk_tooltips_enable", "GtkTooltips*");
  gtk_tooltips_enable(XEN_TO_C_GtkTooltips_(tooltips));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tooltips_disable(XEN tooltips)
{
  #define H_gtk_tooltips_disable "void gtk_tooltips_disable(GtkTooltips* tooltips)"
  XEN_ASSERT_TYPE(XEN_GtkTooltips__P(tooltips), tooltips, 1, "gtk_tooltips_disable", "GtkTooltips*");
  gtk_tooltips_disable(XEN_TO_C_GtkTooltips_(tooltips));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tooltips_set_tip(XEN tooltips, XEN widget, XEN tip_text, XEN tip_private)
{
  #define H_gtk_tooltips_set_tip "void gtk_tooltips_set_tip(GtkTooltips* tooltips, GtkWidget* widget, \
gchar* tip_text, gchar* tip_private)"
  XEN_ASSERT_TYPE(XEN_GtkTooltips__P(tooltips), tooltips, 1, "gtk_tooltips_set_tip", "GtkTooltips*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 2, "gtk_tooltips_set_tip", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_gchar__P(tip_text), tip_text, 3, "gtk_tooltips_set_tip", "gchar*");
  XEN_ASSERT_TYPE(XEN_gchar__P(tip_private), tip_private, 4, "gtk_tooltips_set_tip", "gchar*");
  gtk_tooltips_set_tip(XEN_TO_C_GtkTooltips_(tooltips), XEN_TO_C_GtkWidget_(widget), XEN_TO_C_gchar_(tip_text), XEN_TO_C_gchar_(tip_private));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tooltips_data_get(XEN widget)
{
  #define H_gtk_tooltips_data_get "GtkTooltipsData* gtk_tooltips_data_get(GtkWidget* widget)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_tooltips_data_get", "GtkWidget*");
  return(C_TO_XEN_GtkTooltipsData_(gtk_tooltips_data_get(XEN_TO_C_GtkWidget_(widget))));
}

static XEN gxg_gtk_tooltips_force_window(XEN tooltips)
{
  #define H_gtk_tooltips_force_window "void gtk_tooltips_force_window(GtkTooltips* tooltips)"
  XEN_ASSERT_TYPE(XEN_GtkTooltips__P(tooltips), tooltips, 1, "gtk_tooltips_force_window", "GtkTooltips*");
  gtk_tooltips_force_window(XEN_TO_C_GtkTooltips_(tooltips));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_tree_drag_source_get_type(void)
{
  #define H_gtk_tree_drag_source_get_type "GType gtk_tree_drag_source_get_type( void)"
  return(C_TO_XEN_GType(gtk_tree_drag_source_get_type()));
}

static XEN gxg_gtk_tree_drag_source_row_draggable(XEN drag_source, XEN path)
{
  #define H_gtk_tree_drag_source_row_draggable "gboolean gtk_tree_drag_source_row_draggable(GtkTreeDragSource* drag_source, \
GtkTreePath* path)"
  XEN_ASSERT_TYPE(XEN_GtkTreeDragSource__P(drag_source), drag_source, 1, "gtk_tree_drag_source_row_draggable", "GtkTreeDragSource*");
  XEN_ASSERT_TYPE(XEN_GtkTreePath__P(path), path, 2, "gtk_tree_drag_source_row_draggable", "GtkTreePath*");
  return(C_TO_XEN_gboolean(gtk_tree_drag_source_row_draggable(XEN_TO_C_GtkTreeDragSource_(drag_source), XEN_TO_C_GtkTreePath_(path))));
}

static XEN gxg_gtk_tree_drag_source_drag_data_delete(XEN drag_source, XEN path)
{
  #define H_gtk_tree_drag_source_drag_data_delete "gboolean gtk_tree_drag_source_drag_data_delete(GtkTreeDragSource* drag_source, \
GtkTreePath* path)"
  XEN_ASSERT_TYPE(XEN_GtkTreeDragSource__P(drag_source), drag_source, 1, "gtk_tree_drag_source_drag_data_delete", "GtkTreeDragSource*");
  XEN_ASSERT_TYPE(XEN_GtkTreePath__P(path), path, 2, "gtk_tree_drag_source_drag_data_delete", "GtkTreePath*");
  return(C_TO_XEN_gboolean(gtk_tree_drag_source_drag_data_delete(XEN_TO_C_GtkTreeDragSource_(drag_source), XEN_TO_C_GtkTreePath_(path))));
}

static XEN gxg_gtk_tree_drag_source_drag_data_get(XEN drag_source, XEN path, XEN selection_data)
{
  #define H_gtk_tree_drag_source_drag_data_get "gboolean gtk_tree_drag_source_drag_data_get(GtkTreeDragSource* drag_source, \
GtkTreePath* path, GtkSelectionData* selection_data)"
  XEN_ASSERT_TYPE(XEN_GtkTreeDragSource__P(drag_source), drag_source, 1, "gtk_tree_drag_source_drag_data_get", "GtkTreeDragSource*");
  XEN_ASSERT_TYPE(XEN_GtkTreePath__P(path), path, 2, "gtk_tree_drag_source_drag_data_get", "GtkTreePath*");
  XEN_ASSERT_TYPE(XEN_GtkSelectionData__P(selection_data), selection_data, 3, "gtk_tree_drag_source_drag_data_get", "GtkSelectionData*");
  return(C_TO_XEN_gboolean(gtk_tree_drag_source_drag_data_get(XEN_TO_C_GtkTreeDragSource_(drag_source), XEN_TO_C_GtkTreePath_(path), 
                                                              XEN_TO_C_GtkSelectionData_(selection_data))));
}

static XEN gxg_gtk_tree_drag_dest_get_type(void)
{
  #define H_gtk_tree_drag_dest_get_type "GType gtk_tree_drag_dest_get_type( void)"
  return(C_TO_XEN_GType(gtk_tree_drag_dest_get_type()));
}

static XEN gxg_gtk_tree_drag_dest_drag_data_received(XEN drag_dest, XEN dest, XEN selection_data)
{
  #define H_gtk_tree_drag_dest_drag_data_received "gboolean gtk_tree_drag_dest_drag_data_received(GtkTreeDragDest* drag_dest, \
GtkTreePath* dest, GtkSelectionData* selection_data)"
  XEN_ASSERT_TYPE(XEN_GtkTreeDragDest__P(drag_dest), drag_dest, 1, "gtk_tree_drag_dest_drag_data_received", "GtkTreeDragDest*");
  XEN_ASSERT_TYPE(XEN_GtkTreePath__P(dest), dest, 2, "gtk_tree_drag_dest_drag_data_received", "GtkTreePath*");
  XEN_ASSERT_TYPE(XEN_GtkSelectionData__P(selection_data), selection_data, 3, "gtk_tree_drag_dest_drag_data_received", "GtkSelectionData*");
  return(C_TO_XEN_gboolean(gtk_tree_drag_dest_drag_data_received(XEN_TO_C_GtkTreeDragDest_(drag_dest), XEN_TO_C_GtkTreePath_(dest), 
                                                                 XEN_TO_C_GtkSelectionData_(selection_data))));
}

static XEN gxg_gtk_tree_drag_dest_row_drop_possible(XEN drag_dest, XEN dest_path, XEN selection_data)
{
  #define H_gtk_tree_drag_dest_row_drop_possible "gboolean gtk_tree_drag_dest_row_drop_possible(GtkTreeDragDest* drag_dest, \
GtkTreePath* dest_path, GtkSelectionData* selection_data)"
  XEN_ASSERT_TYPE(XEN_GtkTreeDragDest__P(drag_dest), drag_dest, 1, "gtk_tree_drag_dest_row_drop_possible", "GtkTreeDragDest*");
  XEN_ASSERT_TYPE(XEN_GtkTreePath__P(dest_path), dest_path, 2, "gtk_tree_drag_dest_row_drop_possible", "GtkTreePath*");
  XEN_ASSERT_TYPE(XEN_GtkSelectionData__P(selection_data), selection_data, 3, "gtk_tree_drag_dest_row_drop_possible", "GtkSelectionData*");
  return(C_TO_XEN_gboolean(gtk_tree_drag_dest_row_drop_possible(XEN_TO_C_GtkTreeDragDest_(drag_dest), XEN_TO_C_GtkTreePath_(dest_path), 
                                                                XEN_TO_C_GtkSelectionData_(selection_data))));
}

static XEN gxg_gtk_tree_set_row_drag_data(XEN selection_data, XEN tree_model, XEN path)
{
  #define H_gtk_tree_set_row_drag_data "gboolean gtk_tree_set_row_drag_data(GtkSelectionData* selection_data, \
GtkTreeModel* tree_model, GtkTreePath* path)"
  XEN_ASSERT_TYPE(XEN_GtkSelectionData__P(selection_data), selection_data, 1, "gtk_tree_set_row_drag_data", "GtkSelectionData*");
  XEN_ASSERT_TYPE(XEN_GtkTreeModel__P(tree_model), tree_model, 2, "gtk_tree_set_row_drag_data", "GtkTreeModel*");
  XEN_ASSERT_TYPE(XEN_GtkTreePath__P(path), path, 3, "gtk_tree_set_row_drag_data", "GtkTreePath*");
  return(C_TO_XEN_gboolean(gtk_tree_set_row_drag_data(XEN_TO_C_GtkSelectionData_(selection_data), XEN_TO_C_GtkTreeModel_(tree_model), 
                                                      XEN_TO_C_GtkTreePath_(path))));
}

static XEN gxg_gtk_tree_get_row_drag_data(XEN selection_data, XEN tree_model, XEN path)
{
  #define H_gtk_tree_get_row_drag_data "gboolean gtk_tree_get_row_drag_data(GtkSelectionData* selection_data, \
GtkTreeModel** [tree_model], GtkTreePath** [path])"
  GtkTreeModel* ref_tree_model;
  GtkTreePath* ref_path;
  XEN_ASSERT_TYPE(XEN_GtkSelectionData__P(selection_data), selection_data, 1, "gtk_tree_get_row_drag_data", "GtkSelectionData*");
  {
    XEN result = XEN_FALSE;
    result = C_TO_XEN_gboolean(gtk_tree_get_row_drag_data(XEN_TO_C_GtkSelectionData_(selection_data), &ref_tree_model, &ref_path));
    return(XEN_LIST_3(result, C_TO_XEN_GtkTreeModel_(ref_tree_model), C_TO_XEN_GtkTreePath_(ref_path)));
   }
}

static XEN gxg_gtk_tree_path_new(void)
{
  #define H_gtk_tree_path_new "GtkTreePath* gtk_tree_path_new( void)"
  return(C_TO_XEN_GtkTreePath_(gtk_tree_path_new()));
}

static XEN gxg_gtk_tree_path_new_from_string(XEN path)
{
  #define H_gtk_tree_path_new_from_string "GtkTreePath* gtk_tree_path_new_from_string(gchar* path)"
  XEN_ASSERT_TYPE(XEN_gchar__P(path), path, 1, "gtk_tree_path_new_from_string", "gchar*");
  return(C_TO_XEN_GtkTreePath_(gtk_tree_path_new_from_string(XEN_TO_C_gchar_(path))));
}

static XEN gxg_gtk_tree_path_to_string(XEN path)
{
  #define H_gtk_tree_path_to_string "gchar* gtk_tree_path_to_string(GtkTreePath* path)"
  XEN_ASSERT_TYPE(XEN_GtkTreePath__P(path), path, 1, "gtk_tree_path_to_string", "GtkTreePath*");
  return(C_TO_XEN_gchar_(gtk_tree_path_to_string(XEN_TO_C_GtkTreePath_(path))));
}

static XEN gxg_gtk_tree_path_new_root(void)
{
  #define H_gtk_tree_path_new_root "GtkTreePath* gtk_tree_path_new_root( void)"
  return(C_TO_XEN_GtkTreePath_(gtk_tree_path_new_root()));
}

#if HAVE_GTK_FILE_SELECTION_GET_SELECTIONS
static XEN gxg_gtk_tree_path_new_first(void)
{
  #define H_gtk_tree_path_new_first "GtkTreePath* gtk_tree_path_new_first( void)"
  return(C_TO_XEN_GtkTreePath_(gtk_tree_path_new_first()));
}
#endif

static XEN gxg_gtk_tree_path_append_index(XEN path, XEN index)
{
  #define H_gtk_tree_path_append_index "void gtk_tree_path_append_index(GtkTreePath* path, gint index)"
  XEN_ASSERT_TYPE(XEN_GtkTreePath__P(path), path, 1, "gtk_tree_path_append_index", "GtkTreePath*");
  XEN_ASSERT_TYPE(XEN_gint_P(index), index, 2, "gtk_tree_path_append_index", "gint");
  gtk_tree_path_append_index(XEN_TO_C_GtkTreePath_(path), XEN_TO_C_gint(index));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tree_path_prepend_index(XEN path, XEN index)
{
  #define H_gtk_tree_path_prepend_index "void gtk_tree_path_prepend_index(GtkTreePath* path, gint index)"
  XEN_ASSERT_TYPE(XEN_GtkTreePath__P(path), path, 1, "gtk_tree_path_prepend_index", "GtkTreePath*");
  XEN_ASSERT_TYPE(XEN_gint_P(index), index, 2, "gtk_tree_path_prepend_index", "gint");
  gtk_tree_path_prepend_index(XEN_TO_C_GtkTreePath_(path), XEN_TO_C_gint(index));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tree_path_get_depth(XEN path)
{
  #define H_gtk_tree_path_get_depth "gint gtk_tree_path_get_depth(GtkTreePath* path)"
  XEN_ASSERT_TYPE(XEN_GtkTreePath__P(path), path, 1, "gtk_tree_path_get_depth", "GtkTreePath*");
  return(C_TO_XEN_gint(gtk_tree_path_get_depth(XEN_TO_C_GtkTreePath_(path))));
}

static XEN gxg_gtk_tree_path_get_indices(XEN path)
{
  #define H_gtk_tree_path_get_indices "gint* gtk_tree_path_get_indices(GtkTreePath* path)"
  XEN_ASSERT_TYPE(XEN_GtkTreePath__P(path), path, 1, "gtk_tree_path_get_indices", "GtkTreePath*");
  return(C_TO_XEN_gint_(gtk_tree_path_get_indices(XEN_TO_C_GtkTreePath_(path))));
}

static XEN gxg_gtk_tree_path_free(XEN path)
{
  #define H_gtk_tree_path_free "void gtk_tree_path_free(GtkTreePath* path)"
  XEN_ASSERT_TYPE(XEN_GtkTreePath__P(path), path, 1, "gtk_tree_path_free", "GtkTreePath*");
  gtk_tree_path_free(XEN_TO_C_GtkTreePath_(path));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tree_path_copy(XEN path)
{
  #define H_gtk_tree_path_copy "GtkTreePath* gtk_tree_path_copy(GtkTreePath* path)"
  XEN_ASSERT_TYPE(XEN_GtkTreePath__P(path), path, 1, "gtk_tree_path_copy", "GtkTreePath*");
  return(C_TO_XEN_GtkTreePath_(gtk_tree_path_copy(XEN_TO_C_GtkTreePath_(path))));
}

static XEN gxg_gtk_tree_path_compare(XEN a, XEN b)
{
  #define H_gtk_tree_path_compare "gint gtk_tree_path_compare(GtkTreePath* a, GtkTreePath* b)"
  XEN_ASSERT_TYPE(XEN_GtkTreePath__P(a), a, 1, "gtk_tree_path_compare", "GtkTreePath*");
  XEN_ASSERT_TYPE(XEN_GtkTreePath__P(b), b, 2, "gtk_tree_path_compare", "GtkTreePath*");
  return(C_TO_XEN_gint(gtk_tree_path_compare(XEN_TO_C_GtkTreePath_(a), XEN_TO_C_GtkTreePath_(b))));
}

static XEN gxg_gtk_tree_path_next(XEN path)
{
  #define H_gtk_tree_path_next "void gtk_tree_path_next(GtkTreePath* path)"
  XEN_ASSERT_TYPE(XEN_GtkTreePath__P(path), path, 1, "gtk_tree_path_next", "GtkTreePath*");
  gtk_tree_path_next(XEN_TO_C_GtkTreePath_(path));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tree_path_prev(XEN path)
{
  #define H_gtk_tree_path_prev "gboolean gtk_tree_path_prev(GtkTreePath* path)"
  XEN_ASSERT_TYPE(XEN_GtkTreePath__P(path), path, 1, "gtk_tree_path_prev", "GtkTreePath*");
  return(C_TO_XEN_gboolean(gtk_tree_path_prev(XEN_TO_C_GtkTreePath_(path))));
}

static XEN gxg_gtk_tree_path_up(XEN path)
{
  #define H_gtk_tree_path_up "gboolean gtk_tree_path_up(GtkTreePath* path)"
  XEN_ASSERT_TYPE(XEN_GtkTreePath__P(path), path, 1, "gtk_tree_path_up", "GtkTreePath*");
  return(C_TO_XEN_gboolean(gtk_tree_path_up(XEN_TO_C_GtkTreePath_(path))));
}

static XEN gxg_gtk_tree_path_down(XEN path)
{
  #define H_gtk_tree_path_down "void gtk_tree_path_down(GtkTreePath* path)"
  XEN_ASSERT_TYPE(XEN_GtkTreePath__P(path), path, 1, "gtk_tree_path_down", "GtkTreePath*");
  gtk_tree_path_down(XEN_TO_C_GtkTreePath_(path));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tree_path_is_ancestor(XEN path, XEN descendant)
{
  #define H_gtk_tree_path_is_ancestor "gboolean gtk_tree_path_is_ancestor(GtkTreePath* path, GtkTreePath* descendant)"
  XEN_ASSERT_TYPE(XEN_GtkTreePath__P(path), path, 1, "gtk_tree_path_is_ancestor", "GtkTreePath*");
  XEN_ASSERT_TYPE(XEN_GtkTreePath__P(descendant), descendant, 2, "gtk_tree_path_is_ancestor", "GtkTreePath*");
  return(C_TO_XEN_gboolean(gtk_tree_path_is_ancestor(XEN_TO_C_GtkTreePath_(path), XEN_TO_C_GtkTreePath_(descendant))));
}

static XEN gxg_gtk_tree_path_is_descendant(XEN path, XEN ancestor)
{
  #define H_gtk_tree_path_is_descendant "gboolean gtk_tree_path_is_descendant(GtkTreePath* path, GtkTreePath* ancestor)"
  XEN_ASSERT_TYPE(XEN_GtkTreePath__P(path), path, 1, "gtk_tree_path_is_descendant", "GtkTreePath*");
  XEN_ASSERT_TYPE(XEN_GtkTreePath__P(ancestor), ancestor, 2, "gtk_tree_path_is_descendant", "GtkTreePath*");
  return(C_TO_XEN_gboolean(gtk_tree_path_is_descendant(XEN_TO_C_GtkTreePath_(path), XEN_TO_C_GtkTreePath_(ancestor))));
}

static XEN gxg_gtk_tree_row_reference_new(XEN model, XEN path)
{
  #define H_gtk_tree_row_reference_new "GtkTreeRowReference* gtk_tree_row_reference_new(GtkTreeModel* model, \
GtkTreePath* path)"
  XEN_ASSERT_TYPE(XEN_GtkTreeModel__P(model), model, 1, "gtk_tree_row_reference_new", "GtkTreeModel*");
  XEN_ASSERT_TYPE(XEN_GtkTreePath__P(path), path, 2, "gtk_tree_row_reference_new", "GtkTreePath*");
  return(C_TO_XEN_GtkTreeRowReference_(gtk_tree_row_reference_new(XEN_TO_C_GtkTreeModel_(model), XEN_TO_C_GtkTreePath_(path))));
}

static XEN gxg_gtk_tree_row_reference_new_proxy(XEN proxy, XEN model, XEN path)
{
  #define H_gtk_tree_row_reference_new_proxy "GtkTreeRowReference* gtk_tree_row_reference_new_proxy(GObject* proxy, \
GtkTreeModel* model, GtkTreePath* path)"
  XEN_ASSERT_TYPE(XEN_GObject__P(proxy), proxy, 1, "gtk_tree_row_reference_new_proxy", "GObject*");
  XEN_ASSERT_TYPE(XEN_GtkTreeModel__P(model), model, 2, "gtk_tree_row_reference_new_proxy", "GtkTreeModel*");
  XEN_ASSERT_TYPE(XEN_GtkTreePath__P(path), path, 3, "gtk_tree_row_reference_new_proxy", "GtkTreePath*");
  return(C_TO_XEN_GtkTreeRowReference_(gtk_tree_row_reference_new_proxy(XEN_TO_C_GObject_(proxy), XEN_TO_C_GtkTreeModel_(model), 
                                                                        XEN_TO_C_GtkTreePath_(path))));
}

static XEN gxg_gtk_tree_row_reference_get_path(XEN reference)
{
  #define H_gtk_tree_row_reference_get_path "GtkTreePath* gtk_tree_row_reference_get_path(GtkTreeRowReference* reference)"
  XEN_ASSERT_TYPE(XEN_GtkTreeRowReference__P(reference), reference, 1, "gtk_tree_row_reference_get_path", "GtkTreeRowReference*");
  return(C_TO_XEN_GtkTreePath_(gtk_tree_row_reference_get_path(XEN_TO_C_GtkTreeRowReference_(reference))));
}

static XEN gxg_gtk_tree_row_reference_valid(XEN reference)
{
  #define H_gtk_tree_row_reference_valid "gboolean gtk_tree_row_reference_valid(GtkTreeRowReference* reference)"
  XEN_ASSERT_TYPE(XEN_GtkTreeRowReference__P(reference), reference, 1, "gtk_tree_row_reference_valid", "GtkTreeRowReference*");
  return(C_TO_XEN_gboolean(gtk_tree_row_reference_valid(XEN_TO_C_GtkTreeRowReference_(reference))));
}

static XEN gxg_gtk_tree_row_reference_free(XEN reference)
{
  #define H_gtk_tree_row_reference_free "void gtk_tree_row_reference_free(GtkTreeRowReference* reference)"
  XEN_ASSERT_TYPE(XEN_GtkTreeRowReference__P(reference), reference, 1, "gtk_tree_row_reference_free", "GtkTreeRowReference*");
  gtk_tree_row_reference_free(XEN_TO_C_GtkTreeRowReference_(reference));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tree_row_reference_inserted(XEN proxy, XEN path)
{
  #define H_gtk_tree_row_reference_inserted "void gtk_tree_row_reference_inserted(GObject* proxy, GtkTreePath* path)"
  XEN_ASSERT_TYPE(XEN_GObject__P(proxy), proxy, 1, "gtk_tree_row_reference_inserted", "GObject*");
  XEN_ASSERT_TYPE(XEN_GtkTreePath__P(path), path, 2, "gtk_tree_row_reference_inserted", "GtkTreePath*");
  gtk_tree_row_reference_inserted(XEN_TO_C_GObject_(proxy), XEN_TO_C_GtkTreePath_(path));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tree_row_reference_deleted(XEN proxy, XEN path)
{
  #define H_gtk_tree_row_reference_deleted "void gtk_tree_row_reference_deleted(GObject* proxy, GtkTreePath* path)"
  XEN_ASSERT_TYPE(XEN_GObject__P(proxy), proxy, 1, "gtk_tree_row_reference_deleted", "GObject*");
  XEN_ASSERT_TYPE(XEN_GtkTreePath__P(path), path, 2, "gtk_tree_row_reference_deleted", "GtkTreePath*");
  gtk_tree_row_reference_deleted(XEN_TO_C_GObject_(proxy), XEN_TO_C_GtkTreePath_(path));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tree_row_reference_reordered(XEN proxy, XEN path, XEN iter, XEN new_order)
{
  #define H_gtk_tree_row_reference_reordered "void gtk_tree_row_reference_reordered(GObject* proxy, GtkTreePath* path, \
GtkTreeIter* iter, gint* new_order)"
  XEN_ASSERT_TYPE(XEN_GObject__P(proxy), proxy, 1, "gtk_tree_row_reference_reordered", "GObject*");
  XEN_ASSERT_TYPE(XEN_GtkTreePath__P(path), path, 2, "gtk_tree_row_reference_reordered", "GtkTreePath*");
  XEN_ASSERT_TYPE(XEN_GtkTreeIter__P(iter), iter, 3, "gtk_tree_row_reference_reordered", "GtkTreeIter*");
  XEN_ASSERT_TYPE(XEN_gint__P(new_order), new_order, 4, "gtk_tree_row_reference_reordered", "gint*");
  gtk_tree_row_reference_reordered(XEN_TO_C_GObject_(proxy), XEN_TO_C_GtkTreePath_(path), XEN_TO_C_GtkTreeIter_(iter), XEN_TO_C_gint_(new_order));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tree_iter_copy(XEN iter)
{
  #define H_gtk_tree_iter_copy "GtkTreeIter* gtk_tree_iter_copy(GtkTreeIter* iter)"
  XEN_ASSERT_TYPE(XEN_GtkTreeIter__P(iter), iter, 1, "gtk_tree_iter_copy", "GtkTreeIter*");
  return(C_TO_XEN_GtkTreeIter_(gtk_tree_iter_copy(XEN_TO_C_GtkTreeIter_(iter))));
}

static XEN gxg_gtk_tree_iter_free(XEN iter)
{
  #define H_gtk_tree_iter_free "void gtk_tree_iter_free(GtkTreeIter* iter)"
  XEN_ASSERT_TYPE(XEN_GtkTreeIter__P(iter), iter, 1, "gtk_tree_iter_free", "GtkTreeIter*");
  gtk_tree_iter_free(XEN_TO_C_GtkTreeIter_(iter));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tree_model_get_type(void)
{
  #define H_gtk_tree_model_get_type "GtkType gtk_tree_model_get_type( void)"
  return(C_TO_XEN_GtkType(gtk_tree_model_get_type()));
}

static XEN gxg_gtk_tree_model_get_flags(XEN tree_model)
{
  #define H_gtk_tree_model_get_flags "GtkTreeModelFlags gtk_tree_model_get_flags(GtkTreeModel* tree_model)"
  XEN_ASSERT_TYPE(XEN_GtkTreeModel__P(tree_model), tree_model, 1, "gtk_tree_model_get_flags", "GtkTreeModel*");
  return(C_TO_XEN_GtkTreeModelFlags(gtk_tree_model_get_flags(XEN_TO_C_GtkTreeModel_(tree_model))));
}

static XEN gxg_gtk_tree_model_get_n_columns(XEN tree_model)
{
  #define H_gtk_tree_model_get_n_columns "gint gtk_tree_model_get_n_columns(GtkTreeModel* tree_model)"
  XEN_ASSERT_TYPE(XEN_GtkTreeModel__P(tree_model), tree_model, 1, "gtk_tree_model_get_n_columns", "GtkTreeModel*");
  return(C_TO_XEN_gint(gtk_tree_model_get_n_columns(XEN_TO_C_GtkTreeModel_(tree_model))));
}

static XEN gxg_gtk_tree_model_get_column_type(XEN tree_model, XEN index)
{
  #define H_gtk_tree_model_get_column_type "GType gtk_tree_model_get_column_type(GtkTreeModel* tree_model, \
gint index)"
  XEN_ASSERT_TYPE(XEN_GtkTreeModel__P(tree_model), tree_model, 1, "gtk_tree_model_get_column_type", "GtkTreeModel*");
  XEN_ASSERT_TYPE(XEN_gint_P(index), index, 2, "gtk_tree_model_get_column_type", "gint");
  return(C_TO_XEN_GType(gtk_tree_model_get_column_type(XEN_TO_C_GtkTreeModel_(tree_model), XEN_TO_C_gint(index))));
}

static XEN gxg_gtk_tree_model_get_iter(XEN tree_model, XEN iter, XEN path)
{
  #define H_gtk_tree_model_get_iter "gboolean gtk_tree_model_get_iter(GtkTreeModel* tree_model, GtkTreeIter* iter, \
GtkTreePath* path)"
  XEN_ASSERT_TYPE(XEN_GtkTreeModel__P(tree_model), tree_model, 1, "gtk_tree_model_get_iter", "GtkTreeModel*");
  XEN_ASSERT_TYPE(XEN_GtkTreeIter__P(iter), iter, 2, "gtk_tree_model_get_iter", "GtkTreeIter*");
  XEN_ASSERT_TYPE(XEN_GtkTreePath__P(path), path, 3, "gtk_tree_model_get_iter", "GtkTreePath*");
  return(C_TO_XEN_gboolean(gtk_tree_model_get_iter(XEN_TO_C_GtkTreeModel_(tree_model), XEN_TO_C_GtkTreeIter_(iter), XEN_TO_C_GtkTreePath_(path))));
}

static XEN gxg_gtk_tree_model_get_iter_from_string(XEN tree_model, XEN iter, XEN path_string)
{
  #define H_gtk_tree_model_get_iter_from_string "gboolean gtk_tree_model_get_iter_from_string(GtkTreeModel* tree_model, \
GtkTreeIter* iter, gchar* path_string)"
  XEN_ASSERT_TYPE(XEN_GtkTreeModel__P(tree_model), tree_model, 1, "gtk_tree_model_get_iter_from_string", "GtkTreeModel*");
  XEN_ASSERT_TYPE(XEN_GtkTreeIter__P(iter), iter, 2, "gtk_tree_model_get_iter_from_string", "GtkTreeIter*");
  XEN_ASSERT_TYPE(XEN_gchar__P(path_string), path_string, 3, "gtk_tree_model_get_iter_from_string", "gchar*");
  return(C_TO_XEN_gboolean(gtk_tree_model_get_iter_from_string(XEN_TO_C_GtkTreeModel_(tree_model), XEN_TO_C_GtkTreeIter_(iter), 
                                                               XEN_TO_C_gchar_(path_string))));
}

static XEN gxg_gtk_tree_model_get_iter_root(XEN tree_model, XEN iter)
{
  #define H_gtk_tree_model_get_iter_root "gboolean gtk_tree_model_get_iter_root(GtkTreeModel* tree_model, \
GtkTreeIter* iter)"
  XEN_ASSERT_TYPE(XEN_GtkTreeModel__P(tree_model), tree_model, 1, "gtk_tree_model_get_iter_root", "GtkTreeModel*");
  XEN_ASSERT_TYPE(XEN_GtkTreeIter__P(iter), iter, 2, "gtk_tree_model_get_iter_root", "GtkTreeIter*");
  return(C_TO_XEN_gboolean(gtk_tree_model_get_iter_root(XEN_TO_C_GtkTreeModel_(tree_model), XEN_TO_C_GtkTreeIter_(iter))));
}

#if HAVE_GTK_FILE_SELECTION_GET_SELECTIONS
static XEN gxg_gtk_tree_model_get_iter_first(XEN tree_model, XEN iter)
{
  #define H_gtk_tree_model_get_iter_first "gboolean gtk_tree_model_get_iter_first(GtkTreeModel* tree_model, \
GtkTreeIter* iter)"
  XEN_ASSERT_TYPE(XEN_GtkTreeModel__P(tree_model), tree_model, 1, "gtk_tree_model_get_iter_first", "GtkTreeModel*");
  XEN_ASSERT_TYPE(XEN_GtkTreeIter__P(iter), iter, 2, "gtk_tree_model_get_iter_first", "GtkTreeIter*");
  return(C_TO_XEN_gboolean(gtk_tree_model_get_iter_first(XEN_TO_C_GtkTreeModel_(tree_model), XEN_TO_C_GtkTreeIter_(iter))));
}
#endif

static XEN gxg_gtk_tree_model_get_path(XEN tree_model, XEN iter)
{
  #define H_gtk_tree_model_get_path "GtkTreePath* gtk_tree_model_get_path(GtkTreeModel* tree_model, GtkTreeIter* iter)"
  XEN_ASSERT_TYPE(XEN_GtkTreeModel__P(tree_model), tree_model, 1, "gtk_tree_model_get_path", "GtkTreeModel*");
  XEN_ASSERT_TYPE(XEN_GtkTreeIter__P(iter), iter, 2, "gtk_tree_model_get_path", "GtkTreeIter*");
  return(C_TO_XEN_GtkTreePath_(gtk_tree_model_get_path(XEN_TO_C_GtkTreeModel_(tree_model), XEN_TO_C_GtkTreeIter_(iter))));
}

static XEN gxg_gtk_tree_model_get_value(XEN tree_model, XEN iter, XEN column, XEN value)
{
  #define H_gtk_tree_model_get_value "void gtk_tree_model_get_value(GtkTreeModel* tree_model, GtkTreeIter* iter, \
gint column, GValue* value)"
  XEN_ASSERT_TYPE(XEN_GtkTreeModel__P(tree_model), tree_model, 1, "gtk_tree_model_get_value", "GtkTreeModel*");
  XEN_ASSERT_TYPE(XEN_GtkTreeIter__P(iter), iter, 2, "gtk_tree_model_get_value", "GtkTreeIter*");
  XEN_ASSERT_TYPE(XEN_gint_P(column), column, 3, "gtk_tree_model_get_value", "gint");
  XEN_ASSERT_TYPE(XEN_GValue__P(value), value, 4, "gtk_tree_model_get_value", "GValue*");
  gtk_tree_model_get_value(XEN_TO_C_GtkTreeModel_(tree_model), XEN_TO_C_GtkTreeIter_(iter), XEN_TO_C_gint(column), XEN_TO_C_GValue_(value));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tree_model_iter_next(XEN tree_model, XEN iter)
{
  #define H_gtk_tree_model_iter_next "gboolean gtk_tree_model_iter_next(GtkTreeModel* tree_model, GtkTreeIter* iter)"
  XEN_ASSERT_TYPE(XEN_GtkTreeModel__P(tree_model), tree_model, 1, "gtk_tree_model_iter_next", "GtkTreeModel*");
  XEN_ASSERT_TYPE(XEN_GtkTreeIter__P(iter), iter, 2, "gtk_tree_model_iter_next", "GtkTreeIter*");
  return(C_TO_XEN_gboolean(gtk_tree_model_iter_next(XEN_TO_C_GtkTreeModel_(tree_model), XEN_TO_C_GtkTreeIter_(iter))));
}

static XEN gxg_gtk_tree_model_iter_children(XEN tree_model, XEN iter, XEN parent)
{
  #define H_gtk_tree_model_iter_children "gboolean gtk_tree_model_iter_children(GtkTreeModel* tree_model, \
GtkTreeIter* iter, GtkTreeIter* parent)"
  XEN_ASSERT_TYPE(XEN_GtkTreeModel__P(tree_model), tree_model, 1, "gtk_tree_model_iter_children", "GtkTreeModel*");
  XEN_ASSERT_TYPE(XEN_GtkTreeIter__P(iter), iter, 2, "gtk_tree_model_iter_children", "GtkTreeIter*");
  XEN_ASSERT_TYPE(XEN_GtkTreeIter__P(parent), parent, 3, "gtk_tree_model_iter_children", "GtkTreeIter*");
  return(C_TO_XEN_gboolean(gtk_tree_model_iter_children(XEN_TO_C_GtkTreeModel_(tree_model), XEN_TO_C_GtkTreeIter_(iter), 
                                                        XEN_TO_C_GtkTreeIter_(parent))));
}

static XEN gxg_gtk_tree_model_iter_has_child(XEN tree_model, XEN iter)
{
  #define H_gtk_tree_model_iter_has_child "gboolean gtk_tree_model_iter_has_child(GtkTreeModel* tree_model, \
GtkTreeIter* iter)"
  XEN_ASSERT_TYPE(XEN_GtkTreeModel__P(tree_model), tree_model, 1, "gtk_tree_model_iter_has_child", "GtkTreeModel*");
  XEN_ASSERT_TYPE(XEN_GtkTreeIter__P(iter), iter, 2, "gtk_tree_model_iter_has_child", "GtkTreeIter*");
  return(C_TO_XEN_gboolean(gtk_tree_model_iter_has_child(XEN_TO_C_GtkTreeModel_(tree_model), XEN_TO_C_GtkTreeIter_(iter))));
}

static XEN gxg_gtk_tree_model_iter_n_children(XEN tree_model, XEN iter)
{
  #define H_gtk_tree_model_iter_n_children "gint gtk_tree_model_iter_n_children(GtkTreeModel* tree_model, \
GtkTreeIter* iter)"
  XEN_ASSERT_TYPE(XEN_GtkTreeModel__P(tree_model), tree_model, 1, "gtk_tree_model_iter_n_children", "GtkTreeModel*");
  XEN_ASSERT_TYPE(XEN_GtkTreeIter__P(iter) || XEN_FALSE_P(iter), iter, 2, "gtk_tree_model_iter_n_children", "GtkTreeIter*");
  return(C_TO_XEN_gint(gtk_tree_model_iter_n_children(XEN_TO_C_GtkTreeModel_(tree_model), XEN_TO_C_GtkTreeIter_(iter))));
}

static XEN gxg_gtk_tree_model_iter_nth_child(XEN tree_model, XEN iter, XEN parent, XEN n)
{
  #define H_gtk_tree_model_iter_nth_child "gboolean gtk_tree_model_iter_nth_child(GtkTreeModel* tree_model, \
GtkTreeIter* iter, GtkTreeIter* parent, gint n)"
  XEN_ASSERT_TYPE(XEN_GtkTreeModel__P(tree_model), tree_model, 1, "gtk_tree_model_iter_nth_child", "GtkTreeModel*");
  XEN_ASSERT_TYPE(XEN_GtkTreeIter__P(iter), iter, 2, "gtk_tree_model_iter_nth_child", "GtkTreeIter*");
  XEN_ASSERT_TYPE(XEN_GtkTreeIter__P(parent) || XEN_FALSE_P(parent), parent, 3, "gtk_tree_model_iter_nth_child", "GtkTreeIter*");
  XEN_ASSERT_TYPE(XEN_gint_P(n), n, 4, "gtk_tree_model_iter_nth_child", "gint");
  return(C_TO_XEN_gboolean(gtk_tree_model_iter_nth_child(XEN_TO_C_GtkTreeModel_(tree_model), XEN_TO_C_GtkTreeIter_(iter), 
                                                         XEN_TO_C_GtkTreeIter_(parent), XEN_TO_C_gint(n))));
}

static XEN gxg_gtk_tree_model_iter_parent(XEN tree_model, XEN iter, XEN child)
{
  #define H_gtk_tree_model_iter_parent "gboolean gtk_tree_model_iter_parent(GtkTreeModel* tree_model, \
GtkTreeIter* iter, GtkTreeIter* child)"
  XEN_ASSERT_TYPE(XEN_GtkTreeModel__P(tree_model), tree_model, 1, "gtk_tree_model_iter_parent", "GtkTreeModel*");
  XEN_ASSERT_TYPE(XEN_GtkTreeIter__P(iter), iter, 2, "gtk_tree_model_iter_parent", "GtkTreeIter*");
  XEN_ASSERT_TYPE(XEN_GtkTreeIter__P(child), child, 3, "gtk_tree_model_iter_parent", "GtkTreeIter*");
  return(C_TO_XEN_gboolean(gtk_tree_model_iter_parent(XEN_TO_C_GtkTreeModel_(tree_model), XEN_TO_C_GtkTreeIter_(iter), XEN_TO_C_GtkTreeIter_(child))));
}

static XEN gxg_gtk_tree_model_ref_node(XEN tree_model, XEN iter)
{
  #define H_gtk_tree_model_ref_node "void gtk_tree_model_ref_node(GtkTreeModel* tree_model, GtkTreeIter* iter)"
  XEN_ASSERT_TYPE(XEN_GtkTreeModel__P(tree_model), tree_model, 1, "gtk_tree_model_ref_node", "GtkTreeModel*");
  XEN_ASSERT_TYPE(XEN_GtkTreeIter__P(iter), iter, 2, "gtk_tree_model_ref_node", "GtkTreeIter*");
  gtk_tree_model_ref_node(XEN_TO_C_GtkTreeModel_(tree_model), XEN_TO_C_GtkTreeIter_(iter));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tree_model_unref_node(XEN tree_model, XEN iter)
{
  #define H_gtk_tree_model_unref_node "void gtk_tree_model_unref_node(GtkTreeModel* tree_model, GtkTreeIter* iter)"
  XEN_ASSERT_TYPE(XEN_GtkTreeModel__P(tree_model), tree_model, 1, "gtk_tree_model_unref_node", "GtkTreeModel*");
  XEN_ASSERT_TYPE(XEN_GtkTreeIter__P(iter), iter, 2, "gtk_tree_model_unref_node", "GtkTreeIter*");
  gtk_tree_model_unref_node(XEN_TO_C_GtkTreeModel_(tree_model), XEN_TO_C_GtkTreeIter_(iter));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tree_model_get(XEN tree_model, XEN iter)
{
  #define H_gtk_tree_model_get "void gtk_tree_model_get(GtkTreeModel* tree_model, GtkTreeIter* iter, \
...)"
  XEN_ASSERT_TYPE(XEN_GtkTreeModel__P(tree_model), tree_model, 1, "gtk_tree_model_get", "GtkTreeModel*");
  XEN_ASSERT_TYPE(XEN_GtkTreeIter__P(iter), iter, 2, "gtk_tree_model_get", "GtkTreeIter*");
  gtk_tree_model_get(XEN_TO_C_GtkTreeModel_(tree_model), XEN_TO_C_GtkTreeIter_(iter));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tree_model_foreach(XEN model, XEN func, XEN func_data)
{
  #define H_gtk_tree_model_foreach "void gtk_tree_model_foreach(GtkTreeModel* model, GtkTreeModelForeachFunc func, \
lambda_data func_data)"
  XEN_ASSERT_TYPE(XEN_GtkTreeModel__P(model), model, 1, "gtk_tree_model_foreach", "GtkTreeModel*");
  XEN_ASSERT_TYPE(XEN_GtkTreeModelForeachFunc_P(func), func, 2, "gtk_tree_model_foreach", "GtkTreeModelForeachFunc");
  if (XEN_NOT_BOUND_P(func_data)) func_data = XEN_FALSE; 
  else XEN_ASSERT_TYPE(XEN_lambda_data_P(func_data), func_data, 3, "gtk_tree_model_foreach", "lambda_data");
  {
    int loc;
    XEN gxg_ptr = XEN_LIST_5(func, func_data, XEN_FALSE, XEN_FALSE, XEN_FALSE);
    loc = xm_protect(gxg_ptr);
    XEN_LIST_SET(gxg_ptr, 2, C_TO_XEN_INT(loc));
    gtk_tree_model_foreach(XEN_TO_C_GtkTreeModel_(model), XEN_TO_C_GtkTreeModelForeachFunc(func), XEN_TO_C_lambda_data(func_data));
    xm_unprotect_at(loc);
    return(XEN_FALSE);
   }
}

static XEN gxg_gtk_tree_model_row_changed(XEN tree_model, XEN path, XEN iter)
{
  #define H_gtk_tree_model_row_changed "void gtk_tree_model_row_changed(GtkTreeModel* tree_model, GtkTreePath* path, \
GtkTreeIter* iter)"
  XEN_ASSERT_TYPE(XEN_GtkTreeModel__P(tree_model), tree_model, 1, "gtk_tree_model_row_changed", "GtkTreeModel*");
  XEN_ASSERT_TYPE(XEN_GtkTreePath__P(path), path, 2, "gtk_tree_model_row_changed", "GtkTreePath*");
  XEN_ASSERT_TYPE(XEN_GtkTreeIter__P(iter), iter, 3, "gtk_tree_model_row_changed", "GtkTreeIter*");
  gtk_tree_model_row_changed(XEN_TO_C_GtkTreeModel_(tree_model), XEN_TO_C_GtkTreePath_(path), XEN_TO_C_GtkTreeIter_(iter));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tree_model_row_inserted(XEN tree_model, XEN path, XEN iter)
{
  #define H_gtk_tree_model_row_inserted "void gtk_tree_model_row_inserted(GtkTreeModel* tree_model, GtkTreePath* path, \
GtkTreeIter* iter)"
  XEN_ASSERT_TYPE(XEN_GtkTreeModel__P(tree_model), tree_model, 1, "gtk_tree_model_row_inserted", "GtkTreeModel*");
  XEN_ASSERT_TYPE(XEN_GtkTreePath__P(path), path, 2, "gtk_tree_model_row_inserted", "GtkTreePath*");
  XEN_ASSERT_TYPE(XEN_GtkTreeIter__P(iter), iter, 3, "gtk_tree_model_row_inserted", "GtkTreeIter*");
  gtk_tree_model_row_inserted(XEN_TO_C_GtkTreeModel_(tree_model), XEN_TO_C_GtkTreePath_(path), XEN_TO_C_GtkTreeIter_(iter));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tree_model_row_has_child_toggled(XEN tree_model, XEN path, XEN iter)
{
  #define H_gtk_tree_model_row_has_child_toggled "void gtk_tree_model_row_has_child_toggled(GtkTreeModel* tree_model, \
GtkTreePath* path, GtkTreeIter* iter)"
  XEN_ASSERT_TYPE(XEN_GtkTreeModel__P(tree_model), tree_model, 1, "gtk_tree_model_row_has_child_toggled", "GtkTreeModel*");
  XEN_ASSERT_TYPE(XEN_GtkTreePath__P(path), path, 2, "gtk_tree_model_row_has_child_toggled", "GtkTreePath*");
  XEN_ASSERT_TYPE(XEN_GtkTreeIter__P(iter), iter, 3, "gtk_tree_model_row_has_child_toggled", "GtkTreeIter*");
  gtk_tree_model_row_has_child_toggled(XEN_TO_C_GtkTreeModel_(tree_model), XEN_TO_C_GtkTreePath_(path), XEN_TO_C_GtkTreeIter_(iter));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tree_model_row_deleted(XEN tree_model, XEN path)
{
  #define H_gtk_tree_model_row_deleted "void gtk_tree_model_row_deleted(GtkTreeModel* tree_model, GtkTreePath* path)"
  XEN_ASSERT_TYPE(XEN_GtkTreeModel__P(tree_model), tree_model, 1, "gtk_tree_model_row_deleted", "GtkTreeModel*");
  XEN_ASSERT_TYPE(XEN_GtkTreePath__P(path), path, 2, "gtk_tree_model_row_deleted", "GtkTreePath*");
  gtk_tree_model_row_deleted(XEN_TO_C_GtkTreeModel_(tree_model), XEN_TO_C_GtkTreePath_(path));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tree_model_rows_reordered(XEN tree_model, XEN path, XEN iter, XEN new_order)
{
  #define H_gtk_tree_model_rows_reordered "void gtk_tree_model_rows_reordered(GtkTreeModel* tree_model, \
GtkTreePath* path, GtkTreeIter* iter, gint* new_order)"
  XEN_ASSERT_TYPE(XEN_GtkTreeModel__P(tree_model), tree_model, 1, "gtk_tree_model_rows_reordered", "GtkTreeModel*");
  XEN_ASSERT_TYPE(XEN_GtkTreePath__P(path), path, 2, "gtk_tree_model_rows_reordered", "GtkTreePath*");
  XEN_ASSERT_TYPE(XEN_GtkTreeIter__P(iter), iter, 3, "gtk_tree_model_rows_reordered", "GtkTreeIter*");
  XEN_ASSERT_TYPE(XEN_gint__P(new_order), new_order, 4, "gtk_tree_model_rows_reordered", "gint*");
  gtk_tree_model_rows_reordered(XEN_TO_C_GtkTreeModel_(tree_model), XEN_TO_C_GtkTreePath_(path), XEN_TO_C_GtkTreeIter_(iter), 
                                XEN_TO_C_gint_(new_order));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tree_model_sort_get_type(void)
{
  #define H_gtk_tree_model_sort_get_type "GType gtk_tree_model_sort_get_type( void)"
  return(C_TO_XEN_GType(gtk_tree_model_sort_get_type()));
}

static XEN gxg_gtk_tree_model_sort_new_with_model(XEN child_model)
{
  #define H_gtk_tree_model_sort_new_with_model "GtkTreeModel* gtk_tree_model_sort_new_with_model(GtkTreeModel* child_model)"
  XEN_ASSERT_TYPE(XEN_GtkTreeModel__P(child_model), child_model, 1, "gtk_tree_model_sort_new_with_model", "GtkTreeModel*");
  return(C_TO_XEN_GtkTreeModel_(gtk_tree_model_sort_new_with_model(XEN_TO_C_GtkTreeModel_(child_model))));
}

static XEN gxg_gtk_tree_model_sort_get_model(XEN tree_model)
{
  #define H_gtk_tree_model_sort_get_model "GtkTreeModel* gtk_tree_model_sort_get_model(GtkTreeModelSort* tree_model)"
  XEN_ASSERT_TYPE(XEN_GtkTreeModelSort__P(tree_model), tree_model, 1, "gtk_tree_model_sort_get_model", "GtkTreeModelSort*");
  return(C_TO_XEN_GtkTreeModel_(gtk_tree_model_sort_get_model(XEN_TO_C_GtkTreeModelSort_(tree_model))));
}

static XEN gxg_gtk_tree_model_sort_convert_child_path_to_path(XEN tree_model_sort, XEN child_path)
{
  #define H_gtk_tree_model_sort_convert_child_path_to_path "GtkTreePath* gtk_tree_model_sort_convert_child_path_to_path(GtkTreeModelSort* tree_model_sort, \
GtkTreePath* child_path)"
  XEN_ASSERT_TYPE(XEN_GtkTreeModelSort__P(tree_model_sort), tree_model_sort, 1, "gtk_tree_model_sort_convert_child_path_to_path", "GtkTreeModelSort*");
  XEN_ASSERT_TYPE(XEN_GtkTreePath__P(child_path), child_path, 2, "gtk_tree_model_sort_convert_child_path_to_path", "GtkTreePath*");
  return(C_TO_XEN_GtkTreePath_(gtk_tree_model_sort_convert_child_path_to_path(XEN_TO_C_GtkTreeModelSort_(tree_model_sort), 
                                                                              XEN_TO_C_GtkTreePath_(child_path))));
}

static XEN gxg_gtk_tree_model_sort_convert_child_iter_to_iter(XEN tree_model_sort, XEN sort_iter, XEN child_iter)
{
  #define H_gtk_tree_model_sort_convert_child_iter_to_iter "void gtk_tree_model_sort_convert_child_iter_to_iter(GtkTreeModelSort* tree_model_sort, \
GtkTreeIter* sort_iter, GtkTreeIter* child_iter)"
  XEN_ASSERT_TYPE(XEN_GtkTreeModelSort__P(tree_model_sort), tree_model_sort, 1, "gtk_tree_model_sort_convert_child_iter_to_iter", "GtkTreeModelSort*");
  XEN_ASSERT_TYPE(XEN_GtkTreeIter__P(sort_iter), sort_iter, 2, "gtk_tree_model_sort_convert_child_iter_to_iter", "GtkTreeIter*");
  XEN_ASSERT_TYPE(XEN_GtkTreeIter__P(child_iter), child_iter, 3, "gtk_tree_model_sort_convert_child_iter_to_iter", "GtkTreeIter*");
  gtk_tree_model_sort_convert_child_iter_to_iter(XEN_TO_C_GtkTreeModelSort_(tree_model_sort), XEN_TO_C_GtkTreeIter_(sort_iter), 
                                                 XEN_TO_C_GtkTreeIter_(child_iter));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tree_model_sort_convert_path_to_child_path(XEN tree_model_sort, XEN sorted_path)
{
  #define H_gtk_tree_model_sort_convert_path_to_child_path "GtkTreePath* gtk_tree_model_sort_convert_path_to_child_path(GtkTreeModelSort* tree_model_sort, \
GtkTreePath* sorted_path)"
  XEN_ASSERT_TYPE(XEN_GtkTreeModelSort__P(tree_model_sort), tree_model_sort, 1, "gtk_tree_model_sort_convert_path_to_child_path", "GtkTreeModelSort*");
  XEN_ASSERT_TYPE(XEN_GtkTreePath__P(sorted_path), sorted_path, 2, "gtk_tree_model_sort_convert_path_to_child_path", "GtkTreePath*");
  return(C_TO_XEN_GtkTreePath_(gtk_tree_model_sort_convert_path_to_child_path(XEN_TO_C_GtkTreeModelSort_(tree_model_sort), 
                                                                              XEN_TO_C_GtkTreePath_(sorted_path))));
}

static XEN gxg_gtk_tree_model_sort_convert_iter_to_child_iter(XEN tree_model_sort, XEN child_iter, XEN sorted_iter)
{
  #define H_gtk_tree_model_sort_convert_iter_to_child_iter "void gtk_tree_model_sort_convert_iter_to_child_iter(GtkTreeModelSort* tree_model_sort, \
GtkTreeIter* child_iter, GtkTreeIter* sorted_iter)"
  XEN_ASSERT_TYPE(XEN_GtkTreeModelSort__P(tree_model_sort), tree_model_sort, 1, "gtk_tree_model_sort_convert_iter_to_child_iter", "GtkTreeModelSort*");
  XEN_ASSERT_TYPE(XEN_GtkTreeIter__P(child_iter), child_iter, 2, "gtk_tree_model_sort_convert_iter_to_child_iter", "GtkTreeIter*");
  XEN_ASSERT_TYPE(XEN_GtkTreeIter__P(sorted_iter), sorted_iter, 3, "gtk_tree_model_sort_convert_iter_to_child_iter", "GtkTreeIter*");
  gtk_tree_model_sort_convert_iter_to_child_iter(XEN_TO_C_GtkTreeModelSort_(tree_model_sort), XEN_TO_C_GtkTreeIter_(child_iter), 
                                                 XEN_TO_C_GtkTreeIter_(sorted_iter));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tree_model_sort_reset_default_sort_func(XEN tree_model_sort)
{
  #define H_gtk_tree_model_sort_reset_default_sort_func "void gtk_tree_model_sort_reset_default_sort_func(GtkTreeModelSort* tree_model_sort)"
  XEN_ASSERT_TYPE(XEN_GtkTreeModelSort__P(tree_model_sort), tree_model_sort, 1, "gtk_tree_model_sort_reset_default_sort_func", "GtkTreeModelSort*");
  gtk_tree_model_sort_reset_default_sort_func(XEN_TO_C_GtkTreeModelSort_(tree_model_sort));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tree_model_sort_clear_cache(XEN tree_model_sort)
{
  #define H_gtk_tree_model_sort_clear_cache "void gtk_tree_model_sort_clear_cache(GtkTreeModelSort* tree_model_sort)"
  XEN_ASSERT_TYPE(XEN_GtkTreeModelSort__P(tree_model_sort), tree_model_sort, 1, "gtk_tree_model_sort_clear_cache", "GtkTreeModelSort*");
  gtk_tree_model_sort_clear_cache(XEN_TO_C_GtkTreeModelSort_(tree_model_sort));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tree_selection_get_type(void)
{
  #define H_gtk_tree_selection_get_type "GtkType gtk_tree_selection_get_type( void)"
  return(C_TO_XEN_GtkType(gtk_tree_selection_get_type()));
}

static XEN gxg_gtk_tree_selection_set_mode(XEN selection, XEN type)
{
  #define H_gtk_tree_selection_set_mode "void gtk_tree_selection_set_mode(GtkTreeSelection* selection, \
GtkSelectionMode type)"
  XEN_ASSERT_TYPE(XEN_GtkTreeSelection__P(selection), selection, 1, "gtk_tree_selection_set_mode", "GtkTreeSelection*");
  XEN_ASSERT_TYPE(XEN_GtkSelectionMode_P(type), type, 2, "gtk_tree_selection_set_mode", "GtkSelectionMode");
  gtk_tree_selection_set_mode(XEN_TO_C_GtkTreeSelection_(selection), XEN_TO_C_GtkSelectionMode(type));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tree_selection_get_mode(XEN selection)
{
  #define H_gtk_tree_selection_get_mode "GtkSelectionMode gtk_tree_selection_get_mode(GtkTreeSelection* selection)"
  XEN_ASSERT_TYPE(XEN_GtkTreeSelection__P(selection), selection, 1, "gtk_tree_selection_get_mode", "GtkTreeSelection*");
  return(C_TO_XEN_GtkSelectionMode(gtk_tree_selection_get_mode(XEN_TO_C_GtkTreeSelection_(selection))));
}

static XEN gxg_gtk_tree_selection_set_select_function(XEN selection, XEN func, XEN func_data, XEN destroy)
{
  #define H_gtk_tree_selection_set_select_function "void gtk_tree_selection_set_select_function(GtkTreeSelection* selection, \
GtkTreeSelectionFunc func, lambda_data func_data, GtkDestroyNotify destroy)"
  XEN_ASSERT_TYPE(XEN_GtkTreeSelection__P(selection), selection, 1, "gtk_tree_selection_set_select_function", "GtkTreeSelection*");
  XEN_ASSERT_TYPE(XEN_GtkTreeSelectionFunc_P(func), func, 2, "gtk_tree_selection_set_select_function", "GtkTreeSelectionFunc");
  XEN_ASSERT_TYPE(XEN_lambda_data_P(func_data), func_data, 3, "gtk_tree_selection_set_select_function", "lambda_data");
  XEN_ASSERT_TYPE(XEN_GtkDestroyNotify_P(destroy), destroy, 4, "gtk_tree_selection_set_select_function", "GtkDestroyNotify");
  {
    XEN gxg_ptr = XEN_LIST_5(func, func_data, XEN_FALSE, XEN_FALSE, XEN_FALSE);
    xm_protect(gxg_ptr);
    XEN_LIST_SET(gxg_ptr, 3, destroy);
    gtk_tree_selection_set_select_function(XEN_TO_C_GtkTreeSelection_(selection), XEN_TO_C_GtkTreeSelectionFunc(func), XEN_TO_C_lambda_data(func_data), 
                                       XEN_TO_C_GtkDestroyNotify(destroy));
    return(XEN_FALSE);
   }
}

static XEN gxg_gtk_tree_selection_get_user_data(XEN selection)
{
  #define H_gtk_tree_selection_get_user_data "gpointer gtk_tree_selection_get_user_data(GtkTreeSelection* selection)"
  XEN_ASSERT_TYPE(XEN_GtkTreeSelection__P(selection), selection, 1, "gtk_tree_selection_get_user_data", "GtkTreeSelection*");
  return(C_TO_XEN_gpointer(gtk_tree_selection_get_user_data(XEN_TO_C_GtkTreeSelection_(selection))));
}

static XEN gxg_gtk_tree_selection_get_tree_view(XEN selection)
{
  #define H_gtk_tree_selection_get_tree_view "GtkTreeView* gtk_tree_selection_get_tree_view(GtkTreeSelection* selection)"
  XEN_ASSERT_TYPE(XEN_GtkTreeSelection__P(selection), selection, 1, "gtk_tree_selection_get_tree_view", "GtkTreeSelection*");
  return(C_TO_XEN_GtkTreeView_(gtk_tree_selection_get_tree_view(XEN_TO_C_GtkTreeSelection_(selection))));
}

static XEN gxg_gtk_tree_selection_get_selected(XEN selection, XEN model, XEN iter)
{
  #define H_gtk_tree_selection_get_selected "gboolean gtk_tree_selection_get_selected(GtkTreeSelection* selection, \
GtkTreeModel** model, GtkTreeIter* iter)"
  XEN_ASSERT_TYPE(XEN_GtkTreeSelection__P(selection), selection, 1, "gtk_tree_selection_get_selected", "GtkTreeSelection*");
  XEN_ASSERT_TYPE(XEN_GtkTreeModel___P(model), model, 2, "gtk_tree_selection_get_selected", "GtkTreeModel**");
  XEN_ASSERT_TYPE(XEN_GtkTreeIter__P(iter), iter, 3, "gtk_tree_selection_get_selected", "GtkTreeIter*");
  return(C_TO_XEN_gboolean(gtk_tree_selection_get_selected(XEN_TO_C_GtkTreeSelection_(selection), XEN_TO_C_GtkTreeModel__(model), 
                                                           XEN_TO_C_GtkTreeIter_(iter))));
}

static XEN gxg_gtk_tree_selection_selected_foreach(XEN selection, XEN func, XEN func_data)
{
  #define H_gtk_tree_selection_selected_foreach "void gtk_tree_selection_selected_foreach(GtkTreeSelection* selection, \
GtkTreeSelectionForeachFunc func, lambda_data func_data)"
  XEN_ASSERT_TYPE(XEN_GtkTreeSelection__P(selection), selection, 1, "gtk_tree_selection_selected_foreach", "GtkTreeSelection*");
  XEN_ASSERT_TYPE(XEN_GtkTreeSelectionForeachFunc_P(func), func, 2, "gtk_tree_selection_selected_foreach", "GtkTreeSelectionForeachFunc");
  if (XEN_NOT_BOUND_P(func_data)) func_data = XEN_FALSE; 
  else XEN_ASSERT_TYPE(XEN_lambda_data_P(func_data), func_data, 3, "gtk_tree_selection_selected_foreach", "lambda_data");
  {
    int loc;
    XEN gxg_ptr = XEN_LIST_5(func, func_data, XEN_FALSE, XEN_FALSE, XEN_FALSE);
    loc = xm_protect(gxg_ptr);
    XEN_LIST_SET(gxg_ptr, 2, C_TO_XEN_INT(loc));
    gtk_tree_selection_selected_foreach(XEN_TO_C_GtkTreeSelection_(selection), XEN_TO_C_GtkTreeSelectionForeachFunc(func), XEN_TO_C_lambda_data(func_data));
    xm_unprotect_at(loc);
    return(XEN_FALSE);
   }
}

static XEN gxg_gtk_tree_selection_select_path(XEN selection, XEN path)
{
  #define H_gtk_tree_selection_select_path "void gtk_tree_selection_select_path(GtkTreeSelection* selection, \
GtkTreePath* path)"
  XEN_ASSERT_TYPE(XEN_GtkTreeSelection__P(selection), selection, 1, "gtk_tree_selection_select_path", "GtkTreeSelection*");
  XEN_ASSERT_TYPE(XEN_GtkTreePath__P(path), path, 2, "gtk_tree_selection_select_path", "GtkTreePath*");
  gtk_tree_selection_select_path(XEN_TO_C_GtkTreeSelection_(selection), XEN_TO_C_GtkTreePath_(path));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tree_selection_unselect_path(XEN selection, XEN path)
{
  #define H_gtk_tree_selection_unselect_path "void gtk_tree_selection_unselect_path(GtkTreeSelection* selection, \
GtkTreePath* path)"
  XEN_ASSERT_TYPE(XEN_GtkTreeSelection__P(selection), selection, 1, "gtk_tree_selection_unselect_path", "GtkTreeSelection*");
  XEN_ASSERT_TYPE(XEN_GtkTreePath__P(path), path, 2, "gtk_tree_selection_unselect_path", "GtkTreePath*");
  gtk_tree_selection_unselect_path(XEN_TO_C_GtkTreeSelection_(selection), XEN_TO_C_GtkTreePath_(path));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tree_selection_select_iter(XEN selection, XEN iter)
{
  #define H_gtk_tree_selection_select_iter "void gtk_tree_selection_select_iter(GtkTreeSelection* selection, \
GtkTreeIter* iter)"
  XEN_ASSERT_TYPE(XEN_GtkTreeSelection__P(selection), selection, 1, "gtk_tree_selection_select_iter", "GtkTreeSelection*");
  XEN_ASSERT_TYPE(XEN_GtkTreeIter__P(iter), iter, 2, "gtk_tree_selection_select_iter", "GtkTreeIter*");
  gtk_tree_selection_select_iter(XEN_TO_C_GtkTreeSelection_(selection), XEN_TO_C_GtkTreeIter_(iter));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tree_selection_unselect_iter(XEN selection, XEN iter)
{
  #define H_gtk_tree_selection_unselect_iter "void gtk_tree_selection_unselect_iter(GtkTreeSelection* selection, \
GtkTreeIter* iter)"
  XEN_ASSERT_TYPE(XEN_GtkTreeSelection__P(selection), selection, 1, "gtk_tree_selection_unselect_iter", "GtkTreeSelection*");
  XEN_ASSERT_TYPE(XEN_GtkTreeIter__P(iter), iter, 2, "gtk_tree_selection_unselect_iter", "GtkTreeIter*");
  gtk_tree_selection_unselect_iter(XEN_TO_C_GtkTreeSelection_(selection), XEN_TO_C_GtkTreeIter_(iter));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tree_selection_path_is_selected(XEN selection, XEN path)
{
  #define H_gtk_tree_selection_path_is_selected "gboolean gtk_tree_selection_path_is_selected(GtkTreeSelection* selection, \
GtkTreePath* path)"
  XEN_ASSERT_TYPE(XEN_GtkTreeSelection__P(selection), selection, 1, "gtk_tree_selection_path_is_selected", "GtkTreeSelection*");
  XEN_ASSERT_TYPE(XEN_GtkTreePath__P(path), path, 2, "gtk_tree_selection_path_is_selected", "GtkTreePath*");
  return(C_TO_XEN_gboolean(gtk_tree_selection_path_is_selected(XEN_TO_C_GtkTreeSelection_(selection), XEN_TO_C_GtkTreePath_(path))));
}

static XEN gxg_gtk_tree_selection_iter_is_selected(XEN selection, XEN iter)
{
  #define H_gtk_tree_selection_iter_is_selected "gboolean gtk_tree_selection_iter_is_selected(GtkTreeSelection* selection, \
GtkTreeIter* iter)"
  XEN_ASSERT_TYPE(XEN_GtkTreeSelection__P(selection), selection, 1, "gtk_tree_selection_iter_is_selected", "GtkTreeSelection*");
  XEN_ASSERT_TYPE(XEN_GtkTreeIter__P(iter), iter, 2, "gtk_tree_selection_iter_is_selected", "GtkTreeIter*");
  return(C_TO_XEN_gboolean(gtk_tree_selection_iter_is_selected(XEN_TO_C_GtkTreeSelection_(selection), XEN_TO_C_GtkTreeIter_(iter))));
}

static XEN gxg_gtk_tree_selection_select_all(XEN selection)
{
  #define H_gtk_tree_selection_select_all "void gtk_tree_selection_select_all(GtkTreeSelection* selection)"
  XEN_ASSERT_TYPE(XEN_GtkTreeSelection__P(selection), selection, 1, "gtk_tree_selection_select_all", "GtkTreeSelection*");
  gtk_tree_selection_select_all(XEN_TO_C_GtkTreeSelection_(selection));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tree_selection_unselect_all(XEN selection)
{
  #define H_gtk_tree_selection_unselect_all "void gtk_tree_selection_unselect_all(GtkTreeSelection* selection)"
  XEN_ASSERT_TYPE(XEN_GtkTreeSelection__P(selection), selection, 1, "gtk_tree_selection_unselect_all", "GtkTreeSelection*");
  gtk_tree_selection_unselect_all(XEN_TO_C_GtkTreeSelection_(selection));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tree_selection_select_range(XEN selection, XEN start_path, XEN end_path)
{
  #define H_gtk_tree_selection_select_range "void gtk_tree_selection_select_range(GtkTreeSelection* selection, \
GtkTreePath* start_path, GtkTreePath* end_path)"
  XEN_ASSERT_TYPE(XEN_GtkTreeSelection__P(selection), selection, 1, "gtk_tree_selection_select_range", "GtkTreeSelection*");
  XEN_ASSERT_TYPE(XEN_GtkTreePath__P(start_path), start_path, 2, "gtk_tree_selection_select_range", "GtkTreePath*");
  XEN_ASSERT_TYPE(XEN_GtkTreePath__P(end_path), end_path, 3, "gtk_tree_selection_select_range", "GtkTreePath*");
  gtk_tree_selection_select_range(XEN_TO_C_GtkTreeSelection_(selection), XEN_TO_C_GtkTreePath_(start_path), XEN_TO_C_GtkTreePath_(end_path));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tree_sortable_get_type(void)
{
  #define H_gtk_tree_sortable_get_type "GType gtk_tree_sortable_get_type( void)"
  return(C_TO_XEN_GType(gtk_tree_sortable_get_type()));
}

static XEN gxg_gtk_tree_sortable_sort_column_changed(XEN sortable)
{
  #define H_gtk_tree_sortable_sort_column_changed "void gtk_tree_sortable_sort_column_changed(GtkTreeSortable* sortable)"
  XEN_ASSERT_TYPE(XEN_GtkTreeSortable__P(sortable), sortable, 1, "gtk_tree_sortable_sort_column_changed", "GtkTreeSortable*");
  gtk_tree_sortable_sort_column_changed(XEN_TO_C_GtkTreeSortable_(sortable));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tree_sortable_get_sort_column_id(XEN sortable, XEN sort_column_id, XEN order)
{
  #define H_gtk_tree_sortable_get_sort_column_id "gboolean gtk_tree_sortable_get_sort_column_id(GtkTreeSortable* sortable, \
gint* sort_column_id, GtkSortType* order)"
  XEN_ASSERT_TYPE(XEN_GtkTreeSortable__P(sortable), sortable, 1, "gtk_tree_sortable_get_sort_column_id", "GtkTreeSortable*");
  XEN_ASSERT_TYPE(XEN_gint__P(sort_column_id), sort_column_id, 2, "gtk_tree_sortable_get_sort_column_id", "gint*");
  XEN_ASSERT_TYPE(XEN_GtkSortType__P(order), order, 3, "gtk_tree_sortable_get_sort_column_id", "GtkSortType*");
  return(C_TO_XEN_gboolean(gtk_tree_sortable_get_sort_column_id(XEN_TO_C_GtkTreeSortable_(sortable), XEN_TO_C_gint_(sort_column_id), 
                                                                XEN_TO_C_GtkSortType_(order))));
}

static XEN gxg_gtk_tree_sortable_set_sort_column_id(XEN sortable, XEN sort_column_id, XEN order)
{
  #define H_gtk_tree_sortable_set_sort_column_id "void gtk_tree_sortable_set_sort_column_id(GtkTreeSortable* sortable, \
gint sort_column_id, GtkSortType order)"
  XEN_ASSERT_TYPE(XEN_GtkTreeSortable__P(sortable), sortable, 1, "gtk_tree_sortable_set_sort_column_id", "GtkTreeSortable*");
  XEN_ASSERT_TYPE(XEN_gint_P(sort_column_id), sort_column_id, 2, "gtk_tree_sortable_set_sort_column_id", "gint");
  XEN_ASSERT_TYPE(XEN_GtkSortType_P(order), order, 3, "gtk_tree_sortable_set_sort_column_id", "GtkSortType");
  gtk_tree_sortable_set_sort_column_id(XEN_TO_C_GtkTreeSortable_(sortable), XEN_TO_C_gint(sort_column_id), XEN_TO_C_GtkSortType(order));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tree_sortable_set_sort_func(XEN sortable, XEN sort_column_id, XEN func, XEN func_data, XEN destroy)
{
  #define H_gtk_tree_sortable_set_sort_func "void gtk_tree_sortable_set_sort_func(GtkTreeSortable* sortable, \
gint sort_column_id, GtkTreeIterCompareFunc func, lambda_data func_data, GtkDestroyNotify destroy)"
  XEN_ASSERT_TYPE(XEN_GtkTreeSortable__P(sortable), sortable, 1, "gtk_tree_sortable_set_sort_func", "GtkTreeSortable*");
  XEN_ASSERT_TYPE(XEN_gint_P(sort_column_id), sort_column_id, 2, "gtk_tree_sortable_set_sort_func", "gint");
  XEN_ASSERT_TYPE(XEN_GtkTreeIterCompareFunc_P(func), func, 3, "gtk_tree_sortable_set_sort_func", "GtkTreeIterCompareFunc");
  XEN_ASSERT_TYPE(XEN_lambda_data_P(func_data), func_data, 4, "gtk_tree_sortable_set_sort_func", "lambda_data");
  XEN_ASSERT_TYPE(XEN_GtkDestroyNotify_P(destroy), destroy, 5, "gtk_tree_sortable_set_sort_func", "GtkDestroyNotify");
  {
    XEN gxg_ptr = XEN_LIST_5(func, func_data, XEN_FALSE, XEN_FALSE, XEN_FALSE);
    xm_protect(gxg_ptr);
    XEN_LIST_SET(gxg_ptr, 3, destroy);
    gtk_tree_sortable_set_sort_func(XEN_TO_C_GtkTreeSortable_(sortable), XEN_TO_C_gint(sort_column_id), XEN_TO_C_GtkTreeIterCompareFunc(func), 
                                XEN_TO_C_lambda_data(func_data), XEN_TO_C_GtkDestroyNotify(destroy));
    return(XEN_FALSE);
   }
}

static XEN gxg_gtk_tree_sortable_set_default_sort_func(XEN sortable, XEN func, XEN func_data, XEN destroy)
{
  #define H_gtk_tree_sortable_set_default_sort_func "void gtk_tree_sortable_set_default_sort_func(GtkTreeSortable* sortable, \
GtkTreeIterCompareFunc func, lambda_data func_data, GtkDestroyNotify destroy)"
  XEN_ASSERT_TYPE(XEN_GtkTreeSortable__P(sortable), sortable, 1, "gtk_tree_sortable_set_default_sort_func", "GtkTreeSortable*");
  XEN_ASSERT_TYPE(XEN_GtkTreeIterCompareFunc_P(func), func, 2, "gtk_tree_sortable_set_default_sort_func", "GtkTreeIterCompareFunc");
  XEN_ASSERT_TYPE(XEN_lambda_data_P(func_data), func_data, 3, "gtk_tree_sortable_set_default_sort_func", "lambda_data");
  XEN_ASSERT_TYPE(XEN_GtkDestroyNotify_P(destroy), destroy, 4, "gtk_tree_sortable_set_default_sort_func", "GtkDestroyNotify");
  {
    XEN gxg_ptr = XEN_LIST_5(func, func_data, XEN_FALSE, XEN_FALSE, XEN_FALSE);
    xm_protect(gxg_ptr);
    XEN_LIST_SET(gxg_ptr, 3, destroy);
    gtk_tree_sortable_set_default_sort_func(XEN_TO_C_GtkTreeSortable_(sortable), XEN_TO_C_GtkTreeIterCompareFunc(func), XEN_TO_C_lambda_data(func_data), 
                                        XEN_TO_C_GtkDestroyNotify(destroy));
    return(XEN_FALSE);
   }
}

static XEN gxg_gtk_tree_sortable_has_default_sort_func(XEN sortable)
{
  #define H_gtk_tree_sortable_has_default_sort_func "gboolean gtk_tree_sortable_has_default_sort_func(GtkTreeSortable* sortable)"
  XEN_ASSERT_TYPE(XEN_GtkTreeSortable__P(sortable), sortable, 1, "gtk_tree_sortable_has_default_sort_func", "GtkTreeSortable*");
  return(C_TO_XEN_gboolean(gtk_tree_sortable_has_default_sort_func(XEN_TO_C_GtkTreeSortable_(sortable))));
}

static XEN gxg_gtk_tree_store_get_type(void)
{
  #define H_gtk_tree_store_get_type "GtkType gtk_tree_store_get_type( void)"
  return(C_TO_XEN_GtkType(gtk_tree_store_get_type()));
}

static XEN gxg_gtk_tree_store_new(XEN n_columns)
{
  #define H_gtk_tree_store_new "GtkTreeStore* gtk_tree_store_new(gint n_columns, ...)"
  XEN_ASSERT_TYPE(XEN_gint_P(n_columns), n_columns, 1, "gtk_tree_store_new", "gint");
  return(C_TO_XEN_GtkTreeStore_(gtk_tree_store_new(XEN_TO_C_gint(n_columns))));
}

static XEN gxg_gtk_tree_store_newv(XEN n_columns, XEN types)
{
  #define H_gtk_tree_store_newv "GtkTreeStore* gtk_tree_store_newv(gint n_columns, GType* types)"
  XEN_ASSERT_TYPE(XEN_gint_P(n_columns), n_columns, 1, "gtk_tree_store_newv", "gint");
  XEN_ASSERT_TYPE(XEN_GType__P(types), types, 2, "gtk_tree_store_newv", "GType*");
  return(C_TO_XEN_GtkTreeStore_(gtk_tree_store_newv(XEN_TO_C_gint(n_columns), XEN_TO_C_GType_(types))));
}

static XEN gxg_gtk_tree_store_set_column_types(XEN tree_store, XEN n_columns, XEN types)
{
  #define H_gtk_tree_store_set_column_types "void gtk_tree_store_set_column_types(GtkTreeStore* tree_store, \
gint n_columns, GType* types)"
  XEN_ASSERT_TYPE(XEN_GtkTreeStore__P(tree_store), tree_store, 1, "gtk_tree_store_set_column_types", "GtkTreeStore*");
  XEN_ASSERT_TYPE(XEN_gint_P(n_columns), n_columns, 2, "gtk_tree_store_set_column_types", "gint");
  XEN_ASSERT_TYPE(XEN_GType__P(types), types, 3, "gtk_tree_store_set_column_types", "GType*");
  gtk_tree_store_set_column_types(XEN_TO_C_GtkTreeStore_(tree_store), XEN_TO_C_gint(n_columns), XEN_TO_C_GType_(types));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tree_store_set_value(XEN tree_store, XEN iter, XEN column, XEN value)
{
  #define H_gtk_tree_store_set_value "void gtk_tree_store_set_value(GtkTreeStore* tree_store, GtkTreeIter* iter, \
gint column, GValue* value)"
  XEN_ASSERT_TYPE(XEN_GtkTreeStore__P(tree_store), tree_store, 1, "gtk_tree_store_set_value", "GtkTreeStore*");
  XEN_ASSERT_TYPE(XEN_GtkTreeIter__P(iter), iter, 2, "gtk_tree_store_set_value", "GtkTreeIter*");
  XEN_ASSERT_TYPE(XEN_gint_P(column), column, 3, "gtk_tree_store_set_value", "gint");
  XEN_ASSERT_TYPE(XEN_GValue__P(value), value, 4, "gtk_tree_store_set_value", "GValue*");
  gtk_tree_store_set_value(XEN_TO_C_GtkTreeStore_(tree_store), XEN_TO_C_GtkTreeIter_(iter), XEN_TO_C_gint(column), XEN_TO_C_GValue_(value));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tree_store_set(XEN tree_store, XEN iter)
{
  #define H_gtk_tree_store_set "void gtk_tree_store_set(GtkTreeStore* tree_store, GtkTreeIter* iter, \
...)"
  XEN_ASSERT_TYPE(XEN_GtkTreeStore__P(tree_store), tree_store, 1, "gtk_tree_store_set", "GtkTreeStore*");
  XEN_ASSERT_TYPE(XEN_GtkTreeIter__P(iter), iter, 2, "gtk_tree_store_set", "GtkTreeIter*");
  gtk_tree_store_set(XEN_TO_C_GtkTreeStore_(tree_store), XEN_TO_C_GtkTreeIter_(iter));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tree_store_remove(XEN tree_store, XEN iter)
{
  #define H_gtk_tree_store_remove "void gtk_tree_store_remove(GtkTreeStore* tree_store, GtkTreeIter* iter)"
  XEN_ASSERT_TYPE(XEN_GtkTreeStore__P(tree_store), tree_store, 1, "gtk_tree_store_remove", "GtkTreeStore*");
  XEN_ASSERT_TYPE(XEN_GtkTreeIter__P(iter), iter, 2, "gtk_tree_store_remove", "GtkTreeIter*");
  gtk_tree_store_remove(XEN_TO_C_GtkTreeStore_(tree_store), XEN_TO_C_GtkTreeIter_(iter));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tree_store_insert(XEN tree_store, XEN iter, XEN parent, XEN position)
{
  #define H_gtk_tree_store_insert "void gtk_tree_store_insert(GtkTreeStore* tree_store, GtkTreeIter* iter, \
GtkTreeIter* parent, gint position)"
  XEN_ASSERT_TYPE(XEN_GtkTreeStore__P(tree_store), tree_store, 1, "gtk_tree_store_insert", "GtkTreeStore*");
  XEN_ASSERT_TYPE(XEN_GtkTreeIter__P(iter), iter, 2, "gtk_tree_store_insert", "GtkTreeIter*");
  XEN_ASSERT_TYPE(XEN_GtkTreeIter__P(parent) || XEN_FALSE_P(parent), parent, 3, "gtk_tree_store_insert", "GtkTreeIter*");
  XEN_ASSERT_TYPE(XEN_gint_P(position), position, 4, "gtk_tree_store_insert", "gint");
  gtk_tree_store_insert(XEN_TO_C_GtkTreeStore_(tree_store), XEN_TO_C_GtkTreeIter_(iter), XEN_TO_C_GtkTreeIter_(parent), XEN_TO_C_gint(position));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tree_store_insert_before(XEN tree_store, XEN iter, XEN parent, XEN sibling)
{
  #define H_gtk_tree_store_insert_before "void gtk_tree_store_insert_before(GtkTreeStore* tree_store, \
GtkTreeIter* iter, GtkTreeIter* parent, GtkTreeIter* sibling)"
  XEN_ASSERT_TYPE(XEN_GtkTreeStore__P(tree_store), tree_store, 1, "gtk_tree_store_insert_before", "GtkTreeStore*");
  XEN_ASSERT_TYPE(XEN_GtkTreeIter__P(iter), iter, 2, "gtk_tree_store_insert_before", "GtkTreeIter*");
  XEN_ASSERT_TYPE(XEN_GtkTreeIter__P(parent) || XEN_FALSE_P(parent), parent, 3, "gtk_tree_store_insert_before", "GtkTreeIter*");
  XEN_ASSERT_TYPE(XEN_GtkTreeIter__P(sibling) || XEN_FALSE_P(sibling), sibling, 4, "gtk_tree_store_insert_before", "GtkTreeIter*");
  gtk_tree_store_insert_before(XEN_TO_C_GtkTreeStore_(tree_store), XEN_TO_C_GtkTreeIter_(iter), XEN_TO_C_GtkTreeIter_(parent), 
                               XEN_TO_C_GtkTreeIter_(sibling));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tree_store_insert_after(XEN tree_store, XEN iter, XEN parent, XEN sibling)
{
  #define H_gtk_tree_store_insert_after "void gtk_tree_store_insert_after(GtkTreeStore* tree_store, GtkTreeIter* iter, \
GtkTreeIter* parent, GtkTreeIter* sibling)"
  XEN_ASSERT_TYPE(XEN_GtkTreeStore__P(tree_store), tree_store, 1, "gtk_tree_store_insert_after", "GtkTreeStore*");
  XEN_ASSERT_TYPE(XEN_GtkTreeIter__P(iter), iter, 2, "gtk_tree_store_insert_after", "GtkTreeIter*");
  XEN_ASSERT_TYPE(XEN_GtkTreeIter__P(parent) || XEN_FALSE_P(parent), parent, 3, "gtk_tree_store_insert_after", "GtkTreeIter*");
  XEN_ASSERT_TYPE(XEN_GtkTreeIter__P(sibling) || XEN_FALSE_P(sibling), sibling, 4, "gtk_tree_store_insert_after", "GtkTreeIter*");
  gtk_tree_store_insert_after(XEN_TO_C_GtkTreeStore_(tree_store), XEN_TO_C_GtkTreeIter_(iter), XEN_TO_C_GtkTreeIter_(parent), 
                              XEN_TO_C_GtkTreeIter_(sibling));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tree_store_prepend(XEN tree_store, XEN iter, XEN parent)
{
  #define H_gtk_tree_store_prepend "void gtk_tree_store_prepend(GtkTreeStore* tree_store, GtkTreeIter* iter, \
GtkTreeIter* parent)"
  XEN_ASSERT_TYPE(XEN_GtkTreeStore__P(tree_store), tree_store, 1, "gtk_tree_store_prepend", "GtkTreeStore*");
  XEN_ASSERT_TYPE(XEN_GtkTreeIter__P(iter), iter, 2, "gtk_tree_store_prepend", "GtkTreeIter*");
  XEN_ASSERT_TYPE(XEN_GtkTreeIter__P(parent) || XEN_FALSE_P(parent), parent, 3, "gtk_tree_store_prepend", "GtkTreeIter*");
  gtk_tree_store_prepend(XEN_TO_C_GtkTreeStore_(tree_store), XEN_TO_C_GtkTreeIter_(iter), XEN_TO_C_GtkTreeIter_(parent));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tree_store_append(XEN tree_store, XEN iter, XEN parent)
{
  #define H_gtk_tree_store_append "void gtk_tree_store_append(GtkTreeStore* tree_store, GtkTreeIter* iter, \
GtkTreeIter* parent)"
  XEN_ASSERT_TYPE(XEN_GtkTreeStore__P(tree_store), tree_store, 1, "gtk_tree_store_append", "GtkTreeStore*");
  XEN_ASSERT_TYPE(XEN_GtkTreeIter__P(iter), iter, 2, "gtk_tree_store_append", "GtkTreeIter*");
  XEN_ASSERT_TYPE(XEN_GtkTreeIter__P(parent) || XEN_FALSE_P(parent), parent, 3, "gtk_tree_store_append", "GtkTreeIter*");
  gtk_tree_store_append(XEN_TO_C_GtkTreeStore_(tree_store), XEN_TO_C_GtkTreeIter_(iter), XEN_TO_C_GtkTreeIter_(parent));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tree_store_is_ancestor(XEN tree_store, XEN iter, XEN descendant)
{
  #define H_gtk_tree_store_is_ancestor "gboolean gtk_tree_store_is_ancestor(GtkTreeStore* tree_store, \
GtkTreeIter* iter, GtkTreeIter* descendant)"
  XEN_ASSERT_TYPE(XEN_GtkTreeStore__P(tree_store), tree_store, 1, "gtk_tree_store_is_ancestor", "GtkTreeStore*");
  XEN_ASSERT_TYPE(XEN_GtkTreeIter__P(iter), iter, 2, "gtk_tree_store_is_ancestor", "GtkTreeIter*");
  XEN_ASSERT_TYPE(XEN_GtkTreeIter__P(descendant), descendant, 3, "gtk_tree_store_is_ancestor", "GtkTreeIter*");
  return(C_TO_XEN_gboolean(gtk_tree_store_is_ancestor(XEN_TO_C_GtkTreeStore_(tree_store), XEN_TO_C_GtkTreeIter_(iter), XEN_TO_C_GtkTreeIter_(descendant))));
}

static XEN gxg_gtk_tree_store_iter_depth(XEN tree_store, XEN iter)
{
  #define H_gtk_tree_store_iter_depth "gint gtk_tree_store_iter_depth(GtkTreeStore* tree_store, GtkTreeIter* iter)"
  XEN_ASSERT_TYPE(XEN_GtkTreeStore__P(tree_store), tree_store, 1, "gtk_tree_store_iter_depth", "GtkTreeStore*");
  XEN_ASSERT_TYPE(XEN_GtkTreeIter__P(iter), iter, 2, "gtk_tree_store_iter_depth", "GtkTreeIter*");
  return(C_TO_XEN_gint(gtk_tree_store_iter_depth(XEN_TO_C_GtkTreeStore_(tree_store), XEN_TO_C_GtkTreeIter_(iter))));
}

static XEN gxg_gtk_tree_store_clear(XEN tree_store)
{
  #define H_gtk_tree_store_clear "void gtk_tree_store_clear(GtkTreeStore* tree_store)"
  XEN_ASSERT_TYPE(XEN_GtkTreeStore__P(tree_store), tree_store, 1, "gtk_tree_store_clear", "GtkTreeStore*");
  gtk_tree_store_clear(XEN_TO_C_GtkTreeStore_(tree_store));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tree_view_column_get_type(void)
{
  #define H_gtk_tree_view_column_get_type "GtkType gtk_tree_view_column_get_type( void)"
  return(C_TO_XEN_GtkType(gtk_tree_view_column_get_type()));
}

static XEN gxg_gtk_tree_view_column_new(void)
{
  #define H_gtk_tree_view_column_new "GtkTreeViewColumn* gtk_tree_view_column_new( void)"
  return(C_TO_XEN_GtkTreeViewColumn_(gtk_tree_view_column_new()));
}

static XEN gxg_gtk_tree_view_column_new_with_attributes(XEN title, XEN cell)
{
  #define H_gtk_tree_view_column_new_with_attributes "GtkTreeViewColumn* gtk_tree_view_column_new_with_attributes(gchar* title, \
GtkCellRenderer* cell, ...)"
  XEN_ASSERT_TYPE(XEN_gchar__P(title), title, 1, "gtk_tree_view_column_new_with_attributes", "gchar*");
  XEN_ASSERT_TYPE(XEN_GtkCellRenderer__P(cell), cell, 2, "gtk_tree_view_column_new_with_attributes", "GtkCellRenderer*");
  return(C_TO_XEN_GtkTreeViewColumn_(gtk_tree_view_column_new_with_attributes(XEN_TO_C_gchar_(title), XEN_TO_C_GtkCellRenderer_(cell))));
}

static XEN gxg_gtk_tree_view_column_pack_start(XEN tree_column, XEN cell, XEN expand)
{
  #define H_gtk_tree_view_column_pack_start "void gtk_tree_view_column_pack_start(GtkTreeViewColumn* tree_column, \
GtkCellRenderer* cell, gboolean expand)"
  XEN_ASSERT_TYPE(XEN_GtkTreeViewColumn__P(tree_column), tree_column, 1, "gtk_tree_view_column_pack_start", "GtkTreeViewColumn*");
  XEN_ASSERT_TYPE(XEN_GtkCellRenderer__P(cell), cell, 2, "gtk_tree_view_column_pack_start", "GtkCellRenderer*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(expand), expand, 3, "gtk_tree_view_column_pack_start", "gboolean");
  gtk_tree_view_column_pack_start(XEN_TO_C_GtkTreeViewColumn_(tree_column), XEN_TO_C_GtkCellRenderer_(cell), XEN_TO_C_gboolean(expand));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tree_view_column_pack_end(XEN tree_column, XEN cell, XEN expand)
{
  #define H_gtk_tree_view_column_pack_end "void gtk_tree_view_column_pack_end(GtkTreeViewColumn* tree_column, \
GtkCellRenderer* cell, gboolean expand)"
  XEN_ASSERT_TYPE(XEN_GtkTreeViewColumn__P(tree_column), tree_column, 1, "gtk_tree_view_column_pack_end", "GtkTreeViewColumn*");
  XEN_ASSERT_TYPE(XEN_GtkCellRenderer__P(cell), cell, 2, "gtk_tree_view_column_pack_end", "GtkCellRenderer*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(expand), expand, 3, "gtk_tree_view_column_pack_end", "gboolean");
  gtk_tree_view_column_pack_end(XEN_TO_C_GtkTreeViewColumn_(tree_column), XEN_TO_C_GtkCellRenderer_(cell), XEN_TO_C_gboolean(expand));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tree_view_column_clear(XEN tree_column)
{
  #define H_gtk_tree_view_column_clear "void gtk_tree_view_column_clear(GtkTreeViewColumn* tree_column)"
  XEN_ASSERT_TYPE(XEN_GtkTreeViewColumn__P(tree_column), tree_column, 1, "gtk_tree_view_column_clear", "GtkTreeViewColumn*");
  gtk_tree_view_column_clear(XEN_TO_C_GtkTreeViewColumn_(tree_column));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tree_view_column_get_cell_renderers(XEN tree_column)
{
  #define H_gtk_tree_view_column_get_cell_renderers "GList* gtk_tree_view_column_get_cell_renderers(GtkTreeViewColumn* tree_column)"
  XEN_ASSERT_TYPE(XEN_GtkTreeViewColumn__P(tree_column), tree_column, 1, "gtk_tree_view_column_get_cell_renderers", "GtkTreeViewColumn*");
  return(C_TO_XEN_GList_(gtk_tree_view_column_get_cell_renderers(XEN_TO_C_GtkTreeViewColumn_(tree_column))));
}

static XEN gxg_gtk_tree_view_column_add_attribute(XEN tree_column, XEN cell_renderer, XEN attribute, XEN column)
{
  #define H_gtk_tree_view_column_add_attribute "void gtk_tree_view_column_add_attribute(GtkTreeViewColumn* tree_column, \
GtkCellRenderer* cell_renderer, gchar* attribute, gint column)"
  XEN_ASSERT_TYPE(XEN_GtkTreeViewColumn__P(tree_column), tree_column, 1, "gtk_tree_view_column_add_attribute", "GtkTreeViewColumn*");
  XEN_ASSERT_TYPE(XEN_GtkCellRenderer__P(cell_renderer), cell_renderer, 2, "gtk_tree_view_column_add_attribute", "GtkCellRenderer*");
  XEN_ASSERT_TYPE(XEN_gchar__P(attribute), attribute, 3, "gtk_tree_view_column_add_attribute", "gchar*");
  XEN_ASSERT_TYPE(XEN_gint_P(column), column, 4, "gtk_tree_view_column_add_attribute", "gint");
  gtk_tree_view_column_add_attribute(XEN_TO_C_GtkTreeViewColumn_(tree_column), XEN_TO_C_GtkCellRenderer_(cell_renderer), 
                                     XEN_TO_C_gchar_(attribute), XEN_TO_C_gint(column));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tree_view_column_set_attributes(XEN tree_column, XEN cell_renderer)
{
  #define H_gtk_tree_view_column_set_attributes "void gtk_tree_view_column_set_attributes(GtkTreeViewColumn* tree_column, \
GtkCellRenderer* cell_renderer, ...)"
  XEN_ASSERT_TYPE(XEN_GtkTreeViewColumn__P(tree_column), tree_column, 1, "gtk_tree_view_column_set_attributes", "GtkTreeViewColumn*");
  XEN_ASSERT_TYPE(XEN_GtkCellRenderer__P(cell_renderer), cell_renderer, 2, "gtk_tree_view_column_set_attributes", "GtkCellRenderer*");
  gtk_tree_view_column_set_attributes(XEN_TO_C_GtkTreeViewColumn_(tree_column), XEN_TO_C_GtkCellRenderer_(cell_renderer));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tree_view_column_set_cell_data_func(XEN tree_column, XEN cell_renderer, XEN func, XEN func_data, XEN destroy)
{
  #define H_gtk_tree_view_column_set_cell_data_func "void gtk_tree_view_column_set_cell_data_func(GtkTreeViewColumn* tree_column, \
GtkCellRenderer* cell_renderer, GtkTreeCellDataFunc func, lambda_data func_data, GtkDestroyNotify destroy)"
  XEN_ASSERT_TYPE(XEN_GtkTreeViewColumn__P(tree_column), tree_column, 1, "gtk_tree_view_column_set_cell_data_func", "GtkTreeViewColumn*");
  XEN_ASSERT_TYPE(XEN_GtkCellRenderer__P(cell_renderer), cell_renderer, 2, "gtk_tree_view_column_set_cell_data_func", "GtkCellRenderer*");
  XEN_ASSERT_TYPE(XEN_GtkTreeCellDataFunc_P(func), func, 3, "gtk_tree_view_column_set_cell_data_func", "GtkTreeCellDataFunc");
  XEN_ASSERT_TYPE(XEN_lambda_data_P(func_data), func_data, 4, "gtk_tree_view_column_set_cell_data_func", "lambda_data");
  XEN_ASSERT_TYPE(XEN_GtkDestroyNotify_P(destroy), destroy, 5, "gtk_tree_view_column_set_cell_data_func", "GtkDestroyNotify");
  {
    XEN gxg_ptr = XEN_LIST_5(func, func_data, XEN_FALSE, XEN_FALSE, XEN_FALSE);
    xm_protect(gxg_ptr);
    XEN_LIST_SET(gxg_ptr, 3, destroy);
    gtk_tree_view_column_set_cell_data_func(XEN_TO_C_GtkTreeViewColumn_(tree_column), XEN_TO_C_GtkCellRenderer_(cell_renderer), 
                                        XEN_TO_C_GtkTreeCellDataFunc(func), XEN_TO_C_lambda_data(func_data), XEN_TO_C_GtkDestroyNotify(destroy));
    return(XEN_FALSE);
   }
}

static XEN gxg_gtk_tree_view_column_clear_attributes(XEN tree_column, XEN cell_renderer)
{
  #define H_gtk_tree_view_column_clear_attributes "void gtk_tree_view_column_clear_attributes(GtkTreeViewColumn* tree_column, \
GtkCellRenderer* cell_renderer)"
  XEN_ASSERT_TYPE(XEN_GtkTreeViewColumn__P(tree_column), tree_column, 1, "gtk_tree_view_column_clear_attributes", "GtkTreeViewColumn*");
  XEN_ASSERT_TYPE(XEN_GtkCellRenderer__P(cell_renderer), cell_renderer, 2, "gtk_tree_view_column_clear_attributes", "GtkCellRenderer*");
  gtk_tree_view_column_clear_attributes(XEN_TO_C_GtkTreeViewColumn_(tree_column), XEN_TO_C_GtkCellRenderer_(cell_renderer));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tree_view_column_set_spacing(XEN tree_column, XEN spacing)
{
  #define H_gtk_tree_view_column_set_spacing "void gtk_tree_view_column_set_spacing(GtkTreeViewColumn* tree_column, \
gint spacing)"
  XEN_ASSERT_TYPE(XEN_GtkTreeViewColumn__P(tree_column), tree_column, 1, "gtk_tree_view_column_set_spacing", "GtkTreeViewColumn*");
  XEN_ASSERT_TYPE(XEN_gint_P(spacing), spacing, 2, "gtk_tree_view_column_set_spacing", "gint");
  gtk_tree_view_column_set_spacing(XEN_TO_C_GtkTreeViewColumn_(tree_column), XEN_TO_C_gint(spacing));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tree_view_column_get_spacing(XEN tree_column)
{
  #define H_gtk_tree_view_column_get_spacing "gint gtk_tree_view_column_get_spacing(GtkTreeViewColumn* tree_column)"
  XEN_ASSERT_TYPE(XEN_GtkTreeViewColumn__P(tree_column), tree_column, 1, "gtk_tree_view_column_get_spacing", "GtkTreeViewColumn*");
  return(C_TO_XEN_gint(gtk_tree_view_column_get_spacing(XEN_TO_C_GtkTreeViewColumn_(tree_column))));
}

static XEN gxg_gtk_tree_view_column_set_visible(XEN tree_column, XEN visible)
{
  #define H_gtk_tree_view_column_set_visible "void gtk_tree_view_column_set_visible(GtkTreeViewColumn* tree_column, \
gboolean visible)"
  XEN_ASSERT_TYPE(XEN_GtkTreeViewColumn__P(tree_column), tree_column, 1, "gtk_tree_view_column_set_visible", "GtkTreeViewColumn*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(visible), visible, 2, "gtk_tree_view_column_set_visible", "gboolean");
  gtk_tree_view_column_set_visible(XEN_TO_C_GtkTreeViewColumn_(tree_column), XEN_TO_C_gboolean(visible));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tree_view_column_get_visible(XEN tree_column)
{
  #define H_gtk_tree_view_column_get_visible "gboolean gtk_tree_view_column_get_visible(GtkTreeViewColumn* tree_column)"
  XEN_ASSERT_TYPE(XEN_GtkTreeViewColumn__P(tree_column), tree_column, 1, "gtk_tree_view_column_get_visible", "GtkTreeViewColumn*");
  return(C_TO_XEN_gboolean(gtk_tree_view_column_get_visible(XEN_TO_C_GtkTreeViewColumn_(tree_column))));
}

static XEN gxg_gtk_tree_view_column_set_resizable(XEN tree_column, XEN resizable)
{
  #define H_gtk_tree_view_column_set_resizable "void gtk_tree_view_column_set_resizable(GtkTreeViewColumn* tree_column, \
gboolean resizable)"
  XEN_ASSERT_TYPE(XEN_GtkTreeViewColumn__P(tree_column), tree_column, 1, "gtk_tree_view_column_set_resizable", "GtkTreeViewColumn*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(resizable), resizable, 2, "gtk_tree_view_column_set_resizable", "gboolean");
  gtk_tree_view_column_set_resizable(XEN_TO_C_GtkTreeViewColumn_(tree_column), XEN_TO_C_gboolean(resizable));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tree_view_column_get_resizable(XEN tree_column)
{
  #define H_gtk_tree_view_column_get_resizable "gboolean gtk_tree_view_column_get_resizable(GtkTreeViewColumn* tree_column)"
  XEN_ASSERT_TYPE(XEN_GtkTreeViewColumn__P(tree_column), tree_column, 1, "gtk_tree_view_column_get_resizable", "GtkTreeViewColumn*");
  return(C_TO_XEN_gboolean(gtk_tree_view_column_get_resizable(XEN_TO_C_GtkTreeViewColumn_(tree_column))));
}

static XEN gxg_gtk_tree_view_column_set_sizing(XEN tree_column, XEN type)
{
  #define H_gtk_tree_view_column_set_sizing "void gtk_tree_view_column_set_sizing(GtkTreeViewColumn* tree_column, \
GtkTreeViewColumnSizing type)"
  XEN_ASSERT_TYPE(XEN_GtkTreeViewColumn__P(tree_column), tree_column, 1, "gtk_tree_view_column_set_sizing", "GtkTreeViewColumn*");
  XEN_ASSERT_TYPE(XEN_GtkTreeViewColumnSizing_P(type), type, 2, "gtk_tree_view_column_set_sizing", "GtkTreeViewColumnSizing");
  gtk_tree_view_column_set_sizing(XEN_TO_C_GtkTreeViewColumn_(tree_column), XEN_TO_C_GtkTreeViewColumnSizing(type));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tree_view_column_get_sizing(XEN tree_column)
{
  #define H_gtk_tree_view_column_get_sizing "GtkTreeViewColumnSizing gtk_tree_view_column_get_sizing(GtkTreeViewColumn* tree_column)"
  XEN_ASSERT_TYPE(XEN_GtkTreeViewColumn__P(tree_column), tree_column, 1, "gtk_tree_view_column_get_sizing", "GtkTreeViewColumn*");
  return(C_TO_XEN_GtkTreeViewColumnSizing(gtk_tree_view_column_get_sizing(XEN_TO_C_GtkTreeViewColumn_(tree_column))));
}

static XEN gxg_gtk_tree_view_column_get_width(XEN tree_column)
{
  #define H_gtk_tree_view_column_get_width "gint gtk_tree_view_column_get_width(GtkTreeViewColumn* tree_column)"
  XEN_ASSERT_TYPE(XEN_GtkTreeViewColumn__P(tree_column), tree_column, 1, "gtk_tree_view_column_get_width", "GtkTreeViewColumn*");
  return(C_TO_XEN_gint(gtk_tree_view_column_get_width(XEN_TO_C_GtkTreeViewColumn_(tree_column))));
}

static XEN gxg_gtk_tree_view_column_get_fixed_width(XEN tree_column)
{
  #define H_gtk_tree_view_column_get_fixed_width "gint gtk_tree_view_column_get_fixed_width(GtkTreeViewColumn* tree_column)"
  XEN_ASSERT_TYPE(XEN_GtkTreeViewColumn__P(tree_column), tree_column, 1, "gtk_tree_view_column_get_fixed_width", "GtkTreeViewColumn*");
  return(C_TO_XEN_gint(gtk_tree_view_column_get_fixed_width(XEN_TO_C_GtkTreeViewColumn_(tree_column))));
}

static XEN gxg_gtk_tree_view_column_set_fixed_width(XEN tree_column, XEN fixed_width)
{
  #define H_gtk_tree_view_column_set_fixed_width "void gtk_tree_view_column_set_fixed_width(GtkTreeViewColumn* tree_column, \
gint fixed_width)"
  XEN_ASSERT_TYPE(XEN_GtkTreeViewColumn__P(tree_column), tree_column, 1, "gtk_tree_view_column_set_fixed_width", "GtkTreeViewColumn*");
  XEN_ASSERT_TYPE(XEN_gint_P(fixed_width), fixed_width, 2, "gtk_tree_view_column_set_fixed_width", "gint");
  gtk_tree_view_column_set_fixed_width(XEN_TO_C_GtkTreeViewColumn_(tree_column), XEN_TO_C_gint(fixed_width));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tree_view_column_set_min_width(XEN tree_column, XEN min_width)
{
  #define H_gtk_tree_view_column_set_min_width "void gtk_tree_view_column_set_min_width(GtkTreeViewColumn* tree_column, \
gint min_width)"
  XEN_ASSERT_TYPE(XEN_GtkTreeViewColumn__P(tree_column), tree_column, 1, "gtk_tree_view_column_set_min_width", "GtkTreeViewColumn*");
  XEN_ASSERT_TYPE(XEN_gint_P(min_width), min_width, 2, "gtk_tree_view_column_set_min_width", "gint");
  gtk_tree_view_column_set_min_width(XEN_TO_C_GtkTreeViewColumn_(tree_column), XEN_TO_C_gint(min_width));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tree_view_column_get_min_width(XEN tree_column)
{
  #define H_gtk_tree_view_column_get_min_width "gint gtk_tree_view_column_get_min_width(GtkTreeViewColumn* tree_column)"
  XEN_ASSERT_TYPE(XEN_GtkTreeViewColumn__P(tree_column), tree_column, 1, "gtk_tree_view_column_get_min_width", "GtkTreeViewColumn*");
  return(C_TO_XEN_gint(gtk_tree_view_column_get_min_width(XEN_TO_C_GtkTreeViewColumn_(tree_column))));
}

static XEN gxg_gtk_tree_view_column_set_max_width(XEN tree_column, XEN max_width)
{
  #define H_gtk_tree_view_column_set_max_width "void gtk_tree_view_column_set_max_width(GtkTreeViewColumn* tree_column, \
gint max_width)"
  XEN_ASSERT_TYPE(XEN_GtkTreeViewColumn__P(tree_column), tree_column, 1, "gtk_tree_view_column_set_max_width", "GtkTreeViewColumn*");
  XEN_ASSERT_TYPE(XEN_gint_P(max_width), max_width, 2, "gtk_tree_view_column_set_max_width", "gint");
  gtk_tree_view_column_set_max_width(XEN_TO_C_GtkTreeViewColumn_(tree_column), XEN_TO_C_gint(max_width));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tree_view_column_get_max_width(XEN tree_column)
{
  #define H_gtk_tree_view_column_get_max_width "gint gtk_tree_view_column_get_max_width(GtkTreeViewColumn* tree_column)"
  XEN_ASSERT_TYPE(XEN_GtkTreeViewColumn__P(tree_column), tree_column, 1, "gtk_tree_view_column_get_max_width", "GtkTreeViewColumn*");
  return(C_TO_XEN_gint(gtk_tree_view_column_get_max_width(XEN_TO_C_GtkTreeViewColumn_(tree_column))));
}

static XEN gxg_gtk_tree_view_column_clicked(XEN tree_column)
{
  #define H_gtk_tree_view_column_clicked "void gtk_tree_view_column_clicked(GtkTreeViewColumn* tree_column)"
  XEN_ASSERT_TYPE(XEN_GtkTreeViewColumn__P(tree_column), tree_column, 1, "gtk_tree_view_column_clicked", "GtkTreeViewColumn*");
  gtk_tree_view_column_clicked(XEN_TO_C_GtkTreeViewColumn_(tree_column));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tree_view_column_set_title(XEN tree_column, XEN title)
{
  #define H_gtk_tree_view_column_set_title "void gtk_tree_view_column_set_title(GtkTreeViewColumn* tree_column, \
gchar* title)"
  XEN_ASSERT_TYPE(XEN_GtkTreeViewColumn__P(tree_column), tree_column, 1, "gtk_tree_view_column_set_title", "GtkTreeViewColumn*");
  XEN_ASSERT_TYPE(XEN_gchar__P(title), title, 2, "gtk_tree_view_column_set_title", "gchar*");
  gtk_tree_view_column_set_title(XEN_TO_C_GtkTreeViewColumn_(tree_column), XEN_TO_C_gchar_(title));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tree_view_column_get_title(XEN tree_column)
{
  #define H_gtk_tree_view_column_get_title "gchar* gtk_tree_view_column_get_title(GtkTreeViewColumn* tree_column)"
  XEN_ASSERT_TYPE(XEN_GtkTreeViewColumn__P(tree_column), tree_column, 1, "gtk_tree_view_column_get_title", "GtkTreeViewColumn*");
  return(C_TO_XEN_gchar_(gtk_tree_view_column_get_title(XEN_TO_C_GtkTreeViewColumn_(tree_column))));
}

static XEN gxg_gtk_tree_view_column_set_clickable(XEN tree_column, XEN clickable)
{
  #define H_gtk_tree_view_column_set_clickable "void gtk_tree_view_column_set_clickable(GtkTreeViewColumn* tree_column, \
gboolean clickable)"
  XEN_ASSERT_TYPE(XEN_GtkTreeViewColumn__P(tree_column), tree_column, 1, "gtk_tree_view_column_set_clickable", "GtkTreeViewColumn*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(clickable), clickable, 2, "gtk_tree_view_column_set_clickable", "gboolean");
  gtk_tree_view_column_set_clickable(XEN_TO_C_GtkTreeViewColumn_(tree_column), XEN_TO_C_gboolean(clickable));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tree_view_column_get_clickable(XEN tree_column)
{
  #define H_gtk_tree_view_column_get_clickable "gboolean gtk_tree_view_column_get_clickable(GtkTreeViewColumn* tree_column)"
  XEN_ASSERT_TYPE(XEN_GtkTreeViewColumn__P(tree_column), tree_column, 1, "gtk_tree_view_column_get_clickable", "GtkTreeViewColumn*");
  return(C_TO_XEN_gboolean(gtk_tree_view_column_get_clickable(XEN_TO_C_GtkTreeViewColumn_(tree_column))));
}

static XEN gxg_gtk_tree_view_column_set_widget(XEN tree_column, XEN widget)
{
  #define H_gtk_tree_view_column_set_widget "void gtk_tree_view_column_set_widget(GtkTreeViewColumn* tree_column, \
GtkWidget* widget)"
  XEN_ASSERT_TYPE(XEN_GtkTreeViewColumn__P(tree_column), tree_column, 1, "gtk_tree_view_column_set_widget", "GtkTreeViewColumn*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget) || XEN_FALSE_P(widget), widget, 2, "gtk_tree_view_column_set_widget", "GtkWidget*");
  gtk_tree_view_column_set_widget(XEN_TO_C_GtkTreeViewColumn_(tree_column), XEN_TO_C_GtkWidget_(widget));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tree_view_column_get_widget(XEN tree_column)
{
  #define H_gtk_tree_view_column_get_widget "GtkWidget* gtk_tree_view_column_get_widget(GtkTreeViewColumn* tree_column)"
  XEN_ASSERT_TYPE(XEN_GtkTreeViewColumn__P(tree_column), tree_column, 1, "gtk_tree_view_column_get_widget", "GtkTreeViewColumn*");
  return(C_TO_XEN_GtkWidget_(gtk_tree_view_column_get_widget(XEN_TO_C_GtkTreeViewColumn_(tree_column))));
}

static XEN gxg_gtk_tree_view_column_set_alignment(XEN tree_column, XEN xalign)
{
  #define H_gtk_tree_view_column_set_alignment "void gtk_tree_view_column_set_alignment(GtkTreeViewColumn* tree_column, \
gfloat xalign)"
  XEN_ASSERT_TYPE(XEN_GtkTreeViewColumn__P(tree_column), tree_column, 1, "gtk_tree_view_column_set_alignment", "GtkTreeViewColumn*");
  XEN_ASSERT_TYPE(XEN_gfloat_P(xalign), xalign, 2, "gtk_tree_view_column_set_alignment", "gfloat");
  gtk_tree_view_column_set_alignment(XEN_TO_C_GtkTreeViewColumn_(tree_column), XEN_TO_C_gfloat(xalign));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tree_view_column_get_alignment(XEN tree_column)
{
  #define H_gtk_tree_view_column_get_alignment "gfloat gtk_tree_view_column_get_alignment(GtkTreeViewColumn* tree_column)"
  XEN_ASSERT_TYPE(XEN_GtkTreeViewColumn__P(tree_column), tree_column, 1, "gtk_tree_view_column_get_alignment", "GtkTreeViewColumn*");
  return(C_TO_XEN_gfloat(gtk_tree_view_column_get_alignment(XEN_TO_C_GtkTreeViewColumn_(tree_column))));
}

static XEN gxg_gtk_tree_view_column_set_reorderable(XEN tree_column, XEN reorderable)
{
  #define H_gtk_tree_view_column_set_reorderable "void gtk_tree_view_column_set_reorderable(GtkTreeViewColumn* tree_column, \
gboolean reorderable)"
  XEN_ASSERT_TYPE(XEN_GtkTreeViewColumn__P(tree_column), tree_column, 1, "gtk_tree_view_column_set_reorderable", "GtkTreeViewColumn*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(reorderable), reorderable, 2, "gtk_tree_view_column_set_reorderable", "gboolean");
  gtk_tree_view_column_set_reorderable(XEN_TO_C_GtkTreeViewColumn_(tree_column), XEN_TO_C_gboolean(reorderable));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tree_view_column_get_reorderable(XEN tree_column)
{
  #define H_gtk_tree_view_column_get_reorderable "gboolean gtk_tree_view_column_get_reorderable(GtkTreeViewColumn* tree_column)"
  XEN_ASSERT_TYPE(XEN_GtkTreeViewColumn__P(tree_column), tree_column, 1, "gtk_tree_view_column_get_reorderable", "GtkTreeViewColumn*");
  return(C_TO_XEN_gboolean(gtk_tree_view_column_get_reorderable(XEN_TO_C_GtkTreeViewColumn_(tree_column))));
}

static XEN gxg_gtk_tree_view_column_set_sort_column_id(XEN tree_column, XEN sort_column_id)
{
  #define H_gtk_tree_view_column_set_sort_column_id "void gtk_tree_view_column_set_sort_column_id(GtkTreeViewColumn* tree_column, \
gint sort_column_id)"
  XEN_ASSERT_TYPE(XEN_GtkTreeViewColumn__P(tree_column), tree_column, 1, "gtk_tree_view_column_set_sort_column_id", "GtkTreeViewColumn*");
  XEN_ASSERT_TYPE(XEN_gint_P(sort_column_id), sort_column_id, 2, "gtk_tree_view_column_set_sort_column_id", "gint");
  gtk_tree_view_column_set_sort_column_id(XEN_TO_C_GtkTreeViewColumn_(tree_column), XEN_TO_C_gint(sort_column_id));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tree_view_column_get_sort_column_id(XEN tree_column)
{
  #define H_gtk_tree_view_column_get_sort_column_id "gint gtk_tree_view_column_get_sort_column_id(GtkTreeViewColumn* tree_column)"
  XEN_ASSERT_TYPE(XEN_GtkTreeViewColumn__P(tree_column), tree_column, 1, "gtk_tree_view_column_get_sort_column_id", "GtkTreeViewColumn*");
  return(C_TO_XEN_gint(gtk_tree_view_column_get_sort_column_id(XEN_TO_C_GtkTreeViewColumn_(tree_column))));
}

static XEN gxg_gtk_tree_view_column_set_sort_indicator(XEN tree_column, XEN setting)
{
  #define H_gtk_tree_view_column_set_sort_indicator "void gtk_tree_view_column_set_sort_indicator(GtkTreeViewColumn* tree_column, \
gboolean setting)"
  XEN_ASSERT_TYPE(XEN_GtkTreeViewColumn__P(tree_column), tree_column, 1, "gtk_tree_view_column_set_sort_indicator", "GtkTreeViewColumn*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(setting), setting, 2, "gtk_tree_view_column_set_sort_indicator", "gboolean");
  gtk_tree_view_column_set_sort_indicator(XEN_TO_C_GtkTreeViewColumn_(tree_column), XEN_TO_C_gboolean(setting));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tree_view_column_get_sort_indicator(XEN tree_column)
{
  #define H_gtk_tree_view_column_get_sort_indicator "gboolean gtk_tree_view_column_get_sort_indicator(GtkTreeViewColumn* tree_column)"
  XEN_ASSERT_TYPE(XEN_GtkTreeViewColumn__P(tree_column), tree_column, 1, "gtk_tree_view_column_get_sort_indicator", "GtkTreeViewColumn*");
  return(C_TO_XEN_gboolean(gtk_tree_view_column_get_sort_indicator(XEN_TO_C_GtkTreeViewColumn_(tree_column))));
}

static XEN gxg_gtk_tree_view_column_set_sort_order(XEN tree_column, XEN order)
{
  #define H_gtk_tree_view_column_set_sort_order "void gtk_tree_view_column_set_sort_order(GtkTreeViewColumn* tree_column, \
GtkSortType order)"
  XEN_ASSERT_TYPE(XEN_GtkTreeViewColumn__P(tree_column), tree_column, 1, "gtk_tree_view_column_set_sort_order", "GtkTreeViewColumn*");
  XEN_ASSERT_TYPE(XEN_GtkSortType_P(order), order, 2, "gtk_tree_view_column_set_sort_order", "GtkSortType");
  gtk_tree_view_column_set_sort_order(XEN_TO_C_GtkTreeViewColumn_(tree_column), XEN_TO_C_GtkSortType(order));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tree_view_column_get_sort_order(XEN tree_column)
{
  #define H_gtk_tree_view_column_get_sort_order "GtkSortType gtk_tree_view_column_get_sort_order(GtkTreeViewColumn* tree_column)"
  XEN_ASSERT_TYPE(XEN_GtkTreeViewColumn__P(tree_column), tree_column, 1, "gtk_tree_view_column_get_sort_order", "GtkTreeViewColumn*");
  return(C_TO_XEN_GtkSortType(gtk_tree_view_column_get_sort_order(XEN_TO_C_GtkTreeViewColumn_(tree_column))));
}

static XEN gxg_gtk_tree_view_column_cell_set_cell_data(XEN tree_column, XEN tree_model, XEN iter, XEN is_expander, XEN is_expanded)
{
  #define H_gtk_tree_view_column_cell_set_cell_data "void gtk_tree_view_column_cell_set_cell_data(GtkTreeViewColumn* tree_column, \
GtkTreeModel* tree_model, GtkTreeIter* iter, gboolean is_expander, gboolean is_expanded)"
  XEN_ASSERT_TYPE(XEN_GtkTreeViewColumn__P(tree_column), tree_column, 1, "gtk_tree_view_column_cell_set_cell_data", "GtkTreeViewColumn*");
  XEN_ASSERT_TYPE(XEN_GtkTreeModel__P(tree_model), tree_model, 2, "gtk_tree_view_column_cell_set_cell_data", "GtkTreeModel*");
  XEN_ASSERT_TYPE(XEN_GtkTreeIter__P(iter), iter, 3, "gtk_tree_view_column_cell_set_cell_data", "GtkTreeIter*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(is_expander), is_expander, 4, "gtk_tree_view_column_cell_set_cell_data", "gboolean");
  XEN_ASSERT_TYPE(XEN_gboolean_P(is_expanded), is_expanded, 5, "gtk_tree_view_column_cell_set_cell_data", "gboolean");
  gtk_tree_view_column_cell_set_cell_data(XEN_TO_C_GtkTreeViewColumn_(tree_column), XEN_TO_C_GtkTreeModel_(tree_model), XEN_TO_C_GtkTreeIter_(iter), 
                                          XEN_TO_C_gboolean(is_expander), XEN_TO_C_gboolean(is_expanded));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tree_view_column_cell_get_size(XEN tree_column, XEN cell_area, XEN x_offset, XEN y_offset, XEN width, XEN height)
{
  #define H_gtk_tree_view_column_cell_get_size "void gtk_tree_view_column_cell_get_size(GtkTreeViewColumn* tree_column, \
GdkRectangle* cell_area, gint* x_offset, gint* y_offset, gint* width, gint* height)"
  XEN_ASSERT_TYPE(XEN_GtkTreeViewColumn__P(tree_column), tree_column, 1, "gtk_tree_view_column_cell_get_size", "GtkTreeViewColumn*");
  XEN_ASSERT_TYPE(XEN_GdkRectangle__P(cell_area), cell_area, 2, "gtk_tree_view_column_cell_get_size", "GdkRectangle*");
  XEN_ASSERT_TYPE(XEN_gint__P(x_offset), x_offset, 3, "gtk_tree_view_column_cell_get_size", "gint*");
  XEN_ASSERT_TYPE(XEN_gint__P(y_offset), y_offset, 4, "gtk_tree_view_column_cell_get_size", "gint*");
  XEN_ASSERT_TYPE(XEN_gint__P(width), width, 5, "gtk_tree_view_column_cell_get_size", "gint*");
  XEN_ASSERT_TYPE(XEN_gint__P(height), height, 6, "gtk_tree_view_column_cell_get_size", "gint*");
  gtk_tree_view_column_cell_get_size(XEN_TO_C_GtkTreeViewColumn_(tree_column), XEN_TO_C_GdkRectangle_(cell_area), XEN_TO_C_gint_(x_offset), 
                                     XEN_TO_C_gint_(y_offset), XEN_TO_C_gint_(width), XEN_TO_C_gint_(height));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tree_view_column_cell_is_visible(XEN tree_column)
{
  #define H_gtk_tree_view_column_cell_is_visible "gboolean gtk_tree_view_column_cell_is_visible(GtkTreeViewColumn* tree_column)"
  XEN_ASSERT_TYPE(XEN_GtkTreeViewColumn__P(tree_column), tree_column, 1, "gtk_tree_view_column_cell_is_visible", "GtkTreeViewColumn*");
  return(C_TO_XEN_gboolean(gtk_tree_view_column_cell_is_visible(XEN_TO_C_GtkTreeViewColumn_(tree_column))));
}

static XEN gxg_gtk_tree_view_get_type(void)
{
  #define H_gtk_tree_view_get_type "GtkType gtk_tree_view_get_type( void)"
  return(C_TO_XEN_GtkType(gtk_tree_view_get_type()));
}

static XEN gxg_gtk_tree_view_new(void)
{
  #define H_gtk_tree_view_new "GtkWidget* gtk_tree_view_new( void)"
  return(C_TO_XEN_GtkWidget_(gtk_tree_view_new()));
}

static XEN gxg_gtk_tree_view_new_with_model(XEN model)
{
  #define H_gtk_tree_view_new_with_model "GtkWidget* gtk_tree_view_new_with_model(GtkTreeModel* model)"
  XEN_ASSERT_TYPE(XEN_GtkTreeModel__P(model), model, 1, "gtk_tree_view_new_with_model", "GtkTreeModel*");
  return(C_TO_XEN_GtkWidget_(gtk_tree_view_new_with_model(XEN_TO_C_GtkTreeModel_(model))));
}

static XEN gxg_gtk_tree_view_get_model(XEN tree_view)
{
  #define H_gtk_tree_view_get_model "GtkTreeModel* gtk_tree_view_get_model(GtkTreeView* tree_view)"
  XEN_ASSERT_TYPE(XEN_GtkTreeView__P(tree_view), tree_view, 1, "gtk_tree_view_get_model", "GtkTreeView*");
  return(C_TO_XEN_GtkTreeModel_(gtk_tree_view_get_model(XEN_TO_C_GtkTreeView_(tree_view))));
}

static XEN gxg_gtk_tree_view_set_model(XEN tree_view, XEN model)
{
  #define H_gtk_tree_view_set_model "void gtk_tree_view_set_model(GtkTreeView* tree_view, GtkTreeModel* model)"
  XEN_ASSERT_TYPE(XEN_GtkTreeView__P(tree_view), tree_view, 1, "gtk_tree_view_set_model", "GtkTreeView*");
  XEN_ASSERT_TYPE(XEN_GtkTreeModel__P(model) || XEN_FALSE_P(model), model, 2, "gtk_tree_view_set_model", "GtkTreeModel*");
  gtk_tree_view_set_model(XEN_TO_C_GtkTreeView_(tree_view), XEN_TO_C_GtkTreeModel_(model));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tree_view_get_selection(XEN tree_view)
{
  #define H_gtk_tree_view_get_selection "GtkTreeSelection* gtk_tree_view_get_selection(GtkTreeView* tree_view)"
  XEN_ASSERT_TYPE(XEN_GtkTreeView__P(tree_view), tree_view, 1, "gtk_tree_view_get_selection", "GtkTreeView*");
  return(C_TO_XEN_GtkTreeSelection_(gtk_tree_view_get_selection(XEN_TO_C_GtkTreeView_(tree_view))));
}

static XEN gxg_gtk_tree_view_get_hadjustment(XEN tree_view)
{
  #define H_gtk_tree_view_get_hadjustment "GtkAdjustment* gtk_tree_view_get_hadjustment(GtkTreeView* tree_view)"
  XEN_ASSERT_TYPE(XEN_GtkTreeView__P(tree_view), tree_view, 1, "gtk_tree_view_get_hadjustment", "GtkTreeView*");
  return(C_TO_XEN_GtkAdjustment_(gtk_tree_view_get_hadjustment(XEN_TO_C_GtkTreeView_(tree_view))));
}

static XEN gxg_gtk_tree_view_set_hadjustment(XEN tree_view, XEN adjustment)
{
  #define H_gtk_tree_view_set_hadjustment "void gtk_tree_view_set_hadjustment(GtkTreeView* tree_view, \
GtkAdjustment* adjustment)"
  XEN_ASSERT_TYPE(XEN_GtkTreeView__P(tree_view), tree_view, 1, "gtk_tree_view_set_hadjustment", "GtkTreeView*");
  XEN_ASSERT_TYPE(XEN_GtkAdjustment__P(adjustment) || XEN_FALSE_P(adjustment), adjustment, 2, "gtk_tree_view_set_hadjustment", "GtkAdjustment*");
  gtk_tree_view_set_hadjustment(XEN_TO_C_GtkTreeView_(tree_view), XEN_TO_C_GtkAdjustment_(adjustment));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tree_view_get_vadjustment(XEN tree_view)
{
  #define H_gtk_tree_view_get_vadjustment "GtkAdjustment* gtk_tree_view_get_vadjustment(GtkTreeView* tree_view)"
  XEN_ASSERT_TYPE(XEN_GtkTreeView__P(tree_view), tree_view, 1, "gtk_tree_view_get_vadjustment", "GtkTreeView*");
  return(C_TO_XEN_GtkAdjustment_(gtk_tree_view_get_vadjustment(XEN_TO_C_GtkTreeView_(tree_view))));
}

static XEN gxg_gtk_tree_view_set_vadjustment(XEN tree_view, XEN adjustment)
{
  #define H_gtk_tree_view_set_vadjustment "void gtk_tree_view_set_vadjustment(GtkTreeView* tree_view, \
GtkAdjustment* adjustment)"
  XEN_ASSERT_TYPE(XEN_GtkTreeView__P(tree_view), tree_view, 1, "gtk_tree_view_set_vadjustment", "GtkTreeView*");
  XEN_ASSERT_TYPE(XEN_GtkAdjustment__P(adjustment) || XEN_FALSE_P(adjustment), adjustment, 2, "gtk_tree_view_set_vadjustment", "GtkAdjustment*");
  gtk_tree_view_set_vadjustment(XEN_TO_C_GtkTreeView_(tree_view), XEN_TO_C_GtkAdjustment_(adjustment));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tree_view_get_headers_visible(XEN tree_view)
{
  #define H_gtk_tree_view_get_headers_visible "gboolean gtk_tree_view_get_headers_visible(GtkTreeView* tree_view)"
  XEN_ASSERT_TYPE(XEN_GtkTreeView__P(tree_view), tree_view, 1, "gtk_tree_view_get_headers_visible", "GtkTreeView*");
  return(C_TO_XEN_gboolean(gtk_tree_view_get_headers_visible(XEN_TO_C_GtkTreeView_(tree_view))));
}

static XEN gxg_gtk_tree_view_set_headers_visible(XEN tree_view, XEN headers_visible)
{
  #define H_gtk_tree_view_set_headers_visible "void gtk_tree_view_set_headers_visible(GtkTreeView* tree_view, \
gboolean headers_visible)"
  XEN_ASSERT_TYPE(XEN_GtkTreeView__P(tree_view), tree_view, 1, "gtk_tree_view_set_headers_visible", "GtkTreeView*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(headers_visible), headers_visible, 2, "gtk_tree_view_set_headers_visible", "gboolean");
  gtk_tree_view_set_headers_visible(XEN_TO_C_GtkTreeView_(tree_view), XEN_TO_C_gboolean(headers_visible));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tree_view_columns_autosize(XEN tree_view)
{
  #define H_gtk_tree_view_columns_autosize "void gtk_tree_view_columns_autosize(GtkTreeView* tree_view)"
  XEN_ASSERT_TYPE(XEN_GtkTreeView__P(tree_view), tree_view, 1, "gtk_tree_view_columns_autosize", "GtkTreeView*");
  gtk_tree_view_columns_autosize(XEN_TO_C_GtkTreeView_(tree_view));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tree_view_set_headers_clickable(XEN tree_view, XEN setting)
{
  #define H_gtk_tree_view_set_headers_clickable "void gtk_tree_view_set_headers_clickable(GtkTreeView* tree_view, \
gboolean setting)"
  XEN_ASSERT_TYPE(XEN_GtkTreeView__P(tree_view), tree_view, 1, "gtk_tree_view_set_headers_clickable", "GtkTreeView*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(setting), setting, 2, "gtk_tree_view_set_headers_clickable", "gboolean");
  gtk_tree_view_set_headers_clickable(XEN_TO_C_GtkTreeView_(tree_view), XEN_TO_C_gboolean(setting));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tree_view_set_rules_hint(XEN tree_view, XEN setting)
{
  #define H_gtk_tree_view_set_rules_hint "void gtk_tree_view_set_rules_hint(GtkTreeView* tree_view, gboolean setting)"
  XEN_ASSERT_TYPE(XEN_GtkTreeView__P(tree_view), tree_view, 1, "gtk_tree_view_set_rules_hint", "GtkTreeView*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(setting), setting, 2, "gtk_tree_view_set_rules_hint", "gboolean");
  gtk_tree_view_set_rules_hint(XEN_TO_C_GtkTreeView_(tree_view), XEN_TO_C_gboolean(setting));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tree_view_get_rules_hint(XEN tree_view)
{
  #define H_gtk_tree_view_get_rules_hint "gboolean gtk_tree_view_get_rules_hint(GtkTreeView* tree_view)"
  XEN_ASSERT_TYPE(XEN_GtkTreeView__P(tree_view), tree_view, 1, "gtk_tree_view_get_rules_hint", "GtkTreeView*");
  return(C_TO_XEN_gboolean(gtk_tree_view_get_rules_hint(XEN_TO_C_GtkTreeView_(tree_view))));
}

static XEN gxg_gtk_tree_view_append_column(XEN tree_view, XEN column)
{
  #define H_gtk_tree_view_append_column "gint gtk_tree_view_append_column(GtkTreeView* tree_view, GtkTreeViewColumn* column)"
  XEN_ASSERT_TYPE(XEN_GtkTreeView__P(tree_view), tree_view, 1, "gtk_tree_view_append_column", "GtkTreeView*");
  XEN_ASSERT_TYPE(XEN_GtkTreeViewColumn__P(column), column, 2, "gtk_tree_view_append_column", "GtkTreeViewColumn*");
  return(C_TO_XEN_gint(gtk_tree_view_append_column(XEN_TO_C_GtkTreeView_(tree_view), XEN_TO_C_GtkTreeViewColumn_(column))));
}

static XEN gxg_gtk_tree_view_remove_column(XEN tree_view, XEN column)
{
  #define H_gtk_tree_view_remove_column "gint gtk_tree_view_remove_column(GtkTreeView* tree_view, GtkTreeViewColumn* column)"
  XEN_ASSERT_TYPE(XEN_GtkTreeView__P(tree_view), tree_view, 1, "gtk_tree_view_remove_column", "GtkTreeView*");
  XEN_ASSERT_TYPE(XEN_GtkTreeViewColumn__P(column), column, 2, "gtk_tree_view_remove_column", "GtkTreeViewColumn*");
  return(C_TO_XEN_gint(gtk_tree_view_remove_column(XEN_TO_C_GtkTreeView_(tree_view), XEN_TO_C_GtkTreeViewColumn_(column))));
}

static XEN gxg_gtk_tree_view_insert_column(XEN tree_view, XEN column, XEN position)
{
  #define H_gtk_tree_view_insert_column "gint gtk_tree_view_insert_column(GtkTreeView* tree_view, GtkTreeViewColumn* column, \
gint position)"
  XEN_ASSERT_TYPE(XEN_GtkTreeView__P(tree_view), tree_view, 1, "gtk_tree_view_insert_column", "GtkTreeView*");
  XEN_ASSERT_TYPE(XEN_GtkTreeViewColumn__P(column), column, 2, "gtk_tree_view_insert_column", "GtkTreeViewColumn*");
  XEN_ASSERT_TYPE(XEN_gint_P(position), position, 3, "gtk_tree_view_insert_column", "gint");
  return(C_TO_XEN_gint(gtk_tree_view_insert_column(XEN_TO_C_GtkTreeView_(tree_view), XEN_TO_C_GtkTreeViewColumn_(column), 
                                                   XEN_TO_C_gint(position))));
}

static XEN gxg_gtk_tree_view_insert_column_with_attributes(XEN tree_view, XEN position, XEN title, XEN cell)
{
  #define H_gtk_tree_view_insert_column_with_attributes "gint gtk_tree_view_insert_column_with_attributes(GtkTreeView* tree_view, \
gint position, gchar* title, GtkCellRenderer* cell, ...)"
  XEN_ASSERT_TYPE(XEN_GtkTreeView__P(tree_view), tree_view, 1, "gtk_tree_view_insert_column_with_attributes", "GtkTreeView*");
  XEN_ASSERT_TYPE(XEN_gint_P(position), position, 2, "gtk_tree_view_insert_column_with_attributes", "gint");
  XEN_ASSERT_TYPE(XEN_gchar__P(title), title, 3, "gtk_tree_view_insert_column_with_attributes", "gchar*");
  XEN_ASSERT_TYPE(XEN_GtkCellRenderer__P(cell), cell, 4, "gtk_tree_view_insert_column_with_attributes", "GtkCellRenderer*");
  return(C_TO_XEN_gint(gtk_tree_view_insert_column_with_attributes(XEN_TO_C_GtkTreeView_(tree_view), XEN_TO_C_gint(position), 
                                                                   XEN_TO_C_gchar_(title), XEN_TO_C_GtkCellRenderer_(cell))));
}

static XEN gxg_gtk_tree_view_insert_column_with_data_func(XEN tree_view, XEN position, XEN title, XEN cell, XEN func, XEN func_data, XEN dnotify)
{
  #define H_gtk_tree_view_insert_column_with_data_func "gint gtk_tree_view_insert_column_with_data_func(GtkTreeView* tree_view, \
gint position, gchar* title, GtkCellRenderer* cell, GtkTreeCellDataFunc func, lambda_data func_data, \
GDestroyNotify dnotify)"
  XEN_ASSERT_TYPE(XEN_GtkTreeView__P(tree_view), tree_view, 1, "gtk_tree_view_insert_column_with_data_func", "GtkTreeView*");
  XEN_ASSERT_TYPE(XEN_gint_P(position), position, 2, "gtk_tree_view_insert_column_with_data_func", "gint");
  XEN_ASSERT_TYPE(XEN_gchar__P(title), title, 3, "gtk_tree_view_insert_column_with_data_func", "gchar*");
  XEN_ASSERT_TYPE(XEN_GtkCellRenderer__P(cell), cell, 4, "gtk_tree_view_insert_column_with_data_func", "GtkCellRenderer*");
  XEN_ASSERT_TYPE(XEN_GtkTreeCellDataFunc_P(func), func, 5, "gtk_tree_view_insert_column_with_data_func", "GtkTreeCellDataFunc");
  XEN_ASSERT_TYPE(XEN_lambda_data_P(func_data), func_data, 6, "gtk_tree_view_insert_column_with_data_func", "lambda_data");
  XEN_ASSERT_TYPE(XEN_GDestroyNotify_P(dnotify), dnotify, 7, "gtk_tree_view_insert_column_with_data_func", "GDestroyNotify");
  {
    XEN result = XEN_FALSE;
    XEN gxg_ptr = XEN_LIST_5(func, func_data, XEN_FALSE, XEN_FALSE, XEN_FALSE);
    xm_protect(gxg_ptr);
    result = C_TO_XEN_gint(gtk_tree_view_insert_column_with_data_func(XEN_TO_C_GtkTreeView_(tree_view), XEN_TO_C_gint(position), 
                                                                      XEN_TO_C_gchar_(title), XEN_TO_C_GtkCellRenderer_(cell), 
                                                                      XEN_TO_C_GtkTreeCellDataFunc(func), XEN_TO_C_lambda_data(func_data), 
                                                                      XEN_TO_C_GDestroyNotify(dnotify)));
    return(result);
   }
}

static XEN gxg_gtk_tree_view_get_column(XEN tree_view, XEN n)
{
  #define H_gtk_tree_view_get_column "GtkTreeViewColumn* gtk_tree_view_get_column(GtkTreeView* tree_view, \
gint n)"
  XEN_ASSERT_TYPE(XEN_GtkTreeView__P(tree_view), tree_view, 1, "gtk_tree_view_get_column", "GtkTreeView*");
  XEN_ASSERT_TYPE(XEN_gint_P(n), n, 2, "gtk_tree_view_get_column", "gint");
  return(C_TO_XEN_GtkTreeViewColumn_(gtk_tree_view_get_column(XEN_TO_C_GtkTreeView_(tree_view), XEN_TO_C_gint(n))));
}

static XEN gxg_gtk_tree_view_get_columns(XEN tree_view)
{
  #define H_gtk_tree_view_get_columns "GList* gtk_tree_view_get_columns(GtkTreeView* tree_view)"
  XEN_ASSERT_TYPE(XEN_GtkTreeView__P(tree_view), tree_view, 1, "gtk_tree_view_get_columns", "GtkTreeView*");
  return(C_TO_XEN_GList_(gtk_tree_view_get_columns(XEN_TO_C_GtkTreeView_(tree_view))));
}

static XEN gxg_gtk_tree_view_move_column_after(XEN tree_view, XEN column, XEN base_column)
{
  #define H_gtk_tree_view_move_column_after "void gtk_tree_view_move_column_after(GtkTreeView* tree_view, \
GtkTreeViewColumn* column, GtkTreeViewColumn* base_column)"
  XEN_ASSERT_TYPE(XEN_GtkTreeView__P(tree_view), tree_view, 1, "gtk_tree_view_move_column_after", "GtkTreeView*");
  XEN_ASSERT_TYPE(XEN_GtkTreeViewColumn__P(column), column, 2, "gtk_tree_view_move_column_after", "GtkTreeViewColumn*");
  XEN_ASSERT_TYPE(XEN_GtkTreeViewColumn__P(base_column) || XEN_FALSE_P(base_column), base_column, 3, "gtk_tree_view_move_column_after", "GtkTreeViewColumn*");
  gtk_tree_view_move_column_after(XEN_TO_C_GtkTreeView_(tree_view), XEN_TO_C_GtkTreeViewColumn_(column), XEN_TO_C_GtkTreeViewColumn_(base_column));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tree_view_set_expander_column(XEN tree_view, XEN column)
{
  #define H_gtk_tree_view_set_expander_column "void gtk_tree_view_set_expander_column(GtkTreeView* tree_view, \
GtkTreeViewColumn* column)"
  XEN_ASSERT_TYPE(XEN_GtkTreeView__P(tree_view), tree_view, 1, "gtk_tree_view_set_expander_column", "GtkTreeView*");
  XEN_ASSERT_TYPE(XEN_GtkTreeViewColumn__P(column) || XEN_FALSE_P(column), column, 2, "gtk_tree_view_set_expander_column", "GtkTreeViewColumn*");
  gtk_tree_view_set_expander_column(XEN_TO_C_GtkTreeView_(tree_view), XEN_TO_C_GtkTreeViewColumn_(column));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tree_view_get_expander_column(XEN tree_view)
{
  #define H_gtk_tree_view_get_expander_column "GtkTreeViewColumn* gtk_tree_view_get_expander_column(GtkTreeView* tree_view)"
  XEN_ASSERT_TYPE(XEN_GtkTreeView__P(tree_view), tree_view, 1, "gtk_tree_view_get_expander_column", "GtkTreeView*");
  return(C_TO_XEN_GtkTreeViewColumn_(gtk_tree_view_get_expander_column(XEN_TO_C_GtkTreeView_(tree_view))));
}

static XEN gxg_gtk_tree_view_set_column_drag_function(XEN tree_view, XEN func, XEN func_data, XEN destroy)
{
  #define H_gtk_tree_view_set_column_drag_function "void gtk_tree_view_set_column_drag_function(GtkTreeView* tree_view, \
GtkTreeViewColumnDropFunc func, lambda_data func_data, GtkDestroyNotify destroy)"
  XEN_ASSERT_TYPE(XEN_GtkTreeView__P(tree_view), tree_view, 1, "gtk_tree_view_set_column_drag_function", "GtkTreeView*");
  XEN_ASSERT_TYPE(XEN_GtkTreeViewColumnDropFunc_P(func), func, 2, "gtk_tree_view_set_column_drag_function", "GtkTreeViewColumnDropFunc");
  XEN_ASSERT_TYPE(XEN_lambda_data_P(func_data), func_data, 3, "gtk_tree_view_set_column_drag_function", "lambda_data");
  XEN_ASSERT_TYPE(XEN_GtkDestroyNotify_P(destroy), destroy, 4, "gtk_tree_view_set_column_drag_function", "GtkDestroyNotify");
  {
    int loc;
    XEN gxg_ptr = XEN_LIST_5(func, func_data, XEN_FALSE, XEN_FALSE, XEN_FALSE);
    loc = xm_protect(gxg_ptr);
    XEN_LIST_SET(gxg_ptr, 2, C_TO_XEN_INT(loc));
    XEN_LIST_SET(gxg_ptr, 3, destroy);
    gtk_tree_view_set_column_drag_function(XEN_TO_C_GtkTreeView_(tree_view), XEN_TO_C_GtkTreeViewColumnDropFunc(func), XEN_TO_C_lambda_data(func_data), 
                                       XEN_TO_C_GtkDestroyNotify(destroy));
    xm_unprotect_at(loc);
    return(XEN_FALSE);
   }
}

static XEN gxg_gtk_tree_view_scroll_to_point(XEN tree_view, XEN tree_x, XEN tree_y)
{
  #define H_gtk_tree_view_scroll_to_point "void gtk_tree_view_scroll_to_point(GtkTreeView* tree_view, \
gint tree_x, gint tree_y)"
  XEN_ASSERT_TYPE(XEN_GtkTreeView__P(tree_view), tree_view, 1, "gtk_tree_view_scroll_to_point", "GtkTreeView*");
  XEN_ASSERT_TYPE(XEN_gint_P(tree_x), tree_x, 2, "gtk_tree_view_scroll_to_point", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(tree_y), tree_y, 3, "gtk_tree_view_scroll_to_point", "gint");
  gtk_tree_view_scroll_to_point(XEN_TO_C_GtkTreeView_(tree_view), XEN_TO_C_gint(tree_x), XEN_TO_C_gint(tree_y));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tree_view_scroll_to_cell(XEN tree_view, XEN path, XEN column, XEN use_align, XEN row_align, XEN col_align)
{
  #define H_gtk_tree_view_scroll_to_cell "void gtk_tree_view_scroll_to_cell(GtkTreeView* tree_view, GtkTreePath* path, \
GtkTreeViewColumn* column, gboolean use_align, gfloat row_align, gfloat col_align)"
  XEN_ASSERT_TYPE(XEN_GtkTreeView__P(tree_view), tree_view, 1, "gtk_tree_view_scroll_to_cell", "GtkTreeView*");
  XEN_ASSERT_TYPE(XEN_GtkTreePath__P(path) || XEN_FALSE_P(path), path, 2, "gtk_tree_view_scroll_to_cell", "GtkTreePath*");
  XEN_ASSERT_TYPE(XEN_GtkTreeViewColumn__P(column) || XEN_FALSE_P(column), column, 3, "gtk_tree_view_scroll_to_cell", "GtkTreeViewColumn*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(use_align), use_align, 4, "gtk_tree_view_scroll_to_cell", "gboolean");
  XEN_ASSERT_TYPE(XEN_gfloat_P(row_align), row_align, 5, "gtk_tree_view_scroll_to_cell", "gfloat");
  XEN_ASSERT_TYPE(XEN_gfloat_P(col_align), col_align, 6, "gtk_tree_view_scroll_to_cell", "gfloat");
  gtk_tree_view_scroll_to_cell(XEN_TO_C_GtkTreeView_(tree_view), XEN_TO_C_GtkTreePath_(path), XEN_TO_C_GtkTreeViewColumn_(column), 
                               XEN_TO_C_gboolean(use_align), XEN_TO_C_gfloat(row_align), XEN_TO_C_gfloat(col_align));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tree_view_row_activated(XEN tree_view, XEN path, XEN column)
{
  #define H_gtk_tree_view_row_activated "void gtk_tree_view_row_activated(GtkTreeView* tree_view, GtkTreePath* path, \
GtkTreeViewColumn* column)"
  XEN_ASSERT_TYPE(XEN_GtkTreeView__P(tree_view), tree_view, 1, "gtk_tree_view_row_activated", "GtkTreeView*");
  XEN_ASSERT_TYPE(XEN_GtkTreePath__P(path), path, 2, "gtk_tree_view_row_activated", "GtkTreePath*");
  XEN_ASSERT_TYPE(XEN_GtkTreeViewColumn__P(column), column, 3, "gtk_tree_view_row_activated", "GtkTreeViewColumn*");
  gtk_tree_view_row_activated(XEN_TO_C_GtkTreeView_(tree_view), XEN_TO_C_GtkTreePath_(path), XEN_TO_C_GtkTreeViewColumn_(column));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tree_view_expand_all(XEN tree_view)
{
  #define H_gtk_tree_view_expand_all "void gtk_tree_view_expand_all(GtkTreeView* tree_view)"
  XEN_ASSERT_TYPE(XEN_GtkTreeView__P(tree_view), tree_view, 1, "gtk_tree_view_expand_all", "GtkTreeView*");
  gtk_tree_view_expand_all(XEN_TO_C_GtkTreeView_(tree_view));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tree_view_collapse_all(XEN tree_view)
{
  #define H_gtk_tree_view_collapse_all "void gtk_tree_view_collapse_all(GtkTreeView* tree_view)"
  XEN_ASSERT_TYPE(XEN_GtkTreeView__P(tree_view), tree_view, 1, "gtk_tree_view_collapse_all", "GtkTreeView*");
  gtk_tree_view_collapse_all(XEN_TO_C_GtkTreeView_(tree_view));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tree_view_expand_row(XEN tree_view, XEN path, XEN open_all)
{
  #define H_gtk_tree_view_expand_row "gboolean gtk_tree_view_expand_row(GtkTreeView* tree_view, GtkTreePath* path, \
gboolean open_all)"
  XEN_ASSERT_TYPE(XEN_GtkTreeView__P(tree_view), tree_view, 1, "gtk_tree_view_expand_row", "GtkTreeView*");
  XEN_ASSERT_TYPE(XEN_GtkTreePath__P(path), path, 2, "gtk_tree_view_expand_row", "GtkTreePath*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(open_all), open_all, 3, "gtk_tree_view_expand_row", "gboolean");
  return(C_TO_XEN_gboolean(gtk_tree_view_expand_row(XEN_TO_C_GtkTreeView_(tree_view), XEN_TO_C_GtkTreePath_(path), XEN_TO_C_gboolean(open_all))));
}

static XEN gxg_gtk_tree_view_collapse_row(XEN tree_view, XEN path)
{
  #define H_gtk_tree_view_collapse_row "gboolean gtk_tree_view_collapse_row(GtkTreeView* tree_view, GtkTreePath* path)"
  XEN_ASSERT_TYPE(XEN_GtkTreeView__P(tree_view), tree_view, 1, "gtk_tree_view_collapse_row", "GtkTreeView*");
  XEN_ASSERT_TYPE(XEN_GtkTreePath__P(path), path, 2, "gtk_tree_view_collapse_row", "GtkTreePath*");
  return(C_TO_XEN_gboolean(gtk_tree_view_collapse_row(XEN_TO_C_GtkTreeView_(tree_view), XEN_TO_C_GtkTreePath_(path))));
}

static XEN gxg_gtk_tree_view_map_expanded_rows(XEN tree_view, XEN func, XEN func_data)
{
  #define H_gtk_tree_view_map_expanded_rows "void gtk_tree_view_map_expanded_rows(GtkTreeView* tree_view, \
GtkTreeViewMappingFunc func, lambda_data func_data)"
  XEN_ASSERT_TYPE(XEN_GtkTreeView__P(tree_view), tree_view, 1, "gtk_tree_view_map_expanded_rows", "GtkTreeView*");
  XEN_ASSERT_TYPE(XEN_GtkTreeViewMappingFunc_P(func), func, 2, "gtk_tree_view_map_expanded_rows", "GtkTreeViewMappingFunc");
  if (XEN_NOT_BOUND_P(func_data)) func_data = XEN_FALSE; 
  else XEN_ASSERT_TYPE(XEN_lambda_data_P(func_data), func_data, 3, "gtk_tree_view_map_expanded_rows", "lambda_data");
  {
    int loc;
    XEN gxg_ptr = XEN_LIST_5(func, func_data, XEN_FALSE, XEN_FALSE, XEN_FALSE);
    loc = xm_protect(gxg_ptr);
    XEN_LIST_SET(gxg_ptr, 2, C_TO_XEN_INT(loc));
    gtk_tree_view_map_expanded_rows(XEN_TO_C_GtkTreeView_(tree_view), XEN_TO_C_GtkTreeViewMappingFunc(func), XEN_TO_C_lambda_data(func_data));
    xm_unprotect_at(loc);
    return(XEN_FALSE);
   }
}

static XEN gxg_gtk_tree_view_row_expanded(XEN tree_view, XEN path)
{
  #define H_gtk_tree_view_row_expanded "gboolean gtk_tree_view_row_expanded(GtkTreeView* tree_view, GtkTreePath* path)"
  XEN_ASSERT_TYPE(XEN_GtkTreeView__P(tree_view), tree_view, 1, "gtk_tree_view_row_expanded", "GtkTreeView*");
  XEN_ASSERT_TYPE(XEN_GtkTreePath__P(path), path, 2, "gtk_tree_view_row_expanded", "GtkTreePath*");
  return(C_TO_XEN_gboolean(gtk_tree_view_row_expanded(XEN_TO_C_GtkTreeView_(tree_view), XEN_TO_C_GtkTreePath_(path))));
}

static XEN gxg_gtk_tree_view_set_reorderable(XEN tree_view, XEN reorderable)
{
  #define H_gtk_tree_view_set_reorderable "void gtk_tree_view_set_reorderable(GtkTreeView* tree_view, \
gboolean reorderable)"
  XEN_ASSERT_TYPE(XEN_GtkTreeView__P(tree_view), tree_view, 1, "gtk_tree_view_set_reorderable", "GtkTreeView*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(reorderable), reorderable, 2, "gtk_tree_view_set_reorderable", "gboolean");
  gtk_tree_view_set_reorderable(XEN_TO_C_GtkTreeView_(tree_view), XEN_TO_C_gboolean(reorderable));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tree_view_get_reorderable(XEN tree_view)
{
  #define H_gtk_tree_view_get_reorderable "gboolean gtk_tree_view_get_reorderable(GtkTreeView* tree_view)"
  XEN_ASSERT_TYPE(XEN_GtkTreeView__P(tree_view), tree_view, 1, "gtk_tree_view_get_reorderable", "GtkTreeView*");
  return(C_TO_XEN_gboolean(gtk_tree_view_get_reorderable(XEN_TO_C_GtkTreeView_(tree_view))));
}

static XEN gxg_gtk_tree_view_set_cursor(XEN tree_view, XEN path, XEN focus_column, XEN start_editing)
{
  #define H_gtk_tree_view_set_cursor "void gtk_tree_view_set_cursor(GtkTreeView* tree_view, GtkTreePath* path, \
GtkTreeViewColumn* focus_column, gboolean start_editing)"
  XEN_ASSERT_TYPE(XEN_GtkTreeView__P(tree_view), tree_view, 1, "gtk_tree_view_set_cursor", "GtkTreeView*");
  XEN_ASSERT_TYPE(XEN_GtkTreePath__P(path), path, 2, "gtk_tree_view_set_cursor", "GtkTreePath*");
  XEN_ASSERT_TYPE(XEN_GtkTreeViewColumn__P(focus_column), focus_column, 3, "gtk_tree_view_set_cursor", "GtkTreeViewColumn*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(start_editing), start_editing, 4, "gtk_tree_view_set_cursor", "gboolean");
  gtk_tree_view_set_cursor(XEN_TO_C_GtkTreeView_(tree_view), XEN_TO_C_GtkTreePath_(path), XEN_TO_C_GtkTreeViewColumn_(focus_column), 
                           XEN_TO_C_gboolean(start_editing));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tree_view_get_cursor(XEN tree_view, XEN path, XEN focus_column)
{
  #define H_gtk_tree_view_get_cursor "void gtk_tree_view_get_cursor(GtkTreeView* tree_view, GtkTreePath** [path], \
GtkTreeViewColumn** [focus_column])"
  GtkTreePath* ref_path;
  GtkTreeViewColumn* ref_focus_column;
  XEN_ASSERT_TYPE(XEN_GtkTreeView__P(tree_view), tree_view, 1, "gtk_tree_view_get_cursor", "GtkTreeView*");
  gtk_tree_view_get_cursor(XEN_TO_C_GtkTreeView_(tree_view), &ref_path, &ref_focus_column);
  return(XEN_LIST_2(C_TO_XEN_GtkTreePath_(ref_path), C_TO_XEN_GtkTreeViewColumn_(ref_focus_column)));
}

static XEN gxg_gtk_tree_view_get_bin_window(XEN tree_view)
{
  #define H_gtk_tree_view_get_bin_window "GdkWindow* gtk_tree_view_get_bin_window(GtkTreeView* tree_view)"
  XEN_ASSERT_TYPE(XEN_GtkTreeView__P(tree_view), tree_view, 1, "gtk_tree_view_get_bin_window", "GtkTreeView*");
  return(C_TO_XEN_GdkWindow_(gtk_tree_view_get_bin_window(XEN_TO_C_GtkTreeView_(tree_view))));
}

static XEN gxg_gtk_tree_view_get_path_at_pos(XEN tree_view, XEN x, XEN y, XEN path, XEN column, XEN cell_x, XEN cell_y)
{
  #define H_gtk_tree_view_get_path_at_pos "gboolean gtk_tree_view_get_path_at_pos(GtkTreeView* tree_view, \
gint x, gint y, GtkTreePath** [path], GtkTreeViewColumn** [column], gint* [cell_x], gint* [cell_y])"
  GtkTreePath* ref_path;
  GtkTreeViewColumn* ref_column;
  gint ref_cell_x;
  gint ref_cell_y;
  XEN_ASSERT_TYPE(XEN_GtkTreeView__P(tree_view), tree_view, 1, "gtk_tree_view_get_path_at_pos", "GtkTreeView*");
  XEN_ASSERT_TYPE(XEN_gint_P(x), x, 2, "gtk_tree_view_get_path_at_pos", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(y), y, 3, "gtk_tree_view_get_path_at_pos", "gint");
  {
    XEN result = XEN_FALSE;
    result = C_TO_XEN_gboolean(gtk_tree_view_get_path_at_pos(XEN_TO_C_GtkTreeView_(tree_view), XEN_TO_C_gint(x), XEN_TO_C_gint(y), 
                                                             &ref_path, &ref_column, &ref_cell_x, &ref_cell_y));
    return(XEN_LIST_5(result, C_TO_XEN_GtkTreePath_(ref_path), C_TO_XEN_GtkTreeViewColumn_(ref_column), C_TO_XEN_gint(ref_cell_x), C_TO_XEN_gint(ref_cell_y)));
   }
}

static XEN gxg_gtk_tree_view_get_cell_area(XEN tree_view, XEN path, XEN column, XEN rect)
{
  #define H_gtk_tree_view_get_cell_area "void gtk_tree_view_get_cell_area(GtkTreeView* tree_view, GtkTreePath* path, \
GtkTreeViewColumn* column, GdkRectangle* rect)"
  XEN_ASSERT_TYPE(XEN_GtkTreeView__P(tree_view), tree_view, 1, "gtk_tree_view_get_cell_area", "GtkTreeView*");
  XEN_ASSERT_TYPE(XEN_GtkTreePath__P(path) || XEN_FALSE_P(path), path, 2, "gtk_tree_view_get_cell_area", "GtkTreePath*");
  XEN_ASSERT_TYPE(XEN_GtkTreeViewColumn__P(column) || XEN_FALSE_P(column), column, 3, "gtk_tree_view_get_cell_area", "GtkTreeViewColumn*");
  XEN_ASSERT_TYPE(XEN_GdkRectangle__P(rect), rect, 4, "gtk_tree_view_get_cell_area", "GdkRectangle*");
  gtk_tree_view_get_cell_area(XEN_TO_C_GtkTreeView_(tree_view), XEN_TO_C_GtkTreePath_(path), XEN_TO_C_GtkTreeViewColumn_(column), 
                              XEN_TO_C_GdkRectangle_(rect));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tree_view_get_background_area(XEN tree_view, XEN path, XEN column, XEN rect)
{
  #define H_gtk_tree_view_get_background_area "void gtk_tree_view_get_background_area(GtkTreeView* tree_view, \
GtkTreePath* path, GtkTreeViewColumn* column, GdkRectangle* rect)"
  XEN_ASSERT_TYPE(XEN_GtkTreeView__P(tree_view), tree_view, 1, "gtk_tree_view_get_background_area", "GtkTreeView*");
  XEN_ASSERT_TYPE(XEN_GtkTreePath__P(path) || XEN_FALSE_P(path), path, 2, "gtk_tree_view_get_background_area", "GtkTreePath*");
  XEN_ASSERT_TYPE(XEN_GtkTreeViewColumn__P(column) || XEN_FALSE_P(column), column, 3, "gtk_tree_view_get_background_area", "GtkTreeViewColumn*");
  XEN_ASSERT_TYPE(XEN_GdkRectangle__P(rect), rect, 4, "gtk_tree_view_get_background_area", "GdkRectangle*");
  gtk_tree_view_get_background_area(XEN_TO_C_GtkTreeView_(tree_view), XEN_TO_C_GtkTreePath_(path), XEN_TO_C_GtkTreeViewColumn_(column), 
                                    XEN_TO_C_GdkRectangle_(rect));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tree_view_get_visible_rect(XEN tree_view, XEN visible_rect)
{
  #define H_gtk_tree_view_get_visible_rect "void gtk_tree_view_get_visible_rect(GtkTreeView* tree_view, \
GdkRectangle* visible_rect)"
  XEN_ASSERT_TYPE(XEN_GtkTreeView__P(tree_view), tree_view, 1, "gtk_tree_view_get_visible_rect", "GtkTreeView*");
  XEN_ASSERT_TYPE(XEN_GdkRectangle__P(visible_rect), visible_rect, 2, "gtk_tree_view_get_visible_rect", "GdkRectangle*");
  gtk_tree_view_get_visible_rect(XEN_TO_C_GtkTreeView_(tree_view), XEN_TO_C_GdkRectangle_(visible_rect));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tree_view_widget_to_tree_coords(XEN tree_view, XEN wx, XEN wy, XEN tx, XEN ty)
{
  #define H_gtk_tree_view_widget_to_tree_coords "void gtk_tree_view_widget_to_tree_coords(GtkTreeView* tree_view, \
gint wx, gint wy, gint* [tx], gint* [ty])"
  gint ref_tx;
  gint ref_ty;
  XEN_ASSERT_TYPE(XEN_GtkTreeView__P(tree_view), tree_view, 1, "gtk_tree_view_widget_to_tree_coords", "GtkTreeView*");
  XEN_ASSERT_TYPE(XEN_gint_P(wx), wx, 2, "gtk_tree_view_widget_to_tree_coords", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(wy), wy, 3, "gtk_tree_view_widget_to_tree_coords", "gint");
  gtk_tree_view_widget_to_tree_coords(XEN_TO_C_GtkTreeView_(tree_view), XEN_TO_C_gint(wx), XEN_TO_C_gint(wy), &ref_tx, &ref_ty);
  return(XEN_LIST_2(C_TO_XEN_gint(ref_tx), C_TO_XEN_gint(ref_ty)));
}

static XEN gxg_gtk_tree_view_tree_to_widget_coords(XEN tree_view, XEN tx, XEN ty, XEN wx, XEN wy)
{
  #define H_gtk_tree_view_tree_to_widget_coords "void gtk_tree_view_tree_to_widget_coords(GtkTreeView* tree_view, \
gint tx, gint ty, gint* [wx], gint* [wy])"
  gint ref_wx;
  gint ref_wy;
  XEN_ASSERT_TYPE(XEN_GtkTreeView__P(tree_view), tree_view, 1, "gtk_tree_view_tree_to_widget_coords", "GtkTreeView*");
  XEN_ASSERT_TYPE(XEN_gint_P(tx), tx, 2, "gtk_tree_view_tree_to_widget_coords", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(ty), ty, 3, "gtk_tree_view_tree_to_widget_coords", "gint");
  gtk_tree_view_tree_to_widget_coords(XEN_TO_C_GtkTreeView_(tree_view), XEN_TO_C_gint(tx), XEN_TO_C_gint(ty), &ref_wx, &ref_wy);
  return(XEN_LIST_2(C_TO_XEN_gint(ref_wx), C_TO_XEN_gint(ref_wy)));
}

static XEN gxg_gtk_tree_view_enable_model_drag_source(XEN tree_view, XEN start_button_mask, XEN targets, XEN n_targets, XEN actions)
{
  #define H_gtk_tree_view_enable_model_drag_source "void gtk_tree_view_enable_model_drag_source(GtkTreeView* tree_view, \
GdkModifierType start_button_mask, GtkTargetEntry* targets, gint n_targets, GdkDragAction actions)"
  XEN_ASSERT_TYPE(XEN_GtkTreeView__P(tree_view), tree_view, 1, "gtk_tree_view_enable_model_drag_source", "GtkTreeView*");
  XEN_ASSERT_TYPE(XEN_GdkModifierType_P(start_button_mask), start_button_mask, 2, "gtk_tree_view_enable_model_drag_source", "GdkModifierType");
  XEN_ASSERT_TYPE(XEN_GtkTargetEntry__P(targets), targets, 3, "gtk_tree_view_enable_model_drag_source", "GtkTargetEntry*");
  XEN_ASSERT_TYPE(XEN_gint_P(n_targets), n_targets, 4, "gtk_tree_view_enable_model_drag_source", "gint");
  XEN_ASSERT_TYPE(XEN_GdkDragAction_P(actions), actions, 5, "gtk_tree_view_enable_model_drag_source", "GdkDragAction");
  gtk_tree_view_enable_model_drag_source(XEN_TO_C_GtkTreeView_(tree_view), XEN_TO_C_GdkModifierType(start_button_mask), XEN_TO_C_GtkTargetEntry_(targets), 
                                         XEN_TO_C_gint(n_targets), XEN_TO_C_GdkDragAction(actions));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tree_view_enable_model_drag_dest(XEN tree_view, XEN targets, XEN n_targets, XEN actions)
{
  #define H_gtk_tree_view_enable_model_drag_dest "void gtk_tree_view_enable_model_drag_dest(GtkTreeView* tree_view, \
GtkTargetEntry* targets, gint n_targets, GdkDragAction actions)"
  XEN_ASSERT_TYPE(XEN_GtkTreeView__P(tree_view), tree_view, 1, "gtk_tree_view_enable_model_drag_dest", "GtkTreeView*");
  XEN_ASSERT_TYPE(XEN_GtkTargetEntry__P(targets), targets, 2, "gtk_tree_view_enable_model_drag_dest", "GtkTargetEntry*");
  XEN_ASSERT_TYPE(XEN_gint_P(n_targets), n_targets, 3, "gtk_tree_view_enable_model_drag_dest", "gint");
  XEN_ASSERT_TYPE(XEN_GdkDragAction_P(actions), actions, 4, "gtk_tree_view_enable_model_drag_dest", "GdkDragAction");
  gtk_tree_view_enable_model_drag_dest(XEN_TO_C_GtkTreeView_(tree_view), XEN_TO_C_GtkTargetEntry_(targets), XEN_TO_C_gint(n_targets), 
                                       XEN_TO_C_GdkDragAction(actions));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tree_view_unset_rows_drag_source(XEN tree_view)
{
  #define H_gtk_tree_view_unset_rows_drag_source "void gtk_tree_view_unset_rows_drag_source(GtkTreeView* tree_view)"
  XEN_ASSERT_TYPE(XEN_GtkTreeView__P(tree_view), tree_view, 1, "gtk_tree_view_unset_rows_drag_source", "GtkTreeView*");
  gtk_tree_view_unset_rows_drag_source(XEN_TO_C_GtkTreeView_(tree_view));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tree_view_unset_rows_drag_dest(XEN tree_view)
{
  #define H_gtk_tree_view_unset_rows_drag_dest "void gtk_tree_view_unset_rows_drag_dest(GtkTreeView* tree_view)"
  XEN_ASSERT_TYPE(XEN_GtkTreeView__P(tree_view), tree_view, 1, "gtk_tree_view_unset_rows_drag_dest", "GtkTreeView*");
  gtk_tree_view_unset_rows_drag_dest(XEN_TO_C_GtkTreeView_(tree_view));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tree_view_set_drag_dest_row(XEN tree_view, XEN path, XEN pos)
{
  #define H_gtk_tree_view_set_drag_dest_row "void gtk_tree_view_set_drag_dest_row(GtkTreeView* tree_view, \
GtkTreePath* path, GtkTreeViewDropPosition pos)"
  XEN_ASSERT_TYPE(XEN_GtkTreeView__P(tree_view), tree_view, 1, "gtk_tree_view_set_drag_dest_row", "GtkTreeView*");
  XEN_ASSERT_TYPE(XEN_GtkTreePath__P(path), path, 2, "gtk_tree_view_set_drag_dest_row", "GtkTreePath*");
  XEN_ASSERT_TYPE(XEN_GtkTreeViewDropPosition_P(pos), pos, 3, "gtk_tree_view_set_drag_dest_row", "GtkTreeViewDropPosition");
  gtk_tree_view_set_drag_dest_row(XEN_TO_C_GtkTreeView_(tree_view), XEN_TO_C_GtkTreePath_(path), XEN_TO_C_GtkTreeViewDropPosition(pos));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tree_view_get_drag_dest_row(XEN tree_view, XEN path, XEN pos)
{
  #define H_gtk_tree_view_get_drag_dest_row "void gtk_tree_view_get_drag_dest_row(GtkTreeView* tree_view, \
GtkTreePath** [path], GtkTreeViewDropPosition* [pos])"
  GtkTreePath* ref_path;
  GtkTreeViewDropPosition ref_pos;
  XEN_ASSERT_TYPE(XEN_GtkTreeView__P(tree_view), tree_view, 1, "gtk_tree_view_get_drag_dest_row", "GtkTreeView*");
  gtk_tree_view_get_drag_dest_row(XEN_TO_C_GtkTreeView_(tree_view), &ref_path, &ref_pos);
  return(XEN_LIST_2(C_TO_XEN_GtkTreePath_(ref_path), C_TO_XEN_GtkTreeViewDropPosition(ref_pos)));
}

static XEN gxg_gtk_tree_view_get_dest_row_at_pos(XEN tree_view, XEN drag_x, XEN drag_y, XEN path, XEN pos)
{
  #define H_gtk_tree_view_get_dest_row_at_pos "gboolean gtk_tree_view_get_dest_row_at_pos(GtkTreeView* tree_view, \
gint drag_x, gint drag_y, GtkTreePath** [path], GtkTreeViewDropPosition* [pos])"
  GtkTreePath* ref_path;
  GtkTreeViewDropPosition ref_pos;
  XEN_ASSERT_TYPE(XEN_GtkTreeView__P(tree_view), tree_view, 1, "gtk_tree_view_get_dest_row_at_pos", "GtkTreeView*");
  XEN_ASSERT_TYPE(XEN_gint_P(drag_x), drag_x, 2, "gtk_tree_view_get_dest_row_at_pos", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(drag_y), drag_y, 3, "gtk_tree_view_get_dest_row_at_pos", "gint");
  {
    XEN result = XEN_FALSE;
    result = C_TO_XEN_gboolean(gtk_tree_view_get_dest_row_at_pos(XEN_TO_C_GtkTreeView_(tree_view), XEN_TO_C_gint(drag_x), 
                                                                 XEN_TO_C_gint(drag_y), &ref_path, &ref_pos));
    return(XEN_LIST_3(result, C_TO_XEN_GtkTreePath_(ref_path), C_TO_XEN_GtkTreeViewDropPosition(ref_pos)));
   }
}

static XEN gxg_gtk_tree_view_create_row_drag_icon(XEN tree_view, XEN path)
{
  #define H_gtk_tree_view_create_row_drag_icon "GdkPixmap* gtk_tree_view_create_row_drag_icon(GtkTreeView* tree_view, \
GtkTreePath* path)"
  XEN_ASSERT_TYPE(XEN_GtkTreeView__P(tree_view), tree_view, 1, "gtk_tree_view_create_row_drag_icon", "GtkTreeView*");
  XEN_ASSERT_TYPE(XEN_GtkTreePath__P(path), path, 2, "gtk_tree_view_create_row_drag_icon", "GtkTreePath*");
  return(C_TO_XEN_GdkPixmap_(gtk_tree_view_create_row_drag_icon(XEN_TO_C_GtkTreeView_(tree_view), XEN_TO_C_GtkTreePath_(path))));
}

static XEN gxg_gtk_tree_view_set_enable_search(XEN tree_view, XEN enable_search)
{
  #define H_gtk_tree_view_set_enable_search "void gtk_tree_view_set_enable_search(GtkTreeView* tree_view, \
gboolean enable_search)"
  XEN_ASSERT_TYPE(XEN_GtkTreeView__P(tree_view), tree_view, 1, "gtk_tree_view_set_enable_search", "GtkTreeView*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(enable_search), enable_search, 2, "gtk_tree_view_set_enable_search", "gboolean");
  gtk_tree_view_set_enable_search(XEN_TO_C_GtkTreeView_(tree_view), XEN_TO_C_gboolean(enable_search));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tree_view_get_enable_search(XEN tree_view)
{
  #define H_gtk_tree_view_get_enable_search "gboolean gtk_tree_view_get_enable_search(GtkTreeView* tree_view)"
  XEN_ASSERT_TYPE(XEN_GtkTreeView__P(tree_view), tree_view, 1, "gtk_tree_view_get_enable_search", "GtkTreeView*");
  return(C_TO_XEN_gboolean(gtk_tree_view_get_enable_search(XEN_TO_C_GtkTreeView_(tree_view))));
}

static XEN gxg_gtk_tree_view_get_search_column(XEN tree_view)
{
  #define H_gtk_tree_view_get_search_column "gint gtk_tree_view_get_search_column(GtkTreeView* tree_view)"
  XEN_ASSERT_TYPE(XEN_GtkTreeView__P(tree_view), tree_view, 1, "gtk_tree_view_get_search_column", "GtkTreeView*");
  return(C_TO_XEN_gint(gtk_tree_view_get_search_column(XEN_TO_C_GtkTreeView_(tree_view))));
}

static XEN gxg_gtk_tree_view_set_search_column(XEN tree_view, XEN column)
{
  #define H_gtk_tree_view_set_search_column "void gtk_tree_view_set_search_column(GtkTreeView* tree_view, \
gint column)"
  XEN_ASSERT_TYPE(XEN_GtkTreeView__P(tree_view), tree_view, 1, "gtk_tree_view_set_search_column", "GtkTreeView*");
  XEN_ASSERT_TYPE(XEN_gint_P(column), column, 2, "gtk_tree_view_set_search_column", "gint");
  gtk_tree_view_set_search_column(XEN_TO_C_GtkTreeView_(tree_view), XEN_TO_C_gint(column));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tree_view_get_search_equal_func(XEN tree_view)
{
  #define H_gtk_tree_view_get_search_equal_func "GtkTreeViewSearchEqualFunc gtk_tree_view_get_search_equal_func(GtkTreeView* tree_view)"
  XEN_ASSERT_TYPE(XEN_GtkTreeView__P(tree_view), tree_view, 1, "gtk_tree_view_get_search_equal_func", "GtkTreeView*");
  return(C_TO_XEN_GtkTreeViewSearchEqualFunc(gtk_tree_view_get_search_equal_func(XEN_TO_C_GtkTreeView_(tree_view))));
}

static XEN gxg_gtk_tree_view_set_search_equal_func(XEN tree_view, XEN func, XEN func_data, XEN search_destroy)
{
  #define H_gtk_tree_view_set_search_equal_func "void gtk_tree_view_set_search_equal_func(GtkTreeView* tree_view, \
GtkTreeViewSearchEqualFunc func, lambda_data func_data, GtkDestroyNotify search_destroy)"
  XEN_ASSERT_TYPE(XEN_GtkTreeView__P(tree_view), tree_view, 1, "gtk_tree_view_set_search_equal_func", "GtkTreeView*");
  XEN_ASSERT_TYPE(XEN_GtkTreeViewSearchEqualFunc_P(func), func, 2, "gtk_tree_view_set_search_equal_func", "GtkTreeViewSearchEqualFunc");
  XEN_ASSERT_TYPE(XEN_lambda_data_P(func_data), func_data, 3, "gtk_tree_view_set_search_equal_func", "lambda_data");
  XEN_ASSERT_TYPE(XEN_GtkDestroyNotify_P(search_destroy), search_destroy, 4, "gtk_tree_view_set_search_equal_func", "GtkDestroyNotify");
  {
    int loc;
    XEN gxg_ptr = XEN_LIST_5(func, func_data, XEN_FALSE, XEN_FALSE, XEN_FALSE);
    loc = xm_protect(gxg_ptr);
    XEN_LIST_SET(gxg_ptr, 2, C_TO_XEN_INT(loc));
    XEN_LIST_SET(gxg_ptr, 3, search_destroy);
    gtk_tree_view_set_search_equal_func(XEN_TO_C_GtkTreeView_(tree_view), XEN_TO_C_GtkTreeViewSearchEqualFunc(func), XEN_TO_C_lambda_data(func_data), 
                                    XEN_TO_C_GtkDestroyNotify(search_destroy));
    xm_unprotect_at(loc);
    return(XEN_FALSE);
   }
}
#endif

static XEN gxg_gtk_vbutton_box_get_type(void)
{
  #define H_gtk_vbutton_box_get_type "GtkType gtk_vbutton_box_get_type( void)"
  return(C_TO_XEN_GtkType(gtk_vbutton_box_get_type()));
}

static XEN gxg_gtk_vbutton_box_new(void)
{
  #define H_gtk_vbutton_box_new "GtkWidget* gtk_vbutton_box_new( void)"
  return(C_TO_XEN_GtkWidget_(gtk_vbutton_box_new()));
}

static XEN gxg_gtk_viewport_get_type(void)
{
  #define H_gtk_viewport_get_type "GtkType gtk_viewport_get_type( void)"
  return(C_TO_XEN_GtkType(gtk_viewport_get_type()));
}

static XEN gxg_gtk_viewport_new(XEN hadjustment, XEN vadjustment)
{
  #define H_gtk_viewport_new "GtkWidget* gtk_viewport_new(GtkAdjustment* hadjustment, GtkAdjustment* vadjustment)"
  XEN_ASSERT_TYPE(XEN_GtkAdjustment__P(hadjustment), hadjustment, 1, "gtk_viewport_new", "GtkAdjustment*");
  XEN_ASSERT_TYPE(XEN_GtkAdjustment__P(vadjustment), vadjustment, 2, "gtk_viewport_new", "GtkAdjustment*");
  return(C_TO_XEN_GtkWidget_(gtk_viewport_new(XEN_TO_C_GtkAdjustment_(hadjustment), XEN_TO_C_GtkAdjustment_(vadjustment))));
}

static XEN gxg_gtk_viewport_get_hadjustment(XEN viewport)
{
  #define H_gtk_viewport_get_hadjustment "GtkAdjustment* gtk_viewport_get_hadjustment(GtkViewport* viewport)"
  XEN_ASSERT_TYPE(XEN_GtkViewport__P(viewport), viewport, 1, "gtk_viewport_get_hadjustment", "GtkViewport*");
  return(C_TO_XEN_GtkAdjustment_(gtk_viewport_get_hadjustment(XEN_TO_C_GtkViewport_(viewport))));
}

static XEN gxg_gtk_viewport_get_vadjustment(XEN viewport)
{
  #define H_gtk_viewport_get_vadjustment "GtkAdjustment* gtk_viewport_get_vadjustment(GtkViewport* viewport)"
  XEN_ASSERT_TYPE(XEN_GtkViewport__P(viewport), viewport, 1, "gtk_viewport_get_vadjustment", "GtkViewport*");
  return(C_TO_XEN_GtkAdjustment_(gtk_viewport_get_vadjustment(XEN_TO_C_GtkViewport_(viewport))));
}

static XEN gxg_gtk_viewport_set_hadjustment(XEN viewport, XEN adjustment)
{
  #define H_gtk_viewport_set_hadjustment "void gtk_viewport_set_hadjustment(GtkViewport* viewport, GtkAdjustment* adjustment)"
  XEN_ASSERT_TYPE(XEN_GtkViewport__P(viewport), viewport, 1, "gtk_viewport_set_hadjustment", "GtkViewport*");
  XEN_ASSERT_TYPE(XEN_GtkAdjustment__P(adjustment), adjustment, 2, "gtk_viewport_set_hadjustment", "GtkAdjustment*");
  gtk_viewport_set_hadjustment(XEN_TO_C_GtkViewport_(viewport), XEN_TO_C_GtkAdjustment_(adjustment));
  return(XEN_FALSE);
}

static XEN gxg_gtk_viewport_set_vadjustment(XEN viewport, XEN adjustment)
{
  #define H_gtk_viewport_set_vadjustment "void gtk_viewport_set_vadjustment(GtkViewport* viewport, GtkAdjustment* adjustment)"
  XEN_ASSERT_TYPE(XEN_GtkViewport__P(viewport), viewport, 1, "gtk_viewport_set_vadjustment", "GtkViewport*");
  XEN_ASSERT_TYPE(XEN_GtkAdjustment__P(adjustment), adjustment, 2, "gtk_viewport_set_vadjustment", "GtkAdjustment*");
  gtk_viewport_set_vadjustment(XEN_TO_C_GtkViewport_(viewport), XEN_TO_C_GtkAdjustment_(adjustment));
  return(XEN_FALSE);
}

static XEN gxg_gtk_viewport_set_shadow_type(XEN viewport, XEN type)
{
  #define H_gtk_viewport_set_shadow_type "void gtk_viewport_set_shadow_type(GtkViewport* viewport, GtkShadowType type)"
  XEN_ASSERT_TYPE(XEN_GtkViewport__P(viewport), viewport, 1, "gtk_viewport_set_shadow_type", "GtkViewport*");
  XEN_ASSERT_TYPE(XEN_GtkShadowType_P(type), type, 2, "gtk_viewport_set_shadow_type", "GtkShadowType");
  gtk_viewport_set_shadow_type(XEN_TO_C_GtkViewport_(viewport), XEN_TO_C_GtkShadowType(type));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_viewport_get_shadow_type(XEN viewport)
{
  #define H_gtk_viewport_get_shadow_type "GtkShadowType gtk_viewport_get_shadow_type(GtkViewport* viewport)"
  XEN_ASSERT_TYPE(XEN_GtkViewport__P(viewport), viewport, 1, "gtk_viewport_get_shadow_type", "GtkViewport*");
  return(C_TO_XEN_GtkShadowType(gtk_viewport_get_shadow_type(XEN_TO_C_GtkViewport_(viewport))));
}
#endif

static XEN gxg_gtk_vpaned_get_type(void)
{
  #define H_gtk_vpaned_get_type "GtkType gtk_vpaned_get_type( void)"
  return(C_TO_XEN_GtkType(gtk_vpaned_get_type()));
}

static XEN gxg_gtk_vpaned_new(void)
{
  #define H_gtk_vpaned_new "GtkWidget* gtk_vpaned_new( void)"
  return(C_TO_XEN_GtkWidget_(gtk_vpaned_new()));
}

static XEN gxg_gtk_vruler_get_type(void)
{
  #define H_gtk_vruler_get_type "GtkType gtk_vruler_get_type( void)"
  return(C_TO_XEN_GtkType(gtk_vruler_get_type()));
}

static XEN gxg_gtk_vruler_new(void)
{
  #define H_gtk_vruler_new "GtkWidget* gtk_vruler_new( void)"
  return(C_TO_XEN_GtkWidget_(gtk_vruler_new()));
}

static XEN gxg_gtk_vscale_get_type(void)
{
  #define H_gtk_vscale_get_type "GtkType gtk_vscale_get_type( void)"
  return(C_TO_XEN_GtkType(gtk_vscale_get_type()));
}

static XEN gxg_gtk_vscale_new(XEN adjustment)
{
  #define H_gtk_vscale_new "GtkWidget* gtk_vscale_new(GtkAdjustment* adjustment)"
  XEN_ASSERT_TYPE(XEN_GtkAdjustment__P(adjustment), adjustment, 1, "gtk_vscale_new", "GtkAdjustment*");
  return(C_TO_XEN_GtkWidget_(gtk_vscale_new(XEN_TO_C_GtkAdjustment_(adjustment))));
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_vscale_new_with_range(XEN min, XEN max, XEN step)
{
  #define H_gtk_vscale_new_with_range "GtkWidget* gtk_vscale_new_with_range(gdouble min, gdouble max, \
gdouble step)"
  XEN_ASSERT_TYPE(XEN_gdouble_P(min), min, 1, "gtk_vscale_new_with_range", "gdouble");
  XEN_ASSERT_TYPE(XEN_gdouble_P(max), max, 2, "gtk_vscale_new_with_range", "gdouble");
  XEN_ASSERT_TYPE(XEN_gdouble_P(step), step, 3, "gtk_vscale_new_with_range", "gdouble");
  return(C_TO_XEN_GtkWidget_(gtk_vscale_new_with_range(XEN_TO_C_gdouble(min), XEN_TO_C_gdouble(max), XEN_TO_C_gdouble(step))));
}
#endif

static XEN gxg_gtk_vscrollbar_get_type(void)
{
  #define H_gtk_vscrollbar_get_type "GtkType gtk_vscrollbar_get_type( void)"
  return(C_TO_XEN_GtkType(gtk_vscrollbar_get_type()));
}

static XEN gxg_gtk_vscrollbar_new(XEN adjustment)
{
  #define H_gtk_vscrollbar_new "GtkWidget* gtk_vscrollbar_new(GtkAdjustment* adjustment)"
  XEN_ASSERT_TYPE(XEN_GtkAdjustment__P(adjustment), adjustment, 1, "gtk_vscrollbar_new", "GtkAdjustment*");
  return(C_TO_XEN_GtkWidget_(gtk_vscrollbar_new(XEN_TO_C_GtkAdjustment_(adjustment))));
}

static XEN gxg_gtk_vseparator_get_type(void)
{
  #define H_gtk_vseparator_get_type "GtkType gtk_vseparator_get_type( void)"
  return(C_TO_XEN_GtkType(gtk_vseparator_get_type()));
}

static XEN gxg_gtk_vseparator_new(void)
{
  #define H_gtk_vseparator_new "GtkWidget* gtk_vseparator_new( void)"
  return(C_TO_XEN_GtkWidget_(gtk_vseparator_new()));
}

static XEN gxg_GTK_WIDGET_STATE(XEN wid)
{
  #define H_GTK_WIDGET_STATE "int GTK_WIDGET_STATE(GtkWidget* wid)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(wid), wid, 1, "GTK_WIDGET_STATE", "GtkWidget*");
  return(C_TO_XEN_int(GTK_WIDGET_STATE(XEN_TO_C_GtkWidget_(wid))));
}

static XEN gxg_GTK_WIDGET_SAVED_STATE(XEN wid)
{
  #define H_GTK_WIDGET_SAVED_STATE "int GTK_WIDGET_SAVED_STATE(GtkWidget* wid)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(wid), wid, 1, "GTK_WIDGET_SAVED_STATE", "GtkWidget*");
  return(C_TO_XEN_int(GTK_WIDGET_SAVED_STATE(XEN_TO_C_GtkWidget_(wid))));
}

static XEN gxg_GTK_WIDGET_FLAGS(XEN wid)
{
  #define H_GTK_WIDGET_FLAGS "int GTK_WIDGET_FLAGS(GtkWidget* wid)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(wid), wid, 1, "GTK_WIDGET_FLAGS", "GtkWidget*");
  return(C_TO_XEN_int(GTK_WIDGET_FLAGS(XEN_TO_C_GtkWidget_(wid))));
}

static XEN gxg_GTK_WIDGET_TOPLEVEL(XEN wid)
{
  #define H_GTK_WIDGET_TOPLEVEL "gboolean GTK_WIDGET_TOPLEVEL(GtkWidget* wid)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(wid), wid, 1, "GTK_WIDGET_TOPLEVEL", "GtkWidget*");
  return(C_TO_XEN_gboolean(GTK_WIDGET_TOPLEVEL(XEN_TO_C_GtkWidget_(wid))));
}

static XEN gxg_GTK_WIDGET_NO_WINDOW(XEN wid)
{
  #define H_GTK_WIDGET_NO_WINDOW "gboolean GTK_WIDGET_NO_WINDOW(GtkWidget* wid)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(wid), wid, 1, "GTK_WIDGET_NO_WINDOW", "GtkWidget*");
  return(C_TO_XEN_gboolean(GTK_WIDGET_NO_WINDOW(XEN_TO_C_GtkWidget_(wid))));
}

static XEN gxg_GTK_WIDGET_REALIZED(XEN wid)
{
  #define H_GTK_WIDGET_REALIZED "gboolean GTK_WIDGET_REALIZED(GtkWidget* wid)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(wid), wid, 1, "GTK_WIDGET_REALIZED", "GtkWidget*");
  return(C_TO_XEN_gboolean(GTK_WIDGET_REALIZED(XEN_TO_C_GtkWidget_(wid))));
}

static XEN gxg_GTK_WIDGET_MAPPED(XEN wid)
{
  #define H_GTK_WIDGET_MAPPED "gboolean GTK_WIDGET_MAPPED(GtkWidget* wid)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(wid), wid, 1, "GTK_WIDGET_MAPPED", "GtkWidget*");
  return(C_TO_XEN_gboolean(GTK_WIDGET_MAPPED(XEN_TO_C_GtkWidget_(wid))));
}

static XEN gxg_GTK_WIDGET_VISIBLE(XEN wid)
{
  #define H_GTK_WIDGET_VISIBLE "gboolean GTK_WIDGET_VISIBLE(GtkWidget* wid)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(wid), wid, 1, "GTK_WIDGET_VISIBLE", "GtkWidget*");
  return(C_TO_XEN_gboolean(GTK_WIDGET_VISIBLE(XEN_TO_C_GtkWidget_(wid))));
}

static XEN gxg_GTK_WIDGET_DRAWABLE(XEN wid)
{
  #define H_GTK_WIDGET_DRAWABLE "gboolean GTK_WIDGET_DRAWABLE(GtkWidget* wid)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(wid), wid, 1, "GTK_WIDGET_DRAWABLE", "GtkWidget*");
  return(C_TO_XEN_gboolean(GTK_WIDGET_DRAWABLE(XEN_TO_C_GtkWidget_(wid))));
}

static XEN gxg_GTK_WIDGET_SENSITIVE(XEN wid)
{
  #define H_GTK_WIDGET_SENSITIVE "gboolean GTK_WIDGET_SENSITIVE(GtkWidget* wid)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(wid), wid, 1, "GTK_WIDGET_SENSITIVE", "GtkWidget*");
  return(C_TO_XEN_gboolean(GTK_WIDGET_SENSITIVE(XEN_TO_C_GtkWidget_(wid))));
}

static XEN gxg_GTK_WIDGET_PARENT_SENSITIVE(XEN wid)
{
  #define H_GTK_WIDGET_PARENT_SENSITIVE "gboolean GTK_WIDGET_PARENT_SENSITIVE(GtkWidget* wid)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(wid), wid, 1, "GTK_WIDGET_PARENT_SENSITIVE", "GtkWidget*");
  return(C_TO_XEN_gboolean(GTK_WIDGET_PARENT_SENSITIVE(XEN_TO_C_GtkWidget_(wid))));
}

static XEN gxg_GTK_WIDGET_IS_SENSITIVE(XEN wid)
{
  #define H_GTK_WIDGET_IS_SENSITIVE "gboolean GTK_WIDGET_IS_SENSITIVE(GtkWidget* wid)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(wid), wid, 1, "GTK_WIDGET_IS_SENSITIVE", "GtkWidget*");
  return(C_TO_XEN_gboolean(GTK_WIDGET_IS_SENSITIVE(XEN_TO_C_GtkWidget_(wid))));
}

static XEN gxg_GTK_WIDGET_CAN_FOCUS(XEN wid)
{
  #define H_GTK_WIDGET_CAN_FOCUS "gboolean GTK_WIDGET_CAN_FOCUS(GtkWidget* wid)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(wid), wid, 1, "GTK_WIDGET_CAN_FOCUS", "GtkWidget*");
  return(C_TO_XEN_gboolean(GTK_WIDGET_CAN_FOCUS(XEN_TO_C_GtkWidget_(wid))));
}

static XEN gxg_GTK_WIDGET_HAS_FOCUS(XEN wid)
{
  #define H_GTK_WIDGET_HAS_FOCUS "gboolean GTK_WIDGET_HAS_FOCUS(GtkWidget* wid)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(wid), wid, 1, "GTK_WIDGET_HAS_FOCUS", "GtkWidget*");
  return(C_TO_XEN_gboolean(GTK_WIDGET_HAS_FOCUS(XEN_TO_C_GtkWidget_(wid))));
}

static XEN gxg_GTK_WIDGET_CAN_DEFAULT(XEN wid)
{
  #define H_GTK_WIDGET_CAN_DEFAULT "gboolean GTK_WIDGET_CAN_DEFAULT(GtkWidget* wid)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(wid), wid, 1, "GTK_WIDGET_CAN_DEFAULT", "GtkWidget*");
  return(C_TO_XEN_gboolean(GTK_WIDGET_CAN_DEFAULT(XEN_TO_C_GtkWidget_(wid))));
}

static XEN gxg_GTK_WIDGET_HAS_DEFAULT(XEN wid)
{
  #define H_GTK_WIDGET_HAS_DEFAULT "gboolean GTK_WIDGET_HAS_DEFAULT(GtkWidget* wid)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(wid), wid, 1, "GTK_WIDGET_HAS_DEFAULT", "GtkWidget*");
  return(C_TO_XEN_gboolean(GTK_WIDGET_HAS_DEFAULT(XEN_TO_C_GtkWidget_(wid))));
}

static XEN gxg_GTK_WIDGET_HAS_GRAB(XEN wid)
{
  #define H_GTK_WIDGET_HAS_GRAB "gboolean GTK_WIDGET_HAS_GRAB(GtkWidget* wid)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(wid), wid, 1, "GTK_WIDGET_HAS_GRAB", "GtkWidget*");
  return(C_TO_XEN_gboolean(GTK_WIDGET_HAS_GRAB(XEN_TO_C_GtkWidget_(wid))));
}

static XEN gxg_GTK_WIDGET_RC_STYLE(XEN wid)
{
  #define H_GTK_WIDGET_RC_STYLE "gboolean GTK_WIDGET_RC_STYLE(GtkWidget* wid)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(wid), wid, 1, "GTK_WIDGET_RC_STYLE", "GtkWidget*");
  return(C_TO_XEN_gboolean(GTK_WIDGET_RC_STYLE(XEN_TO_C_GtkWidget_(wid))));
}

static XEN gxg_GTK_WIDGET_COMPOSITE_CHILD(XEN wid)
{
  #define H_GTK_WIDGET_COMPOSITE_CHILD "gboolean GTK_WIDGET_COMPOSITE_CHILD(GtkWidget* wid)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(wid), wid, 1, "GTK_WIDGET_COMPOSITE_CHILD", "GtkWidget*");
  return(C_TO_XEN_gboolean(GTK_WIDGET_COMPOSITE_CHILD(XEN_TO_C_GtkWidget_(wid))));
}

static XEN gxg_GTK_WIDGET_APP_PAINTABLE(XEN wid)
{
  #define H_GTK_WIDGET_APP_PAINTABLE "gboolean GTK_WIDGET_APP_PAINTABLE(GtkWidget* wid)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(wid), wid, 1, "GTK_WIDGET_APP_PAINTABLE", "GtkWidget*");
  return(C_TO_XEN_gboolean(GTK_WIDGET_APP_PAINTABLE(XEN_TO_C_GtkWidget_(wid))));
}

static XEN gxg_GTK_WIDGET_RECEIVES_DEFAULT(XEN wid)
{
  #define H_GTK_WIDGET_RECEIVES_DEFAULT "gboolean GTK_WIDGET_RECEIVES_DEFAULT(GtkWidget* wid)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(wid), wid, 1, "GTK_WIDGET_RECEIVES_DEFAULT", "GtkWidget*");
  return(C_TO_XEN_gboolean(GTK_WIDGET_RECEIVES_DEFAULT(XEN_TO_C_GtkWidget_(wid))));
}

#if (!HAVE_GTK_1)
static XEN gxg_GTK_WIDGET_DOUBLE_BUFFERED(XEN wid)
{
  #define H_GTK_WIDGET_DOUBLE_BUFFERED "gboolean GTK_WIDGET_DOUBLE_BUFFERED(GtkWidget* wid)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(wid), wid, 1, "GTK_WIDGET_DOUBLE_BUFFERED", "GtkWidget*");
  return(C_TO_XEN_gboolean(GTK_WIDGET_DOUBLE_BUFFERED(XEN_TO_C_GtkWidget_(wid))));
}
#endif

static XEN gxg_GTK_WIDGET_SET_FLAGS(XEN wid, XEN flag)
{
  #define H_GTK_WIDGET_SET_FLAGS "void GTK_WIDGET_SET_FLAGS(GtkWidget* wid, int flag)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(wid), wid, 1, "GTK_WIDGET_SET_FLAGS", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_int_P(flag), flag, 2, "GTK_WIDGET_SET_FLAGS", "int");
  GTK_WIDGET_SET_FLAGS(XEN_TO_C_GtkWidget_(wid), XEN_TO_C_int(flag));
  return(XEN_FALSE);
}

static XEN gxg_GTK_WIDGET_UNSET_FLAGS(XEN wid, XEN flag)
{
  #define H_GTK_WIDGET_UNSET_FLAGS "void GTK_WIDGET_UNSET_FLAGS(GtkWidget* wid, int flag)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(wid), wid, 1, "GTK_WIDGET_UNSET_FLAGS", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_int_P(flag), flag, 2, "GTK_WIDGET_UNSET_FLAGS", "int");
  GTK_WIDGET_UNSET_FLAGS(XEN_TO_C_GtkWidget_(wid), XEN_TO_C_int(flag));
  return(XEN_FALSE);
}

static XEN gxg_gtk_widget_get_type(void)
{
  #define H_gtk_widget_get_type "GtkType gtk_widget_get_type( void)"
  return(C_TO_XEN_GtkType(gtk_widget_get_type()));
}

static XEN gxg_gtk_widget_new(XEN type, XEN first_property_name)
{
  #define H_gtk_widget_new "GtkWidget* gtk_widget_new(GtkType type, gchar* first_property_name, ...)"
  XEN_ASSERT_TYPE(XEN_GtkType_P(type), type, 1, "gtk_widget_new", "GtkType");
  XEN_ASSERT_TYPE(XEN_gchar__P(first_property_name), first_property_name, 2, "gtk_widget_new", "gchar*");
  return(C_TO_XEN_GtkWidget_(gtk_widget_new(XEN_TO_C_GtkType(type), XEN_TO_C_gchar_(first_property_name))));
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_widget_ref(XEN widget)
{
  #define H_gtk_widget_ref "GtkWidget* gtk_widget_ref(GtkWidget* widget)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_widget_ref", "GtkWidget*");
  return(C_TO_XEN_GtkWidget_(gtk_widget_ref(XEN_TO_C_GtkWidget_(widget))));
}

static XEN gxg_gtk_widget_unref(XEN widget)
{
  #define H_gtk_widget_unref "void gtk_widget_unref(GtkWidget* widget)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_widget_unref", "GtkWidget*");
  gtk_widget_unref(XEN_TO_C_GtkWidget_(widget));
  return(XEN_FALSE);
}
#endif

static XEN gxg_gtk_widget_destroy(XEN widget)
{
  #define H_gtk_widget_destroy "void gtk_widget_destroy(GtkWidget* widget)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_widget_destroy", "GtkWidget*");
  gtk_widget_destroy(XEN_TO_C_GtkWidget_(widget));
  return(XEN_FALSE);
}

static XEN gxg_gtk_widget_destroyed(XEN widget, XEN widget_pointer)
{
  #define H_gtk_widget_destroyed "void gtk_widget_destroyed(GtkWidget* widget, GtkWidget** [widget_pointer])"
  GtkWidget* ref_widget_pointer;
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_widget_destroyed", "GtkWidget*");
  gtk_widget_destroyed(XEN_TO_C_GtkWidget_(widget), &ref_widget_pointer);
  return(XEN_LIST_1(C_TO_XEN_GtkWidget_(ref_widget_pointer)));
}

static XEN gxg_gtk_widget_unparent(XEN widget)
{
  #define H_gtk_widget_unparent "void gtk_widget_unparent(GtkWidget* widget)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_widget_unparent", "GtkWidget*");
  gtk_widget_unparent(XEN_TO_C_GtkWidget_(widget));
  return(XEN_FALSE);
}

static XEN gxg_gtk_widget_show(XEN widget)
{
  #define H_gtk_widget_show "void gtk_widget_show(GtkWidget* widget)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_widget_show", "GtkWidget*");
  gtk_widget_show(XEN_TO_C_GtkWidget_(widget));
  return(XEN_FALSE);
}

static XEN gxg_gtk_widget_show_now(XEN widget)
{
  #define H_gtk_widget_show_now "void gtk_widget_show_now(GtkWidget* widget)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_widget_show_now", "GtkWidget*");
  gtk_widget_show_now(XEN_TO_C_GtkWidget_(widget));
  return(XEN_FALSE);
}

static XEN gxg_gtk_widget_hide(XEN widget)
{
  #define H_gtk_widget_hide "void gtk_widget_hide(GtkWidget* widget)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_widget_hide", "GtkWidget*");
  gtk_widget_hide(XEN_TO_C_GtkWidget_(widget));
  return(XEN_FALSE);
}

static XEN gxg_gtk_widget_show_all(XEN widget)
{
  #define H_gtk_widget_show_all "void gtk_widget_show_all(GtkWidget* widget)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_widget_show_all", "GtkWidget*");
  gtk_widget_show_all(XEN_TO_C_GtkWidget_(widget));
  return(XEN_FALSE);
}

static XEN gxg_gtk_widget_hide_all(XEN widget)
{
  #define H_gtk_widget_hide_all "void gtk_widget_hide_all(GtkWidget* widget)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_widget_hide_all", "GtkWidget*");
  gtk_widget_hide_all(XEN_TO_C_GtkWidget_(widget));
  return(XEN_FALSE);
}

static XEN gxg_gtk_widget_map(XEN widget)
{
  #define H_gtk_widget_map "void gtk_widget_map(GtkWidget* widget)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_widget_map", "GtkWidget*");
  gtk_widget_map(XEN_TO_C_GtkWidget_(widget));
  return(XEN_FALSE);
}

static XEN gxg_gtk_widget_unmap(XEN widget)
{
  #define H_gtk_widget_unmap "void gtk_widget_unmap(GtkWidget* widget)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_widget_unmap", "GtkWidget*");
  gtk_widget_unmap(XEN_TO_C_GtkWidget_(widget));
  return(XEN_FALSE);
}

static XEN gxg_gtk_widget_realize(XEN widget)
{
  #define H_gtk_widget_realize "void gtk_widget_realize(GtkWidget* widget)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_widget_realize", "GtkWidget*");
  gtk_widget_realize(XEN_TO_C_GtkWidget_(widget));
  return(XEN_FALSE);
}

static XEN gxg_gtk_widget_unrealize(XEN widget)
{
  #define H_gtk_widget_unrealize "void gtk_widget_unrealize(GtkWidget* widget)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_widget_unrealize", "GtkWidget*");
  gtk_widget_unrealize(XEN_TO_C_GtkWidget_(widget));
  return(XEN_FALSE);
}

static XEN gxg_gtk_widget_queue_draw(XEN widget)
{
  #define H_gtk_widget_queue_draw "void gtk_widget_queue_draw(GtkWidget* widget)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_widget_queue_draw", "GtkWidget*");
  gtk_widget_queue_draw(XEN_TO_C_GtkWidget_(widget));
  return(XEN_FALSE);
}

static XEN gxg_gtk_widget_queue_draw_area(XEN widget, XEN x, XEN y, XEN width, XEN height)
{
  #define H_gtk_widget_queue_draw_area "void gtk_widget_queue_draw_area(GtkWidget* widget, gint x, gint y, \
gint width, gint height)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_widget_queue_draw_area", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_gint_P(x), x, 2, "gtk_widget_queue_draw_area", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(y), y, 3, "gtk_widget_queue_draw_area", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(width), width, 4, "gtk_widget_queue_draw_area", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(height), height, 5, "gtk_widget_queue_draw_area", "gint");
  gtk_widget_queue_draw_area(XEN_TO_C_GtkWidget_(widget), XEN_TO_C_gint(x), XEN_TO_C_gint(y), XEN_TO_C_gint(width), XEN_TO_C_gint(height));
  return(XEN_FALSE);
}

static XEN gxg_gtk_widget_queue_resize(XEN widget)
{
  #define H_gtk_widget_queue_resize "void gtk_widget_queue_resize(GtkWidget* widget)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_widget_queue_resize", "GtkWidget*");
  gtk_widget_queue_resize(XEN_TO_C_GtkWidget_(widget));
  return(XEN_FALSE);
}

static XEN gxg_gtk_widget_size_request(XEN widget, XEN requisition)
{
  #define H_gtk_widget_size_request "void gtk_widget_size_request(GtkWidget* widget, GtkRequisition* requisition)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_widget_size_request", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_GtkRequisition__P(requisition), requisition, 2, "gtk_widget_size_request", "GtkRequisition*");
  gtk_widget_size_request(XEN_TO_C_GtkWidget_(widget), XEN_TO_C_GtkRequisition_(requisition));
  return(XEN_FALSE);
}

static XEN gxg_gtk_widget_size_allocate(XEN widget, XEN allocation)
{
  #define H_gtk_widget_size_allocate "void gtk_widget_size_allocate(GtkWidget* widget, GtkAllocation* allocation)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_widget_size_allocate", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_GtkAllocation__P(allocation), allocation, 2, "gtk_widget_size_allocate", "GtkAllocation*");
  gtk_widget_size_allocate(XEN_TO_C_GtkWidget_(widget), XEN_TO_C_GtkAllocation_(allocation));
  return(XEN_FALSE);
}

static XEN gxg_gtk_widget_get_child_requisition(XEN widget, XEN requisition)
{
  #define H_gtk_widget_get_child_requisition "void gtk_widget_get_child_requisition(GtkWidget* widget, \
GtkRequisition* requisition)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_widget_get_child_requisition", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_GtkRequisition__P(requisition), requisition, 2, "gtk_widget_get_child_requisition", "GtkRequisition*");
  gtk_widget_get_child_requisition(XEN_TO_C_GtkWidget_(widget), XEN_TO_C_GtkRequisition_(requisition));
  return(XEN_FALSE);
}

static XEN gxg_gtk_widget_add_accelerator(XEN widget, XEN accel_signal, XEN accel_group, XEN accel_key, XEN accel_mods, XEN accel_flags)
{
  #define H_gtk_widget_add_accelerator "void gtk_widget_add_accelerator(GtkWidget* widget, gchar* accel_signal, \
GtkAccelGroup* accel_group, guint accel_key, GdkModifierType accel_mods, GtkAccelFlags accel_flags)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_widget_add_accelerator", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_gchar__P(accel_signal), accel_signal, 2, "gtk_widget_add_accelerator", "gchar*");
  XEN_ASSERT_TYPE(XEN_GtkAccelGroup__P(accel_group), accel_group, 3, "gtk_widget_add_accelerator", "GtkAccelGroup*");
  XEN_ASSERT_TYPE(XEN_guint_P(accel_key), accel_key, 4, "gtk_widget_add_accelerator", "guint");
  XEN_ASSERT_TYPE(XEN_GdkModifierType_P(accel_mods), accel_mods, 5, "gtk_widget_add_accelerator", "GdkModifierType");
  XEN_ASSERT_TYPE(XEN_GtkAccelFlags_P(accel_flags), accel_flags, 6, "gtk_widget_add_accelerator", "GtkAccelFlags");
  gtk_widget_add_accelerator(XEN_TO_C_GtkWidget_(widget), XEN_TO_C_gchar_(accel_signal), XEN_TO_C_GtkAccelGroup_(accel_group), 
                             XEN_TO_C_guint(accel_key), XEN_TO_C_GdkModifierType(accel_mods), XEN_TO_C_GtkAccelFlags(accel_flags));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_widget_remove_accelerator(XEN widget, XEN accel_group, XEN accel_key, XEN accel_mods)
{
  #define H_gtk_widget_remove_accelerator "gboolean gtk_widget_remove_accelerator(GtkWidget* widget, \
GtkAccelGroup* accel_group, guint accel_key, GdkModifierType accel_mods)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_widget_remove_accelerator", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_GtkAccelGroup__P(accel_group), accel_group, 2, "gtk_widget_remove_accelerator", "GtkAccelGroup*");
  XEN_ASSERT_TYPE(XEN_guint_P(accel_key), accel_key, 3, "gtk_widget_remove_accelerator", "guint");
  XEN_ASSERT_TYPE(XEN_GdkModifierType_P(accel_mods), accel_mods, 4, "gtk_widget_remove_accelerator", "GdkModifierType");
  return(C_TO_XEN_gboolean(gtk_widget_remove_accelerator(XEN_TO_C_GtkWidget_(widget), XEN_TO_C_GtkAccelGroup_(accel_group), 
                                                         XEN_TO_C_guint(accel_key), XEN_TO_C_GdkModifierType(accel_mods))));
}

static XEN gxg_gtk_widget_list_accel_closures(XEN widget)
{
  #define H_gtk_widget_list_accel_closures "GList* gtk_widget_list_accel_closures(GtkWidget* widget)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_widget_list_accel_closures", "GtkWidget*");
  return(C_TO_XEN_GList_(gtk_widget_list_accel_closures(XEN_TO_C_GtkWidget_(widget))));
}

static XEN gxg_gtk_widget_mnemonic_activate(XEN widget, XEN group_cycling)
{
  #define H_gtk_widget_mnemonic_activate "gboolean gtk_widget_mnemonic_activate(GtkWidget* widget, gboolean group_cycling)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_widget_mnemonic_activate", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(group_cycling), group_cycling, 2, "gtk_widget_mnemonic_activate", "gboolean");
  return(C_TO_XEN_gboolean(gtk_widget_mnemonic_activate(XEN_TO_C_GtkWidget_(widget), XEN_TO_C_gboolean(group_cycling))));
}
#endif

static XEN gxg_gtk_widget_event(XEN widget, XEN event)
{
  #define H_gtk_widget_event "gboolean gtk_widget_event(GtkWidget* widget, GdkEvent* event)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_widget_event", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_GdkEvent__P(event), event, 2, "gtk_widget_event", "GdkEvent*");
  return(C_TO_XEN_gboolean(gtk_widget_event(XEN_TO_C_GtkWidget_(widget), XEN_TO_C_GdkEvent_(event))));
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_widget_send_expose(XEN widget, XEN event)
{
  #define H_gtk_widget_send_expose "gint gtk_widget_send_expose(GtkWidget* widget, GdkEvent* event)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_widget_send_expose", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_GdkEvent__P(event), event, 2, "gtk_widget_send_expose", "GdkEvent*");
  return(C_TO_XEN_gint(gtk_widget_send_expose(XEN_TO_C_GtkWidget_(widget), XEN_TO_C_GdkEvent_(event))));
}
#endif

static XEN gxg_gtk_widget_activate(XEN widget)
{
  #define H_gtk_widget_activate "gboolean gtk_widget_activate(GtkWidget* widget)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_widget_activate", "GtkWidget*");
  return(C_TO_XEN_gboolean(gtk_widget_activate(XEN_TO_C_GtkWidget_(widget))));
}

static XEN gxg_gtk_widget_set_scroll_adjustments(XEN widget, XEN hadjustment, XEN vadjustment)
{
  #define H_gtk_widget_set_scroll_adjustments "gboolean gtk_widget_set_scroll_adjustments(GtkWidget* widget, \
GtkAdjustment* hadjustment, GtkAdjustment* vadjustment)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_widget_set_scroll_adjustments", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_GtkAdjustment__P(hadjustment) || XEN_FALSE_P(hadjustment), hadjustment, 2, "gtk_widget_set_scroll_adjustments", "GtkAdjustment*");
  XEN_ASSERT_TYPE(XEN_GtkAdjustment__P(vadjustment) || XEN_FALSE_P(vadjustment), vadjustment, 3, "gtk_widget_set_scroll_adjustments", "GtkAdjustment*");
  return(C_TO_XEN_gboolean(gtk_widget_set_scroll_adjustments(XEN_TO_C_GtkWidget_(widget), XEN_TO_C_GtkAdjustment_(hadjustment), 
                                                             XEN_TO_C_GtkAdjustment_(vadjustment))));
}

static XEN gxg_gtk_widget_reparent(XEN widget, XEN new_parent)
{
  #define H_gtk_widget_reparent "void gtk_widget_reparent(GtkWidget* widget, GtkWidget* new_parent)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_widget_reparent", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(new_parent), new_parent, 2, "gtk_widget_reparent", "GtkWidget*");
  gtk_widget_reparent(XEN_TO_C_GtkWidget_(widget), XEN_TO_C_GtkWidget_(new_parent));
  return(XEN_FALSE);
}

static XEN gxg_gtk_widget_intersect(XEN widget, XEN area, XEN intersection)
{
  #define H_gtk_widget_intersect "gboolean gtk_widget_intersect(GtkWidget* widget, GdkRectangle* area, \
GdkRectangle* intersection)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_widget_intersect", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_GdkRectangle__P(area), area, 2, "gtk_widget_intersect", "GdkRectangle*");
  XEN_ASSERT_TYPE(XEN_GdkRectangle__P(intersection) || XEN_FALSE_P(intersection), intersection, 3, "gtk_widget_intersect", "GdkRectangle*");
  return(C_TO_XEN_gboolean(gtk_widget_intersect(XEN_TO_C_GtkWidget_(widget), XEN_TO_C_GdkRectangle_(area), XEN_TO_C_GdkRectangle_(intersection))));
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_widget_region_intersect(XEN widget, XEN region)
{
  #define H_gtk_widget_region_intersect "GdkRegion* gtk_widget_region_intersect(GtkWidget* widget, GdkRegion* region)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_widget_region_intersect", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_GdkRegion__P(region), region, 2, "gtk_widget_region_intersect", "GdkRegion*");
  return(C_TO_XEN_GdkRegion_(gtk_widget_region_intersect(XEN_TO_C_GtkWidget_(widget), XEN_TO_C_GdkRegion_(region))));
}

static XEN gxg_gtk_widget_freeze_child_notify(XEN widget)
{
  #define H_gtk_widget_freeze_child_notify "void gtk_widget_freeze_child_notify(GtkWidget* widget)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_widget_freeze_child_notify", "GtkWidget*");
  gtk_widget_freeze_child_notify(XEN_TO_C_GtkWidget_(widget));
  return(XEN_FALSE);
}

static XEN gxg_gtk_widget_child_notify(XEN widget, XEN child_property)
{
  #define H_gtk_widget_child_notify "void gtk_widget_child_notify(GtkWidget* widget, gchar* child_property)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_widget_child_notify", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_gchar__P(child_property), child_property, 2, "gtk_widget_child_notify", "gchar*");
  gtk_widget_child_notify(XEN_TO_C_GtkWidget_(widget), XEN_TO_C_gchar_(child_property));
  return(XEN_FALSE);
}

static XEN gxg_gtk_widget_thaw_child_notify(XEN widget)
{
  #define H_gtk_widget_thaw_child_notify "void gtk_widget_thaw_child_notify(GtkWidget* widget)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_widget_thaw_child_notify", "GtkWidget*");
  gtk_widget_thaw_child_notify(XEN_TO_C_GtkWidget_(widget));
  return(XEN_FALSE);
}

static XEN gxg_gtk_widget_is_focus(XEN widget)
{
  #define H_gtk_widget_is_focus "gboolean gtk_widget_is_focus(GtkWidget* widget)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_widget_is_focus", "GtkWidget*");
  return(C_TO_XEN_gboolean(gtk_widget_is_focus(XEN_TO_C_GtkWidget_(widget))));
}
#endif

static XEN gxg_gtk_widget_grab_focus(XEN widget)
{
  #define H_gtk_widget_grab_focus "void gtk_widget_grab_focus(GtkWidget* widget)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_widget_grab_focus", "GtkWidget*");
  gtk_widget_grab_focus(XEN_TO_C_GtkWidget_(widget));
  return(XEN_FALSE);
}

static XEN gxg_gtk_widget_grab_default(XEN widget)
{
  #define H_gtk_widget_grab_default "void gtk_widget_grab_default(GtkWidget* widget)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_widget_grab_default", "GtkWidget*");
  gtk_widget_grab_default(XEN_TO_C_GtkWidget_(widget));
  return(XEN_FALSE);
}

static XEN gxg_gtk_widget_set_name(XEN widget, XEN name)
{
  #define H_gtk_widget_set_name "void gtk_widget_set_name(GtkWidget* widget, gchar* name)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_widget_set_name", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_gchar__P(name), name, 2, "gtk_widget_set_name", "gchar*");
  gtk_widget_set_name(XEN_TO_C_GtkWidget_(widget), XEN_TO_C_gchar_(name));
  return(XEN_FALSE);
}

static XEN gxg_gtk_widget_get_name(XEN widget)
{
  #define H_gtk_widget_get_name "gchar* gtk_widget_get_name(GtkWidget* widget)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_widget_get_name", "GtkWidget*");
  return(C_TO_XEN_gchar_(gtk_widget_get_name(XEN_TO_C_GtkWidget_(widget))));
}

static XEN gxg_gtk_widget_set_state(XEN widget, XEN state)
{
  #define H_gtk_widget_set_state "void gtk_widget_set_state(GtkWidget* widget, GtkStateType state)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_widget_set_state", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_GtkStateType_P(state), state, 2, "gtk_widget_set_state", "GtkStateType");
  gtk_widget_set_state(XEN_TO_C_GtkWidget_(widget), XEN_TO_C_GtkStateType(state));
  return(XEN_FALSE);
}

static XEN gxg_gtk_widget_set_sensitive(XEN widget, XEN sensitive)
{
  #define H_gtk_widget_set_sensitive "void gtk_widget_set_sensitive(GtkWidget* widget, gboolean sensitive)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_widget_set_sensitive", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(sensitive), sensitive, 2, "gtk_widget_set_sensitive", "gboolean");
  gtk_widget_set_sensitive(XEN_TO_C_GtkWidget_(widget), XEN_TO_C_gboolean(sensitive));
  return(XEN_FALSE);
}

static XEN gxg_gtk_widget_set_app_paintable(XEN widget, XEN app_paintable)
{
  #define H_gtk_widget_set_app_paintable "void gtk_widget_set_app_paintable(GtkWidget* widget, gboolean app_paintable)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_widget_set_app_paintable", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(app_paintable), app_paintable, 2, "gtk_widget_set_app_paintable", "gboolean");
  gtk_widget_set_app_paintable(XEN_TO_C_GtkWidget_(widget), XEN_TO_C_gboolean(app_paintable));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_widget_set_double_buffered(XEN widget, XEN double_buffered)
{
  #define H_gtk_widget_set_double_buffered "void gtk_widget_set_double_buffered(GtkWidget* widget, gboolean double_buffered)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_widget_set_double_buffered", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(double_buffered), double_buffered, 2, "gtk_widget_set_double_buffered", "gboolean");
  gtk_widget_set_double_buffered(XEN_TO_C_GtkWidget_(widget), XEN_TO_C_gboolean(double_buffered));
  return(XEN_FALSE);
}

static XEN gxg_gtk_widget_set_redraw_on_allocate(XEN widget, XEN redraw_on_allocate)
{
  #define H_gtk_widget_set_redraw_on_allocate "void gtk_widget_set_redraw_on_allocate(GtkWidget* widget, \
gboolean redraw_on_allocate)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_widget_set_redraw_on_allocate", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(redraw_on_allocate), redraw_on_allocate, 2, "gtk_widget_set_redraw_on_allocate", "gboolean");
  gtk_widget_set_redraw_on_allocate(XEN_TO_C_GtkWidget_(widget), XEN_TO_C_gboolean(redraw_on_allocate));
  return(XEN_FALSE);
}
#endif

static XEN gxg_gtk_widget_set_parent(XEN widget, XEN parent)
{
  #define H_gtk_widget_set_parent "void gtk_widget_set_parent(GtkWidget* widget, GtkWidget* parent)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_widget_set_parent", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(parent), parent, 2, "gtk_widget_set_parent", "GtkWidget*");
  gtk_widget_set_parent(XEN_TO_C_GtkWidget_(widget), XEN_TO_C_GtkWidget_(parent));
  return(XEN_FALSE);
}

static XEN gxg_gtk_widget_set_parent_window(XEN widget, XEN parent_window)
{
  #define H_gtk_widget_set_parent_window "void gtk_widget_set_parent_window(GtkWidget* widget, GdkWindow* parent_window)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_widget_set_parent_window", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(parent_window), parent_window, 2, "gtk_widget_set_parent_window", "GdkWindow*");
  gtk_widget_set_parent_window(XEN_TO_C_GtkWidget_(widget), XEN_TO_C_GdkWindow_(parent_window));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_widget_set_child_visible(XEN widget, XEN is_visible)
{
  #define H_gtk_widget_set_child_visible "void gtk_widget_set_child_visible(GtkWidget* widget, gboolean is_visible)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_widget_set_child_visible", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(is_visible), is_visible, 2, "gtk_widget_set_child_visible", "gboolean");
  gtk_widget_set_child_visible(XEN_TO_C_GtkWidget_(widget), XEN_TO_C_gboolean(is_visible));
  return(XEN_FALSE);
}

static XEN gxg_gtk_widget_set_accel_path(XEN widget, XEN accel_path, XEN accel_group)
{
  #define H_gtk_widget_set_accel_path "void gtk_widget_set_accel_path(GtkWidget* widget, gchar* accel_path, \
GtkAccelGroup* accel_group)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_widget_set_accel_path", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_gchar__P(accel_path), accel_path, 2, "gtk_widget_set_accel_path", "gchar*");
  XEN_ASSERT_TYPE(XEN_GtkAccelGroup__P(accel_group), accel_group, 3, "gtk_widget_set_accel_path", "GtkAccelGroup*");
  gtk_widget_set_accel_path(XEN_TO_C_GtkWidget_(widget), XEN_TO_C_gchar_(accel_path), XEN_TO_C_GtkAccelGroup_(accel_group));
  return(XEN_FALSE);
}

static XEN gxg_gtk_widget_get_child_visible(XEN widget)
{
  #define H_gtk_widget_get_child_visible "gboolean gtk_widget_get_child_visible(GtkWidget* widget)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_widget_get_child_visible", "GtkWidget*");
  return(C_TO_XEN_gboolean(gtk_widget_get_child_visible(XEN_TO_C_GtkWidget_(widget))));
}

static XEN gxg_gtk_widget_get_parent(XEN widget)
{
  #define H_gtk_widget_get_parent "GtkWidget* gtk_widget_get_parent(GtkWidget* widget)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_widget_get_parent", "GtkWidget*");
  return(C_TO_XEN_GtkWidget_(gtk_widget_get_parent(XEN_TO_C_GtkWidget_(widget))));
}
#endif

static XEN gxg_gtk_widget_get_parent_window(XEN widget)
{
  #define H_gtk_widget_get_parent_window "GdkWindow* gtk_widget_get_parent_window(GtkWidget* widget)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_widget_get_parent_window", "GtkWidget*");
  return(C_TO_XEN_GdkWindow_(gtk_widget_get_parent_window(XEN_TO_C_GtkWidget_(widget))));
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_widget_child_focus(XEN widget, XEN direction)
{
  #define H_gtk_widget_child_focus "gboolean gtk_widget_child_focus(GtkWidget* widget, GtkDirectionType direction)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_widget_child_focus", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_GtkDirectionType_P(direction), direction, 2, "gtk_widget_child_focus", "GtkDirectionType");
  return(C_TO_XEN_gboolean(gtk_widget_child_focus(XEN_TO_C_GtkWidget_(widget), XEN_TO_C_GtkDirectionType(direction))));
}

static XEN gxg_gtk_widget_set_size_request(XEN widget, XEN width, XEN height)
{
  #define H_gtk_widget_set_size_request "void gtk_widget_set_size_request(GtkWidget* widget, gint width, \
gint height)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_widget_set_size_request", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_gint_P(width), width, 2, "gtk_widget_set_size_request", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(height), height, 3, "gtk_widget_set_size_request", "gint");
  gtk_widget_set_size_request(XEN_TO_C_GtkWidget_(widget), XEN_TO_C_gint(width), XEN_TO_C_gint(height));
  return(XEN_FALSE);
}

static XEN gxg_gtk_widget_get_size_request(XEN widget, XEN width, XEN height)
{
  #define H_gtk_widget_get_size_request "void gtk_widget_get_size_request(GtkWidget* widget, gint* [width], \
gint* [height])"
  gint ref_width;
  gint ref_height;
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_widget_get_size_request", "GtkWidget*");
  gtk_widget_get_size_request(XEN_TO_C_GtkWidget_(widget), &ref_width, &ref_height);
  return(XEN_LIST_2(C_TO_XEN_gint(ref_width), C_TO_XEN_gint(ref_height)));
}
#endif

static XEN gxg_gtk_widget_set_events(XEN widget, XEN events)
{
  #define H_gtk_widget_set_events "void gtk_widget_set_events(GtkWidget* widget, gint events)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_widget_set_events", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_gint_P(events), events, 2, "gtk_widget_set_events", "gint");
  gtk_widget_set_events(XEN_TO_C_GtkWidget_(widget), XEN_TO_C_gint(events));
  return(XEN_FALSE);
}

static XEN gxg_gtk_widget_add_events(XEN widget, XEN events)
{
  #define H_gtk_widget_add_events "void gtk_widget_add_events(GtkWidget* widget, gint events)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_widget_add_events", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_gint_P(events), events, 2, "gtk_widget_add_events", "gint");
  gtk_widget_add_events(XEN_TO_C_GtkWidget_(widget), XEN_TO_C_gint(events));
  return(XEN_FALSE);
}

static XEN gxg_gtk_widget_set_extension_events(XEN widget, XEN mode)
{
  #define H_gtk_widget_set_extension_events "void gtk_widget_set_extension_events(GtkWidget* widget, \
GdkExtensionMode mode)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_widget_set_extension_events", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_GdkExtensionMode_P(mode), mode, 2, "gtk_widget_set_extension_events", "GdkExtensionMode");
  gtk_widget_set_extension_events(XEN_TO_C_GtkWidget_(widget), XEN_TO_C_GdkExtensionMode(mode));
  return(XEN_FALSE);
}

static XEN gxg_gtk_widget_get_extension_events(XEN widget)
{
  #define H_gtk_widget_get_extension_events "GdkExtensionMode gtk_widget_get_extension_events(GtkWidget* widget)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_widget_get_extension_events", "GtkWidget*");
  return(C_TO_XEN_GdkExtensionMode(gtk_widget_get_extension_events(XEN_TO_C_GtkWidget_(widget))));
}

static XEN gxg_gtk_widget_get_toplevel(XEN widget)
{
  #define H_gtk_widget_get_toplevel "GtkWidget* gtk_widget_get_toplevel(GtkWidget* widget)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_widget_get_toplevel", "GtkWidget*");
  return(C_TO_XEN_GtkWidget_(gtk_widget_get_toplevel(XEN_TO_C_GtkWidget_(widget))));
}

static XEN gxg_gtk_widget_get_ancestor(XEN widget, XEN widget_type)
{
  #define H_gtk_widget_get_ancestor "GtkWidget* gtk_widget_get_ancestor(GtkWidget* widget, GtkType widget_type)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_widget_get_ancestor", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_GtkType_P(widget_type), widget_type, 2, "gtk_widget_get_ancestor", "GtkType");
  return(C_TO_XEN_GtkWidget_(gtk_widget_get_ancestor(XEN_TO_C_GtkWidget_(widget), XEN_TO_C_GtkType(widget_type))));
}

static XEN gxg_gtk_widget_get_colormap(XEN widget)
{
  #define H_gtk_widget_get_colormap "GdkColormap* gtk_widget_get_colormap(GtkWidget* widget)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_widget_get_colormap", "GtkWidget*");
  return(C_TO_XEN_GdkColormap_(gtk_widget_get_colormap(XEN_TO_C_GtkWidget_(widget))));
}

static XEN gxg_gtk_widget_get_visual(XEN widget)
{
  #define H_gtk_widget_get_visual "GdkVisual* gtk_widget_get_visual(GtkWidget* widget)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_widget_get_visual", "GtkWidget*");
  return(C_TO_XEN_GdkVisual_(gtk_widget_get_visual(XEN_TO_C_GtkWidget_(widget))));
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_widget_get_settings(XEN widget)
{
  #define H_gtk_widget_get_settings "GtkSettings* gtk_widget_get_settings(GtkWidget* widget)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_widget_get_settings", "GtkWidget*");
  return(C_TO_XEN_GtkSettings_(gtk_widget_get_settings(XEN_TO_C_GtkWidget_(widget))));
}

static XEN gxg_gtk_widget_get_accessible(XEN widget)
{
  #define H_gtk_widget_get_accessible "AtkObject* gtk_widget_get_accessible(GtkWidget* widget)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_widget_get_accessible", "GtkWidget*");
  return(C_TO_XEN_AtkObject_(gtk_widget_get_accessible(XEN_TO_C_GtkWidget_(widget))));
}
#endif

static XEN gxg_gtk_widget_set_colormap(XEN widget, XEN colormap)
{
  #define H_gtk_widget_set_colormap "void gtk_widget_set_colormap(GtkWidget* widget, GdkColormap* colormap)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_widget_set_colormap", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_GdkColormap__P(colormap), colormap, 2, "gtk_widget_set_colormap", "GdkColormap*");
  gtk_widget_set_colormap(XEN_TO_C_GtkWidget_(widget), XEN_TO_C_GdkColormap_(colormap));
  return(XEN_FALSE);
}

static XEN gxg_gtk_widget_get_events(XEN widget)
{
  #define H_gtk_widget_get_events "gint gtk_widget_get_events(GtkWidget* widget)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_widget_get_events", "GtkWidget*");
  return(C_TO_XEN_gint(gtk_widget_get_events(XEN_TO_C_GtkWidget_(widget))));
}

static XEN gxg_gtk_widget_get_pointer(XEN widget, XEN x, XEN y)
{
  #define H_gtk_widget_get_pointer "void gtk_widget_get_pointer(GtkWidget* widget, gint* [x], gint* [y])"
  gint ref_x;
  gint ref_y;
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_widget_get_pointer", "GtkWidget*");
  gtk_widget_get_pointer(XEN_TO_C_GtkWidget_(widget), &ref_x, &ref_y);
  return(XEN_LIST_2(C_TO_XEN_gint(ref_x), C_TO_XEN_gint(ref_y)));
}

static XEN gxg_gtk_widget_is_ancestor(XEN widget, XEN ancestor)
{
  #define H_gtk_widget_is_ancestor "gboolean gtk_widget_is_ancestor(GtkWidget* widget, GtkWidget* ancestor)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_widget_is_ancestor", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(ancestor), ancestor, 2, "gtk_widget_is_ancestor", "GtkWidget*");
  return(C_TO_XEN_gboolean(gtk_widget_is_ancestor(XEN_TO_C_GtkWidget_(widget), XEN_TO_C_GtkWidget_(ancestor))));
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_widget_translate_coordinates(XEN src_widget, XEN dest_widget, XEN src_x, XEN src_y, XEN dest_x, XEN dest_y)
{
  #define H_gtk_widget_translate_coordinates "gboolean gtk_widget_translate_coordinates(GtkWidget* src_widget, \
GtkWidget* dest_widget, gint src_x, gint src_y, gint* [dest_x], gint* [dest_y])"
  gint ref_dest_x;
  gint ref_dest_y;
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(src_widget), src_widget, 1, "gtk_widget_translate_coordinates", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(dest_widget), dest_widget, 2, "gtk_widget_translate_coordinates", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_gint_P(src_x), src_x, 3, "gtk_widget_translate_coordinates", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(src_y), src_y, 4, "gtk_widget_translate_coordinates", "gint");
  {
    XEN result = XEN_FALSE;
    result = C_TO_XEN_gboolean(gtk_widget_translate_coordinates(XEN_TO_C_GtkWidget_(src_widget), XEN_TO_C_GtkWidget_(dest_widget), 
                                                                XEN_TO_C_gint(src_x), XEN_TO_C_gint(src_y), &ref_dest_x, 
                                                                &ref_dest_y));
    return(XEN_LIST_3(result, C_TO_XEN_gint(ref_dest_x), C_TO_XEN_gint(ref_dest_y)));
   }
}
#endif

static XEN gxg_gtk_widget_hide_on_delete(XEN widget)
{
  #define H_gtk_widget_hide_on_delete "gboolean gtk_widget_hide_on_delete(GtkWidget* widget)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_widget_hide_on_delete", "GtkWidget*");
  return(C_TO_XEN_gboolean(gtk_widget_hide_on_delete(XEN_TO_C_GtkWidget_(widget))));
}

static XEN gxg_gtk_widget_set_style(XEN widget, XEN style)
{
  #define H_gtk_widget_set_style "void gtk_widget_set_style(GtkWidget* widget, GtkStyle* style)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_widget_set_style", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_GtkStyle__P(style), style, 2, "gtk_widget_set_style", "GtkStyle*");
  gtk_widget_set_style(XEN_TO_C_GtkWidget_(widget), XEN_TO_C_GtkStyle_(style));
  return(XEN_FALSE);
}

static XEN gxg_gtk_widget_ensure_style(XEN widget)
{
  #define H_gtk_widget_ensure_style "void gtk_widget_ensure_style(GtkWidget* widget)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_widget_ensure_style", "GtkWidget*");
  gtk_widget_ensure_style(XEN_TO_C_GtkWidget_(widget));
  return(XEN_FALSE);
}

static XEN gxg_gtk_widget_get_style(XEN widget)
{
  #define H_gtk_widget_get_style "GtkStyle* gtk_widget_get_style(GtkWidget* widget)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_widget_get_style", "GtkWidget*");
  return(C_TO_XEN_GtkStyle_(gtk_widget_get_style(XEN_TO_C_GtkWidget_(widget))));
}

static XEN gxg_gtk_widget_modify_style(XEN widget, XEN style)
{
  #define H_gtk_widget_modify_style "void gtk_widget_modify_style(GtkWidget* widget, GtkRcStyle* style)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_widget_modify_style", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_GtkRcStyle__P(style), style, 2, "gtk_widget_modify_style", "GtkRcStyle*");
  gtk_widget_modify_style(XEN_TO_C_GtkWidget_(widget), XEN_TO_C_GtkRcStyle_(style));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_widget_get_modifier_style(XEN widget)
{
  #define H_gtk_widget_get_modifier_style "GtkRcStyle* gtk_widget_get_modifier_style(GtkWidget* widget)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_widget_get_modifier_style", "GtkWidget*");
  return(C_TO_XEN_GtkRcStyle_(gtk_widget_get_modifier_style(XEN_TO_C_GtkWidget_(widget))));
}

static XEN gxg_gtk_widget_modify_fg(XEN widget, XEN state, XEN color)
{
  #define H_gtk_widget_modify_fg "void gtk_widget_modify_fg(GtkWidget* widget, GtkStateType state, GdkColor* color)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_widget_modify_fg", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_GtkStateType_P(state), state, 2, "gtk_widget_modify_fg", "GtkStateType");
  XEN_ASSERT_TYPE(XEN_GdkColor__P(color), color, 3, "gtk_widget_modify_fg", "GdkColor*");
  gtk_widget_modify_fg(XEN_TO_C_GtkWidget_(widget), XEN_TO_C_GtkStateType(state), XEN_TO_C_GdkColor_(color));
  return(XEN_FALSE);
}

static XEN gxg_gtk_widget_modify_bg(XEN widget, XEN state, XEN color)
{
  #define H_gtk_widget_modify_bg "void gtk_widget_modify_bg(GtkWidget* widget, GtkStateType state, GdkColor* color)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_widget_modify_bg", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_GtkStateType_P(state), state, 2, "gtk_widget_modify_bg", "GtkStateType");
  XEN_ASSERT_TYPE(XEN_GdkColor__P(color), color, 3, "gtk_widget_modify_bg", "GdkColor*");
  gtk_widget_modify_bg(XEN_TO_C_GtkWidget_(widget), XEN_TO_C_GtkStateType(state), XEN_TO_C_GdkColor_(color));
  return(XEN_FALSE);
}

static XEN gxg_gtk_widget_modify_text(XEN widget, XEN state, XEN color)
{
  #define H_gtk_widget_modify_text "void gtk_widget_modify_text(GtkWidget* widget, GtkStateType state, \
GdkColor* color)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_widget_modify_text", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_GtkStateType_P(state), state, 2, "gtk_widget_modify_text", "GtkStateType");
  XEN_ASSERT_TYPE(XEN_GdkColor__P(color), color, 3, "gtk_widget_modify_text", "GdkColor*");
  gtk_widget_modify_text(XEN_TO_C_GtkWidget_(widget), XEN_TO_C_GtkStateType(state), XEN_TO_C_GdkColor_(color));
  return(XEN_FALSE);
}

static XEN gxg_gtk_widget_modify_base(XEN widget, XEN state, XEN color)
{
  #define H_gtk_widget_modify_base "void gtk_widget_modify_base(GtkWidget* widget, GtkStateType state, \
GdkColor* color)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_widget_modify_base", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_GtkStateType_P(state), state, 2, "gtk_widget_modify_base", "GtkStateType");
  XEN_ASSERT_TYPE(XEN_GdkColor__P(color), color, 3, "gtk_widget_modify_base", "GdkColor*");
  gtk_widget_modify_base(XEN_TO_C_GtkWidget_(widget), XEN_TO_C_GtkStateType(state), XEN_TO_C_GdkColor_(color));
  return(XEN_FALSE);
}

static XEN gxg_gtk_widget_modify_font(XEN widget, XEN font_desc)
{
  #define H_gtk_widget_modify_font "void gtk_widget_modify_font(GtkWidget* widget, PangoFontDescription* font_desc)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_widget_modify_font", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_PangoFontDescription__P(font_desc), font_desc, 2, "gtk_widget_modify_font", "PangoFontDescription*");
  gtk_widget_modify_font(XEN_TO_C_GtkWidget_(widget), XEN_TO_C_PangoFontDescription_(font_desc));
  return(XEN_FALSE);
}

static XEN gxg_gtk_widget_create_pango_context(XEN widget)
{
  #define H_gtk_widget_create_pango_context "PangoContext* gtk_widget_create_pango_context(GtkWidget* widget)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_widget_create_pango_context", "GtkWidget*");
  return(C_TO_XEN_PangoContext_(gtk_widget_create_pango_context(XEN_TO_C_GtkWidget_(widget))));
}

static XEN gxg_gtk_widget_get_pango_context(XEN widget)
{
  #define H_gtk_widget_get_pango_context "PangoContext* gtk_widget_get_pango_context(GtkWidget* widget)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_widget_get_pango_context", "GtkWidget*");
  return(C_TO_XEN_PangoContext_(gtk_widget_get_pango_context(XEN_TO_C_GtkWidget_(widget))));
}

static XEN gxg_gtk_widget_create_pango_layout(XEN widget, XEN text)
{
  #define H_gtk_widget_create_pango_layout "PangoLayout* gtk_widget_create_pango_layout(GtkWidget* widget, \
gchar* text)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_widget_create_pango_layout", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_gchar__P(text), text, 2, "gtk_widget_create_pango_layout", "gchar*");
  return(C_TO_XEN_PangoLayout_(gtk_widget_create_pango_layout(XEN_TO_C_GtkWidget_(widget), XEN_TO_C_gchar_(text))));
}

static XEN gxg_gtk_widget_render_icon(XEN widget, XEN stock_id, XEN size, XEN detail)
{
  #define H_gtk_widget_render_icon "GdkPixbuf* gtk_widget_render_icon(GtkWidget* widget, gchar* stock_id, \
GtkIconSize size, gchar* detail)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_widget_render_icon", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_gchar__P(stock_id), stock_id, 2, "gtk_widget_render_icon", "gchar*");
  XEN_ASSERT_TYPE(XEN_GtkIconSize_P(size), size, 3, "gtk_widget_render_icon", "GtkIconSize");
  XEN_ASSERT_TYPE(XEN_gchar__P(detail), detail, 4, "gtk_widget_render_icon", "gchar*");
  return(C_TO_XEN_GdkPixbuf_(gtk_widget_render_icon(XEN_TO_C_GtkWidget_(widget), XEN_TO_C_gchar_(stock_id), XEN_TO_C_GtkIconSize(size), 
                                                    XEN_TO_C_gchar_(detail))));
}
#endif

static XEN gxg_gtk_widget_set_composite_name(XEN widget, XEN name)
{
  #define H_gtk_widget_set_composite_name "void gtk_widget_set_composite_name(GtkWidget* widget, gchar* name)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_widget_set_composite_name", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_gchar__P(name), name, 2, "gtk_widget_set_composite_name", "gchar*");
  gtk_widget_set_composite_name(XEN_TO_C_GtkWidget_(widget), XEN_TO_C_gchar_(name));
  return(XEN_FALSE);
}

static XEN gxg_gtk_widget_get_composite_name(XEN widget)
{
  #define H_gtk_widget_get_composite_name "gchar* gtk_widget_get_composite_name(GtkWidget* widget)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_widget_get_composite_name", "GtkWidget*");
  return(C_TO_XEN_gchar_(gtk_widget_get_composite_name(XEN_TO_C_GtkWidget_(widget))));
}

static XEN gxg_gtk_widget_reset_rc_styles(XEN widget)
{
  #define H_gtk_widget_reset_rc_styles "void gtk_widget_reset_rc_styles(GtkWidget* widget)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_widget_reset_rc_styles", "GtkWidget*");
  gtk_widget_reset_rc_styles(XEN_TO_C_GtkWidget_(widget));
  return(XEN_FALSE);
}

static XEN gxg_gtk_widget_push_colormap(XEN cmap)
{
  #define H_gtk_widget_push_colormap "void gtk_widget_push_colormap(GdkColormap* cmap)"
  XEN_ASSERT_TYPE(XEN_GdkColormap__P(cmap), cmap, 1, "gtk_widget_push_colormap", "GdkColormap*");
  gtk_widget_push_colormap(XEN_TO_C_GdkColormap_(cmap));
  return(XEN_FALSE);
}

static XEN gxg_gtk_widget_push_composite_child(void)
{
  #define H_gtk_widget_push_composite_child "void gtk_widget_push_composite_child( void)"
  gtk_widget_push_composite_child();
  return(XEN_FALSE);
}

static XEN gxg_gtk_widget_pop_composite_child(void)
{
  #define H_gtk_widget_pop_composite_child "void gtk_widget_pop_composite_child( void)"
  gtk_widget_pop_composite_child();
  return(XEN_FALSE);
}

static XEN gxg_gtk_widget_pop_colormap(void)
{
  #define H_gtk_widget_pop_colormap "void gtk_widget_pop_colormap( void)"
  gtk_widget_pop_colormap();
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_widget_style_get_property(XEN widget, XEN property_name, XEN value)
{
  #define H_gtk_widget_style_get_property "void gtk_widget_style_get_property(GtkWidget* widget, gchar* property_name, \
GValue* value)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_widget_style_get_property", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_gchar__P(property_name), property_name, 2, "gtk_widget_style_get_property", "gchar*");
  XEN_ASSERT_TYPE(XEN_GValue__P(value), value, 3, "gtk_widget_style_get_property", "GValue*");
  gtk_widget_style_get_property(XEN_TO_C_GtkWidget_(widget), XEN_TO_C_gchar_(property_name), XEN_TO_C_GValue_(value));
  return(XEN_FALSE);
}

static XEN gxg_gtk_widget_style_get(XEN widget, XEN first_property_name)
{
  #define H_gtk_widget_style_get "void gtk_widget_style_get(GtkWidget* widget, gchar* first_property_name, \
...)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_widget_style_get", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_gchar__P(first_property_name), first_property_name, 2, "gtk_widget_style_get", "gchar*");
  gtk_widget_style_get(XEN_TO_C_GtkWidget_(widget), XEN_TO_C_gchar_(first_property_name));
  return(XEN_FALSE);
}
#endif

static XEN gxg_gtk_widget_set_default_colormap(XEN colormap)
{
  #define H_gtk_widget_set_default_colormap "void gtk_widget_set_default_colormap(GdkColormap* colormap)"
  XEN_ASSERT_TYPE(XEN_GdkColormap__P(colormap), colormap, 1, "gtk_widget_set_default_colormap", "GdkColormap*");
  gtk_widget_set_default_colormap(XEN_TO_C_GdkColormap_(colormap));
  return(XEN_FALSE);
}

static XEN gxg_gtk_widget_get_default_style(void)
{
  #define H_gtk_widget_get_default_style "GtkStyle* gtk_widget_get_default_style( void)"
  return(C_TO_XEN_GtkStyle_(gtk_widget_get_default_style()));
}

static XEN gxg_gtk_widget_get_default_colormap(void)
{
  #define H_gtk_widget_get_default_colormap "GdkColormap* gtk_widget_get_default_colormap( void)"
  return(C_TO_XEN_GdkColormap_(gtk_widget_get_default_colormap()));
}

static XEN gxg_gtk_widget_get_default_visual(void)
{
  #define H_gtk_widget_get_default_visual "GdkVisual* gtk_widget_get_default_visual( void)"
  return(C_TO_XEN_GdkVisual_(gtk_widget_get_default_visual()));
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_widget_set_direction(XEN widget, XEN dir)
{
  #define H_gtk_widget_set_direction "void gtk_widget_set_direction(GtkWidget* widget, GtkTextDirection dir)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_widget_set_direction", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_GtkTextDirection_P(dir), dir, 2, "gtk_widget_set_direction", "GtkTextDirection");
  gtk_widget_set_direction(XEN_TO_C_GtkWidget_(widget), XEN_TO_C_GtkTextDirection(dir));
  return(XEN_FALSE);
}

static XEN gxg_gtk_widget_get_direction(XEN widget)
{
  #define H_gtk_widget_get_direction "GtkTextDirection gtk_widget_get_direction(GtkWidget* widget)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_widget_get_direction", "GtkWidget*");
  return(C_TO_XEN_GtkTextDirection(gtk_widget_get_direction(XEN_TO_C_GtkWidget_(widget))));
}

static XEN gxg_gtk_widget_set_default_direction(XEN dir)
{
  #define H_gtk_widget_set_default_direction "void gtk_widget_set_default_direction(GtkTextDirection dir)"
  XEN_ASSERT_TYPE(XEN_GtkTextDirection_P(dir), dir, 1, "gtk_widget_set_default_direction", "GtkTextDirection");
  gtk_widget_set_default_direction(XEN_TO_C_GtkTextDirection(dir));
  return(XEN_FALSE);
}

static XEN gxg_gtk_widget_get_default_direction(void)
{
  #define H_gtk_widget_get_default_direction "GtkTextDirection gtk_widget_get_default_direction( void)"
  return(C_TO_XEN_GtkTextDirection(gtk_widget_get_default_direction()));
}
#endif

static XEN gxg_gtk_widget_shape_combine_mask(XEN widget, XEN shape_mask, XEN offset_x, XEN offset_y)
{
  #define H_gtk_widget_shape_combine_mask "void gtk_widget_shape_combine_mask(GtkWidget* widget, GdkBitmap* shape_mask, \
gint offset_x, gint offset_y)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_widget_shape_combine_mask", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_GdkBitmap__P(shape_mask) || XEN_FALSE_P(shape_mask), shape_mask, 2, "gtk_widget_shape_combine_mask", "GdkBitmap*");
  XEN_ASSERT_TYPE(XEN_gint_P(offset_x), offset_x, 3, "gtk_widget_shape_combine_mask", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(offset_y), offset_y, 4, "gtk_widget_shape_combine_mask", "gint");
  gtk_widget_shape_combine_mask(XEN_TO_C_GtkWidget_(widget), XEN_TO_C_GdkBitmap_(shape_mask), XEN_TO_C_gint(offset_x), XEN_TO_C_gint(offset_y));
  return(XEN_FALSE);
}

static XEN gxg_gtk_widget_reset_shapes(XEN widget)
{
  #define H_gtk_widget_reset_shapes "void gtk_widget_reset_shapes(GtkWidget* widget)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_widget_reset_shapes", "GtkWidget*");
  gtk_widget_reset_shapes(XEN_TO_C_GtkWidget_(widget));
  return(XEN_FALSE);
}

static XEN gxg_gtk_widget_path(XEN widget, XEN path_length, XEN path, XEN path_reversed)
{
  #define H_gtk_widget_path "void gtk_widget_path(GtkWidget* widget, guint* path_length, gchar** [path], \
gchar** [path_reversed])"
  gchar* ref_path;
  gchar* ref_path_reversed;
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_widget_path", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_guint__P(path_length), path_length, 2, "gtk_widget_path", "guint*");
  gtk_widget_path(XEN_TO_C_GtkWidget_(widget), XEN_TO_C_guint_(path_length), &ref_path, &ref_path_reversed);
  return(XEN_LIST_2(C_TO_XEN_gchar_(ref_path), C_TO_XEN_gchar_(ref_path_reversed)));
}

static XEN gxg_gtk_widget_class_path(XEN widget, XEN path_length, XEN path, XEN path_reversed)
{
  #define H_gtk_widget_class_path "void gtk_widget_class_path(GtkWidget* widget, guint* path_length, \
gchar** [path], gchar** [path_reversed])"
  gchar* ref_path;
  gchar* ref_path_reversed;
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_widget_class_path", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_guint__P(path_length), path_length, 2, "gtk_widget_class_path", "guint*");
  gtk_widget_class_path(XEN_TO_C_GtkWidget_(widget), XEN_TO_C_guint_(path_length), &ref_path, &ref_path_reversed);
  return(XEN_LIST_2(C_TO_XEN_gchar_(ref_path), C_TO_XEN_gchar_(ref_path_reversed)));
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_requisition_copy(XEN requisition)
{
  #define H_gtk_requisition_copy "GtkRequisition* gtk_requisition_copy(GtkRequisition* requisition)"
  XEN_ASSERT_TYPE(XEN_GtkRequisition__P(requisition), requisition, 1, "gtk_requisition_copy", "GtkRequisition*");
  return(C_TO_XEN_GtkRequisition_(gtk_requisition_copy(XEN_TO_C_GtkRequisition_(requisition))));
}

static XEN gxg_gtk_requisition_free(XEN requisition)
{
  #define H_gtk_requisition_free "void gtk_requisition_free(GtkRequisition* requisition)"
  XEN_ASSERT_TYPE(XEN_GtkRequisition__P(requisition), requisition, 1, "gtk_requisition_free", "GtkRequisition*");
  gtk_requisition_free(XEN_TO_C_GtkRequisition_(requisition));
  return(XEN_FALSE);
}

static XEN gxg__gtk_widget_get_aux_info(XEN widget, XEN create)
{
  #define H__gtk_widget_get_aux_info "GtkWidgetAuxInfo* _gtk_widget_get_aux_info(GtkWidget* widget, gboolean create)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "_gtk_widget_get_aux_info", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(create), create, 2, "_gtk_widget_get_aux_info", "gboolean");
  return(C_TO_XEN_GtkWidgetAuxInfo_(_gtk_widget_get_aux_info(XEN_TO_C_GtkWidget_(widget), XEN_TO_C_gboolean(create))));
}
#endif

static XEN gxg_gtk_window_get_type(void)
{
  #define H_gtk_window_get_type "GtkType gtk_window_get_type( void)"
  return(C_TO_XEN_GtkType(gtk_window_get_type()));
}

static XEN gxg_gtk_window_new(XEN type)
{
  #define H_gtk_window_new "GtkWidget* gtk_window_new(GtkWindowType type)"
  XEN_ASSERT_TYPE(XEN_GtkWindowType_P(type), type, 1, "gtk_window_new", "GtkWindowType");
  return(C_TO_XEN_GtkWidget_(gtk_window_new(XEN_TO_C_GtkWindowType(type))));
}

static XEN gxg_gtk_window_set_title(XEN window, XEN title)
{
  #define H_gtk_window_set_title "void gtk_window_set_title(GtkWindow* window, gchar* title)"
  XEN_ASSERT_TYPE(XEN_GtkWindow__P(window), window, 1, "gtk_window_set_title", "GtkWindow*");
  XEN_ASSERT_TYPE(XEN_gchar__P(title), title, 2, "gtk_window_set_title", "gchar*");
  gtk_window_set_title(XEN_TO_C_GtkWindow_(window), XEN_TO_C_gchar_(title));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_window_get_title(XEN window)
{
  #define H_gtk_window_get_title "gchar* gtk_window_get_title(GtkWindow* window)"
  XEN_ASSERT_TYPE(XEN_GtkWindow__P(window), window, 1, "gtk_window_get_title", "GtkWindow*");
  return(C_TO_XEN_gchar_(gtk_window_get_title(XEN_TO_C_GtkWindow_(window))));
}
#endif

static XEN gxg_gtk_window_set_wmclass(XEN window, XEN wmclass_name, XEN wmclass_class)
{
  #define H_gtk_window_set_wmclass "void gtk_window_set_wmclass(GtkWindow* window, gchar* wmclass_name, \
gchar* wmclass_class)"
  XEN_ASSERT_TYPE(XEN_GtkWindow__P(window), window, 1, "gtk_window_set_wmclass", "GtkWindow*");
  XEN_ASSERT_TYPE(XEN_gchar__P(wmclass_name), wmclass_name, 2, "gtk_window_set_wmclass", "gchar*");
  XEN_ASSERT_TYPE(XEN_gchar__P(wmclass_class), wmclass_class, 3, "gtk_window_set_wmclass", "gchar*");
  gtk_window_set_wmclass(XEN_TO_C_GtkWindow_(window), XEN_TO_C_gchar_(wmclass_name), XEN_TO_C_gchar_(wmclass_class));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_window_set_role(XEN window, XEN role)
{
  #define H_gtk_window_set_role "void gtk_window_set_role(GtkWindow* window, gchar* role)"
  XEN_ASSERT_TYPE(XEN_GtkWindow__P(window), window, 1, "gtk_window_set_role", "GtkWindow*");
  XEN_ASSERT_TYPE(XEN_gchar__P(role), role, 2, "gtk_window_set_role", "gchar*");
  gtk_window_set_role(XEN_TO_C_GtkWindow_(window), XEN_TO_C_gchar_(role));
  return(XEN_FALSE);
}

static XEN gxg_gtk_window_get_role(XEN window)
{
  #define H_gtk_window_get_role "gchar* gtk_window_get_role(GtkWindow* window)"
  XEN_ASSERT_TYPE(XEN_GtkWindow__P(window), window, 1, "gtk_window_get_role", "GtkWindow*");
  return(C_TO_XEN_gchar_(gtk_window_get_role(XEN_TO_C_GtkWindow_(window))));
}
#endif

static XEN gxg_gtk_window_add_accel_group(XEN window, XEN accel_group)
{
  #define H_gtk_window_add_accel_group "void gtk_window_add_accel_group(GtkWindow* window, GtkAccelGroup* accel_group)"
  XEN_ASSERT_TYPE(XEN_GtkWindow__P(window), window, 1, "gtk_window_add_accel_group", "GtkWindow*");
  XEN_ASSERT_TYPE(XEN_GtkAccelGroup__P(accel_group), accel_group, 2, "gtk_window_add_accel_group", "GtkAccelGroup*");
  gtk_window_add_accel_group(XEN_TO_C_GtkWindow_(window), XEN_TO_C_GtkAccelGroup_(accel_group));
  return(XEN_FALSE);
}

static XEN gxg_gtk_window_remove_accel_group(XEN window, XEN accel_group)
{
  #define H_gtk_window_remove_accel_group "void gtk_window_remove_accel_group(GtkWindow* window, GtkAccelGroup* accel_group)"
  XEN_ASSERT_TYPE(XEN_GtkWindow__P(window), window, 1, "gtk_window_remove_accel_group", "GtkWindow*");
  XEN_ASSERT_TYPE(XEN_GtkAccelGroup__P(accel_group), accel_group, 2, "gtk_window_remove_accel_group", "GtkAccelGroup*");
  gtk_window_remove_accel_group(XEN_TO_C_GtkWindow_(window), XEN_TO_C_GtkAccelGroup_(accel_group));
  return(XEN_FALSE);
}

static XEN gxg_gtk_window_set_position(XEN window, XEN position)
{
  #define H_gtk_window_set_position "void gtk_window_set_position(GtkWindow* window, GtkWindowPosition position)"
  XEN_ASSERT_TYPE(XEN_GtkWindow__P(window), window, 1, "gtk_window_set_position", "GtkWindow*");
  XEN_ASSERT_TYPE(XEN_GtkWindowPosition_P(position), position, 2, "gtk_window_set_position", "GtkWindowPosition");
  gtk_window_set_position(XEN_TO_C_GtkWindow_(window), XEN_TO_C_GtkWindowPosition(position));
  return(XEN_FALSE);
}

static XEN gxg_gtk_window_activate_focus(XEN window)
{
  #define H_gtk_window_activate_focus "gboolean gtk_window_activate_focus(GtkWindow* window)"
  XEN_ASSERT_TYPE(XEN_GtkWindow__P(window), window, 1, "gtk_window_activate_focus", "GtkWindow*");
  return(C_TO_XEN_gboolean(gtk_window_activate_focus(XEN_TO_C_GtkWindow_(window))));
}

static XEN gxg_gtk_window_set_focus(XEN window, XEN focus)
{
  #define H_gtk_window_set_focus "void gtk_window_set_focus(GtkWindow* window, GtkWidget* focus)"
  XEN_ASSERT_TYPE(XEN_GtkWindow__P(window), window, 1, "gtk_window_set_focus", "GtkWindow*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(focus) || XEN_FALSE_P(focus), focus, 2, "gtk_window_set_focus", "GtkWidget*");
  gtk_window_set_focus(XEN_TO_C_GtkWindow_(window), XEN_TO_C_GtkWidget_(focus));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_window_get_focus(XEN window)
{
  #define H_gtk_window_get_focus "GtkWidget* gtk_window_get_focus(GtkWindow* window)"
  XEN_ASSERT_TYPE(XEN_GtkWindow__P(window), window, 1, "gtk_window_get_focus", "GtkWindow*");
  return(C_TO_XEN_GtkWidget_(gtk_window_get_focus(XEN_TO_C_GtkWindow_(window))));
}
#endif

static XEN gxg_gtk_window_set_default(XEN window, XEN default_widget)
{
  #define H_gtk_window_set_default "void gtk_window_set_default(GtkWindow* window, GtkWidget* default_widget)"
  XEN_ASSERT_TYPE(XEN_GtkWindow__P(window), window, 1, "gtk_window_set_default", "GtkWindow*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(default_widget) || XEN_FALSE_P(default_widget), default_widget, 2, "gtk_window_set_default", "GtkWidget*");
  gtk_window_set_default(XEN_TO_C_GtkWindow_(window), XEN_TO_C_GtkWidget_(default_widget));
  return(XEN_FALSE);
}

static XEN gxg_gtk_window_activate_default(XEN window)
{
  #define H_gtk_window_activate_default "gboolean gtk_window_activate_default(GtkWindow* window)"
  XEN_ASSERT_TYPE(XEN_GtkWindow__P(window), window, 1, "gtk_window_activate_default", "GtkWindow*");
  return(C_TO_XEN_gboolean(gtk_window_activate_default(XEN_TO_C_GtkWindow_(window))));
}

static XEN gxg_gtk_window_set_transient_for(XEN window, XEN parent)
{
  #define H_gtk_window_set_transient_for "void gtk_window_set_transient_for(GtkWindow* window, GtkWindow* parent)"
  XEN_ASSERT_TYPE(XEN_GtkWindow__P(window), window, 1, "gtk_window_set_transient_for", "GtkWindow*");
  XEN_ASSERT_TYPE(XEN_GtkWindow__P(parent), parent, 2, "gtk_window_set_transient_for", "GtkWindow*");
  gtk_window_set_transient_for(XEN_TO_C_GtkWindow_(window), XEN_TO_C_GtkWindow_(parent));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_window_get_transient_for(XEN window)
{
  #define H_gtk_window_get_transient_for "GtkWindow* gtk_window_get_transient_for(GtkWindow* window)"
  XEN_ASSERT_TYPE(XEN_GtkWindow__P(window), window, 1, "gtk_window_get_transient_for", "GtkWindow*");
  return(C_TO_XEN_GtkWindow_(gtk_window_get_transient_for(XEN_TO_C_GtkWindow_(window))));
}

static XEN gxg_gtk_window_set_type_hint(XEN window, XEN hint)
{
  #define H_gtk_window_set_type_hint "void gtk_window_set_type_hint(GtkWindow* window, GdkWindowTypeHint hint)"
  XEN_ASSERT_TYPE(XEN_GtkWindow__P(window), window, 1, "gtk_window_set_type_hint", "GtkWindow*");
  XEN_ASSERT_TYPE(XEN_GdkWindowTypeHint_P(hint), hint, 2, "gtk_window_set_type_hint", "GdkWindowTypeHint");
  gtk_window_set_type_hint(XEN_TO_C_GtkWindow_(window), XEN_TO_C_GdkWindowTypeHint(hint));
  return(XEN_FALSE);
}

static XEN gxg_gtk_window_get_type_hint(XEN window)
{
  #define H_gtk_window_get_type_hint "GdkWindowTypeHint gtk_window_get_type_hint(GtkWindow* window)"
  XEN_ASSERT_TYPE(XEN_GtkWindow__P(window), window, 1, "gtk_window_get_type_hint", "GtkWindow*");
  return(C_TO_XEN_GdkWindowTypeHint(gtk_window_get_type_hint(XEN_TO_C_GtkWindow_(window))));
}

static XEN gxg_gtk_window_set_destroy_with_parent(XEN window, XEN setting)
{
  #define H_gtk_window_set_destroy_with_parent "void gtk_window_set_destroy_with_parent(GtkWindow* window, \
gboolean setting)"
  XEN_ASSERT_TYPE(XEN_GtkWindow__P(window), window, 1, "gtk_window_set_destroy_with_parent", "GtkWindow*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(setting), setting, 2, "gtk_window_set_destroy_with_parent", "gboolean");
  gtk_window_set_destroy_with_parent(XEN_TO_C_GtkWindow_(window), XEN_TO_C_gboolean(setting));
  return(XEN_FALSE);
}

static XEN gxg_gtk_window_get_destroy_with_parent(XEN window)
{
  #define H_gtk_window_get_destroy_with_parent "gboolean gtk_window_get_destroy_with_parent(GtkWindow* window)"
  XEN_ASSERT_TYPE(XEN_GtkWindow__P(window), window, 1, "gtk_window_get_destroy_with_parent", "GtkWindow*");
  return(C_TO_XEN_gboolean(gtk_window_get_destroy_with_parent(XEN_TO_C_GtkWindow_(window))));
}

static XEN gxg_gtk_window_set_resizable(XEN window, XEN resizable)
{
  #define H_gtk_window_set_resizable "void gtk_window_set_resizable(GtkWindow* window, gboolean resizable)"
  XEN_ASSERT_TYPE(XEN_GtkWindow__P(window), window, 1, "gtk_window_set_resizable", "GtkWindow*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(resizable), resizable, 2, "gtk_window_set_resizable", "gboolean");
  gtk_window_set_resizable(XEN_TO_C_GtkWindow_(window), XEN_TO_C_gboolean(resizable));
  return(XEN_FALSE);
}

static XEN gxg_gtk_window_get_resizable(XEN window)
{
  #define H_gtk_window_get_resizable "gboolean gtk_window_get_resizable(GtkWindow* window)"
  XEN_ASSERT_TYPE(XEN_GtkWindow__P(window), window, 1, "gtk_window_get_resizable", "GtkWindow*");
  return(C_TO_XEN_gboolean(gtk_window_get_resizable(XEN_TO_C_GtkWindow_(window))));
}

static XEN gxg_gtk_window_set_gravity(XEN window, XEN gravity)
{
  #define H_gtk_window_set_gravity "void gtk_window_set_gravity(GtkWindow* window, GdkGravity gravity)"
  XEN_ASSERT_TYPE(XEN_GtkWindow__P(window), window, 1, "gtk_window_set_gravity", "GtkWindow*");
  XEN_ASSERT_TYPE(XEN_GdkGravity_P(gravity), gravity, 2, "gtk_window_set_gravity", "GdkGravity");
  gtk_window_set_gravity(XEN_TO_C_GtkWindow_(window), XEN_TO_C_GdkGravity(gravity));
  return(XEN_FALSE);
}

static XEN gxg_gtk_window_get_gravity(XEN window)
{
  #define H_gtk_window_get_gravity "GdkGravity gtk_window_get_gravity(GtkWindow* window)"
  XEN_ASSERT_TYPE(XEN_GtkWindow__P(window), window, 1, "gtk_window_get_gravity", "GtkWindow*");
  return(C_TO_XEN_GdkGravity(gtk_window_get_gravity(XEN_TO_C_GtkWindow_(window))));
}
#endif

static XEN gxg_gtk_window_set_geometry_hints(XEN window, XEN geometry_widget, XEN geometry, XEN geom_mask)
{
  #define H_gtk_window_set_geometry_hints "void gtk_window_set_geometry_hints(GtkWindow* window, GtkWidget* geometry_widget, \
GdkGeometry* geometry, GdkWindowHints geom_mask)"
  XEN_ASSERT_TYPE(XEN_GtkWindow__P(window), window, 1, "gtk_window_set_geometry_hints", "GtkWindow*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(geometry_widget), geometry_widget, 2, "gtk_window_set_geometry_hints", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_GdkGeometry__P(geometry), geometry, 3, "gtk_window_set_geometry_hints", "GdkGeometry*");
  XEN_ASSERT_TYPE(XEN_GdkWindowHints_P(geom_mask), geom_mask, 4, "gtk_window_set_geometry_hints", "GdkWindowHints");
  gtk_window_set_geometry_hints(XEN_TO_C_GtkWindow_(window), XEN_TO_C_GtkWidget_(geometry_widget), XEN_TO_C_GdkGeometry_(geometry), 
                                XEN_TO_C_GdkWindowHints(geom_mask));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_window_set_has_frame(XEN window, XEN setting)
{
  #define H_gtk_window_set_has_frame "void gtk_window_set_has_frame(GtkWindow* window, gboolean setting)"
  XEN_ASSERT_TYPE(XEN_GtkWindow__P(window), window, 1, "gtk_window_set_has_frame", "GtkWindow*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(setting), setting, 2, "gtk_window_set_has_frame", "gboolean");
  gtk_window_set_has_frame(XEN_TO_C_GtkWindow_(window), XEN_TO_C_gboolean(setting));
  return(XEN_FALSE);
}

static XEN gxg_gtk_window_get_has_frame(XEN window)
{
  #define H_gtk_window_get_has_frame "gboolean gtk_window_get_has_frame(GtkWindow* window)"
  XEN_ASSERT_TYPE(XEN_GtkWindow__P(window), window, 1, "gtk_window_get_has_frame", "GtkWindow*");
  return(C_TO_XEN_gboolean(gtk_window_get_has_frame(XEN_TO_C_GtkWindow_(window))));
}

static XEN gxg_gtk_window_set_frame_dimensions(XEN window, XEN left, XEN top, XEN right, XEN bottom)
{
  #define H_gtk_window_set_frame_dimensions "void gtk_window_set_frame_dimensions(GtkWindow* window, \
gint left, gint top, gint right, gint bottom)"
  XEN_ASSERT_TYPE(XEN_GtkWindow__P(window), window, 1, "gtk_window_set_frame_dimensions", "GtkWindow*");
  XEN_ASSERT_TYPE(XEN_gint_P(left), left, 2, "gtk_window_set_frame_dimensions", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(top), top, 3, "gtk_window_set_frame_dimensions", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(right), right, 4, "gtk_window_set_frame_dimensions", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(bottom), bottom, 5, "gtk_window_set_frame_dimensions", "gint");
  gtk_window_set_frame_dimensions(XEN_TO_C_GtkWindow_(window), XEN_TO_C_gint(left), XEN_TO_C_gint(top), XEN_TO_C_gint(right), 
                                  XEN_TO_C_gint(bottom));
  return(XEN_FALSE);
}

static XEN gxg_gtk_window_get_frame_dimensions(XEN window, XEN left, XEN top, XEN right, XEN bottom)
{
  #define H_gtk_window_get_frame_dimensions "void gtk_window_get_frame_dimensions(GtkWindow* window, \
gint* [left], gint* [top], gint* [right], gint* [bottom])"
  gint ref_left;
  gint ref_top;
  gint ref_right;
  gint ref_bottom;
  XEN_ASSERT_TYPE(XEN_GtkWindow__P(window), window, 1, "gtk_window_get_frame_dimensions", "GtkWindow*");
  gtk_window_get_frame_dimensions(XEN_TO_C_GtkWindow_(window), &ref_left, &ref_top, &ref_right, &ref_bottom);
  return(XEN_LIST_4(C_TO_XEN_gint(ref_left), C_TO_XEN_gint(ref_top), C_TO_XEN_gint(ref_right), C_TO_XEN_gint(ref_bottom)));
}

static XEN gxg_gtk_window_set_decorated(XEN window, XEN setting)
{
  #define H_gtk_window_set_decorated "void gtk_window_set_decorated(GtkWindow* window, gboolean setting)"
  XEN_ASSERT_TYPE(XEN_GtkWindow__P(window), window, 1, "gtk_window_set_decorated", "GtkWindow*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(setting), setting, 2, "gtk_window_set_decorated", "gboolean");
  gtk_window_set_decorated(XEN_TO_C_GtkWindow_(window), XEN_TO_C_gboolean(setting));
  return(XEN_FALSE);
}

static XEN gxg_gtk_window_get_decorated(XEN window)
{
  #define H_gtk_window_get_decorated "gboolean gtk_window_get_decorated(GtkWindow* window)"
  XEN_ASSERT_TYPE(XEN_GtkWindow__P(window), window, 1, "gtk_window_get_decorated", "GtkWindow*");
  return(C_TO_XEN_gboolean(gtk_window_get_decorated(XEN_TO_C_GtkWindow_(window))));
}

static XEN gxg_gtk_window_set_icon_list(XEN window, XEN list)
{
  #define H_gtk_window_set_icon_list "void gtk_window_set_icon_list(GtkWindow* window, GList* list)"
  XEN_ASSERT_TYPE(XEN_GtkWindow__P(window), window, 1, "gtk_window_set_icon_list", "GtkWindow*");
  XEN_ASSERT_TYPE(XEN_GList__P(list), list, 2, "gtk_window_set_icon_list", "GList*");
  gtk_window_set_icon_list(XEN_TO_C_GtkWindow_(window), XEN_TO_C_GList_(list));
  return(XEN_FALSE);
}

static XEN gxg_gtk_window_get_icon_list(XEN window)
{
  #define H_gtk_window_get_icon_list "GList* gtk_window_get_icon_list(GtkWindow* window)"
  XEN_ASSERT_TYPE(XEN_GtkWindow__P(window), window, 1, "gtk_window_get_icon_list", "GtkWindow*");
  return(C_TO_XEN_GList_(gtk_window_get_icon_list(XEN_TO_C_GtkWindow_(window))));
}

static XEN gxg_gtk_window_set_icon(XEN window, XEN icon)
{
  #define H_gtk_window_set_icon "void gtk_window_set_icon(GtkWindow* window, GdkPixbuf* icon)"
  XEN_ASSERT_TYPE(XEN_GtkWindow__P(window), window, 1, "gtk_window_set_icon", "GtkWindow*");
  XEN_ASSERT_TYPE(XEN_GdkPixbuf__P(icon) || XEN_FALSE_P(icon), icon, 2, "gtk_window_set_icon", "GdkPixbuf*");
  gtk_window_set_icon(XEN_TO_C_GtkWindow_(window), XEN_TO_C_GdkPixbuf_(icon));
  return(XEN_FALSE);
}

static XEN gxg_gtk_window_get_icon(XEN window)
{
  #define H_gtk_window_get_icon "GdkPixbuf* gtk_window_get_icon(GtkWindow* window)"
  XEN_ASSERT_TYPE(XEN_GtkWindow__P(window), window, 1, "gtk_window_get_icon", "GtkWindow*");
  return(C_TO_XEN_GdkPixbuf_(gtk_window_get_icon(XEN_TO_C_GtkWindow_(window))));
}

static XEN gxg_gtk_window_set_default_icon_list(XEN list)
{
  #define H_gtk_window_set_default_icon_list "void gtk_window_set_default_icon_list(GList* list)"
  XEN_ASSERT_TYPE(XEN_GList__P(list), list, 1, "gtk_window_set_default_icon_list", "GList*");
  gtk_window_set_default_icon_list(XEN_TO_C_GList_(list));
  return(XEN_FALSE);
}

static XEN gxg_gtk_window_get_default_icon_list(void)
{
  #define H_gtk_window_get_default_icon_list "GList* gtk_window_get_default_icon_list( void)"
  return(C_TO_XEN_GList_(gtk_window_get_default_icon_list()));
}
#endif

static XEN gxg_gtk_window_set_modal(XEN window, XEN modal)
{
  #define H_gtk_window_set_modal "void gtk_window_set_modal(GtkWindow* window, gboolean modal)"
  XEN_ASSERT_TYPE(XEN_GtkWindow__P(window), window, 1, "gtk_window_set_modal", "GtkWindow*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(modal), modal, 2, "gtk_window_set_modal", "gboolean");
  gtk_window_set_modal(XEN_TO_C_GtkWindow_(window), XEN_TO_C_gboolean(modal));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_window_get_modal(XEN window)
{
  #define H_gtk_window_get_modal "gboolean gtk_window_get_modal(GtkWindow* window)"
  XEN_ASSERT_TYPE(XEN_GtkWindow__P(window), window, 1, "gtk_window_get_modal", "GtkWindow*");
  return(C_TO_XEN_gboolean(gtk_window_get_modal(XEN_TO_C_GtkWindow_(window))));
}

static XEN gxg_gtk_window_list_toplevels(void)
{
  #define H_gtk_window_list_toplevels "GList* gtk_window_list_toplevels( void)"
  return(C_TO_XEN_GList_(gtk_window_list_toplevels()));
}

static XEN gxg_gtk_window_add_mnemonic(XEN window, XEN keyval, XEN target)
{
  #define H_gtk_window_add_mnemonic "void gtk_window_add_mnemonic(GtkWindow* window, guint keyval, GtkWidget* target)"
  XEN_ASSERT_TYPE(XEN_GtkWindow__P(window), window, 1, "gtk_window_add_mnemonic", "GtkWindow*");
  XEN_ASSERT_TYPE(XEN_guint_P(keyval), keyval, 2, "gtk_window_add_mnemonic", "guint");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(target), target, 3, "gtk_window_add_mnemonic", "GtkWidget*");
  gtk_window_add_mnemonic(XEN_TO_C_GtkWindow_(window), XEN_TO_C_guint(keyval), XEN_TO_C_GtkWidget_(target));
  return(XEN_FALSE);
}

static XEN gxg_gtk_window_remove_mnemonic(XEN window, XEN keyval, XEN target)
{
  #define H_gtk_window_remove_mnemonic "void gtk_window_remove_mnemonic(GtkWindow* window, guint keyval, \
GtkWidget* target)"
  XEN_ASSERT_TYPE(XEN_GtkWindow__P(window), window, 1, "gtk_window_remove_mnemonic", "GtkWindow*");
  XEN_ASSERT_TYPE(XEN_guint_P(keyval), keyval, 2, "gtk_window_remove_mnemonic", "guint");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(target), target, 3, "gtk_window_remove_mnemonic", "GtkWidget*");
  gtk_window_remove_mnemonic(XEN_TO_C_GtkWindow_(window), XEN_TO_C_guint(keyval), XEN_TO_C_GtkWidget_(target));
  return(XEN_FALSE);
}

static XEN gxg_gtk_window_mnemonic_activate(XEN window, XEN keyval, XEN modifier)
{
  #define H_gtk_window_mnemonic_activate "gboolean gtk_window_mnemonic_activate(GtkWindow* window, guint keyval, \
GdkModifierType modifier)"
  XEN_ASSERT_TYPE(XEN_GtkWindow__P(window), window, 1, "gtk_window_mnemonic_activate", "GtkWindow*");
  XEN_ASSERT_TYPE(XEN_guint_P(keyval), keyval, 2, "gtk_window_mnemonic_activate", "guint");
  XEN_ASSERT_TYPE(XEN_GdkModifierType_P(modifier), modifier, 3, "gtk_window_mnemonic_activate", "GdkModifierType");
  return(C_TO_XEN_gboolean(gtk_window_mnemonic_activate(XEN_TO_C_GtkWindow_(window), XEN_TO_C_guint(keyval), XEN_TO_C_GdkModifierType(modifier))));
}

static XEN gxg_gtk_window_set_mnemonic_modifier(XEN window, XEN modifier)
{
  #define H_gtk_window_set_mnemonic_modifier "void gtk_window_set_mnemonic_modifier(GtkWindow* window, \
GdkModifierType modifier)"
  XEN_ASSERT_TYPE(XEN_GtkWindow__P(window), window, 1, "gtk_window_set_mnemonic_modifier", "GtkWindow*");
  XEN_ASSERT_TYPE(XEN_GdkModifierType_P(modifier), modifier, 2, "gtk_window_set_mnemonic_modifier", "GdkModifierType");
  gtk_window_set_mnemonic_modifier(XEN_TO_C_GtkWindow_(window), XEN_TO_C_GdkModifierType(modifier));
  return(XEN_FALSE);
}

static XEN gxg_gtk_window_get_mnemonic_modifier(XEN window)
{
  #define H_gtk_window_get_mnemonic_modifier "GdkModifierType gtk_window_get_mnemonic_modifier(GtkWindow* window)"
  XEN_ASSERT_TYPE(XEN_GtkWindow__P(window), window, 1, "gtk_window_get_mnemonic_modifier", "GtkWindow*");
  return(C_TO_XEN_GdkModifierType(gtk_window_get_mnemonic_modifier(XEN_TO_C_GtkWindow_(window))));
}

static XEN gxg_gtk_window_present(XEN window)
{
  #define H_gtk_window_present "void gtk_window_present(GtkWindow* window)"
  XEN_ASSERT_TYPE(XEN_GtkWindow__P(window), window, 1, "gtk_window_present", "GtkWindow*");
  gtk_window_present(XEN_TO_C_GtkWindow_(window));
  return(XEN_FALSE);
}

static XEN gxg_gtk_window_iconify(XEN window)
{
  #define H_gtk_window_iconify "void gtk_window_iconify(GtkWindow* window)"
  XEN_ASSERT_TYPE(XEN_GtkWindow__P(window), window, 1, "gtk_window_iconify", "GtkWindow*");
  gtk_window_iconify(XEN_TO_C_GtkWindow_(window));
  return(XEN_FALSE);
}

static XEN gxg_gtk_window_deiconify(XEN window)
{
  #define H_gtk_window_deiconify "void gtk_window_deiconify(GtkWindow* window)"
  XEN_ASSERT_TYPE(XEN_GtkWindow__P(window), window, 1, "gtk_window_deiconify", "GtkWindow*");
  gtk_window_deiconify(XEN_TO_C_GtkWindow_(window));
  return(XEN_FALSE);
}

static XEN gxg_gtk_window_stick(XEN window)
{
  #define H_gtk_window_stick "void gtk_window_stick(GtkWindow* window)"
  XEN_ASSERT_TYPE(XEN_GtkWindow__P(window), window, 1, "gtk_window_stick", "GtkWindow*");
  gtk_window_stick(XEN_TO_C_GtkWindow_(window));
  return(XEN_FALSE);
}

static XEN gxg_gtk_window_unstick(XEN window)
{
  #define H_gtk_window_unstick "void gtk_window_unstick(GtkWindow* window)"
  XEN_ASSERT_TYPE(XEN_GtkWindow__P(window), window, 1, "gtk_window_unstick", "GtkWindow*");
  gtk_window_unstick(XEN_TO_C_GtkWindow_(window));
  return(XEN_FALSE);
}

static XEN gxg_gtk_window_maximize(XEN window)
{
  #define H_gtk_window_maximize "void gtk_window_maximize(GtkWindow* window)"
  XEN_ASSERT_TYPE(XEN_GtkWindow__P(window), window, 1, "gtk_window_maximize", "GtkWindow*");
  gtk_window_maximize(XEN_TO_C_GtkWindow_(window));
  return(XEN_FALSE);
}

static XEN gxg_gtk_window_unmaximize(XEN window)
{
  #define H_gtk_window_unmaximize "void gtk_window_unmaximize(GtkWindow* window)"
  XEN_ASSERT_TYPE(XEN_GtkWindow__P(window), window, 1, "gtk_window_unmaximize", "GtkWindow*");
  gtk_window_unmaximize(XEN_TO_C_GtkWindow_(window));
  return(XEN_FALSE);
}

static XEN gxg_gtk_window_begin_resize_drag(XEN window, XEN edge, XEN button, XEN root_x, XEN root_y, XEN timestamp)
{
  #define H_gtk_window_begin_resize_drag "void gtk_window_begin_resize_drag(GtkWindow* window, GdkWindowEdge edge, \
gint button, gint root_x, gint root_y, guint32 timestamp)"
  XEN_ASSERT_TYPE(XEN_GtkWindow__P(window), window, 1, "gtk_window_begin_resize_drag", "GtkWindow*");
  XEN_ASSERT_TYPE(XEN_GdkWindowEdge_P(edge), edge, 2, "gtk_window_begin_resize_drag", "GdkWindowEdge");
  XEN_ASSERT_TYPE(XEN_gint_P(button), button, 3, "gtk_window_begin_resize_drag", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(root_x), root_x, 4, "gtk_window_begin_resize_drag", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(root_y), root_y, 5, "gtk_window_begin_resize_drag", "gint");
  XEN_ASSERT_TYPE(XEN_guint32_P(timestamp), timestamp, 6, "gtk_window_begin_resize_drag", "guint32");
  gtk_window_begin_resize_drag(XEN_TO_C_GtkWindow_(window), XEN_TO_C_GdkWindowEdge(edge), XEN_TO_C_gint(button), XEN_TO_C_gint(root_x), 
                               XEN_TO_C_gint(root_y), XEN_TO_C_guint32(timestamp));
  return(XEN_FALSE);
}

static XEN gxg_gtk_window_begin_move_drag(XEN window, XEN button, XEN root_x, XEN root_y, XEN timestamp)
{
  #define H_gtk_window_begin_move_drag "void gtk_window_begin_move_drag(GtkWindow* window, gint button, \
gint root_x, gint root_y, guint32 timestamp)"
  XEN_ASSERT_TYPE(XEN_GtkWindow__P(window), window, 1, "gtk_window_begin_move_drag", "GtkWindow*");
  XEN_ASSERT_TYPE(XEN_gint_P(button), button, 2, "gtk_window_begin_move_drag", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(root_x), root_x, 3, "gtk_window_begin_move_drag", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(root_y), root_y, 4, "gtk_window_begin_move_drag", "gint");
  XEN_ASSERT_TYPE(XEN_guint32_P(timestamp), timestamp, 5, "gtk_window_begin_move_drag", "guint32");
  gtk_window_begin_move_drag(XEN_TO_C_GtkWindow_(window), XEN_TO_C_gint(button), XEN_TO_C_gint(root_x), XEN_TO_C_gint(root_y), 
                             XEN_TO_C_guint32(timestamp));
  return(XEN_FALSE);
}
#endif

static XEN gxg_gtk_window_set_default_size(XEN window, XEN width, XEN height)
{
  #define H_gtk_window_set_default_size "void gtk_window_set_default_size(GtkWindow* window, gint width, \
gint height)"
  XEN_ASSERT_TYPE(XEN_GtkWindow__P(window), window, 1, "gtk_window_set_default_size", "GtkWindow*");
  XEN_ASSERT_TYPE(XEN_gint_P(width), width, 2, "gtk_window_set_default_size", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(height), height, 3, "gtk_window_set_default_size", "gint");
  gtk_window_set_default_size(XEN_TO_C_GtkWindow_(window), XEN_TO_C_gint(width), XEN_TO_C_gint(height));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_window_get_default_size(XEN window, XEN width, XEN height)
{
  #define H_gtk_window_get_default_size "void gtk_window_get_default_size(GtkWindow* window, gint* [width], \
gint* [height])"
  gint ref_width;
  gint ref_height;
  XEN_ASSERT_TYPE(XEN_GtkWindow__P(window), window, 1, "gtk_window_get_default_size", "GtkWindow*");
  gtk_window_get_default_size(XEN_TO_C_GtkWindow_(window), &ref_width, &ref_height);
  return(XEN_LIST_2(C_TO_XEN_gint(ref_width), C_TO_XEN_gint(ref_height)));
}

static XEN gxg_gtk_window_resize(XEN window, XEN width, XEN height)
{
  #define H_gtk_window_resize "void gtk_window_resize(GtkWindow* window, gint width, gint height)"
  XEN_ASSERT_TYPE(XEN_GtkWindow__P(window), window, 1, "gtk_window_resize", "GtkWindow*");
  XEN_ASSERT_TYPE(XEN_gint_P(width), width, 2, "gtk_window_resize", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(height), height, 3, "gtk_window_resize", "gint");
  gtk_window_resize(XEN_TO_C_GtkWindow_(window), XEN_TO_C_gint(width), XEN_TO_C_gint(height));
  return(XEN_FALSE);
}

static XEN gxg_gtk_window_get_size(XEN window, XEN width, XEN height)
{
  #define H_gtk_window_get_size "void gtk_window_get_size(GtkWindow* window, gint* width, gint* height)"
  XEN_ASSERT_TYPE(XEN_GtkWindow__P(window), window, 1, "gtk_window_get_size", "GtkWindow*");
  XEN_ASSERT_TYPE(XEN_gint__P(width), width, 2, "gtk_window_get_size", "gint*");
  XEN_ASSERT_TYPE(XEN_gint__P(height), height, 3, "gtk_window_get_size", "gint*");
  gtk_window_get_size(XEN_TO_C_GtkWindow_(window), XEN_TO_C_gint_(width), XEN_TO_C_gint_(height));
  return(XEN_FALSE);
}

static XEN gxg_gtk_window_move(XEN window, XEN x, XEN y)
{
  #define H_gtk_window_move "void gtk_window_move(GtkWindow* window, gint x, gint y)"
  XEN_ASSERT_TYPE(XEN_GtkWindow__P(window), window, 1, "gtk_window_move", "GtkWindow*");
  XEN_ASSERT_TYPE(XEN_gint_P(x), x, 2, "gtk_window_move", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(y), y, 3, "gtk_window_move", "gint");
  gtk_window_move(XEN_TO_C_GtkWindow_(window), XEN_TO_C_gint(x), XEN_TO_C_gint(y));
  return(XEN_FALSE);
}

static XEN gxg_gtk_window_get_position(XEN window, XEN root_x, XEN root_y)
{
  #define H_gtk_window_get_position "void gtk_window_get_position(GtkWindow* window, gint* [root_x], \
gint* [root_y])"
  gint ref_root_x;
  gint ref_root_y;
  XEN_ASSERT_TYPE(XEN_GtkWindow__P(window), window, 1, "gtk_window_get_position", "GtkWindow*");
  gtk_window_get_position(XEN_TO_C_GtkWindow_(window), &ref_root_x, &ref_root_y);
  return(XEN_LIST_2(C_TO_XEN_gint(ref_root_x), C_TO_XEN_gint(ref_root_y)));
}

static XEN gxg_gtk_window_parse_geometry(XEN window, XEN geometry)
{
  #define H_gtk_window_parse_geometry "gboolean gtk_window_parse_geometry(GtkWindow* window, gchar* geometry)"
  XEN_ASSERT_TYPE(XEN_GtkWindow__P(window), window, 1, "gtk_window_parse_geometry", "GtkWindow*");
  XEN_ASSERT_TYPE(XEN_gchar__P(geometry), geometry, 2, "gtk_window_parse_geometry", "gchar*");
  return(C_TO_XEN_gboolean(gtk_window_parse_geometry(XEN_TO_C_GtkWindow_(window), XEN_TO_C_gchar_(geometry))));
}

static XEN gxg_gtk_window_reshow_with_initial_size(XEN window)
{
  #define H_gtk_window_reshow_with_initial_size "void gtk_window_reshow_with_initial_size(GtkWindow* window)"
  XEN_ASSERT_TYPE(XEN_GtkWindow__P(window), window, 1, "gtk_window_reshow_with_initial_size", "GtkWindow*");
  gtk_window_reshow_with_initial_size(XEN_TO_C_GtkWindow_(window));
  return(XEN_FALSE);
}

static XEN gxg_gtk_window_group_get_type(void)
{
  #define H_gtk_window_group_get_type "GType gtk_window_group_get_type( void)"
  return(C_TO_XEN_GType(gtk_window_group_get_type()));
}

static XEN gxg_gtk_window_group_new(void)
{
  #define H_gtk_window_group_new "GtkWindowGroup* gtk_window_group_new( void)"
  return(C_TO_XEN_GtkWindowGroup_(gtk_window_group_new()));
}

static XEN gxg_pango_color_get_type(void)
{
  #define H_pango_color_get_type "GType pango_color_get_type( void)"
  return(C_TO_XEN_GType(pango_color_get_type()));
}

static XEN gxg_pango_color_copy(XEN src)
{
  #define H_pango_color_copy "PangoColor* pango_color_copy(PangoColor* src)"
  XEN_ASSERT_TYPE(XEN_PangoColor__P(src), src, 1, "pango_color_copy", "PangoColor*");
  return(C_TO_XEN_PangoColor_(pango_color_copy(XEN_TO_C_PangoColor_(src))));
}

static XEN gxg_pango_color_free(XEN color)
{
  #define H_pango_color_free "void pango_color_free(PangoColor* color)"
  XEN_ASSERT_TYPE(XEN_PangoColor__P(color), color, 1, "pango_color_free", "PangoColor*");
  pango_color_free(XEN_TO_C_PangoColor_(color));
  return(XEN_FALSE);
}

static XEN gxg_pango_color_parse(XEN color, XEN spec)
{
  #define H_pango_color_parse "gboolean pango_color_parse(PangoColor* color, char* spec)"
  XEN_ASSERT_TYPE(XEN_PangoColor__P(color), color, 1, "pango_color_parse", "PangoColor*");
  XEN_ASSERT_TYPE(XEN_char__P(spec), spec, 2, "pango_color_parse", "char*");
  return(C_TO_XEN_gboolean(pango_color_parse(XEN_TO_C_PangoColor_(color), XEN_TO_C_char_(spec))));
}

static XEN gxg_pango_attr_type_register(XEN name)
{
  #define H_pango_attr_type_register "PangoAttrType pango_attr_type_register(gchar* name)"
  XEN_ASSERT_TYPE(XEN_gchar__P(name), name, 1, "pango_attr_type_register", "gchar*");
  return(C_TO_XEN_PangoAttrType(pango_attr_type_register(XEN_TO_C_gchar_(name))));
}

static XEN gxg_pango_attribute_copy(XEN attr)
{
  #define H_pango_attribute_copy "PangoAttribute* pango_attribute_copy(PangoAttribute* attr)"
  XEN_ASSERT_TYPE(XEN_PangoAttribute__P(attr), attr, 1, "pango_attribute_copy", "PangoAttribute*");
  return(C_TO_XEN_PangoAttribute_(pango_attribute_copy(XEN_TO_C_PangoAttribute_(attr))));
}

static XEN gxg_pango_attribute_destroy(XEN attr)
{
  #define H_pango_attribute_destroy "void pango_attribute_destroy(PangoAttribute* attr)"
  XEN_ASSERT_TYPE(XEN_PangoAttribute__P(attr), attr, 1, "pango_attribute_destroy", "PangoAttribute*");
  pango_attribute_destroy(XEN_TO_C_PangoAttribute_(attr));
  return(XEN_FALSE);
}

static XEN gxg_pango_attribute_equal(XEN attr1, XEN attr2)
{
  #define H_pango_attribute_equal "gboolean pango_attribute_equal(PangoAttribute* attr1, PangoAttribute* attr2)"
  XEN_ASSERT_TYPE(XEN_PangoAttribute__P(attr1), attr1, 1, "pango_attribute_equal", "PangoAttribute*");
  XEN_ASSERT_TYPE(XEN_PangoAttribute__P(attr2), attr2, 2, "pango_attribute_equal", "PangoAttribute*");
  return(C_TO_XEN_gboolean(pango_attribute_equal(XEN_TO_C_PangoAttribute_(attr1), XEN_TO_C_PangoAttribute_(attr2))));
}

static XEN gxg_pango_attr_language_new(XEN language)
{
  #define H_pango_attr_language_new "PangoAttribute* pango_attr_language_new(PangoLanguage* language)"
  XEN_ASSERT_TYPE(XEN_PangoLanguage__P(language), language, 1, "pango_attr_language_new", "PangoLanguage*");
  return(C_TO_XEN_PangoAttribute_(pango_attr_language_new(XEN_TO_C_PangoLanguage_(language))));
}

static XEN gxg_pango_attr_family_new(XEN family)
{
  #define H_pango_attr_family_new "PangoAttribute* pango_attr_family_new(char* family)"
  XEN_ASSERT_TYPE(XEN_char__P(family), family, 1, "pango_attr_family_new", "char*");
  return(C_TO_XEN_PangoAttribute_(pango_attr_family_new(XEN_TO_C_char_(family))));
}

static XEN gxg_pango_attr_foreground_new(XEN red, XEN green, XEN blue)
{
  #define H_pango_attr_foreground_new "PangoAttribute* pango_attr_foreground_new(guint16 red, guint16 green, \
guint16 blue)"
  XEN_ASSERT_TYPE(XEN_guint16_P(red), red, 1, "pango_attr_foreground_new", "guint16");
  XEN_ASSERT_TYPE(XEN_guint16_P(green), green, 2, "pango_attr_foreground_new", "guint16");
  XEN_ASSERT_TYPE(XEN_guint16_P(blue), blue, 3, "pango_attr_foreground_new", "guint16");
  return(C_TO_XEN_PangoAttribute_(pango_attr_foreground_new(XEN_TO_C_guint16(red), XEN_TO_C_guint16(green), XEN_TO_C_guint16(blue))));
}

static XEN gxg_pango_attr_background_new(XEN red, XEN green, XEN blue)
{
  #define H_pango_attr_background_new "PangoAttribute* pango_attr_background_new(guint16 red, guint16 green, \
guint16 blue)"
  XEN_ASSERT_TYPE(XEN_guint16_P(red), red, 1, "pango_attr_background_new", "guint16");
  XEN_ASSERT_TYPE(XEN_guint16_P(green), green, 2, "pango_attr_background_new", "guint16");
  XEN_ASSERT_TYPE(XEN_guint16_P(blue), blue, 3, "pango_attr_background_new", "guint16");
  return(C_TO_XEN_PangoAttribute_(pango_attr_background_new(XEN_TO_C_guint16(red), XEN_TO_C_guint16(green), XEN_TO_C_guint16(blue))));
}

static XEN gxg_pango_attr_size_new(XEN size)
{
  #define H_pango_attr_size_new "PangoAttribute* pango_attr_size_new(int size)"
  XEN_ASSERT_TYPE(XEN_int_P(size), size, 1, "pango_attr_size_new", "int");
  return(C_TO_XEN_PangoAttribute_(pango_attr_size_new(XEN_TO_C_int(size))));
}

static XEN gxg_pango_attr_style_new(XEN style)
{
  #define H_pango_attr_style_new "PangoAttribute* pango_attr_style_new(PangoStyle style)"
  XEN_ASSERT_TYPE(XEN_PangoStyle_P(style), style, 1, "pango_attr_style_new", "PangoStyle");
  return(C_TO_XEN_PangoAttribute_(pango_attr_style_new(XEN_TO_C_PangoStyle(style))));
}

static XEN gxg_pango_attr_weight_new(XEN weight)
{
  #define H_pango_attr_weight_new "PangoAttribute* pango_attr_weight_new(PangoWeight weight)"
  XEN_ASSERT_TYPE(XEN_PangoWeight_P(weight), weight, 1, "pango_attr_weight_new", "PangoWeight");
  return(C_TO_XEN_PangoAttribute_(pango_attr_weight_new(XEN_TO_C_PangoWeight(weight))));
}

static XEN gxg_pango_attr_variant_new(XEN variant)
{
  #define H_pango_attr_variant_new "PangoAttribute* pango_attr_variant_new(PangoVariant variant)"
  XEN_ASSERT_TYPE(XEN_PangoVariant_P(variant), variant, 1, "pango_attr_variant_new", "PangoVariant");
  return(C_TO_XEN_PangoAttribute_(pango_attr_variant_new(XEN_TO_C_PangoVariant(variant))));
}

static XEN gxg_pango_attr_stretch_new(XEN stretch)
{
  #define H_pango_attr_stretch_new "PangoAttribute* pango_attr_stretch_new(PangoStretch stretch)"
  XEN_ASSERT_TYPE(XEN_PangoStretch_P(stretch), stretch, 1, "pango_attr_stretch_new", "PangoStretch");
  return(C_TO_XEN_PangoAttribute_(pango_attr_stretch_new(XEN_TO_C_PangoStretch(stretch))));
}

static XEN gxg_pango_attr_font_desc_new(XEN desc)
{
  #define H_pango_attr_font_desc_new "PangoAttribute* pango_attr_font_desc_new(PangoFontDescription* desc)"
  XEN_ASSERT_TYPE(XEN_PangoFontDescription__P(desc), desc, 1, "pango_attr_font_desc_new", "PangoFontDescription*");
  return(C_TO_XEN_PangoAttribute_(pango_attr_font_desc_new(XEN_TO_C_PangoFontDescription_(desc))));
}

static XEN gxg_pango_attr_underline_new(XEN underline)
{
  #define H_pango_attr_underline_new "PangoAttribute* pango_attr_underline_new(PangoUnderline underline)"
  XEN_ASSERT_TYPE(XEN_PangoUnderline_P(underline), underline, 1, "pango_attr_underline_new", "PangoUnderline");
  return(C_TO_XEN_PangoAttribute_(pango_attr_underline_new(XEN_TO_C_PangoUnderline(underline))));
}

static XEN gxg_pango_attr_strikethrough_new(XEN strikethrough)
{
  #define H_pango_attr_strikethrough_new "PangoAttribute* pango_attr_strikethrough_new(gboolean strikethrough)"
  XEN_ASSERT_TYPE(XEN_gboolean_P(strikethrough), strikethrough, 1, "pango_attr_strikethrough_new", "gboolean");
  return(C_TO_XEN_PangoAttribute_(pango_attr_strikethrough_new(XEN_TO_C_gboolean(strikethrough))));
}

static XEN gxg_pango_attr_rise_new(XEN rise)
{
  #define H_pango_attr_rise_new "PangoAttribute* pango_attr_rise_new(int rise)"
  XEN_ASSERT_TYPE(XEN_int_P(rise), rise, 1, "pango_attr_rise_new", "int");
  return(C_TO_XEN_PangoAttribute_(pango_attr_rise_new(XEN_TO_C_int(rise))));
}

static XEN gxg_pango_attr_shape_new(XEN ink_rect, XEN logical_rect)
{
  #define H_pango_attr_shape_new "PangoAttribute* pango_attr_shape_new(PangoRectangle* ink_rect, PangoRectangle* logical_rect)"
  XEN_ASSERT_TYPE(XEN_PangoRectangle__P(ink_rect), ink_rect, 1, "pango_attr_shape_new", "PangoRectangle*");
  XEN_ASSERT_TYPE(XEN_PangoRectangle__P(logical_rect), logical_rect, 2, "pango_attr_shape_new", "PangoRectangle*");
  return(C_TO_XEN_PangoAttribute_(pango_attr_shape_new(XEN_TO_C_PangoRectangle_(ink_rect), XEN_TO_C_PangoRectangle_(logical_rect))));
}

static XEN gxg_pango_attr_scale_new(XEN scale_factor)
{
  #define H_pango_attr_scale_new "PangoAttribute* pango_attr_scale_new(double scale_factor)"
  XEN_ASSERT_TYPE(XEN_double_P(scale_factor), scale_factor, 1, "pango_attr_scale_new", "double");
  return(C_TO_XEN_PangoAttribute_(pango_attr_scale_new(XEN_TO_C_double(scale_factor))));
}

static XEN gxg_pango_attr_list_get_type(void)
{
  #define H_pango_attr_list_get_type "GType pango_attr_list_get_type( void)"
  return(C_TO_XEN_GType(pango_attr_list_get_type()));
}

static XEN gxg_pango_attr_list_new(void)
{
  #define H_pango_attr_list_new "PangoAttrList* pango_attr_list_new( void)"
  return(C_TO_XEN_PangoAttrList_(pango_attr_list_new()));
}

static XEN gxg_pango_attr_list_ref(XEN list)
{
  #define H_pango_attr_list_ref "void pango_attr_list_ref(PangoAttrList* list)"
  XEN_ASSERT_TYPE(XEN_PangoAttrList__P(list), list, 1, "pango_attr_list_ref", "PangoAttrList*");
  pango_attr_list_ref(XEN_TO_C_PangoAttrList_(list));
  return(XEN_FALSE);
}

static XEN gxg_pango_attr_list_unref(XEN list)
{
  #define H_pango_attr_list_unref "void pango_attr_list_unref(PangoAttrList* list)"
  XEN_ASSERT_TYPE(XEN_PangoAttrList__P(list), list, 1, "pango_attr_list_unref", "PangoAttrList*");
  pango_attr_list_unref(XEN_TO_C_PangoAttrList_(list));
  return(XEN_FALSE);
}

static XEN gxg_pango_attr_list_copy(XEN list)
{
  #define H_pango_attr_list_copy "PangoAttrList* pango_attr_list_copy(PangoAttrList* list)"
  XEN_ASSERT_TYPE(XEN_PangoAttrList__P(list), list, 1, "pango_attr_list_copy", "PangoAttrList*");
  return(C_TO_XEN_PangoAttrList_(pango_attr_list_copy(XEN_TO_C_PangoAttrList_(list))));
}

static XEN gxg_pango_attr_list_insert(XEN list, XEN attr)
{
  #define H_pango_attr_list_insert "void pango_attr_list_insert(PangoAttrList* list, PangoAttribute* attr)"
  XEN_ASSERT_TYPE(XEN_PangoAttrList__P(list), list, 1, "pango_attr_list_insert", "PangoAttrList*");
  XEN_ASSERT_TYPE(XEN_PangoAttribute__P(attr), attr, 2, "pango_attr_list_insert", "PangoAttribute*");
  pango_attr_list_insert(XEN_TO_C_PangoAttrList_(list), XEN_TO_C_PangoAttribute_(attr));
  return(XEN_FALSE);
}

static XEN gxg_pango_attr_list_insert_before(XEN list, XEN attr)
{
  #define H_pango_attr_list_insert_before "void pango_attr_list_insert_before(PangoAttrList* list, PangoAttribute* attr)"
  XEN_ASSERT_TYPE(XEN_PangoAttrList__P(list), list, 1, "pango_attr_list_insert_before", "PangoAttrList*");
  XEN_ASSERT_TYPE(XEN_PangoAttribute__P(attr), attr, 2, "pango_attr_list_insert_before", "PangoAttribute*");
  pango_attr_list_insert_before(XEN_TO_C_PangoAttrList_(list), XEN_TO_C_PangoAttribute_(attr));
  return(XEN_FALSE);
}

static XEN gxg_pango_attr_list_change(XEN list, XEN attr)
{
  #define H_pango_attr_list_change "void pango_attr_list_change(PangoAttrList* list, PangoAttribute* attr)"
  XEN_ASSERT_TYPE(XEN_PangoAttrList__P(list), list, 1, "pango_attr_list_change", "PangoAttrList*");
  XEN_ASSERT_TYPE(XEN_PangoAttribute__P(attr), attr, 2, "pango_attr_list_change", "PangoAttribute*");
  pango_attr_list_change(XEN_TO_C_PangoAttrList_(list), XEN_TO_C_PangoAttribute_(attr));
  return(XEN_FALSE);
}

static XEN gxg_pango_attr_list_splice(XEN list, XEN other, XEN pos, XEN len)
{
  #define H_pango_attr_list_splice "void pango_attr_list_splice(PangoAttrList* list, PangoAttrList* other, \
gint pos, gint len)"
  XEN_ASSERT_TYPE(XEN_PangoAttrList__P(list), list, 1, "pango_attr_list_splice", "PangoAttrList*");
  XEN_ASSERT_TYPE(XEN_PangoAttrList__P(other), other, 2, "pango_attr_list_splice", "PangoAttrList*");
  XEN_ASSERT_TYPE(XEN_gint_P(pos), pos, 3, "pango_attr_list_splice", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(len), len, 4, "pango_attr_list_splice", "gint");
  pango_attr_list_splice(XEN_TO_C_PangoAttrList_(list), XEN_TO_C_PangoAttrList_(other), XEN_TO_C_gint(pos), XEN_TO_C_gint(len));
  return(XEN_FALSE);
}

static XEN gxg_pango_attr_list_get_iterator(XEN list)
{
  #define H_pango_attr_list_get_iterator "PangoAttrIterator* pango_attr_list_get_iterator(PangoAttrList* list)"
  XEN_ASSERT_TYPE(XEN_PangoAttrList__P(list), list, 1, "pango_attr_list_get_iterator", "PangoAttrList*");
  return(C_TO_XEN_PangoAttrIterator_(pango_attr_list_get_iterator(XEN_TO_C_PangoAttrList_(list))));
}

static XEN gxg_pango_attr_iterator_range(XEN iterator, XEN start, XEN end)
{
  #define H_pango_attr_iterator_range "void pango_attr_iterator_range(PangoAttrIterator* iterator, gint* start, \
gint* end)"
  XEN_ASSERT_TYPE(XEN_PangoAttrIterator__P(iterator), iterator, 1, "pango_attr_iterator_range", "PangoAttrIterator*");
  XEN_ASSERT_TYPE(XEN_gint__P(start), start, 2, "pango_attr_iterator_range", "gint*");
  XEN_ASSERT_TYPE(XEN_gint__P(end), end, 3, "pango_attr_iterator_range", "gint*");
  pango_attr_iterator_range(XEN_TO_C_PangoAttrIterator_(iterator), XEN_TO_C_gint_(start), XEN_TO_C_gint_(end));
  return(XEN_FALSE);
}

static XEN gxg_pango_attr_iterator_next(XEN iterator)
{
  #define H_pango_attr_iterator_next "gboolean pango_attr_iterator_next(PangoAttrIterator* iterator)"
  XEN_ASSERT_TYPE(XEN_PangoAttrIterator__P(iterator), iterator, 1, "pango_attr_iterator_next", "PangoAttrIterator*");
  return(C_TO_XEN_gboolean(pango_attr_iterator_next(XEN_TO_C_PangoAttrIterator_(iterator))));
}

static XEN gxg_pango_attr_iterator_copy(XEN iterator)
{
  #define H_pango_attr_iterator_copy "PangoAttrIterator* pango_attr_iterator_copy(PangoAttrIterator* iterator)"
  XEN_ASSERT_TYPE(XEN_PangoAttrIterator__P(iterator), iterator, 1, "pango_attr_iterator_copy", "PangoAttrIterator*");
  return(C_TO_XEN_PangoAttrIterator_(pango_attr_iterator_copy(XEN_TO_C_PangoAttrIterator_(iterator))));
}

static XEN gxg_pango_attr_iterator_destroy(XEN iterator)
{
  #define H_pango_attr_iterator_destroy "void pango_attr_iterator_destroy(PangoAttrIterator* iterator)"
  XEN_ASSERT_TYPE(XEN_PangoAttrIterator__P(iterator), iterator, 1, "pango_attr_iterator_destroy", "PangoAttrIterator*");
  pango_attr_iterator_destroy(XEN_TO_C_PangoAttrIterator_(iterator));
  return(XEN_FALSE);
}

static XEN gxg_pango_attr_iterator_get(XEN iterator, XEN type)
{
  #define H_pango_attr_iterator_get "PangoAttribute* pango_attr_iterator_get(PangoAttrIterator* iterator, \
PangoAttrType type)"
  XEN_ASSERT_TYPE(XEN_PangoAttrIterator__P(iterator), iterator, 1, "pango_attr_iterator_get", "PangoAttrIterator*");
  XEN_ASSERT_TYPE(XEN_PangoAttrType_P(type), type, 2, "pango_attr_iterator_get", "PangoAttrType");
  return(C_TO_XEN_PangoAttribute_(pango_attr_iterator_get(XEN_TO_C_PangoAttrIterator_(iterator), XEN_TO_C_PangoAttrType(type))));
}

static XEN gxg_pango_attr_iterator_get_font(XEN iterator, XEN desc, XEN language, XEN extra_attrs)
{
  #define H_pango_attr_iterator_get_font "void pango_attr_iterator_get_font(PangoAttrIterator* iterator, \
PangoFontDescription* desc, PangoLanguage** [language], GSList** [extra_attrs])"
  PangoLanguage* ref_language;
  GSList* ref_extra_attrs;
  XEN_ASSERT_TYPE(XEN_PangoAttrIterator__P(iterator), iterator, 1, "pango_attr_iterator_get_font", "PangoAttrIterator*");
  XEN_ASSERT_TYPE(XEN_PangoFontDescription__P(desc), desc, 2, "pango_attr_iterator_get_font", "PangoFontDescription*");
  pango_attr_iterator_get_font(XEN_TO_C_PangoAttrIterator_(iterator), XEN_TO_C_PangoFontDescription_(desc), &ref_language, 
                               &ref_extra_attrs);
  return(XEN_LIST_2(C_TO_XEN_PangoLanguage_(ref_language), C_TO_XEN_GSList_(ref_extra_attrs)));
}

static XEN gxg_pango_parse_markup(XEN markup_text, XEN length, XEN accel_marker, XEN attr_list, XEN text, XEN accel_char, XEN error)
{
  #define H_pango_parse_markup "gboolean pango_parse_markup(char* markup_text, int length, gunichar accel_marker, \
PangoAttrList** attr_list, char** text, gunichar* accel_char, GError** [error])"
  GError* ref_error;
  XEN_ASSERT_TYPE(XEN_char__P(markup_text), markup_text, 1, "pango_parse_markup", "char*");
  XEN_ASSERT_TYPE(XEN_int_P(length), length, 2, "pango_parse_markup", "int");
  XEN_ASSERT_TYPE(XEN_gunichar_P(accel_marker), accel_marker, 3, "pango_parse_markup", "gunichar");
  XEN_ASSERT_TYPE(XEN_PangoAttrList___P(attr_list), attr_list, 4, "pango_parse_markup", "PangoAttrList**");
  XEN_ASSERT_TYPE(XEN_char___P(text), text, 5, "pango_parse_markup", "char**");
  XEN_ASSERT_TYPE(XEN_gunichar__P(accel_char), accel_char, 6, "pango_parse_markup", "gunichar*");
  {
    XEN result = XEN_FALSE;
    result = C_TO_XEN_gboolean(pango_parse_markup(XEN_TO_C_char_(markup_text), XEN_TO_C_int(length), XEN_TO_C_gunichar(accel_marker), 
                                                  XEN_TO_C_PangoAttrList__(attr_list), XEN_TO_C_char__(text), XEN_TO_C_gunichar_(accel_char), 
                                                  &ref_error));
    return(XEN_LIST_2(result, C_TO_XEN_GError_(ref_error)));
   }
}

static XEN gxg_pango_break(XEN text, XEN length, XEN analysis, XEN attrs, XEN attrs_len)
{
  #define H_pango_break "void pango_break(gchar* text, int length, PangoAnalysis* analysis, PangoLogAttr* attrs, \
int attrs_len)"
  XEN_ASSERT_TYPE(XEN_gchar__P(text), text, 1, "pango_break", "gchar*");
  XEN_ASSERT_TYPE(XEN_int_P(length), length, 2, "pango_break", "int");
  XEN_ASSERT_TYPE(XEN_PangoAnalysis__P(analysis), analysis, 3, "pango_break", "PangoAnalysis*");
  XEN_ASSERT_TYPE(XEN_PangoLogAttr__P(attrs), attrs, 4, "pango_break", "PangoLogAttr*");
  XEN_ASSERT_TYPE(XEN_int_P(attrs_len), attrs_len, 5, "pango_break", "int");
  pango_break(XEN_TO_C_gchar_(text), XEN_TO_C_int(length), XEN_TO_C_PangoAnalysis_(analysis), XEN_TO_C_PangoLogAttr_(attrs), 
              XEN_TO_C_int(attrs_len));
  return(XEN_FALSE);
}

static XEN gxg_pango_find_paragraph_boundary(XEN text, XEN length, XEN paragraph_delimiter_index, XEN next_paragraph_start)
{
  #define H_pango_find_paragraph_boundary "void pango_find_paragraph_boundary(gchar* text, gint length, \
gint* paragraph_delimiter_index, gint* next_paragraph_start)"
  XEN_ASSERT_TYPE(XEN_gchar__P(text), text, 1, "pango_find_paragraph_boundary", "gchar*");
  XEN_ASSERT_TYPE(XEN_gint_P(length), length, 2, "pango_find_paragraph_boundary", "gint");
  XEN_ASSERT_TYPE(XEN_gint__P(paragraph_delimiter_index), paragraph_delimiter_index, 3, "pango_find_paragraph_boundary", "gint*");
  XEN_ASSERT_TYPE(XEN_gint__P(next_paragraph_start), next_paragraph_start, 4, "pango_find_paragraph_boundary", "gint*");
  pango_find_paragraph_boundary(XEN_TO_C_gchar_(text), XEN_TO_C_gint(length), XEN_TO_C_gint_(paragraph_delimiter_index), 
                                XEN_TO_C_gint_(next_paragraph_start));
  return(XEN_FALSE);
}

static XEN gxg_pango_get_log_attrs(XEN text, XEN length, XEN level, XEN language, XEN log_attrs, XEN attrs_len)
{
  #define H_pango_get_log_attrs "void pango_get_log_attrs(char* text, int length, int level, PangoLanguage* language, \
PangoLogAttr* log_attrs, int attrs_len)"
  XEN_ASSERT_TYPE(XEN_char__P(text), text, 1, "pango_get_log_attrs", "char*");
  XEN_ASSERT_TYPE(XEN_int_P(length), length, 2, "pango_get_log_attrs", "int");
  XEN_ASSERT_TYPE(XEN_int_P(level), level, 3, "pango_get_log_attrs", "int");
  XEN_ASSERT_TYPE(XEN_PangoLanguage__P(language), language, 4, "pango_get_log_attrs", "PangoLanguage*");
  XEN_ASSERT_TYPE(XEN_PangoLogAttr__P(log_attrs), log_attrs, 5, "pango_get_log_attrs", "PangoLogAttr*");
  XEN_ASSERT_TYPE(XEN_int_P(attrs_len), attrs_len, 6, "pango_get_log_attrs", "int");
  pango_get_log_attrs(XEN_TO_C_char_(text), XEN_TO_C_int(length), XEN_TO_C_int(level), XEN_TO_C_PangoLanguage_(language), 
                      XEN_TO_C_PangoLogAttr_(log_attrs), XEN_TO_C_int(attrs_len));
  return(XEN_FALSE);
}

static XEN gxg_pango_context_get_type(void)
{
  #define H_pango_context_get_type "GType pango_context_get_type( void)"
  return(C_TO_XEN_GType(pango_context_get_type()));
}

static XEN gxg_pango_context_list_families(XEN context, XEN families, XEN n_families)
{
  #define H_pango_context_list_families "void pango_context_list_families(PangoContext* context, PangoFontFamily*** [families], \
int* [n_families])"
  PangoFontFamily** ref_families;
  int ref_n_families;
  XEN_ASSERT_TYPE(XEN_PangoContext__P(context), context, 1, "pango_context_list_families", "PangoContext*");
  pango_context_list_families(XEN_TO_C_PangoContext_(context), &ref_families, &ref_n_families);
  return(XEN_LIST_2(C_TO_XEN_PangoFontFamily__(ref_families), C_TO_XEN_int(ref_n_families)));
}

static XEN gxg_pango_context_load_font(XEN context, XEN desc)
{
  #define H_pango_context_load_font "PangoFont* pango_context_load_font(PangoContext* context, PangoFontDescription* desc)"
  XEN_ASSERT_TYPE(XEN_PangoContext__P(context), context, 1, "pango_context_load_font", "PangoContext*");
  XEN_ASSERT_TYPE(XEN_PangoFontDescription__P(desc), desc, 2, "pango_context_load_font", "PangoFontDescription*");
  return(C_TO_XEN_PangoFont_(pango_context_load_font(XEN_TO_C_PangoContext_(context), XEN_TO_C_PangoFontDescription_(desc))));
}

static XEN gxg_pango_context_load_fontset(XEN context, XEN desc, XEN language)
{
  #define H_pango_context_load_fontset "PangoFontset* pango_context_load_fontset(PangoContext* context, \
PangoFontDescription* desc, PangoLanguage* language)"
  XEN_ASSERT_TYPE(XEN_PangoContext__P(context), context, 1, "pango_context_load_fontset", "PangoContext*");
  XEN_ASSERT_TYPE(XEN_PangoFontDescription__P(desc), desc, 2, "pango_context_load_fontset", "PangoFontDescription*");
  XEN_ASSERT_TYPE(XEN_PangoLanguage__P(language), language, 3, "pango_context_load_fontset", "PangoLanguage*");
  return(C_TO_XEN_PangoFontset_(pango_context_load_fontset(XEN_TO_C_PangoContext_(context), XEN_TO_C_PangoFontDescription_(desc), 
                                                           XEN_TO_C_PangoLanguage_(language))));
}

static XEN gxg_pango_context_get_metrics(XEN context, XEN desc, XEN language)
{
  #define H_pango_context_get_metrics "PangoFontMetrics* pango_context_get_metrics(PangoContext* context, \
PangoFontDescription* desc, PangoLanguage* language)"
  XEN_ASSERT_TYPE(XEN_PangoContext__P(context), context, 1, "pango_context_get_metrics", "PangoContext*");
  XEN_ASSERT_TYPE(XEN_PangoFontDescription__P(desc), desc, 2, "pango_context_get_metrics", "PangoFontDescription*");
  XEN_ASSERT_TYPE(XEN_PangoLanguage__P(language), language, 3, "pango_context_get_metrics", "PangoLanguage*");
  return(C_TO_XEN_PangoFontMetrics_(pango_context_get_metrics(XEN_TO_C_PangoContext_(context), XEN_TO_C_PangoFontDescription_(desc), 
                                                              XEN_TO_C_PangoLanguage_(language))));
}

static XEN gxg_pango_context_set_font_description(XEN context, XEN desc)
{
  #define H_pango_context_set_font_description "void pango_context_set_font_description(PangoContext* context, \
PangoFontDescription* desc)"
  XEN_ASSERT_TYPE(XEN_PangoContext__P(context), context, 1, "pango_context_set_font_description", "PangoContext*");
  XEN_ASSERT_TYPE(XEN_PangoFontDescription__P(desc), desc, 2, "pango_context_set_font_description", "PangoFontDescription*");
  pango_context_set_font_description(XEN_TO_C_PangoContext_(context), XEN_TO_C_PangoFontDescription_(desc));
  return(XEN_FALSE);
}

static XEN gxg_pango_context_get_font_description(XEN context)
{
  #define H_pango_context_get_font_description "PangoFontDescription* pango_context_get_font_description(PangoContext* context)"
  XEN_ASSERT_TYPE(XEN_PangoContext__P(context), context, 1, "pango_context_get_font_description", "PangoContext*");
  return(C_TO_XEN_PangoFontDescription_(pango_context_get_font_description(XEN_TO_C_PangoContext_(context))));
}

static XEN gxg_pango_context_get_language(XEN context)
{
  #define H_pango_context_get_language "PangoLanguage* pango_context_get_language(PangoContext* context)"
  XEN_ASSERT_TYPE(XEN_PangoContext__P(context), context, 1, "pango_context_get_language", "PangoContext*");
  return(C_TO_XEN_PangoLanguage_(pango_context_get_language(XEN_TO_C_PangoContext_(context))));
}

static XEN gxg_pango_context_set_language(XEN context, XEN language)
{
  #define H_pango_context_set_language "void pango_context_set_language(PangoContext* context, PangoLanguage* language)"
  XEN_ASSERT_TYPE(XEN_PangoContext__P(context), context, 1, "pango_context_set_language", "PangoContext*");
  XEN_ASSERT_TYPE(XEN_PangoLanguage__P(language), language, 2, "pango_context_set_language", "PangoLanguage*");
  pango_context_set_language(XEN_TO_C_PangoContext_(context), XEN_TO_C_PangoLanguage_(language));
  return(XEN_FALSE);
}

static XEN gxg_pango_context_set_base_dir(XEN context, XEN direction)
{
  #define H_pango_context_set_base_dir "void pango_context_set_base_dir(PangoContext* context, PangoDirection direction)"
  XEN_ASSERT_TYPE(XEN_PangoContext__P(context), context, 1, "pango_context_set_base_dir", "PangoContext*");
  XEN_ASSERT_TYPE(XEN_PangoDirection_P(direction), direction, 2, "pango_context_set_base_dir", "PangoDirection");
  pango_context_set_base_dir(XEN_TO_C_PangoContext_(context), XEN_TO_C_PangoDirection(direction));
  return(XEN_FALSE);
}

static XEN gxg_pango_context_get_base_dir(XEN context)
{
  #define H_pango_context_get_base_dir "PangoDirection pango_context_get_base_dir(PangoContext* context)"
  XEN_ASSERT_TYPE(XEN_PangoContext__P(context), context, 1, "pango_context_get_base_dir", "PangoContext*");
  return(C_TO_XEN_PangoDirection(pango_context_get_base_dir(XEN_TO_C_PangoContext_(context))));
}

static XEN gxg_pango_itemize(XEN context, XEN text, XEN start_index, XEN length, XEN attrs, XEN cached_iter)
{
  #define H_pango_itemize "GList* pango_itemize(PangoContext* context, char* text, int start_index, int length, \
PangoAttrList* attrs, PangoAttrIterator* cached_iter)"
  XEN_ASSERT_TYPE(XEN_PangoContext__P(context), context, 1, "pango_itemize", "PangoContext*");
  XEN_ASSERT_TYPE(XEN_char__P(text), text, 2, "pango_itemize", "char*");
  XEN_ASSERT_TYPE(XEN_int_P(start_index), start_index, 3, "pango_itemize", "int");
  XEN_ASSERT_TYPE(XEN_int_P(length), length, 4, "pango_itemize", "int");
  XEN_ASSERT_TYPE(XEN_PangoAttrList__P(attrs), attrs, 5, "pango_itemize", "PangoAttrList*");
  XEN_ASSERT_TYPE(XEN_PangoAttrIterator__P(cached_iter), cached_iter, 6, "pango_itemize", "PangoAttrIterator*");
  return(C_TO_XEN_GList_(pango_itemize(XEN_TO_C_PangoContext_(context), XEN_TO_C_char_(text), XEN_TO_C_int(start_index), 
                                       XEN_TO_C_int(length), XEN_TO_C_PangoAttrList_(attrs), XEN_TO_C_PangoAttrIterator_(cached_iter))));
}

static XEN gxg_pango_coverage_new(void)
{
  #define H_pango_coverage_new "PangoCoverage* pango_coverage_new( void)"
  return(C_TO_XEN_PangoCoverage_(pango_coverage_new()));
}

static XEN gxg_pango_coverage_ref(XEN coverage)
{
  #define H_pango_coverage_ref "PangoCoverage* pango_coverage_ref(PangoCoverage* coverage)"
  XEN_ASSERT_TYPE(XEN_PangoCoverage__P(coverage), coverage, 1, "pango_coverage_ref", "PangoCoverage*");
  return(C_TO_XEN_PangoCoverage_(pango_coverage_ref(XEN_TO_C_PangoCoverage_(coverage))));
}

static XEN gxg_pango_coverage_unref(XEN coverage)
{
  #define H_pango_coverage_unref "void pango_coverage_unref(PangoCoverage* coverage)"
  XEN_ASSERT_TYPE(XEN_PangoCoverage__P(coverage), coverage, 1, "pango_coverage_unref", "PangoCoverage*");
  pango_coverage_unref(XEN_TO_C_PangoCoverage_(coverage));
  return(XEN_FALSE);
}

static XEN gxg_pango_coverage_copy(XEN coverage)
{
  #define H_pango_coverage_copy "PangoCoverage* pango_coverage_copy(PangoCoverage* coverage)"
  XEN_ASSERT_TYPE(XEN_PangoCoverage__P(coverage), coverage, 1, "pango_coverage_copy", "PangoCoverage*");
  return(C_TO_XEN_PangoCoverage_(pango_coverage_copy(XEN_TO_C_PangoCoverage_(coverage))));
}

static XEN gxg_pango_coverage_get(XEN coverage, XEN index)
{
  #define H_pango_coverage_get "PangoCoverageLevel pango_coverage_get(PangoCoverage* coverage, int index)"
  XEN_ASSERT_TYPE(XEN_PangoCoverage__P(coverage), coverage, 1, "pango_coverage_get", "PangoCoverage*");
  XEN_ASSERT_TYPE(XEN_int_P(index), index, 2, "pango_coverage_get", "int");
  return(C_TO_XEN_PangoCoverageLevel(pango_coverage_get(XEN_TO_C_PangoCoverage_(coverage), XEN_TO_C_int(index))));
}

static XEN gxg_pango_coverage_set(XEN coverage, XEN index, XEN level)
{
  #define H_pango_coverage_set "void pango_coverage_set(PangoCoverage* coverage, int index, PangoCoverageLevel level)"
  XEN_ASSERT_TYPE(XEN_PangoCoverage__P(coverage), coverage, 1, "pango_coverage_set", "PangoCoverage*");
  XEN_ASSERT_TYPE(XEN_int_P(index), index, 2, "pango_coverage_set", "int");
  XEN_ASSERT_TYPE(XEN_PangoCoverageLevel_P(level), level, 3, "pango_coverage_set", "PangoCoverageLevel");
  pango_coverage_set(XEN_TO_C_PangoCoverage_(coverage), XEN_TO_C_int(index), XEN_TO_C_PangoCoverageLevel(level));
  return(XEN_FALSE);
}

static XEN gxg_pango_coverage_max(XEN coverage, XEN other)
{
  #define H_pango_coverage_max "void pango_coverage_max(PangoCoverage* coverage, PangoCoverage* other)"
  XEN_ASSERT_TYPE(XEN_PangoCoverage__P(coverage), coverage, 1, "pango_coverage_max", "PangoCoverage*");
  XEN_ASSERT_TYPE(XEN_PangoCoverage__P(other), other, 2, "pango_coverage_max", "PangoCoverage*");
  pango_coverage_max(XEN_TO_C_PangoCoverage_(coverage), XEN_TO_C_PangoCoverage_(other));
  return(XEN_FALSE);
}

static XEN gxg_pango_coverage_to_bytes(XEN coverage, XEN bytes, XEN n_bytes)
{
  #define H_pango_coverage_to_bytes "void pango_coverage_to_bytes(PangoCoverage* coverage, guchar** [bytes], \
int* [n_bytes])"
  guchar* ref_bytes;
  int ref_n_bytes;
  XEN_ASSERT_TYPE(XEN_PangoCoverage__P(coverage), coverage, 1, "pango_coverage_to_bytes", "PangoCoverage*");
  pango_coverage_to_bytes(XEN_TO_C_PangoCoverage_(coverage), &ref_bytes, &ref_n_bytes);
  return(XEN_LIST_2(C_TO_XEN_guchar_(ref_bytes), C_TO_XEN_int(ref_n_bytes)));
}

static XEN gxg_pango_coverage_from_bytes(XEN bytes, XEN n_bytes)
{
  #define H_pango_coverage_from_bytes "PangoCoverage* pango_coverage_from_bytes(guchar* bytes, int n_bytes)"
  XEN_ASSERT_TYPE(XEN_guchar__P(bytes), bytes, 1, "pango_coverage_from_bytes", "guchar*");
  XEN_ASSERT_TYPE(XEN_int_P(n_bytes), n_bytes, 2, "pango_coverage_from_bytes", "int");
  return(C_TO_XEN_PangoCoverage_(pango_coverage_from_bytes(XEN_TO_C_guchar_(bytes), XEN_TO_C_int(n_bytes))));
}

static XEN gxg_pango_attr_type_get_type(void)
{
  #define H_pango_attr_type_get_type "GType pango_attr_type_get_type( void)"
  return(C_TO_XEN_GType(pango_attr_type_get_type()));
}

static XEN gxg_pango_underline_get_type(void)
{
  #define H_pango_underline_get_type "GType pango_underline_get_type( void)"
  return(C_TO_XEN_GType(pango_underline_get_type()));
}

static XEN gxg_pango_coverage_level_get_type(void)
{
  #define H_pango_coverage_level_get_type "GType pango_coverage_level_get_type( void)"
  return(C_TO_XEN_GType(pango_coverage_level_get_type()));
}

static XEN gxg_pango_style_get_type(void)
{
  #define H_pango_style_get_type "GType pango_style_get_type( void)"
  return(C_TO_XEN_GType(pango_style_get_type()));
}

static XEN gxg_pango_variant_get_type(void)
{
  #define H_pango_variant_get_type "GType pango_variant_get_type( void)"
  return(C_TO_XEN_GType(pango_variant_get_type()));
}

static XEN gxg_pango_weight_get_type(void)
{
  #define H_pango_weight_get_type "GType pango_weight_get_type( void)"
  return(C_TO_XEN_GType(pango_weight_get_type()));
}

static XEN gxg_pango_stretch_get_type(void)
{
  #define H_pango_stretch_get_type "GType pango_stretch_get_type( void)"
  return(C_TO_XEN_GType(pango_stretch_get_type()));
}

static XEN gxg_pango_font_mask_get_type(void)
{
  #define H_pango_font_mask_get_type "GType pango_font_mask_get_type( void)"
  return(C_TO_XEN_GType(pango_font_mask_get_type()));
}

static XEN gxg_pango_alignment_get_type(void)
{
  #define H_pango_alignment_get_type "GType pango_alignment_get_type( void)"
  return(C_TO_XEN_GType(pango_alignment_get_type()));
}

static XEN gxg_pango_wrap_mode_get_type(void)
{
  #define H_pango_wrap_mode_get_type "GType pango_wrap_mode_get_type( void)"
  return(C_TO_XEN_GType(pango_wrap_mode_get_type()));
}

static XEN gxg_pango_tab_align_get_type(void)
{
  #define H_pango_tab_align_get_type "GType pango_tab_align_get_type( void)"
  return(C_TO_XEN_GType(pango_tab_align_get_type()));
}

static XEN gxg_pango_direction_get_type(void)
{
  #define H_pango_direction_get_type "GType pango_direction_get_type( void)"
  return(C_TO_XEN_GType(pango_direction_get_type()));
}

static XEN gxg_pango_font_description_get_type(void)
{
  #define H_pango_font_description_get_type "GType pango_font_description_get_type( void)"
  return(C_TO_XEN_GType(pango_font_description_get_type()));
}

static XEN gxg_pango_font_description_new(void)
{
  #define H_pango_font_description_new "PangoFontDescription* pango_font_description_new( void)"
  return(C_TO_XEN_PangoFontDescription_(pango_font_description_new()));
}

static XEN gxg_pango_font_description_copy(XEN desc)
{
  #define H_pango_font_description_copy "PangoFontDescription* pango_font_description_copy(PangoFontDescription* desc)"
  XEN_ASSERT_TYPE(XEN_PangoFontDescription__P(desc), desc, 1, "pango_font_description_copy", "PangoFontDescription*");
  return(C_TO_XEN_PangoFontDescription_(pango_font_description_copy(XEN_TO_C_PangoFontDescription_(desc))));
}

static XEN gxg_pango_font_description_copy_static(XEN desc)
{
  #define H_pango_font_description_copy_static "PangoFontDescription* pango_font_description_copy_static(PangoFontDescription* desc)"
  XEN_ASSERT_TYPE(XEN_PangoFontDescription__P(desc), desc, 1, "pango_font_description_copy_static", "PangoFontDescription*");
  return(C_TO_XEN_PangoFontDescription_(pango_font_description_copy_static(XEN_TO_C_PangoFontDescription_(desc))));
}

static XEN gxg_pango_font_description_hash(XEN desc)
{
  #define H_pango_font_description_hash "guint pango_font_description_hash(PangoFontDescription* desc)"
  XEN_ASSERT_TYPE(XEN_PangoFontDescription__P(desc), desc, 1, "pango_font_description_hash", "PangoFontDescription*");
  return(C_TO_XEN_guint(pango_font_description_hash(XEN_TO_C_PangoFontDescription_(desc))));
}

static XEN gxg_pango_font_description_equal(XEN desc1, XEN desc2)
{
  #define H_pango_font_description_equal "gboolean pango_font_description_equal(PangoFontDescription* desc1, \
PangoFontDescription* desc2)"
  XEN_ASSERT_TYPE(XEN_PangoFontDescription__P(desc1), desc1, 1, "pango_font_description_equal", "PangoFontDescription*");
  XEN_ASSERT_TYPE(XEN_PangoFontDescription__P(desc2), desc2, 2, "pango_font_description_equal", "PangoFontDescription*");
  return(C_TO_XEN_gboolean(pango_font_description_equal(XEN_TO_C_PangoFontDescription_(desc1), XEN_TO_C_PangoFontDescription_(desc2))));
}

static XEN gxg_pango_font_description_free(XEN desc)
{
  #define H_pango_font_description_free "void pango_font_description_free(PangoFontDescription* desc)"
  XEN_ASSERT_TYPE(XEN_PangoFontDescription__P(desc), desc, 1, "pango_font_description_free", "PangoFontDescription*");
  pango_font_description_free(XEN_TO_C_PangoFontDescription_(desc));
  return(XEN_FALSE);
}

static XEN gxg_pango_font_descriptions_free(XEN descs, XEN n_descs)
{
  #define H_pango_font_descriptions_free "void pango_font_descriptions_free(PangoFontDescription** descs, \
int n_descs)"
  XEN_ASSERT_TYPE(XEN_PangoFontDescription___P(descs), descs, 1, "pango_font_descriptions_free", "PangoFontDescription**");
  XEN_ASSERT_TYPE(XEN_int_P(n_descs), n_descs, 2, "pango_font_descriptions_free", "int");
  pango_font_descriptions_free(XEN_TO_C_PangoFontDescription__(descs), XEN_TO_C_int(n_descs));
  return(XEN_FALSE);
}

static XEN gxg_pango_font_description_set_family(XEN desc, XEN family)
{
  #define H_pango_font_description_set_family "void pango_font_description_set_family(PangoFontDescription* desc, \
char* family)"
  XEN_ASSERT_TYPE(XEN_PangoFontDescription__P(desc), desc, 1, "pango_font_description_set_family", "PangoFontDescription*");
  XEN_ASSERT_TYPE(XEN_char__P(family), family, 2, "pango_font_description_set_family", "char*");
  pango_font_description_set_family(XEN_TO_C_PangoFontDescription_(desc), XEN_TO_C_char_(family));
  return(XEN_FALSE);
}

static XEN gxg_pango_font_description_set_family_static(XEN desc, XEN family)
{
  #define H_pango_font_description_set_family_static "void pango_font_description_set_family_static(PangoFontDescription* desc, \
char* family)"
  XEN_ASSERT_TYPE(XEN_PangoFontDescription__P(desc), desc, 1, "pango_font_description_set_family_static", "PangoFontDescription*");
  XEN_ASSERT_TYPE(XEN_char__P(family), family, 2, "pango_font_description_set_family_static", "char*");
  pango_font_description_set_family_static(XEN_TO_C_PangoFontDescription_(desc), XEN_TO_C_char_(family));
  return(XEN_FALSE);
}

static XEN gxg_pango_font_description_get_family(XEN desc)
{
  #define H_pango_font_description_get_family "char* pango_font_description_get_family(PangoFontDescription* desc)"
  XEN_ASSERT_TYPE(XEN_PangoFontDescription__P(desc), desc, 1, "pango_font_description_get_family", "PangoFontDescription*");
  return(C_TO_XEN_char_(pango_font_description_get_family(XEN_TO_C_PangoFontDescription_(desc))));
}

static XEN gxg_pango_font_description_set_style(XEN desc, XEN style)
{
  #define H_pango_font_description_set_style "void pango_font_description_set_style(PangoFontDescription* desc, \
PangoStyle style)"
  XEN_ASSERT_TYPE(XEN_PangoFontDescription__P(desc), desc, 1, "pango_font_description_set_style", "PangoFontDescription*");
  XEN_ASSERT_TYPE(XEN_PangoStyle_P(style), style, 2, "pango_font_description_set_style", "PangoStyle");
  pango_font_description_set_style(XEN_TO_C_PangoFontDescription_(desc), XEN_TO_C_PangoStyle(style));
  return(XEN_FALSE);
}

static XEN gxg_pango_font_description_get_style(XEN desc)
{
  #define H_pango_font_description_get_style "PangoStyle pango_font_description_get_style(PangoFontDescription* desc)"
  XEN_ASSERT_TYPE(XEN_PangoFontDescription__P(desc), desc, 1, "pango_font_description_get_style", "PangoFontDescription*");
  return(C_TO_XEN_PangoStyle(pango_font_description_get_style(XEN_TO_C_PangoFontDescription_(desc))));
}

static XEN gxg_pango_font_description_set_variant(XEN desc, XEN variant)
{
  #define H_pango_font_description_set_variant "void pango_font_description_set_variant(PangoFontDescription* desc, \
PangoVariant variant)"
  XEN_ASSERT_TYPE(XEN_PangoFontDescription__P(desc), desc, 1, "pango_font_description_set_variant", "PangoFontDescription*");
  XEN_ASSERT_TYPE(XEN_PangoVariant_P(variant), variant, 2, "pango_font_description_set_variant", "PangoVariant");
  pango_font_description_set_variant(XEN_TO_C_PangoFontDescription_(desc), XEN_TO_C_PangoVariant(variant));
  return(XEN_FALSE);
}

static XEN gxg_pango_font_description_get_variant(XEN desc)
{
  #define H_pango_font_description_get_variant "PangoVariant pango_font_description_get_variant(PangoFontDescription* desc)"
  XEN_ASSERT_TYPE(XEN_PangoFontDescription__P(desc), desc, 1, "pango_font_description_get_variant", "PangoFontDescription*");
  return(C_TO_XEN_PangoVariant(pango_font_description_get_variant(XEN_TO_C_PangoFontDescription_(desc))));
}

static XEN gxg_pango_font_description_set_weight(XEN desc, XEN weight)
{
  #define H_pango_font_description_set_weight "void pango_font_description_set_weight(PangoFontDescription* desc, \
PangoWeight weight)"
  XEN_ASSERT_TYPE(XEN_PangoFontDescription__P(desc), desc, 1, "pango_font_description_set_weight", "PangoFontDescription*");
  XEN_ASSERT_TYPE(XEN_PangoWeight_P(weight), weight, 2, "pango_font_description_set_weight", "PangoWeight");
  pango_font_description_set_weight(XEN_TO_C_PangoFontDescription_(desc), XEN_TO_C_PangoWeight(weight));
  return(XEN_FALSE);
}

static XEN gxg_pango_font_description_get_weight(XEN desc)
{
  #define H_pango_font_description_get_weight "PangoWeight pango_font_description_get_weight(PangoFontDescription* desc)"
  XEN_ASSERT_TYPE(XEN_PangoFontDescription__P(desc), desc, 1, "pango_font_description_get_weight", "PangoFontDescription*");
  return(C_TO_XEN_PangoWeight(pango_font_description_get_weight(XEN_TO_C_PangoFontDescription_(desc))));
}

static XEN gxg_pango_font_description_set_stretch(XEN desc, XEN stretch)
{
  #define H_pango_font_description_set_stretch "void pango_font_description_set_stretch(PangoFontDescription* desc, \
PangoStretch stretch)"
  XEN_ASSERT_TYPE(XEN_PangoFontDescription__P(desc), desc, 1, "pango_font_description_set_stretch", "PangoFontDescription*");
  XEN_ASSERT_TYPE(XEN_PangoStretch_P(stretch), stretch, 2, "pango_font_description_set_stretch", "PangoStretch");
  pango_font_description_set_stretch(XEN_TO_C_PangoFontDescription_(desc), XEN_TO_C_PangoStretch(stretch));
  return(XEN_FALSE);
}

static XEN gxg_pango_font_description_get_stretch(XEN desc)
{
  #define H_pango_font_description_get_stretch "PangoStretch pango_font_description_get_stretch(PangoFontDescription* desc)"
  XEN_ASSERT_TYPE(XEN_PangoFontDescription__P(desc), desc, 1, "pango_font_description_get_stretch", "PangoFontDescription*");
  return(C_TO_XEN_PangoStretch(pango_font_description_get_stretch(XEN_TO_C_PangoFontDescription_(desc))));
}

static XEN gxg_pango_font_description_set_size(XEN desc, XEN size)
{
  #define H_pango_font_description_set_size "void pango_font_description_set_size(PangoFontDescription* desc, \
gint size)"
  XEN_ASSERT_TYPE(XEN_PangoFontDescription__P(desc), desc, 1, "pango_font_description_set_size", "PangoFontDescription*");
  XEN_ASSERT_TYPE(XEN_gint_P(size), size, 2, "pango_font_description_set_size", "gint");
  pango_font_description_set_size(XEN_TO_C_PangoFontDescription_(desc), XEN_TO_C_gint(size));
  return(XEN_FALSE);
}

static XEN gxg_pango_font_description_get_size(XEN desc)
{
  #define H_pango_font_description_get_size "gint pango_font_description_get_size(PangoFontDescription* desc)"
  XEN_ASSERT_TYPE(XEN_PangoFontDescription__P(desc), desc, 1, "pango_font_description_get_size", "PangoFontDescription*");
  return(C_TO_XEN_gint(pango_font_description_get_size(XEN_TO_C_PangoFontDescription_(desc))));
}

static XEN gxg_pango_font_description_get_set_fields(XEN desc)
{
  #define H_pango_font_description_get_set_fields "PangoFontMask pango_font_description_get_set_fields(PangoFontDescription* desc)"
  XEN_ASSERT_TYPE(XEN_PangoFontDescription__P(desc), desc, 1, "pango_font_description_get_set_fields", "PangoFontDescription*");
  return(C_TO_XEN_PangoFontMask(pango_font_description_get_set_fields(XEN_TO_C_PangoFontDescription_(desc))));
}

static XEN gxg_pango_font_description_unset_fields(XEN desc, XEN to_unset)
{
  #define H_pango_font_description_unset_fields "void pango_font_description_unset_fields(PangoFontDescription* desc, \
PangoFontMask to_unset)"
  XEN_ASSERT_TYPE(XEN_PangoFontDescription__P(desc), desc, 1, "pango_font_description_unset_fields", "PangoFontDescription*");
  XEN_ASSERT_TYPE(XEN_PangoFontMask_P(to_unset), to_unset, 2, "pango_font_description_unset_fields", "PangoFontMask");
  pango_font_description_unset_fields(XEN_TO_C_PangoFontDescription_(desc), XEN_TO_C_PangoFontMask(to_unset));
  return(XEN_FALSE);
}

static XEN gxg_pango_font_description_merge(XEN desc, XEN desc_to_merge, XEN replace_existing)
{
  #define H_pango_font_description_merge "void pango_font_description_merge(PangoFontDescription* desc, \
PangoFontDescription* desc_to_merge, gboolean replace_existing)"
  XEN_ASSERT_TYPE(XEN_PangoFontDescription__P(desc), desc, 1, "pango_font_description_merge", "PangoFontDescription*");
  XEN_ASSERT_TYPE(XEN_PangoFontDescription__P(desc_to_merge), desc_to_merge, 2, "pango_font_description_merge", "PangoFontDescription*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(replace_existing), replace_existing, 3, "pango_font_description_merge", "gboolean");
  pango_font_description_merge(XEN_TO_C_PangoFontDescription_(desc), XEN_TO_C_PangoFontDescription_(desc_to_merge), XEN_TO_C_gboolean(replace_existing));
  return(XEN_FALSE);
}

static XEN gxg_pango_font_description_merge_static(XEN desc, XEN desc_to_merge, XEN replace_existing)
{
  #define H_pango_font_description_merge_static "void pango_font_description_merge_static(PangoFontDescription* desc, \
PangoFontDescription* desc_to_merge, gboolean replace_existing)"
  XEN_ASSERT_TYPE(XEN_PangoFontDescription__P(desc), desc, 1, "pango_font_description_merge_static", "PangoFontDescription*");
  XEN_ASSERT_TYPE(XEN_PangoFontDescription__P(desc_to_merge), desc_to_merge, 2, "pango_font_description_merge_static", "PangoFontDescription*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(replace_existing), replace_existing, 3, "pango_font_description_merge_static", "gboolean");
  pango_font_description_merge_static(XEN_TO_C_PangoFontDescription_(desc), XEN_TO_C_PangoFontDescription_(desc_to_merge), 
                                      XEN_TO_C_gboolean(replace_existing));
  return(XEN_FALSE);
}

static XEN gxg_pango_font_description_better_match(XEN desc, XEN old_match, XEN new_match)
{
  #define H_pango_font_description_better_match "gboolean pango_font_description_better_match(PangoFontDescription* desc, \
PangoFontDescription* old_match, PangoFontDescription* new_match)"
  XEN_ASSERT_TYPE(XEN_PangoFontDescription__P(desc), desc, 1, "pango_font_description_better_match", "PangoFontDescription*");
  XEN_ASSERT_TYPE(XEN_PangoFontDescription__P(old_match), old_match, 2, "pango_font_description_better_match", "PangoFontDescription*");
  XEN_ASSERT_TYPE(XEN_PangoFontDescription__P(new_match), new_match, 3, "pango_font_description_better_match", "PangoFontDescription*");
  return(C_TO_XEN_gboolean(pango_font_description_better_match(XEN_TO_C_PangoFontDescription_(desc), XEN_TO_C_PangoFontDescription_(old_match), 
                                                               XEN_TO_C_PangoFontDescription_(new_match))));
}

static XEN gxg_pango_font_description_from_string(XEN str)
{
  #define H_pango_font_description_from_string "PangoFontDescription* pango_font_description_from_string(char* str)"
  XEN_ASSERT_TYPE(XEN_char__P(str), str, 1, "pango_font_description_from_string", "char*");
  return(C_TO_XEN_PangoFontDescription_(pango_font_description_from_string(XEN_TO_C_char_(str))));
}

static XEN gxg_pango_font_description_to_string(XEN desc)
{
  #define H_pango_font_description_to_string "char* pango_font_description_to_string(PangoFontDescription* desc)"
  XEN_ASSERT_TYPE(XEN_PangoFontDescription__P(desc), desc, 1, "pango_font_description_to_string", "PangoFontDescription*");
  return(C_TO_XEN_char_(pango_font_description_to_string(XEN_TO_C_PangoFontDescription_(desc))));
}

static XEN gxg_pango_font_description_to_filename(XEN desc)
{
  #define H_pango_font_description_to_filename "char* pango_font_description_to_filename(PangoFontDescription* desc)"
  XEN_ASSERT_TYPE(XEN_PangoFontDescription__P(desc), desc, 1, "pango_font_description_to_filename", "PangoFontDescription*");
  return(C_TO_XEN_char_(pango_font_description_to_filename(XEN_TO_C_PangoFontDescription_(desc))));
}

static XEN gxg_pango_font_metrics_get_type(void)
{
  #define H_pango_font_metrics_get_type "GType pango_font_metrics_get_type( void)"
  return(C_TO_XEN_GType(pango_font_metrics_get_type()));
}

static XEN gxg_pango_font_metrics_ref(XEN metrics)
{
  #define H_pango_font_metrics_ref "PangoFontMetrics* pango_font_metrics_ref(PangoFontMetrics* metrics)"
  XEN_ASSERT_TYPE(XEN_PangoFontMetrics__P(metrics), metrics, 1, "pango_font_metrics_ref", "PangoFontMetrics*");
  return(C_TO_XEN_PangoFontMetrics_(pango_font_metrics_ref(XEN_TO_C_PangoFontMetrics_(metrics))));
}

static XEN gxg_pango_font_metrics_unref(XEN metrics)
{
  #define H_pango_font_metrics_unref "void pango_font_metrics_unref(PangoFontMetrics* metrics)"
  XEN_ASSERT_TYPE(XEN_PangoFontMetrics__P(metrics), metrics, 1, "pango_font_metrics_unref", "PangoFontMetrics*");
  pango_font_metrics_unref(XEN_TO_C_PangoFontMetrics_(metrics));
  return(XEN_FALSE);
}

static XEN gxg_pango_font_metrics_get_ascent(XEN metrics)
{
  #define H_pango_font_metrics_get_ascent "int pango_font_metrics_get_ascent(PangoFontMetrics* metrics)"
  XEN_ASSERT_TYPE(XEN_PangoFontMetrics__P(metrics), metrics, 1, "pango_font_metrics_get_ascent", "PangoFontMetrics*");
  return(C_TO_XEN_int(pango_font_metrics_get_ascent(XEN_TO_C_PangoFontMetrics_(metrics))));
}

static XEN gxg_pango_font_metrics_get_descent(XEN metrics)
{
  #define H_pango_font_metrics_get_descent "int pango_font_metrics_get_descent(PangoFontMetrics* metrics)"
  XEN_ASSERT_TYPE(XEN_PangoFontMetrics__P(metrics), metrics, 1, "pango_font_metrics_get_descent", "PangoFontMetrics*");
  return(C_TO_XEN_int(pango_font_metrics_get_descent(XEN_TO_C_PangoFontMetrics_(metrics))));
}

static XEN gxg_pango_font_metrics_get_approximate_char_width(XEN metrics)
{
  #define H_pango_font_metrics_get_approximate_char_width "int pango_font_metrics_get_approximate_char_width(PangoFontMetrics* metrics)"
  XEN_ASSERT_TYPE(XEN_PangoFontMetrics__P(metrics), metrics, 1, "pango_font_metrics_get_approximate_char_width", "PangoFontMetrics*");
  return(C_TO_XEN_int(pango_font_metrics_get_approximate_char_width(XEN_TO_C_PangoFontMetrics_(metrics))));
}

static XEN gxg_pango_font_metrics_get_approximate_digit_width(XEN metrics)
{
  #define H_pango_font_metrics_get_approximate_digit_width "int pango_font_metrics_get_approximate_digit_width(PangoFontMetrics* metrics)"
  XEN_ASSERT_TYPE(XEN_PangoFontMetrics__P(metrics), metrics, 1, "pango_font_metrics_get_approximate_digit_width", "PangoFontMetrics*");
  return(C_TO_XEN_int(pango_font_metrics_get_approximate_digit_width(XEN_TO_C_PangoFontMetrics_(metrics))));
}

static XEN gxg_pango_font_face_get_type(void)
{
  #define H_pango_font_face_get_type "GType pango_font_face_get_type( void)"
  return(C_TO_XEN_GType(pango_font_face_get_type()));
}

static XEN gxg_pango_font_face_describe(XEN face)
{
  #define H_pango_font_face_describe "PangoFontDescription* pango_font_face_describe(PangoFontFace* face)"
  XEN_ASSERT_TYPE(XEN_PangoFontFace__P(face), face, 1, "pango_font_face_describe", "PangoFontFace*");
  return(C_TO_XEN_PangoFontDescription_(pango_font_face_describe(XEN_TO_C_PangoFontFace_(face))));
}

static XEN gxg_pango_font_face_get_face_name(XEN face)
{
  #define H_pango_font_face_get_face_name "char* pango_font_face_get_face_name(PangoFontFace* face)"
  XEN_ASSERT_TYPE(XEN_PangoFontFace__P(face), face, 1, "pango_font_face_get_face_name", "PangoFontFace*");
  return(C_TO_XEN_char_(pango_font_face_get_face_name(XEN_TO_C_PangoFontFace_(face))));
}

static XEN gxg_pango_font_get_type(void)
{
  #define H_pango_font_get_type "GType pango_font_get_type( void)"
  return(C_TO_XEN_GType(pango_font_get_type()));
}

static XEN gxg_pango_font_describe(XEN font)
{
  #define H_pango_font_describe "PangoFontDescription* pango_font_describe(PangoFont* font)"
  XEN_ASSERT_TYPE(XEN_PangoFont__P(font), font, 1, "pango_font_describe", "PangoFont*");
  return(C_TO_XEN_PangoFontDescription_(pango_font_describe(XEN_TO_C_PangoFont_(font))));
}

static XEN gxg_pango_font_get_coverage(XEN font, XEN language)
{
  #define H_pango_font_get_coverage "PangoCoverage* pango_font_get_coverage(PangoFont* font, PangoLanguage* language)"
  XEN_ASSERT_TYPE(XEN_PangoFont__P(font), font, 1, "pango_font_get_coverage", "PangoFont*");
  XEN_ASSERT_TYPE(XEN_PangoLanguage__P(language), language, 2, "pango_font_get_coverage", "PangoLanguage*");
  return(C_TO_XEN_PangoCoverage_(pango_font_get_coverage(XEN_TO_C_PangoFont_(font), XEN_TO_C_PangoLanguage_(language))));
}

static XEN gxg_pango_font_find_shaper(XEN font, XEN language, XEN ch)
{
  #define H_pango_font_find_shaper "PangoEngineShape* pango_font_find_shaper(PangoFont* font, PangoLanguage* language, \
guint32 ch)"
  XEN_ASSERT_TYPE(XEN_PangoFont__P(font), font, 1, "pango_font_find_shaper", "PangoFont*");
  XEN_ASSERT_TYPE(XEN_PangoLanguage__P(language), language, 2, "pango_font_find_shaper", "PangoLanguage*");
  XEN_ASSERT_TYPE(XEN_guint32_P(ch), ch, 3, "pango_font_find_shaper", "guint32");
  return(C_TO_XEN_PangoEngineShape_(pango_font_find_shaper(XEN_TO_C_PangoFont_(font), XEN_TO_C_PangoLanguage_(language), 
                                                           XEN_TO_C_guint32(ch))));
}

static XEN gxg_pango_font_get_metrics(XEN font, XEN language)
{
  #define H_pango_font_get_metrics "PangoFontMetrics* pango_font_get_metrics(PangoFont* font, PangoLanguage* language)"
  XEN_ASSERT_TYPE(XEN_PangoFont__P(font), font, 1, "pango_font_get_metrics", "PangoFont*");
  XEN_ASSERT_TYPE(XEN_PangoLanguage__P(language), language, 2, "pango_font_get_metrics", "PangoLanguage*");
  return(C_TO_XEN_PangoFontMetrics_(pango_font_get_metrics(XEN_TO_C_PangoFont_(font), XEN_TO_C_PangoLanguage_(language))));
}

static XEN gxg_pango_font_get_glyph_extents(XEN font, XEN glyph, XEN ink_rect, XEN logical_rect)
{
  #define H_pango_font_get_glyph_extents "void pango_font_get_glyph_extents(PangoFont* font, PangoGlyph glyph, \
PangoRectangle* ink_rect, PangoRectangle* logical_rect)"
  XEN_ASSERT_TYPE(XEN_PangoFont__P(font), font, 1, "pango_font_get_glyph_extents", "PangoFont*");
  XEN_ASSERT_TYPE(XEN_PangoGlyph_P(glyph), glyph, 2, "pango_font_get_glyph_extents", "PangoGlyph");
  XEN_ASSERT_TYPE(XEN_PangoRectangle__P(ink_rect), ink_rect, 3, "pango_font_get_glyph_extents", "PangoRectangle*");
  XEN_ASSERT_TYPE(XEN_PangoRectangle__P(logical_rect), logical_rect, 4, "pango_font_get_glyph_extents", "PangoRectangle*");
  pango_font_get_glyph_extents(XEN_TO_C_PangoFont_(font), XEN_TO_C_PangoGlyph(glyph), XEN_TO_C_PangoRectangle_(ink_rect), 
                               XEN_TO_C_PangoRectangle_(logical_rect));
  return(XEN_FALSE);
}

static XEN gxg_pango_font_map_get_type(void)
{
  #define H_pango_font_map_get_type "GType pango_font_map_get_type( void)"
  return(C_TO_XEN_GType(pango_font_map_get_type()));
}

static XEN gxg_pango_font_map_load_font(XEN fontmap, XEN context, XEN desc)
{
  #define H_pango_font_map_load_font "PangoFont* pango_font_map_load_font(PangoFontMap* fontmap, PangoContext* context, \
PangoFontDescription* desc)"
  XEN_ASSERT_TYPE(XEN_PangoFontMap__P(fontmap), fontmap, 1, "pango_font_map_load_font", "PangoFontMap*");
  XEN_ASSERT_TYPE(XEN_PangoContext__P(context), context, 2, "pango_font_map_load_font", "PangoContext*");
  XEN_ASSERT_TYPE(XEN_PangoFontDescription__P(desc), desc, 3, "pango_font_map_load_font", "PangoFontDescription*");
  return(C_TO_XEN_PangoFont_(pango_font_map_load_font(XEN_TO_C_PangoFontMap_(fontmap), XEN_TO_C_PangoContext_(context), XEN_TO_C_PangoFontDescription_(desc))));
}

static XEN gxg_pango_font_map_load_fontset(XEN fontmap, XEN context, XEN desc, XEN language)
{
  #define H_pango_font_map_load_fontset "PangoFontset* pango_font_map_load_fontset(PangoFontMap* fontmap, \
PangoContext* context, PangoFontDescription* desc, PangoLanguage* language)"
  XEN_ASSERT_TYPE(XEN_PangoFontMap__P(fontmap), fontmap, 1, "pango_font_map_load_fontset", "PangoFontMap*");
  XEN_ASSERT_TYPE(XEN_PangoContext__P(context), context, 2, "pango_font_map_load_fontset", "PangoContext*");
  XEN_ASSERT_TYPE(XEN_PangoFontDescription__P(desc), desc, 3, "pango_font_map_load_fontset", "PangoFontDescription*");
  XEN_ASSERT_TYPE(XEN_PangoLanguage__P(language), language, 4, "pango_font_map_load_fontset", "PangoLanguage*");
  return(C_TO_XEN_PangoFontset_(pango_font_map_load_fontset(XEN_TO_C_PangoFontMap_(fontmap), XEN_TO_C_PangoContext_(context), 
                                                            XEN_TO_C_PangoFontDescription_(desc), XEN_TO_C_PangoLanguage_(language))));
}

static XEN gxg_pango_font_map_list_families(XEN fontmap, XEN families, XEN n_families)
{
  #define H_pango_font_map_list_families "void pango_font_map_list_families(PangoFontMap* fontmap, PangoFontFamily*** [families], \
int* [n_families])"
  PangoFontFamily** ref_families;
  int ref_n_families;
  XEN_ASSERT_TYPE(XEN_PangoFontMap__P(fontmap), fontmap, 1, "pango_font_map_list_families", "PangoFontMap*");
  pango_font_map_list_families(XEN_TO_C_PangoFontMap_(fontmap), &ref_families, &ref_n_families);
  return(XEN_LIST_2(C_TO_XEN_PangoFontFamily__(ref_families), C_TO_XEN_int(ref_n_families)));
}

static XEN gxg_pango_glyph_string_new(void)
{
  #define H_pango_glyph_string_new "PangoGlyphString* pango_glyph_string_new( void)"
  return(C_TO_XEN_PangoGlyphString_(pango_glyph_string_new()));
}

static XEN gxg_pango_glyph_string_set_size(XEN string, XEN new_len)
{
  #define H_pango_glyph_string_set_size "void pango_glyph_string_set_size(PangoGlyphString* string, gint new_len)"
  XEN_ASSERT_TYPE(XEN_PangoGlyphString__P(string), string, 1, "pango_glyph_string_set_size", "PangoGlyphString*");
  XEN_ASSERT_TYPE(XEN_gint_P(new_len), new_len, 2, "pango_glyph_string_set_size", "gint");
  pango_glyph_string_set_size(XEN_TO_C_PangoGlyphString_(string), XEN_TO_C_gint(new_len));
  return(XEN_FALSE);
}

static XEN gxg_pango_glyph_string_get_type(void)
{
  #define H_pango_glyph_string_get_type "GType pango_glyph_string_get_type( void)"
  return(C_TO_XEN_GType(pango_glyph_string_get_type()));
}

static XEN gxg_pango_glyph_string_copy(XEN string)
{
  #define H_pango_glyph_string_copy "PangoGlyphString* pango_glyph_string_copy(PangoGlyphString* string)"
  XEN_ASSERT_TYPE(XEN_PangoGlyphString__P(string), string, 1, "pango_glyph_string_copy", "PangoGlyphString*");
  return(C_TO_XEN_PangoGlyphString_(pango_glyph_string_copy(XEN_TO_C_PangoGlyphString_(string))));
}

static XEN gxg_pango_glyph_string_free(XEN string)
{
  #define H_pango_glyph_string_free "void pango_glyph_string_free(PangoGlyphString* string)"
  XEN_ASSERT_TYPE(XEN_PangoGlyphString__P(string), string, 1, "pango_glyph_string_free", "PangoGlyphString*");
  pango_glyph_string_free(XEN_TO_C_PangoGlyphString_(string));
  return(XEN_FALSE);
}

static XEN gxg_pango_glyph_string_extents(XEN glyphs, XEN font, XEN ink_rect, XEN logical_rect)
{
  #define H_pango_glyph_string_extents "void pango_glyph_string_extents(PangoGlyphString* glyphs, PangoFont* font, \
PangoRectangle* ink_rect, PangoRectangle* logical_rect)"
  XEN_ASSERT_TYPE(XEN_PangoGlyphString__P(glyphs), glyphs, 1, "pango_glyph_string_extents", "PangoGlyphString*");
  XEN_ASSERT_TYPE(XEN_PangoFont__P(font), font, 2, "pango_glyph_string_extents", "PangoFont*");
  XEN_ASSERT_TYPE(XEN_PangoRectangle__P(ink_rect), ink_rect, 3, "pango_glyph_string_extents", "PangoRectangle*");
  XEN_ASSERT_TYPE(XEN_PangoRectangle__P(logical_rect), logical_rect, 4, "pango_glyph_string_extents", "PangoRectangle*");
  pango_glyph_string_extents(XEN_TO_C_PangoGlyphString_(glyphs), XEN_TO_C_PangoFont_(font), XEN_TO_C_PangoRectangle_(ink_rect), 
                             XEN_TO_C_PangoRectangle_(logical_rect));
  return(XEN_FALSE);
}

static XEN gxg_pango_glyph_string_extents_range(XEN glyphs, XEN start, XEN end, XEN font, XEN ink_rect, XEN logical_rect)
{
  #define H_pango_glyph_string_extents_range "void pango_glyph_string_extents_range(PangoGlyphString* glyphs, \
int start, int end, PangoFont* font, PangoRectangle* ink_rect, PangoRectangle* logical_rect)"
  XEN_ASSERT_TYPE(XEN_PangoGlyphString__P(glyphs), glyphs, 1, "pango_glyph_string_extents_range", "PangoGlyphString*");
  XEN_ASSERT_TYPE(XEN_int_P(start), start, 2, "pango_glyph_string_extents_range", "int");
  XEN_ASSERT_TYPE(XEN_int_P(end), end, 3, "pango_glyph_string_extents_range", "int");
  XEN_ASSERT_TYPE(XEN_PangoFont__P(font), font, 4, "pango_glyph_string_extents_range", "PangoFont*");
  XEN_ASSERT_TYPE(XEN_PangoRectangle__P(ink_rect), ink_rect, 5, "pango_glyph_string_extents_range", "PangoRectangle*");
  XEN_ASSERT_TYPE(XEN_PangoRectangle__P(logical_rect), logical_rect, 6, "pango_glyph_string_extents_range", "PangoRectangle*");
  pango_glyph_string_extents_range(XEN_TO_C_PangoGlyphString_(glyphs), XEN_TO_C_int(start), XEN_TO_C_int(end), XEN_TO_C_PangoFont_(font), 
                                   XEN_TO_C_PangoRectangle_(ink_rect), XEN_TO_C_PangoRectangle_(logical_rect));
  return(XEN_FALSE);
}

static XEN gxg_pango_glyph_string_get_logical_widths(XEN glyphs, XEN text, XEN length, XEN embedding_level, XEN logical_widths)
{
  #define H_pango_glyph_string_get_logical_widths "void pango_glyph_string_get_logical_widths(PangoGlyphString* glyphs, \
char* text, int length, int embedding_level, int* logical_widths)"
  XEN_ASSERT_TYPE(XEN_PangoGlyphString__P(glyphs), glyphs, 1, "pango_glyph_string_get_logical_widths", "PangoGlyphString*");
  XEN_ASSERT_TYPE(XEN_char__P(text), text, 2, "pango_glyph_string_get_logical_widths", "char*");
  XEN_ASSERT_TYPE(XEN_int_P(length), length, 3, "pango_glyph_string_get_logical_widths", "int");
  XEN_ASSERT_TYPE(XEN_int_P(embedding_level), embedding_level, 4, "pango_glyph_string_get_logical_widths", "int");
  XEN_ASSERT_TYPE(XEN_int__P(logical_widths), logical_widths, 5, "pango_glyph_string_get_logical_widths", "int*");
  pango_glyph_string_get_logical_widths(XEN_TO_C_PangoGlyphString_(glyphs), XEN_TO_C_char_(text), XEN_TO_C_int(length), XEN_TO_C_int(embedding_level), 
                                        XEN_TO_C_int_(logical_widths));
  return(XEN_FALSE);
}

static XEN gxg_pango_glyph_string_index_to_x(XEN glyphs, XEN text, XEN length, XEN analysis, XEN index, XEN trailing, XEN x_pos)
{
  #define H_pango_glyph_string_index_to_x "void pango_glyph_string_index_to_x(PangoGlyphString* glyphs, \
char* text, int length, PangoAnalysis* analysis, int index, gboolean trailing, int* x_pos)"
  XEN_ASSERT_TYPE(XEN_PangoGlyphString__P(glyphs), glyphs, 1, "pango_glyph_string_index_to_x", "PangoGlyphString*");
  XEN_ASSERT_TYPE(XEN_char__P(text), text, 2, "pango_glyph_string_index_to_x", "char*");
  XEN_ASSERT_TYPE(XEN_int_P(length), length, 3, "pango_glyph_string_index_to_x", "int");
  XEN_ASSERT_TYPE(XEN_PangoAnalysis__P(analysis), analysis, 4, "pango_glyph_string_index_to_x", "PangoAnalysis*");
  XEN_ASSERT_TYPE(XEN_int_P(index), index, 5, "pango_glyph_string_index_to_x", "int");
  XEN_ASSERT_TYPE(XEN_gboolean_P(trailing), trailing, 6, "pango_glyph_string_index_to_x", "gboolean");
  XEN_ASSERT_TYPE(XEN_int__P(x_pos), x_pos, 7, "pango_glyph_string_index_to_x", "int*");
  pango_glyph_string_index_to_x(XEN_TO_C_PangoGlyphString_(glyphs), XEN_TO_C_char_(text), XEN_TO_C_int(length), XEN_TO_C_PangoAnalysis_(analysis), 
                                XEN_TO_C_int(index), XEN_TO_C_gboolean(trailing), XEN_TO_C_int_(x_pos));
  return(XEN_FALSE);
}

static XEN gxg_pango_glyph_string_x_to_index(XEN glyphs, XEN text, XEN length, XEN analysis, XEN x_pos, XEN index, XEN trailing)
{
  #define H_pango_glyph_string_x_to_index "void pango_glyph_string_x_to_index(PangoGlyphString* glyphs, \
char* text, int length, PangoAnalysis* analysis, int x_pos, int* index, int* trailing)"
  XEN_ASSERT_TYPE(XEN_PangoGlyphString__P(glyphs), glyphs, 1, "pango_glyph_string_x_to_index", "PangoGlyphString*");
  XEN_ASSERT_TYPE(XEN_char__P(text), text, 2, "pango_glyph_string_x_to_index", "char*");
  XEN_ASSERT_TYPE(XEN_int_P(length), length, 3, "pango_glyph_string_x_to_index", "int");
  XEN_ASSERT_TYPE(XEN_PangoAnalysis__P(analysis), analysis, 4, "pango_glyph_string_x_to_index", "PangoAnalysis*");
  XEN_ASSERT_TYPE(XEN_int_P(x_pos), x_pos, 5, "pango_glyph_string_x_to_index", "int");
  XEN_ASSERT_TYPE(XEN_int__P(index), index, 6, "pango_glyph_string_x_to_index", "int*");
  XEN_ASSERT_TYPE(XEN_int__P(trailing), trailing, 7, "pango_glyph_string_x_to_index", "int*");
  pango_glyph_string_x_to_index(XEN_TO_C_PangoGlyphString_(glyphs), XEN_TO_C_char_(text), XEN_TO_C_int(length), XEN_TO_C_PangoAnalysis_(analysis), 
                                XEN_TO_C_int(x_pos), XEN_TO_C_int_(index), XEN_TO_C_int_(trailing));
  return(XEN_FALSE);
}

static XEN gxg_pango_shape(XEN text, XEN length, XEN analysis, XEN glyphs)
{
  #define H_pango_shape "void pango_shape(gchar* text, gint length, PangoAnalysis* analysis, PangoGlyphString* glyphs)"
  XEN_ASSERT_TYPE(XEN_gchar__P(text), text, 1, "pango_shape", "gchar*");
  XEN_ASSERT_TYPE(XEN_gint_P(length), length, 2, "pango_shape", "gint");
  XEN_ASSERT_TYPE(XEN_PangoAnalysis__P(analysis), analysis, 3, "pango_shape", "PangoAnalysis*");
  XEN_ASSERT_TYPE(XEN_PangoGlyphString__P(glyphs), glyphs, 4, "pango_shape", "PangoGlyphString*");
  pango_shape(XEN_TO_C_gchar_(text), XEN_TO_C_gint(length), XEN_TO_C_PangoAnalysis_(analysis), XEN_TO_C_PangoGlyphString_(glyphs));
  return(XEN_FALSE);
}

static XEN gxg_pango_reorder_items(XEN logical_items)
{
  #define H_pango_reorder_items "GList* pango_reorder_items(GList* logical_items)"
  XEN_ASSERT_TYPE(XEN_GList__P(logical_items), logical_items, 1, "pango_reorder_items", "GList*");
  return(C_TO_XEN_GList_(pango_reorder_items(XEN_TO_C_GList_(logical_items))));
}

static XEN gxg_pango_item_new(void)
{
  #define H_pango_item_new "PangoItem* pango_item_new( void)"
  return(C_TO_XEN_PangoItem_(pango_item_new()));
}

static XEN gxg_pango_item_copy(XEN item)
{
  #define H_pango_item_copy "PangoItem* pango_item_copy(PangoItem* item)"
  XEN_ASSERT_TYPE(XEN_PangoItem__P(item), item, 1, "pango_item_copy", "PangoItem*");
  return(C_TO_XEN_PangoItem_(pango_item_copy(XEN_TO_C_PangoItem_(item))));
}

static XEN gxg_pango_item_free(XEN item)
{
  #define H_pango_item_free "void pango_item_free(PangoItem* item)"
  XEN_ASSERT_TYPE(XEN_PangoItem__P(item), item, 1, "pango_item_free", "PangoItem*");
  pango_item_free(XEN_TO_C_PangoItem_(item));
  return(XEN_FALSE);
}

static XEN gxg_pango_item_split(XEN orig, XEN split_index, XEN split_offset)
{
  #define H_pango_item_split "PangoItem* pango_item_split(PangoItem* orig, int split_index, int split_offset)"
  XEN_ASSERT_TYPE(XEN_PangoItem__P(orig), orig, 1, "pango_item_split", "PangoItem*");
  XEN_ASSERT_TYPE(XEN_int_P(split_index), split_index, 2, "pango_item_split", "int");
  XEN_ASSERT_TYPE(XEN_int_P(split_offset), split_offset, 3, "pango_item_split", "int");
  return(C_TO_XEN_PangoItem_(pango_item_split(XEN_TO_C_PangoItem_(orig), XEN_TO_C_int(split_index), XEN_TO_C_int(split_offset))));
}

static XEN gxg_pango_layout_get_type(void)
{
  #define H_pango_layout_get_type "GType pango_layout_get_type( void)"
  return(C_TO_XEN_GType(pango_layout_get_type()));
}

static XEN gxg_pango_layout_new(XEN context)
{
  #define H_pango_layout_new "PangoLayout* pango_layout_new(PangoContext* context)"
  XEN_ASSERT_TYPE(XEN_PangoContext__P(context), context, 1, "pango_layout_new", "PangoContext*");
  return(C_TO_XEN_PangoLayout_(pango_layout_new(XEN_TO_C_PangoContext_(context))));
}

static XEN gxg_pango_layout_copy(XEN src)
{
  #define H_pango_layout_copy "PangoLayout* pango_layout_copy(PangoLayout* src)"
  XEN_ASSERT_TYPE(XEN_PangoLayout__P(src), src, 1, "pango_layout_copy", "PangoLayout*");
  return(C_TO_XEN_PangoLayout_(pango_layout_copy(XEN_TO_C_PangoLayout_(src))));
}

static XEN gxg_pango_layout_get_context(XEN layout)
{
  #define H_pango_layout_get_context "PangoContext* pango_layout_get_context(PangoLayout* layout)"
  XEN_ASSERT_TYPE(XEN_PangoLayout__P(layout), layout, 1, "pango_layout_get_context", "PangoLayout*");
  return(C_TO_XEN_PangoContext_(pango_layout_get_context(XEN_TO_C_PangoLayout_(layout))));
}

static XEN gxg_pango_layout_set_attributes(XEN layout, XEN attrs)
{
  #define H_pango_layout_set_attributes "void pango_layout_set_attributes(PangoLayout* layout, PangoAttrList* attrs)"
  XEN_ASSERT_TYPE(XEN_PangoLayout__P(layout), layout, 1, "pango_layout_set_attributes", "PangoLayout*");
  XEN_ASSERT_TYPE(XEN_PangoAttrList__P(attrs), attrs, 2, "pango_layout_set_attributes", "PangoAttrList*");
  pango_layout_set_attributes(XEN_TO_C_PangoLayout_(layout), XEN_TO_C_PangoAttrList_(attrs));
  return(XEN_FALSE);
}

static XEN gxg_pango_layout_get_attributes(XEN layout)
{
  #define H_pango_layout_get_attributes "PangoAttrList* pango_layout_get_attributes(PangoLayout* layout)"
  XEN_ASSERT_TYPE(XEN_PangoLayout__P(layout), layout, 1, "pango_layout_get_attributes", "PangoLayout*");
  return(C_TO_XEN_PangoAttrList_(pango_layout_get_attributes(XEN_TO_C_PangoLayout_(layout))));
}

static XEN gxg_pango_layout_set_text(XEN layout, XEN text, XEN length)
{
  #define H_pango_layout_set_text "void pango_layout_set_text(PangoLayout* layout, char* text, int length)"
  XEN_ASSERT_TYPE(XEN_PangoLayout__P(layout), layout, 1, "pango_layout_set_text", "PangoLayout*");
  XEN_ASSERT_TYPE(XEN_char__P(text), text, 2, "pango_layout_set_text", "char*");
  XEN_ASSERT_TYPE(XEN_int_P(length), length, 3, "pango_layout_set_text", "int");
  pango_layout_set_text(XEN_TO_C_PangoLayout_(layout), XEN_TO_C_char_(text), XEN_TO_C_int(length));
  return(XEN_FALSE);
}

static XEN gxg_pango_layout_get_text(XEN layout)
{
  #define H_pango_layout_get_text "char* pango_layout_get_text(PangoLayout* layout)"
  XEN_ASSERT_TYPE(XEN_PangoLayout__P(layout), layout, 1, "pango_layout_get_text", "PangoLayout*");
  return(C_TO_XEN_char_(pango_layout_get_text(XEN_TO_C_PangoLayout_(layout))));
}

static XEN gxg_pango_layout_set_markup(XEN layout, XEN markup, XEN length)
{
  #define H_pango_layout_set_markup "void pango_layout_set_markup(PangoLayout* layout, char* markup, \
int length)"
  XEN_ASSERT_TYPE(XEN_PangoLayout__P(layout), layout, 1, "pango_layout_set_markup", "PangoLayout*");
  XEN_ASSERT_TYPE(XEN_char__P(markup), markup, 2, "pango_layout_set_markup", "char*");
  XEN_ASSERT_TYPE(XEN_int_P(length), length, 3, "pango_layout_set_markup", "int");
  pango_layout_set_markup(XEN_TO_C_PangoLayout_(layout), XEN_TO_C_char_(markup), XEN_TO_C_int(length));
  return(XEN_FALSE);
}

static XEN gxg_pango_layout_set_markup_with_accel(XEN layout, XEN markup, XEN length, XEN accel_marker, XEN accel_char)
{
  #define H_pango_layout_set_markup_with_accel "void pango_layout_set_markup_with_accel(PangoLayout* layout, \
char* markup, int length, gunichar accel_marker, gunichar* accel_char)"
  XEN_ASSERT_TYPE(XEN_PangoLayout__P(layout), layout, 1, "pango_layout_set_markup_with_accel", "PangoLayout*");
  XEN_ASSERT_TYPE(XEN_char__P(markup), markup, 2, "pango_layout_set_markup_with_accel", "char*");
  XEN_ASSERT_TYPE(XEN_int_P(length), length, 3, "pango_layout_set_markup_with_accel", "int");
  XEN_ASSERT_TYPE(XEN_gunichar_P(accel_marker), accel_marker, 4, "pango_layout_set_markup_with_accel", "gunichar");
  XEN_ASSERT_TYPE(XEN_gunichar__P(accel_char), accel_char, 5, "pango_layout_set_markup_with_accel", "gunichar*");
  pango_layout_set_markup_with_accel(XEN_TO_C_PangoLayout_(layout), XEN_TO_C_char_(markup), XEN_TO_C_int(length), XEN_TO_C_gunichar(accel_marker), 
                                     XEN_TO_C_gunichar_(accel_char));
  return(XEN_FALSE);
}

static XEN gxg_pango_layout_set_font_description(XEN layout, XEN desc)
{
  #define H_pango_layout_set_font_description "void pango_layout_set_font_description(PangoLayout* layout, \
PangoFontDescription* desc)"
  XEN_ASSERT_TYPE(XEN_PangoLayout__P(layout), layout, 1, "pango_layout_set_font_description", "PangoLayout*");
  XEN_ASSERT_TYPE(XEN_PangoFontDescription__P(desc), desc, 2, "pango_layout_set_font_description", "PangoFontDescription*");
  pango_layout_set_font_description(XEN_TO_C_PangoLayout_(layout), XEN_TO_C_PangoFontDescription_(desc));
  return(XEN_FALSE);
}

static XEN gxg_pango_layout_set_width(XEN layout, XEN width)
{
  #define H_pango_layout_set_width "void pango_layout_set_width(PangoLayout* layout, int width)"
  XEN_ASSERT_TYPE(XEN_PangoLayout__P(layout), layout, 1, "pango_layout_set_width", "PangoLayout*");
  XEN_ASSERT_TYPE(XEN_int_P(width), width, 2, "pango_layout_set_width", "int");
  pango_layout_set_width(XEN_TO_C_PangoLayout_(layout), XEN_TO_C_int(width));
  return(XEN_FALSE);
}

static XEN gxg_pango_layout_get_width(XEN layout)
{
  #define H_pango_layout_get_width "int pango_layout_get_width(PangoLayout* layout)"
  XEN_ASSERT_TYPE(XEN_PangoLayout__P(layout), layout, 1, "pango_layout_get_width", "PangoLayout*");
  return(C_TO_XEN_int(pango_layout_get_width(XEN_TO_C_PangoLayout_(layout))));
}

static XEN gxg_pango_layout_set_wrap(XEN layout, XEN wrap)
{
  #define H_pango_layout_set_wrap "void pango_layout_set_wrap(PangoLayout* layout, PangoWrapMode wrap)"
  XEN_ASSERT_TYPE(XEN_PangoLayout__P(layout), layout, 1, "pango_layout_set_wrap", "PangoLayout*");
  XEN_ASSERT_TYPE(XEN_PangoWrapMode_P(wrap), wrap, 2, "pango_layout_set_wrap", "PangoWrapMode");
  pango_layout_set_wrap(XEN_TO_C_PangoLayout_(layout), XEN_TO_C_PangoWrapMode(wrap));
  return(XEN_FALSE);
}

static XEN gxg_pango_layout_get_wrap(XEN layout)
{
  #define H_pango_layout_get_wrap "PangoWrapMode pango_layout_get_wrap(PangoLayout* layout)"
  XEN_ASSERT_TYPE(XEN_PangoLayout__P(layout), layout, 1, "pango_layout_get_wrap", "PangoLayout*");
  return(C_TO_XEN_PangoWrapMode(pango_layout_get_wrap(XEN_TO_C_PangoLayout_(layout))));
}

static XEN gxg_pango_layout_set_indent(XEN layout, XEN indent)
{
  #define H_pango_layout_set_indent "void pango_layout_set_indent(PangoLayout* layout, int indent)"
  XEN_ASSERT_TYPE(XEN_PangoLayout__P(layout), layout, 1, "pango_layout_set_indent", "PangoLayout*");
  XEN_ASSERT_TYPE(XEN_int_P(indent), indent, 2, "pango_layout_set_indent", "int");
  pango_layout_set_indent(XEN_TO_C_PangoLayout_(layout), XEN_TO_C_int(indent));
  return(XEN_FALSE);
}

static XEN gxg_pango_layout_get_indent(XEN layout)
{
  #define H_pango_layout_get_indent "int pango_layout_get_indent(PangoLayout* layout)"
  XEN_ASSERT_TYPE(XEN_PangoLayout__P(layout), layout, 1, "pango_layout_get_indent", "PangoLayout*");
  return(C_TO_XEN_int(pango_layout_get_indent(XEN_TO_C_PangoLayout_(layout))));
}

static XEN gxg_pango_layout_set_spacing(XEN layout, XEN spacing)
{
  #define H_pango_layout_set_spacing "void pango_layout_set_spacing(PangoLayout* layout, int spacing)"
  XEN_ASSERT_TYPE(XEN_PangoLayout__P(layout), layout, 1, "pango_layout_set_spacing", "PangoLayout*");
  XEN_ASSERT_TYPE(XEN_int_P(spacing), spacing, 2, "pango_layout_set_spacing", "int");
  pango_layout_set_spacing(XEN_TO_C_PangoLayout_(layout), XEN_TO_C_int(spacing));
  return(XEN_FALSE);
}

static XEN gxg_pango_layout_get_spacing(XEN layout)
{
  #define H_pango_layout_get_spacing "int pango_layout_get_spacing(PangoLayout* layout)"
  XEN_ASSERT_TYPE(XEN_PangoLayout__P(layout), layout, 1, "pango_layout_get_spacing", "PangoLayout*");
  return(C_TO_XEN_int(pango_layout_get_spacing(XEN_TO_C_PangoLayout_(layout))));
}

static XEN gxg_pango_layout_set_justify(XEN layout, XEN justify)
{
  #define H_pango_layout_set_justify "void pango_layout_set_justify(PangoLayout* layout, gboolean justify)"
  XEN_ASSERT_TYPE(XEN_PangoLayout__P(layout), layout, 1, "pango_layout_set_justify", "PangoLayout*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(justify), justify, 2, "pango_layout_set_justify", "gboolean");
  pango_layout_set_justify(XEN_TO_C_PangoLayout_(layout), XEN_TO_C_gboolean(justify));
  return(XEN_FALSE);
}

static XEN gxg_pango_layout_get_justify(XEN layout)
{
  #define H_pango_layout_get_justify "gboolean pango_layout_get_justify(PangoLayout* layout)"
  XEN_ASSERT_TYPE(XEN_PangoLayout__P(layout), layout, 1, "pango_layout_get_justify", "PangoLayout*");
  return(C_TO_XEN_gboolean(pango_layout_get_justify(XEN_TO_C_PangoLayout_(layout))));
}

static XEN gxg_pango_layout_set_alignment(XEN layout, XEN alignment)
{
  #define H_pango_layout_set_alignment "void pango_layout_set_alignment(PangoLayout* layout, PangoAlignment alignment)"
  XEN_ASSERT_TYPE(XEN_PangoLayout__P(layout), layout, 1, "pango_layout_set_alignment", "PangoLayout*");
  XEN_ASSERT_TYPE(XEN_PangoAlignment_P(alignment), alignment, 2, "pango_layout_set_alignment", "PangoAlignment");
  pango_layout_set_alignment(XEN_TO_C_PangoLayout_(layout), XEN_TO_C_PangoAlignment(alignment));
  return(XEN_FALSE);
}

static XEN gxg_pango_layout_get_alignment(XEN layout)
{
  #define H_pango_layout_get_alignment "PangoAlignment pango_layout_get_alignment(PangoLayout* layout)"
  XEN_ASSERT_TYPE(XEN_PangoLayout__P(layout), layout, 1, "pango_layout_get_alignment", "PangoLayout*");
  return(C_TO_XEN_PangoAlignment(pango_layout_get_alignment(XEN_TO_C_PangoLayout_(layout))));
}

static XEN gxg_pango_layout_set_tabs(XEN layout, XEN tabs)
{
  #define H_pango_layout_set_tabs "void pango_layout_set_tabs(PangoLayout* layout, PangoTabArray* tabs)"
  XEN_ASSERT_TYPE(XEN_PangoLayout__P(layout), layout, 1, "pango_layout_set_tabs", "PangoLayout*");
  XEN_ASSERT_TYPE(XEN_PangoTabArray__P(tabs), tabs, 2, "pango_layout_set_tabs", "PangoTabArray*");
  pango_layout_set_tabs(XEN_TO_C_PangoLayout_(layout), XEN_TO_C_PangoTabArray_(tabs));
  return(XEN_FALSE);
}

static XEN gxg_pango_layout_get_tabs(XEN layout)
{
  #define H_pango_layout_get_tabs "PangoTabArray* pango_layout_get_tabs(PangoLayout* layout)"
  XEN_ASSERT_TYPE(XEN_PangoLayout__P(layout), layout, 1, "pango_layout_get_tabs", "PangoLayout*");
  return(C_TO_XEN_PangoTabArray_(pango_layout_get_tabs(XEN_TO_C_PangoLayout_(layout))));
}

static XEN gxg_pango_layout_set_single_paragraph_mode(XEN layout, XEN setting)
{
  #define H_pango_layout_set_single_paragraph_mode "void pango_layout_set_single_paragraph_mode(PangoLayout* layout, \
gboolean setting)"
  XEN_ASSERT_TYPE(XEN_PangoLayout__P(layout), layout, 1, "pango_layout_set_single_paragraph_mode", "PangoLayout*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(setting), setting, 2, "pango_layout_set_single_paragraph_mode", "gboolean");
  pango_layout_set_single_paragraph_mode(XEN_TO_C_PangoLayout_(layout), XEN_TO_C_gboolean(setting));
  return(XEN_FALSE);
}

static XEN gxg_pango_layout_get_single_paragraph_mode(XEN layout)
{
  #define H_pango_layout_get_single_paragraph_mode "gboolean pango_layout_get_single_paragraph_mode(PangoLayout* layout)"
  XEN_ASSERT_TYPE(XEN_PangoLayout__P(layout), layout, 1, "pango_layout_get_single_paragraph_mode", "PangoLayout*");
  return(C_TO_XEN_gboolean(pango_layout_get_single_paragraph_mode(XEN_TO_C_PangoLayout_(layout))));
}

static XEN gxg_pango_layout_context_changed(XEN layout)
{
  #define H_pango_layout_context_changed "void pango_layout_context_changed(PangoLayout* layout)"
  XEN_ASSERT_TYPE(XEN_PangoLayout__P(layout), layout, 1, "pango_layout_context_changed", "PangoLayout*");
  pango_layout_context_changed(XEN_TO_C_PangoLayout_(layout));
  return(XEN_FALSE);
}

static XEN gxg_pango_layout_get_log_attrs(XEN layout, XEN attrs, XEN n_attrs)
{
  #define H_pango_layout_get_log_attrs "void pango_layout_get_log_attrs(PangoLayout* layout, PangoLogAttr** [attrs], \
gint* [n_attrs])"
  PangoLogAttr* ref_attrs;
  gint ref_n_attrs;
  XEN_ASSERT_TYPE(XEN_PangoLayout__P(layout), layout, 1, "pango_layout_get_log_attrs", "PangoLayout*");
  pango_layout_get_log_attrs(XEN_TO_C_PangoLayout_(layout), &ref_attrs, &ref_n_attrs);
  return(XEN_LIST_2(C_TO_XEN_PangoLogAttr_(ref_attrs), C_TO_XEN_gint(ref_n_attrs)));
}

static XEN gxg_pango_layout_index_to_pos(XEN layout, XEN index, XEN pos)
{
  #define H_pango_layout_index_to_pos "void pango_layout_index_to_pos(PangoLayout* layout, int index, \
PangoRectangle* pos)"
  XEN_ASSERT_TYPE(XEN_PangoLayout__P(layout), layout, 1, "pango_layout_index_to_pos", "PangoLayout*");
  XEN_ASSERT_TYPE(XEN_int_P(index), index, 2, "pango_layout_index_to_pos", "int");
  XEN_ASSERT_TYPE(XEN_PangoRectangle__P(pos), pos, 3, "pango_layout_index_to_pos", "PangoRectangle*");
  pango_layout_index_to_pos(XEN_TO_C_PangoLayout_(layout), XEN_TO_C_int(index), XEN_TO_C_PangoRectangle_(pos));
  return(XEN_FALSE);
}

static XEN gxg_pango_layout_get_cursor_pos(XEN layout, XEN index, XEN strong_pos, XEN weak_pos)
{
  #define H_pango_layout_get_cursor_pos "void pango_layout_get_cursor_pos(PangoLayout* layout, int index, \
PangoRectangle* strong_pos, PangoRectangle* weak_pos)"
  XEN_ASSERT_TYPE(XEN_PangoLayout__P(layout), layout, 1, "pango_layout_get_cursor_pos", "PangoLayout*");
  XEN_ASSERT_TYPE(XEN_int_P(index), index, 2, "pango_layout_get_cursor_pos", "int");
  XEN_ASSERT_TYPE(XEN_PangoRectangle__P(strong_pos), strong_pos, 3, "pango_layout_get_cursor_pos", "PangoRectangle*");
  XEN_ASSERT_TYPE(XEN_PangoRectangle__P(weak_pos), weak_pos, 4, "pango_layout_get_cursor_pos", "PangoRectangle*");
  pango_layout_get_cursor_pos(XEN_TO_C_PangoLayout_(layout), XEN_TO_C_int(index), XEN_TO_C_PangoRectangle_(strong_pos), XEN_TO_C_PangoRectangle_(weak_pos));
  return(XEN_FALSE);
}

static XEN gxg_pango_layout_move_cursor_visually(XEN layout, XEN strong, XEN old_index, XEN old_trailing, XEN direction, XEN new_index, XEN new_trailing)
{
  #define H_pango_layout_move_cursor_visually "void pango_layout_move_cursor_visually(PangoLayout* layout, \
gboolean strong, int old_index, int old_trailing, int direction, int* new_index, int* new_trailing)"
  XEN_ASSERT_TYPE(XEN_PangoLayout__P(layout), layout, 1, "pango_layout_move_cursor_visually", "PangoLayout*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(strong), strong, 2, "pango_layout_move_cursor_visually", "gboolean");
  XEN_ASSERT_TYPE(XEN_int_P(old_index), old_index, 3, "pango_layout_move_cursor_visually", "int");
  XEN_ASSERT_TYPE(XEN_int_P(old_trailing), old_trailing, 4, "pango_layout_move_cursor_visually", "int");
  XEN_ASSERT_TYPE(XEN_int_P(direction), direction, 5, "pango_layout_move_cursor_visually", "int");
  XEN_ASSERT_TYPE(XEN_int__P(new_index), new_index, 6, "pango_layout_move_cursor_visually", "int*");
  XEN_ASSERT_TYPE(XEN_int__P(new_trailing), new_trailing, 7, "pango_layout_move_cursor_visually", "int*");
  pango_layout_move_cursor_visually(XEN_TO_C_PangoLayout_(layout), XEN_TO_C_gboolean(strong), XEN_TO_C_int(old_index), XEN_TO_C_int(old_trailing), 
                                    XEN_TO_C_int(direction), XEN_TO_C_int_(new_index), XEN_TO_C_int_(new_trailing));
  return(XEN_FALSE);
}

static XEN gxg_pango_layout_xy_to_index(XEN layout, XEN x, XEN y, XEN index, XEN trailing)
{
  #define H_pango_layout_xy_to_index "gboolean pango_layout_xy_to_index(PangoLayout* layout, int x, int y, \
int* index, int* trailing)"
  XEN_ASSERT_TYPE(XEN_PangoLayout__P(layout), layout, 1, "pango_layout_xy_to_index", "PangoLayout*");
  XEN_ASSERT_TYPE(XEN_int_P(x), x, 2, "pango_layout_xy_to_index", "int");
  XEN_ASSERT_TYPE(XEN_int_P(y), y, 3, "pango_layout_xy_to_index", "int");
  XEN_ASSERT_TYPE(XEN_int__P(index), index, 4, "pango_layout_xy_to_index", "int*");
  XEN_ASSERT_TYPE(XEN_int__P(trailing), trailing, 5, "pango_layout_xy_to_index", "int*");
  return(C_TO_XEN_gboolean(pango_layout_xy_to_index(XEN_TO_C_PangoLayout_(layout), XEN_TO_C_int(x), XEN_TO_C_int(y), XEN_TO_C_int_(index), 
                                                    XEN_TO_C_int_(trailing))));
}

static XEN gxg_pango_layout_get_extents(XEN layout, XEN ink_rect, XEN logical_rect)
{
  #define H_pango_layout_get_extents "void pango_layout_get_extents(PangoLayout* layout, PangoRectangle* ink_rect, \
PangoRectangle* logical_rect)"
  XEN_ASSERT_TYPE(XEN_PangoLayout__P(layout), layout, 1, "pango_layout_get_extents", "PangoLayout*");
  XEN_ASSERT_TYPE(XEN_PangoRectangle__P(ink_rect), ink_rect, 2, "pango_layout_get_extents", "PangoRectangle*");
  XEN_ASSERT_TYPE(XEN_PangoRectangle__P(logical_rect), logical_rect, 3, "pango_layout_get_extents", "PangoRectangle*");
  pango_layout_get_extents(XEN_TO_C_PangoLayout_(layout), XEN_TO_C_PangoRectangle_(ink_rect), XEN_TO_C_PangoRectangle_(logical_rect));
  return(XEN_FALSE);
}

static XEN gxg_pango_layout_get_pixel_extents(XEN layout, XEN ink_rect, XEN logical_rect)
{
  #define H_pango_layout_get_pixel_extents "void pango_layout_get_pixel_extents(PangoLayout* layout, \
PangoRectangle* ink_rect, PangoRectangle* logical_rect)"
  XEN_ASSERT_TYPE(XEN_PangoLayout__P(layout), layout, 1, "pango_layout_get_pixel_extents", "PangoLayout*");
  XEN_ASSERT_TYPE(XEN_PangoRectangle__P(ink_rect), ink_rect, 2, "pango_layout_get_pixel_extents", "PangoRectangle*");
  XEN_ASSERT_TYPE(XEN_PangoRectangle__P(logical_rect), logical_rect, 3, "pango_layout_get_pixel_extents", "PangoRectangle*");
  pango_layout_get_pixel_extents(XEN_TO_C_PangoLayout_(layout), XEN_TO_C_PangoRectangle_(ink_rect), XEN_TO_C_PangoRectangle_(logical_rect));
  return(XEN_FALSE);
}

static XEN gxg_pango_layout_get_size(XEN layout, XEN width, XEN height)
{
  #define H_pango_layout_get_size "void pango_layout_get_size(PangoLayout* layout, int* width, int* height)"
  XEN_ASSERT_TYPE(XEN_PangoLayout__P(layout), layout, 1, "pango_layout_get_size", "PangoLayout*");
  XEN_ASSERT_TYPE(XEN_int__P(width), width, 2, "pango_layout_get_size", "int*");
  XEN_ASSERT_TYPE(XEN_int__P(height), height, 3, "pango_layout_get_size", "int*");
  pango_layout_get_size(XEN_TO_C_PangoLayout_(layout), XEN_TO_C_int_(width), XEN_TO_C_int_(height));
  return(XEN_FALSE);
}

static XEN gxg_pango_layout_get_pixel_size(XEN layout, XEN width, XEN height)
{
  #define H_pango_layout_get_pixel_size "void pango_layout_get_pixel_size(PangoLayout* layout, int* width, \
int* height)"
  XEN_ASSERT_TYPE(XEN_PangoLayout__P(layout), layout, 1, "pango_layout_get_pixel_size", "PangoLayout*");
  XEN_ASSERT_TYPE(XEN_int__P(width), width, 2, "pango_layout_get_pixel_size", "int*");
  XEN_ASSERT_TYPE(XEN_int__P(height), height, 3, "pango_layout_get_pixel_size", "int*");
  pango_layout_get_pixel_size(XEN_TO_C_PangoLayout_(layout), XEN_TO_C_int_(width), XEN_TO_C_int_(height));
  return(XEN_FALSE);
}

static XEN gxg_pango_layout_get_line_count(XEN layout)
{
  #define H_pango_layout_get_line_count "int pango_layout_get_line_count(PangoLayout* layout)"
  XEN_ASSERT_TYPE(XEN_PangoLayout__P(layout), layout, 1, "pango_layout_get_line_count", "PangoLayout*");
  return(C_TO_XEN_int(pango_layout_get_line_count(XEN_TO_C_PangoLayout_(layout))));
}

static XEN gxg_pango_layout_get_line(XEN layout, XEN line)
{
  #define H_pango_layout_get_line "PangoLayoutLine* pango_layout_get_line(PangoLayout* layout, int line)"
  XEN_ASSERT_TYPE(XEN_PangoLayout__P(layout), layout, 1, "pango_layout_get_line", "PangoLayout*");
  XEN_ASSERT_TYPE(XEN_int_P(line), line, 2, "pango_layout_get_line", "int");
  return(C_TO_XEN_PangoLayoutLine_(pango_layout_get_line(XEN_TO_C_PangoLayout_(layout), XEN_TO_C_int(line))));
}

static XEN gxg_pango_layout_get_lines(XEN layout)
{
  #define H_pango_layout_get_lines "GSList* pango_layout_get_lines(PangoLayout* layout)"
  XEN_ASSERT_TYPE(XEN_PangoLayout__P(layout), layout, 1, "pango_layout_get_lines", "PangoLayout*");
  return(C_TO_XEN_GSList_(pango_layout_get_lines(XEN_TO_C_PangoLayout_(layout))));
}

static XEN gxg_pango_layout_line_ref(XEN line)
{
  #define H_pango_layout_line_ref "void pango_layout_line_ref(PangoLayoutLine* line)"
  XEN_ASSERT_TYPE(XEN_PangoLayoutLine__P(line), line, 1, "pango_layout_line_ref", "PangoLayoutLine*");
  pango_layout_line_ref(XEN_TO_C_PangoLayoutLine_(line));
  return(XEN_FALSE);
}

static XEN gxg_pango_layout_line_unref(XEN line)
{
  #define H_pango_layout_line_unref "void pango_layout_line_unref(PangoLayoutLine* line)"
  XEN_ASSERT_TYPE(XEN_PangoLayoutLine__P(line), line, 1, "pango_layout_line_unref", "PangoLayoutLine*");
  pango_layout_line_unref(XEN_TO_C_PangoLayoutLine_(line));
  return(XEN_FALSE);
}

static XEN gxg_pango_layout_line_x_to_index(XEN line, XEN x_pos, XEN index, XEN trailing)
{
  #define H_pango_layout_line_x_to_index "gboolean pango_layout_line_x_to_index(PangoLayoutLine* line, \
int x_pos, int* index, int* trailing)"
  XEN_ASSERT_TYPE(XEN_PangoLayoutLine__P(line), line, 1, "pango_layout_line_x_to_index", "PangoLayoutLine*");
  XEN_ASSERT_TYPE(XEN_int_P(x_pos), x_pos, 2, "pango_layout_line_x_to_index", "int");
  XEN_ASSERT_TYPE(XEN_int__P(index), index, 3, "pango_layout_line_x_to_index", "int*");
  XEN_ASSERT_TYPE(XEN_int__P(trailing), trailing, 4, "pango_layout_line_x_to_index", "int*");
  return(C_TO_XEN_gboolean(pango_layout_line_x_to_index(XEN_TO_C_PangoLayoutLine_(line), XEN_TO_C_int(x_pos), XEN_TO_C_int_(index), 
                                                        XEN_TO_C_int_(trailing))));
}

static XEN gxg_pango_layout_line_index_to_x(XEN line, XEN index, XEN trailing, XEN x_pos)
{
  #define H_pango_layout_line_index_to_x "void pango_layout_line_index_to_x(PangoLayoutLine* line, int index, \
gboolean trailing, int* x_pos)"
  XEN_ASSERT_TYPE(XEN_PangoLayoutLine__P(line), line, 1, "pango_layout_line_index_to_x", "PangoLayoutLine*");
  XEN_ASSERT_TYPE(XEN_int_P(index), index, 2, "pango_layout_line_index_to_x", "int");
  XEN_ASSERT_TYPE(XEN_gboolean_P(trailing), trailing, 3, "pango_layout_line_index_to_x", "gboolean");
  XEN_ASSERT_TYPE(XEN_int__P(x_pos), x_pos, 4, "pango_layout_line_index_to_x", "int*");
  pango_layout_line_index_to_x(XEN_TO_C_PangoLayoutLine_(line), XEN_TO_C_int(index), XEN_TO_C_gboolean(trailing), XEN_TO_C_int_(x_pos));
  return(XEN_FALSE);
}

static XEN gxg_pango_layout_line_get_x_ranges(XEN line, XEN start_index, XEN end_index, XEN ranges, XEN n_ranges)
{
  #define H_pango_layout_line_get_x_ranges "void pango_layout_line_get_x_ranges(PangoLayoutLine* line, \
int start_index, int end_index, int** [ranges], int* [n_ranges])"
  int* ref_ranges;
  int ref_n_ranges;
  XEN_ASSERT_TYPE(XEN_PangoLayoutLine__P(line), line, 1, "pango_layout_line_get_x_ranges", "PangoLayoutLine*");
  XEN_ASSERT_TYPE(XEN_int_P(start_index), start_index, 2, "pango_layout_line_get_x_ranges", "int");
  XEN_ASSERT_TYPE(XEN_int_P(end_index), end_index, 3, "pango_layout_line_get_x_ranges", "int");
  pango_layout_line_get_x_ranges(XEN_TO_C_PangoLayoutLine_(line), XEN_TO_C_int(start_index), XEN_TO_C_int(end_index), &ref_ranges, 
                                 &ref_n_ranges);
  return(XEN_LIST_2(C_TO_XEN_int_(ref_ranges), C_TO_XEN_int(ref_n_ranges)));
}

static XEN gxg_pango_layout_line_get_extents(XEN line, XEN ink_rect, XEN logical_rect)
{
  #define H_pango_layout_line_get_extents "void pango_layout_line_get_extents(PangoLayoutLine* line, \
PangoRectangle* ink_rect, PangoRectangle* logical_rect)"
  XEN_ASSERT_TYPE(XEN_PangoLayoutLine__P(line), line, 1, "pango_layout_line_get_extents", "PangoLayoutLine*");
  XEN_ASSERT_TYPE(XEN_PangoRectangle__P(ink_rect), ink_rect, 2, "pango_layout_line_get_extents", "PangoRectangle*");
  XEN_ASSERT_TYPE(XEN_PangoRectangle__P(logical_rect), logical_rect, 3, "pango_layout_line_get_extents", "PangoRectangle*");
  pango_layout_line_get_extents(XEN_TO_C_PangoLayoutLine_(line), XEN_TO_C_PangoRectangle_(ink_rect), XEN_TO_C_PangoRectangle_(logical_rect));
  return(XEN_FALSE);
}

static XEN gxg_pango_layout_line_get_pixel_extents(XEN layout_line, XEN ink_rect, XEN logical_rect)
{
  #define H_pango_layout_line_get_pixel_extents "void pango_layout_line_get_pixel_extents(PangoLayoutLine* layout_line, \
PangoRectangle* ink_rect, PangoRectangle* logical_rect)"
  XEN_ASSERT_TYPE(XEN_PangoLayoutLine__P(layout_line), layout_line, 1, "pango_layout_line_get_pixel_extents", "PangoLayoutLine*");
  XEN_ASSERT_TYPE(XEN_PangoRectangle__P(ink_rect), ink_rect, 2, "pango_layout_line_get_pixel_extents", "PangoRectangle*");
  XEN_ASSERT_TYPE(XEN_PangoRectangle__P(logical_rect), logical_rect, 3, "pango_layout_line_get_pixel_extents", "PangoRectangle*");
  pango_layout_line_get_pixel_extents(XEN_TO_C_PangoLayoutLine_(layout_line), XEN_TO_C_PangoRectangle_(ink_rect), XEN_TO_C_PangoRectangle_(logical_rect));
  return(XEN_FALSE);
}

static XEN gxg_pango_layout_get_iter(XEN layout)
{
  #define H_pango_layout_get_iter "PangoLayoutIter* pango_layout_get_iter(PangoLayout* layout)"
  XEN_ASSERT_TYPE(XEN_PangoLayout__P(layout), layout, 1, "pango_layout_get_iter", "PangoLayout*");
  return(C_TO_XEN_PangoLayoutIter_(pango_layout_get_iter(XEN_TO_C_PangoLayout_(layout))));
}

static XEN gxg_pango_layout_iter_free(XEN iter)
{
  #define H_pango_layout_iter_free "void pango_layout_iter_free(PangoLayoutIter* iter)"
  XEN_ASSERT_TYPE(XEN_PangoLayoutIter__P(iter), iter, 1, "pango_layout_iter_free", "PangoLayoutIter*");
  pango_layout_iter_free(XEN_TO_C_PangoLayoutIter_(iter));
  return(XEN_FALSE);
}

static XEN gxg_pango_layout_iter_get_index(XEN iter)
{
  #define H_pango_layout_iter_get_index "int pango_layout_iter_get_index(PangoLayoutIter* iter)"
  XEN_ASSERT_TYPE(XEN_PangoLayoutIter__P(iter), iter, 1, "pango_layout_iter_get_index", "PangoLayoutIter*");
  return(C_TO_XEN_int(pango_layout_iter_get_index(XEN_TO_C_PangoLayoutIter_(iter))));
}

static XEN gxg_pango_layout_iter_get_run(XEN iter)
{
  #define H_pango_layout_iter_get_run "PangoLayoutRun* pango_layout_iter_get_run(PangoLayoutIter* iter)"
  XEN_ASSERT_TYPE(XEN_PangoLayoutIter__P(iter), iter, 1, "pango_layout_iter_get_run", "PangoLayoutIter*");
  return(C_TO_XEN_PangoLayoutRun_(pango_layout_iter_get_run(XEN_TO_C_PangoLayoutIter_(iter))));
}

static XEN gxg_pango_layout_iter_get_line(XEN iter)
{
  #define H_pango_layout_iter_get_line "PangoLayoutLine* pango_layout_iter_get_line(PangoLayoutIter* iter)"
  XEN_ASSERT_TYPE(XEN_PangoLayoutIter__P(iter), iter, 1, "pango_layout_iter_get_line", "PangoLayoutIter*");
  return(C_TO_XEN_PangoLayoutLine_(pango_layout_iter_get_line(XEN_TO_C_PangoLayoutIter_(iter))));
}

static XEN gxg_pango_layout_iter_at_last_line(XEN iter)
{
  #define H_pango_layout_iter_at_last_line "gboolean pango_layout_iter_at_last_line(PangoLayoutIter* iter)"
  XEN_ASSERT_TYPE(XEN_PangoLayoutIter__P(iter), iter, 1, "pango_layout_iter_at_last_line", "PangoLayoutIter*");
  return(C_TO_XEN_gboolean(pango_layout_iter_at_last_line(XEN_TO_C_PangoLayoutIter_(iter))));
}

static XEN gxg_pango_layout_iter_next_char(XEN iter)
{
  #define H_pango_layout_iter_next_char "gboolean pango_layout_iter_next_char(PangoLayoutIter* iter)"
  XEN_ASSERT_TYPE(XEN_PangoLayoutIter__P(iter), iter, 1, "pango_layout_iter_next_char", "PangoLayoutIter*");
  return(C_TO_XEN_gboolean(pango_layout_iter_next_char(XEN_TO_C_PangoLayoutIter_(iter))));
}

static XEN gxg_pango_layout_iter_next_cluster(XEN iter)
{
  #define H_pango_layout_iter_next_cluster "gboolean pango_layout_iter_next_cluster(PangoLayoutIter* iter)"
  XEN_ASSERT_TYPE(XEN_PangoLayoutIter__P(iter), iter, 1, "pango_layout_iter_next_cluster", "PangoLayoutIter*");
  return(C_TO_XEN_gboolean(pango_layout_iter_next_cluster(XEN_TO_C_PangoLayoutIter_(iter))));
}

static XEN gxg_pango_layout_iter_next_run(XEN iter)
{
  #define H_pango_layout_iter_next_run "gboolean pango_layout_iter_next_run(PangoLayoutIter* iter)"
  XEN_ASSERT_TYPE(XEN_PangoLayoutIter__P(iter), iter, 1, "pango_layout_iter_next_run", "PangoLayoutIter*");
  return(C_TO_XEN_gboolean(pango_layout_iter_next_run(XEN_TO_C_PangoLayoutIter_(iter))));
}

static XEN gxg_pango_layout_iter_next_line(XEN iter)
{
  #define H_pango_layout_iter_next_line "gboolean pango_layout_iter_next_line(PangoLayoutIter* iter)"
  XEN_ASSERT_TYPE(XEN_PangoLayoutIter__P(iter), iter, 1, "pango_layout_iter_next_line", "PangoLayoutIter*");
  return(C_TO_XEN_gboolean(pango_layout_iter_next_line(XEN_TO_C_PangoLayoutIter_(iter))));
}

static XEN gxg_pango_layout_iter_get_char_extents(XEN iter, XEN logical_rect)
{
  #define H_pango_layout_iter_get_char_extents "void pango_layout_iter_get_char_extents(PangoLayoutIter* iter, \
PangoRectangle* logical_rect)"
  XEN_ASSERT_TYPE(XEN_PangoLayoutIter__P(iter), iter, 1, "pango_layout_iter_get_char_extents", "PangoLayoutIter*");
  XEN_ASSERT_TYPE(XEN_PangoRectangle__P(logical_rect), logical_rect, 2, "pango_layout_iter_get_char_extents", "PangoRectangle*");
  pango_layout_iter_get_char_extents(XEN_TO_C_PangoLayoutIter_(iter), XEN_TO_C_PangoRectangle_(logical_rect));
  return(XEN_FALSE);
}

static XEN gxg_pango_layout_iter_get_cluster_extents(XEN iter, XEN ink_rect, XEN logical_rect)
{
  #define H_pango_layout_iter_get_cluster_extents "void pango_layout_iter_get_cluster_extents(PangoLayoutIter* iter, \
PangoRectangle* ink_rect, PangoRectangle* logical_rect)"
  XEN_ASSERT_TYPE(XEN_PangoLayoutIter__P(iter), iter, 1, "pango_layout_iter_get_cluster_extents", "PangoLayoutIter*");
  XEN_ASSERT_TYPE(XEN_PangoRectangle__P(ink_rect), ink_rect, 2, "pango_layout_iter_get_cluster_extents", "PangoRectangle*");
  XEN_ASSERT_TYPE(XEN_PangoRectangle__P(logical_rect), logical_rect, 3, "pango_layout_iter_get_cluster_extents", "PangoRectangle*");
  pango_layout_iter_get_cluster_extents(XEN_TO_C_PangoLayoutIter_(iter), XEN_TO_C_PangoRectangle_(ink_rect), XEN_TO_C_PangoRectangle_(logical_rect));
  return(XEN_FALSE);
}

static XEN gxg_pango_layout_iter_get_run_extents(XEN iter, XEN ink_rect, XEN logical_rect)
{
  #define H_pango_layout_iter_get_run_extents "void pango_layout_iter_get_run_extents(PangoLayoutIter* iter, \
PangoRectangle* ink_rect, PangoRectangle* logical_rect)"
  XEN_ASSERT_TYPE(XEN_PangoLayoutIter__P(iter), iter, 1, "pango_layout_iter_get_run_extents", "PangoLayoutIter*");
  XEN_ASSERT_TYPE(XEN_PangoRectangle__P(ink_rect), ink_rect, 2, "pango_layout_iter_get_run_extents", "PangoRectangle*");
  XEN_ASSERT_TYPE(XEN_PangoRectangle__P(logical_rect), logical_rect, 3, "pango_layout_iter_get_run_extents", "PangoRectangle*");
  pango_layout_iter_get_run_extents(XEN_TO_C_PangoLayoutIter_(iter), XEN_TO_C_PangoRectangle_(ink_rect), XEN_TO_C_PangoRectangle_(logical_rect));
  return(XEN_FALSE);
}

static XEN gxg_pango_layout_iter_get_line_extents(XEN iter, XEN ink_rect, XEN logical_rect)
{
  #define H_pango_layout_iter_get_line_extents "void pango_layout_iter_get_line_extents(PangoLayoutIter* iter, \
PangoRectangle* ink_rect, PangoRectangle* logical_rect)"
  XEN_ASSERT_TYPE(XEN_PangoLayoutIter__P(iter), iter, 1, "pango_layout_iter_get_line_extents", "PangoLayoutIter*");
  XEN_ASSERT_TYPE(XEN_PangoRectangle__P(ink_rect), ink_rect, 2, "pango_layout_iter_get_line_extents", "PangoRectangle*");
  XEN_ASSERT_TYPE(XEN_PangoRectangle__P(logical_rect), logical_rect, 3, "pango_layout_iter_get_line_extents", "PangoRectangle*");
  pango_layout_iter_get_line_extents(XEN_TO_C_PangoLayoutIter_(iter), XEN_TO_C_PangoRectangle_(ink_rect), XEN_TO_C_PangoRectangle_(logical_rect));
  return(XEN_FALSE);
}

static XEN gxg_pango_layout_iter_get_line_yrange(XEN iter, XEN y0, XEN y1)
{
  #define H_pango_layout_iter_get_line_yrange "void pango_layout_iter_get_line_yrange(PangoLayoutIter* iter, \
int* y0, int* y1)"
  XEN_ASSERT_TYPE(XEN_PangoLayoutIter__P(iter), iter, 1, "pango_layout_iter_get_line_yrange", "PangoLayoutIter*");
  XEN_ASSERT_TYPE(XEN_int__P(y0), y0, 2, "pango_layout_iter_get_line_yrange", "int*");
  XEN_ASSERT_TYPE(XEN_int__P(y1), y1, 3, "pango_layout_iter_get_line_yrange", "int*");
  pango_layout_iter_get_line_yrange(XEN_TO_C_PangoLayoutIter_(iter), XEN_TO_C_int_(y0), XEN_TO_C_int_(y1));
  return(XEN_FALSE);
}

static XEN gxg_pango_layout_iter_get_layout_extents(XEN iter, XEN ink_rect, XEN logical_rect)
{
  #define H_pango_layout_iter_get_layout_extents "void pango_layout_iter_get_layout_extents(PangoLayoutIter* iter, \
PangoRectangle* ink_rect, PangoRectangle* logical_rect)"
  XEN_ASSERT_TYPE(XEN_PangoLayoutIter__P(iter), iter, 1, "pango_layout_iter_get_layout_extents", "PangoLayoutIter*");
  XEN_ASSERT_TYPE(XEN_PangoRectangle__P(ink_rect), ink_rect, 2, "pango_layout_iter_get_layout_extents", "PangoRectangle*");
  XEN_ASSERT_TYPE(XEN_PangoRectangle__P(logical_rect), logical_rect, 3, "pango_layout_iter_get_layout_extents", "PangoRectangle*");
  pango_layout_iter_get_layout_extents(XEN_TO_C_PangoLayoutIter_(iter), XEN_TO_C_PangoRectangle_(ink_rect), XEN_TO_C_PangoRectangle_(logical_rect));
  return(XEN_FALSE);
}

static XEN gxg_pango_layout_iter_get_baseline(XEN iter)
{
  #define H_pango_layout_iter_get_baseline "int pango_layout_iter_get_baseline(PangoLayoutIter* iter)"
  XEN_ASSERT_TYPE(XEN_PangoLayoutIter__P(iter), iter, 1, "pango_layout_iter_get_baseline", "PangoLayoutIter*");
  return(C_TO_XEN_int(pango_layout_iter_get_baseline(XEN_TO_C_PangoLayoutIter_(iter))));
}

static XEN gxg_pango_language_get_type(void)
{
  #define H_pango_language_get_type "GType pango_language_get_type( void)"
  return(C_TO_XEN_GType(pango_language_get_type()));
}

static XEN gxg_pango_language_from_string(XEN language)
{
  #define H_pango_language_from_string "PangoLanguage* pango_language_from_string(char* language)"
  XEN_ASSERT_TYPE(XEN_char__P(language), language, 1, "pango_language_from_string", "char*");
  return(C_TO_XEN_PangoLanguage_(pango_language_from_string(XEN_TO_C_char_(language))));
}

static XEN gxg_pango_language_matches(XEN language, XEN range_list)
{
  #define H_pango_language_matches "gboolean pango_language_matches(PangoLanguage* language, char* range_list)"
  XEN_ASSERT_TYPE(XEN_PangoLanguage__P(language), language, 1, "pango_language_matches", "PangoLanguage*");
  XEN_ASSERT_TYPE(XEN_char__P(range_list), range_list, 2, "pango_language_matches", "char*");
  return(C_TO_XEN_gboolean(pango_language_matches(XEN_TO_C_PangoLanguage_(language), XEN_TO_C_char_(range_list))));
}
#endif
#if PANGO_ENABLE_ENGINE && PANGO_ENABLE_BACKEND

#if (!HAVE_GTK_1)
static XEN gxg_pango_default_break(XEN text, XEN length, XEN analysis, XEN attrs, XEN attrs_len)
{
  #define H_pango_default_break "void pango_default_break(gchar* text, int length, PangoAnalysis* analysis, \
PangoLogAttr* attrs, int attrs_len)"
  XEN_ASSERT_TYPE(XEN_gchar__P(text), text, 1, "pango_default_break", "gchar*");
  XEN_ASSERT_TYPE(XEN_int_P(length), length, 2, "pango_default_break", "int");
  XEN_ASSERT_TYPE(XEN_PangoAnalysis__P(analysis), analysis, 3, "pango_default_break", "PangoAnalysis*");
  XEN_ASSERT_TYPE(XEN_PangoLogAttr__P(attrs), attrs, 4, "pango_default_break", "PangoLogAttr*");
  XEN_ASSERT_TYPE(XEN_int_P(attrs_len), attrs_len, 5, "pango_default_break", "int");
  pango_default_break(XEN_TO_C_gchar_(text), XEN_TO_C_int(length), XEN_TO_C_PangoAnalysis_(analysis), XEN_TO_C_PangoLogAttr_(attrs), 
                      XEN_TO_C_int(attrs_len));
  return(XEN_FALSE);
}

static XEN gxg_pango_context_new(void)
{
  #define H_pango_context_new "PangoContext* pango_context_new( void)"
  return(C_TO_XEN_PangoContext_(pango_context_new()));
}

static XEN gxg_script_engine_list(XEN engines, XEN n_engines)
{
  #define H_script_engine_list "void script_engine_list(PangoEngineInfo** [engines], int* [n_engines])"
  PangoEngineInfo* ref_engines;
  int ref_n_engines;
  script_engine_list(&ref_engines, &ref_n_engines);
  return(XEN_LIST_2(C_TO_XEN_PangoEngineInfo_(ref_engines), C_TO_XEN_int(ref_n_engines)));
}

static XEN gxg_script_engine_load(XEN id)
{
  #define H_script_engine_load "PangoEngine* script_engine_load(char* id)"
  XEN_ASSERT_TYPE(XEN_char__P(id), id, 1, "script_engine_load", "char*");
  return(C_TO_XEN_PangoEngine_(script_engine_load(XEN_TO_C_char_(id))));
}

static XEN gxg_script_engine_unload(XEN engine)
{
  #define H_script_engine_unload "void script_engine_unload(PangoEngine* engine)"
  XEN_ASSERT_TYPE(XEN_PangoEngine__P(engine), engine, 1, "script_engine_unload", "PangoEngine*");
  script_engine_unload(XEN_TO_C_PangoEngine_(engine));
  return(XEN_FALSE);
}

static XEN gxg_pango_font_metrics_new(void)
{
  #define H_pango_font_metrics_new "PangoFontMetrics* pango_font_metrics_new( void)"
  return(C_TO_XEN_PangoFontMetrics_(pango_font_metrics_new()));
}

static XEN gxg_pango_font_family_get_type(void)
{
  #define H_pango_font_family_get_type "GType pango_font_family_get_type( void)"
  return(C_TO_XEN_GType(pango_font_family_get_type()));
}

static XEN gxg_pango_font_family_list_faces(XEN family, XEN faces, XEN n_faces)
{
  #define H_pango_font_family_list_faces "void pango_font_family_list_faces(PangoFontFamily* family, \
PangoFontFace*** [faces], int* [n_faces])"
  PangoFontFace** ref_faces;
  int ref_n_faces;
  XEN_ASSERT_TYPE(XEN_PangoFontFamily__P(family), family, 1, "pango_font_family_list_faces", "PangoFontFamily*");
  pango_font_family_list_faces(XEN_TO_C_PangoFontFamily_(family), &ref_faces, &ref_n_faces);
  return(XEN_LIST_2(C_TO_XEN_PangoFontFace__(ref_faces), C_TO_XEN_int(ref_n_faces)));
}

static XEN gxg_pango_font_family_get_name(XEN family)
{
  #define H_pango_font_family_get_name "char* pango_font_family_get_name(PangoFontFamily* family)"
  XEN_ASSERT_TYPE(XEN_PangoFontFamily__P(family), family, 1, "pango_font_family_get_name", "PangoFontFamily*");
  return(C_TO_XEN_char_(pango_font_family_get_name(XEN_TO_C_PangoFontFamily_(family))));
}

static XEN gxg_pango_context_set_font_map(XEN context, XEN font_map)
{
  #define H_pango_context_set_font_map "void pango_context_set_font_map(PangoContext* context, PangoFontMap* font_map)"
  XEN_ASSERT_TYPE(XEN_PangoContext__P(context), context, 1, "pango_context_set_font_map", "PangoContext*");
  XEN_ASSERT_TYPE(XEN_PangoFontMap__P(font_map), font_map, 2, "pango_context_set_font_map", "PangoFontMap*");
  pango_context_set_font_map(XEN_TO_C_PangoContext_(context), XEN_TO_C_PangoFontMap_(font_map));
  return(XEN_FALSE);
}
#endif
#endif

#if (!(defined(GDK_DISABLE_DEPRECATED))) && (!(defined(GTK_DISABLE_DEPRECATED))) && (!(defined(GDK_PIXBUF_DISABLE_DEPRECATED)))

static XEN gxg_gdk_window_get_deskrelative_origin(XEN window, XEN x, XEN y)
{
  #define H_gdk_window_get_deskrelative_origin "gboolean gdk_window_get_deskrelative_origin(GdkWindow* window, \
gint* [x], gint* [y])"
  gint ref_x;
  gint ref_y;
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 1, "gdk_window_get_deskrelative_origin", "GdkWindow*");
  {
    XEN result = XEN_FALSE;
    result = C_TO_XEN_gboolean(gdk_window_get_deskrelative_origin(XEN_TO_C_GdkWindow_(window), &ref_x, &ref_y));
    return(XEN_LIST_3(result, C_TO_XEN_gint(ref_x), C_TO_XEN_gint(ref_y)));
   }
}

#if (!HAVE_GTK_1)
static XEN gxg_gdk_pixbuf_ref(XEN pixbuf)
{
  #define H_gdk_pixbuf_ref "GdkPixbuf* gdk_pixbuf_ref(GdkPixbuf* pixbuf)"
  XEN_ASSERT_TYPE(XEN_GdkPixbuf__P(pixbuf), pixbuf, 1, "gdk_pixbuf_ref", "GdkPixbuf*");
  return(C_TO_XEN_GdkPixbuf_(gdk_pixbuf_ref(XEN_TO_C_GdkPixbuf_(pixbuf))));
}

static XEN gxg_gdk_pixbuf_unref(XEN pixbuf)
{
  #define H_gdk_pixbuf_unref "void gdk_pixbuf_unref(GdkPixbuf* pixbuf)"
  XEN_ASSERT_TYPE(XEN_GdkPixbuf__P(pixbuf), pixbuf, 1, "gdk_pixbuf_unref", "GdkPixbuf*");
  gdk_pixbuf_unref(XEN_TO_C_GdkPixbuf_(pixbuf));
  return(XEN_FALSE);
}

static XEN gxg_gdk_pixbuf_animation_ref(XEN animation)
{
  #define H_gdk_pixbuf_animation_ref "GdkPixbufAnimation* gdk_pixbuf_animation_ref(GdkPixbufAnimation* animation)"
  XEN_ASSERT_TYPE(XEN_GdkPixbufAnimation__P(animation), animation, 1, "gdk_pixbuf_animation_ref", "GdkPixbufAnimation*");
  return(C_TO_XEN_GdkPixbufAnimation_(gdk_pixbuf_animation_ref(XEN_TO_C_GdkPixbufAnimation_(animation))));
}

static XEN gxg_gdk_pixbuf_animation_unref(XEN animation)
{
  #define H_gdk_pixbuf_animation_unref "void gdk_pixbuf_animation_unref(GdkPixbufAnimation* animation)"
  XEN_ASSERT_TYPE(XEN_GdkPixbufAnimation__P(animation), animation, 1, "gdk_pixbuf_animation_unref", "GdkPixbufAnimation*");
  gdk_pixbuf_animation_unref(XEN_TO_C_GdkPixbufAnimation_(animation));
  return(XEN_FALSE);
}
#endif

static XEN gxg_gdk_colors_alloc(XEN colormap, XEN contiguous, XEN planes, XEN nplanes, XEN pixels, XEN npixels)
{
  #define H_gdk_colors_alloc "gint gdk_colors_alloc(GdkColormap* colormap, gboolean contiguous, gulong* planes, \
gint nplanes, gulong* pixels, gint npixels)"
  XEN_ASSERT_TYPE(XEN_GdkColormap__P(colormap), colormap, 1, "gdk_colors_alloc", "GdkColormap*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(contiguous), contiguous, 2, "gdk_colors_alloc", "gboolean");
  XEN_ASSERT_TYPE(XEN_gulong__P(planes), planes, 3, "gdk_colors_alloc", "gulong*");
  XEN_ASSERT_TYPE(XEN_gint_P(nplanes), nplanes, 4, "gdk_colors_alloc", "gint");
  XEN_ASSERT_TYPE(XEN_gulong__P(pixels), pixels, 5, "gdk_colors_alloc", "gulong*");
  XEN_ASSERT_TYPE(XEN_gint_P(npixels), npixels, 6, "gdk_colors_alloc", "gint");
  return(C_TO_XEN_gint(gdk_colors_alloc(XEN_TO_C_GdkColormap_(colormap), XEN_TO_C_gboolean(contiguous), XEN_TO_C_gulong_(planes), 
                                        XEN_TO_C_gint(nplanes), XEN_TO_C_gulong_(pixels), XEN_TO_C_gint(npixels))));
}

static XEN gxg_gdk_colors_free(XEN colormap, XEN pixels, XEN npixels, XEN planes)
{
  #define H_gdk_colors_free "void gdk_colors_free(GdkColormap* colormap, gulong* pixels, gint npixels, \
gulong planes)"
  XEN_ASSERT_TYPE(XEN_GdkColormap__P(colormap), colormap, 1, "gdk_colors_free", "GdkColormap*");
  XEN_ASSERT_TYPE(XEN_gulong__P(pixels), pixels, 2, "gdk_colors_free", "gulong*");
  XEN_ASSERT_TYPE(XEN_gint_P(npixels), npixels, 3, "gdk_colors_free", "gint");
  XEN_ASSERT_TYPE(XEN_gulong_P(planes), planes, 4, "gdk_colors_free", "gulong");
  gdk_colors_free(XEN_TO_C_GdkColormap_(colormap), XEN_TO_C_gulong_(pixels), XEN_TO_C_gint(npixels), XEN_TO_C_gulong(planes));
  return(XEN_FALSE);
}

static XEN gxg_gdk_color_white(XEN colormap, XEN color)
{
  #define H_gdk_color_white "gint gdk_color_white(GdkColormap* colormap, GdkColor* color)"
  XEN_ASSERT_TYPE(XEN_GdkColormap__P(colormap), colormap, 1, "gdk_color_white", "GdkColormap*");
  XEN_ASSERT_TYPE(XEN_GdkColor__P(color), color, 2, "gdk_color_white", "GdkColor*");
  return(C_TO_XEN_gint(gdk_color_white(XEN_TO_C_GdkColormap_(colormap), XEN_TO_C_GdkColor_(color))));
}

static XEN gxg_gdk_color_black(XEN colormap, XEN color)
{
  #define H_gdk_color_black "gint gdk_color_black(GdkColormap* colormap, GdkColor* color)"
  XEN_ASSERT_TYPE(XEN_GdkColormap__P(colormap), colormap, 1, "gdk_color_black", "GdkColormap*");
  XEN_ASSERT_TYPE(XEN_GdkColor__P(color), color, 2, "gdk_color_black", "GdkColor*");
  return(C_TO_XEN_gint(gdk_color_black(XEN_TO_C_GdkColormap_(colormap), XEN_TO_C_GdkColor_(color))));
}

static XEN gxg_gdk_color_alloc(XEN colormap, XEN color)
{
  #define H_gdk_color_alloc "gint gdk_color_alloc(GdkColormap* colormap, GdkColor* color)"
  XEN_ASSERT_TYPE(XEN_GdkColormap__P(colormap), colormap, 1, "gdk_color_alloc", "GdkColormap*");
  XEN_ASSERT_TYPE(XEN_GdkColor__P(color), color, 2, "gdk_color_alloc", "GdkColor*");
  return(C_TO_XEN_gint(gdk_color_alloc(XEN_TO_C_GdkColormap_(colormap), XEN_TO_C_GdkColor_(color))));
}

static XEN gxg_gdk_color_change(XEN colormap, XEN color)
{
  #define H_gdk_color_change "gint gdk_color_change(GdkColormap* colormap, GdkColor* color)"
  XEN_ASSERT_TYPE(XEN_GdkColormap__P(colormap), colormap, 1, "gdk_color_change", "GdkColormap*");
  XEN_ASSERT_TYPE(XEN_GdkColor__P(color), color, 2, "gdk_color_change", "GdkColor*");
  return(C_TO_XEN_gint(gdk_color_change(XEN_TO_C_GdkColormap_(colormap), XEN_TO_C_GdkColor_(color))));
}

static XEN gxg_gdk_drag_context_ref(XEN context)
{
  #define H_gdk_drag_context_ref "void gdk_drag_context_ref(GdkDragContext* context)"
  XEN_ASSERT_TYPE(XEN_GdkDragContext__P(context), context, 1, "gdk_drag_context_ref", "GdkDragContext*");
  gdk_drag_context_ref(XEN_TO_C_GdkDragContext_(context));
  return(XEN_FALSE);
}

static XEN gxg_gdk_drag_context_unref(XEN context)
{
  #define H_gdk_drag_context_unref "void gdk_drag_context_unref(GdkDragContext* context)"
  XEN_ASSERT_TYPE(XEN_GdkDragContext__P(context), context, 1, "gdk_drag_context_unref", "GdkDragContext*");
  gdk_drag_context_unref(XEN_TO_C_GdkDragContext_(context));
  return(XEN_FALSE);
}

static XEN gxg_gdk_drawable_set_data(XEN drawable, XEN key, XEN data, XEN destroy_func)
{
  #define H_gdk_drawable_set_data "void gdk_drawable_set_data(GdkDrawable* drawable, gchar* key, gpointer data, \
GDestroyNotify destroy_func)"
  XEN_ASSERT_TYPE(XEN_GdkDrawable__P(drawable), drawable, 1, "gdk_drawable_set_data", "GdkDrawable*");
  XEN_ASSERT_TYPE(XEN_gchar__P(key), key, 2, "gdk_drawable_set_data", "gchar*");
  XEN_ASSERT_TYPE(XEN_gpointer_P(data), data, 3, "gdk_drawable_set_data", "gpointer");
  XEN_ASSERT_TYPE(XEN_GDestroyNotify_P(destroy_func), destroy_func, 4, "gdk_drawable_set_data", "GDestroyNotify");
  gdk_drawable_set_data(XEN_TO_C_GdkDrawable_(drawable), XEN_TO_C_gchar_(key), XEN_TO_C_gpointer(data), XEN_TO_C_GDestroyNotify(destroy_func));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gdk_drawable_get_data(XEN drawable, XEN key)
{
  #define H_gdk_drawable_get_data "gpointer gdk_drawable_get_data(GdkDrawable* drawable, gchar* key)"
  XEN_ASSERT_TYPE(XEN_GdkDrawable__P(drawable), drawable, 1, "gdk_drawable_get_data", "GdkDrawable*");
  XEN_ASSERT_TYPE(XEN_gchar__P(key), key, 2, "gdk_drawable_get_data", "gchar*");
  return(C_TO_XEN_gpointer(gdk_drawable_get_data(XEN_TO_C_GdkDrawable_(drawable), XEN_TO_C_gchar_(key))));
}

static XEN gxg_gdk_drawable_ref(XEN drawable)
{
  #define H_gdk_drawable_ref "GdkDrawable* gdk_drawable_ref(GdkDrawable* drawable)"
  XEN_ASSERT_TYPE(XEN_GdkDrawable__P(drawable), drawable, 1, "gdk_drawable_ref", "GdkDrawable*");
  return(C_TO_XEN_GdkDrawable_(gdk_drawable_ref(XEN_TO_C_GdkDrawable_(drawable))));
}

static XEN gxg_gdk_drawable_unref(XEN drawable)
{
  #define H_gdk_drawable_unref "void gdk_drawable_unref(GdkDrawable* drawable)"
  XEN_ASSERT_TYPE(XEN_GdkDrawable__P(drawable), drawable, 1, "gdk_drawable_unref", "GdkDrawable*");
  gdk_drawable_unref(XEN_TO_C_GdkDrawable_(drawable));
  return(XEN_FALSE);
}
#endif

static XEN gxg_gdk_draw_string(XEN drawable, XEN font, XEN gc, XEN x, XEN y, XEN string)
{
  #define H_gdk_draw_string "void gdk_draw_string(GdkDrawable* drawable, GdkFont* font, GdkGC* gc, gint x, \
gint y, gchar* string)"
  XEN_ASSERT_TYPE(XEN_GdkDrawable__P(drawable), drawable, 1, "gdk_draw_string", "GdkDrawable*");
  XEN_ASSERT_TYPE(XEN_GdkFont__P(font), font, 2, "gdk_draw_string", "GdkFont*");
  XEN_ASSERT_TYPE(XEN_GdkGC__P(gc), gc, 3, "gdk_draw_string", "GdkGC*");
  XEN_ASSERT_TYPE(XEN_gint_P(x), x, 4, "gdk_draw_string", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(y), y, 5, "gdk_draw_string", "gint");
  XEN_ASSERT_TYPE(XEN_gchar__P(string), string, 6, "gdk_draw_string", "gchar*");
  gdk_draw_string(XEN_TO_C_GdkDrawable_(drawable), XEN_TO_C_GdkFont_(font), XEN_TO_C_GdkGC_(gc), XEN_TO_C_gint(x), XEN_TO_C_gint(y), 
                  XEN_TO_C_gchar_(string));
  return(XEN_FALSE);
}

static XEN gxg_gdk_draw_text(XEN drawable, XEN font, XEN gc, XEN x, XEN y, XEN text, XEN text_length)
{
  #define H_gdk_draw_text "void gdk_draw_text(GdkDrawable* drawable, GdkFont* font, GdkGC* gc, gint x, \
gint y, gchar* text, gint text_length)"
  XEN_ASSERT_TYPE(XEN_GdkDrawable__P(drawable), drawable, 1, "gdk_draw_text", "GdkDrawable*");
  XEN_ASSERT_TYPE(XEN_GdkFont__P(font), font, 2, "gdk_draw_text", "GdkFont*");
  XEN_ASSERT_TYPE(XEN_GdkGC__P(gc), gc, 3, "gdk_draw_text", "GdkGC*");
  XEN_ASSERT_TYPE(XEN_gint_P(x), x, 4, "gdk_draw_text", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(y), y, 5, "gdk_draw_text", "gint");
  XEN_ASSERT_TYPE(XEN_gchar__P(text), text, 6, "gdk_draw_text", "gchar*");
  XEN_ASSERT_TYPE(XEN_gint_P(text_length), text_length, 7, "gdk_draw_text", "gint");
  gdk_draw_text(XEN_TO_C_GdkDrawable_(drawable), XEN_TO_C_GdkFont_(font), XEN_TO_C_GdkGC_(gc), XEN_TO_C_gint(x), XEN_TO_C_gint(y), 
                XEN_TO_C_gchar_(text), XEN_TO_C_gint(text_length));
  return(XEN_FALSE);
}

static XEN gxg_gdk_draw_text_wc(XEN drawable, XEN font, XEN gc, XEN x, XEN y, XEN text, XEN text_length)
{
  #define H_gdk_draw_text_wc "void gdk_draw_text_wc(GdkDrawable* drawable, GdkFont* font, GdkGC* gc, \
gint x, gint y, GdkWChar* text, gint text_length)"
  XEN_ASSERT_TYPE(XEN_GdkDrawable__P(drawable), drawable, 1, "gdk_draw_text_wc", "GdkDrawable*");
  XEN_ASSERT_TYPE(XEN_GdkFont__P(font), font, 2, "gdk_draw_text_wc", "GdkFont*");
  XEN_ASSERT_TYPE(XEN_GdkGC__P(gc), gc, 3, "gdk_draw_text_wc", "GdkGC*");
  XEN_ASSERT_TYPE(XEN_gint_P(x), x, 4, "gdk_draw_text_wc", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(y), y, 5, "gdk_draw_text_wc", "gint");
  XEN_ASSERT_TYPE(XEN_GdkWChar__P(text), text, 6, "gdk_draw_text_wc", "GdkWChar*");
  XEN_ASSERT_TYPE(XEN_gint_P(text_length), text_length, 7, "gdk_draw_text_wc", "gint");
  gdk_draw_text_wc(XEN_TO_C_GdkDrawable_(drawable), XEN_TO_C_GdkFont_(font), XEN_TO_C_GdkGC_(gc), XEN_TO_C_gint(x), XEN_TO_C_gint(y), 
                   XEN_TO_C_GdkWChar_(text), XEN_TO_C_gint(text_length));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gdk_font_get_type(void)
{
  #define H_gdk_font_get_type "GType gdk_font_get_type( void)"
  return(C_TO_XEN_GType(gdk_font_get_type()));
}
#endif

static XEN gxg_gdk_font_load(XEN font_name)
{
  #define H_gdk_font_load "GdkFont* gdk_font_load(gchar* font_name)"
  XEN_ASSERT_TYPE(XEN_gchar__P(font_name), font_name, 1, "gdk_font_load", "gchar*");
  return(C_TO_XEN_GdkFont_(gdk_font_load(XEN_TO_C_gchar_(font_name))));
}

static XEN gxg_gdk_fontset_load(XEN fontset_name)
{
  #define H_gdk_fontset_load "GdkFont* gdk_fontset_load(gchar* fontset_name)"
  XEN_ASSERT_TYPE(XEN_gchar__P(fontset_name), fontset_name, 1, "gdk_fontset_load", "gchar*");
  return(C_TO_XEN_GdkFont_(gdk_fontset_load(XEN_TO_C_gchar_(fontset_name))));
}

#if (!HAVE_GTK_1)
static XEN gxg_gdk_font_from_description(XEN font_desc)
{
  #define H_gdk_font_from_description "GdkFont* gdk_font_from_description(PangoFontDescription* font_desc)"
  XEN_ASSERT_TYPE(XEN_PangoFontDescription__P(font_desc), font_desc, 1, "gdk_font_from_description", "PangoFontDescription*");
  return(C_TO_XEN_GdkFont_(gdk_font_from_description(XEN_TO_C_PangoFontDescription_(font_desc))));
}
#endif

static XEN gxg_gdk_font_ref(XEN font)
{
  #define H_gdk_font_ref "GdkFont* gdk_font_ref(GdkFont* font)"
  XEN_ASSERT_TYPE(XEN_GdkFont__P(font), font, 1, "gdk_font_ref", "GdkFont*");
  return(C_TO_XEN_GdkFont_(gdk_font_ref(XEN_TO_C_GdkFont_(font))));
}

static XEN gxg_gdk_font_unref(XEN font)
{
  #define H_gdk_font_unref "void gdk_font_unref(GdkFont* font)"
  XEN_ASSERT_TYPE(XEN_GdkFont__P(font), font, 1, "gdk_font_unref", "GdkFont*");
  gdk_font_unref(XEN_TO_C_GdkFont_(font));
  return(XEN_FALSE);
}

static XEN gxg_gdk_font_id(XEN font)
{
  #define H_gdk_font_id "gint gdk_font_id(GdkFont* font)"
  XEN_ASSERT_TYPE(XEN_GdkFont__P(font), font, 1, "gdk_font_id", "GdkFont*");
  return(C_TO_XEN_gint(gdk_font_id(XEN_TO_C_GdkFont_(font))));
}

static XEN gxg_gdk_font_equal(XEN fonta, XEN fontb)
{
  #define H_gdk_font_equal "gboolean gdk_font_equal(GdkFont* fonta, GdkFont* fontb)"
  XEN_ASSERT_TYPE(XEN_GdkFont__P(fonta), fonta, 1, "gdk_font_equal", "GdkFont*");
  XEN_ASSERT_TYPE(XEN_GdkFont__P(fontb), fontb, 2, "gdk_font_equal", "GdkFont*");
  return(C_TO_XEN_gboolean(gdk_font_equal(XEN_TO_C_GdkFont_(fonta), XEN_TO_C_GdkFont_(fontb))));
}

static XEN gxg_gdk_string_width(XEN font, XEN string)
{
  #define H_gdk_string_width "gint gdk_string_width(GdkFont* font, gchar* string)"
  XEN_ASSERT_TYPE(XEN_GdkFont__P(font), font, 1, "gdk_string_width", "GdkFont*");
  XEN_ASSERT_TYPE(XEN_gchar__P(string), string, 2, "gdk_string_width", "gchar*");
  return(C_TO_XEN_gint(gdk_string_width(XEN_TO_C_GdkFont_(font), XEN_TO_C_gchar_(string))));
}

static XEN gxg_gdk_text_width(XEN font, XEN text, XEN text_length)
{
  #define H_gdk_text_width "gint gdk_text_width(GdkFont* font, gchar* text, gint text_length)"
  XEN_ASSERT_TYPE(XEN_GdkFont__P(font), font, 1, "gdk_text_width", "GdkFont*");
  XEN_ASSERT_TYPE(XEN_gchar__P(text), text, 2, "gdk_text_width", "gchar*");
  XEN_ASSERT_TYPE(XEN_gint_P(text_length), text_length, 3, "gdk_text_width", "gint");
  return(C_TO_XEN_gint(gdk_text_width(XEN_TO_C_GdkFont_(font), XEN_TO_C_gchar_(text), XEN_TO_C_gint(text_length))));
}

static XEN gxg_gdk_text_width_wc(XEN font, XEN text, XEN text_length)
{
  #define H_gdk_text_width_wc "gint gdk_text_width_wc(GdkFont* font, GdkWChar* text, gint text_length)"
  XEN_ASSERT_TYPE(XEN_GdkFont__P(font), font, 1, "gdk_text_width_wc", "GdkFont*");
  XEN_ASSERT_TYPE(XEN_GdkWChar__P(text), text, 2, "gdk_text_width_wc", "GdkWChar*");
  XEN_ASSERT_TYPE(XEN_gint_P(text_length), text_length, 3, "gdk_text_width_wc", "gint");
  return(C_TO_XEN_gint(gdk_text_width_wc(XEN_TO_C_GdkFont_(font), XEN_TO_C_GdkWChar_(text), XEN_TO_C_gint(text_length))));
}

static XEN gxg_gdk_char_width(XEN font, XEN character)
{
  #define H_gdk_char_width "gint gdk_char_width(GdkFont* font, gchar character)"
  XEN_ASSERT_TYPE(XEN_GdkFont__P(font), font, 1, "gdk_char_width", "GdkFont*");
  XEN_ASSERT_TYPE(XEN_gchar_P(character), character, 2, "gdk_char_width", "gchar");
  return(C_TO_XEN_gint(gdk_char_width(XEN_TO_C_GdkFont_(font), XEN_TO_C_gchar(character))));
}

static XEN gxg_gdk_char_width_wc(XEN font, XEN character)
{
  #define H_gdk_char_width_wc "gint gdk_char_width_wc(GdkFont* font, GdkWChar character)"
  XEN_ASSERT_TYPE(XEN_GdkFont__P(font), font, 1, "gdk_char_width_wc", "GdkFont*");
  XEN_ASSERT_TYPE(XEN_GdkWChar_P(character), character, 2, "gdk_char_width_wc", "GdkWChar");
  return(C_TO_XEN_gint(gdk_char_width_wc(XEN_TO_C_GdkFont_(font), XEN_TO_C_GdkWChar(character))));
}

static XEN gxg_gdk_string_measure(XEN font, XEN string)
{
  #define H_gdk_string_measure "gint gdk_string_measure(GdkFont* font, gchar* string)"
  XEN_ASSERT_TYPE(XEN_GdkFont__P(font), font, 1, "gdk_string_measure", "GdkFont*");
  XEN_ASSERT_TYPE(XEN_gchar__P(string), string, 2, "gdk_string_measure", "gchar*");
  return(C_TO_XEN_gint(gdk_string_measure(XEN_TO_C_GdkFont_(font), XEN_TO_C_gchar_(string))));
}

static XEN gxg_gdk_text_measure(XEN font, XEN text, XEN text_length)
{
  #define H_gdk_text_measure "gint gdk_text_measure(GdkFont* font, gchar* text, gint text_length)"
  XEN_ASSERT_TYPE(XEN_GdkFont__P(font), font, 1, "gdk_text_measure", "GdkFont*");
  XEN_ASSERT_TYPE(XEN_gchar__P(text), text, 2, "gdk_text_measure", "gchar*");
  XEN_ASSERT_TYPE(XEN_gint_P(text_length), text_length, 3, "gdk_text_measure", "gint");
  return(C_TO_XEN_gint(gdk_text_measure(XEN_TO_C_GdkFont_(font), XEN_TO_C_gchar_(text), XEN_TO_C_gint(text_length))));
}

static XEN gxg_gdk_char_measure(XEN font, XEN character)
{
  #define H_gdk_char_measure "gint gdk_char_measure(GdkFont* font, gchar character)"
  XEN_ASSERT_TYPE(XEN_GdkFont__P(font), font, 1, "gdk_char_measure", "GdkFont*");
  XEN_ASSERT_TYPE(XEN_gchar_P(character), character, 2, "gdk_char_measure", "gchar");
  return(C_TO_XEN_gint(gdk_char_measure(XEN_TO_C_GdkFont_(font), XEN_TO_C_gchar(character))));
}

static XEN gxg_gdk_string_height(XEN font, XEN string)
{
  #define H_gdk_string_height "gint gdk_string_height(GdkFont* font, gchar* string)"
  XEN_ASSERT_TYPE(XEN_GdkFont__P(font), font, 1, "gdk_string_height", "GdkFont*");
  XEN_ASSERT_TYPE(XEN_gchar__P(string), string, 2, "gdk_string_height", "gchar*");
  return(C_TO_XEN_gint(gdk_string_height(XEN_TO_C_GdkFont_(font), XEN_TO_C_gchar_(string))));
}

static XEN gxg_gdk_text_height(XEN font, XEN text, XEN text_length)
{
  #define H_gdk_text_height "gint gdk_text_height(GdkFont* font, gchar* text, gint text_length)"
  XEN_ASSERT_TYPE(XEN_GdkFont__P(font), font, 1, "gdk_text_height", "GdkFont*");
  XEN_ASSERT_TYPE(XEN_gchar__P(text), text, 2, "gdk_text_height", "gchar*");
  XEN_ASSERT_TYPE(XEN_gint_P(text_length), text_length, 3, "gdk_text_height", "gint");
  return(C_TO_XEN_gint(gdk_text_height(XEN_TO_C_GdkFont_(font), XEN_TO_C_gchar_(text), XEN_TO_C_gint(text_length))));
}

static XEN gxg_gdk_char_height(XEN font, XEN character)
{
  #define H_gdk_char_height "gint gdk_char_height(GdkFont* font, gchar character)"
  XEN_ASSERT_TYPE(XEN_GdkFont__P(font), font, 1, "gdk_char_height", "GdkFont*");
  XEN_ASSERT_TYPE(XEN_gchar_P(character), character, 2, "gdk_char_height", "gchar");
  return(C_TO_XEN_gint(gdk_char_height(XEN_TO_C_GdkFont_(font), XEN_TO_C_gchar(character))));
}

static XEN gxg_gdk_text_extents(XEN font, XEN text, XEN text_length, XEN lbearing, XEN rbearing, XEN width, XEN ascent, XEN descent)
{
  #define H_gdk_text_extents "void gdk_text_extents(GdkFont* font, gchar* text, gint text_length, gint* lbearing, \
gint* rbearing, gint* width, gint* ascent, gint* descent)"
  XEN_ASSERT_TYPE(XEN_GdkFont__P(font), font, 1, "gdk_text_extents", "GdkFont*");
  XEN_ASSERT_TYPE(XEN_gchar__P(text), text, 2, "gdk_text_extents", "gchar*");
  XEN_ASSERT_TYPE(XEN_gint_P(text_length), text_length, 3, "gdk_text_extents", "gint");
  XEN_ASSERT_TYPE(XEN_gint__P(lbearing), lbearing, 4, "gdk_text_extents", "gint*");
  XEN_ASSERT_TYPE(XEN_gint__P(rbearing), rbearing, 5, "gdk_text_extents", "gint*");
  XEN_ASSERT_TYPE(XEN_gint__P(width), width, 6, "gdk_text_extents", "gint*");
  XEN_ASSERT_TYPE(XEN_gint__P(ascent), ascent, 7, "gdk_text_extents", "gint*");
  XEN_ASSERT_TYPE(XEN_gint__P(descent), descent, 8, "gdk_text_extents", "gint*");
  gdk_text_extents(XEN_TO_C_GdkFont_(font), XEN_TO_C_gchar_(text), XEN_TO_C_gint(text_length), XEN_TO_C_gint_(lbearing), 
                   XEN_TO_C_gint_(rbearing), XEN_TO_C_gint_(width), XEN_TO_C_gint_(ascent), XEN_TO_C_gint_(descent));
  return(XEN_FALSE);
}

static XEN gxg_gdk_text_extents_wc(XEN font, XEN text, XEN text_length, XEN lbearing, XEN rbearing, XEN width, XEN ascent, XEN descent)
{
  #define H_gdk_text_extents_wc "void gdk_text_extents_wc(GdkFont* font, GdkWChar* text, gint text_length, \
gint* lbearing, gint* rbearing, gint* width, gint* ascent, gint* descent)"
  XEN_ASSERT_TYPE(XEN_GdkFont__P(font), font, 1, "gdk_text_extents_wc", "GdkFont*");
  XEN_ASSERT_TYPE(XEN_GdkWChar__P(text), text, 2, "gdk_text_extents_wc", "GdkWChar*");
  XEN_ASSERT_TYPE(XEN_gint_P(text_length), text_length, 3, "gdk_text_extents_wc", "gint");
  XEN_ASSERT_TYPE(XEN_gint__P(lbearing), lbearing, 4, "gdk_text_extents_wc", "gint*");
  XEN_ASSERT_TYPE(XEN_gint__P(rbearing), rbearing, 5, "gdk_text_extents_wc", "gint*");
  XEN_ASSERT_TYPE(XEN_gint__P(width), width, 6, "gdk_text_extents_wc", "gint*");
  XEN_ASSERT_TYPE(XEN_gint__P(ascent), ascent, 7, "gdk_text_extents_wc", "gint*");
  XEN_ASSERT_TYPE(XEN_gint__P(descent), descent, 8, "gdk_text_extents_wc", "gint*");
  gdk_text_extents_wc(XEN_TO_C_GdkFont_(font), XEN_TO_C_GdkWChar_(text), XEN_TO_C_gint(text_length), XEN_TO_C_gint_(lbearing), 
                      XEN_TO_C_gint_(rbearing), XEN_TO_C_gint_(width), XEN_TO_C_gint_(ascent), XEN_TO_C_gint_(descent));
  return(XEN_FALSE);
}

static XEN gxg_gdk_string_extents(XEN font, XEN string, XEN lbearing, XEN rbearing, XEN width, XEN ascent, XEN descent)
{
  #define H_gdk_string_extents "void gdk_string_extents(GdkFont* font, gchar* string, gint* lbearing, \
gint* rbearing, gint* width, gint* ascent, gint* descent)"
  XEN_ASSERT_TYPE(XEN_GdkFont__P(font), font, 1, "gdk_string_extents", "GdkFont*");
  XEN_ASSERT_TYPE(XEN_gchar__P(string), string, 2, "gdk_string_extents", "gchar*");
  XEN_ASSERT_TYPE(XEN_gint__P(lbearing), lbearing, 3, "gdk_string_extents", "gint*");
  XEN_ASSERT_TYPE(XEN_gint__P(rbearing), rbearing, 4, "gdk_string_extents", "gint*");
  XEN_ASSERT_TYPE(XEN_gint__P(width), width, 5, "gdk_string_extents", "gint*");
  XEN_ASSERT_TYPE(XEN_gint__P(ascent), ascent, 6, "gdk_string_extents", "gint*");
  XEN_ASSERT_TYPE(XEN_gint__P(descent), descent, 7, "gdk_string_extents", "gint*");
  gdk_string_extents(XEN_TO_C_GdkFont_(font), XEN_TO_C_gchar_(string), XEN_TO_C_gint_(lbearing), XEN_TO_C_gint_(rbearing), 
                     XEN_TO_C_gint_(width), XEN_TO_C_gint_(ascent), XEN_TO_C_gint_(descent));
  return(XEN_FALSE);
}

static XEN gxg_gdk_gc_ref(XEN gc)
{
  #define H_gdk_gc_ref "GdkGC* gdk_gc_ref(GdkGC* gc)"
  XEN_ASSERT_TYPE(XEN_GdkGC__P(gc), gc, 1, "gdk_gc_ref", "GdkGC*");
  return(C_TO_XEN_GdkGC_(gdk_gc_ref(XEN_TO_C_GdkGC_(gc))));
}

static XEN gxg_gdk_gc_unref(XEN gc)
{
  #define H_gdk_gc_unref "void gdk_gc_unref(GdkGC* gc)"
  XEN_ASSERT_TYPE(XEN_GdkGC__P(gc), gc, 1, "gdk_gc_unref", "GdkGC*");
  gdk_gc_unref(XEN_TO_C_GdkGC_(gc));
  return(XEN_FALSE);
}

static XEN gxg_gdk_gc_set_font(XEN gc, XEN font)
{
  #define H_gdk_gc_set_font "void gdk_gc_set_font(GdkGC* gc, GdkFont* font)"
  XEN_ASSERT_TYPE(XEN_GdkGC__P(gc), gc, 1, "gdk_gc_set_font", "GdkGC*");
  XEN_ASSERT_TYPE(XEN_GdkFont__P(font), font, 2, "gdk_gc_set_font", "GdkFont*");
  gdk_gc_set_font(XEN_TO_C_GdkGC_(gc), XEN_TO_C_GdkFont_(font));
  return(XEN_FALSE);
}

static XEN gxg_gdk_exit(XEN error_code)
{
  #define H_gdk_exit "void gdk_exit(gint error_code)"
  XEN_ASSERT_TYPE(XEN_gint_P(error_code), error_code, 1, "gdk_exit", "gint");
  gdk_exit(XEN_TO_C_gint(error_code));
  return(XEN_FALSE);
}

static XEN gxg_gdk_set_use_xshm(XEN use_xshm)
{
  #define H_gdk_set_use_xshm "void gdk_set_use_xshm(gboolean use_xshm)"
  XEN_ASSERT_TYPE(XEN_gboolean_P(use_xshm), use_xshm, 1, "gdk_set_use_xshm", "gboolean");
  gdk_set_use_xshm(XEN_TO_C_gboolean(use_xshm));
  return(XEN_FALSE);
}

static XEN gxg_gdk_get_use_xshm(void)
{
  #define H_gdk_get_use_xshm "gboolean gdk_get_use_xshm( void)"
  return(C_TO_XEN_gboolean(gdk_get_use_xshm()));
}

static XEN gxg_gdk_input_add_full(XEN source, XEN condition, XEN func, XEN func_data, XEN destroy)
{
  #define H_gdk_input_add_full "gint gdk_input_add_full(gint source, GdkInputCondition condition, GdkInputFunction func, \
lambda_data func_data, GdkDestroyNotify destroy)"
  XEN_ASSERT_TYPE(XEN_gint_P(source), source, 1, "gdk_input_add_full", "gint");
  XEN_ASSERT_TYPE(XEN_GdkInputCondition_P(condition), condition, 2, "gdk_input_add_full", "GdkInputCondition");
  XEN_ASSERT_TYPE(XEN_GdkInputFunction_P(func), func, 3, "gdk_input_add_full", "GdkInputFunction");
  XEN_ASSERT_TYPE(XEN_lambda_data_P(func_data), func_data, 4, "gdk_input_add_full", "lambda_data");
  XEN_ASSERT_TYPE(XEN_GdkDestroyNotify_P(destroy), destroy, 5, "gdk_input_add_full", "GdkDestroyNotify");
  {
    XEN result = XEN_FALSE;
    int loc;
    XEN gxg_ptr = XEN_LIST_5(func, func_data, XEN_FALSE, XEN_FALSE, XEN_FALSE);
    loc = xm_protect(gxg_ptr);
    XEN_LIST_SET(gxg_ptr, 2, C_TO_XEN_INT(loc));
    result = C_TO_XEN_gint(gdk_input_add_full(XEN_TO_C_gint(source), XEN_TO_C_GdkInputCondition(condition), XEN_TO_C_GdkInputFunction(func), 
                                              XEN_TO_C_lambda_data(func_data), XEN_TO_C_GdkDestroyNotify(destroy)));
    XEN_LIST_SET(gxg_ptr, 2, XEN_LIST_3(C_STRING_TO_XEN_SYMBOL("idler"), result, C_TO_XEN_INT(loc)));
    return(result);
   }
}

static XEN gxg_gdk_input_add(XEN source, XEN condition, XEN func, XEN func_data)
{
  #define H_gdk_input_add "gint gdk_input_add(gint source, GdkInputCondition condition, GdkInputFunction func, \
lambda_data func_data)"
  XEN_ASSERT_TYPE(XEN_gint_P(source), source, 1, "gdk_input_add", "gint");
  XEN_ASSERT_TYPE(XEN_GdkInputCondition_P(condition), condition, 2, "gdk_input_add", "GdkInputCondition");
  XEN_ASSERT_TYPE(XEN_GdkInputFunction_P(func), func, 3, "gdk_input_add", "GdkInputFunction");
  if (XEN_NOT_BOUND_P(func_data)) func_data = XEN_FALSE; 
  else XEN_ASSERT_TYPE(XEN_lambda_data_P(func_data), func_data, 4, "gdk_input_add", "lambda_data");
  {
    XEN result = XEN_FALSE;
    int loc;
    XEN gxg_ptr = XEN_LIST_5(func, func_data, XEN_FALSE, XEN_FALSE, XEN_FALSE);
    loc = xm_protect(gxg_ptr);
    XEN_LIST_SET(gxg_ptr, 2, C_TO_XEN_INT(loc));
    result = C_TO_XEN_gint(gdk_input_add(XEN_TO_C_gint(source), XEN_TO_C_GdkInputCondition(condition), XEN_TO_C_GdkInputFunction(func), 
                                         XEN_TO_C_lambda_data(func_data)));
    XEN_LIST_SET(gxg_ptr, 2, XEN_LIST_3(C_STRING_TO_XEN_SYMBOL("idler"), result, C_TO_XEN_INT(loc)));
    return(result);
   }
}

static XEN gxg_gdk_input_remove(XEN tag)
{
  #define H_gdk_input_remove "void gdk_input_remove(gint tag)"
  XEN_ASSERT_TYPE(XEN_gint_P(tag), tag, 1, "gdk_input_remove", "gint");
  gdk_input_remove(XEN_TO_C_gint(tag));
  return(XEN_FALSE);
}

static XEN gxg_gdk_image_get(XEN drawable, XEN x, XEN y, XEN width, XEN height)
{
  #define H_gdk_image_get "GdkImage* gdk_image_get(GdkDrawable* drawable, gint x, gint y, gint width, \
gint height)"
  XEN_ASSERT_TYPE(XEN_GdkDrawable__P(drawable), drawable, 1, "gdk_image_get", "GdkDrawable*");
  XEN_ASSERT_TYPE(XEN_gint_P(x), x, 2, "gdk_image_get", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(y), y, 3, "gdk_image_get", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(width), width, 4, "gdk_image_get", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(height), height, 5, "gdk_image_get", "gint");
  return(C_TO_XEN_GdkImage_(gdk_image_get(XEN_TO_C_GdkDrawable_(drawable), XEN_TO_C_gint(x), XEN_TO_C_gint(y), XEN_TO_C_gint(width), 
                                          XEN_TO_C_gint(height))));
}

#if (!HAVE_GTK_1)
static XEN gxg_gdk_image_ref(XEN image)
{
  #define H_gdk_image_ref "GdkImage* gdk_image_ref(GdkImage* image)"
  XEN_ASSERT_TYPE(XEN_GdkImage__P(image), image, 1, "gdk_image_ref", "GdkImage*");
  return(C_TO_XEN_GdkImage_(gdk_image_ref(XEN_TO_C_GdkImage_(image))));
}

static XEN gxg_gdk_image_unref(XEN image)
{
  #define H_gdk_image_unref "void gdk_image_unref(GdkImage* image)"
  XEN_ASSERT_TYPE(XEN_GdkImage__P(image), image, 1, "gdk_image_unref", "GdkImage*");
  gdk_image_unref(XEN_TO_C_GdkImage_(image));
  return(XEN_FALSE);
}
#endif

static XEN gxg_gdk_image_put_pixel(XEN image, XEN x, XEN y, XEN pixel)
{
  #define H_gdk_image_put_pixel "void gdk_image_put_pixel(GdkImage* image, gint x, gint y, guint32 pixel)"
  XEN_ASSERT_TYPE(XEN_GdkImage__P(image), image, 1, "gdk_image_put_pixel", "GdkImage*");
  XEN_ASSERT_TYPE(XEN_gint_P(x), x, 2, "gdk_image_put_pixel", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(y), y, 3, "gdk_image_put_pixel", "gint");
  XEN_ASSERT_TYPE(XEN_guint32_P(pixel), pixel, 4, "gdk_image_put_pixel", "guint32");
  gdk_image_put_pixel(XEN_TO_C_GdkImage_(image), XEN_TO_C_gint(x), XEN_TO_C_gint(y), XEN_TO_C_guint32(pixel));
  return(XEN_FALSE);
}

static XEN gxg_gdk_image_get_pixel(XEN image, XEN x, XEN y)
{
  #define H_gdk_image_get_pixel "guint32 gdk_image_get_pixel(GdkImage* image, gint x, gint y)"
  XEN_ASSERT_TYPE(XEN_GdkImage__P(image), image, 1, "gdk_image_get_pixel", "GdkImage*");
  XEN_ASSERT_TYPE(XEN_gint_P(x), x, 2, "gdk_image_get_pixel", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(y), y, 3, "gdk_image_get_pixel", "gint");
  return(C_TO_XEN_guint32(gdk_image_get_pixel(XEN_TO_C_GdkImage_(image), XEN_TO_C_gint(x), XEN_TO_C_gint(y))));
}

#if (!HAVE_GTK_1)
static XEN gxg_gdk_image_set_colormap(XEN image, XEN colormap)
{
  #define H_gdk_image_set_colormap "void gdk_image_set_colormap(GdkImage* image, GdkColormap* colormap)"
  XEN_ASSERT_TYPE(XEN_GdkImage__P(image), image, 1, "gdk_image_set_colormap", "GdkImage*");
  XEN_ASSERT_TYPE(XEN_GdkColormap__P(colormap), colormap, 2, "gdk_image_set_colormap", "GdkColormap*");
  gdk_image_set_colormap(XEN_TO_C_GdkImage_(image), XEN_TO_C_GdkColormap_(colormap));
  return(XEN_FALSE);
}

static XEN gxg_gdk_image_get_colormap(XEN image)
{
  #define H_gdk_image_get_colormap "GdkColormap* gdk_image_get_colormap(GdkImage* image)"
  XEN_ASSERT_TYPE(XEN_GdkImage__P(image), image, 1, "gdk_image_get_colormap", "GdkImage*");
  return(C_TO_XEN_GdkColormap_(gdk_image_get_colormap(XEN_TO_C_GdkImage_(image))));
}
#endif

static XEN gxg_gdk_rgb_init(void)
{
  #define H_gdk_rgb_init "void gdk_rgb_init( void)"
  gdk_rgb_init();
  return(XEN_FALSE);
}

static XEN gxg_gdk_rgb_xpixel_from_rgb(XEN rgb)
{
  #define H_gdk_rgb_xpixel_from_rgb "gulong gdk_rgb_xpixel_from_rgb(guint32 rgb)"
  XEN_ASSERT_TYPE(XEN_guint32_P(rgb), rgb, 1, "gdk_rgb_xpixel_from_rgb", "guint32");
  return(C_TO_XEN_gulong(gdk_rgb_xpixel_from_rgb(XEN_TO_C_guint32(rgb))));
}

static XEN gxg_gdk_rgb_gc_set_foreground(XEN gc, XEN rgb)
{
  #define H_gdk_rgb_gc_set_foreground "void gdk_rgb_gc_set_foreground(GdkGC* gc, guint32 rgb)"
  XEN_ASSERT_TYPE(XEN_GdkGC__P(gc), gc, 1, "gdk_rgb_gc_set_foreground", "GdkGC*");
  XEN_ASSERT_TYPE(XEN_guint32_P(rgb), rgb, 2, "gdk_rgb_gc_set_foreground", "guint32");
  gdk_rgb_gc_set_foreground(XEN_TO_C_GdkGC_(gc), XEN_TO_C_guint32(rgb));
  return(XEN_FALSE);
}

static XEN gxg_gdk_rgb_gc_set_background(XEN gc, XEN rgb)
{
  #define H_gdk_rgb_gc_set_background "void gdk_rgb_gc_set_background(GdkGC* gc, guint32 rgb)"
  XEN_ASSERT_TYPE(XEN_GdkGC__P(gc), gc, 1, "gdk_rgb_gc_set_background", "GdkGC*");
  XEN_ASSERT_TYPE(XEN_guint32_P(rgb), rgb, 2, "gdk_rgb_gc_set_background", "guint32");
  gdk_rgb_gc_set_background(XEN_TO_C_GdkGC_(gc), XEN_TO_C_guint32(rgb));
  return(XEN_FALSE);
}

static XEN gxg_gdk_window_set_hints(XEN window, XEN x, XEN y, XEN min_width, XEN min_height, XEN max_width, XEN max_height, XEN flags)
{
  #define H_gdk_window_set_hints "void gdk_window_set_hints(GdkWindow* window, gint x, gint y, gint min_width, \
gint min_height, gint max_width, gint max_height, gint flags)"
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 1, "gdk_window_set_hints", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_gint_P(x), x, 2, "gdk_window_set_hints", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(y), y, 3, "gdk_window_set_hints", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(min_width), min_width, 4, "gdk_window_set_hints", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(min_height), min_height, 5, "gdk_window_set_hints", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(max_width), max_width, 6, "gdk_window_set_hints", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(max_height), max_height, 7, "gdk_window_set_hints", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(flags), flags, 8, "gdk_window_set_hints", "gint");
  gdk_window_set_hints(XEN_TO_C_GdkWindow_(window), XEN_TO_C_gint(x), XEN_TO_C_gint(y), XEN_TO_C_gint(min_width), XEN_TO_C_gint(min_height), 
                       XEN_TO_C_gint(max_width), XEN_TO_C_gint(max_height), XEN_TO_C_gint(flags));
  return(XEN_FALSE);
}

static XEN gxg_gtk_button_box_set_child_size(XEN widget, XEN min_width, XEN min_height)
{
  #define H_gtk_button_box_set_child_size "void gtk_button_box_set_child_size(GtkButtonBox* widget, gint min_width, \
gint min_height)"
  XEN_ASSERT_TYPE(XEN_GtkButtonBox__P(widget), widget, 1, "gtk_button_box_set_child_size", "GtkButtonBox*");
  XEN_ASSERT_TYPE(XEN_gint_P(min_width), min_width, 2, "gtk_button_box_set_child_size", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(min_height), min_height, 3, "gtk_button_box_set_child_size", "gint");
  gtk_button_box_set_child_size(XEN_TO_C_GtkButtonBox_(widget), XEN_TO_C_gint(min_width), XEN_TO_C_gint(min_height));
  return(XEN_FALSE);
}

static XEN gxg_gtk_button_box_set_child_ipadding(XEN widget, XEN ipad_x, XEN ipad_y)
{
  #define H_gtk_button_box_set_child_ipadding "void gtk_button_box_set_child_ipadding(GtkButtonBox* widget, \
gint ipad_x, gint ipad_y)"
  XEN_ASSERT_TYPE(XEN_GtkButtonBox__P(widget), widget, 1, "gtk_button_box_set_child_ipadding", "GtkButtonBox*");
  XEN_ASSERT_TYPE(XEN_gint_P(ipad_x), ipad_x, 2, "gtk_button_box_set_child_ipadding", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(ipad_y), ipad_y, 3, "gtk_button_box_set_child_ipadding", "gint");
  gtk_button_box_set_child_ipadding(XEN_TO_C_GtkButtonBox_(widget), XEN_TO_C_gint(ipad_x), XEN_TO_C_gint(ipad_y));
  return(XEN_FALSE);
}

static XEN gxg_gtk_button_box_get_child_size(XEN widget, XEN min_width, XEN min_height)
{
  #define H_gtk_button_box_get_child_size "void gtk_button_box_get_child_size(GtkButtonBox* widget, gint* min_width, \
gint* min_height)"
  XEN_ASSERT_TYPE(XEN_GtkButtonBox__P(widget), widget, 1, "gtk_button_box_get_child_size", "GtkButtonBox*");
  XEN_ASSERT_TYPE(XEN_gint__P(min_width), min_width, 2, "gtk_button_box_get_child_size", "gint*");
  XEN_ASSERT_TYPE(XEN_gint__P(min_height), min_height, 3, "gtk_button_box_get_child_size", "gint*");
  gtk_button_box_get_child_size(XEN_TO_C_GtkButtonBox_(widget), XEN_TO_C_gint_(min_width), XEN_TO_C_gint_(min_height));
  return(XEN_FALSE);
}

static XEN gxg_gtk_button_box_get_child_ipadding(XEN widget, XEN ipad_x, XEN ipad_y)
{
  #define H_gtk_button_box_get_child_ipadding "void gtk_button_box_get_child_ipadding(GtkButtonBox* widget, \
gint* ipad_x, gint* ipad_y)"
  XEN_ASSERT_TYPE(XEN_GtkButtonBox__P(widget), widget, 1, "gtk_button_box_get_child_ipadding", "GtkButtonBox*");
  XEN_ASSERT_TYPE(XEN_gint__P(ipad_x), ipad_x, 2, "gtk_button_box_get_child_ipadding", "gint*");
  XEN_ASSERT_TYPE(XEN_gint__P(ipad_y), ipad_y, 3, "gtk_button_box_get_child_ipadding", "gint*");
  gtk_button_box_get_child_ipadding(XEN_TO_C_GtkButtonBox_(widget), XEN_TO_C_gint_(ipad_x), XEN_TO_C_gint_(ipad_y));
  return(XEN_FALSE);
}

static XEN gxg_gtk_check_menu_item_set_show_toggle(XEN menu_item, XEN always)
{
  #define H_gtk_check_menu_item_set_show_toggle "void gtk_check_menu_item_set_show_toggle(GtkCheckMenuItem* menu_item, \
gboolean always)"
  XEN_ASSERT_TYPE(XEN_GtkCheckMenuItem__P(menu_item), menu_item, 1, "gtk_check_menu_item_set_show_toggle", "GtkCheckMenuItem*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(always), always, 2, "gtk_check_menu_item_set_show_toggle", "gboolean");
  gtk_check_menu_item_set_show_toggle(XEN_TO_C_GtkCheckMenuItem_(menu_item), XEN_TO_C_gboolean(always));
  return(XEN_FALSE);
}

static XEN gxg_gtk_clist_get_type(void)
{
  #define H_gtk_clist_get_type "GtkType gtk_clist_get_type( void)"
  return(C_TO_XEN_GtkType(gtk_clist_get_type()));
}

static XEN gxg_gtk_clist_new(XEN columns)
{
  #define H_gtk_clist_new "GtkWidget* gtk_clist_new(gint columns)"
  XEN_ASSERT_TYPE(XEN_gint_P(columns), columns, 1, "gtk_clist_new", "gint");
  return(C_TO_XEN_GtkWidget_(gtk_clist_new(XEN_TO_C_gint(columns))));
}

static XEN gxg_gtk_clist_new_with_titles(XEN columns, XEN titles)
{
  #define H_gtk_clist_new_with_titles "GtkWidget* gtk_clist_new_with_titles(gint columns, gchar** titles)"
  XEN_ASSERT_TYPE(XEN_gint_P(columns), columns, 1, "gtk_clist_new_with_titles", "gint");
  XEN_ASSERT_TYPE(XEN_gchar___P(titles), titles, 2, "gtk_clist_new_with_titles", "gchar**");
  return(C_TO_XEN_GtkWidget_(gtk_clist_new_with_titles(XEN_TO_C_gint(columns), XEN_TO_C_gchar__(titles))));
}

static XEN gxg_gtk_clist_set_hadjustment(XEN clist, XEN adjustment)
{
  #define H_gtk_clist_set_hadjustment "void gtk_clist_set_hadjustment(GtkCList* clist, GtkAdjustment* adjustment)"
  XEN_ASSERT_TYPE(XEN_GtkCList__P(clist), clist, 1, "gtk_clist_set_hadjustment", "GtkCList*");
  XEN_ASSERT_TYPE(XEN_GtkAdjustment__P(adjustment), adjustment, 2, "gtk_clist_set_hadjustment", "GtkAdjustment*");
  gtk_clist_set_hadjustment(XEN_TO_C_GtkCList_(clist), XEN_TO_C_GtkAdjustment_(adjustment));
  return(XEN_FALSE);
}

static XEN gxg_gtk_clist_set_vadjustment(XEN clist, XEN adjustment)
{
  #define H_gtk_clist_set_vadjustment "void gtk_clist_set_vadjustment(GtkCList* clist, GtkAdjustment* adjustment)"
  XEN_ASSERT_TYPE(XEN_GtkCList__P(clist), clist, 1, "gtk_clist_set_vadjustment", "GtkCList*");
  XEN_ASSERT_TYPE(XEN_GtkAdjustment__P(adjustment), adjustment, 2, "gtk_clist_set_vadjustment", "GtkAdjustment*");
  gtk_clist_set_vadjustment(XEN_TO_C_GtkCList_(clist), XEN_TO_C_GtkAdjustment_(adjustment));
  return(XEN_FALSE);
}

static XEN gxg_gtk_clist_get_hadjustment(XEN clist)
{
  #define H_gtk_clist_get_hadjustment "GtkAdjustment* gtk_clist_get_hadjustment(GtkCList* clist)"
  XEN_ASSERT_TYPE(XEN_GtkCList__P(clist), clist, 1, "gtk_clist_get_hadjustment", "GtkCList*");
  return(C_TO_XEN_GtkAdjustment_(gtk_clist_get_hadjustment(XEN_TO_C_GtkCList_(clist))));
}

static XEN gxg_gtk_clist_get_vadjustment(XEN clist)
{
  #define H_gtk_clist_get_vadjustment "GtkAdjustment* gtk_clist_get_vadjustment(GtkCList* clist)"
  XEN_ASSERT_TYPE(XEN_GtkCList__P(clist), clist, 1, "gtk_clist_get_vadjustment", "GtkCList*");
  return(C_TO_XEN_GtkAdjustment_(gtk_clist_get_vadjustment(XEN_TO_C_GtkCList_(clist))));
}

static XEN gxg_gtk_clist_set_shadow_type(XEN clist, XEN type)
{
  #define H_gtk_clist_set_shadow_type "void gtk_clist_set_shadow_type(GtkCList* clist, GtkShadowType type)"
  XEN_ASSERT_TYPE(XEN_GtkCList__P(clist), clist, 1, "gtk_clist_set_shadow_type", "GtkCList*");
  XEN_ASSERT_TYPE(XEN_GtkShadowType_P(type), type, 2, "gtk_clist_set_shadow_type", "GtkShadowType");
  gtk_clist_set_shadow_type(XEN_TO_C_GtkCList_(clist), XEN_TO_C_GtkShadowType(type));
  return(XEN_FALSE);
}

static XEN gxg_gtk_clist_set_selection_mode(XEN clist, XEN mode)
{
  #define H_gtk_clist_set_selection_mode "void gtk_clist_set_selection_mode(GtkCList* clist, GtkSelectionMode mode)"
  XEN_ASSERT_TYPE(XEN_GtkCList__P(clist), clist, 1, "gtk_clist_set_selection_mode", "GtkCList*");
  XEN_ASSERT_TYPE(XEN_GtkSelectionMode_P(mode), mode, 2, "gtk_clist_set_selection_mode", "GtkSelectionMode");
  gtk_clist_set_selection_mode(XEN_TO_C_GtkCList_(clist), XEN_TO_C_GtkSelectionMode(mode));
  return(XEN_FALSE);
}

static XEN gxg_gtk_clist_set_reorderable(XEN clist, XEN reorderable)
{
  #define H_gtk_clist_set_reorderable "void gtk_clist_set_reorderable(GtkCList* clist, gboolean reorderable)"
  XEN_ASSERT_TYPE(XEN_GtkCList__P(clist), clist, 1, "gtk_clist_set_reorderable", "GtkCList*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(reorderable), reorderable, 2, "gtk_clist_set_reorderable", "gboolean");
  gtk_clist_set_reorderable(XEN_TO_C_GtkCList_(clist), XEN_TO_C_gboolean(reorderable));
  return(XEN_FALSE);
}

static XEN gxg_gtk_clist_set_use_drag_icons(XEN clist, XEN use_icons)
{
  #define H_gtk_clist_set_use_drag_icons "void gtk_clist_set_use_drag_icons(GtkCList* clist, gboolean use_icons)"
  XEN_ASSERT_TYPE(XEN_GtkCList__P(clist), clist, 1, "gtk_clist_set_use_drag_icons", "GtkCList*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(use_icons), use_icons, 2, "gtk_clist_set_use_drag_icons", "gboolean");
  gtk_clist_set_use_drag_icons(XEN_TO_C_GtkCList_(clist), XEN_TO_C_gboolean(use_icons));
  return(XEN_FALSE);
}

static XEN gxg_gtk_clist_set_button_actions(XEN clist, XEN button, XEN button_actions)
{
  #define H_gtk_clist_set_button_actions "void gtk_clist_set_button_actions(GtkCList* clist, guint button, \
guint8 button_actions)"
  XEN_ASSERT_TYPE(XEN_GtkCList__P(clist), clist, 1, "gtk_clist_set_button_actions", "GtkCList*");
  XEN_ASSERT_TYPE(XEN_guint_P(button), button, 2, "gtk_clist_set_button_actions", "guint");
  XEN_ASSERT_TYPE(XEN_guint8_P(button_actions), button_actions, 3, "gtk_clist_set_button_actions", "guint8");
  gtk_clist_set_button_actions(XEN_TO_C_GtkCList_(clist), XEN_TO_C_guint(button), XEN_TO_C_guint8(button_actions));
  return(XEN_FALSE);
}

static XEN gxg_gtk_clist_freeze(XEN clist)
{
  #define H_gtk_clist_freeze "void gtk_clist_freeze(GtkCList* clist)"
  XEN_ASSERT_TYPE(XEN_GtkCList__P(clist), clist, 1, "gtk_clist_freeze", "GtkCList*");
  gtk_clist_freeze(XEN_TO_C_GtkCList_(clist));
  return(XEN_FALSE);
}

static XEN gxg_gtk_clist_thaw(XEN clist)
{
  #define H_gtk_clist_thaw "void gtk_clist_thaw(GtkCList* clist)"
  XEN_ASSERT_TYPE(XEN_GtkCList__P(clist), clist, 1, "gtk_clist_thaw", "GtkCList*");
  gtk_clist_thaw(XEN_TO_C_GtkCList_(clist));
  return(XEN_FALSE);
}

static XEN gxg_gtk_clist_column_titles_show(XEN clist)
{
  #define H_gtk_clist_column_titles_show "void gtk_clist_column_titles_show(GtkCList* clist)"
  XEN_ASSERT_TYPE(XEN_GtkCList__P(clist), clist, 1, "gtk_clist_column_titles_show", "GtkCList*");
  gtk_clist_column_titles_show(XEN_TO_C_GtkCList_(clist));
  return(XEN_FALSE);
}

static XEN gxg_gtk_clist_column_titles_hide(XEN clist)
{
  #define H_gtk_clist_column_titles_hide "void gtk_clist_column_titles_hide(GtkCList* clist)"
  XEN_ASSERT_TYPE(XEN_GtkCList__P(clist), clist, 1, "gtk_clist_column_titles_hide", "GtkCList*");
  gtk_clist_column_titles_hide(XEN_TO_C_GtkCList_(clist));
  return(XEN_FALSE);
}

static XEN gxg_gtk_clist_column_title_active(XEN clist, XEN column)
{
  #define H_gtk_clist_column_title_active "void gtk_clist_column_title_active(GtkCList* clist, gint column)"
  XEN_ASSERT_TYPE(XEN_GtkCList__P(clist), clist, 1, "gtk_clist_column_title_active", "GtkCList*");
  XEN_ASSERT_TYPE(XEN_gint_P(column), column, 2, "gtk_clist_column_title_active", "gint");
  gtk_clist_column_title_active(XEN_TO_C_GtkCList_(clist), XEN_TO_C_gint(column));
  return(XEN_FALSE);
}

static XEN gxg_gtk_clist_column_title_passive(XEN clist, XEN column)
{
  #define H_gtk_clist_column_title_passive "void gtk_clist_column_title_passive(GtkCList* clist, gint column)"
  XEN_ASSERT_TYPE(XEN_GtkCList__P(clist), clist, 1, "gtk_clist_column_title_passive", "GtkCList*");
  XEN_ASSERT_TYPE(XEN_gint_P(column), column, 2, "gtk_clist_column_title_passive", "gint");
  gtk_clist_column_title_passive(XEN_TO_C_GtkCList_(clist), XEN_TO_C_gint(column));
  return(XEN_FALSE);
}

static XEN gxg_gtk_clist_column_titles_active(XEN clist)
{
  #define H_gtk_clist_column_titles_active "void gtk_clist_column_titles_active(GtkCList* clist)"
  XEN_ASSERT_TYPE(XEN_GtkCList__P(clist), clist, 1, "gtk_clist_column_titles_active", "GtkCList*");
  gtk_clist_column_titles_active(XEN_TO_C_GtkCList_(clist));
  return(XEN_FALSE);
}

static XEN gxg_gtk_clist_column_titles_passive(XEN clist)
{
  #define H_gtk_clist_column_titles_passive "void gtk_clist_column_titles_passive(GtkCList* clist)"
  XEN_ASSERT_TYPE(XEN_GtkCList__P(clist), clist, 1, "gtk_clist_column_titles_passive", "GtkCList*");
  gtk_clist_column_titles_passive(XEN_TO_C_GtkCList_(clist));
  return(XEN_FALSE);
}

static XEN gxg_gtk_clist_set_column_title(XEN clist, XEN column, XEN title)
{
  #define H_gtk_clist_set_column_title "void gtk_clist_set_column_title(GtkCList* clist, gint column, \
gchar* title)"
  XEN_ASSERT_TYPE(XEN_GtkCList__P(clist), clist, 1, "gtk_clist_set_column_title", "GtkCList*");
  XEN_ASSERT_TYPE(XEN_gint_P(column), column, 2, "gtk_clist_set_column_title", "gint");
  XEN_ASSERT_TYPE(XEN_gchar__P(title), title, 3, "gtk_clist_set_column_title", "gchar*");
  gtk_clist_set_column_title(XEN_TO_C_GtkCList_(clist), XEN_TO_C_gint(column), XEN_TO_C_gchar_(title));
  return(XEN_FALSE);
}

static XEN gxg_gtk_clist_get_column_title(XEN clist, XEN column)
{
  #define H_gtk_clist_get_column_title "gchar* gtk_clist_get_column_title(GtkCList* clist, gint column)"
  XEN_ASSERT_TYPE(XEN_GtkCList__P(clist), clist, 1, "gtk_clist_get_column_title", "GtkCList*");
  XEN_ASSERT_TYPE(XEN_gint_P(column), column, 2, "gtk_clist_get_column_title", "gint");
  return(C_TO_XEN_gchar_(gtk_clist_get_column_title(XEN_TO_C_GtkCList_(clist), XEN_TO_C_gint(column))));
}

static XEN gxg_gtk_clist_set_column_widget(XEN clist, XEN column, XEN widget)
{
  #define H_gtk_clist_set_column_widget "void gtk_clist_set_column_widget(GtkCList* clist, gint column, \
GtkWidget* widget)"
  XEN_ASSERT_TYPE(XEN_GtkCList__P(clist), clist, 1, "gtk_clist_set_column_widget", "GtkCList*");
  XEN_ASSERT_TYPE(XEN_gint_P(column), column, 2, "gtk_clist_set_column_widget", "gint");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 3, "gtk_clist_set_column_widget", "GtkWidget*");
  gtk_clist_set_column_widget(XEN_TO_C_GtkCList_(clist), XEN_TO_C_gint(column), XEN_TO_C_GtkWidget_(widget));
  return(XEN_FALSE);
}

static XEN gxg_gtk_clist_get_column_widget(XEN clist, XEN column)
{
  #define H_gtk_clist_get_column_widget "GtkWidget* gtk_clist_get_column_widget(GtkCList* clist, gint column)"
  XEN_ASSERT_TYPE(XEN_GtkCList__P(clist), clist, 1, "gtk_clist_get_column_widget", "GtkCList*");
  XEN_ASSERT_TYPE(XEN_gint_P(column), column, 2, "gtk_clist_get_column_widget", "gint");
  return(C_TO_XEN_GtkWidget_(gtk_clist_get_column_widget(XEN_TO_C_GtkCList_(clist), XEN_TO_C_gint(column))));
}

static XEN gxg_gtk_clist_set_column_justification(XEN clist, XEN column, XEN justification)
{
  #define H_gtk_clist_set_column_justification "void gtk_clist_set_column_justification(GtkCList* clist, \
gint column, GtkJustification justification)"
  XEN_ASSERT_TYPE(XEN_GtkCList__P(clist), clist, 1, "gtk_clist_set_column_justification", "GtkCList*");
  XEN_ASSERT_TYPE(XEN_gint_P(column), column, 2, "gtk_clist_set_column_justification", "gint");
  XEN_ASSERT_TYPE(XEN_GtkJustification_P(justification), justification, 3, "gtk_clist_set_column_justification", "GtkJustification");
  gtk_clist_set_column_justification(XEN_TO_C_GtkCList_(clist), XEN_TO_C_gint(column), XEN_TO_C_GtkJustification(justification));
  return(XEN_FALSE);
}

static XEN gxg_gtk_clist_set_column_visibility(XEN clist, XEN column, XEN visible)
{
  #define H_gtk_clist_set_column_visibility "void gtk_clist_set_column_visibility(GtkCList* clist, gint column, \
gboolean visible)"
  XEN_ASSERT_TYPE(XEN_GtkCList__P(clist), clist, 1, "gtk_clist_set_column_visibility", "GtkCList*");
  XEN_ASSERT_TYPE(XEN_gint_P(column), column, 2, "gtk_clist_set_column_visibility", "gint");
  XEN_ASSERT_TYPE(XEN_gboolean_P(visible), visible, 3, "gtk_clist_set_column_visibility", "gboolean");
  gtk_clist_set_column_visibility(XEN_TO_C_GtkCList_(clist), XEN_TO_C_gint(column), XEN_TO_C_gboolean(visible));
  return(XEN_FALSE);
}

static XEN gxg_gtk_clist_set_column_resizeable(XEN clist, XEN column, XEN resizeable)
{
  #define H_gtk_clist_set_column_resizeable "void gtk_clist_set_column_resizeable(GtkCList* clist, gint column, \
gboolean resizeable)"
  XEN_ASSERT_TYPE(XEN_GtkCList__P(clist), clist, 1, "gtk_clist_set_column_resizeable", "GtkCList*");
  XEN_ASSERT_TYPE(XEN_gint_P(column), column, 2, "gtk_clist_set_column_resizeable", "gint");
  XEN_ASSERT_TYPE(XEN_gboolean_P(resizeable), resizeable, 3, "gtk_clist_set_column_resizeable", "gboolean");
  gtk_clist_set_column_resizeable(XEN_TO_C_GtkCList_(clist), XEN_TO_C_gint(column), XEN_TO_C_gboolean(resizeable));
  return(XEN_FALSE);
}

static XEN gxg_gtk_clist_set_column_auto_resize(XEN clist, XEN column, XEN auto_resize)
{
  #define H_gtk_clist_set_column_auto_resize "void gtk_clist_set_column_auto_resize(GtkCList* clist, \
gint column, gboolean auto_resize)"
  XEN_ASSERT_TYPE(XEN_GtkCList__P(clist), clist, 1, "gtk_clist_set_column_auto_resize", "GtkCList*");
  XEN_ASSERT_TYPE(XEN_gint_P(column), column, 2, "gtk_clist_set_column_auto_resize", "gint");
  XEN_ASSERT_TYPE(XEN_gboolean_P(auto_resize), auto_resize, 3, "gtk_clist_set_column_auto_resize", "gboolean");
  gtk_clist_set_column_auto_resize(XEN_TO_C_GtkCList_(clist), XEN_TO_C_gint(column), XEN_TO_C_gboolean(auto_resize));
  return(XEN_FALSE);
}

static XEN gxg_gtk_clist_columns_autosize(XEN clist)
{
  #define H_gtk_clist_columns_autosize "gint gtk_clist_columns_autosize(GtkCList* clist)"
  XEN_ASSERT_TYPE(XEN_GtkCList__P(clist), clist, 1, "gtk_clist_columns_autosize", "GtkCList*");
  return(C_TO_XEN_gint(gtk_clist_columns_autosize(XEN_TO_C_GtkCList_(clist))));
}

static XEN gxg_gtk_clist_optimal_column_width(XEN clist, XEN column)
{
  #define H_gtk_clist_optimal_column_width "gint gtk_clist_optimal_column_width(GtkCList* clist, gint column)"
  XEN_ASSERT_TYPE(XEN_GtkCList__P(clist), clist, 1, "gtk_clist_optimal_column_width", "GtkCList*");
  XEN_ASSERT_TYPE(XEN_gint_P(column), column, 2, "gtk_clist_optimal_column_width", "gint");
  return(C_TO_XEN_gint(gtk_clist_optimal_column_width(XEN_TO_C_GtkCList_(clist), XEN_TO_C_gint(column))));
}

static XEN gxg_gtk_clist_set_column_width(XEN clist, XEN column, XEN width)
{
  #define H_gtk_clist_set_column_width "void gtk_clist_set_column_width(GtkCList* clist, gint column, \
gint width)"
  XEN_ASSERT_TYPE(XEN_GtkCList__P(clist), clist, 1, "gtk_clist_set_column_width", "GtkCList*");
  XEN_ASSERT_TYPE(XEN_gint_P(column), column, 2, "gtk_clist_set_column_width", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(width), width, 3, "gtk_clist_set_column_width", "gint");
  gtk_clist_set_column_width(XEN_TO_C_GtkCList_(clist), XEN_TO_C_gint(column), XEN_TO_C_gint(width));
  return(XEN_FALSE);
}

static XEN gxg_gtk_clist_set_column_min_width(XEN clist, XEN column, XEN min_width)
{
  #define H_gtk_clist_set_column_min_width "void gtk_clist_set_column_min_width(GtkCList* clist, gint column, \
gint min_width)"
  XEN_ASSERT_TYPE(XEN_GtkCList__P(clist), clist, 1, "gtk_clist_set_column_min_width", "GtkCList*");
  XEN_ASSERT_TYPE(XEN_gint_P(column), column, 2, "gtk_clist_set_column_min_width", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(min_width), min_width, 3, "gtk_clist_set_column_min_width", "gint");
  gtk_clist_set_column_min_width(XEN_TO_C_GtkCList_(clist), XEN_TO_C_gint(column), XEN_TO_C_gint(min_width));
  return(XEN_FALSE);
}

static XEN gxg_gtk_clist_set_column_max_width(XEN clist, XEN column, XEN max_width)
{
  #define H_gtk_clist_set_column_max_width "void gtk_clist_set_column_max_width(GtkCList* clist, gint column, \
gint max_width)"
  XEN_ASSERT_TYPE(XEN_GtkCList__P(clist), clist, 1, "gtk_clist_set_column_max_width", "GtkCList*");
  XEN_ASSERT_TYPE(XEN_gint_P(column), column, 2, "gtk_clist_set_column_max_width", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(max_width), max_width, 3, "gtk_clist_set_column_max_width", "gint");
  gtk_clist_set_column_max_width(XEN_TO_C_GtkCList_(clist), XEN_TO_C_gint(column), XEN_TO_C_gint(max_width));
  return(XEN_FALSE);
}

static XEN gxg_gtk_clist_set_row_height(XEN clist, XEN height)
{
  #define H_gtk_clist_set_row_height "void gtk_clist_set_row_height(GtkCList* clist, guint height)"
  XEN_ASSERT_TYPE(XEN_GtkCList__P(clist), clist, 1, "gtk_clist_set_row_height", "GtkCList*");
  XEN_ASSERT_TYPE(XEN_guint_P(height), height, 2, "gtk_clist_set_row_height", "guint");
  gtk_clist_set_row_height(XEN_TO_C_GtkCList_(clist), XEN_TO_C_guint(height));
  return(XEN_FALSE);
}

static XEN gxg_gtk_clist_moveto(XEN clist, XEN row, XEN column, XEN row_align, XEN col_align)
{
  #define H_gtk_clist_moveto "void gtk_clist_moveto(GtkCList* clist, gint row, gint column, gfloat row_align, \
gfloat col_align)"
  XEN_ASSERT_TYPE(XEN_GtkCList__P(clist), clist, 1, "gtk_clist_moveto", "GtkCList*");
  XEN_ASSERT_TYPE(XEN_gint_P(row), row, 2, "gtk_clist_moveto", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(column), column, 3, "gtk_clist_moveto", "gint");
  XEN_ASSERT_TYPE(XEN_gfloat_P(row_align), row_align, 4, "gtk_clist_moveto", "gfloat");
  XEN_ASSERT_TYPE(XEN_gfloat_P(col_align), col_align, 5, "gtk_clist_moveto", "gfloat");
  gtk_clist_moveto(XEN_TO_C_GtkCList_(clist), XEN_TO_C_gint(row), XEN_TO_C_gint(column), XEN_TO_C_gfloat(row_align), XEN_TO_C_gfloat(col_align));
  return(XEN_FALSE);
}

static XEN gxg_gtk_clist_row_is_visible(XEN clist, XEN row)
{
  #define H_gtk_clist_row_is_visible "GtkVisibility gtk_clist_row_is_visible(GtkCList* clist, gint row)"
  XEN_ASSERT_TYPE(XEN_GtkCList__P(clist), clist, 1, "gtk_clist_row_is_visible", "GtkCList*");
  XEN_ASSERT_TYPE(XEN_gint_P(row), row, 2, "gtk_clist_row_is_visible", "gint");
  return(C_TO_XEN_GtkVisibility(gtk_clist_row_is_visible(XEN_TO_C_GtkCList_(clist), XEN_TO_C_gint(row))));
}

static XEN gxg_gtk_clist_get_cell_type(XEN clist, XEN row, XEN column)
{
  #define H_gtk_clist_get_cell_type "GtkCellType gtk_clist_get_cell_type(GtkCList* clist, gint row, gint column)"
  XEN_ASSERT_TYPE(XEN_GtkCList__P(clist), clist, 1, "gtk_clist_get_cell_type", "GtkCList*");
  XEN_ASSERT_TYPE(XEN_gint_P(row), row, 2, "gtk_clist_get_cell_type", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(column), column, 3, "gtk_clist_get_cell_type", "gint");
  return(C_TO_XEN_GtkCellType(gtk_clist_get_cell_type(XEN_TO_C_GtkCList_(clist), XEN_TO_C_gint(row), XEN_TO_C_gint(column))));
}

static XEN gxg_gtk_clist_set_text(XEN clist, XEN row, XEN column, XEN text)
{
  #define H_gtk_clist_set_text "void gtk_clist_set_text(GtkCList* clist, gint row, gint column, gchar* text)"
  XEN_ASSERT_TYPE(XEN_GtkCList__P(clist), clist, 1, "gtk_clist_set_text", "GtkCList*");
  XEN_ASSERT_TYPE(XEN_gint_P(row), row, 2, "gtk_clist_set_text", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(column), column, 3, "gtk_clist_set_text", "gint");
  XEN_ASSERT_TYPE(XEN_gchar__P(text), text, 4, "gtk_clist_set_text", "gchar*");
  gtk_clist_set_text(XEN_TO_C_GtkCList_(clist), XEN_TO_C_gint(row), XEN_TO_C_gint(column), XEN_TO_C_gchar_(text));
  return(XEN_FALSE);
}

static XEN gxg_gtk_clist_get_text(XEN clist, XEN row, XEN column, XEN text)
{
  #define H_gtk_clist_get_text "gint gtk_clist_get_text(GtkCList* clist, gint row, gint column, gchar** [text])"
  gchar* ref_text;
  XEN_ASSERT_TYPE(XEN_GtkCList__P(clist), clist, 1, "gtk_clist_get_text", "GtkCList*");
  XEN_ASSERT_TYPE(XEN_gint_P(row), row, 2, "gtk_clist_get_text", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(column), column, 3, "gtk_clist_get_text", "gint");
  {
    XEN result = XEN_FALSE;
    result = C_TO_XEN_gint(gtk_clist_get_text(XEN_TO_C_GtkCList_(clist), XEN_TO_C_gint(row), XEN_TO_C_gint(column), &ref_text));
    return(XEN_LIST_2(result, C_TO_XEN_gchar_(ref_text)));
   }
}

static XEN gxg_gtk_clist_set_pixmap(XEN clist, XEN row, XEN column, XEN pixmap, XEN mask)
{
  #define H_gtk_clist_set_pixmap "void gtk_clist_set_pixmap(GtkCList* clist, gint row, gint column, GdkPixmap* pixmap, \
GdkBitmap* mask)"
  XEN_ASSERT_TYPE(XEN_GtkCList__P(clist), clist, 1, "gtk_clist_set_pixmap", "GtkCList*");
  XEN_ASSERT_TYPE(XEN_gint_P(row), row, 2, "gtk_clist_set_pixmap", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(column), column, 3, "gtk_clist_set_pixmap", "gint");
  XEN_ASSERT_TYPE(XEN_GdkPixmap__P(pixmap), pixmap, 4, "gtk_clist_set_pixmap", "GdkPixmap*");
  XEN_ASSERT_TYPE(XEN_GdkBitmap__P(mask), mask, 5, "gtk_clist_set_pixmap", "GdkBitmap*");
  gtk_clist_set_pixmap(XEN_TO_C_GtkCList_(clist), XEN_TO_C_gint(row), XEN_TO_C_gint(column), XEN_TO_C_GdkPixmap_(pixmap), 
                       XEN_TO_C_GdkBitmap_(mask));
  return(XEN_FALSE);
}

static XEN gxg_gtk_clist_get_pixmap(XEN clist, XEN row, XEN column, XEN pixmap, XEN mask)
{
  #define H_gtk_clist_get_pixmap "gint gtk_clist_get_pixmap(GtkCList* clist, gint row, gint column, GdkPixmap** [pixmap], \
GdkBitmap** [mask])"
  GdkPixmap* ref_pixmap;
  GdkBitmap* ref_mask;
  XEN_ASSERT_TYPE(XEN_GtkCList__P(clist), clist, 1, "gtk_clist_get_pixmap", "GtkCList*");
  XEN_ASSERT_TYPE(XEN_gint_P(row), row, 2, "gtk_clist_get_pixmap", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(column), column, 3, "gtk_clist_get_pixmap", "gint");
  {
    XEN result = XEN_FALSE;
    result = C_TO_XEN_gint(gtk_clist_get_pixmap(XEN_TO_C_GtkCList_(clist), XEN_TO_C_gint(row), XEN_TO_C_gint(column), &ref_pixmap, 
                                                &ref_mask));
    return(XEN_LIST_3(result, C_TO_XEN_GdkPixmap_(ref_pixmap), C_TO_XEN_GdkBitmap_(ref_mask)));
   }
}

static XEN gxg_gtk_clist_set_pixtext(XEN clist, XEN row, XEN column, XEN text, XEN spacing, XEN pixmap, XEN mask)
{
  #define H_gtk_clist_set_pixtext "void gtk_clist_set_pixtext(GtkCList* clist, gint row, gint column, \
gchar* text, guint8 spacing, GdkPixmap* pixmap, GdkBitmap* mask)"
  XEN_ASSERT_TYPE(XEN_GtkCList__P(clist), clist, 1, "gtk_clist_set_pixtext", "GtkCList*");
  XEN_ASSERT_TYPE(XEN_gint_P(row), row, 2, "gtk_clist_set_pixtext", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(column), column, 3, "gtk_clist_set_pixtext", "gint");
  XEN_ASSERT_TYPE(XEN_gchar__P(text), text, 4, "gtk_clist_set_pixtext", "gchar*");
  XEN_ASSERT_TYPE(XEN_guint8_P(spacing), spacing, 5, "gtk_clist_set_pixtext", "guint8");
  XEN_ASSERT_TYPE(XEN_GdkPixmap__P(pixmap), pixmap, 6, "gtk_clist_set_pixtext", "GdkPixmap*");
  XEN_ASSERT_TYPE(XEN_GdkBitmap__P(mask), mask, 7, "gtk_clist_set_pixtext", "GdkBitmap*");
  gtk_clist_set_pixtext(XEN_TO_C_GtkCList_(clist), XEN_TO_C_gint(row), XEN_TO_C_gint(column), XEN_TO_C_gchar_(text), XEN_TO_C_guint8(spacing), 
                        XEN_TO_C_GdkPixmap_(pixmap), XEN_TO_C_GdkBitmap_(mask));
  return(XEN_FALSE);
}

static XEN gxg_gtk_clist_get_pixtext(XEN clist, XEN row, XEN column, XEN text, XEN spacing, XEN pixmap, XEN mask)
{
  #define H_gtk_clist_get_pixtext "gint gtk_clist_get_pixtext(GtkCList* clist, gint row, gint column, \
gchar** [text], guint8* [spacing], GdkPixmap** [pixmap], GdkBitmap** [mask])"
  gchar* ref_text;
  guint8 ref_spacing;
  GdkPixmap* ref_pixmap;
  GdkBitmap* ref_mask;
  XEN_ASSERT_TYPE(XEN_GtkCList__P(clist), clist, 1, "gtk_clist_get_pixtext", "GtkCList*");
  XEN_ASSERT_TYPE(XEN_gint_P(row), row, 2, "gtk_clist_get_pixtext", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(column), column, 3, "gtk_clist_get_pixtext", "gint");
  {
    XEN result = XEN_FALSE;
    result = C_TO_XEN_gint(gtk_clist_get_pixtext(XEN_TO_C_GtkCList_(clist), XEN_TO_C_gint(row), XEN_TO_C_gint(column), &ref_text, 
                                                 &ref_spacing, &ref_pixmap, &ref_mask));
    return(XEN_LIST_5(result, C_TO_XEN_gchar_(ref_text), C_TO_XEN_guint8(ref_spacing), C_TO_XEN_GdkPixmap_(ref_pixmap), C_TO_XEN_GdkBitmap_(ref_mask)));
   }
}

static XEN gxg_gtk_clist_set_foreground(XEN clist, XEN row, XEN color)
{
  #define H_gtk_clist_set_foreground "void gtk_clist_set_foreground(GtkCList* clist, gint row, GdkColor* color)"
  XEN_ASSERT_TYPE(XEN_GtkCList__P(clist), clist, 1, "gtk_clist_set_foreground", "GtkCList*");
  XEN_ASSERT_TYPE(XEN_gint_P(row), row, 2, "gtk_clist_set_foreground", "gint");
  XEN_ASSERT_TYPE(XEN_GdkColor__P(color), color, 3, "gtk_clist_set_foreground", "GdkColor*");
  gtk_clist_set_foreground(XEN_TO_C_GtkCList_(clist), XEN_TO_C_gint(row), XEN_TO_C_GdkColor_(color));
  return(XEN_FALSE);
}

static XEN gxg_gtk_clist_set_background(XEN clist, XEN row, XEN color)
{
  #define H_gtk_clist_set_background "void gtk_clist_set_background(GtkCList* clist, gint row, GdkColor* color)"
  XEN_ASSERT_TYPE(XEN_GtkCList__P(clist), clist, 1, "gtk_clist_set_background", "GtkCList*");
  XEN_ASSERT_TYPE(XEN_gint_P(row), row, 2, "gtk_clist_set_background", "gint");
  XEN_ASSERT_TYPE(XEN_GdkColor__P(color), color, 3, "gtk_clist_set_background", "GdkColor*");
  gtk_clist_set_background(XEN_TO_C_GtkCList_(clist), XEN_TO_C_gint(row), XEN_TO_C_GdkColor_(color));
  return(XEN_FALSE);
}

static XEN gxg_gtk_clist_set_cell_style(XEN clist, XEN row, XEN column, XEN style)
{
  #define H_gtk_clist_set_cell_style "void gtk_clist_set_cell_style(GtkCList* clist, gint row, gint column, \
GtkStyle* style)"
  XEN_ASSERT_TYPE(XEN_GtkCList__P(clist), clist, 1, "gtk_clist_set_cell_style", "GtkCList*");
  XEN_ASSERT_TYPE(XEN_gint_P(row), row, 2, "gtk_clist_set_cell_style", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(column), column, 3, "gtk_clist_set_cell_style", "gint");
  XEN_ASSERT_TYPE(XEN_GtkStyle__P(style), style, 4, "gtk_clist_set_cell_style", "GtkStyle*");
  gtk_clist_set_cell_style(XEN_TO_C_GtkCList_(clist), XEN_TO_C_gint(row), XEN_TO_C_gint(column), XEN_TO_C_GtkStyle_(style));
  return(XEN_FALSE);
}

static XEN gxg_gtk_clist_get_cell_style(XEN clist, XEN row, XEN column)
{
  #define H_gtk_clist_get_cell_style "GtkStyle* gtk_clist_get_cell_style(GtkCList* clist, gint row, gint column)"
  XEN_ASSERT_TYPE(XEN_GtkCList__P(clist), clist, 1, "gtk_clist_get_cell_style", "GtkCList*");
  XEN_ASSERT_TYPE(XEN_gint_P(row), row, 2, "gtk_clist_get_cell_style", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(column), column, 3, "gtk_clist_get_cell_style", "gint");
  return(C_TO_XEN_GtkStyle_(gtk_clist_get_cell_style(XEN_TO_C_GtkCList_(clist), XEN_TO_C_gint(row), XEN_TO_C_gint(column))));
}

static XEN gxg_gtk_clist_set_row_style(XEN clist, XEN row, XEN style)
{
  #define H_gtk_clist_set_row_style "void gtk_clist_set_row_style(GtkCList* clist, gint row, GtkStyle* style)"
  XEN_ASSERT_TYPE(XEN_GtkCList__P(clist), clist, 1, "gtk_clist_set_row_style", "GtkCList*");
  XEN_ASSERT_TYPE(XEN_gint_P(row), row, 2, "gtk_clist_set_row_style", "gint");
  XEN_ASSERT_TYPE(XEN_GtkStyle__P(style), style, 3, "gtk_clist_set_row_style", "GtkStyle*");
  gtk_clist_set_row_style(XEN_TO_C_GtkCList_(clist), XEN_TO_C_gint(row), XEN_TO_C_GtkStyle_(style));
  return(XEN_FALSE);
}

static XEN gxg_gtk_clist_get_row_style(XEN clist, XEN row)
{
  #define H_gtk_clist_get_row_style "GtkStyle* gtk_clist_get_row_style(GtkCList* clist, gint row)"
  XEN_ASSERT_TYPE(XEN_GtkCList__P(clist), clist, 1, "gtk_clist_get_row_style", "GtkCList*");
  XEN_ASSERT_TYPE(XEN_gint_P(row), row, 2, "gtk_clist_get_row_style", "gint");
  return(C_TO_XEN_GtkStyle_(gtk_clist_get_row_style(XEN_TO_C_GtkCList_(clist), XEN_TO_C_gint(row))));
}

static XEN gxg_gtk_clist_set_shift(XEN clist, XEN row, XEN column, XEN vertical, XEN horizontal)
{
  #define H_gtk_clist_set_shift "void gtk_clist_set_shift(GtkCList* clist, gint row, gint column, gint vertical, \
gint horizontal)"
  XEN_ASSERT_TYPE(XEN_GtkCList__P(clist), clist, 1, "gtk_clist_set_shift", "GtkCList*");
  XEN_ASSERT_TYPE(XEN_gint_P(row), row, 2, "gtk_clist_set_shift", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(column), column, 3, "gtk_clist_set_shift", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(vertical), vertical, 4, "gtk_clist_set_shift", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(horizontal), horizontal, 5, "gtk_clist_set_shift", "gint");
  gtk_clist_set_shift(XEN_TO_C_GtkCList_(clist), XEN_TO_C_gint(row), XEN_TO_C_gint(column), XEN_TO_C_gint(vertical), XEN_TO_C_gint(horizontal));
  return(XEN_FALSE);
}

static XEN gxg_gtk_clist_set_selectable(XEN clist, XEN row, XEN selectable)
{
  #define H_gtk_clist_set_selectable "void gtk_clist_set_selectable(GtkCList* clist, gint row, gboolean selectable)"
  XEN_ASSERT_TYPE(XEN_GtkCList__P(clist), clist, 1, "gtk_clist_set_selectable", "GtkCList*");
  XEN_ASSERT_TYPE(XEN_gint_P(row), row, 2, "gtk_clist_set_selectable", "gint");
  XEN_ASSERT_TYPE(XEN_gboolean_P(selectable), selectable, 3, "gtk_clist_set_selectable", "gboolean");
  gtk_clist_set_selectable(XEN_TO_C_GtkCList_(clist), XEN_TO_C_gint(row), XEN_TO_C_gboolean(selectable));
  return(XEN_FALSE);
}

static XEN gxg_gtk_clist_get_selectable(XEN clist, XEN row)
{
  #define H_gtk_clist_get_selectable "gboolean gtk_clist_get_selectable(GtkCList* clist, gint row)"
  XEN_ASSERT_TYPE(XEN_GtkCList__P(clist), clist, 1, "gtk_clist_get_selectable", "GtkCList*");
  XEN_ASSERT_TYPE(XEN_gint_P(row), row, 2, "gtk_clist_get_selectable", "gint");
  return(C_TO_XEN_gboolean(gtk_clist_get_selectable(XEN_TO_C_GtkCList_(clist), XEN_TO_C_gint(row))));
}

static XEN gxg_gtk_clist_prepend(XEN clist, XEN text)
{
  #define H_gtk_clist_prepend "gint gtk_clist_prepend(GtkCList* clist, gchar** text)"
  XEN_ASSERT_TYPE(XEN_GtkCList__P(clist), clist, 1, "gtk_clist_prepend", "GtkCList*");
  XEN_ASSERT_TYPE(XEN_gchar___P(text), text, 2, "gtk_clist_prepend", "gchar**");
  return(C_TO_XEN_gint(gtk_clist_prepend(XEN_TO_C_GtkCList_(clist), XEN_TO_C_gchar__(text))));
}

static XEN gxg_gtk_clist_append(XEN clist, XEN text)
{
  #define H_gtk_clist_append "gint gtk_clist_append(GtkCList* clist, gchar** text)"
  XEN_ASSERT_TYPE(XEN_GtkCList__P(clist), clist, 1, "gtk_clist_append", "GtkCList*");
  XEN_ASSERT_TYPE(XEN_gchar___P(text), text, 2, "gtk_clist_append", "gchar**");
  return(C_TO_XEN_gint(gtk_clist_append(XEN_TO_C_GtkCList_(clist), XEN_TO_C_gchar__(text))));
}

static XEN gxg_gtk_clist_insert(XEN clist, XEN row, XEN text)
{
  #define H_gtk_clist_insert "gint gtk_clist_insert(GtkCList* clist, gint row, gchar** text)"
  XEN_ASSERT_TYPE(XEN_GtkCList__P(clist), clist, 1, "gtk_clist_insert", "GtkCList*");
  XEN_ASSERT_TYPE(XEN_gint_P(row), row, 2, "gtk_clist_insert", "gint");
  XEN_ASSERT_TYPE(XEN_gchar___P(text), text, 3, "gtk_clist_insert", "gchar**");
  return(C_TO_XEN_gint(gtk_clist_insert(XEN_TO_C_GtkCList_(clist), XEN_TO_C_gint(row), XEN_TO_C_gchar__(text))));
}

static XEN gxg_gtk_clist_remove(XEN clist, XEN row)
{
  #define H_gtk_clist_remove "void gtk_clist_remove(GtkCList* clist, gint row)"
  XEN_ASSERT_TYPE(XEN_GtkCList__P(clist), clist, 1, "gtk_clist_remove", "GtkCList*");
  XEN_ASSERT_TYPE(XEN_gint_P(row), row, 2, "gtk_clist_remove", "gint");
  gtk_clist_remove(XEN_TO_C_GtkCList_(clist), XEN_TO_C_gint(row));
  return(XEN_FALSE);
}

static XEN gxg_gtk_clist_set_row_data(XEN clist, XEN row, XEN data)
{
  #define H_gtk_clist_set_row_data "void gtk_clist_set_row_data(GtkCList* clist, gint row, gpointer data)"
  XEN_ASSERT_TYPE(XEN_GtkCList__P(clist), clist, 1, "gtk_clist_set_row_data", "GtkCList*");
  XEN_ASSERT_TYPE(XEN_gint_P(row), row, 2, "gtk_clist_set_row_data", "gint");
  XEN_ASSERT_TYPE(XEN_gpointer_P(data), data, 3, "gtk_clist_set_row_data", "gpointer");
  gtk_clist_set_row_data(XEN_TO_C_GtkCList_(clist), XEN_TO_C_gint(row), XEN_TO_C_gpointer(data));
  return(XEN_FALSE);
}

static XEN gxg_gtk_clist_set_row_data_full(XEN clist, XEN row, XEN func_data, XEN destroy)
{
  #define H_gtk_clist_set_row_data_full "void gtk_clist_set_row_data_full(GtkCList* clist, gint row, \
gpointer func_data, GtkDestroyNotify destroy)"
  XEN_ASSERT_TYPE(XEN_GtkCList__P(clist), clist, 1, "gtk_clist_set_row_data_full", "GtkCList*");
  XEN_ASSERT_TYPE(XEN_gint_P(row), row, 2, "gtk_clist_set_row_data_full", "gint");
  XEN_ASSERT_TYPE(XEN_gpointer_P(func_data), func_data, 3, "gtk_clist_set_row_data_full", "gpointer");
  XEN_ASSERT_TYPE(XEN_GtkDestroyNotify_P(destroy), destroy, 4, "gtk_clist_set_row_data_full", "GtkDestroyNotify");
  {
    XEN gxg_ptr = XEN_LIST_5(XEN_FALSE, func_data, XEN_FALSE, XEN_FALSE, XEN_FALSE);
    xm_protect(gxg_ptr);
    XEN_LIST_SET(gxg_ptr, 3, destroy);
    gtk_clist_set_row_data_full(XEN_TO_C_GtkCList_(clist), XEN_TO_C_gint(row), XEN_TO_C_gpointer(func_data), XEN_TO_C_GtkDestroyNotify(destroy));
    return(XEN_FALSE);
   }
}

static XEN gxg_gtk_clist_get_row_data(XEN clist, XEN row)
{
  #define H_gtk_clist_get_row_data "gpointer gtk_clist_get_row_data(GtkCList* clist, gint row)"
  XEN_ASSERT_TYPE(XEN_GtkCList__P(clist), clist, 1, "gtk_clist_get_row_data", "GtkCList*");
  XEN_ASSERT_TYPE(XEN_gint_P(row), row, 2, "gtk_clist_get_row_data", "gint");
  return(C_TO_XEN_gpointer(gtk_clist_get_row_data(XEN_TO_C_GtkCList_(clist), XEN_TO_C_gint(row))));
}

static XEN gxg_gtk_clist_find_row_from_data(XEN clist, XEN data)
{
  #define H_gtk_clist_find_row_from_data "gint gtk_clist_find_row_from_data(GtkCList* clist, gpointer data)"
  XEN_ASSERT_TYPE(XEN_GtkCList__P(clist), clist, 1, "gtk_clist_find_row_from_data", "GtkCList*");
  XEN_ASSERT_TYPE(XEN_gpointer_P(data), data, 2, "gtk_clist_find_row_from_data", "gpointer");
  return(C_TO_XEN_gint(gtk_clist_find_row_from_data(XEN_TO_C_GtkCList_(clist), XEN_TO_C_gpointer(data))));
}

static XEN gxg_gtk_clist_select_row(XEN clist, XEN row, XEN column)
{
  #define H_gtk_clist_select_row "void gtk_clist_select_row(GtkCList* clist, gint row, gint column)"
  XEN_ASSERT_TYPE(XEN_GtkCList__P(clist), clist, 1, "gtk_clist_select_row", "GtkCList*");
  XEN_ASSERT_TYPE(XEN_gint_P(row), row, 2, "gtk_clist_select_row", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(column), column, 3, "gtk_clist_select_row", "gint");
  gtk_clist_select_row(XEN_TO_C_GtkCList_(clist), XEN_TO_C_gint(row), XEN_TO_C_gint(column));
  return(XEN_FALSE);
}

static XEN gxg_gtk_clist_unselect_row(XEN clist, XEN row, XEN column)
{
  #define H_gtk_clist_unselect_row "void gtk_clist_unselect_row(GtkCList* clist, gint row, gint column)"
  XEN_ASSERT_TYPE(XEN_GtkCList__P(clist), clist, 1, "gtk_clist_unselect_row", "GtkCList*");
  XEN_ASSERT_TYPE(XEN_gint_P(row), row, 2, "gtk_clist_unselect_row", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(column), column, 3, "gtk_clist_unselect_row", "gint");
  gtk_clist_unselect_row(XEN_TO_C_GtkCList_(clist), XEN_TO_C_gint(row), XEN_TO_C_gint(column));
  return(XEN_FALSE);
}

static XEN gxg_gtk_clist_undo_selection(XEN clist)
{
  #define H_gtk_clist_undo_selection "void gtk_clist_undo_selection(GtkCList* clist)"
  XEN_ASSERT_TYPE(XEN_GtkCList__P(clist), clist, 1, "gtk_clist_undo_selection", "GtkCList*");
  gtk_clist_undo_selection(XEN_TO_C_GtkCList_(clist));
  return(XEN_FALSE);
}

static XEN gxg_gtk_clist_clear(XEN clist)
{
  #define H_gtk_clist_clear "void gtk_clist_clear(GtkCList* clist)"
  XEN_ASSERT_TYPE(XEN_GtkCList__P(clist), clist, 1, "gtk_clist_clear", "GtkCList*");
  gtk_clist_clear(XEN_TO_C_GtkCList_(clist));
  return(XEN_FALSE);
}

static XEN gxg_gtk_clist_get_selection_info(XEN clist, XEN x, XEN y, XEN row, XEN column)
{
  #define H_gtk_clist_get_selection_info "gint gtk_clist_get_selection_info(GtkCList* clist, gint x, \
gint y, gint* row, gint* column)"
  XEN_ASSERT_TYPE(XEN_GtkCList__P(clist), clist, 1, "gtk_clist_get_selection_info", "GtkCList*");
  XEN_ASSERT_TYPE(XEN_gint_P(x), x, 2, "gtk_clist_get_selection_info", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(y), y, 3, "gtk_clist_get_selection_info", "gint");
  XEN_ASSERT_TYPE(XEN_gint__P(row), row, 4, "gtk_clist_get_selection_info", "gint*");
  XEN_ASSERT_TYPE(XEN_gint__P(column), column, 5, "gtk_clist_get_selection_info", "gint*");
  return(C_TO_XEN_gint(gtk_clist_get_selection_info(XEN_TO_C_GtkCList_(clist), XEN_TO_C_gint(x), XEN_TO_C_gint(y), XEN_TO_C_gint_(row), 
                                                    XEN_TO_C_gint_(column))));
}

static XEN gxg_gtk_clist_select_all(XEN clist)
{
  #define H_gtk_clist_select_all "void gtk_clist_select_all(GtkCList* clist)"
  XEN_ASSERT_TYPE(XEN_GtkCList__P(clist), clist, 1, "gtk_clist_select_all", "GtkCList*");
  gtk_clist_select_all(XEN_TO_C_GtkCList_(clist));
  return(XEN_FALSE);
}

static XEN gxg_gtk_clist_unselect_all(XEN clist)
{
  #define H_gtk_clist_unselect_all "void gtk_clist_unselect_all(GtkCList* clist)"
  XEN_ASSERT_TYPE(XEN_GtkCList__P(clist), clist, 1, "gtk_clist_unselect_all", "GtkCList*");
  gtk_clist_unselect_all(XEN_TO_C_GtkCList_(clist));
  return(XEN_FALSE);
}

static XEN gxg_gtk_clist_swap_rows(XEN clist, XEN row1, XEN row2)
{
  #define H_gtk_clist_swap_rows "void gtk_clist_swap_rows(GtkCList* clist, gint row1, gint row2)"
  XEN_ASSERT_TYPE(XEN_GtkCList__P(clist), clist, 1, "gtk_clist_swap_rows", "GtkCList*");
  XEN_ASSERT_TYPE(XEN_gint_P(row1), row1, 2, "gtk_clist_swap_rows", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(row2), row2, 3, "gtk_clist_swap_rows", "gint");
  gtk_clist_swap_rows(XEN_TO_C_GtkCList_(clist), XEN_TO_C_gint(row1), XEN_TO_C_gint(row2));
  return(XEN_FALSE);
}

static XEN gxg_gtk_clist_row_move(XEN clist, XEN source_row, XEN dest_row)
{
  #define H_gtk_clist_row_move "void gtk_clist_row_move(GtkCList* clist, gint source_row, gint dest_row)"
  XEN_ASSERT_TYPE(XEN_GtkCList__P(clist), clist, 1, "gtk_clist_row_move", "GtkCList*");
  XEN_ASSERT_TYPE(XEN_gint_P(source_row), source_row, 2, "gtk_clist_row_move", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(dest_row), dest_row, 3, "gtk_clist_row_move", "gint");
  gtk_clist_row_move(XEN_TO_C_GtkCList_(clist), XEN_TO_C_gint(source_row), XEN_TO_C_gint(dest_row));
  return(XEN_FALSE);
}

static XEN gxg_gtk_clist_set_compare_func(XEN clist, XEN cmp_func)
{
  #define H_gtk_clist_set_compare_func "void gtk_clist_set_compare_func(GtkCList* clist, GtkCListCompareFunc cmp_func)"
  XEN_ASSERT_TYPE(XEN_GtkCList__P(clist), clist, 1, "gtk_clist_set_compare_func", "GtkCList*");
  XEN_ASSERT_TYPE(XEN_GtkCListCompareFunc_P(cmp_func), cmp_func, 2, "gtk_clist_set_compare_func", "GtkCListCompareFunc");
  gtk_clist_set_compare_func(XEN_TO_C_GtkCList_(clist), XEN_TO_C_GtkCListCompareFunc(cmp_func));
  return(XEN_FALSE);
}

static XEN gxg_gtk_clist_set_sort_column(XEN clist, XEN column)
{
  #define H_gtk_clist_set_sort_column "void gtk_clist_set_sort_column(GtkCList* clist, gint column)"
  XEN_ASSERT_TYPE(XEN_GtkCList__P(clist), clist, 1, "gtk_clist_set_sort_column", "GtkCList*");
  XEN_ASSERT_TYPE(XEN_gint_P(column), column, 2, "gtk_clist_set_sort_column", "gint");
  gtk_clist_set_sort_column(XEN_TO_C_GtkCList_(clist), XEN_TO_C_gint(column));
  return(XEN_FALSE);
}

static XEN gxg_gtk_clist_set_sort_type(XEN clist, XEN sort_type)
{
  #define H_gtk_clist_set_sort_type "void gtk_clist_set_sort_type(GtkCList* clist, GtkSortType sort_type)"
  XEN_ASSERT_TYPE(XEN_GtkCList__P(clist), clist, 1, "gtk_clist_set_sort_type", "GtkCList*");
  XEN_ASSERT_TYPE(XEN_GtkSortType_P(sort_type), sort_type, 2, "gtk_clist_set_sort_type", "GtkSortType");
  gtk_clist_set_sort_type(XEN_TO_C_GtkCList_(clist), XEN_TO_C_GtkSortType(sort_type));
  return(XEN_FALSE);
}

static XEN gxg_gtk_clist_sort(XEN clist)
{
  #define H_gtk_clist_sort "void gtk_clist_sort(GtkCList* clist)"
  XEN_ASSERT_TYPE(XEN_GtkCList__P(clist), clist, 1, "gtk_clist_sort", "GtkCList*");
  gtk_clist_sort(XEN_TO_C_GtkCList_(clist));
  return(XEN_FALSE);
}

static XEN gxg_gtk_clist_set_auto_sort(XEN clist, XEN auto_sort)
{
  #define H_gtk_clist_set_auto_sort "void gtk_clist_set_auto_sort(GtkCList* clist, gboolean auto_sort)"
  XEN_ASSERT_TYPE(XEN_GtkCList__P(clist), clist, 1, "gtk_clist_set_auto_sort", "GtkCList*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(auto_sort), auto_sort, 2, "gtk_clist_set_auto_sort", "gboolean");
  gtk_clist_set_auto_sort(XEN_TO_C_GtkCList_(clist), XEN_TO_C_gboolean(auto_sort));
  return(XEN_FALSE);
}

static XEN gxg_gtk_color_selection_set_color(XEN colorsel, XEN color)
{
  #define H_gtk_color_selection_set_color "void gtk_color_selection_set_color(GtkColorSelection* colorsel, \
gdouble* color)"
  XEN_ASSERT_TYPE(XEN_GtkColorSelection__P(colorsel), colorsel, 1, "gtk_color_selection_set_color", "GtkColorSelection*");
  XEN_ASSERT_TYPE(XEN_gdouble__P(color), color, 2, "gtk_color_selection_set_color", "gdouble*");
  gtk_color_selection_set_color(XEN_TO_C_GtkColorSelection_(colorsel), XEN_TO_C_gdouble_(color));
  return(XEN_FALSE);
}

static XEN gxg_gtk_color_selection_get_color(XEN colorsel, XEN color)
{
  #define H_gtk_color_selection_get_color "void gtk_color_selection_get_color(GtkColorSelection* colorsel, \
gdouble* color)"
  XEN_ASSERT_TYPE(XEN_GtkColorSelection__P(colorsel), colorsel, 1, "gtk_color_selection_get_color", "GtkColorSelection*");
  XEN_ASSERT_TYPE(XEN_gdouble__P(color), color, 2, "gtk_color_selection_get_color", "gdouble*");
  gtk_color_selection_get_color(XEN_TO_C_GtkColorSelection_(colorsel), XEN_TO_C_gdouble_(color));
  return(XEN_FALSE);
}

static XEN gxg_gtk_color_selection_set_update_policy(XEN colorsel, XEN policy)
{
  #define H_gtk_color_selection_set_update_policy "void gtk_color_selection_set_update_policy(GtkColorSelection* colorsel, \
GtkUpdateType policy)"
  XEN_ASSERT_TYPE(XEN_GtkColorSelection__P(colorsel), colorsel, 1, "gtk_color_selection_set_update_policy", "GtkColorSelection*");
  XEN_ASSERT_TYPE(XEN_GtkUpdateType_P(policy), policy, 2, "gtk_color_selection_set_update_policy", "GtkUpdateType");
  gtk_color_selection_set_update_policy(XEN_TO_C_GtkColorSelection_(colorsel), XEN_TO_C_GtkUpdateType(policy));
  return(XEN_FALSE);
}

static XEN gxg_gtk_container_foreach_full(XEN container, XEN func, XEN marshal, XEN func_data, XEN notify)
{
  #define H_gtk_container_foreach_full "void gtk_container_foreach_full(GtkContainer* container, GtkCallback func, \
GtkCallbackMarshal marshal, lambda_data func_data, GtkDestroyNotify notify)"
  XEN_ASSERT_TYPE(XEN_GtkContainer__P(container), container, 1, "gtk_container_foreach_full", "GtkContainer*");
  XEN_ASSERT_TYPE(XEN_GtkCallback_P(func), func, 2, "gtk_container_foreach_full", "GtkCallback");
  XEN_ASSERT_TYPE(XEN_GtkCallbackMarshal_P(marshal), marshal, 3, "gtk_container_foreach_full", "GtkCallbackMarshal");
  XEN_ASSERT_TYPE(XEN_lambda_data_P(func_data), func_data, 4, "gtk_container_foreach_full", "lambda_data");
  XEN_ASSERT_TYPE(XEN_GtkDestroyNotify_P(notify), notify, 5, "gtk_container_foreach_full", "GtkDestroyNotify");
  {
    int loc;
    XEN gxg_ptr = XEN_LIST_5(func, func_data, XEN_FALSE, XEN_FALSE, XEN_FALSE);
    loc = xm_protect(gxg_ptr);
    XEN_LIST_SET(gxg_ptr, 2, C_TO_XEN_INT(loc));
    XEN_LIST_SET(gxg_ptr, 3, notify);
    gtk_container_foreach_full(XEN_TO_C_GtkContainer_(container), XEN_TO_C_GtkCallback(func), XEN_TO_C_GtkCallbackMarshal(marshal), 
                           XEN_TO_C_lambda_data(func_data), XEN_TO_C_GtkDestroyNotify(notify));
    xm_unprotect_at(loc);
    return(XEN_FALSE);
   }
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_container_propagate_expose(XEN container, XEN child, XEN event)
{
  #define H_gtk_container_propagate_expose "void gtk_container_propagate_expose(GtkContainer* container, \
GtkWidget* child, GdkEventExpose* event)"
  XEN_ASSERT_TYPE(XEN_GtkContainer__P(container), container, 1, "gtk_container_propagate_expose", "GtkContainer*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(child), child, 2, "gtk_container_propagate_expose", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_GdkEventExpose__P(event), event, 3, "gtk_container_propagate_expose", "GdkEventExpose*");
  gtk_container_propagate_expose(XEN_TO_C_GtkContainer_(container), XEN_TO_C_GtkWidget_(child), XEN_TO_C_GdkEventExpose_(event));
  return(XEN_FALSE);
}

static XEN gxg_gtk_container_set_focus_chain(XEN container, XEN focusable_widgets)
{
  #define H_gtk_container_set_focus_chain "void gtk_container_set_focus_chain(GtkContainer* container, \
GList* focusable_widgets)"
  XEN_ASSERT_TYPE(XEN_GtkContainer__P(container), container, 1, "gtk_container_set_focus_chain", "GtkContainer*");
  XEN_ASSERT_TYPE(XEN_GList__P(focusable_widgets), focusable_widgets, 2, "gtk_container_set_focus_chain", "GList*");
  gtk_container_set_focus_chain(XEN_TO_C_GtkContainer_(container), XEN_TO_C_GList_(focusable_widgets));
  return(XEN_FALSE);
}

static XEN gxg_gtk_container_get_focus_chain(XEN container, XEN focusable_widgets)
{
  #define H_gtk_container_get_focus_chain "gboolean gtk_container_get_focus_chain(GtkContainer* container, \
GList** focusable_widgets)"
  XEN_ASSERT_TYPE(XEN_GtkContainer__P(container), container, 1, "gtk_container_get_focus_chain", "GtkContainer*");
  XEN_ASSERT_TYPE(XEN_GList___P(focusable_widgets), focusable_widgets, 2, "gtk_container_get_focus_chain", "GList**");
  return(C_TO_XEN_gboolean(gtk_container_get_focus_chain(XEN_TO_C_GtkContainer_(container), XEN_TO_C_GList__(focusable_widgets))));
}

static XEN gxg_gtk_container_unset_focus_chain(XEN container)
{
  #define H_gtk_container_unset_focus_chain "void gtk_container_unset_focus_chain(GtkContainer* container)"
  XEN_ASSERT_TYPE(XEN_GtkContainer__P(container), container, 1, "gtk_container_unset_focus_chain", "GtkContainer*");
  gtk_container_unset_focus_chain(XEN_TO_C_GtkContainer_(container));
  return(XEN_FALSE);
}
#endif

static XEN gxg_gtk_container_set_reallocate_redraws(XEN container, XEN needs_redraws)
{
  #define H_gtk_container_set_reallocate_redraws "void gtk_container_set_reallocate_redraws(GtkContainer* container, \
gboolean needs_redraws)"
  XEN_ASSERT_TYPE(XEN_GtkContainer__P(container), container, 1, "gtk_container_set_reallocate_redraws", "GtkContainer*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(needs_redraws), needs_redraws, 2, "gtk_container_set_reallocate_redraws", "gboolean");
  gtk_container_set_reallocate_redraws(XEN_TO_C_GtkContainer_(container), XEN_TO_C_gboolean(needs_redraws));
  return(XEN_FALSE);
}

static XEN gxg_gtk_container_set_focus_child(XEN container, XEN child)
{
  #define H_gtk_container_set_focus_child "void gtk_container_set_focus_child(GtkContainer* container, \
GtkWidget* child)"
  XEN_ASSERT_TYPE(XEN_GtkContainer__P(container), container, 1, "gtk_container_set_focus_child", "GtkContainer*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(child), child, 2, "gtk_container_set_focus_child", "GtkWidget*");
  gtk_container_set_focus_child(XEN_TO_C_GtkContainer_(container), XEN_TO_C_GtkWidget_(child));
  return(XEN_FALSE);
}

static XEN gxg_gtk_container_set_focus_vadjustment(XEN container, XEN adjustment)
{
  #define H_gtk_container_set_focus_vadjustment "void gtk_container_set_focus_vadjustment(GtkContainer* container, \
GtkAdjustment* adjustment)"
  XEN_ASSERT_TYPE(XEN_GtkContainer__P(container), container, 1, "gtk_container_set_focus_vadjustment", "GtkContainer*");
  XEN_ASSERT_TYPE(XEN_GtkAdjustment__P(adjustment), adjustment, 2, "gtk_container_set_focus_vadjustment", "GtkAdjustment*");
  gtk_container_set_focus_vadjustment(XEN_TO_C_GtkContainer_(container), XEN_TO_C_GtkAdjustment_(adjustment));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_container_get_focus_vadjustment(XEN container)
{
  #define H_gtk_container_get_focus_vadjustment "GtkAdjustment* gtk_container_get_focus_vadjustment(GtkContainer* container)"
  XEN_ASSERT_TYPE(XEN_GtkContainer__P(container), container, 1, "gtk_container_get_focus_vadjustment", "GtkContainer*");
  return(C_TO_XEN_GtkAdjustment_(gtk_container_get_focus_vadjustment(XEN_TO_C_GtkContainer_(container))));
}
#endif

static XEN gxg_gtk_container_set_focus_hadjustment(XEN container, XEN adjustment)
{
  #define H_gtk_container_set_focus_hadjustment "void gtk_container_set_focus_hadjustment(GtkContainer* container, \
GtkAdjustment* adjustment)"
  XEN_ASSERT_TYPE(XEN_GtkContainer__P(container), container, 1, "gtk_container_set_focus_hadjustment", "GtkContainer*");
  XEN_ASSERT_TYPE(XEN_GtkAdjustment__P(adjustment), adjustment, 2, "gtk_container_set_focus_hadjustment", "GtkAdjustment*");
  gtk_container_set_focus_hadjustment(XEN_TO_C_GtkContainer_(container), XEN_TO_C_GtkAdjustment_(adjustment));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_container_get_focus_hadjustment(XEN container)
{
  #define H_gtk_container_get_focus_hadjustment "GtkAdjustment* gtk_container_get_focus_hadjustment(GtkContainer* container)"
  XEN_ASSERT_TYPE(XEN_GtkContainer__P(container), container, 1, "gtk_container_get_focus_hadjustment", "GtkContainer*");
  return(C_TO_XEN_GtkAdjustment_(gtk_container_get_focus_hadjustment(XEN_TO_C_GtkContainer_(container))));
}
#endif

static XEN gxg_gtk_container_resize_children(XEN container)
{
  #define H_gtk_container_resize_children "void gtk_container_resize_children(GtkContainer* container)"
  XEN_ASSERT_TYPE(XEN_GtkContainer__P(container), container, 1, "gtk_container_resize_children", "GtkContainer*");
  gtk_container_resize_children(XEN_TO_C_GtkContainer_(container));
  return(XEN_FALSE);
}

static XEN gxg_gtk_container_child_type(XEN container)
{
  #define H_gtk_container_child_type "GtkType gtk_container_child_type(GtkContainer* container)"
  XEN_ASSERT_TYPE(XEN_GtkContainer__P(container), container, 1, "gtk_container_child_type", "GtkContainer*");
  return(C_TO_XEN_GtkType(gtk_container_child_type(XEN_TO_C_GtkContainer_(container))));
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_container_add_with_properties(XEN container, XEN widget, XEN first_prop_name)
{
  #define H_gtk_container_add_with_properties "void gtk_container_add_with_properties(GtkContainer* container, \
GtkWidget* widget, gchar* first_prop_name, ...)"
  XEN_ASSERT_TYPE(XEN_GtkContainer__P(container), container, 1, "gtk_container_add_with_properties", "GtkContainer*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 2, "gtk_container_add_with_properties", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_gchar__P(first_prop_name), first_prop_name, 3, "gtk_container_add_with_properties", "gchar*");
  gtk_container_add_with_properties(XEN_TO_C_GtkContainer_(container), XEN_TO_C_GtkWidget_(widget), XEN_TO_C_gchar_(first_prop_name));
  return(XEN_FALSE);
}
#endif

static XEN gxg_gtk_container_child_set(XEN container, XEN child, XEN first_prop_name)
{
  #define H_gtk_container_child_set "void gtk_container_child_set(GtkContainer* container, GtkWidget* child, \
gchar* first_prop_name, ...)"
  XEN_ASSERT_TYPE(XEN_GtkContainer__P(container), container, 1, "gtk_container_child_set", "GtkContainer*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(child), child, 2, "gtk_container_child_set", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_gchar__P(first_prop_name), first_prop_name, 3, "gtk_container_child_set", "gchar*");
  gtk_container_child_set(XEN_TO_C_GtkContainer_(container), XEN_TO_C_GtkWidget_(child), XEN_TO_C_gchar_(first_prop_name));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_container_child_get(XEN container, XEN child, XEN first_prop_name)
{
  #define H_gtk_container_child_get "void gtk_container_child_get(GtkContainer* container, GtkWidget* child, \
gchar* first_prop_name, ...)"
  XEN_ASSERT_TYPE(XEN_GtkContainer__P(container), container, 1, "gtk_container_child_get", "GtkContainer*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(child), child, 2, "gtk_container_child_get", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_gchar__P(first_prop_name), first_prop_name, 3, "gtk_container_child_get", "gchar*");
  gtk_container_child_get(XEN_TO_C_GtkContainer_(container), XEN_TO_C_GtkWidget_(child), XEN_TO_C_gchar_(first_prop_name));
  return(XEN_FALSE);
}

static XEN gxg_gtk_container_child_set_property(XEN container, XEN child, XEN property_name, XEN value)
{
  #define H_gtk_container_child_set_property "void gtk_container_child_set_property(GtkContainer* container, \
GtkWidget* child, gchar* property_name, GValue* value)"
  XEN_ASSERT_TYPE(XEN_GtkContainer__P(container), container, 1, "gtk_container_child_set_property", "GtkContainer*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(child), child, 2, "gtk_container_child_set_property", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_gchar__P(property_name), property_name, 3, "gtk_container_child_set_property", "gchar*");
  XEN_ASSERT_TYPE(XEN_GValue__P(value), value, 4, "gtk_container_child_set_property", "GValue*");
  gtk_container_child_set_property(XEN_TO_C_GtkContainer_(container), XEN_TO_C_GtkWidget_(child), XEN_TO_C_gchar_(property_name), 
                                   XEN_TO_C_GValue_(value));
  return(XEN_FALSE);
}

static XEN gxg_gtk_container_child_get_property(XEN container, XEN child, XEN property_name, XEN value)
{
  #define H_gtk_container_child_get_property "void gtk_container_child_get_property(GtkContainer* container, \
GtkWidget* child, gchar* property_name, GValue* value)"
  XEN_ASSERT_TYPE(XEN_GtkContainer__P(container), container, 1, "gtk_container_child_get_property", "GtkContainer*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(child), child, 2, "gtk_container_child_get_property", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_gchar__P(property_name), property_name, 3, "gtk_container_child_get_property", "gchar*");
  XEN_ASSERT_TYPE(XEN_GValue__P(value), value, 4, "gtk_container_child_get_property", "GValue*");
  gtk_container_child_get_property(XEN_TO_C_GtkContainer_(container), XEN_TO_C_GtkWidget_(child), XEN_TO_C_gchar_(property_name), 
                                   XEN_TO_C_GValue_(value));
  return(XEN_FALSE);
}
#endif

static XEN gxg_gtk_container_forall(XEN container, XEN func, XEN func_data)
{
  #define H_gtk_container_forall "void gtk_container_forall(GtkContainer* container, GtkCallback func, \
lambda_data func_data)"
  XEN_ASSERT_TYPE(XEN_GtkContainer__P(container), container, 1, "gtk_container_forall", "GtkContainer*");
  XEN_ASSERT_TYPE(XEN_GtkCallback_P(func), func, 2, "gtk_container_forall", "GtkCallback");
  if (XEN_NOT_BOUND_P(func_data)) func_data = XEN_FALSE; 
  else XEN_ASSERT_TYPE(XEN_lambda_data_P(func_data), func_data, 3, "gtk_container_forall", "lambda_data");
  {
    int loc;
    XEN gxg_ptr = XEN_LIST_5(func, func_data, XEN_FALSE, XEN_FALSE, XEN_FALSE);
    loc = xm_protect(gxg_ptr);
    XEN_LIST_SET(gxg_ptr, 2, C_TO_XEN_INT(loc));
    gtk_container_forall(XEN_TO_C_GtkContainer_(container), XEN_TO_C_GtkCallback(func), XEN_TO_C_lambda_data(func_data));
    xm_unprotect_at(loc);
    return(XEN_FALSE);
   }
}

static XEN gxg_gtk_ctree_get_type(void)
{
  #define H_gtk_ctree_get_type "GtkType gtk_ctree_get_type( void)"
  return(C_TO_XEN_GtkType(gtk_ctree_get_type()));
}

static XEN gxg_gtk_ctree_new_with_titles(XEN columns, XEN tree_column, XEN titles)
{
  #define H_gtk_ctree_new_with_titles "GtkWidget* gtk_ctree_new_with_titles(gint columns, gint tree_column, \
gchar** titles)"
  XEN_ASSERT_TYPE(XEN_gint_P(columns), columns, 1, "gtk_ctree_new_with_titles", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(tree_column), tree_column, 2, "gtk_ctree_new_with_titles", "gint");
  XEN_ASSERT_TYPE(XEN_gchar___P(titles), titles, 3, "gtk_ctree_new_with_titles", "gchar**");
  return(C_TO_XEN_GtkWidget_(gtk_ctree_new_with_titles(XEN_TO_C_gint(columns), XEN_TO_C_gint(tree_column), XEN_TO_C_gchar__(titles))));
}

static XEN gxg_gtk_ctree_new(XEN columns, XEN tree_column)
{
  #define H_gtk_ctree_new "GtkWidget* gtk_ctree_new(gint columns, gint tree_column)"
  XEN_ASSERT_TYPE(XEN_gint_P(columns), columns, 1, "gtk_ctree_new", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(tree_column), tree_column, 2, "gtk_ctree_new", "gint");
  return(C_TO_XEN_GtkWidget_(gtk_ctree_new(XEN_TO_C_gint(columns), XEN_TO_C_gint(tree_column))));
}

static XEN gxg_gtk_ctree_insert_node(XEN arglist)
{
  #define H_gtk_ctree_insert_node "GtkCTreeNode* gtk_ctree_insert_node(GtkCTree* ctree, GtkCTreeNode* parent, \
GtkCTreeNode* sibling, gchar** text, guint8 spacing, GdkPixmap* pixmap_closed, GdkBitmap* mask_closed, \
GdkPixmap* pixmap_opened, GdkBitmap* mask_opened, gboolean is_leaf, gboolean expanded)"
  XEN ctree, parent, sibling, text, spacing, pixmap_closed, mask_closed, pixmap_opened, mask_opened, is_leaf, expanded;
  ctree = XEN_LIST_REF(arglist, 0);
  parent = XEN_LIST_REF(arglist, 1);
  sibling = XEN_LIST_REF(arglist, 2);
  text = XEN_LIST_REF(arglist, 3);
  spacing = XEN_LIST_REF(arglist, 4);
  pixmap_closed = XEN_LIST_REF(arglist, 5);
  mask_closed = XEN_LIST_REF(arglist, 6);
  pixmap_opened = XEN_LIST_REF(arglist, 7);
  mask_opened = XEN_LIST_REF(arglist, 8);
  is_leaf = XEN_LIST_REF(arglist, 9);
  expanded = XEN_LIST_REF(arglist, 10);
  XEN_ASSERT_TYPE(XEN_GtkCTree__P(ctree), ctree, 1, "gtk_ctree_insert_node", "GtkCTree*");
  XEN_ASSERT_TYPE(XEN_GtkCTreeNode__P(parent), parent, 2, "gtk_ctree_insert_node", "GtkCTreeNode*");
  XEN_ASSERT_TYPE(XEN_GtkCTreeNode__P(sibling), sibling, 3, "gtk_ctree_insert_node", "GtkCTreeNode*");
  XEN_ASSERT_TYPE(XEN_gchar___P(text), text, 4, "gtk_ctree_insert_node", "gchar**");
  XEN_ASSERT_TYPE(XEN_guint8_P(spacing), spacing, 5, "gtk_ctree_insert_node", "guint8");
  XEN_ASSERT_TYPE(XEN_GdkPixmap__P(pixmap_closed), pixmap_closed, 6, "gtk_ctree_insert_node", "GdkPixmap*");
  XEN_ASSERT_TYPE(XEN_GdkBitmap__P(mask_closed), mask_closed, 7, "gtk_ctree_insert_node", "GdkBitmap*");
  XEN_ASSERT_TYPE(XEN_GdkPixmap__P(pixmap_opened), pixmap_opened, 8, "gtk_ctree_insert_node", "GdkPixmap*");
  XEN_ASSERT_TYPE(XEN_GdkBitmap__P(mask_opened), mask_opened, 9, "gtk_ctree_insert_node", "GdkBitmap*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(is_leaf), is_leaf, 10, "gtk_ctree_insert_node", "gboolean");
  XEN_ASSERT_TYPE(XEN_gboolean_P(expanded), expanded, 11, "gtk_ctree_insert_node", "gboolean");
  return(C_TO_XEN_GtkCTreeNode_(gtk_ctree_insert_node(XEN_TO_C_GtkCTree_(ctree), XEN_TO_C_GtkCTreeNode_(parent), XEN_TO_C_GtkCTreeNode_(sibling), 
                                                      XEN_TO_C_gchar__(text), XEN_TO_C_guint8(spacing), XEN_TO_C_GdkPixmap_(pixmap_closed), 
                                                      XEN_TO_C_GdkBitmap_(mask_closed), XEN_TO_C_GdkPixmap_(pixmap_opened), 
                                                      XEN_TO_C_GdkBitmap_(mask_opened), XEN_TO_C_gboolean(is_leaf), XEN_TO_C_gboolean(expanded))));
}

static XEN gxg_gtk_ctree_remove_node(XEN ctree, XEN node)
{
  #define H_gtk_ctree_remove_node "void gtk_ctree_remove_node(GtkCTree* ctree, GtkCTreeNode* node)"
  XEN_ASSERT_TYPE(XEN_GtkCTree__P(ctree), ctree, 1, "gtk_ctree_remove_node", "GtkCTree*");
  XEN_ASSERT_TYPE(XEN_GtkCTreeNode__P(node), node, 2, "gtk_ctree_remove_node", "GtkCTreeNode*");
  gtk_ctree_remove_node(XEN_TO_C_GtkCTree_(ctree), XEN_TO_C_GtkCTreeNode_(node));
  return(XEN_FALSE);
}

static XEN gxg_gtk_ctree_insert_gnode(XEN ctree, XEN parent, XEN sibling, XEN gnode, XEN func, XEN data)
{
  #define H_gtk_ctree_insert_gnode "GtkCTreeNode* gtk_ctree_insert_gnode(GtkCTree* ctree, GtkCTreeNode* parent, \
GtkCTreeNode* sibling, GNode* gnode, GtkCTreeGNodeFunc func, gpointer data)"
  XEN_ASSERT_TYPE(XEN_GtkCTree__P(ctree), ctree, 1, "gtk_ctree_insert_gnode", "GtkCTree*");
  XEN_ASSERT_TYPE(XEN_GtkCTreeNode__P(parent), parent, 2, "gtk_ctree_insert_gnode", "GtkCTreeNode*");
  XEN_ASSERT_TYPE(XEN_GtkCTreeNode__P(sibling), sibling, 3, "gtk_ctree_insert_gnode", "GtkCTreeNode*");
  XEN_ASSERT_TYPE(XEN_GNode__P(gnode), gnode, 4, "gtk_ctree_insert_gnode", "GNode*");
  XEN_ASSERT_TYPE(XEN_GtkCTreeGNodeFunc_P(func), func, 5, "gtk_ctree_insert_gnode", "GtkCTreeGNodeFunc");
  XEN_ASSERT_TYPE(XEN_gpointer_P(data), data, 6, "gtk_ctree_insert_gnode", "gpointer");
  return(C_TO_XEN_GtkCTreeNode_(gtk_ctree_insert_gnode(XEN_TO_C_GtkCTree_(ctree), XEN_TO_C_GtkCTreeNode_(parent), XEN_TO_C_GtkCTreeNode_(sibling), 
                                                       XEN_TO_C_GNode_(gnode), XEN_TO_C_GtkCTreeGNodeFunc(func), XEN_TO_C_gpointer(data))));
}

static XEN gxg_gtk_ctree_export_to_gnode(XEN ctree, XEN parent, XEN sibling, XEN node, XEN func, XEN data)
{
  #define H_gtk_ctree_export_to_gnode "GNode* gtk_ctree_export_to_gnode(GtkCTree* ctree, GNode* parent, \
GNode* sibling, GtkCTreeNode* node, GtkCTreeGNodeFunc func, gpointer data)"
  XEN_ASSERT_TYPE(XEN_GtkCTree__P(ctree), ctree, 1, "gtk_ctree_export_to_gnode", "GtkCTree*");
  XEN_ASSERT_TYPE(XEN_GNode__P(parent), parent, 2, "gtk_ctree_export_to_gnode", "GNode*");
  XEN_ASSERT_TYPE(XEN_GNode__P(sibling), sibling, 3, "gtk_ctree_export_to_gnode", "GNode*");
  XEN_ASSERT_TYPE(XEN_GtkCTreeNode__P(node), node, 4, "gtk_ctree_export_to_gnode", "GtkCTreeNode*");
  XEN_ASSERT_TYPE(XEN_GtkCTreeGNodeFunc_P(func), func, 5, "gtk_ctree_export_to_gnode", "GtkCTreeGNodeFunc");
  XEN_ASSERT_TYPE(XEN_gpointer_P(data), data, 6, "gtk_ctree_export_to_gnode", "gpointer");
  return(C_TO_XEN_GNode_(gtk_ctree_export_to_gnode(XEN_TO_C_GtkCTree_(ctree), XEN_TO_C_GNode_(parent), XEN_TO_C_GNode_(sibling), 
                                                   XEN_TO_C_GtkCTreeNode_(node), XEN_TO_C_GtkCTreeGNodeFunc(func), XEN_TO_C_gpointer(data))));
}

static XEN gxg_gtk_ctree_post_recursive(XEN ctree, XEN node, XEN func, XEN data)
{
  #define H_gtk_ctree_post_recursive "void gtk_ctree_post_recursive(GtkCTree* ctree, GtkCTreeNode* node, \
GtkCTreeFunc func, gpointer data)"
  XEN_ASSERT_TYPE(XEN_GtkCTree__P(ctree), ctree, 1, "gtk_ctree_post_recursive", "GtkCTree*");
  XEN_ASSERT_TYPE(XEN_GtkCTreeNode__P(node), node, 2, "gtk_ctree_post_recursive", "GtkCTreeNode*");
  XEN_ASSERT_TYPE(XEN_GtkCTreeFunc_P(func), func, 3, "gtk_ctree_post_recursive", "GtkCTreeFunc");
  XEN_ASSERT_TYPE(XEN_gpointer_P(data), data, 4, "gtk_ctree_post_recursive", "gpointer");
  gtk_ctree_post_recursive(XEN_TO_C_GtkCTree_(ctree), XEN_TO_C_GtkCTreeNode_(node), XEN_TO_C_GtkCTreeFunc(func), XEN_TO_C_gpointer(data));
  return(XEN_FALSE);
}

static XEN gxg_gtk_ctree_post_recursive_to_depth(XEN ctree, XEN node, XEN depth, XEN func, XEN data)
{
  #define H_gtk_ctree_post_recursive_to_depth "void gtk_ctree_post_recursive_to_depth(GtkCTree* ctree, \
GtkCTreeNode* node, gint depth, GtkCTreeFunc func, gpointer data)"
  XEN_ASSERT_TYPE(XEN_GtkCTree__P(ctree), ctree, 1, "gtk_ctree_post_recursive_to_depth", "GtkCTree*");
  XEN_ASSERT_TYPE(XEN_GtkCTreeNode__P(node), node, 2, "gtk_ctree_post_recursive_to_depth", "GtkCTreeNode*");
  XEN_ASSERT_TYPE(XEN_gint_P(depth), depth, 3, "gtk_ctree_post_recursive_to_depth", "gint");
  XEN_ASSERT_TYPE(XEN_GtkCTreeFunc_P(func), func, 4, "gtk_ctree_post_recursive_to_depth", "GtkCTreeFunc");
  XEN_ASSERT_TYPE(XEN_gpointer_P(data), data, 5, "gtk_ctree_post_recursive_to_depth", "gpointer");
  gtk_ctree_post_recursive_to_depth(XEN_TO_C_GtkCTree_(ctree), XEN_TO_C_GtkCTreeNode_(node), XEN_TO_C_gint(depth), XEN_TO_C_GtkCTreeFunc(func), 
                                    XEN_TO_C_gpointer(data));
  return(XEN_FALSE);
}

static XEN gxg_gtk_ctree_pre_recursive(XEN ctree, XEN node, XEN func, XEN data)
{
  #define H_gtk_ctree_pre_recursive "void gtk_ctree_pre_recursive(GtkCTree* ctree, GtkCTreeNode* node, \
GtkCTreeFunc func, gpointer data)"
  XEN_ASSERT_TYPE(XEN_GtkCTree__P(ctree), ctree, 1, "gtk_ctree_pre_recursive", "GtkCTree*");
  XEN_ASSERT_TYPE(XEN_GtkCTreeNode__P(node), node, 2, "gtk_ctree_pre_recursive", "GtkCTreeNode*");
  XEN_ASSERT_TYPE(XEN_GtkCTreeFunc_P(func), func, 3, "gtk_ctree_pre_recursive", "GtkCTreeFunc");
  XEN_ASSERT_TYPE(XEN_gpointer_P(data), data, 4, "gtk_ctree_pre_recursive", "gpointer");
  gtk_ctree_pre_recursive(XEN_TO_C_GtkCTree_(ctree), XEN_TO_C_GtkCTreeNode_(node), XEN_TO_C_GtkCTreeFunc(func), XEN_TO_C_gpointer(data));
  return(XEN_FALSE);
}

static XEN gxg_gtk_ctree_pre_recursive_to_depth(XEN ctree, XEN node, XEN depth, XEN func, XEN data)
{
  #define H_gtk_ctree_pre_recursive_to_depth "void gtk_ctree_pre_recursive_to_depth(GtkCTree* ctree, \
GtkCTreeNode* node, gint depth, GtkCTreeFunc func, gpointer data)"
  XEN_ASSERT_TYPE(XEN_GtkCTree__P(ctree), ctree, 1, "gtk_ctree_pre_recursive_to_depth", "GtkCTree*");
  XEN_ASSERT_TYPE(XEN_GtkCTreeNode__P(node), node, 2, "gtk_ctree_pre_recursive_to_depth", "GtkCTreeNode*");
  XEN_ASSERT_TYPE(XEN_gint_P(depth), depth, 3, "gtk_ctree_pre_recursive_to_depth", "gint");
  XEN_ASSERT_TYPE(XEN_GtkCTreeFunc_P(func), func, 4, "gtk_ctree_pre_recursive_to_depth", "GtkCTreeFunc");
  XEN_ASSERT_TYPE(XEN_gpointer_P(data), data, 5, "gtk_ctree_pre_recursive_to_depth", "gpointer");
  gtk_ctree_pre_recursive_to_depth(XEN_TO_C_GtkCTree_(ctree), XEN_TO_C_GtkCTreeNode_(node), XEN_TO_C_gint(depth), XEN_TO_C_GtkCTreeFunc(func), 
                                   XEN_TO_C_gpointer(data));
  return(XEN_FALSE);
}

static XEN gxg_gtk_ctree_is_viewable(XEN ctree, XEN node)
{
  #define H_gtk_ctree_is_viewable "gboolean gtk_ctree_is_viewable(GtkCTree* ctree, GtkCTreeNode* node)"
  XEN_ASSERT_TYPE(XEN_GtkCTree__P(ctree), ctree, 1, "gtk_ctree_is_viewable", "GtkCTree*");
  XEN_ASSERT_TYPE(XEN_GtkCTreeNode__P(node), node, 2, "gtk_ctree_is_viewable", "GtkCTreeNode*");
  return(C_TO_XEN_gboolean(gtk_ctree_is_viewable(XEN_TO_C_GtkCTree_(ctree), XEN_TO_C_GtkCTreeNode_(node))));
}

static XEN gxg_gtk_ctree_last(XEN ctree, XEN node)
{
  #define H_gtk_ctree_last "GtkCTreeNode* gtk_ctree_last(GtkCTree* ctree, GtkCTreeNode* node)"
  XEN_ASSERT_TYPE(XEN_GtkCTree__P(ctree), ctree, 1, "gtk_ctree_last", "GtkCTree*");
  XEN_ASSERT_TYPE(XEN_GtkCTreeNode__P(node), node, 2, "gtk_ctree_last", "GtkCTreeNode*");
  return(C_TO_XEN_GtkCTreeNode_(gtk_ctree_last(XEN_TO_C_GtkCTree_(ctree), XEN_TO_C_GtkCTreeNode_(node))));
}

static XEN gxg_gtk_ctree_find_node_ptr(XEN ctree, XEN ctree_row)
{
  #define H_gtk_ctree_find_node_ptr "GtkCTreeNode* gtk_ctree_find_node_ptr(GtkCTree* ctree, GtkCTreeRow* ctree_row)"
  XEN_ASSERT_TYPE(XEN_GtkCTree__P(ctree), ctree, 1, "gtk_ctree_find_node_ptr", "GtkCTree*");
  XEN_ASSERT_TYPE(XEN_GtkCTreeRow__P(ctree_row), ctree_row, 2, "gtk_ctree_find_node_ptr", "GtkCTreeRow*");
  return(C_TO_XEN_GtkCTreeNode_(gtk_ctree_find_node_ptr(XEN_TO_C_GtkCTree_(ctree), XEN_TO_C_GtkCTreeRow_(ctree_row))));
}

static XEN gxg_gtk_ctree_node_nth(XEN ctree, XEN row)
{
  #define H_gtk_ctree_node_nth "GtkCTreeNode* gtk_ctree_node_nth(GtkCTree* ctree, guint row)"
  XEN_ASSERT_TYPE(XEN_GtkCTree__P(ctree), ctree, 1, "gtk_ctree_node_nth", "GtkCTree*");
  XEN_ASSERT_TYPE(XEN_guint_P(row), row, 2, "gtk_ctree_node_nth", "guint");
  return(C_TO_XEN_GtkCTreeNode_(gtk_ctree_node_nth(XEN_TO_C_GtkCTree_(ctree), XEN_TO_C_guint(row))));
}

static XEN gxg_gtk_ctree_find(XEN ctree, XEN node, XEN child)
{
  #define H_gtk_ctree_find "gboolean gtk_ctree_find(GtkCTree* ctree, GtkCTreeNode* node, GtkCTreeNode* child)"
  XEN_ASSERT_TYPE(XEN_GtkCTree__P(ctree), ctree, 1, "gtk_ctree_find", "GtkCTree*");
  XEN_ASSERT_TYPE(XEN_GtkCTreeNode__P(node), node, 2, "gtk_ctree_find", "GtkCTreeNode*");
  XEN_ASSERT_TYPE(XEN_GtkCTreeNode__P(child), child, 3, "gtk_ctree_find", "GtkCTreeNode*");
  return(C_TO_XEN_gboolean(gtk_ctree_find(XEN_TO_C_GtkCTree_(ctree), XEN_TO_C_GtkCTreeNode_(node), XEN_TO_C_GtkCTreeNode_(child))));
}

static XEN gxg_gtk_ctree_is_ancestor(XEN ctree, XEN node, XEN child)
{
  #define H_gtk_ctree_is_ancestor "gboolean gtk_ctree_is_ancestor(GtkCTree* ctree, GtkCTreeNode* node, \
GtkCTreeNode* child)"
  XEN_ASSERT_TYPE(XEN_GtkCTree__P(ctree), ctree, 1, "gtk_ctree_is_ancestor", "GtkCTree*");
  XEN_ASSERT_TYPE(XEN_GtkCTreeNode__P(node), node, 2, "gtk_ctree_is_ancestor", "GtkCTreeNode*");
  XEN_ASSERT_TYPE(XEN_GtkCTreeNode__P(child), child, 3, "gtk_ctree_is_ancestor", "GtkCTreeNode*");
  return(C_TO_XEN_gboolean(gtk_ctree_is_ancestor(XEN_TO_C_GtkCTree_(ctree), XEN_TO_C_GtkCTreeNode_(node), XEN_TO_C_GtkCTreeNode_(child))));
}

static XEN gxg_gtk_ctree_find_by_row_data(XEN ctree, XEN node, XEN data)
{
  #define H_gtk_ctree_find_by_row_data "GtkCTreeNode* gtk_ctree_find_by_row_data(GtkCTree* ctree, GtkCTreeNode* node, \
gpointer data)"
  XEN_ASSERT_TYPE(XEN_GtkCTree__P(ctree), ctree, 1, "gtk_ctree_find_by_row_data", "GtkCTree*");
  XEN_ASSERT_TYPE(XEN_GtkCTreeNode__P(node), node, 2, "gtk_ctree_find_by_row_data", "GtkCTreeNode*");
  XEN_ASSERT_TYPE(XEN_gpointer_P(data), data, 3, "gtk_ctree_find_by_row_data", "gpointer");
  return(C_TO_XEN_GtkCTreeNode_(gtk_ctree_find_by_row_data(XEN_TO_C_GtkCTree_(ctree), XEN_TO_C_GtkCTreeNode_(node), XEN_TO_C_gpointer(data))));
}

static XEN gxg_gtk_ctree_find_all_by_row_data(XEN ctree, XEN node, XEN data)
{
  #define H_gtk_ctree_find_all_by_row_data "GList* gtk_ctree_find_all_by_row_data(GtkCTree* ctree, GtkCTreeNode* node, \
gpointer data)"
  XEN_ASSERT_TYPE(XEN_GtkCTree__P(ctree), ctree, 1, "gtk_ctree_find_all_by_row_data", "GtkCTree*");
  XEN_ASSERT_TYPE(XEN_GtkCTreeNode__P(node), node, 2, "gtk_ctree_find_all_by_row_data", "GtkCTreeNode*");
  XEN_ASSERT_TYPE(XEN_gpointer_P(data), data, 3, "gtk_ctree_find_all_by_row_data", "gpointer");
  return(C_TO_XEN_GList_(gtk_ctree_find_all_by_row_data(XEN_TO_C_GtkCTree_(ctree), XEN_TO_C_GtkCTreeNode_(node), XEN_TO_C_gpointer(data))));
}

static XEN gxg_gtk_ctree_find_by_row_data_custom(XEN ctree, XEN node, XEN data, XEN func)
{
  #define H_gtk_ctree_find_by_row_data_custom "GtkCTreeNode* gtk_ctree_find_by_row_data_custom(GtkCTree* ctree, \
GtkCTreeNode* node, gpointer data, GCompareFunc func)"
  XEN_ASSERT_TYPE(XEN_GtkCTree__P(ctree), ctree, 1, "gtk_ctree_find_by_row_data_custom", "GtkCTree*");
  XEN_ASSERT_TYPE(XEN_GtkCTreeNode__P(node), node, 2, "gtk_ctree_find_by_row_data_custom", "GtkCTreeNode*");
  XEN_ASSERT_TYPE(XEN_gpointer_P(data), data, 3, "gtk_ctree_find_by_row_data_custom", "gpointer");
  XEN_ASSERT_TYPE(XEN_GCompareFunc_P(func), func, 4, "gtk_ctree_find_by_row_data_custom", "GCompareFunc");
  return(C_TO_XEN_GtkCTreeNode_(gtk_ctree_find_by_row_data_custom(XEN_TO_C_GtkCTree_(ctree), XEN_TO_C_GtkCTreeNode_(node), 
                                                                  XEN_TO_C_gpointer(data), XEN_TO_C_GCompareFunc(func))));
}

static XEN gxg_gtk_ctree_find_all_by_row_data_custom(XEN ctree, XEN node, XEN data, XEN func)
{
  #define H_gtk_ctree_find_all_by_row_data_custom "GList* gtk_ctree_find_all_by_row_data_custom(GtkCTree* ctree, \
GtkCTreeNode* node, gpointer data, GCompareFunc func)"
  XEN_ASSERT_TYPE(XEN_GtkCTree__P(ctree), ctree, 1, "gtk_ctree_find_all_by_row_data_custom", "GtkCTree*");
  XEN_ASSERT_TYPE(XEN_GtkCTreeNode__P(node), node, 2, "gtk_ctree_find_all_by_row_data_custom", "GtkCTreeNode*");
  XEN_ASSERT_TYPE(XEN_gpointer_P(data), data, 3, "gtk_ctree_find_all_by_row_data_custom", "gpointer");
  XEN_ASSERT_TYPE(XEN_GCompareFunc_P(func), func, 4, "gtk_ctree_find_all_by_row_data_custom", "GCompareFunc");
  return(C_TO_XEN_GList_(gtk_ctree_find_all_by_row_data_custom(XEN_TO_C_GtkCTree_(ctree), XEN_TO_C_GtkCTreeNode_(node), XEN_TO_C_gpointer(data), 
                                                               XEN_TO_C_GCompareFunc(func))));
}

static XEN gxg_gtk_ctree_is_hot_spot(XEN ctree, XEN x, XEN y)
{
  #define H_gtk_ctree_is_hot_spot "gboolean gtk_ctree_is_hot_spot(GtkCTree* ctree, gint x, gint y)"
  XEN_ASSERT_TYPE(XEN_GtkCTree__P(ctree), ctree, 1, "gtk_ctree_is_hot_spot", "GtkCTree*");
  XEN_ASSERT_TYPE(XEN_gint_P(x), x, 2, "gtk_ctree_is_hot_spot", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(y), y, 3, "gtk_ctree_is_hot_spot", "gint");
  return(C_TO_XEN_gboolean(gtk_ctree_is_hot_spot(XEN_TO_C_GtkCTree_(ctree), XEN_TO_C_gint(x), XEN_TO_C_gint(y))));
}

static XEN gxg_gtk_ctree_move(XEN ctree, XEN node, XEN new_parent, XEN new_sibling)
{
  #define H_gtk_ctree_move "void gtk_ctree_move(GtkCTree* ctree, GtkCTreeNode* node, GtkCTreeNode* new_parent, \
GtkCTreeNode* new_sibling)"
  XEN_ASSERT_TYPE(XEN_GtkCTree__P(ctree), ctree, 1, "gtk_ctree_move", "GtkCTree*");
  XEN_ASSERT_TYPE(XEN_GtkCTreeNode__P(node), node, 2, "gtk_ctree_move", "GtkCTreeNode*");
  XEN_ASSERT_TYPE(XEN_GtkCTreeNode__P(new_parent), new_parent, 3, "gtk_ctree_move", "GtkCTreeNode*");
  XEN_ASSERT_TYPE(XEN_GtkCTreeNode__P(new_sibling), new_sibling, 4, "gtk_ctree_move", "GtkCTreeNode*");
  gtk_ctree_move(XEN_TO_C_GtkCTree_(ctree), XEN_TO_C_GtkCTreeNode_(node), XEN_TO_C_GtkCTreeNode_(new_parent), XEN_TO_C_GtkCTreeNode_(new_sibling));
  return(XEN_FALSE);
}

static XEN gxg_gtk_ctree_expand(XEN ctree, XEN node)
{
  #define H_gtk_ctree_expand "void gtk_ctree_expand(GtkCTree* ctree, GtkCTreeNode* node)"
  XEN_ASSERT_TYPE(XEN_GtkCTree__P(ctree), ctree, 1, "gtk_ctree_expand", "GtkCTree*");
  XEN_ASSERT_TYPE(XEN_GtkCTreeNode__P(node), node, 2, "gtk_ctree_expand", "GtkCTreeNode*");
  gtk_ctree_expand(XEN_TO_C_GtkCTree_(ctree), XEN_TO_C_GtkCTreeNode_(node));
  return(XEN_FALSE);
}

static XEN gxg_gtk_ctree_expand_recursive(XEN ctree, XEN node)
{
  #define H_gtk_ctree_expand_recursive "void gtk_ctree_expand_recursive(GtkCTree* ctree, GtkCTreeNode* node)"
  XEN_ASSERT_TYPE(XEN_GtkCTree__P(ctree), ctree, 1, "gtk_ctree_expand_recursive", "GtkCTree*");
  XEN_ASSERT_TYPE(XEN_GtkCTreeNode__P(node), node, 2, "gtk_ctree_expand_recursive", "GtkCTreeNode*");
  gtk_ctree_expand_recursive(XEN_TO_C_GtkCTree_(ctree), XEN_TO_C_GtkCTreeNode_(node));
  return(XEN_FALSE);
}

static XEN gxg_gtk_ctree_expand_to_depth(XEN ctree, XEN node, XEN depth)
{
  #define H_gtk_ctree_expand_to_depth "void gtk_ctree_expand_to_depth(GtkCTree* ctree, GtkCTreeNode* node, \
gint depth)"
  XEN_ASSERT_TYPE(XEN_GtkCTree__P(ctree), ctree, 1, "gtk_ctree_expand_to_depth", "GtkCTree*");
  XEN_ASSERT_TYPE(XEN_GtkCTreeNode__P(node), node, 2, "gtk_ctree_expand_to_depth", "GtkCTreeNode*");
  XEN_ASSERT_TYPE(XEN_gint_P(depth), depth, 3, "gtk_ctree_expand_to_depth", "gint");
  gtk_ctree_expand_to_depth(XEN_TO_C_GtkCTree_(ctree), XEN_TO_C_GtkCTreeNode_(node), XEN_TO_C_gint(depth));
  return(XEN_FALSE);
}

static XEN gxg_gtk_ctree_collapse(XEN ctree, XEN node)
{
  #define H_gtk_ctree_collapse "void gtk_ctree_collapse(GtkCTree* ctree, GtkCTreeNode* node)"
  XEN_ASSERT_TYPE(XEN_GtkCTree__P(ctree), ctree, 1, "gtk_ctree_collapse", "GtkCTree*");
  XEN_ASSERT_TYPE(XEN_GtkCTreeNode__P(node), node, 2, "gtk_ctree_collapse", "GtkCTreeNode*");
  gtk_ctree_collapse(XEN_TO_C_GtkCTree_(ctree), XEN_TO_C_GtkCTreeNode_(node));
  return(XEN_FALSE);
}

static XEN gxg_gtk_ctree_collapse_recursive(XEN ctree, XEN node)
{
  #define H_gtk_ctree_collapse_recursive "void gtk_ctree_collapse_recursive(GtkCTree* ctree, GtkCTreeNode* node)"
  XEN_ASSERT_TYPE(XEN_GtkCTree__P(ctree), ctree, 1, "gtk_ctree_collapse_recursive", "GtkCTree*");
  XEN_ASSERT_TYPE(XEN_GtkCTreeNode__P(node), node, 2, "gtk_ctree_collapse_recursive", "GtkCTreeNode*");
  gtk_ctree_collapse_recursive(XEN_TO_C_GtkCTree_(ctree), XEN_TO_C_GtkCTreeNode_(node));
  return(XEN_FALSE);
}

static XEN gxg_gtk_ctree_collapse_to_depth(XEN ctree, XEN node, XEN depth)
{
  #define H_gtk_ctree_collapse_to_depth "void gtk_ctree_collapse_to_depth(GtkCTree* ctree, GtkCTreeNode* node, \
gint depth)"
  XEN_ASSERT_TYPE(XEN_GtkCTree__P(ctree), ctree, 1, "gtk_ctree_collapse_to_depth", "GtkCTree*");
  XEN_ASSERT_TYPE(XEN_GtkCTreeNode__P(node), node, 2, "gtk_ctree_collapse_to_depth", "GtkCTreeNode*");
  XEN_ASSERT_TYPE(XEN_gint_P(depth), depth, 3, "gtk_ctree_collapse_to_depth", "gint");
  gtk_ctree_collapse_to_depth(XEN_TO_C_GtkCTree_(ctree), XEN_TO_C_GtkCTreeNode_(node), XEN_TO_C_gint(depth));
  return(XEN_FALSE);
}

static XEN gxg_gtk_ctree_toggle_expansion(XEN ctree, XEN node)
{
  #define H_gtk_ctree_toggle_expansion "void gtk_ctree_toggle_expansion(GtkCTree* ctree, GtkCTreeNode* node)"
  XEN_ASSERT_TYPE(XEN_GtkCTree__P(ctree), ctree, 1, "gtk_ctree_toggle_expansion", "GtkCTree*");
  XEN_ASSERT_TYPE(XEN_GtkCTreeNode__P(node), node, 2, "gtk_ctree_toggle_expansion", "GtkCTreeNode*");
  gtk_ctree_toggle_expansion(XEN_TO_C_GtkCTree_(ctree), XEN_TO_C_GtkCTreeNode_(node));
  return(XEN_FALSE);
}

static XEN gxg_gtk_ctree_toggle_expansion_recursive(XEN ctree, XEN node)
{
  #define H_gtk_ctree_toggle_expansion_recursive "void gtk_ctree_toggle_expansion_recursive(GtkCTree* ctree, \
GtkCTreeNode* node)"
  XEN_ASSERT_TYPE(XEN_GtkCTree__P(ctree), ctree, 1, "gtk_ctree_toggle_expansion_recursive", "GtkCTree*");
  XEN_ASSERT_TYPE(XEN_GtkCTreeNode__P(node), node, 2, "gtk_ctree_toggle_expansion_recursive", "GtkCTreeNode*");
  gtk_ctree_toggle_expansion_recursive(XEN_TO_C_GtkCTree_(ctree), XEN_TO_C_GtkCTreeNode_(node));
  return(XEN_FALSE);
}

static XEN gxg_gtk_ctree_select(XEN ctree, XEN node)
{
  #define H_gtk_ctree_select "void gtk_ctree_select(GtkCTree* ctree, GtkCTreeNode* node)"
  XEN_ASSERT_TYPE(XEN_GtkCTree__P(ctree), ctree, 1, "gtk_ctree_select", "GtkCTree*");
  XEN_ASSERT_TYPE(XEN_GtkCTreeNode__P(node), node, 2, "gtk_ctree_select", "GtkCTreeNode*");
  gtk_ctree_select(XEN_TO_C_GtkCTree_(ctree), XEN_TO_C_GtkCTreeNode_(node));
  return(XEN_FALSE);
}

static XEN gxg_gtk_ctree_select_recursive(XEN ctree, XEN node)
{
  #define H_gtk_ctree_select_recursive "void gtk_ctree_select_recursive(GtkCTree* ctree, GtkCTreeNode* node)"
  XEN_ASSERT_TYPE(XEN_GtkCTree__P(ctree), ctree, 1, "gtk_ctree_select_recursive", "GtkCTree*");
  XEN_ASSERT_TYPE(XEN_GtkCTreeNode__P(node), node, 2, "gtk_ctree_select_recursive", "GtkCTreeNode*");
  gtk_ctree_select_recursive(XEN_TO_C_GtkCTree_(ctree), XEN_TO_C_GtkCTreeNode_(node));
  return(XEN_FALSE);
}

static XEN gxg_gtk_ctree_unselect(XEN ctree, XEN node)
{
  #define H_gtk_ctree_unselect "void gtk_ctree_unselect(GtkCTree* ctree, GtkCTreeNode* node)"
  XEN_ASSERT_TYPE(XEN_GtkCTree__P(ctree), ctree, 1, "gtk_ctree_unselect", "GtkCTree*");
  XEN_ASSERT_TYPE(XEN_GtkCTreeNode__P(node), node, 2, "gtk_ctree_unselect", "GtkCTreeNode*");
  gtk_ctree_unselect(XEN_TO_C_GtkCTree_(ctree), XEN_TO_C_GtkCTreeNode_(node));
  return(XEN_FALSE);
}

static XEN gxg_gtk_ctree_unselect_recursive(XEN ctree, XEN node)
{
  #define H_gtk_ctree_unselect_recursive "void gtk_ctree_unselect_recursive(GtkCTree* ctree, GtkCTreeNode* node)"
  XEN_ASSERT_TYPE(XEN_GtkCTree__P(ctree), ctree, 1, "gtk_ctree_unselect_recursive", "GtkCTree*");
  XEN_ASSERT_TYPE(XEN_GtkCTreeNode__P(node), node, 2, "gtk_ctree_unselect_recursive", "GtkCTreeNode*");
  gtk_ctree_unselect_recursive(XEN_TO_C_GtkCTree_(ctree), XEN_TO_C_GtkCTreeNode_(node));
  return(XEN_FALSE);
}

static XEN gxg_gtk_ctree_real_select_recursive(XEN ctree, XEN node, XEN state)
{
  #define H_gtk_ctree_real_select_recursive "void gtk_ctree_real_select_recursive(GtkCTree* ctree, GtkCTreeNode* node, \
gint state)"
  XEN_ASSERT_TYPE(XEN_GtkCTree__P(ctree), ctree, 1, "gtk_ctree_real_select_recursive", "GtkCTree*");
  XEN_ASSERT_TYPE(XEN_GtkCTreeNode__P(node), node, 2, "gtk_ctree_real_select_recursive", "GtkCTreeNode*");
  XEN_ASSERT_TYPE(XEN_gint_P(state), state, 3, "gtk_ctree_real_select_recursive", "gint");
  gtk_ctree_real_select_recursive(XEN_TO_C_GtkCTree_(ctree), XEN_TO_C_GtkCTreeNode_(node), XEN_TO_C_gint(state));
  return(XEN_FALSE);
}

static XEN gxg_gtk_ctree_node_set_text(XEN ctree, XEN node, XEN column, XEN text)
{
  #define H_gtk_ctree_node_set_text "void gtk_ctree_node_set_text(GtkCTree* ctree, GtkCTreeNode* node, \
gint column, gchar* text)"
  XEN_ASSERT_TYPE(XEN_GtkCTree__P(ctree), ctree, 1, "gtk_ctree_node_set_text", "GtkCTree*");
  XEN_ASSERT_TYPE(XEN_GtkCTreeNode__P(node), node, 2, "gtk_ctree_node_set_text", "GtkCTreeNode*");
  XEN_ASSERT_TYPE(XEN_gint_P(column), column, 3, "gtk_ctree_node_set_text", "gint");
  XEN_ASSERT_TYPE(XEN_gchar__P(text), text, 4, "gtk_ctree_node_set_text", "gchar*");
  gtk_ctree_node_set_text(XEN_TO_C_GtkCTree_(ctree), XEN_TO_C_GtkCTreeNode_(node), XEN_TO_C_gint(column), XEN_TO_C_gchar_(text));
  return(XEN_FALSE);
}

static XEN gxg_gtk_ctree_node_set_pixmap(XEN ctree, XEN node, XEN column, XEN pixmap, XEN mask)
{
  #define H_gtk_ctree_node_set_pixmap "void gtk_ctree_node_set_pixmap(GtkCTree* ctree, GtkCTreeNode* node, \
gint column, GdkPixmap* pixmap, GdkBitmap* mask)"
  XEN_ASSERT_TYPE(XEN_GtkCTree__P(ctree), ctree, 1, "gtk_ctree_node_set_pixmap", "GtkCTree*");
  XEN_ASSERT_TYPE(XEN_GtkCTreeNode__P(node), node, 2, "gtk_ctree_node_set_pixmap", "GtkCTreeNode*");
  XEN_ASSERT_TYPE(XEN_gint_P(column), column, 3, "gtk_ctree_node_set_pixmap", "gint");
  XEN_ASSERT_TYPE(XEN_GdkPixmap__P(pixmap), pixmap, 4, "gtk_ctree_node_set_pixmap", "GdkPixmap*");
  XEN_ASSERT_TYPE(XEN_GdkBitmap__P(mask), mask, 5, "gtk_ctree_node_set_pixmap", "GdkBitmap*");
  gtk_ctree_node_set_pixmap(XEN_TO_C_GtkCTree_(ctree), XEN_TO_C_GtkCTreeNode_(node), XEN_TO_C_gint(column), XEN_TO_C_GdkPixmap_(pixmap), 
                            XEN_TO_C_GdkBitmap_(mask));
  return(XEN_FALSE);
}

static XEN gxg_gtk_ctree_node_set_pixtext(XEN ctree, XEN node, XEN column, XEN text, XEN spacing, XEN pixmap, XEN mask)
{
  #define H_gtk_ctree_node_set_pixtext "void gtk_ctree_node_set_pixtext(GtkCTree* ctree, GtkCTreeNode* node, \
gint column, gchar* text, guint8 spacing, GdkPixmap* pixmap, GdkBitmap* mask)"
  XEN_ASSERT_TYPE(XEN_GtkCTree__P(ctree), ctree, 1, "gtk_ctree_node_set_pixtext", "GtkCTree*");
  XEN_ASSERT_TYPE(XEN_GtkCTreeNode__P(node), node, 2, "gtk_ctree_node_set_pixtext", "GtkCTreeNode*");
  XEN_ASSERT_TYPE(XEN_gint_P(column), column, 3, "gtk_ctree_node_set_pixtext", "gint");
  XEN_ASSERT_TYPE(XEN_gchar__P(text), text, 4, "gtk_ctree_node_set_pixtext", "gchar*");
  XEN_ASSERT_TYPE(XEN_guint8_P(spacing), spacing, 5, "gtk_ctree_node_set_pixtext", "guint8");
  XEN_ASSERT_TYPE(XEN_GdkPixmap__P(pixmap), pixmap, 6, "gtk_ctree_node_set_pixtext", "GdkPixmap*");
  XEN_ASSERT_TYPE(XEN_GdkBitmap__P(mask), mask, 7, "gtk_ctree_node_set_pixtext", "GdkBitmap*");
  gtk_ctree_node_set_pixtext(XEN_TO_C_GtkCTree_(ctree), XEN_TO_C_GtkCTreeNode_(node), XEN_TO_C_gint(column), XEN_TO_C_gchar_(text), 
                             XEN_TO_C_guint8(spacing), XEN_TO_C_GdkPixmap_(pixmap), XEN_TO_C_GdkBitmap_(mask));
  return(XEN_FALSE);
}

static XEN gxg_gtk_ctree_set_node_info(XEN arglist)
{
  #define H_gtk_ctree_set_node_info "void gtk_ctree_set_node_info(GtkCTree* ctree, GtkCTreeNode* node, \
gchar* text, guint8 spacing, GdkPixmap* pixmap_closed, GdkBitmap* mask_closed, GdkPixmap* pixmap_opened, \
GdkBitmap* mask_opened, gboolean is_leaf, gboolean expanded)"
  XEN ctree, node, text, spacing, pixmap_closed, mask_closed, pixmap_opened, mask_opened, is_leaf, expanded;
  ctree = XEN_LIST_REF(arglist, 0);
  node = XEN_LIST_REF(arglist, 1);
  text = XEN_LIST_REF(arglist, 2);
  spacing = XEN_LIST_REF(arglist, 3);
  pixmap_closed = XEN_LIST_REF(arglist, 4);
  mask_closed = XEN_LIST_REF(arglist, 5);
  pixmap_opened = XEN_LIST_REF(arglist, 6);
  mask_opened = XEN_LIST_REF(arglist, 7);
  is_leaf = XEN_LIST_REF(arglist, 8);
  expanded = XEN_LIST_REF(arglist, 9);
  XEN_ASSERT_TYPE(XEN_GtkCTree__P(ctree), ctree, 1, "gtk_ctree_set_node_info", "GtkCTree*");
  XEN_ASSERT_TYPE(XEN_GtkCTreeNode__P(node), node, 2, "gtk_ctree_set_node_info", "GtkCTreeNode*");
  XEN_ASSERT_TYPE(XEN_gchar__P(text), text, 3, "gtk_ctree_set_node_info", "gchar*");
  XEN_ASSERT_TYPE(XEN_guint8_P(spacing), spacing, 4, "gtk_ctree_set_node_info", "guint8");
  XEN_ASSERT_TYPE(XEN_GdkPixmap__P(pixmap_closed), pixmap_closed, 5, "gtk_ctree_set_node_info", "GdkPixmap*");
  XEN_ASSERT_TYPE(XEN_GdkBitmap__P(mask_closed), mask_closed, 6, "gtk_ctree_set_node_info", "GdkBitmap*");
  XEN_ASSERT_TYPE(XEN_GdkPixmap__P(pixmap_opened), pixmap_opened, 7, "gtk_ctree_set_node_info", "GdkPixmap*");
  XEN_ASSERT_TYPE(XEN_GdkBitmap__P(mask_opened), mask_opened, 8, "gtk_ctree_set_node_info", "GdkBitmap*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(is_leaf), is_leaf, 9, "gtk_ctree_set_node_info", "gboolean");
  XEN_ASSERT_TYPE(XEN_gboolean_P(expanded), expanded, 10, "gtk_ctree_set_node_info", "gboolean");
  gtk_ctree_set_node_info(XEN_TO_C_GtkCTree_(ctree), XEN_TO_C_GtkCTreeNode_(node), XEN_TO_C_gchar_(text), XEN_TO_C_guint8(spacing), 
                          XEN_TO_C_GdkPixmap_(pixmap_closed), XEN_TO_C_GdkBitmap_(mask_closed), XEN_TO_C_GdkPixmap_(pixmap_opened), 
                          XEN_TO_C_GdkBitmap_(mask_opened), XEN_TO_C_gboolean(is_leaf), XEN_TO_C_gboolean(expanded));
  return(XEN_FALSE);
}

static XEN gxg_gtk_ctree_node_set_shift(XEN ctree, XEN node, XEN column, XEN vertical, XEN horizontal)
{
  #define H_gtk_ctree_node_set_shift "void gtk_ctree_node_set_shift(GtkCTree* ctree, GtkCTreeNode* node, \
gint column, gint vertical, gint horizontal)"
  XEN_ASSERT_TYPE(XEN_GtkCTree__P(ctree), ctree, 1, "gtk_ctree_node_set_shift", "GtkCTree*");
  XEN_ASSERT_TYPE(XEN_GtkCTreeNode__P(node), node, 2, "gtk_ctree_node_set_shift", "GtkCTreeNode*");
  XEN_ASSERT_TYPE(XEN_gint_P(column), column, 3, "gtk_ctree_node_set_shift", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(vertical), vertical, 4, "gtk_ctree_node_set_shift", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(horizontal), horizontal, 5, "gtk_ctree_node_set_shift", "gint");
  gtk_ctree_node_set_shift(XEN_TO_C_GtkCTree_(ctree), XEN_TO_C_GtkCTreeNode_(node), XEN_TO_C_gint(column), XEN_TO_C_gint(vertical), 
                           XEN_TO_C_gint(horizontal));
  return(XEN_FALSE);
}

static XEN gxg_gtk_ctree_node_set_selectable(XEN ctree, XEN node, XEN selectable)
{
  #define H_gtk_ctree_node_set_selectable "void gtk_ctree_node_set_selectable(GtkCTree* ctree, GtkCTreeNode* node, \
gboolean selectable)"
  XEN_ASSERT_TYPE(XEN_GtkCTree__P(ctree), ctree, 1, "gtk_ctree_node_set_selectable", "GtkCTree*");
  XEN_ASSERT_TYPE(XEN_GtkCTreeNode__P(node), node, 2, "gtk_ctree_node_set_selectable", "GtkCTreeNode*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(selectable), selectable, 3, "gtk_ctree_node_set_selectable", "gboolean");
  gtk_ctree_node_set_selectable(XEN_TO_C_GtkCTree_(ctree), XEN_TO_C_GtkCTreeNode_(node), XEN_TO_C_gboolean(selectable));
  return(XEN_FALSE);
}

static XEN gxg_gtk_ctree_node_get_selectable(XEN ctree, XEN node)
{
  #define H_gtk_ctree_node_get_selectable "gboolean gtk_ctree_node_get_selectable(GtkCTree* ctree, GtkCTreeNode* node)"
  XEN_ASSERT_TYPE(XEN_GtkCTree__P(ctree), ctree, 1, "gtk_ctree_node_get_selectable", "GtkCTree*");
  XEN_ASSERT_TYPE(XEN_GtkCTreeNode__P(node), node, 2, "gtk_ctree_node_get_selectable", "GtkCTreeNode*");
  return(C_TO_XEN_gboolean(gtk_ctree_node_get_selectable(XEN_TO_C_GtkCTree_(ctree), XEN_TO_C_GtkCTreeNode_(node))));
}

static XEN gxg_gtk_ctree_node_get_cell_type(XEN ctree, XEN node, XEN column)
{
  #define H_gtk_ctree_node_get_cell_type "GtkCellType gtk_ctree_node_get_cell_type(GtkCTree* ctree, GtkCTreeNode* node, \
gint column)"
  XEN_ASSERT_TYPE(XEN_GtkCTree__P(ctree), ctree, 1, "gtk_ctree_node_get_cell_type", "GtkCTree*");
  XEN_ASSERT_TYPE(XEN_GtkCTreeNode__P(node), node, 2, "gtk_ctree_node_get_cell_type", "GtkCTreeNode*");
  XEN_ASSERT_TYPE(XEN_gint_P(column), column, 3, "gtk_ctree_node_get_cell_type", "gint");
  return(C_TO_XEN_GtkCellType(gtk_ctree_node_get_cell_type(XEN_TO_C_GtkCTree_(ctree), XEN_TO_C_GtkCTreeNode_(node), XEN_TO_C_gint(column))));
}

static XEN gxg_gtk_ctree_node_get_text(XEN ctree, XEN node, XEN column, XEN text)
{
  #define H_gtk_ctree_node_get_text "gboolean gtk_ctree_node_get_text(GtkCTree* ctree, GtkCTreeNode* node, \
gint column, gchar** [text])"
  gchar* ref_text;
  XEN_ASSERT_TYPE(XEN_GtkCTree__P(ctree), ctree, 1, "gtk_ctree_node_get_text", "GtkCTree*");
  XEN_ASSERT_TYPE(XEN_GtkCTreeNode__P(node), node, 2, "gtk_ctree_node_get_text", "GtkCTreeNode*");
  XEN_ASSERT_TYPE(XEN_gint_P(column), column, 3, "gtk_ctree_node_get_text", "gint");
  {
    XEN result = XEN_FALSE;
    result = C_TO_XEN_gboolean(gtk_ctree_node_get_text(XEN_TO_C_GtkCTree_(ctree), XEN_TO_C_GtkCTreeNode_(node), XEN_TO_C_gint(column), 
                                                       &ref_text));
    return(XEN_LIST_2(result, C_TO_XEN_gchar_(ref_text)));
   }
}

static XEN gxg_gtk_ctree_node_get_pixmap(XEN ctree, XEN node, XEN column, XEN pixmap, XEN mask)
{
  #define H_gtk_ctree_node_get_pixmap "gboolean gtk_ctree_node_get_pixmap(GtkCTree* ctree, GtkCTreeNode* node, \
gint column, GdkPixmap** [pixmap], GdkBitmap** [mask])"
  GdkPixmap* ref_pixmap;
  GdkBitmap* ref_mask;
  XEN_ASSERT_TYPE(XEN_GtkCTree__P(ctree), ctree, 1, "gtk_ctree_node_get_pixmap", "GtkCTree*");
  XEN_ASSERT_TYPE(XEN_GtkCTreeNode__P(node), node, 2, "gtk_ctree_node_get_pixmap", "GtkCTreeNode*");
  XEN_ASSERT_TYPE(XEN_gint_P(column), column, 3, "gtk_ctree_node_get_pixmap", "gint");
  {
    XEN result = XEN_FALSE;
    result = C_TO_XEN_gboolean(gtk_ctree_node_get_pixmap(XEN_TO_C_GtkCTree_(ctree), XEN_TO_C_GtkCTreeNode_(node), XEN_TO_C_gint(column), 
                                                         &ref_pixmap, &ref_mask));
    return(XEN_LIST_3(result, C_TO_XEN_GdkPixmap_(ref_pixmap), C_TO_XEN_GdkBitmap_(ref_mask)));
   }
}

static XEN gxg_gtk_ctree_node_get_pixtext(XEN ctree, XEN node, XEN column, XEN text, XEN spacing, XEN pixmap, XEN mask)
{
  #define H_gtk_ctree_node_get_pixtext "gboolean gtk_ctree_node_get_pixtext(GtkCTree* ctree, GtkCTreeNode* node, \
gint column, gchar** [text], guint8* [spacing], GdkPixmap** [pixmap], GdkBitmap** [mask])"
  gchar* ref_text;
  guint8 ref_spacing;
  GdkPixmap* ref_pixmap;
  GdkBitmap* ref_mask;
  XEN_ASSERT_TYPE(XEN_GtkCTree__P(ctree), ctree, 1, "gtk_ctree_node_get_pixtext", "GtkCTree*");
  XEN_ASSERT_TYPE(XEN_GtkCTreeNode__P(node), node, 2, "gtk_ctree_node_get_pixtext", "GtkCTreeNode*");
  XEN_ASSERT_TYPE(XEN_gint_P(column), column, 3, "gtk_ctree_node_get_pixtext", "gint");
  {
    XEN result = XEN_FALSE;
    result = C_TO_XEN_gboolean(gtk_ctree_node_get_pixtext(XEN_TO_C_GtkCTree_(ctree), XEN_TO_C_GtkCTreeNode_(node), XEN_TO_C_gint(column), 
                                                          &ref_text, &ref_spacing, &ref_pixmap, &ref_mask));
    return(XEN_LIST_5(result, C_TO_XEN_gchar_(ref_text), C_TO_XEN_guint8(ref_spacing), C_TO_XEN_GdkPixmap_(ref_pixmap), C_TO_XEN_GdkBitmap_(ref_mask)));
   }
}

static XEN gxg_gtk_ctree_get_node_info(XEN arglist)
{
  #define H_gtk_ctree_get_node_info "gboolean gtk_ctree_get_node_info(GtkCTree* ctree, GtkCTreeNode* node, \
gchar** [text], guint8* [spacing], GdkPixmap** [pixmap_closed], GdkBitmap** [mask_closed], GdkPixmap** [pixmap_opened], \
GdkBitmap** [mask_opened], gboolean* [is_leaf], gboolean* [expanded])"
  gchar* ref_text;
  guint8 ref_spacing;
  GdkPixmap* ref_pixmap_closed;
  GdkBitmap* ref_mask_closed;
  GdkPixmap* ref_pixmap_opened;
  GdkBitmap* ref_mask_opened;
  gboolean ref_is_leaf;
  gboolean ref_expanded;
  XEN ctree, node;
  ctree = XEN_LIST_REF(arglist, 0);
  node = XEN_LIST_REF(arglist, 1);
  XEN_ASSERT_TYPE(XEN_GtkCTree__P(ctree), ctree, 1, "gtk_ctree_get_node_info", "GtkCTree*");
  XEN_ASSERT_TYPE(XEN_GtkCTreeNode__P(node), node, 2, "gtk_ctree_get_node_info", "GtkCTreeNode*");
  {
    XEN result = XEN_FALSE;
    result = C_TO_XEN_gboolean(gtk_ctree_get_node_info(XEN_TO_C_GtkCTree_(ctree), XEN_TO_C_GtkCTreeNode_(node), &ref_text, 
                                                       &ref_spacing, &ref_pixmap_closed, &ref_mask_closed, &ref_pixmap_opened, 
                                                       &ref_mask_opened, &ref_is_leaf, &ref_expanded));
    return(XEN_LIST_9(result, C_TO_XEN_gchar_(ref_text), C_TO_XEN_guint8(ref_spacing), C_TO_XEN_GdkPixmap_(ref_pixmap_closed), C_TO_XEN_GdkBitmap_(ref_mask_closed), C_TO_XEN_GdkPixmap_(ref_pixmap_opened), C_TO_XEN_GdkBitmap_(ref_mask_opened), C_TO_XEN_gboolean(ref_is_leaf), C_TO_XEN_gboolean(ref_expanded)));
   }
}

static XEN gxg_gtk_ctree_node_set_row_style(XEN ctree, XEN node, XEN style)
{
  #define H_gtk_ctree_node_set_row_style "void gtk_ctree_node_set_row_style(GtkCTree* ctree, GtkCTreeNode* node, \
GtkStyle* style)"
  XEN_ASSERT_TYPE(XEN_GtkCTree__P(ctree), ctree, 1, "gtk_ctree_node_set_row_style", "GtkCTree*");
  XEN_ASSERT_TYPE(XEN_GtkCTreeNode__P(node), node, 2, "gtk_ctree_node_set_row_style", "GtkCTreeNode*");
  XEN_ASSERT_TYPE(XEN_GtkStyle__P(style), style, 3, "gtk_ctree_node_set_row_style", "GtkStyle*");
  gtk_ctree_node_set_row_style(XEN_TO_C_GtkCTree_(ctree), XEN_TO_C_GtkCTreeNode_(node), XEN_TO_C_GtkStyle_(style));
  return(XEN_FALSE);
}

static XEN gxg_gtk_ctree_node_get_row_style(XEN ctree, XEN node)
{
  #define H_gtk_ctree_node_get_row_style "GtkStyle* gtk_ctree_node_get_row_style(GtkCTree* ctree, GtkCTreeNode* node)"
  XEN_ASSERT_TYPE(XEN_GtkCTree__P(ctree), ctree, 1, "gtk_ctree_node_get_row_style", "GtkCTree*");
  XEN_ASSERT_TYPE(XEN_GtkCTreeNode__P(node), node, 2, "gtk_ctree_node_get_row_style", "GtkCTreeNode*");
  return(C_TO_XEN_GtkStyle_(gtk_ctree_node_get_row_style(XEN_TO_C_GtkCTree_(ctree), XEN_TO_C_GtkCTreeNode_(node))));
}

static XEN gxg_gtk_ctree_node_set_cell_style(XEN ctree, XEN node, XEN column, XEN style)
{
  #define H_gtk_ctree_node_set_cell_style "void gtk_ctree_node_set_cell_style(GtkCTree* ctree, GtkCTreeNode* node, \
gint column, GtkStyle* style)"
  XEN_ASSERT_TYPE(XEN_GtkCTree__P(ctree), ctree, 1, "gtk_ctree_node_set_cell_style", "GtkCTree*");
  XEN_ASSERT_TYPE(XEN_GtkCTreeNode__P(node), node, 2, "gtk_ctree_node_set_cell_style", "GtkCTreeNode*");
  XEN_ASSERT_TYPE(XEN_gint_P(column), column, 3, "gtk_ctree_node_set_cell_style", "gint");
  XEN_ASSERT_TYPE(XEN_GtkStyle__P(style), style, 4, "gtk_ctree_node_set_cell_style", "GtkStyle*");
  gtk_ctree_node_set_cell_style(XEN_TO_C_GtkCTree_(ctree), XEN_TO_C_GtkCTreeNode_(node), XEN_TO_C_gint(column), XEN_TO_C_GtkStyle_(style));
  return(XEN_FALSE);
}

static XEN gxg_gtk_ctree_node_get_cell_style(XEN ctree, XEN node, XEN column)
{
  #define H_gtk_ctree_node_get_cell_style "GtkStyle* gtk_ctree_node_get_cell_style(GtkCTree* ctree, GtkCTreeNode* node, \
gint column)"
  XEN_ASSERT_TYPE(XEN_GtkCTree__P(ctree), ctree, 1, "gtk_ctree_node_get_cell_style", "GtkCTree*");
  XEN_ASSERT_TYPE(XEN_GtkCTreeNode__P(node), node, 2, "gtk_ctree_node_get_cell_style", "GtkCTreeNode*");
  XEN_ASSERT_TYPE(XEN_gint_P(column), column, 3, "gtk_ctree_node_get_cell_style", "gint");
  return(C_TO_XEN_GtkStyle_(gtk_ctree_node_get_cell_style(XEN_TO_C_GtkCTree_(ctree), XEN_TO_C_GtkCTreeNode_(node), XEN_TO_C_gint(column))));
}

static XEN gxg_gtk_ctree_node_set_foreground(XEN ctree, XEN node, XEN color)
{
  #define H_gtk_ctree_node_set_foreground "void gtk_ctree_node_set_foreground(GtkCTree* ctree, GtkCTreeNode* node, \
GdkColor* color)"
  XEN_ASSERT_TYPE(XEN_GtkCTree__P(ctree), ctree, 1, "gtk_ctree_node_set_foreground", "GtkCTree*");
  XEN_ASSERT_TYPE(XEN_GtkCTreeNode__P(node), node, 2, "gtk_ctree_node_set_foreground", "GtkCTreeNode*");
  XEN_ASSERT_TYPE(XEN_GdkColor__P(color), color, 3, "gtk_ctree_node_set_foreground", "GdkColor*");
  gtk_ctree_node_set_foreground(XEN_TO_C_GtkCTree_(ctree), XEN_TO_C_GtkCTreeNode_(node), XEN_TO_C_GdkColor_(color));
  return(XEN_FALSE);
}

static XEN gxg_gtk_ctree_node_set_background(XEN ctree, XEN node, XEN color)
{
  #define H_gtk_ctree_node_set_background "void gtk_ctree_node_set_background(GtkCTree* ctree, GtkCTreeNode* node, \
GdkColor* color)"
  XEN_ASSERT_TYPE(XEN_GtkCTree__P(ctree), ctree, 1, "gtk_ctree_node_set_background", "GtkCTree*");
  XEN_ASSERT_TYPE(XEN_GtkCTreeNode__P(node), node, 2, "gtk_ctree_node_set_background", "GtkCTreeNode*");
  XEN_ASSERT_TYPE(XEN_GdkColor__P(color), color, 3, "gtk_ctree_node_set_background", "GdkColor*");
  gtk_ctree_node_set_background(XEN_TO_C_GtkCTree_(ctree), XEN_TO_C_GtkCTreeNode_(node), XEN_TO_C_GdkColor_(color));
  return(XEN_FALSE);
}

static XEN gxg_gtk_ctree_node_set_row_data(XEN ctree, XEN node, XEN data)
{
  #define H_gtk_ctree_node_set_row_data "void gtk_ctree_node_set_row_data(GtkCTree* ctree, GtkCTreeNode* node, \
gpointer data)"
  XEN_ASSERT_TYPE(XEN_GtkCTree__P(ctree), ctree, 1, "gtk_ctree_node_set_row_data", "GtkCTree*");
  XEN_ASSERT_TYPE(XEN_GtkCTreeNode__P(node), node, 2, "gtk_ctree_node_set_row_data", "GtkCTreeNode*");
  XEN_ASSERT_TYPE(XEN_gpointer_P(data), data, 3, "gtk_ctree_node_set_row_data", "gpointer");
  gtk_ctree_node_set_row_data(XEN_TO_C_GtkCTree_(ctree), XEN_TO_C_GtkCTreeNode_(node), XEN_TO_C_gpointer(data));
  return(XEN_FALSE);
}

static XEN gxg_gtk_ctree_node_set_row_data_full(XEN ctree, XEN node, XEN func_data, XEN destroy)
{
  #define H_gtk_ctree_node_set_row_data_full "void gtk_ctree_node_set_row_data_full(GtkCTree* ctree, \
GtkCTreeNode* node, gpointer func_data, GtkDestroyNotify destroy)"
  XEN_ASSERT_TYPE(XEN_GtkCTree__P(ctree), ctree, 1, "gtk_ctree_node_set_row_data_full", "GtkCTree*");
  XEN_ASSERT_TYPE(XEN_GtkCTreeNode__P(node), node, 2, "gtk_ctree_node_set_row_data_full", "GtkCTreeNode*");
  XEN_ASSERT_TYPE(XEN_gpointer_P(func_data), func_data, 3, "gtk_ctree_node_set_row_data_full", "gpointer");
  XEN_ASSERT_TYPE(XEN_GtkDestroyNotify_P(destroy), destroy, 4, "gtk_ctree_node_set_row_data_full", "GtkDestroyNotify");
  {
    XEN gxg_ptr = XEN_LIST_5(XEN_FALSE, func_data, XEN_FALSE, XEN_FALSE, XEN_FALSE);
    xm_protect(gxg_ptr);
    XEN_LIST_SET(gxg_ptr, 3, destroy);
    gtk_ctree_node_set_row_data_full(XEN_TO_C_GtkCTree_(ctree), XEN_TO_C_GtkCTreeNode_(node), XEN_TO_C_gpointer(func_data), XEN_TO_C_GtkDestroyNotify(destroy));
    return(XEN_FALSE);
   }
}

static XEN gxg_gtk_ctree_node_get_row_data(XEN ctree, XEN node)
{
  #define H_gtk_ctree_node_get_row_data "gpointer gtk_ctree_node_get_row_data(GtkCTree* ctree, GtkCTreeNode* node)"
  XEN_ASSERT_TYPE(XEN_GtkCTree__P(ctree), ctree, 1, "gtk_ctree_node_get_row_data", "GtkCTree*");
  XEN_ASSERT_TYPE(XEN_GtkCTreeNode__P(node), node, 2, "gtk_ctree_node_get_row_data", "GtkCTreeNode*");
  return(C_TO_XEN_gpointer(gtk_ctree_node_get_row_data(XEN_TO_C_GtkCTree_(ctree), XEN_TO_C_GtkCTreeNode_(node))));
}

static XEN gxg_gtk_ctree_node_moveto(XEN ctree, XEN node, XEN column, XEN row_align, XEN col_align)
{
  #define H_gtk_ctree_node_moveto "void gtk_ctree_node_moveto(GtkCTree* ctree, GtkCTreeNode* node, gint column, \
gfloat row_align, gfloat col_align)"
  XEN_ASSERT_TYPE(XEN_GtkCTree__P(ctree), ctree, 1, "gtk_ctree_node_moveto", "GtkCTree*");
  XEN_ASSERT_TYPE(XEN_GtkCTreeNode__P(node), node, 2, "gtk_ctree_node_moveto", "GtkCTreeNode*");
  XEN_ASSERT_TYPE(XEN_gint_P(column), column, 3, "gtk_ctree_node_moveto", "gint");
  XEN_ASSERT_TYPE(XEN_gfloat_P(row_align), row_align, 4, "gtk_ctree_node_moveto", "gfloat");
  XEN_ASSERT_TYPE(XEN_gfloat_P(col_align), col_align, 5, "gtk_ctree_node_moveto", "gfloat");
  gtk_ctree_node_moveto(XEN_TO_C_GtkCTree_(ctree), XEN_TO_C_GtkCTreeNode_(node), XEN_TO_C_gint(column), XEN_TO_C_gfloat(row_align), 
                        XEN_TO_C_gfloat(col_align));
  return(XEN_FALSE);
}

static XEN gxg_gtk_ctree_node_is_visible(XEN ctree, XEN node)
{
  #define H_gtk_ctree_node_is_visible "GtkVisibility gtk_ctree_node_is_visible(GtkCTree* ctree, GtkCTreeNode* node)"
  XEN_ASSERT_TYPE(XEN_GtkCTree__P(ctree), ctree, 1, "gtk_ctree_node_is_visible", "GtkCTree*");
  XEN_ASSERT_TYPE(XEN_GtkCTreeNode__P(node), node, 2, "gtk_ctree_node_is_visible", "GtkCTreeNode*");
  return(C_TO_XEN_GtkVisibility(gtk_ctree_node_is_visible(XEN_TO_C_GtkCTree_(ctree), XEN_TO_C_GtkCTreeNode_(node))));
}

static XEN gxg_gtk_ctree_set_indent(XEN ctree, XEN indent)
{
  #define H_gtk_ctree_set_indent "void gtk_ctree_set_indent(GtkCTree* ctree, gint indent)"
  XEN_ASSERT_TYPE(XEN_GtkCTree__P(ctree), ctree, 1, "gtk_ctree_set_indent", "GtkCTree*");
  XEN_ASSERT_TYPE(XEN_gint_P(indent), indent, 2, "gtk_ctree_set_indent", "gint");
  gtk_ctree_set_indent(XEN_TO_C_GtkCTree_(ctree), XEN_TO_C_gint(indent));
  return(XEN_FALSE);
}

static XEN gxg_gtk_ctree_set_spacing(XEN ctree, XEN spacing)
{
  #define H_gtk_ctree_set_spacing "void gtk_ctree_set_spacing(GtkCTree* ctree, gint spacing)"
  XEN_ASSERT_TYPE(XEN_GtkCTree__P(ctree), ctree, 1, "gtk_ctree_set_spacing", "GtkCTree*");
  XEN_ASSERT_TYPE(XEN_gint_P(spacing), spacing, 2, "gtk_ctree_set_spacing", "gint");
  gtk_ctree_set_spacing(XEN_TO_C_GtkCTree_(ctree), XEN_TO_C_gint(spacing));
  return(XEN_FALSE);
}

static XEN gxg_gtk_ctree_set_show_stub(XEN ctree, XEN show_stub)
{
  #define H_gtk_ctree_set_show_stub "void gtk_ctree_set_show_stub(GtkCTree* ctree, gboolean show_stub)"
  XEN_ASSERT_TYPE(XEN_GtkCTree__P(ctree), ctree, 1, "gtk_ctree_set_show_stub", "GtkCTree*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(show_stub), show_stub, 2, "gtk_ctree_set_show_stub", "gboolean");
  gtk_ctree_set_show_stub(XEN_TO_C_GtkCTree_(ctree), XEN_TO_C_gboolean(show_stub));
  return(XEN_FALSE);
}

static XEN gxg_gtk_ctree_set_line_style(XEN ctree, XEN line_style)
{
  #define H_gtk_ctree_set_line_style "void gtk_ctree_set_line_style(GtkCTree* ctree, GtkCTreeLineStyle line_style)"
  XEN_ASSERT_TYPE(XEN_GtkCTree__P(ctree), ctree, 1, "gtk_ctree_set_line_style", "GtkCTree*");
  XEN_ASSERT_TYPE(XEN_GtkCTreeLineStyle_P(line_style), line_style, 2, "gtk_ctree_set_line_style", "GtkCTreeLineStyle");
  gtk_ctree_set_line_style(XEN_TO_C_GtkCTree_(ctree), XEN_TO_C_GtkCTreeLineStyle(line_style));
  return(XEN_FALSE);
}

static XEN gxg_gtk_ctree_set_expander_style(XEN ctree, XEN expander_style)
{
  #define H_gtk_ctree_set_expander_style "void gtk_ctree_set_expander_style(GtkCTree* ctree, GtkCTreeExpanderStyle expander_style)"
  XEN_ASSERT_TYPE(XEN_GtkCTree__P(ctree), ctree, 1, "gtk_ctree_set_expander_style", "GtkCTree*");
  XEN_ASSERT_TYPE(XEN_GtkCTreeExpanderStyle_P(expander_style), expander_style, 2, "gtk_ctree_set_expander_style", "GtkCTreeExpanderStyle");
  gtk_ctree_set_expander_style(XEN_TO_C_GtkCTree_(ctree), XEN_TO_C_GtkCTreeExpanderStyle(expander_style));
  return(XEN_FALSE);
}

static XEN gxg_gtk_ctree_set_drag_compare_func(XEN ctree, XEN cmp_func)
{
  #define H_gtk_ctree_set_drag_compare_func "void gtk_ctree_set_drag_compare_func(GtkCTree* ctree, GtkCTreeCompareDragFunc cmp_func)"
  XEN_ASSERT_TYPE(XEN_GtkCTree__P(ctree), ctree, 1, "gtk_ctree_set_drag_compare_func", "GtkCTree*");
  XEN_ASSERT_TYPE(XEN_GtkCTreeCompareDragFunc_P(cmp_func), cmp_func, 2, "gtk_ctree_set_drag_compare_func", "GtkCTreeCompareDragFunc");
  gtk_ctree_set_drag_compare_func(XEN_TO_C_GtkCTree_(ctree), XEN_TO_C_GtkCTreeCompareDragFunc(cmp_func));
  return(XEN_FALSE);
}

static XEN gxg_gtk_ctree_sort_node(XEN ctree, XEN node)
{
  #define H_gtk_ctree_sort_node "void gtk_ctree_sort_node(GtkCTree* ctree, GtkCTreeNode* node)"
  XEN_ASSERT_TYPE(XEN_GtkCTree__P(ctree), ctree, 1, "gtk_ctree_sort_node", "GtkCTree*");
  XEN_ASSERT_TYPE(XEN_GtkCTreeNode__P(node), node, 2, "gtk_ctree_sort_node", "GtkCTreeNode*");
  gtk_ctree_sort_node(XEN_TO_C_GtkCTree_(ctree), XEN_TO_C_GtkCTreeNode_(node));
  return(XEN_FALSE);
}

static XEN gxg_gtk_ctree_sort_recursive(XEN ctree, XEN node)
{
  #define H_gtk_ctree_sort_recursive "void gtk_ctree_sort_recursive(GtkCTree* ctree, GtkCTreeNode* node)"
  XEN_ASSERT_TYPE(XEN_GtkCTree__P(ctree), ctree, 1, "gtk_ctree_sort_recursive", "GtkCTree*");
  XEN_ASSERT_TYPE(XEN_GtkCTreeNode__P(node), node, 2, "gtk_ctree_sort_recursive", "GtkCTreeNode*");
  gtk_ctree_sort_recursive(XEN_TO_C_GtkCTree_(ctree), XEN_TO_C_GtkCTreeNode_(node));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_ctree_node_get_type(void)
{
  #define H_gtk_ctree_node_get_type "GType gtk_ctree_node_get_type( void)"
  return(C_TO_XEN_GType(gtk_ctree_node_get_type()));
}
#endif

static XEN gxg_gtk_drawing_area_size(XEN darea, XEN width, XEN height)
{
  #define H_gtk_drawing_area_size "void gtk_drawing_area_size(GtkDrawingArea* darea, gint width, gint height)"
  XEN_ASSERT_TYPE(XEN_GtkDrawingArea__P(darea), darea, 1, "gtk_drawing_area_size", "GtkDrawingArea*");
  XEN_ASSERT_TYPE(XEN_gint_P(width), width, 2, "gtk_drawing_area_size", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(height), height, 3, "gtk_drawing_area_size", "gint");
  gtk_drawing_area_size(XEN_TO_C_GtkDrawingArea_(darea), XEN_TO_C_gint(width), XEN_TO_C_gint(height));
  return(XEN_FALSE);
}

static XEN gxg_gtk_entry_new_with_max_length(XEN max)
{
  #define H_gtk_entry_new_with_max_length "GtkWidget* gtk_entry_new_with_max_length(gint max)"
  XEN_ASSERT_TYPE(XEN_gint_P(max), max, 1, "gtk_entry_new_with_max_length", "gint");
  return(C_TO_XEN_GtkWidget_(gtk_entry_new_with_max_length(XEN_TO_C_gint(max))));
}

static XEN gxg_gtk_entry_append_text(XEN entry, XEN text)
{
  #define H_gtk_entry_append_text "void gtk_entry_append_text(GtkEntry* entry, gchar* text)"
  XEN_ASSERT_TYPE(XEN_GtkEntry__P(entry), entry, 1, "gtk_entry_append_text", "GtkEntry*");
  XEN_ASSERT_TYPE(XEN_gchar__P(text), text, 2, "gtk_entry_append_text", "gchar*");
  gtk_entry_append_text(XEN_TO_C_GtkEntry_(entry), XEN_TO_C_gchar_(text));
  return(XEN_FALSE);
}

static XEN gxg_gtk_entry_prepend_text(XEN entry, XEN text)
{
  #define H_gtk_entry_prepend_text "void gtk_entry_prepend_text(GtkEntry* entry, gchar* text)"
  XEN_ASSERT_TYPE(XEN_GtkEntry__P(entry), entry, 1, "gtk_entry_prepend_text", "GtkEntry*");
  XEN_ASSERT_TYPE(XEN_gchar__P(text), text, 2, "gtk_entry_prepend_text", "gchar*");
  gtk_entry_prepend_text(XEN_TO_C_GtkEntry_(entry), XEN_TO_C_gchar_(text));
  return(XEN_FALSE);
}

static XEN gxg_gtk_entry_set_position(XEN entry, XEN position)
{
  #define H_gtk_entry_set_position "void gtk_entry_set_position(GtkEntry* entry, gint position)"
  XEN_ASSERT_TYPE(XEN_GtkEntry__P(entry), entry, 1, "gtk_entry_set_position", "GtkEntry*");
  XEN_ASSERT_TYPE(XEN_gint_P(position), position, 2, "gtk_entry_set_position", "gint");
  gtk_entry_set_position(XEN_TO_C_GtkEntry_(entry), XEN_TO_C_gint(position));
  return(XEN_FALSE);
}

static XEN gxg_gtk_entry_select_region(XEN entry, XEN start, XEN end)
{
  #define H_gtk_entry_select_region "void gtk_entry_select_region(GtkEntry* entry, gint start, gint end)"
  XEN_ASSERT_TYPE(XEN_GtkEntry__P(entry), entry, 1, "gtk_entry_select_region", "GtkEntry*");
  XEN_ASSERT_TYPE(XEN_gint_P(start), start, 2, "gtk_entry_select_region", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(end), end, 3, "gtk_entry_select_region", "gint");
  gtk_entry_select_region(XEN_TO_C_GtkEntry_(entry), XEN_TO_C_gint(start), XEN_TO_C_gint(end));
  return(XEN_FALSE);
}

static XEN gxg_gtk_entry_set_editable(XEN entry, XEN editable)
{
  #define H_gtk_entry_set_editable "void gtk_entry_set_editable(GtkEntry* entry, gboolean editable)"
  XEN_ASSERT_TYPE(XEN_GtkEntry__P(entry), entry, 1, "gtk_entry_set_editable", "GtkEntry*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(editable), editable, 2, "gtk_entry_set_editable", "gboolean");
  gtk_entry_set_editable(XEN_TO_C_GtkEntry_(entry), XEN_TO_C_gboolean(editable));
  return(XEN_FALSE);
}

static XEN gxg_gtk_font_selection_get_font(XEN fontsel)
{
  #define H_gtk_font_selection_get_font "GdkFont* gtk_font_selection_get_font(GtkFontSelection* fontsel)"
  XEN_ASSERT_TYPE(XEN_GtkFontSelection__P(fontsel), fontsel, 1, "gtk_font_selection_get_font", "GtkFontSelection*");
  return(C_TO_XEN_GdkFont_(gtk_font_selection_get_font(XEN_TO_C_GtkFontSelection_(fontsel))));
}

static XEN gxg_gtk_font_selection_dialog_get_font(XEN fsd)
{
  #define H_gtk_font_selection_dialog_get_font "GdkFont* gtk_font_selection_dialog_get_font(GtkFontSelectionDialog* fsd)"
  XEN_ASSERT_TYPE(XEN_GtkFontSelectionDialog__P(fsd), fsd, 1, "gtk_font_selection_dialog_get_font", "GtkFontSelectionDialog*");
  return(C_TO_XEN_GdkFont_(gtk_font_selection_dialog_get_font(XEN_TO_C_GtkFontSelectionDialog_(fsd))));
}

static XEN gxg_gtk_hbutton_box_get_spacing_default(void)
{
  #define H_gtk_hbutton_box_get_spacing_default "gint gtk_hbutton_box_get_spacing_default( void)"
  return(C_TO_XEN_gint(gtk_hbutton_box_get_spacing_default()));
}

static XEN gxg_gtk_hbutton_box_get_layout_default(void)
{
  #define H_gtk_hbutton_box_get_layout_default "GtkButtonBoxStyle gtk_hbutton_box_get_layout_default( void)"
  return(C_TO_XEN_GtkButtonBoxStyle(gtk_hbutton_box_get_layout_default()));
}

static XEN gxg_gtk_hbutton_box_set_spacing_default(XEN spacing)
{
  #define H_gtk_hbutton_box_set_spacing_default "void gtk_hbutton_box_set_spacing_default(gint spacing)"
  XEN_ASSERT_TYPE(XEN_gint_P(spacing), spacing, 1, "gtk_hbutton_box_set_spacing_default", "gint");
  gtk_hbutton_box_set_spacing_default(XEN_TO_C_gint(spacing));
  return(XEN_FALSE);
}

static XEN gxg_gtk_hbutton_box_set_layout_default(XEN layout)
{
  #define H_gtk_hbutton_box_set_layout_default "void gtk_hbutton_box_set_layout_default(GtkButtonBoxStyle layout)"
  XEN_ASSERT_TYPE(XEN_GtkButtonBoxStyle_P(layout), layout, 1, "gtk_hbutton_box_set_layout_default", "GtkButtonBoxStyle");
  gtk_hbutton_box_set_layout_default(XEN_TO_C_GtkButtonBoxStyle(layout));
  return(XEN_FALSE);
}

static XEN gxg_gtk_image_set(XEN image, XEN val, XEN mask)
{
  #define H_gtk_image_set "void gtk_image_set(GtkImage* image, GdkImage* val, GdkBitmap* mask)"
  XEN_ASSERT_TYPE(XEN_GtkImage__P(image), image, 1, "gtk_image_set", "GtkImage*");
  XEN_ASSERT_TYPE(XEN_GdkImage__P(val), val, 2, "gtk_image_set", "GdkImage*");
  XEN_ASSERT_TYPE(XEN_GdkBitmap__P(mask), mask, 3, "gtk_image_set", "GdkBitmap*");
  gtk_image_set(XEN_TO_C_GtkImage_(image), XEN_TO_C_GdkImage_(val), XEN_TO_C_GdkBitmap_(mask));
  return(XEN_FALSE);
}

static XEN gxg_gtk_image_get(XEN image, XEN val, XEN mask)
{
  #define H_gtk_image_get "void gtk_image_get(GtkImage* image, GdkImage** [val], GdkBitmap** [mask])"
  GdkImage* ref_val;
  GdkBitmap* ref_mask;
  XEN_ASSERT_TYPE(XEN_GtkImage__P(image), image, 1, "gtk_image_get", "GtkImage*");
  gtk_image_get(XEN_TO_C_GtkImage_(image), &ref_val, &ref_mask);
  return(XEN_LIST_2(C_TO_XEN_GdkImage_(ref_val), C_TO_XEN_GdkBitmap_(ref_mask)));
}

static XEN gxg_gtk_item_factory_from_path(XEN path)
{
  #define H_gtk_item_factory_from_path "GtkItemFactory* gtk_item_factory_from_path(gchar* path)"
  XEN_ASSERT_TYPE(XEN_gchar__P(path), path, 1, "gtk_item_factory_from_path", "gchar*");
  return(C_TO_XEN_GtkItemFactory_(gtk_item_factory_from_path(XEN_TO_C_gchar_(path))));
}

static XEN gxg_gtk_item_factory_create_menu_entries(XEN n_entries, XEN entries)
{
  #define H_gtk_item_factory_create_menu_entries "void gtk_item_factory_create_menu_entries(guint n_entries, \
GtkMenuEntry* entries)"
  XEN_ASSERT_TYPE(XEN_guint_P(n_entries), n_entries, 1, "gtk_item_factory_create_menu_entries", "guint");
  XEN_ASSERT_TYPE(XEN_GtkMenuEntry__P(entries), entries, 2, "gtk_item_factory_create_menu_entries", "GtkMenuEntry*");
  gtk_item_factory_create_menu_entries(XEN_TO_C_guint(n_entries), XEN_TO_C_GtkMenuEntry_(entries));
  return(XEN_FALSE);
}

static XEN gxg_gtk_item_factories_path_delete(XEN ifactory_path, XEN path)
{
  #define H_gtk_item_factories_path_delete "void gtk_item_factories_path_delete(gchar* ifactory_path, \
gchar* path)"
  XEN_ASSERT_TYPE(XEN_gchar__P(ifactory_path), ifactory_path, 1, "gtk_item_factories_path_delete", "gchar*");
  XEN_ASSERT_TYPE(XEN_gchar__P(path), path, 2, "gtk_item_factories_path_delete", "gchar*");
  gtk_item_factories_path_delete(XEN_TO_C_gchar_(ifactory_path), XEN_TO_C_gchar_(path));
  return(XEN_FALSE);
}

static XEN gxg_gtk_item_factory_create_items_ac(XEN ifactory, XEN n_entries, XEN entries, XEN callback_data, XEN callback_type)
{
  #define H_gtk_item_factory_create_items_ac "void gtk_item_factory_create_items_ac(GtkItemFactory* ifactory, \
guint n_entries, GtkItemFactoryEntry* entries, gpointer callback_data, guint callback_type)"
  XEN_ASSERT_TYPE(XEN_GtkItemFactory__P(ifactory), ifactory, 1, "gtk_item_factory_create_items_ac", "GtkItemFactory*");
  XEN_ASSERT_TYPE(XEN_guint_P(n_entries), n_entries, 2, "gtk_item_factory_create_items_ac", "guint");
  XEN_ASSERT_TYPE(XEN_GtkItemFactoryEntry__P(entries), entries, 3, "gtk_item_factory_create_items_ac", "GtkItemFactoryEntry*");
  XEN_ASSERT_TYPE(XEN_gpointer_P(callback_data), callback_data, 4, "gtk_item_factory_create_items_ac", "gpointer");
  XEN_ASSERT_TYPE(XEN_guint_P(callback_type), callback_type, 5, "gtk_item_factory_create_items_ac", "guint");
  gtk_item_factory_create_items_ac(XEN_TO_C_GtkItemFactory_(ifactory), XEN_TO_C_guint(n_entries), XEN_TO_C_GtkItemFactoryEntry_(entries), 
                                   XEN_TO_C_gpointer(callback_data), XEN_TO_C_guint(callback_type));
  return(XEN_FALSE);
}

static XEN gxg_gtk_label_get(XEN label, XEN str)
{
  #define H_gtk_label_get "void gtk_label_get(GtkLabel* label, char** [str])"
  char* ref_str;
  XEN_ASSERT_TYPE(XEN_GtkLabel__P(label), label, 1, "gtk_label_get", "GtkLabel*");
  gtk_label_get(XEN_TO_C_GtkLabel_(label), &ref_str);
  return(XEN_LIST_1(C_TO_XEN_char_(ref_str)));
}

static XEN gxg_gtk_label_parse_uline(XEN label, XEN string)
{
  #define H_gtk_label_parse_uline "guint gtk_label_parse_uline(GtkLabel* label, gchar* string)"
  XEN_ASSERT_TYPE(XEN_GtkLabel__P(label), label, 1, "gtk_label_parse_uline", "GtkLabel*");
  XEN_ASSERT_TYPE(XEN_gchar__P(string), string, 2, "gtk_label_parse_uline", "gchar*");
  return(C_TO_XEN_guint(gtk_label_parse_uline(XEN_TO_C_GtkLabel_(label), XEN_TO_C_gchar_(string))));
}

static XEN gxg_gtk_layout_freeze(XEN layout)
{
  #define H_gtk_layout_freeze "void gtk_layout_freeze(GtkLayout* layout)"
  XEN_ASSERT_TYPE(XEN_GtkLayout__P(layout), layout, 1, "gtk_layout_freeze", "GtkLayout*");
  gtk_layout_freeze(XEN_TO_C_GtkLayout_(layout));
  return(XEN_FALSE);
}

static XEN gxg_gtk_layout_thaw(XEN layout)
{
  #define H_gtk_layout_thaw "void gtk_layout_thaw(GtkLayout* layout)"
  XEN_ASSERT_TYPE(XEN_GtkLayout__P(layout), layout, 1, "gtk_layout_thaw", "GtkLayout*");
  gtk_layout_thaw(XEN_TO_C_GtkLayout_(layout));
  return(XEN_FALSE);
}

static XEN gxg_gtk_list_get_type(void)
{
  #define H_gtk_list_get_type "GtkType gtk_list_get_type( void)"
  return(C_TO_XEN_GtkType(gtk_list_get_type()));
}

static XEN gxg_gtk_list_new(void)
{
  #define H_gtk_list_new "GtkWidget* gtk_list_new( void)"
  return(C_TO_XEN_GtkWidget_(gtk_list_new()));
}

static XEN gxg_gtk_list_insert_items(XEN list, XEN items, XEN position)
{
  #define H_gtk_list_insert_items "void gtk_list_insert_items(GtkList* list, GList* items, gint position)"
  XEN_ASSERT_TYPE(XEN_GtkList__P(list), list, 1, "gtk_list_insert_items", "GtkList*");
  XEN_ASSERT_TYPE(XEN_GList__P(items), items, 2, "gtk_list_insert_items", "GList*");
  XEN_ASSERT_TYPE(XEN_gint_P(position), position, 3, "gtk_list_insert_items", "gint");
  gtk_list_insert_items(XEN_TO_C_GtkList_(list), XEN_TO_C_GList_(items), XEN_TO_C_gint(position));
  return(XEN_FALSE);
}

static XEN gxg_gtk_list_append_items(XEN list, XEN items)
{
  #define H_gtk_list_append_items "void gtk_list_append_items(GtkList* list, GList* items)"
  XEN_ASSERT_TYPE(XEN_GtkList__P(list), list, 1, "gtk_list_append_items", "GtkList*");
  XEN_ASSERT_TYPE(XEN_GList__P(items), items, 2, "gtk_list_append_items", "GList*");
  gtk_list_append_items(XEN_TO_C_GtkList_(list), XEN_TO_C_GList_(items));
  return(XEN_FALSE);
}

static XEN gxg_gtk_list_prepend_items(XEN list, XEN items)
{
  #define H_gtk_list_prepend_items "void gtk_list_prepend_items(GtkList* list, GList* items)"
  XEN_ASSERT_TYPE(XEN_GtkList__P(list), list, 1, "gtk_list_prepend_items", "GtkList*");
  XEN_ASSERT_TYPE(XEN_GList__P(items), items, 2, "gtk_list_prepend_items", "GList*");
  gtk_list_prepend_items(XEN_TO_C_GtkList_(list), XEN_TO_C_GList_(items));
  return(XEN_FALSE);
}

static XEN gxg_gtk_list_remove_items(XEN list, XEN items)
{
  #define H_gtk_list_remove_items "void gtk_list_remove_items(GtkList* list, GList* items)"
  XEN_ASSERT_TYPE(XEN_GtkList__P(list), list, 1, "gtk_list_remove_items", "GtkList*");
  XEN_ASSERT_TYPE(XEN_GList__P(items), items, 2, "gtk_list_remove_items", "GList*");
  gtk_list_remove_items(XEN_TO_C_GtkList_(list), XEN_TO_C_GList_(items));
  return(XEN_FALSE);
}

static XEN gxg_gtk_list_remove_items_no_unref(XEN list, XEN items)
{
  #define H_gtk_list_remove_items_no_unref "void gtk_list_remove_items_no_unref(GtkList* list, GList* items)"
  XEN_ASSERT_TYPE(XEN_GtkList__P(list), list, 1, "gtk_list_remove_items_no_unref", "GtkList*");
  XEN_ASSERT_TYPE(XEN_GList__P(items), items, 2, "gtk_list_remove_items_no_unref", "GList*");
  gtk_list_remove_items_no_unref(XEN_TO_C_GtkList_(list), XEN_TO_C_GList_(items));
  return(XEN_FALSE);
}

static XEN gxg_gtk_list_clear_items(XEN list, XEN start, XEN end)
{
  #define H_gtk_list_clear_items "void gtk_list_clear_items(GtkList* list, gint start, gint end)"
  XEN_ASSERT_TYPE(XEN_GtkList__P(list), list, 1, "gtk_list_clear_items", "GtkList*");
  XEN_ASSERT_TYPE(XEN_gint_P(start), start, 2, "gtk_list_clear_items", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(end), end, 3, "gtk_list_clear_items", "gint");
  gtk_list_clear_items(XEN_TO_C_GtkList_(list), XEN_TO_C_gint(start), XEN_TO_C_gint(end));
  return(XEN_FALSE);
}

static XEN gxg_gtk_list_select_item(XEN list, XEN item)
{
  #define H_gtk_list_select_item "void gtk_list_select_item(GtkList* list, gint item)"
  XEN_ASSERT_TYPE(XEN_GtkList__P(list), list, 1, "gtk_list_select_item", "GtkList*");
  XEN_ASSERT_TYPE(XEN_gint_P(item), item, 2, "gtk_list_select_item", "gint");
  gtk_list_select_item(XEN_TO_C_GtkList_(list), XEN_TO_C_gint(item));
  return(XEN_FALSE);
}

static XEN gxg_gtk_list_unselect_item(XEN list, XEN item)
{
  #define H_gtk_list_unselect_item "void gtk_list_unselect_item(GtkList* list, gint item)"
  XEN_ASSERT_TYPE(XEN_GtkList__P(list), list, 1, "gtk_list_unselect_item", "GtkList*");
  XEN_ASSERT_TYPE(XEN_gint_P(item), item, 2, "gtk_list_unselect_item", "gint");
  gtk_list_unselect_item(XEN_TO_C_GtkList_(list), XEN_TO_C_gint(item));
  return(XEN_FALSE);
}

static XEN gxg_gtk_list_select_child(XEN list, XEN child)
{
  #define H_gtk_list_select_child "void gtk_list_select_child(GtkList* list, GtkWidget* child)"
  XEN_ASSERT_TYPE(XEN_GtkList__P(list), list, 1, "gtk_list_select_child", "GtkList*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(child), child, 2, "gtk_list_select_child", "GtkWidget*");
  gtk_list_select_child(XEN_TO_C_GtkList_(list), XEN_TO_C_GtkWidget_(child));
  return(XEN_FALSE);
}

static XEN gxg_gtk_list_unselect_child(XEN list, XEN child)
{
  #define H_gtk_list_unselect_child "void gtk_list_unselect_child(GtkList* list, GtkWidget* child)"
  XEN_ASSERT_TYPE(XEN_GtkList__P(list), list, 1, "gtk_list_unselect_child", "GtkList*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(child), child, 2, "gtk_list_unselect_child", "GtkWidget*");
  gtk_list_unselect_child(XEN_TO_C_GtkList_(list), XEN_TO_C_GtkWidget_(child));
  return(XEN_FALSE);
}

static XEN gxg_gtk_list_child_position(XEN list, XEN child)
{
  #define H_gtk_list_child_position "gint gtk_list_child_position(GtkList* list, GtkWidget* child)"
  XEN_ASSERT_TYPE(XEN_GtkList__P(list), list, 1, "gtk_list_child_position", "GtkList*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(child), child, 2, "gtk_list_child_position", "GtkWidget*");
  return(C_TO_XEN_gint(gtk_list_child_position(XEN_TO_C_GtkList_(list), XEN_TO_C_GtkWidget_(child))));
}

static XEN gxg_gtk_list_set_selection_mode(XEN list, XEN mode)
{
  #define H_gtk_list_set_selection_mode "void gtk_list_set_selection_mode(GtkList* list, GtkSelectionMode mode)"
  XEN_ASSERT_TYPE(XEN_GtkList__P(list), list, 1, "gtk_list_set_selection_mode", "GtkList*");
  XEN_ASSERT_TYPE(XEN_GtkSelectionMode_P(mode), mode, 2, "gtk_list_set_selection_mode", "GtkSelectionMode");
  gtk_list_set_selection_mode(XEN_TO_C_GtkList_(list), XEN_TO_C_GtkSelectionMode(mode));
  return(XEN_FALSE);
}

static XEN gxg_gtk_list_extend_selection(XEN list, XEN scroll_type, XEN position, XEN auto_start_selection)
{
  #define H_gtk_list_extend_selection "void gtk_list_extend_selection(GtkList* list, GtkScrollType scroll_type, \
gfloat position, gboolean auto_start_selection)"
  XEN_ASSERT_TYPE(XEN_GtkList__P(list), list, 1, "gtk_list_extend_selection", "GtkList*");
  XEN_ASSERT_TYPE(XEN_GtkScrollType_P(scroll_type), scroll_type, 2, "gtk_list_extend_selection", "GtkScrollType");
  XEN_ASSERT_TYPE(XEN_gfloat_P(position), position, 3, "gtk_list_extend_selection", "gfloat");
  XEN_ASSERT_TYPE(XEN_gboolean_P(auto_start_selection), auto_start_selection, 4, "gtk_list_extend_selection", "gboolean");
  gtk_list_extend_selection(XEN_TO_C_GtkList_(list), XEN_TO_C_GtkScrollType(scroll_type), XEN_TO_C_gfloat(position), XEN_TO_C_gboolean(auto_start_selection));
  return(XEN_FALSE);
}

static XEN gxg_gtk_list_start_selection(XEN list)
{
  #define H_gtk_list_start_selection "void gtk_list_start_selection(GtkList* list)"
  XEN_ASSERT_TYPE(XEN_GtkList__P(list), list, 1, "gtk_list_start_selection", "GtkList*");
  gtk_list_start_selection(XEN_TO_C_GtkList_(list));
  return(XEN_FALSE);
}

static XEN gxg_gtk_list_end_selection(XEN list)
{
  #define H_gtk_list_end_selection "void gtk_list_end_selection(GtkList* list)"
  XEN_ASSERT_TYPE(XEN_GtkList__P(list), list, 1, "gtk_list_end_selection", "GtkList*");
  gtk_list_end_selection(XEN_TO_C_GtkList_(list));
  return(XEN_FALSE);
}

static XEN gxg_gtk_list_select_all(XEN list)
{
  #define H_gtk_list_select_all "void gtk_list_select_all(GtkList* list)"
  XEN_ASSERT_TYPE(XEN_GtkList__P(list), list, 1, "gtk_list_select_all", "GtkList*");
  gtk_list_select_all(XEN_TO_C_GtkList_(list));
  return(XEN_FALSE);
}

static XEN gxg_gtk_list_unselect_all(XEN list)
{
  #define H_gtk_list_unselect_all "void gtk_list_unselect_all(GtkList* list)"
  XEN_ASSERT_TYPE(XEN_GtkList__P(list), list, 1, "gtk_list_unselect_all", "GtkList*");
  gtk_list_unselect_all(XEN_TO_C_GtkList_(list));
  return(XEN_FALSE);
}

static XEN gxg_gtk_list_scroll_horizontal(XEN list, XEN scroll_type, XEN position)
{
  #define H_gtk_list_scroll_horizontal "void gtk_list_scroll_horizontal(GtkList* list, GtkScrollType scroll_type, \
gfloat position)"
  XEN_ASSERT_TYPE(XEN_GtkList__P(list), list, 1, "gtk_list_scroll_horizontal", "GtkList*");
  XEN_ASSERT_TYPE(XEN_GtkScrollType_P(scroll_type), scroll_type, 2, "gtk_list_scroll_horizontal", "GtkScrollType");
  XEN_ASSERT_TYPE(XEN_gfloat_P(position), position, 3, "gtk_list_scroll_horizontal", "gfloat");
  gtk_list_scroll_horizontal(XEN_TO_C_GtkList_(list), XEN_TO_C_GtkScrollType(scroll_type), XEN_TO_C_gfloat(position));
  return(XEN_FALSE);
}

static XEN gxg_gtk_list_scroll_vertical(XEN list, XEN scroll_type, XEN position)
{
  #define H_gtk_list_scroll_vertical "void gtk_list_scroll_vertical(GtkList* list, GtkScrollType scroll_type, \
gfloat position)"
  XEN_ASSERT_TYPE(XEN_GtkList__P(list), list, 1, "gtk_list_scroll_vertical", "GtkList*");
  XEN_ASSERT_TYPE(XEN_GtkScrollType_P(scroll_type), scroll_type, 2, "gtk_list_scroll_vertical", "GtkScrollType");
  XEN_ASSERT_TYPE(XEN_gfloat_P(position), position, 3, "gtk_list_scroll_vertical", "gfloat");
  gtk_list_scroll_vertical(XEN_TO_C_GtkList_(list), XEN_TO_C_GtkScrollType(scroll_type), XEN_TO_C_gfloat(position));
  return(XEN_FALSE);
}

static XEN gxg_gtk_list_toggle_add_mode(XEN list)
{
  #define H_gtk_list_toggle_add_mode "void gtk_list_toggle_add_mode(GtkList* list)"
  XEN_ASSERT_TYPE(XEN_GtkList__P(list), list, 1, "gtk_list_toggle_add_mode", "GtkList*");
  gtk_list_toggle_add_mode(XEN_TO_C_GtkList_(list));
  return(XEN_FALSE);
}

static XEN gxg_gtk_list_toggle_focus_row(XEN list)
{
  #define H_gtk_list_toggle_focus_row "void gtk_list_toggle_focus_row(GtkList* list)"
  XEN_ASSERT_TYPE(XEN_GtkList__P(list), list, 1, "gtk_list_toggle_focus_row", "GtkList*");
  gtk_list_toggle_focus_row(XEN_TO_C_GtkList_(list));
  return(XEN_FALSE);
}

static XEN gxg_gtk_list_toggle_row(XEN list, XEN item)
{
  #define H_gtk_list_toggle_row "void gtk_list_toggle_row(GtkList* list, GtkWidget* item)"
  XEN_ASSERT_TYPE(XEN_GtkList__P(list), list, 1, "gtk_list_toggle_row", "GtkList*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(item), item, 2, "gtk_list_toggle_row", "GtkWidget*");
  gtk_list_toggle_row(XEN_TO_C_GtkList_(list), XEN_TO_C_GtkWidget_(item));
  return(XEN_FALSE);
}

static XEN gxg_gtk_list_undo_selection(XEN list)
{
  #define H_gtk_list_undo_selection "void gtk_list_undo_selection(GtkList* list)"
  XEN_ASSERT_TYPE(XEN_GtkList__P(list), list, 1, "gtk_list_undo_selection", "GtkList*");
  gtk_list_undo_selection(XEN_TO_C_GtkList_(list));
  return(XEN_FALSE);
}

static XEN gxg_gtk_list_end_drag_selection(XEN list)
{
  #define H_gtk_list_end_drag_selection "void gtk_list_end_drag_selection(GtkList* list)"
  XEN_ASSERT_TYPE(XEN_GtkList__P(list), list, 1, "gtk_list_end_drag_selection", "GtkList*");
  gtk_list_end_drag_selection(XEN_TO_C_GtkList_(list));
  return(XEN_FALSE);
}

static XEN gxg_gtk_list_item_get_type(void)
{
  #define H_gtk_list_item_get_type "GtkType gtk_list_item_get_type( void)"
  return(C_TO_XEN_GtkType(gtk_list_item_get_type()));
}

static XEN gxg_gtk_list_item_new(void)
{
  #define H_gtk_list_item_new "GtkWidget* gtk_list_item_new( void)"
  return(C_TO_XEN_GtkWidget_(gtk_list_item_new()));
}

static XEN gxg_gtk_list_item_new_with_label(XEN label)
{
  #define H_gtk_list_item_new_with_label "GtkWidget* gtk_list_item_new_with_label(gchar* label)"
  XEN_ASSERT_TYPE(XEN_gchar__P(label), label, 1, "gtk_list_item_new_with_label", "gchar*");
  return(C_TO_XEN_GtkWidget_(gtk_list_item_new_with_label(XEN_TO_C_gchar_(label))));
}

static XEN gxg_gtk_list_item_select(XEN list_item)
{
  #define H_gtk_list_item_select "void gtk_list_item_select(GtkListItem* list_item)"
  XEN_ASSERT_TYPE(XEN_GtkListItem__P(list_item), list_item, 1, "gtk_list_item_select", "GtkListItem*");
  gtk_list_item_select(XEN_TO_C_GtkListItem_(list_item));
  return(XEN_FALSE);
}

static XEN gxg_gtk_list_item_deselect(XEN list_item)
{
  #define H_gtk_list_item_deselect "void gtk_list_item_deselect(GtkListItem* list_item)"
  XEN_ASSERT_TYPE(XEN_GtkListItem__P(list_item), list_item, 1, "gtk_list_item_deselect", "GtkListItem*");
  gtk_list_item_deselect(XEN_TO_C_GtkListItem_(list_item));
  return(XEN_FALSE);
}

static XEN gxg_gtk_notebook_set_homogeneous_tabs(XEN notebook, XEN homogeneous)
{
  #define H_gtk_notebook_set_homogeneous_tabs "void gtk_notebook_set_homogeneous_tabs(GtkNotebook* notebook, \
gboolean homogeneous)"
  XEN_ASSERT_TYPE(XEN_GtkNotebook__P(notebook), notebook, 1, "gtk_notebook_set_homogeneous_tabs", "GtkNotebook*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(homogeneous), homogeneous, 2, "gtk_notebook_set_homogeneous_tabs", "gboolean");
  gtk_notebook_set_homogeneous_tabs(XEN_TO_C_GtkNotebook_(notebook), XEN_TO_C_gboolean(homogeneous));
  return(XEN_FALSE);
}

static XEN gxg_gtk_notebook_set_tab_border(XEN notebook, XEN border_width)
{
  #define H_gtk_notebook_set_tab_border "void gtk_notebook_set_tab_border(GtkNotebook* notebook, guint border_width)"
  XEN_ASSERT_TYPE(XEN_GtkNotebook__P(notebook), notebook, 1, "gtk_notebook_set_tab_border", "GtkNotebook*");
  XEN_ASSERT_TYPE(XEN_guint_P(border_width), border_width, 2, "gtk_notebook_set_tab_border", "guint");
  gtk_notebook_set_tab_border(XEN_TO_C_GtkNotebook_(notebook), XEN_TO_C_guint(border_width));
  return(XEN_FALSE);
}

static XEN gxg_gtk_notebook_set_tab_hborder(XEN notebook, XEN tab_hborder)
{
  #define H_gtk_notebook_set_tab_hborder "void gtk_notebook_set_tab_hborder(GtkNotebook* notebook, guint tab_hborder)"
  XEN_ASSERT_TYPE(XEN_GtkNotebook__P(notebook), notebook, 1, "gtk_notebook_set_tab_hborder", "GtkNotebook*");
  XEN_ASSERT_TYPE(XEN_guint_P(tab_hborder), tab_hborder, 2, "gtk_notebook_set_tab_hborder", "guint");
  gtk_notebook_set_tab_hborder(XEN_TO_C_GtkNotebook_(notebook), XEN_TO_C_guint(tab_hborder));
  return(XEN_FALSE);
}

static XEN gxg_gtk_notebook_set_tab_vborder(XEN notebook, XEN tab_vborder)
{
  #define H_gtk_notebook_set_tab_vborder "void gtk_notebook_set_tab_vborder(GtkNotebook* notebook, guint tab_vborder)"
  XEN_ASSERT_TYPE(XEN_GtkNotebook__P(notebook), notebook, 1, "gtk_notebook_set_tab_vborder", "GtkNotebook*");
  XEN_ASSERT_TYPE(XEN_guint_P(tab_vborder), tab_vborder, 2, "gtk_notebook_set_tab_vborder", "guint");
  gtk_notebook_set_tab_vborder(XEN_TO_C_GtkNotebook_(notebook), XEN_TO_C_guint(tab_vborder));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_object_ref(XEN object)
{
  #define H_gtk_object_ref "GtkObject* gtk_object_ref(GtkObject* object)"
  XEN_ASSERT_TYPE(XEN_GtkObject__P(object), object, 1, "gtk_object_ref", "GtkObject*");
  return(C_TO_XEN_GtkObject_(gtk_object_ref(XEN_TO_C_GtkObject_(object))));
}

static XEN gxg_gtk_object_unref(XEN object)
{
  #define H_gtk_object_unref "void gtk_object_unref(GtkObject* object)"
  XEN_ASSERT_TYPE(XEN_GtkObject__P(object), object, 1, "gtk_object_unref", "GtkObject*");
  gtk_object_unref(XEN_TO_C_GtkObject_(object));
  return(XEN_FALSE);
}
#endif

static XEN gxg_gtk_object_weakref(XEN object, XEN notify, XEN func_data)
{
  #define H_gtk_object_weakref "void gtk_object_weakref(GtkObject* object, GtkDestroyNotify notify, gpointer func_data)"
  XEN_ASSERT_TYPE(XEN_GtkObject__P(object), object, 1, "gtk_object_weakref", "GtkObject*");
  XEN_ASSERT_TYPE(XEN_GtkDestroyNotify_P(notify), notify, 2, "gtk_object_weakref", "GtkDestroyNotify");
  XEN_ASSERT_TYPE(XEN_gpointer_P(func_data), func_data, 3, "gtk_object_weakref", "gpointer");
  {
    XEN gxg_ptr = XEN_LIST_5(XEN_FALSE, func_data, XEN_FALSE, XEN_FALSE, XEN_FALSE);
    xm_protect(gxg_ptr);
    XEN_LIST_SET(gxg_ptr, 3, notify);
    gtk_object_weakref(XEN_TO_C_GtkObject_(object), XEN_TO_C_GtkDestroyNotify(notify), XEN_TO_C_gpointer(func_data));
    return(XEN_FALSE);
   }
}

static XEN gxg_gtk_object_weakunref(XEN object, XEN notify, XEN func_data)
{
  #define H_gtk_object_weakunref "void gtk_object_weakunref(GtkObject* object, GtkDestroyNotify notify, \
gpointer func_data)"
  XEN_ASSERT_TYPE(XEN_GtkObject__P(object), object, 1, "gtk_object_weakunref", "GtkObject*");
  XEN_ASSERT_TYPE(XEN_GtkDestroyNotify_P(notify), notify, 2, "gtk_object_weakunref", "GtkDestroyNotify");
  XEN_ASSERT_TYPE(XEN_gpointer_P(func_data), func_data, 3, "gtk_object_weakunref", "gpointer");
  {
    XEN gxg_ptr = XEN_LIST_5(XEN_FALSE, func_data, XEN_FALSE, XEN_FALSE, XEN_FALSE);
    xm_protect(gxg_ptr);
    XEN_LIST_SET(gxg_ptr, 3, notify);
    gtk_object_weakunref(XEN_TO_C_GtkObject_(object), XEN_TO_C_GtkDestroyNotify(notify), XEN_TO_C_gpointer(func_data));
    return(XEN_FALSE);
   }
}

static XEN gxg_gtk_object_set_data(XEN object, XEN key, XEN data)
{
  #define H_gtk_object_set_data "void gtk_object_set_data(GtkObject* object, gchar* key, gpointer data)"
  XEN_ASSERT_TYPE(XEN_GtkObject__P(object), object, 1, "gtk_object_set_data", "GtkObject*");
  XEN_ASSERT_TYPE(XEN_gchar__P(key), key, 2, "gtk_object_set_data", "gchar*");
  XEN_ASSERT_TYPE(XEN_gpointer_P(data), data, 3, "gtk_object_set_data", "gpointer");
  gtk_object_set_data(XEN_TO_C_GtkObject_(object), XEN_TO_C_gchar_(key), XEN_TO_C_gpointer(data));
  return(XEN_FALSE);
}

static XEN gxg_gtk_object_set_data_full(XEN object, XEN key, XEN func_data, XEN destroy)
{
  #define H_gtk_object_set_data_full "void gtk_object_set_data_full(GtkObject* object, gchar* key, gpointer func_data, \
GtkDestroyNotify destroy)"
  XEN_ASSERT_TYPE(XEN_GtkObject__P(object), object, 1, "gtk_object_set_data_full", "GtkObject*");
  XEN_ASSERT_TYPE(XEN_gchar__P(key), key, 2, "gtk_object_set_data_full", "gchar*");
  XEN_ASSERT_TYPE(XEN_gpointer_P(func_data), func_data, 3, "gtk_object_set_data_full", "gpointer");
  XEN_ASSERT_TYPE(XEN_GtkDestroyNotify_P(destroy), destroy, 4, "gtk_object_set_data_full", "GtkDestroyNotify");
  {
    XEN gxg_ptr = XEN_LIST_5(XEN_FALSE, func_data, XEN_FALSE, XEN_FALSE, XEN_FALSE);
    xm_protect(gxg_ptr);
    XEN_LIST_SET(gxg_ptr, 3, destroy);
    gtk_object_set_data_full(XEN_TO_C_GtkObject_(object), XEN_TO_C_gchar_(key), XEN_TO_C_gpointer(func_data), XEN_TO_C_GtkDestroyNotify(destroy));
    return(XEN_FALSE);
   }
}

static XEN gxg_gtk_object_remove_data(XEN object, XEN key)
{
  #define H_gtk_object_remove_data "void gtk_object_remove_data(GtkObject* object, gchar* key)"
  XEN_ASSERT_TYPE(XEN_GtkObject__P(object), object, 1, "gtk_object_remove_data", "GtkObject*");
  XEN_ASSERT_TYPE(XEN_gchar__P(key), key, 2, "gtk_object_remove_data", "gchar*");
  gtk_object_remove_data(XEN_TO_C_GtkObject_(object), XEN_TO_C_gchar_(key));
  return(XEN_FALSE);
}

static XEN gxg_gtk_object_get_data(XEN object, XEN key)
{
  #define H_gtk_object_get_data "gpointer gtk_object_get_data(GtkObject* object, gchar* key)"
  XEN_ASSERT_TYPE(XEN_GtkObject__P(object), object, 1, "gtk_object_get_data", "GtkObject*");
  XEN_ASSERT_TYPE(XEN_gchar__P(key), key, 2, "gtk_object_get_data", "gchar*");
  return(C_TO_XEN_gpointer(gtk_object_get_data(XEN_TO_C_GtkObject_(object), XEN_TO_C_gchar_(key))));
}

static XEN gxg_gtk_object_remove_no_notify(XEN object, XEN key)
{
  #define H_gtk_object_remove_no_notify "void gtk_object_remove_no_notify(GtkObject* object, gchar* key)"
  XEN_ASSERT_TYPE(XEN_GtkObject__P(object), object, 1, "gtk_object_remove_no_notify", "GtkObject*");
  XEN_ASSERT_TYPE(XEN_gchar__P(key), key, 2, "gtk_object_remove_no_notify", "gchar*");
  gtk_object_remove_no_notify(XEN_TO_C_GtkObject_(object), XEN_TO_C_gchar_(key));
  return(XEN_FALSE);
}

static XEN gxg_gtk_object_set_user_data(XEN object, XEN data)
{
  #define H_gtk_object_set_user_data "void gtk_object_set_user_data(GtkObject* object, gpointer data)"
  XEN_ASSERT_TYPE(XEN_GtkObject__P(object), object, 1, "gtk_object_set_user_data", "GtkObject*");
  XEN_ASSERT_TYPE(XEN_gpointer_P(data), data, 2, "gtk_object_set_user_data", "gpointer");
  gtk_object_set_user_data(XEN_TO_C_GtkObject_(object), XEN_TO_C_gpointer(data));
  return(XEN_FALSE);
}

static XEN gxg_gtk_object_get_user_data(XEN object)
{
  #define H_gtk_object_get_user_data "gpointer gtk_object_get_user_data(GtkObject* object)"
  XEN_ASSERT_TYPE(XEN_GtkObject__P(object), object, 1, "gtk_object_get_user_data", "GtkObject*");
  return(C_TO_XEN_gpointer(gtk_object_get_user_data(XEN_TO_C_GtkObject_(object))));
}

static XEN gxg_gtk_object_set_data_by_id(XEN object, XEN data_id, XEN data)
{
  #define H_gtk_object_set_data_by_id "void gtk_object_set_data_by_id(GtkObject* object, GQuark data_id, \
gpointer data)"
  XEN_ASSERT_TYPE(XEN_GtkObject__P(object), object, 1, "gtk_object_set_data_by_id", "GtkObject*");
  XEN_ASSERT_TYPE(XEN_GQuark_P(data_id), data_id, 2, "gtk_object_set_data_by_id", "GQuark");
  XEN_ASSERT_TYPE(XEN_gpointer_P(data), data, 3, "gtk_object_set_data_by_id", "gpointer");
  gtk_object_set_data_by_id(XEN_TO_C_GtkObject_(object), XEN_TO_C_GQuark(data_id), XEN_TO_C_gpointer(data));
  return(XEN_FALSE);
}

static XEN gxg_gtk_object_set_data_by_id_full(XEN object, XEN data_id, XEN func_data, XEN destroy)
{
  #define H_gtk_object_set_data_by_id_full "void gtk_object_set_data_by_id_full(GtkObject* object, GQuark data_id, \
gpointer func_data, GtkDestroyNotify destroy)"
  XEN_ASSERT_TYPE(XEN_GtkObject__P(object), object, 1, "gtk_object_set_data_by_id_full", "GtkObject*");
  XEN_ASSERT_TYPE(XEN_GQuark_P(data_id), data_id, 2, "gtk_object_set_data_by_id_full", "GQuark");
  XEN_ASSERT_TYPE(XEN_gpointer_P(func_data), func_data, 3, "gtk_object_set_data_by_id_full", "gpointer");
  XEN_ASSERT_TYPE(XEN_GtkDestroyNotify_P(destroy), destroy, 4, "gtk_object_set_data_by_id_full", "GtkDestroyNotify");
  {
    XEN gxg_ptr = XEN_LIST_5(XEN_FALSE, func_data, XEN_FALSE, XEN_FALSE, XEN_FALSE);
    xm_protect(gxg_ptr);
    XEN_LIST_SET(gxg_ptr, 3, destroy);
    gtk_object_set_data_by_id_full(XEN_TO_C_GtkObject_(object), XEN_TO_C_GQuark(data_id), XEN_TO_C_gpointer(func_data), XEN_TO_C_GtkDestroyNotify(destroy));
    return(XEN_FALSE);
   }
}

static XEN gxg_gtk_object_get_data_by_id(XEN object, XEN data_id)
{
  #define H_gtk_object_get_data_by_id "gpointer gtk_object_get_data_by_id(GtkObject* object, GQuark data_id)"
  XEN_ASSERT_TYPE(XEN_GtkObject__P(object), object, 1, "gtk_object_get_data_by_id", "GtkObject*");
  XEN_ASSERT_TYPE(XEN_GQuark_P(data_id), data_id, 2, "gtk_object_get_data_by_id", "GQuark");
  return(C_TO_XEN_gpointer(gtk_object_get_data_by_id(XEN_TO_C_GtkObject_(object), XEN_TO_C_GQuark(data_id))));
}

static XEN gxg_gtk_object_remove_data_by_id(XEN object, XEN data_id)
{
  #define H_gtk_object_remove_data_by_id "void gtk_object_remove_data_by_id(GtkObject* object, GQuark data_id)"
  XEN_ASSERT_TYPE(XEN_GtkObject__P(object), object, 1, "gtk_object_remove_data_by_id", "GtkObject*");
  XEN_ASSERT_TYPE(XEN_GQuark_P(data_id), data_id, 2, "gtk_object_remove_data_by_id", "GQuark");
  gtk_object_remove_data_by_id(XEN_TO_C_GtkObject_(object), XEN_TO_C_GQuark(data_id));
  return(XEN_FALSE);
}

static XEN gxg_gtk_object_remove_no_notify_by_id(XEN object, XEN key_id)
{
  #define H_gtk_object_remove_no_notify_by_id "void gtk_object_remove_no_notify_by_id(GtkObject* object, \
GQuark key_id)"
  XEN_ASSERT_TYPE(XEN_GtkObject__P(object), object, 1, "gtk_object_remove_no_notify_by_id", "GtkObject*");
  XEN_ASSERT_TYPE(XEN_GQuark_P(key_id), key_id, 2, "gtk_object_remove_no_notify_by_id", "GQuark");
  gtk_object_remove_no_notify_by_id(XEN_TO_C_GtkObject_(object), XEN_TO_C_GQuark(key_id));
  return(XEN_FALSE);
}

static XEN gxg_gtk_object_get(XEN object, XEN first_property_name)
{
  #define H_gtk_object_get "void gtk_object_get(GtkObject* object, gchar* first_property_name, ...)"
  XEN_ASSERT_TYPE(XEN_GtkObject__P(object), object, 1, "gtk_object_get", "GtkObject*");
  XEN_ASSERT_TYPE(XEN_gchar__P(first_property_name), first_property_name, 2, "gtk_object_get", "gchar*");
  gtk_object_get(XEN_TO_C_GtkObject_(object), XEN_TO_C_gchar_(first_property_name));
  return(XEN_FALSE);
}

static XEN gxg_gtk_object_set(XEN object, XEN first_property_name)
{
  #define H_gtk_object_set "void gtk_object_set(GtkObject* object, gchar* first_property_name, ...)"
  XEN_ASSERT_TYPE(XEN_GtkObject__P(object), object, 1, "gtk_object_set", "GtkObject*");
  XEN_ASSERT_TYPE(XEN_gchar__P(first_property_name), first_property_name, 2, "gtk_object_set", "gchar*");
  gtk_object_set(XEN_TO_C_GtkObject_(object), XEN_TO_C_gchar_(first_property_name));
  return(XEN_FALSE);
}

static XEN gxg_gtk_object_add_arg_type(XEN arg_name, XEN arg_type, XEN arg_flags, XEN arg_id)
{
  #define H_gtk_object_add_arg_type "void gtk_object_add_arg_type(gchar* arg_name, GtkType arg_type, \
guint arg_flags, guint arg_id)"
  XEN_ASSERT_TYPE(XEN_gchar__P(arg_name), arg_name, 1, "gtk_object_add_arg_type", "gchar*");
  XEN_ASSERT_TYPE(XEN_GtkType_P(arg_type), arg_type, 2, "gtk_object_add_arg_type", "GtkType");
  XEN_ASSERT_TYPE(XEN_guint_P(arg_flags), arg_flags, 3, "gtk_object_add_arg_type", "guint");
  XEN_ASSERT_TYPE(XEN_guint_P(arg_id), arg_id, 4, "gtk_object_add_arg_type", "guint");
  gtk_object_add_arg_type(XEN_TO_C_gchar_(arg_name), XEN_TO_C_GtkType(arg_type), XEN_TO_C_guint(arg_flags), XEN_TO_C_guint(arg_id));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_old_editable_get_type(void)
{
  #define H_gtk_old_editable_get_type "GtkType gtk_old_editable_get_type( void)"
  return(C_TO_XEN_GtkType(gtk_old_editable_get_type()));
}

static XEN gxg_gtk_old_editable_claim_selection(XEN old_editable, XEN claim, XEN time)
{
  #define H_gtk_old_editable_claim_selection "void gtk_old_editable_claim_selection(GtkOldEditable* old_editable, \
gboolean claim, guint32 time)"
  XEN_ASSERT_TYPE(XEN_GtkOldEditable__P(old_editable), old_editable, 1, "gtk_old_editable_claim_selection", "GtkOldEditable*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(claim), claim, 2, "gtk_old_editable_claim_selection", "gboolean");
  XEN_ASSERT_TYPE(XEN_guint32_P(time), time, 3, "gtk_old_editable_claim_selection", "guint32");
  gtk_old_editable_claim_selection(XEN_TO_C_GtkOldEditable_(old_editable), XEN_TO_C_gboolean(claim), XEN_TO_C_guint32(time));
  return(XEN_FALSE);
}

static XEN gxg_gtk_old_editable_changed(XEN old_editable)
{
  #define H_gtk_old_editable_changed "void gtk_old_editable_changed(GtkOldEditable* old_editable)"
  XEN_ASSERT_TYPE(XEN_GtkOldEditable__P(old_editable), old_editable, 1, "gtk_old_editable_changed", "GtkOldEditable*");
  gtk_old_editable_changed(XEN_TO_C_GtkOldEditable_(old_editable));
  return(XEN_FALSE);
}
#endif

static XEN gxg_gtk_pixmap_get_type(void)
{
  #define H_gtk_pixmap_get_type "GtkType gtk_pixmap_get_type( void)"
  return(C_TO_XEN_GtkType(gtk_pixmap_get_type()));
}

static XEN gxg_gtk_pixmap_new(XEN pixmap, XEN mask)
{
  #define H_gtk_pixmap_new "GtkWidget* gtk_pixmap_new(GdkPixmap* pixmap, GdkBitmap* mask)"
  XEN_ASSERT_TYPE(XEN_GdkPixmap__P(pixmap), pixmap, 1, "gtk_pixmap_new", "GdkPixmap*");
  XEN_ASSERT_TYPE(XEN_GdkBitmap__P(mask), mask, 2, "gtk_pixmap_new", "GdkBitmap*");
  return(C_TO_XEN_GtkWidget_(gtk_pixmap_new(XEN_TO_C_GdkPixmap_(pixmap), XEN_TO_C_GdkBitmap_(mask))));
}

static XEN gxg_gtk_pixmap_set(XEN pixmap, XEN val, XEN mask)
{
  #define H_gtk_pixmap_set "void gtk_pixmap_set(GtkPixmap* pixmap, GdkPixmap* val, GdkBitmap* mask)"
  XEN_ASSERT_TYPE(XEN_GtkPixmap__P(pixmap), pixmap, 1, "gtk_pixmap_set", "GtkPixmap*");
  XEN_ASSERT_TYPE(XEN_GdkPixmap__P(val), val, 2, "gtk_pixmap_set", "GdkPixmap*");
  XEN_ASSERT_TYPE(XEN_GdkBitmap__P(mask), mask, 3, "gtk_pixmap_set", "GdkBitmap*");
  gtk_pixmap_set(XEN_TO_C_GtkPixmap_(pixmap), XEN_TO_C_GdkPixmap_(val), XEN_TO_C_GdkBitmap_(mask));
  return(XEN_FALSE);
}

static XEN gxg_gtk_pixmap_get(XEN pixmap, XEN val, XEN mask)
{
  #define H_gtk_pixmap_get "void gtk_pixmap_get(GtkPixmap* pixmap, GdkPixmap** [val], GdkBitmap** [mask])"
  GdkPixmap* ref_val;
  GdkBitmap* ref_mask;
  XEN_ASSERT_TYPE(XEN_GtkPixmap__P(pixmap), pixmap, 1, "gtk_pixmap_get", "GtkPixmap*");
  gtk_pixmap_get(XEN_TO_C_GtkPixmap_(pixmap), &ref_val, &ref_mask);
  return(XEN_LIST_2(C_TO_XEN_GdkPixmap_(ref_val), C_TO_XEN_GdkBitmap_(ref_mask)));
}

static XEN gxg_gtk_pixmap_set_build_insensitive(XEN pixmap, XEN build)
{
  #define H_gtk_pixmap_set_build_insensitive "void gtk_pixmap_set_build_insensitive(GtkPixmap* pixmap, \
gboolean build)"
  XEN_ASSERT_TYPE(XEN_GtkPixmap__P(pixmap), pixmap, 1, "gtk_pixmap_set_build_insensitive", "GtkPixmap*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(build), build, 2, "gtk_pixmap_set_build_insensitive", "gboolean");
  gtk_pixmap_set_build_insensitive(XEN_TO_C_GtkPixmap_(pixmap), XEN_TO_C_gboolean(build));
  return(XEN_FALSE);
}

static XEN gxg_gtk_preview_get_type(void)
{
  #define H_gtk_preview_get_type "GtkType gtk_preview_get_type( void)"
  return(C_TO_XEN_GtkType(gtk_preview_get_type()));
}

static XEN gxg_gtk_preview_uninit(void)
{
  #define H_gtk_preview_uninit "void gtk_preview_uninit( void)"
  gtk_preview_uninit();
  return(XEN_FALSE);
}

static XEN gxg_gtk_preview_new(XEN type)
{
  #define H_gtk_preview_new "GtkWidget* gtk_preview_new(GtkPreviewType type)"
  XEN_ASSERT_TYPE(XEN_GtkPreviewType_P(type), type, 1, "gtk_preview_new", "GtkPreviewType");
  return(C_TO_XEN_GtkWidget_(gtk_preview_new(XEN_TO_C_GtkPreviewType(type))));
}

static XEN gxg_gtk_preview_size(XEN preview, XEN width, XEN height)
{
  #define H_gtk_preview_size "void gtk_preview_size(GtkPreview* preview, gint width, gint height)"
  XEN_ASSERT_TYPE(XEN_GtkPreview__P(preview), preview, 1, "gtk_preview_size", "GtkPreview*");
  XEN_ASSERT_TYPE(XEN_gint_P(width), width, 2, "gtk_preview_size", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(height), height, 3, "gtk_preview_size", "gint");
  gtk_preview_size(XEN_TO_C_GtkPreview_(preview), XEN_TO_C_gint(width), XEN_TO_C_gint(height));
  return(XEN_FALSE);
}

static XEN gxg_gtk_preview_put(XEN preview, XEN window, XEN gc, XEN srcx, XEN srcy, XEN destx, XEN desty, XEN width, XEN height)
{
  #define H_gtk_preview_put "void gtk_preview_put(GtkPreview* preview, GdkWindow* window, GdkGC* gc, \
gint srcx, gint srcy, gint destx, gint desty, gint width, gint height)"
  XEN_ASSERT_TYPE(XEN_GtkPreview__P(preview), preview, 1, "gtk_preview_put", "GtkPreview*");
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 2, "gtk_preview_put", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_GdkGC__P(gc), gc, 3, "gtk_preview_put", "GdkGC*");
  XEN_ASSERT_TYPE(XEN_gint_P(srcx), srcx, 4, "gtk_preview_put", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(srcy), srcy, 5, "gtk_preview_put", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(destx), destx, 6, "gtk_preview_put", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(desty), desty, 7, "gtk_preview_put", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(width), width, 8, "gtk_preview_put", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(height), height, 9, "gtk_preview_put", "gint");
  gtk_preview_put(XEN_TO_C_GtkPreview_(preview), XEN_TO_C_GdkWindow_(window), XEN_TO_C_GdkGC_(gc), XEN_TO_C_gint(srcx), XEN_TO_C_gint(srcy), 
                  XEN_TO_C_gint(destx), XEN_TO_C_gint(desty), XEN_TO_C_gint(width), XEN_TO_C_gint(height));
  return(XEN_FALSE);
}

static XEN gxg_gtk_preview_draw_row(XEN preview, XEN data, XEN x, XEN y, XEN w)
{
  #define H_gtk_preview_draw_row "void gtk_preview_draw_row(GtkPreview* preview, guchar* data, gint x, \
gint y, gint w)"
  XEN_ASSERT_TYPE(XEN_GtkPreview__P(preview), preview, 1, "gtk_preview_draw_row", "GtkPreview*");
  XEN_ASSERT_TYPE(XEN_guchar__P(data), data, 2, "gtk_preview_draw_row", "guchar*");
  XEN_ASSERT_TYPE(XEN_gint_P(x), x, 3, "gtk_preview_draw_row", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(y), y, 4, "gtk_preview_draw_row", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(w), w, 5, "gtk_preview_draw_row", "gint");
  gtk_preview_draw_row(XEN_TO_C_GtkPreview_(preview), XEN_TO_C_guchar_(data), XEN_TO_C_gint(x), XEN_TO_C_gint(y), XEN_TO_C_gint(w));
  return(XEN_FALSE);
}

static XEN gxg_gtk_preview_set_expand(XEN preview, XEN expand)
{
  #define H_gtk_preview_set_expand "void gtk_preview_set_expand(GtkPreview* preview, gboolean expand)"
  XEN_ASSERT_TYPE(XEN_GtkPreview__P(preview), preview, 1, "gtk_preview_set_expand", "GtkPreview*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(expand), expand, 2, "gtk_preview_set_expand", "gboolean");
  gtk_preview_set_expand(XEN_TO_C_GtkPreview_(preview), XEN_TO_C_gboolean(expand));
  return(XEN_FALSE);
}

static XEN gxg_gtk_preview_set_gamma(XEN gamma)
{
  #define H_gtk_preview_set_gamma "void gtk_preview_set_gamma(double gamma)"
  XEN_ASSERT_TYPE(XEN_double_P(gamma), gamma, 1, "gtk_preview_set_gamma", "double");
  gtk_preview_set_gamma(XEN_TO_C_double(gamma));
  return(XEN_FALSE);
}

static XEN gxg_gtk_preview_set_color_cube(XEN nred_shades, XEN ngreen_shades, XEN nblue_shades, XEN ngray_shades)
{
  #define H_gtk_preview_set_color_cube "void gtk_preview_set_color_cube(guint nred_shades, guint ngreen_shades, \
guint nblue_shades, guint ngray_shades)"
  XEN_ASSERT_TYPE(XEN_guint_P(nred_shades), nred_shades, 1, "gtk_preview_set_color_cube", "guint");
  XEN_ASSERT_TYPE(XEN_guint_P(ngreen_shades), ngreen_shades, 2, "gtk_preview_set_color_cube", "guint");
  XEN_ASSERT_TYPE(XEN_guint_P(nblue_shades), nblue_shades, 3, "gtk_preview_set_color_cube", "guint");
  XEN_ASSERT_TYPE(XEN_guint_P(ngray_shades), ngray_shades, 4, "gtk_preview_set_color_cube", "guint");
  gtk_preview_set_color_cube(XEN_TO_C_guint(nred_shades), XEN_TO_C_guint(ngreen_shades), XEN_TO_C_guint(nblue_shades), XEN_TO_C_guint(ngray_shades));
  return(XEN_FALSE);
}

static XEN gxg_gtk_preview_set_install_cmap(XEN install_cmap)
{
  #define H_gtk_preview_set_install_cmap "void gtk_preview_set_install_cmap(gint install_cmap)"
  XEN_ASSERT_TYPE(XEN_gint_P(install_cmap), install_cmap, 1, "gtk_preview_set_install_cmap", "gint");
  gtk_preview_set_install_cmap(XEN_TO_C_gint(install_cmap));
  return(XEN_FALSE);
}

static XEN gxg_gtk_preview_set_reserved(XEN nreserved)
{
  #define H_gtk_preview_set_reserved "void gtk_preview_set_reserved(gint nreserved)"
  XEN_ASSERT_TYPE(XEN_gint_P(nreserved), nreserved, 1, "gtk_preview_set_reserved", "gint");
  gtk_preview_set_reserved(XEN_TO_C_gint(nreserved));
  return(XEN_FALSE);
}

static XEN gxg_gtk_preview_set_dither(XEN preview, XEN dither)
{
  #define H_gtk_preview_set_dither "void gtk_preview_set_dither(GtkPreview* preview, GdkRgbDither dither)"
  XEN_ASSERT_TYPE(XEN_GtkPreview__P(preview), preview, 1, "gtk_preview_set_dither", "GtkPreview*");
  XEN_ASSERT_TYPE(XEN_GdkRgbDither_P(dither), dither, 2, "gtk_preview_set_dither", "GdkRgbDither");
  gtk_preview_set_dither(XEN_TO_C_GtkPreview_(preview), XEN_TO_C_GdkRgbDither(dither));
  return(XEN_FALSE);
}

static XEN gxg_gtk_preview_get_visual(void)
{
  #define H_gtk_preview_get_visual "GdkVisual* gtk_preview_get_visual( void)"
  return(C_TO_XEN_GdkVisual_(gtk_preview_get_visual()));
}

static XEN gxg_gtk_preview_get_cmap(void)
{
  #define H_gtk_preview_get_cmap "GdkColormap* gtk_preview_get_cmap( void)"
  return(C_TO_XEN_GdkColormap_(gtk_preview_get_cmap()));
}

static XEN gxg_gtk_preview_get_info(void)
{
  #define H_gtk_preview_get_info "GtkPreviewInfo* gtk_preview_get_info( void)"
  return(C_TO_XEN_GtkPreviewInfo_(gtk_preview_get_info()));
}

static XEN gxg_gtk_preview_reset(void)
{
  #define H_gtk_preview_reset "void gtk_preview_reset( void)"
  gtk_preview_reset();
  return(XEN_FALSE);
}

static XEN gxg_gtk_progress_bar_new_with_adjustment(XEN adjustment)
{
  #define H_gtk_progress_bar_new_with_adjustment "GtkWidget* gtk_progress_bar_new_with_adjustment(GtkAdjustment* adjustment)"
  XEN_ASSERT_TYPE(XEN_GtkAdjustment__P(adjustment), adjustment, 1, "gtk_progress_bar_new_with_adjustment", "GtkAdjustment*");
  return(C_TO_XEN_GtkWidget_(gtk_progress_bar_new_with_adjustment(XEN_TO_C_GtkAdjustment_(adjustment))));
}

static XEN gxg_gtk_progress_bar_set_bar_style(XEN pbar, XEN style)
{
  #define H_gtk_progress_bar_set_bar_style "void gtk_progress_bar_set_bar_style(GtkProgressBar* pbar, \
GtkProgressBarStyle style)"
  XEN_ASSERT_TYPE(XEN_GtkProgressBar__P(pbar), pbar, 1, "gtk_progress_bar_set_bar_style", "GtkProgressBar*");
  XEN_ASSERT_TYPE(XEN_GtkProgressBarStyle_P(style), style, 2, "gtk_progress_bar_set_bar_style", "GtkProgressBarStyle");
  gtk_progress_bar_set_bar_style(XEN_TO_C_GtkProgressBar_(pbar), XEN_TO_C_GtkProgressBarStyle(style));
  return(XEN_FALSE);
}

static XEN gxg_gtk_progress_bar_set_discrete_blocks(XEN pbar, XEN blocks)
{
  #define H_gtk_progress_bar_set_discrete_blocks "void gtk_progress_bar_set_discrete_blocks(GtkProgressBar* pbar, \
guint blocks)"
  XEN_ASSERT_TYPE(XEN_GtkProgressBar__P(pbar), pbar, 1, "gtk_progress_bar_set_discrete_blocks", "GtkProgressBar*");
  XEN_ASSERT_TYPE(XEN_guint_P(blocks), blocks, 2, "gtk_progress_bar_set_discrete_blocks", "guint");
  gtk_progress_bar_set_discrete_blocks(XEN_TO_C_GtkProgressBar_(pbar), XEN_TO_C_guint(blocks));
  return(XEN_FALSE);
}

static XEN gxg_gtk_progress_bar_set_activity_step(XEN pbar, XEN step)
{
  #define H_gtk_progress_bar_set_activity_step "void gtk_progress_bar_set_activity_step(GtkProgressBar* pbar, \
guint step)"
  XEN_ASSERT_TYPE(XEN_GtkProgressBar__P(pbar), pbar, 1, "gtk_progress_bar_set_activity_step", "GtkProgressBar*");
  XEN_ASSERT_TYPE(XEN_guint_P(step), step, 2, "gtk_progress_bar_set_activity_step", "guint");
  gtk_progress_bar_set_activity_step(XEN_TO_C_GtkProgressBar_(pbar), XEN_TO_C_guint(step));
  return(XEN_FALSE);
}

static XEN gxg_gtk_progress_bar_set_activity_blocks(XEN pbar, XEN blocks)
{
  #define H_gtk_progress_bar_set_activity_blocks "void gtk_progress_bar_set_activity_blocks(GtkProgressBar* pbar, \
guint blocks)"
  XEN_ASSERT_TYPE(XEN_GtkProgressBar__P(pbar), pbar, 1, "gtk_progress_bar_set_activity_blocks", "GtkProgressBar*");
  XEN_ASSERT_TYPE(XEN_guint_P(blocks), blocks, 2, "gtk_progress_bar_set_activity_blocks", "guint");
  gtk_progress_bar_set_activity_blocks(XEN_TO_C_GtkProgressBar_(pbar), XEN_TO_C_guint(blocks));
  return(XEN_FALSE);
}

static XEN gxg_gtk_progress_bar_update(XEN pbar, XEN percentage)
{
  #define H_gtk_progress_bar_update "void gtk_progress_bar_update(GtkProgressBar* pbar, gdouble percentage)"
  XEN_ASSERT_TYPE(XEN_GtkProgressBar__P(pbar), pbar, 1, "gtk_progress_bar_update", "GtkProgressBar*");
  XEN_ASSERT_TYPE(XEN_gdouble_P(percentage), percentage, 2, "gtk_progress_bar_update", "gdouble");
  gtk_progress_bar_update(XEN_TO_C_GtkProgressBar_(pbar), XEN_TO_C_gdouble(percentage));
  return(XEN_FALSE);
}

static XEN gxg_gtk_progress_get_type(void)
{
  #define H_gtk_progress_get_type "GtkType gtk_progress_get_type( void)"
  return(C_TO_XEN_GtkType(gtk_progress_get_type()));
}

static XEN gxg_gtk_progress_set_show_text(XEN progress, XEN show_text)
{
  #define H_gtk_progress_set_show_text "void gtk_progress_set_show_text(GtkProgress* progress, gboolean show_text)"
  XEN_ASSERT_TYPE(XEN_GtkProgress__P(progress), progress, 1, "gtk_progress_set_show_text", "GtkProgress*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(show_text), show_text, 2, "gtk_progress_set_show_text", "gboolean");
  gtk_progress_set_show_text(XEN_TO_C_GtkProgress_(progress), XEN_TO_C_gboolean(show_text));
  return(XEN_FALSE);
}

static XEN gxg_gtk_progress_set_text_alignment(XEN progress, XEN x_align, XEN y_align)
{
  #define H_gtk_progress_set_text_alignment "void gtk_progress_set_text_alignment(GtkProgress* progress, \
gfloat x_align, gfloat y_align)"
  XEN_ASSERT_TYPE(XEN_GtkProgress__P(progress), progress, 1, "gtk_progress_set_text_alignment", "GtkProgress*");
  XEN_ASSERT_TYPE(XEN_gfloat_P(x_align), x_align, 2, "gtk_progress_set_text_alignment", "gfloat");
  XEN_ASSERT_TYPE(XEN_gfloat_P(y_align), y_align, 3, "gtk_progress_set_text_alignment", "gfloat");
  gtk_progress_set_text_alignment(XEN_TO_C_GtkProgress_(progress), XEN_TO_C_gfloat(x_align), XEN_TO_C_gfloat(y_align));
  return(XEN_FALSE);
}

static XEN gxg_gtk_progress_set_format_string(XEN progress, XEN format)
{
  #define H_gtk_progress_set_format_string "void gtk_progress_set_format_string(GtkProgress* progress, \
gchar* format)"
  XEN_ASSERT_TYPE(XEN_GtkProgress__P(progress), progress, 1, "gtk_progress_set_format_string", "GtkProgress*");
  XEN_ASSERT_TYPE(XEN_gchar__P(format), format, 2, "gtk_progress_set_format_string", "gchar*");
  gtk_progress_set_format_string(XEN_TO_C_GtkProgress_(progress), XEN_TO_C_gchar_(format));
  return(XEN_FALSE);
}

static XEN gxg_gtk_progress_set_adjustment(XEN progress, XEN adjustment)
{
  #define H_gtk_progress_set_adjustment "void gtk_progress_set_adjustment(GtkProgress* progress, GtkAdjustment* adjustment)"
  XEN_ASSERT_TYPE(XEN_GtkProgress__P(progress), progress, 1, "gtk_progress_set_adjustment", "GtkProgress*");
  XEN_ASSERT_TYPE(XEN_GtkAdjustment__P(adjustment), adjustment, 2, "gtk_progress_set_adjustment", "GtkAdjustment*");
  gtk_progress_set_adjustment(XEN_TO_C_GtkProgress_(progress), XEN_TO_C_GtkAdjustment_(adjustment));
  return(XEN_FALSE);
}

static XEN gxg_gtk_progress_configure(XEN progress, XEN value, XEN min, XEN max)
{
  #define H_gtk_progress_configure "void gtk_progress_configure(GtkProgress* progress, gdouble value, \
gdouble min, gdouble max)"
  XEN_ASSERT_TYPE(XEN_GtkProgress__P(progress), progress, 1, "gtk_progress_configure", "GtkProgress*");
  XEN_ASSERT_TYPE(XEN_gdouble_P(value), value, 2, "gtk_progress_configure", "gdouble");
  XEN_ASSERT_TYPE(XEN_gdouble_P(min), min, 3, "gtk_progress_configure", "gdouble");
  XEN_ASSERT_TYPE(XEN_gdouble_P(max), max, 4, "gtk_progress_configure", "gdouble");
  gtk_progress_configure(XEN_TO_C_GtkProgress_(progress), XEN_TO_C_gdouble(value), XEN_TO_C_gdouble(min), XEN_TO_C_gdouble(max));
  return(XEN_FALSE);
}

static XEN gxg_gtk_progress_set_percentage(XEN progress, XEN percentage)
{
  #define H_gtk_progress_set_percentage "void gtk_progress_set_percentage(GtkProgress* progress, gdouble percentage)"
  XEN_ASSERT_TYPE(XEN_GtkProgress__P(progress), progress, 1, "gtk_progress_set_percentage", "GtkProgress*");
  XEN_ASSERT_TYPE(XEN_gdouble_P(percentage), percentage, 2, "gtk_progress_set_percentage", "gdouble");
  gtk_progress_set_percentage(XEN_TO_C_GtkProgress_(progress), XEN_TO_C_gdouble(percentage));
  return(XEN_FALSE);
}

static XEN gxg_gtk_progress_set_value(XEN progress, XEN value)
{
  #define H_gtk_progress_set_value "void gtk_progress_set_value(GtkProgress* progress, gdouble value)"
  XEN_ASSERT_TYPE(XEN_GtkProgress__P(progress), progress, 1, "gtk_progress_set_value", "GtkProgress*");
  XEN_ASSERT_TYPE(XEN_gdouble_P(value), value, 2, "gtk_progress_set_value", "gdouble");
  gtk_progress_set_value(XEN_TO_C_GtkProgress_(progress), XEN_TO_C_gdouble(value));
  return(XEN_FALSE);
}

static XEN gxg_gtk_progress_get_value(XEN progress)
{
  #define H_gtk_progress_get_value "gdouble gtk_progress_get_value(GtkProgress* progress)"
  XEN_ASSERT_TYPE(XEN_GtkProgress__P(progress), progress, 1, "gtk_progress_get_value", "GtkProgress*");
  return(C_TO_XEN_gdouble(gtk_progress_get_value(XEN_TO_C_GtkProgress_(progress))));
}

static XEN gxg_gtk_progress_set_activity_mode(XEN progress, XEN activity_mode)
{
  #define H_gtk_progress_set_activity_mode "void gtk_progress_set_activity_mode(GtkProgress* progress, \
gboolean activity_mode)"
  XEN_ASSERT_TYPE(XEN_GtkProgress__P(progress), progress, 1, "gtk_progress_set_activity_mode", "GtkProgress*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(activity_mode), activity_mode, 2, "gtk_progress_set_activity_mode", "gboolean");
  gtk_progress_set_activity_mode(XEN_TO_C_GtkProgress_(progress), XEN_TO_C_gboolean(activity_mode));
  return(XEN_FALSE);
}

static XEN gxg_gtk_progress_get_current_text(XEN progress)
{
  #define H_gtk_progress_get_current_text "gchar* gtk_progress_get_current_text(GtkProgress* progress)"
  XEN_ASSERT_TYPE(XEN_GtkProgress__P(progress), progress, 1, "gtk_progress_get_current_text", "GtkProgress*");
  return(C_TO_XEN_gchar_(gtk_progress_get_current_text(XEN_TO_C_GtkProgress_(progress))));
}

static XEN gxg_gtk_progress_get_text_from_value(XEN progress, XEN value)
{
  #define H_gtk_progress_get_text_from_value "gchar* gtk_progress_get_text_from_value(GtkProgress* progress, \
gdouble value)"
  XEN_ASSERT_TYPE(XEN_GtkProgress__P(progress), progress, 1, "gtk_progress_get_text_from_value", "GtkProgress*");
  XEN_ASSERT_TYPE(XEN_gdouble_P(value), value, 2, "gtk_progress_get_text_from_value", "gdouble");
  return(C_TO_XEN_gchar_(gtk_progress_get_text_from_value(XEN_TO_C_GtkProgress_(progress), XEN_TO_C_gdouble(value))));
}

static XEN gxg_gtk_progress_get_current_percentage(XEN progress)
{
  #define H_gtk_progress_get_current_percentage "gdouble gtk_progress_get_current_percentage(GtkProgress* progress)"
  XEN_ASSERT_TYPE(XEN_GtkProgress__P(progress), progress, 1, "gtk_progress_get_current_percentage", "GtkProgress*");
  return(C_TO_XEN_gdouble(gtk_progress_get_current_percentage(XEN_TO_C_GtkProgress_(progress))));
}

static XEN gxg_gtk_progress_get_percentage_from_value(XEN progress, XEN value)
{
  #define H_gtk_progress_get_percentage_from_value "gdouble gtk_progress_get_percentage_from_value(GtkProgress* progress, \
gdouble value)"
  XEN_ASSERT_TYPE(XEN_GtkProgress__P(progress), progress, 1, "gtk_progress_get_percentage_from_value", "GtkProgress*");
  XEN_ASSERT_TYPE(XEN_gdouble_P(value), value, 2, "gtk_progress_get_percentage_from_value", "gdouble");
  return(C_TO_XEN_gdouble(gtk_progress_get_percentage_from_value(XEN_TO_C_GtkProgress_(progress), XEN_TO_C_gdouble(value))));
}

static XEN gxg_gtk_rc_add_widget_name_style(XEN rc_style, XEN pattern)
{
  #define H_gtk_rc_add_widget_name_style "void gtk_rc_add_widget_name_style(GtkRcStyle* rc_style, gchar* pattern)"
  XEN_ASSERT_TYPE(XEN_GtkRcStyle__P(rc_style), rc_style, 1, "gtk_rc_add_widget_name_style", "GtkRcStyle*");
  XEN_ASSERT_TYPE(XEN_gchar__P(pattern), pattern, 2, "gtk_rc_add_widget_name_style", "gchar*");
  gtk_rc_add_widget_name_style(XEN_TO_C_GtkRcStyle_(rc_style), XEN_TO_C_gchar_(pattern));
  return(XEN_FALSE);
}

static XEN gxg_gtk_rc_add_widget_class_style(XEN rc_style, XEN pattern)
{
  #define H_gtk_rc_add_widget_class_style "void gtk_rc_add_widget_class_style(GtkRcStyle* rc_style, gchar* pattern)"
  XEN_ASSERT_TYPE(XEN_GtkRcStyle__P(rc_style), rc_style, 1, "gtk_rc_add_widget_class_style", "GtkRcStyle*");
  XEN_ASSERT_TYPE(XEN_gchar__P(pattern), pattern, 2, "gtk_rc_add_widget_class_style", "gchar*");
  gtk_rc_add_widget_class_style(XEN_TO_C_GtkRcStyle_(rc_style), XEN_TO_C_gchar_(pattern));
  return(XEN_FALSE);
}

static XEN gxg_gtk_rc_add_class_style(XEN rc_style, XEN pattern)
{
  #define H_gtk_rc_add_class_style "void gtk_rc_add_class_style(GtkRcStyle* rc_style, gchar* pattern)"
  XEN_ASSERT_TYPE(XEN_GtkRcStyle__P(rc_style), rc_style, 1, "gtk_rc_add_class_style", "GtkRcStyle*");
  XEN_ASSERT_TYPE(XEN_gchar__P(pattern), pattern, 2, "gtk_rc_add_class_style", "gchar*");
  gtk_rc_add_class_style(XEN_TO_C_GtkRcStyle_(rc_style), XEN_TO_C_gchar_(pattern));
  return(XEN_FALSE);
}

static XEN gxg_gtk_signal_connect(XEN object, XEN name, XEN func, XEN func_data)
{
  #define H_gtk_signal_connect "guint gtk_signal_connect(GtkObject* object, gchar* name, lambda func, \
lambda_data func_data)"
  XEN_ASSERT_TYPE(XEN_GtkObject__P(object), object, 1, "gtk_signal_connect", "GtkObject*");
  XEN_ASSERT_TYPE(XEN_gchar__P(name), name, 2, "gtk_signal_connect", "gchar*");
  XEN_ASSERT_TYPE(XEN_lambda_P(func), func, 3, "gtk_signal_connect", "lambda");
  if (XEN_NOT_BOUND_P(func_data)) func_data = XEN_FALSE; 
  else XEN_ASSERT_TYPE(XEN_lambda_data_P(func_data), func_data, 4, "gtk_signal_connect", "lambda_data");
  {
    XEN gxg_ptr = XEN_LIST_5(func, func_data, XEN_FALSE, XEN_FALSE, XEN_FALSE);
    xm_protect(gxg_ptr);
    if (XEN_REQUIRED_ARGS(func) == 2)
       return(C_TO_XEN_guint(gtk_signal_connect(XEN_TO_C_GtkObject_(object), XEN_TO_C_gchar_(name), (GtkSignalFunc)gxg_func2, 
                                                XEN_TO_C_lambda_data(func_data))));
     else
       return(C_TO_XEN_guint(gtk_signal_connect(XEN_TO_C_GtkObject_(object), XEN_TO_C_gchar_(name), (GtkSignalFunc)gxg_func3, 
                                                XEN_TO_C_lambda_data(func_data))));
  }
}

static XEN gxg_gtk_signal_lookup(XEN name, XEN object_type)
{
  #define H_gtk_signal_lookup "guint gtk_signal_lookup(gchar* name, GtkType object_type)"
  XEN_ASSERT_TYPE(XEN_gchar__P(name), name, 1, "gtk_signal_lookup", "gchar*");
  XEN_ASSERT_TYPE(XEN_GtkType_P(object_type), object_type, 2, "gtk_signal_lookup", "GtkType");
  return(C_TO_XEN_guint(gtk_signal_lookup(XEN_TO_C_gchar_(name), XEN_TO_C_GtkType(object_type))));
}

static XEN gxg_gtk_signal_name(XEN signal_id)
{
  #define H_gtk_signal_name "gchar* gtk_signal_name(guint signal_id)"
  XEN_ASSERT_TYPE(XEN_guint_P(signal_id), signal_id, 1, "gtk_signal_name", "guint");
  return(C_TO_XEN_gchar_(gtk_signal_name(XEN_TO_C_guint(signal_id))));
}

static XEN gxg_gtk_signal_emit_stop(XEN object, XEN signal_id)
{
  #define H_gtk_signal_emit_stop "void gtk_signal_emit_stop(GtkObject* object, guint signal_id)"
  XEN_ASSERT_TYPE(XEN_GtkObject__P(object), object, 1, "gtk_signal_emit_stop", "GtkObject*");
  XEN_ASSERT_TYPE(XEN_guint_P(signal_id), signal_id, 2, "gtk_signal_emit_stop", "guint");
  gtk_signal_emit_stop(XEN_TO_C_GtkObject_(object), XEN_TO_C_guint(signal_id));
  return(XEN_FALSE);
}

static XEN gxg_gtk_signal_emit_stop_by_name(XEN object, XEN name)
{
  #define H_gtk_signal_emit_stop_by_name "void gtk_signal_emit_stop_by_name(GtkObject* object, gchar* name)"
  XEN_ASSERT_TYPE(XEN_GtkObject__P(object), object, 1, "gtk_signal_emit_stop_by_name", "GtkObject*");
  XEN_ASSERT_TYPE(XEN_gchar__P(name), name, 2, "gtk_signal_emit_stop_by_name", "gchar*");
  gtk_signal_emit_stop_by_name(XEN_TO_C_GtkObject_(object), XEN_TO_C_gchar_(name));
  return(XEN_FALSE);
}

static XEN gxg_gtk_signal_connect_after(XEN object, XEN name, XEN func, XEN func_data)
{
  #define H_gtk_signal_connect_after "guint gtk_signal_connect_after(GtkObject* object, gchar* name, \
lambda func, lambda_data func_data)"
  XEN_ASSERT_TYPE(XEN_GtkObject__P(object), object, 1, "gtk_signal_connect_after", "GtkObject*");
  XEN_ASSERT_TYPE(XEN_gchar__P(name), name, 2, "gtk_signal_connect_after", "gchar*");
  XEN_ASSERT_TYPE(XEN_lambda_P(func), func, 3, "gtk_signal_connect_after", "lambda");
  if (XEN_NOT_BOUND_P(func_data)) func_data = XEN_FALSE; 
  else XEN_ASSERT_TYPE(XEN_lambda_data_P(func_data), func_data, 4, "gtk_signal_connect_after", "lambda_data");
  {
    XEN gxg_ptr = XEN_LIST_5(func, func_data, XEN_FALSE, XEN_FALSE, XEN_FALSE);
    xm_protect(gxg_ptr);
    if (XEN_REQUIRED_ARGS(func) == 2)
       return(C_TO_XEN_guint(gtk_signal_connect_after(XEN_TO_C_GtkObject_(object), XEN_TO_C_gchar_(name), (GtkSignalFunc)gxg_func2, 
                                                      XEN_TO_C_lambda_data(func_data))));
     else
       return(C_TO_XEN_guint(gtk_signal_connect_after(XEN_TO_C_GtkObject_(object), XEN_TO_C_gchar_(name), (GtkSignalFunc)gxg_func3, 
                                                      XEN_TO_C_lambda_data(func_data))));
  }
}

static XEN gxg_gtk_signal_connect_full(XEN object, XEN name, XEN func, XEN marshal, XEN func_data, XEN destroy_func, XEN object_signal, XEN after)
{
  #define H_gtk_signal_connect_full "guint gtk_signal_connect_full(GtkObject* object, gchar* name, lambda func, \
GtkCallbackMarshal marshal, lambda_data func_data, GtkDestroyNotify destroy_func, gint object_signal, \
gint after)"
  XEN_ASSERT_TYPE(XEN_GtkObject__P(object), object, 1, "gtk_signal_connect_full", "GtkObject*");
  XEN_ASSERT_TYPE(XEN_gchar__P(name), name, 2, "gtk_signal_connect_full", "gchar*");
  XEN_ASSERT_TYPE(XEN_lambda_P(func), func, 3, "gtk_signal_connect_full", "lambda");
  XEN_ASSERT_TYPE(XEN_GtkCallbackMarshal_P(marshal), marshal, 4, "gtk_signal_connect_full", "GtkCallbackMarshal");
  XEN_ASSERT_TYPE(XEN_lambda_data_P(func_data), func_data, 5, "gtk_signal_connect_full", "lambda_data");
  XEN_ASSERT_TYPE(XEN_GtkDestroyNotify_P(destroy_func), destroy_func, 6, "gtk_signal_connect_full", "GtkDestroyNotify");
  XEN_ASSERT_TYPE(XEN_gint_P(object_signal), object_signal, 7, "gtk_signal_connect_full", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(after), after, 8, "gtk_signal_connect_full", "gint");
  {
    XEN gxg_ptr = XEN_LIST_5(func, func_data, XEN_FALSE, XEN_FALSE, XEN_FALSE);
    xm_protect(gxg_ptr);
    XEN_LIST_SET(gxg_ptr, 3, destroy_func);
    if (XEN_REQUIRED_ARGS(func) == 2)
       return(C_TO_XEN_guint(gtk_signal_connect_full(XEN_TO_C_GtkObject_(object), XEN_TO_C_gchar_(name), (GtkSignalFunc)gxg_func2, 
                                                     XEN_TO_C_GtkCallbackMarshal(marshal), XEN_TO_C_lambda_data(func_data), 
                                                     XEN_TO_C_GtkDestroyNotify(destroy_func), XEN_TO_C_gint(object_signal), 
                                                     XEN_TO_C_gint(after))));
     else
       return(C_TO_XEN_guint(gtk_signal_connect_full(XEN_TO_C_GtkObject_(object), XEN_TO_C_gchar_(name), (GtkSignalFunc)gxg_func3, 
                                                     XEN_TO_C_GtkCallbackMarshal(marshal), XEN_TO_C_lambda_data(func_data), 
                                                     XEN_TO_C_GtkDestroyNotify(destroy_func), XEN_TO_C_gint(object_signal), 
                                                     XEN_TO_C_gint(after))));
  }
}

static XEN gxg_gtk_signal_connect_while_alive(XEN object, XEN signal, XEN func, XEN func_data, XEN alive_object)
{
  #define H_gtk_signal_connect_while_alive "void gtk_signal_connect_while_alive(GtkObject* object, gchar* signal, \
lambda func, lambda_data func_data, GtkObject* alive_object)"
  XEN_ASSERT_TYPE(XEN_GtkObject__P(object), object, 1, "gtk_signal_connect_while_alive", "GtkObject*");
  XEN_ASSERT_TYPE(XEN_gchar__P(signal), signal, 2, "gtk_signal_connect_while_alive", "gchar*");
  XEN_ASSERT_TYPE(XEN_lambda_P(func), func, 3, "gtk_signal_connect_while_alive", "lambda");
  XEN_ASSERT_TYPE(XEN_lambda_data_P(func_data), func_data, 4, "gtk_signal_connect_while_alive", "lambda_data");
  XEN_ASSERT_TYPE(XEN_GtkObject__P(alive_object), alive_object, 5, "gtk_signal_connect_while_alive", "GtkObject*");
  {
    XEN gxg_ptr = XEN_LIST_5(func, func_data, XEN_FALSE, XEN_FALSE, XEN_FALSE);
    xm_protect(gxg_ptr);
    if (XEN_REQUIRED_ARGS(func) == 2)
       gtk_signal_connect_while_alive(XEN_TO_C_GtkObject_(object), XEN_TO_C_gchar_(signal), (GtkSignalFunc)gxg_func2, XEN_TO_C_lambda_data(func_data), 
                                      XEN_TO_C_GtkObject_(alive_object));
     else
       gtk_signal_connect_while_alive(XEN_TO_C_GtkObject_(object), XEN_TO_C_gchar_(signal), (GtkSignalFunc)gxg_func3, XEN_TO_C_lambda_data(func_data), 
                                      XEN_TO_C_GtkObject_(alive_object));
    return(XEN_FALSE);
  }
}

static XEN gxg_gtk_signal_disconnect(XEN object, XEN handler_id)
{
  #define H_gtk_signal_disconnect "void gtk_signal_disconnect(GtkObject* object, guint handler_id)"
  XEN_ASSERT_TYPE(XEN_GtkObject__P(object), object, 1, "gtk_signal_disconnect", "GtkObject*");
  XEN_ASSERT_TYPE(XEN_guint_P(handler_id), handler_id, 2, "gtk_signal_disconnect", "guint");
  gtk_signal_disconnect(XEN_TO_C_GtkObject_(object), XEN_TO_C_guint(handler_id));
  return(XEN_FALSE);
}

static XEN gxg_gtk_signal_disconnect_by_func(XEN object, XEN func, XEN func_data)
{
  #define H_gtk_signal_disconnect_by_func "void gtk_signal_disconnect_by_func(GtkObject* object, lambda func, \
lambda_data func_data)"
  XEN_ASSERT_TYPE(XEN_GtkObject__P(object), object, 1, "gtk_signal_disconnect_by_func", "GtkObject*");
  XEN_ASSERT_TYPE(XEN_lambda_P(func), func, 2, "gtk_signal_disconnect_by_func", "lambda");
  if (XEN_NOT_BOUND_P(func_data)) func_data = XEN_FALSE; 
  else XEN_ASSERT_TYPE(XEN_lambda_data_P(func_data), func_data, 3, "gtk_signal_disconnect_by_func", "lambda_data");
  {
    XEN gxg_ptr = XEN_LIST_5(func, func_data, XEN_FALSE, XEN_FALSE, XEN_FALSE);
    xm_protect(gxg_ptr);
    if (XEN_REQUIRED_ARGS(func) == 2)
       gtk_signal_disconnect_by_func(XEN_TO_C_GtkObject_(object), (GtkSignalFunc)gxg_func2, XEN_TO_C_lambda_data(func_data));
     else
       gtk_signal_disconnect_by_func(XEN_TO_C_GtkObject_(object), (GtkSignalFunc)gxg_func3, XEN_TO_C_lambda_data(func_data));
    return(XEN_FALSE);
  }
}

static XEN gxg_gtk_signal_handler_block(XEN object, XEN handler_id)
{
  #define H_gtk_signal_handler_block "void gtk_signal_handler_block(GtkObject* object, guint handler_id)"
  XEN_ASSERT_TYPE(XEN_GtkObject__P(object), object, 1, "gtk_signal_handler_block", "GtkObject*");
  XEN_ASSERT_TYPE(XEN_guint_P(handler_id), handler_id, 2, "gtk_signal_handler_block", "guint");
  gtk_signal_handler_block(XEN_TO_C_GtkObject_(object), XEN_TO_C_guint(handler_id));
  return(XEN_FALSE);
}

static XEN gxg_gtk_signal_handler_block_by_func(XEN object, XEN func, XEN func_data)
{
  #define H_gtk_signal_handler_block_by_func "void gtk_signal_handler_block_by_func(GtkObject* object, \
lambda func, lambda_data func_data)"
  XEN_ASSERT_TYPE(XEN_GtkObject__P(object), object, 1, "gtk_signal_handler_block_by_func", "GtkObject*");
  XEN_ASSERT_TYPE(XEN_lambda_P(func), func, 2, "gtk_signal_handler_block_by_func", "lambda");
  if (XEN_NOT_BOUND_P(func_data)) func_data = XEN_FALSE; 
  else XEN_ASSERT_TYPE(XEN_lambda_data_P(func_data), func_data, 3, "gtk_signal_handler_block_by_func", "lambda_data");
  {
    XEN gxg_ptr = XEN_LIST_5(func, func_data, XEN_FALSE, XEN_FALSE, XEN_FALSE);
    xm_protect(gxg_ptr);
    if (XEN_REQUIRED_ARGS(func) == 2)
       gtk_signal_handler_block_by_func(XEN_TO_C_GtkObject_(object), (GtkSignalFunc)gxg_func2, XEN_TO_C_lambda_data(func_data));
     else
       gtk_signal_handler_block_by_func(XEN_TO_C_GtkObject_(object), (GtkSignalFunc)gxg_func3, XEN_TO_C_lambda_data(func_data));
    return(XEN_FALSE);
  }
}

static XEN gxg_gtk_signal_handler_unblock(XEN object, XEN handler_id)
{
  #define H_gtk_signal_handler_unblock "void gtk_signal_handler_unblock(GtkObject* object, guint handler_id)"
  XEN_ASSERT_TYPE(XEN_GtkObject__P(object), object, 1, "gtk_signal_handler_unblock", "GtkObject*");
  XEN_ASSERT_TYPE(XEN_guint_P(handler_id), handler_id, 2, "gtk_signal_handler_unblock", "guint");
  gtk_signal_handler_unblock(XEN_TO_C_GtkObject_(object), XEN_TO_C_guint(handler_id));
  return(XEN_FALSE);
}

static XEN gxg_gtk_signal_handler_unblock_by_func(XEN object, XEN func, XEN func_data)
{
  #define H_gtk_signal_handler_unblock_by_func "void gtk_signal_handler_unblock_by_func(GtkObject* object, \
lambda func, lambda_data func_data)"
  XEN_ASSERT_TYPE(XEN_GtkObject__P(object), object, 1, "gtk_signal_handler_unblock_by_func", "GtkObject*");
  XEN_ASSERT_TYPE(XEN_lambda_P(func), func, 2, "gtk_signal_handler_unblock_by_func", "lambda");
  if (XEN_NOT_BOUND_P(func_data)) func_data = XEN_FALSE; 
  else XEN_ASSERT_TYPE(XEN_lambda_data_P(func_data), func_data, 3, "gtk_signal_handler_unblock_by_func", "lambda_data");
  {
    XEN gxg_ptr = XEN_LIST_5(func, func_data, XEN_FALSE, XEN_FALSE, XEN_FALSE);
    xm_protect(gxg_ptr);
    if (XEN_REQUIRED_ARGS(func) == 2)
       gtk_signal_handler_unblock_by_func(XEN_TO_C_GtkObject_(object), (GtkSignalFunc)gxg_func2, XEN_TO_C_lambda_data(func_data));
     else
       gtk_signal_handler_unblock_by_func(XEN_TO_C_GtkObject_(object), (GtkSignalFunc)gxg_func3, XEN_TO_C_lambda_data(func_data));
    return(XEN_FALSE);
  }
}

static XEN gxg_gtk_signal_handler_pending(XEN object, XEN signal_id, XEN may_be_blocked)
{
  #define H_gtk_signal_handler_pending "guint gtk_signal_handler_pending(GtkObject* object, guint signal_id, \
gboolean may_be_blocked)"
  XEN_ASSERT_TYPE(XEN_GtkObject__P(object), object, 1, "gtk_signal_handler_pending", "GtkObject*");
  XEN_ASSERT_TYPE(XEN_guint_P(signal_id), signal_id, 2, "gtk_signal_handler_pending", "guint");
  XEN_ASSERT_TYPE(XEN_gboolean_P(may_be_blocked), may_be_blocked, 3, "gtk_signal_handler_pending", "gboolean");
  return(C_TO_XEN_guint(gtk_signal_handler_pending(XEN_TO_C_GtkObject_(object), XEN_TO_C_guint(signal_id), XEN_TO_C_gboolean(may_be_blocked))));
}

static XEN gxg_gtk_signal_handler_pending_by_func(XEN object, XEN signal_id, XEN may_be_blocked, XEN func, XEN func_data)
{
  #define H_gtk_signal_handler_pending_by_func "guint gtk_signal_handler_pending_by_func(GtkObject* object, \
guint signal_id, gboolean may_be_blocked, lambda func, lambda_data func_data)"
  XEN_ASSERT_TYPE(XEN_GtkObject__P(object), object, 1, "gtk_signal_handler_pending_by_func", "GtkObject*");
  XEN_ASSERT_TYPE(XEN_guint_P(signal_id), signal_id, 2, "gtk_signal_handler_pending_by_func", "guint");
  XEN_ASSERT_TYPE(XEN_gboolean_P(may_be_blocked), may_be_blocked, 3, "gtk_signal_handler_pending_by_func", "gboolean");
  XEN_ASSERT_TYPE(XEN_lambda_P(func), func, 4, "gtk_signal_handler_pending_by_func", "lambda");
  if (XEN_NOT_BOUND_P(func_data)) func_data = XEN_FALSE; 
  else XEN_ASSERT_TYPE(XEN_lambda_data_P(func_data), func_data, 5, "gtk_signal_handler_pending_by_func", "lambda_data");
  {
    XEN gxg_ptr = XEN_LIST_5(func, func_data, XEN_FALSE, XEN_FALSE, XEN_FALSE);
    xm_protect(gxg_ptr);
    if (XEN_REQUIRED_ARGS(func) == 2)
       return(C_TO_XEN_guint(gtk_signal_handler_pending_by_func(XEN_TO_C_GtkObject_(object), XEN_TO_C_guint(signal_id), XEN_TO_C_gboolean(may_be_blocked), 
                                                                (GtkSignalFunc)gxg_func2, XEN_TO_C_lambda_data(func_data))));
     else
       return(C_TO_XEN_guint(gtk_signal_handler_pending_by_func(XEN_TO_C_GtkObject_(object), XEN_TO_C_guint(signal_id), XEN_TO_C_gboolean(may_be_blocked), 
                                                                (GtkSignalFunc)gxg_func3, XEN_TO_C_lambda_data(func_data))));
  }
}

static XEN gxg_gtk_signal_new(XEN name, XEN signal_flags, XEN object_type, XEN function_offset, XEN marshaller, XEN return_val, XEN nparams)
{
  #define H_gtk_signal_new "guint gtk_signal_new(gchar* name, GtkSignalRunType signal_flags, GtkType object_type, \
guint function_offset, GtkSignalMarshaller marshaller, GtkType return_val, guint nparams, ...)"
  XEN_ASSERT_TYPE(XEN_gchar__P(name), name, 1, "gtk_signal_new", "gchar*");
  XEN_ASSERT_TYPE(XEN_GtkSignalRunType_P(signal_flags), signal_flags, 2, "gtk_signal_new", "GtkSignalRunType");
  XEN_ASSERT_TYPE(XEN_GtkType_P(object_type), object_type, 3, "gtk_signal_new", "GtkType");
  XEN_ASSERT_TYPE(XEN_guint_P(function_offset), function_offset, 4, "gtk_signal_new", "guint");
  XEN_ASSERT_TYPE(XEN_GtkSignalMarshaller_P(marshaller), marshaller, 5, "gtk_signal_new", "GtkSignalMarshaller");
  XEN_ASSERT_TYPE(XEN_GtkType_P(return_val), return_val, 6, "gtk_signal_new", "GtkType");
  XEN_ASSERT_TYPE(XEN_guint_P(nparams), nparams, 7, "gtk_signal_new", "guint");
  return(C_TO_XEN_guint(gtk_signal_new(XEN_TO_C_gchar_(name), XEN_TO_C_GtkSignalRunType(signal_flags), XEN_TO_C_GtkType(object_type), 
                                       XEN_TO_C_guint(function_offset), XEN_TO_C_GtkSignalMarshaller(marshaller), XEN_TO_C_GtkType(return_val), 
                                       XEN_TO_C_guint(nparams))));
}

static XEN gxg_gtk_signal_newv(XEN name, XEN signal_flags, XEN object_type, XEN function_offset, XEN marshaller, XEN return_val, XEN nparams, XEN params)
{
  #define H_gtk_signal_newv "guint gtk_signal_newv(gchar* name, GtkSignalRunType signal_flags, GtkType object_type, \
guint function_offset, GtkSignalMarshaller marshaller, GtkType return_val, guint nparams, GtkType* params)"
  XEN_ASSERT_TYPE(XEN_gchar__P(name), name, 1, "gtk_signal_newv", "gchar*");
  XEN_ASSERT_TYPE(XEN_GtkSignalRunType_P(signal_flags), signal_flags, 2, "gtk_signal_newv", "GtkSignalRunType");
  XEN_ASSERT_TYPE(XEN_GtkType_P(object_type), object_type, 3, "gtk_signal_newv", "GtkType");
  XEN_ASSERT_TYPE(XEN_guint_P(function_offset), function_offset, 4, "gtk_signal_newv", "guint");
  XEN_ASSERT_TYPE(XEN_GtkSignalMarshaller_P(marshaller), marshaller, 5, "gtk_signal_newv", "GtkSignalMarshaller");
  XEN_ASSERT_TYPE(XEN_GtkType_P(return_val), return_val, 6, "gtk_signal_newv", "GtkType");
  XEN_ASSERT_TYPE(XEN_guint_P(nparams), nparams, 7, "gtk_signal_newv", "guint");
  XEN_ASSERT_TYPE(XEN_GtkType__P(params), params, 8, "gtk_signal_newv", "GtkType*");
  return(C_TO_XEN_guint(gtk_signal_newv(XEN_TO_C_gchar_(name), XEN_TO_C_GtkSignalRunType(signal_flags), XEN_TO_C_GtkType(object_type), 
                                        XEN_TO_C_guint(function_offset), XEN_TO_C_GtkSignalMarshaller(marshaller), XEN_TO_C_GtkType(return_val), 
                                        XEN_TO_C_guint(nparams), XEN_TO_C_GtkType_(params))));
}

static XEN gxg_gtk_signal_emit(XEN object, XEN signal_id)
{
  #define H_gtk_signal_emit "void gtk_signal_emit(GtkObject* object, guint signal_id, ...)"
  XEN_ASSERT_TYPE(XEN_GtkObject__P(object), object, 1, "gtk_signal_emit", "GtkObject*");
  XEN_ASSERT_TYPE(XEN_guint_P(signal_id), signal_id, 2, "gtk_signal_emit", "guint");
  gtk_signal_emit(XEN_TO_C_GtkObject_(object), XEN_TO_C_guint(signal_id));
  return(XEN_FALSE);
}

static XEN gxg_gtk_signal_emit_by_name(XEN object, XEN name)
{
  #define H_gtk_signal_emit_by_name "void gtk_signal_emit_by_name(GtkObject* object, gchar* name, ...)"
  XEN_ASSERT_TYPE(XEN_GtkObject__P(object), object, 1, "gtk_signal_emit_by_name", "GtkObject*");
  XEN_ASSERT_TYPE(XEN_gchar__P(name), name, 2, "gtk_signal_emit_by_name", "gchar*");
  gtk_signal_emit_by_name(XEN_TO_C_GtkObject_(object), XEN_TO_C_gchar_(name));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_signal_compat_matched(XEN object, XEN func, XEN func_data, XEN match, XEN action)
{
  #define H_gtk_signal_compat_matched "void gtk_signal_compat_matched(GtkObject* object, lambda func, \
lambda_data func_data, GSignalMatchType match, guint action)"
  XEN_ASSERT_TYPE(XEN_GtkObject__P(object), object, 1, "gtk_signal_compat_matched", "GtkObject*");
  XEN_ASSERT_TYPE(XEN_lambda_P(func), func, 2, "gtk_signal_compat_matched", "lambda");
  XEN_ASSERT_TYPE(XEN_lambda_data_P(func_data), func_data, 3, "gtk_signal_compat_matched", "lambda_data");
  XEN_ASSERT_TYPE(XEN_GSignalMatchType_P(match), match, 4, "gtk_signal_compat_matched", "GSignalMatchType");
  XEN_ASSERT_TYPE(XEN_guint_P(action), action, 5, "gtk_signal_compat_matched", "guint");
  {
    XEN gxg_ptr = XEN_LIST_5(func, func_data, XEN_FALSE, XEN_FALSE, XEN_FALSE);
    xm_protect(gxg_ptr);
    if (XEN_REQUIRED_ARGS(func) == 2)
       gtk_signal_compat_matched(XEN_TO_C_GtkObject_(object), (GtkSignalFunc)gxg_func2, XEN_TO_C_lambda_data(func_data), 
                                 XEN_TO_C_GSignalMatchType(match), XEN_TO_C_guint(action));
     else
       gtk_signal_compat_matched(XEN_TO_C_GtkObject_(object), (GtkSignalFunc)gxg_func3, XEN_TO_C_lambda_data(func_data), 
                                 XEN_TO_C_GSignalMatchType(match), XEN_TO_C_guint(action));
    return(XEN_FALSE);
  }
}
#endif

static XEN gxg_gtk_socket_steal(XEN socket, XEN wid)
{
  #define H_gtk_socket_steal "void gtk_socket_steal(GtkSocket* socket, GdkNativeWindow wid)"
  XEN_ASSERT_TYPE(XEN_GtkSocket__P(socket), socket, 1, "gtk_socket_steal", "GtkSocket*");
  XEN_ASSERT_TYPE(XEN_GdkNativeWindow_P(wid), wid, 2, "gtk_socket_steal", "GdkNativeWindow");
  gtk_socket_steal(XEN_TO_C_GtkSocket_(socket), XEN_TO_C_GdkNativeWindow(wid));
  return(XEN_FALSE);
}

static XEN gxg_gtk_style_ref(XEN style)
{
  #define H_gtk_style_ref "GtkStyle* gtk_style_ref(GtkStyle* style)"
  XEN_ASSERT_TYPE(XEN_GtkStyle__P(style), style, 1, "gtk_style_ref", "GtkStyle*");
  return(C_TO_XEN_GtkStyle_(gtk_style_ref(XEN_TO_C_GtkStyle_(style))));
}

static XEN gxg_gtk_style_unref(XEN style)
{
  #define H_gtk_style_unref "void gtk_style_unref(GtkStyle* style)"
  XEN_ASSERT_TYPE(XEN_GtkStyle__P(style), style, 1, "gtk_style_unref", "GtkStyle*");
  gtk_style_unref(XEN_TO_C_GtkStyle_(style));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_style_get_font(XEN style)
{
  #define H_gtk_style_get_font "GdkFont* gtk_style_get_font(GtkStyle* style)"
  XEN_ASSERT_TYPE(XEN_GtkStyle__P(style), style, 1, "gtk_style_get_font", "GtkStyle*");
  return(C_TO_XEN_GdkFont_(gtk_style_get_font(XEN_TO_C_GtkStyle_(style))));
}

static XEN gxg_gtk_style_set_font(XEN style, XEN font)
{
  #define H_gtk_style_set_font "void gtk_style_set_font(GtkStyle* style, GdkFont* font)"
  XEN_ASSERT_TYPE(XEN_GtkStyle__P(style), style, 1, "gtk_style_set_font", "GtkStyle*");
  XEN_ASSERT_TYPE(XEN_GdkFont__P(font) || XEN_FALSE_P(font), font, 2, "gtk_style_set_font", "GdkFont*");
  gtk_style_set_font(XEN_TO_C_GtkStyle_(style), XEN_TO_C_GdkFont_(font));
  return(XEN_FALSE);
}
#endif

static XEN gxg_gtk_draw_hline(XEN style, XEN window, XEN state_type, XEN x1, XEN x2, XEN y)
{
  #define H_gtk_draw_hline "void gtk_draw_hline(GtkStyle* style, GdkWindow* window, GtkStateType state_type, \
gint x1, gint x2, gint y)"
  XEN_ASSERT_TYPE(XEN_GtkStyle__P(style), style, 1, "gtk_draw_hline", "GtkStyle*");
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 2, "gtk_draw_hline", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_GtkStateType_P(state_type), state_type, 3, "gtk_draw_hline", "GtkStateType");
  XEN_ASSERT_TYPE(XEN_gint_P(x1), x1, 4, "gtk_draw_hline", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(x2), x2, 5, "gtk_draw_hline", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(y), y, 6, "gtk_draw_hline", "gint");
  gtk_draw_hline(XEN_TO_C_GtkStyle_(style), XEN_TO_C_GdkWindow_(window), XEN_TO_C_GtkStateType(state_type), XEN_TO_C_gint(x1), 
                 XEN_TO_C_gint(x2), XEN_TO_C_gint(y));
  return(XEN_FALSE);
}

static XEN gxg_gtk_draw_vline(XEN style, XEN window, XEN state_type, XEN y1, XEN y2, XEN x)
{
  #define H_gtk_draw_vline "void gtk_draw_vline(GtkStyle* style, GdkWindow* window, GtkStateType state_type, \
gint y1, gint y2, gint x)"
  XEN_ASSERT_TYPE(XEN_GtkStyle__P(style), style, 1, "gtk_draw_vline", "GtkStyle*");
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 2, "gtk_draw_vline", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_GtkStateType_P(state_type), state_type, 3, "gtk_draw_vline", "GtkStateType");
  XEN_ASSERT_TYPE(XEN_gint_P(y1), y1, 4, "gtk_draw_vline", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(y2), y2, 5, "gtk_draw_vline", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(x), x, 6, "gtk_draw_vline", "gint");
  gtk_draw_vline(XEN_TO_C_GtkStyle_(style), XEN_TO_C_GdkWindow_(window), XEN_TO_C_GtkStateType(state_type), XEN_TO_C_gint(y1), 
                 XEN_TO_C_gint(y2), XEN_TO_C_gint(x));
  return(XEN_FALSE);
}

static XEN gxg_gtk_draw_shadow(XEN style, XEN window, XEN state_type, XEN shadow_type, XEN x, XEN y, XEN width, XEN height)
{
  #define H_gtk_draw_shadow "void gtk_draw_shadow(GtkStyle* style, GdkWindow* window, GtkStateType state_type, \
GtkShadowType shadow_type, gint x, gint y, gint width, gint height)"
  XEN_ASSERT_TYPE(XEN_GtkStyle__P(style), style, 1, "gtk_draw_shadow", "GtkStyle*");
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 2, "gtk_draw_shadow", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_GtkStateType_P(state_type), state_type, 3, "gtk_draw_shadow", "GtkStateType");
  XEN_ASSERT_TYPE(XEN_GtkShadowType_P(shadow_type), shadow_type, 4, "gtk_draw_shadow", "GtkShadowType");
  XEN_ASSERT_TYPE(XEN_gint_P(x), x, 5, "gtk_draw_shadow", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(y), y, 6, "gtk_draw_shadow", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(width), width, 7, "gtk_draw_shadow", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(height), height, 8, "gtk_draw_shadow", "gint");
  gtk_draw_shadow(XEN_TO_C_GtkStyle_(style), XEN_TO_C_GdkWindow_(window), XEN_TO_C_GtkStateType(state_type), XEN_TO_C_GtkShadowType(shadow_type), 
                  XEN_TO_C_gint(x), XEN_TO_C_gint(y), XEN_TO_C_gint(width), XEN_TO_C_gint(height));
  return(XEN_FALSE);
}

static XEN gxg_gtk_draw_polygon(XEN style, XEN window, XEN state_type, XEN shadow_type, XEN points, XEN npoints, XEN fill)
{
  #define H_gtk_draw_polygon "void gtk_draw_polygon(GtkStyle* style, GdkWindow* window, GtkStateType state_type, \
GtkShadowType shadow_type, GdkPoint* points, gint npoints, gboolean fill)"
  XEN_ASSERT_TYPE(XEN_GtkStyle__P(style), style, 1, "gtk_draw_polygon", "GtkStyle*");
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 2, "gtk_draw_polygon", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_GtkStateType_P(state_type), state_type, 3, "gtk_draw_polygon", "GtkStateType");
  XEN_ASSERT_TYPE(XEN_GtkShadowType_P(shadow_type), shadow_type, 4, "gtk_draw_polygon", "GtkShadowType");
  XEN_ASSERT_TYPE(XEN_GdkPoint__P(points), points, 5, "gtk_draw_polygon", "GdkPoint*");
  XEN_ASSERT_TYPE(XEN_gint_P(npoints), npoints, 6, "gtk_draw_polygon", "gint");
  XEN_ASSERT_TYPE(XEN_gboolean_P(fill), fill, 7, "gtk_draw_polygon", "gboolean");
  gtk_draw_polygon(XEN_TO_C_GtkStyle_(style), XEN_TO_C_GdkWindow_(window), XEN_TO_C_GtkStateType(state_type), XEN_TO_C_GtkShadowType(shadow_type), 
                   XEN_TO_C_GdkPoint_(points), XEN_TO_C_gint(npoints), XEN_TO_C_gboolean(fill));
  return(XEN_FALSE);
}

static XEN gxg_gtk_draw_arrow(XEN arglist)
{
  #define H_gtk_draw_arrow "void gtk_draw_arrow(GtkStyle* style, GdkWindow* window, GtkStateType state_type, \
GtkShadowType shadow_type, GtkArrowType arrow_type, gboolean fill, gint x, gint y, gint width, gint height)"
  XEN style, window, state_type, shadow_type, arrow_type, fill, x, y, width, height;
  style = XEN_LIST_REF(arglist, 0);
  window = XEN_LIST_REF(arglist, 1);
  state_type = XEN_LIST_REF(arglist, 2);
  shadow_type = XEN_LIST_REF(arglist, 3);
  arrow_type = XEN_LIST_REF(arglist, 4);
  fill = XEN_LIST_REF(arglist, 5);
  x = XEN_LIST_REF(arglist, 6);
  y = XEN_LIST_REF(arglist, 7);
  width = XEN_LIST_REF(arglist, 8);
  height = XEN_LIST_REF(arglist, 9);
  XEN_ASSERT_TYPE(XEN_GtkStyle__P(style), style, 1, "gtk_draw_arrow", "GtkStyle*");
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 2, "gtk_draw_arrow", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_GtkStateType_P(state_type), state_type, 3, "gtk_draw_arrow", "GtkStateType");
  XEN_ASSERT_TYPE(XEN_GtkShadowType_P(shadow_type), shadow_type, 4, "gtk_draw_arrow", "GtkShadowType");
  XEN_ASSERT_TYPE(XEN_GtkArrowType_P(arrow_type), arrow_type, 5, "gtk_draw_arrow", "GtkArrowType");
  XEN_ASSERT_TYPE(XEN_gboolean_P(fill), fill, 6, "gtk_draw_arrow", "gboolean");
  XEN_ASSERT_TYPE(XEN_gint_P(x), x, 7, "gtk_draw_arrow", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(y), y, 8, "gtk_draw_arrow", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(width), width, 9, "gtk_draw_arrow", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(height), height, 10, "gtk_draw_arrow", "gint");
  gtk_draw_arrow(XEN_TO_C_GtkStyle_(style), XEN_TO_C_GdkWindow_(window), XEN_TO_C_GtkStateType(state_type), XEN_TO_C_GtkShadowType(shadow_type), 
                 XEN_TO_C_GtkArrowType(arrow_type), XEN_TO_C_gboolean(fill), XEN_TO_C_gint(x), XEN_TO_C_gint(y), XEN_TO_C_gint(width), 
                 XEN_TO_C_gint(height));
  return(XEN_FALSE);
}

static XEN gxg_gtk_draw_diamond(XEN style, XEN window, XEN state_type, XEN shadow_type, XEN x, XEN y, XEN width, XEN height)
{
  #define H_gtk_draw_diamond "void gtk_draw_diamond(GtkStyle* style, GdkWindow* window, GtkStateType state_type, \
GtkShadowType shadow_type, gint x, gint y, gint width, gint height)"
  XEN_ASSERT_TYPE(XEN_GtkStyle__P(style), style, 1, "gtk_draw_diamond", "GtkStyle*");
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 2, "gtk_draw_diamond", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_GtkStateType_P(state_type), state_type, 3, "gtk_draw_diamond", "GtkStateType");
  XEN_ASSERT_TYPE(XEN_GtkShadowType_P(shadow_type), shadow_type, 4, "gtk_draw_diamond", "GtkShadowType");
  XEN_ASSERT_TYPE(XEN_gint_P(x), x, 5, "gtk_draw_diamond", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(y), y, 6, "gtk_draw_diamond", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(width), width, 7, "gtk_draw_diamond", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(height), height, 8, "gtk_draw_diamond", "gint");
  gtk_draw_diamond(XEN_TO_C_GtkStyle_(style), XEN_TO_C_GdkWindow_(window), XEN_TO_C_GtkStateType(state_type), XEN_TO_C_GtkShadowType(shadow_type), 
                   XEN_TO_C_gint(x), XEN_TO_C_gint(y), XEN_TO_C_gint(width), XEN_TO_C_gint(height));
  return(XEN_FALSE);
}

static XEN gxg_gtk_draw_box(XEN style, XEN window, XEN state_type, XEN shadow_type, XEN x, XEN y, XEN width, XEN height)
{
  #define H_gtk_draw_box "void gtk_draw_box(GtkStyle* style, GdkWindow* window, GtkStateType state_type, \
GtkShadowType shadow_type, gint x, gint y, gint width, gint height)"
  XEN_ASSERT_TYPE(XEN_GtkStyle__P(style), style, 1, "gtk_draw_box", "GtkStyle*");
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 2, "gtk_draw_box", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_GtkStateType_P(state_type), state_type, 3, "gtk_draw_box", "GtkStateType");
  XEN_ASSERT_TYPE(XEN_GtkShadowType_P(shadow_type), shadow_type, 4, "gtk_draw_box", "GtkShadowType");
  XEN_ASSERT_TYPE(XEN_gint_P(x), x, 5, "gtk_draw_box", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(y), y, 6, "gtk_draw_box", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(width), width, 7, "gtk_draw_box", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(height), height, 8, "gtk_draw_box", "gint");
  gtk_draw_box(XEN_TO_C_GtkStyle_(style), XEN_TO_C_GdkWindow_(window), XEN_TO_C_GtkStateType(state_type), XEN_TO_C_GtkShadowType(shadow_type), 
               XEN_TO_C_gint(x), XEN_TO_C_gint(y), XEN_TO_C_gint(width), XEN_TO_C_gint(height));
  return(XEN_FALSE);
}

static XEN gxg_gtk_draw_flat_box(XEN style, XEN window, XEN state_type, XEN shadow_type, XEN x, XEN y, XEN width, XEN height)
{
  #define H_gtk_draw_flat_box "void gtk_draw_flat_box(GtkStyle* style, GdkWindow* window, GtkStateType state_type, \
GtkShadowType shadow_type, gint x, gint y, gint width, gint height)"
  XEN_ASSERT_TYPE(XEN_GtkStyle__P(style), style, 1, "gtk_draw_flat_box", "GtkStyle*");
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 2, "gtk_draw_flat_box", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_GtkStateType_P(state_type), state_type, 3, "gtk_draw_flat_box", "GtkStateType");
  XEN_ASSERT_TYPE(XEN_GtkShadowType_P(shadow_type), shadow_type, 4, "gtk_draw_flat_box", "GtkShadowType");
  XEN_ASSERT_TYPE(XEN_gint_P(x), x, 5, "gtk_draw_flat_box", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(y), y, 6, "gtk_draw_flat_box", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(width), width, 7, "gtk_draw_flat_box", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(height), height, 8, "gtk_draw_flat_box", "gint");
  gtk_draw_flat_box(XEN_TO_C_GtkStyle_(style), XEN_TO_C_GdkWindow_(window), XEN_TO_C_GtkStateType(state_type), XEN_TO_C_GtkShadowType(shadow_type), 
                    XEN_TO_C_gint(x), XEN_TO_C_gint(y), XEN_TO_C_gint(width), XEN_TO_C_gint(height));
  return(XEN_FALSE);
}

static XEN gxg_gtk_draw_check(XEN style, XEN window, XEN state_type, XEN shadow_type, XEN x, XEN y, XEN width, XEN height)
{
  #define H_gtk_draw_check "void gtk_draw_check(GtkStyle* style, GdkWindow* window, GtkStateType state_type, \
GtkShadowType shadow_type, gint x, gint y, gint width, gint height)"
  XEN_ASSERT_TYPE(XEN_GtkStyle__P(style), style, 1, "gtk_draw_check", "GtkStyle*");
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 2, "gtk_draw_check", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_GtkStateType_P(state_type), state_type, 3, "gtk_draw_check", "GtkStateType");
  XEN_ASSERT_TYPE(XEN_GtkShadowType_P(shadow_type), shadow_type, 4, "gtk_draw_check", "GtkShadowType");
  XEN_ASSERT_TYPE(XEN_gint_P(x), x, 5, "gtk_draw_check", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(y), y, 6, "gtk_draw_check", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(width), width, 7, "gtk_draw_check", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(height), height, 8, "gtk_draw_check", "gint");
  gtk_draw_check(XEN_TO_C_GtkStyle_(style), XEN_TO_C_GdkWindow_(window), XEN_TO_C_GtkStateType(state_type), XEN_TO_C_GtkShadowType(shadow_type), 
                 XEN_TO_C_gint(x), XEN_TO_C_gint(y), XEN_TO_C_gint(width), XEN_TO_C_gint(height));
  return(XEN_FALSE);
}

static XEN gxg_gtk_draw_option(XEN style, XEN window, XEN state_type, XEN shadow_type, XEN x, XEN y, XEN width, XEN height)
{
  #define H_gtk_draw_option "void gtk_draw_option(GtkStyle* style, GdkWindow* window, GtkStateType state_type, \
GtkShadowType shadow_type, gint x, gint y, gint width, gint height)"
  XEN_ASSERT_TYPE(XEN_GtkStyle__P(style), style, 1, "gtk_draw_option", "GtkStyle*");
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 2, "gtk_draw_option", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_GtkStateType_P(state_type), state_type, 3, "gtk_draw_option", "GtkStateType");
  XEN_ASSERT_TYPE(XEN_GtkShadowType_P(shadow_type), shadow_type, 4, "gtk_draw_option", "GtkShadowType");
  XEN_ASSERT_TYPE(XEN_gint_P(x), x, 5, "gtk_draw_option", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(y), y, 6, "gtk_draw_option", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(width), width, 7, "gtk_draw_option", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(height), height, 8, "gtk_draw_option", "gint");
  gtk_draw_option(XEN_TO_C_GtkStyle_(style), XEN_TO_C_GdkWindow_(window), XEN_TO_C_GtkStateType(state_type), XEN_TO_C_GtkShadowType(shadow_type), 
                  XEN_TO_C_gint(x), XEN_TO_C_gint(y), XEN_TO_C_gint(width), XEN_TO_C_gint(height));
  return(XEN_FALSE);
}

static XEN gxg_gtk_draw_tab(XEN style, XEN window, XEN state_type, XEN shadow_type, XEN x, XEN y, XEN width, XEN height)
{
  #define H_gtk_draw_tab "void gtk_draw_tab(GtkStyle* style, GdkWindow* window, GtkStateType state_type, \
GtkShadowType shadow_type, gint x, gint y, gint width, gint height)"
  XEN_ASSERT_TYPE(XEN_GtkStyle__P(style), style, 1, "gtk_draw_tab", "GtkStyle*");
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 2, "gtk_draw_tab", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_GtkStateType_P(state_type), state_type, 3, "gtk_draw_tab", "GtkStateType");
  XEN_ASSERT_TYPE(XEN_GtkShadowType_P(shadow_type), shadow_type, 4, "gtk_draw_tab", "GtkShadowType");
  XEN_ASSERT_TYPE(XEN_gint_P(x), x, 5, "gtk_draw_tab", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(y), y, 6, "gtk_draw_tab", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(width), width, 7, "gtk_draw_tab", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(height), height, 8, "gtk_draw_tab", "gint");
  gtk_draw_tab(XEN_TO_C_GtkStyle_(style), XEN_TO_C_GdkWindow_(window), XEN_TO_C_GtkStateType(state_type), XEN_TO_C_GtkShadowType(shadow_type), 
               XEN_TO_C_gint(x), XEN_TO_C_gint(y), XEN_TO_C_gint(width), XEN_TO_C_gint(height));
  return(XEN_FALSE);
}

static XEN gxg_gtk_draw_shadow_gap(XEN arglist)
{
  #define H_gtk_draw_shadow_gap "void gtk_draw_shadow_gap(GtkStyle* style, GdkWindow* window, GtkStateType state_type, \
GtkShadowType shadow_type, gint x, gint y, gint width, gint height, GtkPositionType gap_side, gint gap_x, \
gint gap_width)"
  XEN style, window, state_type, shadow_type, x, y, width, height, gap_side, gap_x, gap_width;
  style = XEN_LIST_REF(arglist, 0);
  window = XEN_LIST_REF(arglist, 1);
  state_type = XEN_LIST_REF(arglist, 2);
  shadow_type = XEN_LIST_REF(arglist, 3);
  x = XEN_LIST_REF(arglist, 4);
  y = XEN_LIST_REF(arglist, 5);
  width = XEN_LIST_REF(arglist, 6);
  height = XEN_LIST_REF(arglist, 7);
  gap_side = XEN_LIST_REF(arglist, 8);
  gap_x = XEN_LIST_REF(arglist, 9);
  gap_width = XEN_LIST_REF(arglist, 10);
  XEN_ASSERT_TYPE(XEN_GtkStyle__P(style), style, 1, "gtk_draw_shadow_gap", "GtkStyle*");
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 2, "gtk_draw_shadow_gap", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_GtkStateType_P(state_type), state_type, 3, "gtk_draw_shadow_gap", "GtkStateType");
  XEN_ASSERT_TYPE(XEN_GtkShadowType_P(shadow_type), shadow_type, 4, "gtk_draw_shadow_gap", "GtkShadowType");
  XEN_ASSERT_TYPE(XEN_gint_P(x), x, 5, "gtk_draw_shadow_gap", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(y), y, 6, "gtk_draw_shadow_gap", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(width), width, 7, "gtk_draw_shadow_gap", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(height), height, 8, "gtk_draw_shadow_gap", "gint");
  XEN_ASSERT_TYPE(XEN_GtkPositionType_P(gap_side), gap_side, 9, "gtk_draw_shadow_gap", "GtkPositionType");
  XEN_ASSERT_TYPE(XEN_gint_P(gap_x), gap_x, 10, "gtk_draw_shadow_gap", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(gap_width), gap_width, 11, "gtk_draw_shadow_gap", "gint");
  gtk_draw_shadow_gap(XEN_TO_C_GtkStyle_(style), XEN_TO_C_GdkWindow_(window), XEN_TO_C_GtkStateType(state_type), XEN_TO_C_GtkShadowType(shadow_type), 
                      XEN_TO_C_gint(x), XEN_TO_C_gint(y), XEN_TO_C_gint(width), XEN_TO_C_gint(height), XEN_TO_C_GtkPositionType(gap_side), 
                      XEN_TO_C_gint(gap_x), XEN_TO_C_gint(gap_width));
  return(XEN_FALSE);
}

static XEN gxg_gtk_draw_box_gap(XEN arglist)
{
  #define H_gtk_draw_box_gap "void gtk_draw_box_gap(GtkStyle* style, GdkWindow* window, GtkStateType state_type, \
GtkShadowType shadow_type, gint x, gint y, gint width, gint height, GtkPositionType gap_side, gint gap_x, \
gint gap_width)"
  XEN style, window, state_type, shadow_type, x, y, width, height, gap_side, gap_x, gap_width;
  style = XEN_LIST_REF(arglist, 0);
  window = XEN_LIST_REF(arglist, 1);
  state_type = XEN_LIST_REF(arglist, 2);
  shadow_type = XEN_LIST_REF(arglist, 3);
  x = XEN_LIST_REF(arglist, 4);
  y = XEN_LIST_REF(arglist, 5);
  width = XEN_LIST_REF(arglist, 6);
  height = XEN_LIST_REF(arglist, 7);
  gap_side = XEN_LIST_REF(arglist, 8);
  gap_x = XEN_LIST_REF(arglist, 9);
  gap_width = XEN_LIST_REF(arglist, 10);
  XEN_ASSERT_TYPE(XEN_GtkStyle__P(style), style, 1, "gtk_draw_box_gap", "GtkStyle*");
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 2, "gtk_draw_box_gap", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_GtkStateType_P(state_type), state_type, 3, "gtk_draw_box_gap", "GtkStateType");
  XEN_ASSERT_TYPE(XEN_GtkShadowType_P(shadow_type), shadow_type, 4, "gtk_draw_box_gap", "GtkShadowType");
  XEN_ASSERT_TYPE(XEN_gint_P(x), x, 5, "gtk_draw_box_gap", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(y), y, 6, "gtk_draw_box_gap", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(width), width, 7, "gtk_draw_box_gap", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(height), height, 8, "gtk_draw_box_gap", "gint");
  XEN_ASSERT_TYPE(XEN_GtkPositionType_P(gap_side), gap_side, 9, "gtk_draw_box_gap", "GtkPositionType");
  XEN_ASSERT_TYPE(XEN_gint_P(gap_x), gap_x, 10, "gtk_draw_box_gap", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(gap_width), gap_width, 11, "gtk_draw_box_gap", "gint");
  gtk_draw_box_gap(XEN_TO_C_GtkStyle_(style), XEN_TO_C_GdkWindow_(window), XEN_TO_C_GtkStateType(state_type), XEN_TO_C_GtkShadowType(shadow_type), 
                   XEN_TO_C_gint(x), XEN_TO_C_gint(y), XEN_TO_C_gint(width), XEN_TO_C_gint(height), XEN_TO_C_GtkPositionType(gap_side), 
                   XEN_TO_C_gint(gap_x), XEN_TO_C_gint(gap_width));
  return(XEN_FALSE);
}

static XEN gxg_gtk_draw_extension(XEN style, XEN window, XEN state_type, XEN shadow_type, XEN x, XEN y, XEN width, XEN height, XEN gap_side)
{
  #define H_gtk_draw_extension "void gtk_draw_extension(GtkStyle* style, GdkWindow* window, GtkStateType state_type, \
GtkShadowType shadow_type, gint x, gint y, gint width, gint height, GtkPositionType gap_side)"
  XEN_ASSERT_TYPE(XEN_GtkStyle__P(style), style, 1, "gtk_draw_extension", "GtkStyle*");
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 2, "gtk_draw_extension", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_GtkStateType_P(state_type), state_type, 3, "gtk_draw_extension", "GtkStateType");
  XEN_ASSERT_TYPE(XEN_GtkShadowType_P(shadow_type), shadow_type, 4, "gtk_draw_extension", "GtkShadowType");
  XEN_ASSERT_TYPE(XEN_gint_P(x), x, 5, "gtk_draw_extension", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(y), y, 6, "gtk_draw_extension", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(width), width, 7, "gtk_draw_extension", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(height), height, 8, "gtk_draw_extension", "gint");
  XEN_ASSERT_TYPE(XEN_GtkPositionType_P(gap_side), gap_side, 9, "gtk_draw_extension", "GtkPositionType");
  gtk_draw_extension(XEN_TO_C_GtkStyle_(style), XEN_TO_C_GdkWindow_(window), XEN_TO_C_GtkStateType(state_type), XEN_TO_C_GtkShadowType(shadow_type), 
                     XEN_TO_C_gint(x), XEN_TO_C_gint(y), XEN_TO_C_gint(width), XEN_TO_C_gint(height), XEN_TO_C_GtkPositionType(gap_side));
  return(XEN_FALSE);
}

static XEN gxg_gtk_draw_focus(XEN style, XEN window, XEN x, XEN y, XEN width, XEN height)
{
  #define H_gtk_draw_focus "void gtk_draw_focus(GtkStyle* style, GdkWindow* window, gint x, gint y, gint width, \
gint height)"
  XEN_ASSERT_TYPE(XEN_GtkStyle__P(style), style, 1, "gtk_draw_focus", "GtkStyle*");
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 2, "gtk_draw_focus", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_gint_P(x), x, 3, "gtk_draw_focus", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(y), y, 4, "gtk_draw_focus", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(width), width, 5, "gtk_draw_focus", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(height), height, 6, "gtk_draw_focus", "gint");
  gtk_draw_focus(XEN_TO_C_GtkStyle_(style), XEN_TO_C_GdkWindow_(window), XEN_TO_C_gint(x), XEN_TO_C_gint(y), XEN_TO_C_gint(width), 
                 XEN_TO_C_gint(height));
  return(XEN_FALSE);
}

static XEN gxg_gtk_draw_slider(XEN style, XEN window, XEN state_type, XEN shadow_type, XEN x, XEN y, XEN width, XEN height, XEN orientation)
{
  #define H_gtk_draw_slider "void gtk_draw_slider(GtkStyle* style, GdkWindow* window, GtkStateType state_type, \
GtkShadowType shadow_type, gint x, gint y, gint width, gint height, GtkOrientation orientation)"
  XEN_ASSERT_TYPE(XEN_GtkStyle__P(style), style, 1, "gtk_draw_slider", "GtkStyle*");
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 2, "gtk_draw_slider", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_GtkStateType_P(state_type), state_type, 3, "gtk_draw_slider", "GtkStateType");
  XEN_ASSERT_TYPE(XEN_GtkShadowType_P(shadow_type), shadow_type, 4, "gtk_draw_slider", "GtkShadowType");
  XEN_ASSERT_TYPE(XEN_gint_P(x), x, 5, "gtk_draw_slider", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(y), y, 6, "gtk_draw_slider", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(width), width, 7, "gtk_draw_slider", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(height), height, 8, "gtk_draw_slider", "gint");
  XEN_ASSERT_TYPE(XEN_GtkOrientation_P(orientation), orientation, 9, "gtk_draw_slider", "GtkOrientation");
  gtk_draw_slider(XEN_TO_C_GtkStyle_(style), XEN_TO_C_GdkWindow_(window), XEN_TO_C_GtkStateType(state_type), XEN_TO_C_GtkShadowType(shadow_type), 
                  XEN_TO_C_gint(x), XEN_TO_C_gint(y), XEN_TO_C_gint(width), XEN_TO_C_gint(height), XEN_TO_C_GtkOrientation(orientation));
  return(XEN_FALSE);
}

static XEN gxg_gtk_draw_handle(XEN style, XEN window, XEN state_type, XEN shadow_type, XEN x, XEN y, XEN width, XEN height, XEN orientation)
{
  #define H_gtk_draw_handle "void gtk_draw_handle(GtkStyle* style, GdkWindow* window, GtkStateType state_type, \
GtkShadowType shadow_type, gint x, gint y, gint width, gint height, GtkOrientation orientation)"
  XEN_ASSERT_TYPE(XEN_GtkStyle__P(style), style, 1, "gtk_draw_handle", "GtkStyle*");
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 2, "gtk_draw_handle", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_GtkStateType_P(state_type), state_type, 3, "gtk_draw_handle", "GtkStateType");
  XEN_ASSERT_TYPE(XEN_GtkShadowType_P(shadow_type), shadow_type, 4, "gtk_draw_handle", "GtkShadowType");
  XEN_ASSERT_TYPE(XEN_gint_P(x), x, 5, "gtk_draw_handle", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(y), y, 6, "gtk_draw_handle", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(width), width, 7, "gtk_draw_handle", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(height), height, 8, "gtk_draw_handle", "gint");
  XEN_ASSERT_TYPE(XEN_GtkOrientation_P(orientation), orientation, 9, "gtk_draw_handle", "GtkOrientation");
  gtk_draw_handle(XEN_TO_C_GtkStyle_(style), XEN_TO_C_GdkWindow_(window), XEN_TO_C_GtkStateType(state_type), XEN_TO_C_GtkShadowType(shadow_type), 
                  XEN_TO_C_gint(x), XEN_TO_C_gint(y), XEN_TO_C_gint(width), XEN_TO_C_gint(height), XEN_TO_C_GtkOrientation(orientation));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_gtk_draw_expander(XEN style, XEN window, XEN state_type, XEN x, XEN y, XEN expander_style)
{
  #define H_gtk_draw_expander "void gtk_draw_expander(GtkStyle* style, GdkWindow* window, GtkStateType state_type, \
gint x, gint y, GtkExpanderStyle expander_style)"
  XEN_ASSERT_TYPE(XEN_GtkStyle__P(style), style, 1, "gtk_draw_expander", "GtkStyle*");
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 2, "gtk_draw_expander", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_GtkStateType_P(state_type), state_type, 3, "gtk_draw_expander", "GtkStateType");
  XEN_ASSERT_TYPE(XEN_gint_P(x), x, 4, "gtk_draw_expander", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(y), y, 5, "gtk_draw_expander", "gint");
  XEN_ASSERT_TYPE(XEN_GtkExpanderStyle_P(expander_style), expander_style, 6, "gtk_draw_expander", "GtkExpanderStyle");
  gtk_draw_expander(XEN_TO_C_GtkStyle_(style), XEN_TO_C_GdkWindow_(window), XEN_TO_C_GtkStateType(state_type), XEN_TO_C_gint(x), 
                    XEN_TO_C_gint(y), XEN_TO_C_GtkExpanderStyle(expander_style));
  return(XEN_FALSE);
}

static XEN gxg_gtk_draw_layout(XEN style, XEN window, XEN state_type, XEN use_text, XEN x, XEN y, XEN layout)
{
  #define H_gtk_draw_layout "void gtk_draw_layout(GtkStyle* style, GdkWindow* window, GtkStateType state_type, \
gboolean use_text, gint x, gint y, PangoLayout* layout)"
  XEN_ASSERT_TYPE(XEN_GtkStyle__P(style), style, 1, "gtk_draw_layout", "GtkStyle*");
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 2, "gtk_draw_layout", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_GtkStateType_P(state_type), state_type, 3, "gtk_draw_layout", "GtkStateType");
  XEN_ASSERT_TYPE(XEN_gboolean_P(use_text), use_text, 4, "gtk_draw_layout", "gboolean");
  XEN_ASSERT_TYPE(XEN_gint_P(x), x, 5, "gtk_draw_layout", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(y), y, 6, "gtk_draw_layout", "gint");
  XEN_ASSERT_TYPE(XEN_PangoLayout__P(layout), layout, 7, "gtk_draw_layout", "PangoLayout*");
  gtk_draw_layout(XEN_TO_C_GtkStyle_(style), XEN_TO_C_GdkWindow_(window), XEN_TO_C_GtkStateType(state_type), XEN_TO_C_gboolean(use_text), 
                  XEN_TO_C_gint(x), XEN_TO_C_gint(y), XEN_TO_C_PangoLayout_(layout));
  return(XEN_FALSE);
}

static XEN gxg_gtk_draw_resize_grip(XEN style, XEN window, XEN state_type, XEN edge, XEN x, XEN y, XEN width, XEN height)
{
  #define H_gtk_draw_resize_grip "void gtk_draw_resize_grip(GtkStyle* style, GdkWindow* window, GtkStateType state_type, \
GdkWindowEdge edge, gint x, gint y, gint width, gint height)"
  XEN_ASSERT_TYPE(XEN_GtkStyle__P(style), style, 1, "gtk_draw_resize_grip", "GtkStyle*");
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 2, "gtk_draw_resize_grip", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_GtkStateType_P(state_type), state_type, 3, "gtk_draw_resize_grip", "GtkStateType");
  XEN_ASSERT_TYPE(XEN_GdkWindowEdge_P(edge), edge, 4, "gtk_draw_resize_grip", "GdkWindowEdge");
  XEN_ASSERT_TYPE(XEN_gint_P(x), x, 5, "gtk_draw_resize_grip", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(y), y, 6, "gtk_draw_resize_grip", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(width), width, 7, "gtk_draw_resize_grip", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(height), height, 8, "gtk_draw_resize_grip", "gint");
  gtk_draw_resize_grip(XEN_TO_C_GtkStyle_(style), XEN_TO_C_GdkWindow_(window), XEN_TO_C_GtkStateType(state_type), XEN_TO_C_GdkWindowEdge(edge), 
                       XEN_TO_C_gint(x), XEN_TO_C_gint(y), XEN_TO_C_gint(width), XEN_TO_C_gint(height));
  return(XEN_FALSE);
}
#endif

static XEN gxg_gtk_draw_string(XEN style, XEN window, XEN state_type, XEN x, XEN y, XEN string)
{
  #define H_gtk_draw_string "void gtk_draw_string(GtkStyle* style, GdkWindow* window, GtkStateType state_type, \
gint x, gint y, gchar* string)"
  XEN_ASSERT_TYPE(XEN_GtkStyle__P(style), style, 1, "gtk_draw_string", "GtkStyle*");
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 2, "gtk_draw_string", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_GtkStateType_P(state_type), state_type, 3, "gtk_draw_string", "GtkStateType");
  XEN_ASSERT_TYPE(XEN_gint_P(x), x, 4, "gtk_draw_string", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(y), y, 5, "gtk_draw_string", "gint");
  XEN_ASSERT_TYPE(XEN_gchar__P(string), string, 6, "gtk_draw_string", "gchar*");
  gtk_draw_string(XEN_TO_C_GtkStyle_(style), XEN_TO_C_GdkWindow_(window), XEN_TO_C_GtkStateType(state_type), XEN_TO_C_gint(x), 
                  XEN_TO_C_gint(y), XEN_TO_C_gchar_(string));
  return(XEN_FALSE);
}

static XEN gxg_gtk_paint_string(XEN style, XEN window, XEN state_type, XEN area, XEN widget, XEN detail, XEN x, XEN y, XEN string)
{
  #define H_gtk_paint_string "void gtk_paint_string(GtkStyle* style, GdkWindow* window, GtkStateType state_type, \
GdkRectangle* area, GtkWidget* widget, gchar* detail, gint x, gint y, gchar* string)"
  XEN_ASSERT_TYPE(XEN_GtkStyle__P(style), style, 1, "gtk_paint_string", "GtkStyle*");
  XEN_ASSERT_TYPE(XEN_GdkWindow__P(window), window, 2, "gtk_paint_string", "GdkWindow*");
  XEN_ASSERT_TYPE(XEN_GtkStateType_P(state_type), state_type, 3, "gtk_paint_string", "GtkStateType");
  XEN_ASSERT_TYPE(XEN_GdkRectangle__P(area), area, 4, "gtk_paint_string", "GdkRectangle*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 5, "gtk_paint_string", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_gchar__P(detail), detail, 6, "gtk_paint_string", "gchar*");
  XEN_ASSERT_TYPE(XEN_gint_P(x), x, 7, "gtk_paint_string", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(y), y, 8, "gtk_paint_string", "gint");
  XEN_ASSERT_TYPE(XEN_gchar__P(string), string, 9, "gtk_paint_string", "gchar*");
  gtk_paint_string(XEN_TO_C_GtkStyle_(style), XEN_TO_C_GdkWindow_(window), XEN_TO_C_GtkStateType(state_type), XEN_TO_C_GdkRectangle_(area), 
                   XEN_TO_C_GtkWidget_(widget), XEN_TO_C_gchar_(detail), XEN_TO_C_gint(x), XEN_TO_C_gint(y), XEN_TO_C_gchar_(string));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tips_query_get_type(void)
{
  #define H_gtk_tips_query_get_type "GtkType gtk_tips_query_get_type( void)"
  return(C_TO_XEN_GtkType(gtk_tips_query_get_type()));
}

static XEN gxg_gtk_tips_query_new(void)
{
  #define H_gtk_tips_query_new "GtkWidget* gtk_tips_query_new( void)"
  return(C_TO_XEN_GtkWidget_(gtk_tips_query_new()));
}

static XEN gxg_gtk_tips_query_start_query(XEN tips_query)
{
  #define H_gtk_tips_query_start_query "void gtk_tips_query_start_query(GtkTipsQuery* tips_query)"
  XEN_ASSERT_TYPE(XEN_GtkTipsQuery__P(tips_query), tips_query, 1, "gtk_tips_query_start_query", "GtkTipsQuery*");
  gtk_tips_query_start_query(XEN_TO_C_GtkTipsQuery_(tips_query));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tips_query_stop_query(XEN tips_query)
{
  #define H_gtk_tips_query_stop_query "void gtk_tips_query_stop_query(GtkTipsQuery* tips_query)"
  XEN_ASSERT_TYPE(XEN_GtkTipsQuery__P(tips_query), tips_query, 1, "gtk_tips_query_stop_query", "GtkTipsQuery*");
  gtk_tips_query_stop_query(XEN_TO_C_GtkTipsQuery_(tips_query));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tips_query_set_caller(XEN tips_query, XEN caller)
{
  #define H_gtk_tips_query_set_caller "void gtk_tips_query_set_caller(GtkTipsQuery* tips_query, GtkWidget* caller)"
  XEN_ASSERT_TYPE(XEN_GtkTipsQuery__P(tips_query), tips_query, 1, "gtk_tips_query_set_caller", "GtkTipsQuery*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(caller), caller, 2, "gtk_tips_query_set_caller", "GtkWidget*");
  gtk_tips_query_set_caller(XEN_TO_C_GtkTipsQuery_(tips_query), XEN_TO_C_GtkWidget_(caller));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tips_query_set_labels(XEN tips_query, XEN label_inactive, XEN label_no_tip)
{
  #define H_gtk_tips_query_set_labels "void gtk_tips_query_set_labels(GtkTipsQuery* tips_query, gchar* label_inactive, \
gchar* label_no_tip)"
  XEN_ASSERT_TYPE(XEN_GtkTipsQuery__P(tips_query), tips_query, 1, "gtk_tips_query_set_labels", "GtkTipsQuery*");
  XEN_ASSERT_TYPE(XEN_gchar__P(label_inactive), label_inactive, 2, "gtk_tips_query_set_labels", "gchar*");
  XEN_ASSERT_TYPE(XEN_gchar__P(label_no_tip), label_no_tip, 3, "gtk_tips_query_set_labels", "gchar*");
  gtk_tips_query_set_labels(XEN_TO_C_GtkTipsQuery_(tips_query), XEN_TO_C_gchar_(label_inactive), XEN_TO_C_gchar_(label_no_tip));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tooltips_set_delay(XEN tooltips, XEN delay)
{
  #define H_gtk_tooltips_set_delay "void gtk_tooltips_set_delay(GtkTooltips* tooltips, guint delay)"
  XEN_ASSERT_TYPE(XEN_GtkTooltips__P(tooltips), tooltips, 1, "gtk_tooltips_set_delay", "GtkTooltips*");
  XEN_ASSERT_TYPE(XEN_guint_P(delay), delay, 2, "gtk_tooltips_set_delay", "guint");
  gtk_tooltips_set_delay(XEN_TO_C_GtkTooltips_(tooltips), XEN_TO_C_guint(delay));
  return(XEN_FALSE);
}

static XEN gxg_gtk_vbutton_box_get_spacing_default(void)
{
  #define H_gtk_vbutton_box_get_spacing_default "gint gtk_vbutton_box_get_spacing_default( void)"
  return(C_TO_XEN_gint(gtk_vbutton_box_get_spacing_default()));
}

static XEN gxg_gtk_vbutton_box_set_spacing_default(XEN spacing)
{
  #define H_gtk_vbutton_box_set_spacing_default "void gtk_vbutton_box_set_spacing_default(gint spacing)"
  XEN_ASSERT_TYPE(XEN_gint_P(spacing), spacing, 1, "gtk_vbutton_box_set_spacing_default", "gint");
  gtk_vbutton_box_set_spacing_default(XEN_TO_C_gint(spacing));
  return(XEN_FALSE);
}

static XEN gxg_gtk_vbutton_box_get_layout_default(void)
{
  #define H_gtk_vbutton_box_get_layout_default "GtkButtonBoxStyle gtk_vbutton_box_get_layout_default( void)"
  return(C_TO_XEN_GtkButtonBoxStyle(gtk_vbutton_box_get_layout_default()));
}

static XEN gxg_gtk_vbutton_box_set_layout_default(XEN layout)
{
  #define H_gtk_vbutton_box_set_layout_default "void gtk_vbutton_box_set_layout_default(GtkButtonBoxStyle layout)"
  XEN_ASSERT_TYPE(XEN_GtkButtonBoxStyle_P(layout), layout, 1, "gtk_vbutton_box_set_layout_default", "GtkButtonBoxStyle");
  gtk_vbutton_box_set_layout_default(XEN_TO_C_GtkButtonBoxStyle(layout));
  return(XEN_FALSE);
}

static XEN gxg_gtk_widget_set(XEN widget, XEN first_property_name)
{
  #define H_gtk_widget_set "void gtk_widget_set(GtkWidget* widget, gchar* first_property_name, ...)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_widget_set", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_gchar__P(first_property_name), first_property_name, 2, "gtk_widget_set", "gchar*");
  gtk_widget_set(XEN_TO_C_GtkWidget_(widget), XEN_TO_C_gchar_(first_property_name));
  return(XEN_FALSE);
}

static XEN gxg_gtk_widget_queue_clear(XEN widget)
{
  #define H_gtk_widget_queue_clear "void gtk_widget_queue_clear(GtkWidget* widget)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_widget_queue_clear", "GtkWidget*");
  gtk_widget_queue_clear(XEN_TO_C_GtkWidget_(widget));
  return(XEN_FALSE);
}

static XEN gxg_gtk_widget_queue_clear_area(XEN widget, XEN x, XEN y, XEN width, XEN height)
{
  #define H_gtk_widget_queue_clear_area "void gtk_widget_queue_clear_area(GtkWidget* widget, gint x, \
gint y, gint width, gint height)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_widget_queue_clear_area", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_gint_P(x), x, 2, "gtk_widget_queue_clear_area", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(y), y, 3, "gtk_widget_queue_clear_area", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(width), width, 4, "gtk_widget_queue_clear_area", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(height), height, 5, "gtk_widget_queue_clear_area", "gint");
  gtk_widget_queue_clear_area(XEN_TO_C_GtkWidget_(widget), XEN_TO_C_gint(x), XEN_TO_C_gint(y), XEN_TO_C_gint(width), XEN_TO_C_gint(height));
  return(XEN_FALSE);
}

static XEN gxg_gtk_widget_draw(XEN widget, XEN area)
{
  #define H_gtk_widget_draw "void gtk_widget_draw(GtkWidget* widget, GdkRectangle* area)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_widget_draw", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_GdkRectangle__P(area), area, 2, "gtk_widget_draw", "GdkRectangle*");
  gtk_widget_draw(XEN_TO_C_GtkWidget_(widget), XEN_TO_C_GdkRectangle_(area));
  return(XEN_FALSE);
}

static XEN gxg_gtk_widget_set_uposition(XEN widget, XEN x, XEN y)
{
  #define H_gtk_widget_set_uposition "void gtk_widget_set_uposition(GtkWidget* widget, gint x, gint y)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_widget_set_uposition", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_gint_P(x), x, 2, "gtk_widget_set_uposition", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(y), y, 3, "gtk_widget_set_uposition", "gint");
  gtk_widget_set_uposition(XEN_TO_C_GtkWidget_(widget), XEN_TO_C_gint(x), XEN_TO_C_gint(y));
  return(XEN_FALSE);
}

static XEN gxg_gtk_widget_set_usize(XEN widget, XEN width, XEN height)
{
  #define H_gtk_widget_set_usize "void gtk_widget_set_usize(GtkWidget* widget, gint width, gint height)"
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(widget), widget, 1, "gtk_widget_set_usize", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_gint_P(width), width, 2, "gtk_widget_set_usize", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(height), height, 3, "gtk_widget_set_usize", "gint");
  gtk_widget_set_usize(XEN_TO_C_GtkWidget_(widget), XEN_TO_C_gint(width), XEN_TO_C_gint(height));
  return(XEN_FALSE);
}

static XEN gxg_gtk_window_set_policy(XEN window, XEN allow_shrink, XEN allow_grow, XEN auto_shrink)
{
  #define H_gtk_window_set_policy "void gtk_window_set_policy(GtkWindow* window, gint allow_shrink, gint allow_grow, \
gint auto_shrink)"
  XEN_ASSERT_TYPE(XEN_GtkWindow__P(window), window, 1, "gtk_window_set_policy", "GtkWindow*");
  XEN_ASSERT_TYPE(XEN_gint_P(allow_shrink), allow_shrink, 2, "gtk_window_set_policy", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(allow_grow), allow_grow, 3, "gtk_window_set_policy", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(auto_shrink), auto_shrink, 4, "gtk_window_set_policy", "gint");
  gtk_window_set_policy(XEN_TO_C_GtkWindow_(window), XEN_TO_C_gint(allow_shrink), XEN_TO_C_gint(allow_grow), XEN_TO_C_gint(auto_shrink));
  return(XEN_FALSE);
}

static XEN gxg_gdk_colormap_ref(XEN cmap)
{
  #define H_gdk_colormap_ref "GdkColormap* gdk_colormap_ref(GdkColormap* cmap)"
  XEN_ASSERT_TYPE(XEN_GdkColormap__P(cmap), cmap, 1, "gdk_colormap_ref", "GdkColormap*");
  return(C_TO_XEN_GdkColormap_(gdk_colormap_ref(XEN_TO_C_GdkColormap_(cmap))));
}

static XEN gxg_gdk_colormap_unref(XEN cmap)
{
  #define H_gdk_colormap_unref "void gdk_colormap_unref(GdkColormap* cmap)"
  XEN_ASSERT_TYPE(XEN_GdkColormap__P(cmap), cmap, 1, "gdk_colormap_unref", "GdkColormap*");
  gdk_colormap_unref(XEN_TO_C_GdkColormap_(cmap));
  return(XEN_FALSE);
}

static XEN gxg_gdk_colormap_get_system_size(void)
{
  #define H_gdk_colormap_get_system_size "gint gdk_colormap_get_system_size( void)"
  return(C_TO_XEN_gint(gdk_colormap_get_system_size()));
}

static XEN gxg_gdk_colormap_change(XEN colormap, XEN ncolors)
{
  #define H_gdk_colormap_change "void gdk_colormap_change(GdkColormap* colormap, gint ncolors)"
  XEN_ASSERT_TYPE(XEN_GdkColormap__P(colormap), colormap, 1, "gdk_colormap_change", "GdkColormap*");
  XEN_ASSERT_TYPE(XEN_gint_P(ncolors), ncolors, 2, "gdk_colormap_change", "gint");
  gdk_colormap_change(XEN_TO_C_GdkColormap_(colormap), XEN_TO_C_gint(ncolors));
  return(XEN_FALSE);
}

static XEN gxg_gdk_colors_store(XEN colormap, XEN colors, XEN ncolors)
{
  #define H_gdk_colors_store "void gdk_colors_store(GdkColormap* colormap, GdkColor* colors, gint ncolors)"
  XEN_ASSERT_TYPE(XEN_GdkColormap__P(colormap), colormap, 1, "gdk_colors_store", "GdkColormap*");
  XEN_ASSERT_TYPE(XEN_GdkColor__P(colors), colors, 2, "gdk_colors_store", "GdkColor*");
  XEN_ASSERT_TYPE(XEN_gint_P(ncolors), ncolors, 3, "gdk_colors_store", "gint");
  gdk_colors_store(XEN_TO_C_GdkColormap_(colormap), XEN_TO_C_GdkColor_(colors), XEN_TO_C_gint(ncolors));
  return(XEN_FALSE);
}

#if (!HAVE_GTK_1)
static XEN gxg_GDK_ROOT_PARENT(void)
{
  #define H_GDK_ROOT_PARENT "GdkWindow* GDK_ROOT_PARENT( void)"
  return(C_TO_XEN_GdkWindow_(GDK_ROOT_PARENT()));
}
#endif
#endif

#if GDK_ENABLE_BROKEN && GTK_ENABLE_BROKEN

static XEN gxg_gtk_text_set_adjustments(XEN text, XEN hadj, XEN vadj)
{
  #define H_gtk_text_set_adjustments "void gtk_text_set_adjustments(GtkText* text, GtkAdjustment* hadj, \
GtkAdjustment* vadj)"
  XEN_ASSERT_TYPE(XEN_GtkText__P(text), text, 1, "gtk_text_set_adjustments", "GtkText*");
  XEN_ASSERT_TYPE(XEN_GtkAdjustment__P(hadj), hadj, 2, "gtk_text_set_adjustments", "GtkAdjustment*");
  XEN_ASSERT_TYPE(XEN_GtkAdjustment__P(vadj), vadj, 3, "gtk_text_set_adjustments", "GtkAdjustment*");
  gtk_text_set_adjustments(XEN_TO_C_GtkText_(text), XEN_TO_C_GtkAdjustment_(hadj), XEN_TO_C_GtkAdjustment_(vadj));
  return(XEN_FALSE);
}

static XEN gxg_gtk_text_set_point(XEN text, XEN index)
{
  #define H_gtk_text_set_point "void gtk_text_set_point(GtkText* text, guint index)"
  XEN_ASSERT_TYPE(XEN_GtkText__P(text), text, 1, "gtk_text_set_point", "GtkText*");
  XEN_ASSERT_TYPE(XEN_guint_P(index), index, 2, "gtk_text_set_point", "guint");
  gtk_text_set_point(XEN_TO_C_GtkText_(text), XEN_TO_C_guint(index));
  return(XEN_FALSE);
}

static XEN gxg_gtk_text_get_point(XEN text)
{
  #define H_gtk_text_get_point "guint gtk_text_get_point(GtkText* text)"
  XEN_ASSERT_TYPE(XEN_GtkText__P(text), text, 1, "gtk_text_get_point", "GtkText*");
  return(C_TO_XEN_guint(gtk_text_get_point(XEN_TO_C_GtkText_(text))));
}

static XEN gxg_gtk_text_get_length(XEN text)
{
  #define H_gtk_text_get_length "guint gtk_text_get_length(GtkText* text)"
  XEN_ASSERT_TYPE(XEN_GtkText__P(text), text, 1, "gtk_text_get_length", "GtkText*");
  return(C_TO_XEN_guint(gtk_text_get_length(XEN_TO_C_GtkText_(text))));
}

static XEN gxg_gtk_text_freeze(XEN text)
{
  #define H_gtk_text_freeze "void gtk_text_freeze(GtkText* text)"
  XEN_ASSERT_TYPE(XEN_GtkText__P(text), text, 1, "gtk_text_freeze", "GtkText*");
  gtk_text_freeze(XEN_TO_C_GtkText_(text));
  return(XEN_FALSE);
}

static XEN gxg_gtk_text_thaw(XEN text)
{
  #define H_gtk_text_thaw "void gtk_text_thaw(GtkText* text)"
  XEN_ASSERT_TYPE(XEN_GtkText__P(text), text, 1, "gtk_text_thaw", "GtkText*");
  gtk_text_thaw(XEN_TO_C_GtkText_(text));
  return(XEN_FALSE);
}

static XEN gxg_gtk_text_insert(XEN text, XEN font, XEN fore, XEN back, XEN chars, XEN length)
{
  #define H_gtk_text_insert "void gtk_text_insert(GtkText* text, GdkFont* font, GdkColor* fore, GdkColor* back, \
char* chars, gint length)"
  XEN_ASSERT_TYPE(XEN_GtkText__P(text), text, 1, "gtk_text_insert", "GtkText*");
  XEN_ASSERT_TYPE(XEN_GdkFont__P(font), font, 2, "gtk_text_insert", "GdkFont*");
  XEN_ASSERT_TYPE(XEN_GdkColor__P(fore), fore, 3, "gtk_text_insert", "GdkColor*");
  XEN_ASSERT_TYPE(XEN_GdkColor__P(back), back, 4, "gtk_text_insert", "GdkColor*");
  XEN_ASSERT_TYPE(XEN_char__P(chars), chars, 5, "gtk_text_insert", "char*");
  XEN_ASSERT_TYPE(XEN_gint_P(length), length, 6, "gtk_text_insert", "gint");
  gtk_text_insert(XEN_TO_C_GtkText_(text), XEN_TO_C_GdkFont_(font), XEN_TO_C_GdkColor_(fore), XEN_TO_C_GdkColor_(back), XEN_TO_C_char_(chars), 
                  XEN_TO_C_gint(length));
  return(XEN_FALSE);
}

static XEN gxg_gtk_text_backward_delete(XEN text, XEN nchars)
{
  #define H_gtk_text_backward_delete "gboolean gtk_text_backward_delete(GtkText* text, guint nchars)"
  XEN_ASSERT_TYPE(XEN_GtkText__P(text), text, 1, "gtk_text_backward_delete", "GtkText*");
  XEN_ASSERT_TYPE(XEN_guint_P(nchars), nchars, 2, "gtk_text_backward_delete", "guint");
  return(C_TO_XEN_gboolean(gtk_text_backward_delete(XEN_TO_C_GtkText_(text), XEN_TO_C_guint(nchars))));
}

static XEN gxg_gtk_text_forward_delete(XEN text, XEN nchars)
{
  #define H_gtk_text_forward_delete "gboolean gtk_text_forward_delete(GtkText* text, guint nchars)"
  XEN_ASSERT_TYPE(XEN_GtkText__P(text), text, 1, "gtk_text_forward_delete", "GtkText*");
  XEN_ASSERT_TYPE(XEN_guint_P(nchars), nchars, 2, "gtk_text_forward_delete", "guint");
  return(C_TO_XEN_gboolean(gtk_text_forward_delete(XEN_TO_C_GtkText_(text), XEN_TO_C_guint(nchars))));
}

static XEN gxg_gtk_tree_get_type(void)
{
  #define H_gtk_tree_get_type "GtkType gtk_tree_get_type( void)"
  return(C_TO_XEN_GtkType(gtk_tree_get_type()));
}

static XEN gxg_gtk_tree_new(void)
{
  #define H_gtk_tree_new "GtkWidget* gtk_tree_new( void)"
  return(C_TO_XEN_GtkWidget_(gtk_tree_new()));
}

static XEN gxg_gtk_tree_append(XEN tree, XEN tree_item)
{
  #define H_gtk_tree_append "void gtk_tree_append(GtkTree* tree, GtkWidget* tree_item)"
  XEN_ASSERT_TYPE(XEN_GtkTree__P(tree), tree, 1, "gtk_tree_append", "GtkTree*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(tree_item), tree_item, 2, "gtk_tree_append", "GtkWidget*");
  gtk_tree_append(XEN_TO_C_GtkTree_(tree), XEN_TO_C_GtkWidget_(tree_item));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tree_prepend(XEN tree, XEN tree_item)
{
  #define H_gtk_tree_prepend "void gtk_tree_prepend(GtkTree* tree, GtkWidget* tree_item)"
  XEN_ASSERT_TYPE(XEN_GtkTree__P(tree), tree, 1, "gtk_tree_prepend", "GtkTree*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(tree_item), tree_item, 2, "gtk_tree_prepend", "GtkWidget*");
  gtk_tree_prepend(XEN_TO_C_GtkTree_(tree), XEN_TO_C_GtkWidget_(tree_item));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tree_insert(XEN tree, XEN tree_item, XEN position)
{
  #define H_gtk_tree_insert "void gtk_tree_insert(GtkTree* tree, GtkWidget* tree_item, gint position)"
  XEN_ASSERT_TYPE(XEN_GtkTree__P(tree), tree, 1, "gtk_tree_insert", "GtkTree*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(tree_item), tree_item, 2, "gtk_tree_insert", "GtkWidget*");
  XEN_ASSERT_TYPE(XEN_gint_P(position), position, 3, "gtk_tree_insert", "gint");
  gtk_tree_insert(XEN_TO_C_GtkTree_(tree), XEN_TO_C_GtkWidget_(tree_item), XEN_TO_C_gint(position));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tree_remove_items(XEN tree, XEN items)
{
  #define H_gtk_tree_remove_items "void gtk_tree_remove_items(GtkTree* tree, GList* items)"
  XEN_ASSERT_TYPE(XEN_GtkTree__P(tree), tree, 1, "gtk_tree_remove_items", "GtkTree*");
  XEN_ASSERT_TYPE(XEN_GList__P(items), items, 2, "gtk_tree_remove_items", "GList*");
  gtk_tree_remove_items(XEN_TO_C_GtkTree_(tree), XEN_TO_C_GList_(items));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tree_clear_items(XEN tree, XEN start, XEN end)
{
  #define H_gtk_tree_clear_items "void gtk_tree_clear_items(GtkTree* tree, gint start, gint end)"
  XEN_ASSERT_TYPE(XEN_GtkTree__P(tree), tree, 1, "gtk_tree_clear_items", "GtkTree*");
  XEN_ASSERT_TYPE(XEN_gint_P(start), start, 2, "gtk_tree_clear_items", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(end), end, 3, "gtk_tree_clear_items", "gint");
  gtk_tree_clear_items(XEN_TO_C_GtkTree_(tree), XEN_TO_C_gint(start), XEN_TO_C_gint(end));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tree_select_item(XEN tree, XEN item)
{
  #define H_gtk_tree_select_item "void gtk_tree_select_item(GtkTree* tree, gint item)"
  XEN_ASSERT_TYPE(XEN_GtkTree__P(tree), tree, 1, "gtk_tree_select_item", "GtkTree*");
  XEN_ASSERT_TYPE(XEN_gint_P(item), item, 2, "gtk_tree_select_item", "gint");
  gtk_tree_select_item(XEN_TO_C_GtkTree_(tree), XEN_TO_C_gint(item));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tree_unselect_item(XEN tree, XEN item)
{
  #define H_gtk_tree_unselect_item "void gtk_tree_unselect_item(GtkTree* tree, gint item)"
  XEN_ASSERT_TYPE(XEN_GtkTree__P(tree), tree, 1, "gtk_tree_unselect_item", "GtkTree*");
  XEN_ASSERT_TYPE(XEN_gint_P(item), item, 2, "gtk_tree_unselect_item", "gint");
  gtk_tree_unselect_item(XEN_TO_C_GtkTree_(tree), XEN_TO_C_gint(item));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tree_select_child(XEN tree, XEN tree_item)
{
  #define H_gtk_tree_select_child "void gtk_tree_select_child(GtkTree* tree, GtkWidget* tree_item)"
  XEN_ASSERT_TYPE(XEN_GtkTree__P(tree), tree, 1, "gtk_tree_select_child", "GtkTree*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(tree_item), tree_item, 2, "gtk_tree_select_child", "GtkWidget*");
  gtk_tree_select_child(XEN_TO_C_GtkTree_(tree), XEN_TO_C_GtkWidget_(tree_item));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tree_unselect_child(XEN tree, XEN tree_item)
{
  #define H_gtk_tree_unselect_child "void gtk_tree_unselect_child(GtkTree* tree, GtkWidget* tree_item)"
  XEN_ASSERT_TYPE(XEN_GtkTree__P(tree), tree, 1, "gtk_tree_unselect_child", "GtkTree*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(tree_item), tree_item, 2, "gtk_tree_unselect_child", "GtkWidget*");
  gtk_tree_unselect_child(XEN_TO_C_GtkTree_(tree), XEN_TO_C_GtkWidget_(tree_item));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tree_child_position(XEN tree, XEN child)
{
  #define H_gtk_tree_child_position "gint gtk_tree_child_position(GtkTree* tree, GtkWidget* child)"
  XEN_ASSERT_TYPE(XEN_GtkTree__P(tree), tree, 1, "gtk_tree_child_position", "GtkTree*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(child), child, 2, "gtk_tree_child_position", "GtkWidget*");
  return(C_TO_XEN_gint(gtk_tree_child_position(XEN_TO_C_GtkTree_(tree), XEN_TO_C_GtkWidget_(child))));
}

static XEN gxg_gtk_tree_set_selection_mode(XEN tree, XEN mode)
{
  #define H_gtk_tree_set_selection_mode "void gtk_tree_set_selection_mode(GtkTree* tree, GtkSelectionMode mode)"
  XEN_ASSERT_TYPE(XEN_GtkTree__P(tree), tree, 1, "gtk_tree_set_selection_mode", "GtkTree*");
  XEN_ASSERT_TYPE(XEN_GtkSelectionMode_P(mode), mode, 2, "gtk_tree_set_selection_mode", "GtkSelectionMode");
  gtk_tree_set_selection_mode(XEN_TO_C_GtkTree_(tree), XEN_TO_C_GtkSelectionMode(mode));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tree_set_view_mode(XEN tree, XEN mode)
{
  #define H_gtk_tree_set_view_mode "void gtk_tree_set_view_mode(GtkTree* tree, GtkTreeViewMode mode)"
  XEN_ASSERT_TYPE(XEN_GtkTree__P(tree), tree, 1, "gtk_tree_set_view_mode", "GtkTree*");
  XEN_ASSERT_TYPE(XEN_GtkTreeViewMode_P(mode), mode, 2, "gtk_tree_set_view_mode", "GtkTreeViewMode");
  gtk_tree_set_view_mode(XEN_TO_C_GtkTree_(tree), XEN_TO_C_GtkTreeViewMode(mode));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tree_set_view_lines(XEN tree, XEN flag)
{
  #define H_gtk_tree_set_view_lines "void gtk_tree_set_view_lines(GtkTree* tree, gboolean flag)"
  XEN_ASSERT_TYPE(XEN_GtkTree__P(tree), tree, 1, "gtk_tree_set_view_lines", "GtkTree*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(flag), flag, 2, "gtk_tree_set_view_lines", "gboolean");
  gtk_tree_set_view_lines(XEN_TO_C_GtkTree_(tree), XEN_TO_C_gboolean(flag));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tree_remove_item(XEN tree, XEN child)
{
  #define H_gtk_tree_remove_item "void gtk_tree_remove_item(GtkTree* tree, GtkWidget* child)"
  XEN_ASSERT_TYPE(XEN_GtkTree__P(tree), tree, 1, "gtk_tree_remove_item", "GtkTree*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(child), child, 2, "gtk_tree_remove_item", "GtkWidget*");
  gtk_tree_remove_item(XEN_TO_C_GtkTree_(tree), XEN_TO_C_GtkWidget_(child));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tree_item_get_type(void)
{
  #define H_gtk_tree_item_get_type "GtkType gtk_tree_item_get_type( void)"
  return(C_TO_XEN_GtkType(gtk_tree_item_get_type()));
}

static XEN gxg_gtk_tree_item_new(void)
{
  #define H_gtk_tree_item_new "GtkWidget* gtk_tree_item_new( void)"
  return(C_TO_XEN_GtkWidget_(gtk_tree_item_new()));
}

static XEN gxg_gtk_tree_item_new_with_label(XEN label)
{
  #define H_gtk_tree_item_new_with_label "GtkWidget* gtk_tree_item_new_with_label(gchar* label)"
  XEN_ASSERT_TYPE(XEN_gchar__P(label), label, 1, "gtk_tree_item_new_with_label", "gchar*");
  return(C_TO_XEN_GtkWidget_(gtk_tree_item_new_with_label(XEN_TO_C_gchar_(label))));
}

static XEN gxg_gtk_tree_item_set_subtree(XEN tree_item, XEN subtree)
{
  #define H_gtk_tree_item_set_subtree "void gtk_tree_item_set_subtree(GtkTreeItem* tree_item, GtkWidget* subtree)"
  XEN_ASSERT_TYPE(XEN_GtkTreeItem__P(tree_item), tree_item, 1, "gtk_tree_item_set_subtree", "GtkTreeItem*");
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(subtree), subtree, 2, "gtk_tree_item_set_subtree", "GtkWidget*");
  gtk_tree_item_set_subtree(XEN_TO_C_GtkTreeItem_(tree_item), XEN_TO_C_GtkWidget_(subtree));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tree_item_remove_subtree(XEN tree_item)
{
  #define H_gtk_tree_item_remove_subtree "void gtk_tree_item_remove_subtree(GtkTreeItem* tree_item)"
  XEN_ASSERT_TYPE(XEN_GtkTreeItem__P(tree_item), tree_item, 1, "gtk_tree_item_remove_subtree", "GtkTreeItem*");
  gtk_tree_item_remove_subtree(XEN_TO_C_GtkTreeItem_(tree_item));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tree_item_select(XEN tree_item)
{
  #define H_gtk_tree_item_select "void gtk_tree_item_select(GtkTreeItem* tree_item)"
  XEN_ASSERT_TYPE(XEN_GtkTreeItem__P(tree_item), tree_item, 1, "gtk_tree_item_select", "GtkTreeItem*");
  gtk_tree_item_select(XEN_TO_C_GtkTreeItem_(tree_item));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tree_item_deselect(XEN tree_item)
{
  #define H_gtk_tree_item_deselect "void gtk_tree_item_deselect(GtkTreeItem* tree_item)"
  XEN_ASSERT_TYPE(XEN_GtkTreeItem__P(tree_item), tree_item, 1, "gtk_tree_item_deselect", "GtkTreeItem*");
  gtk_tree_item_deselect(XEN_TO_C_GtkTreeItem_(tree_item));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tree_item_expand(XEN tree_item)
{
  #define H_gtk_tree_item_expand "void gtk_tree_item_expand(GtkTreeItem* tree_item)"
  XEN_ASSERT_TYPE(XEN_GtkTreeItem__P(tree_item), tree_item, 1, "gtk_tree_item_expand", "GtkTreeItem*");
  gtk_tree_item_expand(XEN_TO_C_GtkTreeItem_(tree_item));
  return(XEN_FALSE);
}

static XEN gxg_gtk_tree_item_collapse(XEN tree_item)
{
  #define H_gtk_tree_item_collapse "void gtk_tree_item_collapse(GtkTreeItem* tree_item)"
  XEN_ASSERT_TYPE(XEN_GtkTreeItem__P(tree_item), tree_item, 1, "gtk_tree_item_collapse", "GtkTreeItem*");
  gtk_tree_item_collapse(XEN_TO_C_GtkTreeItem_(tree_item));
  return(XEN_FALSE);
}

static XEN gxg_gdk_image_new_bitmap(XEN visual, XEN data, XEN width, XEN height)
{
  #define H_gdk_image_new_bitmap "GdkImage* gdk_image_new_bitmap(GdkVisual* visual, gpointer data, gint width, \
gint height)"
  XEN_ASSERT_TYPE(XEN_GdkVisual__P(visual), visual, 1, "gdk_image_new_bitmap", "GdkVisual*");
  XEN_ASSERT_TYPE(XEN_gpointer_P(data), data, 2, "gdk_image_new_bitmap", "gpointer");
  XEN_ASSERT_TYPE(XEN_gint_P(width), width, 3, "gdk_image_new_bitmap", "gint");
  XEN_ASSERT_TYPE(XEN_gint_P(height), height, 4, "gdk_image_new_bitmap", "gint");
  return(C_TO_XEN_GdkImage_(gdk_image_new_bitmap(XEN_TO_C_GdkVisual_(visual), XEN_TO_C_gpointer(data), XEN_TO_C_gint(width), 
                                                 XEN_TO_C_gint(height))));
}

static XEN gxg_gtk_text_get_type(void)
{
  #define H_gtk_text_get_type "GtkType gtk_text_get_type( void)"
  return(C_TO_XEN_GtkType(gtk_text_get_type()));
}

static XEN gxg_gtk_text_new(XEN hadj, XEN vadj)
{
  #define H_gtk_text_new "GtkWidget* gtk_text_new(GtkAdjustment* hadj, GtkAdjustment* vadj)"
  XEN_ASSERT_TYPE(XEN_GtkAdjustment__P(hadj), hadj, 1, "gtk_text_new", "GtkAdjustment*");
  XEN_ASSERT_TYPE(XEN_GtkAdjustment__P(vadj), vadj, 2, "gtk_text_new", "GtkAdjustment*");
  return(C_TO_XEN_GtkWidget_(gtk_text_new(XEN_TO_C_GtkAdjustment_(hadj), XEN_TO_C_GtkAdjustment_(vadj))));
}

static XEN gxg_gtk_text_set_editable(XEN text, XEN editable)
{
  #define H_gtk_text_set_editable "void gtk_text_set_editable(GtkText* text, gboolean editable)"
  XEN_ASSERT_TYPE(XEN_GtkText__P(text), text, 1, "gtk_text_set_editable", "GtkText*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(editable), editable, 2, "gtk_text_set_editable", "gboolean");
  gtk_text_set_editable(XEN_TO_C_GtkText_(text), XEN_TO_C_gboolean(editable));
  return(XEN_FALSE);
}

static XEN gxg_gtk_text_set_word_wrap(XEN text, XEN word_wrap)
{
  #define H_gtk_text_set_word_wrap "void gtk_text_set_word_wrap(GtkText* text, gboolean word_wrap)"
  XEN_ASSERT_TYPE(XEN_GtkText__P(text), text, 1, "gtk_text_set_word_wrap", "GtkText*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(word_wrap), word_wrap, 2, "gtk_text_set_word_wrap", "gboolean");
  gtk_text_set_word_wrap(XEN_TO_C_GtkText_(text), XEN_TO_C_gboolean(word_wrap));
  return(XEN_FALSE);
}

static XEN gxg_gtk_text_set_line_wrap(XEN text, XEN line_wrap)
{
  #define H_gtk_text_set_line_wrap "void gtk_text_set_line_wrap(GtkText* text, gboolean line_wrap)"
  XEN_ASSERT_TYPE(XEN_GtkText__P(text), text, 1, "gtk_text_set_line_wrap", "GtkText*");
  XEN_ASSERT_TYPE(XEN_gboolean_P(line_wrap), line_wrap, 2, "gtk_text_set_line_wrap", "gboolean");
  gtk_text_set_line_wrap(XEN_TO_C_GtkText_(text), XEN_TO_C_gboolean(line_wrap));
  return(XEN_FALSE);
}
#endif

static XEN gxg_GDK_COLORMAP(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GdkColormap_"), XEN_CADR(obj)));}
static XEN gxg_GDK_DRAG_CONTEXT(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GdkDragContext_"), XEN_CADR(obj)));}
#if (!HAVE_GTK_1)
static XEN gxg_GDK_DRAWABLE(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GdkDrawable_"), XEN_CADR(obj)));}
#endif
static XEN gxg_GDK_GC(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GdkGC_"), XEN_CADR(obj)));}
static XEN gxg_GDK_IMAGE(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GdkImage_"), XEN_CADR(obj)));}
#if (!HAVE_GTK_1)
static XEN gxg_GDK_DEVICE(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GdkDevice_"), XEN_CADR(obj)));}
static XEN gxg_GDK_KEYMAP(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GdkKeymap_"), XEN_CADR(obj)));}
static XEN gxg_GDK_PIXMAP(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GdkPixmap_"), XEN_CADR(obj)));}
#endif
static XEN gxg_GDK_VISUAL(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GdkVisual_"), XEN_CADR(obj)));}
static XEN gxg_GDK_WINDOW(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GdkWindow_"), XEN_CADR(obj)));}
#if (!HAVE_GTK_1)
static XEN gxg_GDK_WINDOW_OBJECT(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GdkWindowObject_"), XEN_CADR(obj)));}
static XEN gxg_GDK_PIXBUF(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GdkPixbuf_"), XEN_CADR(obj)));}
static XEN gxg_GDK_PIXBUF_ANIMATION(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GdkPixbufAnimation_"), XEN_CADR(obj)));}
static XEN gxg_GDK_PIXBUF_ANIMATION_ITER(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GdkPixbufAnimationIter_"), XEN_CADR(obj)));}
static XEN gxg_GDK_PIXBUF_LOADER(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GdkPixbufLoader_"), XEN_CADR(obj)));}
#endif
static XEN gxg_GTK_VBOX(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkVBox_"), XEN_CADR(obj)));}
static XEN gxg_GTK_ACCEL_GROUP(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkAccelGroup_"), XEN_CADR(obj)));}
static XEN gxg_GTK_ACCEL_LABEL(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkAccelLabel_"), XEN_CADR(obj)));}
#if (!HAVE_GTK_1)
static XEN gxg_GTK_ACCESSIBLE(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkAccessible_"), XEN_CADR(obj)));}
#endif
static XEN gxg_GTK_ADJUSTMENT(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkAdjustment_"), XEN_CADR(obj)));}
static XEN gxg_GTK_ALIGNMENT(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkAlignment_"), XEN_CADR(obj)));}
static XEN gxg_GTK_ARROW(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkArrow_"), XEN_CADR(obj)));}
static XEN gxg_GTK_ASPECT_FRAME(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkAspectFrame_"), XEN_CADR(obj)));}
static XEN gxg_GTK_BUTTON_BOX(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkButtonBox_"), XEN_CADR(obj)));}
static XEN gxg_GTK_BIN(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkBin_"), XEN_CADR(obj)));}
static XEN gxg_GTK_BOX(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkBox_"), XEN_CADR(obj)));}
static XEN gxg_GTK_BUTTON(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkButton_"), XEN_CADR(obj)));}
static XEN gxg_GTK_CALENDAR(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkCalendar_"), XEN_CADR(obj)));}
#if (!HAVE_GTK_1)
static XEN gxg_GTK_CELL_EDITABLE(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkCellEditable_"), XEN_CADR(obj)));}
static XEN gxg_GTK_CELL_RENDERER(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkCellRenderer_"), XEN_CADR(obj)));}
static XEN gxg_GTK_CELL_RENDERER_PIXBUF(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkCellRendererPixbuf_"), XEN_CADR(obj)));}
static XEN gxg_GTK_CELL_RENDERER_TEXT(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkCellRendererText_"), XEN_CADR(obj)));}
static XEN gxg_GTK_CELL_RENDERER_TOGGLE(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkCellRendererToggle_"), XEN_CADR(obj)));}
#endif
static XEN gxg_GTK_CHECK_BUTTON(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkCheckButton_"), XEN_CADR(obj)));}
static XEN gxg_GTK_CHECK_MENU_ITEM(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkCheckMenuItem_"), XEN_CADR(obj)));}
static XEN gxg_GTK_COLOR_SELECTION_DIALOG(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkColorSelectionDialog_"), XEN_CADR(obj)));}
static XEN gxg_GTK_COLOR_SELECTION(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkColorSelection_"), XEN_CADR(obj)));}
static XEN gxg_GTK_COMBO(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkCombo_"), XEN_CADR(obj)));}
static XEN gxg_GTK_CONTAINER(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkContainer_"), XEN_CADR(obj)));}
static XEN gxg_GTK_CURVE(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkCurve_"), XEN_CADR(obj)));}
static XEN gxg_GTK_DIALOG(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkDialog_"), XEN_CADR(obj)));}
static XEN gxg_GTK_DRAWING_AREA(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkDrawingArea_"), XEN_CADR(obj)));}
static XEN gxg_GTK_EDITABLE(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkEditable_"), XEN_CADR(obj)));}
static XEN gxg_GTK_ENTRY(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkEntry_"), XEN_CADR(obj)));}
static XEN gxg_GTK_EVENT_BOX(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkEventBox_"), XEN_CADR(obj)));}
static XEN gxg_GTK_FILE_SELECTION(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkFileSelection_"), XEN_CADR(obj)));}
static XEN gxg_GTK_FIXED(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkFixed_"), XEN_CADR(obj)));}
static XEN gxg_GTK_FONT_SELECTION(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkFontSelection_"), XEN_CADR(obj)));}
static XEN gxg_GTK_FONT_SELECTION_DIALOG(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkFontSelectionDialog_"), XEN_CADR(obj)));}
static XEN gxg_GTK_FRAME(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkFrame_"), XEN_CADR(obj)));}
static XEN gxg_GTK_GAMMA_CURVE(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkGammaCurve_"), XEN_CADR(obj)));}
static XEN gxg_GTK_HANDLE_BOX(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkHandleBox_"), XEN_CADR(obj)));}
static XEN gxg_GTK_HBUTTON_BOX(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkHButtonBox_"), XEN_CADR(obj)));}
static XEN gxg_GTK_HBOX(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkHBox_"), XEN_CADR(obj)));}
static XEN gxg_GTK_HPANED(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkHPaned_"), XEN_CADR(obj)));}
static XEN gxg_GTK_HRULER(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkHRuler_"), XEN_CADR(obj)));}
static XEN gxg_GTK_HSCALE(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkHScale_"), XEN_CADR(obj)));}
static XEN gxg_GTK_HSCROLLBAR(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkHScrollbar_"), XEN_CADR(obj)));}
static XEN gxg_GTK_HSEPARATOR(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkHSeparator_"), XEN_CADR(obj)));}
#if (!HAVE_GTK_1)
static XEN gxg_GTK_ICON_FACTORY(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkIconFactory_"), XEN_CADR(obj)));}
#endif
static XEN gxg_GTK_IMAGE(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkImage_"), XEN_CADR(obj)));}
#if (!HAVE_GTK_1)
static XEN gxg_GTK_IMAGE_MENU_ITEM(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkImageMenuItem_"), XEN_CADR(obj)));}
static XEN gxg_GTK_IM_CONTEXT(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkIMContext_"), XEN_CADR(obj)));}
static XEN gxg_GTK_IM_CONTEXT_SIMPLE(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkIMContextSimple_"), XEN_CADR(obj)));}
static XEN gxg_GTK_IM_MULTICONTEXT(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkIMMulticontext_"), XEN_CADR(obj)));}
#endif
static XEN gxg_GTK_INPUT_DIALOG(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkInputDialog_"), XEN_CADR(obj)));}
static XEN gxg_GTK_INVISIBLE(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkInvisible_"), XEN_CADR(obj)));}
static XEN gxg_GTK_ITEM_FACTORY(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkItemFactory_"), XEN_CADR(obj)));}
static XEN gxg_GTK_ITEM(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkItem_"), XEN_CADR(obj)));}
static XEN gxg_GTK_LABEL(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkLabel_"), XEN_CADR(obj)));}
static XEN gxg_GTK_LAYOUT(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkLayout_"), XEN_CADR(obj)));}
#if (!HAVE_GTK_1)
static XEN gxg_GTK_LIST_STORE(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkListStore_"), XEN_CADR(obj)));}
#endif
static XEN gxg_GTK_MENU_BAR(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkMenuBar_"), XEN_CADR(obj)));}
static XEN gxg_GTK_MENU(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkMenu_"), XEN_CADR(obj)));}
static XEN gxg_GTK_MENU_ITEM(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkMenuItem_"), XEN_CADR(obj)));}
static XEN gxg_GTK_MENU_SHELL(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkMenuShell_"), XEN_CADR(obj)));}
#if (!HAVE_GTK_1)
static XEN gxg_GTK_MESSAGE_DIALOG(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkMessageDialog_"), XEN_CADR(obj)));}
#endif
static XEN gxg_GTK_MISC(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkMisc_"), XEN_CADR(obj)));}
static XEN gxg_GTK_NOTEBOOK(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkNotebook_"), XEN_CADR(obj)));}
static XEN gxg_GTK_OBJECT(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkObject_"), XEN_CADR(obj)));}
static XEN gxg_GTK_OPTION_MENU(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkOptionMenu_"), XEN_CADR(obj)));}
static XEN gxg_GTK_PANED(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkPaned_"), XEN_CADR(obj)));}
static XEN gxg_GTK_PLUG(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkPlug_"), XEN_CADR(obj)));}
static XEN gxg_GTK_PROGRESS_BAR(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkProgressBar_"), XEN_CADR(obj)));}
static XEN gxg_GTK_RADIO_BUTTON(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkRadioButton_"), XEN_CADR(obj)));}
static XEN gxg_GTK_RADIO_MENU_ITEM(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkRadioMenuItem_"), XEN_CADR(obj)));}
static XEN gxg_GTK_RANGE(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkRange_"), XEN_CADR(obj)));}
static XEN gxg_GTK_RC_STYLE(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkRcStyle_"), XEN_CADR(obj)));}
static XEN gxg_GTK_RULER(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkRuler_"), XEN_CADR(obj)));}
static XEN gxg_GTK_SCALE(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkScale_"), XEN_CADR(obj)));}
static XEN gxg_GTK_SCROLLBAR(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkScrollbar_"), XEN_CADR(obj)));}
static XEN gxg_GTK_SCROLLED_WINDOW(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkScrolledWindow_"), XEN_CADR(obj)));}
static XEN gxg_GTK_SEPARATOR(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkSeparator_"), XEN_CADR(obj)));}
#if (!HAVE_GTK_1)
static XEN gxg_GTK_SEPARATOR_MENU_ITEM(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkSeparatorMenuItem_"), XEN_CADR(obj)));}
static XEN gxg_GTK_SETTINGS(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkSettings_"), XEN_CADR(obj)));}
static XEN gxg_GTK_SIZE_GROUP(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkSizeGroup_"), XEN_CADR(obj)));}
#endif
static XEN gxg_GTK_SOCKET(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkSocket_"), XEN_CADR(obj)));}
static XEN gxg_GTK_SPIN_BUTTON(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkSpinButton_"), XEN_CADR(obj)));}
static XEN gxg_GTK_STATUSBAR(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkStatusbar_"), XEN_CADR(obj)));}
static XEN gxg_GTK_STYLE(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkStyle_"), XEN_CADR(obj)));}
static XEN gxg_GTK_TABLE(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkTable_"), XEN_CADR(obj)));}
static XEN gxg_GTK_TEAROFF_MENU_ITEM(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkTearoffMenuItem_"), XEN_CADR(obj)));}
#if (!HAVE_GTK_1)
static XEN gxg_GTK_TEXT_BUFFER(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkTextBuffer_"), XEN_CADR(obj)));}
static XEN gxg_GTK_TEXT_CHILD_ANCHOR(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkTextChildAnchor_"), XEN_CADR(obj)));}
static XEN gxg_GTK_TEXT_MARK(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkTextMark_"), XEN_CADR(obj)));}
static XEN gxg_GTK_TEXT_TAG(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkTextTag_"), XEN_CADR(obj)));}
static XEN gxg_GTK_TEXT_TAG_TABLE(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkTextTagTable_"), XEN_CADR(obj)));}
static XEN gxg_GTK_TEXT_VIEW(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkTextView_"), XEN_CADR(obj)));}
static XEN gxg_GTK_THEME_ENGINE(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkThemeEngine_"), XEN_CADR(obj)));}
#endif
static XEN gxg_GTK_TOGGLE_BUTTON(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkToggleButton_"), XEN_CADR(obj)));}
static XEN gxg_GTK_TOOLBAR(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkToolbar_"), XEN_CADR(obj)));}
static XEN gxg_GTK_TOOLTIPS(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkTooltips_"), XEN_CADR(obj)));}
#if (!HAVE_GTK_1)
static XEN gxg_GTK_TREE_DRAG_SOURCE(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkTreeDragSource_"), XEN_CADR(obj)));}
static XEN gxg_GTK_TREE_DRAG_DEST(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkTreeDragDest_"), XEN_CADR(obj)));}
static XEN gxg_GTK_TREE_MODEL(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkTreeModel_"), XEN_CADR(obj)));}
static XEN gxg_GTK_TREE_MODEL_SORT(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkTreeModelSort_"), XEN_CADR(obj)));}
#endif
static XEN gxg_GTK_TREE_SELECTION(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkTreeSelection_"), XEN_CADR(obj)));}
#if (!HAVE_GTK_1)
static XEN gxg_GTK_TREE_SORTABLE(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkTreeSortable_"), XEN_CADR(obj)));}
static XEN gxg_GTK_TREE_STORE(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkTreeStore_"), XEN_CADR(obj)));}
static XEN gxg_GTK_TREE_VIEW_COLUMN(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkTreeViewColumn_"), XEN_CADR(obj)));}
#endif
static XEN gxg_GTK_TREE_VIEW(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkTreeView_"), XEN_CADR(obj)));}
static XEN gxg_GTK_VBUTTON_BOX(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkVButtonBox_"), XEN_CADR(obj)));}
static XEN gxg_GTK_VIEWPORT(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkViewport_"), XEN_CADR(obj)));}
static XEN gxg_GTK_VPANED(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkVPaned_"), XEN_CADR(obj)));}
static XEN gxg_GTK_VRULER(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkVRuler_"), XEN_CADR(obj)));}
static XEN gxg_GTK_VSCALE(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkVScale_"), XEN_CADR(obj)));}
static XEN gxg_GTK_VSCROLLBAR(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkVScrollbar_"), XEN_CADR(obj)));}
static XEN gxg_GTK_VSEPARATOR(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkVSeparator_"), XEN_CADR(obj)));}
static XEN gxg_GTK_WIDGET(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkWidget_"), XEN_CADR(obj)));}
static XEN gxg_GTK_WINDOW(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkWindow_"), XEN_CADR(obj)));}
#if (!HAVE_GTK_1)
static XEN gxg_GTK_WINDOW_GROUP(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkWindowGroup_"), XEN_CADR(obj)));}
static XEN gxg_PANGO_CONTEXT(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("PangoContext_"), XEN_CADR(obj)));}
static XEN gxg_PANGO_FONT_FACE(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("PangoFontFace_"), XEN_CADR(obj)));}
static XEN gxg_PANGO_FONT(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("PangoFont_"), XEN_CADR(obj)));}
static XEN gxg_PANGO_FONT_MAP(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("PangoFontMap_"), XEN_CADR(obj)));}
static XEN gxg_PANGO_LAYOUT(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("PangoLayout_"), XEN_CADR(obj)));}
#endif
#if PANGO_ENABLE_ENGINE && PANGO_ENABLE_BACKEND
#if (!HAVE_GTK_1)
static XEN gxg_PANGO_FONT_FAMILY(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("PangoFontFamily_"), XEN_CADR(obj)));}
#endif
#endif

#if (!(defined(GDK_DISABLE_DEPRECATED))) && (!(defined(GTK_DISABLE_DEPRECATED))) && (!(defined(GDK_PIXBUF_DISABLE_DEPRECATED)))
static XEN gxg_GTK_CLIST(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkCList_"), XEN_CADR(obj)));}
static XEN gxg_GTK_CTREE(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkCTree_"), XEN_CADR(obj)));}
static XEN gxg_GTK_LIST(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkList_"), XEN_CADR(obj)));}
static XEN gxg_GTK_LIST_ITEM(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkListItem_"), XEN_CADR(obj)));}
#if (!HAVE_GTK_1)
static XEN gxg_GTK_OLD_EDITABLE(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkOldEditable_"), XEN_CADR(obj)));}
#endif
static XEN gxg_GTK_PIXMAP(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkPixmap_"), XEN_CADR(obj)));}
static XEN gxg_GTK_PREVIEW(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkPreview_"), XEN_CADR(obj)));}
static XEN gxg_GTK_PROGRESS(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkProgress_"), XEN_CADR(obj)));}
static XEN gxg_GTK_TIPS_QUERY(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkTipsQuery_"), XEN_CADR(obj)));}
#endif

#if GDK_ENABLE_BROKEN && GTK_ENABLE_BROKEN
static XEN gxg_GTK_TREE(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkTree_"), XEN_CADR(obj)));}
static XEN gxg_GTK_TREE_ITEM(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkTreeItem_"), XEN_CADR(obj)));}
static XEN gxg_GTK_TEXT(XEN obj) {return(XEN_LIST_2(C_STRING_TO_XEN_SYMBOL("GtkText_"), XEN_CADR(obj)));}
#endif

#if HAVE_GUILE
static void define_functions(void)
{
  xm_gc_table = XEN_MAKE_VECTOR(1, XEN_FALSE);
  XEN_PROTECT_FROM_GC(xm_gc_table);
  xm_protected_size = 512;
  xm_protected = XEN_MAKE_VECTOR(xm_protected_size, XEN_FALSE);
  XEN_VECTOR_SET(xm_gc_table, 0, xm_protected);

#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "g_type_name" XG_POST, gxg_g_type_name, 1, 0, 0, H_g_type_name);
  XEN_DEFINE_PROCEDURE(XG_PRE "g_type_qname" XG_POST, gxg_g_type_qname, 1, 0, 0, H_g_type_qname);
  XEN_DEFINE_PROCEDURE(XG_PRE "g_type_from_name" XG_POST, gxg_g_type_from_name, 1, 0, 0, H_g_type_from_name);
  XEN_DEFINE_PROCEDURE(XG_PRE "g_type_parent" XG_POST, gxg_g_type_parent, 1, 0, 0, H_g_type_parent);
  XEN_DEFINE_PROCEDURE(XG_PRE "g_type_is_a" XG_POST, gxg_g_type_is_a, 2, 0, 0, H_g_type_is_a);
  XEN_DEFINE_PROCEDURE(XG_PRE "g_signal_newv" XG_POST, gxg_g_signal_newv, 0, 0, 1, H_g_signal_newv);
  XEN_DEFINE_PROCEDURE(XG_PRE "g_signal_new" XG_POST, gxg_g_signal_new, 9, 0, 0, H_g_signal_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "g_signal_emitv" XG_POST, gxg_g_signal_emitv, 4, 0, 0, H_g_signal_emitv);
  XEN_DEFINE_PROCEDURE(XG_PRE "g_signal_emit" XG_POST, gxg_g_signal_emit, 3, 0, 0, H_g_signal_emit);
  XEN_DEFINE_PROCEDURE(XG_PRE "g_signal_emit_by_name" XG_POST, gxg_g_signal_emit_by_name, 2, 0, 0, H_g_signal_emit_by_name);
  XEN_DEFINE_PROCEDURE(XG_PRE "g_signal_lookup" XG_POST, gxg_g_signal_lookup, 2, 0, 0, H_g_signal_lookup);
  XEN_DEFINE_PROCEDURE(XG_PRE "g_signal_name" XG_POST, gxg_g_signal_name, 1, 0, 0, H_g_signal_name);
  XEN_DEFINE_PROCEDURE(XG_PRE "g_signal_query" XG_POST, gxg_g_signal_query, 2, 0, 0, H_g_signal_query);
  XEN_DEFINE_PROCEDURE(XG_PRE "g_signal_list_ids" XG_POST, gxg_g_signal_list_ids, 2, 0, 0, H_g_signal_list_ids);
  XEN_DEFINE_PROCEDURE(XG_PRE "g_signal_parse_name" XG_POST, gxg_g_signal_parse_name, 5, 0, 0, H_g_signal_parse_name);
  XEN_DEFINE_PROCEDURE(XG_PRE "g_signal_get_invocation_hint" XG_POST, gxg_g_signal_get_invocation_hint, 1, 0, 0, H_g_signal_get_invocation_hint);
  XEN_DEFINE_PROCEDURE(XG_PRE "g_signal_stop_emission" XG_POST, gxg_g_signal_stop_emission, 3, 0, 0, H_g_signal_stop_emission);
  XEN_DEFINE_PROCEDURE(XG_PRE "g_signal_stop_emission_by_name" XG_POST, gxg_g_signal_stop_emission_by_name, 2, 0, 0, H_g_signal_stop_emission_by_name);
  XEN_DEFINE_PROCEDURE(XG_PRE "g_signal_add_emission_hook" XG_POST, gxg_g_signal_add_emission_hook, 5, 0, 0, H_g_signal_add_emission_hook);
  XEN_DEFINE_PROCEDURE(XG_PRE "g_signal_remove_emission_hook" XG_POST, gxg_g_signal_remove_emission_hook, 2, 0, 0, H_g_signal_remove_emission_hook);
  XEN_DEFINE_PROCEDURE(XG_PRE "g_signal_has_handler_pending" XG_POST, gxg_g_signal_has_handler_pending, 4, 0, 0, H_g_signal_has_handler_pending);
  XEN_DEFINE_PROCEDURE(XG_PRE "g_signal_connect_closure_by_id" XG_POST, gxg_g_signal_connect_closure_by_id, 5, 0, 0, H_g_signal_connect_closure_by_id);
  XEN_DEFINE_PROCEDURE(XG_PRE "g_signal_connect_closure" XG_POST, gxg_g_signal_connect_closure, 4, 0, 0, H_g_signal_connect_closure);
  XEN_DEFINE_PROCEDURE(XG_PRE "g_signal_connect_data" XG_POST, gxg_g_signal_connect_data, 6, 0, 0, H_g_signal_connect_data);
  XEN_DEFINE_PROCEDURE(XG_PRE "g_signal_handler_block" XG_POST, gxg_g_signal_handler_block, 2, 0, 0, H_g_signal_handler_block);
  XEN_DEFINE_PROCEDURE(XG_PRE "g_signal_handler_unblock" XG_POST, gxg_g_signal_handler_unblock, 2, 0, 0, H_g_signal_handler_unblock);
  XEN_DEFINE_PROCEDURE(XG_PRE "g_signal_handler_disconnect" XG_POST, gxg_g_signal_handler_disconnect, 2, 0, 0, H_g_signal_handler_disconnect);
  XEN_DEFINE_PROCEDURE(XG_PRE "g_signal_handler_is_connected" XG_POST, gxg_g_signal_handler_is_connected, 2, 0, 0, H_g_signal_handler_is_connected);
  XEN_DEFINE_PROCEDURE(XG_PRE "g_signal_handler_find" XG_POST, gxg_g_signal_handler_find, 7, 0, 0, H_g_signal_handler_find);
  XEN_DEFINE_PROCEDURE(XG_PRE "g_signal_handlers_block_matched" XG_POST, gxg_g_signal_handlers_block_matched, 7, 0, 0, H_g_signal_handlers_block_matched);
  XEN_DEFINE_PROCEDURE(XG_PRE "g_signal_handlers_unblock_matched" XG_POST, gxg_g_signal_handlers_unblock_matched, 7, 0, 0, H_g_signal_handlers_unblock_matched);
  XEN_DEFINE_PROCEDURE(XG_PRE "g_signal_handlers_disconnect_matched" XG_POST, gxg_g_signal_handlers_disconnect_matched, 7, 0, 0, H_g_signal_handlers_disconnect_matched);
  XEN_DEFINE_PROCEDURE(XG_PRE "g_signal_handlers_destroy" XG_POST, gxg_g_signal_handlers_destroy, 1, 0, 0, H_g_signal_handlers_destroy);
  XEN_DEFINE_PROCEDURE(XG_PRE "_g_signals_destroy" XG_POST, gxg__g_signals_destroy, 1, 0, 0, H__g_signals_destroy);
  XEN_DEFINE_PROCEDURE(XG_PRE "g_object_ref" XG_POST, gxg_g_object_ref, 1, 0, 0, H_g_object_ref);
  XEN_DEFINE_PROCEDURE(XG_PRE "g_object_unref" XG_POST, gxg_g_object_unref, 1, 0, 0, H_g_object_unref);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_colormap_get_type" XG_POST, gxg_gdk_colormap_get_type, 0, 0, 0, H_gdk_colormap_get_type);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_colormap_new" XG_POST, gxg_gdk_colormap_new, 2, 0, 0, H_gdk_colormap_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_colormap_get_system" XG_POST, gxg_gdk_colormap_get_system, 0, 0, 0, H_gdk_colormap_get_system);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_colormap_alloc_colors" XG_POST, gxg_gdk_colormap_alloc_colors, 6, 0, 0, H_gdk_colormap_alloc_colors);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_colormap_alloc_color" XG_POST, gxg_gdk_colormap_alloc_color, 4, 0, 0, H_gdk_colormap_alloc_color);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_colormap_free_colors" XG_POST, gxg_gdk_colormap_free_colors, 3, 0, 0, H_gdk_colormap_free_colors);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_colormap_query_color" XG_POST, gxg_gdk_colormap_query_color, 3, 0, 0, H_gdk_colormap_query_color);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_colormap_get_visual" XG_POST, gxg_gdk_colormap_get_visual, 1, 0, 0, H_gdk_colormap_get_visual);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_color_copy" XG_POST, gxg_gdk_color_copy, 1, 0, 0, H_gdk_color_copy);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_color_free" XG_POST, gxg_gdk_color_free, 1, 0, 0, H_gdk_color_free);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_color_parse" XG_POST, gxg_gdk_color_parse, 2, 0, 0, H_gdk_color_parse);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_color_hash" XG_POST, gxg_gdk_color_hash, 1, 0, 0, H_gdk_color_hash);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_color_equal" XG_POST, gxg_gdk_color_equal, 2, 0, 0, H_gdk_color_equal);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_color_get_type" XG_POST, gxg_gdk_color_get_type, 0, 0, 0, H_gdk_color_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_cursor_get_type" XG_POST, gxg_gdk_cursor_get_type, 0, 0, 0, H_gdk_cursor_get_type);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_cursor_new" XG_POST, gxg_gdk_cursor_new, 1, 0, 0, H_gdk_cursor_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_cursor_new_from_pixmap" XG_POST, gxg_gdk_cursor_new_from_pixmap, 6, 0, 0, H_gdk_cursor_new_from_pixmap);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_cursor_ref" XG_POST, gxg_gdk_cursor_ref, 1, 0, 0, H_gdk_cursor_ref);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_cursor_unref" XG_POST, gxg_gdk_cursor_unref, 1, 0, 0, H_gdk_cursor_unref);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_drag_context_get_type" XG_POST, gxg_gdk_drag_context_get_type, 0, 0, 0, H_gdk_drag_context_get_type);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_drag_context_new" XG_POST, gxg_gdk_drag_context_new, 0, 0, 0, H_gdk_drag_context_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_drag_status" XG_POST, gxg_gdk_drag_status, 3, 0, 0, H_gdk_drag_status);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_drop_reply" XG_POST, gxg_gdk_drop_reply, 3, 0, 0, H_gdk_drop_reply);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_drop_finish" XG_POST, gxg_gdk_drop_finish, 3, 0, 0, H_gdk_drop_finish);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_drag_get_selection" XG_POST, gxg_gdk_drag_get_selection, 1, 0, 0, H_gdk_drag_get_selection);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_drag_begin" XG_POST, gxg_gdk_drag_begin, 2, 0, 0, H_gdk_drag_begin);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_drag_get_protocol" XG_POST, gxg_gdk_drag_get_protocol, 2, 0, 0, H_gdk_drag_get_protocol);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_drag_find_window" XG_POST, gxg_gdk_drag_find_window, 4, 2, 0, H_gdk_drag_find_window);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_drag_motion" XG_POST, gxg_gdk_drag_motion, 8, 0, 0, H_gdk_drag_motion);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_drag_drop" XG_POST, gxg_gdk_drag_drop, 2, 0, 0, H_gdk_drag_drop);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_drag_abort" XG_POST, gxg_gdk_drag_abort, 2, 0, 0, H_gdk_drag_abort);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_drawable_get_type" XG_POST, gxg_gdk_drawable_get_type, 0, 0, 0, H_gdk_drawable_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_drawable_get_size" XG_POST, gxg_gdk_drawable_get_size, 1, 2, 0, H_gdk_drawable_get_size);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_drawable_set_colormap" XG_POST, gxg_gdk_drawable_set_colormap, 2, 0, 0, H_gdk_drawable_set_colormap);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_drawable_get_colormap" XG_POST, gxg_gdk_drawable_get_colormap, 1, 0, 0, H_gdk_drawable_get_colormap);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_drawable_get_visual" XG_POST, gxg_gdk_drawable_get_visual, 1, 0, 0, H_gdk_drawable_get_visual);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_drawable_get_depth" XG_POST, gxg_gdk_drawable_get_depth, 1, 0, 0, H_gdk_drawable_get_depth);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_draw_point" XG_POST, gxg_gdk_draw_point, 4, 0, 0, H_gdk_draw_point);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_draw_line" XG_POST, gxg_gdk_draw_line, 6, 0, 0, H_gdk_draw_line);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_draw_rectangle" XG_POST, gxg_gdk_draw_rectangle, 7, 0, 0, H_gdk_draw_rectangle);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_draw_arc" XG_POST, gxg_gdk_draw_arc, 9, 0, 0, H_gdk_draw_arc);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_draw_polygon" XG_POST, gxg_gdk_draw_polygon, 5, 0, 0, H_gdk_draw_polygon);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_draw_drawable" XG_POST, gxg_gdk_draw_drawable, 9, 0, 0, H_gdk_draw_drawable);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_draw_image" XG_POST, gxg_gdk_draw_image, 9, 0, 0, H_gdk_draw_image);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_draw_points" XG_POST, gxg_gdk_draw_points, 4, 0, 0, H_gdk_draw_points);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_draw_segments" XG_POST, gxg_gdk_draw_segments, 4, 0, 0, H_gdk_draw_segments);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_draw_lines" XG_POST, gxg_gdk_draw_lines, 4, 0, 0, H_gdk_draw_lines);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_draw_glyphs" XG_POST, gxg_gdk_draw_glyphs, 6, 0, 0, H_gdk_draw_glyphs);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_draw_layout_line" XG_POST, gxg_gdk_draw_layout_line, 5, 0, 0, H_gdk_draw_layout_line);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_draw_layout" XG_POST, gxg_gdk_draw_layout, 5, 0, 0, H_gdk_draw_layout);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_draw_layout_line_with_colors" XG_POST, gxg_gdk_draw_layout_line_with_colors, 7, 0, 0, H_gdk_draw_layout_line_with_colors);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_draw_layout_with_colors" XG_POST, gxg_gdk_draw_layout_with_colors, 7, 0, 0, H_gdk_draw_layout_with_colors);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_drawable_get_image" XG_POST, gxg_gdk_drawable_get_image, 5, 0, 0, H_gdk_drawable_get_image);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_drawable_get_clip_region" XG_POST, gxg_gdk_drawable_get_clip_region, 1, 0, 0, H_gdk_drawable_get_clip_region);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_drawable_get_visible_region" XG_POST, gxg_gdk_drawable_get_visible_region, 1, 0, 0, H_gdk_drawable_get_visible_region);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_cursor_type_get_type" XG_POST, gxg_gdk_cursor_type_get_type, 0, 0, 0, H_gdk_cursor_type_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_drag_action_get_type" XG_POST, gxg_gdk_drag_action_get_type, 0, 0, 0, H_gdk_drag_action_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_drag_protocol_get_type" XG_POST, gxg_gdk_drag_protocol_get_type, 0, 0, 0, H_gdk_drag_protocol_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_filter_return_get_type" XG_POST, gxg_gdk_filter_return_get_type, 0, 0, 0, H_gdk_filter_return_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_event_type_get_type" XG_POST, gxg_gdk_event_type_get_type, 0, 0, 0, H_gdk_event_type_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_event_mask_get_type" XG_POST, gxg_gdk_event_mask_get_type, 0, 0, 0, H_gdk_event_mask_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_visibility_state_get_type" XG_POST, gxg_gdk_visibility_state_get_type, 0, 0, 0, H_gdk_visibility_state_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_scroll_direction_get_type" XG_POST, gxg_gdk_scroll_direction_get_type, 0, 0, 0, H_gdk_scroll_direction_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_notify_type_get_type" XG_POST, gxg_gdk_notify_type_get_type, 0, 0, 0, H_gdk_notify_type_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_crossing_mode_get_type" XG_POST, gxg_gdk_crossing_mode_get_type, 0, 0, 0, H_gdk_crossing_mode_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_property_state_get_type" XG_POST, gxg_gdk_property_state_get_type, 0, 0, 0, H_gdk_property_state_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_window_state_get_type" XG_POST, gxg_gdk_window_state_get_type, 0, 0, 0, H_gdk_window_state_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_setting_action_get_type" XG_POST, gxg_gdk_setting_action_get_type, 0, 0, 0, H_gdk_setting_action_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_font_type_get_type" XG_POST, gxg_gdk_font_type_get_type, 0, 0, 0, H_gdk_font_type_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_cap_style_get_type" XG_POST, gxg_gdk_cap_style_get_type, 0, 0, 0, H_gdk_cap_style_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_fill_get_type" XG_POST, gxg_gdk_fill_get_type, 0, 0, 0, H_gdk_fill_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_function_get_type" XG_POST, gxg_gdk_function_get_type, 0, 0, 0, H_gdk_function_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_join_style_get_type" XG_POST, gxg_gdk_join_style_get_type, 0, 0, 0, H_gdk_join_style_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_line_style_get_type" XG_POST, gxg_gdk_line_style_get_type, 0, 0, 0, H_gdk_line_style_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_subwindow_mode_get_type" XG_POST, gxg_gdk_subwindow_mode_get_type, 0, 0, 0, H_gdk_subwindow_mode_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_gc_values_mask_get_type" XG_POST, gxg_gdk_gc_values_mask_get_type, 0, 0, 0, H_gdk_gc_values_mask_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_image_type_get_type" XG_POST, gxg_gdk_image_type_get_type, 0, 0, 0, H_gdk_image_type_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_extension_mode_get_type" XG_POST, gxg_gdk_extension_mode_get_type, 0, 0, 0, H_gdk_extension_mode_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_input_source_get_type" XG_POST, gxg_gdk_input_source_get_type, 0, 0, 0, H_gdk_input_source_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_input_mode_get_type" XG_POST, gxg_gdk_input_mode_get_type, 0, 0, 0, H_gdk_input_mode_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_axis_use_get_type" XG_POST, gxg_gdk_axis_use_get_type, 0, 0, 0, H_gdk_axis_use_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_prop_mode_get_type" XG_POST, gxg_gdk_prop_mode_get_type, 0, 0, 0, H_gdk_prop_mode_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_fill_rule_get_type" XG_POST, gxg_gdk_fill_rule_get_type, 0, 0, 0, H_gdk_fill_rule_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_overlap_type_get_type" XG_POST, gxg_gdk_overlap_type_get_type, 0, 0, 0, H_gdk_overlap_type_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_rgb_dither_get_type" XG_POST, gxg_gdk_rgb_dither_get_type, 0, 0, 0, H_gdk_rgb_dither_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_byte_order_get_type" XG_POST, gxg_gdk_byte_order_get_type, 0, 0, 0, H_gdk_byte_order_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_modifier_type_get_type" XG_POST, gxg_gdk_modifier_type_get_type, 0, 0, 0, H_gdk_modifier_type_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_input_condition_get_type" XG_POST, gxg_gdk_input_condition_get_type, 0, 0, 0, H_gdk_input_condition_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_status_get_type" XG_POST, gxg_gdk_status_get_type, 0, 0, 0, H_gdk_status_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_grab_status_get_type" XG_POST, gxg_gdk_grab_status_get_type, 0, 0, 0, H_gdk_grab_status_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_visual_type_get_type" XG_POST, gxg_gdk_visual_type_get_type, 0, 0, 0, H_gdk_visual_type_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_window_class_get_type" XG_POST, gxg_gdk_window_class_get_type, 0, 0, 0, H_gdk_window_class_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_window_type_get_type" XG_POST, gxg_gdk_window_type_get_type, 0, 0, 0, H_gdk_window_type_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_window_attributes_type_get_type" XG_POST, gxg_gdk_window_attributes_type_get_type, 0, 0, 0, H_gdk_window_attributes_type_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_window_hints_get_type" XG_POST, gxg_gdk_window_hints_get_type, 0, 0, 0, H_gdk_window_hints_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_window_type_hint_get_type" XG_POST, gxg_gdk_window_type_hint_get_type, 0, 0, 0, H_gdk_window_type_hint_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_wm_decoration_get_type" XG_POST, gxg_gdk_wm_decoration_get_type, 0, 0, 0, H_gdk_wm_decoration_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_wm_function_get_type" XG_POST, gxg_gdk_wm_function_get_type, 0, 0, 0, H_gdk_wm_function_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_gravity_get_type" XG_POST, gxg_gdk_gravity_get_type, 0, 0, 0, H_gdk_gravity_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_window_edge_get_type" XG_POST, gxg_gdk_window_edge_get_type, 0, 0, 0, H_gdk_window_edge_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_event_get_type" XG_POST, gxg_gdk_event_get_type, 0, 0, 0, H_gdk_event_get_type);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_events_pending" XG_POST, gxg_gdk_events_pending, 0, 0, 0, H_gdk_events_pending);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_event_get" XG_POST, gxg_gdk_event_get, 0, 0, 0, H_gdk_event_get);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_event_peek" XG_POST, gxg_gdk_event_peek, 0, 0, 0, H_gdk_event_peek);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_event_get_graphics_expose" XG_POST, gxg_gdk_event_get_graphics_expose, 1, 0, 0, H_gdk_event_get_graphics_expose);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_event_put" XG_POST, gxg_gdk_event_put, 1, 0, 0, H_gdk_event_put);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_event_copy" XG_POST, gxg_gdk_event_copy, 1, 0, 0, H_gdk_event_copy);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_event_free" XG_POST, gxg_gdk_event_free, 1, 0, 0, H_gdk_event_free);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_event_get_time" XG_POST, gxg_gdk_event_get_time, 1, 0, 0, H_gdk_event_get_time);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_event_get_state" XG_POST, gxg_gdk_event_get_state, 2, 0, 0, H_gdk_event_get_state);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_event_get_coords" XG_POST, gxg_gdk_event_get_coords, 1, 2, 0, H_gdk_event_get_coords);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_event_get_root_coords" XG_POST, gxg_gdk_event_get_root_coords, 3, 0, 0, H_gdk_event_get_root_coords);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_event_get_axis" XG_POST, gxg_gdk_event_get_axis, 2, 1, 0, H_gdk_event_get_axis);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_event_handler_set" XG_POST, gxg_gdk_event_handler_set, 3, 0, 0, H_gdk_event_handler_set);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_set_show_events" XG_POST, gxg_gdk_set_show_events, 1, 0, 0, H_gdk_set_show_events);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_get_show_events" XG_POST, gxg_gdk_get_show_events, 0, 0, 0, H_gdk_get_show_events);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_add_client_message_filter" XG_POST, gxg_gdk_add_client_message_filter, 2, 1, 0, H_gdk_add_client_message_filter);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_setting_get" XG_POST, gxg_gdk_setting_get, 2, 0, 0, H_gdk_setting_get);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_gc_get_type" XG_POST, gxg_gdk_gc_get_type, 0, 0, 0, H_gdk_gc_get_type);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_gc_new" XG_POST, gxg_gdk_gc_new, 1, 0, 0, H_gdk_gc_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_gc_new_with_values" XG_POST, gxg_gdk_gc_new_with_values, 3, 0, 0, H_gdk_gc_new_with_values);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_gc_get_values" XG_POST, gxg_gdk_gc_get_values, 2, 0, 0, H_gdk_gc_get_values);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_gc_set_values" XG_POST, gxg_gdk_gc_set_values, 3, 0, 0, H_gdk_gc_set_values);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_gc_set_foreground" XG_POST, gxg_gdk_gc_set_foreground, 2, 0, 0, H_gdk_gc_set_foreground);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_gc_set_background" XG_POST, gxg_gdk_gc_set_background, 2, 0, 0, H_gdk_gc_set_background);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_gc_set_function" XG_POST, gxg_gdk_gc_set_function, 2, 0, 0, H_gdk_gc_set_function);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_gc_set_fill" XG_POST, gxg_gdk_gc_set_fill, 2, 0, 0, H_gdk_gc_set_fill);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_gc_set_tile" XG_POST, gxg_gdk_gc_set_tile, 2, 0, 0, H_gdk_gc_set_tile);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_gc_set_stipple" XG_POST, gxg_gdk_gc_set_stipple, 2, 0, 0, H_gdk_gc_set_stipple);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_gc_set_ts_origin" XG_POST, gxg_gdk_gc_set_ts_origin, 3, 0, 0, H_gdk_gc_set_ts_origin);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_gc_set_clip_origin" XG_POST, gxg_gdk_gc_set_clip_origin, 3, 0, 0, H_gdk_gc_set_clip_origin);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_gc_set_clip_mask" XG_POST, gxg_gdk_gc_set_clip_mask, 2, 0, 0, H_gdk_gc_set_clip_mask);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_gc_set_clip_rectangle" XG_POST, gxg_gdk_gc_set_clip_rectangle, 2, 0, 0, H_gdk_gc_set_clip_rectangle);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_gc_set_clip_region" XG_POST, gxg_gdk_gc_set_clip_region, 2, 0, 0, H_gdk_gc_set_clip_region);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_gc_set_subwindow" XG_POST, gxg_gdk_gc_set_subwindow, 2, 0, 0, H_gdk_gc_set_subwindow);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_gc_set_exposures" XG_POST, gxg_gdk_gc_set_exposures, 2, 0, 0, H_gdk_gc_set_exposures);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_gc_set_line_attributes" XG_POST, gxg_gdk_gc_set_line_attributes, 5, 0, 0, H_gdk_gc_set_line_attributes);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_gc_set_dashes" XG_POST, gxg_gdk_gc_set_dashes, 4, 0, 0, H_gdk_gc_set_dashes);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_gc_offset" XG_POST, gxg_gdk_gc_offset, 3, 0, 0, H_gdk_gc_offset);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_gc_copy" XG_POST, gxg_gdk_gc_copy, 2, 0, 0, H_gdk_gc_copy);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_gc_set_colormap" XG_POST, gxg_gdk_gc_set_colormap, 2, 0, 0, H_gdk_gc_set_colormap);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_gc_get_colormap" XG_POST, gxg_gdk_gc_get_colormap, 1, 0, 0, H_gdk_gc_get_colormap);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_gc_set_rgb_fg_color" XG_POST, gxg_gdk_gc_set_rgb_fg_color, 2, 0, 0, H_gdk_gc_set_rgb_fg_color);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_gc_set_rgb_bg_color" XG_POST, gxg_gdk_gc_set_rgb_bg_color, 2, 0, 0, H_gdk_gc_set_rgb_bg_color);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_init" XG_POST, gxg_gdk_init, 2, 0, 0, H_gdk_init);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_init_check" XG_POST, gxg_gdk_init_check, 2, 0, 0, H_gdk_init_check);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_set_locale" XG_POST, gxg_gdk_set_locale, 0, 0, 0, H_gdk_set_locale);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_get_program_class" XG_POST, gxg_gdk_get_program_class, 0, 0, 0, H_gdk_get_program_class);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_set_program_class" XG_POST, gxg_gdk_set_program_class, 1, 0, 0, H_gdk_set_program_class);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_error_trap_push" XG_POST, gxg_gdk_error_trap_push, 0, 0, 0, H_gdk_error_trap_push);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_error_trap_pop" XG_POST, gxg_gdk_error_trap_pop, 0, 0, 0, H_gdk_error_trap_pop);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_get_display" XG_POST, gxg_gdk_get_display, 0, 0, 0, H_gdk_get_display);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_pointer_grab" XG_POST, gxg_gdk_pointer_grab, 6, 0, 0, H_gdk_pointer_grab);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_pointer_ungrab" XG_POST, gxg_gdk_pointer_ungrab, 1, 0, 0, H_gdk_pointer_ungrab);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_keyboard_grab" XG_POST, gxg_gdk_keyboard_grab, 3, 0, 0, H_gdk_keyboard_grab);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_keyboard_ungrab" XG_POST, gxg_gdk_keyboard_ungrab, 1, 0, 0, H_gdk_keyboard_ungrab);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_pointer_is_grabbed" XG_POST, gxg_gdk_pointer_is_grabbed, 0, 0, 0, H_gdk_pointer_is_grabbed);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_screen_width" XG_POST, gxg_gdk_screen_width, 0, 0, 0, H_gdk_screen_width);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_screen_height" XG_POST, gxg_gdk_screen_height, 0, 0, 0, H_gdk_screen_height);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_screen_width_mm" XG_POST, gxg_gdk_screen_width_mm, 0, 0, 0, H_gdk_screen_width_mm);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_screen_height_mm" XG_POST, gxg_gdk_screen_height_mm, 0, 0, 0, H_gdk_screen_height_mm);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_flush" XG_POST, gxg_gdk_flush, 0, 0, 0, H_gdk_flush);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_beep" XG_POST, gxg_gdk_beep, 0, 0, 0, H_gdk_beep);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_set_double_click_time" XG_POST, gxg_gdk_set_double_click_time, 1, 0, 0, H_gdk_set_double_click_time);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_rectangle_intersect" XG_POST, gxg_gdk_rectangle_intersect, 3, 0, 0, H_gdk_rectangle_intersect);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_rectangle_union" XG_POST, gxg_gdk_rectangle_union, 3, 0, 0, H_gdk_rectangle_union);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_rectangle_get_type" XG_POST, gxg_gdk_rectangle_get_type, 0, 0, 0, H_gdk_rectangle_get_type);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_wcstombs" XG_POST, gxg_gdk_wcstombs, 1, 0, 0, H_gdk_wcstombs);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_mbstowcs" XG_POST, gxg_gdk_mbstowcs, 3, 0, 0, H_gdk_mbstowcs);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_event_send_clientmessage_toall" XG_POST, gxg_gdk_event_send_clientmessage_toall, 1, 0, 0, H_gdk_event_send_clientmessage_toall);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_event_send_client_message" XG_POST, gxg_gdk_event_send_client_message, 2, 0, 0, H_gdk_event_send_client_message);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_threads_enter" XG_POST, gxg_gdk_threads_enter, 0, 0, 0, H_gdk_threads_enter);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_threads_leave" XG_POST, gxg_gdk_threads_leave, 0, 0, 0, H_gdk_threads_leave);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_threads_init" XG_POST, gxg_gdk_threads_init, 0, 0, 0, H_gdk_threads_init);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_image_get_type" XG_POST, gxg_gdk_image_get_type, 0, 0, 0, H_gdk_image_get_type);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_image_new" XG_POST, gxg_gdk_image_new, 4, 0, 0, H_gdk_image_new);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_device_get_type" XG_POST, gxg_gdk_device_get_type, 0, 0, 0, H_gdk_device_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_devices_list" XG_POST, gxg_gdk_devices_list, 0, 0, 0, H_gdk_devices_list);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_device_set_source" XG_POST, gxg_gdk_device_set_source, 2, 0, 0, H_gdk_device_set_source);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_device_set_mode" XG_POST, gxg_gdk_device_set_mode, 2, 0, 0, H_gdk_device_set_mode);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_device_set_key" XG_POST, gxg_gdk_device_set_key, 4, 0, 0, H_gdk_device_set_key);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_device_set_axis_use" XG_POST, gxg_gdk_device_set_axis_use, 3, 0, 0, H_gdk_device_set_axis_use);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_device_get_state" XG_POST, gxg_gdk_device_get_state, 4, 0, 0, H_gdk_device_get_state);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_device_get_history" XG_POST, gxg_gdk_device_get_history, 4, 2, 0, H_gdk_device_get_history);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_device_free_history" XG_POST, gxg_gdk_device_free_history, 2, 0, 0, H_gdk_device_free_history);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_device_get_axis" XG_POST, gxg_gdk_device_get_axis, 4, 0, 0, H_gdk_device_get_axis);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_input_set_extension_events" XG_POST, gxg_gdk_input_set_extension_events, 3, 0, 0, H_gdk_input_set_extension_events);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_device_get_core_pointer" XG_POST, gxg_gdk_device_get_core_pointer, 0, 0, 0, H_gdk_device_get_core_pointer);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_keymap_get_type" XG_POST, gxg_gdk_keymap_get_type, 0, 0, 0, H_gdk_keymap_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_keymap_get_default" XG_POST, gxg_gdk_keymap_get_default, 0, 0, 0, H_gdk_keymap_get_default);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_keymap_lookup_key" XG_POST, gxg_gdk_keymap_lookup_key, 2, 0, 0, H_gdk_keymap_lookup_key);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_keymap_get_entries_for_keyval" XG_POST, gxg_gdk_keymap_get_entries_for_keyval, 2, 2, 0, H_gdk_keymap_get_entries_for_keyval);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_keymap_get_entries_for_keycode" XG_POST, gxg_gdk_keymap_get_entries_for_keycode, 2, 3, 0, H_gdk_keymap_get_entries_for_keycode);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_keymap_get_direction" XG_POST, gxg_gdk_keymap_get_direction, 1, 0, 0, H_gdk_keymap_get_direction);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_keyval_name" XG_POST, gxg_gdk_keyval_name, 1, 0, 0, H_gdk_keyval_name);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_keyval_from_name" XG_POST, gxg_gdk_keyval_from_name, 1, 0, 0, H_gdk_keyval_from_name);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_keyval_convert_case" XG_POST, gxg_gdk_keyval_convert_case, 1, 2, 0, H_gdk_keyval_convert_case);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_keyval_to_upper" XG_POST, gxg_gdk_keyval_to_upper, 1, 0, 0, H_gdk_keyval_to_upper);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_keyval_to_lower" XG_POST, gxg_gdk_keyval_to_lower, 1, 0, 0, H_gdk_keyval_to_lower);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_keyval_is_upper" XG_POST, gxg_gdk_keyval_is_upper, 1, 0, 0, H_gdk_keyval_is_upper);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_keyval_is_lower" XG_POST, gxg_gdk_keyval_is_lower, 1, 0, 0, H_gdk_keyval_is_lower);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_keyval_to_unicode" XG_POST, gxg_gdk_keyval_to_unicode, 1, 0, 0, H_gdk_keyval_to_unicode);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_unicode_to_keyval" XG_POST, gxg_gdk_unicode_to_keyval, 1, 0, 0, H_gdk_unicode_to_keyval);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_pango_context_get" XG_POST, gxg_gdk_pango_context_get, 0, 0, 0, H_gdk_pango_context_get);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_pango_context_set_colormap" XG_POST, gxg_gdk_pango_context_set_colormap, 2, 0, 0, H_gdk_pango_context_set_colormap);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_pango_layout_line_get_clip_region" XG_POST, gxg_gdk_pango_layout_line_get_clip_region, 5, 0, 0, H_gdk_pango_layout_line_get_clip_region);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_pango_layout_get_clip_region" XG_POST, gxg_gdk_pango_layout_get_clip_region, 5, 0, 0, H_gdk_pango_layout_get_clip_region);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_pango_attr_stipple_new" XG_POST, gxg_gdk_pango_attr_stipple_new, 1, 0, 0, H_gdk_pango_attr_stipple_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_pango_attr_embossed_new" XG_POST, gxg_gdk_pango_attr_embossed_new, 1, 0, 0, H_gdk_pango_attr_embossed_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_pixbuf_render_threshold_alpha" XG_POST, gxg_gdk_pixbuf_render_threshold_alpha, 9, 0, 0, H_gdk_pixbuf_render_threshold_alpha);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_pixbuf_render_to_drawable" XG_POST, gxg_gdk_pixbuf_render_to_drawable, 0, 0, 1, H_gdk_pixbuf_render_to_drawable);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_pixbuf_render_to_drawable_alpha" XG_POST, gxg_gdk_pixbuf_render_to_drawable_alpha, 0, 0, 1, H_gdk_pixbuf_render_to_drawable_alpha);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_pixbuf_render_pixmap_and_mask_for_colormap" XG_POST, gxg_gdk_pixbuf_render_pixmap_and_mask_for_colormap, 3, 2, 0, H_gdk_pixbuf_render_pixmap_and_mask_for_colormap);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_pixbuf_render_pixmap_and_mask" XG_POST, gxg_gdk_pixbuf_render_pixmap_and_mask, 2, 2, 0, H_gdk_pixbuf_render_pixmap_and_mask);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_pixbuf_get_from_drawable" XG_POST, gxg_gdk_pixbuf_get_from_drawable, 9, 0, 0, H_gdk_pixbuf_get_from_drawable);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_pixbuf_get_from_image" XG_POST, gxg_gdk_pixbuf_get_from_image, 9, 0, 0, H_gdk_pixbuf_get_from_image);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_pixmap_get_type" XG_POST, gxg_gdk_pixmap_get_type, 0, 0, 0, H_gdk_pixmap_get_type);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_pixmap_new" XG_POST, gxg_gdk_pixmap_new, 4, 0, 0, H_gdk_pixmap_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_bitmap_create_from_data" XG_POST, gxg_gdk_bitmap_create_from_data, 4, 0, 0, H_gdk_bitmap_create_from_data);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_pixmap_create_from_data" XG_POST, gxg_gdk_pixmap_create_from_data, 7, 0, 0, H_gdk_pixmap_create_from_data);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_pixmap_create_from_xpm" XG_POST, gxg_gdk_pixmap_create_from_xpm, 4, 0, 0, H_gdk_pixmap_create_from_xpm);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_pixmap_colormap_create_from_xpm" XG_POST, gxg_gdk_pixmap_colormap_create_from_xpm, 5, 0, 0, H_gdk_pixmap_colormap_create_from_xpm);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_pixmap_create_from_xpm_d" XG_POST, gxg_gdk_pixmap_create_from_xpm_d, 4, 0, 0, H_gdk_pixmap_create_from_xpm_d);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_pixmap_colormap_create_from_xpm_d" XG_POST, gxg_gdk_pixmap_colormap_create_from_xpm_d, 5, 0, 0, H_gdk_pixmap_colormap_create_from_xpm_d);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_pixmap_foreign_new" XG_POST, gxg_gdk_pixmap_foreign_new, 1, 0, 0, H_gdk_pixmap_foreign_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_pixmap_lookup" XG_POST, gxg_gdk_pixmap_lookup, 1, 0, 0, H_gdk_pixmap_lookup);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_atom_intern" XG_POST, gxg_gdk_atom_intern, 2, 0, 0, H_gdk_atom_intern);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_atom_name" XG_POST, gxg_gdk_atom_name, 1, 0, 0, H_gdk_atom_name);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_property_get" XG_POST, gxg_gdk_property_get, 0, 0, 1, H_gdk_property_get);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_property_change" XG_POST, gxg_gdk_property_change, 7, 0, 0, H_gdk_property_change);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_property_delete" XG_POST, gxg_gdk_property_delete, 2, 0, 0, H_gdk_property_delete);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_text_property_to_text_list" XG_POST, gxg_gdk_text_property_to_text_list, 5, 0, 0, H_gdk_text_property_to_text_list);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_text_property_to_utf8_list" XG_POST, gxg_gdk_text_property_to_utf8_list, 5, 0, 0, H_gdk_text_property_to_utf8_list);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_utf8_to_string_target" XG_POST, gxg_gdk_utf8_to_string_target, 1, 0, 0, H_gdk_utf8_to_string_target);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_utf8_to_compound_text" XG_POST, gxg_gdk_utf8_to_compound_text, 3, 2, 0, H_gdk_utf8_to_compound_text);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_free_text_list" XG_POST, gxg_gdk_free_text_list, 1, 0, 0, H_gdk_free_text_list);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_string_to_compound_text" XG_POST, gxg_gdk_string_to_compound_text, 3, 2, 0, H_gdk_string_to_compound_text);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_free_compound_text" XG_POST, gxg_gdk_free_compound_text, 1, 0, 0, H_gdk_free_compound_text);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_region_new" XG_POST, gxg_gdk_region_new, 0, 0, 0, H_gdk_region_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_region_polygon" XG_POST, gxg_gdk_region_polygon, 3, 0, 0, H_gdk_region_polygon);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_region_copy" XG_POST, gxg_gdk_region_copy, 1, 0, 0, H_gdk_region_copy);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_region_rectangle" XG_POST, gxg_gdk_region_rectangle, 1, 0, 0, H_gdk_region_rectangle);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_region_destroy" XG_POST, gxg_gdk_region_destroy, 1, 0, 0, H_gdk_region_destroy);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_region_get_clipbox" XG_POST, gxg_gdk_region_get_clipbox, 2, 0, 0, H_gdk_region_get_clipbox);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_region_get_rectangles" XG_POST, gxg_gdk_region_get_rectangles, 1, 2, 0, H_gdk_region_get_rectangles);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_region_empty" XG_POST, gxg_gdk_region_empty, 1, 0, 0, H_gdk_region_empty);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_region_equal" XG_POST, gxg_gdk_region_equal, 2, 0, 0, H_gdk_region_equal);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_region_point_in" XG_POST, gxg_gdk_region_point_in, 3, 0, 0, H_gdk_region_point_in);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_region_rect_in" XG_POST, gxg_gdk_region_rect_in, 2, 0, 0, H_gdk_region_rect_in);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_region_offset" XG_POST, gxg_gdk_region_offset, 3, 0, 0, H_gdk_region_offset);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_region_shrink" XG_POST, gxg_gdk_region_shrink, 3, 0, 0, H_gdk_region_shrink);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_region_union_with_rect" XG_POST, gxg_gdk_region_union_with_rect, 2, 0, 0, H_gdk_region_union_with_rect);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_region_intersect" XG_POST, gxg_gdk_region_intersect, 2, 0, 0, H_gdk_region_intersect);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_region_union" XG_POST, gxg_gdk_region_union, 2, 0, 0, H_gdk_region_union);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_region_subtract" XG_POST, gxg_gdk_region_subtract, 2, 0, 0, H_gdk_region_subtract);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_region_xor" XG_POST, gxg_gdk_region_xor, 2, 0, 0, H_gdk_region_xor);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_region_spans_intersect_foreach" XG_POST, gxg_gdk_region_spans_intersect_foreach, 5, 1, 0, H_gdk_region_spans_intersect_foreach);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_rgb_find_color" XG_POST, gxg_gdk_rgb_find_color, 2, 0, 0, H_gdk_rgb_find_color);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_draw_rgb_image" XG_POST, gxg_gdk_draw_rgb_image, 9, 0, 0, H_gdk_draw_rgb_image);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_draw_rgb_image_dithalign" XG_POST, gxg_gdk_draw_rgb_image_dithalign, 0, 0, 1, H_gdk_draw_rgb_image_dithalign);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_draw_rgb_32_image" XG_POST, gxg_gdk_draw_rgb_32_image, 9, 0, 0, H_gdk_draw_rgb_32_image);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_draw_rgb_32_image_dithalign" XG_POST, gxg_gdk_draw_rgb_32_image_dithalign, 0, 0, 1, H_gdk_draw_rgb_32_image_dithalign);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_draw_gray_image" XG_POST, gxg_gdk_draw_gray_image, 9, 0, 0, H_gdk_draw_gray_image);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_draw_indexed_image" XG_POST, gxg_gdk_draw_indexed_image, 0, 0, 1, H_gdk_draw_indexed_image);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_rgb_cmap_new" XG_POST, gxg_gdk_rgb_cmap_new, 2, 0, 0, H_gdk_rgb_cmap_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_rgb_cmap_free" XG_POST, gxg_gdk_rgb_cmap_free, 1, 0, 0, H_gdk_rgb_cmap_free);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_rgb_ditherable" XG_POST, gxg_gdk_rgb_ditherable, 0, 0, 0, H_gdk_rgb_ditherable);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_rgb_set_verbose" XG_POST, gxg_gdk_rgb_set_verbose, 1, 0, 0, H_gdk_rgb_set_verbose);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_rgb_set_install" XG_POST, gxg_gdk_rgb_set_install, 1, 0, 0, H_gdk_rgb_set_install);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_rgb_set_min_colors" XG_POST, gxg_gdk_rgb_set_min_colors, 1, 0, 0, H_gdk_rgb_set_min_colors);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_rgb_get_colormap" XG_POST, gxg_gdk_rgb_get_colormap, 0, 0, 0, H_gdk_rgb_get_colormap);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_rgb_get_visual" XG_POST, gxg_gdk_rgb_get_visual, 0, 0, 0, H_gdk_rgb_get_visual);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_selection_owner_set" XG_POST, gxg_gdk_selection_owner_set, 4, 0, 0, H_gdk_selection_owner_set);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_selection_owner_get" XG_POST, gxg_gdk_selection_owner_get, 1, 0, 0, H_gdk_selection_owner_get);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_selection_convert" XG_POST, gxg_gdk_selection_convert, 4, 0, 0, H_gdk_selection_convert);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_selection_property_get" XG_POST, gxg_gdk_selection_property_get, 1, 3, 0, H_gdk_selection_property_get);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_selection_send_notify" XG_POST, gxg_gdk_selection_send_notify, 5, 0, 0, H_gdk_selection_send_notify);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_visual_get_type" XG_POST, gxg_gdk_visual_get_type, 0, 0, 0, H_gdk_visual_get_type);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_visual_get_best_depth" XG_POST, gxg_gdk_visual_get_best_depth, 0, 0, 0, H_gdk_visual_get_best_depth);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_visual_get_best_type" XG_POST, gxg_gdk_visual_get_best_type, 0, 0, 0, H_gdk_visual_get_best_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_visual_get_system" XG_POST, gxg_gdk_visual_get_system, 0, 0, 0, H_gdk_visual_get_system);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_visual_get_best" XG_POST, gxg_gdk_visual_get_best, 0, 0, 0, H_gdk_visual_get_best);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_visual_get_best_with_depth" XG_POST, gxg_gdk_visual_get_best_with_depth, 1, 0, 0, H_gdk_visual_get_best_with_depth);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_visual_get_best_with_type" XG_POST, gxg_gdk_visual_get_best_with_type, 1, 0, 0, H_gdk_visual_get_best_with_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_visual_get_best_with_both" XG_POST, gxg_gdk_visual_get_best_with_both, 2, 0, 0, H_gdk_visual_get_best_with_both);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_query_depths" XG_POST, gxg_gdk_query_depths, 0, 2, 0, H_gdk_query_depths);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_query_visual_types" XG_POST, gxg_gdk_query_visual_types, 0, 2, 0, H_gdk_query_visual_types);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_list_visuals" XG_POST, gxg_gdk_list_visuals, 0, 0, 0, H_gdk_list_visuals);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_window_object_get_type" XG_POST, gxg_gdk_window_object_get_type, 0, 0, 0, H_gdk_window_object_get_type);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_window_new" XG_POST, gxg_gdk_window_new, 3, 0, 0, H_gdk_window_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_window_destroy" XG_POST, gxg_gdk_window_destroy, 1, 0, 0, H_gdk_window_destroy);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_window_get_window_type" XG_POST, gxg_gdk_window_get_window_type, 1, 0, 0, H_gdk_window_get_window_type);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_window_at_pointer" XG_POST, gxg_gdk_window_at_pointer, 0, 2, 0, H_gdk_window_at_pointer);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_window_show" XG_POST, gxg_gdk_window_show, 1, 0, 0, H_gdk_window_show);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_window_hide" XG_POST, gxg_gdk_window_hide, 1, 0, 0, H_gdk_window_hide);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_window_withdraw" XG_POST, gxg_gdk_window_withdraw, 1, 0, 0, H_gdk_window_withdraw);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_window_show_unraised" XG_POST, gxg_gdk_window_show_unraised, 1, 0, 0, H_gdk_window_show_unraised);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_window_move" XG_POST, gxg_gdk_window_move, 3, 0, 0, H_gdk_window_move);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_window_resize" XG_POST, gxg_gdk_window_resize, 3, 0, 0, H_gdk_window_resize);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_window_move_resize" XG_POST, gxg_gdk_window_move_resize, 5, 0, 0, H_gdk_window_move_resize);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_window_reparent" XG_POST, gxg_gdk_window_reparent, 4, 0, 0, H_gdk_window_reparent);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_window_clear" XG_POST, gxg_gdk_window_clear, 1, 0, 0, H_gdk_window_clear);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_window_clear_area" XG_POST, gxg_gdk_window_clear_area, 5, 0, 0, H_gdk_window_clear_area);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_window_clear_area_e" XG_POST, gxg_gdk_window_clear_area_e, 5, 0, 0, H_gdk_window_clear_area_e);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_window_raise" XG_POST, gxg_gdk_window_raise, 1, 0, 0, H_gdk_window_raise);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_window_lower" XG_POST, gxg_gdk_window_lower, 1, 0, 0, H_gdk_window_lower);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_window_focus" XG_POST, gxg_gdk_window_focus, 2, 0, 0, H_gdk_window_focus);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_window_set_user_data" XG_POST, gxg_gdk_window_set_user_data, 2, 0, 0, H_gdk_window_set_user_data);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_window_set_override_redirect" XG_POST, gxg_gdk_window_set_override_redirect, 2, 0, 0, H_gdk_window_set_override_redirect);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_window_add_filter" XG_POST, gxg_gdk_window_add_filter, 2, 1, 0, H_gdk_window_add_filter);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_window_remove_filter" XG_POST, gxg_gdk_window_remove_filter, 2, 1, 0, H_gdk_window_remove_filter);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_window_scroll" XG_POST, gxg_gdk_window_scroll, 3, 0, 0, H_gdk_window_scroll);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_window_shape_combine_mask" XG_POST, gxg_gdk_window_shape_combine_mask, 4, 0, 0, H_gdk_window_shape_combine_mask);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_window_shape_combine_region" XG_POST, gxg_gdk_window_shape_combine_region, 4, 0, 0, H_gdk_window_shape_combine_region);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_window_set_child_shapes" XG_POST, gxg_gdk_window_set_child_shapes, 1, 0, 0, H_gdk_window_set_child_shapes);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_window_merge_child_shapes" XG_POST, gxg_gdk_window_merge_child_shapes, 1, 0, 0, H_gdk_window_merge_child_shapes);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_window_is_visible" XG_POST, gxg_gdk_window_is_visible, 1, 0, 0, H_gdk_window_is_visible);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_window_is_viewable" XG_POST, gxg_gdk_window_is_viewable, 1, 0, 0, H_gdk_window_is_viewable);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_window_get_state" XG_POST, gxg_gdk_window_get_state, 1, 0, 0, H_gdk_window_get_state);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_window_set_static_gravities" XG_POST, gxg_gdk_window_set_static_gravities, 2, 0, 0, H_gdk_window_set_static_gravities);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_window_foreign_new" XG_POST, gxg_gdk_window_foreign_new, 1, 0, 0, H_gdk_window_foreign_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_window_lookup" XG_POST, gxg_gdk_window_lookup, 1, 0, 0, H_gdk_window_lookup);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_window_get_root_origin" XG_POST, gxg_gdk_window_get_root_origin, 1, 2, 0, H_gdk_window_get_root_origin);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_window_get_frame_extents" XG_POST, gxg_gdk_window_get_frame_extents, 2, 0, 0, H_gdk_window_get_frame_extents);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_window_get_pointer" XG_POST, gxg_gdk_window_get_pointer, 1, 3, 0, H_gdk_window_get_pointer);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_window_get_parent" XG_POST, gxg_gdk_window_get_parent, 1, 0, 0, H_gdk_window_get_parent);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_window_get_toplevel" XG_POST, gxg_gdk_window_get_toplevel, 1, 0, 0, H_gdk_window_get_toplevel);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_window_get_children" XG_POST, gxg_gdk_window_get_children, 1, 0, 0, H_gdk_window_get_children);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_window_peek_children" XG_POST, gxg_gdk_window_peek_children, 1, 0, 0, H_gdk_window_peek_children);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_window_get_events" XG_POST, gxg_gdk_window_get_events, 1, 0, 0, H_gdk_window_get_events);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_window_set_events" XG_POST, gxg_gdk_window_set_events, 2, 0, 0, H_gdk_window_set_events);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_window_set_icon_list" XG_POST, gxg_gdk_window_set_icon_list, 2, 0, 0, H_gdk_window_set_icon_list);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_window_set_icon" XG_POST, gxg_gdk_window_set_icon, 4, 0, 0, H_gdk_window_set_icon);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_window_set_icon_name" XG_POST, gxg_gdk_window_set_icon_name, 2, 0, 0, H_gdk_window_set_icon_name);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_window_set_group" XG_POST, gxg_gdk_window_set_group, 2, 0, 0, H_gdk_window_set_group);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_window_set_decorations" XG_POST, gxg_gdk_window_set_decorations, 2, 0, 0, H_gdk_window_set_decorations);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_window_get_decorations" XG_POST, gxg_gdk_window_get_decorations, 2, 0, 0, H_gdk_window_get_decorations);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_window_set_functions" XG_POST, gxg_gdk_window_set_functions, 2, 0, 0, H_gdk_window_set_functions);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_window_get_toplevels" XG_POST, gxg_gdk_window_get_toplevels, 0, 0, 0, H_gdk_window_get_toplevels);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_window_iconify" XG_POST, gxg_gdk_window_iconify, 1, 0, 0, H_gdk_window_iconify);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_window_deiconify" XG_POST, gxg_gdk_window_deiconify, 1, 0, 0, H_gdk_window_deiconify);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_window_stick" XG_POST, gxg_gdk_window_stick, 1, 0, 0, H_gdk_window_stick);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_window_unstick" XG_POST, gxg_gdk_window_unstick, 1, 0, 0, H_gdk_window_unstick);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_window_maximize" XG_POST, gxg_gdk_window_maximize, 1, 0, 0, H_gdk_window_maximize);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_window_unmaximize" XG_POST, gxg_gdk_window_unmaximize, 1, 0, 0, H_gdk_window_unmaximize);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_window_register_dnd" XG_POST, gxg_gdk_window_register_dnd, 1, 0, 0, H_gdk_window_register_dnd);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_window_begin_resize_drag" XG_POST, gxg_gdk_window_begin_resize_drag, 6, 0, 0, H_gdk_window_begin_resize_drag);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_window_begin_move_drag" XG_POST, gxg_gdk_window_begin_move_drag, 5, 0, 0, H_gdk_window_begin_move_drag);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_window_invalidate_rect" XG_POST, gxg_gdk_window_invalidate_rect, 3, 0, 0, H_gdk_window_invalidate_rect);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_window_invalidate_region" XG_POST, gxg_gdk_window_invalidate_region, 3, 0, 0, H_gdk_window_invalidate_region);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_window_invalidate_maybe_recurse" XG_POST, gxg_gdk_window_invalidate_maybe_recurse, 3, 1, 0, H_gdk_window_invalidate_maybe_recurse);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_window_get_update_area" XG_POST, gxg_gdk_window_get_update_area, 1, 0, 0, H_gdk_window_get_update_area);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_window_freeze_updates" XG_POST, gxg_gdk_window_freeze_updates, 1, 0, 0, H_gdk_window_freeze_updates);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_window_thaw_updates" XG_POST, gxg_gdk_window_thaw_updates, 1, 0, 0, H_gdk_window_thaw_updates);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_window_process_all_updates" XG_POST, gxg_gdk_window_process_all_updates, 0, 0, 0, H_gdk_window_process_all_updates);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_window_process_updates" XG_POST, gxg_gdk_window_process_updates, 2, 0, 0, H_gdk_window_process_updates);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_window_set_debug_updates" XG_POST, gxg_gdk_window_set_debug_updates, 1, 0, 0, H_gdk_window_set_debug_updates);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_window_constrain_size" XG_POST, gxg_gdk_window_constrain_size, 4, 2, 0, H_gdk_window_constrain_size);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_window_get_internal_paint_info" XG_POST, gxg_gdk_window_get_internal_paint_info, 1, 3, 0, H_gdk_window_get_internal_paint_info);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_window_group_add_window" XG_POST, gxg_gtk_window_group_add_window, 2, 0, 0, H_gtk_window_group_add_window);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_window_group_remove_window" XG_POST, gxg_gtk_window_group_remove_window, 2, 0, 0, H_gtk_window_group_remove_window);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_window_remove_embedded_xid" XG_POST, gxg_gtk_window_remove_embedded_xid, 2, 0, 0, H_gtk_window_remove_embedded_xid);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_window_add_embedded_xid" XG_POST, gxg_gtk_window_add_embedded_xid, 2, 0, 0, H_gtk_window_add_embedded_xid);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_window_set_type_hint" XG_POST, gxg_gdk_window_set_type_hint, 2, 0, 0, H_gdk_window_set_type_hint);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_window_set_modal_hint" XG_POST, gxg_gdk_window_set_modal_hint, 2, 0, 0, H_gdk_window_set_modal_hint);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_window_set_geometry_hints" XG_POST, gxg_gdk_window_set_geometry_hints, 3, 0, 0, H_gdk_window_set_geometry_hints);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_set_sm_client_id" XG_POST, gxg_gdk_set_sm_client_id, 1, 0, 0, H_gdk_set_sm_client_id);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_window_begin_paint_rect" XG_POST, gxg_gdk_window_begin_paint_rect, 2, 0, 0, H_gdk_window_begin_paint_rect);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_window_begin_paint_region" XG_POST, gxg_gdk_window_begin_paint_region, 2, 0, 0, H_gdk_window_begin_paint_region);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_window_end_paint" XG_POST, gxg_gdk_window_end_paint, 1, 0, 0, H_gdk_window_end_paint);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_window_set_title" XG_POST, gxg_gdk_window_set_title, 2, 0, 0, H_gdk_window_set_title);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_window_set_role" XG_POST, gxg_gdk_window_set_role, 2, 0, 0, H_gdk_window_set_role);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_window_set_transient_for" XG_POST, gxg_gdk_window_set_transient_for, 2, 0, 0, H_gdk_window_set_transient_for);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_window_set_background" XG_POST, gxg_gdk_window_set_background, 2, 0, 0, H_gdk_window_set_background);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_window_set_back_pixmap" XG_POST, gxg_gdk_window_set_back_pixmap, 3, 0, 0, H_gdk_window_set_back_pixmap);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_window_set_cursor" XG_POST, gxg_gdk_window_set_cursor, 2, 0, 0, H_gdk_window_set_cursor);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_window_get_user_data" XG_POST, gxg_gdk_window_get_user_data, 1, 1, 0, H_gdk_window_get_user_data);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_window_get_geometry" XG_POST, gxg_gdk_window_get_geometry, 6, 0, 0, H_gdk_window_get_geometry);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_window_get_position" XG_POST, gxg_gdk_window_get_position, 1, 2, 0, H_gdk_window_get_position);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_window_get_origin" XG_POST, gxg_gdk_window_get_origin, 3, 0, 0, H_gdk_window_get_origin);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_set_pointer_hooks" XG_POST, gxg_gdk_set_pointer_hooks, 1, 0, 0, H_gdk_set_pointer_hooks);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_get_default_root_window" XG_POST, gxg_gdk_get_default_root_window, 0, 0, 0, H_gdk_get_default_root_window);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_pixbuf_error_quark" XG_POST, gxg_gdk_pixbuf_error_quark, 0, 0, 0, H_gdk_pixbuf_error_quark);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_pixbuf_get_type" XG_POST, gxg_gdk_pixbuf_get_type, 0, 0, 0, H_gdk_pixbuf_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_pixbuf_get_colorspace" XG_POST, gxg_gdk_pixbuf_get_colorspace, 1, 0, 0, H_gdk_pixbuf_get_colorspace);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_pixbuf_get_n_channels" XG_POST, gxg_gdk_pixbuf_get_n_channels, 1, 0, 0, H_gdk_pixbuf_get_n_channels);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_pixbuf_get_has_alpha" XG_POST, gxg_gdk_pixbuf_get_has_alpha, 1, 0, 0, H_gdk_pixbuf_get_has_alpha);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_pixbuf_get_bits_per_sample" XG_POST, gxg_gdk_pixbuf_get_bits_per_sample, 1, 0, 0, H_gdk_pixbuf_get_bits_per_sample);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_pixbuf_get_pixels" XG_POST, gxg_gdk_pixbuf_get_pixels, 1, 0, 0, H_gdk_pixbuf_get_pixels);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_pixbuf_get_width" XG_POST, gxg_gdk_pixbuf_get_width, 1, 0, 0, H_gdk_pixbuf_get_width);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_pixbuf_get_height" XG_POST, gxg_gdk_pixbuf_get_height, 1, 0, 0, H_gdk_pixbuf_get_height);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_pixbuf_get_rowstride" XG_POST, gxg_gdk_pixbuf_get_rowstride, 1, 0, 0, H_gdk_pixbuf_get_rowstride);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_pixbuf_new" XG_POST, gxg_gdk_pixbuf_new, 5, 0, 0, H_gdk_pixbuf_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_pixbuf_copy" XG_POST, gxg_gdk_pixbuf_copy, 1, 0, 0, H_gdk_pixbuf_copy);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_pixbuf_new_subpixbuf" XG_POST, gxg_gdk_pixbuf_new_subpixbuf, 5, 0, 0, H_gdk_pixbuf_new_subpixbuf);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_pixbuf_new_from_file" XG_POST, gxg_gdk_pixbuf_new_from_file, 1, 1, 0, H_gdk_pixbuf_new_from_file);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_pixbuf_new_from_data" XG_POST, gxg_gdk_pixbuf_new_from_data, 9, 0, 0, H_gdk_pixbuf_new_from_data);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_pixbuf_new_from_xpm_data" XG_POST, gxg_gdk_pixbuf_new_from_xpm_data, 1, 0, 0, H_gdk_pixbuf_new_from_xpm_data);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_pixbuf_new_from_inline" XG_POST, gxg_gdk_pixbuf_new_from_inline, 3, 1, 0, H_gdk_pixbuf_new_from_inline);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_pixbuf_fill" XG_POST, gxg_gdk_pixbuf_fill, 2, 0, 0, H_gdk_pixbuf_fill);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_pixbuf_save" XG_POST, gxg_gdk_pixbuf_save, 3, 1, 0, H_gdk_pixbuf_save);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_pixbuf_savev" XG_POST, gxg_gdk_pixbuf_savev, 5, 1, 0, H_gdk_pixbuf_savev);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_pixbuf_add_alpha" XG_POST, gxg_gdk_pixbuf_add_alpha, 5, 0, 0, H_gdk_pixbuf_add_alpha);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_pixbuf_copy_area" XG_POST, gxg_gdk_pixbuf_copy_area, 8, 0, 0, H_gdk_pixbuf_copy_area);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_pixbuf_saturate_and_pixelate" XG_POST, gxg_gdk_pixbuf_saturate_and_pixelate, 4, 0, 0, H_gdk_pixbuf_saturate_and_pixelate);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_pixbuf_scale" XG_POST, gxg_gdk_pixbuf_scale, 0, 0, 1, H_gdk_pixbuf_scale);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_pixbuf_composite" XG_POST, gxg_gdk_pixbuf_composite, 0, 0, 1, H_gdk_pixbuf_composite);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_pixbuf_composite_color" XG_POST, gxg_gdk_pixbuf_composite_color, 0, 0, 1, H_gdk_pixbuf_composite_color);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_pixbuf_scale_simple" XG_POST, gxg_gdk_pixbuf_scale_simple, 4, 0, 0, H_gdk_pixbuf_scale_simple);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_pixbuf_composite_color_simple" XG_POST, gxg_gdk_pixbuf_composite_color_simple, 8, 0, 0, H_gdk_pixbuf_composite_color_simple);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_pixbuf_animation_get_type" XG_POST, gxg_gdk_pixbuf_animation_get_type, 0, 0, 0, H_gdk_pixbuf_animation_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_pixbuf_animation_new_from_file" XG_POST, gxg_gdk_pixbuf_animation_new_from_file, 1, 1, 0, H_gdk_pixbuf_animation_new_from_file);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_pixbuf_animation_get_width" XG_POST, gxg_gdk_pixbuf_animation_get_width, 1, 0, 0, H_gdk_pixbuf_animation_get_width);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_pixbuf_animation_get_height" XG_POST, gxg_gdk_pixbuf_animation_get_height, 1, 0, 0, H_gdk_pixbuf_animation_get_height);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_pixbuf_animation_is_static_image" XG_POST, gxg_gdk_pixbuf_animation_is_static_image, 1, 0, 0, H_gdk_pixbuf_animation_is_static_image);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_pixbuf_animation_get_static_image" XG_POST, gxg_gdk_pixbuf_animation_get_static_image, 1, 0, 0, H_gdk_pixbuf_animation_get_static_image);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_pixbuf_animation_get_iter" XG_POST, gxg_gdk_pixbuf_animation_get_iter, 2, 0, 0, H_gdk_pixbuf_animation_get_iter);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_pixbuf_animation_iter_get_type" XG_POST, gxg_gdk_pixbuf_animation_iter_get_type, 0, 0, 0, H_gdk_pixbuf_animation_iter_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_pixbuf_animation_iter_get_delay_time" XG_POST, gxg_gdk_pixbuf_animation_iter_get_delay_time, 1, 0, 0, H_gdk_pixbuf_animation_iter_get_delay_time);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_pixbuf_animation_iter_get_pixbuf" XG_POST, gxg_gdk_pixbuf_animation_iter_get_pixbuf, 1, 0, 0, H_gdk_pixbuf_animation_iter_get_pixbuf);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_pixbuf_animation_iter_on_currently_loading_frame" XG_POST, gxg_gdk_pixbuf_animation_iter_on_currently_loading_frame, 1, 0, 0, H_gdk_pixbuf_animation_iter_on_currently_loading_frame);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_pixbuf_animation_iter_advance" XG_POST, gxg_gdk_pixbuf_animation_iter_advance, 2, 0, 0, H_gdk_pixbuf_animation_iter_advance);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_pixbuf_get_option" XG_POST, gxg_gdk_pixbuf_get_option, 2, 0, 0, H_gdk_pixbuf_get_option);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_pixbuf_loader_get_type" XG_POST, gxg_gdk_pixbuf_loader_get_type, 0, 0, 0, H_gdk_pixbuf_loader_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_pixbuf_loader_new" XG_POST, gxg_gdk_pixbuf_loader_new, 0, 0, 0, H_gdk_pixbuf_loader_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_pixbuf_loader_new_with_type" XG_POST, gxg_gdk_pixbuf_loader_new_with_type, 1, 1, 0, H_gdk_pixbuf_loader_new_with_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_pixbuf_loader_write" XG_POST, gxg_gdk_pixbuf_loader_write, 3, 1, 0, H_gdk_pixbuf_loader_write);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_pixbuf_loader_get_pixbuf" XG_POST, gxg_gdk_pixbuf_loader_get_pixbuf, 1, 0, 0, H_gdk_pixbuf_loader_get_pixbuf);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_pixbuf_loader_get_animation" XG_POST, gxg_gdk_pixbuf_loader_get_animation, 1, 0, 0, H_gdk_pixbuf_loader_get_animation);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_pixbuf_loader_close" XG_POST, gxg_gdk_pixbuf_loader_close, 1, 1, 0, H_gdk_pixbuf_loader_close);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_pixbuf_alpha_mode_get_type" XG_POST, gxg_gdk_pixbuf_alpha_mode_get_type, 0, 0, 0, H_gdk_pixbuf_alpha_mode_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_colorspace_get_type" XG_POST, gxg_gdk_colorspace_get_type, 0, 0, 0, H_gdk_colorspace_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_pixbuf_error_get_type" XG_POST, gxg_gdk_pixbuf_error_get_type, 0, 0, 0, H_gdk_pixbuf_error_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_interp_type_get_type" XG_POST, gxg_gdk_interp_type_get_type, 0, 0, 0, H_gdk_interp_type_get_type);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_vbox_get_type" XG_POST, gxg_gtk_vbox_get_type, 0, 0, 0, H_gtk_vbox_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_vbox_new" XG_POST, gxg_gtk_vbox_new, 2, 0, 0, H_gtk_vbox_new);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_accel_group_get_type" XG_POST, gxg_gtk_accel_group_get_type, 0, 0, 0, H_gtk_accel_group_get_type);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_accel_group_new" XG_POST, gxg_gtk_accel_group_new, 0, 0, 0, H_gtk_accel_group_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_accel_group_lock" XG_POST, gxg_gtk_accel_group_lock, 1, 0, 0, H_gtk_accel_group_lock);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_accel_group_unlock" XG_POST, gxg_gtk_accel_group_unlock, 1, 0, 0, H_gtk_accel_group_unlock);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_accel_group_connect" XG_POST, gxg_gtk_accel_group_connect, 5, 0, 0, H_gtk_accel_group_connect);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_accel_group_connect_by_path" XG_POST, gxg_gtk_accel_group_connect_by_path, 3, 0, 0, H_gtk_accel_group_connect_by_path);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_accel_group_disconnect" XG_POST, gxg_gtk_accel_group_disconnect, 2, 0, 0, H_gtk_accel_group_disconnect);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_accel_group_disconnect_key" XG_POST, gxg_gtk_accel_group_disconnect_key, 3, 0, 0, H_gtk_accel_group_disconnect_key);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_accel_groups_activate" XG_POST, gxg_gtk_accel_groups_activate, 3, 0, 0, H_gtk_accel_groups_activate);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_accel_groups_from_object" XG_POST, gxg_gtk_accel_groups_from_object, 1, 0, 0, H_gtk_accel_groups_from_object);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_accel_group_find" XG_POST, gxg_gtk_accel_group_find, 2, 1, 0, H_gtk_accel_group_find);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_accel_group_from_accel_closure" XG_POST, gxg_gtk_accel_group_from_accel_closure, 1, 0, 0, H_gtk_accel_group_from_accel_closure);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_accelerator_valid" XG_POST, gxg_gtk_accelerator_valid, 2, 0, 0, H_gtk_accelerator_valid);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_accelerator_parse" XG_POST, gxg_gtk_accelerator_parse, 1, 2, 0, H_gtk_accelerator_parse);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_accelerator_name" XG_POST, gxg_gtk_accelerator_name, 2, 0, 0, H_gtk_accelerator_name);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_accelerator_set_default_mod_mask" XG_POST, gxg_gtk_accelerator_set_default_mod_mask, 1, 0, 0, H_gtk_accelerator_set_default_mod_mask);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_accelerator_get_default_mod_mask" XG_POST, gxg_gtk_accelerator_get_default_mod_mask, 0, 0, 0, H_gtk_accelerator_get_default_mod_mask);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_accel_group_query" XG_POST, gxg_gtk_accel_group_query, 3, 1, 0, H_gtk_accel_group_query);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_accel_label_get_type" XG_POST, gxg_gtk_accel_label_get_type, 0, 0, 0, H_gtk_accel_label_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_accel_label_new" XG_POST, gxg_gtk_accel_label_new, 1, 0, 0, H_gtk_accel_label_new);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_accel_label_get_accel_widget" XG_POST, gxg_gtk_accel_label_get_accel_widget, 1, 0, 0, H_gtk_accel_label_get_accel_widget);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_accel_label_get_accel_width" XG_POST, gxg_gtk_accel_label_get_accel_width, 1, 0, 0, H_gtk_accel_label_get_accel_width);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_accel_label_set_accel_widget" XG_POST, gxg_gtk_accel_label_set_accel_widget, 2, 0, 0, H_gtk_accel_label_set_accel_widget);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_accel_label_set_accel_closure" XG_POST, gxg_gtk_accel_label_set_accel_closure, 2, 0, 0, H_gtk_accel_label_set_accel_closure);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_accel_label_refetch" XG_POST, gxg_gtk_accel_label_refetch, 1, 0, 0, H_gtk_accel_label_refetch);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_accel_map_add_entry" XG_POST, gxg_gtk_accel_map_add_entry, 3, 0, 0, H_gtk_accel_map_add_entry);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_accel_map_lookup_entry" XG_POST, gxg_gtk_accel_map_lookup_entry, 2, 0, 0, H_gtk_accel_map_lookup_entry);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_accel_map_change_entry" XG_POST, gxg_gtk_accel_map_change_entry, 4, 0, 0, H_gtk_accel_map_change_entry);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_accel_map_load" XG_POST, gxg_gtk_accel_map_load, 1, 0, 0, H_gtk_accel_map_load);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_accel_map_save" XG_POST, gxg_gtk_accel_map_save, 1, 0, 0, H_gtk_accel_map_save);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_accel_map_foreach" XG_POST, gxg_gtk_accel_map_foreach, 2, 0, 0, H_gtk_accel_map_foreach);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_accel_map_load_fd" XG_POST, gxg_gtk_accel_map_load_fd, 1, 0, 0, H_gtk_accel_map_load_fd);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_accel_map_load_scanner" XG_POST, gxg_gtk_accel_map_load_scanner, 1, 0, 0, H_gtk_accel_map_load_scanner);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_accel_map_save_fd" XG_POST, gxg_gtk_accel_map_save_fd, 1, 0, 0, H_gtk_accel_map_save_fd);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_accel_map_add_filter" XG_POST, gxg_gtk_accel_map_add_filter, 1, 0, 0, H_gtk_accel_map_add_filter);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_accel_map_foreach_unfiltered" XG_POST, gxg_gtk_accel_map_foreach_unfiltered, 2, 0, 0, H_gtk_accel_map_foreach_unfiltered);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_accessible_get_type" XG_POST, gxg_gtk_accessible_get_type, 0, 0, 0, H_gtk_accessible_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_accessible_connect_widget_destroyed" XG_POST, gxg_gtk_accessible_connect_widget_destroyed, 1, 0, 0, H_gtk_accessible_connect_widget_destroyed);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_adjustment_get_type" XG_POST, gxg_gtk_adjustment_get_type, 0, 0, 0, H_gtk_adjustment_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_adjustment_new" XG_POST, gxg_gtk_adjustment_new, 6, 0, 0, H_gtk_adjustment_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_adjustment_changed" XG_POST, gxg_gtk_adjustment_changed, 1, 0, 0, H_gtk_adjustment_changed);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_adjustment_value_changed" XG_POST, gxg_gtk_adjustment_value_changed, 1, 0, 0, H_gtk_adjustment_value_changed);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_adjustment_clamp_page" XG_POST, gxg_gtk_adjustment_clamp_page, 3, 0, 0, H_gtk_adjustment_clamp_page);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_adjustment_get_value" XG_POST, gxg_gtk_adjustment_get_value, 1, 0, 0, H_gtk_adjustment_get_value);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_adjustment_set_value" XG_POST, gxg_gtk_adjustment_set_value, 2, 0, 0, H_gtk_adjustment_set_value);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_alignment_get_type" XG_POST, gxg_gtk_alignment_get_type, 0, 0, 0, H_gtk_alignment_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_alignment_new" XG_POST, gxg_gtk_alignment_new, 4, 0, 0, H_gtk_alignment_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_alignment_set" XG_POST, gxg_gtk_alignment_set, 5, 0, 0, H_gtk_alignment_set);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_arrow_get_type" XG_POST, gxg_gtk_arrow_get_type, 0, 0, 0, H_gtk_arrow_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_arrow_new" XG_POST, gxg_gtk_arrow_new, 2, 0, 0, H_gtk_arrow_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_arrow_set" XG_POST, gxg_gtk_arrow_set, 3, 0, 0, H_gtk_arrow_set);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_aspect_frame_get_type" XG_POST, gxg_gtk_aspect_frame_get_type, 0, 0, 0, H_gtk_aspect_frame_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_aspect_frame_new" XG_POST, gxg_gtk_aspect_frame_new, 5, 0, 0, H_gtk_aspect_frame_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_aspect_frame_set" XG_POST, gxg_gtk_aspect_frame_set, 5, 0, 0, H_gtk_aspect_frame_set);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_button_box_get_type" XG_POST, gxg_gtk_button_box_get_type, 0, 0, 0, H_gtk_button_box_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_button_box_get_layout" XG_POST, gxg_gtk_button_box_get_layout, 1, 0, 0, H_gtk_button_box_get_layout);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_button_box_set_layout" XG_POST, gxg_gtk_button_box_set_layout, 2, 0, 0, H_gtk_button_box_set_layout);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_button_box_set_child_secondary" XG_POST, gxg_gtk_button_box_set_child_secondary, 3, 0, 0, H_gtk_button_box_set_child_secondary);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_binding_set_new" XG_POST, gxg_gtk_binding_set_new, 1, 0, 0, H_gtk_binding_set_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_binding_set_by_class" XG_POST, gxg_gtk_binding_set_by_class, 1, 0, 0, H_gtk_binding_set_by_class);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_binding_set_find" XG_POST, gxg_gtk_binding_set_find, 1, 0, 0, H_gtk_binding_set_find);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_bindings_activate" XG_POST, gxg_gtk_bindings_activate, 3, 0, 0, H_gtk_bindings_activate);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_binding_set_activate" XG_POST, gxg_gtk_binding_set_activate, 4, 0, 0, H_gtk_binding_set_activate);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_binding_entry_clear" XG_POST, gxg_gtk_binding_entry_clear, 3, 0, 0, H_gtk_binding_entry_clear);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_binding_entry_add_signal" XG_POST, gxg_gtk_binding_entry_add_signal, 5, 0, 0, H_gtk_binding_entry_add_signal);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_binding_set_add_path" XG_POST, gxg_gtk_binding_set_add_path, 4, 0, 0, H_gtk_binding_set_add_path);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_binding_entry_remove" XG_POST, gxg_gtk_binding_entry_remove, 3, 0, 0, H_gtk_binding_entry_remove);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_binding_entry_add_signall" XG_POST, gxg_gtk_binding_entry_add_signall, 5, 0, 0, H_gtk_binding_entry_add_signall);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_binding_parse_binding" XG_POST, gxg_gtk_binding_parse_binding, 1, 0, 0, H_gtk_binding_parse_binding);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_bin_get_type" XG_POST, gxg_gtk_bin_get_type, 0, 0, 0, H_gtk_bin_get_type);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_bin_get_child" XG_POST, gxg_gtk_bin_get_child, 1, 0, 0, H_gtk_bin_get_child);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_box_get_type" XG_POST, gxg_gtk_box_get_type, 0, 0, 0, H_gtk_box_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_box_pack_start" XG_POST, gxg_gtk_box_pack_start, 5, 0, 0, H_gtk_box_pack_start);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_box_pack_end" XG_POST, gxg_gtk_box_pack_end, 5, 0, 0, H_gtk_box_pack_end);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_box_pack_start_defaults" XG_POST, gxg_gtk_box_pack_start_defaults, 2, 0, 0, H_gtk_box_pack_start_defaults);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_box_pack_end_defaults" XG_POST, gxg_gtk_box_pack_end_defaults, 2, 0, 0, H_gtk_box_pack_end_defaults);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_box_set_homogeneous" XG_POST, gxg_gtk_box_set_homogeneous, 2, 0, 0, H_gtk_box_set_homogeneous);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_box_get_homogeneous" XG_POST, gxg_gtk_box_get_homogeneous, 1, 0, 0, H_gtk_box_get_homogeneous);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_box_set_spacing" XG_POST, gxg_gtk_box_set_spacing, 2, 0, 0, H_gtk_box_set_spacing);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_box_get_spacing" XG_POST, gxg_gtk_box_get_spacing, 1, 0, 0, H_gtk_box_get_spacing);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_box_reorder_child" XG_POST, gxg_gtk_box_reorder_child, 3, 0, 0, H_gtk_box_reorder_child);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_box_query_child_packing" XG_POST, gxg_gtk_box_query_child_packing, 6, 0, 0, H_gtk_box_query_child_packing);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_box_set_child_packing" XG_POST, gxg_gtk_box_set_child_packing, 6, 0, 0, H_gtk_box_set_child_packing);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_button_get_type" XG_POST, gxg_gtk_button_get_type, 0, 0, 0, H_gtk_button_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_button_new" XG_POST, gxg_gtk_button_new, 0, 0, 0, H_gtk_button_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_button_new_with_label" XG_POST, gxg_gtk_button_new_with_label, 1, 0, 0, H_gtk_button_new_with_label);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_button_new_from_stock" XG_POST, gxg_gtk_button_new_from_stock, 1, 0, 0, H_gtk_button_new_from_stock);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_button_new_with_mnemonic" XG_POST, gxg_gtk_button_new_with_mnemonic, 1, 0, 0, H_gtk_button_new_with_mnemonic);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_button_pressed" XG_POST, gxg_gtk_button_pressed, 1, 0, 0, H_gtk_button_pressed);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_button_released" XG_POST, gxg_gtk_button_released, 1, 0, 0, H_gtk_button_released);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_button_clicked" XG_POST, gxg_gtk_button_clicked, 1, 0, 0, H_gtk_button_clicked);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_button_enter" XG_POST, gxg_gtk_button_enter, 1, 0, 0, H_gtk_button_enter);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_button_leave" XG_POST, gxg_gtk_button_leave, 1, 0, 0, H_gtk_button_leave);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_button_set_relief" XG_POST, gxg_gtk_button_set_relief, 2, 0, 0, H_gtk_button_set_relief);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_button_get_relief" XG_POST, gxg_gtk_button_get_relief, 1, 0, 0, H_gtk_button_get_relief);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_button_set_label" XG_POST, gxg_gtk_button_set_label, 2, 0, 0, H_gtk_button_set_label);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_button_get_label" XG_POST, gxg_gtk_button_get_label, 1, 0, 0, H_gtk_button_get_label);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_button_set_use_underline" XG_POST, gxg_gtk_button_set_use_underline, 2, 0, 0, H_gtk_button_set_use_underline);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_button_get_use_underline" XG_POST, gxg_gtk_button_get_use_underline, 1, 0, 0, H_gtk_button_get_use_underline);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_button_set_use_stock" XG_POST, gxg_gtk_button_set_use_stock, 2, 0, 0, H_gtk_button_set_use_stock);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_button_get_use_stock" XG_POST, gxg_gtk_button_get_use_stock, 1, 0, 0, H_gtk_button_get_use_stock);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_calendar_get_type" XG_POST, gxg_gtk_calendar_get_type, 0, 0, 0, H_gtk_calendar_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_calendar_new" XG_POST, gxg_gtk_calendar_new, 0, 0, 0, H_gtk_calendar_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_calendar_select_month" XG_POST, gxg_gtk_calendar_select_month, 3, 0, 0, H_gtk_calendar_select_month);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_calendar_select_day" XG_POST, gxg_gtk_calendar_select_day, 2, 0, 0, H_gtk_calendar_select_day);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_calendar_mark_day" XG_POST, gxg_gtk_calendar_mark_day, 2, 0, 0, H_gtk_calendar_mark_day);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_calendar_unmark_day" XG_POST, gxg_gtk_calendar_unmark_day, 2, 0, 0, H_gtk_calendar_unmark_day);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_calendar_clear_marks" XG_POST, gxg_gtk_calendar_clear_marks, 1, 0, 0, H_gtk_calendar_clear_marks);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_calendar_display_options" XG_POST, gxg_gtk_calendar_display_options, 2, 0, 0, H_gtk_calendar_display_options);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_calendar_get_date" XG_POST, gxg_gtk_calendar_get_date, 4, 0, 0, H_gtk_calendar_get_date);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_calendar_freeze" XG_POST, gxg_gtk_calendar_freeze, 1, 0, 0, H_gtk_calendar_freeze);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_calendar_thaw" XG_POST, gxg_gtk_calendar_thaw, 1, 0, 0, H_gtk_calendar_thaw);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_cell_editable_get_type" XG_POST, gxg_gtk_cell_editable_get_type, 0, 0, 0, H_gtk_cell_editable_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_cell_editable_start_editing" XG_POST, gxg_gtk_cell_editable_start_editing, 2, 0, 0, H_gtk_cell_editable_start_editing);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_cell_editable_editing_done" XG_POST, gxg_gtk_cell_editable_editing_done, 1, 0, 0, H_gtk_cell_editable_editing_done);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_cell_editable_remove_widget" XG_POST, gxg_gtk_cell_editable_remove_widget, 1, 0, 0, H_gtk_cell_editable_remove_widget);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_cell_renderer_get_type" XG_POST, gxg_gtk_cell_renderer_get_type, 0, 0, 0, H_gtk_cell_renderer_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_cell_renderer_get_size" XG_POST, gxg_gtk_cell_renderer_get_size, 3, 4, 0, H_gtk_cell_renderer_get_size);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_cell_renderer_render" XG_POST, gxg_gtk_cell_renderer_render, 7, 0, 0, H_gtk_cell_renderer_render);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_cell_renderer_activate" XG_POST, gxg_gtk_cell_renderer_activate, 7, 0, 0, H_gtk_cell_renderer_activate);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_cell_renderer_start_editing" XG_POST, gxg_gtk_cell_renderer_start_editing, 7, 0, 0, H_gtk_cell_renderer_start_editing);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_cell_renderer_set_fixed_size" XG_POST, gxg_gtk_cell_renderer_set_fixed_size, 3, 0, 0, H_gtk_cell_renderer_set_fixed_size);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_cell_renderer_get_fixed_size" XG_POST, gxg_gtk_cell_renderer_get_fixed_size, 1, 2, 0, H_gtk_cell_renderer_get_fixed_size);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_cell_renderer_pixbuf_get_type" XG_POST, gxg_gtk_cell_renderer_pixbuf_get_type, 0, 0, 0, H_gtk_cell_renderer_pixbuf_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_cell_renderer_pixbuf_new" XG_POST, gxg_gtk_cell_renderer_pixbuf_new, 0, 0, 0, H_gtk_cell_renderer_pixbuf_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_cell_renderer_text_get_type" XG_POST, gxg_gtk_cell_renderer_text_get_type, 0, 0, 0, H_gtk_cell_renderer_text_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_cell_renderer_text_new" XG_POST, gxg_gtk_cell_renderer_text_new, 0, 0, 0, H_gtk_cell_renderer_text_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_cell_renderer_text_set_fixed_height_from_font" XG_POST, gxg_gtk_cell_renderer_text_set_fixed_height_from_font, 2, 0, 0, H_gtk_cell_renderer_text_set_fixed_height_from_font);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_cell_renderer_toggle_get_type" XG_POST, gxg_gtk_cell_renderer_toggle_get_type, 0, 0, 0, H_gtk_cell_renderer_toggle_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_cell_renderer_toggle_new" XG_POST, gxg_gtk_cell_renderer_toggle_new, 0, 0, 0, H_gtk_cell_renderer_toggle_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_cell_renderer_toggle_get_radio" XG_POST, gxg_gtk_cell_renderer_toggle_get_radio, 1, 0, 0, H_gtk_cell_renderer_toggle_get_radio);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_cell_renderer_toggle_set_radio" XG_POST, gxg_gtk_cell_renderer_toggle_set_radio, 2, 0, 0, H_gtk_cell_renderer_toggle_set_radio);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_cell_renderer_toggle_get_active" XG_POST, gxg_gtk_cell_renderer_toggle_get_active, 1, 0, 0, H_gtk_cell_renderer_toggle_get_active);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_cell_renderer_toggle_set_active" XG_POST, gxg_gtk_cell_renderer_toggle_set_active, 2, 0, 0, H_gtk_cell_renderer_toggle_set_active);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_check_button_get_type" XG_POST, gxg_gtk_check_button_get_type, 0, 0, 0, H_gtk_check_button_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_check_button_new" XG_POST, gxg_gtk_check_button_new, 0, 0, 0, H_gtk_check_button_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_check_button_new_with_label" XG_POST, gxg_gtk_check_button_new_with_label, 1, 0, 0, H_gtk_check_button_new_with_label);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_check_button_new_with_mnemonic" XG_POST, gxg_gtk_check_button_new_with_mnemonic, 1, 0, 0, H_gtk_check_button_new_with_mnemonic);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_check_menu_item_get_type" XG_POST, gxg_gtk_check_menu_item_get_type, 0, 0, 0, H_gtk_check_menu_item_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_check_menu_item_new" XG_POST, gxg_gtk_check_menu_item_new, 0, 0, 0, H_gtk_check_menu_item_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_check_menu_item_new_with_label" XG_POST, gxg_gtk_check_menu_item_new_with_label, 1, 0, 0, H_gtk_check_menu_item_new_with_label);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_check_menu_item_new_with_mnemonic" XG_POST, gxg_gtk_check_menu_item_new_with_mnemonic, 1, 0, 0, H_gtk_check_menu_item_new_with_mnemonic);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_check_menu_item_set_active" XG_POST, gxg_gtk_check_menu_item_set_active, 2, 0, 0, H_gtk_check_menu_item_set_active);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_check_menu_item_get_active" XG_POST, gxg_gtk_check_menu_item_get_active, 1, 0, 0, H_gtk_check_menu_item_get_active);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_check_menu_item_toggled" XG_POST, gxg_gtk_check_menu_item_toggled, 1, 0, 0, H_gtk_check_menu_item_toggled);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_check_menu_item_set_inconsistent" XG_POST, gxg_gtk_check_menu_item_set_inconsistent, 2, 0, 0, H_gtk_check_menu_item_set_inconsistent);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_check_menu_item_get_inconsistent" XG_POST, gxg_gtk_check_menu_item_get_inconsistent, 1, 0, 0, H_gtk_check_menu_item_get_inconsistent);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_clipboard_get" XG_POST, gxg_gtk_clipboard_get, 1, 0, 0, H_gtk_clipboard_get);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_clipboard_set_with_data" XG_POST, gxg_gtk_clipboard_set_with_data, 5, 1, 0, H_gtk_clipboard_set_with_data);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_clipboard_get_owner" XG_POST, gxg_gtk_clipboard_get_owner, 1, 0, 0, H_gtk_clipboard_get_owner);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_clipboard_clear" XG_POST, gxg_gtk_clipboard_clear, 1, 0, 0, H_gtk_clipboard_clear);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_clipboard_set_text" XG_POST, gxg_gtk_clipboard_set_text, 3, 0, 0, H_gtk_clipboard_set_text);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_clipboard_request_contents" XG_POST, gxg_gtk_clipboard_request_contents, 3, 1, 0, H_gtk_clipboard_request_contents);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_clipboard_request_text" XG_POST, gxg_gtk_clipboard_request_text, 2, 1, 0, H_gtk_clipboard_request_text);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_clipboard_wait_for_contents" XG_POST, gxg_gtk_clipboard_wait_for_contents, 2, 0, 0, H_gtk_clipboard_wait_for_contents);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_clipboard_wait_for_text" XG_POST, gxg_gtk_clipboard_wait_for_text, 1, 0, 0, H_gtk_clipboard_wait_for_text);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_clipboard_wait_is_text_available" XG_POST, gxg_gtk_clipboard_wait_is_text_available, 1, 0, 0, H_gtk_clipboard_wait_is_text_available);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_color_selection_dialog_get_type" XG_POST, gxg_gtk_color_selection_dialog_get_type, 0, 0, 0, H_gtk_color_selection_dialog_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_color_selection_dialog_new" XG_POST, gxg_gtk_color_selection_dialog_new, 1, 0, 0, H_gtk_color_selection_dialog_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_color_selection_get_type" XG_POST, gxg_gtk_color_selection_get_type, 0, 0, 0, H_gtk_color_selection_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_color_selection_new" XG_POST, gxg_gtk_color_selection_new, 0, 0, 0, H_gtk_color_selection_new);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_color_selection_get_has_opacity_control" XG_POST, gxg_gtk_color_selection_get_has_opacity_control, 1, 0, 0, H_gtk_color_selection_get_has_opacity_control);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_color_selection_set_has_opacity_control" XG_POST, gxg_gtk_color_selection_set_has_opacity_control, 2, 0, 0, H_gtk_color_selection_set_has_opacity_control);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_color_selection_get_has_palette" XG_POST, gxg_gtk_color_selection_get_has_palette, 1, 0, 0, H_gtk_color_selection_get_has_palette);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_color_selection_set_has_palette" XG_POST, gxg_gtk_color_selection_set_has_palette, 2, 0, 0, H_gtk_color_selection_set_has_palette);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_color_selection_set_current_color" XG_POST, gxg_gtk_color_selection_set_current_color, 2, 0, 0, H_gtk_color_selection_set_current_color);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_color_selection_set_current_alpha" XG_POST, gxg_gtk_color_selection_set_current_alpha, 2, 0, 0, H_gtk_color_selection_set_current_alpha);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_color_selection_get_current_color" XG_POST, gxg_gtk_color_selection_get_current_color, 2, 0, 0, H_gtk_color_selection_get_current_color);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_color_selection_get_current_alpha" XG_POST, gxg_gtk_color_selection_get_current_alpha, 1, 0, 0, H_gtk_color_selection_get_current_alpha);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_color_selection_set_previous_color" XG_POST, gxg_gtk_color_selection_set_previous_color, 2, 0, 0, H_gtk_color_selection_set_previous_color);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_color_selection_set_previous_alpha" XG_POST, gxg_gtk_color_selection_set_previous_alpha, 2, 0, 0, H_gtk_color_selection_set_previous_alpha);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_color_selection_get_previous_color" XG_POST, gxg_gtk_color_selection_get_previous_color, 2, 0, 0, H_gtk_color_selection_get_previous_color);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_color_selection_get_previous_alpha" XG_POST, gxg_gtk_color_selection_get_previous_alpha, 1, 0, 0, H_gtk_color_selection_get_previous_alpha);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_color_selection_is_adjusting" XG_POST, gxg_gtk_color_selection_is_adjusting, 1, 0, 0, H_gtk_color_selection_is_adjusting);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_color_selection_palette_from_string" XG_POST, gxg_gtk_color_selection_palette_from_string, 1, 2, 0, H_gtk_color_selection_palette_from_string);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_color_selection_palette_to_string" XG_POST, gxg_gtk_color_selection_palette_to_string, 2, 0, 0, H_gtk_color_selection_palette_to_string);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_color_selection_set_change_palette_hook" XG_POST, gxg_gtk_color_selection_set_change_palette_hook, 1, 0, 0, H_gtk_color_selection_set_change_palette_hook);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_combo_get_type" XG_POST, gxg_gtk_combo_get_type, 0, 0, 0, H_gtk_combo_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_combo_new" XG_POST, gxg_gtk_combo_new, 0, 0, 0, H_gtk_combo_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_combo_set_value_in_list" XG_POST, gxg_gtk_combo_set_value_in_list, 3, 0, 0, H_gtk_combo_set_value_in_list);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_combo_set_use_arrows" XG_POST, gxg_gtk_combo_set_use_arrows, 2, 0, 0, H_gtk_combo_set_use_arrows);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_combo_set_use_arrows_always" XG_POST, gxg_gtk_combo_set_use_arrows_always, 2, 0, 0, H_gtk_combo_set_use_arrows_always);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_combo_set_case_sensitive" XG_POST, gxg_gtk_combo_set_case_sensitive, 2, 0, 0, H_gtk_combo_set_case_sensitive);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_combo_set_item_string" XG_POST, gxg_gtk_combo_set_item_string, 3, 0, 0, H_gtk_combo_set_item_string);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_combo_set_popdown_strings" XG_POST, gxg_gtk_combo_set_popdown_strings, 2, 0, 0, H_gtk_combo_set_popdown_strings);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_combo_disable_activate" XG_POST, gxg_gtk_combo_disable_activate, 1, 0, 0, H_gtk_combo_disable_activate);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_RESIZE_CONTAINER" XG_POST, gxg_GTK_IS_RESIZE_CONTAINER, 1, 0, 0, H_GTK_IS_RESIZE_CONTAINER);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_container_get_type" XG_POST, gxg_gtk_container_get_type, 0, 0, 0, H_gtk_container_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_container_set_border_width" XG_POST, gxg_gtk_container_set_border_width, 2, 0, 0, H_gtk_container_set_border_width);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_container_get_border_width" XG_POST, gxg_gtk_container_get_border_width, 1, 0, 0, H_gtk_container_get_border_width);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_container_add" XG_POST, gxg_gtk_container_add, 2, 0, 0, H_gtk_container_add);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_container_remove" XG_POST, gxg_gtk_container_remove, 2, 0, 0, H_gtk_container_remove);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_container_set_resize_mode" XG_POST, gxg_gtk_container_set_resize_mode, 2, 0, 0, H_gtk_container_set_resize_mode);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_container_get_resize_mode" XG_POST, gxg_gtk_container_get_resize_mode, 1, 0, 0, H_gtk_container_get_resize_mode);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_container_check_resize" XG_POST, gxg_gtk_container_check_resize, 1, 0, 0, H_gtk_container_check_resize);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_container_foreach" XG_POST, gxg_gtk_container_foreach, 2, 1, 0, H_gtk_container_foreach);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_container_get_children" XG_POST, gxg_gtk_container_get_children, 1, 0, 0, H_gtk_container_get_children);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_curve_get_type" XG_POST, gxg_gtk_curve_get_type, 0, 0, 0, H_gtk_curve_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_curve_new" XG_POST, gxg_gtk_curve_new, 0, 0, 0, H_gtk_curve_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_curve_reset" XG_POST, gxg_gtk_curve_reset, 1, 0, 0, H_gtk_curve_reset);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_curve_set_gamma" XG_POST, gxg_gtk_curve_set_gamma, 2, 0, 0, H_gtk_curve_set_gamma);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_curve_set_range" XG_POST, gxg_gtk_curve_set_range, 5, 0, 0, H_gtk_curve_set_range);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_curve_get_vector" XG_POST, gxg_gtk_curve_get_vector, 3, 0, 0, H_gtk_curve_get_vector);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_curve_set_vector" XG_POST, gxg_gtk_curve_set_vector, 3, 0, 0, H_gtk_curve_set_vector);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_curve_set_curve_type" XG_POST, gxg_gtk_curve_set_curve_type, 2, 0, 0, H_gtk_curve_set_curve_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_dialog_get_type" XG_POST, gxg_gtk_dialog_get_type, 0, 0, 0, H_gtk_dialog_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_dialog_new" XG_POST, gxg_gtk_dialog_new, 0, 0, 0, H_gtk_dialog_new);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_dialog_new_with_buttons" XG_POST, gxg_gtk_dialog_new_with_buttons, 4, 0, 0, H_gtk_dialog_new_with_buttons);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_dialog_add_action_widget" XG_POST, gxg_gtk_dialog_add_action_widget, 3, 0, 0, H_gtk_dialog_add_action_widget);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_dialog_add_button" XG_POST, gxg_gtk_dialog_add_button, 3, 0, 0, H_gtk_dialog_add_button);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_dialog_add_buttons" XG_POST, gxg_gtk_dialog_add_buttons, 2, 0, 0, H_gtk_dialog_add_buttons);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_dialog_set_response_sensitive" XG_POST, gxg_gtk_dialog_set_response_sensitive, 3, 0, 0, H_gtk_dialog_set_response_sensitive);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_dialog_set_default_response" XG_POST, gxg_gtk_dialog_set_default_response, 2, 0, 0, H_gtk_dialog_set_default_response);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_dialog_set_has_separator" XG_POST, gxg_gtk_dialog_set_has_separator, 2, 0, 0, H_gtk_dialog_set_has_separator);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_dialog_get_has_separator" XG_POST, gxg_gtk_dialog_get_has_separator, 1, 0, 0, H_gtk_dialog_get_has_separator);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_dialog_response" XG_POST, gxg_gtk_dialog_response, 2, 0, 0, H_gtk_dialog_response);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_dialog_run" XG_POST, gxg_gtk_dialog_run, 1, 0, 0, H_gtk_dialog_run);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_drag_get_data" XG_POST, gxg_gtk_drag_get_data, 4, 0, 0, H_gtk_drag_get_data);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_drag_finish" XG_POST, gxg_gtk_drag_finish, 4, 0, 0, H_gtk_drag_finish);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_drag_get_source_widget" XG_POST, gxg_gtk_drag_get_source_widget, 1, 0, 0, H_gtk_drag_get_source_widget);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_drag_highlight" XG_POST, gxg_gtk_drag_highlight, 1, 0, 0, H_gtk_drag_highlight);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_drag_unhighlight" XG_POST, gxg_gtk_drag_unhighlight, 1, 0, 0, H_gtk_drag_unhighlight);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_drag_dest_set" XG_POST, gxg_gtk_drag_dest_set, 5, 0, 0, H_gtk_drag_dest_set);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_drag_dest_set_proxy" XG_POST, gxg_gtk_drag_dest_set_proxy, 4, 0, 0, H_gtk_drag_dest_set_proxy);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_drag_dest_unset" XG_POST, gxg_gtk_drag_dest_unset, 1, 0, 0, H_gtk_drag_dest_unset);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_drag_dest_find_target" XG_POST, gxg_gtk_drag_dest_find_target, 3, 0, 0, H_gtk_drag_dest_find_target);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_drag_dest_get_target_list" XG_POST, gxg_gtk_drag_dest_get_target_list, 1, 0, 0, H_gtk_drag_dest_get_target_list);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_drag_dest_set_target_list" XG_POST, gxg_gtk_drag_dest_set_target_list, 2, 0, 0, H_gtk_drag_dest_set_target_list);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_drag_source_set" XG_POST, gxg_gtk_drag_source_set, 5, 0, 0, H_gtk_drag_source_set);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_drag_source_unset" XG_POST, gxg_gtk_drag_source_unset, 1, 0, 0, H_gtk_drag_source_unset);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_drag_source_set_icon" XG_POST, gxg_gtk_drag_source_set_icon, 4, 0, 0, H_gtk_drag_source_set_icon);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_drag_source_set_icon_pixbuf" XG_POST, gxg_gtk_drag_source_set_icon_pixbuf, 2, 0, 0, H_gtk_drag_source_set_icon_pixbuf);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_drag_source_set_icon_stock" XG_POST, gxg_gtk_drag_source_set_icon_stock, 2, 0, 0, H_gtk_drag_source_set_icon_stock);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_drag_begin" XG_POST, gxg_gtk_drag_begin, 5, 0, 0, H_gtk_drag_begin);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_drag_set_icon_widget" XG_POST, gxg_gtk_drag_set_icon_widget, 4, 0, 0, H_gtk_drag_set_icon_widget);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_drag_set_icon_pixmap" XG_POST, gxg_gtk_drag_set_icon_pixmap, 6, 0, 0, H_gtk_drag_set_icon_pixmap);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_drag_set_icon_pixbuf" XG_POST, gxg_gtk_drag_set_icon_pixbuf, 4, 0, 0, H_gtk_drag_set_icon_pixbuf);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_drag_set_icon_stock" XG_POST, gxg_gtk_drag_set_icon_stock, 4, 0, 0, H_gtk_drag_set_icon_stock);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_drag_set_icon_default" XG_POST, gxg_gtk_drag_set_icon_default, 1, 0, 0, H_gtk_drag_set_icon_default);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_drag_check_threshold" XG_POST, gxg_gtk_drag_check_threshold, 5, 0, 0, H_gtk_drag_check_threshold);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_drawing_area_get_type" XG_POST, gxg_gtk_drawing_area_get_type, 0, 0, 0, H_gtk_drawing_area_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_drawing_area_new" XG_POST, gxg_gtk_drawing_area_new, 0, 0, 0, H_gtk_drawing_area_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_editable_get_type" XG_POST, gxg_gtk_editable_get_type, 0, 0, 0, H_gtk_editable_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_editable_select_region" XG_POST, gxg_gtk_editable_select_region, 3, 0, 0, H_gtk_editable_select_region);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_editable_get_selection_bounds" XG_POST, gxg_gtk_editable_get_selection_bounds, 3, 0, 0, H_gtk_editable_get_selection_bounds);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_editable_insert_text" XG_POST, gxg_gtk_editable_insert_text, 4, 0, 0, H_gtk_editable_insert_text);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_editable_delete_text" XG_POST, gxg_gtk_editable_delete_text, 3, 0, 0, H_gtk_editable_delete_text);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_editable_get_chars" XG_POST, gxg_gtk_editable_get_chars, 3, 0, 0, H_gtk_editable_get_chars);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_editable_cut_clipboard" XG_POST, gxg_gtk_editable_cut_clipboard, 1, 0, 0, H_gtk_editable_cut_clipboard);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_editable_copy_clipboard" XG_POST, gxg_gtk_editable_copy_clipboard, 1, 0, 0, H_gtk_editable_copy_clipboard);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_editable_paste_clipboard" XG_POST, gxg_gtk_editable_paste_clipboard, 1, 0, 0, H_gtk_editable_paste_clipboard);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_editable_delete_selection" XG_POST, gxg_gtk_editable_delete_selection, 1, 0, 0, H_gtk_editable_delete_selection);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_editable_set_position" XG_POST, gxg_gtk_editable_set_position, 2, 0, 0, H_gtk_editable_set_position);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_editable_get_position" XG_POST, gxg_gtk_editable_get_position, 1, 0, 0, H_gtk_editable_get_position);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_editable_set_editable" XG_POST, gxg_gtk_editable_set_editable, 2, 0, 0, H_gtk_editable_set_editable);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_editable_get_editable" XG_POST, gxg_gtk_editable_get_editable, 1, 0, 0, H_gtk_editable_get_editable);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_entry_get_type" XG_POST, gxg_gtk_entry_get_type, 0, 0, 0, H_gtk_entry_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_entry_new" XG_POST, gxg_gtk_entry_new, 0, 0, 0, H_gtk_entry_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_entry_set_visibility" XG_POST, gxg_gtk_entry_set_visibility, 2, 0, 0, H_gtk_entry_set_visibility);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_entry_get_visibility" XG_POST, gxg_gtk_entry_get_visibility, 1, 0, 0, H_gtk_entry_get_visibility);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_entry_set_invisible_char" XG_POST, gxg_gtk_entry_set_invisible_char, 2, 0, 0, H_gtk_entry_set_invisible_char);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_entry_get_invisible_char" XG_POST, gxg_gtk_entry_get_invisible_char, 1, 0, 0, H_gtk_entry_get_invisible_char);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_entry_set_has_frame" XG_POST, gxg_gtk_entry_set_has_frame, 2, 0, 0, H_gtk_entry_set_has_frame);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_entry_get_has_frame" XG_POST, gxg_gtk_entry_get_has_frame, 1, 0, 0, H_gtk_entry_get_has_frame);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_entry_set_max_length" XG_POST, gxg_gtk_entry_set_max_length, 2, 0, 0, H_gtk_entry_set_max_length);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_entry_get_max_length" XG_POST, gxg_gtk_entry_get_max_length, 1, 0, 0, H_gtk_entry_get_max_length);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_entry_set_activates_default" XG_POST, gxg_gtk_entry_set_activates_default, 2, 0, 0, H_gtk_entry_set_activates_default);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_entry_get_activates_default" XG_POST, gxg_gtk_entry_get_activates_default, 1, 0, 0, H_gtk_entry_get_activates_default);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_entry_set_width_chars" XG_POST, gxg_gtk_entry_set_width_chars, 2, 0, 0, H_gtk_entry_set_width_chars);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_entry_get_width_chars" XG_POST, gxg_gtk_entry_get_width_chars, 1, 0, 0, H_gtk_entry_get_width_chars);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_entry_set_text" XG_POST, gxg_gtk_entry_set_text, 2, 0, 0, H_gtk_entry_set_text);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_entry_get_text" XG_POST, gxg_gtk_entry_get_text, 1, 0, 0, H_gtk_entry_get_text);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_entry_get_layout" XG_POST, gxg_gtk_entry_get_layout, 1, 0, 0, H_gtk_entry_get_layout);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_entry_get_layout_offsets" XG_POST, gxg_gtk_entry_get_layout_offsets, 1, 2, 0, H_gtk_entry_get_layout_offsets);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_event_box_get_type" XG_POST, gxg_gtk_event_box_get_type, 0, 0, 0, H_gtk_event_box_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_event_box_new" XG_POST, gxg_gtk_event_box_new, 0, 0, 0, H_gtk_event_box_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_file_selection_get_type" XG_POST, gxg_gtk_file_selection_get_type, 0, 0, 0, H_gtk_file_selection_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_file_selection_new" XG_POST, gxg_gtk_file_selection_new, 1, 0, 0, H_gtk_file_selection_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_file_selection_set_filename" XG_POST, gxg_gtk_file_selection_set_filename, 2, 0, 0, H_gtk_file_selection_set_filename);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_file_selection_get_filename" XG_POST, gxg_gtk_file_selection_get_filename, 1, 0, 0, H_gtk_file_selection_get_filename);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_file_selection_complete" XG_POST, gxg_gtk_file_selection_complete, 2, 0, 0, H_gtk_file_selection_complete);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_file_selection_show_fileop_buttons" XG_POST, gxg_gtk_file_selection_show_fileop_buttons, 1, 0, 0, H_gtk_file_selection_show_fileop_buttons);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_file_selection_hide_fileop_buttons" XG_POST, gxg_gtk_file_selection_hide_fileop_buttons, 1, 0, 0, H_gtk_file_selection_hide_fileop_buttons);
#if (!HAVE_GTK_1)
#if HAVE_GTK_FILE_SELECTION_GET_SELECTIONS
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_file_selection_get_selections" XG_POST, gxg_gtk_file_selection_get_selections, 1, 0, 0, H_gtk_file_selection_get_selections);
#endif
#if HAVE_GTK_FILE_SELECTION_GET_SELECTIONS
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_file_selection_set_select_multiple" XG_POST, gxg_gtk_file_selection_set_select_multiple, 2, 0, 0, H_gtk_file_selection_set_select_multiple);
#endif
#if HAVE_GTK_FILE_SELECTION_GET_SELECTIONS
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_file_selection_get_select_multiple" XG_POST, gxg_gtk_file_selection_get_select_multiple, 1, 0, 0, H_gtk_file_selection_get_select_multiple);
#endif
#if HAVE_GTK_ICON_SET_GET_TYPE
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_icon_set_get_type" XG_POST, gxg_gtk_icon_set_get_type, 0, 0, 0, H_gtk_icon_set_get_type);
#endif
#if HAVE_GTK_ICON_SET_GET_TYPE
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_icon_source_get_type" XG_POST, gxg_gtk_icon_source_get_type, 0, 0, 0, H_gtk_icon_source_get_type);
#endif
#if HAVE_GTK_ICON_SET_GET_TYPE
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_selection_data_get_type" XG_POST, gxg_gtk_selection_data_get_type, 0, 0, 0, H_gtk_selection_data_get_type);
#endif
#if HAVE_GTK_ICON_SET_GET_TYPE
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_border_get_type" XG_POST, gxg_gtk_border_get_type, 0, 0, 0, H_gtk_border_get_type);
#endif
#if HAVE_GTK_ICON_SET_GET_TYPE
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_path_get_type" XG_POST, gxg_gtk_tree_path_get_type, 0, 0, 0, H_gtk_tree_path_get_type);
#endif
#if HAVE_GTK_ICON_SET_GET_TYPE
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_iter_get_type" XG_POST, gxg_gtk_tree_iter_get_type, 0, 0, 0, H_gtk_tree_iter_get_type);
#endif
#if HAVE_GTK_ICON_SET_GET_TYPE
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_identifier_get_type" XG_POST, gxg_gtk_identifier_get_type, 0, 0, 0, H_gtk_identifier_get_type);
#endif
#if HAVE_GTK_ICON_SET_GET_TYPE
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_requisition_get_type" XG_POST, gxg_gtk_requisition_get_type, 0, 0, 0, H_gtk_requisition_get_type);
#endif
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_fixed_get_type" XG_POST, gxg_gtk_fixed_get_type, 0, 0, 0, H_gtk_fixed_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_fixed_new" XG_POST, gxg_gtk_fixed_new, 0, 0, 0, H_gtk_fixed_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_fixed_put" XG_POST, gxg_gtk_fixed_put, 4, 0, 0, H_gtk_fixed_put);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_fixed_move" XG_POST, gxg_gtk_fixed_move, 4, 0, 0, H_gtk_fixed_move);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_fixed_set_has_window" XG_POST, gxg_gtk_fixed_set_has_window, 2, 0, 0, H_gtk_fixed_set_has_window);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_fixed_get_has_window" XG_POST, gxg_gtk_fixed_get_has_window, 1, 0, 0, H_gtk_fixed_get_has_window);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_font_selection_get_type" XG_POST, gxg_gtk_font_selection_get_type, 0, 0, 0, H_gtk_font_selection_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_font_selection_new" XG_POST, gxg_gtk_font_selection_new, 0, 0, 0, H_gtk_font_selection_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_font_selection_get_font_name" XG_POST, gxg_gtk_font_selection_get_font_name, 1, 0, 0, H_gtk_font_selection_get_font_name);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_font_selection_set_font_name" XG_POST, gxg_gtk_font_selection_set_font_name, 2, 0, 0, H_gtk_font_selection_set_font_name);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_font_selection_get_preview_text" XG_POST, gxg_gtk_font_selection_get_preview_text, 1, 0, 0, H_gtk_font_selection_get_preview_text);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_font_selection_set_preview_text" XG_POST, gxg_gtk_font_selection_set_preview_text, 2, 0, 0, H_gtk_font_selection_set_preview_text);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_font_selection_dialog_get_type" XG_POST, gxg_gtk_font_selection_dialog_get_type, 0, 0, 0, H_gtk_font_selection_dialog_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_font_selection_dialog_new" XG_POST, gxg_gtk_font_selection_dialog_new, 1, 0, 0, H_gtk_font_selection_dialog_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_font_selection_dialog_get_font_name" XG_POST, gxg_gtk_font_selection_dialog_get_font_name, 1, 0, 0, H_gtk_font_selection_dialog_get_font_name);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_font_selection_dialog_set_font_name" XG_POST, gxg_gtk_font_selection_dialog_set_font_name, 2, 0, 0, H_gtk_font_selection_dialog_set_font_name);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_font_selection_dialog_get_preview_text" XG_POST, gxg_gtk_font_selection_dialog_get_preview_text, 1, 0, 0, H_gtk_font_selection_dialog_get_preview_text);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_font_selection_dialog_set_preview_text" XG_POST, gxg_gtk_font_selection_dialog_set_preview_text, 2, 0, 0, H_gtk_font_selection_dialog_set_preview_text);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_frame_get_type" XG_POST, gxg_gtk_frame_get_type, 0, 0, 0, H_gtk_frame_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_frame_new" XG_POST, gxg_gtk_frame_new, 1, 0, 0, H_gtk_frame_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_frame_set_label" XG_POST, gxg_gtk_frame_set_label, 2, 0, 0, H_gtk_frame_set_label);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_frame_get_label" XG_POST, gxg_gtk_frame_get_label, 1, 0, 0, H_gtk_frame_get_label);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_frame_set_label_widget" XG_POST, gxg_gtk_frame_set_label_widget, 2, 0, 0, H_gtk_frame_set_label_widget);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_frame_get_label_widget" XG_POST, gxg_gtk_frame_get_label_widget, 1, 0, 0, H_gtk_frame_get_label_widget);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_frame_set_label_align" XG_POST, gxg_gtk_frame_set_label_align, 3, 0, 0, H_gtk_frame_set_label_align);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_frame_get_label_align" XG_POST, gxg_gtk_frame_get_label_align, 1, 2, 0, H_gtk_frame_get_label_align);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_frame_set_shadow_type" XG_POST, gxg_gtk_frame_set_shadow_type, 2, 0, 0, H_gtk_frame_set_shadow_type);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_frame_get_shadow_type" XG_POST, gxg_gtk_frame_get_shadow_type, 1, 0, 0, H_gtk_frame_get_shadow_type);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_gamma_curve_get_type" XG_POST, gxg_gtk_gamma_curve_get_type, 0, 0, 0, H_gtk_gamma_curve_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_gamma_curve_new" XG_POST, gxg_gtk_gamma_curve_new, 0, 0, 0, H_gtk_gamma_curve_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_gc_get" XG_POST, gxg_gtk_gc_get, 4, 0, 0, H_gtk_gc_get);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_gc_release" XG_POST, gxg_gtk_gc_release, 1, 0, 0, H_gtk_gc_release);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_handle_box_get_type" XG_POST, gxg_gtk_handle_box_get_type, 0, 0, 0, H_gtk_handle_box_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_handle_box_new" XG_POST, gxg_gtk_handle_box_new, 0, 0, 0, H_gtk_handle_box_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_handle_box_set_shadow_type" XG_POST, gxg_gtk_handle_box_set_shadow_type, 2, 0, 0, H_gtk_handle_box_set_shadow_type);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_handle_box_get_shadow_type" XG_POST, gxg_gtk_handle_box_get_shadow_type, 1, 0, 0, H_gtk_handle_box_get_shadow_type);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_handle_box_set_handle_position" XG_POST, gxg_gtk_handle_box_set_handle_position, 2, 0, 0, H_gtk_handle_box_set_handle_position);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_handle_box_get_handle_position" XG_POST, gxg_gtk_handle_box_get_handle_position, 1, 0, 0, H_gtk_handle_box_get_handle_position);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_handle_box_set_snap_edge" XG_POST, gxg_gtk_handle_box_set_snap_edge, 2, 0, 0, H_gtk_handle_box_set_snap_edge);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_handle_box_get_snap_edge" XG_POST, gxg_gtk_handle_box_get_snap_edge, 1, 0, 0, H_gtk_handle_box_get_snap_edge);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_hbutton_box_get_type" XG_POST, gxg_gtk_hbutton_box_get_type, 0, 0, 0, H_gtk_hbutton_box_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_hbutton_box_new" XG_POST, gxg_gtk_hbutton_box_new, 0, 0, 0, H_gtk_hbutton_box_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_hbox_get_type" XG_POST, gxg_gtk_hbox_get_type, 0, 0, 0, H_gtk_hbox_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_hbox_new" XG_POST, gxg_gtk_hbox_new, 2, 0, 0, H_gtk_hbox_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_hpaned_get_type" XG_POST, gxg_gtk_hpaned_get_type, 0, 0, 0, H_gtk_hpaned_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_hpaned_new" XG_POST, gxg_gtk_hpaned_new, 0, 0, 0, H_gtk_hpaned_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_hruler_get_type" XG_POST, gxg_gtk_hruler_get_type, 0, 0, 0, H_gtk_hruler_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_hruler_new" XG_POST, gxg_gtk_hruler_new, 0, 0, 0, H_gtk_hruler_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_hscale_get_type" XG_POST, gxg_gtk_hscale_get_type, 0, 0, 0, H_gtk_hscale_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_hscale_new" XG_POST, gxg_gtk_hscale_new, 1, 0, 0, H_gtk_hscale_new);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_hscale_new_with_range" XG_POST, gxg_gtk_hscale_new_with_range, 3, 0, 0, H_gtk_hscale_new_with_range);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_hscrollbar_get_type" XG_POST, gxg_gtk_hscrollbar_get_type, 0, 0, 0, H_gtk_hscrollbar_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_hscrollbar_new" XG_POST, gxg_gtk_hscrollbar_new, 1, 0, 0, H_gtk_hscrollbar_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_hseparator_get_type" XG_POST, gxg_gtk_hseparator_get_type, 0, 0, 0, H_gtk_hseparator_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_hseparator_new" XG_POST, gxg_gtk_hseparator_new, 0, 0, 0, H_gtk_hseparator_new);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_icon_factory_get_type" XG_POST, gxg_gtk_icon_factory_get_type, 0, 0, 0, H_gtk_icon_factory_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_icon_factory_new" XG_POST, gxg_gtk_icon_factory_new, 0, 0, 0, H_gtk_icon_factory_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_icon_factory_add" XG_POST, gxg_gtk_icon_factory_add, 3, 0, 0, H_gtk_icon_factory_add);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_icon_factory_lookup" XG_POST, gxg_gtk_icon_factory_lookup, 2, 0, 0, H_gtk_icon_factory_lookup);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_icon_factory_add_default" XG_POST, gxg_gtk_icon_factory_add_default, 1, 0, 0, H_gtk_icon_factory_add_default);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_icon_factory_remove_default" XG_POST, gxg_gtk_icon_factory_remove_default, 1, 0, 0, H_gtk_icon_factory_remove_default);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_icon_factory_lookup_default" XG_POST, gxg_gtk_icon_factory_lookup_default, 1, 0, 0, H_gtk_icon_factory_lookup_default);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_icon_size_lookup" XG_POST, gxg_gtk_icon_size_lookup, 1, 2, 0, H_gtk_icon_size_lookup);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_icon_size_register" XG_POST, gxg_gtk_icon_size_register, 3, 0, 0, H_gtk_icon_size_register);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_icon_size_register_alias" XG_POST, gxg_gtk_icon_size_register_alias, 2, 0, 0, H_gtk_icon_size_register_alias);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_icon_size_from_name" XG_POST, gxg_gtk_icon_size_from_name, 1, 0, 0, H_gtk_icon_size_from_name);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_icon_size_get_name" XG_POST, gxg_gtk_icon_size_get_name, 1, 0, 0, H_gtk_icon_size_get_name);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_icon_set_new" XG_POST, gxg_gtk_icon_set_new, 0, 0, 0, H_gtk_icon_set_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_icon_set_new_from_pixbuf" XG_POST, gxg_gtk_icon_set_new_from_pixbuf, 1, 0, 0, H_gtk_icon_set_new_from_pixbuf);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_icon_set_ref" XG_POST, gxg_gtk_icon_set_ref, 1, 0, 0, H_gtk_icon_set_ref);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_icon_set_unref" XG_POST, gxg_gtk_icon_set_unref, 1, 0, 0, H_gtk_icon_set_unref);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_icon_set_copy" XG_POST, gxg_gtk_icon_set_copy, 1, 0, 0, H_gtk_icon_set_copy);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_icon_set_render_icon" XG_POST, gxg_gtk_icon_set_render_icon, 7, 0, 0, H_gtk_icon_set_render_icon);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_icon_set_add_source" XG_POST, gxg_gtk_icon_set_add_source, 2, 0, 0, H_gtk_icon_set_add_source);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_icon_set_get_sizes" XG_POST, gxg_gtk_icon_set_get_sizes, 1, 2, 0, H_gtk_icon_set_get_sizes);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_icon_source_new" XG_POST, gxg_gtk_icon_source_new, 0, 0, 0, H_gtk_icon_source_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_icon_source_copy" XG_POST, gxg_gtk_icon_source_copy, 1, 0, 0, H_gtk_icon_source_copy);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_icon_source_free" XG_POST, gxg_gtk_icon_source_free, 1, 0, 0, H_gtk_icon_source_free);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_icon_source_set_filename" XG_POST, gxg_gtk_icon_source_set_filename, 2, 0, 0, H_gtk_icon_source_set_filename);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_icon_source_set_pixbuf" XG_POST, gxg_gtk_icon_source_set_pixbuf, 2, 0, 0, H_gtk_icon_source_set_pixbuf);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_icon_source_get_filename" XG_POST, gxg_gtk_icon_source_get_filename, 1, 0, 0, H_gtk_icon_source_get_filename);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_icon_source_get_pixbuf" XG_POST, gxg_gtk_icon_source_get_pixbuf, 1, 0, 0, H_gtk_icon_source_get_pixbuf);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_icon_source_set_direction_wildcarded" XG_POST, gxg_gtk_icon_source_set_direction_wildcarded, 2, 0, 0, H_gtk_icon_source_set_direction_wildcarded);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_icon_source_set_state_wildcarded" XG_POST, gxg_gtk_icon_source_set_state_wildcarded, 2, 0, 0, H_gtk_icon_source_set_state_wildcarded);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_icon_source_set_size_wildcarded" XG_POST, gxg_gtk_icon_source_set_size_wildcarded, 2, 0, 0, H_gtk_icon_source_set_size_wildcarded);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_icon_source_get_size_wildcarded" XG_POST, gxg_gtk_icon_source_get_size_wildcarded, 1, 0, 0, H_gtk_icon_source_get_size_wildcarded);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_icon_source_get_state_wildcarded" XG_POST, gxg_gtk_icon_source_get_state_wildcarded, 1, 0, 0, H_gtk_icon_source_get_state_wildcarded);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_icon_source_get_direction_wildcarded" XG_POST, gxg_gtk_icon_source_get_direction_wildcarded, 1, 0, 0, H_gtk_icon_source_get_direction_wildcarded);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_icon_source_set_direction" XG_POST, gxg_gtk_icon_source_set_direction, 2, 0, 0, H_gtk_icon_source_set_direction);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_icon_source_set_state" XG_POST, gxg_gtk_icon_source_set_state, 2, 0, 0, H_gtk_icon_source_set_state);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_icon_source_set_size" XG_POST, gxg_gtk_icon_source_set_size, 2, 0, 0, H_gtk_icon_source_set_size);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_icon_source_get_direction" XG_POST, gxg_gtk_icon_source_get_direction, 1, 0, 0, H_gtk_icon_source_get_direction);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_icon_source_get_state" XG_POST, gxg_gtk_icon_source_get_state, 1, 0, 0, H_gtk_icon_source_get_state);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_icon_source_get_size" XG_POST, gxg_gtk_icon_source_get_size, 1, 0, 0, H_gtk_icon_source_get_size);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_image_get_type" XG_POST, gxg_gtk_image_get_type, 0, 0, 0, H_gtk_image_get_type);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_image_new" XG_POST, gxg_gtk_image_new, 0, 0, 0, H_gtk_image_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_image_new_from_pixmap" XG_POST, gxg_gtk_image_new_from_pixmap, 2, 0, 0, H_gtk_image_new_from_pixmap);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_image_new_from_image" XG_POST, gxg_gtk_image_new_from_image, 2, 0, 0, H_gtk_image_new_from_image);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_image_new_from_file" XG_POST, gxg_gtk_image_new_from_file, 1, 0, 0, H_gtk_image_new_from_file);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_image_new_from_pixbuf" XG_POST, gxg_gtk_image_new_from_pixbuf, 1, 0, 0, H_gtk_image_new_from_pixbuf);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_image_new_from_stock" XG_POST, gxg_gtk_image_new_from_stock, 2, 0, 0, H_gtk_image_new_from_stock);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_image_new_from_icon_set" XG_POST, gxg_gtk_image_new_from_icon_set, 2, 0, 0, H_gtk_image_new_from_icon_set);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_image_new_from_animation" XG_POST, gxg_gtk_image_new_from_animation, 1, 0, 0, H_gtk_image_new_from_animation);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_image_set_from_pixmap" XG_POST, gxg_gtk_image_set_from_pixmap, 3, 0, 0, H_gtk_image_set_from_pixmap);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_image_set_from_image" XG_POST, gxg_gtk_image_set_from_image, 3, 0, 0, H_gtk_image_set_from_image);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_image_set_from_file" XG_POST, gxg_gtk_image_set_from_file, 2, 0, 0, H_gtk_image_set_from_file);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_image_set_from_pixbuf" XG_POST, gxg_gtk_image_set_from_pixbuf, 2, 0, 0, H_gtk_image_set_from_pixbuf);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_image_set_from_stock" XG_POST, gxg_gtk_image_set_from_stock, 3, 0, 0, H_gtk_image_set_from_stock);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_image_set_from_icon_set" XG_POST, gxg_gtk_image_set_from_icon_set, 3, 0, 0, H_gtk_image_set_from_icon_set);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_image_set_from_animation" XG_POST, gxg_gtk_image_set_from_animation, 2, 0, 0, H_gtk_image_set_from_animation);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_image_get_storage_type" XG_POST, gxg_gtk_image_get_storage_type, 1, 0, 0, H_gtk_image_get_storage_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_image_get_pixmap" XG_POST, gxg_gtk_image_get_pixmap, 1, 2, 0, H_gtk_image_get_pixmap);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_image_get_image" XG_POST, gxg_gtk_image_get_image, 1, 2, 0, H_gtk_image_get_image);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_image_get_pixbuf" XG_POST, gxg_gtk_image_get_pixbuf, 1, 0, 0, H_gtk_image_get_pixbuf);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_image_get_stock" XG_POST, gxg_gtk_image_get_stock, 1, 2, 0, H_gtk_image_get_stock);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_image_get_icon_set" XG_POST, gxg_gtk_image_get_icon_set, 1, 2, 0, H_gtk_image_get_icon_set);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_image_get_animation" XG_POST, gxg_gtk_image_get_animation, 1, 0, 0, H_gtk_image_get_animation);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_image_menu_item_get_type" XG_POST, gxg_gtk_image_menu_item_get_type, 0, 0, 0, H_gtk_image_menu_item_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_image_menu_item_new" XG_POST, gxg_gtk_image_menu_item_new, 0, 0, 0, H_gtk_image_menu_item_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_image_menu_item_new_with_label" XG_POST, gxg_gtk_image_menu_item_new_with_label, 1, 0, 0, H_gtk_image_menu_item_new_with_label);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_image_menu_item_new_with_mnemonic" XG_POST, gxg_gtk_image_menu_item_new_with_mnemonic, 1, 0, 0, H_gtk_image_menu_item_new_with_mnemonic);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_image_menu_item_new_from_stock" XG_POST, gxg_gtk_image_menu_item_new_from_stock, 2, 0, 0, H_gtk_image_menu_item_new_from_stock);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_image_menu_item_set_image" XG_POST, gxg_gtk_image_menu_item_set_image, 2, 0, 0, H_gtk_image_menu_item_set_image);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_image_menu_item_get_image" XG_POST, gxg_gtk_image_menu_item_get_image, 1, 0, 0, H_gtk_image_menu_item_get_image);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_im_context_get_type" XG_POST, gxg_gtk_im_context_get_type, 0, 0, 0, H_gtk_im_context_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_im_context_set_client_window" XG_POST, gxg_gtk_im_context_set_client_window, 2, 0, 0, H_gtk_im_context_set_client_window);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_im_context_get_preedit_string" XG_POST, gxg_gtk_im_context_get_preedit_string, 1, 3, 0, H_gtk_im_context_get_preedit_string);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_im_context_filter_keypress" XG_POST, gxg_gtk_im_context_filter_keypress, 2, 0, 0, H_gtk_im_context_filter_keypress);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_im_context_focus_in" XG_POST, gxg_gtk_im_context_focus_in, 1, 0, 0, H_gtk_im_context_focus_in);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_im_context_focus_out" XG_POST, gxg_gtk_im_context_focus_out, 1, 0, 0, H_gtk_im_context_focus_out);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_im_context_reset" XG_POST, gxg_gtk_im_context_reset, 1, 0, 0, H_gtk_im_context_reset);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_im_context_set_cursor_location" XG_POST, gxg_gtk_im_context_set_cursor_location, 2, 0, 0, H_gtk_im_context_set_cursor_location);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_im_context_set_use_preedit" XG_POST, gxg_gtk_im_context_set_use_preedit, 2, 0, 0, H_gtk_im_context_set_use_preedit);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_im_context_set_surrounding" XG_POST, gxg_gtk_im_context_set_surrounding, 4, 0, 0, H_gtk_im_context_set_surrounding);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_im_context_get_surrounding" XG_POST, gxg_gtk_im_context_get_surrounding, 1, 2, 0, H_gtk_im_context_get_surrounding);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_im_context_delete_surrounding" XG_POST, gxg_gtk_im_context_delete_surrounding, 3, 0, 0, H_gtk_im_context_delete_surrounding);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_im_context_simple_get_type" XG_POST, gxg_gtk_im_context_simple_get_type, 0, 0, 0, H_gtk_im_context_simple_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_im_context_simple_new" XG_POST, gxg_gtk_im_context_simple_new, 0, 0, 0, H_gtk_im_context_simple_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_im_context_simple_add_table" XG_POST, gxg_gtk_im_context_simple_add_table, 4, 0, 0, H_gtk_im_context_simple_add_table);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_im_multicontext_get_type" XG_POST, gxg_gtk_im_multicontext_get_type, 0, 0, 0, H_gtk_im_multicontext_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_im_multicontext_new" XG_POST, gxg_gtk_im_multicontext_new, 0, 0, 0, H_gtk_im_multicontext_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_im_multicontext_append_menuitems" XG_POST, gxg_gtk_im_multicontext_append_menuitems, 2, 0, 0, H_gtk_im_multicontext_append_menuitems);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_input_dialog_get_type" XG_POST, gxg_gtk_input_dialog_get_type, 0, 0, 0, H_gtk_input_dialog_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_input_dialog_new" XG_POST, gxg_gtk_input_dialog_new, 0, 0, 0, H_gtk_input_dialog_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_invisible_get_type" XG_POST, gxg_gtk_invisible_get_type, 0, 0, 0, H_gtk_invisible_get_type);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_invisible_new" XG_POST, gxg_gtk_invisible_new, 0, 0, 0, H_gtk_invisible_new);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_item_factory_get_type" XG_POST, gxg_gtk_item_factory_get_type, 0, 0, 0, H_gtk_item_factory_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_item_factory_new" XG_POST, gxg_gtk_item_factory_new, 3, 0, 0, H_gtk_item_factory_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_item_factory_construct" XG_POST, gxg_gtk_item_factory_construct, 4, 0, 0, H_gtk_item_factory_construct);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_item_factory_add_foreign" XG_POST, gxg_gtk_item_factory_add_foreign, 5, 0, 0, H_gtk_item_factory_add_foreign);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_item_factory_from_widget" XG_POST, gxg_gtk_item_factory_from_widget, 1, 0, 0, H_gtk_item_factory_from_widget);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_item_factory_path_from_widget" XG_POST, gxg_gtk_item_factory_path_from_widget, 1, 0, 0, H_gtk_item_factory_path_from_widget);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_item_factory_get_item" XG_POST, gxg_gtk_item_factory_get_item, 2, 0, 0, H_gtk_item_factory_get_item);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_item_factory_get_widget" XG_POST, gxg_gtk_item_factory_get_widget, 2, 0, 0, H_gtk_item_factory_get_widget);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_item_factory_get_widget_by_action" XG_POST, gxg_gtk_item_factory_get_widget_by_action, 2, 0, 0, H_gtk_item_factory_get_widget_by_action);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_item_factory_get_item_by_action" XG_POST, gxg_gtk_item_factory_get_item_by_action, 2, 0, 0, H_gtk_item_factory_get_item_by_action);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_item_factory_create_item" XG_POST, gxg_gtk_item_factory_create_item, 4, 0, 0, H_gtk_item_factory_create_item);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_item_factory_create_items" XG_POST, gxg_gtk_item_factory_create_items, 4, 0, 0, H_gtk_item_factory_create_items);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_item_factory_delete_item" XG_POST, gxg_gtk_item_factory_delete_item, 2, 0, 0, H_gtk_item_factory_delete_item);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_item_factory_delete_entry" XG_POST, gxg_gtk_item_factory_delete_entry, 2, 0, 0, H_gtk_item_factory_delete_entry);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_item_factory_delete_entries" XG_POST, gxg_gtk_item_factory_delete_entries, 3, 0, 0, H_gtk_item_factory_delete_entries);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_item_factory_popup" XG_POST, gxg_gtk_item_factory_popup, 5, 0, 0, H_gtk_item_factory_popup);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_item_factory_popup_with_data" XG_POST, gxg_gtk_item_factory_popup_with_data, 7, 0, 0, H_gtk_item_factory_popup_with_data);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_item_factory_popup_data" XG_POST, gxg_gtk_item_factory_popup_data, 1, 0, 0, H_gtk_item_factory_popup_data);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_item_factory_popup_data_from_widget" XG_POST, gxg_gtk_item_factory_popup_data_from_widget, 1, 0, 0, H_gtk_item_factory_popup_data_from_widget);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_item_factory_set_translate_func" XG_POST, gxg_gtk_item_factory_set_translate_func, 4, 0, 0, H_gtk_item_factory_set_translate_func);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_item_get_type" XG_POST, gxg_gtk_item_get_type, 0, 0, 0, H_gtk_item_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_item_select" XG_POST, gxg_gtk_item_select, 1, 0, 0, H_gtk_item_select);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_item_deselect" XG_POST, gxg_gtk_item_deselect, 1, 0, 0, H_gtk_item_deselect);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_item_toggle" XG_POST, gxg_gtk_item_toggle, 1, 0, 0, H_gtk_item_toggle);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_label_get_type" XG_POST, gxg_gtk_label_get_type, 0, 0, 0, H_gtk_label_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_label_new" XG_POST, gxg_gtk_label_new, 1, 0, 0, H_gtk_label_new);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_label_new_with_mnemonic" XG_POST, gxg_gtk_label_new_with_mnemonic, 1, 0, 0, H_gtk_label_new_with_mnemonic);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_label_set_text" XG_POST, gxg_gtk_label_set_text, 2, 0, 0, H_gtk_label_set_text);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_label_get_text" XG_POST, gxg_gtk_label_get_text, 1, 0, 0, H_gtk_label_get_text);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_label_set_attributes" XG_POST, gxg_gtk_label_set_attributes, 2, 0, 0, H_gtk_label_set_attributes);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_label_get_attributes" XG_POST, gxg_gtk_label_get_attributes, 1, 0, 0, H_gtk_label_get_attributes);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_label_set_label" XG_POST, gxg_gtk_label_set_label, 2, 0, 0, H_gtk_label_set_label);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_label_get_label" XG_POST, gxg_gtk_label_get_label, 1, 0, 0, H_gtk_label_get_label);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_label_set_markup" XG_POST, gxg_gtk_label_set_markup, 2, 0, 0, H_gtk_label_set_markup);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_label_set_use_markup" XG_POST, gxg_gtk_label_set_use_markup, 2, 0, 0, H_gtk_label_set_use_markup);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_label_get_use_markup" XG_POST, gxg_gtk_label_get_use_markup, 1, 0, 0, H_gtk_label_get_use_markup);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_label_set_use_underline" XG_POST, gxg_gtk_label_set_use_underline, 2, 0, 0, H_gtk_label_set_use_underline);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_label_get_use_underline" XG_POST, gxg_gtk_label_get_use_underline, 1, 0, 0, H_gtk_label_get_use_underline);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_label_set_markup_with_mnemonic" XG_POST, gxg_gtk_label_set_markup_with_mnemonic, 2, 0, 0, H_gtk_label_set_markup_with_mnemonic);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_label_get_mnemonic_keyval" XG_POST, gxg_gtk_label_get_mnemonic_keyval, 1, 0, 0, H_gtk_label_get_mnemonic_keyval);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_label_set_mnemonic_widget" XG_POST, gxg_gtk_label_set_mnemonic_widget, 2, 0, 0, H_gtk_label_set_mnemonic_widget);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_label_get_mnemonic_widget" XG_POST, gxg_gtk_label_get_mnemonic_widget, 1, 0, 0, H_gtk_label_get_mnemonic_widget);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_label_set_text_with_mnemonic" XG_POST, gxg_gtk_label_set_text_with_mnemonic, 2, 0, 0, H_gtk_label_set_text_with_mnemonic);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_label_set_justify" XG_POST, gxg_gtk_label_set_justify, 2, 0, 0, H_gtk_label_set_justify);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_label_get_justify" XG_POST, gxg_gtk_label_get_justify, 1, 0, 0, H_gtk_label_get_justify);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_label_set_pattern" XG_POST, gxg_gtk_label_set_pattern, 2, 0, 0, H_gtk_label_set_pattern);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_label_set_line_wrap" XG_POST, gxg_gtk_label_set_line_wrap, 2, 0, 0, H_gtk_label_set_line_wrap);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_label_get_line_wrap" XG_POST, gxg_gtk_label_get_line_wrap, 1, 0, 0, H_gtk_label_get_line_wrap);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_label_set_selectable" XG_POST, gxg_gtk_label_set_selectable, 2, 0, 0, H_gtk_label_set_selectable);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_label_get_selectable" XG_POST, gxg_gtk_label_get_selectable, 1, 0, 0, H_gtk_label_get_selectable);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_label_select_region" XG_POST, gxg_gtk_label_select_region, 3, 0, 0, H_gtk_label_select_region);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_label_get_selection_bounds" XG_POST, gxg_gtk_label_get_selection_bounds, 1, 2, 0, H_gtk_label_get_selection_bounds);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_label_get_layout" XG_POST, gxg_gtk_label_get_layout, 1, 0, 0, H_gtk_label_get_layout);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_label_get_layout_offsets" XG_POST, gxg_gtk_label_get_layout_offsets, 1, 2, 0, H_gtk_label_get_layout_offsets);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_layout_get_type" XG_POST, gxg_gtk_layout_get_type, 0, 0, 0, H_gtk_layout_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_layout_new" XG_POST, gxg_gtk_layout_new, 2, 0, 0, H_gtk_layout_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_layout_put" XG_POST, gxg_gtk_layout_put, 4, 0, 0, H_gtk_layout_put);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_layout_move" XG_POST, gxg_gtk_layout_move, 4, 0, 0, H_gtk_layout_move);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_layout_set_size" XG_POST, gxg_gtk_layout_set_size, 3, 0, 0, H_gtk_layout_set_size);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_layout_get_size" XG_POST, gxg_gtk_layout_get_size, 1, 2, 0, H_gtk_layout_get_size);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_layout_get_hadjustment" XG_POST, gxg_gtk_layout_get_hadjustment, 1, 0, 0, H_gtk_layout_get_hadjustment);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_layout_get_vadjustment" XG_POST, gxg_gtk_layout_get_vadjustment, 1, 0, 0, H_gtk_layout_get_vadjustment);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_layout_set_hadjustment" XG_POST, gxg_gtk_layout_set_hadjustment, 2, 0, 0, H_gtk_layout_set_hadjustment);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_layout_set_vadjustment" XG_POST, gxg_gtk_layout_set_vadjustment, 2, 0, 0, H_gtk_layout_set_vadjustment);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_list_store_get_type" XG_POST, gxg_gtk_list_store_get_type, 0, 0, 0, H_gtk_list_store_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_list_store_new" XG_POST, gxg_gtk_list_store_new, 1, 0, 0, H_gtk_list_store_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_list_store_newv" XG_POST, gxg_gtk_list_store_newv, 2, 0, 0, H_gtk_list_store_newv);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_list_store_set_column_types" XG_POST, gxg_gtk_list_store_set_column_types, 3, 0, 0, H_gtk_list_store_set_column_types);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_list_store_set_value" XG_POST, gxg_gtk_list_store_set_value, 4, 0, 0, H_gtk_list_store_set_value);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_list_store_set" XG_POST, gxg_gtk_list_store_set, 2, 0, 0, H_gtk_list_store_set);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_list_store_remove" XG_POST, gxg_gtk_list_store_remove, 2, 0, 0, H_gtk_list_store_remove);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_list_store_insert" XG_POST, gxg_gtk_list_store_insert, 3, 0, 0, H_gtk_list_store_insert);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_list_store_insert_before" XG_POST, gxg_gtk_list_store_insert_before, 3, 0, 0, H_gtk_list_store_insert_before);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_list_store_insert_after" XG_POST, gxg_gtk_list_store_insert_after, 3, 0, 0, H_gtk_list_store_insert_after);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_list_store_prepend" XG_POST, gxg_gtk_list_store_prepend, 2, 0, 0, H_gtk_list_store_prepend);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_list_store_append" XG_POST, gxg_gtk_list_store_append, 2, 0, 0, H_gtk_list_store_append);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_list_store_clear" XG_POST, gxg_gtk_list_store_clear, 1, 0, 0, H_gtk_list_store_clear);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_check_version" XG_POST, gxg_gtk_check_version, 3, 0, 0, H_gtk_check_version);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_init" XG_POST, gxg_gtk_init, 2, 0, 0, H_gtk_init);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_init_check" XG_POST, gxg_gtk_init_check, 2, 0, 0, H_gtk_init_check);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_disable_setlocale" XG_POST, gxg_gtk_disable_setlocale, 0, 0, 0, H_gtk_disable_setlocale);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_set_locale" XG_POST, gxg_gtk_set_locale, 0, 0, 0, H_gtk_set_locale);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_get_default_language" XG_POST, gxg_gtk_get_default_language, 0, 0, 0, H_gtk_get_default_language);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_events_pending" XG_POST, gxg_gtk_events_pending, 0, 0, 0, H_gtk_events_pending);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_main_do_event" XG_POST, gxg_gtk_main_do_event, 1, 0, 0, H_gtk_main_do_event);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_main" XG_POST, gxg_gtk_main, 0, 0, 0, H_gtk_main);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_main_level" XG_POST, gxg_gtk_main_level, 0, 0, 0, H_gtk_main_level);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_main_quit" XG_POST, gxg_gtk_main_quit, 0, 0, 0, H_gtk_main_quit);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_main_iteration" XG_POST, gxg_gtk_main_iteration, 0, 0, 0, H_gtk_main_iteration);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_main_iteration_do" XG_POST, gxg_gtk_main_iteration_do, 1, 0, 0, H_gtk_main_iteration_do);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_true" XG_POST, gxg_gtk_true, 0, 0, 0, H_gtk_true);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_false" XG_POST, gxg_gtk_false, 0, 0, 0, H_gtk_false);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_grab_add" XG_POST, gxg_gtk_grab_add, 1, 0, 0, H_gtk_grab_add);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_grab_get_current" XG_POST, gxg_gtk_grab_get_current, 0, 0, 0, H_gtk_grab_get_current);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_grab_remove" XG_POST, gxg_gtk_grab_remove, 1, 0, 0, H_gtk_grab_remove);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_init_add" XG_POST, gxg_gtk_init_add, 1, 1, 0, H_gtk_init_add);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_quit_add_destroy" XG_POST, gxg_gtk_quit_add_destroy, 2, 0, 0, H_gtk_quit_add_destroy);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_quit_add" XG_POST, gxg_gtk_quit_add, 2, 1, 0, H_gtk_quit_add);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_quit_add_full" XG_POST, gxg_gtk_quit_add_full, 5, 0, 0, H_gtk_quit_add_full);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_quit_remove" XG_POST, gxg_gtk_quit_remove, 1, 0, 0, H_gtk_quit_remove);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_quit_remove_by_data" XG_POST, gxg_gtk_quit_remove_by_data, 1, 0, 0, H_gtk_quit_remove_by_data);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_timeout_add" XG_POST, gxg_gtk_timeout_add, 2, 1, 0, H_gtk_timeout_add);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_timeout_add_full" XG_POST, gxg_gtk_timeout_add_full, 5, 0, 0, H_gtk_timeout_add_full);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_timeout_remove" XG_POST, gxg_gtk_timeout_remove, 1, 0, 0, H_gtk_timeout_remove);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_idle_add" XG_POST, gxg_gtk_idle_add, 1, 1, 0, H_gtk_idle_add);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_idle_add_priority" XG_POST, gxg_gtk_idle_add_priority, 2, 1, 0, H_gtk_idle_add_priority);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_idle_add_full" XG_POST, gxg_gtk_idle_add_full, 5, 0, 0, H_gtk_idle_add_full);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_idle_remove" XG_POST, gxg_gtk_idle_remove, 1, 0, 0, H_gtk_idle_remove);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_idle_remove_by_data" XG_POST, gxg_gtk_idle_remove_by_data, 1, 0, 0, H_gtk_idle_remove_by_data);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_input_add_full" XG_POST, gxg_gtk_input_add_full, 6, 0, 0, H_gtk_input_add_full);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_input_remove" XG_POST, gxg_gtk_input_remove, 1, 0, 0, H_gtk_input_remove);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_key_snooper_install" XG_POST, gxg_gtk_key_snooper_install, 1, 1, 0, H_gtk_key_snooper_install);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_key_snooper_remove" XG_POST, gxg_gtk_key_snooper_remove, 1, 0, 0, H_gtk_key_snooper_remove);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_get_current_event" XG_POST, gxg_gtk_get_current_event, 0, 0, 0, H_gtk_get_current_event);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_get_current_event_time" XG_POST, gxg_gtk_get_current_event_time, 0, 0, 0, H_gtk_get_current_event_time);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_get_current_event_state" XG_POST, gxg_gtk_get_current_event_state, 0, 1, 0, H_gtk_get_current_event_state);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_get_event_widget" XG_POST, gxg_gtk_get_event_widget, 1, 0, 0, H_gtk_get_event_widget);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_propagate_event" XG_POST, gxg_gtk_propagate_event, 2, 0, 0, H_gtk_propagate_event);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_menu_bar_get_type" XG_POST, gxg_gtk_menu_bar_get_type, 0, 0, 0, H_gtk_menu_bar_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_menu_bar_new" XG_POST, gxg_gtk_menu_bar_new, 0, 0, 0, H_gtk_menu_bar_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_menu_get_type" XG_POST, gxg_gtk_menu_get_type, 0, 0, 0, H_gtk_menu_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_menu_new" XG_POST, gxg_gtk_menu_new, 0, 0, 0, H_gtk_menu_new);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_menu_popup" XG_POST, gxg_gtk_menu_popup, 7, 0, 0, H_gtk_menu_popup);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_menu_reposition" XG_POST, gxg_gtk_menu_reposition, 1, 0, 0, H_gtk_menu_reposition);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_menu_popdown" XG_POST, gxg_gtk_menu_popdown, 1, 0, 0, H_gtk_menu_popdown);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_menu_get_active" XG_POST, gxg_gtk_menu_get_active, 1, 0, 0, H_gtk_menu_get_active);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_menu_set_active" XG_POST, gxg_gtk_menu_set_active, 2, 0, 0, H_gtk_menu_set_active);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_menu_set_accel_group" XG_POST, gxg_gtk_menu_set_accel_group, 2, 0, 0, H_gtk_menu_set_accel_group);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_menu_get_accel_group" XG_POST, gxg_gtk_menu_get_accel_group, 1, 0, 0, H_gtk_menu_get_accel_group);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_menu_set_accel_path" XG_POST, gxg_gtk_menu_set_accel_path, 2, 0, 0, H_gtk_menu_set_accel_path);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_menu_attach_to_widget" XG_POST, gxg_gtk_menu_attach_to_widget, 3, 0, 0, H_gtk_menu_attach_to_widget);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_menu_detach" XG_POST, gxg_gtk_menu_detach, 1, 0, 0, H_gtk_menu_detach);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_menu_get_attach_widget" XG_POST, gxg_gtk_menu_get_attach_widget, 1, 0, 0, H_gtk_menu_get_attach_widget);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_menu_set_tearoff_state" XG_POST, gxg_gtk_menu_set_tearoff_state, 2, 0, 0, H_gtk_menu_set_tearoff_state);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_menu_get_tearoff_state" XG_POST, gxg_gtk_menu_get_tearoff_state, 1, 0, 0, H_gtk_menu_get_tearoff_state);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_menu_set_title" XG_POST, gxg_gtk_menu_set_title, 2, 0, 0, H_gtk_menu_set_title);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_menu_get_title" XG_POST, gxg_gtk_menu_get_title, 1, 0, 0, H_gtk_menu_get_title);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_menu_reorder_child" XG_POST, gxg_gtk_menu_reorder_child, 3, 0, 0, H_gtk_menu_reorder_child);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_menu_item_get_type" XG_POST, gxg_gtk_menu_item_get_type, 0, 0, 0, H_gtk_menu_item_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_menu_item_new" XG_POST, gxg_gtk_menu_item_new, 0, 0, 0, H_gtk_menu_item_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_menu_item_new_with_label" XG_POST, gxg_gtk_menu_item_new_with_label, 1, 0, 0, H_gtk_menu_item_new_with_label);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_menu_item_new_with_mnemonic" XG_POST, gxg_gtk_menu_item_new_with_mnemonic, 1, 0, 0, H_gtk_menu_item_new_with_mnemonic);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_menu_item_set_submenu" XG_POST, gxg_gtk_menu_item_set_submenu, 2, 0, 0, H_gtk_menu_item_set_submenu);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_menu_item_get_submenu" XG_POST, gxg_gtk_menu_item_get_submenu, 1, 0, 0, H_gtk_menu_item_get_submenu);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_menu_item_remove_submenu" XG_POST, gxg_gtk_menu_item_remove_submenu, 1, 0, 0, H_gtk_menu_item_remove_submenu);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_menu_item_select" XG_POST, gxg_gtk_menu_item_select, 1, 0, 0, H_gtk_menu_item_select);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_menu_item_deselect" XG_POST, gxg_gtk_menu_item_deselect, 1, 0, 0, H_gtk_menu_item_deselect);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_menu_item_activate" XG_POST, gxg_gtk_menu_item_activate, 1, 0, 0, H_gtk_menu_item_activate);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_menu_item_toggle_size_request" XG_POST, gxg_gtk_menu_item_toggle_size_request, 2, 0, 0, H_gtk_menu_item_toggle_size_request);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_menu_item_toggle_size_allocate" XG_POST, gxg_gtk_menu_item_toggle_size_allocate, 2, 0, 0, H_gtk_menu_item_toggle_size_allocate);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_menu_item_set_right_justified" XG_POST, gxg_gtk_menu_item_set_right_justified, 2, 0, 0, H_gtk_menu_item_set_right_justified);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_menu_item_get_right_justified" XG_POST, gxg_gtk_menu_item_get_right_justified, 1, 0, 0, H_gtk_menu_item_get_right_justified);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_menu_item_set_accel_path" XG_POST, gxg_gtk_menu_item_set_accel_path, 2, 0, 0, H_gtk_menu_item_set_accel_path);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_menu_shell_get_type" XG_POST, gxg_gtk_menu_shell_get_type, 0, 0, 0, H_gtk_menu_shell_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_menu_shell_append" XG_POST, gxg_gtk_menu_shell_append, 2, 0, 0, H_gtk_menu_shell_append);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_menu_shell_prepend" XG_POST, gxg_gtk_menu_shell_prepend, 2, 0, 0, H_gtk_menu_shell_prepend);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_menu_shell_insert" XG_POST, gxg_gtk_menu_shell_insert, 3, 0, 0, H_gtk_menu_shell_insert);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_menu_shell_deactivate" XG_POST, gxg_gtk_menu_shell_deactivate, 1, 0, 0, H_gtk_menu_shell_deactivate);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_menu_shell_select_item" XG_POST, gxg_gtk_menu_shell_select_item, 2, 0, 0, H_gtk_menu_shell_select_item);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_menu_shell_deselect" XG_POST, gxg_gtk_menu_shell_deselect, 1, 0, 0, H_gtk_menu_shell_deselect);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_menu_shell_activate_item" XG_POST, gxg_gtk_menu_shell_activate_item, 3, 0, 0, H_gtk_menu_shell_activate_item);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_message_dialog_get_type" XG_POST, gxg_gtk_message_dialog_get_type, 0, 0, 0, H_gtk_message_dialog_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_message_dialog_new" XG_POST, gxg_gtk_message_dialog_new, 5, 0, 0, H_gtk_message_dialog_new);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_misc_get_type" XG_POST, gxg_gtk_misc_get_type, 0, 0, 0, H_gtk_misc_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_misc_set_alignment" XG_POST, gxg_gtk_misc_set_alignment, 3, 0, 0, H_gtk_misc_set_alignment);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_misc_get_alignment" XG_POST, gxg_gtk_misc_get_alignment, 1, 2, 0, H_gtk_misc_get_alignment);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_misc_set_padding" XG_POST, gxg_gtk_misc_set_padding, 3, 0, 0, H_gtk_misc_set_padding);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_misc_get_padding" XG_POST, gxg_gtk_misc_get_padding, 3, 0, 0, H_gtk_misc_get_padding);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_notebook_get_type" XG_POST, gxg_gtk_notebook_get_type, 0, 0, 0, H_gtk_notebook_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_notebook_new" XG_POST, gxg_gtk_notebook_new, 0, 0, 0, H_gtk_notebook_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_notebook_append_page" XG_POST, gxg_gtk_notebook_append_page, 3, 0, 0, H_gtk_notebook_append_page);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_notebook_append_page_menu" XG_POST, gxg_gtk_notebook_append_page_menu, 4, 0, 0, H_gtk_notebook_append_page_menu);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_notebook_prepend_page" XG_POST, gxg_gtk_notebook_prepend_page, 3, 0, 0, H_gtk_notebook_prepend_page);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_notebook_prepend_page_menu" XG_POST, gxg_gtk_notebook_prepend_page_menu, 4, 0, 0, H_gtk_notebook_prepend_page_menu);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_notebook_insert_page" XG_POST, gxg_gtk_notebook_insert_page, 4, 0, 0, H_gtk_notebook_insert_page);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_notebook_insert_page_menu" XG_POST, gxg_gtk_notebook_insert_page_menu, 5, 0, 0, H_gtk_notebook_insert_page_menu);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_notebook_remove_page" XG_POST, gxg_gtk_notebook_remove_page, 2, 0, 0, H_gtk_notebook_remove_page);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_notebook_get_current_page" XG_POST, gxg_gtk_notebook_get_current_page, 1, 0, 0, H_gtk_notebook_get_current_page);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_notebook_get_nth_page" XG_POST, gxg_gtk_notebook_get_nth_page, 2, 0, 0, H_gtk_notebook_get_nth_page);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_notebook_page_num" XG_POST, gxg_gtk_notebook_page_num, 2, 0, 0, H_gtk_notebook_page_num);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_notebook_set_current_page" XG_POST, gxg_gtk_notebook_set_current_page, 2, 0, 0, H_gtk_notebook_set_current_page);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_notebook_next_page" XG_POST, gxg_gtk_notebook_next_page, 1, 0, 0, H_gtk_notebook_next_page);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_notebook_prev_page" XG_POST, gxg_gtk_notebook_prev_page, 1, 0, 0, H_gtk_notebook_prev_page);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_notebook_set_show_border" XG_POST, gxg_gtk_notebook_set_show_border, 2, 0, 0, H_gtk_notebook_set_show_border);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_notebook_get_show_border" XG_POST, gxg_gtk_notebook_get_show_border, 1, 0, 0, H_gtk_notebook_get_show_border);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_notebook_set_show_tabs" XG_POST, gxg_gtk_notebook_set_show_tabs, 2, 0, 0, H_gtk_notebook_set_show_tabs);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_notebook_get_show_tabs" XG_POST, gxg_gtk_notebook_get_show_tabs, 1, 0, 0, H_gtk_notebook_get_show_tabs);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_notebook_set_tab_pos" XG_POST, gxg_gtk_notebook_set_tab_pos, 2, 0, 0, H_gtk_notebook_set_tab_pos);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_notebook_get_tab_pos" XG_POST, gxg_gtk_notebook_get_tab_pos, 1, 0, 0, H_gtk_notebook_get_tab_pos);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_notebook_set_scrollable" XG_POST, gxg_gtk_notebook_set_scrollable, 2, 0, 0, H_gtk_notebook_set_scrollable);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_notebook_get_scrollable" XG_POST, gxg_gtk_notebook_get_scrollable, 1, 0, 0, H_gtk_notebook_get_scrollable);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_notebook_popup_enable" XG_POST, gxg_gtk_notebook_popup_enable, 1, 0, 0, H_gtk_notebook_popup_enable);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_notebook_popup_disable" XG_POST, gxg_gtk_notebook_popup_disable, 1, 0, 0, H_gtk_notebook_popup_disable);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_notebook_get_tab_label" XG_POST, gxg_gtk_notebook_get_tab_label, 2, 0, 0, H_gtk_notebook_get_tab_label);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_notebook_set_tab_label" XG_POST, gxg_gtk_notebook_set_tab_label, 3, 0, 0, H_gtk_notebook_set_tab_label);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_notebook_set_tab_label_text" XG_POST, gxg_gtk_notebook_set_tab_label_text, 3, 0, 0, H_gtk_notebook_set_tab_label_text);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_notebook_get_tab_label_text" XG_POST, gxg_gtk_notebook_get_tab_label_text, 2, 0, 0, H_gtk_notebook_get_tab_label_text);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_notebook_get_menu_label" XG_POST, gxg_gtk_notebook_get_menu_label, 2, 0, 0, H_gtk_notebook_get_menu_label);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_notebook_set_menu_label" XG_POST, gxg_gtk_notebook_set_menu_label, 3, 0, 0, H_gtk_notebook_set_menu_label);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_notebook_set_menu_label_text" XG_POST, gxg_gtk_notebook_set_menu_label_text, 3, 0, 0, H_gtk_notebook_set_menu_label_text);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_notebook_get_menu_label_text" XG_POST, gxg_gtk_notebook_get_menu_label_text, 2, 0, 0, H_gtk_notebook_get_menu_label_text);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_notebook_query_tab_label_packing" XG_POST, gxg_gtk_notebook_query_tab_label_packing, 5, 0, 0, H_gtk_notebook_query_tab_label_packing);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_notebook_set_tab_label_packing" XG_POST, gxg_gtk_notebook_set_tab_label_packing, 5, 0, 0, H_gtk_notebook_set_tab_label_packing);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_notebook_reorder_child" XG_POST, gxg_gtk_notebook_reorder_child, 3, 0, 0, H_gtk_notebook_reorder_child);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_OBJECT_TYPE_NAME" XG_POST, gxg_GTK_OBJECT_TYPE_NAME, 1, 0, 0, H_GTK_OBJECT_TYPE_NAME);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_OBJECT_FLAGS" XG_POST, gxg_GTK_OBJECT_FLAGS, 1, 0, 0, H_GTK_OBJECT_FLAGS);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_OBJECT_FLOATING" XG_POST, gxg_GTK_OBJECT_FLOATING, 1, 0, 0, H_GTK_OBJECT_FLOATING);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_OBJECT_SET_FLAGS" XG_POST, gxg_GTK_OBJECT_SET_FLAGS, 2, 0, 0, H_GTK_OBJECT_SET_FLAGS);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_OBJECT_UNSET_FLAGS" XG_POST, gxg_GTK_OBJECT_UNSET_FLAGS, 2, 0, 0, H_GTK_OBJECT_UNSET_FLAGS);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_object_get_type" XG_POST, gxg_gtk_object_get_type, 0, 0, 0, H_gtk_object_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_object_new" XG_POST, gxg_gtk_object_new, 2, 0, 0, H_gtk_object_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_object_sink" XG_POST, gxg_gtk_object_sink, 1, 0, 0, H_gtk_object_sink);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_object_destroy" XG_POST, gxg_gtk_object_destroy, 1, 0, 0, H_gtk_object_destroy);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_option_menu_get_type" XG_POST, gxg_gtk_option_menu_get_type, 0, 0, 0, H_gtk_option_menu_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_option_menu_new" XG_POST, gxg_gtk_option_menu_new, 0, 0, 0, H_gtk_option_menu_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_option_menu_get_menu" XG_POST, gxg_gtk_option_menu_get_menu, 1, 0, 0, H_gtk_option_menu_get_menu);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_option_menu_set_menu" XG_POST, gxg_gtk_option_menu_set_menu, 2, 0, 0, H_gtk_option_menu_set_menu);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_option_menu_remove_menu" XG_POST, gxg_gtk_option_menu_remove_menu, 1, 0, 0, H_gtk_option_menu_remove_menu);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_option_menu_get_history" XG_POST, gxg_gtk_option_menu_get_history, 1, 0, 0, H_gtk_option_menu_get_history);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_option_menu_set_history" XG_POST, gxg_gtk_option_menu_set_history, 2, 0, 0, H_gtk_option_menu_set_history);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_paned_get_type" XG_POST, gxg_gtk_paned_get_type, 0, 0, 0, H_gtk_paned_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_paned_add1" XG_POST, gxg_gtk_paned_add1, 2, 0, 0, H_gtk_paned_add1);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_paned_add2" XG_POST, gxg_gtk_paned_add2, 2, 0, 0, H_gtk_paned_add2);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_paned_pack1" XG_POST, gxg_gtk_paned_pack1, 4, 0, 0, H_gtk_paned_pack1);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_paned_pack2" XG_POST, gxg_gtk_paned_pack2, 4, 0, 0, H_gtk_paned_pack2);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_paned_get_position" XG_POST, gxg_gtk_paned_get_position, 1, 0, 0, H_gtk_paned_get_position);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_paned_set_position" XG_POST, gxg_gtk_paned_set_position, 2, 0, 0, H_gtk_paned_set_position);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_paned_compute_position" XG_POST, gxg_gtk_paned_compute_position, 4, 0, 0, H_gtk_paned_compute_position);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_plug_get_type" XG_POST, gxg_gtk_plug_get_type, 0, 0, 0, H_gtk_plug_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_plug_construct" XG_POST, gxg_gtk_plug_construct, 2, 0, 0, H_gtk_plug_construct);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_plug_new" XG_POST, gxg_gtk_plug_new, 1, 0, 0, H_gtk_plug_new);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_plug_get_id" XG_POST, gxg_gtk_plug_get_id, 1, 0, 0, H_gtk_plug_get_id);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_progress_bar_get_type" XG_POST, gxg_gtk_progress_bar_get_type, 0, 0, 0, H_gtk_progress_bar_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_progress_bar_new" XG_POST, gxg_gtk_progress_bar_new, 0, 0, 0, H_gtk_progress_bar_new);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_progress_bar_pulse" XG_POST, gxg_gtk_progress_bar_pulse, 1, 0, 0, H_gtk_progress_bar_pulse);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_progress_bar_set_text" XG_POST, gxg_gtk_progress_bar_set_text, 2, 0, 0, H_gtk_progress_bar_set_text);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_progress_bar_set_fraction" XG_POST, gxg_gtk_progress_bar_set_fraction, 2, 0, 0, H_gtk_progress_bar_set_fraction);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_progress_bar_set_pulse_step" XG_POST, gxg_gtk_progress_bar_set_pulse_step, 2, 0, 0, H_gtk_progress_bar_set_pulse_step);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_progress_bar_set_orientation" XG_POST, gxg_gtk_progress_bar_set_orientation, 2, 0, 0, H_gtk_progress_bar_set_orientation);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_progress_bar_get_text" XG_POST, gxg_gtk_progress_bar_get_text, 1, 0, 0, H_gtk_progress_bar_get_text);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_progress_bar_get_fraction" XG_POST, gxg_gtk_progress_bar_get_fraction, 1, 0, 0, H_gtk_progress_bar_get_fraction);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_progress_bar_get_pulse_step" XG_POST, gxg_gtk_progress_bar_get_pulse_step, 1, 0, 0, H_gtk_progress_bar_get_pulse_step);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_progress_bar_get_orientation" XG_POST, gxg_gtk_progress_bar_get_orientation, 1, 0, 0, H_gtk_progress_bar_get_orientation);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_radio_button_get_type" XG_POST, gxg_gtk_radio_button_get_type, 0, 0, 0, H_gtk_radio_button_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_radio_button_new" XG_POST, gxg_gtk_radio_button_new, 1, 0, 0, H_gtk_radio_button_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_radio_button_new_from_widget" XG_POST, gxg_gtk_radio_button_new_from_widget, 1, 0, 0, H_gtk_radio_button_new_from_widget);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_radio_button_new_with_label" XG_POST, gxg_gtk_radio_button_new_with_label, 2, 0, 0, H_gtk_radio_button_new_with_label);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_radio_button_new_with_label_from_widget" XG_POST, gxg_gtk_radio_button_new_with_label_from_widget, 2, 0, 0, H_gtk_radio_button_new_with_label_from_widget);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_radio_button_new_with_mnemonic" XG_POST, gxg_gtk_radio_button_new_with_mnemonic, 2, 0, 0, H_gtk_radio_button_new_with_mnemonic);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_radio_button_new_with_mnemonic_from_widget" XG_POST, gxg_gtk_radio_button_new_with_mnemonic_from_widget, 2, 0, 0, H_gtk_radio_button_new_with_mnemonic_from_widget);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_radio_button_get_group" XG_POST, gxg_gtk_radio_button_get_group, 1, 0, 0, H_gtk_radio_button_get_group);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_radio_button_set_group" XG_POST, gxg_gtk_radio_button_set_group, 2, 0, 0, H_gtk_radio_button_set_group);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_radio_menu_item_get_type" XG_POST, gxg_gtk_radio_menu_item_get_type, 0, 0, 0, H_gtk_radio_menu_item_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_radio_menu_item_new" XG_POST, gxg_gtk_radio_menu_item_new, 1, 0, 0, H_gtk_radio_menu_item_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_radio_menu_item_new_with_label" XG_POST, gxg_gtk_radio_menu_item_new_with_label, 2, 0, 0, H_gtk_radio_menu_item_new_with_label);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_radio_menu_item_new_with_mnemonic" XG_POST, gxg_gtk_radio_menu_item_new_with_mnemonic, 2, 0, 0, H_gtk_radio_menu_item_new_with_mnemonic);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_radio_menu_item_get_group" XG_POST, gxg_gtk_radio_menu_item_get_group, 1, 0, 0, H_gtk_radio_menu_item_get_group);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_radio_menu_item_set_group" XG_POST, gxg_gtk_radio_menu_item_set_group, 2, 0, 0, H_gtk_radio_menu_item_set_group);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_range_get_type" XG_POST, gxg_gtk_range_get_type, 0, 0, 0, H_gtk_range_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_range_set_update_policy" XG_POST, gxg_gtk_range_set_update_policy, 2, 0, 0, H_gtk_range_set_update_policy);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_range_get_update_policy" XG_POST, gxg_gtk_range_get_update_policy, 1, 0, 0, H_gtk_range_get_update_policy);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_range_set_adjustment" XG_POST, gxg_gtk_range_set_adjustment, 2, 0, 0, H_gtk_range_set_adjustment);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_range_get_adjustment" XG_POST, gxg_gtk_range_get_adjustment, 1, 0, 0, H_gtk_range_get_adjustment);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_range_set_inverted" XG_POST, gxg_gtk_range_set_inverted, 2, 0, 0, H_gtk_range_set_inverted);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_range_get_inverted" XG_POST, gxg_gtk_range_get_inverted, 1, 0, 0, H_gtk_range_get_inverted);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_range_set_increments" XG_POST, gxg_gtk_range_set_increments, 3, 0, 0, H_gtk_range_set_increments);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_range_set_range" XG_POST, gxg_gtk_range_set_range, 3, 0, 0, H_gtk_range_set_range);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_range_set_value" XG_POST, gxg_gtk_range_set_value, 2, 0, 0, H_gtk_range_set_value);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_range_get_value" XG_POST, gxg_gtk_range_get_value, 1, 0, 0, H_gtk_range_get_value);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_rc_add_default_file" XG_POST, gxg_gtk_rc_add_default_file, 1, 0, 0, H_gtk_rc_add_default_file);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_rc_set_default_files" XG_POST, gxg_gtk_rc_set_default_files, 1, 0, 0, H_gtk_rc_set_default_files);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_rc_get_default_files" XG_POST, gxg_gtk_rc_get_default_files, 0, 0, 0, H_gtk_rc_get_default_files);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_rc_get_style" XG_POST, gxg_gtk_rc_get_style, 1, 0, 0, H_gtk_rc_get_style);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_rc_get_style_by_paths" XG_POST, gxg_gtk_rc_get_style_by_paths, 4, 0, 0, H_gtk_rc_get_style_by_paths);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_rc_reparse_all_for_settings" XG_POST, gxg_gtk_rc_reparse_all_for_settings, 2, 0, 0, H_gtk_rc_reparse_all_for_settings);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_rc_find_pixmap_in_path" XG_POST, gxg_gtk_rc_find_pixmap_in_path, 3, 0, 0, H_gtk_rc_find_pixmap_in_path);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_rc_parse" XG_POST, gxg_gtk_rc_parse, 1, 0, 0, H_gtk_rc_parse);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_rc_parse_string" XG_POST, gxg_gtk_rc_parse_string, 1, 0, 0, H_gtk_rc_parse_string);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_rc_reparse_all" XG_POST, gxg_gtk_rc_reparse_all, 0, 0, 0, H_gtk_rc_reparse_all);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_rc_style_get_type" XG_POST, gxg_gtk_rc_style_get_type, 0, 0, 0, H_gtk_rc_style_get_type);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_rc_style_new" XG_POST, gxg_gtk_rc_style_new, 0, 0, 0, H_gtk_rc_style_new);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_rc_style_copy" XG_POST, gxg_gtk_rc_style_copy, 1, 0, 0, H_gtk_rc_style_copy);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_rc_style_ref" XG_POST, gxg_gtk_rc_style_ref, 1, 0, 0, H_gtk_rc_style_ref);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_rc_style_unref" XG_POST, gxg_gtk_rc_style_unref, 1, 0, 0, H_gtk_rc_style_unref);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_rc_find_module_in_path" XG_POST, gxg_gtk_rc_find_module_in_path, 1, 0, 0, H_gtk_rc_find_module_in_path);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_rc_get_theme_dir" XG_POST, gxg_gtk_rc_get_theme_dir, 0, 0, 0, H_gtk_rc_get_theme_dir);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_rc_get_module_dir" XG_POST, gxg_gtk_rc_get_module_dir, 0, 0, 0, H_gtk_rc_get_module_dir);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_rc_get_im_module_path" XG_POST, gxg_gtk_rc_get_im_module_path, 0, 0, 0, H_gtk_rc_get_im_module_path);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_rc_get_im_module_file" XG_POST, gxg_gtk_rc_get_im_module_file, 0, 0, 0, H_gtk_rc_get_im_module_file);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_rc_scanner_new" XG_POST, gxg_gtk_rc_scanner_new, 0, 0, 0, H_gtk_rc_scanner_new);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_rc_parse_color" XG_POST, gxg_gtk_rc_parse_color, 2, 0, 0, H_gtk_rc_parse_color);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_rc_parse_state" XG_POST, gxg_gtk_rc_parse_state, 2, 0, 0, H_gtk_rc_parse_state);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_rc_parse_priority" XG_POST, gxg_gtk_rc_parse_priority, 2, 0, 0, H_gtk_rc_parse_priority);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_ruler_get_type" XG_POST, gxg_gtk_ruler_get_type, 0, 0, 0, H_gtk_ruler_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_ruler_set_metric" XG_POST, gxg_gtk_ruler_set_metric, 2, 0, 0, H_gtk_ruler_set_metric);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_ruler_set_range" XG_POST, gxg_gtk_ruler_set_range, 5, 0, 0, H_gtk_ruler_set_range);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_ruler_draw_ticks" XG_POST, gxg_gtk_ruler_draw_ticks, 1, 0, 0, H_gtk_ruler_draw_ticks);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_ruler_draw_pos" XG_POST, gxg_gtk_ruler_draw_pos, 1, 0, 0, H_gtk_ruler_draw_pos);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_ruler_get_metric" XG_POST, gxg_gtk_ruler_get_metric, 1, 0, 0, H_gtk_ruler_get_metric);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_ruler_get_range" XG_POST, gxg_gtk_ruler_get_range, 1, 4, 0, H_gtk_ruler_get_range);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_scale_get_type" XG_POST, gxg_gtk_scale_get_type, 0, 0, 0, H_gtk_scale_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_scale_set_digits" XG_POST, gxg_gtk_scale_set_digits, 2, 0, 0, H_gtk_scale_set_digits);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_scale_get_digits" XG_POST, gxg_gtk_scale_get_digits, 1, 0, 0, H_gtk_scale_get_digits);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_scale_set_draw_value" XG_POST, gxg_gtk_scale_set_draw_value, 2, 0, 0, H_gtk_scale_set_draw_value);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_scale_get_draw_value" XG_POST, gxg_gtk_scale_get_draw_value, 1, 0, 0, H_gtk_scale_get_draw_value);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_scale_set_value_pos" XG_POST, gxg_gtk_scale_set_value_pos, 2, 0, 0, H_gtk_scale_set_value_pos);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_scale_get_value_pos" XG_POST, gxg_gtk_scale_get_value_pos, 1, 0, 0, H_gtk_scale_get_value_pos);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_scrollbar_get_type" XG_POST, gxg_gtk_scrollbar_get_type, 0, 0, 0, H_gtk_scrollbar_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_scrolled_window_get_type" XG_POST, gxg_gtk_scrolled_window_get_type, 0, 0, 0, H_gtk_scrolled_window_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_scrolled_window_new" XG_POST, gxg_gtk_scrolled_window_new, 2, 0, 0, H_gtk_scrolled_window_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_scrolled_window_set_hadjustment" XG_POST, gxg_gtk_scrolled_window_set_hadjustment, 2, 0, 0, H_gtk_scrolled_window_set_hadjustment);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_scrolled_window_set_vadjustment" XG_POST, gxg_gtk_scrolled_window_set_vadjustment, 2, 0, 0, H_gtk_scrolled_window_set_vadjustment);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_scrolled_window_get_hadjustment" XG_POST, gxg_gtk_scrolled_window_get_hadjustment, 1, 0, 0, H_gtk_scrolled_window_get_hadjustment);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_scrolled_window_get_vadjustment" XG_POST, gxg_gtk_scrolled_window_get_vadjustment, 1, 0, 0, H_gtk_scrolled_window_get_vadjustment);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_scrolled_window_set_policy" XG_POST, gxg_gtk_scrolled_window_set_policy, 3, 0, 0, H_gtk_scrolled_window_set_policy);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_scrolled_window_get_policy" XG_POST, gxg_gtk_scrolled_window_get_policy, 1, 2, 0, H_gtk_scrolled_window_get_policy);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_scrolled_window_set_placement" XG_POST, gxg_gtk_scrolled_window_set_placement, 2, 0, 0, H_gtk_scrolled_window_set_placement);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_scrolled_window_get_placement" XG_POST, gxg_gtk_scrolled_window_get_placement, 1, 0, 0, H_gtk_scrolled_window_get_placement);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_scrolled_window_set_shadow_type" XG_POST, gxg_gtk_scrolled_window_set_shadow_type, 2, 0, 0, H_gtk_scrolled_window_set_shadow_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_scrolled_window_get_shadow_type" XG_POST, gxg_gtk_scrolled_window_get_shadow_type, 1, 0, 0, H_gtk_scrolled_window_get_shadow_type);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_scrolled_window_add_with_viewport" XG_POST, gxg_gtk_scrolled_window_add_with_viewport, 2, 0, 0, H_gtk_scrolled_window_add_with_viewport);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_target_list_new" XG_POST, gxg_gtk_target_list_new, 2, 0, 0, H_gtk_target_list_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_target_list_ref" XG_POST, gxg_gtk_target_list_ref, 1, 0, 0, H_gtk_target_list_ref);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_target_list_unref" XG_POST, gxg_gtk_target_list_unref, 1, 0, 0, H_gtk_target_list_unref);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_target_list_add" XG_POST, gxg_gtk_target_list_add, 4, 0, 0, H_gtk_target_list_add);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_target_list_add_table" XG_POST, gxg_gtk_target_list_add_table, 3, 0, 0, H_gtk_target_list_add_table);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_target_list_remove" XG_POST, gxg_gtk_target_list_remove, 2, 0, 0, H_gtk_target_list_remove);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_target_list_find" XG_POST, gxg_gtk_target_list_find, 3, 0, 0, H_gtk_target_list_find);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_selection_owner_set" XG_POST, gxg_gtk_selection_owner_set, 3, 0, 0, H_gtk_selection_owner_set);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_selection_add_target" XG_POST, gxg_gtk_selection_add_target, 4, 0, 0, H_gtk_selection_add_target);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_selection_add_targets" XG_POST, gxg_gtk_selection_add_targets, 4, 0, 0, H_gtk_selection_add_targets);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_selection_clear_targets" XG_POST, gxg_gtk_selection_clear_targets, 2, 0, 0, H_gtk_selection_clear_targets);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_selection_convert" XG_POST, gxg_gtk_selection_convert, 4, 0, 0, H_gtk_selection_convert);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_selection_data_set" XG_POST, gxg_gtk_selection_data_set, 5, 0, 0, H_gtk_selection_data_set);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_selection_data_set_text" XG_POST, gxg_gtk_selection_data_set_text, 3, 0, 0, H_gtk_selection_data_set_text);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_selection_data_get_text" XG_POST, gxg_gtk_selection_data_get_text, 1, 0, 0, H_gtk_selection_data_get_text);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_selection_data_get_targets" XG_POST, gxg_gtk_selection_data_get_targets, 1, 2, 0, H_gtk_selection_data_get_targets);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_selection_data_targets_include_text" XG_POST, gxg_gtk_selection_data_targets_include_text, 1, 0, 0, H_gtk_selection_data_targets_include_text);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_selection_remove_all" XG_POST, gxg_gtk_selection_remove_all, 1, 0, 0, H_gtk_selection_remove_all);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_selection_clear" XG_POST, gxg_gtk_selection_clear, 2, 0, 0, H_gtk_selection_clear);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_selection_request" XG_POST, gxg_gtk_selection_request, 2, 0, 0, H_gtk_selection_request);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_selection_incr_event" XG_POST, gxg_gtk_selection_incr_event, 2, 0, 0, H_gtk_selection_incr_event);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_selection_notify" XG_POST, gxg_gtk_selection_notify, 2, 0, 0, H_gtk_selection_notify);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_selection_property_notify" XG_POST, gxg_gtk_selection_property_notify, 2, 0, 0, H_gtk_selection_property_notify);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_selection_data_copy" XG_POST, gxg_gtk_selection_data_copy, 1, 0, 0, H_gtk_selection_data_copy);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_selection_data_free" XG_POST, gxg_gtk_selection_data_free, 1, 0, 0, H_gtk_selection_data_free);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_separator_get_type" XG_POST, gxg_gtk_separator_get_type, 0, 0, 0, H_gtk_separator_get_type);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_separator_menu_item_get_type" XG_POST, gxg_gtk_separator_menu_item_get_type, 0, 0, 0, H_gtk_separator_menu_item_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_separator_menu_item_new" XG_POST, gxg_gtk_separator_menu_item_new, 0, 0, 0, H_gtk_separator_menu_item_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_settings_get_type" XG_POST, gxg_gtk_settings_get_type, 0, 0, 0, H_gtk_settings_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_settings_get_default" XG_POST, gxg_gtk_settings_get_default, 0, 0, 0, H_gtk_settings_get_default);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_settings_install_property" XG_POST, gxg_gtk_settings_install_property, 1, 0, 0, H_gtk_settings_install_property);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_settings_install_property_parser" XG_POST, gxg_gtk_settings_install_property_parser, 2, 0, 0, H_gtk_settings_install_property_parser);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_rc_property_parse_color" XG_POST, gxg_gtk_rc_property_parse_color, 3, 0, 0, H_gtk_rc_property_parse_color);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_rc_property_parse_enum" XG_POST, gxg_gtk_rc_property_parse_enum, 3, 0, 0, H_gtk_rc_property_parse_enum);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_rc_property_parse_flags" XG_POST, gxg_gtk_rc_property_parse_flags, 3, 0, 0, H_gtk_rc_property_parse_flags);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_rc_property_parse_requisition" XG_POST, gxg_gtk_rc_property_parse_requisition, 3, 0, 0, H_gtk_rc_property_parse_requisition);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_rc_property_parse_border" XG_POST, gxg_gtk_rc_property_parse_border, 3, 0, 0, H_gtk_rc_property_parse_border);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_settings_set_property_value" XG_POST, gxg_gtk_settings_set_property_value, 3, 0, 0, H_gtk_settings_set_property_value);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_settings_set_string_property" XG_POST, gxg_gtk_settings_set_string_property, 4, 0, 0, H_gtk_settings_set_string_property);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_settings_set_long_property" XG_POST, gxg_gtk_settings_set_long_property, 4, 0, 0, H_gtk_settings_set_long_property);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_settings_set_double_property" XG_POST, gxg_gtk_settings_set_double_property, 4, 0, 0, H_gtk_settings_set_double_property);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_size_group_get_type" XG_POST, gxg_gtk_size_group_get_type, 0, 0, 0, H_gtk_size_group_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_size_group_new" XG_POST, gxg_gtk_size_group_new, 1, 0, 0, H_gtk_size_group_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_size_group_set_mode" XG_POST, gxg_gtk_size_group_set_mode, 2, 0, 0, H_gtk_size_group_set_mode);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_size_group_get_mode" XG_POST, gxg_gtk_size_group_get_mode, 1, 0, 0, H_gtk_size_group_get_mode);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_size_group_add_widget" XG_POST, gxg_gtk_size_group_add_widget, 2, 0, 0, H_gtk_size_group_add_widget);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_size_group_remove_widget" XG_POST, gxg_gtk_size_group_remove_widget, 2, 0, 0, H_gtk_size_group_remove_widget);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_socket_new" XG_POST, gxg_gtk_socket_new, 0, 0, 0, H_gtk_socket_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_socket_get_type" XG_POST, gxg_gtk_socket_get_type, 0, 0, 0, H_gtk_socket_get_type);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_socket_add_id" XG_POST, gxg_gtk_socket_add_id, 2, 0, 0, H_gtk_socket_add_id);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_socket_get_id" XG_POST, gxg_gtk_socket_get_id, 1, 0, 0, H_gtk_socket_get_id);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_spin_button_get_type" XG_POST, gxg_gtk_spin_button_get_type, 0, 0, 0, H_gtk_spin_button_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_spin_button_configure" XG_POST, gxg_gtk_spin_button_configure, 4, 0, 0, H_gtk_spin_button_configure);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_spin_button_new" XG_POST, gxg_gtk_spin_button_new, 3, 0, 0, H_gtk_spin_button_new);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_spin_button_new_with_range" XG_POST, gxg_gtk_spin_button_new_with_range, 3, 0, 0, H_gtk_spin_button_new_with_range);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_spin_button_set_adjustment" XG_POST, gxg_gtk_spin_button_set_adjustment, 2, 0, 0, H_gtk_spin_button_set_adjustment);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_spin_button_get_adjustment" XG_POST, gxg_gtk_spin_button_get_adjustment, 1, 0, 0, H_gtk_spin_button_get_adjustment);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_spin_button_set_digits" XG_POST, gxg_gtk_spin_button_set_digits, 2, 0, 0, H_gtk_spin_button_set_digits);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_spin_button_get_digits" XG_POST, gxg_gtk_spin_button_get_digits, 1, 0, 0, H_gtk_spin_button_get_digits);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_spin_button_set_increments" XG_POST, gxg_gtk_spin_button_set_increments, 3, 0, 0, H_gtk_spin_button_set_increments);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_spin_button_get_increments" XG_POST, gxg_gtk_spin_button_get_increments, 1, 2, 0, H_gtk_spin_button_get_increments);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_spin_button_set_range" XG_POST, gxg_gtk_spin_button_set_range, 3, 0, 0, H_gtk_spin_button_set_range);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_spin_button_get_range" XG_POST, gxg_gtk_spin_button_get_range, 1, 2, 0, H_gtk_spin_button_get_range);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_spin_button_get_value" XG_POST, gxg_gtk_spin_button_get_value, 1, 0, 0, H_gtk_spin_button_get_value);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_spin_button_get_value_as_int" XG_POST, gxg_gtk_spin_button_get_value_as_int, 1, 0, 0, H_gtk_spin_button_get_value_as_int);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_spin_button_set_value" XG_POST, gxg_gtk_spin_button_set_value, 2, 0, 0, H_gtk_spin_button_set_value);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_spin_button_set_update_policy" XG_POST, gxg_gtk_spin_button_set_update_policy, 2, 0, 0, H_gtk_spin_button_set_update_policy);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_spin_button_get_update_policy" XG_POST, gxg_gtk_spin_button_get_update_policy, 1, 0, 0, H_gtk_spin_button_get_update_policy);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_spin_button_set_numeric" XG_POST, gxg_gtk_spin_button_set_numeric, 2, 0, 0, H_gtk_spin_button_set_numeric);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_spin_button_get_numeric" XG_POST, gxg_gtk_spin_button_get_numeric, 1, 0, 0, H_gtk_spin_button_get_numeric);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_spin_button_spin" XG_POST, gxg_gtk_spin_button_spin, 3, 0, 0, H_gtk_spin_button_spin);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_spin_button_set_wrap" XG_POST, gxg_gtk_spin_button_set_wrap, 2, 0, 0, H_gtk_spin_button_set_wrap);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_spin_button_get_wrap" XG_POST, gxg_gtk_spin_button_get_wrap, 1, 0, 0, H_gtk_spin_button_get_wrap);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_spin_button_set_snap_to_ticks" XG_POST, gxg_gtk_spin_button_set_snap_to_ticks, 2, 0, 0, H_gtk_spin_button_set_snap_to_ticks);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_spin_button_get_snap_to_ticks" XG_POST, gxg_gtk_spin_button_get_snap_to_ticks, 1, 0, 0, H_gtk_spin_button_get_snap_to_ticks);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_spin_button_update" XG_POST, gxg_gtk_spin_button_update, 1, 0, 0, H_gtk_spin_button_update);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_statusbar_get_type" XG_POST, gxg_gtk_statusbar_get_type, 0, 0, 0, H_gtk_statusbar_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_statusbar_new" XG_POST, gxg_gtk_statusbar_new, 0, 0, 0, H_gtk_statusbar_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_statusbar_get_context_id" XG_POST, gxg_gtk_statusbar_get_context_id, 2, 0, 0, H_gtk_statusbar_get_context_id);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_statusbar_push" XG_POST, gxg_gtk_statusbar_push, 3, 0, 0, H_gtk_statusbar_push);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_statusbar_pop" XG_POST, gxg_gtk_statusbar_pop, 2, 0, 0, H_gtk_statusbar_pop);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_statusbar_remove" XG_POST, gxg_gtk_statusbar_remove, 3, 0, 0, H_gtk_statusbar_remove);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_statusbar_set_has_resize_grip" XG_POST, gxg_gtk_statusbar_set_has_resize_grip, 2, 0, 0, H_gtk_statusbar_set_has_resize_grip);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_statusbar_get_has_resize_grip" XG_POST, gxg_gtk_statusbar_get_has_resize_grip, 1, 0, 0, H_gtk_statusbar_get_has_resize_grip);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_stock_add" XG_POST, gxg_gtk_stock_add, 2, 0, 0, H_gtk_stock_add);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_stock_add_static" XG_POST, gxg_gtk_stock_add_static, 2, 0, 0, H_gtk_stock_add_static);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_stock_lookup" XG_POST, gxg_gtk_stock_lookup, 2, 0, 0, H_gtk_stock_lookup);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_stock_list_ids" XG_POST, gxg_gtk_stock_list_ids, 0, 0, 0, H_gtk_stock_list_ids);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_stock_item_copy" XG_POST, gxg_gtk_stock_item_copy, 1, 0, 0, H_gtk_stock_item_copy);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_stock_item_free" XG_POST, gxg_gtk_stock_item_free, 1, 0, 0, H_gtk_stock_item_free);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_style_get_type" XG_POST, gxg_gtk_style_get_type, 0, 0, 0, H_gtk_style_get_type);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_style_new" XG_POST, gxg_gtk_style_new, 0, 0, 0, H_gtk_style_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_style_copy" XG_POST, gxg_gtk_style_copy, 1, 0, 0, H_gtk_style_copy);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_style_attach" XG_POST, gxg_gtk_style_attach, 2, 0, 0, H_gtk_style_attach);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_style_detach" XG_POST, gxg_gtk_style_detach, 1, 0, 0, H_gtk_style_detach);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_style_set_background" XG_POST, gxg_gtk_style_set_background, 3, 0, 0, H_gtk_style_set_background);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_style_apply_default_background" XG_POST, gxg_gtk_style_apply_default_background, 9, 0, 0, H_gtk_style_apply_default_background);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_style_lookup_icon_set" XG_POST, gxg_gtk_style_lookup_icon_set, 2, 0, 0, H_gtk_style_lookup_icon_set);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_style_render_icon" XG_POST, gxg_gtk_style_render_icon, 7, 0, 0, H_gtk_style_render_icon);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_paint_hline" XG_POST, gxg_gtk_paint_hline, 9, 0, 0, H_gtk_paint_hline);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_paint_vline" XG_POST, gxg_gtk_paint_vline, 9, 0, 0, H_gtk_paint_vline);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_paint_shadow" XG_POST, gxg_gtk_paint_shadow, 0, 0, 1, H_gtk_paint_shadow);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_paint_polygon" XG_POST, gxg_gtk_paint_polygon, 0, 0, 1, H_gtk_paint_polygon);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_paint_arrow" XG_POST, gxg_gtk_paint_arrow, 0, 0, 1, H_gtk_paint_arrow);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_paint_diamond" XG_POST, gxg_gtk_paint_diamond, 0, 0, 1, H_gtk_paint_diamond);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_paint_box" XG_POST, gxg_gtk_paint_box, 0, 0, 1, H_gtk_paint_box);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_paint_flat_box" XG_POST, gxg_gtk_paint_flat_box, 0, 0, 1, H_gtk_paint_flat_box);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_paint_check" XG_POST, gxg_gtk_paint_check, 0, 0, 1, H_gtk_paint_check);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_paint_option" XG_POST, gxg_gtk_paint_option, 0, 0, 1, H_gtk_paint_option);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_paint_tab" XG_POST, gxg_gtk_paint_tab, 0, 0, 1, H_gtk_paint_tab);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_paint_shadow_gap" XG_POST, gxg_gtk_paint_shadow_gap, 0, 0, 1, H_gtk_paint_shadow_gap);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_paint_box_gap" XG_POST, gxg_gtk_paint_box_gap, 0, 0, 1, H_gtk_paint_box_gap);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_paint_extension" XG_POST, gxg_gtk_paint_extension, 0, 0, 1, H_gtk_paint_extension);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_paint_focus" XG_POST, gxg_gtk_paint_focus, 0, 0, 1, H_gtk_paint_focus);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_paint_slider" XG_POST, gxg_gtk_paint_slider, 0, 0, 1, H_gtk_paint_slider);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_paint_handle" XG_POST, gxg_gtk_paint_handle, 0, 0, 1, H_gtk_paint_handle);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_paint_expander" XG_POST, gxg_gtk_paint_expander, 9, 0, 0, H_gtk_paint_expander);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_paint_layout" XG_POST, gxg_gtk_paint_layout, 0, 0, 1, H_gtk_paint_layout);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_paint_resize_grip" XG_POST, gxg_gtk_paint_resize_grip, 0, 0, 1, H_gtk_paint_resize_grip);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_border_copy" XG_POST, gxg_gtk_border_copy, 1, 0, 0, H_gtk_border_copy);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_border_free" XG_POST, gxg_gtk_border_free, 1, 0, 0, H_gtk_border_free);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_table_get_type" XG_POST, gxg_gtk_table_get_type, 0, 0, 0, H_gtk_table_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_table_new" XG_POST, gxg_gtk_table_new, 3, 0, 0, H_gtk_table_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_table_resize" XG_POST, gxg_gtk_table_resize, 3, 0, 0, H_gtk_table_resize);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_table_attach" XG_POST, gxg_gtk_table_attach, 0, 0, 1, H_gtk_table_attach);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_table_attach_defaults" XG_POST, gxg_gtk_table_attach_defaults, 6, 0, 0, H_gtk_table_attach_defaults);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_table_set_row_spacing" XG_POST, gxg_gtk_table_set_row_spacing, 3, 0, 0, H_gtk_table_set_row_spacing);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_table_get_row_spacing" XG_POST, gxg_gtk_table_get_row_spacing, 2, 0, 0, H_gtk_table_get_row_spacing);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_table_set_col_spacing" XG_POST, gxg_gtk_table_set_col_spacing, 3, 0, 0, H_gtk_table_set_col_spacing);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_table_get_col_spacing" XG_POST, gxg_gtk_table_get_col_spacing, 2, 0, 0, H_gtk_table_get_col_spacing);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_table_set_row_spacings" XG_POST, gxg_gtk_table_set_row_spacings, 2, 0, 0, H_gtk_table_set_row_spacings);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_table_get_default_row_spacing" XG_POST, gxg_gtk_table_get_default_row_spacing, 1, 0, 0, H_gtk_table_get_default_row_spacing);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_table_set_col_spacings" XG_POST, gxg_gtk_table_set_col_spacings, 2, 0, 0, H_gtk_table_set_col_spacings);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_table_get_default_col_spacing" XG_POST, gxg_gtk_table_get_default_col_spacing, 1, 0, 0, H_gtk_table_get_default_col_spacing);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_table_set_homogeneous" XG_POST, gxg_gtk_table_set_homogeneous, 2, 0, 0, H_gtk_table_set_homogeneous);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_table_get_homogeneous" XG_POST, gxg_gtk_table_get_homogeneous, 1, 0, 0, H_gtk_table_get_homogeneous);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tearoff_menu_item_get_type" XG_POST, gxg_gtk_tearoff_menu_item_get_type, 0, 0, 0, H_gtk_tearoff_menu_item_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tearoff_menu_item_new" XG_POST, gxg_gtk_tearoff_menu_item_new, 0, 0, 0, H_gtk_tearoff_menu_item_new);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_buffer_get_type" XG_POST, gxg_gtk_text_buffer_get_type, 0, 0, 0, H_gtk_text_buffer_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_buffer_new" XG_POST, gxg_gtk_text_buffer_new, 1, 0, 0, H_gtk_text_buffer_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_buffer_get_line_count" XG_POST, gxg_gtk_text_buffer_get_line_count, 1, 0, 0, H_gtk_text_buffer_get_line_count);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_buffer_get_char_count" XG_POST, gxg_gtk_text_buffer_get_char_count, 1, 0, 0, H_gtk_text_buffer_get_char_count);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_buffer_get_tag_table" XG_POST, gxg_gtk_text_buffer_get_tag_table, 1, 0, 0, H_gtk_text_buffer_get_tag_table);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_buffer_set_text" XG_POST, gxg_gtk_text_buffer_set_text, 3, 0, 0, H_gtk_text_buffer_set_text);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_buffer_insert" XG_POST, gxg_gtk_text_buffer_insert, 4, 0, 0, H_gtk_text_buffer_insert);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_buffer_insert_at_cursor" XG_POST, gxg_gtk_text_buffer_insert_at_cursor, 3, 0, 0, H_gtk_text_buffer_insert_at_cursor);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_buffer_insert_interactive" XG_POST, gxg_gtk_text_buffer_insert_interactive, 5, 0, 0, H_gtk_text_buffer_insert_interactive);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_buffer_insert_interactive_at_cursor" XG_POST, gxg_gtk_text_buffer_insert_interactive_at_cursor, 4, 0, 0, H_gtk_text_buffer_insert_interactive_at_cursor);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_buffer_insert_range" XG_POST, gxg_gtk_text_buffer_insert_range, 4, 0, 0, H_gtk_text_buffer_insert_range);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_buffer_insert_range_interactive" XG_POST, gxg_gtk_text_buffer_insert_range_interactive, 5, 0, 0, H_gtk_text_buffer_insert_range_interactive);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_buffer_insert_with_tags" XG_POST, gxg_gtk_text_buffer_insert_with_tags, 5, 0, 0, H_gtk_text_buffer_insert_with_tags);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_buffer_insert_with_tags_by_name" XG_POST, gxg_gtk_text_buffer_insert_with_tags_by_name, 5, 0, 0, H_gtk_text_buffer_insert_with_tags_by_name);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_buffer_delete" XG_POST, gxg_gtk_text_buffer_delete, 3, 0, 0, H_gtk_text_buffer_delete);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_buffer_delete_interactive" XG_POST, gxg_gtk_text_buffer_delete_interactive, 4, 0, 0, H_gtk_text_buffer_delete_interactive);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_buffer_get_text" XG_POST, gxg_gtk_text_buffer_get_text, 4, 0, 0, H_gtk_text_buffer_get_text);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_buffer_get_slice" XG_POST, gxg_gtk_text_buffer_get_slice, 4, 0, 0, H_gtk_text_buffer_get_slice);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_buffer_insert_pixbuf" XG_POST, gxg_gtk_text_buffer_insert_pixbuf, 3, 0, 0, H_gtk_text_buffer_insert_pixbuf);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_buffer_insert_child_anchor" XG_POST, gxg_gtk_text_buffer_insert_child_anchor, 3, 0, 0, H_gtk_text_buffer_insert_child_anchor);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_buffer_create_child_anchor" XG_POST, gxg_gtk_text_buffer_create_child_anchor, 2, 0, 0, H_gtk_text_buffer_create_child_anchor);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_buffer_create_mark" XG_POST, gxg_gtk_text_buffer_create_mark, 4, 0, 0, H_gtk_text_buffer_create_mark);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_buffer_move_mark" XG_POST, gxg_gtk_text_buffer_move_mark, 3, 0, 0, H_gtk_text_buffer_move_mark);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_buffer_delete_mark" XG_POST, gxg_gtk_text_buffer_delete_mark, 2, 0, 0, H_gtk_text_buffer_delete_mark);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_buffer_get_mark" XG_POST, gxg_gtk_text_buffer_get_mark, 2, 0, 0, H_gtk_text_buffer_get_mark);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_buffer_move_mark_by_name" XG_POST, gxg_gtk_text_buffer_move_mark_by_name, 3, 0, 0, H_gtk_text_buffer_move_mark_by_name);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_buffer_delete_mark_by_name" XG_POST, gxg_gtk_text_buffer_delete_mark_by_name, 2, 0, 0, H_gtk_text_buffer_delete_mark_by_name);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_buffer_get_insert" XG_POST, gxg_gtk_text_buffer_get_insert, 1, 0, 0, H_gtk_text_buffer_get_insert);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_buffer_get_selection_bound" XG_POST, gxg_gtk_text_buffer_get_selection_bound, 1, 0, 0, H_gtk_text_buffer_get_selection_bound);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_buffer_place_cursor" XG_POST, gxg_gtk_text_buffer_place_cursor, 2, 0, 0, H_gtk_text_buffer_place_cursor);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_buffer_apply_tag" XG_POST, gxg_gtk_text_buffer_apply_tag, 4, 0, 0, H_gtk_text_buffer_apply_tag);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_buffer_remove_tag" XG_POST, gxg_gtk_text_buffer_remove_tag, 4, 0, 0, H_gtk_text_buffer_remove_tag);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_buffer_apply_tag_by_name" XG_POST, gxg_gtk_text_buffer_apply_tag_by_name, 4, 0, 0, H_gtk_text_buffer_apply_tag_by_name);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_buffer_remove_tag_by_name" XG_POST, gxg_gtk_text_buffer_remove_tag_by_name, 4, 0, 0, H_gtk_text_buffer_remove_tag_by_name);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_buffer_remove_all_tags" XG_POST, gxg_gtk_text_buffer_remove_all_tags, 3, 0, 0, H_gtk_text_buffer_remove_all_tags);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_buffer_create_tag" XG_POST, gxg_gtk_text_buffer_create_tag, 3, 0, 0, H_gtk_text_buffer_create_tag);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_buffer_get_iter_at_line_offset" XG_POST, gxg_gtk_text_buffer_get_iter_at_line_offset, 4, 0, 0, H_gtk_text_buffer_get_iter_at_line_offset);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_buffer_get_iter_at_line_index" XG_POST, gxg_gtk_text_buffer_get_iter_at_line_index, 4, 0, 0, H_gtk_text_buffer_get_iter_at_line_index);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_buffer_get_iter_at_offset" XG_POST, gxg_gtk_text_buffer_get_iter_at_offset, 3, 0, 0, H_gtk_text_buffer_get_iter_at_offset);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_buffer_get_iter_at_line" XG_POST, gxg_gtk_text_buffer_get_iter_at_line, 3, 0, 0, H_gtk_text_buffer_get_iter_at_line);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_buffer_get_start_iter" XG_POST, gxg_gtk_text_buffer_get_start_iter, 2, 0, 0, H_gtk_text_buffer_get_start_iter);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_buffer_get_end_iter" XG_POST, gxg_gtk_text_buffer_get_end_iter, 2, 0, 0, H_gtk_text_buffer_get_end_iter);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_buffer_get_bounds" XG_POST, gxg_gtk_text_buffer_get_bounds, 3, 0, 0, H_gtk_text_buffer_get_bounds);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_buffer_get_iter_at_mark" XG_POST, gxg_gtk_text_buffer_get_iter_at_mark, 3, 0, 0, H_gtk_text_buffer_get_iter_at_mark);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_buffer_get_iter_at_child_anchor" XG_POST, gxg_gtk_text_buffer_get_iter_at_child_anchor, 3, 0, 0, H_gtk_text_buffer_get_iter_at_child_anchor);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_buffer_get_modified" XG_POST, gxg_gtk_text_buffer_get_modified, 1, 0, 0, H_gtk_text_buffer_get_modified);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_buffer_set_modified" XG_POST, gxg_gtk_text_buffer_set_modified, 2, 0, 0, H_gtk_text_buffer_set_modified);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_buffer_add_selection_clipboard" XG_POST, gxg_gtk_text_buffer_add_selection_clipboard, 2, 0, 0, H_gtk_text_buffer_add_selection_clipboard);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_buffer_remove_selection_clipboard" XG_POST, gxg_gtk_text_buffer_remove_selection_clipboard, 2, 0, 0, H_gtk_text_buffer_remove_selection_clipboard);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_buffer_cut_clipboard" XG_POST, gxg_gtk_text_buffer_cut_clipboard, 3, 0, 0, H_gtk_text_buffer_cut_clipboard);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_buffer_copy_clipboard" XG_POST, gxg_gtk_text_buffer_copy_clipboard, 2, 0, 0, H_gtk_text_buffer_copy_clipboard);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_buffer_paste_clipboard" XG_POST, gxg_gtk_text_buffer_paste_clipboard, 4, 0, 0, H_gtk_text_buffer_paste_clipboard);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_buffer_get_selection_bounds" XG_POST, gxg_gtk_text_buffer_get_selection_bounds, 3, 0, 0, H_gtk_text_buffer_get_selection_bounds);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_buffer_delete_selection" XG_POST, gxg_gtk_text_buffer_delete_selection, 3, 0, 0, H_gtk_text_buffer_delete_selection);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_buffer_begin_user_action" XG_POST, gxg_gtk_text_buffer_begin_user_action, 1, 0, 0, H_gtk_text_buffer_begin_user_action);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_buffer_end_user_action" XG_POST, gxg_gtk_text_buffer_end_user_action, 1, 0, 0, H_gtk_text_buffer_end_user_action);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_child_anchor_get_type" XG_POST, gxg_gtk_text_child_anchor_get_type, 0, 0, 0, H_gtk_text_child_anchor_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_child_anchor_new" XG_POST, gxg_gtk_text_child_anchor_new, 0, 0, 0, H_gtk_text_child_anchor_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_child_anchor_get_widgets" XG_POST, gxg_gtk_text_child_anchor_get_widgets, 1, 0, 0, H_gtk_text_child_anchor_get_widgets);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_child_anchor_get_deleted" XG_POST, gxg_gtk_text_child_anchor_get_deleted, 1, 0, 0, H_gtk_text_child_anchor_get_deleted);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_iter_get_buffer" XG_POST, gxg_gtk_text_iter_get_buffer, 1, 0, 0, H_gtk_text_iter_get_buffer);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_iter_copy" XG_POST, gxg_gtk_text_iter_copy, 1, 0, 0, H_gtk_text_iter_copy);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_iter_free" XG_POST, gxg_gtk_text_iter_free, 1, 0, 0, H_gtk_text_iter_free);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_iter_get_offset" XG_POST, gxg_gtk_text_iter_get_offset, 1, 0, 0, H_gtk_text_iter_get_offset);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_iter_get_line" XG_POST, gxg_gtk_text_iter_get_line, 1, 0, 0, H_gtk_text_iter_get_line);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_iter_get_line_offset" XG_POST, gxg_gtk_text_iter_get_line_offset, 1, 0, 0, H_gtk_text_iter_get_line_offset);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_iter_get_line_index" XG_POST, gxg_gtk_text_iter_get_line_index, 1, 0, 0, H_gtk_text_iter_get_line_index);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_iter_get_visible_line_offset" XG_POST, gxg_gtk_text_iter_get_visible_line_offset, 1, 0, 0, H_gtk_text_iter_get_visible_line_offset);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_iter_get_visible_line_index" XG_POST, gxg_gtk_text_iter_get_visible_line_index, 1, 0, 0, H_gtk_text_iter_get_visible_line_index);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_iter_get_char" XG_POST, gxg_gtk_text_iter_get_char, 1, 0, 0, H_gtk_text_iter_get_char);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_iter_get_slice" XG_POST, gxg_gtk_text_iter_get_slice, 2, 0, 0, H_gtk_text_iter_get_slice);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_iter_get_text" XG_POST, gxg_gtk_text_iter_get_text, 2, 0, 0, H_gtk_text_iter_get_text);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_iter_get_visible_slice" XG_POST, gxg_gtk_text_iter_get_visible_slice, 2, 0, 0, H_gtk_text_iter_get_visible_slice);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_iter_get_visible_text" XG_POST, gxg_gtk_text_iter_get_visible_text, 2, 0, 0, H_gtk_text_iter_get_visible_text);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_iter_get_pixbuf" XG_POST, gxg_gtk_text_iter_get_pixbuf, 1, 0, 0, H_gtk_text_iter_get_pixbuf);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_iter_get_marks" XG_POST, gxg_gtk_text_iter_get_marks, 1, 0, 0, H_gtk_text_iter_get_marks);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_iter_get_child_anchor" XG_POST, gxg_gtk_text_iter_get_child_anchor, 1, 0, 0, H_gtk_text_iter_get_child_anchor);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_iter_get_toggled_tags" XG_POST, gxg_gtk_text_iter_get_toggled_tags, 2, 0, 0, H_gtk_text_iter_get_toggled_tags);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_iter_begins_tag" XG_POST, gxg_gtk_text_iter_begins_tag, 2, 0, 0, H_gtk_text_iter_begins_tag);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_iter_ends_tag" XG_POST, gxg_gtk_text_iter_ends_tag, 2, 0, 0, H_gtk_text_iter_ends_tag);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_iter_toggles_tag" XG_POST, gxg_gtk_text_iter_toggles_tag, 2, 0, 0, H_gtk_text_iter_toggles_tag);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_iter_has_tag" XG_POST, gxg_gtk_text_iter_has_tag, 2, 0, 0, H_gtk_text_iter_has_tag);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_iter_get_tags" XG_POST, gxg_gtk_text_iter_get_tags, 1, 0, 0, H_gtk_text_iter_get_tags);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_iter_editable" XG_POST, gxg_gtk_text_iter_editable, 2, 0, 0, H_gtk_text_iter_editable);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_iter_can_insert" XG_POST, gxg_gtk_text_iter_can_insert, 2, 0, 0, H_gtk_text_iter_can_insert);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_iter_starts_word" XG_POST, gxg_gtk_text_iter_starts_word, 1, 0, 0, H_gtk_text_iter_starts_word);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_iter_ends_word" XG_POST, gxg_gtk_text_iter_ends_word, 1, 0, 0, H_gtk_text_iter_ends_word);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_iter_inside_word" XG_POST, gxg_gtk_text_iter_inside_word, 1, 0, 0, H_gtk_text_iter_inside_word);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_iter_starts_sentence" XG_POST, gxg_gtk_text_iter_starts_sentence, 1, 0, 0, H_gtk_text_iter_starts_sentence);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_iter_ends_sentence" XG_POST, gxg_gtk_text_iter_ends_sentence, 1, 0, 0, H_gtk_text_iter_ends_sentence);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_iter_inside_sentence" XG_POST, gxg_gtk_text_iter_inside_sentence, 1, 0, 0, H_gtk_text_iter_inside_sentence);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_iter_starts_line" XG_POST, gxg_gtk_text_iter_starts_line, 1, 0, 0, H_gtk_text_iter_starts_line);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_iter_ends_line" XG_POST, gxg_gtk_text_iter_ends_line, 1, 0, 0, H_gtk_text_iter_ends_line);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_iter_is_cursor_position" XG_POST, gxg_gtk_text_iter_is_cursor_position, 1, 0, 0, H_gtk_text_iter_is_cursor_position);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_iter_get_chars_in_line" XG_POST, gxg_gtk_text_iter_get_chars_in_line, 1, 0, 0, H_gtk_text_iter_get_chars_in_line);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_iter_get_bytes_in_line" XG_POST, gxg_gtk_text_iter_get_bytes_in_line, 1, 0, 0, H_gtk_text_iter_get_bytes_in_line);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_iter_get_attributes" XG_POST, gxg_gtk_text_iter_get_attributes, 2, 0, 0, H_gtk_text_iter_get_attributes);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_iter_get_language" XG_POST, gxg_gtk_text_iter_get_language, 1, 0, 0, H_gtk_text_iter_get_language);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_iter_is_end" XG_POST, gxg_gtk_text_iter_is_end, 1, 0, 0, H_gtk_text_iter_is_end);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_iter_is_start" XG_POST, gxg_gtk_text_iter_is_start, 1, 0, 0, H_gtk_text_iter_is_start);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_iter_forward_char" XG_POST, gxg_gtk_text_iter_forward_char, 1, 0, 0, H_gtk_text_iter_forward_char);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_iter_backward_char" XG_POST, gxg_gtk_text_iter_backward_char, 1, 0, 0, H_gtk_text_iter_backward_char);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_iter_forward_chars" XG_POST, gxg_gtk_text_iter_forward_chars, 2, 0, 0, H_gtk_text_iter_forward_chars);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_iter_backward_chars" XG_POST, gxg_gtk_text_iter_backward_chars, 2, 0, 0, H_gtk_text_iter_backward_chars);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_iter_forward_line" XG_POST, gxg_gtk_text_iter_forward_line, 1, 0, 0, H_gtk_text_iter_forward_line);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_iter_backward_line" XG_POST, gxg_gtk_text_iter_backward_line, 1, 0, 0, H_gtk_text_iter_backward_line);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_iter_forward_lines" XG_POST, gxg_gtk_text_iter_forward_lines, 2, 0, 0, H_gtk_text_iter_forward_lines);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_iter_backward_lines" XG_POST, gxg_gtk_text_iter_backward_lines, 2, 0, 0, H_gtk_text_iter_backward_lines);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_iter_forward_word_end" XG_POST, gxg_gtk_text_iter_forward_word_end, 1, 0, 0, H_gtk_text_iter_forward_word_end);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_iter_backward_word_start" XG_POST, gxg_gtk_text_iter_backward_word_start, 1, 0, 0, H_gtk_text_iter_backward_word_start);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_iter_forward_word_ends" XG_POST, gxg_gtk_text_iter_forward_word_ends, 2, 0, 0, H_gtk_text_iter_forward_word_ends);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_iter_backward_word_starts" XG_POST, gxg_gtk_text_iter_backward_word_starts, 2, 0, 0, H_gtk_text_iter_backward_word_starts);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_iter_forward_sentence_end" XG_POST, gxg_gtk_text_iter_forward_sentence_end, 1, 0, 0, H_gtk_text_iter_forward_sentence_end);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_iter_backward_sentence_start" XG_POST, gxg_gtk_text_iter_backward_sentence_start, 1, 0, 0, H_gtk_text_iter_backward_sentence_start);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_iter_forward_sentence_ends" XG_POST, gxg_gtk_text_iter_forward_sentence_ends, 2, 0, 0, H_gtk_text_iter_forward_sentence_ends);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_iter_backward_sentence_starts" XG_POST, gxg_gtk_text_iter_backward_sentence_starts, 2, 0, 0, H_gtk_text_iter_backward_sentence_starts);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_iter_forward_cursor_position" XG_POST, gxg_gtk_text_iter_forward_cursor_position, 1, 0, 0, H_gtk_text_iter_forward_cursor_position);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_iter_backward_cursor_position" XG_POST, gxg_gtk_text_iter_backward_cursor_position, 1, 0, 0, H_gtk_text_iter_backward_cursor_position);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_iter_forward_cursor_positions" XG_POST, gxg_gtk_text_iter_forward_cursor_positions, 2, 0, 0, H_gtk_text_iter_forward_cursor_positions);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_iter_backward_cursor_positions" XG_POST, gxg_gtk_text_iter_backward_cursor_positions, 2, 0, 0, H_gtk_text_iter_backward_cursor_positions);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_iter_set_offset" XG_POST, gxg_gtk_text_iter_set_offset, 2, 0, 0, H_gtk_text_iter_set_offset);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_iter_set_line" XG_POST, gxg_gtk_text_iter_set_line, 2, 0, 0, H_gtk_text_iter_set_line);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_iter_set_line_offset" XG_POST, gxg_gtk_text_iter_set_line_offset, 2, 0, 0, H_gtk_text_iter_set_line_offset);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_iter_set_line_index" XG_POST, gxg_gtk_text_iter_set_line_index, 2, 0, 0, H_gtk_text_iter_set_line_index);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_iter_forward_to_end" XG_POST, gxg_gtk_text_iter_forward_to_end, 1, 0, 0, H_gtk_text_iter_forward_to_end);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_iter_forward_to_line_end" XG_POST, gxg_gtk_text_iter_forward_to_line_end, 1, 0, 0, H_gtk_text_iter_forward_to_line_end);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_iter_set_visible_line_offset" XG_POST, gxg_gtk_text_iter_set_visible_line_offset, 2, 0, 0, H_gtk_text_iter_set_visible_line_offset);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_iter_set_visible_line_index" XG_POST, gxg_gtk_text_iter_set_visible_line_index, 2, 0, 0, H_gtk_text_iter_set_visible_line_index);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_iter_forward_to_tag_toggle" XG_POST, gxg_gtk_text_iter_forward_to_tag_toggle, 2, 0, 0, H_gtk_text_iter_forward_to_tag_toggle);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_iter_backward_to_tag_toggle" XG_POST, gxg_gtk_text_iter_backward_to_tag_toggle, 2, 0, 0, H_gtk_text_iter_backward_to_tag_toggle);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_iter_forward_find_char" XG_POST, gxg_gtk_text_iter_forward_find_char, 4, 0, 0, H_gtk_text_iter_forward_find_char);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_iter_backward_find_char" XG_POST, gxg_gtk_text_iter_backward_find_char, 4, 0, 0, H_gtk_text_iter_backward_find_char);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_iter_forward_search" XG_POST, gxg_gtk_text_iter_forward_search, 6, 0, 0, H_gtk_text_iter_forward_search);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_iter_backward_search" XG_POST, gxg_gtk_text_iter_backward_search, 6, 0, 0, H_gtk_text_iter_backward_search);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_iter_equal" XG_POST, gxg_gtk_text_iter_equal, 2, 0, 0, H_gtk_text_iter_equal);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_iter_compare" XG_POST, gxg_gtk_text_iter_compare, 2, 0, 0, H_gtk_text_iter_compare);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_iter_in_range" XG_POST, gxg_gtk_text_iter_in_range, 3, 0, 0, H_gtk_text_iter_in_range);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_iter_order" XG_POST, gxg_gtk_text_iter_order, 2, 0, 0, H_gtk_text_iter_order);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_mark_get_type" XG_POST, gxg_gtk_text_mark_get_type, 0, 0, 0, H_gtk_text_mark_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_mark_set_visible" XG_POST, gxg_gtk_text_mark_set_visible, 2, 0, 0, H_gtk_text_mark_set_visible);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_mark_get_visible" XG_POST, gxg_gtk_text_mark_get_visible, 1, 0, 0, H_gtk_text_mark_get_visible);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_mark_get_name" XG_POST, gxg_gtk_text_mark_get_name, 1, 0, 0, H_gtk_text_mark_get_name);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_mark_get_deleted" XG_POST, gxg_gtk_text_mark_get_deleted, 1, 0, 0, H_gtk_text_mark_get_deleted);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_mark_get_buffer" XG_POST, gxg_gtk_text_mark_get_buffer, 1, 0, 0, H_gtk_text_mark_get_buffer);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_mark_get_left_gravity" XG_POST, gxg_gtk_text_mark_get_left_gravity, 1, 0, 0, H_gtk_text_mark_get_left_gravity);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_tag_get_type" XG_POST, gxg_gtk_text_tag_get_type, 0, 0, 0, H_gtk_text_tag_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_tag_new" XG_POST, gxg_gtk_text_tag_new, 1, 0, 0, H_gtk_text_tag_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_tag_get_priority" XG_POST, gxg_gtk_text_tag_get_priority, 1, 0, 0, H_gtk_text_tag_get_priority);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_tag_set_priority" XG_POST, gxg_gtk_text_tag_set_priority, 2, 0, 0, H_gtk_text_tag_set_priority);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_tag_event" XG_POST, gxg_gtk_text_tag_event, 4, 0, 0, H_gtk_text_tag_event);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_attributes_new" XG_POST, gxg_gtk_text_attributes_new, 0, 0, 0, H_gtk_text_attributes_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_attributes_copy" XG_POST, gxg_gtk_text_attributes_copy, 1, 0, 0, H_gtk_text_attributes_copy);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_attributes_copy_values" XG_POST, gxg_gtk_text_attributes_copy_values, 2, 0, 0, H_gtk_text_attributes_copy_values);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_attributes_unref" XG_POST, gxg_gtk_text_attributes_unref, 1, 0, 0, H_gtk_text_attributes_unref);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_attributes_ref" XG_POST, gxg_gtk_text_attributes_ref, 1, 0, 0, H_gtk_text_attributes_ref);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_tag_table_get_type" XG_POST, gxg_gtk_text_tag_table_get_type, 0, 0, 0, H_gtk_text_tag_table_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_tag_table_new" XG_POST, gxg_gtk_text_tag_table_new, 0, 0, 0, H_gtk_text_tag_table_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_tag_table_add" XG_POST, gxg_gtk_text_tag_table_add, 2, 0, 0, H_gtk_text_tag_table_add);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_tag_table_remove" XG_POST, gxg_gtk_text_tag_table_remove, 2, 0, 0, H_gtk_text_tag_table_remove);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_tag_table_lookup" XG_POST, gxg_gtk_text_tag_table_lookup, 2, 0, 0, H_gtk_text_tag_table_lookup);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_tag_table_foreach" XG_POST, gxg_gtk_text_tag_table_foreach, 3, 0, 0, H_gtk_text_tag_table_foreach);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_tag_table_get_size" XG_POST, gxg_gtk_text_tag_table_get_size, 1, 0, 0, H_gtk_text_tag_table_get_size);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_view_get_type" XG_POST, gxg_gtk_text_view_get_type, 0, 0, 0, H_gtk_text_view_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_view_new" XG_POST, gxg_gtk_text_view_new, 0, 0, 0, H_gtk_text_view_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_view_new_with_buffer" XG_POST, gxg_gtk_text_view_new_with_buffer, 1, 0, 0, H_gtk_text_view_new_with_buffer);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_view_set_buffer" XG_POST, gxg_gtk_text_view_set_buffer, 2, 0, 0, H_gtk_text_view_set_buffer);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_view_get_buffer" XG_POST, gxg_gtk_text_view_get_buffer, 1, 0, 0, H_gtk_text_view_get_buffer);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_view_scroll_to_iter" XG_POST, gxg_gtk_text_view_scroll_to_iter, 6, 0, 0, H_gtk_text_view_scroll_to_iter);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_view_scroll_to_mark" XG_POST, gxg_gtk_text_view_scroll_to_mark, 6, 0, 0, H_gtk_text_view_scroll_to_mark);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_view_scroll_mark_onscreen" XG_POST, gxg_gtk_text_view_scroll_mark_onscreen, 2, 0, 0, H_gtk_text_view_scroll_mark_onscreen);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_view_move_mark_onscreen" XG_POST, gxg_gtk_text_view_move_mark_onscreen, 2, 0, 0, H_gtk_text_view_move_mark_onscreen);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_view_place_cursor_onscreen" XG_POST, gxg_gtk_text_view_place_cursor_onscreen, 1, 0, 0, H_gtk_text_view_place_cursor_onscreen);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_view_get_visible_rect" XG_POST, gxg_gtk_text_view_get_visible_rect, 2, 0, 0, H_gtk_text_view_get_visible_rect);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_view_set_cursor_visible" XG_POST, gxg_gtk_text_view_set_cursor_visible, 2, 0, 0, H_gtk_text_view_set_cursor_visible);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_view_get_cursor_visible" XG_POST, gxg_gtk_text_view_get_cursor_visible, 1, 0, 0, H_gtk_text_view_get_cursor_visible);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_view_get_iter_location" XG_POST, gxg_gtk_text_view_get_iter_location, 3, 0, 0, H_gtk_text_view_get_iter_location);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_view_get_iter_at_location" XG_POST, gxg_gtk_text_view_get_iter_at_location, 4, 0, 0, H_gtk_text_view_get_iter_at_location);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_view_get_line_yrange" XG_POST, gxg_gtk_text_view_get_line_yrange, 4, 0, 0, H_gtk_text_view_get_line_yrange);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_view_get_line_at_y" XG_POST, gxg_gtk_text_view_get_line_at_y, 3, 1, 0, H_gtk_text_view_get_line_at_y);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_view_buffer_to_window_coords" XG_POST, gxg_gtk_text_view_buffer_to_window_coords, 4, 2, 0, H_gtk_text_view_buffer_to_window_coords);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_view_window_to_buffer_coords" XG_POST, gxg_gtk_text_view_window_to_buffer_coords, 4, 2, 0, H_gtk_text_view_window_to_buffer_coords);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_view_get_window" XG_POST, gxg_gtk_text_view_get_window, 2, 0, 0, H_gtk_text_view_get_window);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_view_get_window_type" XG_POST, gxg_gtk_text_view_get_window_type, 2, 0, 0, H_gtk_text_view_get_window_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_view_set_border_window_size" XG_POST, gxg_gtk_text_view_set_border_window_size, 3, 0, 0, H_gtk_text_view_set_border_window_size);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_view_get_border_window_size" XG_POST, gxg_gtk_text_view_get_border_window_size, 2, 0, 0, H_gtk_text_view_get_border_window_size);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_view_forward_display_line" XG_POST, gxg_gtk_text_view_forward_display_line, 2, 0, 0, H_gtk_text_view_forward_display_line);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_view_backward_display_line" XG_POST, gxg_gtk_text_view_backward_display_line, 2, 0, 0, H_gtk_text_view_backward_display_line);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_view_forward_display_line_end" XG_POST, gxg_gtk_text_view_forward_display_line_end, 2, 0, 0, H_gtk_text_view_forward_display_line_end);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_view_backward_display_line_start" XG_POST, gxg_gtk_text_view_backward_display_line_start, 2, 0, 0, H_gtk_text_view_backward_display_line_start);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_view_starts_display_line" XG_POST, gxg_gtk_text_view_starts_display_line, 2, 0, 0, H_gtk_text_view_starts_display_line);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_view_move_visually" XG_POST, gxg_gtk_text_view_move_visually, 3, 0, 0, H_gtk_text_view_move_visually);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_view_add_child_at_anchor" XG_POST, gxg_gtk_text_view_add_child_at_anchor, 3, 0, 0, H_gtk_text_view_add_child_at_anchor);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_view_add_child_in_window" XG_POST, gxg_gtk_text_view_add_child_in_window, 5, 0, 0, H_gtk_text_view_add_child_in_window);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_view_move_child" XG_POST, gxg_gtk_text_view_move_child, 4, 0, 0, H_gtk_text_view_move_child);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_view_set_wrap_mode" XG_POST, gxg_gtk_text_view_set_wrap_mode, 2, 0, 0, H_gtk_text_view_set_wrap_mode);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_view_get_wrap_mode" XG_POST, gxg_gtk_text_view_get_wrap_mode, 1, 0, 0, H_gtk_text_view_get_wrap_mode);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_view_set_editable" XG_POST, gxg_gtk_text_view_set_editable, 2, 0, 0, H_gtk_text_view_set_editable);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_view_get_editable" XG_POST, gxg_gtk_text_view_get_editable, 1, 0, 0, H_gtk_text_view_get_editable);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_view_set_pixels_above_lines" XG_POST, gxg_gtk_text_view_set_pixels_above_lines, 2, 0, 0, H_gtk_text_view_set_pixels_above_lines);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_view_get_pixels_above_lines" XG_POST, gxg_gtk_text_view_get_pixels_above_lines, 1, 0, 0, H_gtk_text_view_get_pixels_above_lines);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_view_set_pixels_below_lines" XG_POST, gxg_gtk_text_view_set_pixels_below_lines, 2, 0, 0, H_gtk_text_view_set_pixels_below_lines);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_view_get_pixels_below_lines" XG_POST, gxg_gtk_text_view_get_pixels_below_lines, 1, 0, 0, H_gtk_text_view_get_pixels_below_lines);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_view_set_pixels_inside_wrap" XG_POST, gxg_gtk_text_view_set_pixels_inside_wrap, 2, 0, 0, H_gtk_text_view_set_pixels_inside_wrap);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_view_get_pixels_inside_wrap" XG_POST, gxg_gtk_text_view_get_pixels_inside_wrap, 1, 0, 0, H_gtk_text_view_get_pixels_inside_wrap);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_view_set_justification" XG_POST, gxg_gtk_text_view_set_justification, 2, 0, 0, H_gtk_text_view_set_justification);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_view_get_justification" XG_POST, gxg_gtk_text_view_get_justification, 1, 0, 0, H_gtk_text_view_get_justification);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_view_set_left_margin" XG_POST, gxg_gtk_text_view_set_left_margin, 2, 0, 0, H_gtk_text_view_set_left_margin);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_view_get_left_margin" XG_POST, gxg_gtk_text_view_get_left_margin, 1, 0, 0, H_gtk_text_view_get_left_margin);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_view_set_right_margin" XG_POST, gxg_gtk_text_view_set_right_margin, 2, 0, 0, H_gtk_text_view_set_right_margin);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_view_get_right_margin" XG_POST, gxg_gtk_text_view_get_right_margin, 1, 0, 0, H_gtk_text_view_get_right_margin);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_view_set_indent" XG_POST, gxg_gtk_text_view_set_indent, 2, 0, 0, H_gtk_text_view_set_indent);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_view_get_indent" XG_POST, gxg_gtk_text_view_get_indent, 1, 0, 0, H_gtk_text_view_get_indent);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_view_set_tabs" XG_POST, gxg_gtk_text_view_set_tabs, 2, 0, 0, H_gtk_text_view_set_tabs);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_view_get_tabs" XG_POST, gxg_gtk_text_view_get_tabs, 1, 0, 0, H_gtk_text_view_get_tabs);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_view_get_default_attributes" XG_POST, gxg_gtk_text_view_get_default_attributes, 1, 0, 0, H_gtk_text_view_get_default_attributes);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_toggle_button_get_type" XG_POST, gxg_gtk_toggle_button_get_type, 0, 0, 0, H_gtk_toggle_button_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_toggle_button_new" XG_POST, gxg_gtk_toggle_button_new, 0, 0, 0, H_gtk_toggle_button_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_toggle_button_new_with_label" XG_POST, gxg_gtk_toggle_button_new_with_label, 1, 0, 0, H_gtk_toggle_button_new_with_label);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_toggle_button_new_with_mnemonic" XG_POST, gxg_gtk_toggle_button_new_with_mnemonic, 1, 0, 0, H_gtk_toggle_button_new_with_mnemonic);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_toggle_button_set_mode" XG_POST, gxg_gtk_toggle_button_set_mode, 2, 0, 0, H_gtk_toggle_button_set_mode);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_toggle_button_get_mode" XG_POST, gxg_gtk_toggle_button_get_mode, 1, 0, 0, H_gtk_toggle_button_get_mode);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_toggle_button_set_active" XG_POST, gxg_gtk_toggle_button_set_active, 2, 0, 0, H_gtk_toggle_button_set_active);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_toggle_button_get_active" XG_POST, gxg_gtk_toggle_button_get_active, 1, 0, 0, H_gtk_toggle_button_get_active);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_toggle_button_toggled" XG_POST, gxg_gtk_toggle_button_toggled, 1, 0, 0, H_gtk_toggle_button_toggled);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_toggle_button_set_inconsistent" XG_POST, gxg_gtk_toggle_button_set_inconsistent, 2, 0, 0, H_gtk_toggle_button_set_inconsistent);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_toggle_button_get_inconsistent" XG_POST, gxg_gtk_toggle_button_get_inconsistent, 1, 0, 0, H_gtk_toggle_button_get_inconsistent);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_toolbar_get_type" XG_POST, gxg_gtk_toolbar_get_type, 0, 0, 0, H_gtk_toolbar_get_type);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_toolbar_new" XG_POST, gxg_gtk_toolbar_new, 0, 0, 0, H_gtk_toolbar_new);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_toolbar_append_item" XG_POST, gxg_gtk_toolbar_append_item, 6, 1, 0, H_gtk_toolbar_append_item);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_toolbar_prepend_item" XG_POST, gxg_gtk_toolbar_prepend_item, 6, 1, 0, H_gtk_toolbar_prepend_item);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_toolbar_insert_item" XG_POST, gxg_gtk_toolbar_insert_item, 8, 0, 0, H_gtk_toolbar_insert_item);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_toolbar_insert_stock" XG_POST, gxg_gtk_toolbar_insert_stock, 7, 0, 0, H_gtk_toolbar_insert_stock);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_toolbar_append_space" XG_POST, gxg_gtk_toolbar_append_space, 1, 0, 0, H_gtk_toolbar_append_space);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_toolbar_prepend_space" XG_POST, gxg_gtk_toolbar_prepend_space, 1, 0, 0, H_gtk_toolbar_prepend_space);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_toolbar_insert_space" XG_POST, gxg_gtk_toolbar_insert_space, 2, 0, 0, H_gtk_toolbar_insert_space);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_toolbar_remove_space" XG_POST, gxg_gtk_toolbar_remove_space, 2, 0, 0, H_gtk_toolbar_remove_space);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_toolbar_append_element" XG_POST, gxg_gtk_toolbar_append_element, 8, 1, 0, H_gtk_toolbar_append_element);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_toolbar_prepend_element" XG_POST, gxg_gtk_toolbar_prepend_element, 8, 1, 0, H_gtk_toolbar_prepend_element);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_toolbar_insert_element" XG_POST, gxg_gtk_toolbar_insert_element, 0, 0, 1, H_gtk_toolbar_insert_element);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_toolbar_append_widget" XG_POST, gxg_gtk_toolbar_append_widget, 4, 0, 0, H_gtk_toolbar_append_widget);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_toolbar_prepend_widget" XG_POST, gxg_gtk_toolbar_prepend_widget, 4, 0, 0, H_gtk_toolbar_prepend_widget);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_toolbar_insert_widget" XG_POST, gxg_gtk_toolbar_insert_widget, 5, 0, 0, H_gtk_toolbar_insert_widget);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_toolbar_set_orientation" XG_POST, gxg_gtk_toolbar_set_orientation, 2, 0, 0, H_gtk_toolbar_set_orientation);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_toolbar_set_style" XG_POST, gxg_gtk_toolbar_set_style, 2, 0, 0, H_gtk_toolbar_set_style);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_toolbar_set_icon_size" XG_POST, gxg_gtk_toolbar_set_icon_size, 2, 0, 0, H_gtk_toolbar_set_icon_size);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_toolbar_set_tooltips" XG_POST, gxg_gtk_toolbar_set_tooltips, 2, 0, 0, H_gtk_toolbar_set_tooltips);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_toolbar_unset_style" XG_POST, gxg_gtk_toolbar_unset_style, 1, 0, 0, H_gtk_toolbar_unset_style);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_toolbar_unset_icon_size" XG_POST, gxg_gtk_toolbar_unset_icon_size, 1, 0, 0, H_gtk_toolbar_unset_icon_size);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_toolbar_get_orientation" XG_POST, gxg_gtk_toolbar_get_orientation, 1, 0, 0, H_gtk_toolbar_get_orientation);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_toolbar_get_style" XG_POST, gxg_gtk_toolbar_get_style, 1, 0, 0, H_gtk_toolbar_get_style);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_toolbar_get_icon_size" XG_POST, gxg_gtk_toolbar_get_icon_size, 1, 0, 0, H_gtk_toolbar_get_icon_size);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_toolbar_get_tooltips" XG_POST, gxg_gtk_toolbar_get_tooltips, 1, 0, 0, H_gtk_toolbar_get_tooltips);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tooltips_get_type" XG_POST, gxg_gtk_tooltips_get_type, 0, 0, 0, H_gtk_tooltips_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tooltips_new" XG_POST, gxg_gtk_tooltips_new, 0, 0, 0, H_gtk_tooltips_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tooltips_enable" XG_POST, gxg_gtk_tooltips_enable, 1, 0, 0, H_gtk_tooltips_enable);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tooltips_disable" XG_POST, gxg_gtk_tooltips_disable, 1, 0, 0, H_gtk_tooltips_disable);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tooltips_set_tip" XG_POST, gxg_gtk_tooltips_set_tip, 4, 0, 0, H_gtk_tooltips_set_tip);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tooltips_data_get" XG_POST, gxg_gtk_tooltips_data_get, 1, 0, 0, H_gtk_tooltips_data_get);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tooltips_force_window" XG_POST, gxg_gtk_tooltips_force_window, 1, 0, 0, H_gtk_tooltips_force_window);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_drag_source_get_type" XG_POST, gxg_gtk_tree_drag_source_get_type, 0, 0, 0, H_gtk_tree_drag_source_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_drag_source_row_draggable" XG_POST, gxg_gtk_tree_drag_source_row_draggable, 2, 0, 0, H_gtk_tree_drag_source_row_draggable);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_drag_source_drag_data_delete" XG_POST, gxg_gtk_tree_drag_source_drag_data_delete, 2, 0, 0, H_gtk_tree_drag_source_drag_data_delete);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_drag_source_drag_data_get" XG_POST, gxg_gtk_tree_drag_source_drag_data_get, 3, 0, 0, H_gtk_tree_drag_source_drag_data_get);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_drag_dest_get_type" XG_POST, gxg_gtk_tree_drag_dest_get_type, 0, 0, 0, H_gtk_tree_drag_dest_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_drag_dest_drag_data_received" XG_POST, gxg_gtk_tree_drag_dest_drag_data_received, 3, 0, 0, H_gtk_tree_drag_dest_drag_data_received);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_drag_dest_row_drop_possible" XG_POST, gxg_gtk_tree_drag_dest_row_drop_possible, 3, 0, 0, H_gtk_tree_drag_dest_row_drop_possible);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_set_row_drag_data" XG_POST, gxg_gtk_tree_set_row_drag_data, 3, 0, 0, H_gtk_tree_set_row_drag_data);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_get_row_drag_data" XG_POST, gxg_gtk_tree_get_row_drag_data, 1, 2, 0, H_gtk_tree_get_row_drag_data);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_path_new" XG_POST, gxg_gtk_tree_path_new, 0, 0, 0, H_gtk_tree_path_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_path_new_from_string" XG_POST, gxg_gtk_tree_path_new_from_string, 1, 0, 0, H_gtk_tree_path_new_from_string);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_path_to_string" XG_POST, gxg_gtk_tree_path_to_string, 1, 0, 0, H_gtk_tree_path_to_string);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_path_new_root" XG_POST, gxg_gtk_tree_path_new_root, 0, 0, 0, H_gtk_tree_path_new_root);
#if HAVE_GTK_FILE_SELECTION_GET_SELECTIONS
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_path_new_first" XG_POST, gxg_gtk_tree_path_new_first, 0, 0, 0, H_gtk_tree_path_new_first);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_path_append_index" XG_POST, gxg_gtk_tree_path_append_index, 2, 0, 0, H_gtk_tree_path_append_index);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_path_prepend_index" XG_POST, gxg_gtk_tree_path_prepend_index, 2, 0, 0, H_gtk_tree_path_prepend_index);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_path_get_depth" XG_POST, gxg_gtk_tree_path_get_depth, 1, 0, 0, H_gtk_tree_path_get_depth);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_path_get_indices" XG_POST, gxg_gtk_tree_path_get_indices, 1, 0, 0, H_gtk_tree_path_get_indices);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_path_free" XG_POST, gxg_gtk_tree_path_free, 1, 0, 0, H_gtk_tree_path_free);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_path_copy" XG_POST, gxg_gtk_tree_path_copy, 1, 0, 0, H_gtk_tree_path_copy);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_path_compare" XG_POST, gxg_gtk_tree_path_compare, 2, 0, 0, H_gtk_tree_path_compare);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_path_next" XG_POST, gxg_gtk_tree_path_next, 1, 0, 0, H_gtk_tree_path_next);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_path_prev" XG_POST, gxg_gtk_tree_path_prev, 1, 0, 0, H_gtk_tree_path_prev);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_path_up" XG_POST, gxg_gtk_tree_path_up, 1, 0, 0, H_gtk_tree_path_up);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_path_down" XG_POST, gxg_gtk_tree_path_down, 1, 0, 0, H_gtk_tree_path_down);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_path_is_ancestor" XG_POST, gxg_gtk_tree_path_is_ancestor, 2, 0, 0, H_gtk_tree_path_is_ancestor);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_path_is_descendant" XG_POST, gxg_gtk_tree_path_is_descendant, 2, 0, 0, H_gtk_tree_path_is_descendant);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_row_reference_new" XG_POST, gxg_gtk_tree_row_reference_new, 2, 0, 0, H_gtk_tree_row_reference_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_row_reference_new_proxy" XG_POST, gxg_gtk_tree_row_reference_new_proxy, 3, 0, 0, H_gtk_tree_row_reference_new_proxy);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_row_reference_get_path" XG_POST, gxg_gtk_tree_row_reference_get_path, 1, 0, 0, H_gtk_tree_row_reference_get_path);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_row_reference_valid" XG_POST, gxg_gtk_tree_row_reference_valid, 1, 0, 0, H_gtk_tree_row_reference_valid);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_row_reference_free" XG_POST, gxg_gtk_tree_row_reference_free, 1, 0, 0, H_gtk_tree_row_reference_free);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_row_reference_inserted" XG_POST, gxg_gtk_tree_row_reference_inserted, 2, 0, 0, H_gtk_tree_row_reference_inserted);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_row_reference_deleted" XG_POST, gxg_gtk_tree_row_reference_deleted, 2, 0, 0, H_gtk_tree_row_reference_deleted);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_row_reference_reordered" XG_POST, gxg_gtk_tree_row_reference_reordered, 4, 0, 0, H_gtk_tree_row_reference_reordered);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_iter_copy" XG_POST, gxg_gtk_tree_iter_copy, 1, 0, 0, H_gtk_tree_iter_copy);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_iter_free" XG_POST, gxg_gtk_tree_iter_free, 1, 0, 0, H_gtk_tree_iter_free);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_model_get_type" XG_POST, gxg_gtk_tree_model_get_type, 0, 0, 0, H_gtk_tree_model_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_model_get_flags" XG_POST, gxg_gtk_tree_model_get_flags, 1, 0, 0, H_gtk_tree_model_get_flags);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_model_get_n_columns" XG_POST, gxg_gtk_tree_model_get_n_columns, 1, 0, 0, H_gtk_tree_model_get_n_columns);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_model_get_column_type" XG_POST, gxg_gtk_tree_model_get_column_type, 2, 0, 0, H_gtk_tree_model_get_column_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_model_get_iter" XG_POST, gxg_gtk_tree_model_get_iter, 3, 0, 0, H_gtk_tree_model_get_iter);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_model_get_iter_from_string" XG_POST, gxg_gtk_tree_model_get_iter_from_string, 3, 0, 0, H_gtk_tree_model_get_iter_from_string);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_model_get_iter_root" XG_POST, gxg_gtk_tree_model_get_iter_root, 2, 0, 0, H_gtk_tree_model_get_iter_root);
#if HAVE_GTK_FILE_SELECTION_GET_SELECTIONS
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_model_get_iter_first" XG_POST, gxg_gtk_tree_model_get_iter_first, 2, 0, 0, H_gtk_tree_model_get_iter_first);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_model_get_path" XG_POST, gxg_gtk_tree_model_get_path, 2, 0, 0, H_gtk_tree_model_get_path);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_model_get_value" XG_POST, gxg_gtk_tree_model_get_value, 4, 0, 0, H_gtk_tree_model_get_value);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_model_iter_next" XG_POST, gxg_gtk_tree_model_iter_next, 2, 0, 0, H_gtk_tree_model_iter_next);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_model_iter_children" XG_POST, gxg_gtk_tree_model_iter_children, 3, 0, 0, H_gtk_tree_model_iter_children);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_model_iter_has_child" XG_POST, gxg_gtk_tree_model_iter_has_child, 2, 0, 0, H_gtk_tree_model_iter_has_child);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_model_iter_n_children" XG_POST, gxg_gtk_tree_model_iter_n_children, 2, 0, 0, H_gtk_tree_model_iter_n_children);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_model_iter_nth_child" XG_POST, gxg_gtk_tree_model_iter_nth_child, 4, 0, 0, H_gtk_tree_model_iter_nth_child);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_model_iter_parent" XG_POST, gxg_gtk_tree_model_iter_parent, 3, 0, 0, H_gtk_tree_model_iter_parent);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_model_ref_node" XG_POST, gxg_gtk_tree_model_ref_node, 2, 0, 0, H_gtk_tree_model_ref_node);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_model_unref_node" XG_POST, gxg_gtk_tree_model_unref_node, 2, 0, 0, H_gtk_tree_model_unref_node);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_model_get" XG_POST, gxg_gtk_tree_model_get, 2, 0, 0, H_gtk_tree_model_get);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_model_foreach" XG_POST, gxg_gtk_tree_model_foreach, 2, 1, 0, H_gtk_tree_model_foreach);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_model_row_changed" XG_POST, gxg_gtk_tree_model_row_changed, 3, 0, 0, H_gtk_tree_model_row_changed);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_model_row_inserted" XG_POST, gxg_gtk_tree_model_row_inserted, 3, 0, 0, H_gtk_tree_model_row_inserted);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_model_row_has_child_toggled" XG_POST, gxg_gtk_tree_model_row_has_child_toggled, 3, 0, 0, H_gtk_tree_model_row_has_child_toggled);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_model_row_deleted" XG_POST, gxg_gtk_tree_model_row_deleted, 2, 0, 0, H_gtk_tree_model_row_deleted);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_model_rows_reordered" XG_POST, gxg_gtk_tree_model_rows_reordered, 4, 0, 0, H_gtk_tree_model_rows_reordered);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_model_sort_get_type" XG_POST, gxg_gtk_tree_model_sort_get_type, 0, 0, 0, H_gtk_tree_model_sort_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_model_sort_new_with_model" XG_POST, gxg_gtk_tree_model_sort_new_with_model, 1, 0, 0, H_gtk_tree_model_sort_new_with_model);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_model_sort_get_model" XG_POST, gxg_gtk_tree_model_sort_get_model, 1, 0, 0, H_gtk_tree_model_sort_get_model);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_model_sort_convert_child_path_to_path" XG_POST, gxg_gtk_tree_model_sort_convert_child_path_to_path, 2, 0, 0, H_gtk_tree_model_sort_convert_child_path_to_path);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_model_sort_convert_child_iter_to_iter" XG_POST, gxg_gtk_tree_model_sort_convert_child_iter_to_iter, 3, 0, 0, H_gtk_tree_model_sort_convert_child_iter_to_iter);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_model_sort_convert_path_to_child_path" XG_POST, gxg_gtk_tree_model_sort_convert_path_to_child_path, 2, 0, 0, H_gtk_tree_model_sort_convert_path_to_child_path);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_model_sort_convert_iter_to_child_iter" XG_POST, gxg_gtk_tree_model_sort_convert_iter_to_child_iter, 3, 0, 0, H_gtk_tree_model_sort_convert_iter_to_child_iter);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_model_sort_reset_default_sort_func" XG_POST, gxg_gtk_tree_model_sort_reset_default_sort_func, 1, 0, 0, H_gtk_tree_model_sort_reset_default_sort_func);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_model_sort_clear_cache" XG_POST, gxg_gtk_tree_model_sort_clear_cache, 1, 0, 0, H_gtk_tree_model_sort_clear_cache);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_selection_get_type" XG_POST, gxg_gtk_tree_selection_get_type, 0, 0, 0, H_gtk_tree_selection_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_selection_set_mode" XG_POST, gxg_gtk_tree_selection_set_mode, 2, 0, 0, H_gtk_tree_selection_set_mode);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_selection_get_mode" XG_POST, gxg_gtk_tree_selection_get_mode, 1, 0, 0, H_gtk_tree_selection_get_mode);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_selection_set_select_function" XG_POST, gxg_gtk_tree_selection_set_select_function, 4, 0, 0, H_gtk_tree_selection_set_select_function);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_selection_get_user_data" XG_POST, gxg_gtk_tree_selection_get_user_data, 1, 0, 0, H_gtk_tree_selection_get_user_data);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_selection_get_tree_view" XG_POST, gxg_gtk_tree_selection_get_tree_view, 1, 0, 0, H_gtk_tree_selection_get_tree_view);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_selection_get_selected" XG_POST, gxg_gtk_tree_selection_get_selected, 3, 0, 0, H_gtk_tree_selection_get_selected);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_selection_selected_foreach" XG_POST, gxg_gtk_tree_selection_selected_foreach, 2, 1, 0, H_gtk_tree_selection_selected_foreach);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_selection_select_path" XG_POST, gxg_gtk_tree_selection_select_path, 2, 0, 0, H_gtk_tree_selection_select_path);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_selection_unselect_path" XG_POST, gxg_gtk_tree_selection_unselect_path, 2, 0, 0, H_gtk_tree_selection_unselect_path);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_selection_select_iter" XG_POST, gxg_gtk_tree_selection_select_iter, 2, 0, 0, H_gtk_tree_selection_select_iter);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_selection_unselect_iter" XG_POST, gxg_gtk_tree_selection_unselect_iter, 2, 0, 0, H_gtk_tree_selection_unselect_iter);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_selection_path_is_selected" XG_POST, gxg_gtk_tree_selection_path_is_selected, 2, 0, 0, H_gtk_tree_selection_path_is_selected);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_selection_iter_is_selected" XG_POST, gxg_gtk_tree_selection_iter_is_selected, 2, 0, 0, H_gtk_tree_selection_iter_is_selected);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_selection_select_all" XG_POST, gxg_gtk_tree_selection_select_all, 1, 0, 0, H_gtk_tree_selection_select_all);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_selection_unselect_all" XG_POST, gxg_gtk_tree_selection_unselect_all, 1, 0, 0, H_gtk_tree_selection_unselect_all);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_selection_select_range" XG_POST, gxg_gtk_tree_selection_select_range, 3, 0, 0, H_gtk_tree_selection_select_range);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_sortable_get_type" XG_POST, gxg_gtk_tree_sortable_get_type, 0, 0, 0, H_gtk_tree_sortable_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_sortable_sort_column_changed" XG_POST, gxg_gtk_tree_sortable_sort_column_changed, 1, 0, 0, H_gtk_tree_sortable_sort_column_changed);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_sortable_get_sort_column_id" XG_POST, gxg_gtk_tree_sortable_get_sort_column_id, 3, 0, 0, H_gtk_tree_sortable_get_sort_column_id);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_sortable_set_sort_column_id" XG_POST, gxg_gtk_tree_sortable_set_sort_column_id, 3, 0, 0, H_gtk_tree_sortable_set_sort_column_id);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_sortable_set_sort_func" XG_POST, gxg_gtk_tree_sortable_set_sort_func, 5, 0, 0, H_gtk_tree_sortable_set_sort_func);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_sortable_set_default_sort_func" XG_POST, gxg_gtk_tree_sortable_set_default_sort_func, 4, 0, 0, H_gtk_tree_sortable_set_default_sort_func);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_sortable_has_default_sort_func" XG_POST, gxg_gtk_tree_sortable_has_default_sort_func, 1, 0, 0, H_gtk_tree_sortable_has_default_sort_func);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_store_get_type" XG_POST, gxg_gtk_tree_store_get_type, 0, 0, 0, H_gtk_tree_store_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_store_new" XG_POST, gxg_gtk_tree_store_new, 1, 0, 0, H_gtk_tree_store_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_store_newv" XG_POST, gxg_gtk_tree_store_newv, 2, 0, 0, H_gtk_tree_store_newv);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_store_set_column_types" XG_POST, gxg_gtk_tree_store_set_column_types, 3, 0, 0, H_gtk_tree_store_set_column_types);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_store_set_value" XG_POST, gxg_gtk_tree_store_set_value, 4, 0, 0, H_gtk_tree_store_set_value);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_store_set" XG_POST, gxg_gtk_tree_store_set, 2, 0, 0, H_gtk_tree_store_set);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_store_remove" XG_POST, gxg_gtk_tree_store_remove, 2, 0, 0, H_gtk_tree_store_remove);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_store_insert" XG_POST, gxg_gtk_tree_store_insert, 4, 0, 0, H_gtk_tree_store_insert);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_store_insert_before" XG_POST, gxg_gtk_tree_store_insert_before, 4, 0, 0, H_gtk_tree_store_insert_before);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_store_insert_after" XG_POST, gxg_gtk_tree_store_insert_after, 4, 0, 0, H_gtk_tree_store_insert_after);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_store_prepend" XG_POST, gxg_gtk_tree_store_prepend, 3, 0, 0, H_gtk_tree_store_prepend);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_store_append" XG_POST, gxg_gtk_tree_store_append, 3, 0, 0, H_gtk_tree_store_append);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_store_is_ancestor" XG_POST, gxg_gtk_tree_store_is_ancestor, 3, 0, 0, H_gtk_tree_store_is_ancestor);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_store_iter_depth" XG_POST, gxg_gtk_tree_store_iter_depth, 2, 0, 0, H_gtk_tree_store_iter_depth);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_store_clear" XG_POST, gxg_gtk_tree_store_clear, 1, 0, 0, H_gtk_tree_store_clear);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_view_column_get_type" XG_POST, gxg_gtk_tree_view_column_get_type, 0, 0, 0, H_gtk_tree_view_column_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_view_column_new" XG_POST, gxg_gtk_tree_view_column_new, 0, 0, 0, H_gtk_tree_view_column_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_view_column_new_with_attributes" XG_POST, gxg_gtk_tree_view_column_new_with_attributes, 2, 0, 0, H_gtk_tree_view_column_new_with_attributes);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_view_column_pack_start" XG_POST, gxg_gtk_tree_view_column_pack_start, 3, 0, 0, H_gtk_tree_view_column_pack_start);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_view_column_pack_end" XG_POST, gxg_gtk_tree_view_column_pack_end, 3, 0, 0, H_gtk_tree_view_column_pack_end);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_view_column_clear" XG_POST, gxg_gtk_tree_view_column_clear, 1, 0, 0, H_gtk_tree_view_column_clear);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_view_column_get_cell_renderers" XG_POST, gxg_gtk_tree_view_column_get_cell_renderers, 1, 0, 0, H_gtk_tree_view_column_get_cell_renderers);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_view_column_add_attribute" XG_POST, gxg_gtk_tree_view_column_add_attribute, 4, 0, 0, H_gtk_tree_view_column_add_attribute);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_view_column_set_attributes" XG_POST, gxg_gtk_tree_view_column_set_attributes, 2, 0, 0, H_gtk_tree_view_column_set_attributes);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_view_column_set_cell_data_func" XG_POST, gxg_gtk_tree_view_column_set_cell_data_func, 5, 0, 0, H_gtk_tree_view_column_set_cell_data_func);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_view_column_clear_attributes" XG_POST, gxg_gtk_tree_view_column_clear_attributes, 2, 0, 0, H_gtk_tree_view_column_clear_attributes);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_view_column_set_spacing" XG_POST, gxg_gtk_tree_view_column_set_spacing, 2, 0, 0, H_gtk_tree_view_column_set_spacing);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_view_column_get_spacing" XG_POST, gxg_gtk_tree_view_column_get_spacing, 1, 0, 0, H_gtk_tree_view_column_get_spacing);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_view_column_set_visible" XG_POST, gxg_gtk_tree_view_column_set_visible, 2, 0, 0, H_gtk_tree_view_column_set_visible);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_view_column_get_visible" XG_POST, gxg_gtk_tree_view_column_get_visible, 1, 0, 0, H_gtk_tree_view_column_get_visible);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_view_column_set_resizable" XG_POST, gxg_gtk_tree_view_column_set_resizable, 2, 0, 0, H_gtk_tree_view_column_set_resizable);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_view_column_get_resizable" XG_POST, gxg_gtk_tree_view_column_get_resizable, 1, 0, 0, H_gtk_tree_view_column_get_resizable);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_view_column_set_sizing" XG_POST, gxg_gtk_tree_view_column_set_sizing, 2, 0, 0, H_gtk_tree_view_column_set_sizing);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_view_column_get_sizing" XG_POST, gxg_gtk_tree_view_column_get_sizing, 1, 0, 0, H_gtk_tree_view_column_get_sizing);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_view_column_get_width" XG_POST, gxg_gtk_tree_view_column_get_width, 1, 0, 0, H_gtk_tree_view_column_get_width);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_view_column_get_fixed_width" XG_POST, gxg_gtk_tree_view_column_get_fixed_width, 1, 0, 0, H_gtk_tree_view_column_get_fixed_width);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_view_column_set_fixed_width" XG_POST, gxg_gtk_tree_view_column_set_fixed_width, 2, 0, 0, H_gtk_tree_view_column_set_fixed_width);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_view_column_set_min_width" XG_POST, gxg_gtk_tree_view_column_set_min_width, 2, 0, 0, H_gtk_tree_view_column_set_min_width);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_view_column_get_min_width" XG_POST, gxg_gtk_tree_view_column_get_min_width, 1, 0, 0, H_gtk_tree_view_column_get_min_width);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_view_column_set_max_width" XG_POST, gxg_gtk_tree_view_column_set_max_width, 2, 0, 0, H_gtk_tree_view_column_set_max_width);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_view_column_get_max_width" XG_POST, gxg_gtk_tree_view_column_get_max_width, 1, 0, 0, H_gtk_tree_view_column_get_max_width);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_view_column_clicked" XG_POST, gxg_gtk_tree_view_column_clicked, 1, 0, 0, H_gtk_tree_view_column_clicked);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_view_column_set_title" XG_POST, gxg_gtk_tree_view_column_set_title, 2, 0, 0, H_gtk_tree_view_column_set_title);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_view_column_get_title" XG_POST, gxg_gtk_tree_view_column_get_title, 1, 0, 0, H_gtk_tree_view_column_get_title);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_view_column_set_clickable" XG_POST, gxg_gtk_tree_view_column_set_clickable, 2, 0, 0, H_gtk_tree_view_column_set_clickable);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_view_column_get_clickable" XG_POST, gxg_gtk_tree_view_column_get_clickable, 1, 0, 0, H_gtk_tree_view_column_get_clickable);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_view_column_set_widget" XG_POST, gxg_gtk_tree_view_column_set_widget, 2, 0, 0, H_gtk_tree_view_column_set_widget);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_view_column_get_widget" XG_POST, gxg_gtk_tree_view_column_get_widget, 1, 0, 0, H_gtk_tree_view_column_get_widget);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_view_column_set_alignment" XG_POST, gxg_gtk_tree_view_column_set_alignment, 2, 0, 0, H_gtk_tree_view_column_set_alignment);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_view_column_get_alignment" XG_POST, gxg_gtk_tree_view_column_get_alignment, 1, 0, 0, H_gtk_tree_view_column_get_alignment);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_view_column_set_reorderable" XG_POST, gxg_gtk_tree_view_column_set_reorderable, 2, 0, 0, H_gtk_tree_view_column_set_reorderable);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_view_column_get_reorderable" XG_POST, gxg_gtk_tree_view_column_get_reorderable, 1, 0, 0, H_gtk_tree_view_column_get_reorderable);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_view_column_set_sort_column_id" XG_POST, gxg_gtk_tree_view_column_set_sort_column_id, 2, 0, 0, H_gtk_tree_view_column_set_sort_column_id);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_view_column_get_sort_column_id" XG_POST, gxg_gtk_tree_view_column_get_sort_column_id, 1, 0, 0, H_gtk_tree_view_column_get_sort_column_id);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_view_column_set_sort_indicator" XG_POST, gxg_gtk_tree_view_column_set_sort_indicator, 2, 0, 0, H_gtk_tree_view_column_set_sort_indicator);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_view_column_get_sort_indicator" XG_POST, gxg_gtk_tree_view_column_get_sort_indicator, 1, 0, 0, H_gtk_tree_view_column_get_sort_indicator);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_view_column_set_sort_order" XG_POST, gxg_gtk_tree_view_column_set_sort_order, 2, 0, 0, H_gtk_tree_view_column_set_sort_order);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_view_column_get_sort_order" XG_POST, gxg_gtk_tree_view_column_get_sort_order, 1, 0, 0, H_gtk_tree_view_column_get_sort_order);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_view_column_cell_set_cell_data" XG_POST, gxg_gtk_tree_view_column_cell_set_cell_data, 5, 0, 0, H_gtk_tree_view_column_cell_set_cell_data);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_view_column_cell_get_size" XG_POST, gxg_gtk_tree_view_column_cell_get_size, 6, 0, 0, H_gtk_tree_view_column_cell_get_size);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_view_column_cell_is_visible" XG_POST, gxg_gtk_tree_view_column_cell_is_visible, 1, 0, 0, H_gtk_tree_view_column_cell_is_visible);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_view_get_type" XG_POST, gxg_gtk_tree_view_get_type, 0, 0, 0, H_gtk_tree_view_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_view_new" XG_POST, gxg_gtk_tree_view_new, 0, 0, 0, H_gtk_tree_view_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_view_new_with_model" XG_POST, gxg_gtk_tree_view_new_with_model, 1, 0, 0, H_gtk_tree_view_new_with_model);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_view_get_model" XG_POST, gxg_gtk_tree_view_get_model, 1, 0, 0, H_gtk_tree_view_get_model);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_view_set_model" XG_POST, gxg_gtk_tree_view_set_model, 2, 0, 0, H_gtk_tree_view_set_model);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_view_get_selection" XG_POST, gxg_gtk_tree_view_get_selection, 1, 0, 0, H_gtk_tree_view_get_selection);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_view_get_hadjustment" XG_POST, gxg_gtk_tree_view_get_hadjustment, 1, 0, 0, H_gtk_tree_view_get_hadjustment);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_view_set_hadjustment" XG_POST, gxg_gtk_tree_view_set_hadjustment, 2, 0, 0, H_gtk_tree_view_set_hadjustment);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_view_get_vadjustment" XG_POST, gxg_gtk_tree_view_get_vadjustment, 1, 0, 0, H_gtk_tree_view_get_vadjustment);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_view_set_vadjustment" XG_POST, gxg_gtk_tree_view_set_vadjustment, 2, 0, 0, H_gtk_tree_view_set_vadjustment);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_view_get_headers_visible" XG_POST, gxg_gtk_tree_view_get_headers_visible, 1, 0, 0, H_gtk_tree_view_get_headers_visible);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_view_set_headers_visible" XG_POST, gxg_gtk_tree_view_set_headers_visible, 2, 0, 0, H_gtk_tree_view_set_headers_visible);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_view_columns_autosize" XG_POST, gxg_gtk_tree_view_columns_autosize, 1, 0, 0, H_gtk_tree_view_columns_autosize);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_view_set_headers_clickable" XG_POST, gxg_gtk_tree_view_set_headers_clickable, 2, 0, 0, H_gtk_tree_view_set_headers_clickable);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_view_set_rules_hint" XG_POST, gxg_gtk_tree_view_set_rules_hint, 2, 0, 0, H_gtk_tree_view_set_rules_hint);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_view_get_rules_hint" XG_POST, gxg_gtk_tree_view_get_rules_hint, 1, 0, 0, H_gtk_tree_view_get_rules_hint);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_view_append_column" XG_POST, gxg_gtk_tree_view_append_column, 2, 0, 0, H_gtk_tree_view_append_column);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_view_remove_column" XG_POST, gxg_gtk_tree_view_remove_column, 2, 0, 0, H_gtk_tree_view_remove_column);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_view_insert_column" XG_POST, gxg_gtk_tree_view_insert_column, 3, 0, 0, H_gtk_tree_view_insert_column);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_view_insert_column_with_attributes" XG_POST, gxg_gtk_tree_view_insert_column_with_attributes, 4, 0, 0, H_gtk_tree_view_insert_column_with_attributes);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_view_insert_column_with_data_func" XG_POST, gxg_gtk_tree_view_insert_column_with_data_func, 7, 0, 0, H_gtk_tree_view_insert_column_with_data_func);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_view_get_column" XG_POST, gxg_gtk_tree_view_get_column, 2, 0, 0, H_gtk_tree_view_get_column);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_view_get_columns" XG_POST, gxg_gtk_tree_view_get_columns, 1, 0, 0, H_gtk_tree_view_get_columns);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_view_move_column_after" XG_POST, gxg_gtk_tree_view_move_column_after, 3, 0, 0, H_gtk_tree_view_move_column_after);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_view_set_expander_column" XG_POST, gxg_gtk_tree_view_set_expander_column, 2, 0, 0, H_gtk_tree_view_set_expander_column);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_view_get_expander_column" XG_POST, gxg_gtk_tree_view_get_expander_column, 1, 0, 0, H_gtk_tree_view_get_expander_column);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_view_set_column_drag_function" XG_POST, gxg_gtk_tree_view_set_column_drag_function, 4, 0, 0, H_gtk_tree_view_set_column_drag_function);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_view_scroll_to_point" XG_POST, gxg_gtk_tree_view_scroll_to_point, 3, 0, 0, H_gtk_tree_view_scroll_to_point);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_view_scroll_to_cell" XG_POST, gxg_gtk_tree_view_scroll_to_cell, 6, 0, 0, H_gtk_tree_view_scroll_to_cell);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_view_row_activated" XG_POST, gxg_gtk_tree_view_row_activated, 3, 0, 0, H_gtk_tree_view_row_activated);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_view_expand_all" XG_POST, gxg_gtk_tree_view_expand_all, 1, 0, 0, H_gtk_tree_view_expand_all);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_view_collapse_all" XG_POST, gxg_gtk_tree_view_collapse_all, 1, 0, 0, H_gtk_tree_view_collapse_all);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_view_expand_row" XG_POST, gxg_gtk_tree_view_expand_row, 3, 0, 0, H_gtk_tree_view_expand_row);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_view_collapse_row" XG_POST, gxg_gtk_tree_view_collapse_row, 2, 0, 0, H_gtk_tree_view_collapse_row);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_view_map_expanded_rows" XG_POST, gxg_gtk_tree_view_map_expanded_rows, 2, 1, 0, H_gtk_tree_view_map_expanded_rows);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_view_row_expanded" XG_POST, gxg_gtk_tree_view_row_expanded, 2, 0, 0, H_gtk_tree_view_row_expanded);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_view_set_reorderable" XG_POST, gxg_gtk_tree_view_set_reorderable, 2, 0, 0, H_gtk_tree_view_set_reorderable);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_view_get_reorderable" XG_POST, gxg_gtk_tree_view_get_reorderable, 1, 0, 0, H_gtk_tree_view_get_reorderable);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_view_set_cursor" XG_POST, gxg_gtk_tree_view_set_cursor, 4, 0, 0, H_gtk_tree_view_set_cursor);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_view_get_cursor" XG_POST, gxg_gtk_tree_view_get_cursor, 1, 2, 0, H_gtk_tree_view_get_cursor);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_view_get_bin_window" XG_POST, gxg_gtk_tree_view_get_bin_window, 1, 0, 0, H_gtk_tree_view_get_bin_window);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_view_get_path_at_pos" XG_POST, gxg_gtk_tree_view_get_path_at_pos, 3, 4, 0, H_gtk_tree_view_get_path_at_pos);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_view_get_cell_area" XG_POST, gxg_gtk_tree_view_get_cell_area, 4, 0, 0, H_gtk_tree_view_get_cell_area);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_view_get_background_area" XG_POST, gxg_gtk_tree_view_get_background_area, 4, 0, 0, H_gtk_tree_view_get_background_area);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_view_get_visible_rect" XG_POST, gxg_gtk_tree_view_get_visible_rect, 2, 0, 0, H_gtk_tree_view_get_visible_rect);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_view_widget_to_tree_coords" XG_POST, gxg_gtk_tree_view_widget_to_tree_coords, 3, 2, 0, H_gtk_tree_view_widget_to_tree_coords);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_view_tree_to_widget_coords" XG_POST, gxg_gtk_tree_view_tree_to_widget_coords, 3, 2, 0, H_gtk_tree_view_tree_to_widget_coords);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_view_enable_model_drag_source" XG_POST, gxg_gtk_tree_view_enable_model_drag_source, 5, 0, 0, H_gtk_tree_view_enable_model_drag_source);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_view_enable_model_drag_dest" XG_POST, gxg_gtk_tree_view_enable_model_drag_dest, 4, 0, 0, H_gtk_tree_view_enable_model_drag_dest);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_view_unset_rows_drag_source" XG_POST, gxg_gtk_tree_view_unset_rows_drag_source, 1, 0, 0, H_gtk_tree_view_unset_rows_drag_source);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_view_unset_rows_drag_dest" XG_POST, gxg_gtk_tree_view_unset_rows_drag_dest, 1, 0, 0, H_gtk_tree_view_unset_rows_drag_dest);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_view_set_drag_dest_row" XG_POST, gxg_gtk_tree_view_set_drag_dest_row, 3, 0, 0, H_gtk_tree_view_set_drag_dest_row);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_view_get_drag_dest_row" XG_POST, gxg_gtk_tree_view_get_drag_dest_row, 1, 2, 0, H_gtk_tree_view_get_drag_dest_row);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_view_get_dest_row_at_pos" XG_POST, gxg_gtk_tree_view_get_dest_row_at_pos, 3, 2, 0, H_gtk_tree_view_get_dest_row_at_pos);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_view_create_row_drag_icon" XG_POST, gxg_gtk_tree_view_create_row_drag_icon, 2, 0, 0, H_gtk_tree_view_create_row_drag_icon);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_view_set_enable_search" XG_POST, gxg_gtk_tree_view_set_enable_search, 2, 0, 0, H_gtk_tree_view_set_enable_search);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_view_get_enable_search" XG_POST, gxg_gtk_tree_view_get_enable_search, 1, 0, 0, H_gtk_tree_view_get_enable_search);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_view_get_search_column" XG_POST, gxg_gtk_tree_view_get_search_column, 1, 0, 0, H_gtk_tree_view_get_search_column);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_view_set_search_column" XG_POST, gxg_gtk_tree_view_set_search_column, 2, 0, 0, H_gtk_tree_view_set_search_column);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_view_get_search_equal_func" XG_POST, gxg_gtk_tree_view_get_search_equal_func, 1, 0, 0, H_gtk_tree_view_get_search_equal_func);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_view_set_search_equal_func" XG_POST, gxg_gtk_tree_view_set_search_equal_func, 4, 0, 0, H_gtk_tree_view_set_search_equal_func);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_vbutton_box_get_type" XG_POST, gxg_gtk_vbutton_box_get_type, 0, 0, 0, H_gtk_vbutton_box_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_vbutton_box_new" XG_POST, gxg_gtk_vbutton_box_new, 0, 0, 0, H_gtk_vbutton_box_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_viewport_get_type" XG_POST, gxg_gtk_viewport_get_type, 0, 0, 0, H_gtk_viewport_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_viewport_new" XG_POST, gxg_gtk_viewport_new, 2, 0, 0, H_gtk_viewport_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_viewport_get_hadjustment" XG_POST, gxg_gtk_viewport_get_hadjustment, 1, 0, 0, H_gtk_viewport_get_hadjustment);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_viewport_get_vadjustment" XG_POST, gxg_gtk_viewport_get_vadjustment, 1, 0, 0, H_gtk_viewport_get_vadjustment);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_viewport_set_hadjustment" XG_POST, gxg_gtk_viewport_set_hadjustment, 2, 0, 0, H_gtk_viewport_set_hadjustment);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_viewport_set_vadjustment" XG_POST, gxg_gtk_viewport_set_vadjustment, 2, 0, 0, H_gtk_viewport_set_vadjustment);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_viewport_set_shadow_type" XG_POST, gxg_gtk_viewport_set_shadow_type, 2, 0, 0, H_gtk_viewport_set_shadow_type);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_viewport_get_shadow_type" XG_POST, gxg_gtk_viewport_get_shadow_type, 1, 0, 0, H_gtk_viewport_get_shadow_type);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_vpaned_get_type" XG_POST, gxg_gtk_vpaned_get_type, 0, 0, 0, H_gtk_vpaned_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_vpaned_new" XG_POST, gxg_gtk_vpaned_new, 0, 0, 0, H_gtk_vpaned_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_vruler_get_type" XG_POST, gxg_gtk_vruler_get_type, 0, 0, 0, H_gtk_vruler_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_vruler_new" XG_POST, gxg_gtk_vruler_new, 0, 0, 0, H_gtk_vruler_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_vscale_get_type" XG_POST, gxg_gtk_vscale_get_type, 0, 0, 0, H_gtk_vscale_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_vscale_new" XG_POST, gxg_gtk_vscale_new, 1, 0, 0, H_gtk_vscale_new);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_vscale_new_with_range" XG_POST, gxg_gtk_vscale_new_with_range, 3, 0, 0, H_gtk_vscale_new_with_range);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_vscrollbar_get_type" XG_POST, gxg_gtk_vscrollbar_get_type, 0, 0, 0, H_gtk_vscrollbar_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_vscrollbar_new" XG_POST, gxg_gtk_vscrollbar_new, 1, 0, 0, H_gtk_vscrollbar_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_vseparator_get_type" XG_POST, gxg_gtk_vseparator_get_type, 0, 0, 0, H_gtk_vseparator_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_vseparator_new" XG_POST, gxg_gtk_vseparator_new, 0, 0, 0, H_gtk_vseparator_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_WIDGET_STATE" XG_POST, gxg_GTK_WIDGET_STATE, 1, 0, 0, H_GTK_WIDGET_STATE);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_WIDGET_SAVED_STATE" XG_POST, gxg_GTK_WIDGET_SAVED_STATE, 1, 0, 0, H_GTK_WIDGET_SAVED_STATE);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_WIDGET_FLAGS" XG_POST, gxg_GTK_WIDGET_FLAGS, 1, 0, 0, H_GTK_WIDGET_FLAGS);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_WIDGET_TOPLEVEL" XG_POST, gxg_GTK_WIDGET_TOPLEVEL, 1, 0, 0, H_GTK_WIDGET_TOPLEVEL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_WIDGET_NO_WINDOW" XG_POST, gxg_GTK_WIDGET_NO_WINDOW, 1, 0, 0, H_GTK_WIDGET_NO_WINDOW);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_WIDGET_REALIZED" XG_POST, gxg_GTK_WIDGET_REALIZED, 1, 0, 0, H_GTK_WIDGET_REALIZED);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_WIDGET_MAPPED" XG_POST, gxg_GTK_WIDGET_MAPPED, 1, 0, 0, H_GTK_WIDGET_MAPPED);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_WIDGET_VISIBLE" XG_POST, gxg_GTK_WIDGET_VISIBLE, 1, 0, 0, H_GTK_WIDGET_VISIBLE);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_WIDGET_DRAWABLE" XG_POST, gxg_GTK_WIDGET_DRAWABLE, 1, 0, 0, H_GTK_WIDGET_DRAWABLE);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_WIDGET_SENSITIVE" XG_POST, gxg_GTK_WIDGET_SENSITIVE, 1, 0, 0, H_GTK_WIDGET_SENSITIVE);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_WIDGET_PARENT_SENSITIVE" XG_POST, gxg_GTK_WIDGET_PARENT_SENSITIVE, 1, 0, 0, H_GTK_WIDGET_PARENT_SENSITIVE);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_WIDGET_IS_SENSITIVE" XG_POST, gxg_GTK_WIDGET_IS_SENSITIVE, 1, 0, 0, H_GTK_WIDGET_IS_SENSITIVE);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_WIDGET_CAN_FOCUS" XG_POST, gxg_GTK_WIDGET_CAN_FOCUS, 1, 0, 0, H_GTK_WIDGET_CAN_FOCUS);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_WIDGET_HAS_FOCUS" XG_POST, gxg_GTK_WIDGET_HAS_FOCUS, 1, 0, 0, H_GTK_WIDGET_HAS_FOCUS);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_WIDGET_CAN_DEFAULT" XG_POST, gxg_GTK_WIDGET_CAN_DEFAULT, 1, 0, 0, H_GTK_WIDGET_CAN_DEFAULT);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_WIDGET_HAS_DEFAULT" XG_POST, gxg_GTK_WIDGET_HAS_DEFAULT, 1, 0, 0, H_GTK_WIDGET_HAS_DEFAULT);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_WIDGET_HAS_GRAB" XG_POST, gxg_GTK_WIDGET_HAS_GRAB, 1, 0, 0, H_GTK_WIDGET_HAS_GRAB);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_WIDGET_RC_STYLE" XG_POST, gxg_GTK_WIDGET_RC_STYLE, 1, 0, 0, H_GTK_WIDGET_RC_STYLE);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_WIDGET_COMPOSITE_CHILD" XG_POST, gxg_GTK_WIDGET_COMPOSITE_CHILD, 1, 0, 0, H_GTK_WIDGET_COMPOSITE_CHILD);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_WIDGET_APP_PAINTABLE" XG_POST, gxg_GTK_WIDGET_APP_PAINTABLE, 1, 0, 0, H_GTK_WIDGET_APP_PAINTABLE);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_WIDGET_RECEIVES_DEFAULT" XG_POST, gxg_GTK_WIDGET_RECEIVES_DEFAULT, 1, 0, 0, H_GTK_WIDGET_RECEIVES_DEFAULT);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_WIDGET_DOUBLE_BUFFERED" XG_POST, gxg_GTK_WIDGET_DOUBLE_BUFFERED, 1, 0, 0, H_GTK_WIDGET_DOUBLE_BUFFERED);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_WIDGET_SET_FLAGS" XG_POST, gxg_GTK_WIDGET_SET_FLAGS, 2, 0, 0, H_GTK_WIDGET_SET_FLAGS);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_WIDGET_UNSET_FLAGS" XG_POST, gxg_GTK_WIDGET_UNSET_FLAGS, 2, 0, 0, H_GTK_WIDGET_UNSET_FLAGS);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_get_type" XG_POST, gxg_gtk_widget_get_type, 0, 0, 0, H_gtk_widget_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_new" XG_POST, gxg_gtk_widget_new, 2, 0, 0, H_gtk_widget_new);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_ref" XG_POST, gxg_gtk_widget_ref, 1, 0, 0, H_gtk_widget_ref);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_unref" XG_POST, gxg_gtk_widget_unref, 1, 0, 0, H_gtk_widget_unref);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_destroy" XG_POST, gxg_gtk_widget_destroy, 1, 0, 0, H_gtk_widget_destroy);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_destroyed" XG_POST, gxg_gtk_widget_destroyed, 1, 1, 0, H_gtk_widget_destroyed);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_unparent" XG_POST, gxg_gtk_widget_unparent, 1, 0, 0, H_gtk_widget_unparent);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_show" XG_POST, gxg_gtk_widget_show, 1, 0, 0, H_gtk_widget_show);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_show_now" XG_POST, gxg_gtk_widget_show_now, 1, 0, 0, H_gtk_widget_show_now);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_hide" XG_POST, gxg_gtk_widget_hide, 1, 0, 0, H_gtk_widget_hide);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_show_all" XG_POST, gxg_gtk_widget_show_all, 1, 0, 0, H_gtk_widget_show_all);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_hide_all" XG_POST, gxg_gtk_widget_hide_all, 1, 0, 0, H_gtk_widget_hide_all);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_map" XG_POST, gxg_gtk_widget_map, 1, 0, 0, H_gtk_widget_map);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_unmap" XG_POST, gxg_gtk_widget_unmap, 1, 0, 0, H_gtk_widget_unmap);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_realize" XG_POST, gxg_gtk_widget_realize, 1, 0, 0, H_gtk_widget_realize);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_unrealize" XG_POST, gxg_gtk_widget_unrealize, 1, 0, 0, H_gtk_widget_unrealize);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_queue_draw" XG_POST, gxg_gtk_widget_queue_draw, 1, 0, 0, H_gtk_widget_queue_draw);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_queue_draw_area" XG_POST, gxg_gtk_widget_queue_draw_area, 5, 0, 0, H_gtk_widget_queue_draw_area);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_queue_resize" XG_POST, gxg_gtk_widget_queue_resize, 1, 0, 0, H_gtk_widget_queue_resize);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_size_request" XG_POST, gxg_gtk_widget_size_request, 2, 0, 0, H_gtk_widget_size_request);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_size_allocate" XG_POST, gxg_gtk_widget_size_allocate, 2, 0, 0, H_gtk_widget_size_allocate);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_get_child_requisition" XG_POST, gxg_gtk_widget_get_child_requisition, 2, 0, 0, H_gtk_widget_get_child_requisition);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_add_accelerator" XG_POST, gxg_gtk_widget_add_accelerator, 6, 0, 0, H_gtk_widget_add_accelerator);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_remove_accelerator" XG_POST, gxg_gtk_widget_remove_accelerator, 4, 0, 0, H_gtk_widget_remove_accelerator);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_list_accel_closures" XG_POST, gxg_gtk_widget_list_accel_closures, 1, 0, 0, H_gtk_widget_list_accel_closures);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_mnemonic_activate" XG_POST, gxg_gtk_widget_mnemonic_activate, 2, 0, 0, H_gtk_widget_mnemonic_activate);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_event" XG_POST, gxg_gtk_widget_event, 2, 0, 0, H_gtk_widget_event);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_send_expose" XG_POST, gxg_gtk_widget_send_expose, 2, 0, 0, H_gtk_widget_send_expose);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_activate" XG_POST, gxg_gtk_widget_activate, 1, 0, 0, H_gtk_widget_activate);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_set_scroll_adjustments" XG_POST, gxg_gtk_widget_set_scroll_adjustments, 3, 0, 0, H_gtk_widget_set_scroll_adjustments);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_reparent" XG_POST, gxg_gtk_widget_reparent, 2, 0, 0, H_gtk_widget_reparent);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_intersect" XG_POST, gxg_gtk_widget_intersect, 3, 0, 0, H_gtk_widget_intersect);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_region_intersect" XG_POST, gxg_gtk_widget_region_intersect, 2, 0, 0, H_gtk_widget_region_intersect);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_freeze_child_notify" XG_POST, gxg_gtk_widget_freeze_child_notify, 1, 0, 0, H_gtk_widget_freeze_child_notify);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_child_notify" XG_POST, gxg_gtk_widget_child_notify, 2, 0, 0, H_gtk_widget_child_notify);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_thaw_child_notify" XG_POST, gxg_gtk_widget_thaw_child_notify, 1, 0, 0, H_gtk_widget_thaw_child_notify);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_is_focus" XG_POST, gxg_gtk_widget_is_focus, 1, 0, 0, H_gtk_widget_is_focus);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_grab_focus" XG_POST, gxg_gtk_widget_grab_focus, 1, 0, 0, H_gtk_widget_grab_focus);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_grab_default" XG_POST, gxg_gtk_widget_grab_default, 1, 0, 0, H_gtk_widget_grab_default);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_set_name" XG_POST, gxg_gtk_widget_set_name, 2, 0, 0, H_gtk_widget_set_name);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_get_name" XG_POST, gxg_gtk_widget_get_name, 1, 0, 0, H_gtk_widget_get_name);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_set_state" XG_POST, gxg_gtk_widget_set_state, 2, 0, 0, H_gtk_widget_set_state);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_set_sensitive" XG_POST, gxg_gtk_widget_set_sensitive, 2, 0, 0, H_gtk_widget_set_sensitive);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_set_app_paintable" XG_POST, gxg_gtk_widget_set_app_paintable, 2, 0, 0, H_gtk_widget_set_app_paintable);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_set_double_buffered" XG_POST, gxg_gtk_widget_set_double_buffered, 2, 0, 0, H_gtk_widget_set_double_buffered);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_set_redraw_on_allocate" XG_POST, gxg_gtk_widget_set_redraw_on_allocate, 2, 0, 0, H_gtk_widget_set_redraw_on_allocate);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_set_parent" XG_POST, gxg_gtk_widget_set_parent, 2, 0, 0, H_gtk_widget_set_parent);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_set_parent_window" XG_POST, gxg_gtk_widget_set_parent_window, 2, 0, 0, H_gtk_widget_set_parent_window);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_set_child_visible" XG_POST, gxg_gtk_widget_set_child_visible, 2, 0, 0, H_gtk_widget_set_child_visible);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_set_accel_path" XG_POST, gxg_gtk_widget_set_accel_path, 3, 0, 0, H_gtk_widget_set_accel_path);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_get_child_visible" XG_POST, gxg_gtk_widget_get_child_visible, 1, 0, 0, H_gtk_widget_get_child_visible);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_get_parent" XG_POST, gxg_gtk_widget_get_parent, 1, 0, 0, H_gtk_widget_get_parent);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_get_parent_window" XG_POST, gxg_gtk_widget_get_parent_window, 1, 0, 0, H_gtk_widget_get_parent_window);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_child_focus" XG_POST, gxg_gtk_widget_child_focus, 2, 0, 0, H_gtk_widget_child_focus);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_set_size_request" XG_POST, gxg_gtk_widget_set_size_request, 3, 0, 0, H_gtk_widget_set_size_request);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_get_size_request" XG_POST, gxg_gtk_widget_get_size_request, 1, 2, 0, H_gtk_widget_get_size_request);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_set_events" XG_POST, gxg_gtk_widget_set_events, 2, 0, 0, H_gtk_widget_set_events);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_add_events" XG_POST, gxg_gtk_widget_add_events, 2, 0, 0, H_gtk_widget_add_events);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_set_extension_events" XG_POST, gxg_gtk_widget_set_extension_events, 2, 0, 0, H_gtk_widget_set_extension_events);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_get_extension_events" XG_POST, gxg_gtk_widget_get_extension_events, 1, 0, 0, H_gtk_widget_get_extension_events);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_get_toplevel" XG_POST, gxg_gtk_widget_get_toplevel, 1, 0, 0, H_gtk_widget_get_toplevel);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_get_ancestor" XG_POST, gxg_gtk_widget_get_ancestor, 2, 0, 0, H_gtk_widget_get_ancestor);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_get_colormap" XG_POST, gxg_gtk_widget_get_colormap, 1, 0, 0, H_gtk_widget_get_colormap);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_get_visual" XG_POST, gxg_gtk_widget_get_visual, 1, 0, 0, H_gtk_widget_get_visual);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_get_settings" XG_POST, gxg_gtk_widget_get_settings, 1, 0, 0, H_gtk_widget_get_settings);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_get_accessible" XG_POST, gxg_gtk_widget_get_accessible, 1, 0, 0, H_gtk_widget_get_accessible);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_set_colormap" XG_POST, gxg_gtk_widget_set_colormap, 2, 0, 0, H_gtk_widget_set_colormap);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_get_events" XG_POST, gxg_gtk_widget_get_events, 1, 0, 0, H_gtk_widget_get_events);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_get_pointer" XG_POST, gxg_gtk_widget_get_pointer, 1, 2, 0, H_gtk_widget_get_pointer);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_is_ancestor" XG_POST, gxg_gtk_widget_is_ancestor, 2, 0, 0, H_gtk_widget_is_ancestor);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_translate_coordinates" XG_POST, gxg_gtk_widget_translate_coordinates, 4, 2, 0, H_gtk_widget_translate_coordinates);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_hide_on_delete" XG_POST, gxg_gtk_widget_hide_on_delete, 1, 0, 0, H_gtk_widget_hide_on_delete);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_set_style" XG_POST, gxg_gtk_widget_set_style, 2, 0, 0, H_gtk_widget_set_style);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_ensure_style" XG_POST, gxg_gtk_widget_ensure_style, 1, 0, 0, H_gtk_widget_ensure_style);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_get_style" XG_POST, gxg_gtk_widget_get_style, 1, 0, 0, H_gtk_widget_get_style);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_modify_style" XG_POST, gxg_gtk_widget_modify_style, 2, 0, 0, H_gtk_widget_modify_style);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_get_modifier_style" XG_POST, gxg_gtk_widget_get_modifier_style, 1, 0, 0, H_gtk_widget_get_modifier_style);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_modify_fg" XG_POST, gxg_gtk_widget_modify_fg, 3, 0, 0, H_gtk_widget_modify_fg);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_modify_bg" XG_POST, gxg_gtk_widget_modify_bg, 3, 0, 0, H_gtk_widget_modify_bg);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_modify_text" XG_POST, gxg_gtk_widget_modify_text, 3, 0, 0, H_gtk_widget_modify_text);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_modify_base" XG_POST, gxg_gtk_widget_modify_base, 3, 0, 0, H_gtk_widget_modify_base);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_modify_font" XG_POST, gxg_gtk_widget_modify_font, 2, 0, 0, H_gtk_widget_modify_font);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_create_pango_context" XG_POST, gxg_gtk_widget_create_pango_context, 1, 0, 0, H_gtk_widget_create_pango_context);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_get_pango_context" XG_POST, gxg_gtk_widget_get_pango_context, 1, 0, 0, H_gtk_widget_get_pango_context);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_create_pango_layout" XG_POST, gxg_gtk_widget_create_pango_layout, 2, 0, 0, H_gtk_widget_create_pango_layout);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_render_icon" XG_POST, gxg_gtk_widget_render_icon, 4, 0, 0, H_gtk_widget_render_icon);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_set_composite_name" XG_POST, gxg_gtk_widget_set_composite_name, 2, 0, 0, H_gtk_widget_set_composite_name);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_get_composite_name" XG_POST, gxg_gtk_widget_get_composite_name, 1, 0, 0, H_gtk_widget_get_composite_name);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_reset_rc_styles" XG_POST, gxg_gtk_widget_reset_rc_styles, 1, 0, 0, H_gtk_widget_reset_rc_styles);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_push_colormap" XG_POST, gxg_gtk_widget_push_colormap, 1, 0, 0, H_gtk_widget_push_colormap);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_push_composite_child" XG_POST, gxg_gtk_widget_push_composite_child, 0, 0, 0, H_gtk_widget_push_composite_child);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_pop_composite_child" XG_POST, gxg_gtk_widget_pop_composite_child, 0, 0, 0, H_gtk_widget_pop_composite_child);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_pop_colormap" XG_POST, gxg_gtk_widget_pop_colormap, 0, 0, 0, H_gtk_widget_pop_colormap);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_style_get_property" XG_POST, gxg_gtk_widget_style_get_property, 3, 0, 0, H_gtk_widget_style_get_property);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_style_get" XG_POST, gxg_gtk_widget_style_get, 2, 0, 0, H_gtk_widget_style_get);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_set_default_colormap" XG_POST, gxg_gtk_widget_set_default_colormap, 1, 0, 0, H_gtk_widget_set_default_colormap);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_get_default_style" XG_POST, gxg_gtk_widget_get_default_style, 0, 0, 0, H_gtk_widget_get_default_style);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_get_default_colormap" XG_POST, gxg_gtk_widget_get_default_colormap, 0, 0, 0, H_gtk_widget_get_default_colormap);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_get_default_visual" XG_POST, gxg_gtk_widget_get_default_visual, 0, 0, 0, H_gtk_widget_get_default_visual);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_set_direction" XG_POST, gxg_gtk_widget_set_direction, 2, 0, 0, H_gtk_widget_set_direction);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_get_direction" XG_POST, gxg_gtk_widget_get_direction, 1, 0, 0, H_gtk_widget_get_direction);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_set_default_direction" XG_POST, gxg_gtk_widget_set_default_direction, 1, 0, 0, H_gtk_widget_set_default_direction);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_get_default_direction" XG_POST, gxg_gtk_widget_get_default_direction, 0, 0, 0, H_gtk_widget_get_default_direction);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_shape_combine_mask" XG_POST, gxg_gtk_widget_shape_combine_mask, 4, 0, 0, H_gtk_widget_shape_combine_mask);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_reset_shapes" XG_POST, gxg_gtk_widget_reset_shapes, 1, 0, 0, H_gtk_widget_reset_shapes);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_path" XG_POST, gxg_gtk_widget_path, 2, 2, 0, H_gtk_widget_path);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_class_path" XG_POST, gxg_gtk_widget_class_path, 2, 2, 0, H_gtk_widget_class_path);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_requisition_copy" XG_POST, gxg_gtk_requisition_copy, 1, 0, 0, H_gtk_requisition_copy);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_requisition_free" XG_POST, gxg_gtk_requisition_free, 1, 0, 0, H_gtk_requisition_free);
  XEN_DEFINE_PROCEDURE(XG_PRE "_gtk_widget_get_aux_info" XG_POST, gxg__gtk_widget_get_aux_info, 2, 0, 0, H__gtk_widget_get_aux_info);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_window_get_type" XG_POST, gxg_gtk_window_get_type, 0, 0, 0, H_gtk_window_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_window_new" XG_POST, gxg_gtk_window_new, 1, 0, 0, H_gtk_window_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_window_set_title" XG_POST, gxg_gtk_window_set_title, 2, 0, 0, H_gtk_window_set_title);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_window_get_title" XG_POST, gxg_gtk_window_get_title, 1, 0, 0, H_gtk_window_get_title);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_window_set_wmclass" XG_POST, gxg_gtk_window_set_wmclass, 3, 0, 0, H_gtk_window_set_wmclass);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_window_set_role" XG_POST, gxg_gtk_window_set_role, 2, 0, 0, H_gtk_window_set_role);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_window_get_role" XG_POST, gxg_gtk_window_get_role, 1, 0, 0, H_gtk_window_get_role);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_window_add_accel_group" XG_POST, gxg_gtk_window_add_accel_group, 2, 0, 0, H_gtk_window_add_accel_group);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_window_remove_accel_group" XG_POST, gxg_gtk_window_remove_accel_group, 2, 0, 0, H_gtk_window_remove_accel_group);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_window_set_position" XG_POST, gxg_gtk_window_set_position, 2, 0, 0, H_gtk_window_set_position);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_window_activate_focus" XG_POST, gxg_gtk_window_activate_focus, 1, 0, 0, H_gtk_window_activate_focus);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_window_set_focus" XG_POST, gxg_gtk_window_set_focus, 2, 0, 0, H_gtk_window_set_focus);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_window_get_focus" XG_POST, gxg_gtk_window_get_focus, 1, 0, 0, H_gtk_window_get_focus);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_window_set_default" XG_POST, gxg_gtk_window_set_default, 2, 0, 0, H_gtk_window_set_default);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_window_activate_default" XG_POST, gxg_gtk_window_activate_default, 1, 0, 0, H_gtk_window_activate_default);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_window_set_transient_for" XG_POST, gxg_gtk_window_set_transient_for, 2, 0, 0, H_gtk_window_set_transient_for);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_window_get_transient_for" XG_POST, gxg_gtk_window_get_transient_for, 1, 0, 0, H_gtk_window_get_transient_for);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_window_set_type_hint" XG_POST, gxg_gtk_window_set_type_hint, 2, 0, 0, H_gtk_window_set_type_hint);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_window_get_type_hint" XG_POST, gxg_gtk_window_get_type_hint, 1, 0, 0, H_gtk_window_get_type_hint);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_window_set_destroy_with_parent" XG_POST, gxg_gtk_window_set_destroy_with_parent, 2, 0, 0, H_gtk_window_set_destroy_with_parent);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_window_get_destroy_with_parent" XG_POST, gxg_gtk_window_get_destroy_with_parent, 1, 0, 0, H_gtk_window_get_destroy_with_parent);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_window_set_resizable" XG_POST, gxg_gtk_window_set_resizable, 2, 0, 0, H_gtk_window_set_resizable);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_window_get_resizable" XG_POST, gxg_gtk_window_get_resizable, 1, 0, 0, H_gtk_window_get_resizable);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_window_set_gravity" XG_POST, gxg_gtk_window_set_gravity, 2, 0, 0, H_gtk_window_set_gravity);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_window_get_gravity" XG_POST, gxg_gtk_window_get_gravity, 1, 0, 0, H_gtk_window_get_gravity);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_window_set_geometry_hints" XG_POST, gxg_gtk_window_set_geometry_hints, 4, 0, 0, H_gtk_window_set_geometry_hints);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_window_set_has_frame" XG_POST, gxg_gtk_window_set_has_frame, 2, 0, 0, H_gtk_window_set_has_frame);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_window_get_has_frame" XG_POST, gxg_gtk_window_get_has_frame, 1, 0, 0, H_gtk_window_get_has_frame);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_window_set_frame_dimensions" XG_POST, gxg_gtk_window_set_frame_dimensions, 5, 0, 0, H_gtk_window_set_frame_dimensions);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_window_get_frame_dimensions" XG_POST, gxg_gtk_window_get_frame_dimensions, 1, 4, 0, H_gtk_window_get_frame_dimensions);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_window_set_decorated" XG_POST, gxg_gtk_window_set_decorated, 2, 0, 0, H_gtk_window_set_decorated);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_window_get_decorated" XG_POST, gxg_gtk_window_get_decorated, 1, 0, 0, H_gtk_window_get_decorated);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_window_set_icon_list" XG_POST, gxg_gtk_window_set_icon_list, 2, 0, 0, H_gtk_window_set_icon_list);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_window_get_icon_list" XG_POST, gxg_gtk_window_get_icon_list, 1, 0, 0, H_gtk_window_get_icon_list);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_window_set_icon" XG_POST, gxg_gtk_window_set_icon, 2, 0, 0, H_gtk_window_set_icon);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_window_get_icon" XG_POST, gxg_gtk_window_get_icon, 1, 0, 0, H_gtk_window_get_icon);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_window_set_default_icon_list" XG_POST, gxg_gtk_window_set_default_icon_list, 1, 0, 0, H_gtk_window_set_default_icon_list);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_window_get_default_icon_list" XG_POST, gxg_gtk_window_get_default_icon_list, 0, 0, 0, H_gtk_window_get_default_icon_list);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_window_set_modal" XG_POST, gxg_gtk_window_set_modal, 2, 0, 0, H_gtk_window_set_modal);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_window_get_modal" XG_POST, gxg_gtk_window_get_modal, 1, 0, 0, H_gtk_window_get_modal);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_window_list_toplevels" XG_POST, gxg_gtk_window_list_toplevels, 0, 0, 0, H_gtk_window_list_toplevels);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_window_add_mnemonic" XG_POST, gxg_gtk_window_add_mnemonic, 3, 0, 0, H_gtk_window_add_mnemonic);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_window_remove_mnemonic" XG_POST, gxg_gtk_window_remove_mnemonic, 3, 0, 0, H_gtk_window_remove_mnemonic);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_window_mnemonic_activate" XG_POST, gxg_gtk_window_mnemonic_activate, 3, 0, 0, H_gtk_window_mnemonic_activate);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_window_set_mnemonic_modifier" XG_POST, gxg_gtk_window_set_mnemonic_modifier, 2, 0, 0, H_gtk_window_set_mnemonic_modifier);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_window_get_mnemonic_modifier" XG_POST, gxg_gtk_window_get_mnemonic_modifier, 1, 0, 0, H_gtk_window_get_mnemonic_modifier);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_window_present" XG_POST, gxg_gtk_window_present, 1, 0, 0, H_gtk_window_present);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_window_iconify" XG_POST, gxg_gtk_window_iconify, 1, 0, 0, H_gtk_window_iconify);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_window_deiconify" XG_POST, gxg_gtk_window_deiconify, 1, 0, 0, H_gtk_window_deiconify);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_window_stick" XG_POST, gxg_gtk_window_stick, 1, 0, 0, H_gtk_window_stick);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_window_unstick" XG_POST, gxg_gtk_window_unstick, 1, 0, 0, H_gtk_window_unstick);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_window_maximize" XG_POST, gxg_gtk_window_maximize, 1, 0, 0, H_gtk_window_maximize);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_window_unmaximize" XG_POST, gxg_gtk_window_unmaximize, 1, 0, 0, H_gtk_window_unmaximize);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_window_begin_resize_drag" XG_POST, gxg_gtk_window_begin_resize_drag, 6, 0, 0, H_gtk_window_begin_resize_drag);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_window_begin_move_drag" XG_POST, gxg_gtk_window_begin_move_drag, 5, 0, 0, H_gtk_window_begin_move_drag);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_window_set_default_size" XG_POST, gxg_gtk_window_set_default_size, 3, 0, 0, H_gtk_window_set_default_size);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_window_get_default_size" XG_POST, gxg_gtk_window_get_default_size, 1, 2, 0, H_gtk_window_get_default_size);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_window_resize" XG_POST, gxg_gtk_window_resize, 3, 0, 0, H_gtk_window_resize);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_window_get_size" XG_POST, gxg_gtk_window_get_size, 3, 0, 0, H_gtk_window_get_size);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_window_move" XG_POST, gxg_gtk_window_move, 3, 0, 0, H_gtk_window_move);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_window_get_position" XG_POST, gxg_gtk_window_get_position, 1, 2, 0, H_gtk_window_get_position);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_window_parse_geometry" XG_POST, gxg_gtk_window_parse_geometry, 2, 0, 0, H_gtk_window_parse_geometry);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_window_reshow_with_initial_size" XG_POST, gxg_gtk_window_reshow_with_initial_size, 1, 0, 0, H_gtk_window_reshow_with_initial_size);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_window_group_get_type" XG_POST, gxg_gtk_window_group_get_type, 0, 0, 0, H_gtk_window_group_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_window_group_new" XG_POST, gxg_gtk_window_group_new, 0, 0, 0, H_gtk_window_group_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_color_get_type" XG_POST, gxg_pango_color_get_type, 0, 0, 0, H_pango_color_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_color_copy" XG_POST, gxg_pango_color_copy, 1, 0, 0, H_pango_color_copy);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_color_free" XG_POST, gxg_pango_color_free, 1, 0, 0, H_pango_color_free);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_color_parse" XG_POST, gxg_pango_color_parse, 2, 0, 0, H_pango_color_parse);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_attr_type_register" XG_POST, gxg_pango_attr_type_register, 1, 0, 0, H_pango_attr_type_register);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_attribute_copy" XG_POST, gxg_pango_attribute_copy, 1, 0, 0, H_pango_attribute_copy);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_attribute_destroy" XG_POST, gxg_pango_attribute_destroy, 1, 0, 0, H_pango_attribute_destroy);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_attribute_equal" XG_POST, gxg_pango_attribute_equal, 2, 0, 0, H_pango_attribute_equal);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_attr_language_new" XG_POST, gxg_pango_attr_language_new, 1, 0, 0, H_pango_attr_language_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_attr_family_new" XG_POST, gxg_pango_attr_family_new, 1, 0, 0, H_pango_attr_family_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_attr_foreground_new" XG_POST, gxg_pango_attr_foreground_new, 3, 0, 0, H_pango_attr_foreground_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_attr_background_new" XG_POST, gxg_pango_attr_background_new, 3, 0, 0, H_pango_attr_background_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_attr_size_new" XG_POST, gxg_pango_attr_size_new, 1, 0, 0, H_pango_attr_size_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_attr_style_new" XG_POST, gxg_pango_attr_style_new, 1, 0, 0, H_pango_attr_style_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_attr_weight_new" XG_POST, gxg_pango_attr_weight_new, 1, 0, 0, H_pango_attr_weight_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_attr_variant_new" XG_POST, gxg_pango_attr_variant_new, 1, 0, 0, H_pango_attr_variant_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_attr_stretch_new" XG_POST, gxg_pango_attr_stretch_new, 1, 0, 0, H_pango_attr_stretch_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_attr_font_desc_new" XG_POST, gxg_pango_attr_font_desc_new, 1, 0, 0, H_pango_attr_font_desc_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_attr_underline_new" XG_POST, gxg_pango_attr_underline_new, 1, 0, 0, H_pango_attr_underline_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_attr_strikethrough_new" XG_POST, gxg_pango_attr_strikethrough_new, 1, 0, 0, H_pango_attr_strikethrough_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_attr_rise_new" XG_POST, gxg_pango_attr_rise_new, 1, 0, 0, H_pango_attr_rise_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_attr_shape_new" XG_POST, gxg_pango_attr_shape_new, 2, 0, 0, H_pango_attr_shape_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_attr_scale_new" XG_POST, gxg_pango_attr_scale_new, 1, 0, 0, H_pango_attr_scale_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_attr_list_get_type" XG_POST, gxg_pango_attr_list_get_type, 0, 0, 0, H_pango_attr_list_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_attr_list_new" XG_POST, gxg_pango_attr_list_new, 0, 0, 0, H_pango_attr_list_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_attr_list_ref" XG_POST, gxg_pango_attr_list_ref, 1, 0, 0, H_pango_attr_list_ref);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_attr_list_unref" XG_POST, gxg_pango_attr_list_unref, 1, 0, 0, H_pango_attr_list_unref);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_attr_list_copy" XG_POST, gxg_pango_attr_list_copy, 1, 0, 0, H_pango_attr_list_copy);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_attr_list_insert" XG_POST, gxg_pango_attr_list_insert, 2, 0, 0, H_pango_attr_list_insert);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_attr_list_insert_before" XG_POST, gxg_pango_attr_list_insert_before, 2, 0, 0, H_pango_attr_list_insert_before);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_attr_list_change" XG_POST, gxg_pango_attr_list_change, 2, 0, 0, H_pango_attr_list_change);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_attr_list_splice" XG_POST, gxg_pango_attr_list_splice, 4, 0, 0, H_pango_attr_list_splice);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_attr_list_get_iterator" XG_POST, gxg_pango_attr_list_get_iterator, 1, 0, 0, H_pango_attr_list_get_iterator);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_attr_iterator_range" XG_POST, gxg_pango_attr_iterator_range, 3, 0, 0, H_pango_attr_iterator_range);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_attr_iterator_next" XG_POST, gxg_pango_attr_iterator_next, 1, 0, 0, H_pango_attr_iterator_next);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_attr_iterator_copy" XG_POST, gxg_pango_attr_iterator_copy, 1, 0, 0, H_pango_attr_iterator_copy);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_attr_iterator_destroy" XG_POST, gxg_pango_attr_iterator_destroy, 1, 0, 0, H_pango_attr_iterator_destroy);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_attr_iterator_get" XG_POST, gxg_pango_attr_iterator_get, 2, 0, 0, H_pango_attr_iterator_get);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_attr_iterator_get_font" XG_POST, gxg_pango_attr_iterator_get_font, 2, 2, 0, H_pango_attr_iterator_get_font);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_parse_markup" XG_POST, gxg_pango_parse_markup, 6, 1, 0, H_pango_parse_markup);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_break" XG_POST, gxg_pango_break, 5, 0, 0, H_pango_break);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_find_paragraph_boundary" XG_POST, gxg_pango_find_paragraph_boundary, 4, 0, 0, H_pango_find_paragraph_boundary);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_get_log_attrs" XG_POST, gxg_pango_get_log_attrs, 6, 0, 0, H_pango_get_log_attrs);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_context_get_type" XG_POST, gxg_pango_context_get_type, 0, 0, 0, H_pango_context_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_context_list_families" XG_POST, gxg_pango_context_list_families, 1, 2, 0, H_pango_context_list_families);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_context_load_font" XG_POST, gxg_pango_context_load_font, 2, 0, 0, H_pango_context_load_font);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_context_load_fontset" XG_POST, gxg_pango_context_load_fontset, 3, 0, 0, H_pango_context_load_fontset);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_context_get_metrics" XG_POST, gxg_pango_context_get_metrics, 3, 0, 0, H_pango_context_get_metrics);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_context_set_font_description" XG_POST, gxg_pango_context_set_font_description, 2, 0, 0, H_pango_context_set_font_description);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_context_get_font_description" XG_POST, gxg_pango_context_get_font_description, 1, 0, 0, H_pango_context_get_font_description);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_context_get_language" XG_POST, gxg_pango_context_get_language, 1, 0, 0, H_pango_context_get_language);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_context_set_language" XG_POST, gxg_pango_context_set_language, 2, 0, 0, H_pango_context_set_language);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_context_set_base_dir" XG_POST, gxg_pango_context_set_base_dir, 2, 0, 0, H_pango_context_set_base_dir);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_context_get_base_dir" XG_POST, gxg_pango_context_get_base_dir, 1, 0, 0, H_pango_context_get_base_dir);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_itemize" XG_POST, gxg_pango_itemize, 6, 0, 0, H_pango_itemize);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_coverage_new" XG_POST, gxg_pango_coverage_new, 0, 0, 0, H_pango_coverage_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_coverage_ref" XG_POST, gxg_pango_coverage_ref, 1, 0, 0, H_pango_coverage_ref);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_coverage_unref" XG_POST, gxg_pango_coverage_unref, 1, 0, 0, H_pango_coverage_unref);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_coverage_copy" XG_POST, gxg_pango_coverage_copy, 1, 0, 0, H_pango_coverage_copy);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_coverage_get" XG_POST, gxg_pango_coverage_get, 2, 0, 0, H_pango_coverage_get);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_coverage_set" XG_POST, gxg_pango_coverage_set, 3, 0, 0, H_pango_coverage_set);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_coverage_max" XG_POST, gxg_pango_coverage_max, 2, 0, 0, H_pango_coverage_max);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_coverage_to_bytes" XG_POST, gxg_pango_coverage_to_bytes, 1, 2, 0, H_pango_coverage_to_bytes);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_coverage_from_bytes" XG_POST, gxg_pango_coverage_from_bytes, 2, 0, 0, H_pango_coverage_from_bytes);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_attr_type_get_type" XG_POST, gxg_pango_attr_type_get_type, 0, 0, 0, H_pango_attr_type_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_underline_get_type" XG_POST, gxg_pango_underline_get_type, 0, 0, 0, H_pango_underline_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_coverage_level_get_type" XG_POST, gxg_pango_coverage_level_get_type, 0, 0, 0, H_pango_coverage_level_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_style_get_type" XG_POST, gxg_pango_style_get_type, 0, 0, 0, H_pango_style_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_variant_get_type" XG_POST, gxg_pango_variant_get_type, 0, 0, 0, H_pango_variant_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_weight_get_type" XG_POST, gxg_pango_weight_get_type, 0, 0, 0, H_pango_weight_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_stretch_get_type" XG_POST, gxg_pango_stretch_get_type, 0, 0, 0, H_pango_stretch_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_font_mask_get_type" XG_POST, gxg_pango_font_mask_get_type, 0, 0, 0, H_pango_font_mask_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_alignment_get_type" XG_POST, gxg_pango_alignment_get_type, 0, 0, 0, H_pango_alignment_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_wrap_mode_get_type" XG_POST, gxg_pango_wrap_mode_get_type, 0, 0, 0, H_pango_wrap_mode_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_tab_align_get_type" XG_POST, gxg_pango_tab_align_get_type, 0, 0, 0, H_pango_tab_align_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_direction_get_type" XG_POST, gxg_pango_direction_get_type, 0, 0, 0, H_pango_direction_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_font_description_get_type" XG_POST, gxg_pango_font_description_get_type, 0, 0, 0, H_pango_font_description_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_font_description_new" XG_POST, gxg_pango_font_description_new, 0, 0, 0, H_pango_font_description_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_font_description_copy" XG_POST, gxg_pango_font_description_copy, 1, 0, 0, H_pango_font_description_copy);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_font_description_copy_static" XG_POST, gxg_pango_font_description_copy_static, 1, 0, 0, H_pango_font_description_copy_static);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_font_description_hash" XG_POST, gxg_pango_font_description_hash, 1, 0, 0, H_pango_font_description_hash);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_font_description_equal" XG_POST, gxg_pango_font_description_equal, 2, 0, 0, H_pango_font_description_equal);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_font_description_free" XG_POST, gxg_pango_font_description_free, 1, 0, 0, H_pango_font_description_free);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_font_descriptions_free" XG_POST, gxg_pango_font_descriptions_free, 2, 0, 0, H_pango_font_descriptions_free);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_font_description_set_family" XG_POST, gxg_pango_font_description_set_family, 2, 0, 0, H_pango_font_description_set_family);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_font_description_set_family_static" XG_POST, gxg_pango_font_description_set_family_static, 2, 0, 0, H_pango_font_description_set_family_static);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_font_description_get_family" XG_POST, gxg_pango_font_description_get_family, 1, 0, 0, H_pango_font_description_get_family);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_font_description_set_style" XG_POST, gxg_pango_font_description_set_style, 2, 0, 0, H_pango_font_description_set_style);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_font_description_get_style" XG_POST, gxg_pango_font_description_get_style, 1, 0, 0, H_pango_font_description_get_style);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_font_description_set_variant" XG_POST, gxg_pango_font_description_set_variant, 2, 0, 0, H_pango_font_description_set_variant);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_font_description_get_variant" XG_POST, gxg_pango_font_description_get_variant, 1, 0, 0, H_pango_font_description_get_variant);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_font_description_set_weight" XG_POST, gxg_pango_font_description_set_weight, 2, 0, 0, H_pango_font_description_set_weight);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_font_description_get_weight" XG_POST, gxg_pango_font_description_get_weight, 1, 0, 0, H_pango_font_description_get_weight);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_font_description_set_stretch" XG_POST, gxg_pango_font_description_set_stretch, 2, 0, 0, H_pango_font_description_set_stretch);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_font_description_get_stretch" XG_POST, gxg_pango_font_description_get_stretch, 1, 0, 0, H_pango_font_description_get_stretch);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_font_description_set_size" XG_POST, gxg_pango_font_description_set_size, 2, 0, 0, H_pango_font_description_set_size);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_font_description_get_size" XG_POST, gxg_pango_font_description_get_size, 1, 0, 0, H_pango_font_description_get_size);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_font_description_get_set_fields" XG_POST, gxg_pango_font_description_get_set_fields, 1, 0, 0, H_pango_font_description_get_set_fields);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_font_description_unset_fields" XG_POST, gxg_pango_font_description_unset_fields, 2, 0, 0, H_pango_font_description_unset_fields);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_font_description_merge" XG_POST, gxg_pango_font_description_merge, 3, 0, 0, H_pango_font_description_merge);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_font_description_merge_static" XG_POST, gxg_pango_font_description_merge_static, 3, 0, 0, H_pango_font_description_merge_static);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_font_description_better_match" XG_POST, gxg_pango_font_description_better_match, 3, 0, 0, H_pango_font_description_better_match);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_font_description_from_string" XG_POST, gxg_pango_font_description_from_string, 1, 0, 0, H_pango_font_description_from_string);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_font_description_to_string" XG_POST, gxg_pango_font_description_to_string, 1, 0, 0, H_pango_font_description_to_string);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_font_description_to_filename" XG_POST, gxg_pango_font_description_to_filename, 1, 0, 0, H_pango_font_description_to_filename);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_font_metrics_get_type" XG_POST, gxg_pango_font_metrics_get_type, 0, 0, 0, H_pango_font_metrics_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_font_metrics_ref" XG_POST, gxg_pango_font_metrics_ref, 1, 0, 0, H_pango_font_metrics_ref);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_font_metrics_unref" XG_POST, gxg_pango_font_metrics_unref, 1, 0, 0, H_pango_font_metrics_unref);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_font_metrics_get_ascent" XG_POST, gxg_pango_font_metrics_get_ascent, 1, 0, 0, H_pango_font_metrics_get_ascent);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_font_metrics_get_descent" XG_POST, gxg_pango_font_metrics_get_descent, 1, 0, 0, H_pango_font_metrics_get_descent);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_font_metrics_get_approximate_char_width" XG_POST, gxg_pango_font_metrics_get_approximate_char_width, 1, 0, 0, H_pango_font_metrics_get_approximate_char_width);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_font_metrics_get_approximate_digit_width" XG_POST, gxg_pango_font_metrics_get_approximate_digit_width, 1, 0, 0, H_pango_font_metrics_get_approximate_digit_width);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_font_face_get_type" XG_POST, gxg_pango_font_face_get_type, 0, 0, 0, H_pango_font_face_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_font_face_describe" XG_POST, gxg_pango_font_face_describe, 1, 0, 0, H_pango_font_face_describe);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_font_face_get_face_name" XG_POST, gxg_pango_font_face_get_face_name, 1, 0, 0, H_pango_font_face_get_face_name);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_font_get_type" XG_POST, gxg_pango_font_get_type, 0, 0, 0, H_pango_font_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_font_describe" XG_POST, gxg_pango_font_describe, 1, 0, 0, H_pango_font_describe);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_font_get_coverage" XG_POST, gxg_pango_font_get_coverage, 2, 0, 0, H_pango_font_get_coverage);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_font_find_shaper" XG_POST, gxg_pango_font_find_shaper, 3, 0, 0, H_pango_font_find_shaper);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_font_get_metrics" XG_POST, gxg_pango_font_get_metrics, 2, 0, 0, H_pango_font_get_metrics);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_font_get_glyph_extents" XG_POST, gxg_pango_font_get_glyph_extents, 4, 0, 0, H_pango_font_get_glyph_extents);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_font_map_get_type" XG_POST, gxg_pango_font_map_get_type, 0, 0, 0, H_pango_font_map_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_font_map_load_font" XG_POST, gxg_pango_font_map_load_font, 3, 0, 0, H_pango_font_map_load_font);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_font_map_load_fontset" XG_POST, gxg_pango_font_map_load_fontset, 4, 0, 0, H_pango_font_map_load_fontset);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_font_map_list_families" XG_POST, gxg_pango_font_map_list_families, 1, 2, 0, H_pango_font_map_list_families);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_glyph_string_new" XG_POST, gxg_pango_glyph_string_new, 0, 0, 0, H_pango_glyph_string_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_glyph_string_set_size" XG_POST, gxg_pango_glyph_string_set_size, 2, 0, 0, H_pango_glyph_string_set_size);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_glyph_string_get_type" XG_POST, gxg_pango_glyph_string_get_type, 0, 0, 0, H_pango_glyph_string_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_glyph_string_copy" XG_POST, gxg_pango_glyph_string_copy, 1, 0, 0, H_pango_glyph_string_copy);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_glyph_string_free" XG_POST, gxg_pango_glyph_string_free, 1, 0, 0, H_pango_glyph_string_free);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_glyph_string_extents" XG_POST, gxg_pango_glyph_string_extents, 4, 0, 0, H_pango_glyph_string_extents);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_glyph_string_extents_range" XG_POST, gxg_pango_glyph_string_extents_range, 6, 0, 0, H_pango_glyph_string_extents_range);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_glyph_string_get_logical_widths" XG_POST, gxg_pango_glyph_string_get_logical_widths, 5, 0, 0, H_pango_glyph_string_get_logical_widths);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_glyph_string_index_to_x" XG_POST, gxg_pango_glyph_string_index_to_x, 7, 0, 0, H_pango_glyph_string_index_to_x);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_glyph_string_x_to_index" XG_POST, gxg_pango_glyph_string_x_to_index, 7, 0, 0, H_pango_glyph_string_x_to_index);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_shape" XG_POST, gxg_pango_shape, 4, 0, 0, H_pango_shape);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_reorder_items" XG_POST, gxg_pango_reorder_items, 1, 0, 0, H_pango_reorder_items);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_item_new" XG_POST, gxg_pango_item_new, 0, 0, 0, H_pango_item_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_item_copy" XG_POST, gxg_pango_item_copy, 1, 0, 0, H_pango_item_copy);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_item_free" XG_POST, gxg_pango_item_free, 1, 0, 0, H_pango_item_free);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_item_split" XG_POST, gxg_pango_item_split, 3, 0, 0, H_pango_item_split);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_layout_get_type" XG_POST, gxg_pango_layout_get_type, 0, 0, 0, H_pango_layout_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_layout_new" XG_POST, gxg_pango_layout_new, 1, 0, 0, H_pango_layout_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_layout_copy" XG_POST, gxg_pango_layout_copy, 1, 0, 0, H_pango_layout_copy);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_layout_get_context" XG_POST, gxg_pango_layout_get_context, 1, 0, 0, H_pango_layout_get_context);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_layout_set_attributes" XG_POST, gxg_pango_layout_set_attributes, 2, 0, 0, H_pango_layout_set_attributes);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_layout_get_attributes" XG_POST, gxg_pango_layout_get_attributes, 1, 0, 0, H_pango_layout_get_attributes);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_layout_set_text" XG_POST, gxg_pango_layout_set_text, 3, 0, 0, H_pango_layout_set_text);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_layout_get_text" XG_POST, gxg_pango_layout_get_text, 1, 0, 0, H_pango_layout_get_text);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_layout_set_markup" XG_POST, gxg_pango_layout_set_markup, 3, 0, 0, H_pango_layout_set_markup);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_layout_set_markup_with_accel" XG_POST, gxg_pango_layout_set_markup_with_accel, 5, 0, 0, H_pango_layout_set_markup_with_accel);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_layout_set_font_description" XG_POST, gxg_pango_layout_set_font_description, 2, 0, 0, H_pango_layout_set_font_description);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_layout_set_width" XG_POST, gxg_pango_layout_set_width, 2, 0, 0, H_pango_layout_set_width);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_layout_get_width" XG_POST, gxg_pango_layout_get_width, 1, 0, 0, H_pango_layout_get_width);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_layout_set_wrap" XG_POST, gxg_pango_layout_set_wrap, 2, 0, 0, H_pango_layout_set_wrap);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_layout_get_wrap" XG_POST, gxg_pango_layout_get_wrap, 1, 0, 0, H_pango_layout_get_wrap);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_layout_set_indent" XG_POST, gxg_pango_layout_set_indent, 2, 0, 0, H_pango_layout_set_indent);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_layout_get_indent" XG_POST, gxg_pango_layout_get_indent, 1, 0, 0, H_pango_layout_get_indent);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_layout_set_spacing" XG_POST, gxg_pango_layout_set_spacing, 2, 0, 0, H_pango_layout_set_spacing);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_layout_get_spacing" XG_POST, gxg_pango_layout_get_spacing, 1, 0, 0, H_pango_layout_get_spacing);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_layout_set_justify" XG_POST, gxg_pango_layout_set_justify, 2, 0, 0, H_pango_layout_set_justify);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_layout_get_justify" XG_POST, gxg_pango_layout_get_justify, 1, 0, 0, H_pango_layout_get_justify);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_layout_set_alignment" XG_POST, gxg_pango_layout_set_alignment, 2, 0, 0, H_pango_layout_set_alignment);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_layout_get_alignment" XG_POST, gxg_pango_layout_get_alignment, 1, 0, 0, H_pango_layout_get_alignment);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_layout_set_tabs" XG_POST, gxg_pango_layout_set_tabs, 2, 0, 0, H_pango_layout_set_tabs);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_layout_get_tabs" XG_POST, gxg_pango_layout_get_tabs, 1, 0, 0, H_pango_layout_get_tabs);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_layout_set_single_paragraph_mode" XG_POST, gxg_pango_layout_set_single_paragraph_mode, 2, 0, 0, H_pango_layout_set_single_paragraph_mode);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_layout_get_single_paragraph_mode" XG_POST, gxg_pango_layout_get_single_paragraph_mode, 1, 0, 0, H_pango_layout_get_single_paragraph_mode);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_layout_context_changed" XG_POST, gxg_pango_layout_context_changed, 1, 0, 0, H_pango_layout_context_changed);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_layout_get_log_attrs" XG_POST, gxg_pango_layout_get_log_attrs, 1, 2, 0, H_pango_layout_get_log_attrs);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_layout_index_to_pos" XG_POST, gxg_pango_layout_index_to_pos, 3, 0, 0, H_pango_layout_index_to_pos);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_layout_get_cursor_pos" XG_POST, gxg_pango_layout_get_cursor_pos, 4, 0, 0, H_pango_layout_get_cursor_pos);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_layout_move_cursor_visually" XG_POST, gxg_pango_layout_move_cursor_visually, 7, 0, 0, H_pango_layout_move_cursor_visually);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_layout_xy_to_index" XG_POST, gxg_pango_layout_xy_to_index, 5, 0, 0, H_pango_layout_xy_to_index);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_layout_get_extents" XG_POST, gxg_pango_layout_get_extents, 3, 0, 0, H_pango_layout_get_extents);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_layout_get_pixel_extents" XG_POST, gxg_pango_layout_get_pixel_extents, 3, 0, 0, H_pango_layout_get_pixel_extents);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_layout_get_size" XG_POST, gxg_pango_layout_get_size, 3, 0, 0, H_pango_layout_get_size);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_layout_get_pixel_size" XG_POST, gxg_pango_layout_get_pixel_size, 3, 0, 0, H_pango_layout_get_pixel_size);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_layout_get_line_count" XG_POST, gxg_pango_layout_get_line_count, 1, 0, 0, H_pango_layout_get_line_count);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_layout_get_line" XG_POST, gxg_pango_layout_get_line, 2, 0, 0, H_pango_layout_get_line);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_layout_get_lines" XG_POST, gxg_pango_layout_get_lines, 1, 0, 0, H_pango_layout_get_lines);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_layout_line_ref" XG_POST, gxg_pango_layout_line_ref, 1, 0, 0, H_pango_layout_line_ref);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_layout_line_unref" XG_POST, gxg_pango_layout_line_unref, 1, 0, 0, H_pango_layout_line_unref);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_layout_line_x_to_index" XG_POST, gxg_pango_layout_line_x_to_index, 4, 0, 0, H_pango_layout_line_x_to_index);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_layout_line_index_to_x" XG_POST, gxg_pango_layout_line_index_to_x, 4, 0, 0, H_pango_layout_line_index_to_x);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_layout_line_get_x_ranges" XG_POST, gxg_pango_layout_line_get_x_ranges, 3, 2, 0, H_pango_layout_line_get_x_ranges);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_layout_line_get_extents" XG_POST, gxg_pango_layout_line_get_extents, 3, 0, 0, H_pango_layout_line_get_extents);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_layout_line_get_pixel_extents" XG_POST, gxg_pango_layout_line_get_pixel_extents, 3, 0, 0, H_pango_layout_line_get_pixel_extents);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_layout_get_iter" XG_POST, gxg_pango_layout_get_iter, 1, 0, 0, H_pango_layout_get_iter);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_layout_iter_free" XG_POST, gxg_pango_layout_iter_free, 1, 0, 0, H_pango_layout_iter_free);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_layout_iter_get_index" XG_POST, gxg_pango_layout_iter_get_index, 1, 0, 0, H_pango_layout_iter_get_index);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_layout_iter_get_run" XG_POST, gxg_pango_layout_iter_get_run, 1, 0, 0, H_pango_layout_iter_get_run);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_layout_iter_get_line" XG_POST, gxg_pango_layout_iter_get_line, 1, 0, 0, H_pango_layout_iter_get_line);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_layout_iter_at_last_line" XG_POST, gxg_pango_layout_iter_at_last_line, 1, 0, 0, H_pango_layout_iter_at_last_line);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_layout_iter_next_char" XG_POST, gxg_pango_layout_iter_next_char, 1, 0, 0, H_pango_layout_iter_next_char);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_layout_iter_next_cluster" XG_POST, gxg_pango_layout_iter_next_cluster, 1, 0, 0, H_pango_layout_iter_next_cluster);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_layout_iter_next_run" XG_POST, gxg_pango_layout_iter_next_run, 1, 0, 0, H_pango_layout_iter_next_run);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_layout_iter_next_line" XG_POST, gxg_pango_layout_iter_next_line, 1, 0, 0, H_pango_layout_iter_next_line);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_layout_iter_get_char_extents" XG_POST, gxg_pango_layout_iter_get_char_extents, 2, 0, 0, H_pango_layout_iter_get_char_extents);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_layout_iter_get_cluster_extents" XG_POST, gxg_pango_layout_iter_get_cluster_extents, 3, 0, 0, H_pango_layout_iter_get_cluster_extents);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_layout_iter_get_run_extents" XG_POST, gxg_pango_layout_iter_get_run_extents, 3, 0, 0, H_pango_layout_iter_get_run_extents);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_layout_iter_get_line_extents" XG_POST, gxg_pango_layout_iter_get_line_extents, 3, 0, 0, H_pango_layout_iter_get_line_extents);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_layout_iter_get_line_yrange" XG_POST, gxg_pango_layout_iter_get_line_yrange, 3, 0, 0, H_pango_layout_iter_get_line_yrange);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_layout_iter_get_layout_extents" XG_POST, gxg_pango_layout_iter_get_layout_extents, 3, 0, 0, H_pango_layout_iter_get_layout_extents);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_layout_iter_get_baseline" XG_POST, gxg_pango_layout_iter_get_baseline, 1, 0, 0, H_pango_layout_iter_get_baseline);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_language_get_type" XG_POST, gxg_pango_language_get_type, 0, 0, 0, H_pango_language_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_language_from_string" XG_POST, gxg_pango_language_from_string, 1, 0, 0, H_pango_language_from_string);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_language_matches" XG_POST, gxg_pango_language_matches, 2, 0, 0, H_pango_language_matches);
#endif
#if PANGO_ENABLE_ENGINE && PANGO_ENABLE_BACKEND
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_default_break" XG_POST, gxg_pango_default_break, 5, 0, 0, H_pango_default_break);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_context_new" XG_POST, gxg_pango_context_new, 0, 0, 0, H_pango_context_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "script_engine_list" XG_POST, gxg_script_engine_list, 0, 2, 0, H_script_engine_list);
  XEN_DEFINE_PROCEDURE(XG_PRE "script_engine_load" XG_POST, gxg_script_engine_load, 1, 0, 0, H_script_engine_load);
  XEN_DEFINE_PROCEDURE(XG_PRE "script_engine_unload" XG_POST, gxg_script_engine_unload, 1, 0, 0, H_script_engine_unload);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_font_metrics_new" XG_POST, gxg_pango_font_metrics_new, 0, 0, 0, H_pango_font_metrics_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_font_family_get_type" XG_POST, gxg_pango_font_family_get_type, 0, 0, 0, H_pango_font_family_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_font_family_list_faces" XG_POST, gxg_pango_font_family_list_faces, 1, 2, 0, H_pango_font_family_list_faces);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_font_family_get_name" XG_POST, gxg_pango_font_family_get_name, 1, 0, 0, H_pango_font_family_get_name);
  XEN_DEFINE_PROCEDURE(XG_PRE "pango_context_set_font_map" XG_POST, gxg_pango_context_set_font_map, 2, 0, 0, H_pango_context_set_font_map);
#endif
#endif

#if (!(defined(GDK_DISABLE_DEPRECATED))) && (!(defined(GTK_DISABLE_DEPRECATED))) && (!(defined(GDK_PIXBUF_DISABLE_DEPRECATED)))
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_window_get_deskrelative_origin" XG_POST, gxg_gdk_window_get_deskrelative_origin, 1, 2, 0, H_gdk_window_get_deskrelative_origin);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_pixbuf_ref" XG_POST, gxg_gdk_pixbuf_ref, 1, 0, 0, H_gdk_pixbuf_ref);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_pixbuf_unref" XG_POST, gxg_gdk_pixbuf_unref, 1, 0, 0, H_gdk_pixbuf_unref);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_pixbuf_animation_ref" XG_POST, gxg_gdk_pixbuf_animation_ref, 1, 0, 0, H_gdk_pixbuf_animation_ref);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_pixbuf_animation_unref" XG_POST, gxg_gdk_pixbuf_animation_unref, 1, 0, 0, H_gdk_pixbuf_animation_unref);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_colors_alloc" XG_POST, gxg_gdk_colors_alloc, 6, 0, 0, H_gdk_colors_alloc);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_colors_free" XG_POST, gxg_gdk_colors_free, 4, 0, 0, H_gdk_colors_free);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_color_white" XG_POST, gxg_gdk_color_white, 2, 0, 0, H_gdk_color_white);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_color_black" XG_POST, gxg_gdk_color_black, 2, 0, 0, H_gdk_color_black);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_color_alloc" XG_POST, gxg_gdk_color_alloc, 2, 0, 0, H_gdk_color_alloc);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_color_change" XG_POST, gxg_gdk_color_change, 2, 0, 0, H_gdk_color_change);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_drag_context_ref" XG_POST, gxg_gdk_drag_context_ref, 1, 0, 0, H_gdk_drag_context_ref);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_drag_context_unref" XG_POST, gxg_gdk_drag_context_unref, 1, 0, 0, H_gdk_drag_context_unref);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_drawable_set_data" XG_POST, gxg_gdk_drawable_set_data, 4, 0, 0, H_gdk_drawable_set_data);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_drawable_get_data" XG_POST, gxg_gdk_drawable_get_data, 2, 0, 0, H_gdk_drawable_get_data);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_drawable_ref" XG_POST, gxg_gdk_drawable_ref, 1, 0, 0, H_gdk_drawable_ref);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_drawable_unref" XG_POST, gxg_gdk_drawable_unref, 1, 0, 0, H_gdk_drawable_unref);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_draw_string" XG_POST, gxg_gdk_draw_string, 6, 0, 0, H_gdk_draw_string);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_draw_text" XG_POST, gxg_gdk_draw_text, 7, 0, 0, H_gdk_draw_text);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_draw_text_wc" XG_POST, gxg_gdk_draw_text_wc, 7, 0, 0, H_gdk_draw_text_wc);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_font_get_type" XG_POST, gxg_gdk_font_get_type, 0, 0, 0, H_gdk_font_get_type);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_font_load" XG_POST, gxg_gdk_font_load, 1, 0, 0, H_gdk_font_load);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_fontset_load" XG_POST, gxg_gdk_fontset_load, 1, 0, 0, H_gdk_fontset_load);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_font_from_description" XG_POST, gxg_gdk_font_from_description, 1, 0, 0, H_gdk_font_from_description);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_font_ref" XG_POST, gxg_gdk_font_ref, 1, 0, 0, H_gdk_font_ref);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_font_unref" XG_POST, gxg_gdk_font_unref, 1, 0, 0, H_gdk_font_unref);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_font_id" XG_POST, gxg_gdk_font_id, 1, 0, 0, H_gdk_font_id);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_font_equal" XG_POST, gxg_gdk_font_equal, 2, 0, 0, H_gdk_font_equal);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_string_width" XG_POST, gxg_gdk_string_width, 2, 0, 0, H_gdk_string_width);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_text_width" XG_POST, gxg_gdk_text_width, 3, 0, 0, H_gdk_text_width);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_text_width_wc" XG_POST, gxg_gdk_text_width_wc, 3, 0, 0, H_gdk_text_width_wc);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_char_width" XG_POST, gxg_gdk_char_width, 2, 0, 0, H_gdk_char_width);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_char_width_wc" XG_POST, gxg_gdk_char_width_wc, 2, 0, 0, H_gdk_char_width_wc);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_string_measure" XG_POST, gxg_gdk_string_measure, 2, 0, 0, H_gdk_string_measure);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_text_measure" XG_POST, gxg_gdk_text_measure, 3, 0, 0, H_gdk_text_measure);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_char_measure" XG_POST, gxg_gdk_char_measure, 2, 0, 0, H_gdk_char_measure);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_string_height" XG_POST, gxg_gdk_string_height, 2, 0, 0, H_gdk_string_height);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_text_height" XG_POST, gxg_gdk_text_height, 3, 0, 0, H_gdk_text_height);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_char_height" XG_POST, gxg_gdk_char_height, 2, 0, 0, H_gdk_char_height);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_text_extents" XG_POST, gxg_gdk_text_extents, 8, 0, 0, H_gdk_text_extents);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_text_extents_wc" XG_POST, gxg_gdk_text_extents_wc, 8, 0, 0, H_gdk_text_extents_wc);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_string_extents" XG_POST, gxg_gdk_string_extents, 7, 0, 0, H_gdk_string_extents);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_gc_ref" XG_POST, gxg_gdk_gc_ref, 1, 0, 0, H_gdk_gc_ref);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_gc_unref" XG_POST, gxg_gdk_gc_unref, 1, 0, 0, H_gdk_gc_unref);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_gc_set_font" XG_POST, gxg_gdk_gc_set_font, 2, 0, 0, H_gdk_gc_set_font);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_exit" XG_POST, gxg_gdk_exit, 1, 0, 0, H_gdk_exit);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_set_use_xshm" XG_POST, gxg_gdk_set_use_xshm, 1, 0, 0, H_gdk_set_use_xshm);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_get_use_xshm" XG_POST, gxg_gdk_get_use_xshm, 0, 0, 0, H_gdk_get_use_xshm);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_input_add_full" XG_POST, gxg_gdk_input_add_full, 5, 0, 0, H_gdk_input_add_full);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_input_add" XG_POST, gxg_gdk_input_add, 3, 1, 0, H_gdk_input_add);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_input_remove" XG_POST, gxg_gdk_input_remove, 1, 0, 0, H_gdk_input_remove);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_image_get" XG_POST, gxg_gdk_image_get, 5, 0, 0, H_gdk_image_get);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_image_ref" XG_POST, gxg_gdk_image_ref, 1, 0, 0, H_gdk_image_ref);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_image_unref" XG_POST, gxg_gdk_image_unref, 1, 0, 0, H_gdk_image_unref);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_image_put_pixel" XG_POST, gxg_gdk_image_put_pixel, 4, 0, 0, H_gdk_image_put_pixel);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_image_get_pixel" XG_POST, gxg_gdk_image_get_pixel, 3, 0, 0, H_gdk_image_get_pixel);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_image_set_colormap" XG_POST, gxg_gdk_image_set_colormap, 2, 0, 0, H_gdk_image_set_colormap);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_image_get_colormap" XG_POST, gxg_gdk_image_get_colormap, 1, 0, 0, H_gdk_image_get_colormap);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_rgb_init" XG_POST, gxg_gdk_rgb_init, 0, 0, 0, H_gdk_rgb_init);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_rgb_xpixel_from_rgb" XG_POST, gxg_gdk_rgb_xpixel_from_rgb, 1, 0, 0, H_gdk_rgb_xpixel_from_rgb);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_rgb_gc_set_foreground" XG_POST, gxg_gdk_rgb_gc_set_foreground, 2, 0, 0, H_gdk_rgb_gc_set_foreground);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_rgb_gc_set_background" XG_POST, gxg_gdk_rgb_gc_set_background, 2, 0, 0, H_gdk_rgb_gc_set_background);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_window_set_hints" XG_POST, gxg_gdk_window_set_hints, 8, 0, 0, H_gdk_window_set_hints);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_button_box_set_child_size" XG_POST, gxg_gtk_button_box_set_child_size, 3, 0, 0, H_gtk_button_box_set_child_size);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_button_box_set_child_ipadding" XG_POST, gxg_gtk_button_box_set_child_ipadding, 3, 0, 0, H_gtk_button_box_set_child_ipadding);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_button_box_get_child_size" XG_POST, gxg_gtk_button_box_get_child_size, 3, 0, 0, H_gtk_button_box_get_child_size);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_button_box_get_child_ipadding" XG_POST, gxg_gtk_button_box_get_child_ipadding, 3, 0, 0, H_gtk_button_box_get_child_ipadding);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_check_menu_item_set_show_toggle" XG_POST, gxg_gtk_check_menu_item_set_show_toggle, 2, 0, 0, H_gtk_check_menu_item_set_show_toggle);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_clist_get_type" XG_POST, gxg_gtk_clist_get_type, 0, 0, 0, H_gtk_clist_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_clist_new" XG_POST, gxg_gtk_clist_new, 1, 0, 0, H_gtk_clist_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_clist_new_with_titles" XG_POST, gxg_gtk_clist_new_with_titles, 2, 0, 0, H_gtk_clist_new_with_titles);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_clist_set_hadjustment" XG_POST, gxg_gtk_clist_set_hadjustment, 2, 0, 0, H_gtk_clist_set_hadjustment);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_clist_set_vadjustment" XG_POST, gxg_gtk_clist_set_vadjustment, 2, 0, 0, H_gtk_clist_set_vadjustment);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_clist_get_hadjustment" XG_POST, gxg_gtk_clist_get_hadjustment, 1, 0, 0, H_gtk_clist_get_hadjustment);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_clist_get_vadjustment" XG_POST, gxg_gtk_clist_get_vadjustment, 1, 0, 0, H_gtk_clist_get_vadjustment);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_clist_set_shadow_type" XG_POST, gxg_gtk_clist_set_shadow_type, 2, 0, 0, H_gtk_clist_set_shadow_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_clist_set_selection_mode" XG_POST, gxg_gtk_clist_set_selection_mode, 2, 0, 0, H_gtk_clist_set_selection_mode);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_clist_set_reorderable" XG_POST, gxg_gtk_clist_set_reorderable, 2, 0, 0, H_gtk_clist_set_reorderable);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_clist_set_use_drag_icons" XG_POST, gxg_gtk_clist_set_use_drag_icons, 2, 0, 0, H_gtk_clist_set_use_drag_icons);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_clist_set_button_actions" XG_POST, gxg_gtk_clist_set_button_actions, 3, 0, 0, H_gtk_clist_set_button_actions);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_clist_freeze" XG_POST, gxg_gtk_clist_freeze, 1, 0, 0, H_gtk_clist_freeze);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_clist_thaw" XG_POST, gxg_gtk_clist_thaw, 1, 0, 0, H_gtk_clist_thaw);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_clist_column_titles_show" XG_POST, gxg_gtk_clist_column_titles_show, 1, 0, 0, H_gtk_clist_column_titles_show);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_clist_column_titles_hide" XG_POST, gxg_gtk_clist_column_titles_hide, 1, 0, 0, H_gtk_clist_column_titles_hide);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_clist_column_title_active" XG_POST, gxg_gtk_clist_column_title_active, 2, 0, 0, H_gtk_clist_column_title_active);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_clist_column_title_passive" XG_POST, gxg_gtk_clist_column_title_passive, 2, 0, 0, H_gtk_clist_column_title_passive);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_clist_column_titles_active" XG_POST, gxg_gtk_clist_column_titles_active, 1, 0, 0, H_gtk_clist_column_titles_active);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_clist_column_titles_passive" XG_POST, gxg_gtk_clist_column_titles_passive, 1, 0, 0, H_gtk_clist_column_titles_passive);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_clist_set_column_title" XG_POST, gxg_gtk_clist_set_column_title, 3, 0, 0, H_gtk_clist_set_column_title);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_clist_get_column_title" XG_POST, gxg_gtk_clist_get_column_title, 2, 0, 0, H_gtk_clist_get_column_title);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_clist_set_column_widget" XG_POST, gxg_gtk_clist_set_column_widget, 3, 0, 0, H_gtk_clist_set_column_widget);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_clist_get_column_widget" XG_POST, gxg_gtk_clist_get_column_widget, 2, 0, 0, H_gtk_clist_get_column_widget);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_clist_set_column_justification" XG_POST, gxg_gtk_clist_set_column_justification, 3, 0, 0, H_gtk_clist_set_column_justification);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_clist_set_column_visibility" XG_POST, gxg_gtk_clist_set_column_visibility, 3, 0, 0, H_gtk_clist_set_column_visibility);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_clist_set_column_resizeable" XG_POST, gxg_gtk_clist_set_column_resizeable, 3, 0, 0, H_gtk_clist_set_column_resizeable);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_clist_set_column_auto_resize" XG_POST, gxg_gtk_clist_set_column_auto_resize, 3, 0, 0, H_gtk_clist_set_column_auto_resize);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_clist_columns_autosize" XG_POST, gxg_gtk_clist_columns_autosize, 1, 0, 0, H_gtk_clist_columns_autosize);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_clist_optimal_column_width" XG_POST, gxg_gtk_clist_optimal_column_width, 2, 0, 0, H_gtk_clist_optimal_column_width);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_clist_set_column_width" XG_POST, gxg_gtk_clist_set_column_width, 3, 0, 0, H_gtk_clist_set_column_width);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_clist_set_column_min_width" XG_POST, gxg_gtk_clist_set_column_min_width, 3, 0, 0, H_gtk_clist_set_column_min_width);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_clist_set_column_max_width" XG_POST, gxg_gtk_clist_set_column_max_width, 3, 0, 0, H_gtk_clist_set_column_max_width);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_clist_set_row_height" XG_POST, gxg_gtk_clist_set_row_height, 2, 0, 0, H_gtk_clist_set_row_height);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_clist_moveto" XG_POST, gxg_gtk_clist_moveto, 5, 0, 0, H_gtk_clist_moveto);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_clist_row_is_visible" XG_POST, gxg_gtk_clist_row_is_visible, 2, 0, 0, H_gtk_clist_row_is_visible);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_clist_get_cell_type" XG_POST, gxg_gtk_clist_get_cell_type, 3, 0, 0, H_gtk_clist_get_cell_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_clist_set_text" XG_POST, gxg_gtk_clist_set_text, 4, 0, 0, H_gtk_clist_set_text);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_clist_get_text" XG_POST, gxg_gtk_clist_get_text, 3, 1, 0, H_gtk_clist_get_text);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_clist_set_pixmap" XG_POST, gxg_gtk_clist_set_pixmap, 5, 0, 0, H_gtk_clist_set_pixmap);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_clist_get_pixmap" XG_POST, gxg_gtk_clist_get_pixmap, 3, 2, 0, H_gtk_clist_get_pixmap);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_clist_set_pixtext" XG_POST, gxg_gtk_clist_set_pixtext, 7, 0, 0, H_gtk_clist_set_pixtext);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_clist_get_pixtext" XG_POST, gxg_gtk_clist_get_pixtext, 3, 4, 0, H_gtk_clist_get_pixtext);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_clist_set_foreground" XG_POST, gxg_gtk_clist_set_foreground, 3, 0, 0, H_gtk_clist_set_foreground);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_clist_set_background" XG_POST, gxg_gtk_clist_set_background, 3, 0, 0, H_gtk_clist_set_background);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_clist_set_cell_style" XG_POST, gxg_gtk_clist_set_cell_style, 4, 0, 0, H_gtk_clist_set_cell_style);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_clist_get_cell_style" XG_POST, gxg_gtk_clist_get_cell_style, 3, 0, 0, H_gtk_clist_get_cell_style);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_clist_set_row_style" XG_POST, gxg_gtk_clist_set_row_style, 3, 0, 0, H_gtk_clist_set_row_style);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_clist_get_row_style" XG_POST, gxg_gtk_clist_get_row_style, 2, 0, 0, H_gtk_clist_get_row_style);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_clist_set_shift" XG_POST, gxg_gtk_clist_set_shift, 5, 0, 0, H_gtk_clist_set_shift);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_clist_set_selectable" XG_POST, gxg_gtk_clist_set_selectable, 3, 0, 0, H_gtk_clist_set_selectable);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_clist_get_selectable" XG_POST, gxg_gtk_clist_get_selectable, 2, 0, 0, H_gtk_clist_get_selectable);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_clist_prepend" XG_POST, gxg_gtk_clist_prepend, 2, 0, 0, H_gtk_clist_prepend);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_clist_append" XG_POST, gxg_gtk_clist_append, 2, 0, 0, H_gtk_clist_append);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_clist_insert" XG_POST, gxg_gtk_clist_insert, 3, 0, 0, H_gtk_clist_insert);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_clist_remove" XG_POST, gxg_gtk_clist_remove, 2, 0, 0, H_gtk_clist_remove);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_clist_set_row_data" XG_POST, gxg_gtk_clist_set_row_data, 3, 0, 0, H_gtk_clist_set_row_data);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_clist_set_row_data_full" XG_POST, gxg_gtk_clist_set_row_data_full, 4, 0, 0, H_gtk_clist_set_row_data_full);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_clist_get_row_data" XG_POST, gxg_gtk_clist_get_row_data, 2, 0, 0, H_gtk_clist_get_row_data);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_clist_find_row_from_data" XG_POST, gxg_gtk_clist_find_row_from_data, 2, 0, 0, H_gtk_clist_find_row_from_data);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_clist_select_row" XG_POST, gxg_gtk_clist_select_row, 3, 0, 0, H_gtk_clist_select_row);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_clist_unselect_row" XG_POST, gxg_gtk_clist_unselect_row, 3, 0, 0, H_gtk_clist_unselect_row);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_clist_undo_selection" XG_POST, gxg_gtk_clist_undo_selection, 1, 0, 0, H_gtk_clist_undo_selection);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_clist_clear" XG_POST, gxg_gtk_clist_clear, 1, 0, 0, H_gtk_clist_clear);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_clist_get_selection_info" XG_POST, gxg_gtk_clist_get_selection_info, 5, 0, 0, H_gtk_clist_get_selection_info);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_clist_select_all" XG_POST, gxg_gtk_clist_select_all, 1, 0, 0, H_gtk_clist_select_all);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_clist_unselect_all" XG_POST, gxg_gtk_clist_unselect_all, 1, 0, 0, H_gtk_clist_unselect_all);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_clist_swap_rows" XG_POST, gxg_gtk_clist_swap_rows, 3, 0, 0, H_gtk_clist_swap_rows);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_clist_row_move" XG_POST, gxg_gtk_clist_row_move, 3, 0, 0, H_gtk_clist_row_move);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_clist_set_compare_func" XG_POST, gxg_gtk_clist_set_compare_func, 2, 0, 0, H_gtk_clist_set_compare_func);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_clist_set_sort_column" XG_POST, gxg_gtk_clist_set_sort_column, 2, 0, 0, H_gtk_clist_set_sort_column);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_clist_set_sort_type" XG_POST, gxg_gtk_clist_set_sort_type, 2, 0, 0, H_gtk_clist_set_sort_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_clist_sort" XG_POST, gxg_gtk_clist_sort, 1, 0, 0, H_gtk_clist_sort);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_clist_set_auto_sort" XG_POST, gxg_gtk_clist_set_auto_sort, 2, 0, 0, H_gtk_clist_set_auto_sort);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_color_selection_set_color" XG_POST, gxg_gtk_color_selection_set_color, 2, 0, 0, H_gtk_color_selection_set_color);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_color_selection_get_color" XG_POST, gxg_gtk_color_selection_get_color, 2, 0, 0, H_gtk_color_selection_get_color);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_color_selection_set_update_policy" XG_POST, gxg_gtk_color_selection_set_update_policy, 2, 0, 0, H_gtk_color_selection_set_update_policy);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_container_foreach_full" XG_POST, gxg_gtk_container_foreach_full, 5, 0, 0, H_gtk_container_foreach_full);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_container_propagate_expose" XG_POST, gxg_gtk_container_propagate_expose, 3, 0, 0, H_gtk_container_propagate_expose);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_container_set_focus_chain" XG_POST, gxg_gtk_container_set_focus_chain, 2, 0, 0, H_gtk_container_set_focus_chain);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_container_get_focus_chain" XG_POST, gxg_gtk_container_get_focus_chain, 2, 0, 0, H_gtk_container_get_focus_chain);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_container_unset_focus_chain" XG_POST, gxg_gtk_container_unset_focus_chain, 1, 0, 0, H_gtk_container_unset_focus_chain);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_container_set_reallocate_redraws" XG_POST, gxg_gtk_container_set_reallocate_redraws, 2, 0, 0, H_gtk_container_set_reallocate_redraws);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_container_set_focus_child" XG_POST, gxg_gtk_container_set_focus_child, 2, 0, 0, H_gtk_container_set_focus_child);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_container_set_focus_vadjustment" XG_POST, gxg_gtk_container_set_focus_vadjustment, 2, 0, 0, H_gtk_container_set_focus_vadjustment);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_container_get_focus_vadjustment" XG_POST, gxg_gtk_container_get_focus_vadjustment, 1, 0, 0, H_gtk_container_get_focus_vadjustment);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_container_set_focus_hadjustment" XG_POST, gxg_gtk_container_set_focus_hadjustment, 2, 0, 0, H_gtk_container_set_focus_hadjustment);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_container_get_focus_hadjustment" XG_POST, gxg_gtk_container_get_focus_hadjustment, 1, 0, 0, H_gtk_container_get_focus_hadjustment);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_container_resize_children" XG_POST, gxg_gtk_container_resize_children, 1, 0, 0, H_gtk_container_resize_children);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_container_child_type" XG_POST, gxg_gtk_container_child_type, 1, 0, 0, H_gtk_container_child_type);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_container_add_with_properties" XG_POST, gxg_gtk_container_add_with_properties, 3, 0, 0, H_gtk_container_add_with_properties);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_container_child_set" XG_POST, gxg_gtk_container_child_set, 3, 0, 0, H_gtk_container_child_set);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_container_child_get" XG_POST, gxg_gtk_container_child_get, 3, 0, 0, H_gtk_container_child_get);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_container_child_set_property" XG_POST, gxg_gtk_container_child_set_property, 4, 0, 0, H_gtk_container_child_set_property);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_container_child_get_property" XG_POST, gxg_gtk_container_child_get_property, 4, 0, 0, H_gtk_container_child_get_property);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_container_forall" XG_POST, gxg_gtk_container_forall, 2, 1, 0, H_gtk_container_forall);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_ctree_get_type" XG_POST, gxg_gtk_ctree_get_type, 0, 0, 0, H_gtk_ctree_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_ctree_new_with_titles" XG_POST, gxg_gtk_ctree_new_with_titles, 3, 0, 0, H_gtk_ctree_new_with_titles);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_ctree_new" XG_POST, gxg_gtk_ctree_new, 2, 0, 0, H_gtk_ctree_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_ctree_insert_node" XG_POST, gxg_gtk_ctree_insert_node, 0, 0, 1, H_gtk_ctree_insert_node);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_ctree_remove_node" XG_POST, gxg_gtk_ctree_remove_node, 2, 0, 0, H_gtk_ctree_remove_node);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_ctree_insert_gnode" XG_POST, gxg_gtk_ctree_insert_gnode, 6, 0, 0, H_gtk_ctree_insert_gnode);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_ctree_export_to_gnode" XG_POST, gxg_gtk_ctree_export_to_gnode, 6, 0, 0, H_gtk_ctree_export_to_gnode);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_ctree_post_recursive" XG_POST, gxg_gtk_ctree_post_recursive, 4, 0, 0, H_gtk_ctree_post_recursive);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_ctree_post_recursive_to_depth" XG_POST, gxg_gtk_ctree_post_recursive_to_depth, 5, 0, 0, H_gtk_ctree_post_recursive_to_depth);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_ctree_pre_recursive" XG_POST, gxg_gtk_ctree_pre_recursive, 4, 0, 0, H_gtk_ctree_pre_recursive);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_ctree_pre_recursive_to_depth" XG_POST, gxg_gtk_ctree_pre_recursive_to_depth, 5, 0, 0, H_gtk_ctree_pre_recursive_to_depth);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_ctree_is_viewable" XG_POST, gxg_gtk_ctree_is_viewable, 2, 0, 0, H_gtk_ctree_is_viewable);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_ctree_last" XG_POST, gxg_gtk_ctree_last, 2, 0, 0, H_gtk_ctree_last);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_ctree_find_node_ptr" XG_POST, gxg_gtk_ctree_find_node_ptr, 2, 0, 0, H_gtk_ctree_find_node_ptr);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_ctree_node_nth" XG_POST, gxg_gtk_ctree_node_nth, 2, 0, 0, H_gtk_ctree_node_nth);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_ctree_find" XG_POST, gxg_gtk_ctree_find, 3, 0, 0, H_gtk_ctree_find);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_ctree_is_ancestor" XG_POST, gxg_gtk_ctree_is_ancestor, 3, 0, 0, H_gtk_ctree_is_ancestor);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_ctree_find_by_row_data" XG_POST, gxg_gtk_ctree_find_by_row_data, 3, 0, 0, H_gtk_ctree_find_by_row_data);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_ctree_find_all_by_row_data" XG_POST, gxg_gtk_ctree_find_all_by_row_data, 3, 0, 0, H_gtk_ctree_find_all_by_row_data);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_ctree_find_by_row_data_custom" XG_POST, gxg_gtk_ctree_find_by_row_data_custom, 4, 0, 0, H_gtk_ctree_find_by_row_data_custom);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_ctree_find_all_by_row_data_custom" XG_POST, gxg_gtk_ctree_find_all_by_row_data_custom, 4, 0, 0, H_gtk_ctree_find_all_by_row_data_custom);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_ctree_is_hot_spot" XG_POST, gxg_gtk_ctree_is_hot_spot, 3, 0, 0, H_gtk_ctree_is_hot_spot);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_ctree_move" XG_POST, gxg_gtk_ctree_move, 4, 0, 0, H_gtk_ctree_move);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_ctree_expand" XG_POST, gxg_gtk_ctree_expand, 2, 0, 0, H_gtk_ctree_expand);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_ctree_expand_recursive" XG_POST, gxg_gtk_ctree_expand_recursive, 2, 0, 0, H_gtk_ctree_expand_recursive);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_ctree_expand_to_depth" XG_POST, gxg_gtk_ctree_expand_to_depth, 3, 0, 0, H_gtk_ctree_expand_to_depth);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_ctree_collapse" XG_POST, gxg_gtk_ctree_collapse, 2, 0, 0, H_gtk_ctree_collapse);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_ctree_collapse_recursive" XG_POST, gxg_gtk_ctree_collapse_recursive, 2, 0, 0, H_gtk_ctree_collapse_recursive);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_ctree_collapse_to_depth" XG_POST, gxg_gtk_ctree_collapse_to_depth, 3, 0, 0, H_gtk_ctree_collapse_to_depth);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_ctree_toggle_expansion" XG_POST, gxg_gtk_ctree_toggle_expansion, 2, 0, 0, H_gtk_ctree_toggle_expansion);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_ctree_toggle_expansion_recursive" XG_POST, gxg_gtk_ctree_toggle_expansion_recursive, 2, 0, 0, H_gtk_ctree_toggle_expansion_recursive);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_ctree_select" XG_POST, gxg_gtk_ctree_select, 2, 0, 0, H_gtk_ctree_select);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_ctree_select_recursive" XG_POST, gxg_gtk_ctree_select_recursive, 2, 0, 0, H_gtk_ctree_select_recursive);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_ctree_unselect" XG_POST, gxg_gtk_ctree_unselect, 2, 0, 0, H_gtk_ctree_unselect);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_ctree_unselect_recursive" XG_POST, gxg_gtk_ctree_unselect_recursive, 2, 0, 0, H_gtk_ctree_unselect_recursive);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_ctree_real_select_recursive" XG_POST, gxg_gtk_ctree_real_select_recursive, 3, 0, 0, H_gtk_ctree_real_select_recursive);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_ctree_node_set_text" XG_POST, gxg_gtk_ctree_node_set_text, 4, 0, 0, H_gtk_ctree_node_set_text);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_ctree_node_set_pixmap" XG_POST, gxg_gtk_ctree_node_set_pixmap, 5, 0, 0, H_gtk_ctree_node_set_pixmap);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_ctree_node_set_pixtext" XG_POST, gxg_gtk_ctree_node_set_pixtext, 7, 0, 0, H_gtk_ctree_node_set_pixtext);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_ctree_set_node_info" XG_POST, gxg_gtk_ctree_set_node_info, 0, 0, 1, H_gtk_ctree_set_node_info);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_ctree_node_set_shift" XG_POST, gxg_gtk_ctree_node_set_shift, 5, 0, 0, H_gtk_ctree_node_set_shift);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_ctree_node_set_selectable" XG_POST, gxg_gtk_ctree_node_set_selectable, 3, 0, 0, H_gtk_ctree_node_set_selectable);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_ctree_node_get_selectable" XG_POST, gxg_gtk_ctree_node_get_selectable, 2, 0, 0, H_gtk_ctree_node_get_selectable);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_ctree_node_get_cell_type" XG_POST, gxg_gtk_ctree_node_get_cell_type, 3, 0, 0, H_gtk_ctree_node_get_cell_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_ctree_node_get_text" XG_POST, gxg_gtk_ctree_node_get_text, 3, 1, 0, H_gtk_ctree_node_get_text);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_ctree_node_get_pixmap" XG_POST, gxg_gtk_ctree_node_get_pixmap, 3, 2, 0, H_gtk_ctree_node_get_pixmap);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_ctree_node_get_pixtext" XG_POST, gxg_gtk_ctree_node_get_pixtext, 3, 4, 0, H_gtk_ctree_node_get_pixtext);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_ctree_get_node_info" XG_POST, gxg_gtk_ctree_get_node_info, 0, 0, 1, H_gtk_ctree_get_node_info);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_ctree_node_set_row_style" XG_POST, gxg_gtk_ctree_node_set_row_style, 3, 0, 0, H_gtk_ctree_node_set_row_style);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_ctree_node_get_row_style" XG_POST, gxg_gtk_ctree_node_get_row_style, 2, 0, 0, H_gtk_ctree_node_get_row_style);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_ctree_node_set_cell_style" XG_POST, gxg_gtk_ctree_node_set_cell_style, 4, 0, 0, H_gtk_ctree_node_set_cell_style);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_ctree_node_get_cell_style" XG_POST, gxg_gtk_ctree_node_get_cell_style, 3, 0, 0, H_gtk_ctree_node_get_cell_style);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_ctree_node_set_foreground" XG_POST, gxg_gtk_ctree_node_set_foreground, 3, 0, 0, H_gtk_ctree_node_set_foreground);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_ctree_node_set_background" XG_POST, gxg_gtk_ctree_node_set_background, 3, 0, 0, H_gtk_ctree_node_set_background);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_ctree_node_set_row_data" XG_POST, gxg_gtk_ctree_node_set_row_data, 3, 0, 0, H_gtk_ctree_node_set_row_data);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_ctree_node_set_row_data_full" XG_POST, gxg_gtk_ctree_node_set_row_data_full, 4, 0, 0, H_gtk_ctree_node_set_row_data_full);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_ctree_node_get_row_data" XG_POST, gxg_gtk_ctree_node_get_row_data, 2, 0, 0, H_gtk_ctree_node_get_row_data);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_ctree_node_moveto" XG_POST, gxg_gtk_ctree_node_moveto, 5, 0, 0, H_gtk_ctree_node_moveto);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_ctree_node_is_visible" XG_POST, gxg_gtk_ctree_node_is_visible, 2, 0, 0, H_gtk_ctree_node_is_visible);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_ctree_set_indent" XG_POST, gxg_gtk_ctree_set_indent, 2, 0, 0, H_gtk_ctree_set_indent);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_ctree_set_spacing" XG_POST, gxg_gtk_ctree_set_spacing, 2, 0, 0, H_gtk_ctree_set_spacing);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_ctree_set_show_stub" XG_POST, gxg_gtk_ctree_set_show_stub, 2, 0, 0, H_gtk_ctree_set_show_stub);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_ctree_set_line_style" XG_POST, gxg_gtk_ctree_set_line_style, 2, 0, 0, H_gtk_ctree_set_line_style);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_ctree_set_expander_style" XG_POST, gxg_gtk_ctree_set_expander_style, 2, 0, 0, H_gtk_ctree_set_expander_style);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_ctree_set_drag_compare_func" XG_POST, gxg_gtk_ctree_set_drag_compare_func, 2, 0, 0, H_gtk_ctree_set_drag_compare_func);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_ctree_sort_node" XG_POST, gxg_gtk_ctree_sort_node, 2, 0, 0, H_gtk_ctree_sort_node);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_ctree_sort_recursive" XG_POST, gxg_gtk_ctree_sort_recursive, 2, 0, 0, H_gtk_ctree_sort_recursive);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_ctree_node_get_type" XG_POST, gxg_gtk_ctree_node_get_type, 0, 0, 0, H_gtk_ctree_node_get_type);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_drawing_area_size" XG_POST, gxg_gtk_drawing_area_size, 3, 0, 0, H_gtk_drawing_area_size);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_entry_new_with_max_length" XG_POST, gxg_gtk_entry_new_with_max_length, 1, 0, 0, H_gtk_entry_new_with_max_length);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_entry_append_text" XG_POST, gxg_gtk_entry_append_text, 2, 0, 0, H_gtk_entry_append_text);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_entry_prepend_text" XG_POST, gxg_gtk_entry_prepend_text, 2, 0, 0, H_gtk_entry_prepend_text);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_entry_set_position" XG_POST, gxg_gtk_entry_set_position, 2, 0, 0, H_gtk_entry_set_position);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_entry_select_region" XG_POST, gxg_gtk_entry_select_region, 3, 0, 0, H_gtk_entry_select_region);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_entry_set_editable" XG_POST, gxg_gtk_entry_set_editable, 2, 0, 0, H_gtk_entry_set_editable);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_font_selection_get_font" XG_POST, gxg_gtk_font_selection_get_font, 1, 0, 0, H_gtk_font_selection_get_font);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_font_selection_dialog_get_font" XG_POST, gxg_gtk_font_selection_dialog_get_font, 1, 0, 0, H_gtk_font_selection_dialog_get_font);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_hbutton_box_get_spacing_default" XG_POST, gxg_gtk_hbutton_box_get_spacing_default, 0, 0, 0, H_gtk_hbutton_box_get_spacing_default);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_hbutton_box_get_layout_default" XG_POST, gxg_gtk_hbutton_box_get_layout_default, 0, 0, 0, H_gtk_hbutton_box_get_layout_default);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_hbutton_box_set_spacing_default" XG_POST, gxg_gtk_hbutton_box_set_spacing_default, 1, 0, 0, H_gtk_hbutton_box_set_spacing_default);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_hbutton_box_set_layout_default" XG_POST, gxg_gtk_hbutton_box_set_layout_default, 1, 0, 0, H_gtk_hbutton_box_set_layout_default);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_image_set" XG_POST, gxg_gtk_image_set, 3, 0, 0, H_gtk_image_set);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_image_get" XG_POST, gxg_gtk_image_get, 1, 2, 0, H_gtk_image_get);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_item_factory_from_path" XG_POST, gxg_gtk_item_factory_from_path, 1, 0, 0, H_gtk_item_factory_from_path);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_item_factory_create_menu_entries" XG_POST, gxg_gtk_item_factory_create_menu_entries, 2, 0, 0, H_gtk_item_factory_create_menu_entries);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_item_factories_path_delete" XG_POST, gxg_gtk_item_factories_path_delete, 2, 0, 0, H_gtk_item_factories_path_delete);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_item_factory_create_items_ac" XG_POST, gxg_gtk_item_factory_create_items_ac, 5, 0, 0, H_gtk_item_factory_create_items_ac);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_label_get" XG_POST, gxg_gtk_label_get, 1, 1, 0, H_gtk_label_get);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_label_parse_uline" XG_POST, gxg_gtk_label_parse_uline, 2, 0, 0, H_gtk_label_parse_uline);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_layout_freeze" XG_POST, gxg_gtk_layout_freeze, 1, 0, 0, H_gtk_layout_freeze);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_layout_thaw" XG_POST, gxg_gtk_layout_thaw, 1, 0, 0, H_gtk_layout_thaw);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_list_get_type" XG_POST, gxg_gtk_list_get_type, 0, 0, 0, H_gtk_list_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_list_new" XG_POST, gxg_gtk_list_new, 0, 0, 0, H_gtk_list_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_list_insert_items" XG_POST, gxg_gtk_list_insert_items, 3, 0, 0, H_gtk_list_insert_items);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_list_append_items" XG_POST, gxg_gtk_list_append_items, 2, 0, 0, H_gtk_list_append_items);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_list_prepend_items" XG_POST, gxg_gtk_list_prepend_items, 2, 0, 0, H_gtk_list_prepend_items);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_list_remove_items" XG_POST, gxg_gtk_list_remove_items, 2, 0, 0, H_gtk_list_remove_items);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_list_remove_items_no_unref" XG_POST, gxg_gtk_list_remove_items_no_unref, 2, 0, 0, H_gtk_list_remove_items_no_unref);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_list_clear_items" XG_POST, gxg_gtk_list_clear_items, 3, 0, 0, H_gtk_list_clear_items);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_list_select_item" XG_POST, gxg_gtk_list_select_item, 2, 0, 0, H_gtk_list_select_item);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_list_unselect_item" XG_POST, gxg_gtk_list_unselect_item, 2, 0, 0, H_gtk_list_unselect_item);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_list_select_child" XG_POST, gxg_gtk_list_select_child, 2, 0, 0, H_gtk_list_select_child);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_list_unselect_child" XG_POST, gxg_gtk_list_unselect_child, 2, 0, 0, H_gtk_list_unselect_child);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_list_child_position" XG_POST, gxg_gtk_list_child_position, 2, 0, 0, H_gtk_list_child_position);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_list_set_selection_mode" XG_POST, gxg_gtk_list_set_selection_mode, 2, 0, 0, H_gtk_list_set_selection_mode);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_list_extend_selection" XG_POST, gxg_gtk_list_extend_selection, 4, 0, 0, H_gtk_list_extend_selection);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_list_start_selection" XG_POST, gxg_gtk_list_start_selection, 1, 0, 0, H_gtk_list_start_selection);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_list_end_selection" XG_POST, gxg_gtk_list_end_selection, 1, 0, 0, H_gtk_list_end_selection);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_list_select_all" XG_POST, gxg_gtk_list_select_all, 1, 0, 0, H_gtk_list_select_all);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_list_unselect_all" XG_POST, gxg_gtk_list_unselect_all, 1, 0, 0, H_gtk_list_unselect_all);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_list_scroll_horizontal" XG_POST, gxg_gtk_list_scroll_horizontal, 3, 0, 0, H_gtk_list_scroll_horizontal);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_list_scroll_vertical" XG_POST, gxg_gtk_list_scroll_vertical, 3, 0, 0, H_gtk_list_scroll_vertical);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_list_toggle_add_mode" XG_POST, gxg_gtk_list_toggle_add_mode, 1, 0, 0, H_gtk_list_toggle_add_mode);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_list_toggle_focus_row" XG_POST, gxg_gtk_list_toggle_focus_row, 1, 0, 0, H_gtk_list_toggle_focus_row);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_list_toggle_row" XG_POST, gxg_gtk_list_toggle_row, 2, 0, 0, H_gtk_list_toggle_row);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_list_undo_selection" XG_POST, gxg_gtk_list_undo_selection, 1, 0, 0, H_gtk_list_undo_selection);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_list_end_drag_selection" XG_POST, gxg_gtk_list_end_drag_selection, 1, 0, 0, H_gtk_list_end_drag_selection);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_list_item_get_type" XG_POST, gxg_gtk_list_item_get_type, 0, 0, 0, H_gtk_list_item_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_list_item_new" XG_POST, gxg_gtk_list_item_new, 0, 0, 0, H_gtk_list_item_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_list_item_new_with_label" XG_POST, gxg_gtk_list_item_new_with_label, 1, 0, 0, H_gtk_list_item_new_with_label);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_list_item_select" XG_POST, gxg_gtk_list_item_select, 1, 0, 0, H_gtk_list_item_select);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_list_item_deselect" XG_POST, gxg_gtk_list_item_deselect, 1, 0, 0, H_gtk_list_item_deselect);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_notebook_set_homogeneous_tabs" XG_POST, gxg_gtk_notebook_set_homogeneous_tabs, 2, 0, 0, H_gtk_notebook_set_homogeneous_tabs);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_notebook_set_tab_border" XG_POST, gxg_gtk_notebook_set_tab_border, 2, 0, 0, H_gtk_notebook_set_tab_border);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_notebook_set_tab_hborder" XG_POST, gxg_gtk_notebook_set_tab_hborder, 2, 0, 0, H_gtk_notebook_set_tab_hborder);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_notebook_set_tab_vborder" XG_POST, gxg_gtk_notebook_set_tab_vborder, 2, 0, 0, H_gtk_notebook_set_tab_vborder);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_object_ref" XG_POST, gxg_gtk_object_ref, 1, 0, 0, H_gtk_object_ref);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_object_unref" XG_POST, gxg_gtk_object_unref, 1, 0, 0, H_gtk_object_unref);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_object_weakref" XG_POST, gxg_gtk_object_weakref, 3, 0, 0, H_gtk_object_weakref);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_object_weakunref" XG_POST, gxg_gtk_object_weakunref, 3, 0, 0, H_gtk_object_weakunref);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_object_set_data" XG_POST, gxg_gtk_object_set_data, 3, 0, 0, H_gtk_object_set_data);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_object_set_data_full" XG_POST, gxg_gtk_object_set_data_full, 4, 0, 0, H_gtk_object_set_data_full);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_object_remove_data" XG_POST, gxg_gtk_object_remove_data, 2, 0, 0, H_gtk_object_remove_data);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_object_get_data" XG_POST, gxg_gtk_object_get_data, 2, 0, 0, H_gtk_object_get_data);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_object_remove_no_notify" XG_POST, gxg_gtk_object_remove_no_notify, 2, 0, 0, H_gtk_object_remove_no_notify);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_object_set_user_data" XG_POST, gxg_gtk_object_set_user_data, 2, 0, 0, H_gtk_object_set_user_data);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_object_get_user_data" XG_POST, gxg_gtk_object_get_user_data, 1, 0, 0, H_gtk_object_get_user_data);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_object_set_data_by_id" XG_POST, gxg_gtk_object_set_data_by_id, 3, 0, 0, H_gtk_object_set_data_by_id);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_object_set_data_by_id_full" XG_POST, gxg_gtk_object_set_data_by_id_full, 4, 0, 0, H_gtk_object_set_data_by_id_full);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_object_get_data_by_id" XG_POST, gxg_gtk_object_get_data_by_id, 2, 0, 0, H_gtk_object_get_data_by_id);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_object_remove_data_by_id" XG_POST, gxg_gtk_object_remove_data_by_id, 2, 0, 0, H_gtk_object_remove_data_by_id);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_object_remove_no_notify_by_id" XG_POST, gxg_gtk_object_remove_no_notify_by_id, 2, 0, 0, H_gtk_object_remove_no_notify_by_id);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_object_get" XG_POST, gxg_gtk_object_get, 2, 0, 0, H_gtk_object_get);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_object_set" XG_POST, gxg_gtk_object_set, 2, 0, 0, H_gtk_object_set);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_object_add_arg_type" XG_POST, gxg_gtk_object_add_arg_type, 4, 0, 0, H_gtk_object_add_arg_type);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_old_editable_get_type" XG_POST, gxg_gtk_old_editable_get_type, 0, 0, 0, H_gtk_old_editable_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_old_editable_claim_selection" XG_POST, gxg_gtk_old_editable_claim_selection, 3, 0, 0, H_gtk_old_editable_claim_selection);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_old_editable_changed" XG_POST, gxg_gtk_old_editable_changed, 1, 0, 0, H_gtk_old_editable_changed);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_pixmap_get_type" XG_POST, gxg_gtk_pixmap_get_type, 0, 0, 0, H_gtk_pixmap_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_pixmap_new" XG_POST, gxg_gtk_pixmap_new, 2, 0, 0, H_gtk_pixmap_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_pixmap_set" XG_POST, gxg_gtk_pixmap_set, 3, 0, 0, H_gtk_pixmap_set);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_pixmap_get" XG_POST, gxg_gtk_pixmap_get, 1, 2, 0, H_gtk_pixmap_get);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_pixmap_set_build_insensitive" XG_POST, gxg_gtk_pixmap_set_build_insensitive, 2, 0, 0, H_gtk_pixmap_set_build_insensitive);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_preview_get_type" XG_POST, gxg_gtk_preview_get_type, 0, 0, 0, H_gtk_preview_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_preview_uninit" XG_POST, gxg_gtk_preview_uninit, 0, 0, 0, H_gtk_preview_uninit);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_preview_new" XG_POST, gxg_gtk_preview_new, 1, 0, 0, H_gtk_preview_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_preview_size" XG_POST, gxg_gtk_preview_size, 3, 0, 0, H_gtk_preview_size);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_preview_put" XG_POST, gxg_gtk_preview_put, 9, 0, 0, H_gtk_preview_put);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_preview_draw_row" XG_POST, gxg_gtk_preview_draw_row, 5, 0, 0, H_gtk_preview_draw_row);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_preview_set_expand" XG_POST, gxg_gtk_preview_set_expand, 2, 0, 0, H_gtk_preview_set_expand);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_preview_set_gamma" XG_POST, gxg_gtk_preview_set_gamma, 1, 0, 0, H_gtk_preview_set_gamma);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_preview_set_color_cube" XG_POST, gxg_gtk_preview_set_color_cube, 4, 0, 0, H_gtk_preview_set_color_cube);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_preview_set_install_cmap" XG_POST, gxg_gtk_preview_set_install_cmap, 1, 0, 0, H_gtk_preview_set_install_cmap);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_preview_set_reserved" XG_POST, gxg_gtk_preview_set_reserved, 1, 0, 0, H_gtk_preview_set_reserved);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_preview_set_dither" XG_POST, gxg_gtk_preview_set_dither, 2, 0, 0, H_gtk_preview_set_dither);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_preview_get_visual" XG_POST, gxg_gtk_preview_get_visual, 0, 0, 0, H_gtk_preview_get_visual);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_preview_get_cmap" XG_POST, gxg_gtk_preview_get_cmap, 0, 0, 0, H_gtk_preview_get_cmap);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_preview_get_info" XG_POST, gxg_gtk_preview_get_info, 0, 0, 0, H_gtk_preview_get_info);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_preview_reset" XG_POST, gxg_gtk_preview_reset, 0, 0, 0, H_gtk_preview_reset);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_progress_bar_new_with_adjustment" XG_POST, gxg_gtk_progress_bar_new_with_adjustment, 1, 0, 0, H_gtk_progress_bar_new_with_adjustment);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_progress_bar_set_bar_style" XG_POST, gxg_gtk_progress_bar_set_bar_style, 2, 0, 0, H_gtk_progress_bar_set_bar_style);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_progress_bar_set_discrete_blocks" XG_POST, gxg_gtk_progress_bar_set_discrete_blocks, 2, 0, 0, H_gtk_progress_bar_set_discrete_blocks);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_progress_bar_set_activity_step" XG_POST, gxg_gtk_progress_bar_set_activity_step, 2, 0, 0, H_gtk_progress_bar_set_activity_step);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_progress_bar_set_activity_blocks" XG_POST, gxg_gtk_progress_bar_set_activity_blocks, 2, 0, 0, H_gtk_progress_bar_set_activity_blocks);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_progress_bar_update" XG_POST, gxg_gtk_progress_bar_update, 2, 0, 0, H_gtk_progress_bar_update);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_progress_get_type" XG_POST, gxg_gtk_progress_get_type, 0, 0, 0, H_gtk_progress_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_progress_set_show_text" XG_POST, gxg_gtk_progress_set_show_text, 2, 0, 0, H_gtk_progress_set_show_text);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_progress_set_text_alignment" XG_POST, gxg_gtk_progress_set_text_alignment, 3, 0, 0, H_gtk_progress_set_text_alignment);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_progress_set_format_string" XG_POST, gxg_gtk_progress_set_format_string, 2, 0, 0, H_gtk_progress_set_format_string);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_progress_set_adjustment" XG_POST, gxg_gtk_progress_set_adjustment, 2, 0, 0, H_gtk_progress_set_adjustment);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_progress_configure" XG_POST, gxg_gtk_progress_configure, 4, 0, 0, H_gtk_progress_configure);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_progress_set_percentage" XG_POST, gxg_gtk_progress_set_percentage, 2, 0, 0, H_gtk_progress_set_percentage);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_progress_set_value" XG_POST, gxg_gtk_progress_set_value, 2, 0, 0, H_gtk_progress_set_value);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_progress_get_value" XG_POST, gxg_gtk_progress_get_value, 1, 0, 0, H_gtk_progress_get_value);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_progress_set_activity_mode" XG_POST, gxg_gtk_progress_set_activity_mode, 2, 0, 0, H_gtk_progress_set_activity_mode);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_progress_get_current_text" XG_POST, gxg_gtk_progress_get_current_text, 1, 0, 0, H_gtk_progress_get_current_text);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_progress_get_text_from_value" XG_POST, gxg_gtk_progress_get_text_from_value, 2, 0, 0, H_gtk_progress_get_text_from_value);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_progress_get_current_percentage" XG_POST, gxg_gtk_progress_get_current_percentage, 1, 0, 0, H_gtk_progress_get_current_percentage);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_progress_get_percentage_from_value" XG_POST, gxg_gtk_progress_get_percentage_from_value, 2, 0, 0, H_gtk_progress_get_percentage_from_value);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_rc_add_widget_name_style" XG_POST, gxg_gtk_rc_add_widget_name_style, 2, 0, 0, H_gtk_rc_add_widget_name_style);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_rc_add_widget_class_style" XG_POST, gxg_gtk_rc_add_widget_class_style, 2, 0, 0, H_gtk_rc_add_widget_class_style);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_rc_add_class_style" XG_POST, gxg_gtk_rc_add_class_style, 2, 0, 0, H_gtk_rc_add_class_style);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_signal_connect" XG_POST, gxg_gtk_signal_connect, 3, 1, 0, H_gtk_signal_connect);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_signal_lookup" XG_POST, gxg_gtk_signal_lookup, 2, 0, 0, H_gtk_signal_lookup);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_signal_name" XG_POST, gxg_gtk_signal_name, 1, 0, 0, H_gtk_signal_name);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_signal_emit_stop" XG_POST, gxg_gtk_signal_emit_stop, 2, 0, 0, H_gtk_signal_emit_stop);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_signal_emit_stop_by_name" XG_POST, gxg_gtk_signal_emit_stop_by_name, 2, 0, 0, H_gtk_signal_emit_stop_by_name);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_signal_connect_after" XG_POST, gxg_gtk_signal_connect_after, 3, 1, 0, H_gtk_signal_connect_after);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_signal_connect_full" XG_POST, gxg_gtk_signal_connect_full, 8, 0, 0, H_gtk_signal_connect_full);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_signal_connect_while_alive" XG_POST, gxg_gtk_signal_connect_while_alive, 5, 0, 0, H_gtk_signal_connect_while_alive);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_signal_disconnect" XG_POST, gxg_gtk_signal_disconnect, 2, 0, 0, H_gtk_signal_disconnect);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_signal_disconnect_by_func" XG_POST, gxg_gtk_signal_disconnect_by_func, 2, 1, 0, H_gtk_signal_disconnect_by_func);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_signal_handler_block" XG_POST, gxg_gtk_signal_handler_block, 2, 0, 0, H_gtk_signal_handler_block);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_signal_handler_block_by_func" XG_POST, gxg_gtk_signal_handler_block_by_func, 2, 1, 0, H_gtk_signal_handler_block_by_func);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_signal_handler_unblock" XG_POST, gxg_gtk_signal_handler_unblock, 2, 0, 0, H_gtk_signal_handler_unblock);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_signal_handler_unblock_by_func" XG_POST, gxg_gtk_signal_handler_unblock_by_func, 2, 1, 0, H_gtk_signal_handler_unblock_by_func);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_signal_handler_pending" XG_POST, gxg_gtk_signal_handler_pending, 3, 0, 0, H_gtk_signal_handler_pending);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_signal_handler_pending_by_func" XG_POST, gxg_gtk_signal_handler_pending_by_func, 4, 1, 0, H_gtk_signal_handler_pending_by_func);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_signal_new" XG_POST, gxg_gtk_signal_new, 7, 0, 0, H_gtk_signal_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_signal_newv" XG_POST, gxg_gtk_signal_newv, 8, 0, 0, H_gtk_signal_newv);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_signal_emit" XG_POST, gxg_gtk_signal_emit, 2, 0, 0, H_gtk_signal_emit);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_signal_emit_by_name" XG_POST, gxg_gtk_signal_emit_by_name, 2, 0, 0, H_gtk_signal_emit_by_name);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_signal_compat_matched" XG_POST, gxg_gtk_signal_compat_matched, 5, 0, 0, H_gtk_signal_compat_matched);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_socket_steal" XG_POST, gxg_gtk_socket_steal, 2, 0, 0, H_gtk_socket_steal);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_style_ref" XG_POST, gxg_gtk_style_ref, 1, 0, 0, H_gtk_style_ref);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_style_unref" XG_POST, gxg_gtk_style_unref, 1, 0, 0, H_gtk_style_unref);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_style_get_font" XG_POST, gxg_gtk_style_get_font, 1, 0, 0, H_gtk_style_get_font);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_style_set_font" XG_POST, gxg_gtk_style_set_font, 2, 0, 0, H_gtk_style_set_font);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_draw_hline" XG_POST, gxg_gtk_draw_hline, 6, 0, 0, H_gtk_draw_hline);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_draw_vline" XG_POST, gxg_gtk_draw_vline, 6, 0, 0, H_gtk_draw_vline);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_draw_shadow" XG_POST, gxg_gtk_draw_shadow, 8, 0, 0, H_gtk_draw_shadow);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_draw_polygon" XG_POST, gxg_gtk_draw_polygon, 7, 0, 0, H_gtk_draw_polygon);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_draw_arrow" XG_POST, gxg_gtk_draw_arrow, 0, 0, 1, H_gtk_draw_arrow);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_draw_diamond" XG_POST, gxg_gtk_draw_diamond, 8, 0, 0, H_gtk_draw_diamond);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_draw_box" XG_POST, gxg_gtk_draw_box, 8, 0, 0, H_gtk_draw_box);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_draw_flat_box" XG_POST, gxg_gtk_draw_flat_box, 8, 0, 0, H_gtk_draw_flat_box);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_draw_check" XG_POST, gxg_gtk_draw_check, 8, 0, 0, H_gtk_draw_check);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_draw_option" XG_POST, gxg_gtk_draw_option, 8, 0, 0, H_gtk_draw_option);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_draw_tab" XG_POST, gxg_gtk_draw_tab, 8, 0, 0, H_gtk_draw_tab);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_draw_shadow_gap" XG_POST, gxg_gtk_draw_shadow_gap, 0, 0, 1, H_gtk_draw_shadow_gap);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_draw_box_gap" XG_POST, gxg_gtk_draw_box_gap, 0, 0, 1, H_gtk_draw_box_gap);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_draw_extension" XG_POST, gxg_gtk_draw_extension, 9, 0, 0, H_gtk_draw_extension);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_draw_focus" XG_POST, gxg_gtk_draw_focus, 6, 0, 0, H_gtk_draw_focus);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_draw_slider" XG_POST, gxg_gtk_draw_slider, 9, 0, 0, H_gtk_draw_slider);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_draw_handle" XG_POST, gxg_gtk_draw_handle, 9, 0, 0, H_gtk_draw_handle);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_draw_expander" XG_POST, gxg_gtk_draw_expander, 6, 0, 0, H_gtk_draw_expander);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_draw_layout" XG_POST, gxg_gtk_draw_layout, 7, 0, 0, H_gtk_draw_layout);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_draw_resize_grip" XG_POST, gxg_gtk_draw_resize_grip, 8, 0, 0, H_gtk_draw_resize_grip);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_draw_string" XG_POST, gxg_gtk_draw_string, 6, 0, 0, H_gtk_draw_string);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_paint_string" XG_POST, gxg_gtk_paint_string, 9, 0, 0, H_gtk_paint_string);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tips_query_get_type" XG_POST, gxg_gtk_tips_query_get_type, 0, 0, 0, H_gtk_tips_query_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tips_query_new" XG_POST, gxg_gtk_tips_query_new, 0, 0, 0, H_gtk_tips_query_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tips_query_start_query" XG_POST, gxg_gtk_tips_query_start_query, 1, 0, 0, H_gtk_tips_query_start_query);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tips_query_stop_query" XG_POST, gxg_gtk_tips_query_stop_query, 1, 0, 0, H_gtk_tips_query_stop_query);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tips_query_set_caller" XG_POST, gxg_gtk_tips_query_set_caller, 2, 0, 0, H_gtk_tips_query_set_caller);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tips_query_set_labels" XG_POST, gxg_gtk_tips_query_set_labels, 3, 0, 0, H_gtk_tips_query_set_labels);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tooltips_set_delay" XG_POST, gxg_gtk_tooltips_set_delay, 2, 0, 0, H_gtk_tooltips_set_delay);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_vbutton_box_get_spacing_default" XG_POST, gxg_gtk_vbutton_box_get_spacing_default, 0, 0, 0, H_gtk_vbutton_box_get_spacing_default);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_vbutton_box_set_spacing_default" XG_POST, gxg_gtk_vbutton_box_set_spacing_default, 1, 0, 0, H_gtk_vbutton_box_set_spacing_default);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_vbutton_box_get_layout_default" XG_POST, gxg_gtk_vbutton_box_get_layout_default, 0, 0, 0, H_gtk_vbutton_box_get_layout_default);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_vbutton_box_set_layout_default" XG_POST, gxg_gtk_vbutton_box_set_layout_default, 1, 0, 0, H_gtk_vbutton_box_set_layout_default);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_set" XG_POST, gxg_gtk_widget_set, 2, 0, 0, H_gtk_widget_set);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_queue_clear" XG_POST, gxg_gtk_widget_queue_clear, 1, 0, 0, H_gtk_widget_queue_clear);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_queue_clear_area" XG_POST, gxg_gtk_widget_queue_clear_area, 5, 0, 0, H_gtk_widget_queue_clear_area);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_draw" XG_POST, gxg_gtk_widget_draw, 2, 0, 0, H_gtk_widget_draw);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_set_uposition" XG_POST, gxg_gtk_widget_set_uposition, 3, 0, 0, H_gtk_widget_set_uposition);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_widget_set_usize" XG_POST, gxg_gtk_widget_set_usize, 3, 0, 0, H_gtk_widget_set_usize);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_window_set_policy" XG_POST, gxg_gtk_window_set_policy, 4, 0, 0, H_gtk_window_set_policy);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_colormap_ref" XG_POST, gxg_gdk_colormap_ref, 1, 0, 0, H_gdk_colormap_ref);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_colormap_unref" XG_POST, gxg_gdk_colormap_unref, 1, 0, 0, H_gdk_colormap_unref);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_colormap_get_system_size" XG_POST, gxg_gdk_colormap_get_system_size, 0, 0, 0, H_gdk_colormap_get_system_size);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_colormap_change" XG_POST, gxg_gdk_colormap_change, 2, 0, 0, H_gdk_colormap_change);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_colors_store" XG_POST, gxg_gdk_colors_store, 3, 0, 0, H_gdk_colors_store);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "GDK_ROOT_PARENT" XG_POST, gxg_GDK_ROOT_PARENT, 0, 0, 0, H_GDK_ROOT_PARENT);
#endif
#endif

#if GDK_ENABLE_BROKEN && GTK_ENABLE_BROKEN
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_set_adjustments" XG_POST, gxg_gtk_text_set_adjustments, 3, 0, 0, H_gtk_text_set_adjustments);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_set_point" XG_POST, gxg_gtk_text_set_point, 2, 0, 0, H_gtk_text_set_point);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_get_point" XG_POST, gxg_gtk_text_get_point, 1, 0, 0, H_gtk_text_get_point);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_get_length" XG_POST, gxg_gtk_text_get_length, 1, 0, 0, H_gtk_text_get_length);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_freeze" XG_POST, gxg_gtk_text_freeze, 1, 0, 0, H_gtk_text_freeze);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_thaw" XG_POST, gxg_gtk_text_thaw, 1, 0, 0, H_gtk_text_thaw);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_insert" XG_POST, gxg_gtk_text_insert, 6, 0, 0, H_gtk_text_insert);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_backward_delete" XG_POST, gxg_gtk_text_backward_delete, 2, 0, 0, H_gtk_text_backward_delete);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_forward_delete" XG_POST, gxg_gtk_text_forward_delete, 2, 0, 0, H_gtk_text_forward_delete);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_get_type" XG_POST, gxg_gtk_tree_get_type, 0, 0, 0, H_gtk_tree_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_new" XG_POST, gxg_gtk_tree_new, 0, 0, 0, H_gtk_tree_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_append" XG_POST, gxg_gtk_tree_append, 2, 0, 0, H_gtk_tree_append);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_prepend" XG_POST, gxg_gtk_tree_prepend, 2, 0, 0, H_gtk_tree_prepend);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_insert" XG_POST, gxg_gtk_tree_insert, 3, 0, 0, H_gtk_tree_insert);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_remove_items" XG_POST, gxg_gtk_tree_remove_items, 2, 0, 0, H_gtk_tree_remove_items);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_clear_items" XG_POST, gxg_gtk_tree_clear_items, 3, 0, 0, H_gtk_tree_clear_items);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_select_item" XG_POST, gxg_gtk_tree_select_item, 2, 0, 0, H_gtk_tree_select_item);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_unselect_item" XG_POST, gxg_gtk_tree_unselect_item, 2, 0, 0, H_gtk_tree_unselect_item);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_select_child" XG_POST, gxg_gtk_tree_select_child, 2, 0, 0, H_gtk_tree_select_child);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_unselect_child" XG_POST, gxg_gtk_tree_unselect_child, 2, 0, 0, H_gtk_tree_unselect_child);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_child_position" XG_POST, gxg_gtk_tree_child_position, 2, 0, 0, H_gtk_tree_child_position);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_set_selection_mode" XG_POST, gxg_gtk_tree_set_selection_mode, 2, 0, 0, H_gtk_tree_set_selection_mode);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_set_view_mode" XG_POST, gxg_gtk_tree_set_view_mode, 2, 0, 0, H_gtk_tree_set_view_mode);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_set_view_lines" XG_POST, gxg_gtk_tree_set_view_lines, 2, 0, 0, H_gtk_tree_set_view_lines);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_remove_item" XG_POST, gxg_gtk_tree_remove_item, 2, 0, 0, H_gtk_tree_remove_item);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_item_get_type" XG_POST, gxg_gtk_tree_item_get_type, 0, 0, 0, H_gtk_tree_item_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_item_new" XG_POST, gxg_gtk_tree_item_new, 0, 0, 0, H_gtk_tree_item_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_item_new_with_label" XG_POST, gxg_gtk_tree_item_new_with_label, 1, 0, 0, H_gtk_tree_item_new_with_label);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_item_set_subtree" XG_POST, gxg_gtk_tree_item_set_subtree, 2, 0, 0, H_gtk_tree_item_set_subtree);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_item_remove_subtree" XG_POST, gxg_gtk_tree_item_remove_subtree, 1, 0, 0, H_gtk_tree_item_remove_subtree);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_item_select" XG_POST, gxg_gtk_tree_item_select, 1, 0, 0, H_gtk_tree_item_select);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_item_deselect" XG_POST, gxg_gtk_tree_item_deselect, 1, 0, 0, H_gtk_tree_item_deselect);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_item_expand" XG_POST, gxg_gtk_tree_item_expand, 1, 0, 0, H_gtk_tree_item_expand);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_tree_item_collapse" XG_POST, gxg_gtk_tree_item_collapse, 1, 0, 0, H_gtk_tree_item_collapse);
  XEN_DEFINE_PROCEDURE(XG_PRE "gdk_image_new_bitmap" XG_POST, gxg_gdk_image_new_bitmap, 4, 0, 0, H_gdk_image_new_bitmap);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_get_type" XG_POST, gxg_gtk_text_get_type, 0, 0, 0, H_gtk_text_get_type);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_new" XG_POST, gxg_gtk_text_new, 2, 0, 0, H_gtk_text_new);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_set_editable" XG_POST, gxg_gtk_text_set_editable, 2, 0, 0, H_gtk_text_set_editable);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_set_word_wrap" XG_POST, gxg_gtk_text_set_word_wrap, 2, 0, 0, H_gtk_text_set_word_wrap);
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_text_set_line_wrap" XG_POST, gxg_gtk_text_set_line_wrap, 2, 0, 0, H_gtk_text_set_line_wrap);
#endif

  XEN_DEFINE_PROCEDURE(XG_PRE "GDK_COLORMAP" XG_POST, gxg_GDK_COLORMAP, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GDK_DRAG_CONTEXT" XG_POST, gxg_GDK_DRAG_CONTEXT, 1, 0, 0, NULL);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "GDK_DRAWABLE" XG_POST, gxg_GDK_DRAWABLE, 1, 0, 0, NULL);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "GDK_GC" XG_POST, gxg_GDK_GC, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GDK_IMAGE" XG_POST, gxg_GDK_IMAGE, 1, 0, 0, NULL);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "GDK_DEVICE" XG_POST, gxg_GDK_DEVICE, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GDK_KEYMAP" XG_POST, gxg_GDK_KEYMAP, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GDK_PIXMAP" XG_POST, gxg_GDK_PIXMAP, 1, 0, 0, NULL);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "GDK_VISUAL" XG_POST, gxg_GDK_VISUAL, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GDK_WINDOW" XG_POST, gxg_GDK_WINDOW, 1, 0, 0, NULL);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "GDK_WINDOW_OBJECT" XG_POST, gxg_GDK_WINDOW_OBJECT, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GDK_PIXBUF" XG_POST, gxg_GDK_PIXBUF, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GDK_PIXBUF_ANIMATION" XG_POST, gxg_GDK_PIXBUF_ANIMATION, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GDK_PIXBUF_ANIMATION_ITER" XG_POST, gxg_GDK_PIXBUF_ANIMATION_ITER, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GDK_PIXBUF_LOADER" XG_POST, gxg_GDK_PIXBUF_LOADER, 1, 0, 0, NULL);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_VBOX" XG_POST, gxg_GTK_VBOX, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_ACCEL_GROUP" XG_POST, gxg_GTK_ACCEL_GROUP, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_ACCEL_LABEL" XG_POST, gxg_GTK_ACCEL_LABEL, 1, 0, 0, NULL);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_ACCESSIBLE" XG_POST, gxg_GTK_ACCESSIBLE, 1, 0, 0, NULL);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_ADJUSTMENT" XG_POST, gxg_GTK_ADJUSTMENT, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_ALIGNMENT" XG_POST, gxg_GTK_ALIGNMENT, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_ARROW" XG_POST, gxg_GTK_ARROW, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_ASPECT_FRAME" XG_POST, gxg_GTK_ASPECT_FRAME, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_BUTTON_BOX" XG_POST, gxg_GTK_BUTTON_BOX, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_BIN" XG_POST, gxg_GTK_BIN, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_BOX" XG_POST, gxg_GTK_BOX, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_BUTTON" XG_POST, gxg_GTK_BUTTON, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_CALENDAR" XG_POST, gxg_GTK_CALENDAR, 1, 0, 0, NULL);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_CELL_EDITABLE" XG_POST, gxg_GTK_CELL_EDITABLE, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_CELL_RENDERER" XG_POST, gxg_GTK_CELL_RENDERER, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_CELL_RENDERER_PIXBUF" XG_POST, gxg_GTK_CELL_RENDERER_PIXBUF, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_CELL_RENDERER_TEXT" XG_POST, gxg_GTK_CELL_RENDERER_TEXT, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_CELL_RENDERER_TOGGLE" XG_POST, gxg_GTK_CELL_RENDERER_TOGGLE, 1, 0, 0, NULL);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_CHECK_BUTTON" XG_POST, gxg_GTK_CHECK_BUTTON, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_CHECK_MENU_ITEM" XG_POST, gxg_GTK_CHECK_MENU_ITEM, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_COLOR_SELECTION_DIALOG" XG_POST, gxg_GTK_COLOR_SELECTION_DIALOG, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_COLOR_SELECTION" XG_POST, gxg_GTK_COLOR_SELECTION, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_COMBO" XG_POST, gxg_GTK_COMBO, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_CONTAINER" XG_POST, gxg_GTK_CONTAINER, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_CURVE" XG_POST, gxg_GTK_CURVE, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_DIALOG" XG_POST, gxg_GTK_DIALOG, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_DRAWING_AREA" XG_POST, gxg_GTK_DRAWING_AREA, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_EDITABLE" XG_POST, gxg_GTK_EDITABLE, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_ENTRY" XG_POST, gxg_GTK_ENTRY, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_EVENT_BOX" XG_POST, gxg_GTK_EVENT_BOX, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_FILE_SELECTION" XG_POST, gxg_GTK_FILE_SELECTION, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_FIXED" XG_POST, gxg_GTK_FIXED, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_FONT_SELECTION" XG_POST, gxg_GTK_FONT_SELECTION, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_FONT_SELECTION_DIALOG" XG_POST, gxg_GTK_FONT_SELECTION_DIALOG, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_FRAME" XG_POST, gxg_GTK_FRAME, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_GAMMA_CURVE" XG_POST, gxg_GTK_GAMMA_CURVE, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_HANDLE_BOX" XG_POST, gxg_GTK_HANDLE_BOX, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_HBUTTON_BOX" XG_POST, gxg_GTK_HBUTTON_BOX, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_HBOX" XG_POST, gxg_GTK_HBOX, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_HPANED" XG_POST, gxg_GTK_HPANED, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_HRULER" XG_POST, gxg_GTK_HRULER, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_HSCALE" XG_POST, gxg_GTK_HSCALE, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_HSCROLLBAR" XG_POST, gxg_GTK_HSCROLLBAR, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_HSEPARATOR" XG_POST, gxg_GTK_HSEPARATOR, 1, 0, 0, NULL);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_ICON_FACTORY" XG_POST, gxg_GTK_ICON_FACTORY, 1, 0, 0, NULL);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IMAGE" XG_POST, gxg_GTK_IMAGE, 1, 0, 0, NULL);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IMAGE_MENU_ITEM" XG_POST, gxg_GTK_IMAGE_MENU_ITEM, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IM_CONTEXT" XG_POST, gxg_GTK_IM_CONTEXT, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IM_CONTEXT_SIMPLE" XG_POST, gxg_GTK_IM_CONTEXT_SIMPLE, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IM_MULTICONTEXT" XG_POST, gxg_GTK_IM_MULTICONTEXT, 1, 0, 0, NULL);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_INPUT_DIALOG" XG_POST, gxg_GTK_INPUT_DIALOG, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_INVISIBLE" XG_POST, gxg_GTK_INVISIBLE, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_ITEM_FACTORY" XG_POST, gxg_GTK_ITEM_FACTORY, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_ITEM" XG_POST, gxg_GTK_ITEM, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_LABEL" XG_POST, gxg_GTK_LABEL, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_LAYOUT" XG_POST, gxg_GTK_LAYOUT, 1, 0, 0, NULL);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_LIST_STORE" XG_POST, gxg_GTK_LIST_STORE, 1, 0, 0, NULL);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_MENU_BAR" XG_POST, gxg_GTK_MENU_BAR, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_MENU" XG_POST, gxg_GTK_MENU, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_MENU_ITEM" XG_POST, gxg_GTK_MENU_ITEM, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_MENU_SHELL" XG_POST, gxg_GTK_MENU_SHELL, 1, 0, 0, NULL);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_MESSAGE_DIALOG" XG_POST, gxg_GTK_MESSAGE_DIALOG, 1, 0, 0, NULL);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_MISC" XG_POST, gxg_GTK_MISC, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_NOTEBOOK" XG_POST, gxg_GTK_NOTEBOOK, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_OBJECT" XG_POST, gxg_GTK_OBJECT, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_OPTION_MENU" XG_POST, gxg_GTK_OPTION_MENU, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_PANED" XG_POST, gxg_GTK_PANED, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_PLUG" XG_POST, gxg_GTK_PLUG, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_PROGRESS_BAR" XG_POST, gxg_GTK_PROGRESS_BAR, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_RADIO_BUTTON" XG_POST, gxg_GTK_RADIO_BUTTON, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_RADIO_MENU_ITEM" XG_POST, gxg_GTK_RADIO_MENU_ITEM, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_RANGE" XG_POST, gxg_GTK_RANGE, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_RC_STYLE" XG_POST, gxg_GTK_RC_STYLE, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_RULER" XG_POST, gxg_GTK_RULER, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_SCALE" XG_POST, gxg_GTK_SCALE, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_SCROLLBAR" XG_POST, gxg_GTK_SCROLLBAR, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_SCROLLED_WINDOW" XG_POST, gxg_GTK_SCROLLED_WINDOW, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_SEPARATOR" XG_POST, gxg_GTK_SEPARATOR, 1, 0, 0, NULL);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_SEPARATOR_MENU_ITEM" XG_POST, gxg_GTK_SEPARATOR_MENU_ITEM, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_SETTINGS" XG_POST, gxg_GTK_SETTINGS, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_SIZE_GROUP" XG_POST, gxg_GTK_SIZE_GROUP, 1, 0, 0, NULL);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_SOCKET" XG_POST, gxg_GTK_SOCKET, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_SPIN_BUTTON" XG_POST, gxg_GTK_SPIN_BUTTON, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_STATUSBAR" XG_POST, gxg_GTK_STATUSBAR, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_STYLE" XG_POST, gxg_GTK_STYLE, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_TABLE" XG_POST, gxg_GTK_TABLE, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_TEAROFF_MENU_ITEM" XG_POST, gxg_GTK_TEAROFF_MENU_ITEM, 1, 0, 0, NULL);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_TEXT_BUFFER" XG_POST, gxg_GTK_TEXT_BUFFER, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_TEXT_CHILD_ANCHOR" XG_POST, gxg_GTK_TEXT_CHILD_ANCHOR, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_TEXT_MARK" XG_POST, gxg_GTK_TEXT_MARK, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_TEXT_TAG" XG_POST, gxg_GTK_TEXT_TAG, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_TEXT_TAG_TABLE" XG_POST, gxg_GTK_TEXT_TAG_TABLE, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_TEXT_VIEW" XG_POST, gxg_GTK_TEXT_VIEW, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_THEME_ENGINE" XG_POST, gxg_GTK_THEME_ENGINE, 1, 0, 0, NULL);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_TOGGLE_BUTTON" XG_POST, gxg_GTK_TOGGLE_BUTTON, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_TOOLBAR" XG_POST, gxg_GTK_TOOLBAR, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_TOOLTIPS" XG_POST, gxg_GTK_TOOLTIPS, 1, 0, 0, NULL);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_TREE_DRAG_SOURCE" XG_POST, gxg_GTK_TREE_DRAG_SOURCE, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_TREE_DRAG_DEST" XG_POST, gxg_GTK_TREE_DRAG_DEST, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_TREE_MODEL" XG_POST, gxg_GTK_TREE_MODEL, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_TREE_MODEL_SORT" XG_POST, gxg_GTK_TREE_MODEL_SORT, 1, 0, 0, NULL);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_TREE_SELECTION" XG_POST, gxg_GTK_TREE_SELECTION, 1, 0, 0, NULL);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_TREE_SORTABLE" XG_POST, gxg_GTK_TREE_SORTABLE, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_TREE_STORE" XG_POST, gxg_GTK_TREE_STORE, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_TREE_VIEW_COLUMN" XG_POST, gxg_GTK_TREE_VIEW_COLUMN, 1, 0, 0, NULL);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_TREE_VIEW" XG_POST, gxg_GTK_TREE_VIEW, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_VBUTTON_BOX" XG_POST, gxg_GTK_VBUTTON_BOX, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_VIEWPORT" XG_POST, gxg_GTK_VIEWPORT, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_VPANED" XG_POST, gxg_GTK_VPANED, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_VRULER" XG_POST, gxg_GTK_VRULER, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_VSCALE" XG_POST, gxg_GTK_VSCALE, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_VSCROLLBAR" XG_POST, gxg_GTK_VSCROLLBAR, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_VSEPARATOR" XG_POST, gxg_GTK_VSEPARATOR, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_WIDGET" XG_POST, gxg_GTK_WIDGET, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_WINDOW" XG_POST, gxg_GTK_WINDOW, 1, 0, 0, NULL);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_WINDOW_GROUP" XG_POST, gxg_GTK_WINDOW_GROUP, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "PANGO_CONTEXT" XG_POST, gxg_PANGO_CONTEXT, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "PANGO_FONT_FACE" XG_POST, gxg_PANGO_FONT_FACE, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "PANGO_FONT" XG_POST, gxg_PANGO_FONT, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "PANGO_FONT_MAP" XG_POST, gxg_PANGO_FONT_MAP, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "PANGO_LAYOUT" XG_POST, gxg_PANGO_LAYOUT, 1, 0, 0, NULL);
#endif
#if PANGO_ENABLE_ENGINE && PANGO_ENABLE_BACKEND
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "PANGO_FONT_FAMILY" XG_POST, gxg_PANGO_FONT_FAMILY, 1, 0, 0, NULL);
#endif
#endif

#if (!(defined(GDK_DISABLE_DEPRECATED))) && (!(defined(GTK_DISABLE_DEPRECATED))) && (!(defined(GDK_PIXBUF_DISABLE_DEPRECATED)))
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_CLIST" XG_POST, gxg_GTK_CLIST, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_CTREE" XG_POST, gxg_GTK_CTREE, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_LIST" XG_POST, gxg_GTK_LIST, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_LIST_ITEM" XG_POST, gxg_GTK_LIST_ITEM, 1, 0, 0, NULL);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_OLD_EDITABLE" XG_POST, gxg_GTK_OLD_EDITABLE, 1, 0, 0, NULL);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_PIXMAP" XG_POST, gxg_GTK_PIXMAP, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_PREVIEW" XG_POST, gxg_GTK_PREVIEW, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_PROGRESS" XG_POST, gxg_GTK_PROGRESS, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_TIPS_QUERY" XG_POST, gxg_GTK_TIPS_QUERY, 1, 0, 0, NULL);
#endif

#if GDK_ENABLE_BROKEN && GTK_ENABLE_BROKEN
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_TREE" XG_POST, gxg_GTK_TREE, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_TREE_ITEM" XG_POST, gxg_GTK_TREE_ITEM, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_TEXT" XG_POST, gxg_GTK_TEXT, 1, 0, 0, NULL);
#endif

  XEN_DEFINE_PROCEDURE(XG_PRE "GDK_IS_COLORMAP" XG_POST, XEN_GdkColormap__p, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GDK_IS_DRAG_CONTEXT" XG_POST, XEN_GdkDragContext__p, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GDK_IS_DRAWABLE" XG_POST, XEN_GdkDrawable__p, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GDK_IS_GC" XG_POST, XEN_GdkGC__p, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GDK_IS_IMAGE" XG_POST, XEN_GdkImage__p, 1, 0, 0, NULL);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "GDK_IS_DEVICE" XG_POST, XEN_GdkDevice__p, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GDK_IS_KEYMAP" XG_POST, XEN_GdkKeymap__p, 1, 0, 0, NULL);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "GDK_IS_PIXMAP" XG_POST, XEN_GdkPixmap__p, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GDK_IS_VISUAL" XG_POST, XEN_GdkVisual__p, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GDK_IS_WINDOW" XG_POST, XEN_GdkWindow__p, 1, 0, 0, NULL);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "GDK_IS_PIXBUF" XG_POST, XEN_GdkPixbuf__p, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GDK_IS_PIXBUF_ANIMATION" XG_POST, XEN_GdkPixbufAnimation__p, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GDK_IS_PIXBUF_ANIMATION_ITER" XG_POST, XEN_GdkPixbufAnimationIter__p, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GDK_IS_PIXBUF_LOADER" XG_POST, XEN_GdkPixbufLoader__p, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_VBOX" XG_POST, XEN_GtkVBox__p, 1, 0, 0, NULL);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_ACCEL_GROUP" XG_POST, XEN_GtkAccelGroup__p, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_ACCEL_LABEL" XG_POST, XEN_GtkAccelLabel__p, 1, 0, 0, NULL);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_ACCESSIBLE" XG_POST, XEN_GtkAccessible__p, 1, 0, 0, NULL);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_ADJUSTMENT" XG_POST, XEN_GtkAdjustment__p, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_ALIGNMENT" XG_POST, XEN_GtkAlignment__p, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_ARROW" XG_POST, XEN_GtkArrow__p, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_ASPECT_FRAME" XG_POST, XEN_GtkAspectFrame__p, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_BUTTON_BOX" XG_POST, XEN_GtkButtonBox__p, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_BIN" XG_POST, XEN_GtkBin__p, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_BOX" XG_POST, XEN_GtkBox__p, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_BUTTON" XG_POST, XEN_GtkButton__p, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_CALENDAR" XG_POST, XEN_GtkCalendar__p, 1, 0, 0, NULL);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_CELL_EDITABLE" XG_POST, XEN_GtkCellEditable__p, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_CELL_RENDERER" XG_POST, XEN_GtkCellRenderer__p, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_CELL_RENDERER_PIXBUF" XG_POST, XEN_GtkCellRendererPixbuf__p, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_CELL_RENDERER_TEXT" XG_POST, XEN_GtkCellRendererText__p, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_CELL_RENDERER_TOGGLE" XG_POST, XEN_GtkCellRendererToggle__p, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_CHECK_BUTTON" XG_POST, XEN_GtkCheckButton__p, 1, 0, 0, NULL);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_CHECK_MENU_ITEM" XG_POST, XEN_GtkCheckMenuItem__p, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_COLOR_SELECTION_DIALOG" XG_POST, XEN_GtkColorSelectionDialog__p, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_COLOR_SELECTION" XG_POST, XEN_GtkColorSelection__p, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_COMBO" XG_POST, XEN_GtkCombo__p, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_CONTAINER" XG_POST, XEN_GtkContainer__p, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_CURVE" XG_POST, XEN_GtkCurve__p, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_DIALOG" XG_POST, XEN_GtkDialog__p, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_DRAWING_AREA" XG_POST, XEN_GtkDrawingArea__p, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_EDITABLE" XG_POST, XEN_GtkEditable__p, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_ENTRY" XG_POST, XEN_GtkEntry__p, 1, 0, 0, NULL);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_EVENT_BOX" XG_POST, XEN_GtkEventBox__p, 1, 0, 0, NULL);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_FILE_SELECTION" XG_POST, XEN_GtkFileSelection__p, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_FIXED" XG_POST, XEN_GtkFixed__p, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_FONT_SELECTION" XG_POST, XEN_GtkFontSelection__p, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_FONT_SELECTION_DIALOG" XG_POST, XEN_GtkFontSelectionDialog__p, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_FRAME" XG_POST, XEN_GtkFrame__p, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_GAMMA_CURVE" XG_POST, XEN_GtkGammaCurve__p, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_HANDLE_BOX" XG_POST, XEN_GtkHandleBox__p, 1, 0, 0, NULL);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_HBUTTON_BOX" XG_POST, XEN_GtkHButtonBox__p, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_HBOX" XG_POST, XEN_GtkHBox__p, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_HPANED" XG_POST, XEN_GtkHPaned__p, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_HRULER" XG_POST, XEN_GtkHRuler__p, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_HSCALE" XG_POST, XEN_GtkHScale__p, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_HSCROLLBAR" XG_POST, XEN_GtkHScrollbar__p, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_HSEPARATOR" XG_POST, XEN_GtkHSeparator__p, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_ICON_FACTORY" XG_POST, XEN_GtkIconFactory__p, 1, 0, 0, NULL);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_IMAGE" XG_POST, XEN_GtkImage__p, 1, 0, 0, NULL);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_IMAGE_MENU_ITEM" XG_POST, XEN_GtkImageMenuItem__p, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_IM_CONTEXT" XG_POST, XEN_GtkIMContext__p, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_IM_CONTEXT_SIMPLE" XG_POST, XEN_GtkIMContextSimple__p, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_IM_MULTICONTEXT" XG_POST, XEN_GtkIMMulticontext__p, 1, 0, 0, NULL);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_INPUT_DIALOG" XG_POST, XEN_GtkInputDialog__p, 1, 0, 0, NULL);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_INVISIBLE" XG_POST, XEN_GtkInvisible__p, 1, 0, 0, NULL);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_ITEM_FACTORY" XG_POST, XEN_GtkItemFactory__p, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_ITEM" XG_POST, XEN_GtkItem__p, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_LABEL" XG_POST, XEN_GtkLabel__p, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_LAYOUT" XG_POST, XEN_GtkLayout__p, 1, 0, 0, NULL);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_LIST_STORE" XG_POST, XEN_GtkListStore__p, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_MENU_BAR" XG_POST, XEN_GtkMenuBar__p, 1, 0, 0, NULL);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_MENU" XG_POST, XEN_GtkMenu__p, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_MENU_ITEM" XG_POST, XEN_GtkMenuItem__p, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_MENU_SHELL" XG_POST, XEN_GtkMenuShell__p, 1, 0, 0, NULL);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_MESSAGE_DIALOG" XG_POST, XEN_GtkMessageDialog__p, 1, 0, 0, NULL);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_MISC" XG_POST, XEN_GtkMisc__p, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_NOTEBOOK" XG_POST, XEN_GtkNotebook__p, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_OBJECT" XG_POST, XEN_GtkObject__p, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_OPTION_MENU" XG_POST, XEN_GtkOptionMenu__p, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_PANED" XG_POST, XEN_GtkPaned__p, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_PLUG" XG_POST, XEN_GtkPlug__p, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_PROGRESS_BAR" XG_POST, XEN_GtkProgressBar__p, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_RADIO_BUTTON" XG_POST, XEN_GtkRadioButton__p, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_RADIO_MENU_ITEM" XG_POST, XEN_GtkRadioMenuItem__p, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_RANGE" XG_POST, XEN_GtkRange__p, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_RC_STYLE" XG_POST, XEN_GtkRcStyle__p, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_RULER" XG_POST, XEN_GtkRuler__p, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_SCALE" XG_POST, XEN_GtkScale__p, 1, 0, 0, NULL);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_SCROLLBAR" XG_POST, XEN_GtkScrollbar__p, 1, 0, 0, NULL);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_SCROLLED_WINDOW" XG_POST, XEN_GtkScrolledWindow__p, 1, 0, 0, NULL);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_SEPARATOR" XG_POST, XEN_GtkSeparator__p, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_SEPARATOR_MENU_ITEM" XG_POST, XEN_GtkSeparatorMenuItem__p, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_SETTINGS" XG_POST, XEN_GtkSettings__p, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_SIZE_GROUP" XG_POST, XEN_GtkSizeGroup__p, 1, 0, 0, NULL);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_SOCKET" XG_POST, XEN_GtkSocket__p, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_SPIN_BUTTON" XG_POST, XEN_GtkSpinButton__p, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_STATUSBAR" XG_POST, XEN_GtkStatusbar__p, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_STYLE" XG_POST, XEN_GtkStyle__p, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_TABLE" XG_POST, XEN_GtkTable__p, 1, 0, 0, NULL);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_TEAROFF_MENU_ITEM" XG_POST, XEN_GtkTearoffMenuItem__p, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_TEXT_BUFFER" XG_POST, XEN_GtkTextBuffer__p, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_TEXT_CHILD_ANCHOR" XG_POST, XEN_GtkTextChildAnchor__p, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_TEXT_MARK" XG_POST, XEN_GtkTextMark__p, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_TEXT_TAG" XG_POST, XEN_GtkTextTag__p, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_TEXT_TAG_TABLE" XG_POST, XEN_GtkTextTagTable__p, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_TEXT_VIEW" XG_POST, XEN_GtkTextView__p, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_THEME_ENGINE" XG_POST, XEN_GtkThemeEngine__p, 1, 0, 0, NULL);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_TOGGLE_BUTTON" XG_POST, XEN_GtkToggleButton__p, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_TOOLBAR" XG_POST, XEN_GtkToolbar__p, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_TOOLTIPS" XG_POST, XEN_GtkTooltips__p, 1, 0, 0, NULL);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_TREE_DRAG_SOURCE" XG_POST, XEN_GtkTreeDragSource__p, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_TREE_DRAG_DEST" XG_POST, XEN_GtkTreeDragDest__p, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_TREE_MODEL" XG_POST, XEN_GtkTreeModel__p, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_TREE_MODEL_SORT" XG_POST, XEN_GtkTreeModelSort__p, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_TREE_SELECTION" XG_POST, XEN_GtkTreeSelection__p, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_TREE_SORTABLE" XG_POST, XEN_GtkTreeSortable__p, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_TREE_STORE" XG_POST, XEN_GtkTreeStore__p, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_TREE_VIEW_COLUMN" XG_POST, XEN_GtkTreeViewColumn__p, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_TREE_VIEW" XG_POST, XEN_GtkTreeView__p, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_VBUTTON_BOX" XG_POST, XEN_GtkVButtonBox__p, 1, 0, 0, NULL);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_VIEWPORT" XG_POST, XEN_GtkViewport__p, 1, 0, 0, NULL);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_VPANED" XG_POST, XEN_GtkVPaned__p, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_VRULER" XG_POST, XEN_GtkVRuler__p, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_VSCALE" XG_POST, XEN_GtkVScale__p, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_VSCROLLBAR" XG_POST, XEN_GtkVScrollbar__p, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_VSEPARATOR" XG_POST, XEN_GtkVSeparator__p, 1, 0, 0, NULL);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_WIDGET" XG_POST, XEN_GtkWidget__p, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_WINDOW" XG_POST, XEN_GtkWindow__p, 1, 0, 0, NULL);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_WINDOW_GROUP" XG_POST, XEN_GtkWindowGroup__p, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "PANGO_IS_CONTEXT" XG_POST, XEN_PangoContext__p, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "PANGO_IS_FONT_FACE" XG_POST, XEN_PangoFontFace__p, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "PANGO_IS_FONT" XG_POST, XEN_PangoFont__p, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "PANGO_IS_FONT_MAP" XG_POST, XEN_PangoFontMap__p, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "PANGO_IS_LAYOUT" XG_POST, XEN_PangoLayout__p, 1, 0, 0, NULL);
#endif
#if PANGO_ENABLE_ENGINE && PANGO_ENABLE_BACKEND
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "PANGO_IS_FONT_FAMILY" XG_POST, XEN_PangoFontFamily__p, 1, 0, 0, NULL);
#endif
#endif

#if (!(defined(GDK_DISABLE_DEPRECATED))) && (!(defined(GTK_DISABLE_DEPRECATED))) && (!(defined(GDK_PIXBUF_DISABLE_DEPRECATED)))
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_CLIST" XG_POST, XEN_GtkCList__p, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_CTREE" XG_POST, XEN_GtkCTree__p, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_LIST" XG_POST, XEN_GtkList__p, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_LIST_ITEM" XG_POST, XEN_GtkListItem__p, 1, 0, 0, NULL);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_OLD_EDITABLE" XG_POST, XEN_GtkOldEditable__p, 1, 0, 0, NULL);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_PIXMAP" XG_POST, XEN_GtkPixmap__p, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_PREVIEW" XG_POST, XEN_GtkPreview__p, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_PROGRESS" XG_POST, XEN_GtkProgress__p, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_TIPS_QUERY" XG_POST, XEN_GtkTipsQuery__p, 1, 0, 0, NULL);
#endif

#if GDK_ENABLE_BROKEN && GTK_ENABLE_BROKEN
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_TREE" XG_POST, XEN_GtkTree__p, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_TREE_ITEM" XG_POST, XEN_GtkTreeItem__p, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GTK_IS_TEXT" XG_POST, XEN_GtkText__p, 1, 0, 0, NULL);
#endif

}

#endif
/* ---------------------------------------- structs ---------------------------------------- */

/* conversions */
static XEN c_array_to_xen_list(XEN val, XEN clen)
{
  XEN result = XEN_EMPTY_LIST;
  int i, len;
  char *ctype;
  len = XEN_TO_C_INT(clen);
  ctype = XEN_SYMBOL_TO_C_STRING(XEN_CAR(val));
  if (strcmp(ctype, "gint*") == 0)
    {
      gint* arr; arr = (gint*)XEN_CADR(val); 
      for (i = len - 1; i >= 0; i--) result = XEN_CONS(C_TO_XEN_gint(arr[i]), result);
    }
  if (strcmp(ctype, "gboolean*") == 0)
    {
      gboolean* arr; arr = (gboolean*)XEN_CADR(val); 
      for (i = len - 1; i >= 0; i--) result = XEN_CONS(C_TO_XEN_gboolean(arr[i]), result);
    }
#if (!HAVE_GTK_1)
  if (strcmp(ctype, "GType*") == 0)
    {
      GType* arr; arr = (GType*)XEN_CADR(val); 
      for (i = len - 1; i >= 0; i--) result = XEN_CONS(C_TO_XEN_GType(arr[i]), result);
    }
#endif
  if (strcmp(ctype, "guint*") == 0)
    {
      guint* arr; arr = (guint*)XEN_CADR(val); 
      for (i = len - 1; i >= 0; i--) result = XEN_CONS(C_TO_XEN_guint(arr[i]), result);
    }
  if (strcmp(ctype, "GQuark*") == 0)
    {
      GQuark* arr; arr = (GQuark*)XEN_CADR(val); 
      for (i = len - 1; i >= 0; i--) result = XEN_CONS(C_TO_XEN_GQuark(arr[i]), result);
    }
  if (strcmp(ctype, "GdkDragProtocol*") == 0)
    {
      GdkDragProtocol* arr; arr = (GdkDragProtocol*)XEN_CADR(val); 
      for (i = len - 1; i >= 0; i--) result = XEN_CONS(C_TO_XEN_GdkDragProtocol(arr[i]), result);
    }
  if (strcmp(ctype, "GdkModifierType*") == 0)
    {
      GdkModifierType* arr; arr = (GdkModifierType*)XEN_CADR(val); 
      for (i = len - 1; i >= 0; i--) result = XEN_CONS(C_TO_XEN_GdkModifierType(arr[i]), result);
    }
  if (strcmp(ctype, "gdouble*") == 0)
    {
      gdouble* arr; arr = (gdouble*)XEN_CADR(val); 
      for (i = len - 1; i >= 0; i--) result = XEN_CONS(C_TO_XEN_gdouble(arr[i]), result);
    }
  if (strcmp(ctype, "gint8*") == 0)
    {
      gint8* arr; arr = (gint8*)XEN_CADR(val); 
      for (i = len - 1; i >= 0; i--) result = XEN_CONS(C_TO_XEN_gint8(arr[i]), result);
    }
  if (strcmp(ctype, "gchar**") == 0)
    {
      gchar** arr; arr = (gchar**)XEN_CADR(val); 
      for (i = len - 1; i >= 0; i--) result = XEN_CONS(C_TO_XEN_gchar_(arr[i]), result);
    }
  if (strcmp(ctype, "guchar*") == 0)
    {
      guchar* arr; arr = (guchar*)XEN_CADR(val); 
      for (i = len - 1; i >= 0; i--) result = XEN_CONS(C_TO_XEN_guchar(arr[i]), result);
    }
  if (strcmp(ctype, "guint32*") == 0)
    {
      guint32* arr; arr = (guint32*)XEN_CADR(val); 
      for (i = len - 1; i >= 0; i--) result = XEN_CONS(C_TO_XEN_guint32(arr[i]), result);
    }
  if (strcmp(ctype, "GdkVisualType*") == 0)
    {
      GdkVisualType* arr; arr = (GdkVisualType*)XEN_CADR(val); 
      for (i = len - 1; i >= 0; i--) result = XEN_CONS(C_TO_XEN_GdkVisualType(arr[i]), result);
    }
  if (strcmp(ctype, "GdkWMDecoration*") == 0)
    {
      GdkWMDecoration* arr; arr = (GdkWMDecoration*)XEN_CADR(val); 
      for (i = len - 1; i >= 0; i--) result = XEN_CONS(C_TO_XEN_GdkWMDecoration(arr[i]), result);
    }
  if (strcmp(ctype, "char**") == 0)
    {
      char** arr; arr = (char**)XEN_CADR(val); 
      for (i = len - 1; i >= 0; i--) result = XEN_CONS(C_TO_XEN_char_(arr[i]), result);
    }
  if (strcmp(ctype, "guint8*") == 0)
    {
      guint8* arr; arr = (guint8*)XEN_CADR(val); 
      for (i = len - 1; i >= 0; i--) result = XEN_CONS(C_TO_XEN_guint8(arr[i]), result);
    }
  if (strcmp(ctype, "GtkPackType*") == 0)
    {
      GtkPackType* arr; arr = (GtkPackType*)XEN_CADR(val); 
      for (i = len - 1; i >= 0; i--) result = XEN_CONS(C_TO_XEN_GtkPackType(arr[i]), result);
    }
  if (strcmp(ctype, "gfloat*") == 0)
    {
      gfloat* arr; arr = (gfloat*)XEN_CADR(val); 
      for (i = len - 1; i >= 0; i--) result = XEN_CONS(C_TO_XEN_gfloat(arr[i]), result);
    }
#if (!HAVE_GTK_1)
  if (strcmp(ctype, "GtkIconSize*") == 0)
    {
      GtkIconSize* arr; arr = (GtkIconSize*)XEN_CADR(val); 
      for (i = len - 1; i >= 0; i--) result = XEN_CONS(C_TO_XEN_GtkIconSize(arr[i]), result);
    }
#endif
  if (strcmp(ctype, "guint16*") == 0)
    {
      guint16* arr; arr = (guint16*)XEN_CADR(val); 
      for (i = len - 1; i >= 0; i--) result = XEN_CONS(C_TO_XEN_guint16(arr[i]), result);
    }
  if (strcmp(ctype, "int*") == 0)
    {
      int* arr; arr = (int*)XEN_CADR(val); 
      for (i = len - 1; i >= 0; i--) result = XEN_CONS(C_TO_XEN_int(arr[i]), result);
    }
  if (strcmp(ctype, "GtkStateType*") == 0)
    {
      GtkStateType* arr; arr = (GtkStateType*)XEN_CADR(val); 
      for (i = len - 1; i >= 0; i--) result = XEN_CONS(C_TO_XEN_GtkStateType(arr[i]), result);
    }
  if (strcmp(ctype, "GtkPathPriorityType*") == 0)
    {
      GtkPathPriorityType* arr; arr = (GtkPathPriorityType*)XEN_CADR(val); 
      for (i = len - 1; i >= 0; i--) result = XEN_CONS(C_TO_XEN_GtkPathPriorityType(arr[i]), result);
    }
  if (strcmp(ctype, "GtkSortType*") == 0)
    {
      GtkSortType* arr; arr = (GtkSortType*)XEN_CADR(val); 
      for (i = len - 1; i >= 0; i--) result = XEN_CONS(C_TO_XEN_GtkSortType(arr[i]), result);
    }
#if (!HAVE_GTK_1)
  if (strcmp(ctype, "gunichar*") == 0)
    {
      gunichar* arr; arr = (gunichar*)XEN_CADR(val); 
      for (i = len - 1; i >= 0; i--) result = XEN_CONS(C_TO_XEN_gunichar(arr[i]), result);
    }
#endif
  if (strcmp(ctype, "GdkPixmap**") == 0)
    {
      GdkPixmap** arr; arr = (GdkPixmap**)XEN_CADR(val); 
      for (i = len - 1; i >= 0; i--) result = XEN_CONS(C_TO_XEN_GdkPixmap_(arr[i]), result);
    }
  if (strcmp(ctype, "GdkGC**") == 0)
    {
      GdkGC** arr; arr = (GdkGC**)XEN_CADR(val); 
      for (i = len - 1; i >= 0; i--) result = XEN_CONS(C_TO_XEN_GdkGC_(arr[i]), result);
    }
  if (strcmp(ctype, "GtkWidget**") == 0)
    {
      GtkWidget** arr; arr = (GtkWidget**)XEN_CADR(val); 
      for (i = len - 1; i >= 0; i--) result = XEN_CONS(C_TO_XEN_GtkWidget_(arr[i]), result);
    }
#if (!HAVE_GTK_1)
  if (strcmp(ctype, "PangoFontDescription**") == 0)
    {
      PangoFontDescription** arr; arr = (PangoFontDescription**)XEN_CADR(val); 
      for (i = len - 1; i >= 0; i--) result = XEN_CONS(C_TO_XEN_PangoFontDescription_(arr[i]), result);
    }
  if (strcmp(ctype, "PangoAttrList**") == 0)
    {
      PangoAttrList** arr; arr = (PangoAttrList**)XEN_CADR(val); 
      for (i = len - 1; i >= 0; i--) result = XEN_CONS(C_TO_XEN_PangoAttrList_(arr[i]), result);
    }
  if (strcmp(ctype, "GtkTreeModel**") == 0)
    {
      GtkTreeModel** arr; arr = (GtkTreeModel**)XEN_CADR(val); 
      for (i = len - 1; i >= 0; i--) result = XEN_CONS(C_TO_XEN_GtkTreeModel_(arr[i]), result);
    }
#endif
  if (strcmp(ctype, "char***") == 0)
    {
      char*** arr; arr = (char***)XEN_CADR(val); 
      for (i = len - 1; i >= 0; i--) result = XEN_CONS(C_TO_XEN_char__(arr[i]), result);
    }
  if (strcmp(ctype, "char**") == 0)
    {
      char** arr; arr = (char**)XEN_CADR(val); 
      for (i = len - 1; i >= 0; i--) result = XEN_CONS(C_TO_XEN_char_(arr[i]), result);
    }
  if (strcmp(ctype, "gchar**") == 0)
    {
      gchar** arr; arr = (gchar**)XEN_CADR(val); 
      for (i = len - 1; i >= 0; i--) result = XEN_CONS(C_TO_XEN_gchar_(arr[i]), result);
    }
  if (strcmp(ctype, "GdkBitmap**") == 0)
    {
      GdkBitmap** arr; arr = (GdkBitmap**)XEN_CADR(val); 
      for (i = len - 1; i >= 0; i--) result = XEN_CONS(C_TO_XEN_GdkBitmap_(arr[i]), result);
    }
  if (strcmp(ctype, "gchar***") == 0)
    {
      gchar*** arr; arr = (gchar***)XEN_CADR(val); 
      for (i = len - 1; i >= 0; i--) result = XEN_CONS(C_TO_XEN_gchar__(arr[i]), result);
    }
  if (strcmp(ctype, "GList*") == 0)
    { /* tagging these pointers is currently up to the caller */
      GList* lst;
      lst = (GList*)XEN_CADR(val);
      len = g_list_length(lst);
      for (i = len - 1; i >= 0; i--) result = XEN_CONS(C_TO_XEN_ULONG(g_list_nth_data(lst, i)), result);
    }
  return(result);
}

static XEN xen_list_to_c_array(XEN val, XEN type)
{
  int i, len;
  char *ctype;
  len = XEN_LIST_LENGTH(val);
  ctype = XEN_TO_C_STRING(type);
  if (strcmp(ctype, "gint*") == 0)
    {
      gint* arr; arr = (gint*)CALLOC(len, sizeof(gint));
      for (i = 0; i < len; i++, val = XEN_CDR(val)) arr[i] = XEN_TO_C_gint(XEN_CAR(val));
      return(XEN_LIST_3(C_STRING_TO_XEN_SYMBOL("gint_"), C_TO_XEN_ULONG((unsigned long)arr), make_xm_obj(arr)));
    }
  if (strcmp(ctype, "gboolean*") == 0)
    {
      gboolean* arr; arr = (gboolean*)CALLOC(len, sizeof(gboolean));
      for (i = 0; i < len; i++, val = XEN_CDR(val)) arr[i] = XEN_TO_C_gboolean(XEN_CAR(val));
      return(XEN_LIST_3(C_STRING_TO_XEN_SYMBOL("gboolean_"), C_TO_XEN_ULONG((unsigned long)arr), make_xm_obj(arr)));
    }
#if (!HAVE_GTK_1)
  if (strcmp(ctype, "GType*") == 0)
    {
      GType* arr; arr = (GType*)CALLOC(len, sizeof(GType));
      for (i = 0; i < len; i++, val = XEN_CDR(val)) arr[i] = XEN_TO_C_GType(XEN_CAR(val));
      return(XEN_LIST_3(C_STRING_TO_XEN_SYMBOL("GType_"), C_TO_XEN_ULONG((unsigned long)arr), make_xm_obj(arr)));
    }
#endif
  if (strcmp(ctype, "guint*") == 0)
    {
      guint* arr; arr = (guint*)CALLOC(len, sizeof(guint));
      for (i = 0; i < len; i++, val = XEN_CDR(val)) arr[i] = XEN_TO_C_guint(XEN_CAR(val));
      return(XEN_LIST_3(C_STRING_TO_XEN_SYMBOL("guint_"), C_TO_XEN_ULONG((unsigned long)arr), make_xm_obj(arr)));
    }
  if (strcmp(ctype, "GQuark*") == 0)
    {
      GQuark* arr; arr = (GQuark*)CALLOC(len, sizeof(GQuark));
      for (i = 0; i < len; i++, val = XEN_CDR(val)) arr[i] = XEN_TO_C_GQuark(XEN_CAR(val));
      return(XEN_LIST_3(C_STRING_TO_XEN_SYMBOL("GQuark_"), C_TO_XEN_ULONG((unsigned long)arr), make_xm_obj(arr)));
    }
  if (strcmp(ctype, "GdkDragProtocol*") == 0)
    {
      GdkDragProtocol* arr; arr = (GdkDragProtocol*)CALLOC(len, sizeof(GdkDragProtocol));
      for (i = 0; i < len; i++, val = XEN_CDR(val)) arr[i] = XEN_TO_C_GdkDragProtocol(XEN_CAR(val));
      return(XEN_LIST_3(C_STRING_TO_XEN_SYMBOL("GdkDragProtocol_"), C_TO_XEN_ULONG((unsigned long)arr), make_xm_obj(arr)));
    }
  if (strcmp(ctype, "GdkModifierType*") == 0)
    {
      GdkModifierType* arr; arr = (GdkModifierType*)CALLOC(len, sizeof(GdkModifierType));
      for (i = 0; i < len; i++, val = XEN_CDR(val)) arr[i] = XEN_TO_C_GdkModifierType(XEN_CAR(val));
      return(XEN_LIST_3(C_STRING_TO_XEN_SYMBOL("GdkModifierType_"), C_TO_XEN_ULONG((unsigned long)arr), make_xm_obj(arr)));
    }
  if (strcmp(ctype, "gdouble*") == 0)
    {
      gdouble* arr; arr = (gdouble*)CALLOC(len, sizeof(gdouble));
      for (i = 0; i < len; i++, val = XEN_CDR(val)) arr[i] = XEN_TO_C_gdouble(XEN_CAR(val));
      return(XEN_LIST_3(C_STRING_TO_XEN_SYMBOL("gdouble_"), C_TO_XEN_ULONG((unsigned long)arr), make_xm_obj(arr)));
    }
  if (strcmp(ctype, "gint8*") == 0)
    {
      gint8* arr; arr = (gint8*)CALLOC(len, sizeof(gint8));
      for (i = 0; i < len; i++, val = XEN_CDR(val)) arr[i] = XEN_TO_C_gint8(XEN_CAR(val));
      return(XEN_LIST_3(C_STRING_TO_XEN_SYMBOL("gint8_"), C_TO_XEN_ULONG((unsigned long)arr), make_xm_obj(arr)));
    }
  if (strcmp(ctype, "gchar**") == 0)
    {
      gchar** arr; arr = (gchar**)CALLOC(len, sizeof(gchar*));
      for (i = 0; i < len; i++, val = XEN_CDR(val)) arr[i] = XEN_TO_C_gchar_(XEN_CAR(val));
      return(XEN_LIST_3(C_STRING_TO_XEN_SYMBOL("gchar__"), C_TO_XEN_ULONG((unsigned long)arr), make_xm_obj(arr)));
    }
  if (strcmp(ctype, "guchar*") == 0)
    {
      guchar* arr; arr = (guchar*)CALLOC(len, sizeof(guchar));
      for (i = 0; i < len; i++, val = XEN_CDR(val)) arr[i] = XEN_TO_C_guchar(XEN_CAR(val));
      return(XEN_LIST_3(C_STRING_TO_XEN_SYMBOL("guchar_"), C_TO_XEN_ULONG((unsigned long)arr), make_xm_obj(arr)));
    }
  if (strcmp(ctype, "guint32*") == 0)
    {
      guint32* arr; arr = (guint32*)CALLOC(len, sizeof(guint32));
      for (i = 0; i < len; i++, val = XEN_CDR(val)) arr[i] = XEN_TO_C_guint32(XEN_CAR(val));
      return(XEN_LIST_3(C_STRING_TO_XEN_SYMBOL("guint32_"), C_TO_XEN_ULONG((unsigned long)arr), make_xm_obj(arr)));
    }
  if (strcmp(ctype, "GdkVisualType*") == 0)
    {
      GdkVisualType* arr; arr = (GdkVisualType*)CALLOC(len, sizeof(GdkVisualType));
      for (i = 0; i < len; i++, val = XEN_CDR(val)) arr[i] = XEN_TO_C_GdkVisualType(XEN_CAR(val));
      return(XEN_LIST_3(C_STRING_TO_XEN_SYMBOL("GdkVisualType_"), C_TO_XEN_ULONG((unsigned long)arr), make_xm_obj(arr)));
    }
  if (strcmp(ctype, "GdkWMDecoration*") == 0)
    {
      GdkWMDecoration* arr; arr = (GdkWMDecoration*)CALLOC(len, sizeof(GdkWMDecoration));
      for (i = 0; i < len; i++, val = XEN_CDR(val)) arr[i] = XEN_TO_C_GdkWMDecoration(XEN_CAR(val));
      return(XEN_LIST_3(C_STRING_TO_XEN_SYMBOL("GdkWMDecoration_"), C_TO_XEN_ULONG((unsigned long)arr), make_xm_obj(arr)));
    }
  if (strcmp(ctype, "char**") == 0)
    {
      char** arr; arr = (char**)CALLOC(len, sizeof(char*));
      for (i = 0; i < len; i++, val = XEN_CDR(val)) arr[i] = XEN_TO_C_char_(XEN_CAR(val));
      return(XEN_LIST_3(C_STRING_TO_XEN_SYMBOL("char__"), C_TO_XEN_ULONG((unsigned long)arr), make_xm_obj(arr)));
    }
  if (strcmp(ctype, "guint8*") == 0)
    {
      guint8* arr; arr = (guint8*)CALLOC(len, sizeof(guint8));
      for (i = 0; i < len; i++, val = XEN_CDR(val)) arr[i] = XEN_TO_C_guint8(XEN_CAR(val));
      return(XEN_LIST_3(C_STRING_TO_XEN_SYMBOL("guint8_"), C_TO_XEN_ULONG((unsigned long)arr), make_xm_obj(arr)));
    }
  if (strcmp(ctype, "GtkPackType*") == 0)
    {
      GtkPackType* arr; arr = (GtkPackType*)CALLOC(len, sizeof(GtkPackType));
      for (i = 0; i < len; i++, val = XEN_CDR(val)) arr[i] = XEN_TO_C_GtkPackType(XEN_CAR(val));
      return(XEN_LIST_3(C_STRING_TO_XEN_SYMBOL("GtkPackType_"), C_TO_XEN_ULONG((unsigned long)arr), make_xm_obj(arr)));
    }
  if (strcmp(ctype, "gfloat*") == 0)
    {
      gfloat* arr; arr = (gfloat*)CALLOC(len, sizeof(gfloat));
      for (i = 0; i < len; i++, val = XEN_CDR(val)) arr[i] = XEN_TO_C_gfloat(XEN_CAR(val));
      return(XEN_LIST_3(C_STRING_TO_XEN_SYMBOL("gfloat_"), C_TO_XEN_ULONG((unsigned long)arr), make_xm_obj(arr)));
    }
#if (!HAVE_GTK_1)
  if (strcmp(ctype, "GtkIconSize*") == 0)
    {
      GtkIconSize* arr; arr = (GtkIconSize*)CALLOC(len, sizeof(GtkIconSize));
      for (i = 0; i < len; i++, val = XEN_CDR(val)) arr[i] = XEN_TO_C_GtkIconSize(XEN_CAR(val));
      return(XEN_LIST_3(C_STRING_TO_XEN_SYMBOL("GtkIconSize_"), C_TO_XEN_ULONG((unsigned long)arr), make_xm_obj(arr)));
    }
#endif
  if (strcmp(ctype, "guint16*") == 0)
    {
      guint16* arr; arr = (guint16*)CALLOC(len, sizeof(guint16));
      for (i = 0; i < len; i++, val = XEN_CDR(val)) arr[i] = XEN_TO_C_guint16(XEN_CAR(val));
      return(XEN_LIST_3(C_STRING_TO_XEN_SYMBOL("guint16_"), C_TO_XEN_ULONG((unsigned long)arr), make_xm_obj(arr)));
    }
  if (strcmp(ctype, "int*") == 0)
    {
      int* arr; arr = (int*)CALLOC(len, sizeof(int));
      for (i = 0; i < len; i++, val = XEN_CDR(val)) arr[i] = XEN_TO_C_int(XEN_CAR(val));
      return(XEN_LIST_3(C_STRING_TO_XEN_SYMBOL("int_"), C_TO_XEN_ULONG((unsigned long)arr), make_xm_obj(arr)));
    }
  if (strcmp(ctype, "GtkStateType*") == 0)
    {
      GtkStateType* arr; arr = (GtkStateType*)CALLOC(len, sizeof(GtkStateType));
      for (i = 0; i < len; i++, val = XEN_CDR(val)) arr[i] = XEN_TO_C_GtkStateType(XEN_CAR(val));
      return(XEN_LIST_3(C_STRING_TO_XEN_SYMBOL("GtkStateType_"), C_TO_XEN_ULONG((unsigned long)arr), make_xm_obj(arr)));
    }
  if (strcmp(ctype, "GtkPathPriorityType*") == 0)
    {
      GtkPathPriorityType* arr; arr = (GtkPathPriorityType*)CALLOC(len, sizeof(GtkPathPriorityType));
      for (i = 0; i < len; i++, val = XEN_CDR(val)) arr[i] = XEN_TO_C_GtkPathPriorityType(XEN_CAR(val));
      return(XEN_LIST_3(C_STRING_TO_XEN_SYMBOL("GtkPathPriorityType_"), C_TO_XEN_ULONG((unsigned long)arr), make_xm_obj(arr)));
    }
  if (strcmp(ctype, "GtkSortType*") == 0)
    {
      GtkSortType* arr; arr = (GtkSortType*)CALLOC(len, sizeof(GtkSortType));
      for (i = 0; i < len; i++, val = XEN_CDR(val)) arr[i] = XEN_TO_C_GtkSortType(XEN_CAR(val));
      return(XEN_LIST_3(C_STRING_TO_XEN_SYMBOL("GtkSortType_"), C_TO_XEN_ULONG((unsigned long)arr), make_xm_obj(arr)));
    }
#if (!HAVE_GTK_1)
  if (strcmp(ctype, "gunichar*") == 0)
    {
      gunichar* arr; arr = (gunichar*)CALLOC(len, sizeof(gunichar));
      for (i = 0; i < len; i++, val = XEN_CDR(val)) arr[i] = XEN_TO_C_gunichar(XEN_CAR(val));
      return(XEN_LIST_3(C_STRING_TO_XEN_SYMBOL("gunichar_"), C_TO_XEN_ULONG((unsigned long)arr), make_xm_obj(arr)));
    }
#endif
  if (strcmp(ctype, "GdkPixmap**") == 0)
    {
      GdkPixmap** arr; arr = (GdkPixmap**)CALLOC(len, sizeof(GdkPixmap*));
      for (i = 0; i < len; i++, val = XEN_CDR(val)) arr[i] = XEN_TO_C_GdkPixmap_(XEN_CAR(val));
      return(XEN_LIST_3(C_STRING_TO_XEN_SYMBOL("GdkPixmap__"), C_TO_XEN_ULONG((unsigned long)arr), make_xm_obj(arr)));
    }
  if (strcmp(ctype, "GdkGC**") == 0)
    {
      GdkGC** arr; arr = (GdkGC**)CALLOC(len, sizeof(GdkGC*));
      for (i = 0; i < len; i++, val = XEN_CDR(val)) arr[i] = XEN_TO_C_GdkGC_(XEN_CAR(val));
      return(XEN_LIST_3(C_STRING_TO_XEN_SYMBOL("GdkGC__"), C_TO_XEN_ULONG((unsigned long)arr), make_xm_obj(arr)));
    }
  if (strcmp(ctype, "GtkWidget**") == 0)
    {
      GtkWidget** arr; arr = (GtkWidget**)CALLOC(len, sizeof(GtkWidget*));
      for (i = 0; i < len; i++, val = XEN_CDR(val)) arr[i] = XEN_TO_C_GtkWidget_(XEN_CAR(val));
      return(XEN_LIST_3(C_STRING_TO_XEN_SYMBOL("GtkWidget__"), C_TO_XEN_ULONG((unsigned long)arr), make_xm_obj(arr)));
    }
#if (!HAVE_GTK_1)
  if (strcmp(ctype, "PangoFontDescription**") == 0)
    {
      PangoFontDescription** arr; arr = (PangoFontDescription**)CALLOC(len, sizeof(PangoFontDescription*));
      for (i = 0; i < len; i++, val = XEN_CDR(val)) arr[i] = XEN_TO_C_PangoFontDescription_(XEN_CAR(val));
      return(XEN_LIST_3(C_STRING_TO_XEN_SYMBOL("PangoFontDescription__"), C_TO_XEN_ULONG((unsigned long)arr), make_xm_obj(arr)));
    }
  if (strcmp(ctype, "PangoAttrList**") == 0)
    {
      PangoAttrList** arr; arr = (PangoAttrList**)CALLOC(len, sizeof(PangoAttrList*));
      for (i = 0; i < len; i++, val = XEN_CDR(val)) arr[i] = XEN_TO_C_PangoAttrList_(XEN_CAR(val));
      return(XEN_LIST_3(C_STRING_TO_XEN_SYMBOL("PangoAttrList__"), C_TO_XEN_ULONG((unsigned long)arr), make_xm_obj(arr)));
    }
  if (strcmp(ctype, "GtkTreeModel**") == 0)
    {
      GtkTreeModel** arr; arr = (GtkTreeModel**)CALLOC(len, sizeof(GtkTreeModel*));
      for (i = 0; i < len; i++, val = XEN_CDR(val)) arr[i] = XEN_TO_C_GtkTreeModel_(XEN_CAR(val));
      return(XEN_LIST_3(C_STRING_TO_XEN_SYMBOL("GtkTreeModel__"), C_TO_XEN_ULONG((unsigned long)arr), make_xm_obj(arr)));
    }
#endif
  if (strcmp(ctype, "char***") == 0)
    {
      char*** arr; arr = (char***)CALLOC(len, sizeof(char**));
      for (i = 0; i < len; i++, val = XEN_CDR(val)) arr[i] = XEN_TO_C_char__(XEN_CAR(val));
      return(XEN_LIST_3(C_STRING_TO_XEN_SYMBOL("char___"), C_TO_XEN_ULONG((unsigned long)arr), make_xm_obj(arr)));
    }
  if (strcmp(ctype, "char**") == 0)
    {
      char** arr; arr = (char**)CALLOC(len, sizeof(char*));
      for (i = 0; i < len; i++, val = XEN_CDR(val)) arr[i] = XEN_TO_C_char_(XEN_CAR(val));
      return(XEN_LIST_3(C_STRING_TO_XEN_SYMBOL("char__"), C_TO_XEN_ULONG((unsigned long)arr), make_xm_obj(arr)));
    }
  if (strcmp(ctype, "gchar**") == 0)
    {
      gchar** arr; arr = (gchar**)CALLOC(len, sizeof(gchar*));
      for (i = 0; i < len; i++, val = XEN_CDR(val)) arr[i] = XEN_TO_C_gchar_(XEN_CAR(val));
      return(XEN_LIST_3(C_STRING_TO_XEN_SYMBOL("gchar__"), C_TO_XEN_ULONG((unsigned long)arr), make_xm_obj(arr)));
    }
  if (strcmp(ctype, "GdkBitmap**") == 0)
    {
      GdkBitmap** arr; arr = (GdkBitmap**)CALLOC(len, sizeof(GdkBitmap*));
      for (i = 0; i < len; i++, val = XEN_CDR(val)) arr[i] = XEN_TO_C_GdkBitmap_(XEN_CAR(val));
      return(XEN_LIST_3(C_STRING_TO_XEN_SYMBOL("GdkBitmap__"), C_TO_XEN_ULONG((unsigned long)arr), make_xm_obj(arr)));
    }
  if (strcmp(ctype, "gchar***") == 0)
    {
      gchar*** arr; arr = (gchar***)CALLOC(len, sizeof(gchar**));
      for (i = 0; i < len; i++, val = XEN_CDR(val)) arr[i] = XEN_TO_C_gchar__(XEN_CAR(val));
      return(XEN_LIST_3(C_STRING_TO_XEN_SYMBOL("gchar___"), C_TO_XEN_ULONG((unsigned long)arr), make_xm_obj(arr)));
    }
  return(XEN_FALSE);
}


static XEN gxg_pixel(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GdkColor__P(ptr), ptr, XEN_ONLY_ARG, "pixel", "GdkColor");
  return(C_TO_XEN_guint32((guint32)((XEN_TO_C_GdkColor_(ptr))->pixel)));
}

static XEN gxg_red(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GdkColor__P(ptr), ptr, XEN_ONLY_ARG, "red", "GdkColor");
  return(C_TO_XEN_guint16((guint16)((XEN_TO_C_GdkColor_(ptr))->red)));
}

static XEN gxg_green(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GdkColor__P(ptr), ptr, XEN_ONLY_ARG, "green", "GdkColor");
  return(C_TO_XEN_guint16((guint16)((XEN_TO_C_GdkColor_(ptr))->green)));
}

static XEN gxg_blue(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GdkColor__P(ptr), ptr, XEN_ONLY_ARG, "blue", "GdkColor");
  return(C_TO_XEN_guint16((guint16)((XEN_TO_C_GdkColor_(ptr))->blue)));
}

static XEN gxg_size(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GdkColormap__P(ptr) || XEN_GtkFontSelection__P(ptr), ptr, XEN_ONLY_ARG, "size", "GdkColormap" " or " "GtkFontSelection");
  if (XEN_GdkColormap__P(ptr)) return(C_TO_XEN_gint((gint)((XEN_TO_C_GdkColormap_(ptr))->size)));
  return(C_TO_XEN_gint((gint)((XEN_TO_C_GtkFontSelection_(ptr))->size)));
}

static XEN gxg_colors(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GdkColormap__P(ptr) || XEN_GdkRgbCmap__P(ptr), ptr, XEN_ONLY_ARG, "colors", "GdkColormap" " or " "GdkRgbCmap");
  if (XEN_GdkColormap__P(ptr)) return(C_TO_XEN_GdkColor_((GdkColor*)((XEN_TO_C_GdkColormap_(ptr))->colors)));
  return(C_TO_XEN_guint32_((guint32*)((XEN_TO_C_GdkRgbCmap_(ptr))->colors)));
}

#if (!HAVE_GTK_1)
static XEN gxg_visual(XEN ptr)
{
  if (XEN_GdkColormap__P(ptr)) return(C_TO_XEN_GdkVisual_((GdkVisual*)((XEN_TO_C_GdkColormap_(ptr))->visual)));
  if (XEN_GdkImage__P(ptr)) return(C_TO_XEN_GdkVisual_((GdkVisual*)((XEN_TO_C_GdkImage_(ptr))->visual)));
  if (XEN_GdkWindowAttr__P(ptr)) return(C_TO_XEN_GdkVisual_((GdkVisual*)((XEN_TO_C_GdkWindowAttr_(ptr))->visual)));
  XEN_ASSERT_TYPE(0, ptr, XEN_ONLY_ARG, "visual", "pointer to struct with visual field");
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_windowing_data(XEN ptr)
{
  if (XEN_GdkColormap__P(ptr)) return(C_TO_XEN_gpointer((gpointer)((XEN_TO_C_GdkColormap_(ptr))->windowing_data)));
  if (XEN_GdkDragContext__P(ptr)) return(C_TO_XEN_gpointer((gpointer)((XEN_TO_C_GdkDragContext_(ptr))->windowing_data)));
  if (XEN_GdkImage__P(ptr)) return(C_TO_XEN_gpointer((gpointer)((XEN_TO_C_GdkImage_(ptr))->windowing_data)));
  XEN_ASSERT_TYPE(0, ptr, XEN_ONLY_ARG, "windowing_data", "pointer to struct with windowing_data field");
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_type(XEN ptr)
{
  if (XEN_GdkCursor__P(ptr)) return(C_TO_XEN_GdkCursorType((GdkCursorType)((XEN_TO_C_GdkCursor_(ptr))->type)));
  if (XEN_GdkEventAny__P(ptr)) return(C_TO_XEN_GdkEventType((GdkEventType)((XEN_TO_C_GdkEventAny_(ptr))->type)));
  if (XEN_GdkEventExpose__P(ptr)) return(C_TO_XEN_GdkEventType((GdkEventType)((XEN_TO_C_GdkEventExpose_(ptr))->type)));
  if (XEN_GdkEventNoExpose__P(ptr)) return(C_TO_XEN_GdkEventType((GdkEventType)((XEN_TO_C_GdkEventNoExpose_(ptr))->type)));
  if (XEN_GdkEventVisibility__P(ptr)) return(C_TO_XEN_GdkEventType((GdkEventType)((XEN_TO_C_GdkEventVisibility_(ptr))->type)));
  if (XEN_GdkEventMotion__P(ptr)) return(C_TO_XEN_GdkEventType((GdkEventType)((XEN_TO_C_GdkEventMotion_(ptr))->type)));
  if (XEN_GdkEventButton__P(ptr)) return(C_TO_XEN_GdkEventType((GdkEventType)((XEN_TO_C_GdkEventButton_(ptr))->type)));
  if (XEN_GdkEventScroll__P(ptr)) return(C_TO_XEN_GdkEventType((GdkEventType)((XEN_TO_C_GdkEventScroll_(ptr))->type)));
  if (XEN_GdkEventKey__P(ptr)) return(C_TO_XEN_GdkEventType((GdkEventType)((XEN_TO_C_GdkEventKey_(ptr))->type)));
  if (XEN_GdkEventCrossing__P(ptr)) return(C_TO_XEN_GdkEventType((GdkEventType)((XEN_TO_C_GdkEventCrossing_(ptr))->type)));
  if (XEN_GdkEventFocus__P(ptr)) return(C_TO_XEN_GdkEventType((GdkEventType)((XEN_TO_C_GdkEventFocus_(ptr))->type)));
  if (XEN_GdkEventConfigure__P(ptr)) return(C_TO_XEN_GdkEventType((GdkEventType)((XEN_TO_C_GdkEventConfigure_(ptr))->type)));
  if (XEN_GdkEventProperty__P(ptr)) return(C_TO_XEN_GdkEventType((GdkEventType)((XEN_TO_C_GdkEventProperty_(ptr))->type)));
  if (XEN_GdkEventSelection__P(ptr)) return(C_TO_XEN_GdkEventType((GdkEventType)((XEN_TO_C_GdkEventSelection_(ptr))->type)));
  if (XEN_GdkEventProximity__P(ptr)) return(C_TO_XEN_GdkEventType((GdkEventType)((XEN_TO_C_GdkEventProximity_(ptr))->type)));
  if (XEN_GdkEventSetting__P(ptr)) return(C_TO_XEN_GdkEventType((GdkEventType)((XEN_TO_C_GdkEventSetting_(ptr))->type)));
  if (XEN_GdkEventWindowState__P(ptr)) return(C_TO_XEN_GdkEventType((GdkEventType)((XEN_TO_C_GdkEventWindowState_(ptr))->type)));
  if (XEN_GdkEventDND__P(ptr)) return(C_TO_XEN_GdkEventType((GdkEventType)((XEN_TO_C_GdkEventDND_(ptr))->type)));
  if (XEN_GdkImage__P(ptr)) return(C_TO_XEN_GdkImageType((GdkImageType)((XEN_TO_C_GdkImage_(ptr))->type)));
  if (XEN_GdkVisual__P(ptr)) return(C_TO_XEN_GdkVisualType((GdkVisualType)((XEN_TO_C_GdkVisual_(ptr))->type)));
  if (XEN_GtkSelectionData__P(ptr)) return(C_TO_XEN_GdkAtom((GdkAtom)((XEN_TO_C_GtkSelectionData_(ptr))->type)));
  if (XEN_GtkTreeSelection__P(ptr)) return(C_TO_XEN_GtkSelectionMode((GtkSelectionMode)((XEN_TO_C_GtkTreeSelection_(ptr))->type)));
  if (XEN_GtkWindow__P(ptr)) return(C_TO_XEN_GtkWindowType((GtkWindowType)((XEN_TO_C_GtkWindow_(ptr))->type)));
  XEN_ASSERT_TYPE(0, ptr, XEN_ONLY_ARG, "type", "pointer to struct with type field");
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_ref_count(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GdkCursor__P(ptr) || XEN_GtkTargetList__P(ptr), ptr, XEN_ONLY_ARG, "ref_count", "GdkCursor" " or " "GtkTargetList");
  if (XEN_GdkCursor__P(ptr)) return(C_TO_XEN_guint((guint)((XEN_TO_C_GdkCursor_(ptr))->ref_count)));
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkTargetList_(ptr))->ref_count)));
}
#endif

static XEN gxg_protocol(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GdkDragContext__P(ptr), ptr, XEN_ONLY_ARG, "protocol", "GdkDragContext");
  return(C_TO_XEN_GdkDragProtocol((GdkDragProtocol)((XEN_TO_C_GdkDragContext_(ptr))->protocol)));
}

static XEN gxg_is_source(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GdkDragContext__P(ptr), ptr, XEN_ONLY_ARG, "is_source", "GdkDragContext");
  return(C_TO_XEN_gboolean((gboolean)((XEN_TO_C_GdkDragContext_(ptr))->is_source)));
}

static XEN gxg_source_window(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GdkDragContext__P(ptr), ptr, XEN_ONLY_ARG, "source_window", "GdkDragContext");
  return(C_TO_XEN_GdkWindow_((GdkWindow*)((XEN_TO_C_GdkDragContext_(ptr))->source_window)));
}

static XEN gxg_dest_window(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GdkDragContext__P(ptr), ptr, XEN_ONLY_ARG, "dest_window", "GdkDragContext");
  return(C_TO_XEN_GdkWindow_((GdkWindow*)((XEN_TO_C_GdkDragContext_(ptr))->dest_window)));
}

static XEN gxg_targets(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GdkDragContext__P(ptr), ptr, XEN_ONLY_ARG, "targets", "GdkDragContext");
  return(C_TO_XEN_GList_((GList*)((XEN_TO_C_GdkDragContext_(ptr))->targets)));
}

static XEN gxg_actions(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GdkDragContext__P(ptr), ptr, XEN_ONLY_ARG, "actions", "GdkDragContext");
  return(C_TO_XEN_GdkDragAction((GdkDragAction)((XEN_TO_C_GdkDragContext_(ptr))->actions)));
}

static XEN gxg_suggested_action(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GdkDragContext__P(ptr), ptr, XEN_ONLY_ARG, "suggested_action", "GdkDragContext");
  return(C_TO_XEN_GdkDragAction((GdkDragAction)((XEN_TO_C_GdkDragContext_(ptr))->suggested_action)));
}

#if (!HAVE_GTK_1)
static XEN gxg_action(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GdkDragContext__P(ptr) || XEN_GdkEventSetting__P(ptr), ptr, XEN_ONLY_ARG, "action", "GdkDragContext" " or " "GdkEventSetting");
  if (XEN_GdkDragContext__P(ptr)) return(C_TO_XEN_GdkDragAction((GdkDragAction)((XEN_TO_C_GdkDragContext_(ptr))->action)));
  return(C_TO_XEN_GdkSettingAction((GdkSettingAction)((XEN_TO_C_GdkEventSetting_(ptr))->action)));
}
#endif

static XEN gxg_start_time(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GdkDragContext__P(ptr), ptr, XEN_ONLY_ARG, "start_time", "GdkDragContext");
  return(C_TO_XEN_guint32((guint32)((XEN_TO_C_GdkDragContext_(ptr))->start_time)));
}

#if (!HAVE_GTK_1)
static XEN gxg_window(XEN ptr)
{
  if (XEN_GdkEventAny__P(ptr)) return(C_TO_XEN_GdkWindow_((GdkWindow*)((XEN_TO_C_GdkEventAny_(ptr))->window)));
  if (XEN_GdkEventExpose__P(ptr)) return(C_TO_XEN_GdkWindow_((GdkWindow*)((XEN_TO_C_GdkEventExpose_(ptr))->window)));
  if (XEN_GdkEventNoExpose__P(ptr)) return(C_TO_XEN_GdkWindow_((GdkWindow*)((XEN_TO_C_GdkEventNoExpose_(ptr))->window)));
  if (XEN_GdkEventVisibility__P(ptr)) return(C_TO_XEN_GdkWindow_((GdkWindow*)((XEN_TO_C_GdkEventVisibility_(ptr))->window)));
  if (XEN_GdkEventMotion__P(ptr)) return(C_TO_XEN_GdkWindow_((GdkWindow*)((XEN_TO_C_GdkEventMotion_(ptr))->window)));
  if (XEN_GdkEventButton__P(ptr)) return(C_TO_XEN_GdkWindow_((GdkWindow*)((XEN_TO_C_GdkEventButton_(ptr))->window)));
  if (XEN_GdkEventScroll__P(ptr)) return(C_TO_XEN_GdkWindow_((GdkWindow*)((XEN_TO_C_GdkEventScroll_(ptr))->window)));
  if (XEN_GdkEventKey__P(ptr)) return(C_TO_XEN_GdkWindow_((GdkWindow*)((XEN_TO_C_GdkEventKey_(ptr))->window)));
  if (XEN_GdkEventCrossing__P(ptr)) return(C_TO_XEN_GdkWindow_((GdkWindow*)((XEN_TO_C_GdkEventCrossing_(ptr))->window)));
  if (XEN_GdkEventFocus__P(ptr)) return(C_TO_XEN_GdkWindow_((GdkWindow*)((XEN_TO_C_GdkEventFocus_(ptr))->window)));
  if (XEN_GdkEventConfigure__P(ptr)) return(C_TO_XEN_GdkWindow_((GdkWindow*)((XEN_TO_C_GdkEventConfigure_(ptr))->window)));
  if (XEN_GdkEventProperty__P(ptr)) return(C_TO_XEN_GdkWindow_((GdkWindow*)((XEN_TO_C_GdkEventProperty_(ptr))->window)));
  if (XEN_GdkEventSelection__P(ptr)) return(C_TO_XEN_GdkWindow_((GdkWindow*)((XEN_TO_C_GdkEventSelection_(ptr))->window)));
  if (XEN_GdkEventProximity__P(ptr)) return(C_TO_XEN_GdkWindow_((GdkWindow*)((XEN_TO_C_GdkEventProximity_(ptr))->window)));
  if (XEN_GdkEventSetting__P(ptr)) return(C_TO_XEN_GdkWindow_((GdkWindow*)((XEN_TO_C_GdkEventSetting_(ptr))->window)));
  if (XEN_GdkEventWindowState__P(ptr)) return(C_TO_XEN_GdkWindow_((GdkWindow*)((XEN_TO_C_GdkEventWindowState_(ptr))->window)));
  if (XEN_GdkEventDND__P(ptr)) return(C_TO_XEN_GdkWindow_((GdkWindow*)((XEN_TO_C_GdkEventDND_(ptr))->window)));
  if (XEN_GtkTreeViewColumn__P(ptr)) return(C_TO_XEN_GdkWindow_((GdkWindow*)((XEN_TO_C_GtkTreeViewColumn_(ptr))->window)));
  if (XEN_GtkWidget__P(ptr)) return(C_TO_XEN_GdkWindow_((GdkWindow*)((XEN_TO_C_GtkWidget_(ptr))->window)));
  XEN_ASSERT_TYPE(0, ptr, XEN_ONLY_ARG, "window", "pointer to struct with window field");
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_send_event(XEN ptr)
{
  if (XEN_GdkEventAny__P(ptr)) return(C_TO_XEN_gint8((gint8)((XEN_TO_C_GdkEventAny_(ptr))->send_event)));
  if (XEN_GdkEventExpose__P(ptr)) return(C_TO_XEN_gint8((gint8)((XEN_TO_C_GdkEventExpose_(ptr))->send_event)));
  if (XEN_GdkEventNoExpose__P(ptr)) return(C_TO_XEN_gint8((gint8)((XEN_TO_C_GdkEventNoExpose_(ptr))->send_event)));
  if (XEN_GdkEventVisibility__P(ptr)) return(C_TO_XEN_gint8((gint8)((XEN_TO_C_GdkEventVisibility_(ptr))->send_event)));
  if (XEN_GdkEventMotion__P(ptr)) return(C_TO_XEN_gint8((gint8)((XEN_TO_C_GdkEventMotion_(ptr))->send_event)));
  if (XEN_GdkEventButton__P(ptr)) return(C_TO_XEN_gint8((gint8)((XEN_TO_C_GdkEventButton_(ptr))->send_event)));
  if (XEN_GdkEventScroll__P(ptr)) return(C_TO_XEN_gint8((gint8)((XEN_TO_C_GdkEventScroll_(ptr))->send_event)));
  if (XEN_GdkEventKey__P(ptr)) return(C_TO_XEN_gint8((gint8)((XEN_TO_C_GdkEventKey_(ptr))->send_event)));
  if (XEN_GdkEventCrossing__P(ptr)) return(C_TO_XEN_gint8((gint8)((XEN_TO_C_GdkEventCrossing_(ptr))->send_event)));
  if (XEN_GdkEventFocus__P(ptr)) return(C_TO_XEN_gint8((gint8)((XEN_TO_C_GdkEventFocus_(ptr))->send_event)));
  if (XEN_GdkEventConfigure__P(ptr)) return(C_TO_XEN_gint8((gint8)((XEN_TO_C_GdkEventConfigure_(ptr))->send_event)));
  if (XEN_GdkEventProperty__P(ptr)) return(C_TO_XEN_gint8((gint8)((XEN_TO_C_GdkEventProperty_(ptr))->send_event)));
  if (XEN_GdkEventSelection__P(ptr)) return(C_TO_XEN_gint8((gint8)((XEN_TO_C_GdkEventSelection_(ptr))->send_event)));
  if (XEN_GdkEventProximity__P(ptr)) return(C_TO_XEN_gint8((gint8)((XEN_TO_C_GdkEventProximity_(ptr))->send_event)));
  if (XEN_GdkEventSetting__P(ptr)) return(C_TO_XEN_gint8((gint8)((XEN_TO_C_GdkEventSetting_(ptr))->send_event)));
  if (XEN_GdkEventWindowState__P(ptr)) return(C_TO_XEN_gint8((gint8)((XEN_TO_C_GdkEventWindowState_(ptr))->send_event)));
  if (XEN_GdkEventDND__P(ptr)) return(C_TO_XEN_gint8((gint8)((XEN_TO_C_GdkEventDND_(ptr))->send_event)));
  XEN_ASSERT_TYPE(0, ptr, XEN_ONLY_ARG, "send_event", "pointer to struct with send_event field");
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_region(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GdkEventExpose__P(ptr), ptr, XEN_ONLY_ARG, "region", "GdkEventExpose");
  return(C_TO_XEN_GdkRegion_((GdkRegion*)((XEN_TO_C_GdkEventExpose_(ptr))->region)));
}
#endif

static XEN gxg_count(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GdkEventExpose__P(ptr), ptr, XEN_ONLY_ARG, "count", "GdkEventExpose");
  return(C_TO_XEN_gint((gint)((XEN_TO_C_GdkEventExpose_(ptr))->count)));
}

#if (!HAVE_GTK_1)
static XEN gxg_state(XEN ptr)
{
  if (XEN_GdkEventVisibility__P(ptr)) return(C_TO_XEN_GdkVisibilityState((GdkVisibilityState)((XEN_TO_C_GdkEventVisibility_(ptr))->state)));
  if (XEN_GdkEventMotion__P(ptr)) return(C_TO_XEN_guint((guint)((XEN_TO_C_GdkEventMotion_(ptr))->state)));
  if (XEN_GdkEventButton__P(ptr)) return(C_TO_XEN_guint((guint)((XEN_TO_C_GdkEventButton_(ptr))->state)));
  if (XEN_GdkEventScroll__P(ptr)) return(C_TO_XEN_guint((guint)((XEN_TO_C_GdkEventScroll_(ptr))->state)));
  if (XEN_GdkEventKey__P(ptr)) return(C_TO_XEN_guint((guint)((XEN_TO_C_GdkEventKey_(ptr))->state)));
  if (XEN_GdkEventCrossing__P(ptr)) return(C_TO_XEN_guint((guint)((XEN_TO_C_GdkEventCrossing_(ptr))->state)));
  if (XEN_GdkEventProperty__P(ptr)) return(C_TO_XEN_guint((guint)((XEN_TO_C_GdkEventProperty_(ptr))->state)));
  if (XEN_GdkWindowObject__P(ptr)) return(C_TO_XEN_GdkWindowState((GdkWindowState)((XEN_TO_C_GdkWindowObject_(ptr))->state)));
  if (XEN_GtkWidget__P(ptr)) return(C_TO_XEN_guint8((guint8)((XEN_TO_C_GtkWidget_(ptr))->state)));
  XEN_ASSERT_TYPE(0, ptr, XEN_ONLY_ARG, "state", "pointer to struct with state field");
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_time(XEN ptr)
{
  if (XEN_GdkEventMotion__P(ptr)) return(C_TO_XEN_guint32((guint32)((XEN_TO_C_GdkEventMotion_(ptr))->time)));
  if (XEN_GdkEventButton__P(ptr)) return(C_TO_XEN_guint32((guint32)((XEN_TO_C_GdkEventButton_(ptr))->time)));
  if (XEN_GdkEventScroll__P(ptr)) return(C_TO_XEN_guint32((guint32)((XEN_TO_C_GdkEventScroll_(ptr))->time)));
  if (XEN_GdkEventKey__P(ptr)) return(C_TO_XEN_guint32((guint32)((XEN_TO_C_GdkEventKey_(ptr))->time)));
  if (XEN_GdkEventCrossing__P(ptr)) return(C_TO_XEN_guint32((guint32)((XEN_TO_C_GdkEventCrossing_(ptr))->time)));
  if (XEN_GdkEventProperty__P(ptr)) return(C_TO_XEN_guint32((guint32)((XEN_TO_C_GdkEventProperty_(ptr))->time)));
  if (XEN_GdkEventSelection__P(ptr)) return(C_TO_XEN_guint32((guint32)((XEN_TO_C_GdkEventSelection_(ptr))->time)));
  if (XEN_GdkEventProximity__P(ptr)) return(C_TO_XEN_guint32((guint32)((XEN_TO_C_GdkEventProximity_(ptr))->time)));
  if (XEN_GdkEventDND__P(ptr)) return(C_TO_XEN_guint32((guint32)((XEN_TO_C_GdkEventDND_(ptr))->time)));
  XEN_ASSERT_TYPE(0, ptr, XEN_ONLY_ARG, "time", "pointer to struct with time field");
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_x(XEN ptr)
{
  if (XEN_GdkEventMotion__P(ptr)) return(C_TO_XEN_gdouble((gdouble)((XEN_TO_C_GdkEventMotion_(ptr))->x)));
  if (XEN_GdkEventButton__P(ptr)) return(C_TO_XEN_gdouble((gdouble)((XEN_TO_C_GdkEventButton_(ptr))->x)));
  if (XEN_GdkEventScroll__P(ptr)) return(C_TO_XEN_gdouble((gdouble)((XEN_TO_C_GdkEventScroll_(ptr))->x)));
  if (XEN_GdkEventCrossing__P(ptr)) return(C_TO_XEN_gdouble((gdouble)((XEN_TO_C_GdkEventCrossing_(ptr))->x)));
  if (XEN_GdkEventConfigure__P(ptr)) return(C_TO_XEN_gint((gint)((XEN_TO_C_GdkEventConfigure_(ptr))->x)));
  if (XEN_GdkPoint__P(ptr)) return(C_TO_XEN_gint((gint)((XEN_TO_C_GdkPoint_(ptr))->x)));
  if (XEN_GdkRectangle__P(ptr)) return(C_TO_XEN_gint((gint)((XEN_TO_C_GdkRectangle_(ptr))->x)));
  if (XEN_GdkSpan__P(ptr)) return(C_TO_XEN_gint((gint)((XEN_TO_C_GdkSpan_(ptr))->x)));
  if (XEN_GdkWindowAttr__P(ptr)) return(C_TO_XEN_gint((gint)((XEN_TO_C_GdkWindowAttr_(ptr))->x)));
  if (XEN_GdkWindowObject__P(ptr)) return(C_TO_XEN_gint((gint)((XEN_TO_C_GdkWindowObject_(ptr))->x)));
  XEN_ASSERT_TYPE(0, ptr, XEN_ONLY_ARG, "x", "pointer to struct with x field");
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_y(XEN ptr)
{
  if (XEN_GdkEventMotion__P(ptr)) return(C_TO_XEN_gdouble((gdouble)((XEN_TO_C_GdkEventMotion_(ptr))->y)));
  if (XEN_GdkEventButton__P(ptr)) return(C_TO_XEN_gdouble((gdouble)((XEN_TO_C_GdkEventButton_(ptr))->y)));
  if (XEN_GdkEventScroll__P(ptr)) return(C_TO_XEN_gdouble((gdouble)((XEN_TO_C_GdkEventScroll_(ptr))->y)));
  if (XEN_GdkEventCrossing__P(ptr)) return(C_TO_XEN_gdouble((gdouble)((XEN_TO_C_GdkEventCrossing_(ptr))->y)));
  if (XEN_GdkEventConfigure__P(ptr)) return(C_TO_XEN_gint((gint)((XEN_TO_C_GdkEventConfigure_(ptr))->y)));
  if (XEN_GdkPoint__P(ptr)) return(C_TO_XEN_gint((gint)((XEN_TO_C_GdkPoint_(ptr))->y)));
  if (XEN_GdkRectangle__P(ptr)) return(C_TO_XEN_gint((gint)((XEN_TO_C_GdkRectangle_(ptr))->y)));
  if (XEN_GdkSpan__P(ptr)) return(C_TO_XEN_gint((gint)((XEN_TO_C_GdkSpan_(ptr))->y)));
  if (XEN_GdkWindowAttr__P(ptr)) return(C_TO_XEN_gint((gint)((XEN_TO_C_GdkWindowAttr_(ptr))->y)));
  if (XEN_GdkWindowObject__P(ptr)) return(C_TO_XEN_gint((gint)((XEN_TO_C_GdkWindowObject_(ptr))->y)));
  XEN_ASSERT_TYPE(0, ptr, XEN_ONLY_ARG, "y", "pointer to struct with y field");
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_axes(XEN ptr)
{
  if (XEN_GdkEventMotion__P(ptr)) return(C_TO_XEN_gdouble_((gdouble*)((XEN_TO_C_GdkEventMotion_(ptr))->axes)));
  if (XEN_GdkEventButton__P(ptr)) return(C_TO_XEN_gdouble_((gdouble*)((XEN_TO_C_GdkEventButton_(ptr))->axes)));
  if (XEN_GdkDevice__P(ptr)) return(C_TO_XEN_GdkDeviceAxis_((GdkDeviceAxis*)((XEN_TO_C_GdkDevice_(ptr))->axes)));
  XEN_ASSERT_TYPE(0, ptr, XEN_ONLY_ARG, "axes", "pointer to struct with axes field");
}
#endif

static XEN gxg_is_hint(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GdkEventMotion__P(ptr), ptr, XEN_ONLY_ARG, "is_hint", "GdkEventMotion");
  return(C_TO_XEN_gint16((gint16)((XEN_TO_C_GdkEventMotion_(ptr))->is_hint)));
}

#if (!HAVE_GTK_1)
static XEN gxg_device(XEN ptr)
{
  if (XEN_GdkEventMotion__P(ptr)) return(C_TO_XEN_GdkDevice_((GdkDevice*)((XEN_TO_C_GdkEventMotion_(ptr))->device)));
  if (XEN_GdkEventButton__P(ptr)) return(C_TO_XEN_GdkDevice_((GdkDevice*)((XEN_TO_C_GdkEventButton_(ptr))->device)));
  if (XEN_GdkEventScroll__P(ptr)) return(C_TO_XEN_GdkDevice_((GdkDevice*)((XEN_TO_C_GdkEventScroll_(ptr))->device)));
  if (XEN_GdkEventProximity__P(ptr)) return(C_TO_XEN_GdkDevice_((GdkDevice*)((XEN_TO_C_GdkEventProximity_(ptr))->device)));
  XEN_ASSERT_TYPE(0, ptr, XEN_ONLY_ARG, "device", "pointer to struct with device field");
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_x_root(XEN ptr)
{
  if (XEN_GdkEventMotion__P(ptr)) return(C_TO_XEN_gdouble((gdouble)((XEN_TO_C_GdkEventMotion_(ptr))->x_root)));
  if (XEN_GdkEventButton__P(ptr)) return(C_TO_XEN_gdouble((gdouble)((XEN_TO_C_GdkEventButton_(ptr))->x_root)));
  if (XEN_GdkEventScroll__P(ptr)) return(C_TO_XEN_gdouble((gdouble)((XEN_TO_C_GdkEventScroll_(ptr))->x_root)));
  if (XEN_GdkEventCrossing__P(ptr)) return(C_TO_XEN_gdouble((gdouble)((XEN_TO_C_GdkEventCrossing_(ptr))->x_root)));
  if (XEN_GdkEventDND__P(ptr)) return(C_TO_XEN_gshort((gshort)((XEN_TO_C_GdkEventDND_(ptr))->x_root)));
  XEN_ASSERT_TYPE(0, ptr, XEN_ONLY_ARG, "x_root", "pointer to struct with x_root field");
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_y_root(XEN ptr)
{
  if (XEN_GdkEventMotion__P(ptr)) return(C_TO_XEN_gdouble((gdouble)((XEN_TO_C_GdkEventMotion_(ptr))->y_root)));
  if (XEN_GdkEventButton__P(ptr)) return(C_TO_XEN_gdouble((gdouble)((XEN_TO_C_GdkEventButton_(ptr))->y_root)));
  if (XEN_GdkEventScroll__P(ptr)) return(C_TO_XEN_gdouble((gdouble)((XEN_TO_C_GdkEventScroll_(ptr))->y_root)));
  if (XEN_GdkEventCrossing__P(ptr)) return(C_TO_XEN_gdouble((gdouble)((XEN_TO_C_GdkEventCrossing_(ptr))->y_root)));
  if (XEN_GdkEventDND__P(ptr)) return(C_TO_XEN_gshort((gshort)((XEN_TO_C_GdkEventDND_(ptr))->y_root)));
  XEN_ASSERT_TYPE(0, ptr, XEN_ONLY_ARG, "y_root", "pointer to struct with y_root field");
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_button(XEN ptr)
{
  if (XEN_GdkEventButton__P(ptr)) return(C_TO_XEN_guint((guint)((XEN_TO_C_GdkEventButton_(ptr))->button)));
  if (XEN_GtkCombo__P(ptr)) return(C_TO_XEN_GtkWidget_((GtkWidget*)((XEN_TO_C_GtkCombo_(ptr))->button)));
  if (XEN_GtkEntry__P(ptr)) return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkEntry_(ptr))->button)));
  if (XEN_GtkGammaCurve__P(ptr)) return(C_TO_XEN_GtkWidget__((GtkWidget**)((XEN_TO_C_GtkGammaCurve_(ptr))->button)));
  if (XEN_GtkMenuShell__P(ptr)) return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkMenuShell_(ptr))->button)));
  if (XEN_GtkNotebook__P(ptr)) return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkNotebook_(ptr))->button)));
  if (XEN_GtkSpinButton__P(ptr)) return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkSpinButton_(ptr))->button)));
  if (XEN_GtkTreeViewColumn__P(ptr)) return(C_TO_XEN_GtkWidget_((GtkWidget*)((XEN_TO_C_GtkTreeViewColumn_(ptr))->button)));
  XEN_ASSERT_TYPE(0, ptr, XEN_ONLY_ARG, "button", "pointer to struct with button field");
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_direction(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GdkEventScroll__P(ptr), ptr, XEN_ONLY_ARG, "direction", "GdkEventScroll");
  return(C_TO_XEN_GdkScrollDirection((GdkScrollDirection)((XEN_TO_C_GdkEventScroll_(ptr))->direction)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_keyval(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GdkEventKey__P(ptr) || XEN_GtkStockItem__P(ptr), ptr, XEN_ONLY_ARG, "keyval", "GdkEventKey" " or " "GtkStockItem");
  if (XEN_GdkEventKey__P(ptr)) return(C_TO_XEN_guint((guint)((XEN_TO_C_GdkEventKey_(ptr))->keyval)));
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkStockItem_(ptr))->keyval)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_length(XEN ptr)
{
  if (XEN_GdkEventKey__P(ptr)) return(C_TO_XEN_gint((gint)((XEN_TO_C_GdkEventKey_(ptr))->length)));
  if (XEN_GtkListStore__P(ptr)) return(C_TO_XEN_gint((gint)((XEN_TO_C_GtkListStore_(ptr))->length)));
  if (XEN_GtkSelectionData__P(ptr)) return(C_TO_XEN_gint((gint)((XEN_TO_C_GtkSelectionData_(ptr))->length)));
  XEN_ASSERT_TYPE(0, ptr, XEN_ONLY_ARG, "length", "pointer to struct with length field");
}
#endif

static XEN gxg_string(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GdkEventKey__P(ptr), ptr, XEN_ONLY_ARG, "string", "GdkEventKey");
  return(C_TO_XEN_gchar_((gchar*)((XEN_TO_C_GdkEventKey_(ptr))->string)));
}

#if (!HAVE_GTK_1)
static XEN gxg_hardware_keycode(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GdkEventKey__P(ptr), ptr, XEN_ONLY_ARG, "hardware_keycode", "GdkEventKey");
  return(C_TO_XEN_guint16((guint16)((XEN_TO_C_GdkEventKey_(ptr))->hardware_keycode)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_group(XEN ptr)
{
  if (XEN_GdkEventKey__P(ptr)) return(C_TO_XEN_guint8((guint8)((XEN_TO_C_GdkEventKey_(ptr))->group)));
  if (XEN_GdkKeymapKey__P(ptr)) return(C_TO_XEN_gint((gint)((XEN_TO_C_GdkKeymapKey_(ptr))->group)));
  if (XEN_GtkRadioButton__P(ptr)) return(C_TO_XEN_GSList_((GSList*)((XEN_TO_C_GtkRadioButton_(ptr))->group)));
  if (XEN_GtkRadioMenuItem__P(ptr)) return(C_TO_XEN_GSList_((GSList*)((XEN_TO_C_GtkRadioMenuItem_(ptr))->group)));
  if (XEN_GtkWindow__P(ptr)) return(C_TO_XEN_GtkWindowGroup_((GtkWindowGroup*)((XEN_TO_C_GtkWindow_(ptr))->group)));
  XEN_ASSERT_TYPE(0, ptr, XEN_ONLY_ARG, "group", "pointer to struct with group field");
}
#endif

static XEN gxg_subwindow(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GdkEventCrossing__P(ptr), ptr, XEN_ONLY_ARG, "subwindow", "GdkEventCrossing");
  return(C_TO_XEN_GdkWindow_((GdkWindow*)((XEN_TO_C_GdkEventCrossing_(ptr))->subwindow)));
}

#if (!HAVE_GTK_1)
static XEN gxg_mode(XEN ptr)
{
  if (XEN_GdkEventCrossing__P(ptr)) return(C_TO_XEN_GdkCrossingMode((GdkCrossingMode)((XEN_TO_C_GdkEventCrossing_(ptr))->mode)));
  if (XEN_GdkDevice__P(ptr)) return(C_TO_XEN_GdkInputMode((GdkInputMode)((XEN_TO_C_GdkDevice_(ptr))->mode)));
  if (XEN_GtkCellRenderer__P(ptr)) return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkCellRenderer_(ptr))->mode)));
  if (XEN_GtkSizeGroup__P(ptr)) return(C_TO_XEN_guint8((guint8)((XEN_TO_C_GtkSizeGroup_(ptr))->mode)));
  XEN_ASSERT_TYPE(0, ptr, XEN_ONLY_ARG, "mode", "pointer to struct with mode field");
}
#endif

static XEN gxg_detail(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GdkEventCrossing__P(ptr), ptr, XEN_ONLY_ARG, "detail", "GdkEventCrossing");
  return(C_TO_XEN_GdkNotifyType((GdkNotifyType)((XEN_TO_C_GdkEventCrossing_(ptr))->detail)));
}

static XEN gxg_focus(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GdkEventCrossing__P(ptr), ptr, XEN_ONLY_ARG, "focus", "GdkEventCrossing");
  return(C_TO_XEN_gboolean((gboolean)((XEN_TO_C_GdkEventCrossing_(ptr))->focus)));
}

static XEN gxg_in(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GdkEventFocus__P(ptr), ptr, XEN_ONLY_ARG, "in", "GdkEventFocus");
  return(C_TO_XEN_gint16((gint16)((XEN_TO_C_GdkEventFocus_(ptr))->in)));
}

#if (!HAVE_GTK_1)
static XEN gxg_width(XEN ptr)
{
  if (XEN_GdkEventConfigure__P(ptr)) return(C_TO_XEN_gint((gint)((XEN_TO_C_GdkEventConfigure_(ptr))->width)));
  if (XEN_GdkImage__P(ptr)) return(C_TO_XEN_gint((gint)((XEN_TO_C_GdkImage_(ptr))->width)));
  if (XEN_GdkRectangle__P(ptr)) return(C_TO_XEN_gint((gint)((XEN_TO_C_GdkRectangle_(ptr))->width)));
  if (XEN_GdkSpan__P(ptr)) return(C_TO_XEN_gint((gint)((XEN_TO_C_GdkSpan_(ptr))->width)));
  if (XEN_GdkWindowAttr__P(ptr)) return(C_TO_XEN_gint((gint)((XEN_TO_C_GdkWindowAttr_(ptr))->width)));
  if (XEN_GtkCellRenderer__P(ptr)) return(C_TO_XEN_gint((gint)((XEN_TO_C_GtkCellRenderer_(ptr))->width)));
  if (XEN_GtkLayout__P(ptr)) return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkLayout_(ptr))->width)));
  if (XEN_GtkOptionMenu__P(ptr)) return(C_TO_XEN_guint16((guint16)((XEN_TO_C_GtkOptionMenu_(ptr))->width)));
  if (XEN_GtkTreeViewColumn__P(ptr)) return(C_TO_XEN_gint((gint)((XEN_TO_C_GtkTreeViewColumn_(ptr))->width)));
  if (XEN_GtkRequisition__P(ptr)) return(C_TO_XEN_gint((gint)((XEN_TO_C_GtkRequisition_(ptr))->width)));
  XEN_ASSERT_TYPE(0, ptr, XEN_ONLY_ARG, "width", "pointer to struct with width field");
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_height(XEN ptr)
{
  if (XEN_GdkEventConfigure__P(ptr)) return(C_TO_XEN_gint((gint)((XEN_TO_C_GdkEventConfigure_(ptr))->height)));
  if (XEN_GdkImage__P(ptr)) return(C_TO_XEN_gint((gint)((XEN_TO_C_GdkImage_(ptr))->height)));
  if (XEN_GdkRectangle__P(ptr)) return(C_TO_XEN_gint((gint)((XEN_TO_C_GdkRectangle_(ptr))->height)));
  if (XEN_GdkWindowAttr__P(ptr)) return(C_TO_XEN_gint((gint)((XEN_TO_C_GdkWindowAttr_(ptr))->height)));
  if (XEN_GtkCellRenderer__P(ptr)) return(C_TO_XEN_gint((gint)((XEN_TO_C_GtkCellRenderer_(ptr))->height)));
  if (XEN_GtkCurve__P(ptr)) return(C_TO_XEN_gint((gint)((XEN_TO_C_GtkCurve_(ptr))->height)));
  if (XEN_GtkLayout__P(ptr)) return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkLayout_(ptr))->height)));
  if (XEN_GtkOptionMenu__P(ptr)) return(C_TO_XEN_guint16((guint16)((XEN_TO_C_GtkOptionMenu_(ptr))->height)));
  if (XEN_GtkRequisition__P(ptr)) return(C_TO_XEN_gint((gint)((XEN_TO_C_GtkRequisition_(ptr))->height)));
  XEN_ASSERT_TYPE(0, ptr, XEN_ONLY_ARG, "height", "pointer to struct with height field");
}
#endif

static XEN gxg_atom(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GdkEventProperty__P(ptr), ptr, XEN_ONLY_ARG, "atom", "GdkEventProperty");
  return(C_TO_XEN_GdkAtom((GdkAtom)((XEN_TO_C_GdkEventProperty_(ptr))->atom)));
}

static XEN gxg_selection(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GdkEventSelection__P(ptr) || XEN_GtkSelectionData__P(ptr), ptr, XEN_ONLY_ARG, "selection", "GdkEventSelection" " or " "GtkSelectionData");
  if (XEN_GdkEventSelection__P(ptr)) return(C_TO_XEN_GdkAtom((GdkAtom)((XEN_TO_C_GdkEventSelection_(ptr))->selection)));
  return(C_TO_XEN_GdkAtom((GdkAtom)((XEN_TO_C_GtkSelectionData_(ptr))->selection)));
}

static XEN gxg_target(XEN ptr)
{
  if (XEN_GdkEventSelection__P(ptr)) return(C_TO_XEN_GdkAtom((GdkAtom)((XEN_TO_C_GdkEventSelection_(ptr))->target)));
  if (XEN_GtkSelectionData__P(ptr)) return(C_TO_XEN_GdkAtom((GdkAtom)((XEN_TO_C_GtkSelectionData_(ptr))->target)));
  if (XEN_GtkTargetEntry__P(ptr)) return(C_TO_XEN_gchar_((gchar*)((XEN_TO_C_GtkTargetEntry_(ptr))->target)));
  XEN_ASSERT_TYPE(0, ptr, XEN_ONLY_ARG, "target", "pointer to struct with target field");
}

static XEN gxg_property(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GdkEventSelection__P(ptr), ptr, XEN_ONLY_ARG, "property", "GdkEventSelection");
  return(C_TO_XEN_GdkAtom((GdkAtom)((XEN_TO_C_GdkEventSelection_(ptr))->property)));
}

static XEN gxg_requestor(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GdkEventSelection__P(ptr), ptr, XEN_ONLY_ARG, "requestor", "GdkEventSelection");
#if (!HAVE_GTK_1)
  return(C_TO_XEN_GdkNativeWindow((GdkNativeWindow)((XEN_TO_C_GdkEventSelection_(ptr))->requestor)));
#endif
}

#if (!HAVE_GTK_1)
static XEN gxg_name(XEN ptr)
{
  if (XEN_GdkEventSetting__P(ptr)) return(C_TO_XEN_char_((char*)((XEN_TO_C_GdkEventSetting_(ptr))->name)));
  if (XEN_GdkDevice__P(ptr)) return(C_TO_XEN_gchar_((gchar*)((XEN_TO_C_GdkDevice_(ptr))->name)));
  if (XEN_GtkWidget__P(ptr)) return(C_TO_XEN_gchar_((gchar*)((XEN_TO_C_GtkWidget_(ptr))->name)));
  XEN_ASSERT_TYPE(0, ptr, XEN_ONLY_ARG, "name", "pointer to struct with name field");
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_changed_mask(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GdkEventWindowState__P(ptr), ptr, XEN_ONLY_ARG, "changed_mask", "GdkEventWindowState");
  return(C_TO_XEN_GdkWindowState((GdkWindowState)((XEN_TO_C_GdkEventWindowState_(ptr))->changed_mask)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_new_window_state(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GdkEventWindowState__P(ptr), ptr, XEN_ONLY_ARG, "new_window_state", "GdkEventWindowState");
  return(C_TO_XEN_GdkWindowState((GdkWindowState)((XEN_TO_C_GdkEventWindowState_(ptr))->new_window_state)));
}
#endif

static XEN gxg_context(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GdkEventDND__P(ptr), ptr, XEN_ONLY_ARG, "context", "GdkEventDND");
  return(C_TO_XEN_GdkDragContext_((GdkDragContext*)((XEN_TO_C_GdkEventDND_(ptr))->context)));
}

static XEN gxg_function(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GdkGCValues__P(ptr), ptr, XEN_ONLY_ARG, "function", "GdkGCValues");
  return(C_TO_XEN_GdkFunction((GdkFunction)((XEN_TO_C_GdkGCValues_(ptr))->function)));
}

static XEN gxg_fill(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GdkGCValues__P(ptr), ptr, XEN_ONLY_ARG, "fill", "GdkGCValues");
  return(C_TO_XEN_GdkFill((GdkFill)((XEN_TO_C_GdkGCValues_(ptr))->fill)));
}

static XEN gxg_tile(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GdkGCValues__P(ptr), ptr, XEN_ONLY_ARG, "tile", "GdkGCValues");
  return(C_TO_XEN_GdkPixmap_((GdkPixmap*)((XEN_TO_C_GdkGCValues_(ptr))->tile)));
}

static XEN gxg_stipple(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GdkGCValues__P(ptr), ptr, XEN_ONLY_ARG, "stipple", "GdkGCValues");
  return(C_TO_XEN_GdkPixmap_((GdkPixmap*)((XEN_TO_C_GdkGCValues_(ptr))->stipple)));
}

static XEN gxg_clip_mask(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GdkGCValues__P(ptr), ptr, XEN_ONLY_ARG, "clip_mask", "GdkGCValues");
  return(C_TO_XEN_GdkPixmap_((GdkPixmap*)((XEN_TO_C_GdkGCValues_(ptr))->clip_mask)));
}

static XEN gxg_subwindow_mode(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GdkGCValues__P(ptr), ptr, XEN_ONLY_ARG, "subwindow_mode", "GdkGCValues");
  return(C_TO_XEN_GdkSubwindowMode((GdkSubwindowMode)((XEN_TO_C_GdkGCValues_(ptr))->subwindow_mode)));
}

#if (!HAVE_GTK_1)
static XEN gxg_ts_x_origin(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GdkGCValues__P(ptr) || XEN_GdkGC__P(ptr), ptr, XEN_ONLY_ARG, "ts_x_origin", "GdkGCValues" " or " "GdkGC");
  if (XEN_GdkGCValues__P(ptr)) return(C_TO_XEN_gint((gint)((XEN_TO_C_GdkGCValues_(ptr))->ts_x_origin)));
  return(C_TO_XEN_gint((gint)((XEN_TO_C_GdkGC_(ptr))->ts_x_origin)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_ts_y_origin(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GdkGCValues__P(ptr) || XEN_GdkGC__P(ptr), ptr, XEN_ONLY_ARG, "ts_y_origin", "GdkGCValues" " or " "GdkGC");
  if (XEN_GdkGCValues__P(ptr)) return(C_TO_XEN_gint((gint)((XEN_TO_C_GdkGCValues_(ptr))->ts_y_origin)));
  return(C_TO_XEN_gint((gint)((XEN_TO_C_GdkGC_(ptr))->ts_y_origin)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_clip_x_origin(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GdkGCValues__P(ptr) || XEN_GdkGC__P(ptr), ptr, XEN_ONLY_ARG, "clip_x_origin", "GdkGCValues" " or " "GdkGC");
  if (XEN_GdkGCValues__P(ptr)) return(C_TO_XEN_gint((gint)((XEN_TO_C_GdkGCValues_(ptr))->clip_x_origin)));
  return(C_TO_XEN_gint((gint)((XEN_TO_C_GdkGC_(ptr))->clip_x_origin)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_clip_y_origin(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GdkGCValues__P(ptr) || XEN_GdkGC__P(ptr), ptr, XEN_ONLY_ARG, "clip_y_origin", "GdkGCValues" " or " "GdkGC");
  if (XEN_GdkGCValues__P(ptr)) return(C_TO_XEN_gint((gint)((XEN_TO_C_GdkGCValues_(ptr))->clip_y_origin)));
  return(C_TO_XEN_gint((gint)((XEN_TO_C_GdkGC_(ptr))->clip_y_origin)));
}
#endif

static XEN gxg_graphics_exposures(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GdkGCValues__P(ptr), ptr, XEN_ONLY_ARG, "graphics_exposures", "GdkGCValues");
  return(C_TO_XEN_gint((gint)((XEN_TO_C_GdkGCValues_(ptr))->graphics_exposures)));
}

static XEN gxg_line_width(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GdkGCValues__P(ptr), ptr, XEN_ONLY_ARG, "line_width", "GdkGCValues");
  return(C_TO_XEN_gint((gint)((XEN_TO_C_GdkGCValues_(ptr))->line_width)));
}

static XEN gxg_line_style(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GdkGCValues__P(ptr), ptr, XEN_ONLY_ARG, "line_style", "GdkGCValues");
  return(C_TO_XEN_GdkLineStyle((GdkLineStyle)((XEN_TO_C_GdkGCValues_(ptr))->line_style)));
}

static XEN gxg_cap_style(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GdkGCValues__P(ptr), ptr, XEN_ONLY_ARG, "cap_style", "GdkGCValues");
  return(C_TO_XEN_GdkCapStyle((GdkCapStyle)((XEN_TO_C_GdkGCValues_(ptr))->cap_style)));
}

static XEN gxg_join_style(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GdkGCValues__P(ptr), ptr, XEN_ONLY_ARG, "join_style", "GdkGCValues");
  return(C_TO_XEN_GdkJoinStyle((GdkJoinStyle)((XEN_TO_C_GdkGCValues_(ptr))->join_style)));
}

#if (!HAVE_GTK_1)
static XEN gxg_colormap(XEN ptr)
{
  if (XEN_GdkGC__P(ptr)) return(C_TO_XEN_GdkColormap_((GdkColormap*)((XEN_TO_C_GdkGC_(ptr))->colormap)));
  if (XEN_GdkImage__P(ptr)) return(C_TO_XEN_GdkColormap_((GdkColormap*)((XEN_TO_C_GdkImage_(ptr))->colormap)));
  if (XEN_GdkWindowAttr__P(ptr)) return(C_TO_XEN_GdkColormap_((GdkColormap*)((XEN_TO_C_GdkWindowAttr_(ptr))->colormap)));
  if (XEN_GtkStyle__P(ptr)) return(C_TO_XEN_GdkColormap_((GdkColormap*)((XEN_TO_C_GtkStyle_(ptr))->colormap)));
  XEN_ASSERT_TYPE(0, ptr, XEN_ONLY_ARG, "colormap", "pointer to struct with colormap field");
}
#endif

static XEN gxg_byte_order(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GdkImage__P(ptr) || XEN_GdkVisual__P(ptr), ptr, XEN_ONLY_ARG, "byte_order", "GdkImage" " or " "GdkVisual");
  if (XEN_GdkImage__P(ptr)) return(C_TO_XEN_GdkByteOrder((GdkByteOrder)((XEN_TO_C_GdkImage_(ptr))->byte_order)));
  return(C_TO_XEN_GdkByteOrder((GdkByteOrder)((XEN_TO_C_GdkVisual_(ptr))->byte_order)));
}

#if (!HAVE_GTK_1)
static XEN gxg_depth(XEN ptr)
{
  if (XEN_GdkImage__P(ptr)) return(C_TO_XEN_guint16((guint16)((XEN_TO_C_GdkImage_(ptr))->depth)));
  if (XEN_GdkVisual__P(ptr)) return(C_TO_XEN_gint((gint)((XEN_TO_C_GdkVisual_(ptr))->depth)));
  if (XEN_GdkWindowObject__P(ptr)) return(C_TO_XEN_guint8((guint8)((XEN_TO_C_GdkWindowObject_(ptr))->depth)));
  if (XEN_GtkStyle__P(ptr)) return(C_TO_XEN_gint((gint)((XEN_TO_C_GtkStyle_(ptr))->depth)));
  XEN_ASSERT_TYPE(0, ptr, XEN_ONLY_ARG, "depth", "pointer to struct with depth field");
}
#endif

static XEN gxg_bpp(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GdkImage__P(ptr), ptr, XEN_ONLY_ARG, "bpp", "GdkImage");
  return(C_TO_XEN_guint16((guint16)((XEN_TO_C_GdkImage_(ptr))->bpp)));
}

static XEN gxg_bpl(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GdkImage__P(ptr), ptr, XEN_ONLY_ARG, "bpl", "GdkImage");
  return(C_TO_XEN_guint16((guint16)((XEN_TO_C_GdkImage_(ptr))->bpl)));
}

#if (!HAVE_GTK_1)
static XEN gxg_bits_per_pixel(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GdkImage__P(ptr), ptr, XEN_ONLY_ARG, "bits_per_pixel", "GdkImage");
  return(C_TO_XEN_guint16((guint16)((XEN_TO_C_GdkImage_(ptr))->bits_per_pixel)));
}
#endif

static XEN gxg_mem(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GdkImage__P(ptr), ptr, XEN_ONLY_ARG, "mem", "GdkImage");
  return(C_TO_XEN_gpointer((gpointer)((XEN_TO_C_GdkImage_(ptr))->mem)));
}

#if (!HAVE_GTK_1)
static XEN gxg_source(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GdkDevice__P(ptr), ptr, XEN_ONLY_ARG, "source", "GdkDevice");
  return(C_TO_XEN_GdkInputSource((GdkInputSource)((XEN_TO_C_GdkDevice_(ptr))->source)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_has_cursor(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GdkDevice__P(ptr), ptr, XEN_ONLY_ARG, "has_cursor", "GdkDevice");
  return(C_TO_XEN_gboolean((gboolean)((XEN_TO_C_GdkDevice_(ptr))->has_cursor)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_num_axes(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GdkDevice__P(ptr), ptr, XEN_ONLY_ARG, "num_axes", "GdkDevice");
  return(C_TO_XEN_gint((gint)((XEN_TO_C_GdkDevice_(ptr))->num_axes)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_num_keys(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GdkDevice__P(ptr), ptr, XEN_ONLY_ARG, "num_keys", "GdkDevice");
  return(C_TO_XEN_gint((gint)((XEN_TO_C_GdkDevice_(ptr))->num_keys)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_keys(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GdkDevice__P(ptr) || XEN_GtkStatusbar__P(ptr), ptr, XEN_ONLY_ARG, "keys", "GdkDevice" " or " "GtkStatusbar");
  if (XEN_GdkDevice__P(ptr)) return(C_TO_XEN_GdkDeviceKey_((GdkDeviceKey*)((XEN_TO_C_GdkDevice_(ptr))->keys)));
  return(C_TO_XEN_GSList_((GSList*)((XEN_TO_C_GtkStatusbar_(ptr))->keys)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_keycode(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GdkKeymapKey__P(ptr), ptr, XEN_ONLY_ARG, "keycode", "GdkKeymapKey");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GdkKeymapKey_(ptr))->keycode)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_level(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GdkKeymapKey__P(ptr), ptr, XEN_ONLY_ARG, "level", "GdkKeymapKey");
  return(C_TO_XEN_gint((gint)((XEN_TO_C_GdkKeymapKey_(ptr))->level)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_n_colors(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GdkRgbCmap__P(ptr), ptr, XEN_ONLY_ARG, "n_colors", "GdkRgbCmap");
  return(C_TO_XEN_gint((gint)((XEN_TO_C_GdkRgbCmap_(ptr))->n_colors)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_info_list(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GdkRgbCmap__P(ptr), ptr, XEN_ONLY_ARG, "info_list", "GdkRgbCmap");
  return(C_TO_XEN_GSList_((GSList*)((XEN_TO_C_GdkRgbCmap_(ptr))->info_list)));
}
#endif

static XEN gxg_x1(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GdkSegment__P(ptr), ptr, XEN_ONLY_ARG, "x1", "GdkSegment");
  return(C_TO_XEN_gint((gint)((XEN_TO_C_GdkSegment_(ptr))->x1)));
}

static XEN gxg_y1(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GdkSegment__P(ptr), ptr, XEN_ONLY_ARG, "y1", "GdkSegment");
  return(C_TO_XEN_gint((gint)((XEN_TO_C_GdkSegment_(ptr))->y1)));
}

static XEN gxg_x2(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GdkSegment__P(ptr), ptr, XEN_ONLY_ARG, "x2", "GdkSegment");
  return(C_TO_XEN_gint((gint)((XEN_TO_C_GdkSegment_(ptr))->x2)));
}

static XEN gxg_y2(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GdkSegment__P(ptr), ptr, XEN_ONLY_ARG, "y2", "GdkSegment");
  return(C_TO_XEN_gint((gint)((XEN_TO_C_GdkSegment_(ptr))->y2)));
}

static XEN gxg_colormap_size(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GdkVisual__P(ptr), ptr, XEN_ONLY_ARG, "colormap_size", "GdkVisual");
  return(C_TO_XEN_gint((gint)((XEN_TO_C_GdkVisual_(ptr))->colormap_size)));
}

static XEN gxg_bits_per_rgb(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GdkVisual__P(ptr), ptr, XEN_ONLY_ARG, "bits_per_rgb", "GdkVisual");
  return(C_TO_XEN_gint((gint)((XEN_TO_C_GdkVisual_(ptr))->bits_per_rgb)));
}

static XEN gxg_red_mask(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GdkVisual__P(ptr), ptr, XEN_ONLY_ARG, "red_mask", "GdkVisual");
  return(C_TO_XEN_guint32((guint32)((XEN_TO_C_GdkVisual_(ptr))->red_mask)));
}

static XEN gxg_red_shift(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GdkVisual__P(ptr), ptr, XEN_ONLY_ARG, "red_shift", "GdkVisual");
  return(C_TO_XEN_gint((gint)((XEN_TO_C_GdkVisual_(ptr))->red_shift)));
}

static XEN gxg_red_prec(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GdkVisual__P(ptr), ptr, XEN_ONLY_ARG, "red_prec", "GdkVisual");
  return(C_TO_XEN_gint((gint)((XEN_TO_C_GdkVisual_(ptr))->red_prec)));
}

static XEN gxg_green_mask(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GdkVisual__P(ptr), ptr, XEN_ONLY_ARG, "green_mask", "GdkVisual");
  return(C_TO_XEN_guint32((guint32)((XEN_TO_C_GdkVisual_(ptr))->green_mask)));
}

static XEN gxg_green_shift(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GdkVisual__P(ptr), ptr, XEN_ONLY_ARG, "green_shift", "GdkVisual");
  return(C_TO_XEN_gint((gint)((XEN_TO_C_GdkVisual_(ptr))->green_shift)));
}

static XEN gxg_green_prec(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GdkVisual__P(ptr), ptr, XEN_ONLY_ARG, "green_prec", "GdkVisual");
  return(C_TO_XEN_gint((gint)((XEN_TO_C_GdkVisual_(ptr))->green_prec)));
}

static XEN gxg_blue_mask(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GdkVisual__P(ptr), ptr, XEN_ONLY_ARG, "blue_mask", "GdkVisual");
  return(C_TO_XEN_guint32((guint32)((XEN_TO_C_GdkVisual_(ptr))->blue_mask)));
}

static XEN gxg_blue_shift(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GdkVisual__P(ptr), ptr, XEN_ONLY_ARG, "blue_shift", "GdkVisual");
  return(C_TO_XEN_gint((gint)((XEN_TO_C_GdkVisual_(ptr))->blue_shift)));
}

static XEN gxg_blue_prec(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GdkVisual__P(ptr), ptr, XEN_ONLY_ARG, "blue_prec", "GdkVisual");
  return(C_TO_XEN_gint((gint)((XEN_TO_C_GdkVisual_(ptr))->blue_prec)));
}

#if (!HAVE_GTK_1)
static XEN gxg_title(XEN ptr)
{
  if (XEN_GdkWindowAttr__P(ptr)) return(C_TO_XEN_gchar_((gchar*)((XEN_TO_C_GdkWindowAttr_(ptr))->title)));
  if (XEN_GtkTreeViewColumn__P(ptr)) return(C_TO_XEN_gchar_((gchar*)((XEN_TO_C_GtkTreeViewColumn_(ptr))->title)));
  if (XEN_GtkWindow__P(ptr)) return(C_TO_XEN_gchar_((gchar*)((XEN_TO_C_GtkWindow_(ptr))->title)));
  XEN_ASSERT_TYPE(0, ptr, XEN_ONLY_ARG, "title", "pointer to struct with title field");
}
#endif

static XEN gxg_event_mask(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GdkWindowAttr__P(ptr), ptr, XEN_ONLY_ARG, "event_mask", "GdkWindowAttr");
  return(C_TO_XEN_gint((gint)((XEN_TO_C_GdkWindowAttr_(ptr))->event_mask)));
}

#if (!HAVE_GTK_1)
static XEN gxg_window_type(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GdkWindowAttr__P(ptr) || XEN_GdkWindowObject__P(ptr), ptr, XEN_ONLY_ARG, "window_type", "GdkWindowAttr" " or " "GdkWindowObject");
  if (XEN_GdkWindowAttr__P(ptr)) return(C_TO_XEN_GdkWindowType((GdkWindowType)((XEN_TO_C_GdkWindowAttr_(ptr))->window_type)));
  return(C_TO_XEN_guint8((guint8)((XEN_TO_C_GdkWindowObject_(ptr))->window_type)));
}
#endif

static XEN gxg_cursor(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GdkWindowAttr__P(ptr), ptr, XEN_ONLY_ARG, "cursor", "GdkWindowAttr");
  return(C_TO_XEN_GdkCursor_((GdkCursor*)((XEN_TO_C_GdkWindowAttr_(ptr))->cursor)));
}

static XEN gxg_wmclass_name(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GdkWindowAttr__P(ptr) || XEN_GtkWindow__P(ptr), ptr, XEN_ONLY_ARG, "wmclass_name", "GdkWindowAttr" " or " "GtkWindow");
  if (XEN_GdkWindowAttr__P(ptr)) return(C_TO_XEN_gchar_((gchar*)((XEN_TO_C_GdkWindowAttr_(ptr))->wmclass_name)));
  return(C_TO_XEN_gchar_((gchar*)((XEN_TO_C_GtkWindow_(ptr))->wmclass_name)));
}

static XEN gxg_wmclass_class(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GdkWindowAttr__P(ptr) || XEN_GtkWindow__P(ptr), ptr, XEN_ONLY_ARG, "wmclass_class", "GdkWindowAttr" " or " "GtkWindow");
  if (XEN_GdkWindowAttr__P(ptr)) return(C_TO_XEN_gchar_((gchar*)((XEN_TO_C_GdkWindowAttr_(ptr))->wmclass_class)));
  return(C_TO_XEN_gchar_((gchar*)((XEN_TO_C_GtkWindow_(ptr))->wmclass_class)));
}

static XEN gxg_override_redirect(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GdkWindowAttr__P(ptr), ptr, XEN_ONLY_ARG, "override_redirect", "GdkWindowAttr");
  return(C_TO_XEN_gboolean((gboolean)((XEN_TO_C_GdkWindowAttr_(ptr))->override_redirect)));
}

#if (!HAVE_GTK_1)
static XEN gxg_min_width(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GdkGeometry__P(ptr) || XEN_GtkTreeViewColumn__P(ptr), ptr, XEN_ONLY_ARG, "min_width", "GdkGeometry" " or " "GtkTreeViewColumn");
  if (XEN_GdkGeometry__P(ptr)) return(C_TO_XEN_gint((gint)((XEN_TO_C_GdkGeometry_(ptr))->min_width)));
  return(C_TO_XEN_gint((gint)((XEN_TO_C_GtkTreeViewColumn_(ptr))->min_width)));
}
#endif

static XEN gxg_min_height(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GdkGeometry__P(ptr), ptr, XEN_ONLY_ARG, "min_height", "GdkGeometry");
  return(C_TO_XEN_gint((gint)((XEN_TO_C_GdkGeometry_(ptr))->min_height)));
}

#if (!HAVE_GTK_1)
static XEN gxg_max_width(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GdkGeometry__P(ptr) || XEN_GtkTreeViewColumn__P(ptr), ptr, XEN_ONLY_ARG, "max_width", "GdkGeometry" " or " "GtkTreeViewColumn");
  if (XEN_GdkGeometry__P(ptr)) return(C_TO_XEN_gint((gint)((XEN_TO_C_GdkGeometry_(ptr))->max_width)));
  return(C_TO_XEN_gint((gint)((XEN_TO_C_GtkTreeViewColumn_(ptr))->max_width)));
}
#endif

static XEN gxg_max_height(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GdkGeometry__P(ptr), ptr, XEN_ONLY_ARG, "max_height", "GdkGeometry");
  return(C_TO_XEN_gint((gint)((XEN_TO_C_GdkGeometry_(ptr))->max_height)));
}

static XEN gxg_base_width(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GdkGeometry__P(ptr), ptr, XEN_ONLY_ARG, "base_width", "GdkGeometry");
  return(C_TO_XEN_gint((gint)((XEN_TO_C_GdkGeometry_(ptr))->base_width)));
}

static XEN gxg_base_height(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GdkGeometry__P(ptr), ptr, XEN_ONLY_ARG, "base_height", "GdkGeometry");
  return(C_TO_XEN_gint((gint)((XEN_TO_C_GdkGeometry_(ptr))->base_height)));
}

static XEN gxg_width_inc(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GdkGeometry__P(ptr), ptr, XEN_ONLY_ARG, "width_inc", "GdkGeometry");
  return(C_TO_XEN_gint((gint)((XEN_TO_C_GdkGeometry_(ptr))->width_inc)));
}

static XEN gxg_height_inc(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GdkGeometry__P(ptr), ptr, XEN_ONLY_ARG, "height_inc", "GdkGeometry");
  return(C_TO_XEN_gint((gint)((XEN_TO_C_GdkGeometry_(ptr))->height_inc)));
}

static XEN gxg_min_aspect(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GdkGeometry__P(ptr), ptr, XEN_ONLY_ARG, "min_aspect", "GdkGeometry");
  return(C_TO_XEN_gdouble((gdouble)((XEN_TO_C_GdkGeometry_(ptr))->min_aspect)));
}

static XEN gxg_max_aspect(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GdkGeometry__P(ptr), ptr, XEN_ONLY_ARG, "max_aspect", "GdkGeometry");
  return(C_TO_XEN_gdouble((gdouble)((XEN_TO_C_GdkGeometry_(ptr))->max_aspect)));
}

#if (!HAVE_GTK_1)
static XEN gxg_win_gravity(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GdkGeometry__P(ptr), ptr, XEN_ONLY_ARG, "win_gravity", "GdkGeometry");
  return(C_TO_XEN_GdkGravity((GdkGravity)((XEN_TO_C_GdkGeometry_(ptr))->win_gravity)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_impl(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GdkWindowObject__P(ptr), ptr, XEN_ONLY_ARG, "impl", "GdkWindowObject");
  return(C_TO_XEN_GdkDrawable_((GdkDrawable*)((XEN_TO_C_GdkWindowObject_(ptr))->impl)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_parent(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GdkWindowObject__P(ptr) || XEN_GtkWidget__P(ptr), ptr, XEN_ONLY_ARG, "parent", "GdkWindowObject" " or " "GtkWidget");
  if (XEN_GdkWindowObject__P(ptr)) return(C_TO_XEN_GdkWindowObject_((GdkWindowObject*)((XEN_TO_C_GdkWindowObject_(ptr))->parent)));
  return(C_TO_XEN_GtkWidget_((GtkWidget*)((XEN_TO_C_GtkWidget_(ptr))->parent)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_user_data(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GdkWindowObject__P(ptr) || XEN_GtkTreeSelection__P(ptr), ptr, XEN_ONLY_ARG, "user_data", "GdkWindowObject" " or " "GtkTreeSelection");
  if (XEN_GdkWindowObject__P(ptr)) return(C_TO_XEN_gpointer((gpointer)((XEN_TO_C_GdkWindowObject_(ptr))->user_data)));
  return(C_TO_XEN_gpointer((gpointer)((XEN_TO_C_GtkTreeSelection_(ptr))->user_data)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_extension_events(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GdkWindowObject__P(ptr), ptr, XEN_ONLY_ARG, "extension_events", "GdkWindowObject");
  return(C_TO_XEN_gint((gint)((XEN_TO_C_GdkWindowObject_(ptr))->extension_events)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_filters(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GdkWindowObject__P(ptr), ptr, XEN_ONLY_ARG, "filters", "GdkWindowObject");
  return(C_TO_XEN_GList_((GList*)((XEN_TO_C_GdkWindowObject_(ptr))->filters)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_children(XEN ptr)
{
  if (XEN_GdkWindowObject__P(ptr)) return(C_TO_XEN_GList_((GList*)((XEN_TO_C_GdkWindowObject_(ptr))->children)));
  if (XEN_GtkBox__P(ptr)) return(C_TO_XEN_GList_((GList*)((XEN_TO_C_GtkBox_(ptr))->children)));
  if (XEN_GtkFixed__P(ptr)) return(C_TO_XEN_GList_((GList*)((XEN_TO_C_GtkFixed_(ptr))->children)));
  if (XEN_GtkLayout__P(ptr)) return(C_TO_XEN_GList_((GList*)((XEN_TO_C_GtkLayout_(ptr))->children)));
  if (XEN_GtkMenuShell__P(ptr)) return(C_TO_XEN_GList_((GList*)((XEN_TO_C_GtkMenuShell_(ptr))->children)));
  if (XEN_GtkNotebook__P(ptr)) return(C_TO_XEN_GList_((GList*)((XEN_TO_C_GtkNotebook_(ptr))->children)));
  if (XEN_GtkTable__P(ptr)) return(C_TO_XEN_GList_((GList*)((XEN_TO_C_GtkTable_(ptr))->children)));
  if (XEN_GtkToolbar__P(ptr)) return(C_TO_XEN_GList_((GList*)((XEN_TO_C_GtkToolbar_(ptr))->children)));
  XEN_ASSERT_TYPE(0, ptr, XEN_ONLY_ARG, "children", "pointer to struct with children field");
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_bg_pixmap(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GdkWindowObject__P(ptr) || XEN_GtkStyle__P(ptr), ptr, XEN_ONLY_ARG, "bg_pixmap", "GdkWindowObject" " or " "GtkStyle");
  if (XEN_GdkWindowObject__P(ptr)) return(C_TO_XEN_GdkPixmap_((GdkPixmap*)((XEN_TO_C_GdkWindowObject_(ptr))->bg_pixmap)));
  return(C_TO_XEN_GdkPixmap__((GdkPixmap**)((XEN_TO_C_GtkStyle_(ptr))->bg_pixmap)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_paint_stack(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GdkWindowObject__P(ptr), ptr, XEN_ONLY_ARG, "paint_stack", "GdkWindowObject");
  return(C_TO_XEN_GSList_((GSList*)((XEN_TO_C_GdkWindowObject_(ptr))->paint_stack)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_update_area(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GdkWindowObject__P(ptr), ptr, XEN_ONLY_ARG, "update_area", "GdkWindowObject");
  return(C_TO_XEN_GdkRegion_((GdkRegion*)((XEN_TO_C_GdkWindowObject_(ptr))->update_area)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_update_freeze_count(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GdkWindowObject__P(ptr), ptr, XEN_ONLY_ARG, "update_freeze_count", "GdkWindowObject");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GdkWindowObject_(ptr))->update_freeze_count)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_resize_count(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GdkWindowObject__P(ptr), ptr, XEN_ONLY_ARG, "resize_count", "GdkWindowObject");
  return(C_TO_XEN_guint8((guint8)((XEN_TO_C_GdkWindowObject_(ptr))->resize_count)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_guffaw_gravity(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GdkWindowObject__P(ptr), ptr, XEN_ONLY_ARG, "guffaw_gravity", "GdkWindowObject");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GdkWindowObject_(ptr))->guffaw_gravity)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_input_only(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GdkWindowObject__P(ptr), ptr, XEN_ONLY_ARG, "input_only", "GdkWindowObject");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GdkWindowObject_(ptr))->input_only)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_modal_hint(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GdkWindowObject__P(ptr), ptr, XEN_ONLY_ARG, "modal_hint", "GdkWindowObject");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GdkWindowObject_(ptr))->modal_hint)));
}
#endif

static XEN gxg_lock_count(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkAccelGroup__P(ptr), ptr, XEN_ONLY_ARG, "lock_count", "GtkAccelGroup");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkAccelGroup_(ptr))->lock_count)));
}

static XEN gxg_modifier_mask(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkAccelGroup__P(ptr), ptr, XEN_ONLY_ARG, "modifier_mask", "GtkAccelGroup");
  return(C_TO_XEN_GdkModifierType((GdkModifierType)((XEN_TO_C_GtkAccelGroup_(ptr))->modifier_mask)));
}

#if (!HAVE_GTK_1)
static XEN gxg_acceleratables(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkAccelGroup__P(ptr), ptr, XEN_ONLY_ARG, "acceleratables", "GtkAccelGroup");
  return(C_TO_XEN_GSList_((GSList*)((XEN_TO_C_GtkAccelGroup_(ptr))->acceleratables)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_n_accels(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkAccelGroup__P(ptr), ptr, XEN_ONLY_ARG, "n_accels", "GtkAccelGroup");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkAccelGroup_(ptr))->n_accels)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_priv_accels(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkAccelGroup__P(ptr), ptr, XEN_ONLY_ARG, "priv_accels", "GtkAccelGroup");
  return(C_TO_XEN_GtkAccelGroupEntry_((GtkAccelGroupEntry*)((XEN_TO_C_GtkAccelGroup_(ptr))->priv_accels)));
}
#endif

static XEN gxg_queue_id(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkAccelLabel__P(ptr), ptr, XEN_ONLY_ARG, "queue_id", "GtkAccelLabel");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkAccelLabel_(ptr))->queue_id)));
}

static XEN gxg_accel_padding(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkAccelLabel__P(ptr), ptr, XEN_ONLY_ARG, "accel_padding", "GtkAccelLabel");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkAccelLabel_(ptr))->accel_padding)));
}

static XEN gxg_accel_widget(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkAccelLabel__P(ptr), ptr, XEN_ONLY_ARG, "accel_widget", "GtkAccelLabel");
  return(C_TO_XEN_GtkWidget_((GtkWidget*)((XEN_TO_C_GtkAccelLabel_(ptr))->accel_widget)));
}

#if (!HAVE_GTK_1)
static XEN gxg_accel_closure(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkAccelLabel__P(ptr), ptr, XEN_ONLY_ARG, "accel_closure", "GtkAccelLabel");
  return(C_TO_XEN_GClosure_((GClosure*)((XEN_TO_C_GtkAccelLabel_(ptr))->accel_closure)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_accel_group(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkAccelLabel__P(ptr) || XEN_GtkMenu__P(ptr), ptr, XEN_ONLY_ARG, "accel_group", "GtkAccelLabel" " or " "GtkMenu");
  if (XEN_GtkAccelLabel__P(ptr)) return(C_TO_XEN_GtkAccelGroup_((GtkAccelGroup*)((XEN_TO_C_GtkAccelLabel_(ptr))->accel_group)));
  return(C_TO_XEN_GtkAccelGroup_((GtkAccelGroup*)((XEN_TO_C_GtkMenu_(ptr))->accel_group)));
}
#endif

static XEN gxg_accel_string(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkAccelLabel__P(ptr), ptr, XEN_ONLY_ARG, "accel_string", "GtkAccelLabel");
  return(C_TO_XEN_gchar_((gchar*)((XEN_TO_C_GtkAccelLabel_(ptr))->accel_string)));
}

static XEN gxg_accel_string_width(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkAccelLabel__P(ptr), ptr, XEN_ONLY_ARG, "accel_string_width", "GtkAccelLabel");
  return(C_TO_XEN_guint16((guint16)((XEN_TO_C_GtkAccelLabel_(ptr))->accel_string_width)));
}

static XEN gxg_lower(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkAdjustment__P(ptr) || XEN_GtkRuler__P(ptr), ptr, XEN_ONLY_ARG, "lower", "GtkAdjustment" " or " "GtkRuler");
  if (XEN_GtkAdjustment__P(ptr)) return(C_TO_XEN_gdouble((gdouble)((XEN_TO_C_GtkAdjustment_(ptr))->lower)));
  return(C_TO_XEN_gdouble((gdouble)((XEN_TO_C_GtkRuler_(ptr))->lower)));
}

static XEN gxg_upper(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkAdjustment__P(ptr) || XEN_GtkRuler__P(ptr), ptr, XEN_ONLY_ARG, "upper", "GtkAdjustment" " or " "GtkRuler");
  if (XEN_GtkAdjustment__P(ptr)) return(C_TO_XEN_gdouble((gdouble)((XEN_TO_C_GtkAdjustment_(ptr))->upper)));
  return(C_TO_XEN_gdouble((gdouble)((XEN_TO_C_GtkRuler_(ptr))->upper)));
}

static XEN gxg_value(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkAdjustment__P(ptr), ptr, XEN_ONLY_ARG, "value", "GtkAdjustment");
  return(C_TO_XEN_gdouble((gdouble)((XEN_TO_C_GtkAdjustment_(ptr))->value)));
}

static XEN gxg_step_increment(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkAdjustment__P(ptr), ptr, XEN_ONLY_ARG, "step_increment", "GtkAdjustment");
  return(C_TO_XEN_gdouble((gdouble)((XEN_TO_C_GtkAdjustment_(ptr))->step_increment)));
}

static XEN gxg_page_increment(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkAdjustment__P(ptr), ptr, XEN_ONLY_ARG, "page_increment", "GtkAdjustment");
  return(C_TO_XEN_gdouble((gdouble)((XEN_TO_C_GtkAdjustment_(ptr))->page_increment)));
}

static XEN gxg_page_size(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkAdjustment__P(ptr), ptr, XEN_ONLY_ARG, "page_size", "GtkAdjustment");
  return(C_TO_XEN_gdouble((gdouble)((XEN_TO_C_GtkAdjustment_(ptr))->page_size)));
}

#if (!HAVE_GTK_1)
static XEN gxg_xalign(XEN ptr)
{
  if (XEN_GtkAlignment__P(ptr)) return(C_TO_XEN_gfloat((gfloat)((XEN_TO_C_GtkAlignment_(ptr))->xalign)));
  if (XEN_GtkAspectFrame__P(ptr)) return(C_TO_XEN_gfloat((gfloat)((XEN_TO_C_GtkAspectFrame_(ptr))->xalign)));
  if (XEN_GtkCellRenderer__P(ptr)) return(C_TO_XEN_gfloat((gfloat)((XEN_TO_C_GtkCellRenderer_(ptr))->xalign)));
  if (XEN_GtkMisc__P(ptr)) return(C_TO_XEN_gfloat((gfloat)((XEN_TO_C_GtkMisc_(ptr))->xalign)));
  if (XEN_GtkTreeViewColumn__P(ptr)) return(C_TO_XEN_gfloat((gfloat)((XEN_TO_C_GtkTreeViewColumn_(ptr))->xalign)));
  XEN_ASSERT_TYPE(0, ptr, XEN_ONLY_ARG, "xalign", "pointer to struct with xalign field");
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_yalign(XEN ptr)
{
  if (XEN_GtkAlignment__P(ptr)) return(C_TO_XEN_gfloat((gfloat)((XEN_TO_C_GtkAlignment_(ptr))->yalign)));
  if (XEN_GtkAspectFrame__P(ptr)) return(C_TO_XEN_gfloat((gfloat)((XEN_TO_C_GtkAspectFrame_(ptr))->yalign)));
  if (XEN_GtkCellRenderer__P(ptr)) return(C_TO_XEN_gfloat((gfloat)((XEN_TO_C_GtkCellRenderer_(ptr))->yalign)));
  if (XEN_GtkMisc__P(ptr)) return(C_TO_XEN_gfloat((gfloat)((XEN_TO_C_GtkMisc_(ptr))->yalign)));
  XEN_ASSERT_TYPE(0, ptr, XEN_ONLY_ARG, "yalign", "pointer to struct with yalign field");
}
#endif

static XEN gxg_xscale(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkAlignment__P(ptr), ptr, XEN_ONLY_ARG, "xscale", "GtkAlignment");
  return(C_TO_XEN_gfloat((gfloat)((XEN_TO_C_GtkAlignment_(ptr))->xscale)));
}

static XEN gxg_yscale(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkAlignment__P(ptr), ptr, XEN_ONLY_ARG, "yscale", "GtkAlignment");
  return(C_TO_XEN_gfloat((gfloat)((XEN_TO_C_GtkAlignment_(ptr))->yscale)));
}

static XEN gxg_ratio(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkAspectFrame__P(ptr), ptr, XEN_ONLY_ARG, "ratio", "GtkAspectFrame");
  return(C_TO_XEN_gfloat((gfloat)((XEN_TO_C_GtkAspectFrame_(ptr))->ratio)));
}

static XEN gxg_obey_child(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkAspectFrame__P(ptr), ptr, XEN_ONLY_ARG, "obey_child", "GtkAspectFrame");
  return(C_TO_XEN_gboolean((gboolean)((XEN_TO_C_GtkAspectFrame_(ptr))->obey_child)));
}

static XEN gxg_arrow_type(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkArrow__P(ptr), ptr, XEN_ONLY_ARG, "arrow_type", "GtkArrow");
  return(C_TO_XEN_gint16((gint16)((XEN_TO_C_GtkArrow_(ptr))->arrow_type)));
}

#if (!HAVE_GTK_1)
static XEN gxg_shadow_type(XEN ptr)
{
  if (XEN_GtkArrow__P(ptr)) return(C_TO_XEN_gint16((gint16)((XEN_TO_C_GtkArrow_(ptr))->shadow_type)));
  if (XEN_GtkFrame__P(ptr)) return(C_TO_XEN_gint16((gint16)((XEN_TO_C_GtkFrame_(ptr))->shadow_type)));
  if (XEN_GtkHandleBox__P(ptr)) return(C_TO_XEN_GtkShadowType((GtkShadowType)((XEN_TO_C_GtkHandleBox_(ptr))->shadow_type)));
  if (XEN_GtkScrolledWindow__P(ptr)) return(C_TO_XEN_guint16((guint16)((XEN_TO_C_GtkScrolledWindow_(ptr))->shadow_type)));
  if (XEN_GtkViewport__P(ptr)) return(C_TO_XEN_GtkShadowType((GtkShadowType)((XEN_TO_C_GtkViewport_(ptr))->shadow_type)));
  XEN_ASSERT_TYPE(0, ptr, XEN_ONLY_ARG, "shadow_type", "pointer to struct with shadow_type field");
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_child(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkBin__P(ptr) || XEN_GtkTreeViewColumn__P(ptr), ptr, XEN_ONLY_ARG, "child", "GtkBin" " or " "GtkTreeViewColumn");
  if (XEN_GtkBin__P(ptr)) return(C_TO_XEN_GtkWidget_((GtkWidget*)((XEN_TO_C_GtkBin_(ptr))->child)));
  return(C_TO_XEN_GtkWidget_((GtkWidget*)((XEN_TO_C_GtkTreeViewColumn_(ptr))->child)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_spacing(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkBox__P(ptr) || XEN_GtkTreeViewColumn__P(ptr), ptr, XEN_ONLY_ARG, "spacing", "GtkBox" " or " "GtkTreeViewColumn");
  if (XEN_GtkBox__P(ptr)) return(C_TO_XEN_gint16((gint16)((XEN_TO_C_GtkBox_(ptr))->spacing)));
  return(C_TO_XEN_gint((gint)((XEN_TO_C_GtkTreeViewColumn_(ptr))->spacing)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_event_window(XEN ptr)
{
  if (XEN_GtkButton__P(ptr)) return(C_TO_XEN_GdkWindow_((GdkWindow*)((XEN_TO_C_GtkButton_(ptr))->event_window)));
  if (XEN_GtkMenuItem__P(ptr)) return(C_TO_XEN_GdkWindow_((GdkWindow*)((XEN_TO_C_GtkMenuItem_(ptr))->event_window)));
  if (XEN_GtkNotebook__P(ptr)) return(C_TO_XEN_GdkWindow_((GdkWindow*)((XEN_TO_C_GtkNotebook_(ptr))->event_window)));
  if (XEN_GtkRange__P(ptr)) return(C_TO_XEN_GdkWindow_((GdkWindow*)((XEN_TO_C_GtkRange_(ptr))->event_window)));
  XEN_ASSERT_TYPE(0, ptr, XEN_ONLY_ARG, "event_window", "pointer to struct with event_window field");
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_label_text(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkButton__P(ptr), ptr, XEN_ONLY_ARG, "label_text", "GtkButton");
  return(C_TO_XEN_gchar_((gchar*)((XEN_TO_C_GtkButton_(ptr))->label_text)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_activate_timeout(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkButton__P(ptr), ptr, XEN_ONLY_ARG, "activate_timeout", "GtkButton");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkButton_(ptr))->activate_timeout)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_constructed(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkButton__P(ptr), ptr, XEN_ONLY_ARG, "constructed", "GtkButton");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkButton_(ptr))->constructed)));
}
#endif

static XEN gxg_in_button(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkButton__P(ptr), ptr, XEN_ONLY_ARG, "in_button", "GtkButton");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkButton_(ptr))->in_button)));
}

static XEN gxg_button_down(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkButton__P(ptr), ptr, XEN_ONLY_ARG, "button_down", "GtkButton");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkButton_(ptr))->button_down)));
}

static XEN gxg_relief(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkButton__P(ptr), ptr, XEN_ONLY_ARG, "relief", "GtkButton");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkButton_(ptr))->relief)));
}

#if (!HAVE_GTK_1)
static XEN gxg_use_underline(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkButton__P(ptr) || XEN_GtkLabel__P(ptr), ptr, XEN_ONLY_ARG, "use_underline", "GtkButton" " or " "GtkLabel");
  if (XEN_GtkButton__P(ptr)) return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkButton_(ptr))->use_underline)));
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkLabel_(ptr))->use_underline)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_use_stock(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkButton__P(ptr), ptr, XEN_ONLY_ARG, "use_stock", "GtkButton");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkButton_(ptr))->use_stock)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_xpad(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkCellRenderer__P(ptr) || XEN_GtkMisc__P(ptr), ptr, XEN_ONLY_ARG, "xpad", "GtkCellRenderer" " or " "GtkMisc");
  if (XEN_GtkCellRenderer__P(ptr)) return(C_TO_XEN_guint16((guint16)((XEN_TO_C_GtkCellRenderer_(ptr))->xpad)));
  return(C_TO_XEN_guint16((guint16)((XEN_TO_C_GtkMisc_(ptr))->xpad)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_ypad(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkCellRenderer__P(ptr) || XEN_GtkMisc__P(ptr), ptr, XEN_ONLY_ARG, "ypad", "GtkCellRenderer" " or " "GtkMisc");
  if (XEN_GtkCellRenderer__P(ptr)) return(C_TO_XEN_guint16((guint16)((XEN_TO_C_GtkCellRenderer_(ptr))->ypad)));
  return(C_TO_XEN_guint16((guint16)((XEN_TO_C_GtkMisc_(ptr))->ypad)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_visible(XEN ptr)
{
  if (XEN_GtkCellRenderer__P(ptr)) return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkCellRenderer_(ptr))->visible)));
  if (XEN_GtkEntry__P(ptr)) return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkEntry_(ptr))->visible)));
  if (XEN_GtkTreeViewColumn__P(ptr)) return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkTreeViewColumn_(ptr))->visible)));
  XEN_ASSERT_TYPE(0, ptr, XEN_ONLY_ARG, "visible", "pointer to struct with visible field");
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_is_expander(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkCellRenderer__P(ptr), ptr, XEN_ONLY_ARG, "is_expander", "GtkCellRenderer");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkCellRenderer_(ptr))->is_expander)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_text(XEN ptr)
{
  if (XEN_GtkCellRendererText__P(ptr)) return(C_TO_XEN_gchar_((gchar*)((XEN_TO_C_GtkCellRendererText_(ptr))->text)));
  if (XEN_GtkEntry__P(ptr)) return(C_TO_XEN_gchar_((gchar*)((XEN_TO_C_GtkEntry_(ptr))->text)));
  if (XEN_GtkLabel__P(ptr)) return(C_TO_XEN_gchar_((gchar*)((XEN_TO_C_GtkLabel_(ptr))->text)));
  if (XEN_GtkStyle__P(ptr)) return(C_TO_XEN_GdkColor_((GdkColor*)((XEN_TO_C_GtkStyle_(ptr))->text)));
  XEN_ASSERT_TYPE(0, ptr, XEN_ONLY_ARG, "text", "pointer to struct with text field");
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_font(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkCellRendererText__P(ptr), ptr, XEN_ONLY_ARG, "font", "GtkCellRendererText");
  return(C_TO_XEN_PangoFontDescription_((PangoFontDescription*)((XEN_TO_C_GtkCellRendererText_(ptr))->font)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_font_scale(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkCellRendererText__P(ptr), ptr, XEN_ONLY_ARG, "font_scale", "GtkCellRendererText");
  return(C_TO_XEN_gdouble((gdouble)((XEN_TO_C_GtkCellRendererText_(ptr))->font_scale)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_extra_attrs(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkCellRendererText__P(ptr), ptr, XEN_ONLY_ARG, "extra_attrs", "GtkCellRendererText");
  return(C_TO_XEN_PangoAttrList_((PangoAttrList*)((XEN_TO_C_GtkCellRendererText_(ptr))->extra_attrs)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_underline_style(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkCellRendererText__P(ptr), ptr, XEN_ONLY_ARG, "underline_style", "GtkCellRendererText");
  return(C_TO_XEN_PangoUnderline((PangoUnderline)((XEN_TO_C_GtkCellRendererText_(ptr))->underline_style)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_rise(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkCellRendererText__P(ptr), ptr, XEN_ONLY_ARG, "rise", "GtkCellRendererText");
  return(C_TO_XEN_gint((gint)((XEN_TO_C_GtkCellRendererText_(ptr))->rise)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_fixed_height_rows(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkCellRendererText__P(ptr), ptr, XEN_ONLY_ARG, "fixed_height_rows", "GtkCellRendererText");
  return(C_TO_XEN_gint((gint)((XEN_TO_C_GtkCellRendererText_(ptr))->fixed_height_rows)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_strikethrough(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkCellRendererText__P(ptr), ptr, XEN_ONLY_ARG, "strikethrough", "GtkCellRendererText");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkCellRendererText_(ptr))->strikethrough)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_editable(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkCellRendererText__P(ptr) || XEN_GtkEntry__P(ptr), ptr, XEN_ONLY_ARG, "editable", "GtkCellRendererText" " or " "GtkEntry");
  if (XEN_GtkCellRendererText__P(ptr)) return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkCellRendererText_(ptr))->editable)));
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkEntry_(ptr))->editable)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_scale_set(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkCellRendererText__P(ptr), ptr, XEN_ONLY_ARG, "scale_set", "GtkCellRendererText");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkCellRendererText_(ptr))->scale_set)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_foreground_set(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkCellRendererText__P(ptr), ptr, XEN_ONLY_ARG, "foreground_set", "GtkCellRendererText");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkCellRendererText_(ptr))->foreground_set)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_background_set(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkCellRendererText__P(ptr), ptr, XEN_ONLY_ARG, "background_set", "GtkCellRendererText");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkCellRendererText_(ptr))->background_set)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_underline_set(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkCellRendererText__P(ptr), ptr, XEN_ONLY_ARG, "underline_set", "GtkCellRendererText");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkCellRendererText_(ptr))->underline_set)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_rise_set(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkCellRendererText__P(ptr), ptr, XEN_ONLY_ARG, "rise_set", "GtkCellRendererText");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkCellRendererText_(ptr))->rise_set)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_strikethrough_set(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkCellRendererText__P(ptr), ptr, XEN_ONLY_ARG, "strikethrough_set", "GtkCellRendererText");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkCellRendererText_(ptr))->strikethrough_set)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_editable_set(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkCellRendererText__P(ptr), ptr, XEN_ONLY_ARG, "editable_set", "GtkCellRendererText");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkCellRendererText_(ptr))->editable_set)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_active(XEN ptr)
{
  if (XEN_GtkCellRendererToggle__P(ptr)) return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkCellRendererToggle_(ptr))->active)));
  if (XEN_GtkCheckMenuItem__P(ptr)) return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkCheckMenuItem_(ptr))->active)));
  if (XEN_GtkMenuShell__P(ptr)) return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkMenuShell_(ptr))->active)));
  if (XEN_GtkToggleButton__P(ptr)) return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkToggleButton_(ptr))->active)));
  XEN_ASSERT_TYPE(0, ptr, XEN_ONLY_ARG, "active", "pointer to struct with active field");
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_activatable(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkCellRendererToggle__P(ptr), ptr, XEN_ONLY_ARG, "activatable", "GtkCellRendererToggle");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkCellRendererToggle_(ptr))->activatable)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_pixbuf(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkCellRendererPixbuf__P(ptr), ptr, XEN_ONLY_ARG, "pixbuf", "GtkCellRendererPixbuf");
  return(C_TO_XEN_GdkPixbuf_((GdkPixbuf*)((XEN_TO_C_GtkCellRendererPixbuf_(ptr))->pixbuf)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_pixbuf_expander_open(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkCellRendererPixbuf__P(ptr), ptr, XEN_ONLY_ARG, "pixbuf_expander_open", "GtkCellRendererPixbuf");
  return(C_TO_XEN_GdkPixbuf_((GdkPixbuf*)((XEN_TO_C_GtkCellRendererPixbuf_(ptr))->pixbuf_expander_open)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_pixbuf_expander_closed(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkCellRendererPixbuf__P(ptr), ptr, XEN_ONLY_ARG, "pixbuf_expander_closed", "GtkCellRendererPixbuf");
  return(C_TO_XEN_GdkPixbuf_((GdkPixbuf*)((XEN_TO_C_GtkCellRendererPixbuf_(ptr))->pixbuf_expander_closed)));
}
#endif

static XEN gxg_always_show_toggle(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkCheckMenuItem__P(ptr), ptr, XEN_ONLY_ARG, "always_show_toggle", "GtkCheckMenuItem");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkCheckMenuItem_(ptr))->always_show_toggle)));
}

static XEN gxg_colorsel(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkColorSelectionDialog__P(ptr), ptr, XEN_ONLY_ARG, "colorsel", "GtkColorSelectionDialog");
  return(C_TO_XEN_GtkWidget_((GtkWidget*)((XEN_TO_C_GtkColorSelectionDialog_(ptr))->colorsel)));
}

static XEN gxg_ok_button(XEN ptr)
{
  if (XEN_GtkColorSelectionDialog__P(ptr)) return(C_TO_XEN_GtkWidget_((GtkWidget*)((XEN_TO_C_GtkColorSelectionDialog_(ptr))->ok_button)));
  if (XEN_GtkFileSelection__P(ptr)) return(C_TO_XEN_GtkWidget_((GtkWidget*)((XEN_TO_C_GtkFileSelection_(ptr))->ok_button)));
  if (XEN_GtkFontSelectionDialog__P(ptr)) return(C_TO_XEN_GtkWidget_((GtkWidget*)((XEN_TO_C_GtkFontSelectionDialog_(ptr))->ok_button)));
  XEN_ASSERT_TYPE(0, ptr, XEN_ONLY_ARG, "ok_button", "pointer to struct with ok_button field");
}

static XEN gxg_cancel_button(XEN ptr)
{
  if (XEN_GtkColorSelectionDialog__P(ptr)) return(C_TO_XEN_GtkWidget_((GtkWidget*)((XEN_TO_C_GtkColorSelectionDialog_(ptr))->cancel_button)));
  if (XEN_GtkFileSelection__P(ptr)) return(C_TO_XEN_GtkWidget_((GtkWidget*)((XEN_TO_C_GtkFileSelection_(ptr))->cancel_button)));
  if (XEN_GtkFontSelectionDialog__P(ptr)) return(C_TO_XEN_GtkWidget_((GtkWidget*)((XEN_TO_C_GtkFontSelectionDialog_(ptr))->cancel_button)));
  XEN_ASSERT_TYPE(0, ptr, XEN_ONLY_ARG, "cancel_button", "pointer to struct with cancel_button field");
}

static XEN gxg_help_button(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkColorSelectionDialog__P(ptr) || XEN_GtkFileSelection__P(ptr), ptr, XEN_ONLY_ARG, "help_button", "GtkColorSelectionDialog" " or " "GtkFileSelection");
  if (XEN_GtkColorSelectionDialog__P(ptr)) return(C_TO_XEN_GtkWidget_((GtkWidget*)((XEN_TO_C_GtkColorSelectionDialog_(ptr))->help_button)));
  return(C_TO_XEN_GtkWidget_((GtkWidget*)((XEN_TO_C_GtkFileSelection_(ptr))->help_button)));
}

static XEN gxg_entry(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkCombo__P(ptr), ptr, XEN_ONLY_ARG, "entry", "GtkCombo");
  return(C_TO_XEN_GtkWidget_((GtkWidget*)((XEN_TO_C_GtkCombo_(ptr))->entry)));
}

static XEN gxg_popup(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkCombo__P(ptr), ptr, XEN_ONLY_ARG, "popup", "GtkCombo");
  return(C_TO_XEN_GtkWidget_((GtkWidget*)((XEN_TO_C_GtkCombo_(ptr))->popup)));
}

static XEN gxg_popwin(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkCombo__P(ptr), ptr, XEN_ONLY_ARG, "popwin", "GtkCombo");
  return(C_TO_XEN_GtkWidget_((GtkWidget*)((XEN_TO_C_GtkCombo_(ptr))->popwin)));
}

static XEN gxg_list(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkCombo__P(ptr) || XEN_GtkTargetList__P(ptr), ptr, XEN_ONLY_ARG, "list", "GtkCombo" " or " "GtkTargetList");
  if (XEN_GtkCombo__P(ptr)) return(C_TO_XEN_GtkWidget_((GtkWidget*)((XEN_TO_C_GtkCombo_(ptr))->list)));
  return(C_TO_XEN_GList_((GList*)((XEN_TO_C_GtkTargetList_(ptr))->list)));
}

static XEN gxg_entry_change_id(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkCombo__P(ptr), ptr, XEN_ONLY_ARG, "entry_change_id", "GtkCombo");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkCombo_(ptr))->entry_change_id)));
}

static XEN gxg_list_change_id(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkCombo__P(ptr), ptr, XEN_ONLY_ARG, "list_change_id", "GtkCombo");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkCombo_(ptr))->list_change_id)));
}

static XEN gxg_value_in_list(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkCombo__P(ptr), ptr, XEN_ONLY_ARG, "value_in_list", "GtkCombo");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkCombo_(ptr))->value_in_list)));
}

static XEN gxg_ok_if_empty(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkCombo__P(ptr), ptr, XEN_ONLY_ARG, "ok_if_empty", "GtkCombo");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkCombo_(ptr))->ok_if_empty)));
}

static XEN gxg_case_sensitive(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkCombo__P(ptr), ptr, XEN_ONLY_ARG, "case_sensitive", "GtkCombo");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkCombo_(ptr))->case_sensitive)));
}

static XEN gxg_use_arrows(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkCombo__P(ptr), ptr, XEN_ONLY_ARG, "use_arrows", "GtkCombo");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkCombo_(ptr))->use_arrows)));
}

static XEN gxg_use_arrows_always(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkCombo__P(ptr), ptr, XEN_ONLY_ARG, "use_arrows_always", "GtkCombo");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkCombo_(ptr))->use_arrows_always)));
}

static XEN gxg_current_button(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkCombo__P(ptr), ptr, XEN_ONLY_ARG, "current_button", "GtkCombo");
  return(C_TO_XEN_guint16((guint16)((XEN_TO_C_GtkCombo_(ptr))->current_button)));
}

static XEN gxg_activate_id(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkCombo__P(ptr), ptr, XEN_ONLY_ARG, "activate_id", "GtkCombo");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkCombo_(ptr))->activate_id)));
}

static XEN gxg_focus_child(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkContainer__P(ptr), ptr, XEN_ONLY_ARG, "focus_child", "GtkContainer");
  return(C_TO_XEN_GtkWidget_((GtkWidget*)((XEN_TO_C_GtkContainer_(ptr))->focus_child)));
}

static XEN gxg_border_width(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkContainer__P(ptr), ptr, XEN_ONLY_ARG, "border_width", "GtkContainer");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkContainer_(ptr))->border_width)));
}

static XEN gxg_need_resize(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkContainer__P(ptr), ptr, XEN_ONLY_ARG, "need_resize", "GtkContainer");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkContainer_(ptr))->need_resize)));
}

static XEN gxg_resize_mode(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkContainer__P(ptr), ptr, XEN_ONLY_ARG, "resize_mode", "GtkContainer");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkContainer_(ptr))->resize_mode)));
}

static XEN gxg_reallocate_redraws(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkContainer__P(ptr), ptr, XEN_ONLY_ARG, "reallocate_redraws", "GtkContainer");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkContainer_(ptr))->reallocate_redraws)));
}

#if (!HAVE_GTK_1)
static XEN gxg_cursor_type(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkCurve__P(ptr) || XEN_GtkPaned__P(ptr), ptr, XEN_ONLY_ARG, "cursor_type", "GtkCurve" " or " "GtkPaned");
  if (XEN_GtkCurve__P(ptr)) return(C_TO_XEN_gint((gint)((XEN_TO_C_GtkCurve_(ptr))->cursor_type)));
  return(C_TO_XEN_GdkCursorType((GdkCursorType)((XEN_TO_C_GtkPaned_(ptr))->cursor_type)));
}
#endif

static XEN gxg_min_x(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkCurve__P(ptr), ptr, XEN_ONLY_ARG, "min_x", "GtkCurve");
  return(C_TO_XEN_gfloat((gfloat)((XEN_TO_C_GtkCurve_(ptr))->min_x)));
}

static XEN gxg_max_x(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkCurve__P(ptr), ptr, XEN_ONLY_ARG, "max_x", "GtkCurve");
  return(C_TO_XEN_gfloat((gfloat)((XEN_TO_C_GtkCurve_(ptr))->max_x)));
}

static XEN gxg_min_y(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkCurve__P(ptr), ptr, XEN_ONLY_ARG, "min_y", "GtkCurve");
  return(C_TO_XEN_gfloat((gfloat)((XEN_TO_C_GtkCurve_(ptr))->min_y)));
}

static XEN gxg_max_y(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkCurve__P(ptr), ptr, XEN_ONLY_ARG, "max_y", "GtkCurve");
  return(C_TO_XEN_gfloat((gfloat)((XEN_TO_C_GtkCurve_(ptr))->max_y)));
}

static XEN gxg_pixmap(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkCurve__P(ptr), ptr, XEN_ONLY_ARG, "pixmap", "GtkCurve");
  return(C_TO_XEN_GdkPixmap_((GdkPixmap*)((XEN_TO_C_GtkCurve_(ptr))->pixmap)));
}

static XEN gxg_curve_type(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkCurve__P(ptr), ptr, XEN_ONLY_ARG, "curve_type", "GtkCurve");
  return(C_TO_XEN_GtkCurveType((GtkCurveType)((XEN_TO_C_GtkCurve_(ptr))->curve_type)));
}

static XEN gxg_grab_point(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkCurve__P(ptr), ptr, XEN_ONLY_ARG, "grab_point", "GtkCurve");
  return(C_TO_XEN_gint((gint)((XEN_TO_C_GtkCurve_(ptr))->grab_point)));
}

#if (!HAVE_GTK_1)
static XEN gxg_last(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkCurve__P(ptr) || XEN_GtkTreeStore__P(ptr), ptr, XEN_ONLY_ARG, "last", "GtkCurve" " or " "GtkTreeStore");
  if (XEN_GtkCurve__P(ptr)) return(C_TO_XEN_gint((gint)((XEN_TO_C_GtkCurve_(ptr))->last)));
  return(C_TO_XEN_gpointer((gpointer)((XEN_TO_C_GtkTreeStore_(ptr))->last)));
}
#endif

static XEN gxg_num_points(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkCurve__P(ptr), ptr, XEN_ONLY_ARG, "num_points", "GtkCurve");
  return(C_TO_XEN_gint((gint)((XEN_TO_C_GtkCurve_(ptr))->num_points)));
}

static XEN gxg_point(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkCurve__P(ptr), ptr, XEN_ONLY_ARG, "point", "GtkCurve");
  return(C_TO_XEN_GdkPoint_((GdkPoint*)((XEN_TO_C_GtkCurve_(ptr))->point)));
}

static XEN gxg_num_ctlpoints(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkCurve__P(ptr), ptr, XEN_ONLY_ARG, "num_ctlpoints", "GtkCurve");
  return(C_TO_XEN_gint((gint)((XEN_TO_C_GtkCurve_(ptr))->num_ctlpoints)));
}

static XEN gxg_vbox(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkDialog__P(ptr), ptr, XEN_ONLY_ARG, "vbox", "GtkDialog");
  return(C_TO_XEN_GtkWidget_((GtkWidget*)((XEN_TO_C_GtkDialog_(ptr))->vbox)));
}

static XEN gxg_action_area(XEN ptr)
{
  if (XEN_GtkDialog__P(ptr)) return(C_TO_XEN_GtkWidget_((GtkWidget*)((XEN_TO_C_GtkDialog_(ptr))->action_area)));
  if (XEN_GtkFileSelection__P(ptr)) return(C_TO_XEN_GtkWidget_((GtkWidget*)((XEN_TO_C_GtkFileSelection_(ptr))->action_area)));
  if (XEN_GtkFontSelectionDialog__P(ptr)) return(C_TO_XEN_GtkWidget_((GtkWidget*)((XEN_TO_C_GtkFontSelectionDialog_(ptr))->action_area)));
  XEN_ASSERT_TYPE(0, ptr, XEN_ONLY_ARG, "action_area", "pointer to struct with action_area field");
}

#if (!HAVE_GTK_1)
static XEN gxg_separator(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkDialog__P(ptr), ptr, XEN_ONLY_ARG, "separator", "GtkDialog");
  return(C_TO_XEN_GtkWidget_((GtkWidget*)((XEN_TO_C_GtkDialog_(ptr))->separator)));
}
#endif

static XEN gxg_draw_data(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkDrawingArea__P(ptr), ptr, XEN_ONLY_ARG, "draw_data", "GtkDrawingArea");
  return(C_TO_XEN_gpointer((gpointer)((XEN_TO_C_GtkDrawingArea_(ptr))->draw_data)));
}

#if (!HAVE_GTK_1)
static XEN gxg_overwrite_mode(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkEntry__P(ptr), ptr, XEN_ONLY_ARG, "overwrite_mode", "GtkEntry");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkEntry_(ptr))->overwrite_mode)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_in_drag(XEN ptr)
{
  if (XEN_GtkEntry__P(ptr)) return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkEntry_(ptr))->in_drag)));
  if (XEN_GtkHandleBox__P(ptr)) return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkHandleBox_(ptr))->in_drag)));
  if (XEN_GtkPaned__P(ptr)) return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkPaned_(ptr))->in_drag)));
  XEN_ASSERT_TYPE(0, ptr, XEN_ONLY_ARG, "in_drag", "pointer to struct with in_drag field");
}
#endif

static XEN gxg_text_length(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkEntry__P(ptr), ptr, XEN_ONLY_ARG, "text_length", "GtkEntry");
  return(C_TO_XEN_guint16((guint16)((XEN_TO_C_GtkEntry_(ptr))->text_length)));
}

static XEN gxg_text_max_length(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkEntry__P(ptr), ptr, XEN_ONLY_ARG, "text_max_length", "GtkEntry");
  return(C_TO_XEN_guint16((guint16)((XEN_TO_C_GtkEntry_(ptr))->text_max_length)));
}

static XEN gxg_text_area(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkEntry__P(ptr), ptr, XEN_ONLY_ARG, "text_area", "GtkEntry");
  return(C_TO_XEN_GdkWindow_((GdkWindow*)((XEN_TO_C_GtkEntry_(ptr))->text_area)));
}

#if (!HAVE_GTK_1)
static XEN gxg_im_context(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkEntry__P(ptr), ptr, XEN_ONLY_ARG, "im_context", "GtkEntry");
  return(C_TO_XEN_GtkIMContext_((GtkIMContext*)((XEN_TO_C_GtkEntry_(ptr))->im_context)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_popup_menu(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkEntry__P(ptr), ptr, XEN_ONLY_ARG, "popup_menu", "GtkEntry");
  return(C_TO_XEN_GtkWidget_((GtkWidget*)((XEN_TO_C_GtkEntry_(ptr))->popup_menu)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_current_pos(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkEntry__P(ptr), ptr, XEN_ONLY_ARG, "current_pos", "GtkEntry");
  return(C_TO_XEN_gint((gint)((XEN_TO_C_GtkEntry_(ptr))->current_pos)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_selection_bound(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkEntry__P(ptr), ptr, XEN_ONLY_ARG, "selection_bound", "GtkEntry");
  return(C_TO_XEN_gint((gint)((XEN_TO_C_GtkEntry_(ptr))->selection_bound)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_cached_layout(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkEntry__P(ptr), ptr, XEN_ONLY_ARG, "cached_layout", "GtkEntry");
  return(C_TO_XEN_PangoLayout_((PangoLayout*)((XEN_TO_C_GtkEntry_(ptr))->cached_layout)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_cache_includes_preedit(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkEntry__P(ptr), ptr, XEN_ONLY_ARG, "cache_includes_preedit", "GtkEntry");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkEntry_(ptr))->cache_includes_preedit)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_need_im_reset(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkEntry__P(ptr), ptr, XEN_ONLY_ARG, "need_im_reset", "GtkEntry");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkEntry_(ptr))->need_im_reset)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_has_frame(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkEntry__P(ptr) || XEN_GtkWindow__P(ptr), ptr, XEN_ONLY_ARG, "has_frame", "GtkEntry" " or " "GtkWindow");
  if (XEN_GtkEntry__P(ptr)) return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkEntry_(ptr))->has_frame)));
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkWindow_(ptr))->has_frame)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_activates_default(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkEntry__P(ptr), ptr, XEN_ONLY_ARG, "activates_default", "GtkEntry");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkEntry_(ptr))->activates_default)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_cursor_visible(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkEntry__P(ptr), ptr, XEN_ONLY_ARG, "cursor_visible", "GtkEntry");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkEntry_(ptr))->cursor_visible)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_in_click(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkEntry__P(ptr), ptr, XEN_ONLY_ARG, "in_click", "GtkEntry");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkEntry_(ptr))->in_click)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_is_cell_renderer(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkEntry__P(ptr), ptr, XEN_ONLY_ARG, "is_cell_renderer", "GtkEntry");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkEntry_(ptr))->is_cell_renderer)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_editing_canceled(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkEntry__P(ptr), ptr, XEN_ONLY_ARG, "editing_canceled", "GtkEntry");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkEntry_(ptr))->editing_canceled)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_mouse_cursor_obscured(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkEntry__P(ptr), ptr, XEN_ONLY_ARG, "mouse_cursor_obscured", "GtkEntry");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkEntry_(ptr))->mouse_cursor_obscured)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_blink_timeout(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkEntry__P(ptr), ptr, XEN_ONLY_ARG, "blink_timeout", "GtkEntry");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkEntry_(ptr))->blink_timeout)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_recompute_idle(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkEntry__P(ptr), ptr, XEN_ONLY_ARG, "recompute_idle", "GtkEntry");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkEntry_(ptr))->recompute_idle)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_scroll_offset(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkEntry__P(ptr) || XEN_GtkMenu__P(ptr), ptr, XEN_ONLY_ARG, "scroll_offset", "GtkEntry" " or " "GtkMenu");
  if (XEN_GtkEntry__P(ptr)) return(C_TO_XEN_gint((gint)((XEN_TO_C_GtkEntry_(ptr))->scroll_offset)));
  return(C_TO_XEN_gint((gint)((XEN_TO_C_GtkMenu_(ptr))->scroll_offset)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_ascent(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkEntry__P(ptr), ptr, XEN_ONLY_ARG, "ascent", "GtkEntry");
  return(C_TO_XEN_gint((gint)((XEN_TO_C_GtkEntry_(ptr))->ascent)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_descent(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkEntry__P(ptr), ptr, XEN_ONLY_ARG, "descent", "GtkEntry");
  return(C_TO_XEN_gint((gint)((XEN_TO_C_GtkEntry_(ptr))->descent)));
}
#endif

static XEN gxg_text_size(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkEntry__P(ptr), ptr, XEN_ONLY_ARG, "text_size", "GtkEntry");
  return(C_TO_XEN_guint16((guint16)((XEN_TO_C_GtkEntry_(ptr))->text_size)));
}

#if (!HAVE_GTK_1)
static XEN gxg_n_bytes(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkEntry__P(ptr), ptr, XEN_ONLY_ARG, "n_bytes", "GtkEntry");
  return(C_TO_XEN_guint16((guint16)((XEN_TO_C_GtkEntry_(ptr))->n_bytes)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_preedit_length(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkEntry__P(ptr), ptr, XEN_ONLY_ARG, "preedit_length", "GtkEntry");
  return(C_TO_XEN_guint16((guint16)((XEN_TO_C_GtkEntry_(ptr))->preedit_length)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_preedit_cursor(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkEntry__P(ptr), ptr, XEN_ONLY_ARG, "preedit_cursor", "GtkEntry");
  return(C_TO_XEN_guint16((guint16)((XEN_TO_C_GtkEntry_(ptr))->preedit_cursor)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_dnd_position(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkEntry__P(ptr), ptr, XEN_ONLY_ARG, "dnd_position", "GtkEntry");
  return(C_TO_XEN_gint((gint)((XEN_TO_C_GtkEntry_(ptr))->dnd_position)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_drag_start_x(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkEntry__P(ptr), ptr, XEN_ONLY_ARG, "drag_start_x", "GtkEntry");
  return(C_TO_XEN_gint((gint)((XEN_TO_C_GtkEntry_(ptr))->drag_start_x)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_drag_start_y(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkEntry__P(ptr), ptr, XEN_ONLY_ARG, "drag_start_y", "GtkEntry");
  return(C_TO_XEN_gint((gint)((XEN_TO_C_GtkEntry_(ptr))->drag_start_y)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_invisible_char(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkEntry__P(ptr), ptr, XEN_ONLY_ARG, "invisible_char", "GtkEntry");
  return(C_TO_XEN_gunichar((gunichar)((XEN_TO_C_GtkEntry_(ptr))->invisible_char)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_width_chars(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkEntry__P(ptr), ptr, XEN_ONLY_ARG, "width_chars", "GtkEntry");
  return(C_TO_XEN_gint((gint)((XEN_TO_C_GtkEntry_(ptr))->width_chars)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_cursor_gc(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkEntry__P(ptr), ptr, XEN_ONLY_ARG, "cursor_gc", "GtkEntry");
  return(C_TO_XEN_GdkGC_((GdkGC*)((XEN_TO_C_GtkEntry_(ptr))->cursor_gc)));
}
#endif

static XEN gxg_dir_list(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkFileSelection__P(ptr), ptr, XEN_ONLY_ARG, "dir_list", "GtkFileSelection");
  return(C_TO_XEN_GtkWidget_((GtkWidget*)((XEN_TO_C_GtkFileSelection_(ptr))->dir_list)));
}

static XEN gxg_file_list(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkFileSelection__P(ptr), ptr, XEN_ONLY_ARG, "file_list", "GtkFileSelection");
  return(C_TO_XEN_GtkWidget_((GtkWidget*)((XEN_TO_C_GtkFileSelection_(ptr))->file_list)));
}

static XEN gxg_selection_entry(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkFileSelection__P(ptr), ptr, XEN_ONLY_ARG, "selection_entry", "GtkFileSelection");
  return(C_TO_XEN_GtkWidget_((GtkWidget*)((XEN_TO_C_GtkFileSelection_(ptr))->selection_entry)));
}

static XEN gxg_selection_text(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkFileSelection__P(ptr), ptr, XEN_ONLY_ARG, "selection_text", "GtkFileSelection");
  return(C_TO_XEN_GtkWidget_((GtkWidget*)((XEN_TO_C_GtkFileSelection_(ptr))->selection_text)));
}

static XEN gxg_main_vbox(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkFileSelection__P(ptr) || XEN_GtkFontSelectionDialog__P(ptr), ptr, XEN_ONLY_ARG, "main_vbox", "GtkFileSelection" " or " "GtkFontSelectionDialog");
  if (XEN_GtkFileSelection__P(ptr)) return(C_TO_XEN_GtkWidget_((GtkWidget*)((XEN_TO_C_GtkFileSelection_(ptr))->main_vbox)));
  return(C_TO_XEN_GtkWidget_((GtkWidget*)((XEN_TO_C_GtkFontSelectionDialog_(ptr))->main_vbox)));
}

static XEN gxg_history_pulldown(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkFileSelection__P(ptr), ptr, XEN_ONLY_ARG, "history_pulldown", "GtkFileSelection");
  return(C_TO_XEN_GtkWidget_((GtkWidget*)((XEN_TO_C_GtkFileSelection_(ptr))->history_pulldown)));
}

static XEN gxg_history_menu(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkFileSelection__P(ptr), ptr, XEN_ONLY_ARG, "history_menu", "GtkFileSelection");
  return(C_TO_XEN_GtkWidget_((GtkWidget*)((XEN_TO_C_GtkFileSelection_(ptr))->history_menu)));
}

static XEN gxg_history_list(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkFileSelection__P(ptr), ptr, XEN_ONLY_ARG, "history_list", "GtkFileSelection");
  return(C_TO_XEN_GList_((GList*)((XEN_TO_C_GtkFileSelection_(ptr))->history_list)));
}

static XEN gxg_fileop_dialog(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkFileSelection__P(ptr), ptr, XEN_ONLY_ARG, "fileop_dialog", "GtkFileSelection");
  return(C_TO_XEN_GtkWidget_((GtkWidget*)((XEN_TO_C_GtkFileSelection_(ptr))->fileop_dialog)));
}

static XEN gxg_fileop_entry(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkFileSelection__P(ptr), ptr, XEN_ONLY_ARG, "fileop_entry", "GtkFileSelection");
  return(C_TO_XEN_GtkWidget_((GtkWidget*)((XEN_TO_C_GtkFileSelection_(ptr))->fileop_entry)));
}

static XEN gxg_fileop_file(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkFileSelection__P(ptr), ptr, XEN_ONLY_ARG, "fileop_file", "GtkFileSelection");
  return(C_TO_XEN_gchar_((gchar*)((XEN_TO_C_GtkFileSelection_(ptr))->fileop_file)));
}

static XEN gxg_cmpl_state(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkFileSelection__P(ptr), ptr, XEN_ONLY_ARG, "cmpl_state", "GtkFileSelection");
  return(C_TO_XEN_gpointer((gpointer)((XEN_TO_C_GtkFileSelection_(ptr))->cmpl_state)));
}

static XEN gxg_fileop_c_dir(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkFileSelection__P(ptr), ptr, XEN_ONLY_ARG, "fileop_c_dir", "GtkFileSelection");
  return(C_TO_XEN_GtkWidget_((GtkWidget*)((XEN_TO_C_GtkFileSelection_(ptr))->fileop_c_dir)));
}

static XEN gxg_fileop_del_file(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkFileSelection__P(ptr), ptr, XEN_ONLY_ARG, "fileop_del_file", "GtkFileSelection");
  return(C_TO_XEN_GtkWidget_((GtkWidget*)((XEN_TO_C_GtkFileSelection_(ptr))->fileop_del_file)));
}

static XEN gxg_fileop_ren_file(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkFileSelection__P(ptr), ptr, XEN_ONLY_ARG, "fileop_ren_file", "GtkFileSelection");
  return(C_TO_XEN_GtkWidget_((GtkWidget*)((XEN_TO_C_GtkFileSelection_(ptr))->fileop_ren_file)));
}

static XEN gxg_button_area(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkFileSelection__P(ptr), ptr, XEN_ONLY_ARG, "button_area", "GtkFileSelection");
  return(C_TO_XEN_GtkWidget_((GtkWidget*)((XEN_TO_C_GtkFileSelection_(ptr))->button_area)));
}

static XEN gxg_font_entry(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkFontSelection__P(ptr), ptr, XEN_ONLY_ARG, "font_entry", "GtkFontSelection");
  return(C_TO_XEN_GtkWidget_((GtkWidget*)((XEN_TO_C_GtkFontSelection_(ptr))->font_entry)));
}

#if (!HAVE_GTK_1)
static XEN gxg_family_list(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkFontSelection__P(ptr), ptr, XEN_ONLY_ARG, "family_list", "GtkFontSelection");
  return(C_TO_XEN_GtkWidget_((GtkWidget*)((XEN_TO_C_GtkFontSelection_(ptr))->family_list)));
}
#endif

static XEN gxg_font_style_entry(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkFontSelection__P(ptr), ptr, XEN_ONLY_ARG, "font_style_entry", "GtkFontSelection");
  return(C_TO_XEN_GtkWidget_((GtkWidget*)((XEN_TO_C_GtkFontSelection_(ptr))->font_style_entry)));
}

#if (!HAVE_GTK_1)
static XEN gxg_face_list(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkFontSelection__P(ptr), ptr, XEN_ONLY_ARG, "face_list", "GtkFontSelection");
  return(C_TO_XEN_GtkWidget_((GtkWidget*)((XEN_TO_C_GtkFontSelection_(ptr))->face_list)));
}
#endif

static XEN gxg_size_entry(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkFontSelection__P(ptr), ptr, XEN_ONLY_ARG, "size_entry", "GtkFontSelection");
  return(C_TO_XEN_GtkWidget_((GtkWidget*)((XEN_TO_C_GtkFontSelection_(ptr))->size_entry)));
}

#if (!HAVE_GTK_1)
static XEN gxg_size_list(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkFontSelection__P(ptr), ptr, XEN_ONLY_ARG, "size_list", "GtkFontSelection");
  return(C_TO_XEN_GtkWidget_((GtkWidget*)((XEN_TO_C_GtkFontSelection_(ptr))->size_list)));
}
#endif

static XEN gxg_pixels_button(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkFontSelection__P(ptr), ptr, XEN_ONLY_ARG, "pixels_button", "GtkFontSelection");
  return(C_TO_XEN_GtkWidget_((GtkWidget*)((XEN_TO_C_GtkFontSelection_(ptr))->pixels_button)));
}

static XEN gxg_points_button(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkFontSelection__P(ptr), ptr, XEN_ONLY_ARG, "points_button", "GtkFontSelection");
  return(C_TO_XEN_GtkWidget_((GtkWidget*)((XEN_TO_C_GtkFontSelection_(ptr))->points_button)));
}

static XEN gxg_filter_button(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkFontSelection__P(ptr), ptr, XEN_ONLY_ARG, "filter_button", "GtkFontSelection");
  return(C_TO_XEN_GtkWidget_((GtkWidget*)((XEN_TO_C_GtkFontSelection_(ptr))->filter_button)));
}

static XEN gxg_preview_entry(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkFontSelection__P(ptr), ptr, XEN_ONLY_ARG, "preview_entry", "GtkFontSelection");
  return(C_TO_XEN_GtkWidget_((GtkWidget*)((XEN_TO_C_GtkFontSelection_(ptr))->preview_entry)));
}

static XEN gxg_face(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkFontSelection__P(ptr), ptr, XEN_ONLY_ARG, "face", "GtkFontSelection");
#if (!HAVE_GTK_1)
  return(C_TO_XEN_PangoFontFace_((PangoFontFace*)((XEN_TO_C_GtkFontSelection_(ptr))->face)));
#endif
}

static XEN gxg_fontsel(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkFontSelectionDialog__P(ptr), ptr, XEN_ONLY_ARG, "fontsel", "GtkFontSelectionDialog");
  return(C_TO_XEN_GtkWidget_((GtkWidget*)((XEN_TO_C_GtkFontSelectionDialog_(ptr))->fontsel)));
}

static XEN gxg_apply_button(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkFontSelectionDialog__P(ptr), ptr, XEN_ONLY_ARG, "apply_button", "GtkFontSelectionDialog");
  return(C_TO_XEN_GtkWidget_((GtkWidget*)((XEN_TO_C_GtkFontSelectionDialog_(ptr))->apply_button)));
}

static XEN gxg_dialog_width(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkFontSelectionDialog__P(ptr), ptr, XEN_ONLY_ARG, "dialog_width", "GtkFontSelectionDialog");
  return(C_TO_XEN_gint((gint)((XEN_TO_C_GtkFontSelectionDialog_(ptr))->dialog_width)));
}

static XEN gxg_auto_resize(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkFontSelectionDialog__P(ptr), ptr, XEN_ONLY_ARG, "auto_resize", "GtkFontSelectionDialog");
  return(C_TO_XEN_gboolean((gboolean)((XEN_TO_C_GtkFontSelectionDialog_(ptr))->auto_resize)));
}

#if (!HAVE_GTK_1)
static XEN gxg_label_widget(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkFrame__P(ptr), ptr, XEN_ONLY_ARG, "label_widget", "GtkFrame");
  return(C_TO_XEN_GtkWidget_((GtkWidget*)((XEN_TO_C_GtkFrame_(ptr))->label_widget)));
}
#endif

static XEN gxg_label_xalign(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkFrame__P(ptr), ptr, XEN_ONLY_ARG, "label_xalign", "GtkFrame");
  return(C_TO_XEN_gfloat((gfloat)((XEN_TO_C_GtkFrame_(ptr))->label_xalign)));
}

static XEN gxg_label_yalign(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkFrame__P(ptr), ptr, XEN_ONLY_ARG, "label_yalign", "GtkFrame");
  return(C_TO_XEN_gfloat((gfloat)((XEN_TO_C_GtkFrame_(ptr))->label_yalign)));
}

static XEN gxg_table(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkGammaCurve__P(ptr), ptr, XEN_ONLY_ARG, "table", "GtkGammaCurve");
  return(C_TO_XEN_GtkWidget_((GtkWidget*)((XEN_TO_C_GtkGammaCurve_(ptr))->table)));
}

static XEN gxg_curve(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkGammaCurve__P(ptr), ptr, XEN_ONLY_ARG, "curve", "GtkGammaCurve");
  return(C_TO_XEN_GtkWidget_((GtkWidget*)((XEN_TO_C_GtkGammaCurve_(ptr))->curve)));
}

static XEN gxg_gamma(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkGammaCurve__P(ptr), ptr, XEN_ONLY_ARG, "gamma", "GtkGammaCurve");
  return(C_TO_XEN_gfloat((gfloat)((XEN_TO_C_GtkGammaCurve_(ptr))->gamma)));
}

static XEN gxg_gamma_dialog(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkGammaCurve__P(ptr), ptr, XEN_ONLY_ARG, "gamma_dialog", "GtkGammaCurve");
  return(C_TO_XEN_GtkWidget_((GtkWidget*)((XEN_TO_C_GtkGammaCurve_(ptr))->gamma_dialog)));
}

static XEN gxg_gamma_text(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkGammaCurve__P(ptr), ptr, XEN_ONLY_ARG, "gamma_text", "GtkGammaCurve");
  return(C_TO_XEN_GtkWidget_((GtkWidget*)((XEN_TO_C_GtkGammaCurve_(ptr))->gamma_text)));
}

#if (!HAVE_GTK_1)
static XEN gxg_bin_window(XEN ptr)
{
  if (XEN_GtkHandleBox__P(ptr)) return(C_TO_XEN_GdkWindow_((GdkWindow*)((XEN_TO_C_GtkHandleBox_(ptr))->bin_window)));
  if (XEN_GtkLayout__P(ptr)) return(C_TO_XEN_GdkWindow_((GdkWindow*)((XEN_TO_C_GtkLayout_(ptr))->bin_window)));
  if (XEN_GtkMenu__P(ptr)) return(C_TO_XEN_GdkWindow_((GdkWindow*)((XEN_TO_C_GtkMenu_(ptr))->bin_window)));
  if (XEN_GtkViewport__P(ptr)) return(C_TO_XEN_GdkWindow_((GdkWindow*)((XEN_TO_C_GtkViewport_(ptr))->bin_window)));
  XEN_ASSERT_TYPE(0, ptr, XEN_ONLY_ARG, "bin_window", "pointer to struct with bin_window field");
}
#endif

static XEN gxg_float_window(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkHandleBox__P(ptr), ptr, XEN_ONLY_ARG, "float_window", "GtkHandleBox");
  return(C_TO_XEN_GdkWindow_((GdkWindow*)((XEN_TO_C_GtkHandleBox_(ptr))->float_window)));
}

static XEN gxg_handle_position(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkHandleBox__P(ptr), ptr, XEN_ONLY_ARG, "handle_position", "GtkHandleBox");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkHandleBox_(ptr))->handle_position)));
}

static XEN gxg_float_window_mapped(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkHandleBox__P(ptr), ptr, XEN_ONLY_ARG, "float_window_mapped", "GtkHandleBox");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkHandleBox_(ptr))->float_window_mapped)));
}

static XEN gxg_child_detached(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkHandleBox__P(ptr), ptr, XEN_ONLY_ARG, "child_detached", "GtkHandleBox");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkHandleBox_(ptr))->child_detached)));
}

static XEN gxg_shrink_on_detach(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkHandleBox__P(ptr), ptr, XEN_ONLY_ARG, "shrink_on_detach", "GtkHandleBox");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkHandleBox_(ptr))->shrink_on_detach)));
}

static XEN gxg_snap_edge(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkHandleBox__P(ptr), ptr, XEN_ONLY_ARG, "snap_edge", "GtkHandleBox");
  return(C_TO_XEN_gint((gint)((XEN_TO_C_GtkHandleBox_(ptr))->snap_edge)));
}

static XEN gxg_deskoff_x(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkHandleBox__P(ptr), ptr, XEN_ONLY_ARG, "deskoff_x", "GtkHandleBox");
  return(C_TO_XEN_gint((gint)((XEN_TO_C_GtkHandleBox_(ptr))->deskoff_x)));
}

static XEN gxg_deskoff_y(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkHandleBox__P(ptr), ptr, XEN_ONLY_ARG, "deskoff_y", "GtkHandleBox");
  return(C_TO_XEN_gint((gint)((XEN_TO_C_GtkHandleBox_(ptr))->deskoff_y)));
}

#if (!HAVE_GTK_1)
static XEN gxg_image(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkImageMenuItem__P(ptr) || XEN_GtkMessageDialog__P(ptr), ptr, XEN_ONLY_ARG, "image", "GtkImageMenuItem" " or " "GtkMessageDialog");
  if (XEN_GtkImageMenuItem__P(ptr)) return(C_TO_XEN_GtkWidget_((GtkWidget*)((XEN_TO_C_GtkImageMenuItem_(ptr))->image)));
  return(C_TO_XEN_GtkWidget_((GtkWidget*)((XEN_TO_C_GtkMessageDialog_(ptr))->image)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_slave(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkIMMulticontext__P(ptr), ptr, XEN_ONLY_ARG, "slave", "GtkIMMulticontext");
  return(C_TO_XEN_GtkIMContext_((GtkIMContext*)((XEN_TO_C_GtkIMMulticontext_(ptr))->slave)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_client_window(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkIMMulticontext__P(ptr), ptr, XEN_ONLY_ARG, "client_window", "GtkIMMulticontext");
  return(C_TO_XEN_GdkWindow_((GdkWindow*)((XEN_TO_C_GtkIMMulticontext_(ptr))->client_window)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_context_id(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkIMMulticontext__P(ptr), ptr, XEN_ONLY_ARG, "context_id", "GtkIMMulticontext");
  return(C_TO_XEN_gchar_((gchar*)((XEN_TO_C_GtkIMMulticontext_(ptr))->context_id)));
}
#endif

static XEN gxg_axis_list(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkInputDialog__P(ptr), ptr, XEN_ONLY_ARG, "axis_list", "GtkInputDialog");
  return(C_TO_XEN_GtkWidget_((GtkWidget*)((XEN_TO_C_GtkInputDialog_(ptr))->axis_list)));
}

static XEN gxg_axis_listbox(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkInputDialog__P(ptr), ptr, XEN_ONLY_ARG, "axis_listbox", "GtkInputDialog");
  return(C_TO_XEN_GtkWidget_((GtkWidget*)((XEN_TO_C_GtkInputDialog_(ptr))->axis_listbox)));
}

static XEN gxg_mode_optionmenu(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkInputDialog__P(ptr), ptr, XEN_ONLY_ARG, "mode_optionmenu", "GtkInputDialog");
  return(C_TO_XEN_GtkWidget_((GtkWidget*)((XEN_TO_C_GtkInputDialog_(ptr))->mode_optionmenu)));
}

static XEN gxg_close_button(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkInputDialog__P(ptr), ptr, XEN_ONLY_ARG, "close_button", "GtkInputDialog");
  return(C_TO_XEN_GtkWidget_((GtkWidget*)((XEN_TO_C_GtkInputDialog_(ptr))->close_button)));
}

static XEN gxg_save_button(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkInputDialog__P(ptr), ptr, XEN_ONLY_ARG, "save_button", "GtkInputDialog");
  return(C_TO_XEN_GtkWidget_((GtkWidget*)((XEN_TO_C_GtkInputDialog_(ptr))->save_button)));
}

static XEN gxg_axis_items(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkInputDialog__P(ptr), ptr, XEN_ONLY_ARG, "axis_items", "GtkInputDialog");
  return(C_TO_XEN_GtkWidget__((GtkWidget**)((XEN_TO_C_GtkInputDialog_(ptr))->axis_items)));
}

static XEN gxg_current_device(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkInputDialog__P(ptr), ptr, XEN_ONLY_ARG, "current_device", "GtkInputDialog");
#if (!HAVE_GTK_1)
  return(C_TO_XEN_GdkDevice_((GdkDevice*)((XEN_TO_C_GtkInputDialog_(ptr))->current_device)));
#endif
}

static XEN gxg_keys_list(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkInputDialog__P(ptr), ptr, XEN_ONLY_ARG, "keys_list", "GtkInputDialog");
  return(C_TO_XEN_GtkWidget_((GtkWidget*)((XEN_TO_C_GtkInputDialog_(ptr))->keys_list)));
}

static XEN gxg_keys_listbox(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkInputDialog__P(ptr), ptr, XEN_ONLY_ARG, "keys_listbox", "GtkInputDialog");
  return(C_TO_XEN_GtkWidget_((GtkWidget*)((XEN_TO_C_GtkInputDialog_(ptr))->keys_listbox)));
}

#if (!HAVE_GTK_1)
static XEN gxg_has_user_ref_count(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkInvisible__P(ptr) || XEN_GtkWindow__P(ptr), ptr, XEN_ONLY_ARG, "has_user_ref_count", "GtkInvisible" " or " "GtkWindow");
  if (XEN_GtkInvisible__P(ptr)) return(C_TO_XEN_gboolean((gboolean)((XEN_TO_C_GtkInvisible_(ptr))->has_user_ref_count)));
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkWindow_(ptr))->has_user_ref_count)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_gtk_reserved1(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkInvisible__P(ptr) || XEN_GtkWindow__P(ptr), ptr, XEN_ONLY_ARG, "gtk_reserved1", "GtkInvisible" " or " "GtkWindow");
  if (XEN_GtkInvisible__P(ptr)) return(C_TO_XEN_gpointer((gpointer)((XEN_TO_C_GtkInvisible_(ptr))->gtk_reserved1)));
  return(C_TO_XEN_gpointer((gpointer)((XEN_TO_C_GtkWindow_(ptr))->gtk_reserved1)));
}
#endif

static XEN gxg_path(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkItemFactoryEntry__P(ptr), ptr, XEN_ONLY_ARG, "path", "GtkItemFactoryEntry");
  return(C_TO_XEN_gchar_((gchar*)((XEN_TO_C_GtkItemFactoryEntry_(ptr))->path)));
}

static XEN gxg_accelerator(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkItemFactoryEntry__P(ptr), ptr, XEN_ONLY_ARG, "accelerator", "GtkItemFactoryEntry");
  return(C_TO_XEN_gchar_((gchar*)((XEN_TO_C_GtkItemFactoryEntry_(ptr))->accelerator)));
}

static XEN gxg_callback(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkItemFactoryEntry__P(ptr), ptr, XEN_ONLY_ARG, "callback", "GtkItemFactoryEntry");
  return(C_TO_XEN_GtkItemFactoryCallback((GtkItemFactoryCallback)((XEN_TO_C_GtkItemFactoryEntry_(ptr))->callback)));
}

static XEN gxg_callback_action(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkItemFactoryEntry__P(ptr), ptr, XEN_ONLY_ARG, "callback_action", "GtkItemFactoryEntry");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkItemFactoryEntry_(ptr))->callback_action)));
}

static XEN gxg_item_type(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkItemFactoryEntry__P(ptr), ptr, XEN_ONLY_ARG, "item_type", "GtkItemFactoryEntry");
  return(C_TO_XEN_gchar_((gchar*)((XEN_TO_C_GtkItemFactoryEntry_(ptr))->item_type)));
}

#if (!HAVE_GTK_1)
static XEN gxg_extra_data(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkItemFactoryEntry__P(ptr), ptr, XEN_ONLY_ARG, "extra_data", "GtkItemFactoryEntry");
  return(C_TO_XEN_gpointer((gpointer)((XEN_TO_C_GtkItemFactoryEntry_(ptr))->extra_data)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_label(XEN ptr)
{
  if (XEN_GtkLabel__P(ptr)) return(C_TO_XEN_gchar_((gchar*)((XEN_TO_C_GtkLabel_(ptr))->label)));
  if (XEN_GtkMessageDialog__P(ptr)) return(C_TO_XEN_GtkWidget_((GtkWidget*)((XEN_TO_C_GtkMessageDialog_(ptr))->label)));
  if (XEN_GtkStockItem__P(ptr)) return(C_TO_XEN_gchar_((gchar*)((XEN_TO_C_GtkStockItem_(ptr))->label)));
  if (XEN_GtkStatusbar__P(ptr)) return(C_TO_XEN_GtkWidget_((GtkWidget*)((XEN_TO_C_GtkStatusbar_(ptr))->label)));
  XEN_ASSERT_TYPE(0, ptr, XEN_ONLY_ARG, "label", "pointer to struct with label field");
}
#endif

static XEN gxg_jtype(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkLabel__P(ptr), ptr, XEN_ONLY_ARG, "jtype", "GtkLabel");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkLabel_(ptr))->jtype)));
}

static XEN gxg_wrap(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkLabel__P(ptr) || XEN_GtkSpinButton__P(ptr), ptr, XEN_ONLY_ARG, "wrap", "GtkLabel" " or " "GtkSpinButton");
  if (XEN_GtkLabel__P(ptr)) return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkLabel_(ptr))->wrap)));
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkSpinButton_(ptr))->wrap)));
}

#if (!HAVE_GTK_1)
static XEN gxg_use_markup(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkLabel__P(ptr), ptr, XEN_ONLY_ARG, "use_markup", "GtkLabel");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkLabel_(ptr))->use_markup)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_mnemonic_keyval(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkLabel__P(ptr), ptr, XEN_ONLY_ARG, "mnemonic_keyval", "GtkLabel");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkLabel_(ptr))->mnemonic_keyval)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_attrs(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkLabel__P(ptr), ptr, XEN_ONLY_ARG, "attrs", "GtkLabel");
  return(C_TO_XEN_PangoAttrList_((PangoAttrList*)((XEN_TO_C_GtkLabel_(ptr))->attrs)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_effective_attrs(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkLabel__P(ptr), ptr, XEN_ONLY_ARG, "effective_attrs", "GtkLabel");
  return(C_TO_XEN_PangoAttrList_((PangoAttrList*)((XEN_TO_C_GtkLabel_(ptr))->effective_attrs)));
}
#endif

static XEN gxg_layout(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkLabel__P(ptr) || XEN_GtkRange__P(ptr), ptr, XEN_ONLY_ARG, "layout", "GtkLabel" " or " "GtkRange");
#if (!HAVE_GTK_1)
  if (XEN_GtkLabel__P(ptr)) return(C_TO_XEN_PangoLayout_((PangoLayout*)((XEN_TO_C_GtkLabel_(ptr))->layout)));
#endif
#if (!HAVE_GTK_1)
  return(C_TO_XEN_GtkRangeLayout_((GtkRangeLayout*)((XEN_TO_C_GtkRange_(ptr))->layout)));
#endif
}

#if (!HAVE_GTK_1)
static XEN gxg_mnemonic_widget(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkLabel__P(ptr), ptr, XEN_ONLY_ARG, "mnemonic_widget", "GtkLabel");
  return(C_TO_XEN_GtkWidget_((GtkWidget*)((XEN_TO_C_GtkLabel_(ptr))->mnemonic_widget)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_mnemonic_window(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkLabel__P(ptr), ptr, XEN_ONLY_ARG, "mnemonic_window", "GtkLabel");
  return(C_TO_XEN_GtkWindow_((GtkWindow*)((XEN_TO_C_GtkLabel_(ptr))->mnemonic_window)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_select_info(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkLabel__P(ptr), ptr, XEN_ONLY_ARG, "select_info", "GtkLabel");
  return(C_TO_XEN_GtkLabelSelectionInfo_((GtkLabelSelectionInfo*)((XEN_TO_C_GtkLabel_(ptr))->select_info)));
}
#endif

static XEN gxg_hadjustment(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkLayout__P(ptr) || XEN_GtkViewport__P(ptr), ptr, XEN_ONLY_ARG, "hadjustment", "GtkLayout" " or " "GtkViewport");
  if (XEN_GtkLayout__P(ptr)) return(C_TO_XEN_GtkAdjustment_((GtkAdjustment*)((XEN_TO_C_GtkLayout_(ptr))->hadjustment)));
  return(C_TO_XEN_GtkAdjustment_((GtkAdjustment*)((XEN_TO_C_GtkViewport_(ptr))->hadjustment)));
}

static XEN gxg_vadjustment(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkLayout__P(ptr) || XEN_GtkViewport__P(ptr), ptr, XEN_ONLY_ARG, "vadjustment", "GtkLayout" " or " "GtkViewport");
  if (XEN_GtkLayout__P(ptr)) return(C_TO_XEN_GtkAdjustment_((GtkAdjustment*)((XEN_TO_C_GtkLayout_(ptr))->vadjustment)));
  return(C_TO_XEN_GtkAdjustment_((GtkAdjustment*)((XEN_TO_C_GtkViewport_(ptr))->vadjustment)));
}

static XEN gxg_visibility(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkLayout__P(ptr), ptr, XEN_ONLY_ARG, "visibility", "GtkLayout");
  return(C_TO_XEN_GdkVisibilityState((GdkVisibilityState)((XEN_TO_C_GtkLayout_(ptr))->visibility)));
}

static XEN gxg_scroll_x(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkLayout__P(ptr), ptr, XEN_ONLY_ARG, "scroll_x", "GtkLayout");
  return(C_TO_XEN_gint((gint)((XEN_TO_C_GtkLayout_(ptr))->scroll_x)));
}

static XEN gxg_scroll_y(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkLayout__P(ptr), ptr, XEN_ONLY_ARG, "scroll_y", "GtkLayout");
  return(C_TO_XEN_gint((gint)((XEN_TO_C_GtkLayout_(ptr))->scroll_y)));
}

static XEN gxg_freeze_count(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkLayout__P(ptr), ptr, XEN_ONLY_ARG, "freeze_count", "GtkLayout");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkLayout_(ptr))->freeze_count)));
}

#if (!HAVE_GTK_1)
static XEN gxg_stamp(XEN ptr)
{
  if (XEN_GtkListStore__P(ptr)) return(C_TO_XEN_gint((gint)((XEN_TO_C_GtkListStore_(ptr))->stamp)));
  if (XEN_GtkTreeModelSort__P(ptr)) return(C_TO_XEN_gint((gint)((XEN_TO_C_GtkTreeModelSort_(ptr))->stamp)));
  if (XEN_GtkTreeStore__P(ptr)) return(C_TO_XEN_gint((gint)((XEN_TO_C_GtkTreeStore_(ptr))->stamp)));
  XEN_ASSERT_TYPE(0, ptr, XEN_ONLY_ARG, "stamp", "pointer to struct with stamp field");
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_root(XEN ptr)
{
  if (XEN_GtkListStore__P(ptr)) return(C_TO_XEN_gpointer((gpointer)((XEN_TO_C_GtkListStore_(ptr))->root)));
  if (XEN_GtkTreeModelSort__P(ptr)) return(C_TO_XEN_gpointer((gpointer)((XEN_TO_C_GtkTreeModelSort_(ptr))->root)));
  if (XEN_GtkTreeStore__P(ptr)) return(C_TO_XEN_gpointer((gpointer)((XEN_TO_C_GtkTreeStore_(ptr))->root)));
  XEN_ASSERT_TYPE(0, ptr, XEN_ONLY_ARG, "root", "pointer to struct with root field");
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_tail(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkListStore__P(ptr), ptr, XEN_ONLY_ARG, "tail", "GtkListStore");
  return(C_TO_XEN_gpointer((gpointer)((XEN_TO_C_GtkListStore_(ptr))->tail)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_sort_list(XEN ptr)
{
  if (XEN_GtkListStore__P(ptr)) return(C_TO_XEN_GList_((GList*)((XEN_TO_C_GtkListStore_(ptr))->sort_list)));
  if (XEN_GtkTreeModelSort__P(ptr)) return(C_TO_XEN_GList_((GList*)((XEN_TO_C_GtkTreeModelSort_(ptr))->sort_list)));
  if (XEN_GtkTreeStore__P(ptr)) return(C_TO_XEN_GList_((GList*)((XEN_TO_C_GtkTreeStore_(ptr))->sort_list)));
  XEN_ASSERT_TYPE(0, ptr, XEN_ONLY_ARG, "sort_list", "pointer to struct with sort_list field");
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_n_columns(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkListStore__P(ptr) || XEN_GtkTreeStore__P(ptr), ptr, XEN_ONLY_ARG, "n_columns", "GtkListStore" " or " "GtkTreeStore");
  if (XEN_GtkListStore__P(ptr)) return(C_TO_XEN_gint((gint)((XEN_TO_C_GtkListStore_(ptr))->n_columns)));
  return(C_TO_XEN_gint((gint)((XEN_TO_C_GtkTreeStore_(ptr))->n_columns)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_sort_column_id(XEN ptr)
{
  if (XEN_GtkListStore__P(ptr)) return(C_TO_XEN_gint((gint)((XEN_TO_C_GtkListStore_(ptr))->sort_column_id)));
  if (XEN_GtkTreeModelSort__P(ptr)) return(C_TO_XEN_gint((gint)((XEN_TO_C_GtkTreeModelSort_(ptr))->sort_column_id)));
  if (XEN_GtkTreeStore__P(ptr)) return(C_TO_XEN_gint((gint)((XEN_TO_C_GtkTreeStore_(ptr))->sort_column_id)));
  if (XEN_GtkTreeViewColumn__P(ptr)) return(C_TO_XEN_gint((gint)((XEN_TO_C_GtkTreeViewColumn_(ptr))->sort_column_id)));
  XEN_ASSERT_TYPE(0, ptr, XEN_ONLY_ARG, "sort_column_id", "pointer to struct with sort_column_id field");
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_order(XEN ptr)
{
  if (XEN_GtkListStore__P(ptr)) return(C_TO_XEN_GtkSortType((GtkSortType)((XEN_TO_C_GtkListStore_(ptr))->order)));
  if (XEN_GtkTreeModelSort__P(ptr)) return(C_TO_XEN_GtkSortType((GtkSortType)((XEN_TO_C_GtkTreeModelSort_(ptr))->order)));
  if (XEN_GtkTreeStore__P(ptr)) return(C_TO_XEN_GtkSortType((GtkSortType)((XEN_TO_C_GtkTreeStore_(ptr))->order)));
  XEN_ASSERT_TYPE(0, ptr, XEN_ONLY_ARG, "order", "pointer to struct with order field");
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_column_headers(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkListStore__P(ptr) || XEN_GtkTreeStore__P(ptr), ptr, XEN_ONLY_ARG, "column_headers", "GtkListStore" " or " "GtkTreeStore");
  if (XEN_GtkListStore__P(ptr)) return(C_TO_XEN_GType_((GType*)((XEN_TO_C_GtkListStore_(ptr))->column_headers)));
  return(C_TO_XEN_GType_((GType*)((XEN_TO_C_GtkTreeStore_(ptr))->column_headers)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_default_sort_func(XEN ptr)
{
  if (XEN_GtkListStore__P(ptr)) return(C_TO_XEN_GtkTreeIterCompareFunc((GtkTreeIterCompareFunc)((XEN_TO_C_GtkListStore_(ptr))->default_sort_func)));
  if (XEN_GtkTreeModelSort__P(ptr)) return(C_TO_XEN_GtkTreeIterCompareFunc((GtkTreeIterCompareFunc)((XEN_TO_C_GtkTreeModelSort_(ptr))->default_sort_func)));
  if (XEN_GtkTreeStore__P(ptr)) return(C_TO_XEN_GtkTreeIterCompareFunc((GtkTreeIterCompareFunc)((XEN_TO_C_GtkTreeStore_(ptr))->default_sort_func)));
  XEN_ASSERT_TYPE(0, ptr, XEN_ONLY_ARG, "default_sort_func", "pointer to struct with default_sort_func field");
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_default_sort_data(XEN ptr)
{
  if (XEN_GtkListStore__P(ptr)) return(C_TO_XEN_gpointer((gpointer)((XEN_TO_C_GtkListStore_(ptr))->default_sort_data)));
  if (XEN_GtkTreeModelSort__P(ptr)) return(C_TO_XEN_gpointer((gpointer)((XEN_TO_C_GtkTreeModelSort_(ptr))->default_sort_data)));
  if (XEN_GtkTreeStore__P(ptr)) return(C_TO_XEN_gpointer((gpointer)((XEN_TO_C_GtkTreeStore_(ptr))->default_sort_data)));
  XEN_ASSERT_TYPE(0, ptr, XEN_ONLY_ARG, "default_sort_data", "pointer to struct with default_sort_data field");
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_default_sort_destroy(XEN ptr)
{
  if (XEN_GtkListStore__P(ptr)) return(C_TO_XEN_GtkDestroyNotify((GtkDestroyNotify)((XEN_TO_C_GtkListStore_(ptr))->default_sort_destroy)));
  if (XEN_GtkTreeModelSort__P(ptr)) return(C_TO_XEN_GtkDestroyNotify((GtkDestroyNotify)((XEN_TO_C_GtkTreeModelSort_(ptr))->default_sort_destroy)));
  if (XEN_GtkTreeStore__P(ptr)) return(C_TO_XEN_GtkDestroyNotify((GtkDestroyNotify)((XEN_TO_C_GtkTreeStore_(ptr))->default_sort_destroy)));
  XEN_ASSERT_TYPE(0, ptr, XEN_ONLY_ARG, "default_sort_destroy", "pointer to struct with default_sort_destroy field");
}
#endif

static XEN gxg_parent_menu_item(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkMenu__P(ptr), ptr, XEN_ONLY_ARG, "parent_menu_item", "GtkMenu");
  return(C_TO_XEN_GtkWidget_((GtkWidget*)((XEN_TO_C_GtkMenu_(ptr))->parent_menu_item)));
}

static XEN gxg_old_active_menu_item(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkMenu__P(ptr), ptr, XEN_ONLY_ARG, "old_active_menu_item", "GtkMenu");
  return(C_TO_XEN_GtkWidget_((GtkWidget*)((XEN_TO_C_GtkMenu_(ptr))->old_active_menu_item)));
}

#if (!HAVE_GTK_1)
static XEN gxg_accel_path(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkMenu__P(ptr) || XEN_GtkMenuItem__P(ptr), ptr, XEN_ONLY_ARG, "accel_path", "GtkMenu" " or " "GtkMenuItem");
  if (XEN_GtkMenu__P(ptr)) return(C_TO_XEN_gchar_((gchar*)((XEN_TO_C_GtkMenu_(ptr))->accel_path)));
  return(C_TO_XEN_gchar_((gchar*)((XEN_TO_C_GtkMenuItem_(ptr))->accel_path)));
}
#endif

static XEN gxg_position_func(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkMenu__P(ptr), ptr, XEN_ONLY_ARG, "position_func", "GtkMenu");
  return(C_TO_XEN_GtkMenuPositionFunc((GtkMenuPositionFunc)((XEN_TO_C_GtkMenu_(ptr))->position_func)));
}

static XEN gxg_position_func_data(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkMenu__P(ptr), ptr, XEN_ONLY_ARG, "position_func_data", "GtkMenu");
  return(C_TO_XEN_gpointer((gpointer)((XEN_TO_C_GtkMenu_(ptr))->position_func_data)));
}

#if (!HAVE_GTK_1)
static XEN gxg_toggle_size(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkMenu__P(ptr) || XEN_GtkMenuItem__P(ptr), ptr, XEN_ONLY_ARG, "toggle_size", "GtkMenu" " or " "GtkMenuItem");
  if (XEN_GtkMenu__P(ptr)) return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkMenu_(ptr))->toggle_size)));
  return(C_TO_XEN_guint16((guint16)((XEN_TO_C_GtkMenuItem_(ptr))->toggle_size)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_toplevel(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkMenu__P(ptr) || XEN_GtkSocket__P(ptr), ptr, XEN_ONLY_ARG, "toplevel", "GtkMenu" " or " "GtkSocket");
  if (XEN_GtkMenu__P(ptr)) return(C_TO_XEN_GtkWidget_((GtkWidget*)((XEN_TO_C_GtkMenu_(ptr))->toplevel)));
  return(C_TO_XEN_GtkWidget_((GtkWidget*)((XEN_TO_C_GtkSocket_(ptr))->toplevel)));
}
#endif

static XEN gxg_tearoff_window(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkMenu__P(ptr), ptr, XEN_ONLY_ARG, "tearoff_window", "GtkMenu");
  return(C_TO_XEN_GtkWidget_((GtkWidget*)((XEN_TO_C_GtkMenu_(ptr))->tearoff_window)));
}

#if (!HAVE_GTK_1)
static XEN gxg_tearoff_hbox(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkMenu__P(ptr), ptr, XEN_ONLY_ARG, "tearoff_hbox", "GtkMenu");
  return(C_TO_XEN_GtkWidget_((GtkWidget*)((XEN_TO_C_GtkMenu_(ptr))->tearoff_hbox)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_tearoff_scrollbar(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkMenu__P(ptr), ptr, XEN_ONLY_ARG, "tearoff_scrollbar", "GtkMenu");
  return(C_TO_XEN_GtkWidget_((GtkWidget*)((XEN_TO_C_GtkMenu_(ptr))->tearoff_scrollbar)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_tearoff_adjustment(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkMenu__P(ptr), ptr, XEN_ONLY_ARG, "tearoff_adjustment", "GtkMenu");
  return(C_TO_XEN_GtkAdjustment_((GtkAdjustment*)((XEN_TO_C_GtkMenu_(ptr))->tearoff_adjustment)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_view_window(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkMenu__P(ptr) || XEN_GtkViewport__P(ptr), ptr, XEN_ONLY_ARG, "view_window", "GtkMenu" " or " "GtkViewport");
  if (XEN_GtkMenu__P(ptr)) return(C_TO_XEN_GdkWindow_((GdkWindow*)((XEN_TO_C_GtkMenu_(ptr))->view_window)));
  return(C_TO_XEN_GdkWindow_((GdkWindow*)((XEN_TO_C_GtkViewport_(ptr))->view_window)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_saved_scroll_offset(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkMenu__P(ptr), ptr, XEN_ONLY_ARG, "saved_scroll_offset", "GtkMenu");
  return(C_TO_XEN_gint((gint)((XEN_TO_C_GtkMenu_(ptr))->saved_scroll_offset)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_scroll_step(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkMenu__P(ptr), ptr, XEN_ONLY_ARG, "scroll_step", "GtkMenu");
  return(C_TO_XEN_gint((gint)((XEN_TO_C_GtkMenu_(ptr))->scroll_step)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_timeout_id(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkMenu__P(ptr), ptr, XEN_ONLY_ARG, "timeout_id", "GtkMenu");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkMenu_(ptr))->timeout_id)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_navigation_region(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkMenu__P(ptr), ptr, XEN_ONLY_ARG, "navigation_region", "GtkMenu");
  return(C_TO_XEN_GdkRegion_((GdkRegion*)((XEN_TO_C_GtkMenu_(ptr))->navigation_region)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_navigation_timeout(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkMenu__P(ptr), ptr, XEN_ONLY_ARG, "navigation_timeout", "GtkMenu");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkMenu_(ptr))->navigation_timeout)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_needs_destruction_ref_count(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkMenu__P(ptr), ptr, XEN_ONLY_ARG, "needs_destruction_ref_count", "GtkMenu");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkMenu_(ptr))->needs_destruction_ref_count)));
}
#endif

static XEN gxg_torn_off(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkMenu__P(ptr), ptr, XEN_ONLY_ARG, "torn_off", "GtkMenu");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkMenu_(ptr))->torn_off)));
}

#if (!HAVE_GTK_1)
static XEN gxg_tearoff_active(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkMenu__P(ptr), ptr, XEN_ONLY_ARG, "tearoff_active", "GtkMenu");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkMenu_(ptr))->tearoff_active)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_scroll_fast(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkMenu__P(ptr), ptr, XEN_ONLY_ARG, "scroll_fast", "GtkMenu");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkMenu_(ptr))->scroll_fast)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_upper_arrow_visible(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkMenu__P(ptr), ptr, XEN_ONLY_ARG, "upper_arrow_visible", "GtkMenu");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkMenu_(ptr))->upper_arrow_visible)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_lower_arrow_visible(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkMenu__P(ptr), ptr, XEN_ONLY_ARG, "lower_arrow_visible", "GtkMenu");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkMenu_(ptr))->lower_arrow_visible)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_upper_arrow_prelight(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkMenu__P(ptr), ptr, XEN_ONLY_ARG, "upper_arrow_prelight", "GtkMenu");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkMenu_(ptr))->upper_arrow_prelight)));
}
#endif

static XEN gxg_submenu(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkMenuItem__P(ptr), ptr, XEN_ONLY_ARG, "submenu", "GtkMenuItem");
  return(C_TO_XEN_GtkWidget_((GtkWidget*)((XEN_TO_C_GtkMenuItem_(ptr))->submenu)));
}

static XEN gxg_accelerator_width(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkMenuItem__P(ptr), ptr, XEN_ONLY_ARG, "accelerator_width", "GtkMenuItem");
  return(C_TO_XEN_guint16((guint16)((XEN_TO_C_GtkMenuItem_(ptr))->accelerator_width)));
}

static XEN gxg_show_submenu_indicator(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkMenuItem__P(ptr), ptr, XEN_ONLY_ARG, "show_submenu_indicator", "GtkMenuItem");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkMenuItem_(ptr))->show_submenu_indicator)));
}

static XEN gxg_submenu_placement(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkMenuItem__P(ptr), ptr, XEN_ONLY_ARG, "submenu_placement", "GtkMenuItem");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkMenuItem_(ptr))->submenu_placement)));
}

static XEN gxg_submenu_direction(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkMenuItem__P(ptr), ptr, XEN_ONLY_ARG, "submenu_direction", "GtkMenuItem");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkMenuItem_(ptr))->submenu_direction)));
}

static XEN gxg_right_justify(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkMenuItem__P(ptr), ptr, XEN_ONLY_ARG, "right_justify", "GtkMenuItem");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkMenuItem_(ptr))->right_justify)));
}

static XEN gxg_timer(XEN ptr)
{
  if (XEN_GtkMenuItem__P(ptr)) return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkMenuItem_(ptr))->timer)));
  if (XEN_GtkNotebook__P(ptr)) return(C_TO_XEN_guint32((guint32)((XEN_TO_C_GtkNotebook_(ptr))->timer)));
#if (!HAVE_GTK_1)
  if (XEN_GtkRange__P(ptr)) return(C_TO_XEN_GtkRangeStepTimer_((GtkRangeStepTimer*)((XEN_TO_C_GtkRange_(ptr))->timer)));
#endif
  if (XEN_GtkSpinButton__P(ptr)) return(C_TO_XEN_guint32((guint32)((XEN_TO_C_GtkSpinButton_(ptr))->timer)));
  XEN_ASSERT_TYPE(0, ptr, XEN_ONLY_ARG, "timer", "pointer to struct with timer field");
}

static XEN gxg_active_menu_item(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkMenuShell__P(ptr), ptr, XEN_ONLY_ARG, "active_menu_item", "GtkMenuShell");
  return(C_TO_XEN_GtkWidget_((GtkWidget*)((XEN_TO_C_GtkMenuShell_(ptr))->active_menu_item)));
}

static XEN gxg_parent_menu_shell(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkMenuShell__P(ptr), ptr, XEN_ONLY_ARG, "parent_menu_shell", "GtkMenuShell");
  return(C_TO_XEN_GtkWidget_((GtkWidget*)((XEN_TO_C_GtkMenuShell_(ptr))->parent_menu_shell)));
}

static XEN gxg_activate_time(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkMenuShell__P(ptr), ptr, XEN_ONLY_ARG, "activate_time", "GtkMenuShell");
  return(C_TO_XEN_guint32((guint32)((XEN_TO_C_GtkMenuShell_(ptr))->activate_time)));
}

#if (!HAVE_GTK_1)
static XEN gxg_have_grab(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkMenuShell__P(ptr) || XEN_GtkTooltips__P(ptr), ptr, XEN_ONLY_ARG, "have_grab", "GtkMenuShell" " or " "GtkTooltips");
  if (XEN_GtkMenuShell__P(ptr)) return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkMenuShell_(ptr))->have_grab)));
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkTooltips_(ptr))->have_grab)));
}
#endif

static XEN gxg_have_xgrab(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkMenuShell__P(ptr), ptr, XEN_ONLY_ARG, "have_xgrab", "GtkMenuShell");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkMenuShell_(ptr))->have_xgrab)));
}

static XEN gxg_ignore_leave(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkMenuShell__P(ptr), ptr, XEN_ONLY_ARG, "ignore_leave", "GtkMenuShell");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkMenuShell_(ptr))->ignore_leave)));
}

static XEN gxg_menu_flag(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkMenuShell__P(ptr), ptr, XEN_ONLY_ARG, "menu_flag", "GtkMenuShell");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkMenuShell_(ptr))->menu_flag)));
}

static XEN gxg_ignore_enter(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkMenuShell__P(ptr), ptr, XEN_ONLY_ARG, "ignore_enter", "GtkMenuShell");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkMenuShell_(ptr))->ignore_enter)));
}

static XEN gxg_cur_page(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkNotebook__P(ptr), ptr, XEN_ONLY_ARG, "cur_page", "GtkNotebook");
  return(C_TO_XEN_GtkNotebookPage_((GtkNotebookPage*)((XEN_TO_C_GtkNotebook_(ptr))->cur_page)));
}

static XEN gxg_first_tab(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkNotebook__P(ptr), ptr, XEN_ONLY_ARG, "first_tab", "GtkNotebook");
  return(C_TO_XEN_GList_((GList*)((XEN_TO_C_GtkNotebook_(ptr))->first_tab)));
}

static XEN gxg_focus_tab(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkNotebook__P(ptr), ptr, XEN_ONLY_ARG, "focus_tab", "GtkNotebook");
  return(C_TO_XEN_GList_((GList*)((XEN_TO_C_GtkNotebook_(ptr))->focus_tab)));
}

static XEN gxg_menu(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkNotebook__P(ptr) || XEN_GtkOptionMenu__P(ptr), ptr, XEN_ONLY_ARG, "menu", "GtkNotebook" " or " "GtkOptionMenu");
  if (XEN_GtkNotebook__P(ptr)) return(C_TO_XEN_GtkWidget_((GtkWidget*)((XEN_TO_C_GtkNotebook_(ptr))->menu)));
  return(C_TO_XEN_GtkWidget_((GtkWidget*)((XEN_TO_C_GtkOptionMenu_(ptr))->menu)));
}

static XEN gxg_tab_hborder(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkNotebook__P(ptr), ptr, XEN_ONLY_ARG, "tab_hborder", "GtkNotebook");
  return(C_TO_XEN_guint16((guint16)((XEN_TO_C_GtkNotebook_(ptr))->tab_hborder)));
}

static XEN gxg_tab_vborder(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkNotebook__P(ptr), ptr, XEN_ONLY_ARG, "tab_vborder", "GtkNotebook");
  return(C_TO_XEN_guint16((guint16)((XEN_TO_C_GtkNotebook_(ptr))->tab_vborder)));
}

static XEN gxg_show_tabs(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkNotebook__P(ptr), ptr, XEN_ONLY_ARG, "show_tabs", "GtkNotebook");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkNotebook_(ptr))->show_tabs)));
}

static XEN gxg_homogeneous(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkNotebook__P(ptr), ptr, XEN_ONLY_ARG, "homogeneous", "GtkNotebook");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkNotebook_(ptr))->homogeneous)));
}

static XEN gxg_show_border(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkNotebook__P(ptr), ptr, XEN_ONLY_ARG, "show_border", "GtkNotebook");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkNotebook_(ptr))->show_border)));
}

static XEN gxg_tab_pos(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkNotebook__P(ptr), ptr, XEN_ONLY_ARG, "tab_pos", "GtkNotebook");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkNotebook_(ptr))->tab_pos)));
}

static XEN gxg_scrollable(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkNotebook__P(ptr), ptr, XEN_ONLY_ARG, "scrollable", "GtkNotebook");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkNotebook_(ptr))->scrollable)));
}

static XEN gxg_in_child(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkNotebook__P(ptr) || XEN_GtkSpinButton__P(ptr), ptr, XEN_ONLY_ARG, "in_child", "GtkNotebook" " or " "GtkSpinButton");
  if (XEN_GtkNotebook__P(ptr)) return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkNotebook_(ptr))->in_child)));
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkSpinButton_(ptr))->in_child)));
}

static XEN gxg_click_child(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkNotebook__P(ptr) || XEN_GtkSpinButton__P(ptr), ptr, XEN_ONLY_ARG, "click_child", "GtkNotebook" " or " "GtkSpinButton");
  if (XEN_GtkNotebook__P(ptr)) return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkNotebook_(ptr))->click_child)));
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkSpinButton_(ptr))->click_child)));
}

static XEN gxg_need_timer(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkNotebook__P(ptr) || XEN_GtkSpinButton__P(ptr), ptr, XEN_ONLY_ARG, "need_timer", "GtkNotebook" " or " "GtkSpinButton");
  if (XEN_GtkNotebook__P(ptr)) return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkNotebook_(ptr))->need_timer)));
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkSpinButton_(ptr))->need_timer)));
}

static XEN gxg_child_has_focus(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkNotebook__P(ptr), ptr, XEN_ONLY_ARG, "child_has_focus", "GtkNotebook");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkNotebook_(ptr))->child_has_focus)));
}

static XEN gxg_flags(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkObject__P(ptr) || XEN_GtkTargetEntry__P(ptr), ptr, XEN_ONLY_ARG, "flags", "GtkObject" " or " "GtkTargetEntry");
  if (XEN_GtkObject__P(ptr)) return(C_TO_XEN_guint32((guint32)((XEN_TO_C_GtkObject_(ptr))->flags)));
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkTargetEntry_(ptr))->flags)));
}

static XEN gxg_menu_item(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkOptionMenu__P(ptr), ptr, XEN_ONLY_ARG, "menu_item", "GtkOptionMenu");
  return(C_TO_XEN_GtkWidget_((GtkWidget*)((XEN_TO_C_GtkOptionMenu_(ptr))->menu_item)));
}

static XEN gxg_child1(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkPaned__P(ptr), ptr, XEN_ONLY_ARG, "child1", "GtkPaned");
  return(C_TO_XEN_GtkWidget_((GtkWidget*)((XEN_TO_C_GtkPaned_(ptr))->child1)));
}

static XEN gxg_child2(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkPaned__P(ptr), ptr, XEN_ONLY_ARG, "child2", "GtkPaned");
  return(C_TO_XEN_GtkWidget_((GtkWidget*)((XEN_TO_C_GtkPaned_(ptr))->child2)));
}

static XEN gxg_handle(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkPaned__P(ptr), ptr, XEN_ONLY_ARG, "handle", "GtkPaned");
  return(C_TO_XEN_GdkWindow_((GdkWindow*)((XEN_TO_C_GtkPaned_(ptr))->handle)));
}

static XEN gxg_xor_gc(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkPaned__P(ptr), ptr, XEN_ONLY_ARG, "xor_gc", "GtkPaned");
  return(C_TO_XEN_GdkGC_((GdkGC*)((XEN_TO_C_GtkPaned_(ptr))->xor_gc)));
}

static XEN gxg_child1_size(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkPaned__P(ptr), ptr, XEN_ONLY_ARG, "child1_size", "GtkPaned");
  return(C_TO_XEN_gint((gint)((XEN_TO_C_GtkPaned_(ptr))->child1_size)));
}

static XEN gxg_last_allocation(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkPaned__P(ptr), ptr, XEN_ONLY_ARG, "last_allocation", "GtkPaned");
  return(C_TO_XEN_gint((gint)((XEN_TO_C_GtkPaned_(ptr))->last_allocation)));
}

static XEN gxg_min_position(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkPaned__P(ptr), ptr, XEN_ONLY_ARG, "min_position", "GtkPaned");
  return(C_TO_XEN_gint((gint)((XEN_TO_C_GtkPaned_(ptr))->min_position)));
}

static XEN gxg_max_position(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkPaned__P(ptr), ptr, XEN_ONLY_ARG, "max_position", "GtkPaned");
  return(C_TO_XEN_gint((gint)((XEN_TO_C_GtkPaned_(ptr))->max_position)));
}

static XEN gxg_position_set(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkPaned__P(ptr), ptr, XEN_ONLY_ARG, "position_set", "GtkPaned");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkPaned_(ptr))->position_set)));
}

static XEN gxg_child1_shrink(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkPaned__P(ptr), ptr, XEN_ONLY_ARG, "child1_shrink", "GtkPaned");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkPaned_(ptr))->child1_shrink)));
}

static XEN gxg_child1_resize(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkPaned__P(ptr), ptr, XEN_ONLY_ARG, "child1_resize", "GtkPaned");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkPaned_(ptr))->child1_resize)));
}

static XEN gxg_child2_shrink(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkPaned__P(ptr), ptr, XEN_ONLY_ARG, "child2_shrink", "GtkPaned");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkPaned_(ptr))->child2_shrink)));
}

static XEN gxg_child2_resize(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkPaned__P(ptr), ptr, XEN_ONLY_ARG, "child2_resize", "GtkPaned");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkPaned_(ptr))->child2_resize)));
}

static XEN gxg_socket_window(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkPlug__P(ptr), ptr, XEN_ONLY_ARG, "socket_window", "GtkPlug");
  return(C_TO_XEN_GdkWindow_((GdkWindow*)((XEN_TO_C_GtkPlug_(ptr))->socket_window)));
}

#if (!HAVE_GTK_1)
static XEN gxg_modality_window(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkPlug__P(ptr), ptr, XEN_ONLY_ARG, "modality_window", "GtkPlug");
  return(C_TO_XEN_GtkWidget_((GtkWidget*)((XEN_TO_C_GtkPlug_(ptr))->modality_window)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_modality_group(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkPlug__P(ptr), ptr, XEN_ONLY_ARG, "modality_group", "GtkPlug");
  return(C_TO_XEN_GtkWindowGroup_((GtkWindowGroup*)((XEN_TO_C_GtkPlug_(ptr))->modality_group)));
}
#endif

static XEN gxg_adjustment(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkRange__P(ptr) || XEN_GtkSpinButton__P(ptr), ptr, XEN_ONLY_ARG, "adjustment", "GtkRange" " or " "GtkSpinButton");
  if (XEN_GtkRange__P(ptr)) return(C_TO_XEN_GtkAdjustment_((GtkAdjustment*)((XEN_TO_C_GtkRange_(ptr))->adjustment)));
  return(C_TO_XEN_GtkAdjustment_((GtkAdjustment*)((XEN_TO_C_GtkSpinButton_(ptr))->adjustment)));
}

#if (!HAVE_GTK_1)
static XEN gxg_update_policy(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkRange__P(ptr) || XEN_GtkSpinButton__P(ptr), ptr, XEN_ONLY_ARG, "update_policy", "GtkRange" " or " "GtkSpinButton");
  if (XEN_GtkRange__P(ptr)) return(C_TO_XEN_GtkUpdateType((GtkUpdateType)((XEN_TO_C_GtkRange_(ptr))->update_policy)));
  return(C_TO_XEN_GtkSpinButtonUpdatePolicy((GtkSpinButtonUpdatePolicy)((XEN_TO_C_GtkSpinButton_(ptr))->update_policy)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_inverted(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkRange__P(ptr), ptr, XEN_ONLY_ARG, "inverted", "GtkRange");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkRange_(ptr))->inverted)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_flippable(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkRange__P(ptr), ptr, XEN_ONLY_ARG, "flippable", "GtkRange");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkRange_(ptr))->flippable)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_has_stepper_a(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkRange__P(ptr), ptr, XEN_ONLY_ARG, "has_stepper_a", "GtkRange");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkRange_(ptr))->has_stepper_a)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_has_stepper_b(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkRange__P(ptr), ptr, XEN_ONLY_ARG, "has_stepper_b", "GtkRange");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkRange_(ptr))->has_stepper_b)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_has_stepper_c(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkRange__P(ptr), ptr, XEN_ONLY_ARG, "has_stepper_c", "GtkRange");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkRange_(ptr))->has_stepper_c)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_has_stepper_d(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkRange__P(ptr), ptr, XEN_ONLY_ARG, "has_stepper_d", "GtkRange");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkRange_(ptr))->has_stepper_d)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_need_recalc(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkRange__P(ptr), ptr, XEN_ONLY_ARG, "need_recalc", "GtkRange");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkRange_(ptr))->need_recalc)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_slider_size_fixed(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkRange__P(ptr), ptr, XEN_ONLY_ARG, "slider_size_fixed", "GtkRange");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkRange_(ptr))->slider_size_fixed)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_min_slider_size(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkRange__P(ptr), ptr, XEN_ONLY_ARG, "min_slider_size", "GtkRange");
  return(C_TO_XEN_gint((gint)((XEN_TO_C_GtkRange_(ptr))->min_slider_size)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_orientation(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkRange__P(ptr) || XEN_GtkToolbar__P(ptr), ptr, XEN_ONLY_ARG, "orientation", "GtkRange" " or " "GtkToolbar");
  if (XEN_GtkRange__P(ptr)) return(C_TO_XEN_GtkOrientation((GtkOrientation)((XEN_TO_C_GtkRange_(ptr))->orientation)));
  return(C_TO_XEN_GtkOrientation((GtkOrientation)((XEN_TO_C_GtkToolbar_(ptr))->orientation)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_slider_start(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkRange__P(ptr), ptr, XEN_ONLY_ARG, "slider_start", "GtkRange");
  return(C_TO_XEN_gint((gint)((XEN_TO_C_GtkRange_(ptr))->slider_start)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_slider_end(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkRange__P(ptr), ptr, XEN_ONLY_ARG, "slider_end", "GtkRange");
  return(C_TO_XEN_gint((gint)((XEN_TO_C_GtkRange_(ptr))->slider_end)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_round_digits(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkRange__P(ptr), ptr, XEN_ONLY_ARG, "round_digits", "GtkRange");
  return(C_TO_XEN_gint((gint)((XEN_TO_C_GtkRange_(ptr))->round_digits)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_trough_click_forward(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkRange__P(ptr), ptr, XEN_ONLY_ARG, "trough_click_forward", "GtkRange");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkRange_(ptr))->trough_click_forward)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_update_pending(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkRange__P(ptr), ptr, XEN_ONLY_ARG, "update_pending", "GtkRange");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkRange_(ptr))->update_pending)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_slide_initial_slider_position(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkRange__P(ptr), ptr, XEN_ONLY_ARG, "slide_initial_slider_position", "GtkRange");
  return(C_TO_XEN_gint((gint)((XEN_TO_C_GtkRange_(ptr))->slide_initial_slider_position)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_slide_initial_coordinate(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkRange__P(ptr), ptr, XEN_ONLY_ARG, "slide_initial_coordinate", "GtkRange");
  return(C_TO_XEN_gint((gint)((XEN_TO_C_GtkRange_(ptr))->slide_initial_coordinate)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_update_timeout_id(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkRange__P(ptr), ptr, XEN_ONLY_ARG, "update_timeout_id", "GtkRange");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkRange_(ptr))->update_timeout_id)));
}
#endif

static XEN gxg_metric_name(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkRulerMetric__P(ptr), ptr, XEN_ONLY_ARG, "metric_name", "GtkRulerMetric");
  return(C_TO_XEN_gchar_((gchar*)((XEN_TO_C_GtkRulerMetric_(ptr))->metric_name)));
}

static XEN gxg_abbrev(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkRulerMetric__P(ptr), ptr, XEN_ONLY_ARG, "abbrev", "GtkRulerMetric");
  return(C_TO_XEN_gchar_((gchar*)((XEN_TO_C_GtkRulerMetric_(ptr))->abbrev)));
}

static XEN gxg_pixels_per_unit(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkRulerMetric__P(ptr), ptr, XEN_ONLY_ARG, "pixels_per_unit", "GtkRulerMetric");
  return(C_TO_XEN_gdouble((gdouble)((XEN_TO_C_GtkRulerMetric_(ptr))->pixels_per_unit)));
}

static XEN gxg_ruler_scale(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkRulerMetric__P(ptr), ptr, XEN_ONLY_ARG, "ruler_scale", "GtkRulerMetric");
  return(C_TO_XEN_gdouble_((gdouble*)((XEN_TO_C_GtkRulerMetric_(ptr))->ruler_scale)));
}

static XEN gxg_subdivide(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkRulerMetric__P(ptr), ptr, XEN_ONLY_ARG, "subdivide", "GtkRulerMetric");
  return(C_TO_XEN_gint_((gint*)((XEN_TO_C_GtkRulerMetric_(ptr))->subdivide)));
}

static XEN gxg_backing_store(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkRuler__P(ptr), ptr, XEN_ONLY_ARG, "backing_store", "GtkRuler");
  return(C_TO_XEN_GdkPixmap_((GdkPixmap*)((XEN_TO_C_GtkRuler_(ptr))->backing_store)));
}

static XEN gxg_non_gr_exp_gc(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkRuler__P(ptr), ptr, XEN_ONLY_ARG, "non_gr_exp_gc", "GtkRuler");
  return(C_TO_XEN_GdkGC_((GdkGC*)((XEN_TO_C_GtkRuler_(ptr))->non_gr_exp_gc)));
}

static XEN gxg_metric(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkRuler__P(ptr), ptr, XEN_ONLY_ARG, "metric", "GtkRuler");
  return(C_TO_XEN_GtkRulerMetric_((GtkRulerMetric*)((XEN_TO_C_GtkRuler_(ptr))->metric)));
}

static XEN gxg_xsrc(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkRuler__P(ptr), ptr, XEN_ONLY_ARG, "xsrc", "GtkRuler");
  return(C_TO_XEN_gint((gint)((XEN_TO_C_GtkRuler_(ptr))->xsrc)));
}

static XEN gxg_ysrc(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkRuler__P(ptr), ptr, XEN_ONLY_ARG, "ysrc", "GtkRuler");
  return(C_TO_XEN_gint((gint)((XEN_TO_C_GtkRuler_(ptr))->ysrc)));
}

static XEN gxg_slider_size(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkRuler__P(ptr), ptr, XEN_ONLY_ARG, "slider_size", "GtkRuler");
  return(C_TO_XEN_gint((gint)((XEN_TO_C_GtkRuler_(ptr))->slider_size)));
}

static XEN gxg_position(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkRuler__P(ptr) || XEN_GtkWindow__P(ptr), ptr, XEN_ONLY_ARG, "position", "GtkRuler" " or " "GtkWindow");
  if (XEN_GtkRuler__P(ptr)) return(C_TO_XEN_gdouble((gdouble)((XEN_TO_C_GtkRuler_(ptr))->position)));
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkWindow_(ptr))->position)));
}

static XEN gxg_max_size(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkRuler__P(ptr), ptr, XEN_ONLY_ARG, "max_size", "GtkRuler");
  return(C_TO_XEN_gdouble((gdouble)((XEN_TO_C_GtkRuler_(ptr))->max_size)));
}

#if (!HAVE_GTK_1)
static XEN gxg_digits(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkScale__P(ptr) || XEN_GtkSpinButton__P(ptr), ptr, XEN_ONLY_ARG, "digits", "GtkScale" " or " "GtkSpinButton");
  if (XEN_GtkScale__P(ptr)) return(C_TO_XEN_gint((gint)((XEN_TO_C_GtkScale_(ptr))->digits)));
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkSpinButton_(ptr))->digits)));
}
#endif

static XEN gxg_draw_value(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkScale__P(ptr), ptr, XEN_ONLY_ARG, "draw_value", "GtkScale");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkScale_(ptr))->draw_value)));
}

static XEN gxg_hscrollbar(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkScrolledWindow__P(ptr), ptr, XEN_ONLY_ARG, "hscrollbar", "GtkScrolledWindow");
  return(C_TO_XEN_GtkWidget_((GtkWidget*)((XEN_TO_C_GtkScrolledWindow_(ptr))->hscrollbar)));
}

static XEN gxg_vscrollbar(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkScrolledWindow__P(ptr), ptr, XEN_ONLY_ARG, "vscrollbar", "GtkScrolledWindow");
  return(C_TO_XEN_GtkWidget_((GtkWidget*)((XEN_TO_C_GtkScrolledWindow_(ptr))->vscrollbar)));
}

static XEN gxg_hscrollbar_policy(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkScrolledWindow__P(ptr), ptr, XEN_ONLY_ARG, "hscrollbar_policy", "GtkScrolledWindow");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkScrolledWindow_(ptr))->hscrollbar_policy)));
}

static XEN gxg_vscrollbar_policy(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkScrolledWindow__P(ptr), ptr, XEN_ONLY_ARG, "vscrollbar_policy", "GtkScrolledWindow");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkScrolledWindow_(ptr))->vscrollbar_policy)));
}

static XEN gxg_hscrollbar_visible(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkScrolledWindow__P(ptr), ptr, XEN_ONLY_ARG, "hscrollbar_visible", "GtkScrolledWindow");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkScrolledWindow_(ptr))->hscrollbar_visible)));
}

static XEN gxg_vscrollbar_visible(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkScrolledWindow__P(ptr), ptr, XEN_ONLY_ARG, "vscrollbar_visible", "GtkScrolledWindow");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkScrolledWindow_(ptr))->vscrollbar_visible)));
}

static XEN gxg_window_placement(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkScrolledWindow__P(ptr), ptr, XEN_ONLY_ARG, "window_placement", "GtkScrolledWindow");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkScrolledWindow_(ptr))->window_placement)));
}

static XEN gxg_format(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkSelectionData__P(ptr), ptr, XEN_ONLY_ARG, "format", "GtkSelectionData");
  return(C_TO_XEN_gint((gint)((XEN_TO_C_GtkSelectionData_(ptr))->format)));
}

static XEN gxg_data(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkSelectionData__P(ptr), ptr, XEN_ONLY_ARG, "data", "GtkSelectionData");
  return(C_TO_XEN_guchar_((guchar*)((XEN_TO_C_GtkSelectionData_(ptr))->data)));
}

static XEN gxg_info(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkTargetEntry__P(ptr), ptr, XEN_ONLY_ARG, "info", "GtkTargetEntry");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkTargetEntry_(ptr))->info)));
}

#if (!HAVE_GTK_1)
static XEN gxg_queued_settings(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkSettings__P(ptr), ptr, XEN_ONLY_ARG, "queued_settings", "GtkSettings");
  return(C_TO_XEN_GData_((GData*)((XEN_TO_C_GtkSettings_(ptr))->queued_settings)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_property_values(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkSettings__P(ptr), ptr, XEN_ONLY_ARG, "property_values", "GtkSettings");
  return(C_TO_XEN_GValue_((GValue*)((XEN_TO_C_GtkSettings_(ptr))->property_values)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_rc_context(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkSettings__P(ptr), ptr, XEN_ONLY_ARG, "rc_context", "GtkSettings");
  return(C_TO_XEN_GtkRcContext_((GtkRcContext*)((XEN_TO_C_GtkSettings_(ptr))->rc_context)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_widgets(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkSizeGroup__P(ptr), ptr, XEN_ONLY_ARG, "widgets", "GtkSizeGroup");
  return(C_TO_XEN_GSList_((GSList*)((XEN_TO_C_GtkSizeGroup_(ptr))->widgets)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_have_width(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkSizeGroup__P(ptr), ptr, XEN_ONLY_ARG, "have_width", "GtkSizeGroup");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkSizeGroup_(ptr))->have_width)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_have_height(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkSizeGroup__P(ptr), ptr, XEN_ONLY_ARG, "have_height", "GtkSizeGroup");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkSizeGroup_(ptr))->have_height)));
}
#endif

static XEN gxg_request_width(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkSocket__P(ptr), ptr, XEN_ONLY_ARG, "request_width", "GtkSocket");
  return(C_TO_XEN_guint16((guint16)((XEN_TO_C_GtkSocket_(ptr))->request_width)));
}

static XEN gxg_request_height(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkSocket__P(ptr), ptr, XEN_ONLY_ARG, "request_height", "GtkSocket");
  return(C_TO_XEN_guint16((guint16)((XEN_TO_C_GtkSocket_(ptr))->request_height)));
}

static XEN gxg_current_width(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkSocket__P(ptr), ptr, XEN_ONLY_ARG, "current_width", "GtkSocket");
  return(C_TO_XEN_guint16((guint16)((XEN_TO_C_GtkSocket_(ptr))->current_width)));
}

static XEN gxg_current_height(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkSocket__P(ptr), ptr, XEN_ONLY_ARG, "current_height", "GtkSocket");
  return(C_TO_XEN_guint16((guint16)((XEN_TO_C_GtkSocket_(ptr))->current_height)));
}

static XEN gxg_plug_window(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkSocket__P(ptr), ptr, XEN_ONLY_ARG, "plug_window", "GtkSocket");
  return(C_TO_XEN_GdkWindow_((GdkWindow*)((XEN_TO_C_GtkSocket_(ptr))->plug_window)));
}

#if (!HAVE_GTK_1)
static XEN gxg_plug_widget(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkSocket__P(ptr), ptr, XEN_ONLY_ARG, "plug_widget", "GtkSocket");
  return(C_TO_XEN_GtkWidget_((GtkWidget*)((XEN_TO_C_GtkSocket_(ptr))->plug_widget)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_xembed_version(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkSocket__P(ptr), ptr, XEN_ONLY_ARG, "xembed_version", "GtkSocket");
  return(C_TO_XEN_gshort((gshort)((XEN_TO_C_GtkSocket_(ptr))->xembed_version)));
}
#endif

static XEN gxg_same_app(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkSocket__P(ptr), ptr, XEN_ONLY_ARG, "same_app", "GtkSocket");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkSocket_(ptr))->same_app)));
}

static XEN gxg_focus_in(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkSocket__P(ptr), ptr, XEN_ONLY_ARG, "focus_in", "GtkSocket");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkSocket_(ptr))->focus_in)));
}

static XEN gxg_have_size(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkSocket__P(ptr), ptr, XEN_ONLY_ARG, "have_size", "GtkSocket");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkSocket_(ptr))->have_size)));
}

static XEN gxg_need_map(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkSocket__P(ptr), ptr, XEN_ONLY_ARG, "need_map", "GtkSocket");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkSocket_(ptr))->need_map)));
}

#if (!HAVE_GTK_1)
static XEN gxg_is_mapped(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkSocket__P(ptr), ptr, XEN_ONLY_ARG, "is_mapped", "GtkSocket");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkSocket_(ptr))->is_mapped)));
}
#endif

static XEN gxg_panel(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkSpinButton__P(ptr), ptr, XEN_ONLY_ARG, "panel", "GtkSpinButton");
  return(C_TO_XEN_GdkWindow_((GdkWindow*)((XEN_TO_C_GtkSpinButton_(ptr))->panel)));
}

static XEN gxg_climb_rate(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkSpinButton__P(ptr), ptr, XEN_ONLY_ARG, "climb_rate", "GtkSpinButton");
  return(C_TO_XEN_gdouble((gdouble)((XEN_TO_C_GtkSpinButton_(ptr))->climb_rate)));
}

static XEN gxg_timer_step(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkSpinButton__P(ptr), ptr, XEN_ONLY_ARG, "timer_step", "GtkSpinButton");
  return(C_TO_XEN_gdouble((gdouble)((XEN_TO_C_GtkSpinButton_(ptr))->timer_step)));
}

static XEN gxg_timer_calls(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkSpinButton__P(ptr), ptr, XEN_ONLY_ARG, "timer_calls", "GtkSpinButton");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkSpinButton_(ptr))->timer_calls)));
}

static XEN gxg_numeric(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkSpinButton__P(ptr), ptr, XEN_ONLY_ARG, "numeric", "GtkSpinButton");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkSpinButton_(ptr))->numeric)));
}

#if (!HAVE_GTK_1)
static XEN gxg_stock_id(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkStockItem__P(ptr), ptr, XEN_ONLY_ARG, "stock_id", "GtkStockItem");
  return(C_TO_XEN_gchar_((gchar*)((XEN_TO_C_GtkStockItem_(ptr))->stock_id)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_modifier(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkStockItem__P(ptr), ptr, XEN_ONLY_ARG, "modifier", "GtkStockItem");
  return(C_TO_XEN_GdkModifierType((GdkModifierType)((XEN_TO_C_GtkStockItem_(ptr))->modifier)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_translation_domain(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkStockItem__P(ptr), ptr, XEN_ONLY_ARG, "translation_domain", "GtkStockItem");
  return(C_TO_XEN_gchar_((gchar*)((XEN_TO_C_GtkStockItem_(ptr))->translation_domain)));
}
#endif

static XEN gxg_fg(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkStyle__P(ptr), ptr, XEN_ONLY_ARG, "fg", "GtkStyle");
  return(C_TO_XEN_GdkColor_((GdkColor*)((XEN_TO_C_GtkStyle_(ptr))->fg)));
}

static XEN gxg_bg(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkStyle__P(ptr), ptr, XEN_ONLY_ARG, "bg", "GtkStyle");
  return(C_TO_XEN_GdkColor_((GdkColor*)((XEN_TO_C_GtkStyle_(ptr))->bg)));
}

static XEN gxg_light(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkStyle__P(ptr), ptr, XEN_ONLY_ARG, "light", "GtkStyle");
  return(C_TO_XEN_GdkColor_((GdkColor*)((XEN_TO_C_GtkStyle_(ptr))->light)));
}

static XEN gxg_dark(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkStyle__P(ptr), ptr, XEN_ONLY_ARG, "dark", "GtkStyle");
  return(C_TO_XEN_GdkColor_((GdkColor*)((XEN_TO_C_GtkStyle_(ptr))->dark)));
}

static XEN gxg_mid(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkStyle__P(ptr), ptr, XEN_ONLY_ARG, "mid", "GtkStyle");
  return(C_TO_XEN_GdkColor_((GdkColor*)((XEN_TO_C_GtkStyle_(ptr))->mid)));
}

static XEN gxg_base(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkStyle__P(ptr), ptr, XEN_ONLY_ARG, "base", "GtkStyle");
  return(C_TO_XEN_GdkColor_((GdkColor*)((XEN_TO_C_GtkStyle_(ptr))->base)));
}

#if (!HAVE_GTK_1)
static XEN gxg_text_aa(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkStyle__P(ptr), ptr, XEN_ONLY_ARG, "text_aa", "GtkStyle");
  return(C_TO_XEN_GdkColor_((GdkColor*)((XEN_TO_C_GtkStyle_(ptr))->text_aa)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_font_desc(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkStyle__P(ptr), ptr, XEN_ONLY_ARG, "font_desc", "GtkStyle");
  return(C_TO_XEN_PangoFontDescription_((PangoFontDescription*)((XEN_TO_C_GtkStyle_(ptr))->font_desc)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_xthickness(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkStyle__P(ptr), ptr, XEN_ONLY_ARG, "xthickness", "GtkStyle");
  return(C_TO_XEN_gint((gint)((XEN_TO_C_GtkStyle_(ptr))->xthickness)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_ythickness(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkStyle__P(ptr), ptr, XEN_ONLY_ARG, "ythickness", "GtkStyle");
  return(C_TO_XEN_gint((gint)((XEN_TO_C_GtkStyle_(ptr))->ythickness)));
}
#endif

static XEN gxg_fg_gc(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkStyle__P(ptr), ptr, XEN_ONLY_ARG, "fg_gc", "GtkStyle");
  return(C_TO_XEN_GdkGC__((GdkGC**)((XEN_TO_C_GtkStyle_(ptr))->fg_gc)));
}

static XEN gxg_bg_gc(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkStyle__P(ptr), ptr, XEN_ONLY_ARG, "bg_gc", "GtkStyle");
  return(C_TO_XEN_GdkGC__((GdkGC**)((XEN_TO_C_GtkStyle_(ptr))->bg_gc)));
}

static XEN gxg_light_gc(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkStyle__P(ptr), ptr, XEN_ONLY_ARG, "light_gc", "GtkStyle");
  return(C_TO_XEN_GdkGC__((GdkGC**)((XEN_TO_C_GtkStyle_(ptr))->light_gc)));
}

static XEN gxg_dark_gc(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkStyle__P(ptr), ptr, XEN_ONLY_ARG, "dark_gc", "GtkStyle");
  return(C_TO_XEN_GdkGC__((GdkGC**)((XEN_TO_C_GtkStyle_(ptr))->dark_gc)));
}

static XEN gxg_mid_gc(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkStyle__P(ptr), ptr, XEN_ONLY_ARG, "mid_gc", "GtkStyle");
  return(C_TO_XEN_GdkGC__((GdkGC**)((XEN_TO_C_GtkStyle_(ptr))->mid_gc)));
}

static XEN gxg_text_gc(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkStyle__P(ptr), ptr, XEN_ONLY_ARG, "text_gc", "GtkStyle");
  return(C_TO_XEN_GdkGC__((GdkGC**)((XEN_TO_C_GtkStyle_(ptr))->text_gc)));
}

static XEN gxg_base_gc(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkStyle__P(ptr), ptr, XEN_ONLY_ARG, "base_gc", "GtkStyle");
  return(C_TO_XEN_GdkGC__((GdkGC**)((XEN_TO_C_GtkStyle_(ptr))->base_gc)));
}

#if (!HAVE_GTK_1)
static XEN gxg_text_aa_gc(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkStyle__P(ptr), ptr, XEN_ONLY_ARG, "text_aa_gc", "GtkStyle");
  return(C_TO_XEN_GdkGC__((GdkGC**)((XEN_TO_C_GtkStyle_(ptr))->text_aa_gc)));
}
#endif

static XEN gxg_black_gc(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkStyle__P(ptr), ptr, XEN_ONLY_ARG, "black_gc", "GtkStyle");
  return(C_TO_XEN_GdkGC_((GdkGC*)((XEN_TO_C_GtkStyle_(ptr))->black_gc)));
}

static XEN gxg_white_gc(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkStyle__P(ptr), ptr, XEN_ONLY_ARG, "white_gc", "GtkStyle");
  return(C_TO_XEN_GdkGC_((GdkGC*)((XEN_TO_C_GtkStyle_(ptr))->white_gc)));
}

static XEN gxg_attach_count(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkStyle__P(ptr), ptr, XEN_ONLY_ARG, "attach_count", "GtkStyle");
  return(C_TO_XEN_gint((gint)((XEN_TO_C_GtkStyle_(ptr))->attach_count)));
}

static XEN gxg_rc_style(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkStyle__P(ptr), ptr, XEN_ONLY_ARG, "rc_style", "GtkStyle");
  return(C_TO_XEN_GtkRcStyle_((GtkRcStyle*)((XEN_TO_C_GtkStyle_(ptr))->rc_style)));
}

static XEN gxg_styles(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkStyle__P(ptr), ptr, XEN_ONLY_ARG, "styles", "GtkStyle");
  return(C_TO_XEN_GSList_((GSList*)((XEN_TO_C_GtkStyle_(ptr))->styles)));
}

#if (!HAVE_GTK_1)
static XEN gxg_property_cache(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkStyle__P(ptr), ptr, XEN_ONLY_ARG, "property_cache", "GtkStyle");
  return(C_TO_XEN_GArray_((GArray*)((XEN_TO_C_GtkStyle_(ptr))->property_cache)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_icon_factories(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkStyle__P(ptr), ptr, XEN_ONLY_ARG, "icon_factories", "GtkStyle");
  return(C_TO_XEN_GSList_((GSList*)((XEN_TO_C_GtkStyle_(ptr))->icon_factories)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_left(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkBorder__P(ptr), ptr, XEN_ONLY_ARG, "left", "GtkBorder");
  return(C_TO_XEN_gint((gint)((XEN_TO_C_GtkBorder_(ptr))->left)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_right(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkBorder__P(ptr), ptr, XEN_ONLY_ARG, "right", "GtkBorder");
  return(C_TO_XEN_gint((gint)((XEN_TO_C_GtkBorder_(ptr))->right)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_top(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkBorder__P(ptr), ptr, XEN_ONLY_ARG, "top", "GtkBorder");
  return(C_TO_XEN_gint((gint)((XEN_TO_C_GtkBorder_(ptr))->top)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_bottom(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkBorder__P(ptr), ptr, XEN_ONLY_ARG, "bottom", "GtkBorder");
  return(C_TO_XEN_gint((gint)((XEN_TO_C_GtkBorder_(ptr))->bottom)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_frame(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkStatusbar__P(ptr) || XEN_GtkWindow__P(ptr), ptr, XEN_ONLY_ARG, "frame", "GtkStatusbar" " or " "GtkWindow");
  if (XEN_GtkStatusbar__P(ptr)) return(C_TO_XEN_GtkWidget_((GtkWidget*)((XEN_TO_C_GtkStatusbar_(ptr))->frame)));
  return(C_TO_XEN_GdkWindow_((GdkWindow*)((XEN_TO_C_GtkWindow_(ptr))->frame)));
}
#endif

static XEN gxg_messages(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkStatusbar__P(ptr), ptr, XEN_ONLY_ARG, "messages", "GtkStatusbar");
  return(C_TO_XEN_GSList_((GSList*)((XEN_TO_C_GtkStatusbar_(ptr))->messages)));
}

static XEN gxg_seq_context_id(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkStatusbar__P(ptr), ptr, XEN_ONLY_ARG, "seq_context_id", "GtkStatusbar");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkStatusbar_(ptr))->seq_context_id)));
}

static XEN gxg_seq_message_id(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkStatusbar__P(ptr), ptr, XEN_ONLY_ARG, "seq_message_id", "GtkStatusbar");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkStatusbar_(ptr))->seq_message_id)));
}

#if (!HAVE_GTK_1)
static XEN gxg_grip_window(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkStatusbar__P(ptr), ptr, XEN_ONLY_ARG, "grip_window", "GtkStatusbar");
  return(C_TO_XEN_GdkWindow_((GdkWindow*)((XEN_TO_C_GtkStatusbar_(ptr))->grip_window)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_has_resize_grip(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkStatusbar__P(ptr), ptr, XEN_ONLY_ARG, "has_resize_grip", "GtkStatusbar");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkStatusbar_(ptr))->has_resize_grip)));
}
#endif

static XEN gxg_rows(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkTable__P(ptr), ptr, XEN_ONLY_ARG, "rows", "GtkTable");
  return(C_TO_XEN_GtkTableRowCol_((GtkTableRowCol*)((XEN_TO_C_GtkTable_(ptr))->rows)));
}

static XEN gxg_cols(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkTable__P(ptr), ptr, XEN_ONLY_ARG, "cols", "GtkTable");
  return(C_TO_XEN_GtkTableRowCol_((GtkTableRowCol*)((XEN_TO_C_GtkTable_(ptr))->cols)));
}

static XEN gxg_nrows(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkTable__P(ptr), ptr, XEN_ONLY_ARG, "nrows", "GtkTable");
  return(C_TO_XEN_guint16((guint16)((XEN_TO_C_GtkTable_(ptr))->nrows)));
}

static XEN gxg_ncols(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkTable__P(ptr), ptr, XEN_ONLY_ARG, "ncols", "GtkTable");
  return(C_TO_XEN_guint16((guint16)((XEN_TO_C_GtkTable_(ptr))->ncols)));
}

static XEN gxg_column_spacing(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkTable__P(ptr), ptr, XEN_ONLY_ARG, "column_spacing", "GtkTable");
  return(C_TO_XEN_guint16((guint16)((XEN_TO_C_GtkTable_(ptr))->column_spacing)));
}

static XEN gxg_row_spacing(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkTable__P(ptr), ptr, XEN_ONLY_ARG, "row_spacing", "GtkTable");
  return(C_TO_XEN_guint16((guint16)((XEN_TO_C_GtkTable_(ptr))->row_spacing)));
}

#if (!HAVE_GTK_1)
static XEN gxg_tag_table(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkTextBuffer__P(ptr), ptr, XEN_ONLY_ARG, "tag_table", "GtkTextBuffer");
  return(C_TO_XEN_GtkTextTagTable_((GtkTextTagTable*)((XEN_TO_C_GtkTextBuffer_(ptr))->tag_table)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_btree(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkTextBuffer__P(ptr), ptr, XEN_ONLY_ARG, "btree", "GtkTextBuffer");
  return(C_TO_XEN_GtkTextBTree_((GtkTextBTree*)((XEN_TO_C_GtkTextBuffer_(ptr))->btree)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_clipboard_contents_buffers(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkTextBuffer__P(ptr), ptr, XEN_ONLY_ARG, "clipboard_contents_buffers", "GtkTextBuffer");
  return(C_TO_XEN_GSList_((GSList*)((XEN_TO_C_GtkTextBuffer_(ptr))->clipboard_contents_buffers)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_selection_clipboards(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkTextBuffer__P(ptr), ptr, XEN_ONLY_ARG, "selection_clipboards", "GtkTextBuffer");
  return(C_TO_XEN_GSList_((GSList*)((XEN_TO_C_GtkTextBuffer_(ptr))->selection_clipboards)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_log_attr_cache(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkTextBuffer__P(ptr), ptr, XEN_ONLY_ARG, "log_attr_cache", "GtkTextBuffer");
  return(C_TO_XEN_GtkTextLogAttrCache_((GtkTextLogAttrCache*)((XEN_TO_C_GtkTextBuffer_(ptr))->log_attr_cache)));
}
#endif

static XEN gxg_draw_indicator(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkToggleButton__P(ptr), ptr, XEN_ONLY_ARG, "draw_indicator", "GtkToggleButton");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkToggleButton_(ptr))->draw_indicator)));
}

static XEN gxg_num_children(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkToolbar__P(ptr), ptr, XEN_ONLY_ARG, "num_children", "GtkToolbar");
  return(C_TO_XEN_gint((gint)((XEN_TO_C_GtkToolbar_(ptr))->num_children)));
}

static XEN gxg_style(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkToolbar__P(ptr) || XEN_GtkWidget__P(ptr), ptr, XEN_ONLY_ARG, "style", "GtkToolbar" " or " "GtkWidget");
  if (XEN_GtkToolbar__P(ptr)) return(C_TO_XEN_GtkToolbarStyle((GtkToolbarStyle)((XEN_TO_C_GtkToolbar_(ptr))->style)));
  return(C_TO_XEN_GtkStyle_((GtkStyle*)((XEN_TO_C_GtkWidget_(ptr))->style)));
}

#if (!HAVE_GTK_1)
static XEN gxg_icon_size(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkToolbar__P(ptr), ptr, XEN_ONLY_ARG, "icon_size", "GtkToolbar");
  return(C_TO_XEN_GtkIconSize((GtkIconSize)((XEN_TO_C_GtkToolbar_(ptr))->icon_size)));
}
#endif

static XEN gxg_tooltips(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkToolbar__P(ptr) || XEN_GtkTooltipsData__P(ptr), ptr, XEN_ONLY_ARG, "tooltips", "GtkToolbar" " or " "GtkTooltipsData");
  if (XEN_GtkToolbar__P(ptr)) return(C_TO_XEN_GtkTooltips_((GtkTooltips*)((XEN_TO_C_GtkToolbar_(ptr))->tooltips)));
  return(C_TO_XEN_GtkTooltips_((GtkTooltips*)((XEN_TO_C_GtkTooltipsData_(ptr))->tooltips)));
}

static XEN gxg_button_maxw(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkToolbar__P(ptr), ptr, XEN_ONLY_ARG, "button_maxw", "GtkToolbar");
  return(C_TO_XEN_gint((gint)((XEN_TO_C_GtkToolbar_(ptr))->button_maxw)));
}

static XEN gxg_button_maxh(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkToolbar__P(ptr), ptr, XEN_ONLY_ARG, "button_maxh", "GtkToolbar");
  return(C_TO_XEN_gint((gint)((XEN_TO_C_GtkToolbar_(ptr))->button_maxh)));
}

#if (!HAVE_GTK_1)
static XEN gxg_style_set_connection(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkToolbar__P(ptr), ptr, XEN_ONLY_ARG, "style_set_connection", "GtkToolbar");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkToolbar_(ptr))->style_set_connection)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_icon_size_connection(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkToolbar__P(ptr), ptr, XEN_ONLY_ARG, "icon_size_connection", "GtkToolbar");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkToolbar_(ptr))->icon_size_connection)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_style_set(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkToolbar__P(ptr), ptr, XEN_ONLY_ARG, "style_set", "GtkToolbar");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkToolbar_(ptr))->style_set)));
}
#endif

static XEN gxg_widget(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkTooltipsData__P(ptr), ptr, XEN_ONLY_ARG, "widget", "GtkTooltipsData");
  return(C_TO_XEN_GtkWidget_((GtkWidget*)((XEN_TO_C_GtkTooltipsData_(ptr))->widget)));
}

static XEN gxg_tip_text(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkTooltipsData__P(ptr), ptr, XEN_ONLY_ARG, "tip_text", "GtkTooltipsData");
  return(C_TO_XEN_gchar_((gchar*)((XEN_TO_C_GtkTooltipsData_(ptr))->tip_text)));
}

static XEN gxg_tip_window(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkTooltips__P(ptr), ptr, XEN_ONLY_ARG, "tip_window", "GtkTooltips");
  return(C_TO_XEN_GtkWidget_((GtkWidget*)((XEN_TO_C_GtkTooltips_(ptr))->tip_window)));
}

#if (!HAVE_GTK_1)
static XEN gxg_tip_label(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkTooltips__P(ptr), ptr, XEN_ONLY_ARG, "tip_label", "GtkTooltips");
  return(C_TO_XEN_GtkWidget_((GtkWidget*)((XEN_TO_C_GtkTooltips_(ptr))->tip_label)));
}
#endif

static XEN gxg_active_tips_data(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkTooltips__P(ptr), ptr, XEN_ONLY_ARG, "active_tips_data", "GtkTooltips");
  return(C_TO_XEN_GtkTooltipsData_((GtkTooltipsData*)((XEN_TO_C_GtkTooltips_(ptr))->active_tips_data)));
}

static XEN gxg_tips_data_list(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkTooltips__P(ptr), ptr, XEN_ONLY_ARG, "tips_data_list", "GtkTooltips");
  return(C_TO_XEN_GList_((GList*)((XEN_TO_C_GtkTooltips_(ptr))->tips_data_list)));
}

static XEN gxg_delay(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkTooltips__P(ptr), ptr, XEN_ONLY_ARG, "delay", "GtkTooltips");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkTooltips_(ptr))->delay)));
}

static XEN gxg_enabled(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkTooltips__P(ptr), ptr, XEN_ONLY_ARG, "enabled", "GtkTooltips");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkTooltips_(ptr))->enabled)));
}

static XEN gxg_timer_tag(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkTooltips__P(ptr), ptr, XEN_ONLY_ARG, "timer_tag", "GtkTooltips");
  return(C_TO_XEN_gint((gint)((XEN_TO_C_GtkTooltips_(ptr))->timer_tag)));
}

#if (!HAVE_GTK_1)
static XEN gxg_use_sticky_delay(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkTooltips__P(ptr), ptr, XEN_ONLY_ARG, "use_sticky_delay", "GtkTooltips");
  return(C_TO_XEN_gboolean((gboolean)((XEN_TO_C_GtkTooltips_(ptr))->use_sticky_delay)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_child_flags(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkTreeModelSort__P(ptr), ptr, XEN_ONLY_ARG, "child_flags", "GtkTreeModelSort");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkTreeModelSort_(ptr))->child_flags)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_child_model(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkTreeModelSort__P(ptr), ptr, XEN_ONLY_ARG, "child_model", "GtkTreeModelSort");
  return(C_TO_XEN_GtkTreeModel_((GtkTreeModel*)((XEN_TO_C_GtkTreeModelSort_(ptr))->child_model)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_zero_ref_count(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkTreeModelSort__P(ptr), ptr, XEN_ONLY_ARG, "zero_ref_count", "GtkTreeModelSort");
  return(C_TO_XEN_gint((gint)((XEN_TO_C_GtkTreeModelSort_(ptr))->zero_ref_count)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_changed_id(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkTreeModelSort__P(ptr), ptr, XEN_ONLY_ARG, "changed_id", "GtkTreeModelSort");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkTreeModelSort_(ptr))->changed_id)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_inserted_id(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkTreeModelSort__P(ptr), ptr, XEN_ONLY_ARG, "inserted_id", "GtkTreeModelSort");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkTreeModelSort_(ptr))->inserted_id)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_has_child_toggled_id(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkTreeModelSort__P(ptr), ptr, XEN_ONLY_ARG, "has_child_toggled_id", "GtkTreeModelSort");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkTreeModelSort_(ptr))->has_child_toggled_id)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_deleted_id(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkTreeModelSort__P(ptr), ptr, XEN_ONLY_ARG, "deleted_id", "GtkTreeModelSort");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkTreeModelSort_(ptr))->deleted_id)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_reordered_id(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkTreeModelSort__P(ptr), ptr, XEN_ONLY_ARG, "reordered_id", "GtkTreeModelSort");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkTreeModelSort_(ptr))->reordered_id)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_tree_view(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkTreeSelection__P(ptr) || XEN_GtkTreeViewColumn__P(ptr), ptr, XEN_ONLY_ARG, "tree_view", "GtkTreeSelection" " or " "GtkTreeViewColumn");
  if (XEN_GtkTreeSelection__P(ptr)) return(C_TO_XEN_GtkTreeView_((GtkTreeView*)((XEN_TO_C_GtkTreeSelection_(ptr))->tree_view)));
  return(C_TO_XEN_GtkWidget_((GtkWidget*)((XEN_TO_C_GtkTreeViewColumn_(ptr))->tree_view)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_user_func(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkTreeSelection__P(ptr), ptr, XEN_ONLY_ARG, "user_func", "GtkTreeSelection");
  return(C_TO_XEN_GtkTreeSelectionFunc((GtkTreeSelectionFunc)((XEN_TO_C_GtkTreeSelection_(ptr))->user_func)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_destroy(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkTreeSelection__P(ptr), ptr, XEN_ONLY_ARG, "destroy", "GtkTreeSelection");
  return(C_TO_XEN_GtkDestroyNotify((GtkDestroyNotify)((XEN_TO_C_GtkTreeSelection_(ptr))->destroy)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_arrow(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkTreeViewColumn__P(ptr), ptr, XEN_ONLY_ARG, "arrow", "GtkTreeViewColumn");
  return(C_TO_XEN_GtkWidget_((GtkWidget*)((XEN_TO_C_GtkTreeViewColumn_(ptr))->arrow)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_alignment(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkTreeViewColumn__P(ptr), ptr, XEN_ONLY_ARG, "alignment", "GtkTreeViewColumn");
  return(C_TO_XEN_GtkWidget_((GtkWidget*)((XEN_TO_C_GtkTreeViewColumn_(ptr))->alignment)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_editable_widget(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkTreeViewColumn__P(ptr), ptr, XEN_ONLY_ARG, "editable_widget", "GtkTreeViewColumn");
  return(C_TO_XEN_GtkCellEditable_((GtkCellEditable*)((XEN_TO_C_GtkTreeViewColumn_(ptr))->editable_widget)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_property_changed_signal(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkTreeViewColumn__P(ptr), ptr, XEN_ONLY_ARG, "property_changed_signal", "GtkTreeViewColumn");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkTreeViewColumn_(ptr))->property_changed_signal)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_column_type(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkTreeViewColumn__P(ptr), ptr, XEN_ONLY_ARG, "column_type", "GtkTreeViewColumn");
  return(C_TO_XEN_GtkTreeViewColumnSizing((GtkTreeViewColumnSizing)((XEN_TO_C_GtkTreeViewColumn_(ptr))->column_type)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_requested_width(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkTreeViewColumn__P(ptr), ptr, XEN_ONLY_ARG, "requested_width", "GtkTreeViewColumn");
  return(C_TO_XEN_gint((gint)((XEN_TO_C_GtkTreeViewColumn_(ptr))->requested_width)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_button_request(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkTreeViewColumn__P(ptr), ptr, XEN_ONLY_ARG, "button_request", "GtkTreeViewColumn");
  return(C_TO_XEN_gint((gint)((XEN_TO_C_GtkTreeViewColumn_(ptr))->button_request)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_resized_width(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkTreeViewColumn__P(ptr), ptr, XEN_ONLY_ARG, "resized_width", "GtkTreeViewColumn");
  return(C_TO_XEN_gint((gint)((XEN_TO_C_GtkTreeViewColumn_(ptr))->resized_width)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_fixed_width(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkTreeViewColumn__P(ptr), ptr, XEN_ONLY_ARG, "fixed_width", "GtkTreeViewColumn");
  return(C_TO_XEN_gint((gint)((XEN_TO_C_GtkTreeViewColumn_(ptr))->fixed_width)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_drag_x(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkTreeViewColumn__P(ptr), ptr, XEN_ONLY_ARG, "drag_x", "GtkTreeViewColumn");
  return(C_TO_XEN_gint((gint)((XEN_TO_C_GtkTreeViewColumn_(ptr))->drag_x)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_drag_y(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkTreeViewColumn__P(ptr), ptr, XEN_ONLY_ARG, "drag_y", "GtkTreeViewColumn");
  return(C_TO_XEN_gint((gint)((XEN_TO_C_GtkTreeViewColumn_(ptr))->drag_y)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_cell_list(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkTreeViewColumn__P(ptr), ptr, XEN_ONLY_ARG, "cell_list", "GtkTreeViewColumn");
  return(C_TO_XEN_GList_((GList*)((XEN_TO_C_GtkTreeViewColumn_(ptr))->cell_list)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_sort_clicked_signal(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkTreeViewColumn__P(ptr), ptr, XEN_ONLY_ARG, "sort_clicked_signal", "GtkTreeViewColumn");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkTreeViewColumn_(ptr))->sort_clicked_signal)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_sort_column_changed_signal(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkTreeViewColumn__P(ptr), ptr, XEN_ONLY_ARG, "sort_column_changed_signal", "GtkTreeViewColumn");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkTreeViewColumn_(ptr))->sort_column_changed_signal)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_sort_order(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkTreeViewColumn__P(ptr), ptr, XEN_ONLY_ARG, "sort_order", "GtkTreeViewColumn");
  return(C_TO_XEN_GtkSortType((GtkSortType)((XEN_TO_C_GtkTreeViewColumn_(ptr))->sort_order)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_resizable(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkTreeViewColumn__P(ptr), ptr, XEN_ONLY_ARG, "resizable", "GtkTreeViewColumn");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkTreeViewColumn_(ptr))->resizable)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_clickable(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkTreeViewColumn__P(ptr), ptr, XEN_ONLY_ARG, "clickable", "GtkTreeViewColumn");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkTreeViewColumn_(ptr))->clickable)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_dirty(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkTreeViewColumn__P(ptr), ptr, XEN_ONLY_ARG, "dirty", "GtkTreeViewColumn");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkTreeViewColumn_(ptr))->dirty)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_show_sort_indicator(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkTreeViewColumn__P(ptr), ptr, XEN_ONLY_ARG, "show_sort_indicator", "GtkTreeViewColumn");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkTreeViewColumn_(ptr))->show_sort_indicator)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_maybe_reordered(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkTreeViewColumn__P(ptr), ptr, XEN_ONLY_ARG, "maybe_reordered", "GtkTreeViewColumn");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkTreeViewColumn_(ptr))->maybe_reordered)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_reorderable(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkTreeViewColumn__P(ptr), ptr, XEN_ONLY_ARG, "reorderable", "GtkTreeViewColumn");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkTreeViewColumn_(ptr))->reorderable)));
}
#endif

static XEN gxg_saved_state(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkWidget__P(ptr), ptr, XEN_ONLY_ARG, "saved_state", "GtkWidget");
  return(C_TO_XEN_guint8((guint8)((XEN_TO_C_GtkWidget_(ptr))->saved_state)));
}

#if (!HAVE_GTK_1)
static XEN gxg_wm_role(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkWindow__P(ptr), ptr, XEN_ONLY_ARG, "wm_role", "GtkWindow");
  return(C_TO_XEN_gchar_((gchar*)((XEN_TO_C_GtkWindow_(ptr))->wm_role)));
}
#endif

static XEN gxg_focus_widget(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkWindow__P(ptr), ptr, XEN_ONLY_ARG, "focus_widget", "GtkWindow");
  return(C_TO_XEN_GtkWidget_((GtkWidget*)((XEN_TO_C_GtkWindow_(ptr))->focus_widget)));
}

static XEN gxg_default_widget(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkWindow__P(ptr), ptr, XEN_ONLY_ARG, "default_widget", "GtkWindow");
  return(C_TO_XEN_GtkWidget_((GtkWidget*)((XEN_TO_C_GtkWindow_(ptr))->default_widget)));
}

static XEN gxg_transient_parent(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkWindow__P(ptr), ptr, XEN_ONLY_ARG, "transient_parent", "GtkWindow");
  return(C_TO_XEN_GtkWindow_((GtkWindow*)((XEN_TO_C_GtkWindow_(ptr))->transient_parent)));
}

#if (!HAVE_GTK_1)
static XEN gxg_configure_request_count(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkWindow__P(ptr), ptr, XEN_ONLY_ARG, "configure_request_count", "GtkWindow");
  return(C_TO_XEN_guint16((guint16)((XEN_TO_C_GtkWindow_(ptr))->configure_request_count)));
}
#endif

static XEN gxg_allow_shrink(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkWindow__P(ptr), ptr, XEN_ONLY_ARG, "allow_shrink", "GtkWindow");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkWindow_(ptr))->allow_shrink)));
}

static XEN gxg_allow_grow(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkWindow__P(ptr), ptr, XEN_ONLY_ARG, "allow_grow", "GtkWindow");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkWindow_(ptr))->allow_grow)));
}

#if (!HAVE_GTK_1)
static XEN gxg_configure_notify_received(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkWindow__P(ptr), ptr, XEN_ONLY_ARG, "configure_notify_received", "GtkWindow");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkWindow_(ptr))->configure_notify_received)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_need_default_position(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkWindow__P(ptr), ptr, XEN_ONLY_ARG, "need_default_position", "GtkWindow");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkWindow_(ptr))->need_default_position)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_need_default_size(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkWindow__P(ptr), ptr, XEN_ONLY_ARG, "need_default_size", "GtkWindow");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkWindow_(ptr))->need_default_size)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_has_focus(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkWindow__P(ptr), ptr, XEN_ONLY_ARG, "has_focus", "GtkWindow");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkWindow_(ptr))->has_focus)));
}
#endif

static XEN gxg_modal(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkWindow__P(ptr), ptr, XEN_ONLY_ARG, "modal", "GtkWindow");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkWindow_(ptr))->modal)));
}

#if (!HAVE_GTK_1)
static XEN gxg_destroy_with_parent(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkWindow__P(ptr), ptr, XEN_ONLY_ARG, "destroy_with_parent", "GtkWindow");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkWindow_(ptr))->destroy_with_parent)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_iconify_initially(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkWindow__P(ptr), ptr, XEN_ONLY_ARG, "iconify_initially", "GtkWindow");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkWindow_(ptr))->iconify_initially)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_stick_initially(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkWindow__P(ptr), ptr, XEN_ONLY_ARG, "stick_initially", "GtkWindow");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkWindow_(ptr))->stick_initially)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_maximize_initially(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkWindow__P(ptr), ptr, XEN_ONLY_ARG, "maximize_initially", "GtkWindow");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkWindow_(ptr))->maximize_initially)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_decorated(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkWindow__P(ptr), ptr, XEN_ONLY_ARG, "decorated", "GtkWindow");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkWindow_(ptr))->decorated)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_type_hint(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkWindow__P(ptr), ptr, XEN_ONLY_ARG, "type_hint", "GtkWindow");
  return(C_TO_XEN_GdkWindowTypeHint((GdkWindowTypeHint)((XEN_TO_C_GtkWindow_(ptr))->type_hint)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_gravity(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkWindow__P(ptr), ptr, XEN_ONLY_ARG, "gravity", "GtkWindow");
  return(C_TO_XEN_GdkGravity((GdkGravity)((XEN_TO_C_GtkWindow_(ptr))->gravity)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_frame_left(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkWindow__P(ptr), ptr, XEN_ONLY_ARG, "frame_left", "GtkWindow");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkWindow_(ptr))->frame_left)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_frame_top(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkWindow__P(ptr), ptr, XEN_ONLY_ARG, "frame_top", "GtkWindow");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkWindow_(ptr))->frame_top)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_frame_right(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkWindow__P(ptr), ptr, XEN_ONLY_ARG, "frame_right", "GtkWindow");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkWindow_(ptr))->frame_right)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_frame_bottom(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkWindow__P(ptr), ptr, XEN_ONLY_ARG, "frame_bottom", "GtkWindow");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkWindow_(ptr))->frame_bottom)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_keys_changed_handler(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkWindow__P(ptr), ptr, XEN_ONLY_ARG, "keys_changed_handler", "GtkWindow");
  return(C_TO_XEN_guint((guint)((XEN_TO_C_GtkWindow_(ptr))->keys_changed_handler)));
}
#endif

#if (!HAVE_GTK_1)
static XEN gxg_mnemonic_modifier(XEN ptr)
{
  XEN_ASSERT_TYPE(XEN_GtkWindow__P(ptr), ptr, XEN_ONLY_ARG, "mnemonic_modifier", "GtkWindow");
  return(C_TO_XEN_GdkModifierType((GdkModifierType)((XEN_TO_C_GtkWindow_(ptr))->mnemonic_modifier)));
}
#endif
#if (!HAVE_GTK_1)
static XEN gxg_make_GdkCursor(XEN arglist)
{
  GdkCursor* result;
  int i, len;
  result = (GdkCursor*)CALLOC(1, sizeof(GdkCursor));
  len = XEN_LIST_LENGTH(arglist);
  for (i = 0; i < len; i++)
    switch (i)
      {
      case 0: result->type = XEN_TO_C_GdkCursorType(XEN_LIST_REF(arglist, 0));
      case 1: result->ref_count = XEN_TO_C_guint(XEN_LIST_REF(arglist, 1));
      }
  return(XEN_LIST_3(C_STRING_TO_XEN_SYMBOL("GdkCursor_"), C_TO_XEN_ULONG((unsigned long)result), make_xm_obj(result)));
}

static XEN gxg_make_GdkPoint(XEN arglist)
{
  GdkPoint* result;
  int i, len;
  result = (GdkPoint*)CALLOC(1, sizeof(GdkPoint));
  len = XEN_LIST_LENGTH(arglist);
  for (i = 0; i < len; i++)
    switch (i)
      {
      case 0: result->x = XEN_TO_C_gint(XEN_LIST_REF(arglist, 0));
      case 1: result->y = XEN_TO_C_gint(XEN_LIST_REF(arglist, 1));
      }
  return(XEN_LIST_3(C_STRING_TO_XEN_SYMBOL("GdkPoint_"), C_TO_XEN_ULONG((unsigned long)result), make_xm_obj(result)));
}

static XEN gxg_make_GdkRectangle(XEN arglist)
{
  GdkRectangle* result;
  int i, len;
  result = (GdkRectangle*)CALLOC(1, sizeof(GdkRectangle));
  len = XEN_LIST_LENGTH(arglist);
  for (i = 0; i < len; i++)
    switch (i)
      {
      case 0: result->x = XEN_TO_C_gint(XEN_LIST_REF(arglist, 0));
      case 1: result->y = XEN_TO_C_gint(XEN_LIST_REF(arglist, 1));
      case 2: result->width = XEN_TO_C_gint(XEN_LIST_REF(arglist, 2));
      case 3: result->height = XEN_TO_C_gint(XEN_LIST_REF(arglist, 3));
      }
  return(XEN_LIST_3(C_STRING_TO_XEN_SYMBOL("GdkRectangle_"), C_TO_XEN_ULONG((unsigned long)result), make_xm_obj(result)));
}

static XEN gxg_make_GdkSegment(XEN arglist)
{
  GdkSegment* result;
  int i, len;
  result = (GdkSegment*)CALLOC(1, sizeof(GdkSegment));
  len = XEN_LIST_LENGTH(arglist);
  for (i = 0; i < len; i++)
    switch (i)
      {
      case 0: result->x1 = XEN_TO_C_gint(XEN_LIST_REF(arglist, 0));
      case 1: result->y1 = XEN_TO_C_gint(XEN_LIST_REF(arglist, 1));
      case 2: result->x2 = XEN_TO_C_gint(XEN_LIST_REF(arglist, 2));
      case 3: result->y2 = XEN_TO_C_gint(XEN_LIST_REF(arglist, 3));
      }
  return(XEN_LIST_3(C_STRING_TO_XEN_SYMBOL("GdkSegment_"), C_TO_XEN_ULONG((unsigned long)result), make_xm_obj(result)));
}

static XEN gxg_make_GdkSpan(XEN arglist)
{
  GdkSpan* result;
  int i, len;
  result = (GdkSpan*)CALLOC(1, sizeof(GdkSpan));
  len = XEN_LIST_LENGTH(arglist);
  for (i = 0; i < len; i++)
    switch (i)
      {
      case 0: result->x = XEN_TO_C_gint(XEN_LIST_REF(arglist, 0));
      case 1: result->y = XEN_TO_C_gint(XEN_LIST_REF(arglist, 1));
      case 2: result->width = XEN_TO_C_gint(XEN_LIST_REF(arglist, 2));
      }
  return(XEN_LIST_3(C_STRING_TO_XEN_SYMBOL("GdkSpan_"), C_TO_XEN_ULONG((unsigned long)result), make_xm_obj(result)));
}

static XEN gxg_make_GtkTextIter(void)
{
  GtkTextIter* result;
  result = (GtkTextIter*)CALLOC(1, sizeof(GtkTextIter));
  return(XEN_LIST_3(C_STRING_TO_XEN_SYMBOL("GtkTextIter_"), C_TO_XEN_ULONG((unsigned long)result), make_xm_obj(result)));
}

static XEN gxg_make_GtkTextMark(void)
{
  GtkTextMark* result;
  result = (GtkTextMark*)CALLOC(1, sizeof(GtkTextMark));
  return(XEN_LIST_3(C_STRING_TO_XEN_SYMBOL("GtkTextMark_"), C_TO_XEN_ULONG((unsigned long)result), make_xm_obj(result)));
}

static XEN gxg_make_GtkTextChildAnchor(void)
{
  GtkTextChildAnchor* result;
  result = (GtkTextChildAnchor*)CALLOC(1, sizeof(GtkTextChildAnchor));
  return(XEN_LIST_3(C_STRING_TO_XEN_SYMBOL("GtkTextChildAnchor_"), C_TO_XEN_ULONG((unsigned long)result), make_xm_obj(result)));
}

#endif

#if HAVE_GUILE
static void define_structs(void)
{

  XEN_DEFINE_PROCEDURE(XG_PRE "c-array->list" XG_POST, c_array_to_xen_list, 2, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "list->c-array" XG_POST, xen_list_to_c_array, 2, 0, 0, NULL);

#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "mnemonic_modifier" XG_POST, gxg_mnemonic_modifier, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "keys_changed_handler" XG_POST, gxg_keys_changed_handler, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "frame_bottom" XG_POST, gxg_frame_bottom, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "frame_right" XG_POST, gxg_frame_right, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "frame_top" XG_POST, gxg_frame_top, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "frame_left" XG_POST, gxg_frame_left, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "gravity" XG_POST, gxg_gravity, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "type_hint" XG_POST, gxg_type_hint, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "decorated" XG_POST, gxg_decorated, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "maximize_initially" XG_POST, gxg_maximize_initially, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "stick_initially" XG_POST, gxg_stick_initially, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "iconify_initially" XG_POST, gxg_iconify_initially, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "destroy_with_parent" XG_POST, gxg_destroy_with_parent, 1, 0, 0, NULL);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "modal" XG_POST, gxg_modal, 1, 0, 0, NULL);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "has_focus" XG_POST, gxg_has_focus, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "need_default_size" XG_POST, gxg_need_default_size, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "need_default_position" XG_POST, gxg_need_default_position, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "configure_notify_received" XG_POST, gxg_configure_notify_received, 1, 0, 0, NULL);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "allow_grow" XG_POST, gxg_allow_grow, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "allow_shrink" XG_POST, gxg_allow_shrink, 1, 0, 0, NULL);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "configure_request_count" XG_POST, gxg_configure_request_count, 1, 0, 0, NULL);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "transient_parent" XG_POST, gxg_transient_parent, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "default_widget" XG_POST, gxg_default_widget, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "focus_widget" XG_POST, gxg_focus_widget, 1, 0, 0, NULL);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "wm_role" XG_POST, gxg_wm_role, 1, 0, 0, NULL);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "saved_state" XG_POST, gxg_saved_state, 1, 0, 0, NULL);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "reorderable" XG_POST, gxg_reorderable, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "maybe_reordered" XG_POST, gxg_maybe_reordered, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "show_sort_indicator" XG_POST, gxg_show_sort_indicator, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "dirty" XG_POST, gxg_dirty, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "clickable" XG_POST, gxg_clickable, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "resizable" XG_POST, gxg_resizable, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "sort_order" XG_POST, gxg_sort_order, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "sort_column_changed_signal" XG_POST, gxg_sort_column_changed_signal, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "sort_clicked_signal" XG_POST, gxg_sort_clicked_signal, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "cell_list" XG_POST, gxg_cell_list, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "drag_y" XG_POST, gxg_drag_y, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "drag_x" XG_POST, gxg_drag_x, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "fixed_width" XG_POST, gxg_fixed_width, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "resized_width" XG_POST, gxg_resized_width, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "button_request" XG_POST, gxg_button_request, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "requested_width" XG_POST, gxg_requested_width, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "column_type" XG_POST, gxg_column_type, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "property_changed_signal" XG_POST, gxg_property_changed_signal, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "editable_widget" XG_POST, gxg_editable_widget, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "alignment" XG_POST, gxg_alignment, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "arrow" XG_POST, gxg_arrow, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "destroy" XG_POST, gxg_destroy, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "user_func" XG_POST, gxg_user_func, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "tree_view" XG_POST, gxg_tree_view, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "reordered_id" XG_POST, gxg_reordered_id, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "deleted_id" XG_POST, gxg_deleted_id, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "has_child_toggled_id" XG_POST, gxg_has_child_toggled_id, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "inserted_id" XG_POST, gxg_inserted_id, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "changed_id" XG_POST, gxg_changed_id, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "zero_ref_count" XG_POST, gxg_zero_ref_count, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "child_model" XG_POST, gxg_child_model, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "child_flags" XG_POST, gxg_child_flags, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "use_sticky_delay" XG_POST, gxg_use_sticky_delay, 1, 0, 0, NULL);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "timer_tag" XG_POST, gxg_timer_tag, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "enabled" XG_POST, gxg_enabled, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "delay" XG_POST, gxg_delay, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "tips_data_list" XG_POST, gxg_tips_data_list, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "active_tips_data" XG_POST, gxg_active_tips_data, 1, 0, 0, NULL);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "tip_label" XG_POST, gxg_tip_label, 1, 0, 0, NULL);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "tip_window" XG_POST, gxg_tip_window, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "tip_text" XG_POST, gxg_tip_text, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "widget" XG_POST, gxg_widget, 1, 0, 0, NULL);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "style_set" XG_POST, gxg_style_set, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "icon_size_connection" XG_POST, gxg_icon_size_connection, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "style_set_connection" XG_POST, gxg_style_set_connection, 1, 0, 0, NULL);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "button_maxh" XG_POST, gxg_button_maxh, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "button_maxw" XG_POST, gxg_button_maxw, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "tooltips" XG_POST, gxg_tooltips, 1, 0, 0, NULL);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "icon_size" XG_POST, gxg_icon_size, 1, 0, 0, NULL);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "style" XG_POST, gxg_style, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "num_children" XG_POST, gxg_num_children, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "draw_indicator" XG_POST, gxg_draw_indicator, 1, 0, 0, NULL);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "log_attr_cache" XG_POST, gxg_log_attr_cache, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "selection_clipboards" XG_POST, gxg_selection_clipboards, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "clipboard_contents_buffers" XG_POST, gxg_clipboard_contents_buffers, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "btree" XG_POST, gxg_btree, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "tag_table" XG_POST, gxg_tag_table, 1, 0, 0, NULL);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "row_spacing" XG_POST, gxg_row_spacing, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "column_spacing" XG_POST, gxg_column_spacing, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "ncols" XG_POST, gxg_ncols, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "nrows" XG_POST, gxg_nrows, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "cols" XG_POST, gxg_cols, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "rows" XG_POST, gxg_rows, 1, 0, 0, NULL);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "has_resize_grip" XG_POST, gxg_has_resize_grip, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "grip_window" XG_POST, gxg_grip_window, 1, 0, 0, NULL);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "seq_message_id" XG_POST, gxg_seq_message_id, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "seq_context_id" XG_POST, gxg_seq_context_id, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "messages" XG_POST, gxg_messages, 1, 0, 0, NULL);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "frame" XG_POST, gxg_frame, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "bottom" XG_POST, gxg_bottom, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "top" XG_POST, gxg_top, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "right" XG_POST, gxg_right, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "left" XG_POST, gxg_left, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "icon_factories" XG_POST, gxg_icon_factories, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "property_cache" XG_POST, gxg_property_cache, 1, 0, 0, NULL);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "styles" XG_POST, gxg_styles, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "rc_style" XG_POST, gxg_rc_style, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "attach_count" XG_POST, gxg_attach_count, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "white_gc" XG_POST, gxg_white_gc, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "black_gc" XG_POST, gxg_black_gc, 1, 0, 0, NULL);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "text_aa_gc" XG_POST, gxg_text_aa_gc, 1, 0, 0, NULL);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "base_gc" XG_POST, gxg_base_gc, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "text_gc" XG_POST, gxg_text_gc, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "mid_gc" XG_POST, gxg_mid_gc, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "dark_gc" XG_POST, gxg_dark_gc, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "light_gc" XG_POST, gxg_light_gc, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "bg_gc" XG_POST, gxg_bg_gc, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "fg_gc" XG_POST, gxg_fg_gc, 1, 0, 0, NULL);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "ythickness" XG_POST, gxg_ythickness, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "xthickness" XG_POST, gxg_xthickness, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "font_desc" XG_POST, gxg_font_desc, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "text_aa" XG_POST, gxg_text_aa, 1, 0, 0, NULL);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "base" XG_POST, gxg_base, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "mid" XG_POST, gxg_mid, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "dark" XG_POST, gxg_dark, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "light" XG_POST, gxg_light, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "bg" XG_POST, gxg_bg, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "fg" XG_POST, gxg_fg, 1, 0, 0, NULL);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "translation_domain" XG_POST, gxg_translation_domain, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "modifier" XG_POST, gxg_modifier, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "stock_id" XG_POST, gxg_stock_id, 1, 0, 0, NULL);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "numeric" XG_POST, gxg_numeric, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "timer_calls" XG_POST, gxg_timer_calls, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "timer_step" XG_POST, gxg_timer_step, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "climb_rate" XG_POST, gxg_climb_rate, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "panel" XG_POST, gxg_panel, 1, 0, 0, NULL);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "is_mapped" XG_POST, gxg_is_mapped, 1, 0, 0, NULL);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "need_map" XG_POST, gxg_need_map, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "have_size" XG_POST, gxg_have_size, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "focus_in" XG_POST, gxg_focus_in, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "same_app" XG_POST, gxg_same_app, 1, 0, 0, NULL);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "xembed_version" XG_POST, gxg_xembed_version, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "plug_widget" XG_POST, gxg_plug_widget, 1, 0, 0, NULL);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "plug_window" XG_POST, gxg_plug_window, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "current_height" XG_POST, gxg_current_height, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "current_width" XG_POST, gxg_current_width, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "request_height" XG_POST, gxg_request_height, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "request_width" XG_POST, gxg_request_width, 1, 0, 0, NULL);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "have_height" XG_POST, gxg_have_height, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "have_width" XG_POST, gxg_have_width, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "widgets" XG_POST, gxg_widgets, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "rc_context" XG_POST, gxg_rc_context, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "property_values" XG_POST, gxg_property_values, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "queued_settings" XG_POST, gxg_queued_settings, 1, 0, 0, NULL);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "info" XG_POST, gxg_info, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "data" XG_POST, gxg_data, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "format" XG_POST, gxg_format, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "window_placement" XG_POST, gxg_window_placement, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "vscrollbar_visible" XG_POST, gxg_vscrollbar_visible, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "hscrollbar_visible" XG_POST, gxg_hscrollbar_visible, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "vscrollbar_policy" XG_POST, gxg_vscrollbar_policy, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "hscrollbar_policy" XG_POST, gxg_hscrollbar_policy, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "vscrollbar" XG_POST, gxg_vscrollbar, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "hscrollbar" XG_POST, gxg_hscrollbar, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "draw_value" XG_POST, gxg_draw_value, 1, 0, 0, NULL);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "digits" XG_POST, gxg_digits, 1, 0, 0, NULL);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "max_size" XG_POST, gxg_max_size, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "position" XG_POST, gxg_position, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "slider_size" XG_POST, gxg_slider_size, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "ysrc" XG_POST, gxg_ysrc, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "xsrc" XG_POST, gxg_xsrc, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "metric" XG_POST, gxg_metric, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "non_gr_exp_gc" XG_POST, gxg_non_gr_exp_gc, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "backing_store" XG_POST, gxg_backing_store, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "subdivide" XG_POST, gxg_subdivide, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "ruler_scale" XG_POST, gxg_ruler_scale, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "pixels_per_unit" XG_POST, gxg_pixels_per_unit, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "abbrev" XG_POST, gxg_abbrev, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "metric_name" XG_POST, gxg_metric_name, 1, 0, 0, NULL);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "update_timeout_id" XG_POST, gxg_update_timeout_id, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "slide_initial_coordinate" XG_POST, gxg_slide_initial_coordinate, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "slide_initial_slider_position" XG_POST, gxg_slide_initial_slider_position, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "update_pending" XG_POST, gxg_update_pending, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "trough_click_forward" XG_POST, gxg_trough_click_forward, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "round_digits" XG_POST, gxg_round_digits, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "slider_end" XG_POST, gxg_slider_end, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "slider_start" XG_POST, gxg_slider_start, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "orientation" XG_POST, gxg_orientation, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "min_slider_size" XG_POST, gxg_min_slider_size, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "slider_size_fixed" XG_POST, gxg_slider_size_fixed, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "need_recalc" XG_POST, gxg_need_recalc, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "has_stepper_d" XG_POST, gxg_has_stepper_d, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "has_stepper_c" XG_POST, gxg_has_stepper_c, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "has_stepper_b" XG_POST, gxg_has_stepper_b, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "has_stepper_a" XG_POST, gxg_has_stepper_a, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "flippable" XG_POST, gxg_flippable, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "inverted" XG_POST, gxg_inverted, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "update_policy" XG_POST, gxg_update_policy, 1, 0, 0, NULL);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "adjustment" XG_POST, gxg_adjustment, 1, 0, 0, NULL);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "modality_group" XG_POST, gxg_modality_group, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "modality_window" XG_POST, gxg_modality_window, 1, 0, 0, NULL);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "socket_window" XG_POST, gxg_socket_window, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "child2_resize" XG_POST, gxg_child2_resize, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "child2_shrink" XG_POST, gxg_child2_shrink, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "child1_resize" XG_POST, gxg_child1_resize, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "child1_shrink" XG_POST, gxg_child1_shrink, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "position_set" XG_POST, gxg_position_set, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "max_position" XG_POST, gxg_max_position, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "min_position" XG_POST, gxg_min_position, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "last_allocation" XG_POST, gxg_last_allocation, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "child1_size" XG_POST, gxg_child1_size, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "xor_gc" XG_POST, gxg_xor_gc, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "handle" XG_POST, gxg_handle, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "child2" XG_POST, gxg_child2, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "child1" XG_POST, gxg_child1, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "menu_item" XG_POST, gxg_menu_item, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "flags" XG_POST, gxg_flags, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "child_has_focus" XG_POST, gxg_child_has_focus, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "need_timer" XG_POST, gxg_need_timer, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "click_child" XG_POST, gxg_click_child, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "in_child" XG_POST, gxg_in_child, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "scrollable" XG_POST, gxg_scrollable, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "tab_pos" XG_POST, gxg_tab_pos, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "show_border" XG_POST, gxg_show_border, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "homogeneous" XG_POST, gxg_homogeneous, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "show_tabs" XG_POST, gxg_show_tabs, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "tab_vborder" XG_POST, gxg_tab_vborder, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "tab_hborder" XG_POST, gxg_tab_hborder, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "menu" XG_POST, gxg_menu, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "focus_tab" XG_POST, gxg_focus_tab, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "first_tab" XG_POST, gxg_first_tab, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "cur_page" XG_POST, gxg_cur_page, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "ignore_enter" XG_POST, gxg_ignore_enter, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "menu_flag" XG_POST, gxg_menu_flag, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "ignore_leave" XG_POST, gxg_ignore_leave, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "have_xgrab" XG_POST, gxg_have_xgrab, 1, 0, 0, NULL);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "have_grab" XG_POST, gxg_have_grab, 1, 0, 0, NULL);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "activate_time" XG_POST, gxg_activate_time, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "parent_menu_shell" XG_POST, gxg_parent_menu_shell, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "active_menu_item" XG_POST, gxg_active_menu_item, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "timer" XG_POST, gxg_timer, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "right_justify" XG_POST, gxg_right_justify, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "submenu_direction" XG_POST, gxg_submenu_direction, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "submenu_placement" XG_POST, gxg_submenu_placement, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "show_submenu_indicator" XG_POST, gxg_show_submenu_indicator, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "accelerator_width" XG_POST, gxg_accelerator_width, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "submenu" XG_POST, gxg_submenu, 1, 0, 0, NULL);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "upper_arrow_prelight" XG_POST, gxg_upper_arrow_prelight, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "lower_arrow_visible" XG_POST, gxg_lower_arrow_visible, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "upper_arrow_visible" XG_POST, gxg_upper_arrow_visible, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "scroll_fast" XG_POST, gxg_scroll_fast, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "tearoff_active" XG_POST, gxg_tearoff_active, 1, 0, 0, NULL);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "torn_off" XG_POST, gxg_torn_off, 1, 0, 0, NULL);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "needs_destruction_ref_count" XG_POST, gxg_needs_destruction_ref_count, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "navigation_timeout" XG_POST, gxg_navigation_timeout, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "navigation_region" XG_POST, gxg_navigation_region, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "timeout_id" XG_POST, gxg_timeout_id, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "scroll_step" XG_POST, gxg_scroll_step, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "saved_scroll_offset" XG_POST, gxg_saved_scroll_offset, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "view_window" XG_POST, gxg_view_window, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "tearoff_adjustment" XG_POST, gxg_tearoff_adjustment, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "tearoff_scrollbar" XG_POST, gxg_tearoff_scrollbar, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "tearoff_hbox" XG_POST, gxg_tearoff_hbox, 1, 0, 0, NULL);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "tearoff_window" XG_POST, gxg_tearoff_window, 1, 0, 0, NULL);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "toplevel" XG_POST, gxg_toplevel, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "toggle_size" XG_POST, gxg_toggle_size, 1, 0, 0, NULL);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "position_func_data" XG_POST, gxg_position_func_data, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "position_func" XG_POST, gxg_position_func, 1, 0, 0, NULL);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "accel_path" XG_POST, gxg_accel_path, 1, 0, 0, NULL);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "old_active_menu_item" XG_POST, gxg_old_active_menu_item, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "parent_menu_item" XG_POST, gxg_parent_menu_item, 1, 0, 0, NULL);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "default_sort_destroy" XG_POST, gxg_default_sort_destroy, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "default_sort_data" XG_POST, gxg_default_sort_data, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "default_sort_func" XG_POST, gxg_default_sort_func, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "column_headers" XG_POST, gxg_column_headers, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "order" XG_POST, gxg_order, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "sort_column_id" XG_POST, gxg_sort_column_id, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "n_columns" XG_POST, gxg_n_columns, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "sort_list" XG_POST, gxg_sort_list, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "tail" XG_POST, gxg_tail, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "root" XG_POST, gxg_root, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "stamp" XG_POST, gxg_stamp, 1, 0, 0, NULL);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "freeze_count" XG_POST, gxg_freeze_count, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "scroll_y" XG_POST, gxg_scroll_y, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "scroll_x" XG_POST, gxg_scroll_x, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "visibility" XG_POST, gxg_visibility, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "vadjustment" XG_POST, gxg_vadjustment, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "hadjustment" XG_POST, gxg_hadjustment, 1, 0, 0, NULL);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "select_info" XG_POST, gxg_select_info, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "mnemonic_window" XG_POST, gxg_mnemonic_window, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "mnemonic_widget" XG_POST, gxg_mnemonic_widget, 1, 0, 0, NULL);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "layout" XG_POST, gxg_layout, 1, 0, 0, NULL);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "effective_attrs" XG_POST, gxg_effective_attrs, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "attrs" XG_POST, gxg_attrs, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "mnemonic_keyval" XG_POST, gxg_mnemonic_keyval, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "use_markup" XG_POST, gxg_use_markup, 1, 0, 0, NULL);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "wrap" XG_POST, gxg_wrap, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "jtype" XG_POST, gxg_jtype, 1, 0, 0, NULL);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "label" XG_POST, gxg_label, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "extra_data" XG_POST, gxg_extra_data, 1, 0, 0, NULL);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "item_type" XG_POST, gxg_item_type, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "callback_action" XG_POST, gxg_callback_action, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "callback" XG_POST, gxg_callback, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "accelerator" XG_POST, gxg_accelerator, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "path" XG_POST, gxg_path, 1, 0, 0, NULL);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "gtk_reserved1" XG_POST, gxg_gtk_reserved1, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "has_user_ref_count" XG_POST, gxg_has_user_ref_count, 1, 0, 0, NULL);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "keys_listbox" XG_POST, gxg_keys_listbox, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "keys_list" XG_POST, gxg_keys_list, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "current_device" XG_POST, gxg_current_device, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "axis_items" XG_POST, gxg_axis_items, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "save_button" XG_POST, gxg_save_button, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "close_button" XG_POST, gxg_close_button, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "mode_optionmenu" XG_POST, gxg_mode_optionmenu, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "axis_listbox" XG_POST, gxg_axis_listbox, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "axis_list" XG_POST, gxg_axis_list, 1, 0, 0, NULL);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "context_id" XG_POST, gxg_context_id, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "client_window" XG_POST, gxg_client_window, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "slave" XG_POST, gxg_slave, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "image" XG_POST, gxg_image, 1, 0, 0, NULL);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "deskoff_y" XG_POST, gxg_deskoff_y, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "deskoff_x" XG_POST, gxg_deskoff_x, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "snap_edge" XG_POST, gxg_snap_edge, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "shrink_on_detach" XG_POST, gxg_shrink_on_detach, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "child_detached" XG_POST, gxg_child_detached, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "float_window_mapped" XG_POST, gxg_float_window_mapped, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "handle_position" XG_POST, gxg_handle_position, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "float_window" XG_POST, gxg_float_window, 1, 0, 0, NULL);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "bin_window" XG_POST, gxg_bin_window, 1, 0, 0, NULL);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "gamma_text" XG_POST, gxg_gamma_text, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "gamma_dialog" XG_POST, gxg_gamma_dialog, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "gamma" XG_POST, gxg_gamma, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "curve" XG_POST, gxg_curve, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "table" XG_POST, gxg_table, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "label_yalign" XG_POST, gxg_label_yalign, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "label_xalign" XG_POST, gxg_label_xalign, 1, 0, 0, NULL);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "label_widget" XG_POST, gxg_label_widget, 1, 0, 0, NULL);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "auto_resize" XG_POST, gxg_auto_resize, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "dialog_width" XG_POST, gxg_dialog_width, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "apply_button" XG_POST, gxg_apply_button, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "fontsel" XG_POST, gxg_fontsel, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "face" XG_POST, gxg_face, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "preview_entry" XG_POST, gxg_preview_entry, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "filter_button" XG_POST, gxg_filter_button, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "points_button" XG_POST, gxg_points_button, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "pixels_button" XG_POST, gxg_pixels_button, 1, 0, 0, NULL);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "size_list" XG_POST, gxg_size_list, 1, 0, 0, NULL);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "size_entry" XG_POST, gxg_size_entry, 1, 0, 0, NULL);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "face_list" XG_POST, gxg_face_list, 1, 0, 0, NULL);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "font_style_entry" XG_POST, gxg_font_style_entry, 1, 0, 0, NULL);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "family_list" XG_POST, gxg_family_list, 1, 0, 0, NULL);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "font_entry" XG_POST, gxg_font_entry, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "button_area" XG_POST, gxg_button_area, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "fileop_ren_file" XG_POST, gxg_fileop_ren_file, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "fileop_del_file" XG_POST, gxg_fileop_del_file, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "fileop_c_dir" XG_POST, gxg_fileop_c_dir, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "cmpl_state" XG_POST, gxg_cmpl_state, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "fileop_file" XG_POST, gxg_fileop_file, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "fileop_entry" XG_POST, gxg_fileop_entry, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "fileop_dialog" XG_POST, gxg_fileop_dialog, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "history_list" XG_POST, gxg_history_list, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "history_menu" XG_POST, gxg_history_menu, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "history_pulldown" XG_POST, gxg_history_pulldown, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "main_vbox" XG_POST, gxg_main_vbox, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "selection_text" XG_POST, gxg_selection_text, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "selection_entry" XG_POST, gxg_selection_entry, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "file_list" XG_POST, gxg_file_list, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "dir_list" XG_POST, gxg_dir_list, 1, 0, 0, NULL);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "cursor_gc" XG_POST, gxg_cursor_gc, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "width_chars" XG_POST, gxg_width_chars, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "invisible_char" XG_POST, gxg_invisible_char, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "drag_start_y" XG_POST, gxg_drag_start_y, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "drag_start_x" XG_POST, gxg_drag_start_x, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "dnd_position" XG_POST, gxg_dnd_position, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "preedit_cursor" XG_POST, gxg_preedit_cursor, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "preedit_length" XG_POST, gxg_preedit_length, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "n_bytes" XG_POST, gxg_n_bytes, 1, 0, 0, NULL);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "text_size" XG_POST, gxg_text_size, 1, 0, 0, NULL);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "descent" XG_POST, gxg_descent, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "ascent" XG_POST, gxg_ascent, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "scroll_offset" XG_POST, gxg_scroll_offset, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "recompute_idle" XG_POST, gxg_recompute_idle, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "blink_timeout" XG_POST, gxg_blink_timeout, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "mouse_cursor_obscured" XG_POST, gxg_mouse_cursor_obscured, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "editing_canceled" XG_POST, gxg_editing_canceled, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "is_cell_renderer" XG_POST, gxg_is_cell_renderer, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "in_click" XG_POST, gxg_in_click, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "cursor_visible" XG_POST, gxg_cursor_visible, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "activates_default" XG_POST, gxg_activates_default, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "has_frame" XG_POST, gxg_has_frame, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "need_im_reset" XG_POST, gxg_need_im_reset, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "cache_includes_preedit" XG_POST, gxg_cache_includes_preedit, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "cached_layout" XG_POST, gxg_cached_layout, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "selection_bound" XG_POST, gxg_selection_bound, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "current_pos" XG_POST, gxg_current_pos, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "popup_menu" XG_POST, gxg_popup_menu, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "im_context" XG_POST, gxg_im_context, 1, 0, 0, NULL);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "text_area" XG_POST, gxg_text_area, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "text_max_length" XG_POST, gxg_text_max_length, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "text_length" XG_POST, gxg_text_length, 1, 0, 0, NULL);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "in_drag" XG_POST, gxg_in_drag, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "overwrite_mode" XG_POST, gxg_overwrite_mode, 1, 0, 0, NULL);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "draw_data" XG_POST, gxg_draw_data, 1, 0, 0, NULL);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "separator" XG_POST, gxg_separator, 1, 0, 0, NULL);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "action_area" XG_POST, gxg_action_area, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "vbox" XG_POST, gxg_vbox, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "num_ctlpoints" XG_POST, gxg_num_ctlpoints, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "point" XG_POST, gxg_point, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "num_points" XG_POST, gxg_num_points, 1, 0, 0, NULL);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "last" XG_POST, gxg_last, 1, 0, 0, NULL);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "grab_point" XG_POST, gxg_grab_point, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "curve_type" XG_POST, gxg_curve_type, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "pixmap" XG_POST, gxg_pixmap, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "max_y" XG_POST, gxg_max_y, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "min_y" XG_POST, gxg_min_y, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "max_x" XG_POST, gxg_max_x, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "min_x" XG_POST, gxg_min_x, 1, 0, 0, NULL);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "cursor_type" XG_POST, gxg_cursor_type, 1, 0, 0, NULL);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "reallocate_redraws" XG_POST, gxg_reallocate_redraws, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "resize_mode" XG_POST, gxg_resize_mode, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "need_resize" XG_POST, gxg_need_resize, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "border_width" XG_POST, gxg_border_width, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "focus_child" XG_POST, gxg_focus_child, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "activate_id" XG_POST, gxg_activate_id, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "current_button" XG_POST, gxg_current_button, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "use_arrows_always" XG_POST, gxg_use_arrows_always, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "use_arrows" XG_POST, gxg_use_arrows, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "case_sensitive" XG_POST, gxg_case_sensitive, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "ok_if_empty" XG_POST, gxg_ok_if_empty, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "value_in_list" XG_POST, gxg_value_in_list, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "list_change_id" XG_POST, gxg_list_change_id, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "entry_change_id" XG_POST, gxg_entry_change_id, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "list" XG_POST, gxg_list, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "popwin" XG_POST, gxg_popwin, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "popup" XG_POST, gxg_popup, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "entry" XG_POST, gxg_entry, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "help_button" XG_POST, gxg_help_button, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "cancel_button" XG_POST, gxg_cancel_button, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "ok_button" XG_POST, gxg_ok_button, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "colorsel" XG_POST, gxg_colorsel, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "always_show_toggle" XG_POST, gxg_always_show_toggle, 1, 0, 0, NULL);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "pixbuf_expander_closed" XG_POST, gxg_pixbuf_expander_closed, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "pixbuf_expander_open" XG_POST, gxg_pixbuf_expander_open, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "pixbuf" XG_POST, gxg_pixbuf, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "activatable" XG_POST, gxg_activatable, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "active" XG_POST, gxg_active, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "editable_set" XG_POST, gxg_editable_set, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "strikethrough_set" XG_POST, gxg_strikethrough_set, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "rise_set" XG_POST, gxg_rise_set, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "underline_set" XG_POST, gxg_underline_set, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "background_set" XG_POST, gxg_background_set, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "foreground_set" XG_POST, gxg_foreground_set, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "scale_set" XG_POST, gxg_scale_set, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "editable" XG_POST, gxg_editable, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "strikethrough" XG_POST, gxg_strikethrough, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "fixed_height_rows" XG_POST, gxg_fixed_height_rows, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "rise" XG_POST, gxg_rise, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "underline_style" XG_POST, gxg_underline_style, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "extra_attrs" XG_POST, gxg_extra_attrs, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "font_scale" XG_POST, gxg_font_scale, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "font" XG_POST, gxg_font, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "text" XG_POST, gxg_text, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "is_expander" XG_POST, gxg_is_expander, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "visible" XG_POST, gxg_visible, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "ypad" XG_POST, gxg_ypad, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "xpad" XG_POST, gxg_xpad, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "use_stock" XG_POST, gxg_use_stock, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "use_underline" XG_POST, gxg_use_underline, 1, 0, 0, NULL);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "relief" XG_POST, gxg_relief, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "button_down" XG_POST, gxg_button_down, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "in_button" XG_POST, gxg_in_button, 1, 0, 0, NULL);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "constructed" XG_POST, gxg_constructed, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "activate_timeout" XG_POST, gxg_activate_timeout, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "label_text" XG_POST, gxg_label_text, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "event_window" XG_POST, gxg_event_window, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "spacing" XG_POST, gxg_spacing, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "child" XG_POST, gxg_child, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "shadow_type" XG_POST, gxg_shadow_type, 1, 0, 0, NULL);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "arrow_type" XG_POST, gxg_arrow_type, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "obey_child" XG_POST, gxg_obey_child, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "ratio" XG_POST, gxg_ratio, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "yscale" XG_POST, gxg_yscale, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "xscale" XG_POST, gxg_xscale, 1, 0, 0, NULL);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "yalign" XG_POST, gxg_yalign, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "xalign" XG_POST, gxg_xalign, 1, 0, 0, NULL);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "page_size" XG_POST, gxg_page_size, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "page_increment" XG_POST, gxg_page_increment, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "step_increment" XG_POST, gxg_step_increment, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "value" XG_POST, gxg_value, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "upper" XG_POST, gxg_upper, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "lower" XG_POST, gxg_lower, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "accel_string_width" XG_POST, gxg_accel_string_width, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "accel_string" XG_POST, gxg_accel_string, 1, 0, 0, NULL);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "accel_group" XG_POST, gxg_accel_group, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "accel_closure" XG_POST, gxg_accel_closure, 1, 0, 0, NULL);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "accel_widget" XG_POST, gxg_accel_widget, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "accel_padding" XG_POST, gxg_accel_padding, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "queue_id" XG_POST, gxg_queue_id, 1, 0, 0, NULL);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "priv_accels" XG_POST, gxg_priv_accels, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "n_accels" XG_POST, gxg_n_accels, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "acceleratables" XG_POST, gxg_acceleratables, 1, 0, 0, NULL);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "modifier_mask" XG_POST, gxg_modifier_mask, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "lock_count" XG_POST, gxg_lock_count, 1, 0, 0, NULL);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "modal_hint" XG_POST, gxg_modal_hint, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "input_only" XG_POST, gxg_input_only, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "guffaw_gravity" XG_POST, gxg_guffaw_gravity, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "resize_count" XG_POST, gxg_resize_count, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "update_freeze_count" XG_POST, gxg_update_freeze_count, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "update_area" XG_POST, gxg_update_area, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "paint_stack" XG_POST, gxg_paint_stack, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "bg_pixmap" XG_POST, gxg_bg_pixmap, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "children" XG_POST, gxg_children, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "filters" XG_POST, gxg_filters, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "extension_events" XG_POST, gxg_extension_events, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "user_data" XG_POST, gxg_user_data, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "parent" XG_POST, gxg_parent, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "impl" XG_POST, gxg_impl, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "win_gravity" XG_POST, gxg_win_gravity, 1, 0, 0, NULL);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "max_aspect" XG_POST, gxg_max_aspect, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "min_aspect" XG_POST, gxg_min_aspect, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "height_inc" XG_POST, gxg_height_inc, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "width_inc" XG_POST, gxg_width_inc, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "base_height" XG_POST, gxg_base_height, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "base_width" XG_POST, gxg_base_width, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "max_height" XG_POST, gxg_max_height, 1, 0, 0, NULL);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "max_width" XG_POST, gxg_max_width, 1, 0, 0, NULL);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "min_height" XG_POST, gxg_min_height, 1, 0, 0, NULL);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "min_width" XG_POST, gxg_min_width, 1, 0, 0, NULL);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "override_redirect" XG_POST, gxg_override_redirect, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "wmclass_class" XG_POST, gxg_wmclass_class, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "wmclass_name" XG_POST, gxg_wmclass_name, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "cursor" XG_POST, gxg_cursor, 1, 0, 0, NULL);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "window_type" XG_POST, gxg_window_type, 1, 0, 0, NULL);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "event_mask" XG_POST, gxg_event_mask, 1, 0, 0, NULL);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "title" XG_POST, gxg_title, 1, 0, 0, NULL);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "blue_prec" XG_POST, gxg_blue_prec, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "blue_shift" XG_POST, gxg_blue_shift, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "blue_mask" XG_POST, gxg_blue_mask, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "green_prec" XG_POST, gxg_green_prec, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "green_shift" XG_POST, gxg_green_shift, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "green_mask" XG_POST, gxg_green_mask, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "red_prec" XG_POST, gxg_red_prec, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "red_shift" XG_POST, gxg_red_shift, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "red_mask" XG_POST, gxg_red_mask, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "bits_per_rgb" XG_POST, gxg_bits_per_rgb, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "colormap_size" XG_POST, gxg_colormap_size, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "y2" XG_POST, gxg_y2, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "x2" XG_POST, gxg_x2, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "y1" XG_POST, gxg_y1, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "x1" XG_POST, gxg_x1, 1, 0, 0, NULL);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "info_list" XG_POST, gxg_info_list, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "n_colors" XG_POST, gxg_n_colors, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "level" XG_POST, gxg_level, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "keycode" XG_POST, gxg_keycode, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "keys" XG_POST, gxg_keys, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "num_keys" XG_POST, gxg_num_keys, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "num_axes" XG_POST, gxg_num_axes, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "has_cursor" XG_POST, gxg_has_cursor, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "source" XG_POST, gxg_source, 1, 0, 0, NULL);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "mem" XG_POST, gxg_mem, 1, 0, 0, NULL);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "bits_per_pixel" XG_POST, gxg_bits_per_pixel, 1, 0, 0, NULL);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "bpl" XG_POST, gxg_bpl, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "bpp" XG_POST, gxg_bpp, 1, 0, 0, NULL);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "depth" XG_POST, gxg_depth, 1, 0, 0, NULL);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "byte_order" XG_POST, gxg_byte_order, 1, 0, 0, NULL);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "colormap" XG_POST, gxg_colormap, 1, 0, 0, NULL);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "join_style" XG_POST, gxg_join_style, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "cap_style" XG_POST, gxg_cap_style, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "line_style" XG_POST, gxg_line_style, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "line_width" XG_POST, gxg_line_width, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "graphics_exposures" XG_POST, gxg_graphics_exposures, 1, 0, 0, NULL);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "clip_y_origin" XG_POST, gxg_clip_y_origin, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "clip_x_origin" XG_POST, gxg_clip_x_origin, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "ts_y_origin" XG_POST, gxg_ts_y_origin, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "ts_x_origin" XG_POST, gxg_ts_x_origin, 1, 0, 0, NULL);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "subwindow_mode" XG_POST, gxg_subwindow_mode, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "clip_mask" XG_POST, gxg_clip_mask, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "stipple" XG_POST, gxg_stipple, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "tile" XG_POST, gxg_tile, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "fill" XG_POST, gxg_fill, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "function" XG_POST, gxg_function, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "context" XG_POST, gxg_context, 1, 0, 0, NULL);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "new_window_state" XG_POST, gxg_new_window_state, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "changed_mask" XG_POST, gxg_changed_mask, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "name" XG_POST, gxg_name, 1, 0, 0, NULL);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "requestor" XG_POST, gxg_requestor, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "property" XG_POST, gxg_property, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "target" XG_POST, gxg_target, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "selection" XG_POST, gxg_selection, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "atom" XG_POST, gxg_atom, 1, 0, 0, NULL);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "height" XG_POST, gxg_height, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "width" XG_POST, gxg_width, 1, 0, 0, NULL);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "in" XG_POST, gxg_in, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "focus" XG_POST, gxg_focus, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "detail" XG_POST, gxg_detail, 1, 0, 0, NULL);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "mode" XG_POST, gxg_mode, 1, 0, 0, NULL);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "subwindow" XG_POST, gxg_subwindow, 1, 0, 0, NULL);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "group" XG_POST, gxg_group, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "hardware_keycode" XG_POST, gxg_hardware_keycode, 1, 0, 0, NULL);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "string" XG_POST, gxg_string, 1, 0, 0, NULL);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "length" XG_POST, gxg_length, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "keyval" XG_POST, gxg_keyval, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "direction" XG_POST, gxg_direction, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "button" XG_POST, gxg_button, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "y_root" XG_POST, gxg_y_root, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "x_root" XG_POST, gxg_x_root, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "device" XG_POST, gxg_device, 1, 0, 0, NULL);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "is_hint" XG_POST, gxg_is_hint, 1, 0, 0, NULL);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "axes" XG_POST, gxg_axes, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "y" XG_POST, gxg_y, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "x" XG_POST, gxg_x, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "time" XG_POST, gxg_time, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "state" XG_POST, gxg_state, 1, 0, 0, NULL);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "count" XG_POST, gxg_count, 1, 0, 0, NULL);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "region" XG_POST, gxg_region, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "send_event" XG_POST, gxg_send_event, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "window" XG_POST, gxg_window, 1, 0, 0, NULL);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "start_time" XG_POST, gxg_start_time, 1, 0, 0, NULL);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "action" XG_POST, gxg_action, 1, 0, 0, NULL);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "suggested_action" XG_POST, gxg_suggested_action, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "actions" XG_POST, gxg_actions, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "targets" XG_POST, gxg_targets, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "dest_window" XG_POST, gxg_dest_window, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "source_window" XG_POST, gxg_source_window, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "is_source" XG_POST, gxg_is_source, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "protocol" XG_POST, gxg_protocol, 1, 0, 0, NULL);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "ref_count" XG_POST, gxg_ref_count, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "type" XG_POST, gxg_type, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "windowing_data" XG_POST, gxg_windowing_data, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "visual" XG_POST, gxg_visual, 1, 0, 0, NULL);
#endif
  XEN_DEFINE_PROCEDURE(XG_PRE "colors" XG_POST, gxg_colors, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "size" XG_POST, gxg_size, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "blue" XG_POST, gxg_blue, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "green" XG_POST, gxg_green, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "red" XG_POST, gxg_red, 1, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "pixel" XG_POST, gxg_pixel, 1, 0, 0, NULL);
#if (!HAVE_GTK_1)
  XEN_DEFINE_PROCEDURE(XG_PRE "GdkCursor" XG_POST, gxg_make_GdkCursor, 0, 0, 1, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GdkPoint" XG_POST, gxg_make_GdkPoint, 0, 0, 1, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GdkRectangle" XG_POST, gxg_make_GdkRectangle, 0, 0, 1, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GdkSegment" XG_POST, gxg_make_GdkSegment, 0, 0, 1, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GdkSpan" XG_POST, gxg_make_GdkSpan, 0, 0, 1, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GtkTextIter" XG_POST, gxg_make_GtkTextIter, 0, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GtkTextMark" XG_POST, gxg_make_GtkTextMark, 0, 0, 0, NULL);
  XEN_DEFINE_PROCEDURE(XG_PRE "GtkTextChildAnchor" XG_POST, gxg_make_GtkTextChildAnchor, 0, 0, 0, NULL);
#endif

}

#else
  #include "xg-ruby.c"
#endif
/* ---------------------------------------- macros ---------------------------------------- */

static void define_macros(void)
{
#if (!HAVE_GTK_1)
#if (!(defined(GDK_DISABLE_DEPRECATED))) && (!(defined(GTK_DISABLE_DEPRECATED))) && (!(defined(GDK_PIXBUF_DISABLE_DEPRECATED)))
  XEN_EVAL_C_STRING("(define " XG_PRE "gdk_cursor_destroy" XG_POST " " XG_PRE "gdk_cursor_unref" XG_POST ")");
  XEN_EVAL_C_STRING("(define " XG_PRE "gdk_draw_pixmap" XG_POST " " XG_PRE "gdk_draw_drawable" XG_POST ")");
  XEN_EVAL_C_STRING("(define " XG_PRE "gdk_draw_bitmap" XG_POST " " XG_PRE "gdk_draw_drawable" XG_POST ")");
  XEN_EVAL_C_STRING("(define " XG_PRE "gdk_gc_destroy" XG_POST " " XG_PRE "gdk_gc_unref" XG_POST ")");
  XEN_EVAL_C_STRING("(define " XG_PRE "gdk_image_destroy" XG_POST " " XG_PRE "gdk_image_unref" XG_POST ")");
  XEN_EVAL_C_STRING("(define " XG_PRE "gdk_rgb_get_cmap" XG_POST " " XG_PRE "gdk_rgb_get_colormap" XG_POST ")");
  XEN_EVAL_C_STRING("(define " XG_PRE "gdk_visual_ref" XG_POST " " XG_PRE "g_object_ref" XG_POST ")");
  XEN_EVAL_C_STRING("(define " XG_PRE "gdk_visual_unref" XG_POST " " XG_PRE "g_object_unref" XG_POST ")");
  XEN_EVAL_C_STRING("(define " XG_PRE "gdk_window_get_size" XG_POST " " XG_PRE "gdk_drawable_get_size" XG_POST ")");
  XEN_EVAL_C_STRING("(define " XG_PRE "gdk_window_get_type" XG_POST " " XG_PRE "gdk_window_get_window_type" XG_POST ")");
  XEN_EVAL_C_STRING("(define " XG_PRE "gdk_window_get_colormap" XG_POST " " XG_PRE "gdk_drawable_get_colormap" XG_POST ")");
  XEN_EVAL_C_STRING("(define " XG_PRE "gdk_window_set_colormap" XG_POST " " XG_PRE "gdk_drawable_set_colormap" XG_POST ")");
  XEN_EVAL_C_STRING("(define " XG_PRE "gdk_window_get_visual" XG_POST " " XG_PRE "gdk_drawable_get_visual" XG_POST ")");
  XEN_EVAL_C_STRING("(define " XG_PRE "gdk_window_ref" XG_POST " " XG_PRE "gdk_drawable_ref" XG_POST ")");
  XEN_EVAL_C_STRING("(define " XG_PRE "gdk_window_unref" XG_POST " " XG_PRE "gdk_drawable_unref" XG_POST ")");
  XEN_EVAL_C_STRING("(define " XG_PRE "gtk_accel_group_ref" XG_POST " " XG_PRE "g_object_ref" XG_POST ")");
  XEN_EVAL_C_STRING("(define " XG_PRE "gtk_accel_group_unref" XG_POST " " XG_PRE "g_object_unref" XG_POST ")");
  XEN_EVAL_C_STRING("(define " XG_PRE "gtk_accel_label_accelerator_width" XG_POST " " XG_PRE "gtk_accel_label_get_accel_width" XG_POST ")");
  XEN_EVAL_C_STRING("(define " XG_PRE "gtk_binding_entry_add" XG_POST " " XG_PRE "gtk_binding_entry_clear" XG_POST ")");
  XEN_EVAL_C_STRING("(define " XG_PRE "gtk_check_menu_item_set_state" XG_POST " " XG_PRE "gtk_check_menu_item_set_active" XG_POST ")");
  XEN_EVAL_C_STRING("(define " XG_PRE "gtk_container_children" XG_POST " " XG_PRE "gtk_container_get_children" XG_POST ")");
  XEN_EVAL_C_STRING("(define " XG_PRE "gtk_label_set" XG_POST " " XG_PRE "gtk_label_set_text" XG_POST ")");
  XEN_EVAL_C_STRING("(define " XG_PRE "gtk_notebook_current_page" XG_POST " " XG_PRE "gtk_notebook_get_current_page" XG_POST ")");
  XEN_EVAL_C_STRING("(define " XG_PRE "gtk_notebook_set_page" XG_POST " " XG_PRE "gtk_notebook_set_current_page" XG_POST ")");
  XEN_EVAL_C_STRING("(define " XG_PRE "gtk_radio_button_group" XG_POST " " XG_PRE "gtk_radio_button_get_group" XG_POST ")");
  XEN_EVAL_C_STRING("(define " XG_PRE "gtk_radio_menu_item_group" XG_POST " " XG_PRE "gtk_radio_menu_item_get_group" XG_POST ")");
  XEN_EVAL_C_STRING("(define " XG_PRE "gtk_spin_button_get_value_as_float" XG_POST " " XG_PRE "gtk_spin_button_get_value" XG_POST ")");
  XEN_EVAL_C_STRING("(define " XG_PRE "gtk_toggle_button_set_state" XG_POST " " XG_PRE "gtk_toggle_button_set_active" XG_POST ")");
  XEN_EVAL_C_STRING("(define " XG_PRE "gtk_window_position" XG_POST " " XG_PRE "gtk_window_set_position" XG_POST ")");
#endif

#endif
}

/* ---------------------------------------- constants ---------------------------------------- */

static void define_integers(void)
{

#if HAVE_GUILE
#if HAVE_SCM_C_DEFINE
  #define DEFINE_INTEGER(Name, Value) scm_c_define(Name, scm_long2num(Value))
  #define DEFINE_ULONG(Name, Value) scm_c_define(Name, scm_ulong2num(Value))
#else
  #define DEFINE_INTEGER(Name, Value) gh_define(Name, scm_long2num(Value))
  #define DEFINE_ULONG(Name, Value) gh_define(Name, scm_ulong2num(Value))
#endif
#else
  #define DEFINE_INTEGER(Name, Value) rb_define_global_const(Name, C_TO_XEN_INT(Value))
  #define DEFINE_ULONG(Name, Value) rb_define_global_const(Name, C_TO_XEN_ULONG(Value))
#endif

#if (!HAVE_GTK_1)
  g_type_init();
#endif

#if (!HAVE_GTK_1)
  DEFINE_INTEGER(XG_PRE "G_SIGNAL_RUN_FIRST" XG_POST,                           G_SIGNAL_RUN_FIRST);
  DEFINE_INTEGER(XG_PRE "G_SIGNAL_RUN_LAST" XG_POST,                            G_SIGNAL_RUN_LAST);
  DEFINE_INTEGER(XG_PRE "G_SIGNAL_RUN_CLEANUP" XG_POST,                         G_SIGNAL_RUN_CLEANUP);
  DEFINE_INTEGER(XG_PRE "G_SIGNAL_NO_RECURSE" XG_POST,                          G_SIGNAL_NO_RECURSE);
  DEFINE_INTEGER(XG_PRE "G_SIGNAL_DETAILED" XG_POST,                            G_SIGNAL_DETAILED);
  DEFINE_INTEGER(XG_PRE "G_SIGNAL_ACTION" XG_POST,                              G_SIGNAL_ACTION);
  DEFINE_INTEGER(XG_PRE "G_SIGNAL_NO_HOOKS" XG_POST,                            G_SIGNAL_NO_HOOKS);
  DEFINE_INTEGER(XG_PRE "G_CONNECT_AFTER" XG_POST,                              G_CONNECT_AFTER);
  DEFINE_INTEGER(XG_PRE "G_CONNECT_SWAPPED" XG_POST,                            G_CONNECT_SWAPPED);
  DEFINE_INTEGER(XG_PRE "G_SIGNAL_MATCH_ID" XG_POST,                            G_SIGNAL_MATCH_ID);
  DEFINE_INTEGER(XG_PRE "G_SIGNAL_MATCH_DETAIL" XG_POST,                        G_SIGNAL_MATCH_DETAIL);
  DEFINE_INTEGER(XG_PRE "G_SIGNAL_MATCH_CLOSURE" XG_POST,                       G_SIGNAL_MATCH_CLOSURE);
  DEFINE_INTEGER(XG_PRE "G_SIGNAL_MATCH_FUNC" XG_POST,                          G_SIGNAL_MATCH_FUNC);
  DEFINE_INTEGER(XG_PRE "G_SIGNAL_MATCH_DATA" XG_POST,                          G_SIGNAL_MATCH_DATA);
  DEFINE_INTEGER(XG_PRE "G_SIGNAL_MATCH_UNBLOCKED" XG_POST,                     G_SIGNAL_MATCH_UNBLOCKED);
#endif
  DEFINE_INTEGER(XG_PRE "GDK_X_CURSOR" XG_POST,                                 GDK_X_CURSOR);
  DEFINE_INTEGER(XG_PRE "GDK_ARROW" XG_POST,                                    GDK_ARROW);
  DEFINE_INTEGER(XG_PRE "GDK_BASED_ARROW_DOWN" XG_POST,                         GDK_BASED_ARROW_DOWN);
  DEFINE_INTEGER(XG_PRE "GDK_BASED_ARROW_UP" XG_POST,                           GDK_BASED_ARROW_UP);
  DEFINE_INTEGER(XG_PRE "GDK_BOAT" XG_POST,                                     GDK_BOAT);
  DEFINE_INTEGER(XG_PRE "GDK_BOGOSITY" XG_POST,                                 GDK_BOGOSITY);
  DEFINE_INTEGER(XG_PRE "GDK_BOTTOM_LEFT_CORNER" XG_POST,                       GDK_BOTTOM_LEFT_CORNER);
  DEFINE_INTEGER(XG_PRE "GDK_BOTTOM_RIGHT_CORNER" XG_POST,                      GDK_BOTTOM_RIGHT_CORNER);
  DEFINE_INTEGER(XG_PRE "GDK_BOTTOM_SIDE" XG_POST,                              GDK_BOTTOM_SIDE);
  DEFINE_INTEGER(XG_PRE "GDK_BOTTOM_TEE" XG_POST,                               GDK_BOTTOM_TEE);
  DEFINE_INTEGER(XG_PRE "GDK_BOX_SPIRAL" XG_POST,                               GDK_BOX_SPIRAL);
  DEFINE_INTEGER(XG_PRE "GDK_CENTER_PTR" XG_POST,                               GDK_CENTER_PTR);
  DEFINE_INTEGER(XG_PRE "GDK_CIRCLE" XG_POST,                                   GDK_CIRCLE);
  DEFINE_INTEGER(XG_PRE "GDK_CLOCK" XG_POST,                                    GDK_CLOCK);
  DEFINE_INTEGER(XG_PRE "GDK_COFFEE_MUG" XG_POST,                               GDK_COFFEE_MUG);
  DEFINE_INTEGER(XG_PRE "GDK_CROSS" XG_POST,                                    GDK_CROSS);
  DEFINE_INTEGER(XG_PRE "GDK_CROSS_REVERSE" XG_POST,                            GDK_CROSS_REVERSE);
  DEFINE_INTEGER(XG_PRE "GDK_CROSSHAIR" XG_POST,                                GDK_CROSSHAIR);
  DEFINE_INTEGER(XG_PRE "GDK_DIAMOND_CROSS" XG_POST,                            GDK_DIAMOND_CROSS);
  DEFINE_INTEGER(XG_PRE "GDK_DOT" XG_POST,                                      GDK_DOT);
  DEFINE_INTEGER(XG_PRE "GDK_DOTBOX" XG_POST,                                   GDK_DOTBOX);
  DEFINE_INTEGER(XG_PRE "GDK_DOUBLE_ARROW" XG_POST,                             GDK_DOUBLE_ARROW);
  DEFINE_INTEGER(XG_PRE "GDK_DRAFT_LARGE" XG_POST,                              GDK_DRAFT_LARGE);
  DEFINE_INTEGER(XG_PRE "GDK_DRAFT_SMALL" XG_POST,                              GDK_DRAFT_SMALL);
  DEFINE_INTEGER(XG_PRE "GDK_DRAPED_BOX" XG_POST,                               GDK_DRAPED_BOX);
  DEFINE_INTEGER(XG_PRE "GDK_EXCHANGE" XG_POST,                                 GDK_EXCHANGE);
  DEFINE_INTEGER(XG_PRE "GDK_FLEUR" XG_POST,                                    GDK_FLEUR);
  DEFINE_INTEGER(XG_PRE "GDK_GOBBLER" XG_POST,                                  GDK_GOBBLER);
  DEFINE_INTEGER(XG_PRE "GDK_GUMBY" XG_POST,                                    GDK_GUMBY);
  DEFINE_INTEGER(XG_PRE "GDK_HAND1" XG_POST,                                    GDK_HAND1);
  DEFINE_INTEGER(XG_PRE "GDK_HAND2" XG_POST,                                    GDK_HAND2);
  DEFINE_INTEGER(XG_PRE "GDK_HEART" XG_POST,                                    GDK_HEART);
  DEFINE_INTEGER(XG_PRE "GDK_ICON" XG_POST,                                     GDK_ICON);
  DEFINE_INTEGER(XG_PRE "GDK_IRON_CROSS" XG_POST,                               GDK_IRON_CROSS);
  DEFINE_INTEGER(XG_PRE "GDK_LEFT_PTR" XG_POST,                                 GDK_LEFT_PTR);
  DEFINE_INTEGER(XG_PRE "GDK_LEFT_SIDE" XG_POST,                                GDK_LEFT_SIDE);
  DEFINE_INTEGER(XG_PRE "GDK_LEFT_TEE" XG_POST,                                 GDK_LEFT_TEE);
  DEFINE_INTEGER(XG_PRE "GDK_LEFTBUTTON" XG_POST,                               GDK_LEFTBUTTON);
  DEFINE_INTEGER(XG_PRE "GDK_LL_ANGLE" XG_POST,                                 GDK_LL_ANGLE);
  DEFINE_INTEGER(XG_PRE "GDK_LR_ANGLE" XG_POST,                                 GDK_LR_ANGLE);
  DEFINE_INTEGER(XG_PRE "GDK_MAN" XG_POST,                                      GDK_MAN);
  DEFINE_INTEGER(XG_PRE "GDK_MIDDLEBUTTON" XG_POST,                             GDK_MIDDLEBUTTON);
  DEFINE_INTEGER(XG_PRE "GDK_MOUSE" XG_POST,                                    GDK_MOUSE);
  DEFINE_INTEGER(XG_PRE "GDK_PENCIL" XG_POST,                                   GDK_PENCIL);
  DEFINE_INTEGER(XG_PRE "GDK_PIRATE" XG_POST,                                   GDK_PIRATE);
  DEFINE_INTEGER(XG_PRE "GDK_PLUS" XG_POST,                                     GDK_PLUS);
  DEFINE_INTEGER(XG_PRE "GDK_QUESTION_ARROW" XG_POST,                           GDK_QUESTION_ARROW);
  DEFINE_INTEGER(XG_PRE "GDK_RIGHT_PTR" XG_POST,                                GDK_RIGHT_PTR);
  DEFINE_INTEGER(XG_PRE "GDK_RIGHT_SIDE" XG_POST,                               GDK_RIGHT_SIDE);
  DEFINE_INTEGER(XG_PRE "GDK_RIGHT_TEE" XG_POST,                                GDK_RIGHT_TEE);
  DEFINE_INTEGER(XG_PRE "GDK_RIGHTBUTTON" XG_POST,                              GDK_RIGHTBUTTON);
  DEFINE_INTEGER(XG_PRE "GDK_RTL_LOGO" XG_POST,                                 GDK_RTL_LOGO);
  DEFINE_INTEGER(XG_PRE "GDK_SAILBOAT" XG_POST,                                 GDK_SAILBOAT);
  DEFINE_INTEGER(XG_PRE "GDK_SB_DOWN_ARROW" XG_POST,                            GDK_SB_DOWN_ARROW);
  DEFINE_INTEGER(XG_PRE "GDK_SB_H_DOUBLE_ARROW" XG_POST,                        GDK_SB_H_DOUBLE_ARROW);
  DEFINE_INTEGER(XG_PRE "GDK_SB_LEFT_ARROW" XG_POST,                            GDK_SB_LEFT_ARROW);
  DEFINE_INTEGER(XG_PRE "GDK_SB_RIGHT_ARROW" XG_POST,                           GDK_SB_RIGHT_ARROW);
  DEFINE_INTEGER(XG_PRE "GDK_SB_UP_ARROW" XG_POST,                              GDK_SB_UP_ARROW);
  DEFINE_INTEGER(XG_PRE "GDK_SB_V_DOUBLE_ARROW" XG_POST,                        GDK_SB_V_DOUBLE_ARROW);
  DEFINE_INTEGER(XG_PRE "GDK_SHUTTLE" XG_POST,                                  GDK_SHUTTLE);
  DEFINE_INTEGER(XG_PRE "GDK_SIZING" XG_POST,                                   GDK_SIZING);
  DEFINE_INTEGER(XG_PRE "GDK_SPIDER" XG_POST,                                   GDK_SPIDER);
  DEFINE_INTEGER(XG_PRE "GDK_SPRAYCAN" XG_POST,                                 GDK_SPRAYCAN);
  DEFINE_INTEGER(XG_PRE "GDK_STAR" XG_POST,                                     GDK_STAR);
  DEFINE_INTEGER(XG_PRE "GDK_TARGET" XG_POST,                                   GDK_TARGET);
  DEFINE_INTEGER(XG_PRE "GDK_TCROSS" XG_POST,                                   GDK_TCROSS);
  DEFINE_INTEGER(XG_PRE "GDK_TOP_LEFT_ARROW" XG_POST,                           GDK_TOP_LEFT_ARROW);
  DEFINE_INTEGER(XG_PRE "GDK_TOP_LEFT_CORNER" XG_POST,                          GDK_TOP_LEFT_CORNER);
  DEFINE_INTEGER(XG_PRE "GDK_TOP_RIGHT_CORNER" XG_POST,                         GDK_TOP_RIGHT_CORNER);
  DEFINE_INTEGER(XG_PRE "GDK_TOP_SIDE" XG_POST,                                 GDK_TOP_SIDE);
  DEFINE_INTEGER(XG_PRE "GDK_TOP_TEE" XG_POST,                                  GDK_TOP_TEE);
  DEFINE_INTEGER(XG_PRE "GDK_TREK" XG_POST,                                     GDK_TREK);
  DEFINE_INTEGER(XG_PRE "GDK_UL_ANGLE" XG_POST,                                 GDK_UL_ANGLE);
  DEFINE_INTEGER(XG_PRE "GDK_UMBRELLA" XG_POST,                                 GDK_UMBRELLA);
  DEFINE_INTEGER(XG_PRE "GDK_UR_ANGLE" XG_POST,                                 GDK_UR_ANGLE);
  DEFINE_INTEGER(XG_PRE "GDK_WATCH" XG_POST,                                    GDK_WATCH);
  DEFINE_INTEGER(XG_PRE "GDK_XTERM" XG_POST,                                    GDK_XTERM);
#if (!HAVE_GTK_1)
  DEFINE_INTEGER(XG_PRE "GDK_LAST_CURSOR " XG_POST,                             GDK_LAST_CURSOR );
#endif
  DEFINE_INTEGER(XG_PRE "GDK_CURSOR_IS_PIXMAP" XG_POST,                         GDK_CURSOR_IS_PIXMAP);
  DEFINE_INTEGER(XG_PRE "GDK_ACTION_DEFAULT" XG_POST,                           GDK_ACTION_DEFAULT);
  DEFINE_INTEGER(XG_PRE "GDK_ACTION_COPY" XG_POST,                              GDK_ACTION_COPY);
  DEFINE_INTEGER(XG_PRE "GDK_ACTION_MOVE" XG_POST,                              GDK_ACTION_MOVE);
  DEFINE_INTEGER(XG_PRE "GDK_ACTION_LINK" XG_POST,                              GDK_ACTION_LINK);
  DEFINE_INTEGER(XG_PRE "GDK_ACTION_PRIVATE" XG_POST,                           GDK_ACTION_PRIVATE);
  DEFINE_INTEGER(XG_PRE "GDK_ACTION_ASK" XG_POST,                               GDK_ACTION_ASK);
  DEFINE_INTEGER(XG_PRE "GDK_DRAG_PROTO_MOTIF" XG_POST,                         GDK_DRAG_PROTO_MOTIF);
  DEFINE_INTEGER(XG_PRE "GDK_DRAG_PROTO_XDND" XG_POST,                          GDK_DRAG_PROTO_XDND);
  DEFINE_INTEGER(XG_PRE "GDK_DRAG_PROTO_ROOTWIN" XG_POST,                       GDK_DRAG_PROTO_ROOTWIN);
  DEFINE_INTEGER(XG_PRE "GDK_DRAG_PROTO_NONE" XG_POST,                          GDK_DRAG_PROTO_NONE);
#if (!HAVE_GTK_1)
  DEFINE_INTEGER(XG_PRE "GDK_DRAG_PROTO_WIN32_DROPFILES" XG_POST,               GDK_DRAG_PROTO_WIN32_DROPFILES);
  DEFINE_INTEGER(XG_PRE "GDK_DRAG_PROTO_OLE2" XG_POST,                          GDK_DRAG_PROTO_OLE2);
  DEFINE_INTEGER(XG_PRE "GDK_DRAG_PROTO_LOCAL" XG_POST,                         GDK_DRAG_PROTO_LOCAL);
#endif
  DEFINE_INTEGER(XG_PRE "GDK_PRIORITY_EVENTS" XG_POST,                          GDK_PRIORITY_EVENTS);
#if (!HAVE_GTK_1)
  DEFINE_INTEGER(XG_PRE "GDK_PRIORITY_REDRAW" XG_POST,                          GDK_PRIORITY_REDRAW);
#endif
  DEFINE_INTEGER(XG_PRE "GDK_NOTHING" XG_POST,                                  GDK_NOTHING);
  DEFINE_INTEGER(XG_PRE "GDK_DELETE" XG_POST,                                   GDK_DELETE);
  DEFINE_INTEGER(XG_PRE "GDK_DESTROY" XG_POST,                                  GDK_DESTROY);
  DEFINE_INTEGER(XG_PRE "GDK_EXPOSE" XG_POST,                                   GDK_EXPOSE);
  DEFINE_INTEGER(XG_PRE "GDK_MOTION_NOTIFY" XG_POST,                            GDK_MOTION_NOTIFY);
  DEFINE_INTEGER(XG_PRE "GDK_BUTTON_PRESS" XG_POST,                             GDK_BUTTON_PRESS);
  DEFINE_INTEGER(XG_PRE "GDK_2BUTTON_PRESS" XG_POST,                            GDK_2BUTTON_PRESS);
  DEFINE_INTEGER(XG_PRE "GDK_3BUTTON_PRESS" XG_POST,                            GDK_3BUTTON_PRESS);
  DEFINE_INTEGER(XG_PRE "GDK_BUTTON_RELEASE" XG_POST,                           GDK_BUTTON_RELEASE);
  DEFINE_INTEGER(XG_PRE "GDK_KEY_PRESS" XG_POST,                                GDK_KEY_PRESS);
  DEFINE_INTEGER(XG_PRE "GDK_KEY_RELEASE" XG_POST,                              GDK_KEY_RELEASE);
  DEFINE_INTEGER(XG_PRE "GDK_ENTER_NOTIFY" XG_POST,                             GDK_ENTER_NOTIFY);
  DEFINE_INTEGER(XG_PRE "GDK_LEAVE_NOTIFY" XG_POST,                             GDK_LEAVE_NOTIFY);
  DEFINE_INTEGER(XG_PRE "GDK_FOCUS_CHANGE" XG_POST,                             GDK_FOCUS_CHANGE);
  DEFINE_INTEGER(XG_PRE "GDK_CONFIGURE" XG_POST,                                GDK_CONFIGURE);
  DEFINE_INTEGER(XG_PRE "GDK_MAP" XG_POST,                                      GDK_MAP);
  DEFINE_INTEGER(XG_PRE "GDK_UNMAP" XG_POST,                                    GDK_UNMAP);
  DEFINE_INTEGER(XG_PRE "GDK_PROPERTY_NOTIFY" XG_POST,                          GDK_PROPERTY_NOTIFY);
  DEFINE_INTEGER(XG_PRE "GDK_SELECTION_CLEAR" XG_POST,                          GDK_SELECTION_CLEAR);
  DEFINE_INTEGER(XG_PRE "GDK_SELECTION_REQUEST" XG_POST,                        GDK_SELECTION_REQUEST);
  DEFINE_INTEGER(XG_PRE "GDK_SELECTION_NOTIFY" XG_POST,                         GDK_SELECTION_NOTIFY);
  DEFINE_INTEGER(XG_PRE "GDK_PROXIMITY_IN" XG_POST,                             GDK_PROXIMITY_IN);
  DEFINE_INTEGER(XG_PRE "GDK_PROXIMITY_OUT" XG_POST,                            GDK_PROXIMITY_OUT);
  DEFINE_INTEGER(XG_PRE "GDK_DRAG_ENTER" XG_POST,                               GDK_DRAG_ENTER);
  DEFINE_INTEGER(XG_PRE "GDK_DRAG_LEAVE" XG_POST,                               GDK_DRAG_LEAVE);
  DEFINE_INTEGER(XG_PRE "GDK_DRAG_MOTION" XG_POST,                              GDK_DRAG_MOTION);
  DEFINE_INTEGER(XG_PRE "GDK_DRAG_STATUS" XG_POST,                              GDK_DRAG_STATUS);
  DEFINE_INTEGER(XG_PRE "GDK_DROP_START" XG_POST,                               GDK_DROP_START);
  DEFINE_INTEGER(XG_PRE "GDK_DROP_FINISHED" XG_POST,                            GDK_DROP_FINISHED);
  DEFINE_INTEGER(XG_PRE "GDK_CLIENT_EVENT" XG_POST,                             GDK_CLIENT_EVENT);
  DEFINE_INTEGER(XG_PRE "GDK_VISIBILITY_NOTIFY" XG_POST,                        GDK_VISIBILITY_NOTIFY);
  DEFINE_INTEGER(XG_PRE "GDK_NO_EXPOSE" XG_POST,                                GDK_NO_EXPOSE);
#if (!HAVE_GTK_1)
  DEFINE_INTEGER(XG_PRE "GDK_SCROLL" XG_POST,                                   GDK_SCROLL);
  DEFINE_INTEGER(XG_PRE "GDK_WINDOW_STATE" XG_POST,                             GDK_WINDOW_STATE);
  DEFINE_INTEGER(XG_PRE "GDK_SETTING" XG_POST,                                  GDK_SETTING);
#endif
  DEFINE_INTEGER(XG_PRE "GDK_EXPOSURE_MASK" XG_POST,                            GDK_EXPOSURE_MASK);
  DEFINE_INTEGER(XG_PRE "GDK_POINTER_MOTION_MASK" XG_POST,                      GDK_POINTER_MOTION_MASK);
  DEFINE_INTEGER(XG_PRE "GDK_POINTER_MOTION_HINT_MASK" XG_POST,                 GDK_POINTER_MOTION_HINT_MASK);
  DEFINE_INTEGER(XG_PRE "GDK_BUTTON_MOTION_MASK" XG_POST,                       GDK_BUTTON_MOTION_MASK);
  DEFINE_INTEGER(XG_PRE "GDK_BUTTON1_MOTION_MASK" XG_POST,                      GDK_BUTTON1_MOTION_MASK);
  DEFINE_INTEGER(XG_PRE "GDK_BUTTON2_MOTION_MASK" XG_POST,                      GDK_BUTTON2_MOTION_MASK);
  DEFINE_INTEGER(XG_PRE "GDK_BUTTON3_MOTION_MASK" XG_POST,                      GDK_BUTTON3_MOTION_MASK);
  DEFINE_INTEGER(XG_PRE "GDK_BUTTON_PRESS_MASK" XG_POST,                        GDK_BUTTON_PRESS_MASK);
  DEFINE_INTEGER(XG_PRE "GDK_BUTTON_RELEASE_MASK" XG_POST,                      GDK_BUTTON_RELEASE_MASK);
  DEFINE_INTEGER(XG_PRE "GDK_KEY_PRESS_MASK" XG_POST,                           GDK_KEY_PRESS_MASK);
  DEFINE_INTEGER(XG_PRE "GDK_KEY_RELEASE_MASK" XG_POST,                         GDK_KEY_RELEASE_MASK);
  DEFINE_INTEGER(XG_PRE "GDK_ENTER_NOTIFY_MASK" XG_POST,                        GDK_ENTER_NOTIFY_MASK);
  DEFINE_INTEGER(XG_PRE "GDK_LEAVE_NOTIFY_MASK" XG_POST,                        GDK_LEAVE_NOTIFY_MASK);
  DEFINE_INTEGER(XG_PRE "GDK_FOCUS_CHANGE_MASK" XG_POST,                        GDK_FOCUS_CHANGE_MASK);
  DEFINE_INTEGER(XG_PRE "GDK_STRUCTURE_MASK" XG_POST,                           GDK_STRUCTURE_MASK);
  DEFINE_INTEGER(XG_PRE "GDK_PROPERTY_CHANGE_MASK" XG_POST,                     GDK_PROPERTY_CHANGE_MASK);
  DEFINE_INTEGER(XG_PRE "GDK_VISIBILITY_NOTIFY_MASK" XG_POST,                   GDK_VISIBILITY_NOTIFY_MASK);
  DEFINE_INTEGER(XG_PRE "GDK_PROXIMITY_IN_MASK" XG_POST,                        GDK_PROXIMITY_IN_MASK);
  DEFINE_INTEGER(XG_PRE "GDK_PROXIMITY_OUT_MASK" XG_POST,                       GDK_PROXIMITY_OUT_MASK);
  DEFINE_INTEGER(XG_PRE "GDK_SUBSTRUCTURE_MASK" XG_POST,                        GDK_SUBSTRUCTURE_MASK);
#if (!HAVE_GTK_1)
  DEFINE_INTEGER(XG_PRE "GDK_SCROLL_MASK" XG_POST,                              GDK_SCROLL_MASK);
#endif
  DEFINE_INTEGER(XG_PRE "GDK_ALL_EVENTS_MASK" XG_POST,                          GDK_ALL_EVENTS_MASK);
#if (!HAVE_GTK_1)
  DEFINE_INTEGER(XG_PRE "GDK_VISIBILITY_UNOBSCURED " XG_POST,                   GDK_VISIBILITY_UNOBSCURED );
  DEFINE_INTEGER(XG_PRE "GDK_VISIBILITY_PARTIAL " XG_POST,                      GDK_VISIBILITY_PARTIAL );
#endif
  DEFINE_INTEGER(XG_PRE "GDK_VISIBILITY_FULLY_OBSCURED" XG_POST,                GDK_VISIBILITY_FULLY_OBSCURED);
#if (!HAVE_GTK_1)
  DEFINE_INTEGER(XG_PRE "GDK_SCROLL_UP" XG_POST,                                GDK_SCROLL_UP);
  DEFINE_INTEGER(XG_PRE "GDK_SCROLL_DOWN" XG_POST,                              GDK_SCROLL_DOWN);
  DEFINE_INTEGER(XG_PRE "GDK_SCROLL_LEFT" XG_POST,                              GDK_SCROLL_LEFT);
  DEFINE_INTEGER(XG_PRE "GDK_SCROLL_RIGHT" XG_POST,                             GDK_SCROLL_RIGHT);
#endif
  DEFINE_INTEGER(XG_PRE "GDK_NOTIFY_ANCESTOR" XG_POST,                          GDK_NOTIFY_ANCESTOR);
  DEFINE_INTEGER(XG_PRE "GDK_NOTIFY_VIRTUAL" XG_POST,                           GDK_NOTIFY_VIRTUAL);
  DEFINE_INTEGER(XG_PRE "GDK_NOTIFY_INFERIOR" XG_POST,                          GDK_NOTIFY_INFERIOR);
  DEFINE_INTEGER(XG_PRE "GDK_NOTIFY_NONLINEAR" XG_POST,                         GDK_NOTIFY_NONLINEAR);
  DEFINE_INTEGER(XG_PRE "GDK_NOTIFY_NONLINEAR_VIRTUAL" XG_POST,                 GDK_NOTIFY_NONLINEAR_VIRTUAL);
  DEFINE_INTEGER(XG_PRE "GDK_NOTIFY_UNKNOWN" XG_POST,                           GDK_NOTIFY_UNKNOWN);
  DEFINE_INTEGER(XG_PRE "GDK_CROSSING_NORMAL" XG_POST,                          GDK_CROSSING_NORMAL);
  DEFINE_INTEGER(XG_PRE "GDK_CROSSING_GRAB" XG_POST,                            GDK_CROSSING_GRAB);
  DEFINE_INTEGER(XG_PRE "GDK_CROSSING_UNGRAB" XG_POST,                          GDK_CROSSING_UNGRAB);
  DEFINE_INTEGER(XG_PRE "GDK_PROPERTY_NEW_VALUE" XG_POST,                       GDK_PROPERTY_NEW_VALUE);
  DEFINE_INTEGER(XG_PRE "GDK_PROPERTY_DELETE" XG_POST,                          GDK_PROPERTY_DELETE);
#if (!HAVE_GTK_1)
  DEFINE_INTEGER(XG_PRE "GDK_WINDOW_STATE_WITHDRAWN" XG_POST,                   GDK_WINDOW_STATE_WITHDRAWN);
  DEFINE_INTEGER(XG_PRE "GDK_WINDOW_STATE_ICONIFIED" XG_POST,                   GDK_WINDOW_STATE_ICONIFIED);
  DEFINE_INTEGER(XG_PRE "GDK_WINDOW_STATE_MAXIMIZED" XG_POST,                   GDK_WINDOW_STATE_MAXIMIZED);
  DEFINE_INTEGER(XG_PRE "GDK_WINDOW_STATE_STICKY" XG_POST,                      GDK_WINDOW_STATE_STICKY);
  DEFINE_INTEGER(XG_PRE "GDK_SETTING_ACTION_NEW" XG_POST,                       GDK_SETTING_ACTION_NEW);
  DEFINE_INTEGER(XG_PRE "GDK_SETTING_ACTION_CHANGED" XG_POST,                   GDK_SETTING_ACTION_CHANGED);
  DEFINE_INTEGER(XG_PRE "GDK_SETTING_ACTION_DELETED" XG_POST,                   GDK_SETTING_ACTION_DELETED);
#endif
  DEFINE_INTEGER(XG_PRE "GDK_CAP_NOT_LAST" XG_POST,                             GDK_CAP_NOT_LAST);
  DEFINE_INTEGER(XG_PRE "GDK_CAP_BUTT" XG_POST,                                 GDK_CAP_BUTT);
  DEFINE_INTEGER(XG_PRE "GDK_CAP_ROUND" XG_POST,                                GDK_CAP_ROUND);
  DEFINE_INTEGER(XG_PRE "GDK_CAP_PROJECTING" XG_POST,                           GDK_CAP_PROJECTING);
  DEFINE_INTEGER(XG_PRE "GDK_SOLID" XG_POST,                                    GDK_SOLID);
  DEFINE_INTEGER(XG_PRE "GDK_TILED" XG_POST,                                    GDK_TILED);
  DEFINE_INTEGER(XG_PRE "GDK_STIPPLED" XG_POST,                                 GDK_STIPPLED);
  DEFINE_INTEGER(XG_PRE "GDK_OPAQUE_STIPPLED" XG_POST,                          GDK_OPAQUE_STIPPLED);
  DEFINE_INTEGER(XG_PRE "GDK_COPY" XG_POST,                                     GDK_COPY);
  DEFINE_INTEGER(XG_PRE "GDK_INVERT" XG_POST,                                   GDK_INVERT);
  DEFINE_INTEGER(XG_PRE "GDK_XOR" XG_POST,                                      GDK_XOR);
  DEFINE_INTEGER(XG_PRE "GDK_CLEAR" XG_POST,                                    GDK_CLEAR);
  DEFINE_INTEGER(XG_PRE "GDK_AND" XG_POST,                                      GDK_AND);
  DEFINE_INTEGER(XG_PRE "GDK_AND_REVERSE" XG_POST,                              GDK_AND_REVERSE);
  DEFINE_INTEGER(XG_PRE "GDK_AND_INVERT" XG_POST,                               GDK_AND_INVERT);
  DEFINE_INTEGER(XG_PRE "GDK_NOOP" XG_POST,                                     GDK_NOOP);
  DEFINE_INTEGER(XG_PRE "GDK_OR" XG_POST,                                       GDK_OR);
  DEFINE_INTEGER(XG_PRE "GDK_EQUIV" XG_POST,                                    GDK_EQUIV);
  DEFINE_INTEGER(XG_PRE "GDK_OR_REVERSE" XG_POST,                               GDK_OR_REVERSE);
  DEFINE_INTEGER(XG_PRE "GDK_COPY_INVERT" XG_POST,                              GDK_COPY_INVERT);
  DEFINE_INTEGER(XG_PRE "GDK_OR_INVERT" XG_POST,                                GDK_OR_INVERT);
  DEFINE_INTEGER(XG_PRE "GDK_NAND" XG_POST,                                     GDK_NAND);
#if (!HAVE_GTK_1)
  DEFINE_INTEGER(XG_PRE "GDK_NOR" XG_POST,                                      GDK_NOR);
#endif
  DEFINE_INTEGER(XG_PRE "GDK_SET" XG_POST,                                      GDK_SET);
  DEFINE_INTEGER(XG_PRE "GDK_JOIN_MITER" XG_POST,                               GDK_JOIN_MITER);
  DEFINE_INTEGER(XG_PRE "GDK_JOIN_ROUND" XG_POST,                               GDK_JOIN_ROUND);
  DEFINE_INTEGER(XG_PRE "GDK_JOIN_BEVEL" XG_POST,                               GDK_JOIN_BEVEL);
  DEFINE_INTEGER(XG_PRE "GDK_LINE_SOLID" XG_POST,                               GDK_LINE_SOLID);
  DEFINE_INTEGER(XG_PRE "GDK_LINE_ON_OFF_DASH" XG_POST,                         GDK_LINE_ON_OFF_DASH);
  DEFINE_INTEGER(XG_PRE "GDK_LINE_DOUBLE_DASH" XG_POST,                         GDK_LINE_DOUBLE_DASH);
  DEFINE_INTEGER(XG_PRE "GDK_CLIP_BY_CHILDREN" XG_POST,                         GDK_CLIP_BY_CHILDREN);
  DEFINE_INTEGER(XG_PRE "GDK_INCLUDE_INFERIORS" XG_POST,                        GDK_INCLUDE_INFERIORS);
  DEFINE_INTEGER(XG_PRE "GDK_GC_FOREGROUND" XG_POST,                            GDK_GC_FOREGROUND);
  DEFINE_INTEGER(XG_PRE "GDK_GC_BACKGROUND" XG_POST,                            GDK_GC_BACKGROUND);
  DEFINE_INTEGER(XG_PRE "GDK_GC_FONT" XG_POST,                                  GDK_GC_FONT);
  DEFINE_INTEGER(XG_PRE "GDK_GC_FUNCTION" XG_POST,                              GDK_GC_FUNCTION);
  DEFINE_INTEGER(XG_PRE "GDK_GC_FILL" XG_POST,                                  GDK_GC_FILL);
  DEFINE_INTEGER(XG_PRE "GDK_GC_TILE" XG_POST,                                  GDK_GC_TILE);
  DEFINE_INTEGER(XG_PRE "GDK_GC_STIPPLE" XG_POST,                               GDK_GC_STIPPLE);
  DEFINE_INTEGER(XG_PRE "GDK_GC_CLIP_MASK" XG_POST,                             GDK_GC_CLIP_MASK);
  DEFINE_INTEGER(XG_PRE "GDK_GC_SUBWINDOW" XG_POST,                             GDK_GC_SUBWINDOW);
  DEFINE_INTEGER(XG_PRE "GDK_GC_TS_X_ORIGIN" XG_POST,                           GDK_GC_TS_X_ORIGIN);
  DEFINE_INTEGER(XG_PRE "GDK_GC_TS_Y_ORIGIN" XG_POST,                           GDK_GC_TS_Y_ORIGIN);
  DEFINE_INTEGER(XG_PRE "GDK_GC_CLIP_X_ORIGIN" XG_POST,                         GDK_GC_CLIP_X_ORIGIN);
  DEFINE_INTEGER(XG_PRE "GDK_GC_CLIP_Y_ORIGIN" XG_POST,                         GDK_GC_CLIP_Y_ORIGIN);
  DEFINE_INTEGER(XG_PRE "GDK_GC_EXPOSURES" XG_POST,                             GDK_GC_EXPOSURES);
  DEFINE_INTEGER(XG_PRE "GDK_GC_LINE_WIDTH" XG_POST,                            GDK_GC_LINE_WIDTH);
  DEFINE_INTEGER(XG_PRE "GDK_GC_LINE_STYLE" XG_POST,                            GDK_GC_LINE_STYLE);
  DEFINE_INTEGER(XG_PRE "GDK_GC_CAP_STYLE" XG_POST,                             GDK_GC_CAP_STYLE);
  DEFINE_INTEGER(XG_PRE "GDK_GC_JOIN_STYLE" XG_POST,                            GDK_GC_JOIN_STYLE);
  DEFINE_INTEGER(XG_PRE "GDK_IMAGE_NORMAL" XG_POST,                             GDK_IMAGE_NORMAL);
  DEFINE_INTEGER(XG_PRE "GDK_IMAGE_SHARED" XG_POST,                             GDK_IMAGE_SHARED);
  DEFINE_INTEGER(XG_PRE "GDK_IMAGE_FASTEST" XG_POST,                            GDK_IMAGE_FASTEST);
  DEFINE_INTEGER(XG_PRE "GDK_EXTENSION_EVENTS_NONE" XG_POST,                    GDK_EXTENSION_EVENTS_NONE);
  DEFINE_INTEGER(XG_PRE "GDK_EXTENSION_EVENTS_ALL" XG_POST,                     GDK_EXTENSION_EVENTS_ALL);
  DEFINE_INTEGER(XG_PRE "GDK_EXTENSION_EVENTS_CURSOR" XG_POST,                  GDK_EXTENSION_EVENTS_CURSOR);
  DEFINE_INTEGER(XG_PRE "GDK_SOURCE_MOUSE" XG_POST,                             GDK_SOURCE_MOUSE);
  DEFINE_INTEGER(XG_PRE "GDK_SOURCE_PEN" XG_POST,                               GDK_SOURCE_PEN);
  DEFINE_INTEGER(XG_PRE "GDK_SOURCE_ERASER" XG_POST,                            GDK_SOURCE_ERASER);
  DEFINE_INTEGER(XG_PRE "GDK_SOURCE_CURSOR" XG_POST,                            GDK_SOURCE_CURSOR);
  DEFINE_INTEGER(XG_PRE "GDK_MODE_DISABLED" XG_POST,                            GDK_MODE_DISABLED);
  DEFINE_INTEGER(XG_PRE "GDK_MODE_SCREEN" XG_POST,                              GDK_MODE_SCREEN);
  DEFINE_INTEGER(XG_PRE "GDK_MODE_WINDOW" XG_POST,                              GDK_MODE_WINDOW);
  DEFINE_INTEGER(XG_PRE "GDK_AXIS_IGNORE" XG_POST,                              GDK_AXIS_IGNORE);
  DEFINE_INTEGER(XG_PRE "GDK_AXIS_X" XG_POST,                                   GDK_AXIS_X);
  DEFINE_INTEGER(XG_PRE "GDK_AXIS_Y" XG_POST,                                   GDK_AXIS_Y);
  DEFINE_INTEGER(XG_PRE "GDK_AXIS_PRESSURE" XG_POST,                            GDK_AXIS_PRESSURE);
  DEFINE_INTEGER(XG_PRE "GDK_AXIS_XTILT" XG_POST,                               GDK_AXIS_XTILT);
  DEFINE_INTEGER(XG_PRE "GDK_AXIS_YTILT" XG_POST,                               GDK_AXIS_YTILT);
#if (!HAVE_GTK_1)
  DEFINE_INTEGER(XG_PRE "GDK_AXIS_WHEEL" XG_POST,                               GDK_AXIS_WHEEL);
#endif
  DEFINE_INTEGER(XG_PRE "GDK_AXIS_LAST" XG_POST,                                GDK_AXIS_LAST);
#if (!HAVE_GTK_1)
  DEFINE_INTEGER(XG_PRE "GDK_MAX_TIMECOORD_AXES" XG_POST,                       GDK_MAX_TIMECOORD_AXES);
#endif
  DEFINE_INTEGER(XG_PRE "GDK_VoidSymbol" XG_POST,                               GDK_VoidSymbol);
  DEFINE_INTEGER(XG_PRE "GDK_BackSpace" XG_POST,                                GDK_BackSpace);
  DEFINE_INTEGER(XG_PRE "GDK_Tab" XG_POST,                                      GDK_Tab);
  DEFINE_INTEGER(XG_PRE "GDK_Linefeed" XG_POST,                                 GDK_Linefeed);
  DEFINE_INTEGER(XG_PRE "GDK_Clear" XG_POST,                                    GDK_Clear);
  DEFINE_INTEGER(XG_PRE "GDK_Return" XG_POST,                                   GDK_Return);
  DEFINE_INTEGER(XG_PRE "GDK_Pause" XG_POST,                                    GDK_Pause);
  DEFINE_INTEGER(XG_PRE "GDK_Scroll_Lock" XG_POST,                              GDK_Scroll_Lock);
  DEFINE_INTEGER(XG_PRE "GDK_Sys_Req" XG_POST,                                  GDK_Sys_Req);
  DEFINE_INTEGER(XG_PRE "GDK_Escape" XG_POST,                                   GDK_Escape);
  DEFINE_INTEGER(XG_PRE "GDK_Delete" XG_POST,                                   GDK_Delete);
  DEFINE_INTEGER(XG_PRE "GDK_Home" XG_POST,                                     GDK_Home);
  DEFINE_INTEGER(XG_PRE "GDK_Left" XG_POST,                                     GDK_Left);
  DEFINE_INTEGER(XG_PRE "GDK_Up" XG_POST,                                       GDK_Up);
  DEFINE_INTEGER(XG_PRE "GDK_Right" XG_POST,                                    GDK_Right);
  DEFINE_INTEGER(XG_PRE "GDK_Down" XG_POST,                                     GDK_Down);
  DEFINE_INTEGER(XG_PRE "GDK_Prior" XG_POST,                                    GDK_Prior);
  DEFINE_INTEGER(XG_PRE "GDK_Page_Up" XG_POST,                                  GDK_Page_Up);
  DEFINE_INTEGER(XG_PRE "GDK_Next" XG_POST,                                     GDK_Next);
  DEFINE_INTEGER(XG_PRE "GDK_Page_Down" XG_POST,                                GDK_Page_Down);
  DEFINE_INTEGER(XG_PRE "GDK_End" XG_POST,                                      GDK_End);
  DEFINE_INTEGER(XG_PRE "GDK_Begin" XG_POST,                                    GDK_Begin);
  DEFINE_INTEGER(XG_PRE "GDK_Select" XG_POST,                                   GDK_Select);
  DEFINE_INTEGER(XG_PRE "GDK_Print" XG_POST,                                    GDK_Print);
  DEFINE_INTEGER(XG_PRE "GDK_Execute" XG_POST,                                  GDK_Execute);
  DEFINE_INTEGER(XG_PRE "GDK_Insert" XG_POST,                                   GDK_Insert);
  DEFINE_INTEGER(XG_PRE "GDK_Undo" XG_POST,                                     GDK_Undo);
  DEFINE_INTEGER(XG_PRE "GDK_Redo" XG_POST,                                     GDK_Redo);
  DEFINE_INTEGER(XG_PRE "GDK_Menu" XG_POST,                                     GDK_Menu);
  DEFINE_INTEGER(XG_PRE "GDK_Find" XG_POST,                                     GDK_Find);
  DEFINE_INTEGER(XG_PRE "GDK_Cancel" XG_POST,                                   GDK_Cancel);
  DEFINE_INTEGER(XG_PRE "GDK_Help" XG_POST,                                     GDK_Help);
  DEFINE_INTEGER(XG_PRE "GDK_Break" XG_POST,                                    GDK_Break);
  DEFINE_INTEGER(XG_PRE "GDK_Mode_switch" XG_POST,                              GDK_Mode_switch);
  DEFINE_INTEGER(XG_PRE "GDK_script_switch" XG_POST,                            GDK_script_switch);
  DEFINE_INTEGER(XG_PRE "GDK_Num_Lock" XG_POST,                                 GDK_Num_Lock);
  DEFINE_INTEGER(XG_PRE "GDK_KP_Space" XG_POST,                                 GDK_KP_Space);
  DEFINE_INTEGER(XG_PRE "GDK_KP_Tab" XG_POST,                                   GDK_KP_Tab);
  DEFINE_INTEGER(XG_PRE "GDK_KP_Enter" XG_POST,                                 GDK_KP_Enter);
  DEFINE_INTEGER(XG_PRE "GDK_KP_F1" XG_POST,                                    GDK_KP_F1);
  DEFINE_INTEGER(XG_PRE "GDK_KP_F2" XG_POST,                                    GDK_KP_F2);
  DEFINE_INTEGER(XG_PRE "GDK_KP_F3" XG_POST,                                    GDK_KP_F3);
  DEFINE_INTEGER(XG_PRE "GDK_KP_F4" XG_POST,                                    GDK_KP_F4);
  DEFINE_INTEGER(XG_PRE "GDK_KP_Home" XG_POST,                                  GDK_KP_Home);
  DEFINE_INTEGER(XG_PRE "GDK_KP_Left" XG_POST,                                  GDK_KP_Left);
  DEFINE_INTEGER(XG_PRE "GDK_KP_Up" XG_POST,                                    GDK_KP_Up);
  DEFINE_INTEGER(XG_PRE "GDK_KP_Right" XG_POST,                                 GDK_KP_Right);
  DEFINE_INTEGER(XG_PRE "GDK_KP_Down" XG_POST,                                  GDK_KP_Down);
  DEFINE_INTEGER(XG_PRE "GDK_KP_Prior" XG_POST,                                 GDK_KP_Prior);
  DEFINE_INTEGER(XG_PRE "GDK_KP_Page_Up" XG_POST,                               GDK_KP_Page_Up);
  DEFINE_INTEGER(XG_PRE "GDK_KP_Next" XG_POST,                                  GDK_KP_Next);
  DEFINE_INTEGER(XG_PRE "GDK_KP_Page_Down" XG_POST,                             GDK_KP_Page_Down);
  DEFINE_INTEGER(XG_PRE "GDK_KP_End" XG_POST,                                   GDK_KP_End);
  DEFINE_INTEGER(XG_PRE "GDK_KP_Begin" XG_POST,                                 GDK_KP_Begin);
  DEFINE_INTEGER(XG_PRE "GDK_KP_Insert" XG_POST,                                GDK_KP_Insert);
  DEFINE_INTEGER(XG_PRE "GDK_KP_Delete" XG_POST,                                GDK_KP_Delete);
  DEFINE_INTEGER(XG_PRE "GDK_KP_Equal" XG_POST,                                 GDK_KP_Equal);
  DEFINE_INTEGER(XG_PRE "GDK_KP_Multiply" XG_POST,                              GDK_KP_Multiply);
  DEFINE_INTEGER(XG_PRE "GDK_KP_Add" XG_POST,                                   GDK_KP_Add);
  DEFINE_INTEGER(XG_PRE "GDK_KP_Separator" XG_POST,                             GDK_KP_Separator);
  DEFINE_INTEGER(XG_PRE "GDK_KP_Subtract" XG_POST,                              GDK_KP_Subtract);
  DEFINE_INTEGER(XG_PRE "GDK_KP_Decimal" XG_POST,                               GDK_KP_Decimal);
  DEFINE_INTEGER(XG_PRE "GDK_KP_Divide" XG_POST,                                GDK_KP_Divide);
  DEFINE_INTEGER(XG_PRE "GDK_KP_0" XG_POST,                                     GDK_KP_0);
  DEFINE_INTEGER(XG_PRE "GDK_KP_1" XG_POST,                                     GDK_KP_1);
  DEFINE_INTEGER(XG_PRE "GDK_KP_2" XG_POST,                                     GDK_KP_2);
  DEFINE_INTEGER(XG_PRE "GDK_KP_3" XG_POST,                                     GDK_KP_3);
  DEFINE_INTEGER(XG_PRE "GDK_KP_4" XG_POST,                                     GDK_KP_4);
  DEFINE_INTEGER(XG_PRE "GDK_KP_5" XG_POST,                                     GDK_KP_5);
  DEFINE_INTEGER(XG_PRE "GDK_KP_6" XG_POST,                                     GDK_KP_6);
  DEFINE_INTEGER(XG_PRE "GDK_KP_7" XG_POST,                                     GDK_KP_7);
  DEFINE_INTEGER(XG_PRE "GDK_KP_8" XG_POST,                                     GDK_KP_8);
  DEFINE_INTEGER(XG_PRE "GDK_KP_9" XG_POST,                                     GDK_KP_9);
  DEFINE_INTEGER(XG_PRE "GDK_F1" XG_POST,                                       GDK_F1);
  DEFINE_INTEGER(XG_PRE "GDK_F2" XG_POST,                                       GDK_F2);
  DEFINE_INTEGER(XG_PRE "GDK_F3" XG_POST,                                       GDK_F3);
  DEFINE_INTEGER(XG_PRE "GDK_F4" XG_POST,                                       GDK_F4);
  DEFINE_INTEGER(XG_PRE "GDK_F5" XG_POST,                                       GDK_F5);
  DEFINE_INTEGER(XG_PRE "GDK_F6" XG_POST,                                       GDK_F6);
  DEFINE_INTEGER(XG_PRE "GDK_F7" XG_POST,                                       GDK_F7);
  DEFINE_INTEGER(XG_PRE "GDK_F8" XG_POST,                                       GDK_F8);
  DEFINE_INTEGER(XG_PRE "GDK_F9" XG_POST,                                       GDK_F9);
  DEFINE_INTEGER(XG_PRE "GDK_F10" XG_POST,                                      GDK_F10);
  DEFINE_INTEGER(XG_PRE "GDK_F11" XG_POST,                                      GDK_F11);
  DEFINE_INTEGER(XG_PRE "GDK_L1" XG_POST,                                       GDK_L1);
  DEFINE_INTEGER(XG_PRE "GDK_F12" XG_POST,                                      GDK_F12);
  DEFINE_INTEGER(XG_PRE "GDK_L2" XG_POST,                                       GDK_L2);
  DEFINE_INTEGER(XG_PRE "GDK_F13" XG_POST,                                      GDK_F13);
  DEFINE_INTEGER(XG_PRE "GDK_L3" XG_POST,                                       GDK_L3);
  DEFINE_INTEGER(XG_PRE "GDK_F14" XG_POST,                                      GDK_F14);
  DEFINE_INTEGER(XG_PRE "GDK_L4" XG_POST,                                       GDK_L4);
  DEFINE_INTEGER(XG_PRE "GDK_F15" XG_POST,                                      GDK_F15);
  DEFINE_INTEGER(XG_PRE "GDK_L5" XG_POST,                                       GDK_L5);
  DEFINE_INTEGER(XG_PRE "GDK_F16" XG_POST,                                      GDK_F16);
  DEFINE_INTEGER(XG_PRE "GDK_L6" XG_POST,                                       GDK_L6);
  DEFINE_INTEGER(XG_PRE "GDK_F17" XG_POST,                                      GDK_F17);
  DEFINE_INTEGER(XG_PRE "GDK_L7" XG_POST,                                       GDK_L7);
  DEFINE_INTEGER(XG_PRE "GDK_F18" XG_POST,                                      GDK_F18);
  DEFINE_INTEGER(XG_PRE "GDK_L8" XG_POST,                                       GDK_L8);
  DEFINE_INTEGER(XG_PRE "GDK_F19" XG_POST,                                      GDK_F19);
  DEFINE_INTEGER(XG_PRE "GDK_L9" XG_POST,                                       GDK_L9);
  DEFINE_INTEGER(XG_PRE "GDK_F20" XG_POST,                                      GDK_F20);
  DEFINE_INTEGER(XG_PRE "GDK_L10" XG_POST,                                      GDK_L10);
  DEFINE_INTEGER(XG_PRE "GDK_F21" XG_POST,                                      GDK_F21);
  DEFINE_INTEGER(XG_PRE "GDK_R1" XG_POST,                                       GDK_R1);
  DEFINE_INTEGER(XG_PRE "GDK_F22" XG_POST,                                      GDK_F22);
  DEFINE_INTEGER(XG_PRE "GDK_R2" XG_POST,                                       GDK_R2);
  DEFINE_INTEGER(XG_PRE "GDK_F23" XG_POST,                                      GDK_F23);
  DEFINE_INTEGER(XG_PRE "GDK_R3" XG_POST,                                       GDK_R3);
  DEFINE_INTEGER(XG_PRE "GDK_F24" XG_POST,                                      GDK_F24);
  DEFINE_INTEGER(XG_PRE "GDK_R4" XG_POST,                                       GDK_R4);
  DEFINE_INTEGER(XG_PRE "GDK_F25" XG_POST,                                      GDK_F25);
  DEFINE_INTEGER(XG_PRE "GDK_R5" XG_POST,                                       GDK_R5);
  DEFINE_INTEGER(XG_PRE "GDK_F26" XG_POST,                                      GDK_F26);
  DEFINE_INTEGER(XG_PRE "GDK_R6" XG_POST,                                       GDK_R6);
  DEFINE_INTEGER(XG_PRE "GDK_F27" XG_POST,                                      GDK_F27);
  DEFINE_INTEGER(XG_PRE "GDK_R7" XG_POST,                                       GDK_R7);
  DEFINE_INTEGER(XG_PRE "GDK_F28" XG_POST,                                      GDK_F28);
  DEFINE_INTEGER(XG_PRE "GDK_R8" XG_POST,                                       GDK_R8);
  DEFINE_INTEGER(XG_PRE "GDK_F29" XG_POST,                                      GDK_F29);
  DEFINE_INTEGER(XG_PRE "GDK_R9" XG_POST,                                       GDK_R9);
  DEFINE_INTEGER(XG_PRE "GDK_F30" XG_POST,                                      GDK_F30);
  DEFINE_INTEGER(XG_PRE "GDK_R10" XG_POST,                                      GDK_R10);
  DEFINE_INTEGER(XG_PRE "GDK_F31" XG_POST,                                      GDK_F31);
  DEFINE_INTEGER(XG_PRE "GDK_R11" XG_POST,                                      GDK_R11);
  DEFINE_INTEGER(XG_PRE "GDK_F32" XG_POST,                                      GDK_F32);
  DEFINE_INTEGER(XG_PRE "GDK_R12" XG_POST,                                      GDK_R12);
  DEFINE_INTEGER(XG_PRE "GDK_F33" XG_POST,                                      GDK_F33);
  DEFINE_INTEGER(XG_PRE "GDK_R13" XG_POST,                                      GDK_R13);
  DEFINE_INTEGER(XG_PRE "GDK_F34" XG_POST,                                      GDK_F34);
  DEFINE_INTEGER(XG_PRE "GDK_R14" XG_POST,                                      GDK_R14);
  DEFINE_INTEGER(XG_PRE "GDK_F35" XG_POST,                                      GDK_F35);
  DEFINE_INTEGER(XG_PRE "GDK_R15" XG_POST,                                      GDK_R15);
  DEFINE_INTEGER(XG_PRE "GDK_Shift_L" XG_POST,                                  GDK_Shift_L);
  DEFINE_INTEGER(XG_PRE "GDK_Shift_R" XG_POST,                                  GDK_Shift_R);
  DEFINE_INTEGER(XG_PRE "GDK_Control_L" XG_POST,                                GDK_Control_L);
  DEFINE_INTEGER(XG_PRE "GDK_Control_R" XG_POST,                                GDK_Control_R);
  DEFINE_INTEGER(XG_PRE "GDK_Caps_Lock" XG_POST,                                GDK_Caps_Lock);
  DEFINE_INTEGER(XG_PRE "GDK_Shift_Lock" XG_POST,                               GDK_Shift_Lock);
  DEFINE_INTEGER(XG_PRE "GDK_Meta_L" XG_POST,                                   GDK_Meta_L);
  DEFINE_INTEGER(XG_PRE "GDK_Meta_R" XG_POST,                                   GDK_Meta_R);
  DEFINE_INTEGER(XG_PRE "GDK_Alt_L" XG_POST,                                    GDK_Alt_L);
  DEFINE_INTEGER(XG_PRE "GDK_Alt_R" XG_POST,                                    GDK_Alt_R);
  DEFINE_INTEGER(XG_PRE "GDK_Super_L" XG_POST,                                  GDK_Super_L);
  DEFINE_INTEGER(XG_PRE "GDK_Super_R" XG_POST,                                  GDK_Super_R);
  DEFINE_INTEGER(XG_PRE "GDK_Hyper_L" XG_POST,                                  GDK_Hyper_L);
  DEFINE_INTEGER(XG_PRE "GDK_Hyper_R" XG_POST,                                  GDK_Hyper_R);
  DEFINE_INTEGER(XG_PRE "GDK_space" XG_POST,                                    GDK_space);
  DEFINE_INTEGER(XG_PRE "GDK_exclam" XG_POST,                                   GDK_exclam);
  DEFINE_INTEGER(XG_PRE "GDK_quotedbl" XG_POST,                                 GDK_quotedbl);
  DEFINE_INTEGER(XG_PRE "GDK_numbersign" XG_POST,                               GDK_numbersign);
  DEFINE_INTEGER(XG_PRE "GDK_dollar" XG_POST,                                   GDK_dollar);
  DEFINE_INTEGER(XG_PRE "GDK_percent" XG_POST,                                  GDK_percent);
  DEFINE_INTEGER(XG_PRE "GDK_ampersand" XG_POST,                                GDK_ampersand);
  DEFINE_INTEGER(XG_PRE "GDK_apostrophe" XG_POST,                               GDK_apostrophe);
  DEFINE_INTEGER(XG_PRE "GDK_quoteright" XG_POST,                               GDK_quoteright);
  DEFINE_INTEGER(XG_PRE "GDK_parenleft" XG_POST,                                GDK_parenleft);
  DEFINE_INTEGER(XG_PRE "GDK_parenright" XG_POST,                               GDK_parenright);
  DEFINE_INTEGER(XG_PRE "GDK_asterisk" XG_POST,                                 GDK_asterisk);
  DEFINE_INTEGER(XG_PRE "GDK_plus" XG_POST,                                     GDK_plus);
  DEFINE_INTEGER(XG_PRE "GDK_comma" XG_POST,                                    GDK_comma);
  DEFINE_INTEGER(XG_PRE "GDK_minus" XG_POST,                                    GDK_minus);
  DEFINE_INTEGER(XG_PRE "GDK_period" XG_POST,                                   GDK_period);
  DEFINE_INTEGER(XG_PRE "GDK_slash" XG_POST,                                    GDK_slash);
  DEFINE_INTEGER(XG_PRE "GDK_0" XG_POST,                                        GDK_0);
  DEFINE_INTEGER(XG_PRE "GDK_1" XG_POST,                                        GDK_1);
  DEFINE_INTEGER(XG_PRE "GDK_2" XG_POST,                                        GDK_2);
  DEFINE_INTEGER(XG_PRE "GDK_3" XG_POST,                                        GDK_3);
  DEFINE_INTEGER(XG_PRE "GDK_4" XG_POST,                                        GDK_4);
  DEFINE_INTEGER(XG_PRE "GDK_5" XG_POST,                                        GDK_5);
  DEFINE_INTEGER(XG_PRE "GDK_6" XG_POST,                                        GDK_6);
  DEFINE_INTEGER(XG_PRE "GDK_7" XG_POST,                                        GDK_7);
  DEFINE_INTEGER(XG_PRE "GDK_8" XG_POST,                                        GDK_8);
  DEFINE_INTEGER(XG_PRE "GDK_9" XG_POST,                                        GDK_9);
  DEFINE_INTEGER(XG_PRE "GDK_colon" XG_POST,                                    GDK_colon);
  DEFINE_INTEGER(XG_PRE "GDK_semicolon" XG_POST,                                GDK_semicolon);
  DEFINE_INTEGER(XG_PRE "GDK_less" XG_POST,                                     GDK_less);
  DEFINE_INTEGER(XG_PRE "GDK_equal" XG_POST,                                    GDK_equal);
  DEFINE_INTEGER(XG_PRE "GDK_greater" XG_POST,                                  GDK_greater);
  DEFINE_INTEGER(XG_PRE "GDK_question" XG_POST,                                 GDK_question);
  DEFINE_INTEGER(XG_PRE "GDK_at" XG_POST,                                       GDK_at);
  DEFINE_INTEGER(XG_PRE "GDK_A" XG_POST,                                        GDK_A);
  DEFINE_INTEGER(XG_PRE "GDK_B" XG_POST,                                        GDK_B);
  DEFINE_INTEGER(XG_PRE "GDK_C" XG_POST,                                        GDK_C);
  DEFINE_INTEGER(XG_PRE "GDK_D" XG_POST,                                        GDK_D);
  DEFINE_INTEGER(XG_PRE "GDK_E" XG_POST,                                        GDK_E);
  DEFINE_INTEGER(XG_PRE "GDK_F" XG_POST,                                        GDK_F);
  DEFINE_INTEGER(XG_PRE "GDK_G" XG_POST,                                        GDK_G);
  DEFINE_INTEGER(XG_PRE "GDK_H" XG_POST,                                        GDK_H);
  DEFINE_INTEGER(XG_PRE "GDK_I" XG_POST,                                        GDK_I);
  DEFINE_INTEGER(XG_PRE "GDK_J" XG_POST,                                        GDK_J);
  DEFINE_INTEGER(XG_PRE "GDK_K" XG_POST,                                        GDK_K);
  DEFINE_INTEGER(XG_PRE "GDK_L" XG_POST,                                        GDK_L);
  DEFINE_INTEGER(XG_PRE "GDK_M" XG_POST,                                        GDK_M);
  DEFINE_INTEGER(XG_PRE "GDK_N" XG_POST,                                        GDK_N);
  DEFINE_INTEGER(XG_PRE "GDK_O" XG_POST,                                        GDK_O);
  DEFINE_INTEGER(XG_PRE "GDK_P" XG_POST,                                        GDK_P);
  DEFINE_INTEGER(XG_PRE "GDK_Q" XG_POST,                                        GDK_Q);
  DEFINE_INTEGER(XG_PRE "GDK_R" XG_POST,                                        GDK_R);
  DEFINE_INTEGER(XG_PRE "GDK_S" XG_POST,                                        GDK_S);
  DEFINE_INTEGER(XG_PRE "GDK_T" XG_POST,                                        GDK_T);
  DEFINE_INTEGER(XG_PRE "GDK_U" XG_POST,                                        GDK_U);
  DEFINE_INTEGER(XG_PRE "GDK_V" XG_POST,                                        GDK_V);
  DEFINE_INTEGER(XG_PRE "GDK_W" XG_POST,                                        GDK_W);
  DEFINE_INTEGER(XG_PRE "GDK_X" XG_POST,                                        GDK_X);
  DEFINE_INTEGER(XG_PRE "GDK_Y" XG_POST,                                        GDK_Y);
  DEFINE_INTEGER(XG_PRE "GDK_Z" XG_POST,                                        GDK_Z);
  DEFINE_INTEGER(XG_PRE "GDK_bracketleft" XG_POST,                              GDK_bracketleft);
  DEFINE_INTEGER(XG_PRE "GDK_backslash" XG_POST,                                GDK_backslash);
  DEFINE_INTEGER(XG_PRE "GDK_bracketright" XG_POST,                             GDK_bracketright);
  DEFINE_INTEGER(XG_PRE "GDK_asciicircum" XG_POST,                              GDK_asciicircum);
  DEFINE_INTEGER(XG_PRE "GDK_underscore" XG_POST,                               GDK_underscore);
  DEFINE_INTEGER(XG_PRE "GDK_grave" XG_POST,                                    GDK_grave);
  DEFINE_INTEGER(XG_PRE "GDK_quoteleft" XG_POST,                                GDK_quoteleft);
  DEFINE_INTEGER(XG_PRE "GDK_a" XG_POST,                                        GDK_a);
  DEFINE_INTEGER(XG_PRE "GDK_b" XG_POST,                                        GDK_b);
  DEFINE_INTEGER(XG_PRE "GDK_c" XG_POST,                                        GDK_c);
  DEFINE_INTEGER(XG_PRE "GDK_d" XG_POST,                                        GDK_d);
  DEFINE_INTEGER(XG_PRE "GDK_e" XG_POST,                                        GDK_e);
  DEFINE_INTEGER(XG_PRE "GDK_f" XG_POST,                                        GDK_f);
  DEFINE_INTEGER(XG_PRE "GDK_g" XG_POST,                                        GDK_g);
  DEFINE_INTEGER(XG_PRE "GDK_h" XG_POST,                                        GDK_h);
  DEFINE_INTEGER(XG_PRE "GDK_i" XG_POST,                                        GDK_i);
  DEFINE_INTEGER(XG_PRE "GDK_j" XG_POST,                                        GDK_j);
  DEFINE_INTEGER(XG_PRE "GDK_k" XG_POST,                                        GDK_k);
  DEFINE_INTEGER(XG_PRE "GDK_l" XG_POST,                                        GDK_l);
  DEFINE_INTEGER(XG_PRE "GDK_m" XG_POST,                                        GDK_m);
  DEFINE_INTEGER(XG_PRE "GDK_n" XG_POST,                                        GDK_n);
  DEFINE_INTEGER(XG_PRE "GDK_o" XG_POST,                                        GDK_o);
  DEFINE_INTEGER(XG_PRE "GDK_p" XG_POST,                                        GDK_p);
  DEFINE_INTEGER(XG_PRE "GDK_q" XG_POST,                                        GDK_q);
  DEFINE_INTEGER(XG_PRE "GDK_r" XG_POST,                                        GDK_r);
  DEFINE_INTEGER(XG_PRE "GDK_s" XG_POST,                                        GDK_s);
  DEFINE_INTEGER(XG_PRE "GDK_t" XG_POST,                                        GDK_t);
  DEFINE_INTEGER(XG_PRE "GDK_u" XG_POST,                                        GDK_u);
  DEFINE_INTEGER(XG_PRE "GDK_v" XG_POST,                                        GDK_v);
  DEFINE_INTEGER(XG_PRE "GDK_w" XG_POST,                                        GDK_w);
  DEFINE_INTEGER(XG_PRE "GDK_x" XG_POST,                                        GDK_x);
  DEFINE_INTEGER(XG_PRE "GDK_y" XG_POST,                                        GDK_y);
  DEFINE_INTEGER(XG_PRE "GDK_z" XG_POST,                                        GDK_z);
  DEFINE_INTEGER(XG_PRE "GDK_braceleft" XG_POST,                                GDK_braceleft);
  DEFINE_INTEGER(XG_PRE "GDK_bar" XG_POST,                                      GDK_bar);
  DEFINE_INTEGER(XG_PRE "GDK_braceright" XG_POST,                               GDK_braceright);
  DEFINE_INTEGER(XG_PRE "GDK_asciitilde" XG_POST,                               GDK_asciitilde);
  DEFINE_INTEGER(XG_PRE "GDK_nobreakspace" XG_POST,                             GDK_nobreakspace);
  DEFINE_INTEGER(XG_PRE "GDK_exclamdown" XG_POST,                               GDK_exclamdown);
  DEFINE_INTEGER(XG_PRE "GDK_cent" XG_POST,                                     GDK_cent);
  DEFINE_INTEGER(XG_PRE "GDK_sterling" XG_POST,                                 GDK_sterling);
  DEFINE_INTEGER(XG_PRE "GDK_currency" XG_POST,                                 GDK_currency);
  DEFINE_INTEGER(XG_PRE "GDK_yen" XG_POST,                                      GDK_yen);
  DEFINE_INTEGER(XG_PRE "GDK_brokenbar" XG_POST,                                GDK_brokenbar);
  DEFINE_INTEGER(XG_PRE "GDK_section" XG_POST,                                  GDK_section);
  DEFINE_INTEGER(XG_PRE "GDK_diaeresis" XG_POST,                                GDK_diaeresis);
  DEFINE_INTEGER(XG_PRE "GDK_copyright" XG_POST,                                GDK_copyright);
  DEFINE_INTEGER(XG_PRE "GDK_ordfeminine" XG_POST,                              GDK_ordfeminine);
  DEFINE_INTEGER(XG_PRE "GDK_guillemotleft" XG_POST,                            GDK_guillemotleft);
  DEFINE_INTEGER(XG_PRE "GDK_notsign" XG_POST,                                  GDK_notsign);
  DEFINE_INTEGER(XG_PRE "GDK_hyphen" XG_POST,                                   GDK_hyphen);
  DEFINE_INTEGER(XG_PRE "GDK_registered" XG_POST,                               GDK_registered);
  DEFINE_INTEGER(XG_PRE "GDK_macron" XG_POST,                                   GDK_macron);
  DEFINE_INTEGER(XG_PRE "GDK_degree" XG_POST,                                   GDK_degree);
  DEFINE_INTEGER(XG_PRE "GDK_plusminus" XG_POST,                                GDK_plusminus);
  DEFINE_INTEGER(XG_PRE "GDK_twosuperior" XG_POST,                              GDK_twosuperior);
  DEFINE_INTEGER(XG_PRE "GDK_threesuperior" XG_POST,                            GDK_threesuperior);
  DEFINE_INTEGER(XG_PRE "GDK_acute" XG_POST,                                    GDK_acute);
  DEFINE_INTEGER(XG_PRE "GDK_mu" XG_POST,                                       GDK_mu);
  DEFINE_INTEGER(XG_PRE "GDK_paragraph" XG_POST,                                GDK_paragraph);
  DEFINE_INTEGER(XG_PRE "GDK_periodcentered" XG_POST,                           GDK_periodcentered);
  DEFINE_INTEGER(XG_PRE "GDK_cedilla" XG_POST,                                  GDK_cedilla);
  DEFINE_INTEGER(XG_PRE "GDK_onesuperior" XG_POST,                              GDK_onesuperior);
  DEFINE_INTEGER(XG_PRE "GDK_masculine" XG_POST,                                GDK_masculine);
  DEFINE_INTEGER(XG_PRE "GDK_guillemotright" XG_POST,                           GDK_guillemotright);
  DEFINE_INTEGER(XG_PRE "GDK_onequarter" XG_POST,                               GDK_onequarter);
  DEFINE_INTEGER(XG_PRE "GDK_onehalf" XG_POST,                                  GDK_onehalf);
  DEFINE_INTEGER(XG_PRE "GDK_threequarters" XG_POST,                            GDK_threequarters);
  DEFINE_INTEGER(XG_PRE "GDK_questiondown" XG_POST,                             GDK_questiondown);
  DEFINE_INTEGER(XG_PRE "GDK_Agrave" XG_POST,                                   GDK_Agrave);
  DEFINE_INTEGER(XG_PRE "GDK_Aacute" XG_POST,                                   GDK_Aacute);
  DEFINE_INTEGER(XG_PRE "GDK_Acircumflex" XG_POST,                              GDK_Acircumflex);
  DEFINE_INTEGER(XG_PRE "GDK_Atilde" XG_POST,                                   GDK_Atilde);
  DEFINE_INTEGER(XG_PRE "GDK_Adiaeresis" XG_POST,                               GDK_Adiaeresis);
  DEFINE_INTEGER(XG_PRE "GDK_Aring" XG_POST,                                    GDK_Aring);
  DEFINE_INTEGER(XG_PRE "GDK_AE" XG_POST,                                       GDK_AE);
  DEFINE_INTEGER(XG_PRE "GDK_Ccedilla" XG_POST,                                 GDK_Ccedilla);
  DEFINE_INTEGER(XG_PRE "GDK_Egrave" XG_POST,                                   GDK_Egrave);
  DEFINE_INTEGER(XG_PRE "GDK_Eacute" XG_POST,                                   GDK_Eacute);
  DEFINE_INTEGER(XG_PRE "GDK_Ecircumflex" XG_POST,                              GDK_Ecircumflex);
  DEFINE_INTEGER(XG_PRE "GDK_Ediaeresis" XG_POST,                               GDK_Ediaeresis);
  DEFINE_INTEGER(XG_PRE "GDK_Igrave" XG_POST,                                   GDK_Igrave);
  DEFINE_INTEGER(XG_PRE "GDK_Iacute" XG_POST,                                   GDK_Iacute);
  DEFINE_INTEGER(XG_PRE "GDK_Icircumflex" XG_POST,                              GDK_Icircumflex);
  DEFINE_INTEGER(XG_PRE "GDK_Idiaeresis" XG_POST,                               GDK_Idiaeresis);
  DEFINE_INTEGER(XG_PRE "GDK_ETH" XG_POST,                                      GDK_ETH);
  DEFINE_INTEGER(XG_PRE "GDK_Eth" XG_POST,                                      GDK_Eth);
  DEFINE_INTEGER(XG_PRE "GDK_Ntilde" XG_POST,                                   GDK_Ntilde);
  DEFINE_INTEGER(XG_PRE "GDK_Ograve" XG_POST,                                   GDK_Ograve);
  DEFINE_INTEGER(XG_PRE "GDK_Oacute" XG_POST,                                   GDK_Oacute);
  DEFINE_INTEGER(XG_PRE "GDK_Ocircumflex" XG_POST,                              GDK_Ocircumflex);
  DEFINE_INTEGER(XG_PRE "GDK_Otilde" XG_POST,                                   GDK_Otilde);
  DEFINE_INTEGER(XG_PRE "GDK_Odiaeresis" XG_POST,                               GDK_Odiaeresis);
  DEFINE_INTEGER(XG_PRE "GDK_multiply" XG_POST,                                 GDK_multiply);
  DEFINE_INTEGER(XG_PRE "GDK_Ooblique" XG_POST,                                 GDK_Ooblique);
  DEFINE_INTEGER(XG_PRE "GDK_Ugrave" XG_POST,                                   GDK_Ugrave);
  DEFINE_INTEGER(XG_PRE "GDK_Uacute" XG_POST,                                   GDK_Uacute);
  DEFINE_INTEGER(XG_PRE "GDK_Ucircumflex" XG_POST,                              GDK_Ucircumflex);
  DEFINE_INTEGER(XG_PRE "GDK_Udiaeresis" XG_POST,                               GDK_Udiaeresis);
  DEFINE_INTEGER(XG_PRE "GDK_Yacute" XG_POST,                                   GDK_Yacute);
  DEFINE_INTEGER(XG_PRE "GDK_THORN" XG_POST,                                    GDK_THORN);
  DEFINE_INTEGER(XG_PRE "GDK_Thorn" XG_POST,                                    GDK_Thorn);
  DEFINE_INTEGER(XG_PRE "GDK_ssharp" XG_POST,                                   GDK_ssharp);
  DEFINE_INTEGER(XG_PRE "GDK_agrave" XG_POST,                                   GDK_agrave);
  DEFINE_INTEGER(XG_PRE "GDK_aacute" XG_POST,                                   GDK_aacute);
  DEFINE_INTEGER(XG_PRE "GDK_acircumflex" XG_POST,                              GDK_acircumflex);
  DEFINE_INTEGER(XG_PRE "GDK_atilde" XG_POST,                                   GDK_atilde);
  DEFINE_INTEGER(XG_PRE "GDK_adiaeresis" XG_POST,                               GDK_adiaeresis);
  DEFINE_INTEGER(XG_PRE "GDK_aring" XG_POST,                                    GDK_aring);
  DEFINE_INTEGER(XG_PRE "GDK_ae" XG_POST,                                       GDK_ae);
  DEFINE_INTEGER(XG_PRE "GDK_ccedilla" XG_POST,                                 GDK_ccedilla);
  DEFINE_INTEGER(XG_PRE "GDK_egrave" XG_POST,                                   GDK_egrave);
  DEFINE_INTEGER(XG_PRE "GDK_eacute" XG_POST,                                   GDK_eacute);
  DEFINE_INTEGER(XG_PRE "GDK_ecircumflex" XG_POST,                              GDK_ecircumflex);
  DEFINE_INTEGER(XG_PRE "GDK_ediaeresis" XG_POST,                               GDK_ediaeresis);
  DEFINE_INTEGER(XG_PRE "GDK_igrave" XG_POST,                                   GDK_igrave);
  DEFINE_INTEGER(XG_PRE "GDK_iacute" XG_POST,                                   GDK_iacute);
  DEFINE_INTEGER(XG_PRE "GDK_icircumflex" XG_POST,                              GDK_icircumflex);
  DEFINE_INTEGER(XG_PRE "GDK_idiaeresis" XG_POST,                               GDK_idiaeresis);
  DEFINE_INTEGER(XG_PRE "GDK_eth" XG_POST,                                      GDK_eth);
  DEFINE_INTEGER(XG_PRE "GDK_ntilde" XG_POST,                                   GDK_ntilde);
  DEFINE_INTEGER(XG_PRE "GDK_ograve" XG_POST,                                   GDK_ograve);
  DEFINE_INTEGER(XG_PRE "GDK_oacute" XG_POST,                                   GDK_oacute);
  DEFINE_INTEGER(XG_PRE "GDK_ocircumflex" XG_POST,                              GDK_ocircumflex);
  DEFINE_INTEGER(XG_PRE "GDK_otilde" XG_POST,                                   GDK_otilde);
  DEFINE_INTEGER(XG_PRE "GDK_odiaeresis" XG_POST,                               GDK_odiaeresis);
  DEFINE_INTEGER(XG_PRE "GDK_division" XG_POST,                                 GDK_division);
  DEFINE_INTEGER(XG_PRE "GDK_oslash" XG_POST,                                   GDK_oslash);
  DEFINE_INTEGER(XG_PRE "GDK_ugrave" XG_POST,                                   GDK_ugrave);
  DEFINE_INTEGER(XG_PRE "GDK_uacute" XG_POST,                                   GDK_uacute);
  DEFINE_INTEGER(XG_PRE "GDK_ucircumflex" XG_POST,                              GDK_ucircumflex);
  DEFINE_INTEGER(XG_PRE "GDK_udiaeresis" XG_POST,                               GDK_udiaeresis);
  DEFINE_INTEGER(XG_PRE "GDK_yacute" XG_POST,                                   GDK_yacute);
  DEFINE_INTEGER(XG_PRE "GDK_thorn" XG_POST,                                    GDK_thorn);
  DEFINE_INTEGER(XG_PRE "GDK_ydiaeresis" XG_POST,                               GDK_ydiaeresis);
  DEFINE_INTEGER(XG_PRE "GDK_Aogonek" XG_POST,                                  GDK_Aogonek);
  DEFINE_INTEGER(XG_PRE "GDK_breve" XG_POST,                                    GDK_breve);
  DEFINE_INTEGER(XG_PRE "GDK_Lstroke" XG_POST,                                  GDK_Lstroke);
  DEFINE_INTEGER(XG_PRE "GDK_Lcaron" XG_POST,                                   GDK_Lcaron);
  DEFINE_INTEGER(XG_PRE "GDK_Sacute" XG_POST,                                   GDK_Sacute);
  DEFINE_INTEGER(XG_PRE "GDK_Scaron" XG_POST,                                   GDK_Scaron);
  DEFINE_INTEGER(XG_PRE "GDK_Scedilla" XG_POST,                                 GDK_Scedilla);
  DEFINE_INTEGER(XG_PRE "GDK_Tcaron" XG_POST,                                   GDK_Tcaron);
  DEFINE_INTEGER(XG_PRE "GDK_Zacute" XG_POST,                                   GDK_Zacute);
  DEFINE_INTEGER(XG_PRE "GDK_Zcaron" XG_POST,                                   GDK_Zcaron);
  DEFINE_INTEGER(XG_PRE "GDK_Zabovedot" XG_POST,                                GDK_Zabovedot);
  DEFINE_INTEGER(XG_PRE "GDK_aogonek" XG_POST,                                  GDK_aogonek);
  DEFINE_INTEGER(XG_PRE "GDK_ogonek" XG_POST,                                   GDK_ogonek);
  DEFINE_INTEGER(XG_PRE "GDK_lstroke" XG_POST,                                  GDK_lstroke);
  DEFINE_INTEGER(XG_PRE "GDK_lcaron" XG_POST,                                   GDK_lcaron);
  DEFINE_INTEGER(XG_PRE "GDK_sacute" XG_POST,                                   GDK_sacute);
  DEFINE_INTEGER(XG_PRE "GDK_caron" XG_POST,                                    GDK_caron);
  DEFINE_INTEGER(XG_PRE "GDK_scaron" XG_POST,                                   GDK_scaron);
  DEFINE_INTEGER(XG_PRE "GDK_scedilla" XG_POST,                                 GDK_scedilla);
  DEFINE_INTEGER(XG_PRE "GDK_tcaron" XG_POST,                                   GDK_tcaron);
  DEFINE_INTEGER(XG_PRE "GDK_zacute" XG_POST,                                   GDK_zacute);
  DEFINE_INTEGER(XG_PRE "GDK_doubleacute" XG_POST,                              GDK_doubleacute);
  DEFINE_INTEGER(XG_PRE "GDK_zcaron" XG_POST,                                   GDK_zcaron);
  DEFINE_INTEGER(XG_PRE "GDK_zabovedot" XG_POST,                                GDK_zabovedot);
  DEFINE_INTEGER(XG_PRE "GDK_Racute" XG_POST,                                   GDK_Racute);
  DEFINE_INTEGER(XG_PRE "GDK_Abreve" XG_POST,                                   GDK_Abreve);
  DEFINE_INTEGER(XG_PRE "GDK_Lacute" XG_POST,                                   GDK_Lacute);
  DEFINE_INTEGER(XG_PRE "GDK_Cacute" XG_POST,                                   GDK_Cacute);
  DEFINE_INTEGER(XG_PRE "GDK_Ccaron" XG_POST,                                   GDK_Ccaron);
  DEFINE_INTEGER(XG_PRE "GDK_Eogonek" XG_POST,                                  GDK_Eogonek);
  DEFINE_INTEGER(XG_PRE "GDK_Ecaron" XG_POST,                                   GDK_Ecaron);
  DEFINE_INTEGER(XG_PRE "GDK_Dcaron" XG_POST,                                   GDK_Dcaron);
  DEFINE_INTEGER(XG_PRE "GDK_Dstroke" XG_POST,                                  GDK_Dstroke);
  DEFINE_INTEGER(XG_PRE "GDK_Nacute" XG_POST,                                   GDK_Nacute);
  DEFINE_INTEGER(XG_PRE "GDK_Ncaron" XG_POST,                                   GDK_Ncaron);
  DEFINE_INTEGER(XG_PRE "GDK_Odoubleacute" XG_POST,                             GDK_Odoubleacute);
  DEFINE_INTEGER(XG_PRE "GDK_Rcaron" XG_POST,                                   GDK_Rcaron);
  DEFINE_INTEGER(XG_PRE "GDK_Uring" XG_POST,                                    GDK_Uring);
  DEFINE_INTEGER(XG_PRE "GDK_Udoubleacute" XG_POST,                             GDK_Udoubleacute);
  DEFINE_INTEGER(XG_PRE "GDK_Tcedilla" XG_POST,                                 GDK_Tcedilla);
  DEFINE_INTEGER(XG_PRE "GDK_racute" XG_POST,                                   GDK_racute);
  DEFINE_INTEGER(XG_PRE "GDK_abreve" XG_POST,                                   GDK_abreve);
  DEFINE_INTEGER(XG_PRE "GDK_lacute" XG_POST,                                   GDK_lacute);
  DEFINE_INTEGER(XG_PRE "GDK_cacute" XG_POST,                                   GDK_cacute);
  DEFINE_INTEGER(XG_PRE "GDK_ccaron" XG_POST,                                   GDK_ccaron);
  DEFINE_INTEGER(XG_PRE "GDK_eogonek" XG_POST,                                  GDK_eogonek);
  DEFINE_INTEGER(XG_PRE "GDK_ecaron" XG_POST,                                   GDK_ecaron);
  DEFINE_INTEGER(XG_PRE "GDK_dcaron" XG_POST,                                   GDK_dcaron);
  DEFINE_INTEGER(XG_PRE "GDK_dstroke" XG_POST,                                  GDK_dstroke);
  DEFINE_INTEGER(XG_PRE "GDK_nacute" XG_POST,                                   GDK_nacute);
  DEFINE_INTEGER(XG_PRE "GDK_ncaron" XG_POST,                                   GDK_ncaron);
  DEFINE_INTEGER(XG_PRE "GDK_odoubleacute" XG_POST,                             GDK_odoubleacute);
  DEFINE_INTEGER(XG_PRE "GDK_udoubleacute" XG_POST,                             GDK_udoubleacute);
  DEFINE_INTEGER(XG_PRE "GDK_rcaron" XG_POST,                                   GDK_rcaron);
  DEFINE_INTEGER(XG_PRE "GDK_uring" XG_POST,                                    GDK_uring);
  DEFINE_INTEGER(XG_PRE "GDK_tcedilla" XG_POST,                                 GDK_tcedilla);
  DEFINE_INTEGER(XG_PRE "GDK_abovedot" XG_POST,                                 GDK_abovedot);
  DEFINE_INTEGER(XG_PRE "GDK_Hstroke" XG_POST,                                  GDK_Hstroke);
  DEFINE_INTEGER(XG_PRE "GDK_Hcircumflex" XG_POST,                              GDK_Hcircumflex);
  DEFINE_INTEGER(XG_PRE "GDK_Iabovedot" XG_POST,                                GDK_Iabovedot);
  DEFINE_INTEGER(XG_PRE "GDK_Gbreve" XG_POST,                                   GDK_Gbreve);
  DEFINE_INTEGER(XG_PRE "GDK_Jcircumflex" XG_POST,                              GDK_Jcircumflex);
  DEFINE_INTEGER(XG_PRE "GDK_hstroke" XG_POST,                                  GDK_hstroke);
  DEFINE_INTEGER(XG_PRE "GDK_hcircumflex" XG_POST,                              GDK_hcircumflex);
  DEFINE_INTEGER(XG_PRE "GDK_idotless" XG_POST,                                 GDK_idotless);
  DEFINE_INTEGER(XG_PRE "GDK_gbreve" XG_POST,                                   GDK_gbreve);
  DEFINE_INTEGER(XG_PRE "GDK_jcircumflex" XG_POST,                              GDK_jcircumflex);
  DEFINE_INTEGER(XG_PRE "GDK_Cabovedot" XG_POST,                                GDK_Cabovedot);
  DEFINE_INTEGER(XG_PRE "GDK_Ccircumflex" XG_POST,                              GDK_Ccircumflex);
  DEFINE_INTEGER(XG_PRE "GDK_Gabovedot" XG_POST,                                GDK_Gabovedot);
  DEFINE_INTEGER(XG_PRE "GDK_Gcircumflex" XG_POST,                              GDK_Gcircumflex);
  DEFINE_INTEGER(XG_PRE "GDK_Ubreve" XG_POST,                                   GDK_Ubreve);
  DEFINE_INTEGER(XG_PRE "GDK_Scircumflex" XG_POST,                              GDK_Scircumflex);
  DEFINE_INTEGER(XG_PRE "GDK_cabovedot" XG_POST,                                GDK_cabovedot);
  DEFINE_INTEGER(XG_PRE "GDK_ccircumflex" XG_POST,                              GDK_ccircumflex);
  DEFINE_INTEGER(XG_PRE "GDK_gabovedot" XG_POST,                                GDK_gabovedot);
  DEFINE_INTEGER(XG_PRE "GDK_gcircumflex" XG_POST,                              GDK_gcircumflex);
  DEFINE_INTEGER(XG_PRE "GDK_ubreve" XG_POST,                                   GDK_ubreve);
  DEFINE_INTEGER(XG_PRE "GDK_scircumflex" XG_POST,                              GDK_scircumflex);
  DEFINE_INTEGER(XG_PRE "GDK_kra" XG_POST,                                      GDK_kra);
  DEFINE_INTEGER(XG_PRE "GDK_kappa" XG_POST,                                    GDK_kappa);
  DEFINE_INTEGER(XG_PRE "GDK_Rcedilla" XG_POST,                                 GDK_Rcedilla);
  DEFINE_INTEGER(XG_PRE "GDK_Itilde" XG_POST,                                   GDK_Itilde);
  DEFINE_INTEGER(XG_PRE "GDK_Lcedilla" XG_POST,                                 GDK_Lcedilla);
  DEFINE_INTEGER(XG_PRE "GDK_Emacron" XG_POST,                                  GDK_Emacron);
  DEFINE_INTEGER(XG_PRE "GDK_Gcedilla" XG_POST,                                 GDK_Gcedilla);
  DEFINE_INTEGER(XG_PRE "GDK_Tslash" XG_POST,                                   GDK_Tslash);
  DEFINE_INTEGER(XG_PRE "GDK_rcedilla" XG_POST,                                 GDK_rcedilla);
  DEFINE_INTEGER(XG_PRE "GDK_itilde" XG_POST,                                   GDK_itilde);
  DEFINE_INTEGER(XG_PRE "GDK_lcedilla" XG_POST,                                 GDK_lcedilla);
  DEFINE_INTEGER(XG_PRE "GDK_emacron" XG_POST,                                  GDK_emacron);
  DEFINE_INTEGER(XG_PRE "GDK_gcedilla" XG_POST,                                 GDK_gcedilla);
  DEFINE_INTEGER(XG_PRE "GDK_tslash" XG_POST,                                   GDK_tslash);
  DEFINE_INTEGER(XG_PRE "GDK_ENG" XG_POST,                                      GDK_ENG);
  DEFINE_INTEGER(XG_PRE "GDK_eng" XG_POST,                                      GDK_eng);
  DEFINE_INTEGER(XG_PRE "GDK_Amacron" XG_POST,                                  GDK_Amacron);
  DEFINE_INTEGER(XG_PRE "GDK_Iogonek" XG_POST,                                  GDK_Iogonek);
  DEFINE_INTEGER(XG_PRE "GDK_Eabovedot" XG_POST,                                GDK_Eabovedot);
  DEFINE_INTEGER(XG_PRE "GDK_Imacron" XG_POST,                                  GDK_Imacron);
  DEFINE_INTEGER(XG_PRE "GDK_Ncedilla" XG_POST,                                 GDK_Ncedilla);
  DEFINE_INTEGER(XG_PRE "GDK_Omacron" XG_POST,                                  GDK_Omacron);
  DEFINE_INTEGER(XG_PRE "GDK_Kcedilla" XG_POST,                                 GDK_Kcedilla);
  DEFINE_INTEGER(XG_PRE "GDK_Uogonek" XG_POST,                                  GDK_Uogonek);
  DEFINE_INTEGER(XG_PRE "GDK_Utilde" XG_POST,                                   GDK_Utilde);
  DEFINE_INTEGER(XG_PRE "GDK_Umacron" XG_POST,                                  GDK_Umacron);
  DEFINE_INTEGER(XG_PRE "GDK_amacron" XG_POST,                                  GDK_amacron);
  DEFINE_INTEGER(XG_PRE "GDK_iogonek" XG_POST,                                  GDK_iogonek);
  DEFINE_INTEGER(XG_PRE "GDK_eabovedot" XG_POST,                                GDK_eabovedot);
  DEFINE_INTEGER(XG_PRE "GDK_imacron" XG_POST,                                  GDK_imacron);
  DEFINE_INTEGER(XG_PRE "GDK_ncedilla" XG_POST,                                 GDK_ncedilla);
  DEFINE_INTEGER(XG_PRE "GDK_omacron" XG_POST,                                  GDK_omacron);
  DEFINE_INTEGER(XG_PRE "GDK_kcedilla" XG_POST,                                 GDK_kcedilla);
  DEFINE_INTEGER(XG_PRE "GDK_uogonek" XG_POST,                                  GDK_uogonek);
  DEFINE_INTEGER(XG_PRE "GDK_utilde" XG_POST,                                   GDK_utilde);
  DEFINE_INTEGER(XG_PRE "GDK_umacron" XG_POST,                                  GDK_umacron);
#if (!HAVE_GTK_1)
  DEFINE_INTEGER(XG_PRE "GDK_OE" XG_POST,                                       GDK_OE);
  DEFINE_INTEGER(XG_PRE "GDK_oe" XG_POST,                                       GDK_oe);
  DEFINE_INTEGER(XG_PRE "GDK_Ydiaeresis" XG_POST,                               GDK_Ydiaeresis);
#endif
  DEFINE_INTEGER(XG_PRE "GDK_overline" XG_POST,                                 GDK_overline);
  DEFINE_INTEGER(XG_PRE "GDK_PROP_MODE_REPLACE" XG_POST,                        GDK_PROP_MODE_REPLACE);
  DEFINE_INTEGER(XG_PRE "GDK_PROP_MODE_PREPEND" XG_POST,                        GDK_PROP_MODE_PREPEND);
  DEFINE_INTEGER(XG_PRE "GDK_PROP_MODE_APPEND" XG_POST,                         GDK_PROP_MODE_APPEND);
  DEFINE_INTEGER(XG_PRE "GDK_RGB_DITHER_NONE" XG_POST,                          GDK_RGB_DITHER_NONE);
  DEFINE_INTEGER(XG_PRE "GDK_RGB_DITHER_NORMAL" XG_POST,                        GDK_RGB_DITHER_NORMAL);
  DEFINE_INTEGER(XG_PRE "GDK_RGB_DITHER_MAX" XG_POST,                           GDK_RGB_DITHER_MAX);
  DEFINE_INTEGER(XG_PRE "GDK_CURRENT_TIME" XG_POST,                             GDK_CURRENT_TIME);
  DEFINE_INTEGER(XG_PRE "GDK_PARENT_RELATIVE" XG_POST,                          GDK_PARENT_RELATIVE);
  DEFINE_INTEGER(XG_PRE "GDK_LSB_FIRST" XG_POST,                                GDK_LSB_FIRST);
  DEFINE_INTEGER(XG_PRE "GDK_MSB_FIRST" XG_POST,                                GDK_MSB_FIRST);
  DEFINE_INTEGER(XG_PRE "GDK_SHIFT_MASK" XG_POST,                               GDK_SHIFT_MASK);
  DEFINE_INTEGER(XG_PRE "GDK_LOCK_MASK" XG_POST,                                GDK_LOCK_MASK);
  DEFINE_INTEGER(XG_PRE "GDK_CONTROL_MASK" XG_POST,                             GDK_CONTROL_MASK);
  DEFINE_INTEGER(XG_PRE "GDK_MOD1_MASK" XG_POST,                                GDK_MOD1_MASK);
  DEFINE_INTEGER(XG_PRE "GDK_MOD2_MASK" XG_POST,                                GDK_MOD2_MASK);
  DEFINE_INTEGER(XG_PRE "GDK_MOD3_MASK" XG_POST,                                GDK_MOD3_MASK);
  DEFINE_INTEGER(XG_PRE "GDK_MOD4_MASK" XG_POST,                                GDK_MOD4_MASK);
  DEFINE_INTEGER(XG_PRE "GDK_MOD5_MASK" XG_POST,                                GDK_MOD5_MASK);
  DEFINE_INTEGER(XG_PRE "GDK_BUTTON1_MASK" XG_POST,                             GDK_BUTTON1_MASK);
  DEFINE_INTEGER(XG_PRE "GDK_BUTTON2_MASK" XG_POST,                             GDK_BUTTON2_MASK);
  DEFINE_INTEGER(XG_PRE "GDK_BUTTON3_MASK" XG_POST,                             GDK_BUTTON3_MASK);
  DEFINE_INTEGER(XG_PRE "GDK_BUTTON4_MASK" XG_POST,                             GDK_BUTTON4_MASK);
  DEFINE_INTEGER(XG_PRE "GDK_BUTTON5_MASK" XG_POST,                             GDK_BUTTON5_MASK);
  DEFINE_INTEGER(XG_PRE "GDK_RELEASE_MASK" XG_POST,                             GDK_RELEASE_MASK);
  DEFINE_INTEGER(XG_PRE "GDK_MODIFIER_MASK" XG_POST,                            GDK_MODIFIER_MASK);
  DEFINE_INTEGER(XG_PRE "GDK_INPUT_READ" XG_POST,                               GDK_INPUT_READ);
  DEFINE_INTEGER(XG_PRE "GDK_INPUT_WRITE" XG_POST,                              GDK_INPUT_WRITE);
  DEFINE_INTEGER(XG_PRE "GDK_INPUT_EXCEPTION" XG_POST,                          GDK_INPUT_EXCEPTION);
  DEFINE_INTEGER(XG_PRE "GDK_OK" XG_POST,                                       GDK_OK);
  DEFINE_INTEGER(XG_PRE "GDK_ERROR" XG_POST,                                    GDK_ERROR);
  DEFINE_INTEGER(XG_PRE "GDK_ERROR_PARAM" XG_POST,                              GDK_ERROR_PARAM);
  DEFINE_INTEGER(XG_PRE "GDK_ERROR_FILE" XG_POST,                               GDK_ERROR_FILE);
  DEFINE_INTEGER(XG_PRE "GDK_ERROR_MEM" XG_POST,                                GDK_ERROR_MEM);
#if (!HAVE_GTK_1)
  DEFINE_INTEGER(XG_PRE "GDK_GRAB_SUCCESS" XG_POST,                             GDK_GRAB_SUCCESS);
  DEFINE_INTEGER(XG_PRE "GDK_GRAB_ALREADY_GRABBED" XG_POST,                     GDK_GRAB_ALREADY_GRABBED);
  DEFINE_INTEGER(XG_PRE "GDK_GRAB_INVALID_TIME" XG_POST,                        GDK_GRAB_INVALID_TIME);
  DEFINE_INTEGER(XG_PRE "GDK_GRAB_NOT_VIEWABLE" XG_POST,                        GDK_GRAB_NOT_VIEWABLE);
  DEFINE_INTEGER(XG_PRE "GDK_GRAB_FROZEN" XG_POST,                              GDK_GRAB_FROZEN);
#endif
  DEFINE_INTEGER(XG_PRE "GDK_VISUAL_STATIC_GRAY" XG_POST,                       GDK_VISUAL_STATIC_GRAY);
  DEFINE_INTEGER(XG_PRE "GDK_VISUAL_GRAYSCALE" XG_POST,                         GDK_VISUAL_GRAYSCALE);
  DEFINE_INTEGER(XG_PRE "GDK_VISUAL_STATIC_COLOR" XG_POST,                      GDK_VISUAL_STATIC_COLOR);
  DEFINE_INTEGER(XG_PRE "GDK_VISUAL_PSEUDO_COLOR" XG_POST,                      GDK_VISUAL_PSEUDO_COLOR);
  DEFINE_INTEGER(XG_PRE "GDK_VISUAL_TRUE_COLOR" XG_POST,                        GDK_VISUAL_TRUE_COLOR);
  DEFINE_INTEGER(XG_PRE "GDK_VISUAL_DIRECT_COLOR" XG_POST,                      GDK_VISUAL_DIRECT_COLOR);
  DEFINE_INTEGER(XG_PRE "GDK_INPUT_OUTPUT" XG_POST,                             GDK_INPUT_OUTPUT);
  DEFINE_INTEGER(XG_PRE "GDK_INPUT_ONLY" XG_POST,                               GDK_INPUT_ONLY);
  DEFINE_INTEGER(XG_PRE "GDK_WINDOW_ROOT" XG_POST,                              GDK_WINDOW_ROOT);
  DEFINE_INTEGER(XG_PRE "GDK_WINDOW_TOPLEVEL" XG_POST,                          GDK_WINDOW_TOPLEVEL);
  DEFINE_INTEGER(XG_PRE "GDK_WINDOW_CHILD" XG_POST,                             GDK_WINDOW_CHILD);
  DEFINE_INTEGER(XG_PRE "GDK_WINDOW_DIALOG" XG_POST,                            GDK_WINDOW_DIALOG);
  DEFINE_INTEGER(XG_PRE "GDK_WINDOW_TEMP" XG_POST,                              GDK_WINDOW_TEMP);
  DEFINE_INTEGER(XG_PRE "GDK_WINDOW_FOREIGN" XG_POST,                           GDK_WINDOW_FOREIGN);
  DEFINE_INTEGER(XG_PRE "GDK_WA_TITLE" XG_POST,                                 GDK_WA_TITLE);
  DEFINE_INTEGER(XG_PRE "GDK_WA_X" XG_POST,                                     GDK_WA_X);
  DEFINE_INTEGER(XG_PRE "GDK_WA_Y" XG_POST,                                     GDK_WA_Y);
  DEFINE_INTEGER(XG_PRE "GDK_WA_CURSOR" XG_POST,                                GDK_WA_CURSOR);
  DEFINE_INTEGER(XG_PRE "GDK_WA_COLORMAP" XG_POST,                              GDK_WA_COLORMAP);
  DEFINE_INTEGER(XG_PRE "GDK_WA_VISUAL" XG_POST,                                GDK_WA_VISUAL);
  DEFINE_INTEGER(XG_PRE "GDK_WA_WMCLASS" XG_POST,                               GDK_WA_WMCLASS);
  DEFINE_INTEGER(XG_PRE "GDK_WA_NOREDIR" XG_POST,                               GDK_WA_NOREDIR);
  DEFINE_INTEGER(XG_PRE "GDK_HINT_POS" XG_POST,                                 GDK_HINT_POS);
  DEFINE_INTEGER(XG_PRE "GDK_HINT_MIN_SIZE" XG_POST,                            GDK_HINT_MIN_SIZE);
  DEFINE_INTEGER(XG_PRE "GDK_HINT_MAX_SIZE" XG_POST,                            GDK_HINT_MAX_SIZE);
  DEFINE_INTEGER(XG_PRE "GDK_HINT_BASE_SIZE" XG_POST,                           GDK_HINT_BASE_SIZE);
  DEFINE_INTEGER(XG_PRE "GDK_HINT_ASPECT" XG_POST,                              GDK_HINT_ASPECT);
  DEFINE_INTEGER(XG_PRE "GDK_HINT_RESIZE_INC" XG_POST,                          GDK_HINT_RESIZE_INC);
#if (!HAVE_GTK_1)
  DEFINE_INTEGER(XG_PRE "GDK_HINT_WIN_GRAVITY" XG_POST,                         GDK_HINT_WIN_GRAVITY);
  DEFINE_INTEGER(XG_PRE "GDK_HINT_USER_POS" XG_POST,                            GDK_HINT_USER_POS);
  DEFINE_INTEGER(XG_PRE "GDK_HINT_USER_SIZE" XG_POST,                           GDK_HINT_USER_SIZE);
  DEFINE_INTEGER(XG_PRE "GDK_WINDOW_TYPE_HINT_NORMAL" XG_POST,                  GDK_WINDOW_TYPE_HINT_NORMAL);
  DEFINE_INTEGER(XG_PRE "GDK_WINDOW_TYPE_HINT_DIALOG" XG_POST,                  GDK_WINDOW_TYPE_HINT_DIALOG);
  DEFINE_INTEGER(XG_PRE "GDK_WINDOW_TYPE_HINT_MENU" XG_POST,                    GDK_WINDOW_TYPE_HINT_MENU);
  DEFINE_INTEGER(XG_PRE "GDK_WINDOW_TYPE_HINT_TOOLBAR" XG_POST,                 GDK_WINDOW_TYPE_HINT_TOOLBAR);
#endif
  DEFINE_INTEGER(XG_PRE "GDK_DECOR_ALL" XG_POST,                                GDK_DECOR_ALL);
  DEFINE_INTEGER(XG_PRE "GDK_DECOR_BORDER" XG_POST,                             GDK_DECOR_BORDER);
  DEFINE_INTEGER(XG_PRE "GDK_DECOR_RESIZEH" XG_POST,                            GDK_DECOR_RESIZEH);
  DEFINE_INTEGER(XG_PRE "GDK_DECOR_TITLE" XG_POST,                              GDK_DECOR_TITLE);
  DEFINE_INTEGER(XG_PRE "GDK_DECOR_MENU" XG_POST,                               GDK_DECOR_MENU);
  DEFINE_INTEGER(XG_PRE "GDK_DECOR_MINIMIZE" XG_POST,                           GDK_DECOR_MINIMIZE);
  DEFINE_INTEGER(XG_PRE "GDK_DECOR_MAXIMIZE" XG_POST,                           GDK_DECOR_MAXIMIZE);
  DEFINE_INTEGER(XG_PRE "GDK_FUNC_ALL" XG_POST,                                 GDK_FUNC_ALL);
  DEFINE_INTEGER(XG_PRE "GDK_FUNC_RESIZE" XG_POST,                              GDK_FUNC_RESIZE);
  DEFINE_INTEGER(XG_PRE "GDK_FUNC_MOVE" XG_POST,                                GDK_FUNC_MOVE);
  DEFINE_INTEGER(XG_PRE "GDK_FUNC_MINIMIZE" XG_POST,                            GDK_FUNC_MINIMIZE);
  DEFINE_INTEGER(XG_PRE "GDK_FUNC_MAXIMIZE" XG_POST,                            GDK_FUNC_MAXIMIZE);
  DEFINE_INTEGER(XG_PRE "GDK_FUNC_CLOSE" XG_POST,                               GDK_FUNC_CLOSE);
#if (!HAVE_GTK_1)
  DEFINE_INTEGER(XG_PRE "GDK_GRAVITY_NORTH_WEST" XG_POST,                       GDK_GRAVITY_NORTH_WEST);
  DEFINE_INTEGER(XG_PRE "GDK_GRAVITY_NORTH" XG_POST,                            GDK_GRAVITY_NORTH);
  DEFINE_INTEGER(XG_PRE "GDK_GRAVITY_NORTH_EAST" XG_POST,                       GDK_GRAVITY_NORTH_EAST);
  DEFINE_INTEGER(XG_PRE "GDK_GRAVITY_WEST" XG_POST,                             GDK_GRAVITY_WEST);
  DEFINE_INTEGER(XG_PRE "GDK_GRAVITY_CENTER" XG_POST,                           GDK_GRAVITY_CENTER);
  DEFINE_INTEGER(XG_PRE "GDK_GRAVITY_EAST" XG_POST,                             GDK_GRAVITY_EAST);
  DEFINE_INTEGER(XG_PRE "GDK_GRAVITY_SOUTH_WEST" XG_POST,                       GDK_GRAVITY_SOUTH_WEST);
  DEFINE_INTEGER(XG_PRE "GDK_GRAVITY_SOUTH" XG_POST,                            GDK_GRAVITY_SOUTH);
  DEFINE_INTEGER(XG_PRE "GDK_GRAVITY_SOUTH_EAST" XG_POST,                       GDK_GRAVITY_SOUTH_EAST);
  DEFINE_INTEGER(XG_PRE "GDK_GRAVITY_STATIC" XG_POST,                           GDK_GRAVITY_STATIC);
  DEFINE_INTEGER(XG_PRE "GDK_WINDOW_EDGE_NORTH_WEST" XG_POST,                   GDK_WINDOW_EDGE_NORTH_WEST);
  DEFINE_INTEGER(XG_PRE "GDK_WINDOW_EDGE_NORTH" XG_POST,                        GDK_WINDOW_EDGE_NORTH);
  DEFINE_INTEGER(XG_PRE "GDK_WINDOW_EDGE_NORTH_EAST" XG_POST,                   GDK_WINDOW_EDGE_NORTH_EAST);
  DEFINE_INTEGER(XG_PRE "GDK_WINDOW_EDGE_WEST" XG_POST,                         GDK_WINDOW_EDGE_WEST);
  DEFINE_INTEGER(XG_PRE "GDK_WINDOW_EDGE_EAST" XG_POST,                         GDK_WINDOW_EDGE_EAST);
  DEFINE_INTEGER(XG_PRE "GDK_WINDOW_EDGE_SOUTH_WEST" XG_POST,                   GDK_WINDOW_EDGE_SOUTH_WEST);
  DEFINE_INTEGER(XG_PRE "GDK_WINDOW_EDGE_SOUTH" XG_POST,                        GDK_WINDOW_EDGE_SOUTH);
  DEFINE_INTEGER(XG_PRE "GDK_WINDOW_EDGE_SOUTH_EAST" XG_POST,                   GDK_WINDOW_EDGE_SOUTH_EAST);
  DEFINE_INTEGER(XG_PRE "GDK_PIXBUF_ALPHA_BILEVEL" XG_POST,                     GDK_PIXBUF_ALPHA_BILEVEL);
  DEFINE_INTEGER(XG_PRE "GDK_PIXBUF_ALPHA_FULL" XG_POST,                        GDK_PIXBUF_ALPHA_FULL);
  DEFINE_INTEGER(XG_PRE "GDK_COLORSPACE_RGB" XG_POST,                           GDK_COLORSPACE_RGB);
  DEFINE_INTEGER(XG_PRE "GDK_PIXBUF_ERROR_CORRUPT_IMAGE" XG_POST,               GDK_PIXBUF_ERROR_CORRUPT_IMAGE);
  DEFINE_INTEGER(XG_PRE "GDK_PIXBUF_ERROR_INSUFFICIENT_MEMORY" XG_POST,         GDK_PIXBUF_ERROR_INSUFFICIENT_MEMORY);
  DEFINE_INTEGER(XG_PRE "GDK_PIXBUF_ERROR_BAD_OPTION" XG_POST,                  GDK_PIXBUF_ERROR_BAD_OPTION);
  DEFINE_INTEGER(XG_PRE "GDK_PIXBUF_ERROR_UNKNOWN_TYPE" XG_POST,                GDK_PIXBUF_ERROR_UNKNOWN_TYPE);
  DEFINE_INTEGER(XG_PRE "GDK_PIXBUF_ERROR_UNSUPPORTED_OPERATION" XG_POST,       GDK_PIXBUF_ERROR_UNSUPPORTED_OPERATION);
  DEFINE_INTEGER(XG_PRE "GDK_PIXBUF_ERROR_FAILED" XG_POST,                      GDK_PIXBUF_ERROR_FAILED);
  DEFINE_INTEGER(XG_PRE "GDK_INTERP_NEAREST" XG_POST,                           GDK_INTERP_NEAREST);
  DEFINE_INTEGER(XG_PRE "GDK_INTERP_TILES" XG_POST,                             GDK_INTERP_TILES);
  DEFINE_INTEGER(XG_PRE "GDK_INTERP_BILINEAR" XG_POST,                          GDK_INTERP_BILINEAR);
  DEFINE_INTEGER(XG_PRE "GDK_INTERP_HYPER" XG_POST,                             GDK_INTERP_HYPER);
  DEFINE_INTEGER(XG_PRE "GDK_PIXBUF_MAJOR" XG_POST,                             GDK_PIXBUF_MAJOR);
  DEFINE_INTEGER(XG_PRE "GDK_PIXBUF_MINOR" XG_POST,                             GDK_PIXBUF_MINOR);
  DEFINE_INTEGER(XG_PRE "GDK_PIXBUF_MICRO" XG_POST,                             GDK_PIXBUF_MICRO);
#endif
  DEFINE_INTEGER(XG_PRE "GTK_ACCEL_VISIBLE" XG_POST,                            GTK_ACCEL_VISIBLE);
  DEFINE_INTEGER(XG_PRE "GTK_ACCEL_LOCKED" XG_POST,                             GTK_ACCEL_LOCKED);
  DEFINE_INTEGER(XG_PRE "GTK_ACCEL_MASK" XG_POST,                               GTK_ACCEL_MASK);
  DEFINE_INTEGER(XG_PRE "GTK_BUTTONBOX_DEFAULT" XG_POST,                        GTK_BUTTONBOX_DEFAULT);
  DEFINE_INTEGER(XG_PRE "GTK_CALENDAR_SHOW_HEADING" XG_POST,                    GTK_CALENDAR_SHOW_HEADING);
  DEFINE_INTEGER(XG_PRE "GTK_CALENDAR_SHOW_DAY_NAMES" XG_POST,                  GTK_CALENDAR_SHOW_DAY_NAMES);
  DEFINE_INTEGER(XG_PRE "GTK_CALENDAR_NO_MONTH_CHANGE" XG_POST,                 GTK_CALENDAR_NO_MONTH_CHANGE);
  DEFINE_INTEGER(XG_PRE "GTK_CALENDAR_SHOW_WEEK_NUMBERS" XG_POST,               GTK_CALENDAR_SHOW_WEEK_NUMBERS);
  DEFINE_INTEGER(XG_PRE "GTK_CALENDAR_WEEK_START_MONDAY" XG_POST,               GTK_CALENDAR_WEEK_START_MONDAY);
#if (!HAVE_GTK_1)
  DEFINE_INTEGER(XG_PRE "GTK_CELL_RENDERER_SELECTED" XG_POST,                   GTK_CELL_RENDERER_SELECTED);
  DEFINE_INTEGER(XG_PRE "GTK_CELL_RENDERER_PRELIT" XG_POST,                     GTK_CELL_RENDERER_PRELIT);
  DEFINE_INTEGER(XG_PRE "GTK_CELL_RENDERER_INSENSITIVE" XG_POST,                GTK_CELL_RENDERER_INSENSITIVE);
  DEFINE_INTEGER(XG_PRE "GTK_CELL_RENDERER_SORTED" XG_POST,                     GTK_CELL_RENDERER_SORTED);
  DEFINE_INTEGER(XG_PRE "GTK_CELL_RENDERER_MODE_INERT" XG_POST,                 GTK_CELL_RENDERER_MODE_INERT);
  DEFINE_INTEGER(XG_PRE "GTK_CELL_RENDERER_MODE_ACTIVATABLE" XG_POST,           GTK_CELL_RENDERER_MODE_ACTIVATABLE);
  DEFINE_INTEGER(XG_PRE "GTK_CELL_RENDERER_MODE_EDITABLE" XG_POST,              GTK_CELL_RENDERER_MODE_EDITABLE);
  DEFINE_INTEGER(XG_PRE "GTK_DIALOG_MODAL" XG_POST,                             GTK_DIALOG_MODAL);
  DEFINE_INTEGER(XG_PRE "GTK_DIALOG_DESTROY_WITH_PARENT" XG_POST,               GTK_DIALOG_DESTROY_WITH_PARENT);
  DEFINE_INTEGER(XG_PRE "GTK_DIALOG_NO_SEPARATOR" XG_POST,                      GTK_DIALOG_NO_SEPARATOR);
  DEFINE_INTEGER(XG_PRE "GTK_RESPONSE_NONE" XG_POST,                            GTK_RESPONSE_NONE);
  DEFINE_INTEGER(XG_PRE "GTK_RESPONSE_REJECT" XG_POST,                          GTK_RESPONSE_REJECT);
  DEFINE_INTEGER(XG_PRE "GTK_RESPONSE_ACCEPT" XG_POST,                          GTK_RESPONSE_ACCEPT);
  DEFINE_INTEGER(XG_PRE "GTK_RESPONSE_DELETE_EVENT" XG_POST,                    GTK_RESPONSE_DELETE_EVENT);
  DEFINE_INTEGER(XG_PRE "GTK_RESPONSE_OK" XG_POST,                              GTK_RESPONSE_OK);
  DEFINE_INTEGER(XG_PRE "GTK_RESPONSE_CANCEL" XG_POST,                          GTK_RESPONSE_CANCEL);
  DEFINE_INTEGER(XG_PRE "GTK_RESPONSE_CLOSE" XG_POST,                           GTK_RESPONSE_CLOSE);
  DEFINE_INTEGER(XG_PRE "GTK_RESPONSE_YES" XG_POST,                             GTK_RESPONSE_YES);
  DEFINE_INTEGER(XG_PRE "GTK_RESPONSE_NO" XG_POST,                              GTK_RESPONSE_NO);
  DEFINE_INTEGER(XG_PRE "GTK_RESPONSE_APPLY" XG_POST,                           GTK_RESPONSE_APPLY);
  DEFINE_INTEGER(XG_PRE "GTK_RESPONSE_HELP" XG_POST,                            GTK_RESPONSE_HELP);
#endif
  DEFINE_INTEGER(XG_PRE "GTK_DEST_DEFAULT_MOTION" XG_POST,                      GTK_DEST_DEFAULT_MOTION);
  DEFINE_INTEGER(XG_PRE "GTK_DEST_DEFAULT_HIGHLIGHT" XG_POST,                   GTK_DEST_DEFAULT_HIGHLIGHT);
  DEFINE_INTEGER(XG_PRE "GTK_DEST_DEFAULT_DROP" XG_POST,                        GTK_DEST_DEFAULT_DROP);
  DEFINE_INTEGER(XG_PRE "GTK_DEST_DEFAULT_ALL" XG_POST,                         GTK_DEST_DEFAULT_ALL);
  DEFINE_INTEGER(XG_PRE "GTK_TARGET_SAME_APP" XG_POST,                          GTK_TARGET_SAME_APP);
  DEFINE_INTEGER(XG_PRE "GTK_TARGET_SAME_WIDGET" XG_POST,                       GTK_TARGET_SAME_WIDGET);
  DEFINE_INTEGER(XG_PRE "GTK_ANCHOR_CENTER" XG_POST,                            GTK_ANCHOR_CENTER);
  DEFINE_INTEGER(XG_PRE "GTK_ANCHOR_NORTH" XG_POST,                             GTK_ANCHOR_NORTH);
  DEFINE_INTEGER(XG_PRE "GTK_ANCHOR_NORTH_WEST" XG_POST,                        GTK_ANCHOR_NORTH_WEST);
  DEFINE_INTEGER(XG_PRE "GTK_ANCHOR_NORTH_EAST" XG_POST,                        GTK_ANCHOR_NORTH_EAST);
  DEFINE_INTEGER(XG_PRE "GTK_ANCHOR_SOUTH" XG_POST,                             GTK_ANCHOR_SOUTH);
  DEFINE_INTEGER(XG_PRE "GTK_ANCHOR_SOUTH_WEST" XG_POST,                        GTK_ANCHOR_SOUTH_WEST);
  DEFINE_INTEGER(XG_PRE "GTK_ANCHOR_SOUTH_EAST" XG_POST,                        GTK_ANCHOR_SOUTH_EAST);
  DEFINE_INTEGER(XG_PRE "GTK_ANCHOR_WEST" XG_POST,                              GTK_ANCHOR_WEST);
  DEFINE_INTEGER(XG_PRE "GTK_ANCHOR_EAST" XG_POST,                              GTK_ANCHOR_EAST);
  DEFINE_INTEGER(XG_PRE "GTK_ANCHOR_N" XG_POST,                                 GTK_ANCHOR_N);
  DEFINE_INTEGER(XG_PRE "GTK_ANCHOR_NW" XG_POST,                                GTK_ANCHOR_NW);
  DEFINE_INTEGER(XG_PRE "GTK_ANCHOR_NE" XG_POST,                                GTK_ANCHOR_NE);
  DEFINE_INTEGER(XG_PRE "GTK_ANCHOR_S" XG_POST,                                 GTK_ANCHOR_S);
  DEFINE_INTEGER(XG_PRE "GTK_ANCHOR_SW" XG_POST,                                GTK_ANCHOR_SW);
  DEFINE_INTEGER(XG_PRE "GTK_ANCHOR_SE" XG_POST,                                GTK_ANCHOR_SE);
  DEFINE_INTEGER(XG_PRE "GTK_ANCHOR_W" XG_POST,                                 GTK_ANCHOR_W);
  DEFINE_INTEGER(XG_PRE "GTK_ANCHOR_E" XG_POST,                                 GTK_ANCHOR_E);
  DEFINE_INTEGER(XG_PRE "GTK_ARROW_UP" XG_POST,                                 GTK_ARROW_UP);
  DEFINE_INTEGER(XG_PRE "GTK_ARROW_DOWN" XG_POST,                               GTK_ARROW_DOWN);
  DEFINE_INTEGER(XG_PRE "GTK_ARROW_LEFT" XG_POST,                               GTK_ARROW_LEFT);
  DEFINE_INTEGER(XG_PRE "GTK_ARROW_RIGHT" XG_POST,                              GTK_ARROW_RIGHT);
  DEFINE_INTEGER(XG_PRE "GTK_EXPAND" XG_POST,                                   GTK_EXPAND);
  DEFINE_INTEGER(XG_PRE "GTK_SHRINK" XG_POST,                                   GTK_SHRINK);
  DEFINE_INTEGER(XG_PRE "GTK_FILL" XG_POST,                                     GTK_FILL);
  DEFINE_INTEGER(XG_PRE "GTK_BUTTONBOX_DEFAULT_STYLE" XG_POST,                  GTK_BUTTONBOX_DEFAULT_STYLE);
  DEFINE_INTEGER(XG_PRE "GTK_BUTTONBOX_SPREAD" XG_POST,                         GTK_BUTTONBOX_SPREAD);
  DEFINE_INTEGER(XG_PRE "GTK_BUTTONBOX_EDGE" XG_POST,                           GTK_BUTTONBOX_EDGE);
  DEFINE_INTEGER(XG_PRE "GTK_BUTTONBOX_START" XG_POST,                          GTK_BUTTONBOX_START);
  DEFINE_INTEGER(XG_PRE "GTK_BUTTONBOX_END" XG_POST,                            GTK_BUTTONBOX_END);
#if (!HAVE_GTK_1)
  DEFINE_INTEGER(XG_PRE "GTK_CURVE_TYPE_LINEAR" XG_POST,                        GTK_CURVE_TYPE_LINEAR);
  DEFINE_INTEGER(XG_PRE "GTK_CURVE_TYPE_SPLINE" XG_POST,                        GTK_CURVE_TYPE_SPLINE);
  DEFINE_INTEGER(XG_PRE "GTK_CURVE_TYPE_FREE" XG_POST,                          GTK_CURVE_TYPE_FREE);
  DEFINE_INTEGER(XG_PRE "GTK_DELETE_CHARS" XG_POST,                             GTK_DELETE_CHARS);
  DEFINE_INTEGER(XG_PRE "GTK_DELETE_WORD_ENDS" XG_POST,                         GTK_DELETE_WORD_ENDS);
  DEFINE_INTEGER(XG_PRE "GTK_DELETE_WORDS" XG_POST,                             GTK_DELETE_WORDS);
  DEFINE_INTEGER(XG_PRE "GTK_DELETE_DISPLAY_LINES" XG_POST,                     GTK_DELETE_DISPLAY_LINES);
  DEFINE_INTEGER(XG_PRE "GTK_DELETE_DISPLAY_LINE_ENDS" XG_POST,                 GTK_DELETE_DISPLAY_LINE_ENDS);
  DEFINE_INTEGER(XG_PRE "GTK_DELETE_PARAGRAPH_ENDS" XG_POST,                    GTK_DELETE_PARAGRAPH_ENDS);
  DEFINE_INTEGER(XG_PRE "GTK_DELETE_PARAGRAPHS" XG_POST,                        GTK_DELETE_PARAGRAPHS);
  DEFINE_INTEGER(XG_PRE "GTK_DELETE_WHITESPACE" XG_POST,                        GTK_DELETE_WHITESPACE);
#endif
  DEFINE_INTEGER(XG_PRE "GTK_DIR_TAB_FORWARD" XG_POST,                          GTK_DIR_TAB_FORWARD);
  DEFINE_INTEGER(XG_PRE "GTK_DIR_TAB_BACKWARD" XG_POST,                         GTK_DIR_TAB_BACKWARD);
  DEFINE_INTEGER(XG_PRE "GTK_DIR_UP" XG_POST,                                   GTK_DIR_UP);
  DEFINE_INTEGER(XG_PRE "GTK_DIR_DOWN" XG_POST,                                 GTK_DIR_DOWN);
  DEFINE_INTEGER(XG_PRE "GTK_DIR_LEFT" XG_POST,                                 GTK_DIR_LEFT);
  DEFINE_INTEGER(XG_PRE "GTK_DIR_RIGHT" XG_POST,                                GTK_DIR_RIGHT);
#if (!HAVE_GTK_1)
  DEFINE_INTEGER(XG_PRE "GTK_EXPANDER_COLLAPSED" XG_POST,                       GTK_EXPANDER_COLLAPSED);
  DEFINE_INTEGER(XG_PRE "GTK_EXPANDER_SEMI_COLLAPSED" XG_POST,                  GTK_EXPANDER_SEMI_COLLAPSED);
  DEFINE_INTEGER(XG_PRE "GTK_EXPANDER_SEMI_EXPANDED" XG_POST,                   GTK_EXPANDER_SEMI_EXPANDED);
  DEFINE_INTEGER(XG_PRE "GTK_EXPANDER_EXPANDED" XG_POST,                        GTK_EXPANDER_EXPANDED);
  DEFINE_INTEGER(XG_PRE "GTK_ICON_SIZE_INVALID" XG_POST,                        GTK_ICON_SIZE_INVALID);
  DEFINE_INTEGER(XG_PRE "GTK_ICON_SIZE_MENU" XG_POST,                           GTK_ICON_SIZE_MENU);
  DEFINE_INTEGER(XG_PRE "GTK_ICON_SIZE_SMALL_TOOLBAR" XG_POST,                  GTK_ICON_SIZE_SMALL_TOOLBAR);
  DEFINE_INTEGER(XG_PRE "GTK_ICON_SIZE_LARGE_TOOLBAR" XG_POST,                  GTK_ICON_SIZE_LARGE_TOOLBAR);
  DEFINE_INTEGER(XG_PRE "GTK_ICON_SIZE_BUTTON" XG_POST,                         GTK_ICON_SIZE_BUTTON);
  DEFINE_INTEGER(XG_PRE "GTK_ICON_SIZE_DND" XG_POST,                            GTK_ICON_SIZE_DND);
  DEFINE_INTEGER(XG_PRE "GTK_ICON_SIZE_DIALOG" XG_POST,                         GTK_ICON_SIZE_DIALOG);
  DEFINE_INTEGER(XG_PRE "GTK_TEXT_DIR_NONE" XG_POST,                            GTK_TEXT_DIR_NONE);
  DEFINE_INTEGER(XG_PRE "GTK_TEXT_DIR_LTR" XG_POST,                             GTK_TEXT_DIR_LTR);
  DEFINE_INTEGER(XG_PRE "GTK_TEXT_DIR_RTL" XG_POST,                             GTK_TEXT_DIR_RTL);
#endif
  DEFINE_INTEGER(XG_PRE "GTK_JUSTIFY_LEFT" XG_POST,                             GTK_JUSTIFY_LEFT);
  DEFINE_INTEGER(XG_PRE "GTK_JUSTIFY_RIGHT" XG_POST,                            GTK_JUSTIFY_RIGHT);
  DEFINE_INTEGER(XG_PRE "GTK_JUSTIFY_CENTER" XG_POST,                           GTK_JUSTIFY_CENTER);
  DEFINE_INTEGER(XG_PRE "GTK_JUSTIFY_FILL" XG_POST,                             GTK_JUSTIFY_FILL);
  DEFINE_INTEGER(XG_PRE "GTK_MENU_DIR_PARENT" XG_POST,                          GTK_MENU_DIR_PARENT);
  DEFINE_INTEGER(XG_PRE "GTK_MENU_DIR_CHILD" XG_POST,                           GTK_MENU_DIR_CHILD);
  DEFINE_INTEGER(XG_PRE "GTK_MENU_DIR_NEXT" XG_POST,                            GTK_MENU_DIR_NEXT);
  DEFINE_INTEGER(XG_PRE "GTK_MENU_DIR_PREV" XG_POST,                            GTK_MENU_DIR_PREV);
  DEFINE_INTEGER(XG_PRE "GTK_PIXELS" XG_POST,                                   GTK_PIXELS);
  DEFINE_INTEGER(XG_PRE "GTK_INCHES" XG_POST,                                   GTK_INCHES);
  DEFINE_INTEGER(XG_PRE "GTK_CENTIMETERS" XG_POST,                              GTK_CENTIMETERS);
#if (!HAVE_GTK_1)
  DEFINE_INTEGER(XG_PRE "GTK_MOVEMENT_LOGICAL_POSITIONS" XG_POST,               GTK_MOVEMENT_LOGICAL_POSITIONS);
  DEFINE_INTEGER(XG_PRE "GTK_MOVEMENT_VISUAL_POSITIONS" XG_POST,                GTK_MOVEMENT_VISUAL_POSITIONS);
  DEFINE_INTEGER(XG_PRE "GTK_MOVEMENT_WORDS" XG_POST,                           GTK_MOVEMENT_WORDS);
  DEFINE_INTEGER(XG_PRE "GTK_MOVEMENT_DISPLAY_LINES" XG_POST,                   GTK_MOVEMENT_DISPLAY_LINES);
  DEFINE_INTEGER(XG_PRE "GTK_MOVEMENT_DISPLAY_LINE_ENDS" XG_POST,               GTK_MOVEMENT_DISPLAY_LINE_ENDS);
  DEFINE_INTEGER(XG_PRE "GTK_MOVEMENT_PARAGRAPHS" XG_POST,                      GTK_MOVEMENT_PARAGRAPHS);
  DEFINE_INTEGER(XG_PRE "GTK_MOVEMENT_PARAGRAPH_ENDS" XG_POST,                  GTK_MOVEMENT_PARAGRAPH_ENDS);
  DEFINE_INTEGER(XG_PRE "GTK_MOVEMENT_PAGES" XG_POST,                           GTK_MOVEMENT_PAGES);
  DEFINE_INTEGER(XG_PRE "GTK_MOVEMENT_BUFFER_ENDS" XG_POST,                     GTK_MOVEMENT_BUFFER_ENDS);
#endif
  DEFINE_INTEGER(XG_PRE "GTK_ORIENTATION_HORIZONTAL" XG_POST,                   GTK_ORIENTATION_HORIZONTAL);
  DEFINE_INTEGER(XG_PRE "GTK_ORIENTATION_VERTICAL" XG_POST,                     GTK_ORIENTATION_VERTICAL);
  DEFINE_INTEGER(XG_PRE "GTK_CORNER_TOP_LEFT" XG_POST,                          GTK_CORNER_TOP_LEFT);
  DEFINE_INTEGER(XG_PRE "GTK_CORNER_BOTTOM_LEFT" XG_POST,                       GTK_CORNER_BOTTOM_LEFT);
  DEFINE_INTEGER(XG_PRE "GTK_CORNER_TOP_RIGHT" XG_POST,                         GTK_CORNER_TOP_RIGHT);
  DEFINE_INTEGER(XG_PRE "GTK_CORNER_BOTTOM_RIGHT" XG_POST,                      GTK_CORNER_BOTTOM_RIGHT);
  DEFINE_INTEGER(XG_PRE "GTK_PACK_START" XG_POST,                               GTK_PACK_START);
  DEFINE_INTEGER(XG_PRE "GTK_PACK_END" XG_POST,                                 GTK_PACK_END);
  DEFINE_INTEGER(XG_PRE "GTK_PATH_PRIO_LOWEST" XG_POST,                         GTK_PATH_PRIO_LOWEST);
  DEFINE_INTEGER(XG_PRE "GTK_PATH_PRIO_GTK" XG_POST,                            GTK_PATH_PRIO_GTK);
  DEFINE_INTEGER(XG_PRE "GTK_PATH_PRIO_APPLICATION" XG_POST,                    GTK_PATH_PRIO_APPLICATION);
#if (!HAVE_GTK_1)
  DEFINE_INTEGER(XG_PRE "GTK_PATH_PRIO_THEME" XG_POST,                          GTK_PATH_PRIO_THEME);
#endif
  DEFINE_INTEGER(XG_PRE "GTK_PATH_PRIO_RC" XG_POST,                             GTK_PATH_PRIO_RC);
  DEFINE_INTEGER(XG_PRE "GTK_PATH_PRIO_HIGHEST" XG_POST,                        GTK_PATH_PRIO_HIGHEST);
  DEFINE_INTEGER(XG_PRE "GTK_PATH_PRIO_MASK" XG_POST,                           GTK_PATH_PRIO_MASK);
  DEFINE_INTEGER(XG_PRE "GTK_PATH_WIDGET" XG_POST,                              GTK_PATH_WIDGET);
  DEFINE_INTEGER(XG_PRE "GTK_PATH_WIDGET_CLASS" XG_POST,                        GTK_PATH_WIDGET_CLASS);
  DEFINE_INTEGER(XG_PRE "GTK_PATH_CLASS" XG_POST,                               GTK_PATH_CLASS);
  DEFINE_INTEGER(XG_PRE "GTK_POLICY_ALWAYS" XG_POST,                            GTK_POLICY_ALWAYS);
  DEFINE_INTEGER(XG_PRE "GTK_POLICY_AUTOMATIC" XG_POST,                         GTK_POLICY_AUTOMATIC);
  DEFINE_INTEGER(XG_PRE "GTK_POLICY_NEVER" XG_POST,                             GTK_POLICY_NEVER);
  DEFINE_INTEGER(XG_PRE "GTK_POS_LEFT" XG_POST,                                 GTK_POS_LEFT);
  DEFINE_INTEGER(XG_PRE "GTK_POS_RIGHT" XG_POST,                                GTK_POS_RIGHT);
  DEFINE_INTEGER(XG_PRE "GTK_POS_TOP" XG_POST,                                  GTK_POS_TOP);
  DEFINE_INTEGER(XG_PRE "GTK_POS_BOTTOM" XG_POST,                               GTK_POS_BOTTOM);
  DEFINE_INTEGER(XG_PRE "GTK_RELIEF_NORMAL" XG_POST,                            GTK_RELIEF_NORMAL);
  DEFINE_INTEGER(XG_PRE "GTK_RELIEF_HALF" XG_POST,                              GTK_RELIEF_HALF);
  DEFINE_INTEGER(XG_PRE "GTK_RELIEF_NONE" XG_POST,                              GTK_RELIEF_NONE);
  DEFINE_INTEGER(XG_PRE "GTK_RESIZE_PARENT" XG_POST,                            GTK_RESIZE_PARENT);
  DEFINE_INTEGER(XG_PRE "GTK_RESIZE_QUEUE" XG_POST,                             GTK_RESIZE_QUEUE);
  DEFINE_INTEGER(XG_PRE "GTK_RESIZE_IMMEDIATE" XG_POST,                         GTK_RESIZE_IMMEDIATE);
  DEFINE_INTEGER(XG_PRE "GTK_SCROLL_NONE" XG_POST,                              GTK_SCROLL_NONE);
  DEFINE_INTEGER(XG_PRE "GTK_SCROLL_JUMP" XG_POST,                              GTK_SCROLL_JUMP);
  DEFINE_INTEGER(XG_PRE "GTK_SCROLL_STEP_BACKWARD" XG_POST,                     GTK_SCROLL_STEP_BACKWARD);
  DEFINE_INTEGER(XG_PRE "GTK_SCROLL_STEP_FORWARD" XG_POST,                      GTK_SCROLL_STEP_FORWARD);
  DEFINE_INTEGER(XG_PRE "GTK_SCROLL_PAGE_BACKWARD" XG_POST,                     GTK_SCROLL_PAGE_BACKWARD);
  DEFINE_INTEGER(XG_PRE "GTK_SCROLL_PAGE_FORWARD" XG_POST,                      GTK_SCROLL_PAGE_FORWARD);
#if (!HAVE_GTK_1)
  DEFINE_INTEGER(XG_PRE "GTK_SCROLL_STEP_UP" XG_POST,                           GTK_SCROLL_STEP_UP);
  DEFINE_INTEGER(XG_PRE "GTK_SCROLL_STEP_DOWN" XG_POST,                         GTK_SCROLL_STEP_DOWN);
  DEFINE_INTEGER(XG_PRE "GTK_SCROLL_PAGE_UP" XG_POST,                           GTK_SCROLL_PAGE_UP);
  DEFINE_INTEGER(XG_PRE "GTK_SCROLL_PAGE_DOWN" XG_POST,                         GTK_SCROLL_PAGE_DOWN);
  DEFINE_INTEGER(XG_PRE "GTK_SCROLL_STEP_LEFT" XG_POST,                         GTK_SCROLL_STEP_LEFT);
  DEFINE_INTEGER(XG_PRE "GTK_SCROLL_STEP_RIGHT" XG_POST,                        GTK_SCROLL_STEP_RIGHT);
  DEFINE_INTEGER(XG_PRE "GTK_SCROLL_PAGE_LEFT" XG_POST,                         GTK_SCROLL_PAGE_LEFT);
  DEFINE_INTEGER(XG_PRE "GTK_SCROLL_PAGE_RIGHT" XG_POST,                        GTK_SCROLL_PAGE_RIGHT);
  DEFINE_INTEGER(XG_PRE "GTK_SCROLL_START" XG_POST,                             GTK_SCROLL_START);
  DEFINE_INTEGER(XG_PRE "GTK_SCROLL_END" XG_POST,                               GTK_SCROLL_END);
  DEFINE_INTEGER(XG_PRE "GTK_SELECTION_NONE" XG_POST,                           GTK_SELECTION_NONE);
#endif
  DEFINE_INTEGER(XG_PRE "GTK_SELECTION_SINGLE" XG_POST,                         GTK_SELECTION_SINGLE);
  DEFINE_INTEGER(XG_PRE "GTK_SELECTION_BROWSE" XG_POST,                         GTK_SELECTION_BROWSE);
  DEFINE_INTEGER(XG_PRE "GTK_SELECTION_MULTIPLE" XG_POST,                       GTK_SELECTION_MULTIPLE);
  DEFINE_INTEGER(XG_PRE "GTK_SELECTION_EXTENDED" XG_POST,                       GTK_SELECTION_EXTENDED);
  DEFINE_INTEGER(XG_PRE "GTK_SHADOW_NONE" XG_POST,                              GTK_SHADOW_NONE);
  DEFINE_INTEGER(XG_PRE "GTK_SHADOW_IN" XG_POST,                                GTK_SHADOW_IN);
  DEFINE_INTEGER(XG_PRE "GTK_SHADOW_OUT" XG_POST,                               GTK_SHADOW_OUT);
  DEFINE_INTEGER(XG_PRE "GTK_SHADOW_ETCHED_IN" XG_POST,                         GTK_SHADOW_ETCHED_IN);
  DEFINE_INTEGER(XG_PRE "GTK_SHADOW_ETCHED_OUT" XG_POST,                        GTK_SHADOW_ETCHED_OUT);
  DEFINE_INTEGER(XG_PRE "GTK_STATE_NORMAL" XG_POST,                             GTK_STATE_NORMAL);
  DEFINE_INTEGER(XG_PRE "GTK_STATE_ACTIVE" XG_POST,                             GTK_STATE_ACTIVE);
  DEFINE_INTEGER(XG_PRE "GTK_STATE_PRELIGHT" XG_POST,                           GTK_STATE_PRELIGHT);
  DEFINE_INTEGER(XG_PRE "GTK_STATE_SELECTED" XG_POST,                           GTK_STATE_SELECTED);
  DEFINE_INTEGER(XG_PRE "GTK_STATE_INSENSITIVE" XG_POST,                        GTK_STATE_INSENSITIVE);
  DEFINE_INTEGER(XG_PRE "GTK_TOOLBAR_ICONS" XG_POST,                            GTK_TOOLBAR_ICONS);
  DEFINE_INTEGER(XG_PRE "GTK_TOOLBAR_TEXT" XG_POST,                             GTK_TOOLBAR_TEXT);
  DEFINE_INTEGER(XG_PRE "GTK_TOOLBAR_BOTH" XG_POST,                             GTK_TOOLBAR_BOTH);
#if (!HAVE_GTK_1)
  DEFINE_INTEGER(XG_PRE "GTK_TOOLBAR_BOTH_HORIZ" XG_POST,                       GTK_TOOLBAR_BOTH_HORIZ);
#endif
  DEFINE_INTEGER(XG_PRE "GTK_UPDATE_CONTINUOUS" XG_POST,                        GTK_UPDATE_CONTINUOUS);
  DEFINE_INTEGER(XG_PRE "GTK_UPDATE_DISCONTINUOUS" XG_POST,                     GTK_UPDATE_DISCONTINUOUS);
  DEFINE_INTEGER(XG_PRE "GTK_UPDATE_DELAYED" XG_POST,                           GTK_UPDATE_DELAYED);
  DEFINE_INTEGER(XG_PRE "GTK_VISIBILITY_NONE" XG_POST,                          GTK_VISIBILITY_NONE);
  DEFINE_INTEGER(XG_PRE "GTK_VISIBILITY_PARTIAL" XG_POST,                       GTK_VISIBILITY_PARTIAL);
  DEFINE_INTEGER(XG_PRE "GTK_VISIBILITY_FULL" XG_POST,                          GTK_VISIBILITY_FULL);
  DEFINE_INTEGER(XG_PRE "GTK_WIN_POS_NONE" XG_POST,                             GTK_WIN_POS_NONE);
  DEFINE_INTEGER(XG_PRE "GTK_WIN_POS_CENTER" XG_POST,                           GTK_WIN_POS_CENTER);
  DEFINE_INTEGER(XG_PRE "GTK_WIN_POS_MOUSE" XG_POST,                            GTK_WIN_POS_MOUSE);
  DEFINE_INTEGER(XG_PRE "GTK_WIN_POS_CENTER_ALWAYS" XG_POST,                    GTK_WIN_POS_CENTER_ALWAYS);
#if (!HAVE_GTK_1)
  DEFINE_INTEGER(XG_PRE "GTK_WIN_POS_CENTER_ON_PARENT" XG_POST,                 GTK_WIN_POS_CENTER_ON_PARENT);
#endif
  DEFINE_INTEGER(XG_PRE "GTK_WINDOW_TOPLEVEL" XG_POST,                          GTK_WINDOW_TOPLEVEL);
  DEFINE_INTEGER(XG_PRE "GTK_WINDOW_POPUP" XG_POST,                             GTK_WINDOW_POPUP);
#if (!HAVE_GTK_1)
  DEFINE_INTEGER(XG_PRE "GTK_WRAP_NONE" XG_POST,                                GTK_WRAP_NONE);
  DEFINE_INTEGER(XG_PRE "GTK_WRAP_CHAR" XG_POST,                                GTK_WRAP_CHAR);
  DEFINE_INTEGER(XG_PRE "GTK_WRAP_WORD" XG_POST,                                GTK_WRAP_WORD);
#endif
  DEFINE_INTEGER(XG_PRE "GTK_SORT_ASCENDING" XG_POST,                           GTK_SORT_ASCENDING);
  DEFINE_INTEGER(XG_PRE "GTK_SORT_DESCENDING" XG_POST,                          GTK_SORT_DESCENDING);
#if (!HAVE_GTK_1)
  DEFINE_INTEGER(XG_PRE "GTK_IMAGE_EMPTY" XG_POST,                              GTK_IMAGE_EMPTY);
  DEFINE_INTEGER(XG_PRE "GTK_IMAGE_PIXMAP" XG_POST,                             GTK_IMAGE_PIXMAP);
  DEFINE_INTEGER(XG_PRE "GTK_IMAGE_IMAGE" XG_POST,                              GTK_IMAGE_IMAGE);
  DEFINE_INTEGER(XG_PRE "GTK_IMAGE_PIXBUF" XG_POST,                             GTK_IMAGE_PIXBUF);
  DEFINE_INTEGER(XG_PRE "GTK_IMAGE_STOCK" XG_POST,                              GTK_IMAGE_STOCK);
  DEFINE_INTEGER(XG_PRE "GTK_IMAGE_ICON_SET" XG_POST,                           GTK_IMAGE_ICON_SET);
  DEFINE_INTEGER(XG_PRE "GTK_IMAGE_ANIMATION" XG_POST,                          GTK_IMAGE_ANIMATION);
  DEFINE_INTEGER(XG_PRE "GTK_MAX_COMPOSE_LEN" XG_POST,                          GTK_MAX_COMPOSE_LEN);
#endif
  DEFINE_INTEGER(XG_PRE "GTK_PRIORITY_RESIZE" XG_POST,                          GTK_PRIORITY_RESIZE);
#if (!HAVE_GTK_1)
  DEFINE_INTEGER(XG_PRE "GTK_MESSAGE_INFO" XG_POST,                             GTK_MESSAGE_INFO);
  DEFINE_INTEGER(XG_PRE "GTK_MESSAGE_WARNING" XG_POST,                          GTK_MESSAGE_WARNING);
  DEFINE_INTEGER(XG_PRE "GTK_MESSAGE_QUESTION" XG_POST,                         GTK_MESSAGE_QUESTION);
  DEFINE_INTEGER(XG_PRE "GTK_MESSAGE_ERROR" XG_POST,                            GTK_MESSAGE_ERROR);
  DEFINE_INTEGER(XG_PRE "GTK_BUTTONS_NONE" XG_POST,                             GTK_BUTTONS_NONE);
  DEFINE_INTEGER(XG_PRE "GTK_BUTTONS_OK" XG_POST,                               GTK_BUTTONS_OK);
  DEFINE_INTEGER(XG_PRE "GTK_BUTTONS_CLOSE" XG_POST,                            GTK_BUTTONS_CLOSE);
  DEFINE_INTEGER(XG_PRE "GTK_BUTTONS_CANCEL" XG_POST,                           GTK_BUTTONS_CANCEL);
  DEFINE_INTEGER(XG_PRE "GTK_BUTTONS_YES_NO" XG_POST,                           GTK_BUTTONS_YES_NO);
  DEFINE_INTEGER(XG_PRE "GTK_BUTTONS_OK_CANCEL" XG_POST,                        GTK_BUTTONS_OK_CANCEL);
  DEFINE_INTEGER(XG_PRE "GTK_NOTEBOOK_TAB_FIRST" XG_POST,                       GTK_NOTEBOOK_TAB_FIRST);
  DEFINE_INTEGER(XG_PRE "GTK_NOTEBOOK_TAB_LAST" XG_POST,                        GTK_NOTEBOOK_TAB_LAST);
  DEFINE_INTEGER(XG_PRE "GTK_IN_DESTRUCTION" XG_POST,                           GTK_IN_DESTRUCTION);
#endif
  DEFINE_INTEGER(XG_PRE "GTK_FLOATING" XG_POST,                                 GTK_FLOATING);
#if (!HAVE_GTK_1)
  DEFINE_INTEGER(XG_PRE "GTK_RESERVED_1" XG_POST,                               GTK_RESERVED_1);
  DEFINE_INTEGER(XG_PRE "GTK_RESERVED_2" XG_POST,                               GTK_RESERVED_2);
#endif
  DEFINE_INTEGER(XG_PRE "GTK_PROGRESS_CONTINUOUS" XG_POST,                      GTK_PROGRESS_CONTINUOUS);
  DEFINE_INTEGER(XG_PRE "GTK_PROGRESS_DISCRETE" XG_POST,                        GTK_PROGRESS_DISCRETE);
  DEFINE_INTEGER(XG_PRE "GTK_PROGRESS_LEFT_TO_RIGHT" XG_POST,                   GTK_PROGRESS_LEFT_TO_RIGHT);
  DEFINE_INTEGER(XG_PRE "GTK_PROGRESS_RIGHT_TO_LEFT" XG_POST,                   GTK_PROGRESS_RIGHT_TO_LEFT);
  DEFINE_INTEGER(XG_PRE "GTK_PROGRESS_BOTTOM_TO_TOP" XG_POST,                   GTK_PROGRESS_BOTTOM_TO_TOP);
  DEFINE_INTEGER(XG_PRE "GTK_PROGRESS_TOP_TO_BOTTOM" XG_POST,                   GTK_PROGRESS_TOP_TO_BOTTOM);
  DEFINE_INTEGER(XG_PRE "GTK_RC_FG" XG_POST,                                    GTK_RC_FG);
  DEFINE_INTEGER(XG_PRE "GTK_RC_BG" XG_POST,                                    GTK_RC_BG);
  DEFINE_INTEGER(XG_PRE "GTK_RC_TEXT" XG_POST,                                  GTK_RC_TEXT);
  DEFINE_INTEGER(XG_PRE "GTK_RC_BASE" XG_POST,                                  GTK_RC_BASE);
  DEFINE_INTEGER(XG_PRE "GTK_RC_TOKEN_INVALID" XG_POST,                         GTK_RC_TOKEN_INVALID);
  DEFINE_INTEGER(XG_PRE "GTK_RC_TOKEN_INCLUDE" XG_POST,                         GTK_RC_TOKEN_INCLUDE);
  DEFINE_INTEGER(XG_PRE "GTK_RC_TOKEN_NORMAL" XG_POST,                          GTK_RC_TOKEN_NORMAL);
  DEFINE_INTEGER(XG_PRE "GTK_RC_TOKEN_ACTIVE" XG_POST,                          GTK_RC_TOKEN_ACTIVE);
  DEFINE_INTEGER(XG_PRE "GTK_RC_TOKEN_PRELIGHT" XG_POST,                        GTK_RC_TOKEN_PRELIGHT);
  DEFINE_INTEGER(XG_PRE "GTK_RC_TOKEN_SELECTED" XG_POST,                        GTK_RC_TOKEN_SELECTED);
  DEFINE_INTEGER(XG_PRE "GTK_RC_TOKEN_INSENSITIVE" XG_POST,                     GTK_RC_TOKEN_INSENSITIVE);
  DEFINE_INTEGER(XG_PRE "GTK_RC_TOKEN_FG" XG_POST,                              GTK_RC_TOKEN_FG);
  DEFINE_INTEGER(XG_PRE "GTK_RC_TOKEN_BG" XG_POST,                              GTK_RC_TOKEN_BG);
  DEFINE_INTEGER(XG_PRE "GTK_RC_TOKEN_TEXT" XG_POST,                            GTK_RC_TOKEN_TEXT);
  DEFINE_INTEGER(XG_PRE "GTK_RC_TOKEN_BASE" XG_POST,                            GTK_RC_TOKEN_BASE);
#if (!HAVE_GTK_1)
  DEFINE_INTEGER(XG_PRE "GTK_RC_TOKEN_XTHICKNESS" XG_POST,                      GTK_RC_TOKEN_XTHICKNESS);
  DEFINE_INTEGER(XG_PRE "GTK_RC_TOKEN_YTHICKNESS" XG_POST,                      GTK_RC_TOKEN_YTHICKNESS);
#endif
  DEFINE_INTEGER(XG_PRE "GTK_RC_TOKEN_FONT" XG_POST,                            GTK_RC_TOKEN_FONT);
  DEFINE_INTEGER(XG_PRE "GTK_RC_TOKEN_FONTSET" XG_POST,                         GTK_RC_TOKEN_FONTSET);
#if (!HAVE_GTK_1)
  DEFINE_INTEGER(XG_PRE "GTK_RC_TOKEN_FONT_NAME" XG_POST,                       GTK_RC_TOKEN_FONT_NAME);
#endif
  DEFINE_INTEGER(XG_PRE "GTK_RC_TOKEN_BG_PIXMAP" XG_POST,                       GTK_RC_TOKEN_BG_PIXMAP);
  DEFINE_INTEGER(XG_PRE "GTK_RC_TOKEN_PIXMAP_PATH" XG_POST,                     GTK_RC_TOKEN_PIXMAP_PATH);
  DEFINE_INTEGER(XG_PRE "GTK_RC_TOKEN_STYLE" XG_POST,                           GTK_RC_TOKEN_STYLE);
  DEFINE_INTEGER(XG_PRE "GTK_RC_TOKEN_BINDING" XG_POST,                         GTK_RC_TOKEN_BINDING);
  DEFINE_INTEGER(XG_PRE "GTK_RC_TOKEN_BIND" XG_POST,                            GTK_RC_TOKEN_BIND);
  DEFINE_INTEGER(XG_PRE "GTK_RC_TOKEN_WIDGET" XG_POST,                          GTK_RC_TOKEN_WIDGET);
  DEFINE_INTEGER(XG_PRE "GTK_RC_TOKEN_WIDGET_CLASS" XG_POST,                    GTK_RC_TOKEN_WIDGET_CLASS);
  DEFINE_INTEGER(XG_PRE "GTK_RC_TOKEN_CLASS" XG_POST,                           GTK_RC_TOKEN_CLASS);
  DEFINE_INTEGER(XG_PRE "GTK_RC_TOKEN_LOWEST" XG_POST,                          GTK_RC_TOKEN_LOWEST);
  DEFINE_INTEGER(XG_PRE "GTK_RC_TOKEN_GTK" XG_POST,                             GTK_RC_TOKEN_GTK);
  DEFINE_INTEGER(XG_PRE "GTK_RC_TOKEN_APPLICATION" XG_POST,                     GTK_RC_TOKEN_APPLICATION);
#if (!HAVE_GTK_1)
  DEFINE_INTEGER(XG_PRE "GTK_RC_TOKEN_THEME" XG_POST,                           GTK_RC_TOKEN_THEME);
#endif
  DEFINE_INTEGER(XG_PRE "GTK_RC_TOKEN_RC" XG_POST,                              GTK_RC_TOKEN_RC);
  DEFINE_INTEGER(XG_PRE "GTK_RC_TOKEN_HIGHEST" XG_POST,                         GTK_RC_TOKEN_HIGHEST);
  DEFINE_INTEGER(XG_PRE "GTK_RC_TOKEN_ENGINE" XG_POST,                          GTK_RC_TOKEN_ENGINE);
  DEFINE_INTEGER(XG_PRE "GTK_RC_TOKEN_MODULE_PATH" XG_POST,                     GTK_RC_TOKEN_MODULE_PATH);
#if (!HAVE_GTK_1)
  DEFINE_INTEGER(XG_PRE "GTK_RC_TOKEN_IM_MODULE_PATH" XG_POST,                  GTK_RC_TOKEN_IM_MODULE_PATH);
  DEFINE_INTEGER(XG_PRE "GTK_RC_TOKEN_IM_MODULE_FILE" XG_POST,                  GTK_RC_TOKEN_IM_MODULE_FILE);
  DEFINE_INTEGER(XG_PRE "GTK_RC_TOKEN_STOCK" XG_POST,                           GTK_RC_TOKEN_STOCK);
  DEFINE_INTEGER(XG_PRE "GTK_RC_TOKEN_LTR" XG_POST,                             GTK_RC_TOKEN_LTR);
  DEFINE_INTEGER(XG_PRE "GTK_RC_TOKEN_RTL" XG_POST,                             GTK_RC_TOKEN_RTL);
#endif
  DEFINE_INTEGER(XG_PRE "GTK_RC_TOKEN_LAST" XG_POST,                            GTK_RC_TOKEN_LAST);
#if (!HAVE_GTK_1)
  DEFINE_INTEGER(XG_PRE "GTK_SIZE_GROUP_NONE" XG_POST,                          GTK_SIZE_GROUP_NONE);
  DEFINE_INTEGER(XG_PRE "GTK_SIZE_GROUP_HORIZONTAL" XG_POST,                    GTK_SIZE_GROUP_HORIZONTAL);
  DEFINE_INTEGER(XG_PRE "GTK_SIZE_GROUP_VERTICAL" XG_POST,                      GTK_SIZE_GROUP_VERTICAL);
  DEFINE_INTEGER(XG_PRE "GTK_SIZE_GROUP_BOTH" XG_POST,                          GTK_SIZE_GROUP_BOTH);
  DEFINE_INTEGER(XG_PRE "GTK_INPUT_ERROR" XG_POST,                              GTK_INPUT_ERROR);
#endif
  DEFINE_INTEGER(XG_PRE "GTK_UPDATE_ALWAYS" XG_POST,                            GTK_UPDATE_ALWAYS);
  DEFINE_INTEGER(XG_PRE "GTK_UPDATE_IF_VALID" XG_POST,                          GTK_UPDATE_IF_VALID);
  DEFINE_INTEGER(XG_PRE "GTK_SPIN_STEP_FORWARD" XG_POST,                        GTK_SPIN_STEP_FORWARD);
  DEFINE_INTEGER(XG_PRE "GTK_SPIN_STEP_BACKWARD" XG_POST,                       GTK_SPIN_STEP_BACKWARD);
  DEFINE_INTEGER(XG_PRE "GTK_SPIN_PAGE_FORWARD" XG_POST,                        GTK_SPIN_PAGE_FORWARD);
  DEFINE_INTEGER(XG_PRE "GTK_SPIN_PAGE_BACKWARD" XG_POST,                       GTK_SPIN_PAGE_BACKWARD);
  DEFINE_INTEGER(XG_PRE "GTK_SPIN_HOME" XG_POST,                                GTK_SPIN_HOME);
  DEFINE_INTEGER(XG_PRE "GTK_SPIN_END" XG_POST,                                 GTK_SPIN_END);
  DEFINE_INTEGER(XG_PRE "GTK_SPIN_USER_DEFINED" XG_POST,                        GTK_SPIN_USER_DEFINED);
#if (!HAVE_GTK_1)
  DEFINE_INTEGER(XG_PRE "GTK_TEXT_WINDOW_PRIVATE" XG_POST,                      GTK_TEXT_WINDOW_PRIVATE);
  DEFINE_INTEGER(XG_PRE "GTK_TEXT_WINDOW_WIDGET" XG_POST,                       GTK_TEXT_WINDOW_WIDGET);
  DEFINE_INTEGER(XG_PRE "GTK_TEXT_WINDOW_TEXT" XG_POST,                         GTK_TEXT_WINDOW_TEXT);
  DEFINE_INTEGER(XG_PRE "GTK_TEXT_WINDOW_LEFT" XG_POST,                         GTK_TEXT_WINDOW_LEFT);
  DEFINE_INTEGER(XG_PRE "GTK_TEXT_WINDOW_RIGHT" XG_POST,                        GTK_TEXT_WINDOW_RIGHT);
  DEFINE_INTEGER(XG_PRE "GTK_TEXT_WINDOW_TOP" XG_POST,                          GTK_TEXT_WINDOW_TOP);
  DEFINE_INTEGER(XG_PRE "GTK_TEXT_WINDOW_BOTTOM" XG_POST,                       GTK_TEXT_WINDOW_BOTTOM);
  DEFINE_INTEGER(XG_PRE "GTK_TEXT_VIEW_PRIORITY_VALIDATE" XG_POST,              GTK_TEXT_VIEW_PRIORITY_VALIDATE);
#endif
  DEFINE_INTEGER(XG_PRE "GTK_TOOLBAR_CHILD_SPACE" XG_POST,                      GTK_TOOLBAR_CHILD_SPACE);
  DEFINE_INTEGER(XG_PRE "GTK_TOOLBAR_CHILD_BUTTON" XG_POST,                     GTK_TOOLBAR_CHILD_BUTTON);
  DEFINE_INTEGER(XG_PRE "GTK_TOOLBAR_CHILD_TOGGLEBUTTON" XG_POST,               GTK_TOOLBAR_CHILD_TOGGLEBUTTON);
  DEFINE_INTEGER(XG_PRE "GTK_TOOLBAR_CHILD_RADIOBUTTON" XG_POST,                GTK_TOOLBAR_CHILD_RADIOBUTTON);
  DEFINE_INTEGER(XG_PRE "GTK_TOOLBAR_CHILD_WIDGET" XG_POST,                     GTK_TOOLBAR_CHILD_WIDGET);
  DEFINE_INTEGER(XG_PRE "GTK_TOOLBAR_SPACE_EMPTY" XG_POST,                      GTK_TOOLBAR_SPACE_EMPTY);
  DEFINE_INTEGER(XG_PRE "GTK_TOOLBAR_SPACE_LINE" XG_POST,                       GTK_TOOLBAR_SPACE_LINE);
#if (!HAVE_GTK_1)
  DEFINE_INTEGER(XG_PRE "GTK_TREE_MODEL_ITERS_PERSIST" XG_POST,                 GTK_TREE_MODEL_ITERS_PERSIST);
  DEFINE_INTEGER(XG_PRE "GTK_TREE_MODEL_LIST_ONLY" XG_POST,                     GTK_TREE_MODEL_LIST_ONLY);
  DEFINE_INTEGER(XG_PRE "GTK_TREE_VIEW_COLUMN_GROW_ONLY" XG_POST,               GTK_TREE_VIEW_COLUMN_GROW_ONLY);
  DEFINE_INTEGER(XG_PRE "GTK_TREE_VIEW_COLUMN_AUTOSIZE" XG_POST,                GTK_TREE_VIEW_COLUMN_AUTOSIZE);
  DEFINE_INTEGER(XG_PRE "GTK_TREE_VIEW_COLUMN_FIXED" XG_POST,                   GTK_TREE_VIEW_COLUMN_FIXED);
  DEFINE_INTEGER(XG_PRE "GTK_TREE_VIEW_DROP_BEFORE" XG_POST,                    GTK_TREE_VIEW_DROP_BEFORE);
  DEFINE_INTEGER(XG_PRE "GTK_TREE_VIEW_DROP_AFTER" XG_POST,                     GTK_TREE_VIEW_DROP_AFTER);
  DEFINE_INTEGER(XG_PRE "GTK_TREE_VIEW_DROP_INTO_OR_BEFORE" XG_POST,            GTK_TREE_VIEW_DROP_INTO_OR_BEFORE);
  DEFINE_INTEGER(XG_PRE "GTK_TREE_VIEW_DROP_INTO_OR_AFTER" XG_POST,             GTK_TREE_VIEW_DROP_INTO_OR_AFTER);
#endif
  DEFINE_INTEGER(XG_PRE "GTK_MAJOR_VERSION" XG_POST,                            GTK_MAJOR_VERSION);
  DEFINE_INTEGER(XG_PRE "GTK_MINOR_VERSION" XG_POST,                            GTK_MINOR_VERSION);
  DEFINE_INTEGER(XG_PRE "GTK_MICRO_VERSION" XG_POST,                            GTK_MICRO_VERSION);
  DEFINE_INTEGER(XG_PRE "GTK_BINARY_AGE" XG_POST,                               GTK_BINARY_AGE);
  DEFINE_INTEGER(XG_PRE "GTK_INTERFACE_AGE" XG_POST,                            GTK_INTERFACE_AGE);
  DEFINE_INTEGER(XG_PRE "GTK_TOPLEVEL" XG_POST,                                 GTK_TOPLEVEL);
  DEFINE_INTEGER(XG_PRE "GTK_NO_WINDOW" XG_POST,                                GTK_NO_WINDOW);
  DEFINE_INTEGER(XG_PRE "GTK_REALIZED" XG_POST,                                 GTK_REALIZED);
  DEFINE_INTEGER(XG_PRE "GTK_MAPPED" XG_POST,                                   GTK_MAPPED);
  DEFINE_INTEGER(XG_PRE "GTK_VISIBLE" XG_POST,                                  GTK_VISIBLE);
  DEFINE_INTEGER(XG_PRE "GTK_SENSITIVE" XG_POST,                                GTK_SENSITIVE);
  DEFINE_INTEGER(XG_PRE "GTK_PARENT_SENSITIVE" XG_POST,                         GTK_PARENT_SENSITIVE);
  DEFINE_INTEGER(XG_PRE "GTK_CAN_FOCUS" XG_POST,                                GTK_CAN_FOCUS);
  DEFINE_INTEGER(XG_PRE "GTK_HAS_FOCUS" XG_POST,                                GTK_HAS_FOCUS);
  DEFINE_INTEGER(XG_PRE "GTK_CAN_DEFAULT" XG_POST,                              GTK_CAN_DEFAULT);
  DEFINE_INTEGER(XG_PRE "GTK_HAS_DEFAULT" XG_POST,                              GTK_HAS_DEFAULT);
  DEFINE_INTEGER(XG_PRE "GTK_HAS_GRAB" XG_POST,                                 GTK_HAS_GRAB);
  DEFINE_INTEGER(XG_PRE "GTK_RC_STYLE" XG_POST,                                 GTK_RC_STYLE);
  DEFINE_INTEGER(XG_PRE "GTK_COMPOSITE_CHILD" XG_POST,                          GTK_COMPOSITE_CHILD);
  DEFINE_INTEGER(XG_PRE "GTK_NO_REPARENT" XG_POST,                              GTK_NO_REPARENT);
  DEFINE_INTEGER(XG_PRE "GTK_APP_PAINTABLE" XG_POST,                            GTK_APP_PAINTABLE);
  DEFINE_INTEGER(XG_PRE "GTK_RECEIVES_DEFAULT" XG_POST,                         GTK_RECEIVES_DEFAULT);
#if (!HAVE_GTK_1)
  DEFINE_INTEGER(XG_PRE "GTK_DOUBLE_BUFFERED" XG_POST,                          GTK_DOUBLE_BUFFERED);
  DEFINE_INTEGER(XG_PRE "GTK_WIDGET_HELP_TOOLTIP" XG_POST,                      GTK_WIDGET_HELP_TOOLTIP);
  DEFINE_INTEGER(XG_PRE "GTK_WIDGET_HELP_WHATS_THIS" XG_POST,                   GTK_WIDGET_HELP_WHATS_THIS);
  DEFINE_INTEGER(XG_PRE "PANGO_ATTR_INVALID" XG_POST,                           PANGO_ATTR_INVALID);
  DEFINE_INTEGER(XG_PRE "PANGO_ATTR_LANGUAGE" XG_POST,                          PANGO_ATTR_LANGUAGE);
  DEFINE_INTEGER(XG_PRE "PANGO_ATTR_FAMILY" XG_POST,                            PANGO_ATTR_FAMILY);
  DEFINE_INTEGER(XG_PRE "PANGO_ATTR_STYLE" XG_POST,                             PANGO_ATTR_STYLE);
  DEFINE_INTEGER(XG_PRE "PANGO_ATTR_WEIGHT" XG_POST,                            PANGO_ATTR_WEIGHT);
  DEFINE_INTEGER(XG_PRE "PANGO_ATTR_VARIANT" XG_POST,                           PANGO_ATTR_VARIANT);
  DEFINE_INTEGER(XG_PRE "PANGO_ATTR_STRETCH" XG_POST,                           PANGO_ATTR_STRETCH);
  DEFINE_INTEGER(XG_PRE "PANGO_ATTR_SIZE" XG_POST,                              PANGO_ATTR_SIZE);
  DEFINE_INTEGER(XG_PRE "PANGO_ATTR_FONT_DESC" XG_POST,                         PANGO_ATTR_FONT_DESC);
  DEFINE_INTEGER(XG_PRE "PANGO_ATTR_FOREGROUND" XG_POST,                        PANGO_ATTR_FOREGROUND);
  DEFINE_INTEGER(XG_PRE "PANGO_ATTR_BACKGROUND" XG_POST,                        PANGO_ATTR_BACKGROUND);
  DEFINE_INTEGER(XG_PRE "PANGO_ATTR_UNDERLINE" XG_POST,                         PANGO_ATTR_UNDERLINE);
  DEFINE_INTEGER(XG_PRE "PANGO_ATTR_STRIKETHROUGH" XG_POST,                     PANGO_ATTR_STRIKETHROUGH);
  DEFINE_INTEGER(XG_PRE "PANGO_ATTR_RISE" XG_POST,                              PANGO_ATTR_RISE);
  DEFINE_INTEGER(XG_PRE "PANGO_ATTR_SHAPE" XG_POST,                             PANGO_ATTR_SHAPE);
  DEFINE_INTEGER(XG_PRE "PANGO_ATTR_SCALE" XG_POST,                             PANGO_ATTR_SCALE);
  DEFINE_INTEGER(XG_PRE "PANGO_UNDERLINE_NONE" XG_POST,                         PANGO_UNDERLINE_NONE);
  DEFINE_INTEGER(XG_PRE "PANGO_UNDERLINE_SINGLE" XG_POST,                       PANGO_UNDERLINE_SINGLE);
  DEFINE_INTEGER(XG_PRE "PANGO_UNDERLINE_DOUBLE" XG_POST,                       PANGO_UNDERLINE_DOUBLE);
  DEFINE_INTEGER(XG_PRE "PANGO_UNDERLINE_LOW" XG_POST,                          PANGO_UNDERLINE_LOW);
  DEFINE_INTEGER(XG_PRE "PANGO_COVERAGE_NONE" XG_POST,                          PANGO_COVERAGE_NONE);
  DEFINE_INTEGER(XG_PRE "PANGO_COVERAGE_FALLBACK" XG_POST,                      PANGO_COVERAGE_FALLBACK);
  DEFINE_INTEGER(XG_PRE "PANGO_COVERAGE_APPROXIMATE" XG_POST,                   PANGO_COVERAGE_APPROXIMATE);
  DEFINE_INTEGER(XG_PRE "PANGO_COVERAGE_EXACT" XG_POST,                         PANGO_COVERAGE_EXACT);
  DEFINE_INTEGER(XG_PRE "PANGO_STYLE_NORMAL" XG_POST,                           PANGO_STYLE_NORMAL);
  DEFINE_INTEGER(XG_PRE "PANGO_STYLE_OBLIQUE" XG_POST,                          PANGO_STYLE_OBLIQUE);
  DEFINE_INTEGER(XG_PRE "PANGO_STYLE_ITALIC" XG_POST,                           PANGO_STYLE_ITALIC);
  DEFINE_INTEGER(XG_PRE "PANGO_VARIANT_NORMAL" XG_POST,                         PANGO_VARIANT_NORMAL);
  DEFINE_INTEGER(XG_PRE "PANGO_VARIANT_SMALL_CAPS" XG_POST,                     PANGO_VARIANT_SMALL_CAPS);
  DEFINE_INTEGER(XG_PRE "PANGO_WEIGHT_ULTRALIGHT" XG_POST,                      PANGO_WEIGHT_ULTRALIGHT);
  DEFINE_INTEGER(XG_PRE "PANGO_WEIGHT_LIGHT" XG_POST,                           PANGO_WEIGHT_LIGHT);
  DEFINE_INTEGER(XG_PRE "PANGO_WEIGHT_NORMAL" XG_POST,                          PANGO_WEIGHT_NORMAL);
  DEFINE_INTEGER(XG_PRE "PANGO_WEIGHT_BOLD" XG_POST,                            PANGO_WEIGHT_BOLD);
  DEFINE_INTEGER(XG_PRE "PANGO_WEIGHT_ULTRABOLD" XG_POST,                       PANGO_WEIGHT_ULTRABOLD);
  DEFINE_INTEGER(XG_PRE "PANGO_WEIGHT_HEAVY" XG_POST,                           PANGO_WEIGHT_HEAVY);
  DEFINE_INTEGER(XG_PRE "PANGO_STRETCH_ULTRA_CONDENSED" XG_POST,                PANGO_STRETCH_ULTRA_CONDENSED);
  DEFINE_INTEGER(XG_PRE "PANGO_STRETCH_EXTRA_CONDENSED" XG_POST,                PANGO_STRETCH_EXTRA_CONDENSED);
  DEFINE_INTEGER(XG_PRE "PANGO_STRETCH_CONDENSED" XG_POST,                      PANGO_STRETCH_CONDENSED);
  DEFINE_INTEGER(XG_PRE "PANGO_STRETCH_SEMI_CONDENSED" XG_POST,                 PANGO_STRETCH_SEMI_CONDENSED);
  DEFINE_INTEGER(XG_PRE "PANGO_STRETCH_NORMAL" XG_POST,                         PANGO_STRETCH_NORMAL);
  DEFINE_INTEGER(XG_PRE "PANGO_STRETCH_SEMI_EXPANDED" XG_POST,                  PANGO_STRETCH_SEMI_EXPANDED);
  DEFINE_INTEGER(XG_PRE "PANGO_STRETCH_EXPANDED" XG_POST,                       PANGO_STRETCH_EXPANDED);
  DEFINE_INTEGER(XG_PRE "PANGO_STRETCH_EXTRA_EXPANDED" XG_POST,                 PANGO_STRETCH_EXTRA_EXPANDED);
  DEFINE_INTEGER(XG_PRE "PANGO_STRETCH_ULTRA_EXPANDED" XG_POST,                 PANGO_STRETCH_ULTRA_EXPANDED);
  DEFINE_INTEGER(XG_PRE "PANGO_FONT_MASK_FAMILY" XG_POST,                       PANGO_FONT_MASK_FAMILY);
  DEFINE_INTEGER(XG_PRE "PANGO_FONT_MASK_STYLE" XG_POST,                        PANGO_FONT_MASK_STYLE);
  DEFINE_INTEGER(XG_PRE "PANGO_FONT_MASK_VARIANT" XG_POST,                      PANGO_FONT_MASK_VARIANT);
  DEFINE_INTEGER(XG_PRE "PANGO_FONT_MASK_WEIGHT" XG_POST,                       PANGO_FONT_MASK_WEIGHT);
  DEFINE_INTEGER(XG_PRE "PANGO_FONT_MASK_STRETCH" XG_POST,                      PANGO_FONT_MASK_STRETCH);
  DEFINE_INTEGER(XG_PRE "PANGO_FONT_MASK_SIZE" XG_POST,                         PANGO_FONT_MASK_SIZE);
  DEFINE_INTEGER(XG_PRE "PANGO_ALIGN_LEFT" XG_POST,                             PANGO_ALIGN_LEFT);
  DEFINE_INTEGER(XG_PRE "PANGO_ALIGN_CENTER" XG_POST,                           PANGO_ALIGN_CENTER);
  DEFINE_INTEGER(XG_PRE "PANGO_ALIGN_RIGHT" XG_POST,                            PANGO_ALIGN_RIGHT);
  DEFINE_INTEGER(XG_PRE "PANGO_WRAP_WORD" XG_POST,                              PANGO_WRAP_WORD);
  DEFINE_INTEGER(XG_PRE "PANGO_WRAP_CHAR" XG_POST,                              PANGO_WRAP_CHAR);
  DEFINE_INTEGER(XG_PRE "PANGO_DIRECTION_LTR" XG_POST,                          PANGO_DIRECTION_LTR);
  DEFINE_INTEGER(XG_PRE "PANGO_DIRECTION_RTL" XG_POST,                          PANGO_DIRECTION_RTL);
  DEFINE_INTEGER(XG_PRE "PANGO_DIRECTION_TTB_LTR" XG_POST,                      PANGO_DIRECTION_TTB_LTR);
  DEFINE_INTEGER(XG_PRE "PANGO_DIRECTION_TTB_RTL" XG_POST,                      PANGO_DIRECTION_TTB_RTL);
  DEFINE_INTEGER(XG_PRE "G_NORMALIZE_DEFAULT" XG_POST,                          G_NORMALIZE_DEFAULT);
  DEFINE_INTEGER(XG_PRE "G_NORMALIZE_NFD" XG_POST,                              G_NORMALIZE_NFD);
  DEFINE_INTEGER(XG_PRE "G_NORMALIZE_DEFAULT_COMPOSE" XG_POST,                  G_NORMALIZE_DEFAULT_COMPOSE);
  DEFINE_INTEGER(XG_PRE "G_NORMALIZE_NFC" XG_POST,                              G_NORMALIZE_NFC);
  DEFINE_INTEGER(XG_PRE "G_NORMALIZE_ALL" XG_POST,                              G_NORMALIZE_ALL);
  DEFINE_INTEGER(XG_PRE "G_NORMALIZE_NFKD" XG_POST,                             G_NORMALIZE_NFKD);
  DEFINE_INTEGER(XG_PRE "G_NORMALIZE_ALL_COMPOSE" XG_POST,                      G_NORMALIZE_ALL_COMPOSE);
  DEFINE_INTEGER(XG_PRE "G_NORMALIZE_NFKC" XG_POST,                             G_NORMALIZE_NFKC);
  DEFINE_ULONG(XG_PRE "GDK_TYPE_PIXBUF" XG_POST,                                GDK_TYPE_PIXBUF);
  DEFINE_ULONG(XG_PRE "GDK_TYPE_PIXBUF_ANIMATION" XG_POST,                      GDK_TYPE_PIXBUF_ANIMATION);
  DEFINE_ULONG(XG_PRE "GDK_TYPE_PIXBUF_ANIMATION_ITER" XG_POST,                 GDK_TYPE_PIXBUF_ANIMATION_ITER);
  DEFINE_ULONG(XG_PRE "GDK_PIXBUF_ERROR" XG_POST,                               GDK_PIXBUF_ERROR);
  DEFINE_ULONG(XG_PRE "GDK_TYPE_PIXBUF_LOADER" XG_POST,                         GDK_TYPE_PIXBUF_LOADER);
  DEFINE_ULONG(XG_PRE "GDK_TYPE_PIXBUF_ALPHA_MODE" XG_POST,                     GDK_TYPE_PIXBUF_ALPHA_MODE);
  DEFINE_ULONG(XG_PRE "GDK_TYPE_COLORSPACE" XG_POST,                            GDK_TYPE_COLORSPACE);
  DEFINE_ULONG(XG_PRE "GDK_TYPE_PIXBUF_ERROR" XG_POST,                          GDK_TYPE_PIXBUF_ERROR);
  DEFINE_ULONG(XG_PRE "GDK_TYPE_INTERP_TYPE" XG_POST,                           GDK_TYPE_INTERP_TYPE);
#if HAVE_GTK_ICON_SET_GET_TYPE
  DEFINE_ULONG(XG_PRE "GTK_TYPE_ICON_SET" XG_POST,                              GTK_TYPE_ICON_SET);
#endif
#if HAVE_GTK_ICON_SET_GET_TYPE
  DEFINE_ULONG(XG_PRE "GTK_TYPE_ICON_SOURCE" XG_POST,                           GTK_TYPE_ICON_SOURCE);
#endif
#if HAVE_GTK_ICON_SET_GET_TYPE
  DEFINE_ULONG(XG_PRE "GTK_TYPE_SELECTION_DATA" XG_POST,                        GTK_TYPE_SELECTION_DATA);
#endif
#if HAVE_GTK_ICON_SET_GET_TYPE
  DEFINE_ULONG(XG_PRE "GTK_TYPE_BORDER" XG_POST,                                GTK_TYPE_BORDER);
#endif
#if HAVE_GTK_ICON_SET_GET_TYPE
  DEFINE_ULONG(XG_PRE "GTK_TYPE_TREE_ITER" XG_POST,                             GTK_TYPE_TREE_ITER);
#endif
#if HAVE_GTK_ICON_SET_GET_TYPE
  DEFINE_ULONG(XG_PRE "GTK_TYPE_TREE_PATH" XG_POST,                             GTK_TYPE_TREE_PATH);
#endif
#if HAVE_GTK_ICON_SET_GET_TYPE
  DEFINE_ULONG(XG_PRE "GTK_TYPE_IDENTIFIER" XG_POST,                            GTK_TYPE_IDENTIFIER);
#endif
#if HAVE_GTK_ICON_SET_GET_TYPE
  DEFINE_ULONG(XG_PRE "GTK_TYPE_REQUISITION" XG_POST,                           GTK_TYPE_REQUISITION);
#endif
  DEFINE_ULONG(XG_PRE "GTK_TYPE_ACCEL_GROUP" XG_POST,                           GTK_TYPE_ACCEL_GROUP);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_ACCEL_LABEL" XG_POST,                           GTK_TYPE_ACCEL_LABEL);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_ACCESSIBLE" XG_POST,                            GTK_TYPE_ACCESSIBLE);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_ADJUSTMENT" XG_POST,                            GTK_TYPE_ADJUSTMENT);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_ALIGNMENT" XG_POST,                             GTK_TYPE_ALIGNMENT);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_ARROW" XG_POST,                                 GTK_TYPE_ARROW);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_ASPECT_FRAME" XG_POST,                          GTK_TYPE_ASPECT_FRAME);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_BUTTON_BOX" XG_POST,                            GTK_TYPE_BUTTON_BOX);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_BIN" XG_POST,                                   GTK_TYPE_BIN);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_BOX" XG_POST,                                   GTK_TYPE_BOX);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_BUTTON" XG_POST,                                GTK_TYPE_BUTTON);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_CALENDAR" XG_POST,                              GTK_TYPE_CALENDAR);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_CELL_EDITABLE" XG_POST,                         GTK_TYPE_CELL_EDITABLE);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_CELL_RENDERER" XG_POST,                         GTK_TYPE_CELL_RENDERER);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_CELL_RENDERER_PIXBUF" XG_POST,                  GTK_TYPE_CELL_RENDERER_PIXBUF);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_CELL_RENDERER_TEXT" XG_POST,                    GTK_TYPE_CELL_RENDERER_TEXT);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_CELL_RENDERER_TOGGLE" XG_POST,                  GTK_TYPE_CELL_RENDERER_TOGGLE);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_CHECK_BUTTON" XG_POST,                          GTK_TYPE_CHECK_BUTTON);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_CHECK_MENU_ITEM" XG_POST,                       GTK_TYPE_CHECK_MENU_ITEM);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_COLOR_SELECTION_DIALOG" XG_POST,                GTK_TYPE_COLOR_SELECTION_DIALOG);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_COLOR_SELECTION" XG_POST,                       GTK_TYPE_COLOR_SELECTION);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_COMBO" XG_POST,                                 GTK_TYPE_COMBO);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_CONTAINER" XG_POST,                             GTK_TYPE_CONTAINER);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_CURVE" XG_POST,                                 GTK_TYPE_CURVE);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_DIALOG" XG_POST,                                GTK_TYPE_DIALOG);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_DRAWING_AREA" XG_POST,                          GTK_TYPE_DRAWING_AREA);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_EDITABLE" XG_POST,                              GTK_TYPE_EDITABLE);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_ENTRY" XG_POST,                                 GTK_TYPE_ENTRY);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_EVENT_BOX" XG_POST,                             GTK_TYPE_EVENT_BOX);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_FILE_SELECTION" XG_POST,                        GTK_TYPE_FILE_SELECTION);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_FIXED" XG_POST,                                 GTK_TYPE_FIXED);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_FONT_SELECTION" XG_POST,                        GTK_TYPE_FONT_SELECTION);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_FONT_SELECTION_DIALOG" XG_POST,                 GTK_TYPE_FONT_SELECTION_DIALOG);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_FRAME" XG_POST,                                 GTK_TYPE_FRAME);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_GAMMA_CURVE" XG_POST,                           GTK_TYPE_GAMMA_CURVE);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_HANDLE_BOX" XG_POST,                            GTK_TYPE_HANDLE_BOX);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_HBUTTON_BOX" XG_POST,                           GTK_TYPE_HBUTTON_BOX);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_HBOX" XG_POST,                                  GTK_TYPE_HBOX);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_HPANED" XG_POST,                                GTK_TYPE_HPANED);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_HRULER" XG_POST,                                GTK_TYPE_HRULER);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_HSCALE" XG_POST,                                GTK_TYPE_HSCALE);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_HSCROLLBAR" XG_POST,                            GTK_TYPE_HSCROLLBAR);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_HSEPARATOR" XG_POST,                            GTK_TYPE_HSEPARATOR);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_ICON_FACTORY" XG_POST,                          GTK_TYPE_ICON_FACTORY);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_IMAGE" XG_POST,                                 GTK_TYPE_IMAGE);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_IMAGE_MENU_ITEM" XG_POST,                       GTK_TYPE_IMAGE_MENU_ITEM);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_IM_CONTEXT" XG_POST,                            GTK_TYPE_IM_CONTEXT);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_IM_CONTEXT_SIMPLE" XG_POST,                     GTK_TYPE_IM_CONTEXT_SIMPLE);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_IM_MULTICONTEXT" XG_POST,                       GTK_TYPE_IM_MULTICONTEXT);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_INPUT_DIALOG" XG_POST,                          GTK_TYPE_INPUT_DIALOG);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_INVISIBLE" XG_POST,                             GTK_TYPE_INVISIBLE);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_ITEM_FACTORY" XG_POST,                          GTK_TYPE_ITEM_FACTORY);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_ITEM" XG_POST,                                  GTK_TYPE_ITEM);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_LABEL" XG_POST,                                 GTK_TYPE_LABEL);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_LAYOUT" XG_POST,                                GTK_TYPE_LAYOUT);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_LIST_STORE" XG_POST,                            GTK_TYPE_LIST_STORE);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_MENU_BAR" XG_POST,                              GTK_TYPE_MENU_BAR);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_MENU" XG_POST,                                  GTK_TYPE_MENU);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_MENU_ITEM" XG_POST,                             GTK_TYPE_MENU_ITEM);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_MENU_SHELL" XG_POST,                            GTK_TYPE_MENU_SHELL);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_MESSAGE_DIALOG" XG_POST,                        GTK_TYPE_MESSAGE_DIALOG);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_MISC" XG_POST,                                  GTK_TYPE_MISC);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_NOTEBOOK" XG_POST,                              GTK_TYPE_NOTEBOOK);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_OBJECT" XG_POST,                                GTK_TYPE_OBJECT);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_OPTION_MENU" XG_POST,                           GTK_TYPE_OPTION_MENU);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_PANED" XG_POST,                                 GTK_TYPE_PANED);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_PLUG" XG_POST,                                  GTK_TYPE_PLUG);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_PROGRESS_BAR" XG_POST,                          GTK_TYPE_PROGRESS_BAR);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_RADIO_BUTTON" XG_POST,                          GTK_TYPE_RADIO_BUTTON);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_RADIO_MENU_ITEM" XG_POST,                       GTK_TYPE_RADIO_MENU_ITEM);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_RANGE" XG_POST,                                 GTK_TYPE_RANGE);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_RC_STYLE" XG_POST,                              GTK_TYPE_RC_STYLE);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_RULER" XG_POST,                                 GTK_TYPE_RULER);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_SCALE" XG_POST,                                 GTK_TYPE_SCALE);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_SCROLLBAR" XG_POST,                             GTK_TYPE_SCROLLBAR);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_SCROLLED_WINDOW" XG_POST,                       GTK_TYPE_SCROLLED_WINDOW);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_SEPARATOR" XG_POST,                             GTK_TYPE_SEPARATOR);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_SEPARATOR_MENU_ITEM" XG_POST,                   GTK_TYPE_SEPARATOR_MENU_ITEM);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_SETTINGS" XG_POST,                              GTK_TYPE_SETTINGS);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_SIZE_GROUP" XG_POST,                            GTK_TYPE_SIZE_GROUP);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_SOCKET" XG_POST,                                GTK_TYPE_SOCKET);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_SPIN_BUTTON" XG_POST,                           GTK_TYPE_SPIN_BUTTON);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_STATUSBAR" XG_POST,                             GTK_TYPE_STATUSBAR);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_STYLE" XG_POST,                                 GTK_TYPE_STYLE);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_TABLE" XG_POST,                                 GTK_TYPE_TABLE);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_TEAROFF_MENU_ITEM" XG_POST,                     GTK_TYPE_TEAROFF_MENU_ITEM);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_TEXT_BUFFER" XG_POST,                           GTK_TYPE_TEXT_BUFFER);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_TEXT_CHILD_ANCHOR" XG_POST,                     GTK_TYPE_TEXT_CHILD_ANCHOR);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_TEXT_MARK" XG_POST,                             GTK_TYPE_TEXT_MARK);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_TEXT_TAG" XG_POST,                              GTK_TYPE_TEXT_TAG);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_TEXT_TAG_TABLE" XG_POST,                        GTK_TYPE_TEXT_TAG_TABLE);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_TEXT_VIEW" XG_POST,                             GTK_TYPE_TEXT_VIEW);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_TOGGLE_BUTTON" XG_POST,                         GTK_TYPE_TOGGLE_BUTTON);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_TOOLBAR" XG_POST,                               GTK_TYPE_TOOLBAR);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_TOOLTIPS" XG_POST,                              GTK_TYPE_TOOLTIPS);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_TREE_DRAG_SOURCE" XG_POST,                      GTK_TYPE_TREE_DRAG_SOURCE);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_TREE_DRAG_DEST" XG_POST,                        GTK_TYPE_TREE_DRAG_DEST);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_TREE_MODEL" XG_POST,                            GTK_TYPE_TREE_MODEL);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_TREE_MODEL_SORT" XG_POST,                       GTK_TYPE_TREE_MODEL_SORT);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_TREE_SELECTION" XG_POST,                        GTK_TYPE_TREE_SELECTION);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_TREE_SORTABLE" XG_POST,                         GTK_TYPE_TREE_SORTABLE);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_TREE_STORE" XG_POST,                            GTK_TYPE_TREE_STORE);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_TREE_VIEW_COLUMN" XG_POST,                      GTK_TYPE_TREE_VIEW_COLUMN);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_TREE_VIEW" XG_POST,                             GTK_TYPE_TREE_VIEW);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_VBUTTON_BOX" XG_POST,                           GTK_TYPE_VBUTTON_BOX);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_VBOX" XG_POST,                                  GTK_TYPE_VBOX);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_VIEWPORT" XG_POST,                              GTK_TYPE_VIEWPORT);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_VPANED" XG_POST,                                GTK_TYPE_VPANED);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_VRULER" XG_POST,                                GTK_TYPE_VRULER);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_VSCALE" XG_POST,                                GTK_TYPE_VSCALE);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_VSCROLLBAR" XG_POST,                            GTK_TYPE_VSCROLLBAR);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_VSEPARATOR" XG_POST,                            GTK_TYPE_VSEPARATOR);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_WIDGET" XG_POST,                                GTK_TYPE_WIDGET);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_WINDOW" XG_POST,                                GTK_TYPE_WINDOW);
  DEFINE_ULONG(XG_PRE "GTK_TYPE_WINDOW_GROUP" XG_POST,                          GTK_TYPE_WINDOW_GROUP);
  DEFINE_ULONG(XG_PRE "GDK_TYPE_COLORMAP" XG_POST,                              GDK_TYPE_COLORMAP);
  DEFINE_ULONG(XG_PRE "GDK_TYPE_COLOR" XG_POST,                                 GDK_TYPE_COLOR);
  DEFINE_ULONG(XG_PRE "GDK_TYPE_CURSOR" XG_POST,                                GDK_TYPE_CURSOR);
  DEFINE_ULONG(XG_PRE "GDK_TYPE_DRAG_CONTEXT" XG_POST,                          GDK_TYPE_DRAG_CONTEXT);
  DEFINE_ULONG(XG_PRE "GDK_TYPE_DRAWABLE" XG_POST,                              GDK_TYPE_DRAWABLE);
  DEFINE_ULONG(XG_PRE "GDK_TYPE_CURSOR_TYPE" XG_POST,                           GDK_TYPE_CURSOR_TYPE);
  DEFINE_ULONG(XG_PRE "GDK_TYPE_DRAG_ACTION" XG_POST,                           GDK_TYPE_DRAG_ACTION);
  DEFINE_ULONG(XG_PRE "GDK_TYPE_DRAG_PROTOCOL" XG_POST,                         GDK_TYPE_DRAG_PROTOCOL);
  DEFINE_ULONG(XG_PRE "GDK_TYPE_FILTER_RETURN" XG_POST,                         GDK_TYPE_FILTER_RETURN);
  DEFINE_ULONG(XG_PRE "GDK_TYPE_EVENT_TYPE" XG_POST,                            GDK_TYPE_EVENT_TYPE);
  DEFINE_ULONG(XG_PRE "GDK_TYPE_EVENT_MASK" XG_POST,                            GDK_TYPE_EVENT_MASK);
  DEFINE_ULONG(XG_PRE "GDK_TYPE_VISIBILITY_STATE" XG_POST,                      GDK_TYPE_VISIBILITY_STATE);
  DEFINE_ULONG(XG_PRE "GDK_TYPE_SCROLL_DIRECTION" XG_POST,                      GDK_TYPE_SCROLL_DIRECTION);
  DEFINE_ULONG(XG_PRE "GDK_TYPE_NOTIFY_TYPE" XG_POST,                           GDK_TYPE_NOTIFY_TYPE);
  DEFINE_ULONG(XG_PRE "GDK_TYPE_CROSSING_MODE" XG_POST,                         GDK_TYPE_CROSSING_MODE);
  DEFINE_ULONG(XG_PRE "GDK_TYPE_PROPERTY_STATE" XG_POST,                        GDK_TYPE_PROPERTY_STATE);
  DEFINE_ULONG(XG_PRE "GDK_TYPE_WINDOW_STATE" XG_POST,                          GDK_TYPE_WINDOW_STATE);
  DEFINE_ULONG(XG_PRE "GDK_TYPE_SETTING_ACTION" XG_POST,                        GDK_TYPE_SETTING_ACTION);
  DEFINE_ULONG(XG_PRE "GDK_TYPE_FONT_TYPE" XG_POST,                             GDK_TYPE_FONT_TYPE);
  DEFINE_ULONG(XG_PRE "GDK_TYPE_CAP_STYLE" XG_POST,                             GDK_TYPE_CAP_STYLE);
  DEFINE_ULONG(XG_PRE "GDK_TYPE_FILL" XG_POST,                                  GDK_TYPE_FILL);
  DEFINE_ULONG(XG_PRE "GDK_TYPE_FUNCTION" XG_POST,                              GDK_TYPE_FUNCTION);
  DEFINE_ULONG(XG_PRE "GDK_TYPE_JOIN_STYLE" XG_POST,                            GDK_TYPE_JOIN_STYLE);
  DEFINE_ULONG(XG_PRE "GDK_TYPE_LINE_STYLE" XG_POST,                            GDK_TYPE_LINE_STYLE);
  DEFINE_ULONG(XG_PRE "GDK_TYPE_SUBWINDOW_MODE" XG_POST,                        GDK_TYPE_SUBWINDOW_MODE);
  DEFINE_ULONG(XG_PRE "GDK_TYPE_GC_VALUES_MASK" XG_POST,                        GDK_TYPE_GC_VALUES_MASK);
  DEFINE_ULONG(XG_PRE "GDK_TYPE_IMAGE_TYPE" XG_POST,                            GDK_TYPE_IMAGE_TYPE);
  DEFINE_ULONG(XG_PRE "GDK_TYPE_EXTENSION_MODE" XG_POST,                        GDK_TYPE_EXTENSION_MODE);
  DEFINE_ULONG(XG_PRE "GDK_TYPE_INPUT_SOURCE" XG_POST,                          GDK_TYPE_INPUT_SOURCE);
  DEFINE_ULONG(XG_PRE "GDK_TYPE_INPUT_MODE" XG_POST,                            GDK_TYPE_INPUT_MODE);
  DEFINE_ULONG(XG_PRE "GDK_TYPE_AXIS_USE" XG_POST,                              GDK_TYPE_AXIS_USE);
  DEFINE_ULONG(XG_PRE "GDK_TYPE_PROP_MODE" XG_POST,                             GDK_TYPE_PROP_MODE);
  DEFINE_ULONG(XG_PRE "GDK_TYPE_FILL_RULE" XG_POST,                             GDK_TYPE_FILL_RULE);
  DEFINE_ULONG(XG_PRE "GDK_TYPE_OVERLAP_TYPE" XG_POST,                          GDK_TYPE_OVERLAP_TYPE);
  DEFINE_ULONG(XG_PRE "GDK_TYPE_RGB_DITHER" XG_POST,                            GDK_TYPE_RGB_DITHER);
  DEFINE_ULONG(XG_PRE "GDK_TYPE_BYTE_ORDER" XG_POST,                            GDK_TYPE_BYTE_ORDER);
  DEFINE_ULONG(XG_PRE "GDK_TYPE_MODIFIER_TYPE" XG_POST,                         GDK_TYPE_MODIFIER_TYPE);
  DEFINE_ULONG(XG_PRE "GDK_TYPE_INPUT_CONDITION" XG_POST,                       GDK_TYPE_INPUT_CONDITION);
  DEFINE_ULONG(XG_PRE "GDK_TYPE_STATUS" XG_POST,                                GDK_TYPE_STATUS);
  DEFINE_ULONG(XG_PRE "GDK_TYPE_GRAB_STATUS" XG_POST,                           GDK_TYPE_GRAB_STATUS);
  DEFINE_ULONG(XG_PRE "GDK_TYPE_VISUAL_TYPE" XG_POST,                           GDK_TYPE_VISUAL_TYPE);
  DEFINE_ULONG(XG_PRE "GDK_TYPE_WINDOW_CLASS" XG_POST,                          GDK_TYPE_WINDOW_CLASS);
  DEFINE_ULONG(XG_PRE "GDK_TYPE_WINDOW_TYPE" XG_POST,                           GDK_TYPE_WINDOW_TYPE);
  DEFINE_ULONG(XG_PRE "GDK_TYPE_WINDOW_ATTRIBUTES_TYPE" XG_POST,                GDK_TYPE_WINDOW_ATTRIBUTES_TYPE);
  DEFINE_ULONG(XG_PRE "GDK_TYPE_WINDOW_HINTS" XG_POST,                          GDK_TYPE_WINDOW_HINTS);
  DEFINE_ULONG(XG_PRE "GDK_TYPE_WINDOW_TYPE_HINT" XG_POST,                      GDK_TYPE_WINDOW_TYPE_HINT);
  DEFINE_ULONG(XG_PRE "GDK_TYPE_WM_DECORATION" XG_POST,                         GDK_TYPE_WM_DECORATION);
  DEFINE_ULONG(XG_PRE "GDK_TYPE_WM_FUNCTION" XG_POST,                           GDK_TYPE_WM_FUNCTION);
  DEFINE_ULONG(XG_PRE "GDK_TYPE_GRAVITY" XG_POST,                               GDK_TYPE_GRAVITY);
  DEFINE_ULONG(XG_PRE "GDK_TYPE_WINDOW_EDGE" XG_POST,                           GDK_TYPE_WINDOW_EDGE);
  DEFINE_ULONG(XG_PRE "GDK_TYPE_EVENT" XG_POST,                                 GDK_TYPE_EVENT);
  DEFINE_ULONG(XG_PRE "GDK_TYPE_GC" XG_POST,                                    GDK_TYPE_GC);
  DEFINE_ULONG(XG_PRE "GDK_TYPE_RECTANGLE" XG_POST,                             GDK_TYPE_RECTANGLE);
  DEFINE_ULONG(XG_PRE "GDK_TYPE_IMAGE" XG_POST,                                 GDK_TYPE_IMAGE);
  DEFINE_ULONG(XG_PRE "GDK_TYPE_DEVICE" XG_POST,                                GDK_TYPE_DEVICE);
  DEFINE_ULONG(XG_PRE "GDK_TYPE_KEYMAP" XG_POST,                                GDK_TYPE_KEYMAP);
  DEFINE_ULONG(XG_PRE "GDK_TYPE_PIXMAP" XG_POST,                                GDK_TYPE_PIXMAP);
  DEFINE_ULONG(XG_PRE "GDK_TYPE_VISUAL" XG_POST,                                GDK_TYPE_VISUAL);
  DEFINE_ULONG(XG_PRE "GDK_TYPE_WINDOW" XG_POST,                                GDK_TYPE_WINDOW);
  DEFINE_ULONG(XG_PRE "G_TYPE_IO_CONDITION" XG_POST,                            G_TYPE_IO_CONDITION);
  DEFINE_ULONG(XG_PRE "G_TYPE_INVALID" XG_POST,                                 G_TYPE_INVALID);
  DEFINE_ULONG(XG_PRE "G_TYPE_NONE" XG_POST,                                    G_TYPE_NONE);
  DEFINE_ULONG(XG_PRE "G_TYPE_INTERFACE" XG_POST,                               G_TYPE_INTERFACE);
  DEFINE_ULONG(XG_PRE "G_TYPE_CHAR" XG_POST,                                    G_TYPE_CHAR);
  DEFINE_ULONG(XG_PRE "G_TYPE_UCHAR" XG_POST,                                   G_TYPE_UCHAR);
  DEFINE_ULONG(XG_PRE "G_TYPE_BOOLEAN" XG_POST,                                 G_TYPE_BOOLEAN);
  DEFINE_ULONG(XG_PRE "G_TYPE_INT" XG_POST,                                     G_TYPE_INT);
  DEFINE_ULONG(XG_PRE "G_TYPE_UINT" XG_POST,                                    G_TYPE_UINT);
  DEFINE_ULONG(XG_PRE "G_TYPE_LONG" XG_POST,                                    G_TYPE_LONG);
  DEFINE_ULONG(XG_PRE "G_TYPE_ULONG" XG_POST,                                   G_TYPE_ULONG);
  DEFINE_ULONG(XG_PRE "G_TYPE_INT64" XG_POST,                                   G_TYPE_INT64);
  DEFINE_ULONG(XG_PRE "G_TYPE_UINT64" XG_POST,                                  G_TYPE_UINT64);
  DEFINE_ULONG(XG_PRE "G_TYPE_ENUM" XG_POST,                                    G_TYPE_ENUM);
  DEFINE_ULONG(XG_PRE "G_TYPE_FLAGS" XG_POST,                                   G_TYPE_FLAGS);
  DEFINE_ULONG(XG_PRE "G_TYPE_FLOAT" XG_POST,                                   G_TYPE_FLOAT);
  DEFINE_ULONG(XG_PRE "G_TYPE_DOUBLE" XG_POST,                                  G_TYPE_DOUBLE);
  DEFINE_ULONG(XG_PRE "G_TYPE_STRING" XG_POST,                                  G_TYPE_STRING);
  DEFINE_ULONG(XG_PRE "G_TYPE_POINTER" XG_POST,                                 G_TYPE_POINTER);
  DEFINE_ULONG(XG_PRE "G_TYPE_BOXED" XG_POST,                                   G_TYPE_BOXED);
  DEFINE_ULONG(XG_PRE "G_TYPE_PARAM" XG_POST,                                   G_TYPE_PARAM);
  DEFINE_ULONG(XG_PRE "G_TYPE_OBJECT" XG_POST,                                  G_TYPE_OBJECT);
  DEFINE_ULONG(XG_PRE "G_UNICODE_CONTROL" XG_POST,                              G_UNICODE_CONTROL);
  DEFINE_ULONG(XG_PRE "G_UNICODE_FORMAT" XG_POST,                               G_UNICODE_FORMAT);
  DEFINE_ULONG(XG_PRE "G_UNICODE_UNASSIGNED" XG_POST,                           G_UNICODE_UNASSIGNED);
  DEFINE_ULONG(XG_PRE "G_UNICODE_PRIVATE_USE" XG_POST,                          G_UNICODE_PRIVATE_USE);
  DEFINE_ULONG(XG_PRE "G_UNICODE_SURROGATE" XG_POST,                            G_UNICODE_SURROGATE);
  DEFINE_ULONG(XG_PRE "G_UNICODE_LOWERCASE_LETTER" XG_POST,                     G_UNICODE_LOWERCASE_LETTER);
  DEFINE_ULONG(XG_PRE "G_UNICODE_MODIFIER_LETTER" XG_POST,                      G_UNICODE_MODIFIER_LETTER);
  DEFINE_ULONG(XG_PRE "G_UNICODE_OTHER_LETTER" XG_POST,                         G_UNICODE_OTHER_LETTER);
  DEFINE_ULONG(XG_PRE "G_UNICODE_TITLECASE_LETTER" XG_POST,                     G_UNICODE_TITLECASE_LETTER);
  DEFINE_ULONG(XG_PRE "G_UNICODE_UPPERCASE_LETTER" XG_POST,                     G_UNICODE_UPPERCASE_LETTER);
  DEFINE_ULONG(XG_PRE "G_UNICODE_COMBINING_MARK" XG_POST,                       G_UNICODE_COMBINING_MARK);
  DEFINE_ULONG(XG_PRE "G_UNICODE_ENCLOSING_MARK" XG_POST,                       G_UNICODE_ENCLOSING_MARK);
  DEFINE_ULONG(XG_PRE "G_UNICODE_NON_SPACING_MARK" XG_POST,                     G_UNICODE_NON_SPACING_MARK);
  DEFINE_ULONG(XG_PRE "G_UNICODE_DECIMAL_NUMBER" XG_POST,                       G_UNICODE_DECIMAL_NUMBER);
  DEFINE_ULONG(XG_PRE "G_UNICODE_LETTER_NUMBER" XG_POST,                        G_UNICODE_LETTER_NUMBER);
  DEFINE_ULONG(XG_PRE "G_UNICODE_OTHER_NUMBER" XG_POST,                         G_UNICODE_OTHER_NUMBER);
  DEFINE_ULONG(XG_PRE "G_UNICODE_CONNECT_PUNCTUATION" XG_POST,                  G_UNICODE_CONNECT_PUNCTUATION);
  DEFINE_ULONG(XG_PRE "G_UNICODE_DASH_PUNCTUATION" XG_POST,                     G_UNICODE_DASH_PUNCTUATION);
  DEFINE_ULONG(XG_PRE "G_UNICODE_CLOSE_PUNCTUATION" XG_POST,                    G_UNICODE_CLOSE_PUNCTUATION);
  DEFINE_ULONG(XG_PRE "G_UNICODE_FINAL_PUNCTUATION" XG_POST,                    G_UNICODE_FINAL_PUNCTUATION);
  DEFINE_ULONG(XG_PRE "G_UNICODE_INITIAL_PUNCTUATION" XG_POST,                  G_UNICODE_INITIAL_PUNCTUATION);
  DEFINE_ULONG(XG_PRE "G_UNICODE_OTHER_PUNCTUATION" XG_POST,                    G_UNICODE_OTHER_PUNCTUATION);
  DEFINE_ULONG(XG_PRE "G_UNICODE_OPEN_PUNCTUATION" XG_POST,                     G_UNICODE_OPEN_PUNCTUATION);
  DEFINE_ULONG(XG_PRE "G_UNICODE_CURRENCY_SYMBOL" XG_POST,                      G_UNICODE_CURRENCY_SYMBOL);
  DEFINE_ULONG(XG_PRE "G_UNICODE_MODIFIER_SYMBOL" XG_POST,                      G_UNICODE_MODIFIER_SYMBOL);
  DEFINE_ULONG(XG_PRE "G_UNICODE_MATH_SYMBOL" XG_POST,                          G_UNICODE_MATH_SYMBOL);
  DEFINE_ULONG(XG_PRE "G_UNICODE_OTHER_SYMBOL" XG_POST,                         G_UNICODE_OTHER_SYMBOL);
  DEFINE_ULONG(XG_PRE "G_UNICODE_LINE_SEPARATOR" XG_POST,                       G_UNICODE_LINE_SEPARATOR);
  DEFINE_ULONG(XG_PRE "G_UNICODE_PARAGRAPH_SEPARATOR" XG_POST,                  G_UNICODE_PARAGRAPH_SEPARATOR);
  DEFINE_ULONG(XG_PRE "G_UNICODE_SPACE_SEPARATOR" XG_POST,                      G_UNICODE_SPACE_SEPARATOR);
  DEFINE_ULONG(XG_PRE "G_UNICODE_BREAK_MANDATORY" XG_POST,                      G_UNICODE_BREAK_MANDATORY);
  DEFINE_ULONG(XG_PRE "G_UNICODE_BREAK_CARRIAGE_RETURN" XG_POST,                G_UNICODE_BREAK_CARRIAGE_RETURN);
  DEFINE_ULONG(XG_PRE "G_UNICODE_BREAK_LINE_FEED" XG_POST,                      G_UNICODE_BREAK_LINE_FEED);
  DEFINE_ULONG(XG_PRE "G_UNICODE_BREAK_COMBINING_MARK" XG_POST,                 G_UNICODE_BREAK_COMBINING_MARK);
  DEFINE_ULONG(XG_PRE "G_UNICODE_BREAK_SURROGATE" XG_POST,                      G_UNICODE_BREAK_SURROGATE);
  DEFINE_ULONG(XG_PRE "G_UNICODE_BREAK_ZERO_WIDTH_SPACE" XG_POST,               G_UNICODE_BREAK_ZERO_WIDTH_SPACE);
  DEFINE_ULONG(XG_PRE "G_UNICODE_BREAK_INSEPARABLE" XG_POST,                    G_UNICODE_BREAK_INSEPARABLE);
  DEFINE_ULONG(XG_PRE "G_UNICODE_BREAK_NON_BREAKING_GLUE" XG_POST,              G_UNICODE_BREAK_NON_BREAKING_GLUE);
  DEFINE_ULONG(XG_PRE "G_UNICODE_BREAK_CONTINGENT" XG_POST,                     G_UNICODE_BREAK_CONTINGENT);
  DEFINE_ULONG(XG_PRE "G_UNICODE_BREAK_SPACE" XG_POST,                          G_UNICODE_BREAK_SPACE);
  DEFINE_ULONG(XG_PRE "G_UNICODE_BREAK_AFTER" XG_POST,                          G_UNICODE_BREAK_AFTER);
  DEFINE_ULONG(XG_PRE "G_UNICODE_BREAK_BEFORE" XG_POST,                         G_UNICODE_BREAK_BEFORE);
  DEFINE_ULONG(XG_PRE "G_UNICODE_BREAK_BEFORE_AND_AFTER" XG_POST,               G_UNICODE_BREAK_BEFORE_AND_AFTER);
  DEFINE_ULONG(XG_PRE "G_UNICODE_BREAK_HYPHEN" XG_POST,                         G_UNICODE_BREAK_HYPHEN);
  DEFINE_ULONG(XG_PRE "G_UNICODE_BREAK_NON_STARTER" XG_POST,                    G_UNICODE_BREAK_NON_STARTER);
  DEFINE_ULONG(XG_PRE "G_UNICODE_BREAK_OPEN_PUNCTUATION" XG_POST,               G_UNICODE_BREAK_OPEN_PUNCTUATION);
  DEFINE_ULONG(XG_PRE "G_UNICODE_BREAK_CLOSE_PUNCTUATION" XG_POST,              G_UNICODE_BREAK_CLOSE_PUNCTUATION);
  DEFINE_ULONG(XG_PRE "G_UNICODE_BREAK_QUOTATION" XG_POST,                      G_UNICODE_BREAK_QUOTATION);
  DEFINE_ULONG(XG_PRE "G_UNICODE_BREAK_EXCLAMATION" XG_POST,                    G_UNICODE_BREAK_EXCLAMATION);
  DEFINE_ULONG(XG_PRE "G_UNICODE_BREAK_IDEOGRAPHIC" XG_POST,                    G_UNICODE_BREAK_IDEOGRAPHIC);
  DEFINE_ULONG(XG_PRE "G_UNICODE_BREAK_NUMERIC" XG_POST,                        G_UNICODE_BREAK_NUMERIC);
  DEFINE_ULONG(XG_PRE "G_UNICODE_BREAK_INFIX_SEPARATOR" XG_POST,                G_UNICODE_BREAK_INFIX_SEPARATOR);
  DEFINE_ULONG(XG_PRE "G_UNICODE_BREAK_SYMBOL" XG_POST,                         G_UNICODE_BREAK_SYMBOL);
  DEFINE_ULONG(XG_PRE "G_UNICODE_BREAK_ALPHABETIC" XG_POST,                     G_UNICODE_BREAK_ALPHABETIC);
  DEFINE_ULONG(XG_PRE "G_UNICODE_BREAK_PREFIX" XG_POST,                         G_UNICODE_BREAK_PREFIX);
  DEFINE_ULONG(XG_PRE "G_UNICODE_BREAK_POSTFIX" XG_POST,                        G_UNICODE_BREAK_POSTFIX);
  DEFINE_ULONG(XG_PRE "G_UNICODE_BREAK_COMPLEX_CONTEXT" XG_POST,                G_UNICODE_BREAK_COMPLEX_CONTEXT);
  DEFINE_ULONG(XG_PRE "G_UNICODE_BREAK_AMBIGUOUS" XG_POST,                      G_UNICODE_BREAK_AMBIGUOUS);
  DEFINE_ULONG(XG_PRE "G_UNICODE_BREAK_UNKNOWN" XG_POST,                        G_UNICODE_BREAK_UNKNOWN);
#endif
#if (!(defined(GDK_DISABLE_DEPRECATED))) && (!(defined(GTK_DISABLE_DEPRECATED))) && (!(defined(GDK_PIXBUF_DISABLE_DEPRECATED)))
  DEFINE_INTEGER(XG_PRE "GDK_FONT_FONT" XG_POST,                                GDK_FONT_FONT);
  DEFINE_INTEGER(XG_PRE "GDK_FONT_FONTSET" XG_POST,                             GDK_FONT_FONTSET);
  DEFINE_INTEGER(XG_PRE "GTK_CLIST_DRAG_NONE" XG_POST,                          GTK_CLIST_DRAG_NONE);
  DEFINE_INTEGER(XG_PRE "GTK_CLIST_DRAG_BEFORE" XG_POST,                        GTK_CLIST_DRAG_BEFORE);
  DEFINE_INTEGER(XG_PRE "GTK_CLIST_DRAG_INTO" XG_POST,                          GTK_CLIST_DRAG_INTO);
  DEFINE_INTEGER(XG_PRE "GTK_CLIST_DRAG_AFTER" XG_POST,                         GTK_CLIST_DRAG_AFTER);
  DEFINE_INTEGER(XG_PRE "GTK_BUTTON_IGNORED" XG_POST,                           GTK_BUTTON_IGNORED);
  DEFINE_INTEGER(XG_PRE "GTK_BUTTON_SELECTS" XG_POST,                           GTK_BUTTON_SELECTS);
  DEFINE_INTEGER(XG_PRE "GTK_BUTTON_DRAGS" XG_POST,                             GTK_BUTTON_DRAGS);
  DEFINE_INTEGER(XG_PRE "GTK_BUTTON_EXPANDS" XG_POST,                           GTK_BUTTON_EXPANDS);
  DEFINE_INTEGER(XG_PRE "GTK_CTREE_POS_BEFORE" XG_POST,                         GTK_CTREE_POS_BEFORE);
  DEFINE_INTEGER(XG_PRE "GTK_CTREE_POS_AS_CHILD" XG_POST,                       GTK_CTREE_POS_AS_CHILD);
  DEFINE_INTEGER(XG_PRE "GTK_CTREE_POS_AFTER" XG_POST,                          GTK_CTREE_POS_AFTER);
  DEFINE_INTEGER(XG_PRE "GTK_CTREE_LINES_NONE" XG_POST,                         GTK_CTREE_LINES_NONE);
  DEFINE_INTEGER(XG_PRE "GTK_CTREE_LINES_SOLID" XG_POST,                        GTK_CTREE_LINES_SOLID);
  DEFINE_INTEGER(XG_PRE "GTK_CTREE_LINES_DOTTED" XG_POST,                       GTK_CTREE_LINES_DOTTED);
  DEFINE_INTEGER(XG_PRE "GTK_CTREE_LINES_TABBED" XG_POST,                       GTK_CTREE_LINES_TABBED);
  DEFINE_INTEGER(XG_PRE "GTK_CTREE_EXPANDER_NONE" XG_POST,                      GTK_CTREE_EXPANDER_NONE);
  DEFINE_INTEGER(XG_PRE "GTK_CTREE_EXPANDER_SQUARE" XG_POST,                    GTK_CTREE_EXPANDER_SQUARE);
  DEFINE_INTEGER(XG_PRE "GTK_CTREE_EXPANDER_TRIANGLE" XG_POST,                  GTK_CTREE_EXPANDER_TRIANGLE);
  DEFINE_INTEGER(XG_PRE "GTK_CTREE_EXPANDER_CIRCULAR" XG_POST,                  GTK_CTREE_EXPANDER_CIRCULAR);
  DEFINE_INTEGER(XG_PRE "GTK_CTREE_EXPANSION_EXPAND" XG_POST,                   GTK_CTREE_EXPANSION_EXPAND);
  DEFINE_INTEGER(XG_PRE "GTK_CTREE_EXPANSION_EXPAND_RECURSIVE" XG_POST,         GTK_CTREE_EXPANSION_EXPAND_RECURSIVE);
  DEFINE_INTEGER(XG_PRE "GTK_CTREE_EXPANSION_COLLAPSE" XG_POST,                 GTK_CTREE_EXPANSION_COLLAPSE);
  DEFINE_INTEGER(XG_PRE "GTK_CTREE_EXPANSION_COLLAPSE_RECURSIVE" XG_POST,       GTK_CTREE_EXPANSION_COLLAPSE_RECURSIVE);
  DEFINE_INTEGER(XG_PRE "GTK_CTREE_EXPANSION_TOGGLE" XG_POST,                   GTK_CTREE_EXPANSION_TOGGLE);
  DEFINE_INTEGER(XG_PRE "GTK_CTREE_EXPANSION_TOGGLE_RECURSIVE" XG_POST,         GTK_CTREE_EXPANSION_TOGGLE_RECURSIVE);
  DEFINE_INTEGER(XG_PRE "GTK_SIDE_TOP" XG_POST,                                 GTK_SIDE_TOP);
  DEFINE_INTEGER(XG_PRE "GTK_SIDE_BOTTOM" XG_POST,                              GTK_SIDE_BOTTOM);
  DEFINE_INTEGER(XG_PRE "GTK_SIDE_LEFT" XG_POST,                                GTK_SIDE_LEFT);
  DEFINE_INTEGER(XG_PRE "GTK_SIDE_RIGHT" XG_POST,                               GTK_SIDE_RIGHT);
  DEFINE_INTEGER(XG_PRE "GTK_MATCH_ALL" XG_POST,                                GTK_MATCH_ALL);
  DEFINE_INTEGER(XG_PRE "GTK_MATCH_ALL_TAIL" XG_POST,                           GTK_MATCH_ALL_TAIL);
  DEFINE_INTEGER(XG_PRE "GTK_MATCH_HEAD" XG_POST,                               GTK_MATCH_HEAD);
  DEFINE_INTEGER(XG_PRE "GTK_MATCH_TAIL" XG_POST,                               GTK_MATCH_TAIL);
  DEFINE_INTEGER(XG_PRE "GTK_MATCH_EXACT" XG_POST,                              GTK_MATCH_EXACT);
  DEFINE_INTEGER(XG_PRE "GTK_MATCH_LAST" XG_POST,                               GTK_MATCH_LAST);
  DEFINE_INTEGER(XG_PRE "GTK_PREVIEW_COLOR" XG_POST,                            GTK_PREVIEW_COLOR);
  DEFINE_INTEGER(XG_PRE "GTK_PREVIEW_GRAYSCALE" XG_POST,                        GTK_PREVIEW_GRAYSCALE);
  DEFINE_INTEGER(XG_PRE "GTK_RUN_FIRST" XG_POST,                                GTK_RUN_FIRST);
  DEFINE_INTEGER(XG_PRE "GTK_RUN_LAST" XG_POST,                                 GTK_RUN_LAST);
  DEFINE_INTEGER(XG_PRE "GTK_RUN_BOTH" XG_POST,                                 GTK_RUN_BOTH);
  DEFINE_INTEGER(XG_PRE "GTK_RUN_NO_RECURSE" XG_POST,                           GTK_RUN_NO_RECURSE);
  DEFINE_INTEGER(XG_PRE "GTK_RUN_ACTION" XG_POST,                               GTK_RUN_ACTION);
  DEFINE_INTEGER(XG_PRE "GTK_RUN_NO_HOOKS" XG_POST,                             GTK_RUN_NO_HOOKS);
  DEFINE_INTEGER(XG_PRE "GTK_DIRECTION_LEFT" XG_POST,                           GTK_DIRECTION_LEFT);
  DEFINE_INTEGER(XG_PRE "GTK_DIRECTION_RIGHT" XG_POST,                          GTK_DIRECTION_RIGHT);
  DEFINE_INTEGER(XG_PRE "GTK_TOP_BOTTOM" XG_POST,                               GTK_TOP_BOTTOM);
  DEFINE_INTEGER(XG_PRE "GTK_LEFT_RIGHT" XG_POST,                               GTK_LEFT_RIGHT);
  DEFINE_INTEGER(XG_PRE "GTK_PRIORITY_REDRAW" XG_POST,                          GTK_PRIORITY_REDRAW);
  DEFINE_INTEGER(XG_PRE "GTK_PRIORITY_HIGH" XG_POST,                            GTK_PRIORITY_HIGH);
  DEFINE_INTEGER(XG_PRE "GTK_PRIORITY_INTERNAL" XG_POST,                        GTK_PRIORITY_INTERNAL);
  DEFINE_INTEGER(XG_PRE "GTK_PRIORITY_DEFAULT" XG_POST,                         GTK_PRIORITY_DEFAULT);
  DEFINE_INTEGER(XG_PRE "GTK_PRIORITY_LOW" XG_POST,                             GTK_PRIORITY_LOW);
  DEFINE_INTEGER(XG_PRE "GTK_ARG_READABLE" XG_POST,                             GTK_ARG_READABLE);
  DEFINE_INTEGER(XG_PRE "GTK_ARG_WRITABLE" XG_POST,                             GTK_ARG_WRITABLE);
  DEFINE_INTEGER(XG_PRE "GTK_ARG_CONSTRUCT" XG_POST,                            GTK_ARG_CONSTRUCT);
  DEFINE_INTEGER(XG_PRE "GTK_ARG_CONSTRUCT_ONLY" XG_POST,                       GTK_ARG_CONSTRUCT_ONLY);
  DEFINE_INTEGER(XG_PRE "GTK_ARG_CHILD_ARG" XG_POST,                            GTK_ARG_CHILD_ARG);
  DEFINE_INTEGER(XG_PRE "GTK_ARG_READWRITE" XG_POST,                            GTK_ARG_READWRITE);
#endif

#if GDK_ENABLE_BROKEN && GTK_ENABLE_BROKEN
  DEFINE_INTEGER(XG_PRE "GTK_TREE_VIEW_LINE" XG_POST,                           GTK_TREE_VIEW_LINE);
  DEFINE_INTEGER(XG_PRE "GTK_TREE_VIEW_ITEM" XG_POST,                           GTK_TREE_VIEW_ITEM);
#endif

}

static void define_doubles(void)
{

#if HAVE_GUILE
#if HAVE_SCM_C_DEFINE
  #define DEFINE_DOUBLE(Name, Value) scm_c_define(Name, C_TO_XEN_DOUBLE(Value))
#else
  #define DEFINE_DOUBLE(Name, Value) gh_define(Name, C_TO_XEN_DOUBLE(Value))
#endif
#else
  #define DEFINE_DOUBLE(Name, Value) rb_define_global_const(Name, C_TO_XEN_DOUBLE(Value))
#endif

#if (!HAVE_GTK_1)
  DEFINE_DOUBLE(XG_PRE "PANGO_SCALE_XX_SMALL" XG_POST,                          PANGO_SCALE_XX_SMALL);
  DEFINE_DOUBLE(XG_PRE "PANGO_SCALE_X_SMALL" XG_POST,                           PANGO_SCALE_X_SMALL);
  DEFINE_DOUBLE(XG_PRE "PANGO_SCALE_SMALL" XG_POST,                             PANGO_SCALE_SMALL);
  DEFINE_DOUBLE(XG_PRE "PANGO_SCALE_MEDIUM" XG_POST,                            PANGO_SCALE_MEDIUM);
  DEFINE_DOUBLE(XG_PRE "PANGO_SCALE_LARGE" XG_POST,                             PANGO_SCALE_LARGE);
  DEFINE_DOUBLE(XG_PRE "PANGO_SCALE_X_LARGE" XG_POST,                           PANGO_SCALE_X_LARGE);
  DEFINE_DOUBLE(XG_PRE "PANGO_SCALE_XX_LARGE" XG_POST,                          PANGO_SCALE_XX_LARGE);
#endif
}

/* -------------------------------- predefined Atoms -------------------------------- */

static void define_atoms(void)
{
#if HAVE_GUILE
#if HAVE_SCM_C_DEFINE
  #define DEFINE_ATOM(Name, Value) scm_permanent_object(scm_c_define(Name, C_TO_XEN_GdkAtom(Value)))
#else
  #define DEFINE_ATOM(Name, Value) gh_define(Name, C_TO_XEN_GdkAtom(Value))
#endif
#else
  #define DEFINE_ATOM(Name, Value) rb_define_global_const(Name, C_TO_XEN_GdkAtom(Value))
#endif

  DEFINE_ATOM(XG_PRE "GDK_SELECTION_PRIMARY" XG_POST,                           GDK_SELECTION_PRIMARY);
  DEFINE_ATOM(XG_PRE "GDK_SELECTION_SECONDARY" XG_POST,                         GDK_SELECTION_SECONDARY);
#if (!HAVE_GTK_1)
  DEFINE_ATOM(XG_PRE "GDK_SELECTION_CLIPBOARD" XG_POST,                         GDK_SELECTION_CLIPBOARD);
#endif
  DEFINE_ATOM(XG_PRE "GDK_TARGET_BITMAP" XG_POST,                               GDK_TARGET_BITMAP);
  DEFINE_ATOM(XG_PRE "GDK_TARGET_COLORMAP" XG_POST,                             GDK_TARGET_COLORMAP);
  DEFINE_ATOM(XG_PRE "GDK_TARGET_DRAWABLE" XG_POST,                             GDK_TARGET_DRAWABLE);
  DEFINE_ATOM(XG_PRE "GDK_TARGET_PIXMAP" XG_POST,                               GDK_TARGET_PIXMAP);
  DEFINE_ATOM(XG_PRE "GDK_TARGET_STRING" XG_POST,                               GDK_TARGET_STRING);
  DEFINE_ATOM(XG_PRE "GDK_SELECTION_TYPE_ATOM" XG_POST,                         GDK_SELECTION_TYPE_ATOM);
  DEFINE_ATOM(XG_PRE "GDK_SELECTION_TYPE_BITMAP" XG_POST,                       GDK_SELECTION_TYPE_BITMAP);
  DEFINE_ATOM(XG_PRE "GDK_SELECTION_TYPE_COLORMAP" XG_POST,                     GDK_SELECTION_TYPE_COLORMAP);
  DEFINE_ATOM(XG_PRE "GDK_SELECTION_TYPE_DRAWABLE" XG_POST,                     GDK_SELECTION_TYPE_DRAWABLE);
  DEFINE_ATOM(XG_PRE "GDK_SELECTION_TYPE_INTEGER" XG_POST,                      GDK_SELECTION_TYPE_INTEGER);
  DEFINE_ATOM(XG_PRE "GDK_SELECTION_TYPE_PIXMAP" XG_POST,                       GDK_SELECTION_TYPE_PIXMAP);
  DEFINE_ATOM(XG_PRE "GDK_SELECTION_TYPE_WINDOW" XG_POST,                       GDK_SELECTION_TYPE_WINDOW);
  DEFINE_ATOM(XG_PRE "GDK_SELECTION_TYPE_STRING" XG_POST,                       GDK_SELECTION_TYPE_STRING);
  DEFINE_ATOM(XG_PRE "GDK_NONE" XG_POST,                                        GDK_NONE);
}

/* -------------------------------- strings -------------------------------- */

static void define_strings(void)
{
  
#if HAVE_GUILE
#if HAVE_SCM_C_DEFINE
  #define DEFINE_STRING(Name, Value) scm_c_define(Name, scm_makfrom0str(Value))
#else
  #define DEFINE_STRING(Name, Value) gh_define(Name, scm_makfrom0str(Value))
#endif
#else
  #define DEFINE_STRING(Name, Value) rb_define_global_const(Name, C_TO_XEN_STRING(Value))
#endif
#if (!HAVE_GTK_1)
  DEFINE_STRING(XG_PRE "GDK_PIXBUF_VERSION" XG_POST,                            GDK_PIXBUF_VERSION);
  DEFINE_STRING(XG_PRE "GTK_STOCK_DIALOG_INFO" XG_POST,                         GTK_STOCK_DIALOG_INFO);
  DEFINE_STRING(XG_PRE "GTK_STOCK_DIALOG_WARNING" XG_POST,                      GTK_STOCK_DIALOG_WARNING);
  DEFINE_STRING(XG_PRE "GTK_STOCK_DIALOG_ERROR" XG_POST,                        GTK_STOCK_DIALOG_ERROR);
  DEFINE_STRING(XG_PRE "GTK_STOCK_DIALOG_QUESTION" XG_POST,                     GTK_STOCK_DIALOG_QUESTION);
  DEFINE_STRING(XG_PRE "GTK_STOCK_DND" XG_POST,                                 GTK_STOCK_DND);
  DEFINE_STRING(XG_PRE "GTK_STOCK_DND_MULTIPLE" XG_POST,                        GTK_STOCK_DND_MULTIPLE);
  DEFINE_STRING(XG_PRE "GTK_STOCK_ADD" XG_POST,                                 GTK_STOCK_ADD);
  DEFINE_STRING(XG_PRE "GTK_STOCK_APPLY" XG_POST,                               GTK_STOCK_APPLY);
  DEFINE_STRING(XG_PRE "GTK_STOCK_BOLD" XG_POST,                                GTK_STOCK_BOLD);
  DEFINE_STRING(XG_PRE "GTK_STOCK_CANCEL" XG_POST,                              GTK_STOCK_CANCEL);
  DEFINE_STRING(XG_PRE "GTK_STOCK_CDROM" XG_POST,                               GTK_STOCK_CDROM);
  DEFINE_STRING(XG_PRE "GTK_STOCK_CLEAR" XG_POST,                               GTK_STOCK_CLEAR);
  DEFINE_STRING(XG_PRE "GTK_STOCK_CLOSE" XG_POST,                               GTK_STOCK_CLOSE);
  DEFINE_STRING(XG_PRE "GTK_STOCK_CONVERT" XG_POST,                             GTK_STOCK_CONVERT);
  DEFINE_STRING(XG_PRE "GTK_STOCK_COPY" XG_POST,                                GTK_STOCK_COPY);
  DEFINE_STRING(XG_PRE "GTK_STOCK_CUT" XG_POST,                                 GTK_STOCK_CUT);
  DEFINE_STRING(XG_PRE "GTK_STOCK_DELETE" XG_POST,                              GTK_STOCK_DELETE);
  DEFINE_STRING(XG_PRE "GTK_STOCK_EXECUTE" XG_POST,                             GTK_STOCK_EXECUTE);
  DEFINE_STRING(XG_PRE "GTK_STOCK_FIND" XG_POST,                                GTK_STOCK_FIND);
  DEFINE_STRING(XG_PRE "GTK_STOCK_FIND_AND_REPLACE" XG_POST,                    GTK_STOCK_FIND_AND_REPLACE);
  DEFINE_STRING(XG_PRE "GTK_STOCK_FLOPPY" XG_POST,                              GTK_STOCK_FLOPPY);
  DEFINE_STRING(XG_PRE "GTK_STOCK_GOTO_BOTTOM" XG_POST,                         GTK_STOCK_GOTO_BOTTOM);
  DEFINE_STRING(XG_PRE "GTK_STOCK_GOTO_FIRST" XG_POST,                          GTK_STOCK_GOTO_FIRST);
  DEFINE_STRING(XG_PRE "GTK_STOCK_GOTO_LAST" XG_POST,                           GTK_STOCK_GOTO_LAST);
  DEFINE_STRING(XG_PRE "GTK_STOCK_GOTO_TOP" XG_POST,                            GTK_STOCK_GOTO_TOP);
  DEFINE_STRING(XG_PRE "GTK_STOCK_GO_BACK" XG_POST,                             GTK_STOCK_GO_BACK);
  DEFINE_STRING(XG_PRE "GTK_STOCK_GO_DOWN" XG_POST,                             GTK_STOCK_GO_DOWN);
  DEFINE_STRING(XG_PRE "GTK_STOCK_GO_FORWARD" XG_POST,                          GTK_STOCK_GO_FORWARD);
  DEFINE_STRING(XG_PRE "GTK_STOCK_GO_UP" XG_POST,                               GTK_STOCK_GO_UP);
  DEFINE_STRING(XG_PRE "GTK_STOCK_HELP" XG_POST,                                GTK_STOCK_HELP);
  DEFINE_STRING(XG_PRE "GTK_STOCK_HOME" XG_POST,                                GTK_STOCK_HOME);
  DEFINE_STRING(XG_PRE "GTK_STOCK_INDEX" XG_POST,                               GTK_STOCK_INDEX);
  DEFINE_STRING(XG_PRE "GTK_STOCK_ITALIC" XG_POST,                              GTK_STOCK_ITALIC);
  DEFINE_STRING(XG_PRE "GTK_STOCK_JUMP_TO" XG_POST,                             GTK_STOCK_JUMP_TO);
  DEFINE_STRING(XG_PRE "GTK_STOCK_JUSTIFY_CENTER" XG_POST,                      GTK_STOCK_JUSTIFY_CENTER);
  DEFINE_STRING(XG_PRE "GTK_STOCK_JUSTIFY_FILL" XG_POST,                        GTK_STOCK_JUSTIFY_FILL);
  DEFINE_STRING(XG_PRE "GTK_STOCK_JUSTIFY_LEFT" XG_POST,                        GTK_STOCK_JUSTIFY_LEFT);
  DEFINE_STRING(XG_PRE "GTK_STOCK_JUSTIFY_RIGHT" XG_POST,                       GTK_STOCK_JUSTIFY_RIGHT);
  DEFINE_STRING(XG_PRE "GTK_STOCK_MISSING_IMAGE" XG_POST,                       GTK_STOCK_MISSING_IMAGE);
  DEFINE_STRING(XG_PRE "GTK_STOCK_NEW" XG_POST,                                 GTK_STOCK_NEW);
  DEFINE_STRING(XG_PRE "GTK_STOCK_NO" XG_POST,                                  GTK_STOCK_NO);
  DEFINE_STRING(XG_PRE "GTK_STOCK_OK" XG_POST,                                  GTK_STOCK_OK);
  DEFINE_STRING(XG_PRE "GTK_STOCK_OPEN" XG_POST,                                GTK_STOCK_OPEN);
  DEFINE_STRING(XG_PRE "GTK_STOCK_PASTE" XG_POST,                               GTK_STOCK_PASTE);
  DEFINE_STRING(XG_PRE "GTK_STOCK_PREFERENCES" XG_POST,                         GTK_STOCK_PREFERENCES);
  DEFINE_STRING(XG_PRE "GTK_STOCK_PRINT" XG_POST,                               GTK_STOCK_PRINT);
  DEFINE_STRING(XG_PRE "GTK_STOCK_PRINT_PREVIEW" XG_POST,                       GTK_STOCK_PRINT_PREVIEW);
  DEFINE_STRING(XG_PRE "GTK_STOCK_PROPERTIES" XG_POST,                          GTK_STOCK_PROPERTIES);
  DEFINE_STRING(XG_PRE "GTK_STOCK_QUIT" XG_POST,                                GTK_STOCK_QUIT);
  DEFINE_STRING(XG_PRE "GTK_STOCK_REDO" XG_POST,                                GTK_STOCK_REDO);
  DEFINE_STRING(XG_PRE "GTK_STOCK_REFRESH" XG_POST,                             GTK_STOCK_REFRESH);
  DEFINE_STRING(XG_PRE "GTK_STOCK_REMOVE" XG_POST,                              GTK_STOCK_REMOVE);
  DEFINE_STRING(XG_PRE "GTK_STOCK_REVERT_TO_SAVED" XG_POST,                     GTK_STOCK_REVERT_TO_SAVED);
  DEFINE_STRING(XG_PRE "GTK_STOCK_SAVE" XG_POST,                                GTK_STOCK_SAVE);
  DEFINE_STRING(XG_PRE "GTK_STOCK_SAVE_AS" XG_POST,                             GTK_STOCK_SAVE_AS);
  DEFINE_STRING(XG_PRE "GTK_STOCK_SELECT_COLOR" XG_POST,                        GTK_STOCK_SELECT_COLOR);
  DEFINE_STRING(XG_PRE "GTK_STOCK_SELECT_FONT" XG_POST,                         GTK_STOCK_SELECT_FONT);
  DEFINE_STRING(XG_PRE "GTK_STOCK_SORT_ASCENDING" XG_POST,                      GTK_STOCK_SORT_ASCENDING);
  DEFINE_STRING(XG_PRE "GTK_STOCK_SORT_DESCENDING" XG_POST,                     GTK_STOCK_SORT_DESCENDING);
  DEFINE_STRING(XG_PRE "GTK_STOCK_SPELL_CHECK" XG_POST,                         GTK_STOCK_SPELL_CHECK);
  DEFINE_STRING(XG_PRE "GTK_STOCK_STOP" XG_POST,                                GTK_STOCK_STOP);
  DEFINE_STRING(XG_PRE "GTK_STOCK_STRIKETHROUGH" XG_POST,                       GTK_STOCK_STRIKETHROUGH);
  DEFINE_STRING(XG_PRE "GTK_STOCK_UNDELETE" XG_POST,                            GTK_STOCK_UNDELETE);
  DEFINE_STRING(XG_PRE "GTK_STOCK_UNDERLINE" XG_POST,                           GTK_STOCK_UNDERLINE);
  DEFINE_STRING(XG_PRE "GTK_STOCK_UNDO" XG_POST,                                GTK_STOCK_UNDO);
  DEFINE_STRING(XG_PRE "GTK_STOCK_YES" XG_POST,                                 GTK_STOCK_YES);
  DEFINE_STRING(XG_PRE "GTK_STOCK_ZOOM_100" XG_POST,                            GTK_STOCK_ZOOM_100);
  DEFINE_STRING(XG_PRE "GTK_STOCK_ZOOM_FIT" XG_POST,                            GTK_STOCK_ZOOM_FIT);
  DEFINE_STRING(XG_PRE "GTK_STOCK_ZOOM_IN" XG_POST,                             GTK_STOCK_ZOOM_IN);
  DEFINE_STRING(XG_PRE "GTK_STOCK_ZOOM_OUT" XG_POST,                            GTK_STOCK_ZOOM_OUT);
#endif
#if PANGO_ENABLE_ENGINE && PANGO_ENABLE_BACKEND
  DEFINE_STRING(XG_PRE "PANGO_ENGINE_TYPE_LANG" XG_POST,                        PANGO_ENGINE_TYPE_LANG);
  DEFINE_STRING(XG_PRE "PANGO_ENGINE_TYPE_SHAPE" XG_POST,                       PANGO_ENGINE_TYPE_SHAPE);
  DEFINE_STRING(XG_PRE "PANGO_RENDER_TYPE_NONE" XG_POST,                        PANGO_RENDER_TYPE_NONE);
#endif

}

/* -------------------------------- initialization -------------------------------- */

static int xg_already_inited = 0;

#if HAVE_GUILE
 XEN init_xm(void);
 XEN init_xm(void)
#else
 XEN Init_libxm(void);
 XEN Init_libxm(void)
#endif
{
  if (!xg_already_inited)
    {
      define_xm_obj();
      define_integers();
      define_doubles();
      define_functions();
      define_structs();
      define_macros();
      define_structs();
      define_atoms();
      define_strings();
      XEN_YES_WE_HAVE("xg");
#if HAVE_GUILE
      XEN_EVAL_C_STRING("(define xm-version \"12-Mar-02\")");
#endif
      xg_already_inited = 1;
    }
  return(XEN_FALSE);
}
